/***--------------------------------------------------------------------------
 * NAME:        artmap.c
 * PURPOSE: driver for fuzzy artmap classifier
 *
 * SEE ALSO: artmap_pack.c
 * GLOBALS:  int debug;
 *
 * DESCRIPTION: 
 *  Functions in this file:
 *      artmap() -- driver for fuzzy artmap classifier
 *           -- handles train, create flags
 * 
 * BUGS AND LIMITATIONS:
 * 
 *       Note that we now compute the norm of the input vector each time
 * we classify it because precision limitation problems gave wacky results
 * when vigilance for the "a" ART module was set to 1.0. -RKC
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_cv.c,v 11.4 2001/08/23 19:56:02 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_cv.c,v $
 * CREATED: 13-Nov-91 rkc               MIT LINCOLN LABORATORY
 *----------------------- pseudocode of driver follows -----------------------
 artmap_build - allocate storage
 artmap_init - fills in values, put labels in structure
 if normalize read_norm_file - get normalization information
 
 nn_fopen - open data file
 allcoate space for data and indices
 read the data
 count the patterns in each class
 calculate the number of patterns in each class in each "fold"
 of the N-fold testing
 create index labeling which "fold" each pattern is in
 for each "fold"
 initialize artmap model
 initialize fold error reporting
 if complement_coding
 learn complement mapping
 complement code the data
 endif
 if(ARTnorm)
 for all vectors
 artmap_vector_norm - make sum_i(inputs) = 1
 
 artmap_epochs_ctrain
 artmap_gc_f2 - remove unused nodes
 do any reinitialization needed
 loop through n_trials
 if this pattern is in this fold's test set
 artmap_classify
 endif
 record any errors
 end loop
 end loop
 deallocate space for indices and data
 print out final error results
 free allocated space
 ---------------------- end of pseudocode ---------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "artmap.h"
#include "artmap_pack.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)
/* Here we avoid the norm calculation for all cases where the norm is know
   in advance.  Note that we need to compute the norm when vigilance is
   equal to 1.0, because we have floating point representation errors. 
   
   */
#define FAST_NORM(ro,v) ((ro == 1.0)? artmap_norm((v),ninputs): \
             (ARTnorm? 1.0: complement_coding?nfinputs: \
                (artmap_norm((v),ninputs))))

/* globals */

static char prog_name[MAXSTR];  /* static variables */

void
artmap_cross_valid(DEFAULTS_FILE_STRUCT *df_args,
           int nfinputs, /* number of inputs to read from file */
           int *features, int normalize, int npatterns, int debg,
           int verbosity, int verror, 
           /* New Cross-valid things */
           int cv_nfolds, int cv_nsplits,
           int use_cv_splits, int *cv_splits, char **cv_train_mask,
           char **cv_test_mask, char *data_path, char *data_fname,
           char *norm_fname, char *exp_path,
           char *error_fname, char *param_fname,
           char *comment, int nepochs, int random_flag,
           int adjust_priors, int priors_npatterns, float *priors_list,
           int cv_random,
           int train_cycle,
           /* Artmap specific goodies */
           float ro_a_train, float ro_a_test, float ro_b, float beta,
           float alpha, int ARTnorm, int complement_coding,
           int is_fast_commit, int gc_threshold)
{
    /*
     * Because complement coding requires 2*ninputs for each vector line (where
     * a(i) is at i and a^c(i) is at ninputs+i), we allocate double the amount
     * of needed space and fill the complements in after we've read in the
     * regular line.
     */
    int ninputs = (complement_coding?nfinputs<<1:nfinputs);
    int model, normalization = 0, err_model, err_part_model, data_model;
    int t_index;
    int desired_class, class;
    int n_read, n_trials;
    int db_type = 0;    /* needed for error reporting routines */
    int fold;
    float *inputs, *outputs;
    float **coded_matrix;
    float *dummy_outputs; /* No match case */
    float *im_norm;
    char *tmp_path;
    FILE *error_file = NULL, *param_file;
    int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());

    dummy_outputs = allocate_vector(ninputs);

    /*
     * do the initialization of the training data structure, reading the
     * previously stored training vectors in the case of retrain and test also
     * initialize the normalization routines, if needed.
     */

    /* allocate and initialize model structure */
    model = artmap_build(ninputs, df_args->noutputs, 0, 0 ,0);
    artmap_init(model,db_type,features,complement_coding,ARTnorm,
        ro_a_train,beta,alpha,ro_b,beta,alpha);
    
    if (normalize) 
      normalization = read_norm_file(path_build(data_path, norm_fname));

    if (verror>0) {     /* open file for classification errors */
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    }
    /* set the outputs pointer */
    outputs = *(float **)artmap_param(model, "outputs");

        /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, df_args->noutputs+1);/* For Novel */
    report_init (err_model, db_type, ninputs, ARTMAP_COST,
         FALSE, post_sum_to_one);
    err_part_model = report_build (verbosity, 0, df_args->noutputs+1);

    if(df_args->output_labels != NULL){
    char **nlabels;
    char *new_label_len;
    int i;

    /* copy the lable list, adding the class "novel" */
    nlabels = (char **)allocate_pvector(df_args->noutputs+2);
    /* add one to the strlen for the \0 and one in case the
       number is bigger */
    new_label_len = nn_malloc(strlen(df_args->output_labels[0])+2,
                  "for new labels list");
    sprintf(new_label_len,"%d",df_args->noutputs+1);
    nlabels[0] = new_label_len;
    for(i = 1; i <= df_args->noutputs; i++)
      nlabels[i] = df_args->output_labels[i];
    nlabels[df_args->noutputs+1] = nn_strdup("Novel");
    /* free the old label list.  This must be ALLOCATED above */
    nn_free(df_args->output_labels[0], "for bigger copy of output labels");
    deallocate_pvector((void **)df_args->output_labels);
    df_args->output_labels = nlabels;
    }

    im_norm = allocate_vector(npatterns);

    /* read data */
    data_model = class_data_read(data_path, data_fname, npatterns,
                 df_args->ninputs, nfinputs,
                 df_args->noutputs, features, normalization,
                 verbosity);
    n_trials = class_data_get_npatterns(data_model);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
               cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
      class_data_priors_init(data_model, priors_list, priors_npatterns);
    
    /* this is maybe a little bit of a cheat.  I am complement coding
       on all of the data at once, rather than doing it on just
       the training data for a fold -- lck Oct 1993 */
    if(complement_coding){/* Need only do it once */
    float **input_matrix = class_data_get_input_matrix(data_model);

    coded_matrix = allocate_matrix(npatterns, ninputs);

    artmap_learn_complement_mapping(model, input_matrix, n_trials,
                    nfinputs);
    artmap_complement_code(model,input_matrix,n_trials,nfinputs,
                   coded_matrix);
    } else {
    coded_matrix = class_data_get_input_matrix(data_model);
    }
    /* You probably don't want to do this with complement coding. */
    if(ARTnorm) /* Do ARTnorm just once */
      for(n_read=0;n_read<n_trials;n_read++)
    artmap_vector_norm(coded_matrix[n_read],ninputs);
    
    for(n_read=0;n_read<n_trials;n_read++)
      /*Compute limited-prcsn norm*/
      im_norm[n_read] = FAST_NORM(MAX(ro_a_test,ro_a_train)
                  ,coded_matrix[n_read]);

    if(ro_a_train < 0 || ro_a_train > 1)
      fatal_error("missing ro_a parameter");
    if(ro_a_test < 0 || ro_a_test > 1)
      fatal_error("missing ro_a parameter");
    
    /* store normalization once */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w",
              "for write of fuzzy artmap model file");
    /* write comment */
    fprintf(param_file, "%s", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    
    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);
        
    artmap_init(model, db_type, features,complement_coding, ARTnorm,
            ro_a_train, beta, alpha, ro_b, beta, alpha);
    
    class_data_start_cv_fold(data_model, fold, TRUE);
    n_trials = class_data_get_npatterns(data_model);
    
    /* initialize error reporting for training fold */
    report_init(err_part_model, db_type, ninputs, ARTMAP_COST,
            FALSE, post_sum_to_one);
    report_epoch_header(err_part_model, stdout, n_trials);
    
    /*******************************/
    /* taken from artmap_epoch_ctrain */
    { /* needed for stopping training when train_cycle is set */
        int errs_1ago = -1, errs_now = 0, err_cnt = 0;
        int epoch;
        
        for(epoch=0; epoch < nepochs; epoch++){
        class_data_epoch_init(data_model);
        /* train_cycle flag says to continue training until
           errs doesn't change 2 times in a row */
        if(train_cycle && errs_now == errs_1ago) break;
        errs_1ago = errs_now;
        err_cnt = errs_now = 0;
        /***********************************/
        /* taken from artmap_batch_ctrain */
        for(n_read = 0; n_read < n_trials; n_read++){
            /* have class_data pick a pattern */
            inputs =
              class_data_get_next(data_model, &desired_class);
            if(inputs == NULL) break;
            t_index = class_data_get_input_num(data_model);
            inputs = coded_matrix[t_index];
            class=
              artmap_inc_ctrain(model, inputs, desired_class,
                    im_norm[t_index],
                    is_fast_commit);
            report_class_trial(err_part_model, error_file,
                       inputs, outputs, desired_class,
                       class, t_index);
            if(class != desired_class){
            errs_now ++;
            if(class != -1)err_cnt++;
            }
        }
        artmap_batch_update(model,n_trials);
        report_artmap_epoch(err_part_model, stdout, epoch, err_cnt,
                    *(int*)artmap_param(model,"num_f2a"),
                    *(int*)artmap_param(model,"num_f2b"));
        report_init(err_part_model, db_type, ninputs, ARTMAP_COST,
                FALSE, post_sum_to_one);
        /******************************/
        }
        /* do garbage collecting */
        class_data_epoch_init(data_model);
        artmap_gc_init(model);
        for(n_read = 0; n_read < n_trials; n_read++){
        /* have class_data pick a pattern */
        inputs = class_data_get_next(data_model, &desired_class);
        if(inputs == NULL) break;
        t_index = class_data_get_input_num(data_model);
        inputs = coded_matrix[t_index];
        artmap_gc_mark(model, inputs, desired_class,
                   im_norm[t_index], gc_threshold);
        }
        if(!artmap_gc_f2(model, n_trials, ninputs,
                 gc_threshold, verbosity) && verbosity>1){
        int arta_nodes, artb_nodes;
        
        arta_nodes = *(int *)artmap_param(model, "num_f2a");
        artb_nodes = *(int *)artmap_param(model, "num_f2b");
        printf("Number of f2 Nodes, a=%9d, b=%9d\n",
               arta_nodes, artb_nodes);
        }
    }
    class_data_end_cv_fold(data_model);
    
    /***********************************************/     
    /* print out the model being used for testing */
    fprintf(param_file, "artmap model %d\n", fold);
    artmap_write_model(model, param_file);   
        
    /* reinitialize reporting for testing */
    report_init(err_part_model, db_type, ninputs, ARTMAP_COST,
            FALSE, post_sum_to_one);
        
    /* Set ro_a to input */
    *(float*)artmap_param(model,"ro_a") = ro_a_test;
    
    class_data_start_cv_fold(data_model, fold, FALSE);
    n_trials = class_data_get_npatterns(data_model);
    for(n_read = 0; n_read < n_trials; n_read++){
        /* have class_data pick a pattern */
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        t_index = class_data_get_input_num(data_model);
        inputs = coded_matrix[t_index];
        class=artmap_classify(model, inputs, &outputs,
                  im_norm[t_index]);
        if(class == -1){
        class = df_args->noutputs;
        outputs = dummy_outputs;
        }
        report_class_trial(err_part_model, error_file, inputs, outputs,
                   desired_class, class, t_index);
        report_class_trial(err_model, error_file, inputs, outputs,
                   desired_class, class, t_index);
    }
    /* report on this fold */
    report_overall(err_part_model, stdout, error_fname, df_args->output_labels);
    
    class_data_end_cv_fold(data_model);
    }
    nn_fclose(param_file, tmp_path,
          "for write of fuzzy artmap model file");
    if ( verbosity > 0)
      printf("\nFinished -- fuzzy artmap model saved in \"%s\"\n",
         tmp_path);
    
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, df_args->output_labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }

    /*free space and close files, for neatness sake */
    artmap_free(model);
    norm_free();
    deallocate_vector(im_norm);
    deallocate_vector(dummy_outputs);
}
