/****------------------------------------------------------------------------
 * NAME:        artmap_pack.c
 * PURPOSE: Functions specific to artmap.c
 *
 * SEE ALSO: artmap.c artmap_main.c
 *
 * GLOBALS:
 *
 * DESCRIPTION:
 *    Functions in this file
 *  artmap_build()      Allocate storage for model parameters.
 *  artmap_init()       Initialize model parameters.
 *  artmap_free()       Deallocates model storage
 *  artmap_read_model() Initializes a artmap model from file
 *  artmap_write_model()    Writes a artmap model to output file
 *  artmap_param()      Returns pointers to elements in model
 *  artmap_inc_ctrain() does training on a vector
 *  artmap_classify()   Classifies input and returns an integer class
 *
 * ALGORITHM:
 *  Training:  
 *
 *  Testing:  
 *      
 * REFERENCES: Fuzzy ARTMAP: A Neural Network Architecture for Incremental
 *             Supervised Learning of Analog Multidimensional Maps 
 *             Gail Carpenter, Stephen Grossberg, Natalya Markuzon, 
 *             John H. Reynolds, and David B. Rosen.
 *
 *             Fuzzy ART: Fast Stable Learning And Categorization of Analog
 *                 Patterns by an Adaptive Resonance System
 *                 Gail Carpenter, Stephen Grossberg, and David B. Rosen.
 *
 *      ARTMAP: Supervised Real-Time Learning and Classification of
 *          Nonstationary Data by a Self-Organizing Neural Network
 *          Gail A. Carpenter, Stephen Grossberg, and John H. Reynolds
 *          Neural Networks, Vol. 4, pp. 565-588, 1991
 *
 * BUGS AND LIMITATIONS:
 *    I implement only the fast learning case; I suspect the slow learning
 *  case would do better.
 *
 *   LNKnet stores its trained weights in ascii, which leads to quantization 
 * errors.  Fuzzy ARTMAP seems sensitive to quantization errors because it 
 * encodes the maximum size of each exemplar.  The only way to fix this is
 * to get LNKnet to store its data in binary form, which has many drawbacks
 * of it own (although most of these relate to higher level issues like 
 * the weights become less portable, and debugging becomes more difficult).
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_pack.c,v 11.4 2001/08/23 19:56:03 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_pack.c,v $
 * CREATED: 13-Nov-91 rkc                             MIT LINCOLN LABORATORY
 * -----------------------------------------------------------------------**/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#include <stdio.h>
#include <nnlib.h>
#define artmap_pack_h_internal_types
#define artmap_pack_h_param_types
#define artmap_pack_statics
#include "artmap_pack.h"

/* constant definitions */
static int debug = 0;
#define TRACE_CODE
#ifdef TRACE_CODE
#define d_printf(level, stuff) if(level<=debug){ \
                     printf stuff; \
                       } 
#define d_fprintf(level, stuff) if(level<=debug){ \
                     fprintf stuff; \
                       } 
#define whentrace(level,stuff) ((level<=debug)? stuff:0)
#else /* TRACE_CODE */
#define d_printf(level,stuff) /*noop*/
#define d_fprintf(level, stuff)  /*noop*/
#endif /* TRACE_CODE*/

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define TINY (1e-3) /* Set to zero and last f2 node was diabled */
#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))
#define F2_STORE_SIZE 10
  
#define UNASSOCIATED (-1)/*Class_maps depends on this being -1,don't modify*/
#define artmap_fast_map_learn(fmf,chosen_f2a,chosen_f2b) \
  (fmf)->wj_ab[(chosen_f2a)] = (chosen_f2b);

#define artmap_fast_map_correct_size(fmf,size)           \
  if(fmf->alloc_wj < (size)){                            \
    int i;                                               \
    fmf->wj_ab = (int*)nn_realloc((char*)fmf->wj_ab,sizeof(int)*(size),\
                    "for expanding fast map's weights");  \
    for(i=fmf->alloc_wj;i<size;i++)                      \
      fmf->wj_ab[i] = UNASSOCIATED;                      \
    fmf->alloc_wj = (size);                              \
  }


    /* Internal Globals */
static ARTMAP_PARAMS **artmappp;       /*array of parameters for each model*/
static int max_models = 0;               /* current length of artmappp */

#define is_complement_coding(ampp) (ampp->nfinputs == (ampp->ninputs>>(1)))

/* Boy do I wish I were coding in C++ 
   The sacrifices I make for the good of the world...
*/
#define FUZZY_AND(a,b) ((a)<(b)?(a):(b)) /* Why not just call it min? */
#define FUZZY_OR(a,b) ((a)>(b)?(a):(b))  /* Why not just call it max? */
#define ABS(a) ((a) >= 0? (a) : -(a))    /* I dislike fabs, iabs */
#define ClassIndex(ampp,chosen_f2a) \
  (chosen_f2a==UNASSOCIATED)?UNASSOCIATED:ampp->mf->wj_ab[chosen_f2a]

#define artmap_fuzzy_wts(f,i) \
{ int j; \
  for(j=0;j<(f)->ninputs;j++)  \
    (f)->w[i][j] = 1.0;        \
  (f)->inv_norm_w_alpha[i] = 1.0/(float)((f)->ninputs+f->alpha); \
}
/* I assume ninputs is the same as was in the call to fuzzy_build */
#define artmap_fuzzy_init(f,alpha_v,beta_v,ro_v) \
{ int i; \
  (f)->alpha = alpha_v; \
  (f)->beta = beta_v; \
  (f)->ro = ro_v; \
  (f)->num_f2 = 0; \
  (f)->which_y = 0; \
  for(i=0;i<f->alloc_f2;i++){ \
    artmap_fuzzy_wts(f,i); \
  } \
  }
/* I assume ninputs is the same as was in the call to fuzzy_build
   Teach these class mapping now so the class numbers correspond */ 
#define artmap_fuzzy_class_init(ampp,f,alpha_v,beta_v,ro_v) \
{ int i; \
    artmap_fuzzy_init(f,alpha_v,beta_v,ro_v); \
    for(i=0;i<(ampp)->nclasses;i++)  \
      artmap_fuzzy_learning(f,ampp->class_maps[i],1.0,UNASSOCIATED,TRUE,TRUE);\
  }

/***********
  Debug routine
*/
static void
p10(char *s, float *ptr, int n)
{
  int i;
  if (n > 10)    n=10;

  printf(s);
  for(i=0;i<n;i++)
    printf(" %g",ptr[i]);
  printf("\n");
}

/****************************************************************************
 * Name: artmap_norm(p,plen)
 *    p                     - vector pointer
 *    plen                  - length of the vector
 * DESCRIPTION:
 *  Computes the sum of the absolute values of the elements of the vector.
 *  This should probably be integrated into norm_subs, but I didn't want to 
 *  modify a file that might be changing. -RKC
 */
float
artmap_norm(float *p, int plen)
{
  float sum,t;

  for(sum=0,plen--;plen>=0;plen--){
    t = *p++;
    sum+= ABS(t);
  }

  return sum;
}
/****************************************************************************
 * Name: artmap_vector_norm(p,plen)
 *    p                     - vector pointer
 *    plen                  - length of the vector
 * DESCRIPTION:
 *  Forces the sum of all elements in the vector to equal 1
 */
void
artmap_vector_norm(float *p, int plen)
{
  int len = plen;
  float sum;

  for(sum=0,len=0;len<plen;len++){
    sum+= p[len];
  }
  for(plen--;plen>=0;plen--){
    *p++ /= sum;
  }
}

/*====================== Complement Coding code ===========================*/
/****************************************************************************
 * Name: artmap_new_scale(model,ncol)
 *    model                 - the model we`re using
 *    ncol                  - number of columns in the INPUT matrix
 *
 * DESCRIPTION:
 * Create the matricies for the specified model
 * When complement coding is set, we need to set each vector element into the
 * range of [0:1].  The way we do this is to get the range of the data and
 * linearly interpolate into the [0:1] region.  We then remember the scale
 * factor and the offset used to do this.  Later when we use test data, we
 * assume the same scalefactor and offset, and truncate at the limit values.
 * If you wanted to logarithmic mapping, you should have done it ahead of time.
 * 
 */
void
artmap_new_scale(int model, int ncol)
{
    ARTMAP_PARAMS *ampp;        /* temporary pointer to artmappp[] */
    ampp = artmappp[model];
    if(ampp->slope){
    warning("Changing Model %d's slope vector\n",model);
    deallocate_vector(ampp->slope);
    }
    ampp->slope=allocate_vector(ncol); /* the slope of the rescaling */
  
    if(ampp->intercept){
    warning("Changing Model %d's intercept vector\n",model);
    deallocate_vector(ampp->intercept);
    }
    ampp->intercept = allocate_vector(ncol); /* the offset of the rescaling */
}
/****************************************************************************
 * Name:artmap_learn_split_extrema(m1,nrow,ncol,cross_index,cv_nsplits,min,max)
 *    m1                    - input matrix ptr
 *    nrow                  - number of rows
 *    ncol                  - number of columns in the INPUT matrix
 *    cv_train_mask         - who is in the training set? [cv_nsplits]
 *    cross_index           - which fold does each input fall in?
 *    cv_nsplits            - the number of splits
 *    min                   - the min array [cv_nsplits x ncol]
 *    max                   - the max array [cv_nsplits x ncol]
 *
 * DESCRIPTION:
 * Find the min and max of each component, returns 1 if you must 
 * rescale the data.  (If not, you can just complement code once and be done!)
 * 
 */
static int
artmap_learn_split_extrema(float **m1, int nrow, int ncol, int *cross_index,
               int cv_nsplits, float **min, float **max)
{
  int fold,i,input;

  /* Initialize the min and max arrays to dummy values that will be replaced
     as soon as one data point is found */
  for(fold=0;fold<cv_nsplits;fold++){
    for(i=0;i<ncol;i++){
      min[fold][i] = MAXFLOAT;
      max[fold][i] = MINFLOAT;
    }
  }
  /* Find the min and max of each component in each input */
  for(input=0;input<nrow;input++){
    fold = cross_index[input];
    for(i=0;i<ncol;i++){
      if(m1[input][i] > max[fold][i])
    max[fold][i] = m1[input][i];
      if(m1[input][i] < min[fold][i])
    min[fold][i] = m1[input][i];
    }
  }
  for(fold=0;fold<cv_nsplits;fold++){
    for(i=0;i<ncol;i++){
      if(min[fold][i] < 0.0 || max[fold][i] > 1.0){
    if(debug >= 1 ) printf("Must complement code\n");
    return(1);
      }
    }
  }
  return(0);
  /* Returns min,max */
}
/****************************************************************************
 * Name: artmap_complement_mapping_from_extrema(model,min,max,ncol,
 *                      train_mask,nsplits)
 *    model                 - the model we`re using
 *    ncol                  - number of columns in the INPUT matrix
 *    train_mask            - who is in the training set for this fold?
 *    cv_nsplits            - the number of splits
 *    min                   - the min array [cv_nsplits x ncol]
 *    max                   - the max array [cv_nsplits x ncol]
 *
 * DESCRIPTION:
 * When complement coding is set, we need to set each vector element into the
 * range of [0:1].  The way we do this is to get the range of the data and
 * linearly interpolate into the [0:1] region.  We then remember the scale
 * factor and the offset used to do this.  Later when we use test data, we
 * assume the same scalefactor and offset, and truncate at the limit values.
 * If you wanted to logarithmic mapping, you should have done it ahead of time.
 * 
 */
static void
artmap_complement_mapping_from_extrema(int model, float **min, float **max,
                       int ncol, char *train_mask,
                       int cv_nsplits)
{
  float *fold_min,*fold_max;
  int i,split;
  ARTMAP_PARAMS *ampp;      /* temporary pointer to artmappp[] */

  ampp = artmappp[model];
  artmap_new_scale(model,ncol);

  fold_min = ampp->intercept;
  fold_max = ampp->slope;/* We'll convert this at the end */
  /* Set up the min and max so we'll find the first input */ 
  for(i=0;i<ncol;i++){
    fold_min[i] = MAXFLOAT;
    fold_max[i] = MINFLOAT;
  }

  for(split=0;split<cv_nsplits;split++){
    if(train_mask[split]!='1')continue;/* Not used in this fold */
    for(i=0;i<ncol;i++){
      if(max[split][i] > fold_max[i])
    fold_max[i] = max[split][i];
      if(min[split][i] < fold_min[i])
    fold_min[i] = min[split][i];
    }
  }
  for(i=0;i<ncol;i++){
    if(fold_min[i] < 0.0 || fold_max[i] > 1.0){
      warning("Input column %d is of range [%g:%g], linearly mapping to [0:1]\n",
          i,fold_min[i],fold_max[i]);
      fold_max[i] = 1/(fold_max[i]-fold_min[i]);/* set slope */
    }
    else {/* We don't need to remap */
      warning("Possible ARTMAP bug--shouldn't be remapping inputs in range\n");
      fold_min[i] = 0;/* set slope */
      fold_max[i] = 1;/* set slope */
    }
  }
}



/****************************************************************************
 * Name: artmap_learn_complement_mapping(model,m1,nrow,ncol)
 *    model                 - the model we`re using
 *    m1                    - input matrix ptr
 *    nrow                  - number of rows
 *    ncol                  - number of columns in the INPUT matrix
 *
 * DESCRIPTION:
 * When complement coding is set, we need to set each vector element into the
 * range of [0:1].  The way we do this is to get the range of the data and
 * linearly interpolate into the [0:1] region.  We then remember the scale
 * factor and the offset used to do this.  Later when we use test data, we
 * assume the same scalefactor and offset, and truncate at the limit values.
 * If you wanted to logarithmic mapping, you should have done it ahead of time.
 * 
 */
void
artmap_learn_complement_mapping(int model,float **m1, int nrow, int ncol)
{
  int i,j;
  float *sl,*inter;
  int remapped = FALSE;
  float max,min;
  ARTMAP_PARAMS *ampp;      /* temporary pointer to artmappp[] */

  ampp = artmappp[model];
  artmap_new_scale(model,ncol);

  sl = ampp->slope;
  inter = ampp->intercept;

  for(i=0;i<ncol;i++){
    min = max = m1[0][i];
    for(j=1;j<nrow;j++){
      if(m1[j][i] > max)
    max = m1[j][i];
      if(m1[j][i] < min)
    min = m1[j][i];
    }
    if(min < 0.0 || max > 1.0){
      warning("Input column %d is of range [%g:%g], linearly mapping to [0:1]\n",
          i,min,max);
      inter[i] = min;
      sl[i] = 1/(max-min);
      remapped = TRUE;
    }
    else {
      inter[i] = 0;
      sl[i] = 1;
    }
  }
  if(!remapped){/* Save nfinputs*n_read multiplications */
    deallocate_vector(ampp->slope);
    ampp->slope = NULL;
    deallocate_vector(ampp->intercept);
    ampp->intercept = NULL;
  }
}
/****************************************************************************
 * Name: artmap_complement_code(model,m1,nrow,ncol,m2)
 *    model                 - the artmap model we're using
 *    m1                    - input matrix ptr
 *    m2                    - output matrix ptr
 *    nrow                  - number of rows
 *    ncol                  - number of columns in the INPUT matrix
 *
 * DESCRIPTION:
 * Complement code the input using values calculated by
 * learn_complement_mapping.  
 *
 */
void
artmap_complement_code(int model, float **m1, int nrow, int ncol,float **m2)
{
  int i,j;
  float *intercept,*slope;
  ARTMAP_PARAMS *ampp;      /* temporary pointer to artmappp[] */

  ampp = artmappp[model];
  intercept = ampp->intercept;
  slope = ampp->slope;

  if(intercept && slope){
    /* First convert each input variable to the range of [0:1] */
    for(i=0;i<nrow;i++){
      for(j=0;j<ncol;j++){
    m2[i][j] = (m1[i][j]-intercept[j]) * slope[j];
    if(m2[i][j]>1.0){
      warning("Complement code truncating m[%d][%d]=%f to 1.0\n",
          i,j,m2[i][j]);
      m2[i][j] = 1.0;
    }
    else if (m2[i][j]<0.0){
      warning("Complement code truncating m[%d][%d]=%f to 0.0\n",
          i,j,m2[i][j]);
      m2[i][j] = 0.0;
    }
      }
    }
  }
  else if (m1 != m2){/* Only copy if we need to */
    for(i=0;i<nrow;i++)
      for(j=0;j<ncol;j++)
    m2[i][j] = m1[i][j];
  }
  /* Now create the complement values */
  for(i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      m2[i][j+ncol] = 1.0 - m2[i][j];
    }
  }
  /* no return value !!! */
}
/**************************************************************************
 *   Name:  artmap_renormalize_vector(model,v1,ncol,v2)
 *        model              - the model we're working on.
 *        v1                 - input vector, with [0..ncol-1] filled with 
 *                data
 *        ncol               - number of columns in the INPUT matrix
 *        v2             - output vector, with [0..ncol-1] filled with 
 *                 the remapped data
 *
 *   DESCRIPTION:
 *   This code will complement code a single input vector.  It is written
 *   so Linda can make the plot program work with Fuzzy ARTmap.   To use this
 *   properly, you should probably have either:
 *      1. run the entire input matrix through artmap_complement_code 
 *   OR     2. be plotting this data for a stored set of information in
 *      which the slope and intercept of the input data has already
 *      been calculated (and read in along with the model).  As far
 *      as I can remember, the calculation should always be stored
 *      with the model provided complement coding was done on the
 *      original testing.
 *
 *      If no slope or intercept is provided, we simply copy the
 *      input, and pray for the best.
 *                -RKC
 */
static void
artmap_renormalize_vector(int model, float *v1, int ncol, float *v2)
{
  int j;
  float *intercept,*slope;
  ARTMAP_PARAMS *ampp;      /* temporary pointer to artmappp[] */

  ampp = artmappp[model];
  intercept = ampp->intercept;
  slope = ampp->slope;

  if(intercept && slope){
    /* First convert each input variable to the range of [0:1] */
    for(j=0;j<ncol;j++){
      v2[j] = (v1[j]-intercept[j]) * slope[j];
      if(v2[j]>1.0){
      warning("Complement code truncating value to range [0:1]");
#ifdef DEBUG
      warning("Complement code truncating v[%d]=%f to 1.0\n",j,v2[j]);
#endif
      v2[j] = 1.0;
      }
      else if (v2[j]<0.0){
      warning("Complement code truncating value to range [0:1]");
#ifdef DEBUG
      warning("Complement code truncating v[%d]=%f to 0.0\n",j,v2[j]);
#endif
      v2[j] = 0.0;
      }
    }
  }
  else if(v1!=v2){/* Only copy if necessary */
    for(j=0;j<ncol;j++)
      v2[j] = v1[j];
  }
}

/**************************************************************************
 *   Name:  artmap_complement_code_vector(model,v1,ncol,v2)
 *        model              - the model we're working on.
 *        v1                 - input vector, with [0..ncol-1] filled with 
 *                data
 *        ncol               - number of columns in the INPUT matrix
 *        v2             - output vector, with [0..ncol-1] filled with 
 *                 the remapped data, and [ncol..2*ncol-1] to be
 *                 filled with the complement coded data.
 *           NOTE: v2 should have 2*ncol spaces allocated for it.
 *
 *   DESCRIPTION:
 *   This code will complement code a single input vector.  It is written
 *   so Linda can make the plot program work with Fuzzy ARTmap.   To use this
 *   properly, you should probably have either:
 *      1. run the entire input matrix through artmap_complement_code 
 *   OR     2. be plotting this data for a stored set of information in
 *      which the slope and intercept of the input data has already
 *      been calculated (and read in along with the model).  As far
 *      as I can remember, the calculation should always be stored
 *      with the model provided complement coding was done on the
 *      original testing.
 *
 *      If no slope or intercept is provided, we simply copy the
 *      input, and pray for the best.
 *                -RKC
 */
void
artmap_complement_code_vector(int model, float *v1, int ncol, float *v2)
{
  int j;
  artmap_renormalize_vector(model,v1,ncol,v2);
  /* Now create the complement values */
  for(j=0;j<ncol;j++){
    v2[j+ncol] = 1.0 - v2[j];
  }
}
/**************************************************************************
 *   Name:  artmap_renormalize_element(model,v1,feature_index)
 *        model              - the model we're working on.
 *        v1                 - input vector, with [0..ncol-1] filled with 
 *                data
 *        feature_index      - index of the feature we're working on
 *           NOTE: v2 should have 2*ncol spaces allocated for it.
 *
 *   DESCRIPTION:
 *   This code will renormalize a single input vector.  It is written
 *   so Linda can make the plot program work with Fuzzy ARTmap.   To use this
 *   properly, you should probably be plotting this data for a stored set of 
 *   information in which the slope and intercept of the input data has already
 *   been calculated (and read in along with the model).  As far as I can 
 *   remember, the calculation should always be stored with the model provided
 *   complement coding was done on the original testing.
 *
 *      If no slope or intercept is provided, we simply copy the
 *      input, and pray for the best.
 *                -RKC
 */
float
artmap_renormalize_element(int model, float e1, int feature_index)
{
  float *intercept,*slope;
  ARTMAP_PARAMS *ampp;      /* temporary pointer to artmappp[] */

  ampp = artmappp[model];       /* this is the slow way...*/
  intercept = ampp->intercept;
  slope = ampp->slope;

  if(intercept && slope){
    /* First convert each input variable to the range of [0:1] */
    e1 = (e1-intercept[feature_index]) * slope[feature_index];
    if(e1>1.0){
      warning("Artmap renormalizing truncating %f to 1.0\n",e1);
      e1 = 1.0;
    }
    else if (e1<0.0){
      warning("Artmap renormalizing truncating %f to 0.0\n",e1);
      e1 = 0.0;
    }
  }
  return e1;
}
/**************************************************************************
 *   Name:  artmap_unrenormalize_element(model,v1,feature_index)
 *        model              - the model we're working on.
 *        v1                 - input vector, with [0..ncol-1] filled with 
 *                data
 *        feature_index      - index of the feature we're working on
 *           NOTE: v2 should have 2*ncol spaces allocated for it.
 *
 *   DESCRIPTION:
 *   This code will renormalize a single input vector.  It is written
 *   so Linda can make the plot program work with Fuzzy ARTmap.   To use this
 *   properly, you should probably be plotting this data for a stored set of 
 *   information in which the slope and intercept of the input data has already
 *   been calculated (and read in along with the model).  As far as I can 
 *   remember, the calculation should always be stored with the model provided
 *   complement coding was done on the original testing.
 *
 *      If no slope or intercept is provided, we simply copy the
 *      input, and pray for the best.
 *                -RKC
 */
float
artmap_unrenormalize_element(int model, float e1, int feature_index)
{
  float *intercept,*slope;
  ARTMAP_PARAMS *ampp;      /* temporary pointer to artmappp[] */

  ampp = artmappp[model];       /* this is the slow way...*/
  intercept = ampp->intercept;
  slope = ampp->slope;

  if(intercept && slope){
    /* First convert each input variable to the range of [0:1] */
    e1 =  e1/slope[feature_index] + intercept[feature_index];
  }
  return e1;
}

static float 
artmap_complement_element(float e1)
{
  return 1.0 - e1;
}

/*=================== END: Complement Coding code ==========================*/

/*********************************************************************
* NAME: artmap_arrival_order(n,xarray,index)
* DESCRIPTION:
*        Inputs sorted xarray[0...n-1] and provides output indexes
*        in an array index[0...n-1] such that
*        xarray[index[i]] goes from largest to smallest as i goes from 
*        0 to n-1, _AND_ equal values of xarray have the lowest index value 
*        appear first. 
* Returns the lowest equal index, or -1 if none were modified.
*
*************************************************************************/
static int artmap_arrival_order(int n, float *xarray, int *index)
{
  int i,j,cnt;
  int *copy = allocate_ivector(n);
  int *index2 = allocate_ivector(n);
  int first_modified = -1;

  for(i=n-1;i>0;){
    if(xarray[index[i]]==xarray[index[i-1]]){/* Two equal consecutive values*/
      /* Start counting */
      for(cnt=2;i-cnt>=0 && xarray[index[i]]==xarray[index[i-cnt]];cnt++)
    ;
      first_modified = i-(cnt-1);

      for(j=cnt-1;j>=0;j--){
    copy[j] = index[i-j];
      }
      heap_isort(cnt,copy,index2);
      for(j=cnt-1;j>=0;j--){
    index[i--] = copy[index2[j]];
      }
    }
    else
      i--;
  }

  deallocate_ivector(index2);
  deallocate_ivector(copy);
  return(first_modified);
}
      
/*****************************************************************************
 * Name: artmap_fuzzy_thinking(f,I,ro_scaled)
 *    f                     - fuzzy ART net
 *    I                     - input to network
 *    ro_scaled             - ro scaled to ro * norm(I);
 *
 * DESCRIPTION:
 *  Searches for an f2 node in resonance with the input
 *  The data is assumed to be > 0.
 */
static int
artmap_fuzzy_thinking(FUZZY_ART *f, float *I, float ro_scaled)
{
  register int j;
  float *wj;/* Sticking to paper's notation */
  float *Tj;
  float norm_x;            /* |I^wj| */
  int chosen_y;
  float *I_s,*I_e = &I[f->ninputs];

  if(f->num_f2 == 0){
      /*f->num_f2 is zero and allocate won't like it. since we do nothing
    in that case, just return */
      return 0;
  }
  Tj = allocate_vector(f->num_f2);
  /* Compute Tj(I) */
  for(j=0;j<f->num_f2;j++){
    norm_x = 0;
    for(wj=f->w[j],I_s=I;I_s<I_e;I_s++,wj++)
      norm_x += FUZZY_AND(*I_s,*wj);
    (f)->norm_x[j]=norm_x;
    Tj[j] = norm_x*f->inv_norm_w_alpha[j];
  }
  
  rev_heap_sort(f->num_f2, Tj, f->yidx);          /* Sort on Tj */
  artmap_arrival_order(f->num_f2, Tj, f->yidx);   

  /* Search for something above resonance */
  for(chosen_y=0;chosen_y<f->num_f2;chosen_y++){/* Vigilance on norm */
    if(ro_scaled<=f->norm_x[f->yidx[chosen_y]]){
      f->which_y = chosen_y;
      chosen_y = f->yidx[chosen_y];
      d_printf(2,("chosen match[%d]=%g\n",chosen_y,f->norm_x[chosen_y]));
      break; 
    }
    else if (3 <= debug){
      chosen_y = f->yidx[chosen_y];
      printf("Search mismatch on node [%d]= %g\n",
         chosen_y,f->norm_x[chosen_y]);
    }
  }
  if(chosen_y>=f->num_f2)/* Noone above resonance */
    chosen_y = UNASSOCIATED;
  
  deallocate_vector(Tj);
  return(chosen_y);
}
/*****************************************************************************
 * Name: artmap_match_track(f,ro_scaled)
 *    f                     - fuzzy ART net
 *    ro                    - the new ro
 *
 * DESCRIPTION:
 *  Searches for the next best f2 node in resonance with the input
 *  This is called "Match Tracking" in the artmap papers.
 *  Note that because we use fast mapping, we assume ro_ab = 1.0.
 * 
 */

static int
artmap_match_track(FUZZY_ART *f, float ro)
{
  int chosen_y;/* Used to move through the sorted index, and to hold chosen_y*/

  for(chosen_y=f->which_y+1;chosen_y<f->num_f2;chosen_y++){
    if(ro<=f->norm_x[f->yidx[chosen_y]]){
      f->which_y = chosen_y;
      chosen_y = f->yidx[chosen_y];
      d_printf(2,("match_track: chosen match[%d]=%g\n",
          chosen_y,f->norm_x[chosen_y]));
      if(4<=debug) p10("Chosen F2:",f->w[chosen_y],f->ninputs);
      break; 
    }
    else if (3 <= debug){
      printf("match_track: Search mismatch on node [%d]= %g\n",
         f->yidx[chosen_y],f->norm_x[f->yidx[chosen_y]]);
    }
  }
  if(chosen_y>=f->num_f2){/* Noone above resonance */
    chosen_y = UNASSOCIATED;
  }

  return chosen_y;
}

  

/*****************************************************************************
 * Name: artmap_update_wt(beta,wj,nwj,ninputs)
 *  beta                    - Amount of new that is mixed in
 *  wj                      - vector of (old) weights being updated
 *  I                       - vector of (new) weights being mixed in
 *  ninputs                    - length of the vector 
 *
 * DESCRIPTION:
 *  Updates the selected weight vector
 * 
 */
static void
artmap_update_wt(float beta, float *wj, float *I, int ninputs)
{
  register int i;
  float t;
  float beta_complement = 1.0-beta;

  for(i=0;i<ninputs;i++,wj++,I++){
    t = FUZZY_AND(*I,*wj); /* Beware of side-effects */
    *wj = beta*t + beta_complement* *wj;
  }
}
 
/*****************************************************************************
 * Name: artmap_fuzzy_learning(f)
 *    f                     - fuzzy ART net
 *    chosen                - index to chosen weight vector
 *                            UNASSOCIATED if none found
 *    islearning            - are we permitted to learn a new node?
 *    isfastcommit          - are we quickly learning the input?
 *
 * DESCRIPTION:
 *  Determines the rule to update the selected weight vector
 * 
 */
 
static int
artmap_fuzzy_learning(FUZZY_ART *f, float *I, float norm_I,
              int chosen, /* Don't test for range--better be ok */
              int is_learning, int is_fast_commit)
{
  register int i;
  register float *wj;

  if(chosen!=UNASSOCIATED){
    wj = f->w[chosen];
    artmap_update_wt(f->beta,wj,I,f->ninputs);
    f->inv_norm_w_alpha[chosen] = 1/(artmap_norm(wj,f->ninputs)+f->alpha);
  }
  else if(is_learning){/* Create uncommitted node */

    d_printf(2,("Adding new f2 node 0x%x:%d\n",(int)f,f->num_f2));
    chosen = artmap_fuzzy_new_f2(f);
    wj = f->w[chosen];

    if(is_fast_commit){/* Avoid multiplies */
      for(i=0;i<f->ninputs;i++,I++,wj++)
    *wj = *I;
      f->inv_norm_w_alpha[chosen] = 1/(norm_I+f->alpha);
    }
    else {
      wj = f->w[chosen];
      artmap_update_wt(f->beta,wj,I,f->ninputs);
      f->inv_norm_w_alpha[chosen] = 1/(artmap_norm(wj,f->ninputs)+f->alpha);
    }
  }
  else /* chosen = UNASSOCIATED, !is_learning */
    warning("Learning deactivated and no existing node achieves resonance.");

  return chosen;
}
/*****************************************************************************
* NAME: artmap_fuzzy_build(w,ninputs,f2_size)
*       ninputs        - number of dimensions in training vectors
*       w              - pre-defined wt values
*       f2_size        - number of entries in the wt matrix
*
* DESCRIPTION:
*       Allocate storage for fuzzy ART.
*
******************************************************************************/
static FUZZY_ART *
artmap_fuzzy_build(int f2_size, int ninputs)
{
  FUZZY_ART *f;

  f = (FUZZY_ART *)nn_malloc(sizeof(FUZZY_ART),"for fuzzy ART structure");

  if(f2_size==0){
    f2_size = F2_STORE_SIZE;
  }
  (f)->w = allocate_matrix(f2_size,ninputs);
  (f)->inv_norm_w_alpha = allocate_vector(f2_size);

  (f)->alloc_f2 = f2_size;
  (f)->ninputs = ninputs;

  (f)->yidx = allocate_iindex(f2_size);
  (f)->norm_x= allocate_vector(f2_size);
  
  return f;
}
/*****************************************************************************
* NAME: artmap_fuzzy_new_f2(f)
*       ninputs        - number of dimensions in training vectors
*
* DESCRIPTION:
*       Allocate storage for fuzzy ART.  This is inefficient because the
*  matrix allocators set their initial values to 0, while I have to then
*  set it to 1.
*  
* Warning:
*       This routine expects ampp to be set.  Do not call it from outside
*       this file.
*
****************************************************************************/
static int
artmap_fuzzy_new_f2(FUZZY_ART *f) 
{ 
  int chosen = f->num_f2;
  int i;

  if((f)->num_f2==(f)->alloc_f2){
    (f)->alloc_f2 += F2_STORE_SIZE;
    (f)->w = reallocate_matrix((f)->w,(f)->num_f2,(f)->ninputs,(f)->alloc_f2);
    (f)->inv_norm_w_alpha = reallocate_vector((f)->inv_norm_w_alpha,
                          (f)->num_f2,(f)->alloc_f2);
    for(i = f->num_f2;i<f->alloc_f2;i++){
      artmap_fuzzy_wts(f,i);
    }
    (f)->yidx = reallocate_ivector(f->yidx, (f)->num_f2, (f)->alloc_f2);
    for(i = f->num_f2; i < f->alloc_f2; i++){
    f->yidx[i] = i;
    }
    (f)->norm_x = reallocate_vector(f->norm_x, (f)->num_f2,(f)->alloc_f2);
  } 
  (f)->num_f2++; 

  return chosen;
}

/*****************************************************************************
* NAME: artmap_fuzzy_free(f)
*       f              - fuzzy ART net
*
* DESCRIPTION:
*       free storage for fuzzy ART.
*
******************************************************************************/
static void
artmap_fuzzy_free(FUZZY_ART *f)
{
  deallocate_matrix((f)->w);
  deallocate_vector((f)->inv_norm_w_alpha); 
  nn_free((char*)(f)->yidx,"for fuzzy idx");
  nn_free((char*)(f)->norm_x,"for fuzzy norm_x"); 
  nn_free((char*)f,"for fuzzy ART net");
}

/*****************************************************************************
* NAME: artmap_fast_map_resonance(fmf,chosen_f2a,chosen_f2b)
* fmf                               - fast map field
* chosen_f2a,chosen_f2b             - index of chosen f2 nodes
*
* DESCRIPTION:
*       Tests to see if we have resonance between forward and backward ARTs.
*
******************************************************************************/
static int
artmap_fast_map_resonance(FAST_MAP_FIELD *fmf, int chosen_f2a, int chosen_f2b)
{
  artmap_fast_map_correct_size(fmf,chosen_f2a+1);

  if(fmf->wj_ab[chosen_f2a] == UNASSOCIATED){
    artmap_fast_map_learn(fmf,chosen_f2a,chosen_f2b);
    return TRUE;
  }
  if(fmf->wj_ab[chosen_f2a] == chosen_f2b)
    return TRUE;
  else 
    return FALSE;
}
/*****************************************************************************
* NAME: artmap_fast_map_build()
*
* DESCRIPTION:
*       Allocate storage for fast map field
*
******************************************************************************/
static FAST_MAP_FIELD *
artmap_fast_map_build(int alloc_wj)
{
  FAST_MAP_FIELD *fmf;

  if(!alloc_wj) alloc_wj = F2_STORE_SIZE;

  fmf = (FAST_MAP_FIELD *)nn_malloc(sizeof(FAST_MAP_FIELD),
                    "for fast map structure");
  
  fmf->wj_ab = allocate_ivector(alloc_wj);

  fmf->alloc_wj = alloc_wj;
  
  return fmf;
}
/*****************************************************************************
* NAME: artmap_fast_map_free(fmf)
*       fmf              - Fast Map Field
*
* DESCRIPTION:
*       free storage for fast map field
*
******************************************************************************/
static void
artmap_fast_map_free(FAST_MAP_FIELD *fmf)
{
  nn_free((char *)(fmf)->wj_ab,"for wj_ab"); 
  nn_free((char *)fmf,"for fast map field");
}
/*
 * ****************************************************************************
 * NAME: artmap_build(ninputs, nclasses, wa,f2a_size,wb,f2b_size)
 *       ninputs        - number of dimensions in training vectors
 *       nclasses       - number of classes the data falls into
 *       alloc_wj       - size of the inter-ART vector
 *       f2a_size        - size of the wt vector
 *       f2b_size        - size of the wt vector
 *
 * DESCRIPTION:
 *       Allocate storage for model parameters.
 *       This routine used allocate vector to allocate space because other 
 *       LNKnet routines did.  I'm not certain this is a good idea.
 *
 * EXAMPLE:
 *
 *o**************************************************************************/
int
artmap_build(int ninputs, int nclasses, int alloc_wj,
         int f2a_size, int f2b_size)
{
    register int new_model;
    register int i,j;
    ARTMAP_PARAMS *ampp;        /* temporary pointer to artmappp[] */

    new_model =
      model_get_number((void ***)&artmappp,
               &max_models, sizeof(ARTMAP_PARAMS));
    ampp = artmappp[new_model];

    /* Null out things working with complement coding--we'll fill them in
     * if we really need them
     */
    ampp->slope = NULL;
    ampp->intercept = NULL;
    /* set fields which are array sizes */
    ampp->ninputs = ninputs;
    ampp->nfinputs = ninputs;/* Assume until we're told otherwise*/
    ampp->accum_trials = 0;
    ampp->epoch = 0;

    /* set fields that deal with garbage collection, in case we don't
       call gc_init */
    ampp->strict_err_cnt = 0;
    ampp->err_cnt = 0;
    ampp->mark = NULL;

    ampp->ARTa = artmap_fuzzy_build(f2a_size,ninputs);
    ampp->ARTb = artmap_fuzzy_build(f2b_size,nclasses);
    ampp->mf = artmap_fast_map_build(alloc_wj);
    ampp->features = allocate_ivector(ninputs+1);/* This may be twice as large
                            as is strictly necessary,
                            because complement_coding
                            doubles nfinputs. */
    ampp->outputs = allocate_vector(nclasses);

    /* We initialize this here, because we never need to change it, 
     * even when things are re-initialized.  I suspect I'll need to change this
     * when I actually do mappings rather than classification.
     */
    ampp->nclasses = nclasses;
    /* Class_maps are set up so cm[-1] = unassociated */
    ampp->class_maps = allocate_matrix(nclasses+1,nclasses)+1;
    for(i=0;i<nclasses;i++){
      for(j=0;j<nclasses;j++)
    ampp->class_maps[i][j] = (i==j?1.0:0.0);
    }
    return new_model;
}
/*
 * ****************************************************************************
 * NAME: artmap_init()
 * DESCRIPTION:
 *       Initialize model parameters not set in build.
 * Here I allow alpha to differ for the two models if we ever need to make
 * them differ.
 *
 * EXAMPLE:
 *
 *****************************************************************************/
void
artmap_init(int model, int map_flag, int *nn_features, int complement_code,
        int ARTnorm, float ro_a, float beta_a, float alpha_a,
        float ro_b, float beta_b, float alpha_b)
{
    int i;
    ARTMAP_PARAMS *ampp;        /* temporary pointer to artmappp[] */

    ampp = artmappp[model];

    ampp->accum_trials = 0;       /* Definition of initialization */
    ampp->epoch = 0;              /* Definition of initialization */

    ampp->ARTnorm = ARTnorm;

    if(ampp->slope)deallocate_vector(ampp->slope);
    ampp->slope = NULL;
    if(ampp->intercept)deallocate_vector(ampp->intercept);
    ampp->intercept = NULL;
    /* copy feature data into feature list */
    if(complement_code) ampp->nfinputs = (ampp->ninputs)>>1;
      
    for(i = 0; i <= nn_features[0]; i++)
      ampp->features[i] = nn_features[i];

    artmap_fuzzy_init(ampp->ARTa,alpha_a,beta_a,ro_a);
    if(!map_flag){
      artmap_fuzzy_class_init(ampp,ampp->ARTb,alpha_b,beta_b,ro_b);
    }
    else{
      artmap_fuzzy_init(ampp->ARTb,alpha_b,beta_b,ro_b);
    }
    /* Make the map field start out unassociated */
    for(i=0;i<ampp->mf->alloc_wj;i++)
      ampp->mf->wj_ab[i] = UNASSOCIATED;
}
/*
 * ****************************************************************************
 * NAME: artmap_set_model_params()
 * DESCRIPTION:
 *       Initialize model parameters not set in build.
 * Here I allow alpha to differ for the two models if we ever need to make
 * them differ.
 *
 * Note that ro a is not set here--I do it elsewhere, since it's much more
 * common to change that.
 *
 * EXAMPLE:
 *
 *****************************************************************************/
void
artmap_set_model_params(int model, float beta_a, float alpha_a,
            float ro_b, float beta_b, float alpha_b,
            int ARTnorm)
{
    ARTMAP_PARAMS *ampp;        /* temporary pointer to artmappp[] */
    ampp = artmappp[model];
    ampp->ARTnorm = ARTnorm;
    ampp->ARTa->alpha = alpha_a;
    ampp->ARTb->alpha = alpha_b;
    ampp->ARTa->beta = beta_a;
    ampp->ARTb->beta = beta_b;
    ampp->ARTb->ro = ro_b;
}

/*****************************************************************************
* NAME: artmap_free()
* DESCRIPTION:
*       Deallocate storage used in a model 
*
* EXAMPLE:
*
******************************************************************************/
void
artmap_free(int model)
{
    ARTMAP_PARAMS *ampp;        /* temporary pointer to artmappp[] */
    ampp = artmappp[model];
    
    /* free vectors and matrices */
    deallocate_ivector(ampp->features);

    if(ampp->slope)deallocate_vector(ampp->slope);
    if(ampp->intercept)deallocate_vector(ampp->intercept);
    
    deallocate_matrix(ampp->class_maps-1);
    artmap_fuzzy_free(ampp->ARTa);
    artmap_fuzzy_free(ampp->ARTb);
    artmap_fast_map_free(ampp->mf);

    deallocate_vector(ampp->outputs);
    /* allocated in artmap_gc_init() */
    if(ampp->mark != NULL) deallocate_ivector(ampp->mark);

    nn_free((char *)ampp, "for artmap parameter structure");
    artmappp[model] = NULL;
}

/*****************************************************************************
* NAME: artmap_read_fuzzy_art(f,ft,fp)
*   f                                - fuzzy art model
*   fp                               - file pointer
*
* DESCRIPTION:
*   writes parameters and exemplars for a model to a file
*   in ASCII format
*
* EXAMPLE:
*
******************************************************************************/
static int
artmap_read_fuzzy_art(FUZZY_ART *f, int num_f2, int ninputs, FILE *fp)
{
  int rv;
  int i;

  param_setvarptr("alpha",(int *)&f->alpha, fuzzy_sparam, N_SFUZZY);
  param_setvarptr("beta",(int *)&f->beta, fuzzy_sparam, N_SFUZZY);
  param_setvarptr("ro",(int *)&f->ro, fuzzy_sparam, N_SFUZZY);

  if((rv = param_getblk(fp, fuzzy_sparam, N_SFUZZY))<0)
    return rv;
  
  param_setvarptr("w",(int*)&f->w[0][0], fuzzy_aparam, N_AFUZZY);
  param_setdim("w", num_f2, ninputs, fuzzy_aparam, N_AFUZZY);
  
  if((rv = param_getblk(fp, fuzzy_aparam, N_AFUZZY)) >0){
    for(i=0;i<num_f2;i++)
      f->inv_norm_w_alpha[i] = 1/(artmap_norm(f->w[i],f->ninputs)+f->alpha);
  }

  f->num_f2 = num_f2;
  return rv;
}

/*****************************************************************************
* NAME: artmap_read_model()
* 
* Version 2--not compatible with prior file format.
*
* DESCRIPTION:
*       Loads parameters from ascii input file
*   allocates space for and loads data into vectors and matricies
*
*   On error from param_getblk, returns the error value
*   (there is already an error check on the allocation, so we don't
*   have to do that here)
*
******************************************************************************/
int
artmap_read_model(FILE *fp, int *complement_coding,
          float *ro_a, float *beta_a, float *alpha_a,
          float *ro_b, float *beta_b, float *alpha_b,
          int *ARTnorm)
{
    int accum_trials,epoch;
    int model;
    int rv,ninputs,nfinputs,nclasses;
    int alloc_wj;
    int neq;
    int f2a_size,f2b_size;
    ARTMAP_PARAMS *ampp;        /* temporary pointer to artmappp[] */

    /* read in scalars from model file */
    param_setvarptr("ARTnorm",ARTnorm, artmap_scalars, N_SARTMAP);
    param_setvarptr("num_f2a",&f2a_size, artmap_scalars, N_SARTMAP);
    param_setvarptr("num_f2b",&f2b_size, artmap_scalars, N_SARTMAP);
    param_setvarptr("nfinputs",&nfinputs, artmap_scalars, N_SARTMAP);
    param_setvarptr("ninputs",&ninputs, artmap_scalars, N_SARTMAP);
    param_setvarptr("neq",&neq, artmap_scalars, N_SARTMAP);
    param_setvarptr("nclasses",&nclasses,artmap_scalars,N_SARTMAP);
    param_setvarptr("alloc_wj",&alloc_wj,artmap_scalars,N_SARTMAP);
    param_setvarptr("accum_trials", &accum_trials, artmap_scalars, 
            N_SARTMAP);
    param_setvarptr("epoch", &epoch, artmap_scalars,  N_SARTMAP);
    if((rv = param_getblk(fp, artmap_scalars, N_SARTMAP)) < 0) return rv;

    
    /* allocate space for model and vectors*/
    model = artmap_build(ninputs, nclasses, alloc_wj,f2a_size,f2b_size);
    
    ampp = artmappp[model];
       
    if(neq>0){
      ampp->slope = allocate_vector(neq);
      ampp->intercept = allocate_vector(neq);
    }
    ampp->features=allocate_ivector(nfinputs+1);
    ampp->nfinputs = nfinputs;
    ampp->ARTnorm = *ARTnorm;
    ampp->accum_trials = accum_trials;
    ampp->epoch = epoch;
    
    /* read vectors and matrices */
    param_setvarptr ("features", ampp->features, artmap_arrays, N_AARTMAP);
    param_setdim ("features", nfinputs+1, 0, artmap_arrays, N_AARTMAP);
    param_setvarptr ("wj_ab", ampp->mf->wj_ab, artmap_arrays, N_AARTMAP);
    param_setdim ("wj_ab", alloc_wj, 0, artmap_arrays, N_AARTMAP);
    param_setvarptr ("slope", (int *)ampp->slope, artmap_arrays, N_AARTMAP);
    param_setdim ("slope", neq, 0, artmap_arrays, N_AARTMAP);
    param_setvarptr ("intercept", (int *)ampp->intercept,
             artmap_arrays, N_AARTMAP);
    param_setdim ("intercept", neq, 0, artmap_arrays, N_AARTMAP);
    /* read in vectors from model file */
    if((rv = param_getblk(fp, artmap_arrays, N_AARTMAP)) < 0) return rv;
    
    /* read ARTa */
    if((rv = artmap_read_fuzzy_art(ampp->ARTa,f2a_size,ninputs,fp))<0) 
      return rv;

    *alpha_a = ampp->ARTa->alpha;
    *beta_a = ampp->ARTa->beta;
    *ro_a = ampp->ARTa->ro;
    
    /* read ARTb */
    if((rv = artmap_read_fuzzy_art(ampp->ARTb,f2b_size,nclasses,fp))<0) 
      return rv;

    *alpha_b = ampp->ARTb->alpha;
    *beta_b = ampp->ARTb->beta;
    *ro_b = ampp->ARTb->ro;
    *complement_coding = is_complement_coding(ampp);
    return(model);
}
/*****************************************************************************
* NAME: artmap_write_fuzzy_art(f,ft,fp)
*   f                                - fuzzy art model
*   fp                               - file pointer
*
* DESCRIPTION:
*   writes parameters and exemplars for a model to a file
*   in ASCII format
*
* EXAMPLE:
*
****************************************************************************/
int
artmap_write_fuzzy_art(FUZZY_ART *f, FILE *fp)
{
  int rv;

  param_setvarptr("alpha",(int *)&f->alpha, fuzzy_sparam, N_SFUZZY);
  param_setvarptr("beta",(int *)&f->beta, fuzzy_sparam, N_SFUZZY);
  param_setvarptr("ro",(int *)&f->ro, fuzzy_sparam, N_SFUZZY);

  param_setvarptr("w",(int *)&f->w[0][0], fuzzy_aparam, N_AFUZZY);
  param_setdim("w", f->num_f2, f->ninputs,fuzzy_aparam, N_AFUZZY);
  
  if((rv = param_putblk(fp, fuzzy_sparam, N_SFUZZY))<0)
    return rv;

  rv = param_putblk(fp, fuzzy_aparam, N_AFUZZY);
  return rv;
}
/*****************************************************************************
* NAME: artmap_write_model()
* DESCRIPTION:
*   writes parameters and exemplars for a model to a file
*   in ASCII format
* 
* Version 2--not compatible with prior file format.
* EXAMPLE:
*
******************************************************************************/
int
artmap_write_model(int model, FILE *fp)
{
    int rv;
    ARTMAP_PARAMS *ampp;    /* temporary pointer to artmappp[] */
    int alloc_wj;
    int neq;

    ampp = artmappp[model];
    alloc_wj = ampp->ARTa->num_f2;
    neq = ampp->slope?ampp->nfinputs:0;
    /* save scalars */
    param_setvarptr("ARTnorm",&ampp->ARTnorm, artmap_scalars, N_SARTMAP);
    param_setvarptr("num_f2a",&ampp->ARTa->num_f2, artmap_scalars, N_SARTMAP);
    param_setvarptr("num_f2b",&ampp->ARTb->num_f2, artmap_scalars, N_SARTMAP);
    param_setvarptr("nfinputs",&ampp->nfinputs, artmap_scalars, N_SARTMAP);
    param_setvarptr("ninputs",&ampp->ninputs, artmap_scalars, N_SARTMAP);
    param_setvarptr("neq",&neq, artmap_scalars, N_SARTMAP);
    param_setvarptr("nclasses",&ampp->nclasses,artmap_scalars,N_SARTMAP);
    param_setvarptr("alloc_wj",&alloc_wj,artmap_scalars,N_SARTMAP);
    param_setvarptr("accum_trials", &ampp->accum_trials, artmap_scalars, 
            N_SARTMAP);
    param_setvarptr("epoch", &ampp->epoch, artmap_scalars,  N_SARTMAP);
    if((rv = param_putblk(fp, artmap_scalars, N_SARTMAP)) < 0) return rv;

    /* save vectors and matrices */
    param_setvarptr ("features", ampp->features, artmap_arrays, N_AARTMAP);
    param_setdim ("features", ampp->features[0]+1,0,artmap_arrays, N_AARTMAP);
    param_setvarptr ("wj_ab", ampp->mf->wj_ab, artmap_arrays, N_AARTMAP);
    param_setdim ("wj_ab", alloc_wj, 0, artmap_arrays, N_AARTMAP);
    param_setvarptr ("slope", (int *)ampp->slope, artmap_arrays, N_AARTMAP);
    param_setdim ("slope", neq, 0, artmap_arrays, N_AARTMAP);
    param_setvarptr ("intercept",(int *) ampp->intercept, artmap_arrays, N_AARTMAP);
    param_setdim ("intercept", neq, 0, artmap_arrays, N_AARTMAP);
    if((rv = param_putblk(fp, artmap_arrays, N_AARTMAP)) < 0) return rv;

    /* save ARTa */
    if((rv = artmap_write_fuzzy_art(ampp->ARTa, fp)) < 0) return rv;
    /* save ARTb */
    if((rv = artmap_write_fuzzy_art(ampp->ARTb, fp)) < 0) return rv;

    printf("Num ARTa categories stored = %d\n",ampp->ARTa->num_f2);
    printf("Num ARTb categories stored = %d\n",ampp->ARTb->num_f2);

    return(0);
}

/*
 * ***********************************************************************
 * NAME: artmap_param(model, param_name)
 * DESCRIPTION:
 *  Takes model number and parameter name.
 *  Returns a pointer to the named parameter.
 *
 * OBNOXIOUS COMMENTARY:
 *       I cannot imagine why anyone would want to do this, as it only
 *       appears to slow things down.  Why not access the thing via 
 *       a struct enum {...} form.  I can only hope Linda will see this,
 *       decide it makes sense and change this and other code...[RKC]
 *
 * ***********************************************************************/
void *
artmap_param(int model, char *param_name)
{
    void *param_address;
    ARTMAP_PARAMS *ampp;    /* temporary pointer to artmappp[] */
                
    ampp = artmappp[model];
    if (PEQ("accum_trials")) param_address = (void*)&ampp->accum_trials;
    else if (PEQ("ro_a")) param_address = (void*)&ampp->ARTa->ro;
    else if (PEQ("exemplars_a")) param_address = (void*)&ampp->ARTa->w;
    else if (PEQ("features")) param_address = (void*)&ampp->features;
    else if (PEQ("ninputs"))  param_address = (void*)&ampp->ninputs;
    else if (PEQ("nfinputs"))  param_address = (void*)&ampp->nfinputs;
    else if (PEQ("nclasses")) param_address = (void*)&ampp->nclasses;
    else if (PEQ("outputs")) param_address = (void*)&ampp->outputs;
    else if (PEQ("num_f2a")) param_address = (void*)&ampp->ARTa->num_f2;
    else if (PEQ("num_f2b")) param_address = (void*)&ampp->ARTb->num_f2;
    else if (PEQ("ro_b")) param_address = (void*)&ampp->ARTb->ro;
    else if (PEQ("wj_ab")) param_address = (void*)&ampp->mf->wj_ab;
    else {
    warning("Parameter %s not defined", param_name);
    param_address = 0;
    }
    return param_address;
} /* END artmap_param() */

/*****************************************************************************
 * NAME: artmap_inc_ctrain()
 *
 * model                    - which model do we use
 * desired_class                - what is the desired class?
 * input                        - ...for each input vector
 * input_norm                                   - input normalization val.
 * npatterns                    - how many patterns?
 * ninputs                  - how many inputs?
 * is_fast_commit               - type of learning
 *
 * DESCRIPTION:
 *       A subset of this routine could form the basis for a mapping operation
 *  that should be done in artmap_batch_mtrain, since my classes are just 
 *  binary maps.
 * 
 *  The first guess is returned to allow a feeling for the number of errors
 *  that would come if it were forced to classify now.
 *
 * EXAMPLE:
 *       
 * *************************************************************************/
int
artmap_inc_ctrain(int model, float *input, int desired_class, float input_norm,
          int is_fast_commit)
{
  int chosen_f2a,chosen_f2b;
  int first_guess;
  float ro_a_norm;
  ARTMAP_PARAMS *ampp;  /* temporary pointer to artmappp[] */
  ampp = artmappp[model];
    
  /* Learn on output */
  d_printf(2,("\tARTb... class %d (scaled ro = %g)\n",
          desired_class,ampp->ARTb->ro/* totals 1 */));

  chosen_f2b = artmap_fuzzy_thinking(ampp->ARTb,
                     ampp->class_maps[desired_class],
                     ampp->ARTb->ro/* totals 1 */);

  if(chosen_f2b == UNASSOCIATED){/* We save multiplies because the classes
                    never change */
    chosen_f2b = artmap_fuzzy_learning(ampp->ARTb,
                       ampp->class_maps[desired_class],
                       1.0,chosen_f2b,TRUE,is_fast_commit);
  }

  /* Learn on input */
  ro_a_norm = ampp->ARTa->ro*input_norm;
  d_printf(2,("\t\tARTa...\t\t(scaled ro = %g)\n",ro_a_norm));
  if(4<=debug){
    p10("Input:",input,ampp->ninputs);
  }
  chosen_f2a = artmap_fuzzy_thinking(ampp->ARTa, input, ro_a_norm);
  first_guess = ClassIndex(ampp,chosen_f2a);

  while((chosen_f2a != UNASSOCIATED) &&
    !artmap_fast_map_resonance(ampp->mf,chosen_f2a,chosen_f2b)){
    ro_a_norm = ampp->ARTa->norm_x[chosen_f2a] + TINY;
    chosen_f2a = artmap_match_track(ampp->ARTa,ro_a_norm);
  }

  if(chosen_f2a == UNASSOCIATED){/* Search failed gotta learn a new one */
    chosen_f2a = artmap_fuzzy_learning(ampp->ARTa,input,input_norm,
                       chosen_f2a,TRUE,is_fast_commit);    
    artmap_fast_map_correct_size(ampp->mf,chosen_f2a+1);
    artmap_fast_map_learn(ampp->mf,chosen_f2a,chosen_f2b);
  }
  else
    chosen_f2a = artmap_fuzzy_learning(ampp->ARTa,input,input_norm,
                       chosen_f2a,TRUE,is_fast_commit);

  return(first_guess);
}
/*****************************************************************************
 * NAME: artmap_batch_update()
 *
 * model                    - which model do we use
 * npatterns                    - how many patterns?`
 * DESCRIPTION:
 *   updates the model after all data was learned.
 */
int
artmap_batch_update(int model, int npatterns)
{
  ARTMAP_PARAMS *ampp;  /* temporary pointer to artmappp[] */
  ampp = artmappp[model];
  ampp->epoch++;

  ampp->accum_trials+=npatterns;
  
  return(ampp->accum_trials);
}
/****************************************************************************
 * NAME: artmap_gc_init(model)
 *                    
 *
 *   model                          - which ARTmap are we using
 *   input_matrix                   - all input data
 *   desired_classes                - the desired classes, bozo.
 *   n_read                         - number of inputs in matrix
 *   
 * DESCRIPTION:
 * set things up for garbage collection
 *   Mark-and-sweep style garbage collection algorithm to remove unnecessary
 *   f2 nodes.
 *   I assume everything was properly initialized.
 *
 * EXAMPLE:
 *        
 ***************************************************************************/
void
artmap_gc_init(int model)
{
    ARTMAP_PARAMS *ampp;    /* temporary pointer to artmappp[] */
    ampp = artmappp[model];
    
    ampp->mark = allocate_ivector(ampp->ARTa->num_f2);
    ampp->strict_err_cnt = 0;
    ampp->err_cnt = 0;
}
/****************************************************************************
 * NAME: artmap_gc_mark(model,inputs, desired_class, im_norm, threshold)
 *
 * DESCRIPTION:
 *   Mark-and-sweep style garbage collection algorithm to remove unnecessary
 *   f2 nodes.
 *   I assume everything was properly initialized.
 *
 * EXAMPLE:
 *        
 ***************************************************************************/
void artmap_gc_mark(int model, float *inputs, int desired_class,
            float im_norm, int threshold)
{
    int chosen_f2a;
    int class;
    FUZZY_ART *fa;
    float ro_a_norm;
    ARTMAP_PARAMS *ampp;    /* temporary pointer to artmappp[] */

    if(threshold < 0) return;
    
    ampp = artmappp[model];
    fa = ampp->ARTa;
    
    ro_a_norm = fa->ro*im_norm;
    chosen_f2a=artmap_fuzzy_thinking(fa,inputs,ro_a_norm);
        
    class = ClassIndex(ampp,chosen_f2a);
    if(class != desired_class){
    ampp->strict_err_cnt ++;
    if(class != -1)ampp->err_cnt++;
    }
    else{
    ampp->mark[chosen_f2a] ++;
    }
}
/****************************************************************************
 * NAME: artmap_gc_f2(model,input_matrix,n_trials,ninputs,
 *       desired_classes,im_norm,verbosity,threshold)
 *                    
 *
 *   model                          - which ARTmap are we using
 *   input_matrix                   - all input data
 *   desired_classes                - the desired classes, bozo.
 *   n_read                         - number of inputs in matrix
 *   
 * DESCRIPTION:
 *   Mark-and-sweep style garbage collection algorithm to remove unnecessary
 *   f2 nodes.
 *   I assume everything was properly initialized.
 *
 * EXAMPLE:
 *        
 ***************************************************************************/
int
artmap_gc_f2(int model, int n_trials, int ninputs,
         int threshold, int verbosity)
{
    int offset, node_num;
    FUZZY_ART *fa;
    int num_f2a;
    ARTMAP_PARAMS *ampp;    /* temporary pointer to artmappp[] */

    if(threshold < 0) return 0;

    ampp = artmappp[model];
    fa = ampp->ARTa;

    num_f2a = fa->num_f2;
    if (verbosity > 1){
    printf("Deleting unused f2 nodes\n");
    
    printf("BEFORE%5.0f%7.0f%9d%9d\n",
           ((float) ampp->strict_err_cnt / ((float) n_trials)) * 100.0,
           ((float) ampp->err_cnt / ((float) n_trials)) * 100.0,num_f2a,
           ampp->ARTb->num_f2);
    }
    
    /* Sweep up those things not marked */
    for(offset=0, node_num=0; node_num + offset < num_f2a; ){
    /* find a node which is not in use */
    if(ampp->mark[node_num]<=threshold){
        int j;
        do { /* find f2a node in use */
        offset++;
        } while(node_num+offset < num_f2a &&
            (ampp->mark[node_num+offset] <= threshold));
        
        for(; node_num+offset < num_f2a; node_num++){
        if(ampp->mark[node_num+offset] <= threshold){
            /* Make sure we copy to this hole */
            ampp->mark[node_num] = 0;
            break;
        }
        ampp->mf->wj_ab[node_num] = ampp->mf->wj_ab[node_num+offset];
        fa->inv_norm_w_alpha[node_num] =
          fa->inv_norm_w_alpha[node_num+offset];
        for(j=0;j<ninputs;j++)
          fa->w[node_num][j] = fa->w[node_num+offset][j];
        }
    }
    else{
        node_num++;
    }
    }
    /* mark remaining nodes as empty */
    for(; node_num < num_f2a; node_num++){
    ampp->mf->wj_ab[node_num] = UNASSOCIATED;
    artmap_fuzzy_wts(fa, node_num);
    }      
    fa->num_f2 -= offset;
    if(verbosity > 1) 
      printf("Removed %d categories\n",offset);
    
    deallocate_ivector(ampp->mark);
    return(1);
}
/****************************************************************************
 * NAME: artmap_classify()
 * DESCRIPTION:
 *       Implements the artmap to return an integer class and the class
 *       mapping.  Note that with fast learning, this results in perfect
 *       classes rather than probabilistic classes.
 *       
 *   I assume everything was properly initialized.
 *
 * EXAMPLE:
 *        
 ***************************************************************************/
int
artmap_classify (int model, float *inputs, float **outputs, float norm)
{
    int chosen_f2a;
    int class;
    float ro_a_norm;
    ARTMAP_PARAMS *ampp;    /* temporary pointer to artmappp[] */

    ampp = artmappp[model];

    ro_a_norm = ampp->ARTa->ro*norm;
    d_printf(2,("(scaled ro = %g) ",ro_a_norm));
    chosen_f2a=artmap_fuzzy_thinking(ampp->ARTa,inputs,ro_a_norm);

    class = ClassIndex(ampp,chosen_f2a);
    *outputs = ampp->class_maps[class];
    return(class);
}
