/***--------------------------------------------------------------------------
* NAME:     artmap.c
* PURPOSE:  driver for fuzzy artmap classifier
*
* SEE ALSO: artmap_pack.c
* GLOBALS:  int debug;
*
* DESCRIPTION: 
*   Functions in this file:
*       artmap() -- driver for fuzzy artmap classifier
*            -- handles train, create flags
* 
* BUGS AND LIMITATIONS:
* 
*       Note that we now compute the norm of the input vector each time
* we classify it because precision limitation problems gave wacky results
* when vigilance for the "a" ART module was set to 1.0. -RKC
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_test.c,v 11.4 2001/08/23 19:56:03 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_test.c,v $
* CREATED: 13-Nov-91 rkc                MIT LINCOLN LABORATORY
*----------------------- pseudocode of driver follows -----------------------
    norm_read_paramfile - get normalization information from artmap_model
    artmap_read_model - allocates storage, gets parameters from model
    if (one of the basic parameters changed)
       issue a warning--this isn't normally done
    if (one of the critical paramters changed)
       fatal_error - we cannot continue

nn_fopen - open data file

    load_trial_data - read patterns in data file until EOF or npatterns
    
    if complement_coding
       learn complement mapping
       complement code the data
    endif
    if(ARTnorm)
      for all vectors
        artmap_vector_norm - make sum_i(inputs) = 1
    
    for all inputs
        artmap_classify
        add to error file
    end loop
print out final error results
free allocated space
---------------------- end of pseudocode ---------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "artmap.h"
#include "artmap_pack.h"

#define MAXSTR (256)
/* Here we avoid the norm calculation for all cases where the norm is know
   in advance.  Note that we need to compute the norm when vigilance is
   equal to 1.0, because we have floating point representation errors. 

*/
#define FAST_NORM(ro,v) ((ro == 1.0)? artmap_norm((v),ninputs): \
              (ARTnorm? 1.0: complement_coding?nfinputs: \
                (artmap_norm((v),ninputs))))

static char prog_name[MAXSTR];  /* static variables */

void
artmap_test(DEFAULTS_FILE_STRUCT *df_args, int create_flag,
        int nfinputs, /* How many inputs will we read from the file */
        int *features, int npatterns, int debg, int verbosity, int verror, 
        char *data_path, char *data_fname, char *exp_path,
        char *error_fname, char *param_fname,
        /* Artmap specific goodies */
        float ro_a_test, float ro_b, float beta, float alpha, int ARTnorm,
        int complement_coding)
{
  /*
   * Because complement coding requires 2*ninputs for each vector line (where
   * a(i) is at i and a^c(i) is at ninputs+i), we allocate double the amount
   * of needed space and fill the complements in after we've read in the
   * regular line.
   */
    int ninputs = (complement_coding?nfinputs<<1:nfinputs);
    int model, err_model, normalization = 0, data_model;
    int desired_class, class;
    int map_flag = 0;   /* needed for error reporting routines */
    float *inputs, *outputs;
    float *dummy_outputs; /* No match case */
    char *tmp_path, *open_type;
    FILE *error_file, *param_file;
    int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());

    dummy_outputs = allocate_vector(ninputs);

    /*
     * do the initialization of the training data structure, reading the
     * previously stored training vectors in the case of retrain and test also
     * initialize the normalization routines, if needed.
     */
    {
      int rv; /* Local Error read value */
      int t_complement_coding;
      float t_ro_a,t_ro_b,t_beta,t_alpha;
      int t_ARTnorm;

    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of artmap model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
        fatal_error ("%s: read error %d %s %s",
          prog_name, rv, "on normalization parameter file", tmp_path);
    model = artmap_read_model(param_file,&t_complement_coding,
                  &t_ro_a,&t_beta,&t_alpha,
                  &t_ro_b,&t_beta,&t_alpha,&t_ARTnorm);
    if(model < 0)
      fatal_error("%s: error reading model file %s", prog_name, tmp_path);

      if(t_complement_coding!=complement_coding)
    fatal_error(
         "%s: %scomplement coding in train, %scomplement coding in test\n",
            prog_name,(t_complement_coding?"":"no "),
            (complement_coding?"":"no "));

      if(t_ro_b!=ro_b || t_beta!=beta || t_alpha!=alpha || t_ARTnorm!=ARTnorm){
    warning("Changing one of ro_b=%f->%f\n alpha=%f->%f,beta=%f->%f ARTnorm=%d->%d\n", 
        ro_b,t_ro_b,alpha,t_alpha,beta,t_beta,ARTnorm,t_ARTnorm);
    artmap_set_model_params(model,beta,alpha,ro_b,beta,
                alpha,ARTnorm);
      }

    nn_fclose(param_file, tmp_path, "for read of artmap model file");

    verify_model_params (model,
        nfinputs, *(int *) artmap_param(model, "nfinputs"),
        features, *(int **) artmap_param(model, "features"));
    if(df_args->noutputs != *(int *)artmap_param(model,"nclasses"))
      fatal_error("%s: nclasses (=%d) of model is not the same as %d\n",
              prog_name,(int *)artmap_param(model,"nclasses"),df_args->noutputs);
  }

    if (verror>0) {     /* open file for classification errors */
    if (create_flag) open_type = "w";
    else open_type = "a";
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  open_type, "for classification errors");
    } else {
    error_file = NULL;
    }
    /* do remaining initializations and allocations */
    
    err_model = report_build (verbosity, verror, df_args->noutputs+1);/* For Novel */
    report_init (err_model, map_flag, ninputs, CENTERS, 0, post_sum_to_one);

    if(df_args->output_labels != NULL){
    char **nlabels;
    char *new_label_len;
    int i;

    /* copy the lable list, adding the class "novel" */
    nlabels = (char **)allocate_pvector(df_args->noutputs+2);
    /* add one to the strlen for the \0 and one in case the
       number is bigger */
    new_label_len = nn_malloc(strlen(df_args->output_labels[0])+2,
                  "for new labels list");
    sprintf(new_label_len,"%d",df_args->noutputs+1);
    nlabels[0] = new_label_len;
    for(i = 1; i <= df_args->noutputs; i++)
      nlabels[i] = df_args->output_labels[i];
    nlabels[df_args->noutputs+1] = nn_strdup("Novel");
    /* free the old label list.  This must be ALLOCATED above */
    nn_free(df_args->output_labels[0], "for bigger copy of output labels");
    deallocate_pvector((void **)df_args->output_labels);
    df_args->output_labels = nlabels;
    }

    /* set up for reading data */
    data_model =
      class_data_single_init(data_path, data_fname, npatterns,
                 df_args->ninputs, ninputs, df_args->noutputs, features,
                 normalization, verbosity);

    if(ro_a_test < 0 || ro_a_test > 1)
      fatal_error("missing ro_a parameter");
    *(float*)artmap_param(model,"ro_a") = ro_a_test;/* Set ro_a to input */

    /* read and classify input from data file */
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
      if(complement_coding){/* slope and intercept read elsewhere */
    artmap_complement_code_vector(model,inputs,nfinputs,inputs);
      }
/*
 * You probably don't want to do this with complement coding, but who am I to
 * say? 
 */
      if(ARTnorm) artmap_vector_norm(inputs,ninputs);

      /* classify pattern and report errors */
      class=artmap_classify(model, inputs, &outputs,FAST_NORM(ro_a_test,inputs));
    if(class == -1){
      class = df_args->noutputs;
      outputs = dummy_outputs;
    }
      report_class_trial(err_model, error_file, inputs, outputs,
             desired_class, class,
             class_data_get_input_num(data_model));
    }

    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, df_args->output_labels);
    report_test_result(err_model, stdout, data_fname);
    report_free(err_model);
    if(verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
        "for error file");
    }
    
    class_data_exit(data_model);

    /*free space and close files, for neatness sake */
    artmap_free(model);
    norm_free();
    deallocate_vector(dummy_outputs);
}
