/***--------------------------------------------------------------------------
 * NAME:        artmap_train.c
 * PURPOSE: training driver for fuzzy artmap classifier
 *
 * SEE ALSO: artmap_pack.c
 * GLOBALS:  int debug;
 *
 * DESCRIPTION: 
 *  Functions in this file:
 *      artmap() -- driver for fuzzy artmap classifier
 *           -- handles train, create flags
 * 
 * BUGS AND LIMITATIONS:
 * 
 *       Note that we now compute the norm of the input vector each time
 * we classify it because precision limitation problems gave wacky results
 * when vigilance for the "a" ART module was set to 1.0. -RKC
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_train.c,v 11.4 2001/08/23 19:56:03 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/artmap/artmap_train.c,v $
 * CREATED: 13-Nov-91 rkc               MIT LINCOLN LABORATORY
 *----------------------- pseudocode of driver follows -----------------------
 if create_flag
 artmap_build - allocate storage
 artmap_init - fills in values, put labels in structure
 if normalize read_norm_file - get normalization information
 else
 norm_read_paramfile - get normalization information from artmap_model
 artmap_read_model - allocates storage, gets parameters from model
 if (one of the basic parameters changed)
 issue a warning--this isn't normally done
 if (one of the critical paramters changed)
 fatal_error - we cannot continue
 endif
 
 nn_fopen - open data file
 
 allocate storage for data
 load_trial_data - read patterns in data file until EOF or npatterns
 if complement_coding
 learn complement mapping
 complement code the data
 endif
 if(ARTnorm)
 for all vectors
 artmap_vector_norm - make sum_i(inputs) = 1
 
 artmap_epochs_ctrain
 artmap_gc_f2 - remove unused nodes
 norm_write_paramfile - write normalization information
 artmap_write_model - write model parameters
 print out final error results
 free allocated space
 ---------------------- end of pseudocode -------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "artmap.h"
#include "artmap_pack.h"

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)
/* Here we avoid the norm calculation for all cases where the norm is know
   in advance.  Note that we need to compute the norm when vigilance is
   equal to 1.0, because we have floating point representation errors. 
   
   */
#define FAST_NORM(ro,v) ((ro == 1.0)? artmap_norm((v),ninputs): \
             (ARTnorm? 1.0: complement_coding?nfinputs: \
              (artmap_norm((v),ninputs))))

static char prog_name[MAXSTR];  /* static variables */

void
artmap_train(DEFAULTS_FILE_STRUCT *df_args, int create_flag,
         int nfinputs, /* How many inputs will we read from the file */
         int *features, int normalize, int npatterns, int debg,
         int verbosity, int verror, char *data_path, char *data_fname,
         char *norm_fname, char *exp_path, char *error_fname,
         char *param_fname, char *comment, int adjust_priors,
         int priors_npatterns, float *priors_list, int nepochs,
         int random_flag, int train_cycle,
         /* Artmap specific goodies */
         float ro_a_train, float ro_b, float beta, float alpha,
         int ARTnorm, int complement_coding, int is_fast_commit,
         int gc_threshold)
{
    /*
     * Because complement coding requires 2*ninputs for each vector line (where
     * a(i) is at i and a^c(i) is at ninputs+i), we allocate double the amount
     * of needed space and fill the complements in after we've read in the
     * regular line.
     */
    int ninputs = (complement_coding?nfinputs<<1:nfinputs);
    int model, normalization = 0, err_model, err_part_model, data_model;
    int desired_class, class, t_index;
    int n_read, n_trials;
    int db_type = CLASS_DATA;   /* needed for error reporting routines */
    float *outputs; /* No match case */
    float *inputs;
    float **coded_matrix;
    float *im_norm;
    char *tmp_path;
    FILE *error_file, *param_file;
    int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());
    
    /*
     * do the initialization of the training data structure, reading the
     * previously stored training vectors in the case of retrain and test also
     * initialize the normalization routines, if needed.
     */
    
    if(create_flag){
    /* allocate and initialize model structure */
    model = artmap_build(ninputs, df_args->noutputs, 0, 0 ,0);
    artmap_init(model,db_type,features,complement_coding,ARTnorm,
            ro_a_train,beta,alpha,ro_b,beta,alpha);
    
    if (normalize) 
      normalization = read_norm_file(path_build(data_path, norm_fname));
    } else {            /* test or continue training */
    int rv; /* Local Error read value */
    int t_complement_coding;
    float t_ro_a,t_ro_b,t_beta,t_alpha;
    int t_ARTnorm;

    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of artmap model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error ("%s: read error %d %s %s",
               prog_name, rv, "on normalization parameter file",
               tmp_path);
    model = artmap_read_model(param_file,&t_complement_coding,
                  &t_ro_a,&t_beta,&t_alpha,
                  &t_ro_b,&t_beta,&t_alpha,&t_ARTnorm);
    if(model < 0)
      fatal_error("%s: error reading model file %s", prog_name, tmp_path);
    
    if(t_complement_coding != complement_coding)
      fatal_error(
        "%s: %scomplement coding in train, %scomplement coding in test\n",
              prog_name,(t_complement_coding?"":"no "),
              (complement_coding?"":"no "));
    
    if(t_ro_b != ro_b || t_beta != beta || t_alpha != alpha ||
       t_ARTnorm != ARTnorm){
        warning("Changing one of ro_b=%f->%f\n alpha=%f->%f,beta=%f->%f ARTnorm=%d->%d\n", 
            ro_b,t_ro_b,alpha,t_alpha,beta,t_beta,ARTnorm,t_ARTnorm);
        artmap_set_model_params(model,beta,alpha,ro_b,beta,
                    alpha,ARTnorm);
    }
    
    nn_fclose(param_file, tmp_path, "for read of artmap model file");
    
    verify_model_params (model, nfinputs,
                 *(int *) artmap_param(model, "nfinputs"),
                 features,
                 *(int **) artmap_param(model, "features"));
    if(df_args->noutputs != *(int *)artmap_param(model,"nclasses"))
      fatal_error("%s: nclasses (=%d) of model is not the same as %d\n",
        prog_name,(int *)artmap_param(model,"nclasses"),df_args->noutputs);
    }
    
    if (verror>0) {     /* open file for classification errors */
    if (create_flag)
      error_file = nn_fopen(path_build(exp_path, error_fname),
                "w", "for classification errors");
    else
      error_file = nn_fopen(path_build(exp_path, error_fname),
                "a", "for classification errors");
    } else error_file = NULL;
    /* initialize error reporting */
    /* add an entry to the error report for the class "Novel" */
    err_model = report_build (verbosity, verror, df_args->noutputs+1);
    report_init (err_model, db_type, ninputs, ARTMAP_COST,
         FALSE, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, df_args->noutputs+1);

    if(df_args->output_labels != NULL){
    char **nlabels;
    char *new_label_len;
    int i;

    /* copy the lable list, adding the class "novel" */
    nlabels = (char **)allocate_pvector(df_args->noutputs+2);
    /* add one to the strlen for the '\0' and one in case the
       number is bigger */
    new_label_len = nn_malloc(strlen(df_args->output_labels[0])+2,
                  "for new label list");
    sprintf(new_label_len,"%d",df_args->noutputs+1);
    nlabels[0] = new_label_len;
    for(i = 1; i <= df_args->noutputs; i++)
      nlabels[i] = df_args->output_labels[i];
    nlabels[df_args->noutputs+1] = nn_strdup("Novel");
    /* free the old label list.  This must be ALLOCATED above */
    nn_free(df_args->output_labels[0], "for bigger copy of output labels");
    deallocate_pvector((void **)df_args->output_labels);
    df_args->output_labels = nlabels;
    }
    
    outputs = allocate_vector(ninputs);
    im_norm = allocate_vector(npatterns);

    /* open data file */
    /* put together the full path to the data file */
    data_model = class_data_read(data_path, data_fname, npatterns, df_args->ninputs,
                 nfinputs, df_args->noutputs, features, normalization,
                 verbosity);
    n_trials = class_data_get_npatterns(data_model);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
      class_data_priors_init(data_model, priors_list, priors_npatterns);

    /* fix up normalization for complement coding and inputs
       between 0 and 1 */
    { 
    float **input_matrix = class_data_get_input_matrix(data_model);
    if(complement_coding)
      coded_matrix = allocate_matrix(npatterns, ninputs);
    else coded_matrix = input_matrix;

    *(float*)artmap_param(model,"ro_a") = ro_a_train;/*Set ro_a to input*/
    if(complement_coding){
        artmap_learn_complement_mapping(model,
                        input_matrix,npatterns,nfinputs);
        artmap_complement_code(model,input_matrix,
                   npatterns,nfinputs,coded_matrix);
    }
    
    /* You probably don't want to do this with complement coding. */
    if(ARTnorm){
        for(n_read=0;n_read<n_trials;n_read++)
          artmap_vector_norm(coded_matrix[n_read],ninputs);
    }
    /*Compute limited-precision norm*/
    for(n_read=0;n_read<n_trials;n_read++){
        im_norm[n_read] = FAST_NORM(ro_a_train,coded_matrix[n_read]);
    }
    }
    
    /*********************************************/
    /* priors sampling may have changed n_trials */
    n_trials = class_data_get_npatterns(data_model);
    report_epoch_header(err_model, stdout, n_trials);

    /* code taken from artmap_epochs_ctrain */
    { /* needed for stopping training when train_cycle is set */
    int epoch;
    int errs_1ago = -1, errs_now, err_cnt;
    errs_now = 0;

    for(epoch=0; epoch < nepochs; epoch++){
        /* train_cycle flag says to continue training until
           errs doesn't change 2 times in a row */
        if(train_cycle && errs_now == errs_1ago) break;
        errs_1ago = errs_now; /* ignore if train_cycle is not set */
        err_cnt = errs_now = 0;

        class_data_epoch_init(data_model);
        report_init(err_part_model, db_type, ninputs, ARTMAP_COST,
            FALSE, post_sum_to_one);
        /* code taken from artmap_batch_ctrain */
        for(n_read = 0; n_read < n_trials; n_read++){
        /* have class_data pick a pattern for us */
        inputs = class_data_get_next(data_model, &desired_class);
        if(inputs == NULL) break;
        t_index = class_data_get_input_num(data_model);
        /* switch back to the complement coded data */
        inputs = coded_matrix[t_index];
        class = artmap_inc_ctrain(model,inputs, desired_class,
                      im_norm[t_index],is_fast_commit);
        report_class_trial(err_model, error_file,
                   inputs, outputs, desired_class, class,
                   t_index);
        report_class_trial(err_part_model, error_file,
                   inputs, outputs, desired_class, class,
                   t_index);
        /* keep track of errors for stopping loop and
           for reporting errors assuming Novel is correct */
        if(class != desired_class){
            errs_now ++;
            if(class != -1)err_cnt++;
        }
        }
        artmap_batch_update(model,n_trials);
        
        report_artmap_epoch(err_part_model, stdout, epoch, err_cnt,
                *(int*)artmap_param(model,"num_f2a"),
                *(int*)artmap_param(model,"num_f2b"));
        /*******************************************************/
    }
    report_last_epoch(err_part_model, stdout, epoch-1);
    }
    /* pruning */
    /* run through the data one last time to remove unused f2 nodes */
    class_data_epoch_init(data_model);
    artmap_gc_init(model);
    for(n_read = 0; n_read < n_trials; n_read++){
    /* have class_data pick pattern for us */
    inputs = class_data_get_next(data_model, &desired_class);
    if(inputs == NULL) break;
    t_index = class_data_get_input_num(data_model);
    /* go back to the complement coded version */
    inputs = coded_matrix[t_index];
    artmap_gc_mark(model, inputs, desired_class,
               im_norm[t_index], gc_threshold);
    }
    if(!artmap_gc_f2(model, n_read, ninputs, gc_threshold,verbosity)
       && verbosity>1) {
    int arta_nodes, artb_nodes;
      
    arta_nodes = *(int *)artmap_param(model, "num_f2a");
    artb_nodes = *(int *)artmap_param(model, "num_f2b");
    printf("Number of f2 Nodes, a=%9d, b=%9d\n", arta_nodes, artb_nodes);
    }
/*****************************************************/    
    /* write training data to model file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of artmap model file");
    /* write comment */
    fprintf(param_file, "%s", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "artmap model\n");
    artmap_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of artmap model file");
    
    if ( verbosity > 0)
      printf("\nFinished -- artmap model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, df_args->output_labels);
    report_free(err_part_model);
    report_free(err_model);
    if(verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }
    
    /*free space and close files, for neatness sake */
    norm_free();
    artmap_free(model);
    class_data_exit(data_model);
    deallocate_vector(im_norm);
    deallocate_vector(outputs);
    if(complement_coding) deallocate_matrix(coded_matrix);
}
