/* (c) Copyright 1999 M.I.T. Lincoln Laboratory 

   Permission is hereby granted, without written agreement or royalty fee, 
   to use, copy, modify, and distribute this software and its 
   documentation for any purpose, provided that the above copyright notice 
   and the following three paragraphs appear in all copies of this 
   software. 

     IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR 
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF 
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN 
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  M.I.T. LINCOLN 
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT. 

        THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN 
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, 
ENHANCEMENTS, OR MODIFICATIONS. 
*/

/* bintree2c.c */
/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#include <stdio.h>
#include <nnlib.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif

#define bintree_pack_h_internal_types
#include <bintree_pack.h>
#define GEN2CVARS
#include <gen_2c_args.h>

static int max_nodes = 0;
static int prune_tree = 0;

static ARG argtab2[] = 
{
    {"prune_tree", BOOLEAN, &prune_tree,
       "prune tree to max_nodes nodes. Default, use full tree"},
    {"max_nodes", INT, &max_nodes, "maximum non-terminal nodes after pruning"},
};
#define TABSIZE2 (sizeof(argtab2)/sizeof(ARG))

/* bintree specific variables */
static float *train_priors_list;
static int linear;
static int nbins;
static BTNODE **bins;
static char nbins_name[64];

/* function declerations */

int read_model_file(char **argv);
void free_model(int model);

void print_comments(void);
void print_declarations(void);
void print_macros(void);
void clear_macros(void);
void print_classify_code(void);
void print_dot_is_leq_code(void);

void set_priors(int model);

void get_params(int model);
void get_general_params(int model);
void get_specific_params(int model);

void compact_pruned_tree (void);

int
main(int argc, char **argv)
{
    int model;

    /* process arguments */
    ARG *argtab;
    int tabsize;

    /* process arguments */
    nnargs_combine_argtabs(argtab2c, TABSIZE2C, argtab2, TABSIZE2,
               &argtab, &tabsize);
    argc = nnargs(argc, argv, argtab, tabsize);

    /* check arguments */
    if(strlen(class_suff) > 53){
    fatal_error("suffix (%s) is too long.  use at most 50 characters",
            class_suff);
    }

    /* set up subroutine names */
    strcpy(classify_name, "classify");
    strcpy(normalize_name, "normalize");
    strcpy(nraw_name, "NRAW");
    strcpy(ninputs_name, "NINPUTS");
    strcpy(noutputs_name, "NCLASSES");
    strcpy(nbins_name, "NNODES");
    if(class_suff[0] != '\0'){
        strcpy(class_exten, "_");
    strcat(class_exten, class_suff);
    strcat(classify_name, class_exten);
    strcat(normalize_name, class_exten);
    strcat(nraw_name, class_exten);
    strcat(ninputs_name, class_exten);
    strcat(noutputs_name, class_exten);
    strcat(nbins_name, class_exten);
    } else {
      strcpy(class_exten, "");
    }

    /* read in parameters */
    model = read_model_file(argv);

    /* prune tree */
    fprintf(stderr, "prune to %d nodes:", max_nodes);
    max_nodes = bintree_test_init(model, !prune_tree, max_nodes, 0, NULL);
    fprintf(stderr, " %d nodes in pruned tree\n", max_nodes);

    get_params(model);
    if(prune_tree){
    compact_pruned_tree();
    }

    set_priors(model);

    /* print the source file */
    print_comments();
    print_declarations();
    print_macros();
    print_classify_code();
    print_normalize_code(normalize_name);

    /* print bintree only functions */
    if(linear) print_dot_is_leq_code();

    clear_macros();
    free_model(model);
    nnargs_free_argtab(argtab, tabsize);
    exit(0);
}

int
read_model_file(char **argv)
{
    int   model, norm_model, normalization;
    FILE *param_file;

    param_file = nn_fopen(param_fname, "r", "for read of model file");
    norm_model = norm_read_paramfile(param_file, &normalization);
    if (norm_model < 0)
      fatal_error("%s: read error %d %s %s", argv[0], norm_model,
          "on normalization parameter file", param_fname);
    
    model = bintree_read_model(param_file);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s",
          argv[0], model, param_fname);
    nn_fclose(param_file, param_fname, "for read of model file");

    return((int) model);
}

void
free_model(int model)
{
  norm_free();
  bintree_free(model);
}

void
get_params(int model)
{
    get_specific_params(model);
    get_general_params(model);
    get_norm_params(model, ninputs, features);
}

void
set_priors(int model)
{
    int i;
    float total;

    /* now that we know noutputs, we can check the priors settings */
    if(adjust_test){
    if(test_uniform_priors){
        if(test_priors_list != NULL)
          deallocate_vector(test_priors_list);
        test_priors_list = allocate_vector(noutputs+1);
        test_priors_list[0] = noutputs;
        for(i = 1; i <= noutputs; i++)
          test_priors_list[i] = 1./(float)noutputs;
    } else {
        if(test_priors_list[0] != (float)noutputs)
          fatal_error("%s: %s %.0f, should be %d",
              "test_priors_list length is", test_priors_list[0],
              noutputs);
        total = 0.;
        for(i = 1; i <= noutputs; i++) total += test_priors_list[i];
        for(i = 1; i <= noutputs; i++) test_priors_list[i] /= total;
    }
    train_priors_list = *(float **)bintree_param(model, "class_cnt");
    total = (float)*(int *)bintree_param(model, "total_npatterns");
    for(i = 0; i < noutputs; i++) train_priors_list[i] /= total;
    }
}

void
print_comments(void)
{
    printf("/* %s */\n", param_fname);
    print_norm_comments();
}

void
print_declarations(void)
{
    printf("\n");
    printf("#include <stdio.h>\n");
    printf("#include <math.h>\n");
    printf("\n");

    printf("\n /* function declarations */\n");
    printf("extern int %s(/* float *,float * */);\n", classify_name);
    printf("extern int %s(/* float * */);\n", normalize_name);
    if(linear){
    printf("static int dot_is_leq(/* float *,float *,float,int*/);\n");
    }

    printf("\n\n");
}

void
print_macros(void)
{
    print_norm_macros(nraw_name, ninputs_name);
    printf("#define %s %d\n", noutputs_name, noutputs);
    printf("#define %s %d\n", nbins_name, nbins);
    printf("\n");
}    

void
clear_macros(void)
{
    clear_norm_macros();
    printf("#undef %s\n", noutputs_name);
    printf("#undef %s\n", nbins_name);
    printf("\n");
}

void
print_classify_code(void)
{
    int   i, j;
    BTLINK node;

    printf("int %s (inputs, outputs)\n", classify_name);
    printf("   float *inputs,*outputs;\n");
    printf("{\n");
    printf("  int id;\n");
    printf("  struct Node {\n");
    printf("    int id, yesid, noid;\n");
    if(linear){
    printf("    float beta[%s];\n", ninputs_name);
    } else {
    printf("    int feature;\n");
    }
    printf("    float constant;\n");
    printf("    int class;\n");
    printf("    float outputs[%s];\n", noutputs_name);
    printf("  };\n");

    printf("  static struct Node node_list[%s]={\n", nbins_name);
    for(i = 0; i < nbins; i++){
    node = bins[i];
    printf("  { %d, %d, %d,", node->selfno, node->yesno, node->nono);
    if(linear){
        printf(" {");
        for(j = 0; j < ninputs; j++)
          printf("%g,", node->beta[j]);
        printf("},");
    } else {
        printf(" %d,", node->split_input);
    }
    printf(" %g, %d, {",node->split_constant, node->class);
    for(j = 0; j < noutputs; j++) printf("%g,", node->class_distribs[j]);
    printf("},},\n");
    }
    printf("};\n");
    printf("  struct Node *nodep;\n");
    if(adjust_test){
    print_vector_initialization("new_priors_list", &test_priors_list[1],
                    noutputs, noutputs_name);
    print_vector_initialization("old_priors_list", train_priors_list,
                    noutputs, noutputs_name);
    printf("  float total;\n");
    }
    printf("  int best;\n");

    /* copy and normlize inputs */
    print_copy_inputs();

    printf("\n  /* normalize loaded data */\n");
    printf("  %s(x);\n", normalize_name);

    printf("\n  /* find the terminal node this pattern falls in */\n");
    printf("  id = 0;\n");
    printf("  nodep = &node_list[id];\n");
    printf("  while(nodep->yesid != -1 && nodep->noid != -1){\n");
    if(linear){
    printf("    if(dot_is_leq(x, nodep->beta,nodep->constant, %s))\n",
           ninputs_name);
    } else {
    printf("    if(x[nodep->feature] <= nodep->constant)\n");
    }
    printf("      id = nodep->yesid;\n");
    printf("    else id = nodep->noid;\n");
    printf("    nodep = &node_list[id];\n");
    printf("  }\n");
    printf("\n  /* nodep is a terminal node */\n");
    printf("  /* set outputs */\n");
    printf("  for(i = 0; i < %s; i++) ", noutputs_name);
    printf("outputs[i] = nodep->outputs[i];\n");

    if(adjust_test){
    printf("\n  /* scale outputs by priors */\n");
    printf("  for(i = 0; i < %s; i++){\n", noutputs_name);
    printf("    outputs[i] *= new_priors_list[i];\n");
    printf("    outputs[i] /= old_priors_list[i];\n");
    printf("  }\n");
    printf("  /* rescale the new outputs to sum to 1 again */\n");
    printf("  total = 0;\n");
    printf("  for(i = 0; i < %s; i++) total += outputs[i];\n",
           noutputs_name);
    printf("  for(i = 0; i < %s; i++) outputs[i] /= total;\n",
           noutputs_name);
    printf("\n  /* find highest output */\n");
    printf("  for(best = i = 0; i < %s; i++)\n", noutputs_name);
    printf("    if(outputs[best] < outputs[i]) best = i;\n");
    printf("  return(best);\n");
    printf("}\n");    
    } else {
    printf("\n");
    printf("  return(nodep->class);\n");
    printf("}\n");    
    }
}

/* definitions of functions which initialize  *
 * the static (global) variables of this file */

void
get_specific_params(int model)
{
    /* get bintree specific parameters */

    /* (1) get parameters from the arguments list of bintree */
    /* (2) get parameters from bintree model */
    bins = *(BTNODE ***)bintree_param(model, "bins");
    nbins = *(int *)bintree_param(model, "nbins");
    linear = *(int *)bintree_param(model, "linear");
}

void
get_general_params(int model)
{
    void *bintree_param(int model, char *param_name);

    /* get general parameters */

    ninputs = *(int *)bintree_param(model, "ninputs");
    noutputs = *(int *)bintree_param(model, "nclasses");
    features = *(int **)bintree_param(model, "features");
}

void
print_dot_is_leq_code(void)
{
    printf("static int\n");
    printf("dot_is_leq(x, beta, c, ninputs)\n");
    printf("  float *x, *beta, c;\n");
    printf("  int ninputs;\n");
    printf("{\n");
    printf("  int i;\n");
    printf("  float c2 = 0.;\n");
    printf("\n  for(i = 0; i < ninputs; i++)\n");
    printf("    c2 += x[i]*beta[i];\n");
    printf("  return(c2 <= c);\n");
    printf("}\n");
}

static BTNODE **new_bins;

void
compact_pruned_tree(void)
{
    int i, j;
    BTNODE *node;

    new_bins = (BTNODE **)allocate_pvector(nbins);

    i = 0; j = 1;
    new_bins[0] = bins[0];

    while(i < j){
    node = new_bins[i];
    if(node->yeschild != NULL && node->nochild != NULL){
        node->yeschild->selfno = node->yesno = j;
        new_bins[j] = node->yeschild;
        j++;
        node->nochild->selfno = node->nono = j;
        new_bins[j] = node->nochild;
        j++;
    } else {
        node->nono = node->yesno = -1;
    }
    i++;
    }
    bins = new_bins;
    nbins = j;
}
