/***-------------------------------------------------------------------------
 * NAME:    bintree2mlp.c
 * PURPOSE: Initialize an mlp with bintree parameters
 * USAGE:
 *
 * SEE ALSO:    
 *
 * DESCRIPTION: Reads in a bintree parameter file, prunes the tree,
 *      translates the parameters into an mlp and writes the mlp
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree2mlp.c,v 11.4 2001/08/23 20:11:25 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree2mlp.c,v $
 * CREATED: 2-Nov-93 lck            MIT LINCOLN LABORATORY
 -------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree2mlp.c,v 11.4 2001/08/23 20:11:25 craig Exp $";

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define bintree_pack_h_internal_types
#define bintree_pack_h_internal_macros
#include <bintree_pack.h>
#include <mlp_pack.h>

/* bintree args */
static char *bin_param_fname = "bintree.param";
static int max_nodes = 0;
static int prune_tree = 0;
/* mlp args */
static char *mlp_param_fname = "mlp.param";
static int *nnodes = NULL;
static float    etta=0.1;
static int  use_etta_list = 0;
static float    *etta_list = NULL;
static int  hfunction = 0;
static int  ofunction = 0;
static float    max_magnitude = .1;
static int  etta_tau = 0;
static int  step_change=0;  /* 0=no mas, 1=multiple adaptive stepsize,
                 2=constant then reduce*/
static int  seed = 0;

static ARG bintree2mlp_argtab[] = 
{
    /************* bintree args ******************/
    {"bin_fparam", STRING, (int *)&bin_param_fname,
       "name of bintree parameter file"},
    {"max_nodes", INT, &max_nodes, "maximum non-terminal nodes after pruning"},
    {"prune_tree", BOOLEAN, &prune_tree,
       "prune tree to max_nodes nodes. Default, use full tree"},
    /************* mlp args ************/
    {"mlp_fparam", STRING, (int *)&mlp_param_fname,
       "name of mlp parameter file"},
    {"nodes", INT_ALLOC_LIST, (int *)&nnodes,
       "number of nodes in each layer of mlp"},
    {"etta_change_type", INT, &step_change,
       "0: no change 1: constant then c/ntrials reduction 2: multiple-adaptive step size"},
    {"etta", FLOAT, (int *) &etta, "initial stepsize for mas"},
    {"use_etta_list", BOOLEAN, &use_etta_list,
       "use etta_list as stepsize for each layer. default: use etta for all"},
    {"etta_list", FLOAT_ALLOC_LIST, (int *)&etta_list,
       "list of initial stepsizes each mlp layer for mas"},
    {"etta_tau", INT, &etta_tau,
       "number of trials to hold etta constant before reduction (etta_nepochs is used for classifier)"},
    {"hfunction", INT, &hfunction,
       "hidden layer sigmoid 0=sigm(0,1),1=sym sigm(-1,1), 3=step, 4=linear"},
    {"ofunction", INT, &ofunction,
       "output layer sigmoid 0=sigm(0,1),1=sym sigm(-1,1), 3=step, 4=linear"},
    {"init_mag", FLOAT, (int *)&max_magnitude,
       "maximum magnitude of initial weights"},
    {"seed", INT, &seed, "seed for all random number generation"},
};
#define BINTREE2MLP_TABSIZE (sizeof(bintree2mlp_argtab)/sizeof(ARG))

int
main(int argc, char **argv)
{
    int i, nbins, bn, mn;
    int linear, normalization;
    int bin_model, mlp_model;
    int *features, ninputs, nclasses;
    int etta_tau;
    BTNODE **bins, *node;
    FILE *param_file;
    float **weights;

    init_nnsubs(argc, argv, rcsid);
    /********* read command line arguments ************/
    argc = nnargs(argc, argv, bintree2mlp_argtab, BINTREE2MLP_TABSIZE);

    srandom(seed);      /* initialize random numbers */

    /******** check arguments ***********/
    if(nnodes == NULL || nnodes[0] == 0)
      fatal_error("The mlp structure must be specified (-nodes)");

    /* work out the etta_list flags.  by the end, etta_list should be
       filled either with what the user asked for or with etta on each level
       */
    /* I would prefer to do this without a goto,
       but this seemed a little cleaner for now */
    if(etta_list != NULL){
    if(use_etta_list){
        if((int)etta_list[0] >= nnodes[0]-1) goto etta_list_ok;
    }
    /* we need to free the old etta_list so we can build the new one */
    deallocate_vector(etta_list);
    }
    /* if etta list is ok, we should have jumped ahead a little bit */
    etta_list = allocate_vector(nnodes[0]);
    for(i = 0; i < nnodes[0]; i++) etta_list[i] = etta;
  etta_list_ok:

    /********** open and read bintree parameter file ************/
    /********** prune the tree and get the features list ********/
    param_file = nn_fopen(bin_param_fname, "r", "for read of bintree parameters");
    bin_model = norm_read_paramfile(param_file, &normalization);
    bin_model = bintree_read_model(param_file);
    nn_fclose(param_file, bin_param_fname, "for read of bintree parameters");

    printf("\nbintree model read from \"%s\"\n", bin_param_fname);

    nbins = bintree_test_init(bin_model, !prune_tree, max_nodes, 0, NULL);
    printf("\n   %d non-terminal nodes in pruned tree", nbins);
    if(prune_tree)printf(" (%d requested)",max_nodes);
    printf("\n");
    max_nodes = nbins;

    features = *(int **)bintree_param(bin_model, "features");
    ninputs = *(int *)bintree_param(bin_model, "ninputs");
    nclasses = *(int *)bintree_param(bin_model, "noutputs");

    /********* build mlp model *************/
    if(nnodes[1] != ninputs){
    warning("Number of inputs for mlp must match bintree model");
    nnodes[1] = ninputs;
    }
    if(nnodes[0] < 2){
    fatal_error("An input and output layer must be specified in -nodes");
    }
    if(nnodes[nnodes[0]] != nclasses){
    warning("Number of classes for mlp must match bintree model");
    nnodes[nnodes[0]] = nclasses;
    }
    
    mlp_model = mlp_build(nnodes[0], &nnodes[1], step_change);
    mlp_init(mlp_model, max_magnitude, hfunction, ofunction,
         etta_list, features, etta_tau);

    /********* copy bintree parameters into mlp model ********/
    bins = *(BTNODE ***)bintree_param(bin_model, "bins");
    nbins = *(int *)bintree_param(bin_model, "nbins");
    linear = *(int *)bintree_param(bin_model, "linear");
    /* get weights to first hidden layer */
    weights = *(float ***)mlp_param(mlp_model, "weights", 1);
    
    /* check that the number of non-terminal nodes is not greater than
       the number of first layer hidden nodes */
    if(max_nodes > nnodes[2]){
    fatal_error("%s (%d) %s (%d)\n%s",
            "Binary tree has more node tests", max_nodes,
            "than MLP has first layer nodes", nnodes[2],
            "Prune the tree or increase the number of MLP nodes");
    }
    /* do the actual copy */
    for(bn = mn = 0; bn < nbins && mn < nnodes[2]; bn++){
    node = bins[bn];
    if(!TRMNL_NODE(node)){
        if(linear)
          for(i = 0; i < ninputs; i++)
        weights[mn][i] = node->beta[i];
        else{
        for(i = 0; i < ninputs; i++) weights[mn][i] = 0.;
        weights[mn][node->split_input] = 1.;
        }
        /* because the equations are slightly different, there is a sign
           change here */
        /* i = ninputs */       
        weights[mn][i] = -(node->split_constant);
        mn++;
    }
    }
    
    /*************** write mlp parameter file ************/
    param_file = nn_fopen(mlp_param_fname, "w", "for write of model file");
    fprintf(param_file, "Generated from bintree file: %s\n", bin_param_fname);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "mlp model\n");
    mlp_write_model(mlp_model, param_file);
    nn_fclose(param_file, mlp_param_fname, "for write of model file");
    printf("\nFinished -- mlp model saved in \"%s\"\n", mlp_param_fname);

    /************* free space ***************/
    norm_free();
    mlp_free(mlp_model);
    bintree_free(bin_model);
    return(0);
}
