/***-------------------------------------------------------------------------
* NAME:     bintree_feat_sel_main.c
* PURPOSE:  argument parsing and preparation for bintree based
*       feature search driver
*
* USAGE:    bintree_feat_sel -<flags>
* EXAMPLE:  bintree_feat_sel --     * shows all available flags
*
* SEE ALSO: bintree_feat_sel.c bintree_pack.c
*
* GLOBALS: <flags> defined in bintree_feat_sel_args.h
*
* DESCRIPTION:  entry routine for Feature Selection using
*       bintree classifier
*
* RETURNS:
*   creates a parameter file with an ordered list of features
*       and the error rate when that feature was introduced
*   depending on the verbosity, information is printed on stdout:
*   verbosity 0) best feature set found in search and its error rate
*         1) all level 0 printouts plus best feature set found for
*           each number of features and their error rates
*           also, the command line used
*         2) all level 1 printouts plus each feature set tried
*           and their error rates
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree_feat_sel_main.c,v 11.5 2001/08/23 20:11:25 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree_feat_sel_main.c,v $
* CREATED: 21-Mar-94 lck             MIT LINCOLN LABORATORY
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char     rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree_feat_sel_main.c,v 11.5 2001/08/23 20:11:25 craig Exp $";

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <nnlib.h>
#define NO_CV_TABLE
#include <cv_args.h>
#include <gen_class_args.h>
#include "bintree_args.h"
#include <gen_search_args.h>
#include "bintree_feat_sel.h"

int
main(int argc, char **argv)
{
    char *comment;
    int use_splits;
    int train_full_tree, test_full_tree;
    /* arguments for reading defaults file */
    DEFAULTS_FILE_STRUCT *df_args;
    /* argument table with all entries */
    ARG *argtab, *argtab2;
    int tabsize, tabsize2;

    /******************* read argument list **********************/
    init_nnsubs(argc, argv, rcsid);
    nnargs_combine_argtabs(gen_tab, GEN_TABSIZE, bintree_tab, BINTREE_TABSIZE,
               &argtab, &tabsize);
    nnargs_combine_argtabs(argtab, tabsize, search_tab, SEARCH_TABSIZE,
               &argtab2, &tabsize2);
    argc = nnargs(argc, argv, argtab2, tabsize2);

    print_log_header();

    /************* read defaults and set input data parameters ********/
    if(strcmp(desc_fname, "") == 0){
    fatal_error("%s: -fdescribe must be set", argv[0]);
    }
    df_args = (DEFAULTS_FILE_STRUCT *)
      nn_malloc(sizeof(DEFAULTS_FILE_STRUCT), "for the defaults file struct");
    read_defaults_file(path_build(data_path, desc_fname), df_args);
    if(df_args->noutputs == 0)
      fatal_error("noutputs in %s must be greater than 0", desc_fname);
    if(df_args->ninputs == 0)
      fatal_error("ninputs in %s must be greater than 0", desc_fname);

    /* set the number of patterns to no more than the number of lines
       in the data file */
    if(strcmp(data_fname, "") == 0){ /* data_fname is unset */
    fatal_error("%s: -finput must be set", argv[0]);
    }
    check_file_length(path_build(data_path, data_fname), &npatterns);
    if(npatterns == 0)
      fatal_error("%s: Empty data file %s", argv[0],
          path_build(data_path, data_fname));

    if (df_args->db_type != CLASS_DATA)
      fatal_error("class data required, check defaults file");

    /*************** zero out feature selection arguments ***********/
    features[0] = 0;
    ninputs = fix_features(df_args->ninputs, 0, features);

    /******** zero out other output arguments for cv driver ***********/
    param_fname = NULL;
    verror = 0;
    verbosity = -1;

    /************** set up cross validation arguments ***************/
    if(cross_valid < 2)
      fatal_error("%s: -cross_valid must be >= 2", argv[0]);
    cross_valid_from_file = 0; /* no cv_from_file here */
    use_splits =
      fix_cross_valid_args(&cross_valid, &cv_nfolds, &cv_nsplits, cv_splits,
               cv_train_mask, cv_test_mask, cross_valid_from_file,
               path_build(data_path, cross_valid_fname));

    /********* fix up priors adjustment arguments *********************/
    fix_priors(adjust_train, train_uniform_priors, &train_priors_list,
           adjust_test, test_uniform_priors, &test_priors_list,
           train_flag, cross_valid, df_args->noutputs);

    /****************** check search arguments ********************/
    if(search_n > df_args->ninputs){
    warning("%s: %s\nDoing full search", argv[0],
        "Searching for more features than there are.");
    stop_early = 0;
    }
    /* if search_n is 0, feat_sel does the whole search */
    if(!stop_early) search_n = 0;

    /********** check bintree args *****************/
    train_full_tree = !stop_train_early;
    test_full_tree = !prune_tree;

    /************** make and print header comment *******************/
    comment = make_comment_string(nnargs_get_flags(argtab2, tabsize2));
    if (search_verbosity > FLAGS_VERBOSITY) {
    printf("%s", comment);
    fflush(stdout);
    }

    srandom(seed);      /* initialize random numbers */

    /* call the feature selection driver */
    bintree_feat_sel(search_direction, search_n, search_verbosity,
             search_param_fname,
             roc_cost, roc_target, roc_lower, roc_upper,
             df_args->categories, 
             /* bintree_cross_valid_arguments */
             df_args->ninputs, ninputs, features, df_args->noutputs, linear,
             train_full_tree, test_full_tree,
             leaf_min_npatterns, max_nodes, normalize,
             npatterns, debug, verbosity, verror,
             random_flag, cv_nfolds, cv_nsplits,
             use_splits, cv_splits, &cv_train_mask[1],
             &cv_test_mask[1], data_path, data_fname,
             norm_fname, exp_path, error_fname,
             param_fname, df_args->output_labels, comment, adjust_train,
             adjust_test, priors_npatterns,
             &train_priors_list[1], &test_priors_list[1],
             cv_random, log_outputs);

    print_log_footer();
    nnargs_free_argtab(argtab, tabsize);
    nnargs_free_argtab(argtab2, tabsize2);

    /* free the vectors allocated when the defaults file was read */
    free_defaults_file_struct(df_args);

    exit_nnsubs();
    exit(0);
}
