/***-------------------------------------------------------------------------
* NAME: bintree_pack.c
* PURPOSE:     binary tree classification and prediction routines
* USAGE:
*
* SEE ALSO:     
* DESCRIPTION:  binary tree classification and prediction routines.
*       An example of their use is in bintree.c. These routines were
*       modified by deniz to use nnlib libraries to make the source 
*       compatible with other LNKnet codes.
*
* GLOBAL FUNCTIONS:
*   bintree_build()     constructs a bintree model
*       bintree_init_tree()     initializes root node of the tree
*   bintree_free()      deallocates model storage
*   bintree_write_model()   writes a model to parameter file
*   bintree_read_model()    reads a model from parameter file
*   bintree_param()      returns value of parameter
*   bintree_store_pattern()    loads single patterns into tree
*       bintree_grow()
*       bintree_prune()
*       bintree_eval()      evaluates one input, fills ouputs vector
*   bintree_classify()  evaluates one input, returns class
* 
* LOCAL FUNCTIONS:
*     bintree_write_tree()       writes the node parameters to model file
*     bintree_calculate_nleaves_and_strength()  Sets the strength of all nodes
*     bintree_allocate_node_vectors()allocates vectors in a node
*     bintree_allocate_tree_patterns() allocates patterns in tree
*     bintree_set_node_class()   determines the class of a node
*     bintree_make_child()   creates a new node and assigns it to mother
*     bintree_calculate_nterminals() calculates # of terminal nodes
*     bintree_initialize_node()  sets all elements of a node to zero
*     bintree_need_to_split()    returns 1 if node needs to be split else 0
*     bintree_grow_node()        grows a subtree under a given node
*     bintree_find_best_delta_and_gamma() used for linear combination
*     bintree_set_univariate_split() sets univariate split variables
*     bintree_set_linear_combination_values() calculates linear combinations
*     bintree_set_linear_split_values() used for linear combination params
*     bintree_find_best_split()  finds the best split point and max "Phi"
*     bintree_split_criterion()  returns a split criterion "Phi" for a node
*     bintree_add_new_node()     adds a new node to the tree nodes array
*     bintree_split_node()       called from bintree_grow_node()
*     bintree_divide_patterns()  divides patterns between yes- and nochildren
*     bintree_display_node()    displays node properties going down tree
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree_pack.c,v 11.4 2001/08/23 20:11:26 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/bintree/bintree_pack.c,v $
* CREATED: 21-JUL-93 deniz                             MIT LINCOLN LABORATORY
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory 

   Permission is hereby granted, without written agreement or royalty fee, 
   to use, copy, modify, and distribute this software and its 
   documentation for any purpose, provided that the above copyright notice 
   and the following three paragraphs appear in all copies of this 
   software. 
     
     

     IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR 
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF 
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN 
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  M.I.T. LINCOLN 
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT. 

        THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN 
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, 
ENHANCEMENTS, OR MODIFICATIONS. 
*/

/* Include necessary header files */
/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>

/* to include internal declerations and definitions from `bintree_pack.h' */
#define   bintree_pack_h_internal_types
#define   bintree_pack_h_internal_macros
#define   bintree_pack_h_internal_functions
#include "bintree_pack.h"

#define PEQ(y) (0 == strcmp(param_name,y))
#define ABS(x) (((x) < 0)? -(x) : (x))
#define TRUE 1
#define FALSE 0

/* Static Variables (for this file use only) */
static BTREEP *bintpp;  /* bintree_model set */

/**********************************************************************
 **********************************************************************
 *****           DEFINITIONS OF EXTERNAL FUNCTIONS                *****
 **********************************************************************
 **********************************************************************/

/************************************************************************
* NAME:bintree_build(ninputs, nclasses, verbosity)
* DESCRIPTION:
*   Allocates memory for the first node (root) of the tree.
*       Initializes the elements and reads in the data file.
*************************************************************************/
int
bintree_build(int ninputs, int nclasses, int verbosity)
{
    int new_model;
    static int max_models=0;
    BTREEP tree; /* temporary pointer to bintpp[] */

    /* find entry in model array for new model */
    new_model = model_get_number((void ***)&bintpp,
                 &max_models, sizeof(BTREE));
    tree = bintpp[new_model];

    /* set vector sizes */
    tree->params.ninputs = ninputs;
    tree->params.nclasses = nclasses;

    tree->params.verbosity = verbosity;

    /* initialize pattern parameters */
    tree->params.pattern_size = 0;
    tree->patterns = NULL; /* allocated in bintree_allocate_tree_patterns */

    /* initialize unallocated array pointers */
    tree->body.bins = NULL; /* allocated in bintree_add_new_node */
    tree->body.bins_size = 0;
    tree->body.bin_index = NULL; /* allocated in bintree_prune */

    /* initialize working array pointers for growing */
    /* allocated and freed in grow_node */
    tree->body.classes = NULL;
    tree->body.search_array = NULL;
    tree->body.yes_class_count = NULL;
    tree->body.no_class_count = NULL;
    tree->body.linear_classes = NULL;
    tree->body.linear_search_array = NULL;
    tree->body.linear_node_data = NULL;
    tree->body.median = NULL;
    tree->body.range = NULL;
    tree->body.linear_beta = NULL;

    /* allocate model vectors */
    /*  memory for class count */
    if(nclasses > 0) tree->params.class_count = allocate_vector(nclasses);

    /* allocating memory for features list */
    if(ninputs >= 0) tree->params.features = allocate_ivector(ninputs+1);

    /* allocating memory for priors list for evaluating used in classify */
    if(nclasses > 0) tree->params.priors_list = allocate_vector(nclasses);

    /*allocating memory for outputs vector for evaluating used in classify */
    if(nclasses > 0) tree->params.outputs = allocate_vector(nclasses);

    /* allocate space for recording leaves */
    tree->body.leaves_index = allocate_ivector(MEMORY_BLOCK);
    tree->body.leaves_size = MEMORY_BLOCK;

    return(new_model);
}

/************************************************************************
* NAME: bintree_init(model, linear, full_tree, leaf_min_npatterns, features)
* DESCRIPTION:
*   Initializes tree, for starting training
************************************************************************/
void
bintree_init(int model, int linear, int full_tree, int leaf_min_npatterns,
         int *features)
{
    int i;
    BTREEP tree; /* temporary pointer for model */

    tree = bintpp[model];

    /* initialize tree splitting parameters */
    if(leaf_min_npatterns < 1) leaf_min_npatterns = 1;
    if(full_tree) tree->params.leaf_min_npatterns = 1;
    else tree->params.leaf_min_npatterns = leaf_min_npatterns;

    tree->params.linear = linear;
    tree->params.adjust_priors = 0;
    tree->body.tree_error = (float) 0.0;

    /* zero out data and node array lengths, so we can grow the
       next tree in the currently allocated space */
    tree->params.npatterns = 0;
    tree->body.nbins = 0;
    tree->body.nnodes = 0;
    tree->body.nlevels = 0;
    tree->body.nleaves = 0;

    /* copy feature list */
    for(i = 0; i <= features[0]; i++){
    tree->params.features[i] = features[i];
    }
}
/************************************************************************
* NAME: bintree_free(model)
* DESCRIPTION:
*   Deallocates all memory used for a tree
************************************************************************/
void
bintree_free(int model)
{
    BTREEP tree; /* temporary pointer for model */
    BTLINK node;
    int binnum;

    tree = bintpp[model];

    node = &tree->root;
    
    /* don't walk the tree to do this. It might have been pruned */
    for (binnum = 0; binnum < tree->body.bins_size; binnum++){
      node = tree->body.bins[binnum];
      bintree_free_node(node);
    }

#ifdef notdef
    /* travel to leaves to free nodes bottom up */
    while(!TRMNL_NODE(node) || !ROOT_NODE(node)){
    while(!TRMNL_NODE(node)){    
        if(node->yeschild != NULL)
          node = node->yeschild;
        else if(node->nochild != NULL)
          node = node->nochild;
    }
    buffer_node = node;
    node = node->mother;

    if(node->yeschild != NULL)
      node->yeschild = NULL; /* buffer node is yeschild and a leaf */
    else if(node->nochild != NULL)
      node->nochild = NULL; /* buffer node is nochild and a leaf */
    bintree_free_node(buffer_node);
    }

    /* free the root node (which has become terminal and root node) */
    bintree_free_node(node);
#endif

    /* deallocate tree vectors, etc */
    bintree_free_tree(tree);

    /* free main model pointer */
    nn_free((char*) bintpp[model], "CART tree model");
    bintpp[model] = NULL;
}

/****************************************************************************
* NAME: bintree_store_pattern(model, ntrial, desired_class, inputs)
* DESCRIPTION:
*       incremental training for a single input vector
* EXAMPLE:
*
****************************************************************************/
void
bintree_store_pattern(int model, int desired_class, float *inputs)
{
    int j;
    BTREEP tree;    /* temporary pointer to bintpp[] */

    tree = bintpp[model];

    /* check if desired class is in correct range */
    if(desired_class < 0 || desired_class >= tree->params.nclasses)
      fatal_error("\n ERROR in bintree_store_pattern: desired_class = %d",
          desired_class);

    /* update the class distributions */
    tree->params.class_count[desired_class] += (float) 1.0;

    /* make sure there is space for the pattern */
    bintree_allocate_tree_patterns(tree);

    /* load pattern data into the tree structure */
    tree->patterns[tree->params.npatterns].class = desired_class;
    for(j = 0; j < tree->params.ninputs; j++)
      tree->patterns[tree->params.npatterns].inputs[j] = inputs[j];

    tree->params.npatterns++;
}

/************************************************************************
* NAME: bintree_init_tree(model)
* DESCRIPTION:
*   Allocates memory for the first node (root) of the tree.
*       Initializes the elements and reads in the data file.
*************************************************************************/
void
bintree_init_tree(int model)
{
    int i;
    BTREEP tree;

    tree = bintpp[model];   /* temporary pointer to the model */

    /* initialize root and allocate memory for vectors in it */
    bintree_initialize_node(tree, &tree->root);

    /* record root address */
    bintree_add_new_node(tree, &tree->root, "root");

    /* copy addresses for all patterns into root's pattern list */
    /* do we need a node's pattern list after splitting? */
    tree->root.npatterns = tree->params.npatterns;
    for(i = 0; i < tree->params.npatterns; i++)
      tree->root.patterns[i] = &tree->patterns[i];

    /* calculate class_distribs and set class of root node */
    bintree_set_node_class(tree, &tree->root);
}

/************************************************************************
* NAME: bintree_grow_tree(model)
* DESCRIPTION:
*   Calls the bintree_grow_node(tree, node) function which
*   splits the node as goes down the tree creating new nodes
*       while the splitting is necessary.
*************************************************************************/
int
bintree_grow_tree(int model)
{
    int max_level=0;
    BTREEP tree;

    tree = bintpp[model];          /* temporary pointer to bintpp[] */

    bintree_display_node_head(tree, TRUE);

    /* initialize tree_error */
    tree->body.tree_error = tree->root.node_misclass*tree->root.probability;
    bintree_display_node(tree, &tree->root, TRUE);

    max_level = bintree_grow_node(tree, &tree->root);

    /* number of levels = maximum level number + 1 : first level = 0 */
    tree->body.nlevels = max_level + 1;

    return(max_level);
}

/************************************************************************
* NAME: bintree_prune(model)
* DESCRIPTION:
*   Calculates the sub misclassifications and then the strength of
*   non-terminal nodes. Then the adresses of the nodes of the tree
*   are stored in an array in strength order.
*************************************************************************/
void
bintree_prune(int model)
{
    BTREEP tree;
    BTLINK wnode;
    int prune_num, weakesti;

    tree = bintpp[model]; /* temporary pointer to bintpp[] */

    bintree_display_node_head(tree, FALSE);

    /* initialize nleaves and strength for all nodes, calculate the
       tree error */
    bintree_calculate_nleaves_and_strength(tree);
    bintree_calculate_tree_error(tree);

    bintree_display_node(tree, &tree->root, FALSE);

    /* allocate space for pruning index */
    if(tree->body.bin_index != NULL) /* free old arrays first */
      deallocate_ivector(tree->body.bin_index);
    tree->body.bin_index = allocate_ivector(tree->body.nnodes);

    prune_num = tree->body.nnodes-1;
    while(prune_num >= 0){
    /* find the weakest node */
    weakesti = bintree_find_weakest_node(tree);
    wnode = tree->body.bins[weakesti];

    /* put the weakest node in the pruning array */
    tree->body.bin_index[prune_num] = weakesti;
    prune_num--;

    /* mark children as pruned, add any that are non-terminal nodes
       to the prune array */
    prune_num = bintree_prune_children(tree, wnode, prune_num);

    /* decrement the number of leaves and non-terminal nodes */
    /* an extra leaf is added for wnode, which has become a leaf */
    tree->body.nleaves -= (wnode->nleaves-1);
    tree->body.nnodes -= (wnode->nleaves-1);

    /* make wnode a leaf and add its error to the tree error */
    /* we are preserving the yesno and nono fields, so the
       whole tree is recoverable */
    wnode->yeschild = wnode->nochild = NULL;
    wnode->nleaves = 1;
    tree->body.tree_error += wnode->node_misclass*wnode->probability;
    wnode->sub_misclass = wnode->node_misclass;
    wnode->strength = -MINFLOAT;

    /* print out "epoch" information */
    bintree_display_node(tree, wnode, FALSE);

    /* adjust the strength of the ancestor nodes of wnode */
    if(!ROOT_NODE(wnode)) do{
        /* go up to the mother node */
        wnode = wnode->mother;
          
        /* adjust nleaves, sub_misclass, and strength */
        wnode->nleaves =
          wnode->yeschild->nleaves + wnode->nochild->nleaves;
        wnode->sub_misclass =
          wnode->yeschild->sub_misclass + wnode->nochild->sub_misclass;
        wnode->strength =
          ((wnode->node_misclass - wnode->sub_misclass))/
        ((float) wnode->nleaves - 1.);
    } while(!ROOT_NODE(wnode));
    }
    /* reconnect the children to their pruned ancestors */
    bintree_link_tree(tree);
}

/************************************************************************
* NAME: bintree_test_init(model, full_tree, max_nodes,
*                         adjust_priors, priors_list)
* DESCRIPTION:
*   sets initializations for testing:
*   loads priors list if adjust_priors
*   sets the node adresses of mother and children
*   sets the pruning (changes pruning flags appropriately)
*************************************************************************/
int
bintree_test_init(int model, int full_tree, int max_nodes, int adjust_priors,
          float *priors_list)
{
    BTREEP tree;

    tree = bintpp[model];   /* temporary pointer to the model */

    /* load priors list */
    tree->params.adjust_priors = adjust_priors;
    if(adjust_priors) {
    int i;
    tree->params.adjust_priors = 1;
    for(i = 0; i < tree->params.nclasses; i++){ 
        tree->params.priors_list[i] = priors_list[i];
    }
    }
    
    /* sets tree pruning */
    if(!full_tree) max_nodes = bintree_set_pruning(tree, max_nodes);
    else max_nodes = tree->body.nnodes;
    return max_nodes;
}

/************************************************************************
* NAME: bintree_classify(model, inputs)
* DESCRIPTION:
*   classifies a pattern returns class fills in the output probabilities
*************************************************************************/
int
bintree_classify(int model, float *inputs)
{
    int i, noutputs, output_class;
    float  *outputs, total;
    BTREEP tree;    /* temporary pointer to bintpp[] */

    tree = bintpp[model];

    outputs = tree->params.outputs;
    noutputs = tree->params.nclasses;

    bintree_find_pattern_node(tree, inputs, outputs, &output_class);

    /* if adjust priors; adjust output vector */
    if(tree->params.adjust_priors){
    /* adjust priors */
    for(i = 0; i < noutputs; i++){
        outputs[i] *= tree->params.priors_list[i];
        outputs[i] /= (float)tree->params.class_count[i]/
          (float)tree->params.npatterns;
    }
    /* scale adjusted outputs to sum to 1 */
    total = 0.;
    for(i = 0; i < noutputs; i++) total += outputs[i];
    for(i = 0; i < noutputs; i++) outputs[i] /= total;
    output_class = max_pick(outputs, noutputs);
    }

    return(output_class);
}

/************************************************************************
* NAME: bintree_eval(model, inputs, outputs)
* DESCRIPTION:
*   classifies a pattern, fills in the output probabilities
*************************************************************************/
void
bintree_eval(int model, float *inputs, float *outputs)
{
    int output_class;
    BTREEP tree;    /* temporary pointer to bintpp[] */

    tree = bintpp[model];

    bintree_find_pattern_node(tree, inputs, outputs, &output_class);
}

/****************************************************************************
* NAME: bintree_write_model(model, fp)
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
*
* EXAMPLE:
*
****************************************************************************/
int
bintree_write_model(int model, FILE *fp)
{
    int             rv;
    BTREEP tree;    /* temporary pointer to bintpp[] */

    tree = bintpp[model];

    fprintf(fp, "bintree model\n");
    rv = bintree_write_model_params(tree, fp);
    if (rv < 0) return(rv);

    rv = bintree_write_tree_model(tree, fp);
    if(rv < 0) return(rv);

    return (0);
}

/***************************************************************************
* NAME: bintree_read_model(fp)
* DESCRIPTION:
*       reads parameters and exemplars for a model to a file
*       in ASCII format
*
* EXAMPLE:
*
***************************************************************************/
int
bintree_read_model(FILE *fp)
{
    int    rv, model;
    BTREEP tree;    /* temporary pointer to bintpp[] */


    model = bintree_read_model_params(fp);
    if (model < 0)  return(model);
    tree = bintpp[model];

    rv = bintree_read_tree_model(tree, fp);
    if(rv < 0) return(rv);

    return(model);
}

/************************************************************************
* NAME: bintree_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/
void *
bintree_param(int model, char *param_name)
{
    void   *param_address;
    BTREEP  tree;      /* temporary pointer to bintpp[] */

    tree = bintpp[model];
    if (PEQ("features"))
      param_address = (void *)&tree->params.features;
    else if (PEQ("ninputs"))
      param_address = (void *)&tree->params.ninputs;
    else if (PEQ("noutputs"))
      param_address = (void *)&tree->params.nclasses;
    else if (PEQ("nclasses"))
      param_address = (void *)&tree->params.nclasses;
    else if (PEQ("outputs"))
      param_address = (void *)&tree->params.outputs;
    else if (PEQ("class_cnt"))
      param_address = (void *)&tree->params.class_count;
    else if (PEQ("class_count"))
      param_address = (void *)&tree->params.class_count;
    else if (PEQ("npatterns"))
      param_address = (void *)&tree->params.npatterns;
    else if (PEQ("total_npatterns"))
      param_address = (void *)&tree->params.npatterns;
    else if (PEQ("linear"))
      param_address = (void *)&tree->params.linear;
    else if (PEQ("max_nodes"))
      param_address = (void *)&tree->body.nnodes;
    else if (PEQ("nbins"))
      param_address = (void *)&tree->body.nbins;
    else if (PEQ("bins"))
      param_address = (void *)&tree->body.bins;
    else if (PEQ("nlevels"))
      param_address = (void *)&tree->body.nlevels;
    else {
    warning("bintree_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return(param_address);
}

/**********************************************************************
 **********************************************************************
 *****      DEFINITIONS OF LOCAL (static declared) FUNCTIONS      *****
 **********************************************************************
 **********************************************************************/

/*****************************
 ROUTINES USED BY FREE
*******************************/
/************************************************************************
* NAME: bintree_free_node(tree, node)
* DESCRIPTION:
*   Clears memory of one node
*************************************************************************/
/* called by bintree_free */
static void
bintree_free_node(BTLINK node)
{
    if(node != NULL){
    /* free node vectors allocated in bintree_initialize_node() */
    if(node->class_distribs != NULL){
        deallocate_vector(node->class_distribs);
        node->class_distribs = NULL;
    }
    if(node->beta != NULL){
        deallocate_vector(node->beta);
        node->beta = NULL;
    }
    if(node->patterns != NULL){
        deallocate_pvector((void **)node->patterns);
        node->patterns = NULL;
    }
    
    /* free node allocated in bintree_make_child()*/
    if(!ROOT_NODE(node)){
        deallocate_structvector(node);
        node = NULL;
    }
    }
}
/************************************************************************
* NAME: bintree_free_tree(tree)
* DESCRIPTION:
*   Clears memory of the tree  model
*************************************************************************/
/* called by bintree_free */
static void
bintree_free_tree(BTREEP tree)
{
    int i;

    /* free list of node pointers */
    /* allocated in bintree_add_new_node and bintree_make_bins */
    if(tree->body.bins != NULL){
    deallocate_pvector((void **)tree->body.bins);
    tree->body.bins = NULL;
    }
    /* free index of node pointers */
    /* allocted in bintree_prune and bintree_make_bins */
    if(tree->body.bin_index != NULL){
    deallocate_ivector(tree->body.bin_index);
    tree->body.bin_index = NULL;
    }
    /* free vectors allocated in bintree_build */
    if(tree->body.leaves_index != NULL){
    deallocate_ivector(tree->body.leaves_index);
    tree->body.leaves_index = NULL;
    }
    if(tree->params.features != NULL){
    deallocate_ivector(tree->params.features);
    tree->params.features = NULL;
    }
    if(tree->params.class_count != NULL){
    deallocate_vector(tree->params.class_count);
    tree->params.class_count = NULL;
    }
    if(tree->params.priors_list != NULL){
    deallocate_vector(tree->params.priors_list);
    tree->params.priors_list = NULL;
    }
    if(tree->params.outputs != NULL){
    deallocate_vector(tree->params.outputs);
    tree->params.outputs = NULL;
    }
    /* free patterns allocated in bintree_allocate_tree_patterns */
    if(tree->patterns != NULL){
    for(i = 0; i < tree->params.pattern_size; i++){
        if(tree->patterns[i].inputs != NULL){
        deallocate_vector(tree->patterns[i].inputs);
        tree->patterns[i].inputs = NULL;
        }
    }
    deallocate_structvector(tree->patterns);
    }
    /* the actual tree is freed in bintree_free */
}

/***********************************
 ROUTINES FOR INITIAL STORAGE OF TREE DATA
 *************************************/
/************************************************************************
* NAME: bintree_allocate_tree_patterns(tree)
* DESCRIPTION:
*   Allocates memory for to store pointers to the patterns of
*       the node.
*************************************************************************/
/* called by bintree_store_pattern */
static void
bintree_allocate_tree_patterns(BTREEP tree)
{
    int i;
    int ninputs = tree->params.ninputs;
    int newsize = tree->params.pattern_size;
    int oldsize = tree->params.pattern_size;
    int size = tree->params.npatterns; 

    /* make sure the array sizes are all positive */
    if(size < 0) size = 0;
    if(oldsize < 0) oldsize = 0;
    if(newsize < 0) newsize = 0;

    if(size >= oldsize){
    newsize += MEMORY_BLOCK;

    /* allocate more space for patterns array */
    if(oldsize <= 0)
      tree->patterns = (PATTERN*)
          allocate_structvector(newsize, sizeof(PATTERN));
    else
      tree->patterns = (PATTERN*)
        reallocate_structvector(tree->patterns, oldsize, newsize,
                    sizeof(PATTERN));

    /* allocate space for inputs */
    for(i = oldsize; i < newsize; i++){
        if(ninputs > 0)
          tree->patterns[i].inputs = allocate_vector(ninputs);
        else
          tree->patterns[i].inputs = NULL;
    }
    }
    tree->params.pattern_size = newsize;
}

/************************************
  ROUTINES USED TO START TREE GROWTH
*************************************/
/************************************************************************
* NAME: bintree_initialize_node(tree, node)
* DESCRIPTION:
*   Resets all elements of the node.
*************************************************************************/
/* called by bintree_read_node_params, bintree_make_child, bintree_init_tree */
static void
bintree_initialize_node(BTREEP tree, BTLINK node)
{
    /* invalidate the node numbers */
    node->motherno = EOTREE;
    node->selfno = EOTREE;
    node->yesno = EOTREE;
    node->nono = EOTREE;

    /* set everything equal to zero */
    node->npatterns = 0;
    node->patterns = NULL;

    node->class = UNKNOWN_CLASS;
    node->level = 0;
    node->nnodes = 0;
    node->nleaves = 0;

    node->max_distrib = (float) 0.0;
    node->class_distribs = NULL;
    node->split_input = 0;
    node->beta = NULL;

    node->prune = 0;
    node->topflag = 0;
    node->visit = 0;
    node->yesvisit = 0;
    node->novisit = 0;

    node->node_misclass = (float) 0.0;
    node->sub_misclass = (float) 0.0;
    node->split_constant = (float) 0.0;
    node->probability = (float) 0.0;
    node->strength = (float) 0.0;

    node->x = 0; /* this is safe enough, the first node in a tree
            is assigned here */

    node->yeschild = NULL;
    node->nochild = NULL;
    node->mother = NULL;



    /* allocate memory for the float and int arrays of the node */
    {
    int nclasses = tree->params.nclasses;
    int ninputs = tree->params.ninputs;
    int npatterns;
    
    /* allocating memory to store class distribs in node */
    if(nclasses > 0) node->class_distribs = allocate_vector(nclasses);
    
    /* allocating memory for beta: linear combination parameters */
    if(ninputs > 0) node->beta = allocate_vector(ninputs);

    /* allocate memory to store pointers to patterns in the node */
    if(ROOT_NODE(node))
      npatterns = tree->params.npatterns;
    else
      npatterns = node->mother->npatterns;
    
    if(npatterns > 0)
      node->patterns = (PATTERN**) allocate_pvector(npatterns);
    }
}

/************************************************************************
* NAME: bintree_add_new_node(tree, node, child_kind)
* DESCRIPTION:
*   adds the address of the new node to the nodes array of the tree
*   and sets the new appropriate indexes for that array, updates
*   the total nodes count: `tree->body.nbins++'
*   
*************************************************************************/
/* called by bintree_make_child and bintree_init_tree */
static void
bintree_add_new_node(BTREEP tree, BTLINK node, char *child_kind)
{
    int old_size, new_size;

    /* set the node number and increment the total number of nodes */

    switch(WHICH_CHILD(child_kind))  /* WHICH_CHILD defined as macro */
      {
       case YES:
      node->motherno = node->mother->selfno;
      node->mother->yesno = node->selfno = tree->body.nbins;
      tree->body.nbins++;
      break;

    case NO:
      node->motherno = node->mother->selfno;
      node->mother->nono = node->selfno = tree->body.nbins;
      tree->body.nbins++;
      break;

    default:
      if(ROOT_NODE(node)){
          node->motherno = EOTREE;
          node->selfno = tree->body.nbins;
          tree->body.nbins++;
      }
      else
        fatal_error("\n  ERROR: child kind %s is unknown. \n", child_kind);
      }

    /* store address of new node */
    if(tree->body.bins_size <= tree->body.nbins){
    old_size = tree->body.bins_size;
    new_size = old_size + MEMORY_BLOCK;

    if(old_size == 0)
      tree->body.bins = (BTLINK*) allocate_pvector(new_size);
    else
      tree->body.bins = (BTLINK*)
        reallocate_pvector((void **)tree->body.bins, old_size, new_size);

    tree->body.bins_size = new_size;
    }

    tree->body.bins[node->selfno] = node;
}

/************************************************************************
* NAME: bintree_set_node_class(tree, node)
* DESCRIPTION:
*   Counts the occurances of different classes in the node.
*       Counts the number of classes present in the node.
*       Finds the probability distributions of each class in the node.
*       Finds the class with maximum distribution in the node.
*       Calculates node misclassification R(t) from  CART chapt 2 p.35:
*       R = (1 - max class distrib) * npatterns in node  / npatterns in model 
*       Finds and returns the number of classes with maximum distribution.
*************************************************************************/
/* called by bintree_split_node and bintree_init_tree */
static void
bintree_set_node_class(BTREEP tree, BTLINK node)
{
    int i;

    /* the node counting elements must be initialized to zero beforehand */   
    for(i = 0; i < tree->params.nclasses; i++) node->class_distribs[i] = 0.;
    node->max_distrib = MINFLOAT;
    node->class = EOTREE;

    /* counting the number of elements in node from each class */
    for(i = 0; i < node->npatterns; i++)
      node->class_distribs[node->patterns[i]->class] += 1.;

    /* calculate the class distributions and determine the class of node */
    if(node->npatterns > 0){
    for(i = 0; i < tree->params.nclasses; i++){
        node->class_distribs[i] /= (float) node->npatterns;
        if(node->max_distrib < node->class_distribs[i]){
        node->max_distrib = node->class_distribs[i];
        node->class = i;
        }
    }
    } else
      warning("There are %d patterns at node #(%d); class = %d",
          node->npatterns, node->selfno, node->class);

    /* calculate probability of any given pattern to be in this node */
    if(tree->params.npatterns > 0 && node->npatterns > 0){
    node->probability =
      ((float) node->npatterns / (float) tree->params.npatterns);
    } else{
    node->probability = -1;
    warning("There are %d patterns in tree; %d patterns in bin #(%d)\n",
        tree->params.npatterns, node->npatterns, node->selfno);
    }

    /* calculate node misclassification of node */
    node->node_misclass = (1.0 - node->max_distrib) *node->probability;

    return;
}

/****************************
 ROUTINES USED BY GROW
 ****************************/
/************************************************************************
* NAME: bintree_grow_node(tree, node)
* DESCRIPTION:
*   Splits the node as goes down the tree creating new nodes
*       while the splitting is necessary.
*************************************************************************/
/* called by bintree_grow_tree */
static int
bintree_grow_node(BTREEP tree, BTLINK node)
{
    int max_level=0, terminal = 0;
    BTLINK split_node;

    /* initialize queue of leaves in need of splitting */
    tree->body.leaves_index[0] = node->selfno;
    tree->body.nleaves = 1;

    /* allocate space for working arrays */
    tree->body.classes = allocate_ivector(tree->params.npatterns);
    tree->body.search_array = allocate_vector(tree->params.npatterns);
    tree->body.yes_class_count = allocate_ivector(tree->params.nclasses);
    tree->body.no_class_count = allocate_ivector(tree->params.nclasses);

    if(tree->params.linear){
    /* allocate space for search for best delta and gamma */
    tree->body.linear_classes = allocate_ivector(tree->params.npatterns);
    tree->body.linear_search_array =
      allocate_vector(tree->params.npatterns);
    tree->body.linear_beta = allocate_vector(tree->params.ninputs);

    /* allocate space for normalizing data */
    tree->body.linear_node_data = 
      allocate_matrix(tree->params.ninputs, tree->params.npatterns);
    tree->body.median = allocate_vector(tree->params.ninputs);
    tree->body.range = allocate_vector(tree->params.ninputs);
    }

    /* go through the queue of leaves to split */
    while((split_node = bintree_get_split_node(tree)) != NULL){
    terminal = 0;
    if(bintree_need_to_split(tree, split_node)){
        if(bintree_split_node(tree, split_node) == -1)
          /* splitting of patterns is not possible */
          terminal = 1;
        else
          /* increment the number of splitting nodes */
          tree->body.nnodes++;
    } else
      /* splitting of patterns is not required */
      terminal = 1;

    if(terminal){
        /* node is a terminal node: see if the level is maximum */
        /* MAX is defined as macro */
        max_level = MAX(max_level, split_node->level);   
    } else {
        /* display grown node for monitoring */
        bintree_display_node(tree, split_node, TRUE);
    }
    }
    
    /* free working arrays */
    deallocate_ivector(tree->body.classes);
    deallocate_vector(tree->body.search_array);
    deallocate_ivector(tree->body.yes_class_count);
    deallocate_ivector(tree->body.no_class_count);
    if(tree->params.linear){
    deallocate_ivector(tree->body.linear_classes);
    deallocate_vector(tree->body.linear_search_array);
    deallocate_vector(tree->body.linear_beta);
    deallocate_matrix(tree->body.linear_node_data);
    deallocate_vector(tree->body.median);
    deallocate_vector(tree->body.range);
    }
    /* reinitialize array pointers */
    tree->body.classes = NULL;
    tree->body.search_array = NULL;
    tree->body.yes_class_count = NULL;
    tree->body.no_class_count = NULL;
    tree->body.linear_classes = NULL;
    tree->body.linear_search_array = NULL;
    tree->body.linear_node_data = NULL;
    tree->body.median = NULL;
    tree->body.range = NULL;

    return(max_level);
}

/************************************************************************
* NAME: bintree_get_split_node(tree, node_p)
* DESCRIPTION:
*   Finds the terminal node with highest error rate.
*************************************************************************/
/* called by bintree_grow_node */
static BTLINK
bintree_get_split_node(BTREEP tree)
{
    int i;
    BTLINK split_node = NULL;

    if(tree->body.nleaves > 0){
    /* return the first entry in the leaves index queue */
    split_node = tree->body.bins[tree->body.leaves_index[0]];

    /* shift the leaves index queue and decrement its length */
    for(i = 1; i < tree->body.nleaves; i++){
        tree->body.leaves_index[i-1] = tree->body.leaves_index[i];
    }
    tree->body.nleaves -= 1;
    }
    return(split_node);
}

/************************************************************************
* NAME: bintree_need_to_split(tree, node)
* DESCRIPTION:
*   Returns 1 if the node needs to be split else 0.
*************************************************************************/
/* called by bintree_do_splitting and bintree_get_split_node */
static int
bintree_need_to_split(BTREEP tree, BTLINK node)
{
    int node_ok;

    /* a node is ok if it is pure OR
       if leaf_min_npatterns has been set AND
       the node has the minimum patterns or fewer */
    node_ok = (node->max_distrib >= 1. ||
           (tree->params.leaf_min_npatterns > 1 &&
        node->npatterns <= tree->params.leaf_min_npatterns));

    return(!node_ok);
}

/************************************************************************
* NAME: bintree_split_node(tree, node)
* DESCRIPTION:
*   splits a node: finds the splitting criteria, creates two subnodes
*   assigns patterns to the subnodes sets the class and misclass
*   rates of the subnodes, calculates the strength of splitting
*   .
*************************************************************************/
/* called by bintree_do_splitting */
static int
bintree_split_node(BTREEP tree, BTLINK node)
{
    if(bintree_set_univariate_split(tree, node) == -1) return(-1);

    if(tree->params.linear)
      /* if there is one binary (univariate) split, then there is
       * at least one linear split - the binary split itself */
      bintree_set_linear_combination_parameters(tree, node);

    /* check that this will really split the data */
    if(bintree_bad_division(tree, node)) return -1;

    bintree_make_child(tree, node, "yes");
    bintree_make_child(tree, node, "no");
    bintree_divide_patterns(tree, node);
    bintree_set_node_class(tree, node->yeschild);
    bintree_set_node_class(tree, node->nochild);
    bintree_update_tree_error(tree, node);
    /* add new nodes to queue of nodes that need splitting */
    /* if doing the sorting by misclass, new leaves must come after
       set_node_class */
    bintree_add_new_leaf(tree, node->yeschild);
    bintree_add_new_leaf(tree, node->nochild);

    return(1);
}

/************************************************************************
* NAME: bintree_set_univariate_split(tree, node)
* DESCRIPTION:
*   Calculates the best univariate split for a node
*************************************************************************/
/* called by bintree_split_node */
static int
bintree_set_univariate_split(BTREEP tree, BTLINK node)
{
    int    i, split_point, coordinate, split_coordinate=0;
    float  split_constant=0.0, max_phi = -MINFLOAT;
 
    if(tree->params.linear){
    bintree_copy_and_normalize_data(tree, node);
    }
    for(coordinate = 0; coordinate < tree->params.ninputs; coordinate++){
    /* initialize linear combination parameters (beta[coordinate])*/
    node->beta[coordinate] = 0;

    /* load search array and sort it */
    for(i = 0; i < node->npatterns; i++){
        if(tree->params.linear)
          tree->body.search_array[i] =
        tree->body.linear_node_data[coordinate][i];
        else
          tree->body.search_array[i] =
        node->patterns[i]->inputs[coordinate];
        tree->body.classes[i] = node->patterns[i]->class;
    }
    heap_sort2(node->npatterns, tree->body.search_array,
               tree->body.classes);

    /* finding the split constant with the best split performance */
    split_point =
      bintree_find_best_split(node->npatterns, tree->params.nclasses,
                  tree->body.classes, &max_phi,
                  tree->body.search_array, tree);
    if(split_point > -1){
        split_constant =
          bintree_split_value(tree->body.search_array, node->npatterns, 
                  split_point);
        split_coordinate = coordinate;
    }
    }

    if(max_phi != 0.0){
    node->split_constant = split_constant;
    node->beta[split_coordinate] = 1;
    node->split_input = split_coordinate;
    } else{
    node->split_constant = 0.0;
    node->beta[split_coordinate] = 0.0;
    node->split_input = split_coordinate = -1;
    }
    
    return(split_coordinate);
}

/************************************************************************
* NAME: bintree_copy_and_normalize_data(tree, node)
* DESCRIPTION:
*   copies data for this node into linear_node_data,
*   normalizes data using inter-quartile distances
*************************************************************************/
static void
bintree_copy_and_normalize_data(BTREEP tree, BTLINK node)
{
    int coor, pat;
    float one_quart, three_quart;

    for(coor = 0; coor < tree->params.ninputs; coor++){
    /* copy data into norm matrix */
    for(pat = 0; pat < node->npatterns; pat++){
        tree->body.linear_node_data[coor][pat] =
          node->patterns[pat]->inputs[coor];
    }
    /* sort and pick out patterns at 1/4, 1/2, and 3/4 */
    /* we are using classes as the index because it is the right
       length and hasn't been used yet at this point */
    heap_sort(node->npatterns, &tree->body.linear_node_data[coor][0],
          tree->body.linear_classes);
    one_quart = tree->body.linear_node_data[coor]
      [tree->body.linear_classes[node->npatterns/4]];
    three_quart = tree->body.linear_node_data[coor]
      [tree->body.linear_classes[(3*node->npatterns)/4]];
    tree->body.median[coor] = tree->body.linear_node_data[coor]
      [tree->body.linear_classes[node->npatterns/2]];
    tree->body.range[coor] = ABS(three_quart - one_quart);

    /* normalize copied patterns */
    if(tree->body.range[coor] != 0){
        tree->body.range[coor] = 1./tree->body.range[coor];
        for(pat = 0; pat < node->npatterns; pat++){
        tree->body.linear_node_data[coor][pat] -=
          tree->body.median[coor];
        tree->body.linear_node_data[coor][pat] *=
          tree->body.range[coor];
        }
    }
    }
}

/************************************************************************
* NAME: bintree_find_best_split
*          (npatterns, nclasses, class, max_phi, split_values, tree)
* DESCRIPTION:
*   Takes an array 'split_values' with 'npatterns' number of elements
*   sorted by their absolute values. Going down the array finds the
*   splitting points and calculates a splitting criteria phi for
*   each such point looking for the maximum value of phi. Returns the
*       index for the best (with maximum phi) splitting point.
*************************************************************************/
/* called by bintree_find_best_delta_and_gamma, bintree_set_univariate_split,
  bintree_set_linear_combination_parameters */
static int
bintree_find_best_split (int npatterns, int nclasses, int *class,
             float *max_phi, float *split_values, BTREEP tree)
{
    int i;
    int split_point, nyes, nno;
    int *yes = tree->body.yes_class_count;
    int *no = tree->body.no_class_count;
    float phi, split_value;

    /* resetting the counting numbers */
    nyes = 0; nno = npatterns;
    for(i = 0; i < nclasses; i++){ yes[i] = 0; no[i] = 0; }
    for(i = 0; i < npatterns; i++) no[class[i]]++;
    split_point = -1;

    for(i = 0; i < npatterns; i++){
    int j;

    split_value = bintree_split_value(split_values, npatterns,i);
    while(i < npatterns && split_values[i] <= split_value){
        yes[class[i]]++; nyes++;
        no[class[i]]--;  nno--;
        i++;
    }
    i--;

    /*
     * This part searches for the best performer
     * with maximum splitting crtiterion 'phi' calculated
     * according to the delta splitting rule described in CART
     * chapt 11 Appendix.
     */
    phi = 0.;
    for(j = 0; j < nclasses; j++){
        phi += ABS((float)yes[j]/((float)nyes + 1.0e-5)-
               (float)no[j]/((float)nno + 1.0e-5));
    }
    phi *= 2.0 * nyes * nno/((float)npatterns*(float)npatterns+1.0e-5);

    if(phi > *max_phi){
        *max_phi = phi;
        split_point = i;
    }
    }

    return(split_point);
}

/************************************************************************
* NAME: bintree_split_value(split_values, npatterns, patternnum)
* DESCRIPTION:
*   C
*************************************************************************/
/* called by bintree_find_best_split, bintree_set_univariate_split,
   bintree_set_linear_combination_parameters */
static float
bintree_split_value(float *split_values, int npatterns, int patternnum)
{
    float split_value=0.0;
    float upper, lower;

    if(patternnum < npatterns-1){
    split_value = split_values[patternnum] + split_values[patternnum+1];
    split_value *= .5;
    /* check to see if these are reflections across 0 */
    upper = split_values[patternnum+1] + split_values[patternnum];
    lower = split_values[patternnum+1] - split_values[patternnum];
    if(lower != 0.)
      if(ABS(upper/lower) < 1e-4)
        split_value = 0.0;
    } else {
    split_value = split_values[patternnum];
    if(split_value < 0) split_value -= split_value*0.01;
    else split_value += split_value*0.01;
    }
    return(split_value);
}

/************************************************************************
* NAME: bintree_set_linear_combination_parameters(tree, node)
* DESCRIPTION:
*   Calculates the best linear combination parameters for a node
*   has the best univariate split as a preliminary answer
*************************************************************************/
/* called by bintree_split_node */
static void
bintree_set_linear_combination_parameters(BTREEP tree, BTLINK node)
{
    int coordinate, split_point, i;
    int counter=0;
    float sum = 0.;
    float delta, gamma;
    float max_phi=0.0, previous_max_phi=0.0;
    float delta_phi=0.;
    float epsilon = (float) 0.00001;

    do{ /* until series converges */

    for(coordinate = 0; coordinate < tree->params.ninputs; coordinate++){
        /* copy betas and split constant */
        for(i = 0; i < tree->params.ninputs; i++){
        tree->body.linear_beta[i] = node->beta[i];
        }
        tree->body.linear_split_constant = node->split_constant;

        /* initialize search_array and classes */
        bintree_set_linear_combination_values(tree, node);
        /* search_array must not be sorted yet for
           find_best_delta_and_gamma to work */

        /* find best delta and gamma to update linear combination params */
        max_phi =
          bintree_find_best_delta_and_gamma(coordinate, &delta, &gamma,
                        tree, node);
 
        /* update and normalize lin. comb. parameters & split constant */
        node->beta[coordinate] -= delta;
        node->split_constant += delta * gamma;

        /* normalize betas to sum to 1 and adjust split_constant */
        for(i = 0; i < tree->params.ninputs; i++)
          sum += ABS(node->beta[i]);
        if(sum != 0){
        for( i = 0; i < tree->params.ninputs; i++)
          node->beta[i] /= sum;
        node->split_constant /= sum;
        }

        /* calculate and sort linear combination values */
        bintree_set_linear_combination_values(tree, node);
        /* search_array must be sorted for find_best_split */
        heap_sort2(node->npatterns, tree->body.search_array,
               tree->body.classes);

        /* reset max phi */
        max_phi = 0.0;

        /* finding the split constant with the best split performance */
        split_point =
          bintree_find_best_split(node->npatterns, tree->params.nclasses,
                      tree->body.classes, &max_phi,
                      tree->body.search_array, tree);

        if(max_phi > previous_max_phi){
        /* if phi has improved, use the new betas */
        node->split_constant =
          bintree_split_value(tree->body.search_array, node->npatterns,
                      split_point);
        } else {
        /* copy the old betas and split_constant */
        for(i = 0; i < tree->params.ninputs; i++){
            node->beta[i] = tree->body.linear_beta[i];
        }
        node->split_constant = tree->body.linear_split_constant;
        }
    }/* for each coordinate */

    /* calculating difference to check convergence */
    delta_phi = max_phi - previous_max_phi;

    counter++;

    /* resetting the previous values */
    previous_max_phi = max_phi;

    } while(delta_phi > epsilon && counter < 5);
    /* check for convergence. Do at most five times */

    /* un-do normalization for betas and delta, update split_constant */
    /* this is odd looking */
    for(i = 0; i < tree->params.ninputs; i++){
    node->beta[i] *= tree->body.range[i];
    node->split_constant += node->beta[i]*tree->body.median[i];
    }
}

/************************************************************************
* NAME: bintree_set_linear_combination_values(class, v, tree, node)
* DESCRIPTION:
*   Calculates the linear combination values 'v[i]' for all patterns
*   of the node. The parameter name 'beta' is taken from CART
*   chapt. 5 Appendix.
*************************************************************************/
/* called by bintree_set_linear_combination_parameters */
static void
bintree_set_linear_combination_values(BTREEP tree, BTLINK node)
{
    int i, j;

    /* make a fresh copy of the data and classes in the original order */
    for(i = 0; i < node->npatterns; i++){
    /* calculating the linear combinations */
    tree->body.search_array[i] = 0.0;
    for(j = 0; j < tree->params.ninputs; j++){
        /* because we do the indices backwards, this could
           cause paging problems */
        tree->body.search_array[i] +=
          node->beta[j] * tree->body.linear_node_data[j][i];
    }
    tree->body.classes[i] = node->patterns[i]->class;
    }
}

/************************************************************************
 * NAME: bintree_find_best_delta_and_gamma (coordinate, class,
 *              best_delta, best_gamma, tree, node)
 * DESCRIPTION:
 *  Calculates the linear combination parameters as described
 *      in CART chapt 5 Appendix. The variable names are taken 
 *  from the same section.
 *
 * WARNING: tree->body.search_array must not be sorted before use here
 *************************************************************************/
/* called by bintree_set_linear_combination_parameters */
static float
bintree_find_best_delta_and_gamma (int coordinate, float *best_delta,
                   float *best_gamma, BTREEP tree, BTLINK node)
{
    int split_point;
    float gamma, max_phi=0;
    float *u;
    int   *class;
    float *v;
    float x_and_gamma;

    u = tree->body.linear_search_array;
    /* filled in set_linear_combination_values */
    class = tree->body.linear_classes;
    v = tree->body.search_array;

    for(gamma = -0.25; gamma <= 0.25; gamma += 0.25){
    int n_in_u, i;
    
    /*
     * Calculates the split values 'u[i]' as described in CART chapt. 5
     * Appendix using the linear combination values 'v[i]'.
     */
    /* find best positive split (x_and_gamma >= 0) */
    /* I have taken out the values for x = gamma, because of the
       singularity involved, lck */
    for(n_in_u = i = 0; i < node->npatterns; i++){
        x_and_gamma = tree->body.linear_node_data[coordinate][i] + gamma;
        if(x_and_gamma > 0.0){
        u[n_in_u] = (v[i] - node->split_constant)/x_and_gamma;
        class[n_in_u] = tree->body.classes[i];
        n_in_u++;
        }
    }
    heap_sort2(n_in_u, u, class);

    split_point =
      bintree_find_best_split(n_in_u, tree->params.nclasses,  
                  class, &max_phi, u, tree);
    if(split_point != -1){
        *best_delta = bintree_split_value(u, n_in_u, split_point);
        *best_gamma = gamma;
    }

    /* find best negative split (x_and_gamma < 0)*/
    for(n_in_u = i = 0; i < node->npatterns; i++){
        x_and_gamma = tree->body.linear_node_data[coordinate][i] + gamma;
        if(x_and_gamma < 0.0){
        u[n_in_u] = (v[i] - node->split_constant)/x_and_gamma;
        class[n_in_u] = tree->body.classes[i];
        n_in_u++;
        }
    }
    heap_sort2(n_in_u, u, class);

    split_point =
      bintree_find_best_split(n_in_u, tree->params.nclasses,  
                  class, &max_phi, u, tree);
    if(split_point != -1){
        *best_delta = bintree_split_value(u, n_in_u, split_point);
        *best_gamma = gamma;
    }
    }
    return(max_phi);
}

/************************************************************************
 * NAME: bintree_make_child(tree, node, child_kind)
 * DESCRIPTION:
 *  Creates a node and assigns it to the mother node as specified
 *       by child kind. Returns the addres of the new node.
 *************************************************************************/
/* called by bintree_split_node */
static BTLINK
bintree_make_child(BTREEP tree, BTLINK node, char *child_kind)
{
    BTLINK child;
    
    child = (BTLINK) allocate_structvector(1, sizeof(BTNODE));

    /* initialize elements of new the node */
    bintree_initialize_node(tree, child);

    /* set the mother node */
    child->mother = node;

    /* make it a yes or no child of the mother node */
    switch(WHICH_CHILD(child_kind))  /* WHICH_CHILD defined as macro */
      {
    case YES: child->mother->yeschild = child; break;

    case NO: child->mother->nochild = child; break;

      /* this line could be here, except that root is 
         allocated as a part of the structure tree, not here */
      /* case ROOT: tree->root = *child; break; */
    default:
      fatal_error("\n ERROR in make_child: child kind %s is unknown \n",
         child_kind);
      }
      
    /* calculate level of new node */
    child->level = child->mother->level + 1;

    /* add the new node to the tree bin list */
    bintree_add_new_node(tree, child, child_kind); 

    return((BTLINK) child);
}

/************************************************************************
* NAME: bintree_bad_division(tree, node)
* DESCRIPTION:
*   checks that this split will actually divide patterns.
*   if it doesn't, we will infinite loop trying to split right here
*   returns 0 if this is an ok split
*   return -1 if this is a bad split
*
*************************************************************************/
/* called by bintree_split_node */
static int
bintree_bad_division(BTREEP tree, BTLINK node)
{
    int   i, nyes, nno;
    float v; /* splitting criteria (lin comb val) calcd for every pattern */
    float c; /* split constant (comparison constant) for the node */

    /*
     * Splitting criteria v is calculated using linear combination
     * parameters.
     * If the split is suppoused to be univariate, only the split input is 
     * taken as the splitting criteria.
     */

    /* initializing pattern numbers at branching nodes */
    nyes = nno = 0;

    for(i = 0; i < node->npatterns; i++){
    /* set linar combination value v and comparison constant c */
    v = bintree_node_linear_combination_value(tree, node,
                          node->patterns[i]->inputs);
    c = node->split_constant;

    if(v <= c) nyes++;
    else nno++;
    /* we don't actually need to count them all, just until we
       have evidence that it will split a little bit */
    if(nyes > 0 && nno > 0) return 0;
    }
    if(nyes > 0 && nno > 0) return 0;
    else return -1;
}

/************************************************************************
* NAME: bintree_divide_patterns(tree, node)
* DESCRIPTION:
*   divides patterns of a node between the yes- and nochild according
*   the splitting parameters set before this function is called.
*
*************************************************************************/
/* called by bintree_split_node */
static void
bintree_divide_patterns(BTREEP tree, BTLINK node)
{
    int   i;
    float v; /* splitting criteria (lin comb val) calcd for every pattern */
    float c; /* split constant (comparison constant) for the node */

    /*
     * Splitting criteria v is calculated using linear combination
     * parameters.
     * If the split is suppoused to be univariate, only the split input is 
     * taken as the splitting criteria.
     */

    /* initializing pattern numbers at branching nodes */
    node->yeschild->npatterns = node->nochild->npatterns = 0;

    for(i = 0; i < node->npatterns; i++){
    /* set linar combination value v and comparison constant c */
    v = bintree_node_linear_combination_value(tree, node,
                          node->patterns[i]->inputs);
    c = node->split_constant;

    if(v <= c){
        node->yeschild->patterns[node->yeschild->npatterns] =
          node->patterns[i];
        node->yeschild->npatterns++;
    } else{
        node->nochild->patterns[node->nochild->npatterns] =
          node->patterns[i];
        node->nochild->npatterns++;
    }
    }
}

/************************************************************************
* NAME: bintree_update_tree_error(tree, node)
* DESCRIPTION:
*   Sets the total error rate of tree grown so far.
*   Before calling this function the node_misclassifications
*       of the yes and no children must be calculated.
*   node_misclassifications are calculated in set_node_class().
*************************************************************************/
/* called by bintree_split_node */
/* used for epoch print out */
static void
bintree_update_tree_error(BTREEP tree, BTLINK node)
{
    if(!TRMNL_NODE(node)){
    tree->body.tree_error -= node->node_misclass * node->probability;
    tree->body.tree_error +=
      ((node->yeschild->node_misclass * node->yeschild->probability) +
       (node->nochild->node_misclass * node->nochild->probability));
    }
}

/************************************************************************
* NAME: bintree_add_new_leaf(tree, node)
* DESCRIPTION:
*   adds the address of the new terminal nodes to the leaves array 
*   and sets the new appropriate indexes for that array
*   
*************************************************************************/
/* called by bintree_split_node */
static void
bintree_add_new_leaf(BTREEP tree, BTLINK node)
{
    int old_size=0, new_size=0;

    /* check if we need to increase the queue of leaves waiting to be split */
    /* it was allocated the first time in bintree_build */
    if(tree->body.leaves_size <= tree->body.nleaves){
    old_size = tree->body.leaves_size;
    new_size = old_size + MEMORY_BLOCK;

    if(old_size == 0)
      tree->body.leaves_index = allocate_ivector(new_size);
    else
      tree->body.leaves_index =
        reallocate_ivector(tree->body.leaves_index, old_size, new_size);

    tree->body.leaves_size = new_size;
    }
    tree->body.leaves_index[tree->body.nleaves] = node->selfno;
    tree->body.nleaves++;

#if 0
    /* an alternate way of doing the sort */
    /* this messes up the breadth first order of the nodes in body.bins
       which could be a problem.  Having a strict breadth first order
       means that all the nodes on level i are written before any from
       level i+1.  You can travel up the tree from the leaves by going
       through body.bins backwards when it is in breadth first order */
    {
    int i;
    for(i = tree->body.nleaves; i > 0; i--){
        /* check if we are at our position in the list */
        if(tree->body.bins[tree->body.leaves_index[i-1]]->node_misclass >=
           node->node_misclass) break;
        /* shift the lesser nodes up in the queue */
        tree->body.leaves_index[i] = tree->body.leaves_index[i-1];
    }
    tree->body.leaves_index[i] = node->selfno;
    }
#endif
}


/***************************************
 ROUTINES USED IN PRUNING
***********************/

/************************************************************************
* NAME: bintree_calculate_nleaves_and_strength(tree)
* DESCRIPTION:
*   sets the number of leaves and non-terminal nodes for each subtree
*   of tree
*   Calculates submisclassifications and node strength of
*   non-terminal nodes using values set by bintree_set_node_class() 
*   The misclassifications are calculated by formulas explained
*   in CART chapt 2.
*   R(t) = node_misclassification
*   R(T') = sub_misclass
*   T' is the set of all terminal nodes in tree T(t)
*   |T'| = nterminals
*   strength = (R(t) - R(T')) / (nterminals - 1)
*************************************************************************/
/* called by bintree_prune */
static void
bintree_calculate_nleaves_and_strength(BTREEP tree)
{
    BTLINK node;

    node = &tree->root;

    /* set the total number of nodes and leaves */
    tree->body.nnodes = (tree->body.nbins-1)/2;
    tree->body.nleaves = tree->body.nnodes+1;

    /* this is a slightly different tree traversal, we must go all the
       way to the bottom of each branch to sum up the number of leaves,
       non-terminal nodes, and calculate the strength for each subtree */
    while(!ROOT_NODE(node) || !VISITED_NODE(node)){
    /* VISITED_NODE stops for terminal nodes */
    while(!VISITED_NODE(node)){
        if(!node->yesvisit) node = node->yeschild;
        else if(!node->novisit) node = node->nochild;
    }

    if(TRMNL_NODE(node)){
        node->nleaves = 1;
        node->sub_misclass = node->node_misclass;
        node->strength = -MINFLOAT;
    } else {
        /* these calculations are done for the children */
        node->nleaves = node->yeschild->nleaves + node->nochild->nleaves;
        node->sub_misclass =
          node->yeschild->sub_misclass + node->nochild->sub_misclass;
        node->strength = 
          ((node->node_misclass - node->sub_misclass))/
        ((float) node->nleaves - (float) 1.0);
    }
    node->visit = 1;
    node = node->mother;

    if(node->nochild->visit) node->novisit = 1;
    else if(node->yeschild->visit) node->yesvisit = 1;
    }

    /* calculate strength for the root node */
    if(TRMNL_NODE(node)){
    node->nleaves = 1;
    node->sub_misclass = node->node_misclass;
    node->strength = -MINFLOAT;
    } else {
    node->nleaves = node->yeschild->nleaves + node->nochild->nleaves;
    node->sub_misclass =
      node->yeschild->sub_misclass + node->nochild->sub_misclass;
    node->strength = 
      ((node->node_misclass - node->sub_misclass))/
        ((float) node->nleaves - (float) 1.0);
    }
    tree->body.nleaves = node->nleaves;

    /* resetting all visit flags to enable traveling along the tree again */
    bintree_reset_visit_flags(node);  /* here node is the root node */

}

/************************************************************************
* NAME: bintree_calculate_tree_error(tree, node)
* DESCRIPTION:
*   sets tree_error by checking the errors of the pruned nodes
*       
*************************************************************************/
/* called by bintree_prune */
static void
bintree_calculate_tree_error(BTREEP tree)
{
    BTLINK node;

    tree->body.tree_error = (float) 0.0;

    node = &tree->root;
    node->visit = 1;

    while(!ROOT_NODE(node) || !VISITED_NODE(node)){
    while(!VISITED_NODE(node)){
        if(!node->yesvisit){
        node->yesvisit = 1;
        node = node->yeschild;
        } else if(!node->nochild->visit){
        node->novisit = 1;
        node = node->nochild;
        }
        if(TRMNL_NODE(node))
          tree->body.tree_error += node->node_misclass * node->probability;
        node->visit = 1;
    }
    node = node->mother;
    }
    if(TRMNL_NODE(node))
      tree->body.tree_error += node->node_misclass*node->probability;
    
    /* resetting all visit flags to enable traveling along the tree again */
    bintree_reset_visit_flags(node);   /* here node is the top node */
}

/************************************************************************
* NAME: bintree_find_weakest(tree)
* DESCRIPTION:
*   returns the index to the non-terminal node with the least strength
*************************************************************************/
static int
bintree_find_weakest_node(BTREEP tree)
{
    int i, weakesti;
    BTLINK node;
    
    weakesti = 0; /* start with the root node */
    for(i = 0; i < tree->body.nbins; i++){
    node = tree->body.bins[i];
    if(!TRMNL_NODE(node)){
        if(node->strength < tree->body.bins[weakesti]->strength)
          weakesti = i;
    }
    }
    return weakesti;
}

/************************************************************************
* NAME: bintree_prune_children(tree, wnode, prune_num)
* DESCRIPTION:
*   prunes the children of wnode, placeing any non-terminal nodes
*   on the pruning list
*   returns the current number of non-terminal nodes after pruning
*************************************************************************/
static int
bintree_prune_children(BTREEP tree, BTLINK node, int prune_num)
{
    int prune_top_id;

    prune_top_id = node->selfno;

    node->topflag = 1;
    node->visit = 1;
    while(!TOP_NODE(node) || !VISITED_NODE(node)){
    while(!VISITED_NODE(node)){/* also checks for non-terminal */
        if(!node->yesvisit){
        node->yesvisit = 1;
        node = node->yeschild;
        node->mother->yeschild = NULL; /* break link to this child */
        }
        else if(!node->novisit){
        node->novisit = 1;
        node = node->nochild;
        node->mother->nochild = NULL;
        }
        if(TRMNL_NODE(node)){
        /* fix the tree error */
        tree->body.tree_error -= node->node_misclass*node->probability;
        } else {
        /* add a space for this node in the pruned list */
        tree->body.bin_index[prune_num] = prune_top_id;
        prune_num--;
        }
        node->prune = 1;
        node->visit = 1;
    }
    node = node->mother;
    }
    node->topflag = 0;
    node->prune = 1;
    bintree_reset_visit_flags(node);
    return prune_num;
}

/************************************************************************
* NAME: bintree_link_tree(tree)
* DESCRIPTION:
*   sets the mother and child links for all the nodes in a tree
*   all nodes must already be allocated and be in tree->body.bins
*************************************************************************/
/* called by bintree_prune */
static void
bintree_link_tree(BTREEP tree)
{
    int binnum;
    BTLINK node;

    for(binnum = 0; binnum < tree->body.nbins; binnum++){
    node = tree->body.bins[binnum];
    bintree_set_node_addresses(tree, node);
    node->prune = 0;
    }
    tree->body.nnodes = (tree->body.nbins-1)/2;
    tree->body.nleaves = tree->body.nnodes+1;
}

/************************/
/* ROUTINES FOR TESTING */
/************************/

/************************************************************************
* NAME: bintree_set_pruning(tree)
* DESCRIPTION:
*   prunes nodes, making them leaves, in order given in body.bin_index
*************************************************************************/
/* called by bintree_test_init */
static int
bintree_set_pruning(BTREEP tree, int max_nodes)
{
    int binnum, nodei;
    BTLINK node;

    if(max_nodes > tree->body.nnodes || max_nodes < 0)
      max_nodes = tree->body.nnodes;

    binnum = tree->body.nnodes-1;
    while(binnum >= max_nodes){
    /* pick out the next node to be pruned */
    nodei = tree->body.bin_index[binnum];
    node = tree->body.bins[nodei];
    /* make our node and all its children leaves */
    if(!TRMNL_NODE(node)){
        node->topflag = 1;
        while(!TOP_NODE(node) || !VISITED_NODE(node)){
        while(!VISITED_NODE(node)){
            if(!node->yesvisit) node = node->yeschild;
            else if(!node->novisit) node = node->nochild;
        }
        if(VISITED_NODE(node))
          node->yeschild = node->nochild = NULL;
        node->visit = 1;
        node->prune = 1;
        node = node->mother;
        if(node->nochild->visit) node->novisit = 1;
        else if(node->yeschild->visit) node->yesvisit = 1;
        }
        node->yeschild = node->nochild = NULL;
    }
    /* skip over spaces for non-terminal children of node in bin_index */
    while(binnum >= 0 && tree->body.bin_index[binnum] == nodei) binnum--;
    }
    return(binnum+1);
}

/****************************************************************************
* NAME: bintree_find_pattern_node(tree, pattern,inputs,outputs,output_class)
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_classify, bintree_eval */
static BTLINK
bintree_find_pattern_node(BTREEP tree, float *inputs, float *outputs,
              int *output_class)
{
    int j;
    float v=0, c=0;
    BTLINK node, next_node;

    node = &tree->root;
    
    
    while(!PRUNED_NODE(node)){
    /* set linar combination value v and comparison constant c */
    v = bintree_node_linear_combination_value(tree, node, inputs);
    c = node->split_constant;

    if(v <= c)
      next_node = node->yeschild;
    else
      next_node = node->nochild;

    node = next_node;
    }
    *output_class = node->class;
    for(j = 0; j < tree->params.nclasses; j++)
      outputs[j] = node->class_distribs[j];

    return((BTLINK) node);
}

/************************************************************************
* NAME: bintree_node_linear_combination_value(tree, node, inputs)
* DESCRIPTION:
*   Calculates the linear combination value for a given pattern
*   input vector in case of a linear search. 
*   If the search is binary only the value of the
*   split feature (coordinate) is taken instead of the sum.
*************************************************************************/
/* called by bintree_find_pattern_node, bintree_divide_patterns */
static float
bintree_node_linear_combination_value(BTREEP tree, BTLINK node, float *inputs)
{
    int j;
    float v = 0.0;

    if(tree->params.linear)
      for(j = 0; j < tree->params.ninputs; j++)
    v += node->beta[j] * inputs[j];
    else
      v = inputs[node->split_input];

    return(v);
}

/********************************
  READING AND WRITING PARAMETER FILES 
**********************************/

/*** Parameter Table ***/
static PARAM    bintscalars[] = {
    {"linear", P_INT},
    {"ninputs", P_INT},
    {"nclasses", P_INT},
    {"npatterns", P_INT},
};
#define NBINTSCALARS (sizeof(bintscalars)/sizeof(PARAM))

static PARAM    bintarrays[] = {
    {"features", P_INT_VECTOR, NULL, 0, 0},
    {"class_count", P_FLOAT_VECTOR, NULL, 0, 0},
};
#define NBINTARRAYS (sizeof(bintarrays)/sizeof(PARAM))

static PARAM    bintreeparams[] = {
    {"nbins", P_INT},
    {"nlevels", P_INT},
};
#define NBINTREEPARAMS (sizeof(bintreeparams)/sizeof(PARAM))   

static PARAM    bintnodeids[] = {
    {"binnum", P_INT},
    {"yeschildnum", P_INT},
    {"nochildnum", P_INT},
    {"mothernum", P_INT},
    {"class", P_INT},
    {"probability", P_FLOAT},
    {"npatterns", P_INT},
    {"level", P_INT},
    {"split_constant", P_FLOAT},
    {"split_feature", P_INT},
};
#define NBINTNODEIDS (sizeof(bintnodeids)/sizeof(PARAM))   

static PARAM    bintnodeoutputarrays[] = {
    {"lin_parameters", P_FLOAT_VECTOR, NULL, 0, 0},
    {"class_outputs", P_FLOAT_VECTOR, NULL, 0, 0},
};
#define NBINTNODEOUTPUTARRAYS (sizeof(bintnodeoutputarrays)/sizeof(PARAM))   

static PARAM    bintreepruning[] = {
    {"pruning_order", P_INT_VECTOR, NULL, 0, 0},
};
#define NBINTREEPRUNING (sizeof(bintreepruning)/sizeof(PARAM)) 

/***************************************************************************
* NAME: bintree_write_model_params(tree, fp)
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
***************************************************************************/
/* called by bintree_write_model */
static int
bintree_write_model_params(BTREEP tree, FILE *fp)
{
    int rv;

    /* set scalars */
    bintree_set_model_scalar_params(tree);

    /* set vectors and matrices */
    bintree_set_model_array_params(tree);

                    /* save scalars */
    rv = param_putblk(fp, bintscalars, NBINTSCALARS); 
    if (rv < 0) return(rv);
                    /* save vectors and matrices */
    rv = param_putblk(fp, bintarrays, NBINTARRAYS);
    if (rv < 0) return(rv);

    return(0);
}

/***************************************************************************
* NAME: bintree_write_tree_model(tree, fp)
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
***************************************************************************/
/* called by bintree_write_model */
static int
bintree_write_tree_model(BTREEP tree, FILE *fp)
{
    int rv;

    fprintf(fp, "tree parameters\n");
    rv = bintree_write_tree_params(tree, fp);
    if(rv < 0) return rv;

    fprintf(fp, "node parameters\n");
    rv = bintree_write_node_params(tree, fp);
    if(rv < 0) return rv;

    fprintf(fp, "pruning order\n");
    rv = bintree_write_pruning_order(tree, fp);
    if(rv < 0) return rv;

    return(0);
}

/***************************************************************************
* NAME: bintree_write_tree_params(tree, fp)
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_write_tree_model */
static int
bintree_write_tree_params(BTREEP tree, FILE *fp)
{
    int rv;

    /* sets necessary pointers for param_putblk to write tree parameters */
    bintree_set_tree_params(tree);

    /* write total number of nodes, split nodes terminal nodes and levels */
    rv = param_putblk(fp, bintreeparams, NBINTREEPARAMS);
    if (rv < 0) return(rv);

    return(0);
}

/***************************************************************************
* NAME: bintree_write_node_params(tree, fp)
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_write_tree_model */
static int
bintree_write_node_params(BTREEP tree, FILE *fp)
{
    int rv;
    int binnum;
    BTLINK node;

    /* write information for each individual node */
    for(binnum = 0; binnum < tree->body.nbins; binnum++){
    node = tree->body.bins[binnum];

    /* sets pointers for param_putblk to write node parameters */
    bintree_set_node_params(tree, node);
    
    /* write node numbers */
    rv = param_putblk(fp, bintnodeids, NBINTNODEIDS);
    if (rv < 0) return(rv);

    /* write linear comb params (if linear tree), and class outputs */
    rv = param_putblk(fp, bintnodeoutputarrays, NBINTNODEOUTPUTARRAYS);
    if (rv < 0) return(rv);
    }

    return(0);
}

/****************************************************************************
* NAME: bintree_write_pruning_order(tree, fp)
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_write_tree_model */
static int
bintree_write_pruning_order(BTREEP tree, FILE *fp)
{
    int rv;

    /* sets necessarry pointers for param_putblk() to write pruning order */
    bintree_set_pruning_pointer(tree);

    /* write pruning order */
    rv = param_putblk(fp, bintreepruning, NBINTREEPRUNING);

    if (rv < 0) return(rv);

    return(0);
}

/***************************************************************************
* NAME: bintree_read_model_params(fp)
* DESCRIPTION:
*       reads parameters and exemplars for a model to a file
*       in ASCII format
*       builds a model and return the model number
***************************************************************************/
/* called by bintree_read_model */
static int
bintree_read_model_params(FILE *fp)
{
    int    rv, model;
    BTREE  temp_tree;
    /* temporary tree structure to read in initial params */

    BTREEP tree;      /* temporary pointer for model */

    /* sets pointers for param_getblk to load model scalar parameters */
    bintree_set_model_scalar_params(&temp_tree);

    /* load scalars: initial parameters needed to build a new model */
    rv = param_getblk(fp, bintscalars, NBINTSCALARS); 
    if (rv < 0) return(rv);

    /* build a new model */
    model = bintree_build(temp_tree.params.ninputs,
              temp_tree.params.nclasses, EPOCH_VERBOSITY+1);

    /* set temporary pointer for model */
    tree = bintpp[model];

    /* load number of total patterns used to train model */
    tree->params.linear = temp_tree.params.linear;
    tree->params.npatterns = temp_tree.params.npatterns;

    /* sets pointers for param_getblk to load model array parameters */
    bintree_set_model_array_params(tree);

    /* reset the length on features */
    param_setdim("features", tree->params.ninputs+1, 0,
         bintarrays, NBINTARRAYS);

    /* load vectors and matrices */
    rv = param_getblk(fp, bintarrays, NBINTARRAYS);
    if (rv < 0) return(rv);

    return(model);
}

/***************************************************************************
* NAME: bintree_read_tree_model(tree, fp)
* DESCRIPTION:
*       reads parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_read_model */
static int
bintree_read_tree_model(BTREEP tree, FILE *fp)
{
    int rv;

    rv = bintree_read_tree_params(tree, fp);
    if(rv < 0) return rv;

    rv = bintree_read_node_params(tree, fp);
    if(rv < 0) return rv;

    rv = bintree_read_pruning_order(tree, fp);
    if(rv < 0) return rv;

    return(0);
}

/****************************************************************************
* NAME: bintree_read_tree_params(tree, fp)
* DESCRIPTION:
*       reads parameters and exemplars for a model to a file
*       in ASCII format
***************************************************************************/
/* called by bintree_read_tree_model */
static int
bintree_read_tree_params(BTREEP tree, FILE *fp)
{
    int rv;

    /* sets necessary pointers for param_getblk to read tree parameters */
    bintree_set_tree_params(tree);

    /* read total number of nodes, split nodes terminal nodes and levels */
    rv = param_getblk(fp, bintreeparams, NBINTREEPARAMS);
    if (rv < 0) return(rv);

    /* set nnodes and nleaves based on nbins */
    tree->body.nnodes = (tree->body.nbins-1)/2;
    tree->body.nleaves = tree->body.nnodes+1;

    return(0);
}

/***************************************************************************
* NAME: bintree_read_node_params(tree, fp)
* DESCRIPTION:
*       reads parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_read_tree_model */
static int
bintree_read_node_params(BTREEP tree, FILE *fp)
{
    int rv;
    int binnum;
    BTNODE bin, *node; /* temporary structure to load in node data */

    node = &bin;

    /* allocates memory for bins and stores their addresses in an array */
    bintree_make_bins(tree);

    /* read information for each individual node */

    for(binnum = 0; binnum < tree->body.nbins; binnum++){
    bintree_initialize_node(tree, node);

    /* set pointers for bintnodeids, bintnodeoutputarray */
    bintree_set_node_params(tree, &bin);

    /* read node numbers */
    rv = param_getblk(fp, bintnodeids, NBINTNODEIDS);
    if (rv < 0) return(rv);

    /* read linear comb params (if linear tree), and class outputs */
    rv = param_getblk(fp, bintnodeoutputarrays, NBINTNODEOUTPUTARRAYS);
    if (rv < 0) return rv;

    /* calculates some node parameters not read from parameter file */
    bintree_initialize_read_node(tree, node);

    /* sets addresses for children and mother nodes */
    bintree_set_node_addresses(tree, node);

    /* copy the information from temporary structure */
    /* this assignment of a structure (not the pointer!!) is in the
       original definition of C, but is permitted by newer compilers.
       If this needs to run with an old compiler, the fields must
       be copied one by one */
    *tree->body.bins[node->selfno]  = *node; /* = bin */
    }
    return(0);
}

/***************************************************************************
* NAME: bintree_read_pruning_order(tree, fp)
* DESCRIPTION:
*       reads parameters and exemplars for a model to a file
*       in ASCII format
***************************************************************************/
/* called by bintree_read_tree_model */
static int
bintree_read_pruning_order(BTREEP tree, FILE *fp)
{
    int rv;

    /* sets necessary pointers for param_getblk() to read pruning order */
    bintree_set_pruning_pointer(tree);

    /* read pruning order */
    rv = param_getblk(fp, bintreepruning, NBINTREEPRUNING);

    if (rv < 0) return(rv);

    return(0);
}
/************************************************************************
* NAME: bintree_make_bins(tree)
* DESCRIPTION:
*   allocates memory for nbins bins initializes them 
*   and stores their addresses in bins[] array
*************************************************************************/
/* called by bintree_read_node_params */
static void
bintree_make_bins(BTREEP tree)
{
    int binnum, nbins;

    nbins = tree->body.nbins;

    /* need to store nbins addresses */
    tree->body.bins = (BTLINK*) allocate_pvector(nbins);
    tree->body.bins_size = nbins;
    tree->body.bin_index = allocate_ivector(tree->body.nnodes);

    /* set addresses */
    tree->body.bins[0] = &tree->root;
    for(binnum = 1; binnum < nbins; binnum++)
      tree->body.bins[binnum] =
    (BTNODE *)allocate_structvector(1,sizeof(BTNODE));
}

/************************************************************************
* NAME: bintree_initialize_read_node(tree, node)
* DESCRIPTION:
*   Calculates some elements of node not read from parameter file.
*************************************************************************/
/* bintree_read_node_params */
static void
bintree_initialize_read_node(BTREEP tree, BTLINK node)
{
    int i;
    int nclasses = tree->params.nclasses;
    float sum=0.0;
    
    /* sets the linear combination value if binary tree */
    if(!tree->params.linear){
      if(node->split_input > 0 && node->split_input < tree->params.ninputs){
    node->beta[node->split_input] = 1;
      } else {
    node->beta[0] = 1;
      }
    }

    /* calculates the node misclassification */
    for(i = 0; i < nclasses; i++) sum += node->class_distribs[i];
    sum -= node->class_distribs[node->class];
    node->node_misclass = sum * node->probability;
}

/************************************************************************
* NAME: bintree_set_node_addresses(tree, node)
* DESCRIPTION:
*   allocates memory for nbins bins initializes them 
*   and stores their addresses in bins[] array
*************************************************************************/
/* called by bintree_read_node_params and bintree_link_tree */
static void
bintree_set_node_addresses(BTREEP tree, BTLINK node)
{
    
    if(node->motherno >= 0){
    if(node->motherno < tree->body.nbins){
        node->mother = tree->body.bins[node->motherno];
    } else{
        fatal_error("no allocation for: motherno = %d; max = %d; bin = %d",
            node->motherno, tree->body.nbins-1, node->selfno);
    }
    } else{
    node->mother = NULL;
    }

    if(node->yesno >= 0){
    if(node->yesno < tree->body.nbins){
        node->yeschild = tree->body.bins[node->yesno];
    } else{
        fatal_error("no allocation for: yesno = %d; max = %d; bin = %d",
              node->motherno, tree->body.nbins-1, node->selfno);
    }
    } else{
    node->yeschild = NULL;
    }

    if(node->nono >= 0){
    if(node->nono < tree->body.nbins){
        node->nochild = tree->body.bins[node->nono];
    } else{
        fatal_error("no allocation for: nono = %d; max = %d; bin = %d",
              node->motherno, tree->body.nbins-1, node->selfno);
    }
    } else{
    node->nochild = NULL;
    }
}

/***************************************************************************
* NAME: bintree_set_model_scalar_params(tree)
* DESCRIPTION:
*       sets pointers for parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_read_model_params, bintree_set_model_params */
static void
bintree_set_model_scalar_params(BTREEP tree)
{
                    /* set scalars */
    param_setvarptr("linear",  &tree->params.linear, 
            bintscalars, NBINTSCALARS);

    param_setvarptr("ninputs",  &tree->params.ninputs,
            bintscalars, NBINTSCALARS);

    param_setvarptr("nclasses", &tree->params.nclasses,
        bintscalars, NBINTSCALARS);

    param_setvarptr("npatterns",  &tree->params.npatterns,
            bintscalars, NBINTSCALARS);
}

/***************************************************************************
* NAME: bintree_set_model_array_params(tree)
* DESCRIPTION:
*       sets pointers for parameters and exemplars for a model to a file
*       in ASCII format
L****************************************************************************/
/* called by bintree_read_model_params, bintree_set_model_params */
static void
bintree_set_model_array_params(BTREEP tree)
{
                    /* set vectors and matrices */
    param_setvarptr("features", &tree->params.features[0],
            bintarrays, NBINTARRAYS);
    param_setdim("features", tree->params.features[0]+1, 0,
         bintarrays, NBINTARRAYS);

    param_setvarptr("class_count", (int *) &tree->params.class_count[0],
            bintarrays, NBINTARRAYS);
    param_setdim("class_count", tree->params.nclasses, 0,
         bintarrays, NBINTARRAYS);
}

/***************************************************************************
* NAME: bintree_set_tree_params(tree)
* DESCRIPTION:
*       sets pointers for parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_read_tree_params, bintree_write_tree_params */
static void
bintree_set_tree_params(BTREEP tree)
{
    /*set total number of nodes, split nodes terminal nodes and levels */

    param_setvarptr("nbins", &tree->body.nbins,
            bintreeparams, NBINTREEPARAMS);
    param_setvarptr("nlevels", &tree->body.nlevels,
            bintreeparams, NBINTREEPARAMS);
}

/***************************************************************************
* NAME: bintree_set_node_params(tree, node)
* DESCRIPTION:
*       sets pointers for parameters and exemplars for a model to a file
*       in ASCII format
****************************************************************************/
/* called by bintree_read__node_params, bintree_write_node_params */
static void
bintree_set_node_params(BTREEP tree, BTLINK node)
{
    /* set information for each individual node */
    
    /* set node numbers */
    param_setvarptr("binnum", &node->selfno,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("yeschildnum", &node->yesno,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("nochildnum", &node->nono,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("mothernum", &node->motherno,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("class", &node->class,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("probability", (int *) &node->probability,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("split_constant", (int *) &node->split_constant,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("split_feature", &node->split_input,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("npatterns", &node->npatterns,
            bintnodeids, NBINTNODEIDS);
    param_setvarptr("level", &node->level,
            bintnodeids, NBINTNODEIDS);

    /* set linear comb parameters (if linear tree), and class outputs */
    
    param_setvarptr("lin_parameters", (int *) &node->beta[0],
            bintnodeoutputarrays, NBINTNODEOUTPUTARRAYS);
    if(tree->params.linear){
    param_setdim("lin_parameters", tree->params.ninputs, 0,
             bintnodeoutputarrays, NBINTNODEOUTPUTARRAYS);
    } else{
    param_setdim("lin_parameters", 0, 0,
             bintnodeoutputarrays, NBINTNODEOUTPUTARRAYS);
    }
    
    param_setvarptr("class_outputs", (int *) &node->class_distribs[0],
            bintnodeoutputarrays, NBINTNODEOUTPUTARRAYS);
    param_setdim("class_outputs", tree->params.nclasses, 0,
         bintnodeoutputarrays, NBINTNODEOUTPUTARRAYS);

}

/***************************************************************************
* NAME: bintree_set_pruning_pointer(tree)
* DESCRIPTION:
*       sets pointers for parameters and exemplars for a model to a file
*       in ASCII format
***************************************************************************/
/* called by bintree_read_pruning order, bintree_write_pruning order */
static void
bintree_set_pruning_pointer(BTREEP tree)
{
    param_setvarptr("pruning_order", (int *) &tree->body.bin_index[0],
            bintreepruning, NBINTREEPRUNING);
    param_setdim("pruning_order", (tree->body.nbins-1)/2, 0,
             bintreepruning, NBINTREEPRUNING);
}


/**********************************
GENERAL ROUTINES
******************************************/

/************************************************************************
 * NAME: bintree_reset_visit_flags(node)
 * DESCRIPTION:
 *  travels around visited tree and resets all visit flags to zero
 *************************************************************************/
/* called by bintree_set_node_sub_misclass, bintree_calculaate_nterminals
   bintree_grow_node, bintree_calculate_all_strength,
   bintree_calculate_tree_error */
void
bintree_reset_visit_flags(BTLINK node)
{
    node->topflag = 1;

    while(!TOP_NODE(node) || RESET_NODE(node)){
    while(RESET_NODE(node)){
        if(node->yesvisit) node = node->yeschild;
        else if(node->novisit) node = node->nochild;
    }

    node->visit = 0;
    node = node->mother;

    if(!node->nochild->visit) node->novisit = 0;
    else if(!node->yeschild->visit) node->yesvisit = 0;
   }

    node->visit = 0;

    /* reset the top flag */
    node->topflag = 0;
}

/***************************************************************************
* NAME:       bintree_display_node_head(tree, grow)
* DESCRIPTION:
*             to display new node for monitoring the program flow
***************************************************************************/
/* called by bintree_grow_node and bintree_prune */
static void
bintree_display_node_head(BTREEP tree, int grow)
{
    if(tree->params.verbosity >= EPOCH_VERBOSITY){
    printf("\n          OVERALL");
    if(grow)printf("    NODE       NODE");
    printf("\n");
    printf("  NODES   ERROR %%");
    if(grow)printf("   ERROR %%   NERR/NPAT");
    printf("\n");
    printf(  " -------  --------");
    if(grow)printf("  --------  ---------\n");
    else printf("\n");
    }
}

/***************************************************************************
* NAME:       bintree_display_node(tree, node, grow)
* DESCRIPTION:
*             to display node information for either growing or pruning
***************************************************************************/
/* called by bintree_grow_node and bintree_prune */
static void
bintree_display_node(BTREEP tree, BTLINK node, int grow)
{
    float node_error;
    node_error = 1. - node->max_distrib;

    if(tree->params.verbosity >= EPOCH_VERBOSITY){
    printf("%5d  %10.2f", tree->body.nnodes,tree->body.tree_error*100.);
    if(grow)
      printf("%10.2f %6d/%d\n", 100.* node_error,
         (int)(.5 + node_error * node->npatterns), node->npatterns);
    else printf("\n");
    }
}
