/***-------------------------------------------------------------------------
 * NAME:        cknn.c
 * PURPOSE: driver for Condensed K Nearest Neighbor classifier
 *
 * SEE ALSO: cknn_pack.c
 *
 * DESCRIPTION: 
 *  Functions in this file:
 *      cknn()  -- driver for k nearest neighbor classifier
 *          -- handles train, create flags
 *  -- description of algorithm in cknn_main.c
 * 
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/cknn/cknn_cv.c,v 11.4 2001/08/23 20:11:28 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/cknn/cknn_cv.c,v $
 * CREATED: 7-Mar-91 lck            MIT LINCOLN LABORATORY
 **************************************************************************/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "cknn.h"
#include "cknn_pack.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

static char prog_name[MAXSTR];  /* static variables */

float
cknn_cross_valid(int nraw, int ninputs, int *features, int noutputs,
         int normalize, int npatterns, int debg, int verbosity,
         int verror, int nepochs, int random_flag, int cv_nfolds,
         int cv_nsplits, int use_cv_splits, int *cv_splits,
         char **cv_train_mask, char **cv_test_mask, char *data_path,
         char *data_fname, char *norm_fname, char *exp_path,
         char *error_fname, char *param_fname, char **labels,
         char *comment, int adjust_priors, int priors_npatterns,
         float *priors_list, int cv_random, int k)
{
    int     model, err_model, err_part_model, data_model;
    int     normalization = 0, epoch;
    int     desired_class, class;
    int     n_read, n_trials, trial_num;
    int     accum_trials;
    int     map_flag = 0;   /* needed for error reporting routines */
    int     fold;
    float   *inputs, *outputs, percent_error;
    char    *tmp_path = NULL;
    FILE    *error_file = NULL, *param_file = NULL;
        int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());
    
    /* allocate and initialize model structure */
    accum_trials = 0;
    model = cknn_build(ninputs, noutputs, npatterns, k);
    cknn_init(model, accum_trials, features);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    outputs = *(float **) cknn_param(model, "outputs");
    
    if (verror > 0) {       /* open file for classification errors */
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    }
    
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, CENTERS, 0, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);
    
    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
                 ninputs, noutputs, features, normalization,
                 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
               cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
    class_data_priors_init(data_model, priors_list, priors_npatterns);

    /* write normalization data */
    if(param_fname != NULL){
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of cknn model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    }
    
    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);
    cknn_init(model, 0, features);

    /* train */
    class_data_start_cv_fold(data_model, fold, TRUE);
    n_trials = class_data_get_npatterns(data_model);

    /* read first k entries, storing them */
    for(n_read = 0; n_read < k && n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        cknn_store_inputs(model, inputs, desired_class);
    }
    /* verror for training is 0 */
    report_init(err_part_model, map_flag, ninputs, CENTERS,
            0, post_sum_to_one);
    report_epoch_header(err_part_model, stdout, n_trials);
    
    for(epoch = 0; epoch < nepochs; epoch++) {
        for(; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        
        class = cknn_inc_ctrain(model, inputs, desired_class);
        report_class_trial(err_part_model, error_file, inputs,
                   outputs, desired_class, class,
                   class_data_get_input_num(data_model));
        }
        accum_trials = *(int *)cknn_param(model, "accum_trials");
        report_centers_epoch(err_part_model, stdout, epoch, accum_trials);
        report_init(err_part_model, map_flag, ninputs, CENTERS,
            0, post_sum_to_one);
        
        /* because the first k of the first epoch are stored,
           initialize data_model and n_read here, before the next epoch */
        class_data_epoch_init(data_model);
        n_read=0;
    }
    
    class_data_end_cv_fold(data_model);
    
    /* print out cknn model */
    if(param_fname != NULL){
        fprintf(param_file, "cknn model %d\n", fold);
        cknn_write_model(model, param_file);
    }

    /* reinitialize reporting for testing */
    report_init(err_part_model, map_flag, ninputs, CENTERS,
            0, post_sum_to_one);

    /* test */
    class_data_start_cv_fold(data_model, fold, FALSE);
    n_trials = class_data_get_npatterns(data_model);
    for (n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        class = cknn_classify(model, inputs);
        trial_num = class_data_get_input_num(data_model);
        report_class_trial(err_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
        report_class_trial(err_part_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
    }
    /* report on this fold */
    report_overall(err_part_model, stdout, error_fname, labels);
    class_data_end_cv_fold(data_model);
    } /* end fold */

    /********************** free space and close files *********************/
    class_data_exit(data_model);
    
    /* close parameter file */
    if(param_fname != NULL){
    nn_fclose(param_file, tmp_path, "for write of cknn model file");
    if (verbosity > 0)
      printf("\nFinished -- cknn model saved in \"%s\"\n",tmp_path);
    }

    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }
    
    cknn_free(model);
    norm_free();

    return (percent_error);
}
