/***---------------------------------------------------------------------------
* NAME:     cknn_pack.c
* PURPOSE:  Functions specific to cknn.c
*
* SEE ALSO: cknn.c cknn_subs.c
* GLOBALS:
*
* DESCRIPTION:
*   Functions in this file
*       cknn_build()    Allocate storage for model parameters.
*       cknn_init()     Initialize model parameters.
*       cknn_free()     Deallocates model storage
*       cknn_read_model()   Initializes a cknn model from file
*       cknn_write_model()  Writes a cknn model to output file
*       cknn_param()    Returns pointers to elements in model
*       cknn_inc_ctrain()   does training on a vector
*       cknn_classify() Implements cknn to return an integer class
*       cknn_store_inputs() stores a vector in training set
*       cknn_copy_exemplar() returns class of exemplar copied
*
* ALGORITHM:
*   Training: Training vectors are first normalized and randomized, if
*       requested.  The first k (using the new random order) are taken
*       as a core set of vectors.  For each of the remaining training
*       vectors the vector is classified as though we were doing cknn
*       testing.  If the vector is incorrectly classified, the vector
*       is included in the exemplar set.  If correctly classified, it
*       is ignored.
*
*   Testing:  For each test vector, the distance to each training vector
*       is calculated in order to find the K closest training vectors.
*       The majority class of those training vectors is returned as the
*       class of the test vector
*       
* REFERENCE: Kenney Ng, "A Comparative Study of the Practical Characteristics
*   of Neural Network and Conventional Pattern Classifiers", May 1990
*   Master's Thesis for the MIT Department of Electrical Engineering and
*   Computer Science
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/cknn/cknn_pack.c,v 11.3 2001/08/23 20:11:28 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/cknn/cknn_pack.c,v $
* CREATED: 7-Mar-91 lck                             MIT LINCOLN LABORATORY
-------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define cknn_pack_h_internal_types
#include "cknn_pack.h"

#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))
#define SPACE 1024

                /* Internal Globals */
static CKNN_PARAMS **cknnpp;    /* array of parameters for each model */
static int max_models = 0;  /* current length of cknnpp */


/*****************************************************************************
* NAME: cknn_build(ninputs, nclasses, npatterns, k)
*   ninputs - number of dimensions in training vectors
*   nclasses    - number of classes the data falls into
*   npatterns   - number of vectors to be stored
*   k       - number of nearest neighbors used in classification
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*   cknn_build(2,10,338,5)
*       build a model for a 5 nearest neighbor classifier with
*       338 stored data points.
*       The data has 2 dimensions and falls into 10 classes
*
******************************************************************************/
int
cknn_build(int ninputs, int nclasses, int npatterns, int k)
{
    register int new_model;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    
    new_model = model_get_number((void ***)&cknnpp,
                 &max_models, sizeof(CKNN_PARAMS));
    ckpp = cknnpp[new_model];

    /* set fields which are array sizes */
    ckpp->ninputs = ninputs;
    ckpp->nclasses = nclasses;
    ckpp->max_exemplars = npatterns - 1;
    if(k > ninputs) k = ninputs;
    ckpp->k = k;

    /* allocate space for vectors */
    ckpp->features = allocate_ivector(ninputs+1);
    ckpp->outputs = allocate_vector(nclasses);
    ckpp->classes = allocate_ivector(npatterns);
    ckpp->exemplars = allocate_matrix(npatterns, ninputs);
    ckpp->class_vote = allocate_ivector(nclasses);
    ckpp->near_dist = allocate_vector(npatterns);
    ckpp->near_index = allocate_ivector(npatterns);

    return new_model;
}
/*****************************************************************************
* NAME: cknn_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
******************************************************************************/
void
cknn_init(int model, int accum_trials, int *nn_features)
{
    int             i;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    ckpp = cknnpp[model];
    ckpp->accum_trials = accum_trials;

    /* copy feature data into feature list */
    for(i = 0; i <= nn_features[0]; i++)
    ckpp->features[i] = nn_features[i];

}
/*****************************************************************************
* NAME: cknn_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
******************************************************************************/
void
cknn_free(int model)
{
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    ckpp = cknnpp[model];
        /* free vectors and matrices */
    deallocate_ivector(ckpp->classes);
    deallocate_ivector(ckpp->features);
    deallocate_vector(ckpp->outputs);
    deallocate_matrix(ckpp->exemplars);
    deallocate_ivector(ckpp->class_vote);
    deallocate_vector(ckpp->near_dist);
    deallocate_ivector(ckpp->near_index);
        /* free structure itself */
    nn_free((char *) ckpp, "for cknn parameter structure");
    cknnpp[model] = NULL;
}


/*****************************************************************************
* NAME: cknn_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*   allocates space for and loads data into vectors and matricies
*
*   On error from param_getblk, returns the error value
*   (there is already an error check on the allocation, so we don't
*   have to do that here)
*
* EXAMPLE:
******************************************************************************/
int 
cknn_read_model(FILE *fp, int *accum_trials)
{
    int             model;
    int             rv, k, ninputs, nclasses;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    /* read in scalars from model file */
    param_setvarptr("k", &k, paramtab, TABSIZE);
    param_setvarptr("accum_trials", accum_trials, paramtab, TABSIZE);
    param_setvarptr("ninputs", &ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &nclasses, paramtab, TABSIZE);

    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    /* allocate space for model and vectors */
    model = cknn_build(ninputs, nclasses, *accum_trials, k);
    ckpp = cknnpp[model];

    /* read in vectors from model file */
    param_setvarptr("classes", &ckpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim("classes", ckpp->max_exemplars + 1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("exemplars", (int *)&ckpp->exemplars[0][0],
            paramtab_l, TABSIZE_L);
    param_setdim("exemplars", ckpp->max_exemplars + 1, ckpp->ninputs,
         paramtab_l, TABSIZE_L);
    param_setvarptr("features", &ckpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", ckpp->ninputs+1, 0, paramtab_l, TABSIZE_L);

    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    /* set remaining parameters */
    ckpp->accum_trials = *accum_trials;

    return (model);
}

/*****************************************************************************
* NAME: cknn_write_model()
* DESCRIPTION:
*   writes parameters and exemplars for a model to a file
*   in ASCII format
*
* EXAMPLE:
*
******************************************************************************/
int
cknn_write_model(int model, FILE *fp)
{
    int             rv;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    ckpp = cknnpp[model];

    /* save scalars */
    param_setvarptr("k", &ckpp->k, paramtab, TABSIZE);
    param_setvarptr("accum_trials", &ckpp->accum_trials, paramtab, TABSIZE);
    param_setvarptr("ninputs", &ckpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &ckpp->nclasses, paramtab, TABSIZE);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    /* save vectors and matrices */
    param_setvarptr("classes", &ckpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim("classes", ckpp->accum_trials, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("exemplars", (int *)&ckpp->exemplars[0][0],
            paramtab_l, TABSIZE_L);
    param_setdim("exemplars", ckpp->accum_trials, ckpp->ninputs,
         paramtab_l, TABSIZE_L);
    param_setvarptr("features", &ckpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", ckpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    return (0);
}

/************************************************************************
* NAME: cknn_param(model, param_name)
* DESCRIPTION:
*   Takes model number and parameter name.
*   Returns a pointer to the named parameter.
*
*************************************************************************/

void           *
cknn_param(int model, char *param_name)
{
    void           *param_address;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    ckpp = cknnpp[model];
    if (PEQ("k")) param_address = &ckpp->k;
    else if (PEQ("accum_trials")) param_address = &ckpp->accum_trials;
    else if (PEQ("exemplars")) param_address = &ckpp->exemplars;
    else if (PEQ("classes")) param_address = &ckpp->classes;
    else if (PEQ("features")) param_address = &ckpp->features;
    else if (PEQ("ninputs")) param_address = &ckpp->ninputs;
    else if (PEQ("nclasses")) param_address = &ckpp->nclasses;
    else if (PEQ("outputs")) param_address = &ckpp->outputs;
    else {
    warning("cknn_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}               /* END cknn_param() */

/*****************************************************************************
* NAME: cknn_store_inputs()
* DESCRIPTION:
*       assigns a class to the input vector.
*   If the class assigned is wrong, the vector is added to the
*       exemplar set
*   returns the class selected
* EXAMPLE:
*
******************************************************************************/
void 
cknn_store_inputs(int model, float *inputs, int desired_class)
{
    register int    i, newlength;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    ckpp = cknnpp[model];

    /* check to see if we have the room in exemplars for new vector */
    if (ckpp->accum_trials > ckpp->max_exemplars) {
        /* figure out how many vectors we can add to exemplar */
    ckpp->max_exemplars++;  /* now the number of vectors in exemplars */
    newlength = ckpp->max_exemplars +
        SPACE / (sizeof(float) * ckpp->ninputs);
    ckpp->exemplars = reallocate_matrix(ckpp->exemplars,
            ckpp->max_exemplars, ckpp->ninputs, newlength);
    ckpp->classes = reallocate_ivector(ckpp->classes,
            ckpp->max_exemplars, newlength);
    ckpp->near_index = reallocate_ivector(ckpp->near_index,
            ckpp->max_exemplars, newlength);
    ckpp->near_dist = reallocate_vector(ckpp->near_dist,
            ckpp->max_exemplars, newlength);
    ckpp->max_exemplars = newlength - 1;
    }
    /* store input in exemplars, class in classes, and increment counter */
    for (i = 0; i < ckpp->ninputs; i++) {
    ckpp->exemplars[ckpp->accum_trials][i] = inputs[i];
    }
    ckpp->classes[ckpp->accum_trials] = desired_class;
    ckpp->accum_trials++;

}
/*****************************************************************************
* NAME: cknn_inc_ctrain()
* DESCRIPTION:
*       assigns a class to the input vector.
*   If the class assigned is wrong, the vector is added to the
*       exemplar set
*   returns the class selected
* EXAMPLE:
*
******************************************************************************/
int 
cknn_inc_ctrain(int model, float *inputs, int desired_class)
{
    register int    class;

    class = cknn_classify(model, inputs);

    /* if we got it right, ignore it, else, include it */
    if (class != desired_class)
    cknn_store_inputs(model, inputs, desired_class);
    return (class);

}


/*****************************************************************************
* NAME: cknn_classify()
* DESCRIPTION:
*       Implements the cknn to return an integer class.
*
* EXAMPLE:
*
******************************************************************************/
int 
cknn_classify(int model, float *inputs)
{
    int             winner;
    int             i;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    ckpp = cknnpp[model];

    find_k_nearest_neighbors(inputs, ckpp->exemplars, ckpp->near_dist,
                 ckpp->near_index, ckpp->ninputs,
                 ckpp->accum_trials, ckpp->k);

    /* make a histogram of the classes of the k nearest neighbors */
    for (i = 0; i < ckpp->nclasses; i++) ckpp->class_vote[i] = 0;
    for (i = 0; i < ckpp->k; i++) {
    if(ckpp->near_index[i] < 0 ||
       ckpp->near_index[i] >= ckpp->accum_trials)
      continue;
    ckpp->class_vote[ckpp->classes[ckpp->near_index[i]]]++;
    }
    /* find most common class */
    winner = int_max_pick(ckpp->class_vote, ckpp->nclasses);

    /* set outputs to the normalized number of patterns in each class */
    for(i = 0; i < ckpp->nclasses; i++)
      ckpp->outputs[i] = (float)ckpp->class_vote[i]/(float)ckpp->k;

    return (winner);
}


/************************************************************************
* NAME: cknn_copy_exemplar(model, pattern, inputs)
* DESCRIPTION:
*   copies exemplar[pattern] into inputs
*   returns classes[pattern]
*
*************************************************************************/
int 
cknn_copy_exemplar(int model, int pattern, float *inputs)
{
    int             i;
    CKNN_PARAMS *ckpp;  /* temporary pointer to cknnpp[] */

    ckpp = cknnpp[model];
    for (i = 0; i < ckpp->ninputs; i++) {
    inputs[i] = ckpp->exemplars[pattern][i];
    }
    return (ckpp->classes[pattern]);
}
