/***-------------------------------------------------------------------------
* NAME:     committee.c
* PURPOSE:  cross validation driver for committee classifier
*
* SEE ALSO:     committee_pack.c
*
* FUNCTIONS:
*   committee   driver for committee classifier
*       handles model creation, training and testing of data
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/committee/committee_cv.c,v 11.4 2001/08/23 20:11:31 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/committee/committee_cv.c,v $
* CREATED: 18-JUN-91 lck                             MIT LINCOLN LABORATORY
---------------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "committee.h"
#include "committee_pack.h"

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

static char prog_name[MAXSTR];  /* static variables */

float
committee_cross_valid(int db_type, int nraw, int ninputs, int *features,
              int noutputs, int npatterns, int verbosity, int verror,
              int cv_nfolds, int cv_nsplits, int use_cv_splits,
              int *cv_splits, char **cv_train_mask,
              char **cv_test_mask, char *data_path, char *data_fname,
              char *exp_path, char *error_fname, char **labels,
              int cv_random, int log_outputs, int class_type)
{
    int     model, err_model, err_part_model, normalization=0, data_model;
    int     desired_class, class;
    int     n_read, fold, n_trials, trial_num;
    float   *inputs, *outputs;
    float   percent_error;
    FILE    *error_file;
    int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());

    /* create the committee model */
    model = committee_build(ninputs, noutputs, class_type);
    /* there is no normalization on committee data files, so don't read
       the normalization files, and set normalization to NONE */
    norm_setup_none(nraw, noutputs);
    normalization = NONE;

    /* use features from the model */
    ninputs = *(int *) committee_param(model, "ninputs");
    outputs = *(float **) committee_param(model, "outputs");

    if (verror > 0) {   /* open file for classification errors */
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    }
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, db_type, ninputs, SQ_ERR,
         log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);
    init_cost(SQ_ERR, 1.);

    /* set up for reading data */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
                 ninputs, noutputs, features, normalization,
                 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
               cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    /* don't set random flags or the priors adjustment because we don't do
       training */

    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);

    /* we would save the model here, if we had to */

    report_init(err_part_model, db_type, ninputs, SQ_ERR,
            0, post_sum_to_one);
    
    /* prepare for testing fold */
    class_data_start_cv_fold(data_model, fold, FALSE);
    n_trials = class_data_get_npatterns(data_model);
    
    for(n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        class = committee_classify(model, inputs);
        trial_num = class_data_get_input_num(data_model);

        report_class_trial(err_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
        report_class_trial(err_part_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
    }
    /* report on this fold */
    report_overall(err_part_model, stdout, error_fname, labels);
    
    class_data_end_cv_fold(data_model);
    } /* end fold */

    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }

    /* free space and close files */
    class_data_exit(data_model);
    committee_free(model);

    return(percent_error);
}
