/***---------------------------------------------------------------------------
* NAME:         committee_pack.c
* PURPOSE:      Functions specific to committee.c
*
* SEE ALSO: committee.c committee_main.c
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   committee_build     Allocate storage for model parameters.
*   committee_init      Initialize model parameters.
*   committee_free      Deallocates model storage
*   committee_read_model    Initializes model from file
*   committee_write_model   Writes model to output file
*   committee_param     Returns pointers to elements in model
*   committee_inc_ctrain    Does training on a vector
*   committee_classify  Classifies a vector using committee algortihm
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/committee/committee_pack.c,v 11.5 2001/08/23 20:11:31 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/committee/committee_pack.c,v $
* CREATED: 12-Dec-90 lck                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define committee_pack_h_internal_types
#define committee_pack_h_statics
#include "committee_pack.h"

                /* constant definitions */
#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))

#define AVG_TYPE 0
#define MED_TYPE 1
#define VOTE_TYPE 2
                /* Internal Globals */
static COMMITTEE_PARAMS **genpp;/* array of parameters for each model */
static int      max_models = 0; /* current length of genpp */

/*****************************************************************************
* NAME: committee_build(ninputs,noutputs)
*   ninputs - number of dimensions in training vectors
*   noutputs - number of classes
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
*****************************************************************************/
int
committee_build(int ninputs, int noutputs, int class_type)
{
    register int    new_model;
    COMMITTEE_PARAMS *gpp;

                /* find entry in model array for new model */
    new_model = model_get_number((void ***)&genpp,
                 &max_models, sizeof(COMMITTEE_PARAMS));
    gpp = genpp[new_model];

                    /* set fields which are array sizes */
    gpp->ninputs = ninputs;
    gpp->noutputs = noutputs;
    if(ninputs >= noutputs && noutputs != 0 && ninputs % noutputs == 0)
      gpp->nclassifiers = ninputs / noutputs;
    else
      fatal_error("ninputs (%d) must be a multiple of noutputs (%d)",
          ninputs, noutputs);

    gpp->class_type = class_type;
    if(class_type != AVG_TYPE && class_type != MED_TYPE &&
       class_type != VOTE_TYPE){
    fatal_error("%s\n%s(%d) %s(%d) %s(%d)",
            "committee_build: classifer type must be one of:",
            "Average", AVG_TYPE, "Median", MED_TYPE,
            "Vote", VOTE_TYPE);
    }

                    /* allocate space for vectors */
    gpp->outputs = allocate_vector(noutputs);
    gpp->class_cnt = allocate_vector(noutputs);
    gpp->priors_list = allocate_vector(noutputs);
    if(class_type == VOTE_TYPE){
    gpp->vote_cnt = allocate_ivector(noutputs);
    } else if(class_type == MED_TYPE){
    gpp->median_array = allocate_vector(gpp->nclassifiers);
    }

    return new_model;
}
/*****************************************************************************
* NAME: committee_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
*****************************************************************************/
void
committee_init(int model)
{
    int             i;
    COMMITTEE_PARAMS *gpp;

    gpp = genpp[model];

    for (i = 0; i < gpp->noutputs; i++)
      gpp->class_cnt[i] = 0.;
    gpp->total_npatterns = 0.;
    gpp->adjust_priors = 0;
}

/*****************************************************************************
* NAME: committee_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
*****************************************************************************/
void
committee_free(int model)
{
    COMMITTEE_PARAMS *gpp;

    gpp = genpp[model];
                    /* free vectors and matrices */
    deallocate_vector(gpp->outputs);
    deallocate_vector(gpp->class_cnt);
    deallocate_vector(gpp->priors_list);
    if(gpp->class_type == VOTE_TYPE)
      deallocate_ivector(gpp->vote_cnt);
    else if(gpp->class_type == MED_TYPE)
      deallocate_vector(gpp->median_array);

                    /* free structure itself */
    nn_free((char *) gpp, "for committee parameter structure");
    genpp[model] = NULL;
}

static void
committee_set_scalar_params(COMMITTEE_PARAMS *gpp)
{
    param_setvarptr("ninputs", &gpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &gpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("total_npatterns", (int *)&gpp->total_npatterns,
            paramtab, TABSIZE);
    param_setvarptr("class_type", &gpp->class_type, paramtab, TABSIZE);
    param_setvarptr("nclassifiers", &gpp->nclassifiers, paramtab, TABSIZE);
}

static void
committee_set_vector_params(COMMITTEE_PARAMS *gpp)
{
    param_setvarptr("class_cnt", (int *)&gpp->class_cnt[0],
            paramtab_l, TABSIZE_L);
    param_setdim("class_cnt", gpp->noutputs, 0, paramtab_l, TABSIZE_L);

}

/****************************************************************************
* NAME: committee_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*       allocates space for and loads data into vectors and matricies
*
*       On error from param_getblk, returns the error value
*       (there is already an error check on the allocation, so we don't
*       have to do that here)
*
* EXAMPLE:
****************************************************************************/
int
committee_read_model(FILE *fp)
{
    int             model;
    int             rv;
    COMMITTEE_PARAMS temp_gpp, *gpp;

    committee_set_scalar_params(&temp_gpp);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

                    /* allocate space for model and vectors */
    model = committee_build(temp_gpp.ninputs, temp_gpp.noutputs,
                temp_gpp.class_type);
    gpp = genpp[model];

    /* copy parameters from temporary structure to model */
    gpp->total_npatterns = temp_gpp.total_npatterns;
    gpp->nclassifiers = temp_gpp.nclassifiers;

    committee_set_vector_params(gpp);

    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    return (model);
}

/****************************************************************************
* NAME: committee_write_model()
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
*
* EXAMPLE:
*
****************************************************************************/
int
committee_write_model(int model, FILE *fp)
{
    int             rv;
    COMMITTEE_PARAMS *gpp;

    gpp = genpp[model];
                    /* save scalars */
    committee_set_scalar_params(gpp);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

                    /* save vectors and matrices */
    committee_set_vector_params(gpp);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    return (0);
}


/************************************************************************
* NAME: committee_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/

void *
committee_param(int model, char *param_name)
{
    void  *param_address;
    COMMITTEE_PARAMS *gpp;

    gpp = genpp[model];
    if (PEQ("ninputs"))
    param_address = &gpp->ninputs;
    else if (PEQ("noutputs"))
    param_address = &gpp->noutputs;
    else if (PEQ("outputs"))
    param_address = &gpp->outputs;
    else if (PEQ("class_cnt"))
      param_address = &gpp->class_cnt;
    else if (PEQ("total_npatterns"))
      param_address = &gpp->total_npatterns;
    else {
    warning("committee_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}


/*****************************************************************************
* NAME: committee_inc_ctrain()
* DESCRIPTION:
*       incremental training for a single input vector
* EXAMPLE:
*
*****************************************************************************/
int
committee_inc_ctrain(int model, float *inputs, int desired_class)
{
    COMMITTEE_PARAMS *gpp;

    gpp = genpp[model];
    gpp->class_cnt[desired_class] += 1.;
    gpp->total_npatterns += 1.;
    return (rand() % gpp->noutputs);
}
/*****************************************************************************
* NAME: committee_classify()
* DESCRIPTION:
*       Implements the committee classifier to return an integer class.
*
* EXAMPLE:
*
*****************************************************************************/
int
committee_classify(int model, float *inputs)
{
    COMMITTEE_PARAMS *gpp;
    int i, j, k, l, class;
    float total;

    gpp = genpp[model];
    /*    if(gpp->adjust_priors) use priors list to modify outputs */
    /* can only do that if outputs are continuous */
    switch(gpp->class_type){
      case AVG_TYPE:
    for(i = 0; i < gpp->noutputs; i++){
        total = 0.;
        for(j = i; j < gpp->ninputs; j += gpp->noutputs)
          total += inputs[j];
        gpp->outputs[i] = total/(float)gpp->nclassifiers;
    }
    break;
      case MED_TYPE:
    for(i = 0; i < gpp->noutputs; i++){
        for(k = 0; k < gpp->nclassifiers; k++){
        gpp->median_array[k] = MAXFLOAT;
        }
        for(j = i; j < gpp->ninputs; j+= gpp->noutputs){
        total = inputs[j];

        /* do an N-squared sort because N is probably small */
        for(k = 0; k < gpp->nclassifiers; k++)
          if(total < gpp->median_array[k]) break;
        for(l = gpp->nclassifiers-1; l > k; l--)
          gpp->median_array[l] = gpp->median_array[l-1];
        if(k < gpp->nclassifiers){
          gpp->median_array[k] = total;
        } else {
          gpp->median_array[gpp->nclassifiers-1] = total;
        }
        }
        /* pick the output in the middle */
        gpp->outputs[i] = gpp->median_array[gpp->nclassifiers/2];
    }
    break;
      case VOTE_TYPE:
    /* find the class of each classifier, add it to the votes */
    for(j = 0; j < gpp->noutputs; j++)
      gpp->vote_cnt[j] = 0;
    for(k = 0; k < gpp->nclassifiers; k++){
        class = max_pick(&inputs[k*gpp->noutputs], gpp->noutputs);
        gpp->vote_cnt[class]++;
    }
    for(j = 0; j < gpp->noutputs; j++)
      gpp->outputs[j] = (float)gpp->vote_cnt[j]/(float)gpp->nclassifiers;
    break;
    }
    return(max_pick(gpp->outputs, gpp->noutputs));
}
