/***--------------------------------------------------------------------------
* NAME:         em_clus_pack.c
* PURPOSE:      Functions specific to em_clus.c
*
* SEE ALSO: em_clus.c em_clus_main.c
*
* USAGE:
*   model = em_clus_build(...)
*   em_clus_init(model,...)
*       report_stop_flag=0;
*   while (report_stop_flag<=0)
*       {
*       for(i=0; i<all_patterns; i++)
*           em_clus_batch_train(model, input[i]);
*           report_stop_flag = em_clus_batch_update(model,
*           &iterations, &actual_ncenters, &ave_squared_error);
*       }
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   em_clus_build       Allocate storage for model parameters.
*   em_clus_init        Initialize model parameters.
*   em_clus_free        Deallocates model storage
*   em_clus_read_model  Initializes model from file
*   em_clus_write_model Writes model to output file
*   em_clus_param       Returns pointers to elements in model
*   em_clus_batch_train find nearest cluster center to one input
*               pattern and update that clusters counts
*   em_clus_batch_update    calculate new cluster centers from counts,
*               check for convergence, split centers to
*               double the number of centers after convergence
*               with a given number of centers or terminate
*               by setting the output to the number of centers
*               used (a power of 2)
*   em_clus_cluster     returns the cluster an input pattern falls in
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/em_clus/em_clus_pack.c,v 11.3 2001/08/23 20:11:35 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/em_clus/em_clus_pack.c,v $
* CREATED: 16-Dec-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#define em_clus_pack_h_internal_types
#define em_clus_pack_h_statics
#include "em_clus_pack.h"

                /* constant definitions */
#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))
#define SMALLEST 1.e-30

                /* Internal Globals */
static EM_CLUS_PARAMS **genpp;  /* array of parameters for each model */
static int      max_models = 0; /* current length of genpp */

/*****************************************************************************
* NAME: em_clus_build(ninputs, ncenters)
*   ninputs - number of dimensions in training vectors
*   ncenters - number of cluster centers
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
******************************************************************************/
int
em_clus_build(int ninputs, int ncenters)
{
    register int    new_model;
    EM_CLUS_PARAMS *gpp;
    int             i;

                /* find entry in model array for new model */
    new_model = model_get_number((void ***)&genpp,
                 &max_models, sizeof(EM_CLUS_PARAMS));
    gpp = genpp[new_model];

    for (i = 1; i < ncenters;)  /* determine maximum number of centers */
    i *= 2;         /* (must be power of 2 ) */
    gpp->max_centers = i;

                    /* set fields which are array sizes */
    gpp->ninputs = ninputs;
    gpp->ncenters = ncenters;
                    /* allocate space */
    gpp->features = allocate_ivector (ninputs+1);
    gpp->outputs = allocate_vector (gpp->max_centers);

    gpp->weighted_gauss_output = allocate_vector(gpp->max_centers);
    gpp->weight = allocate_vector(gpp->max_centers);
    gpp->new_weight = allocate_vector(gpp->max_centers);
    gpp->log_weight = allocate_vector(gpp->max_centers);
    gpp->means = allocate_matrix (gpp->max_centers, ninputs);
    gpp->new_means = allocate_matrix (gpp->max_centers, ninputs);
    gpp->diag_var = allocate_matrix (gpp->max_centers, ninputs);
    gpp->diag_new_var = allocate_matrix (gpp->max_centers, ninputs);
    gpp->diag_inv_var = allocate_matrix (gpp->max_centers, ninputs);
    gpp->log_det_covar = allocate_vector(gpp->max_centers);

    return new_model;
}

/*****************************************************************************
* NAME: em_clus_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
******************************************************************************/
void
em_clus_init (int model, int *nn_features, float variance_scale,
          float minimum_variance,
          float stop_percentage,/* stop iteration if rms change < this */
          float split_percentage,/* +/- sdev this percentage on split */
          int max_iteration,/* max iterations after any split */
          int reduce_step, int add_random_offset)
{
    int             i;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];
                    /* copy feature data into feature list */
    for (i = 0; i <= nn_features[0]; i++)
    gpp->features[i] = nn_features[i];

                    /* set global variables */
    gpp->variance_scale = variance_scale;
    gpp->minimum_variance = minimum_variance;
    gpp->stop_percentage = stop_percentage;
    gpp->split_percentage = split_percentage;
    gpp->max_iteration = max_iteration;
    gpp->reduce_step = reduce_step;
    gpp->add_random_offset = add_random_offset;

                    /* special startup initialization */
    gpp->ave_logl = -(MAXFLOAT);
    gpp->splits = 0;            /* number of splits so far */
    gpp->total_iteration = 0;   /* total iterations so far */
    gpp->current_iteration = 0; /* number of iterations for this split */
    if(gpp->ncenters != 0)
    gpp->current_ncenters = 1;  /* current number of cluster centers */
    else gpp->current_ncenters = 0;
    gpp->stop_flag = 0;

            /* common initialization before all iterations*/
    em_clus_local_init(model);

}


/*****************************************************************************
* NAME: em_clus_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
******************************************************************************/
void
em_clus_free(int model)
{
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];
                        /* free vectors and matrices */
    deallocate_ivector (gpp->features);
    deallocate_vector (gpp->outputs);

    deallocate_vector (gpp->weighted_gauss_output);
    deallocate_vector (gpp->new_weight);
    deallocate_vector (gpp->weight);
    deallocate_vector (gpp->log_weight);
    deallocate_matrix (gpp->means);
    deallocate_matrix (gpp->new_means);
    deallocate_matrix (gpp->diag_var);
    deallocate_matrix (gpp->diag_new_var);
    deallocate_matrix (gpp->diag_inv_var);
    deallocate_vector (gpp->log_det_covar);

                        /* free structure itself */
    nn_free((char *) gpp, "for em_clus parameter structure");
    genpp[model] = NULL;
}


/*****************************************************************************
* NAME: em_clus_read_model()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*   (currently used only by em_clus_plot_bound)
*
* EXAMPLE:
*
******************************************************************************/
int
em_clus_read_model(FILE *fp)
{
    int clus_model, model, noutputs, i, j;
    float **pcenter_part, **pvariance_part;
    EM_CLUS_PARAMS *epp;
    int ninputs, ncenters, *features;

    clus_model = cluster_read_model(fp);
    if(clus_model < 0) return(clus_model);
    cluster_merge_centers(clus_model);
    ninputs = *(int *)cluster_param(clus_model, "ninputs", 0);
    noutputs = *(int *)cluster_param(clus_model, "noutputs", 0);
    ncenters = *(int *)cluster_param(clus_model, "ncenters", 0);

    model = em_clus_build(ninputs, ncenters);
    epp = genpp[model];

    features = *(int **)cluster_param(clus_model, "features", 0);
    pcenter_part = *(float ***)cluster_param(model, "pcenter", 0);
    pvariance_part = *(float ***)cluster_param(model, "pvariance", 0);

    /* copy vectors and matrices into epp */
    for(i = 0; i <= features[0]; i++){
    epp->features[i] = features[i];
    }
    for(j = 0; j < ncenters; j++){
    epp->log_det_covar[j] = 0.;
    for(i = 0; i < ninputs; i++){
        epp->means[j][i] = pcenter_part[j][i];
        epp->diag_var[j][i] = pvariance_part[j][i];
        epp->diag_inv_var[j][i] = 1./epp->diag_var[j][i];
        epp->log_det_covar[j] +=
          (float)log10((double)epp->diag_var[j][i]);
    }
    }

    cluster_free(clus_model);
    return(model);
}

/*****************************************************************************
* NAME: em_clus_write_model()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int
em_clus_write_model(int model, int noutputs, FILE *fp)
{
    int rv, cluster_by_class = 0, *ncenter_list = NULL;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];

    rv = cluster_write_paramfile(fp, gpp->ninputs, gpp->ncenters,
                 noutputs, &(gpp->means), &(gpp->diag_var),
                 gpp->features, cluster_by_class,
                 ncenter_list);
    return(rv);
}

/*****************************************************************************
* NAME: em_clus_write_model_list()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int
em_clus_write_model_list(int *model_list, int noutputs, FILE *fp)
{
    int rv, cluster_by_class = 1, i, *ncenter_list, ncenters;
    float ***center_list, ***variance_list;
    EM_CLUS_PARAMS *gpp;

    ncenter_list = allocate_ivector(noutputs);
    center_list = (float ***)allocate_pvector(noutputs);
    variance_list = (float ***)allocate_pvector(noutputs);

    for(ncenters = 0, i = 0; i < noutputs; i++){
    gpp = genpp[model_list[i]];
    ncenter_list[i] = gpp->ncenters;
    ncenters += gpp->ncenters;
    center_list[i] = gpp->means;
    variance_list[i] = gpp->diag_var;
    }

    rv = cluster_write_paramfile(fp, gpp->ninputs, gpp->ncenters,
                 noutputs, center_list, variance_list,
                 gpp->features, cluster_by_class,
                 ncenter_list);
    deallocate_ivector(ncenter_list);
    deallocate_pvector((void **)center_list);
    deallocate_pvector((void **)variance_list);

    return(rv);
}

/************************************************************************
* NAME: em_clus_param(model, param_name)
* DESCRIPTION:
*       Takes parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/
void           *
em_clus_param(int model, char *param_name)
{
    void           *param_address;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];
    if (PEQ("features"))
    param_address = &gpp->features;
    else if (PEQ("ninputs"))
    param_address = &gpp->ninputs;
    else if (PEQ("ncenters"))
    param_address = &gpp->ncenters;
    else if (PEQ("current_ncenters"))
      param_address = &gpp->current_ncenters;
    else if (PEQ("outputs"))
    param_address = &gpp->outputs;
    else if (PEQ("means"))
    param_address = &gpp->means;
    else if (PEQ("diag_var"))
    param_address = &gpp->diag_var;
    else if (PEQ("inv_var"))
      param_address = &gpp->diag_inv_var;
    else if (PEQ("log_det_covar"))
      param_address = &gpp->log_det_covar;
    else {
    warning("em_clus_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}


/*****************************************************************************
* NAME:     em_clus_batch_train()
* DESCRIPTION:
*           provide another vector for em_clus training
* EXAMPLE:
*
******************************************************************************/
void
em_clus_batch_train(int model, float *input)
{
    int i, g;
    float  log_total_pdf, log_output;
    float  tau;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];
                        /* find centroid of all data  */
    if (gpp->total_iteration == 0) {
    for (i = 0; i < gpp->ninputs; i++) {
        gpp->new_means[0][i] += input[i];
        gpp->diag_new_var[0][i] += input[i] * input[i];
    }
    gpp->total_vectors++;
    gpp->new_weight[0]++;   /* this way, we can use the same code
                   in em_clus_batch_update as for
                   later iterations */
    } else {
    log_total_pdf = -(MAXFLOAT);
    /* find the output of each gaussian */
    for(g = 0; g < gpp->current_ncenters; g++){
        log_output =
          get_log_gauss_output_diag(gpp->ninputs, input,
                    gpp->means[g],
                    gpp->diag_inv_var[g],
                    gpp->log_det_covar[g]);
        gpp->weighted_gauss_output[g] =
          gpp->log_weight[g] + log_output;
        log_total_pdf =
          logadd(log_total_pdf, gpp->weighted_gauss_output[g]);
    }
    gpp->total_logl += log_total_pdf;

    for (g = 0; g < gpp->current_ncenters; g++) {
        tau = nn_exp10(gpp->weighted_gauss_output[g] - log_total_pdf);
        gpp->new_weight[g] += tau;

        for(i = 0; i < gpp->ninputs; i++){
        gpp->new_means[g][i] += tau * input[i];
        gpp->diag_new_var[g][i] += input[i]*input[i]*tau;
        }
    }
    gpp->total_vectors++;
    }
}

/*****************************************************************************
* NAME:     em_clus_batch_update()
* DESCRIPTION:
*           compute new centers, split if necessary, determine if done
*               Return 0 if not done, and number of centers if done.
* EXAMPLE:
*
*****************************************************************************/
int 
em_clus_batch_update(int model, int *report_iterations,
             int *report_actual_ncenters, float *report_ave_logl,
             int *report_nvectors)
{
    float           percent_change;
    int             i, g, percent_flag, diverging_flag;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];
                /* The iteration is done now label it. First is 1 */
    gpp->current_iteration++;
    gpp->total_iteration++;

    /* compute new weights */
    for (g = 0; g < gpp->current_ncenters; g++) {
    gpp->weight[g] = gpp->new_weight[g]/gpp->total_vectors;
    if(gpp->weight[g] > SMALLEST)
      gpp->log_weight[g] =
        (float)log10((double)gpp->weight[g]);
    else
      gpp->log_weight[g] = 0.;

    gpp->log_det_covar[g] = 0.;
    for (i = 0; i < gpp->ninputs; i++) {
        gpp->means[g][i] = gpp->new_means[g][i] / gpp->new_weight[g];

        gpp->diag_var[g][i] = gpp->diag_new_var[g][i]/gpp->new_weight[g];
        gpp->diag_var[g][i] -= gpp->means[g][i]*gpp->means[g][i];
        if(gpp->diag_var[g][i] < gpp->minimum_variance)
          gpp->diag_var[g][i] = gpp->minimum_variance;
        if(gpp->diag_var[g][i] > SMALLEST)
          gpp->log_det_covar[g] +=
        (float)log10((double)gpp->diag_var[g][i]);
        gpp->diag_inv_var[g][i] = 1./gpp->diag_var[g][i];
    }
    }
                /* compute average log likelihood */
    if (gpp->total_vectors > 0)
    gpp->ave_logl = gpp->total_logl / gpp->total_vectors;
    else
    fatal_error("%s %s",
      "em_clus_batch_update: calling this routine without first calling",
            "em_clus_batch_train");

    /* set return values to current iteration information */
    *report_iterations = gpp->total_iteration;
    *report_actual_ncenters = gpp->current_ncenters;
    *report_ave_logl = gpp->ave_logl;
    *report_nvectors = gpp->total_vectors;

    if(gpp->ncenters == 0){ /* we don't want any centers */
    return 1; /* they look for a non-zero return for stopping */
    }
    if (gpp->total_iteration == 1) {
    /* first overall iteration force a split or quit */
    if(gpp->ncenters > 1){
        em_clus_split(model);
        gpp->ave_logl = -(MAXFLOAT);
    }else{
        gpp->current_iteration = 0;
    }
    } else if (gpp->current_iteration == 1) {
    /* first iteration after a split, force another iteration */
    } else {
    percent_flag = diverging_flag = 0;
    /* more than one iteration without a split */
    percent_change =
      100.0 * (gpp->ave_logl -  gpp->last_ave_logl) /
        (float)fabs((double)gpp->last_ave_logl);
    if((float)fabs((double)percent_change) <= gpp->stop_percentage)
      percent_flag = 1;

    /* error, squared error is increasing */
    if (percent_change < 0.0){
        warning("%s\n%s %g\n%s %g\n%s %g",
            "em_clus_batch_update diverging",
            "Last ave log likelihood", gpp->last_ave_logl,
            "Current log likelihood", gpp->ave_logl,
            "Percentage change", percent_change);
/*
        diverging_flag = 1;
*/
    }

    if(gpp->current_ncenters < gpp->ncenters){
        if (gpp->current_iteration >= gpp->max_iteration ||
        percent_flag || diverging_flag){
        em_clus_split(model);
        }
    } else if(gpp->current_ncenters > gpp->ncenters){
        if (gpp->current_iteration >= gpp->max_iteration ||
        percent_flag || diverging_flag){
        if((gpp->current_ncenters - gpp->ncenters) >
           gpp->reduce_step){
            i = em_clus_reduce_ncenter(model,
               (gpp->current_ncenters - gpp->reduce_step));
        } else {
            i = em_clus_reduce_ncenter(model, gpp->ncenters);
        }
        }
    } else {
        if (gpp->current_iteration > gpp->max_iteration ||
        percent_flag || diverging_flag){
        gpp->stop_flag = gpp->current_ncenters;
        }
    }
    }

                        /* all done, clean up  */
    if (gpp->stop_flag > 0) {
    return (gpp->current_ncenters);
                        /* get ready for another iteration  */
    } else {
    em_clus_local_init(model);
    return (0);
    }
}


/*****************************************************************************
* NAME:     em_clus_split()
* DESCRIPTION:
*           split centers to create twice as many
* EXAMPLE:
*
******************************************************************************/
static void 
em_clus_split(int model)
{
    int             i, g, split_g;
    float           mean, sdev, delta;
    static int split_count = 0;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];
    split_count++;

    for (g = 0; g < gpp->current_ncenters; g++){
    split_g = g + gpp->current_ncenters;
    gpp->log_det_covar[g] = gpp->log_det_covar[split_g] = 0.;
    for (i = 0; i < gpp->ninputs; i++) {
        mean = gpp->means[g][i];
        sdev = (float) sqrt((double) gpp->diag_var[g][i]);
        /* randomness added to break symetry when two centers have
           exactly the same distance to their nearest neighbor */
        if(gpp->add_random_offset){
        delta = gpp->split_percentage * sdev *
          ((9.5 + random_uniform())/1000.);
        } else {
        delta = gpp->split_percentage * sdev /100.0;
        }

        gpp->means[split_g][i] = mean + delta;
        gpp->means[g][i] = mean - delta;

        gpp->diag_var[g][i] *= gpp->variance_scale;
        gpp->diag_var[split_g][i] = gpp->diag_var[g][i];
        if(gpp->diag_var[g][i] > SMALLEST){
        gpp->log_det_covar[g] +=
          (float)log10((double)gpp->diag_var[g][i]);
        }
        gpp->diag_inv_var[split_g][i] = gpp->diag_inv_var[g][i] =
          1./gpp->diag_var[g][i];
    }
    gpp->log_det_covar[split_g] = gpp->log_det_covar[g];

    gpp->weight[g] = gpp->weight[split_g] = gpp->weight[g]/2.;
    if(gpp->weight[g] > SMALLEST)
      gpp->log_weight[split_g] = gpp->log_weight[g] =
        (float)log10((double)gpp->weight[g]);
    else
      gpp->log_weight[split_g] = gpp->log_weight[g] = 0.;

    }
    gpp->current_ncenters *= 2;
    gpp->current_iteration = 0;
    gpp->last_ave_logl = gpp->ave_logl;
    gpp->splits++;
}


/*****************************************************************************
* NAME:     em_clus_local_init()
* DESCRIPTION:
*           initialize variables at beginning of every iteration
* EXAMPLE:
*
******************************************************************************/
static void 
em_clus_local_init(int model)
{
    int             i, g;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];
    gpp->total_vectors = 0;     /* total number of vectors */
    gpp->total_logl = 0.0;  /* total vq error */
    gpp->last_ave_logl = gpp->ave_logl;
                        /* initialize sums and counters  */
    for (g = 0; g < gpp->current_ncenters; g++) {
    gpp->new_weight[g] = 0.;
    for (i = 0; i < gpp->ninputs; i++) {
        gpp->diag_new_var[g][i] = 0.0;
        gpp->new_means[g][i] = 0.0;
    }
    }
}

/*****************************************************************************
 * NAME:    em_clus_remove_ncenter()
 * DESCRIPTION:
 *      remove the designated center from the model, give its
 *      counter to its nearest center
 * EXAMPLE:
 *
 ****************************************************************************/
void
em_clus_remove_ncenter(int model, int center, int nearest_center)
{
    int             i, g;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];

    gpp->weight[nearest_center] += gpp->weight[center];
    if(gpp->weight[nearest_center] > SMALLEST)
      gpp->log_weight[nearest_center] =
    (float)log10((double)gpp->weight[nearest_center]);
    else
      gpp->log_weight[nearest_center] = 0.;

    for (g = center; g < (gpp->current_ncenters - 1); g++) {
    gpp->weight[g] = gpp->weight[g+1];
    gpp->log_weight[g] = gpp->log_weight[g+1];
    for (i = 0; i < gpp->ninputs; i++) {
        gpp->means[g][i] = gpp->means[g+1][i];
        gpp->diag_var[g][i] = gpp->diag_var[g+1][i];
        gpp->diag_inv_var[g][i] = gpp->diag_inv_var[g+1][i];
    }
    gpp->log_det_covar[g] = gpp->log_det_covar[g+1];
    }
    gpp->current_ncenters--;
    gpp->current_iteration = 0;
}

/*****************************************************************************
* NAME:     em_clus_reduce_ncenter()
* DESCRIPTION:
*           calculate the distance between centers, depending on the
*               center and the number of points in the center, determine
*               the center that can be removed with least increase to overall
*               error.
*               reduce the number of center to the goal specified
* EXAMPLE:
*
*****************************************************************************/
int 
em_clus_reduce_ncenter(int model, int goal)
{
    double          dist, min_dist, min_weight;
    int             g, min_g, reduce_goal, nearest_g;
    EM_CLUS_PARAMS *gpp;
    
    gpp = genpp[model];
    
    reduce_goal = gpp->current_ncenters - goal;
    if (reduce_goal <= 0) {
    warning("em_clus_reduce_centers: the number to remove >= ncenter\n");
    }
    
    while(reduce_goal > 0) {
    /* find center with smallest weight */
    min_weight = gpp->new_weight[0];
    min_g = 0;
    for(g = 1; g < gpp->current_ncenters; g++){
        if(gpp->new_weight[g] < min_weight){
        min_g = g;
        min_weight = gpp->new_weight[g];
        }
    }
    min_dist = MAXFLOAT;
    nearest_g = 0;
    /* find center closest to one with smallest weight */
    for(g = 0; g < gpp->current_ncenters; g++){
        if(g == min_g) continue;
        dist = l2dist(gpp->ninputs, gpp->means[min_g], gpp->means[g]);
        if(dist < min_dist){
        nearest_g = g;
        min_dist = dist;
        }
    }

    em_clus_remove_ncenter(model, min_g, nearest_g);
    reduce_goal--;
    } /* end while */
    return(min_g);
}
/*****************************************************************************
* NAME: em_clus_cluster()
* DESCRIPTION:
*   Uses Mahalanobis distances to determine the cluster a pattern
*   belongs to. In the The distances are weighted by a relative size of the
*   clusters as well.
*
* EXAMPLE:
*       
******************************************************************************/
int
em_clus_cluster (int model, float *input)
{
    int max_i, i;
    float max_output;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];

    max_output = -MAXFLOAT;
    max_i = 0;
    for(i = 0; i < gpp->ncenters; i++){
    gpp->outputs[i] =
      get_log_gauss_output_diag(gpp->ninputs, input,
                    gpp->means[i], gpp->diag_inv_var[i],
                    gpp->log_det_covar[i]);
    if(gpp->outputs[i] > max_output){
        max_output = gpp->outputs[i];
        max_i = i;
    }
    }

    return(max_i);
}

void
em_clus_copy_centers(int model, int ncenters, float **means, float **variances)
{
    int i, j;
    EM_CLUS_PARAMS *gpp;

    gpp = genpp[model];

    if(gpp->max_centers < ncenters)
      fatal_error("%s: %s\nModel %d has space for %d centers. %d to copy\n",
          "em_clus_copy_centers", "Too many centers",
          model, gpp->max_centers, ncenters);
    gpp->current_ncenters = ncenters;
    gpp->total_iteration = 1;

    for(i = 0; i < ncenters; i++){
    gpp->log_det_covar[i] = 0.;
    for(j = 0; j < gpp->ninputs; j++){
        gpp->means[i][j] = means[i][j];
        gpp->diag_var[i][j] = variances[i][j];
        if(gpp->diag_var[i][j] < gpp->minimum_variance)
          gpp->diag_var[i][j] = gpp->minimum_variance;
        if(gpp->diag_var[i][j] > SMALLEST)
          gpp->log_det_covar[i] +=
        (float)log10((double)gpp->diag_var[i][j]);
        gpp->diag_inv_var[i][j] = 1./gpp->diag_var[i][j];
    }
    }
}

