/***--------------------------------------------------------------------------
 * NAME:        em_clus_train.c
 * PURPOSE: training driver for em_clus clustering algorithm
 *
 * SEE ALSO:    em_clus_pack.c
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/em_clus/em_clus_train.c,v 11.4 2001/08/23 20:11:35 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/em_clus/em_clus_train.c,v $
 * CREATED: 10-12-91 lck                MIT LINCOLN LABORATORY
 *****************************************************************************/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <nnlib.h>
#include "em_clus.h"
#include "em_clus_pack.h"
#include <kmeans_pack.h>

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define MAXSTR (256)

/* static variables */
static char prog_name[MAXSTR];
static int *model_list, *kmeans_model_list;

void
em_clus_train(int nraw, int ninputs, int *features, int noutputs,
          int normalize, int npatterns, int debg, int verbosity,
          int random_flag, char *data_path, char *data_fname,
          char *norm_fname, char *exp_path, char *param_fname,
          char *comment, int adjust_priors, int priors_npatterns,
          float *priors_list, int max_iterations, float split_percentage,
          float stop_percentage, int cluster_by_class,
          int add_random_offset, int reduce_step, int *desired_ncenter,
          int use_kmeans, float kmeans_stop_percentage,
          float variance_scale, float minimum_variance)
{
    int     model, err_model, kmeans_model, normalization=0, data_model;
    int     current_class, num_in_class, desired_class;
    int     n_read, i, j, n_trials, final_round, first_time;
    float   *inputs;
    char    *tmp_path;
    FILE    *param_file;
    int     iteration, actual_ncenters, *ncenters;
    int     report_stop_flag, report_nvectors;
    float   cost;
    float   **pcenter;      /* center values */
    float   **pvariance;        /* variance values */
    float   *pnorm_variance;    /* for kmeans */
    int     post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    /* create model structures */
    model_list = allocate_ivector(noutputs);
    if(use_kmeans) kmeans_model_list = allocate_ivector(noutputs);

    if(use_kmeans){
    pnorm_variance = allocate_vector(ninputs);
    for(i = 0; i < ninputs; i++) pnorm_variance[i] = 1.;
    }
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    current_class = 0;
    do{
    if(use_kmeans){
        kmeans_model_list[current_class] =
          kmeans_build(ninputs, desired_ncenter[current_class]);
        kmeans_init(kmeans_model_list[current_class], features,
            pnorm_variance, kmeans_stop_percentage,
            split_percentage, max_iterations, reduce_step,
            add_random_offset);
    }
    model_list[current_class] =
      em_clus_build(ninputs, desired_ncenter[current_class]);
    model = model_list[current_class];
    em_clus_init(model, features, variance_scale, minimum_variance,
             stop_percentage, split_percentage, max_iterations,
             reduce_step, add_random_offset);
    current_class++;
    } while (cluster_by_class && current_class < noutputs);
    if(use_kmeans) deallocate_vector(pnorm_variance);

    model = model_list[0];

    err_model = report_build(verbosity, 0, noutputs);
    if(use_kmeans) report_init(err_model, 0, ninputs, CLUS_SQ_ERR, 0, post_sum_to_one);
    else report_init(err_model, 0, ninputs, CLUS_LOG_LIKE, 0, post_sum_to_one);

    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
                 ninputs, noutputs, features, normalization,
                 verbosity);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
    class_data_priors_init(data_model, priors_list, priors_npatterns);

    n_trials = class_data_get_npatterns(data_model);

    current_class = 0;
    do{
    model = model_list[current_class];
    if(use_kmeans){
        kmeans_model = kmeans_model_list[current_class];
        final_round = FALSE;
        first_time = FALSE;
    }
    if (verbosity > 1 && use_kmeans)
      printf("using kmeans to initialize\n");
    if(cluster_by_class){
        if(verbosity > 1)
          printf("training centers for class %d\n", current_class);
        num_in_class = class_data_get_npatterns_class(data_model,
                              current_class);
        report_epoch_header(err_model, stdout, num_in_class);
    } else {
        report_epoch_header(err_model, stdout, n_trials);
    }
    
    report_stop_flag = FALSE;
    while (!report_stop_flag) { /* for as many epochs as necessary */
        num_in_class = 0;
        class_data_epoch_init(data_model);
        for (n_read = 0; n_read < n_trials; n_read++) {
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        
        if(!cluster_by_class ||
           (cluster_by_class && desired_class == current_class)){
            if(use_kmeans && !final_round)
              kmeans_batch_train(kmeans_model, inputs);
            else
              em_clus_batch_train (model, inputs);
            num_in_class++;
        }
        }
        if(num_in_class != 0){
        if(use_kmeans && !final_round){
            report_stop_flag =
              kmeans_batch_update(kmeans_model, &iteration,
                      &actual_ncenters, &cost,
                      &report_nvectors);
            if(report_stop_flag){
            /* set up for em_clus final_round */
            final_round = TRUE;
            first_time = TRUE;
            pcenter =
              *(float ***)kmeans_param(kmeans_model, "pcenter");
            pvariance =
              *(float ***)kmeans_param(kmeans_model, "pvariance");
            em_clus_copy_centers(model, actual_ncenters,
                         pcenter, pvariance);
            report_stop_flag = FALSE;
            }
        } else
          report_stop_flag =
            em_clus_batch_update (model, &iteration, &actual_ncenters,
                      &cost, &report_nvectors);
        } else {
        warning("%s %d.\n%s", "There was no data in class",
            current_class, "Eliminating those centers");
        actual_ncenters = iteration = 0;
        ncenters = (int *)em_clus_param(model, "ncenters");
        *ncenters = 0;
        report_stop_flag = 1;
        cost = 0.;
        }
        report_cluster_epoch(err_model, stdout, iteration, actual_ncenters,
                 cost);
        if(use_kmeans && first_time){
        cost = 0;
        if (verbosity>1) {
            printf("running EM algorithm\n");
            report_init(err_model, 0, ninputs,
                CLUS_LOG_LIKE, 0, post_sum_to_one);
            report_epoch_header(err_model, stdout, n_trials);
            first_time = FALSE;
        }
        }
    }
    current_class++;
    } while(current_class < noutputs && cluster_by_class);
    
    /* write out model or models */
    tmp_path = path_build(exp_path, param_fname);
    /* store model in param file */
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    
    if(cluster_by_class){
    em_clus_write_model_list(model_list, noutputs, param_file);
    } else {
    em_clus_write_model(model_list[0], noutputs, param_file);
    }
    
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    if ( verbosity > 3) {
    current_class = 0;
    do{
        pcenter = *(float ***) em_clus_param(model_list[current_class],
                         "means");
        pvariance = *(float ***) em_clus_param(model_list[current_class],
                           "diag_var");
        printf("Center# Input#  center variance\n");  
        for(i=0; i< actual_ncenters; i++){
        for(j=0; j< ninputs; j++){
            printf("%3d %7d %10.3f %7.3f\n", 
               i, j, pcenter[i][j], pvariance[i][j]);
        }
        }
        current_class++;
    } while (current_class < noutputs && cluster_by_class);
    }
    
    /* free model structures */
    report_free(err_model);
    current_class = 0;
    if(cluster_by_class){
    for(current_class = 0; current_class < noutputs; current_class++)
      em_clus_free(model_list[current_class]);
    } else {
    em_clus_free(model_list[0]);
    }
    norm_free();
    class_data_exit(data_model);
}
