/***---------------------------------------------------------------------------
* NAME:     feat_sel.c
* PURPOSE:  driver for nearest neighbor feature selection
*
* SEE ALSO: knn_pack.c
* GLOBALS:  int debug;
*
* DESCRIPTION:
*   feat_sel()
*       does a forward search or a backward search
*       for the best set of features for classifying a given data set
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/feat_sel/feat_sel.c,v 11.5 2001/08/23 20:11:38 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/feat_sel/feat_sel.c,v $
* CREATED: 18-Apr-91 lck                             MIT LINCOLN LABORATORY
*----------------------- pseudocode of driver follows ----------------------
*   Read in data, normalizing it if necessary
*   build a model structure for a knn classifier
*
*   forward search:
*   initially, no features are selected, so you are choosing the best one
*
*   loop:   initialize the model to use selected features
*       for each remaining feature:
*           try adding it to the selected features:
*           for each vector in the input data: (train)
*               make a knn exemplar using selected features
*           for each knn exemplar: (test)
*               classify (throwing out nearest neighbor)
*           keep track of the feature that gives the minimum
*               percent error
*       add best feature to the array of selected features
*       keep track of the best combination of features
*
*   backward search:
*   initially, all features are selected, so you are choosing the worst one
*
*   loop:   initialize the model to use all remaining features
*       for each remaining feature:
*           try leaving this feature out:
*           train nearest neighbor, using remaining features
*           test on stored exemplars, ignoring closest (cvtest)
*           keep track of feature we are best off leaving out
*       add worst feature to list to leave out
*       keep track of the best combination of features
*
*   save results
---------------------- end of pseudocode ---------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <nnlib.h>
#include <knn_pack.h>
#include "feat_sel.h"

#define MAXSTR (256)
#define FORWARD 0
#define BACKWARD 1
#define FOR_AND_BACK 2

static char prog_name[MAXSTR];

void
feat_sel(int ninputs, int noutputs, int normalize, int npatterns, int debg,
     int verbosity, char *data_path, char *data_fname, char *norm_fname,
     char *exp_path, char *param_fname, char *comment, int search_type,
     int k, int search_n, int *categories)
{
    int model, normalization = 0;
    int desired_class, class, *desired_classes;
    int i, n_read, n_trials;
    int error_cnt;
    int nfeatures;
    int *features;
    int cvtest_flag = 1;
    float **input_matrix, *inputs;
    float percent_error;
    char data_fullname[256];
    FILE *data_file;

    strcpy(prog_name, get_program_name());

            /* put together the full path to the data file */
    strcpy(data_fullname, path_build(data_path, data_fname));

    /* read in data */
    if (normalize) {
    normalization = read_norm_file(path_build(data_path, norm_fname));
    }

    /* allocate space for data */
    inputs = allocate_vector(ninputs);
    desired_classes = allocate_ivector(npatterns);
    input_matrix = allocate_matrix(npatterns, ninputs);

    /* read feature data */
    data_file = nn_fopen(data_fullname, "r", "for reading data");

    n_trials = load_class_data(data_file, npatterns, ninputs, ninputs,
                   desired_classes, input_matrix,
                   normalization, data_fname, NULL);

    /* close training data file */
    nn_fclose(data_file, data_fullname, "for scan of training data");

    /* initialize routines for generating search feature lists */
    cat_sel_build(ninputs, search_type, search_n, verbosity, categories);
    feat_sel_print_head();

    while(!feat_sel_search_done()){
    feat_sel_init_nfeatures();
    while((features = feat_sel_next_features()) != NULL){
        /* features is ninputs+1 long, features[0] = nfeatures */
        nfeatures = features[0];

        /* make knn model */
        model = knn_build(nfeatures, noutputs, n_trials, 1);
        /* train */
        for (n_read = 0; n_read < n_trials; n_read++) {
        /* build input vector */
        desired_class = desired_classes[n_read];
        for (i = 0; i < nfeatures; i++) {
            inputs[i] = input_matrix[n_read][features[i+1]];
        }
        knn_inc_ctrain(model, inputs, desired_class);
        }

        /* initialize fastdist structure */
        knn_init(model, n_trials, features);

        /* test */
        error_cnt = 0;
        for (n_read = 0; n_read < n_trials; n_read++) {
        /* read input vector */
        desired_class = knn_copy_exemplar(model, n_read, inputs);

        /* classify pattern and report errors */
        class = knn_classify(model, inputs, k, cvtest_flag);
        if (class != desired_class) {
            error_cnt++;
        }
        }
        percent_error = 100. * (float) error_cnt/(float)n_trials;
        feat_sel_check_best(percent_error);

        knn_free(model);
    }
    /* add the best feature to the list */
    feat_sel_pick_best();
    }

    /* print out results */
    feat_sel_print_final_result();
    
    /* save results in parameter file */
    feat_sel_save_list(path_build(exp_path, param_fname), comment);

    deallocate_vector(inputs);
    deallocate_matrix(input_matrix);
    deallocate_ivector(desired_classes);
    if(normalize) norm_free();
    feat_sel_free();
}
