/***---------------------------------------------------------------------------
* NAME:     feat_sel_main.c
* PURPOSE:  Driver for K Nearest Neighbor program
*
* USAGE:    feat_sel -<flags>
* EXAMPLE:  feat_sel --     * shows all available flags
*
* RETURN VALUE: dependent on -verbose
*   if testing, at a minimum the percent error is written to standard out
*
*
* SEE ALSO: feat_sel.c knn_pack.c feat_sel.run
*
* GLOBALS: <flags> defined in feat_sel_args.h
*
* DESCRIPTION:  entry routine for Feature Selection using
*       Nearest Neighbor classifier
*
* RETURNS:
*   creates a parameter file with an ordered list of parameters
*       and the error rate when that parameter was introduced
*   depending on the verbosity, information is printed on stdout:
*   verbosity 0) best feature set found in search and its error rate
*         1) all level 0 printouts plus best feature set found for
*           each number of features and their error rates
*           also, the command line used
*         2) all level 1 printouts plus each feature set tried
*           and their error rates
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/feat_sel/feat_sel_main.c,v 11.5 2001/08/23 20:11:38 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/feat_sel/feat_sel_main.c,v $
* CREATED: 12-Dec-90 lck                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char     rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/feat_sel/feat_sel_main.c,v 11.5 2001/08/23 20:11:38 craig Exp $";

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include <gen_search_args.h>
#include "feat_sel_args.h"
#include "feat_sel.h"

int
main(int argc, char **argv)
{
    char *comment;
    /* arguments for reading defaults file */
    DEFAULTS_FILE_STRUCT *df_args;
    /* argument table with all entries */
    ARG *argtab2;
    int tabsize2;

    init_nnsubs(argc, argv, rcsid);

    /* read command line */
    nnargs_combine_argtabs(argtab, TABSIZE, search_tab, SEARCH_TABSIZE,
               &argtab2, &tabsize2);
    argc = nnargs(argc, argv, argtab2, tabsize2);

    print_log_header();

    /* read the defaults file */
    if(strcmp(desc_fname, "") == 0){
    fatal_error("%s: -fdescribe must be set", argv[0]);
    }
    df_args = (DEFAULTS_FILE_STRUCT *)
      nn_malloc(sizeof(DEFAULTS_FILE_STRUCT), "for the defaults file struct");
    read_defaults_file(path_build(data_path, desc_fname), df_args);
    if(df_args->noutputs == 0)
      fatal_error("noutputs in %s must be greater than 0", desc_fname);
    if(df_args->ninputs == 0)
      fatal_error("ninputs in %s must be greater than 0", desc_fname);

    /* set the number of patterns to no more than the number of lines
       in the data file */
    if(strcmp(data_fname, "") == 0){ /* data_fname is unset */
    fatal_error("%s: -finput must be set", argv[0]);
    }
    check_file_length(path_build(data_path, data_fname), &npatterns);
    if(npatterns == 0)
      fatal_error("%s: Empty data file %s", argv[0],
          path_build(data_path, data_fname));

    if (df_args->db_type != CLASS_DATA)
      fatal_error("class data required, check defaults file");

    if(search_n > df_args->ninputs){
    warning("%s: %s\nDoing full search", "feat_sel_main:",
        "Searching for more features than there are.");
    stop_early = 0;
    }
    /* if search_n is 0, feat_sel does the whole search */
    if(!stop_early) search_n = 0;

    comment = make_comment_string(nnargs_get_flags(argtab2, tabsize2));
    if (search_verbosity > FLAGS_VERBOSITY) {
    printf("%s", comment);
    fflush(stdout);
    }

    feat_sel(df_args->ninputs, df_args->noutputs, normalize, npatterns, debug,
         search_verbosity,
         data_path, data_fname, norm_fname, exp_path, search_param_fname,
         comment, search_direction, k, search_n, df_args->categories);

    print_log_footer();
    nnargs_free_argtab(argtab2, tabsize2);

    /* free the vectors allocated when the defaults file was read */
    free_defaults_file_struct(df_args);

    exit_nnsubs();
    exit(0);
}
