/***-------------------------------------------------------------------------
 * NAME:        gauss_pack.c
 * PURPOSE: Functions specific to gauss.c 
 *          
 * SEE ALSO:    gauss.c gauss_main.c
 *
 * GLOBALS: int debug;
 *
 * DESCRIPTION:
 *    Functions in this file
 *  gauss_build     Allocate storage for model parameters
 *  gauss_init      Initialize model parameters
 *  gauss_free      Deallocates model storage
 *  gauss_read_model    Initializes a gauss model from file
 *  gauss_write_model   Writes a gauss model to output file
 *  gauss_param     Returns pointers to elements in model
 *  gauss_batch_ctrain  does training on a vector
 *  gauss_batch_update  updates parameters of model after training
 *  gauss_test_build    Allocate storage for test mode parameters
 *  gauss_test_free     Deallocates test mode storage
 *  gauss_test_init     Initialize test mode parameters
 *  gauss_classify      classifies a vector using the gauss algortihm
 *
 * ALGORITHM:
 * REFERENCE:
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_pack.c,v 11.4 2001/08/23 20:11:43 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_pack.c,v $
 * CREATED: 26-MAR-91 nation            MIT LINCOLN LABORATORY
 ------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#define gauss_pack_h_internal_types
#include "gauss_pack.h"

/* constant definitions */
#define PEQ(y) (0==strcmp(param_name,y))
#define SMALLEST 1.e-30 /* variance is not allowed to be less than SMALLEST */

/* Internal Globals */
static GAUSS_PARAMS **gausspp;  /* array of parameters for each model */
static int max_models = 0;  /* next new model number to be assigned */

/************************************************************************
 * NAME: gauss_build(ninputs, noutputs, per_class, full)
 * DESCRIPTION:
 *  Allocates space for a gauss model 
 *
 *************************************************************************/
int
gauss_build(int ninputs, int noutputs, int per_class_variance, int full_covar)
{
    register int    new_model;
    int i;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    new_model = model_get_number((void ***)&gausspp,
                 &max_models, sizeof(GAUSS_PARAMS));
    gpp = gausspp[new_model];
    
    /* set fields which are array sizes */
    gpp->ninputs = ninputs;
    gpp->noutputs = noutputs;
    
    /* allocate space for vectors */
    gpp->features = allocate_ivector(ninputs+1);
    gpp->class_trials = allocate_ivector (noutputs);
    gpp->cv_class = (PCLCOV *)
      nn_malloc(sizeof(PCLCOV)*noutputs,"for covariance structure");
    for(i = 0; i < noutputs; i++){
    gpp->cv_class[i].mean = allocate_vector(ninputs);
    }
    gpp->per_class = per_class_variance;
    gpp->full_covar = full_covar;
    for (i = 0; i < noutputs; i++) {
    if(full_covar)
      gpp->cv_class[i].covar = allocate_matrix (ninputs, ninputs);
    else
      gpp->cv_class[i].var = allocate_vector(ninputs);
    }
    if(!per_class_variance){
    if(full_covar)
      gpp->covar = allocate_matrix (ninputs, ninputs);
    else
      gpp->var = allocate_vector(ninputs);
    }
    return new_model;
}

/*****************************************************************************
 * NAME: gauss_init ()
 * DESCRIPTION:     
 *      Initialize gaussian classifier variables
 *
 *****************************************************************************/
void
gauss_init (int model, int *nn_features, float minimum_variance)
{
    int i, j, k;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    /* copy features into allocated array */
    for (i = 0; i <= nn_features[0]; i++)
      gpp->features[i] = nn_features[i];
    
    gpp->minimum_variance = minimum_variance;
    gpp->total_trials = 0;
    
    for(i = 0; i < gpp->noutputs; i++){
    gpp->class_trials[i] = 0;
    for(j = 0; j < gpp->ninputs; j++){
        gpp->cv_class[i].mean[j] = 0.0;
    }
    }
    if(gpp->full_covar){
    for(i = 0; i < gpp->ninputs; i++){
        for(j = 0; j < gpp->ninputs; j++){
        for(k = 0; k < gpp->noutputs; k++)
          gpp->cv_class[k].covar[i][j] = 0.0;
        if(!gpp->per_class) gpp->covar[i][j] = 0.0;
        }
    }
    } else {
    /* diagonal */
    for(i = 0; i < gpp->ninputs; i++){
        for(k = 0; k < gpp->noutputs; k++)
          gpp->cv_class[k].var[i] = 0.;
        if(!gpp->per_class) gpp->var[i] = 0.;
    }
    }
}      

/************************************************************************
 * NAME: gauss_free()
 * DESCRIPTION:
 *      Deallocates space used in a gauss model 
 *
 *************************************************************************/
void
gauss_free(int model)
{
    int i;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    /* does not free those things allocated in gauss_test_build */
    gpp = gausspp[model];
    deallocate_ivector (gpp->class_trials);
    deallocate_ivector( gpp->features );
    for (i = 0; i < gpp->noutputs; i++) {
    deallocate_vector(gpp->cv_class[i].mean);
    }
    for(i = 0; i < gpp->noutputs; i++)
      if(gpp->full_covar)
    deallocate_matrix(gpp->cv_class[i].covar);
      else deallocate_vector(gpp->cv_class[i].var);
    if(!gpp->per_class){
    if(gpp->full_covar) deallocate_matrix(gpp->covar);
    else deallocate_vector(gpp->var);
    }
    nn_free((char *)gpp->cv_class, "gauss per class variance structure");
    nn_free((char *)gpp, "gauss model");
    gausspp[model] = NULL;
}

/***---------------------------------------------------------------------------
 * NAME:     gauss_read_model()
 * DESCRIPTION:
 *       Loads parameters from ascii input file
 *       allocates space for and loads data into vectors and matricies
 *
 *       on error from param_getblk, returns the error value
 *       (there is already an error check on the allocation, so we don't
 *       have to do that here)
 *
 * example:
 --------------------------------------------------------------------------***/
int
gauss_read_model (FILE *fp)
{
    int i, rv, model;
    int ninputs, noutputs, total_trials, full, per_class;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    /* get the gauss model scalars */
    param_setvarptr("ninputs", &ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &noutputs, paramtab, TABSIZE);
    param_setvarptr("total_trials", &total_trials, paramtab, TABSIZE);
    param_setvarptr("full_covar", &full, paramtab, TABSIZE);
    param_setvarptr("per_class", &per_class, paramtab, TABSIZE);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0) return rv;
    
    /* allocate space for model and vectors */
    model = gauss_build(ninputs, noutputs, per_class, full);
    gpp = gausspp[model];
    gpp->total_trials = total_trials;
    
    /* fill in parameters set during training or by gauss_init() */
    
    /* get arrays */
    param_setvarptr("features", &gpp->features[0],  paramtab_l,  TABSIZE_L);
    /* set feature list length to the max we should need to read */
    param_setdim("features", gpp->ninputs+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("class_trials", (int *)&gpp->class_trials[0],
            paramtab_l, TABSIZE_L);
    param_setdim("class_trials",gpp->noutputs,0, paramtab_l, TABSIZE_L);
    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0) return rv;

    /* get means */
    for (i = 0; i < gpp->noutputs; i++) {
    param_setvarptr("mean", (int *)&gpp->cv_class[i].mean[0],
            means, NMEANS);
    param_setdim("mean", gpp->ninputs, 0, means, NMEANS);
        if ((rv = param_getblk(fp, means, NMEANS)) < 0) return rv;
    }
    /* get variances or covariances */
    if(per_class){
    for(i = 0; i < gpp->noutputs; i++){
        if(full){
        param_setvarptr("log_det",
                (int *)&gpp->cv_class[i].log_det,
                covar_tab, NCOVAR);
        param_setvarptr("inv_covar",
                (int *)&gpp->cv_class[i].covar[0][0],
                covar_tab, NCOVAR);
        param_setdim("inv_covar", gpp->ninputs, gpp->ninputs,
                 covar_tab, NCOVAR);
        if((rv = param_getblk(fp, covar_tab, NCOVAR)) < 0) return rv;
        } else {
        param_setvarptr("log_det",
                (int *)&gpp->cv_class[i].log_det,
                var_tab, NVAR);
        param_setvarptr("inv_var",
                 (int *)&gpp->cv_class[i].var[0],
                 var_tab, NVAR);
        param_setdim("inv_var", gpp->ninputs, 0, var_tab, NVAR);
        if((rv = param_getblk(fp, var_tab, NVAR)) < 0) return rv;
        }
    }
    } else {
    if(full){
        param_setvarptr("log_det", (int *)&gpp->log_det,
                covar_tab, NCOVAR);
        param_setvarptr("inv_covar", (int *)&gpp->covar[0][0],
                covar_tab, NCOVAR);
        param_setdim("inv_covar", gpp->ninputs, gpp->ninputs,
             covar_tab, NCOVAR);
        if((rv = param_getblk(fp, covar_tab, NCOVAR)) < 0) return rv;
    } else {
        param_setvarptr("log_det", (int *)&gpp->log_det,
                var_tab, NVAR);
        param_setvarptr("inv_var",
                (int *)&gpp->var[0], var_tab, NVAR);
        param_setdim("inv_var", gpp->ninputs, 0, var_tab, NVAR);
        if((rv = param_getblk(fp, var_tab, NVAR)) < 0) return rv;
    }
    }
    return (model);
}               /* gauss_read_model() */
/***--------------------------------------------------------------------------
 * NAME:     gauss_write_model()
 * DESCRIPTION:
 *       Writes parameters and exemplars for a model to a file
 *       in ascii format
 *
 * example:
 *
 --------------------------------------------------------------------------***/
int
gauss_write_model (int model, FILE *fp)
{
    int i, rv;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    /* save scalars */
    /* get the gauss model scalars */
    param_setvarptr("ninputs", &gpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &gpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("total_trials", &gpp->total_trials, paramtab, TABSIZE);
    param_setvarptr("full_covar", &gpp->full_covar, paramtab, TABSIZE);
    param_setvarptr("per_class", &gpp->per_class, paramtab, TABSIZE);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0) return rv;
    
    /* save vectors */
    param_setvarptr("features", &gpp->features[0],  paramtab_l,  TABSIZE_L);
    param_setdim("features",gpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("class_trials", (int *)&gpp->class_trials[0],
            paramtab_l, TABSIZE_L);
    param_setdim("class_trials",gpp->noutputs,0, paramtab_l, TABSIZE_L);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0) return rv;
    
    /* save means */
    for (i = 0; i < gpp->noutputs; i++) {
    param_setvarptr("mean", (int *)&gpp->cv_class[i].mean[0],
            means, NMEANS);
    param_setdim("mean", gpp->ninputs, 0, means, NMEANS);
        if ((rv = param_putblk(fp, means, NMEANS)) < 0) return rv;
    }

    /* save variances or covariances */
    if(gpp->per_class){
    for(i = 0; i < gpp->noutputs; i++){
        if(gpp->full_covar){
        param_setvarptr("log_det",
                (int *)&gpp->cv_class[i].log_det,
                covar_tab, NCOVAR);
        param_setvarptr("inv_covar",
                (int *)&gpp->cv_class[i].covar[0][0],
                covar_tab, NCOVAR);
        param_setdim("inv_covar", gpp->ninputs, gpp->ninputs,
                 covar_tab, NCOVAR);
        if((rv = param_putblk(fp, covar_tab, NCOVAR)) < 0) return rv;
        } else {
        param_setvarptr("log_det",
                (int *)&gpp->cv_class[i].log_det,
                var_tab, NVAR);
        param_setvarptr("inv_var",
                 (int *)&gpp->cv_class[i].var[0],
                 var_tab, NVAR);
        param_setdim("inv_var", gpp->ninputs, 0, var_tab, NVAR);
        if((rv = param_putblk(fp, var_tab, NVAR)) < 0) return rv;
        }
    }
    } else { /* grand */
    if(gpp->full_covar){
        param_setvarptr("log_det", (int *)&gpp->log_det,
                covar_tab, NCOVAR);
        param_setvarptr("inv_covar", (int *)&gpp->covar[0][0],
                covar_tab, NCOVAR);
        param_setdim("inv_covar", gpp->ninputs, gpp->ninputs,
             covar_tab, NCOVAR);
        if((rv = param_putblk(fp, covar_tab, NCOVAR)) < 0) return rv;
    } else {
        param_setvarptr("log_det", (int *)&gpp->log_det,
                var_tab, NVAR);
        param_setvarptr("inv_var", (int *)&gpp->var[0], var_tab, NVAR);
        param_setdim("inv_var", gpp->ninputs, 0, var_tab, NVAR);
        if((rv = param_putblk(fp, var_tab, NVAR)) < 0) return rv;
    }
    }
    return (0);
}               /* gauss_write_model() */


/************************************************************************
 * NAME: gauss_param(param_name)
 * DESCRIPTION:
 *  takes parameter name.
 *  returns a floating point pointer to the parameter.
 *  single value parameters are read only.
 *  function really not necessary if not interested in accessing
 *  parameters (reading) from main()
 *************************************************************************/
void *
gauss_param(int model, char *param_name, int class)
{
    void *param_address;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    if (PEQ("features"))
      param_address = &gpp->features;
    else if (PEQ("ninputs"))
      param_address = &gpp->ninputs;
    else if (PEQ("noutputs"))
      param_address = &gpp->noutputs;
    else if (PEQ("outputs"))
      param_address = &gpp->outputs;
    else if (PEQ("total_trials"))
      param_address = &gpp->total_trials;
    else if (PEQ("class_trials"))
      param_address = &gpp->class_trials[class];
    else if (PEQ("covariance"))
      param_address = &gpp->covar;
    else if (PEQ("variance"))
      param_address = &gpp->var;
    else if (PEQ("per_class_covariance"))
      param_address = &gpp->cv_class[class].covar;
    else if (PEQ("per_class_variance"))
      param_address = &gpp->cv_class[class].var;
    else if (PEQ("determinant"))
      param_address = &gpp->log_det;
    else if (PEQ("per_class_determinant"))
      param_address = &gpp->cv_class[class].log_det;
    else if (PEQ("mean"))
      param_address = &gpp->cv_class[class].mean;
    else if (PEQ("full_covar"))
      param_address = &gpp->full_covar;
    else if (PEQ("per_class"))
      param_address = &gpp->per_class;
    else {
    warning("gauss_param: parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
} 

/************************************************************************
 * NAME: gauss_batch_ctrain(model, inputs, desired_class);
 * DESCRIPTION:
 *  updates the values of the means and variances using inputs
 *
 *        t=n        t=n-1
 *         _          _
 *  m(t) = > x(t)/n = > x(t)(n-1)/(n(n-1)) + x(t)/n = m(t-1)(n-1)/n + x(t)/n
 *         -          -
 *        t=1        t=1
 *
 *  x(i) - m(t) = x(i) - m(t-1) - (x(t) - m(t-1))/n {i < t}
 *
 *  x(n) - m(t) = (x(t) - m(t-1))(n-1)/n
 *
 *        i=t
 *         _
 *  V(t) = > (x(i) - m(t))^2/t = ((t-1)/t)(V(t-1) + (((x(t) - m(t-1))^2)/t))
 *         -
 *        i=1
 *
 *                               t-1
 *                               -
 * V(t) = (x(t)-m(t))^2/t + 1/t* > (x(i)-m(t))^2
 *                               -
 *                              i=1
 *                                           t-1
 *                                            -
 * = ((x(t)-(t-1)m(t-1)/t - x(t)/t)^2)t + 1/t*>(x(i)-m(t-1)(t-1)/t-x(t)/t)^2
 *                                            -
 *                                           i=1
 * 
 *                                       t-1
 *                                        -
 * = ((t-1)/t)^2*(x(t)-m(t-1))^2)/t + 1/t*>((x(i)-m(t-1))-(x(t)-m(t-1))/t)^2
 *                                        -
 *                                       i=1
 *
 * = ((t-1)/t)^2*(x(t)-m(t-1))^2)/t + 
 *
 *    t-1
 *     -
 * 1/t*>((x(i)-m(t-1))^2 -(2/t)(x(t)-m(t-1))(x(i)-m(t-1)) + ((x(t)-m(t-1))/t)^2
 *     -
 *    i=1
 *
 *                                                                      t-1
 *                                                                       -
 * = (((t-1)^2+(t-1)/t^2)/t)(x(t)-m(t-1)^2 + (t-1)V(t-1) - 2/t(x(t)-m(t-1))>(x(i)-m(t-1))
 *                                                                       -
 *                                                                      i=1
 *
 * = (((x(t)-m(t-1))^2)(t^2+t)/t^2)+(t-1)V(T-1) - 2/t(x(t)-m(t-1))(t-1)(m(t-1)-m(t-1)))/t
 *
 * = ((t-1)/t)(V(t-1)+(x(t)-m(t-1))^2/t)
 *
 * 
 *************************************************************************/
void
gauss_batch_ctrain(int model, float *inputs, int desired_class)
{
    register float tmp;
    register int i, j;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    float n;
    
    gpp = gausspp[model];
    gpp->total_trials++;
    gpp->class_trials[desired_class]++;
    /* update variances */
    n = (float)gpp->class_trials[desired_class];
    if(gpp->full_covar){
    for(i = 0; i < gpp->ninputs; i++){
        for (j = 0; j < gpp->ninputs; j++) {
        tmp = inputs[i] - gpp->cv_class[desired_class].mean[i];
        tmp *= inputs[j] - gpp->cv_class[desired_class].mean[j];
        tmp /= n;
        gpp->cv_class[desired_class].covar[i][j] += tmp;
        gpp->cv_class[desired_class].covar[i][j] *= (n-1.)/n;
        }
    }
    } else { /* diagonal */
    for(i = 0; i < gpp->ninputs; i++){
        tmp = inputs[i] - gpp->cv_class[desired_class].mean[i];
        tmp *= tmp/n;
        gpp->cv_class[desired_class].var[i] += tmp;
        gpp->cv_class[desired_class].var[i] *= (n-1.)/n;
    }
    }
    /* update means */
    for (i = 0; i < gpp->ninputs; i++) {
    n = (float)gpp->class_trials[desired_class];
    gpp->cv_class[desired_class].mean[i] *= (n-1.)/n;
        gpp->cv_class[desired_class].mean[i] += inputs[i]/n;
    }
}

/************************************************************************
 * NAME: gauss_batch_update()
 * DESCRIPTION:
 *    the variances are calculated all at once, so this is just
 *    a final check that the minimum variances are ok and taking averages
 *    for the grand versions.  It's too hard to do the overall average
 *    variance from the per class incremental means all at once, so per
 *    class variances are calculated then averaged
 *    
 *    inversion can be done here, now, so it is.
 *
 *************************************************************************/
void
gauss_batch_update(int model, float max_ratio)
{
    register int i, j, k;
    float v_tmp;
    float *eigenvalues;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    if(gpp->per_class){
    /* check on minimum variances */
    for (k = 0; k < gpp->noutputs; k++) {
        if(gpp->full_covar){
        for (i = 0; i < gpp->ninputs; i++){
            if(gpp->cv_class[k].covar[i][i] < gpp->minimum_variance)
              gpp->cv_class[k].covar[i][i] = gpp->minimum_variance;
        }
        } else {
        for(i = 0; i < gpp->ninputs; i++){
            if(gpp->cv_class[k].var[i] < gpp->minimum_variance)
              gpp->cv_class[k].var[i] = gpp->minimum_variance;
        }
        }
    }
    } else {
    /* take averages of per_class variances, then check for min_var */
    if(gpp->full_covar){
        for(i = 0; i < gpp->ninputs; i++){
        for(j = 0; j < gpp->ninputs; j++){
            for(k = 0; k < gpp->noutputs; k++){
            gpp->covar[i][j] += gpp->cv_class[k].covar[i][j];
            }
            gpp->covar[i][j] /= (float)gpp->noutputs;
        }
        }
        for(i = 0; i < gpp->ninputs; i++){
        if(gpp->covar[i][i] < gpp->minimum_variance)
          gpp->covar[i][i] = gpp->minimum_variance;
        }
    } else {
        for(i = 0; i < gpp->ninputs; i++){
        for(k = 0; k < gpp->noutputs; k++){
            gpp->var[i] += gpp->cv_class[k].var[i];
        }
        gpp->var[i] /= (float)gpp->noutputs;
        }
        for(i = 0; i < gpp->ninputs; i++){
        if(gpp->var[i] < gpp->minimum_variance)
          gpp->var[i] = gpp->minimum_variance;
        }
    }
    }
    /* invert the variance or covariance matrices */
    if (gpp->full_covar == 0) { /* diagonal covariance */
    if (gpp->per_class == 0) {  /* grand variance */
        gpp->log_det = 0.;
        for(j = 0; j < gpp->ninputs; j++){
        v_tmp = gpp->var[j];
        if(v_tmp > SMALLEST)
          gpp->log_det += log10(v_tmp);
        gpp->var[j] = 1./v_tmp;
        }
    } else {        /* per class variance*/ 
        for (i = 0; i < gpp->noutputs; i++) {
        gpp->cv_class[i].log_det = 0.;
        for (j = 0; j < gpp->ninputs; j++) {
            v_tmp = gpp->cv_class[i].var[j];
            if(v_tmp > SMALLEST)
              gpp->cv_class[i].log_det += log10(v_tmp);
            gpp->cv_class[i].var[j] = 1./v_tmp;
        }
        }
    }
    } else {                        /* Full Covariance */
    eigenvalues = allocate_vector (gpp->ninputs);
    if (gpp->per_class == 0) {      /* Grand Variance */
        invert_and_calc_det(gpp->ninputs, gpp->covar,
                eigenvalues, max_ratio, &gpp->log_det);
    } else {                /* Per Class Variance*/
        for (i = 0; i < gpp->noutputs; i++) {
        invert_and_calc_det(gpp->ninputs, gpp->cv_class[i].covar,
                    eigenvalues, max_ratio,
                    &gpp->cv_class[i].log_det);
        }
    }
    deallocate_vector (eigenvalues);
    }
}

/*****************************************************************************
 * NAME: gauss_test_build ()
 * DESCRIPTION:     
 *  allocates storage for test mode of gaussian classifier software
 *
 *****************************************************************************/
void
gauss_test_build (int model)
{
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    gpp->outputs = allocate_vector (gpp->noutputs);
    gpp->log_class_priors = allocate_vector(gpp->noutputs);
}


/************************************************************************
 * NAME: gauss_test_free()
 * DESCRIPTION:
 *  deallocates space used in a gauss model 
 *
 *************************************************************************/
void
gauss_test_free(int model)
{
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    deallocate_vector (gpp->outputs);
    deallocate_vector (gpp->log_class_priors);
}

/*****************************************************************************
 * NAME: gauss_test_init ()
 * DESCRIPTION:     
 *  invert the covariance matrices, if full covariance
 *  store the inverse of the diagonal variance if diagonal
 *  store the log_class_priors
 *  
 *  Side effect:  If full, grand, the grand covariance matrix is inverted
 *      If full, per class, the per class covarainces are inverted
 *      otherwise, these matrices are left alone.
 *****************************************************************************/
void
gauss_test_init (int model, int adjust_priors, float *priors_list)
{
    int i;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    
    if(adjust_priors)
      for (i = 0; i < gpp->noutputs; i++) 
    gpp->log_class_priors[i] = log10(priors_list[i]);
    else
      for (i = 0; i < gpp->noutputs; i++){
      if(gpp->class_trials[i] != 0.)
        gpp->log_class_priors[i] = log10(gpp->class_trials[i] / 
                         ((double)gpp->total_trials));
      else
        gpp->log_class_priors[i] = 0.;
      }
}

/************************************************************************
 * NAME: gauss_eval(model, inputs)
 * DESCRIPTION:
 *  Takes gauss model number and a pointer to an input vector.
 *  fills in the output vector.
 *
 *************************************************************************/
void
gauss_eval(int model, float *inputs)
{
    int i;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    for(i = 0; i < gpp->noutputs; i++){
    /* the var and covar entries have been inverted */
    if (gpp->per_class == 0) {          /* Grand Variance */
        if(gpp->full_covar == 0){   /* diagonal */
        gpp->outputs[i] =
          get_log_gauss_output_diag(gpp->ninputs, inputs,
                        gpp->cv_class[i].mean,
                        gpp->var,
                        gpp->log_det);
        } else {
        gpp->outputs[i] =
          get_log_gauss_output_full(gpp->ninputs, inputs,
                        gpp->cv_class[i].mean,
                        gpp->covar,
                        gpp->log_det);
        }
    } else {               /* Per Class Variance */
        if (gpp->full_covar == 0) { /* Diagonal Covariance*/
        gpp->outputs[i] =
          get_log_gauss_output_diag(gpp->ninputs, inputs,
                        gpp->cv_class[i].mean,
                        gpp->cv_class[i].var,
                        gpp->cv_class[i].log_det);
        } else {
        gpp->outputs[i] =
          get_log_gauss_output_full(gpp->ninputs, inputs,
                        gpp->cv_class[i].mean,
                        gpp->cv_class[i].covar,
                        gpp->cv_class[i].log_det);
        }
    }
    gpp->outputs[i] += gpp->log_class_priors[i];
    }
}
/************************************************************************
 * NAME: gauss_classify(model, inputs)
 * DESCRIPTION:
 *  Takes gauss model number and a pointer to an input vector.
 *  Returns the class assigned to the input vector and fills in the
 *      output vector.
 *
 *************************************************************************/
int
gauss_classify(int model, float *inputs)
{
    int class;
    GAUSS_PARAMS *gpp;  /* temporary pointer to genpp[] */
    
    gpp = gausspp[model];
    gauss_eval(model, inputs);
    class = (max_pick (gpp->outputs, gpp->noutputs));
    
    return class;
}
