/***---------------------------------------------------------------------------
* NAME:     gauss_seq_cv.c 
* PURPOSE:  cross validation driver for Gaussian sequence classifier
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_seq_cv.c,v 11.4 2001/08/23 20:11:43 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_seq_cv.c,v $
* CREATED: 10-Aug-93 lck            MIT LINCOLN LABORATORY
****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <nnlib.h>
#include "gauss_seq.h"
#include "gauss_seq_pack.h"
#include "gauss_pack.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

void
gauss_seq_cross_valid(int db_type, int nraw, int ninputs, int *features,
              int nwindow_frames, int *window_frames, int noutputs,
              int normalize, int nsequences, int debg, int verbosity,
              int verror, int cv_nfolds, int cv_nsplits,
              int use_splits, int *cv_splits, char **cv_train_mask,
              char **cv_test_mask, int cv_random, char *data_path,
              char *data_fname, char *norm_fname, char *exp_path,
              char *error_fname, char *param_fname, char **labels,
              char *comment, int covariance_type, int variance_type,
              float minimum_variance, float max_ratio)
{
    int     model, seq_model, normalization = 0;
    int     err_model, err_part_model, data_model, i;
    int     log_outputs = 1, post_sum_to_one = 0, fold, n_read, nframes, seq_num;
    int     model_ninputs, n_files;
    int     desired_class, class;
    float   *inputs, *seq_sums, *priors_list;
    char    *tmp_path;
    FILE    *param_file, *error_file = NULL;

    /* allocate and initialize model structure */
    model_ninputs = ninputs*nwindow_frames;
    model = gauss_build(model_ninputs, noutputs,
            variance_type, covariance_type);
    gauss_init(model, features, minimum_variance);
    gauss_test_build(model);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else norm_setup_none(ninputs, noutputs);

    seq_model = gauss_seq_build(noutputs);
    gauss_seq_init(seq_model, model);
    seq_sums = *(float **) gauss_seq_param(seq_model, "seq_sums");

    /* gauss should adjust for uniform priors for the frames */
    priors_list = allocate_vector(noutputs);
    for(i = 0; i < noutputs; i++) priors_list[i] = 1.;

    /* prepare error handling */
    if(verror > 0) {
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    }
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, db_type, model_ninputs, LOG_LIKE, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize seq_data routines */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
                   ninputs, noutputs, features, normalization,
                   labels, verbosity, nwindow_frames,
                   window_frames);
    
    seq_data_cv_init(data_model, cv_nsplits, use_splits, cv_splits,
             cv_nfolds, cv_train_mask, cv_test_mask, cv_random);

    /* write normalization parameters to model file */
    param_file = nn_fopen(path_build(exp_path, param_fname),
              "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing_parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);

    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);
    gauss_init(model, features, minimum_variance);

    /* set up the data routines for this fold */
    seq_data_start_cv_fold(data_model, fold, TRUE);

    nframes = seq_data_get_nframes(data_model);
    nsequences = seq_data_get_nfiles(data_model);

    for(n_files = 0; n_files < nsequences; n_files++){
        if(seq_data_get_next_file(data_model) == -1) break;
        desired_class = seq_data_get_class(data_model);
        for(nframes = 0;; nframes++){
        inputs = seq_data_get_next_frame(data_model);
        if(inputs == NULL) break;
        gauss_batch_ctrain(model, inputs, desired_class);
        }
    }
    gauss_batch_update (model, max_ratio);

    /* write gauss model to parameter file */
    fprintf(param_file, "gauss model %d\n", fold);
    gauss_write_model(model, param_file);

    /* reinitialize model for testing */
    gauss_test_init(model, 1, priors_list);
    gauss_seq_init(seq_model, model);
    report_init(err_part_model, db_type, model_ninputs, LOG_LIKE,
            log_outputs, post_sum_to_one);

    /* finish the training fold, start the testing fold */
    seq_data_end_cv_fold(data_model);
    seq_data_start_cv_fold(data_model, fold, FALSE);
    nsequences = seq_data_get_nfiles(data_model);
    for(n_read = 0; n_read < nsequences; n_read++){
        if(seq_data_get_next_file(data_model) == -1) break;
        seq_num = seq_data_get_file_num(data_model);
        desired_class = seq_data_get_class(data_model);
        gauss_seq_init(seq_model, model);
        for(nframes = 0; ; nframes++){
        inputs = seq_data_get_next_frame(data_model);
        if(inputs == NULL) break;
        gauss_seq_batch_classify(seq_model, inputs);
        }
        if(nframes > 0){
        class = gauss_seq_update_classify(seq_model);
        report_seq_trial(err_part_model, error_file, seq_num, seq_sums,
                 desired_class, class);
        report_seq_trial(err_model, error_file, seq_num, seq_sums,
                 desired_class, class);
        }
    }
    /* report on this fold's results */
    report_overall(err_part_model, stdout, error_fname, labels);
    
    /* finish the testing fold */
    seq_data_end_cv_fold(data_model);
    }/* end fold */

    /* close parameter file */
    tmp_path = path_build(exp_path, param_fname);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- gauss model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0)
      nn_fclose(error_file, path_build(exp_path, error_fname),
        "for error file");

    /* free space */
    gauss_free (model);
    gauss_seq_free(seq_model);
    norm_free();
    seq_data_exit(data_model);
}
