/***--------------------------------------------------------------------------
* NAME:     gauss_seq_main.c
* PURPOSE:  Entry routine for Gaussian sequence classifier
*
* USAGE:    gauss_seq -<flags>
*       
* EXAMPLE:  gauss_seq --        shows all available flags
*
* SEE ALSO: gauss_seq_train.c gauss_seq_test.c gauss_seq_cv.c
*       gauss_seq_pack.c gauss_pack.c
*
* GLOBALS:  <flags> defined in gauss_seq_args.h
*
* DESCRIPTION: 
*       main    Processes command line parameters and calls
*       gauss_seq_train(), gauss_seq_test, gauss_seq_cross_valid
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_seq_main.c,v 11.4 2001/08/23 20:11:43 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_seq_main.c,v $
*-------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_seq_main.c,v 11.4 2001/08/23 20:11:43 craig Exp $";

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define NO_CV_TABLE
#include <cv_args.h>
#include <gen_seq_args.h>
#include "gauss_seq_args.h"         /* parameter definitions */
#include "gauss_seq.h"
  
int
main(int argc, char **argv)
{
    char *comment;
    int use_splits, nwindow_frames;
    /* arguments for reading defaults file */
    DEFAULTS_FILE_STRUCT *df_args;
    /* argument table with all entries */
    ARG *argtab;
    int tabsize;

    init_nnsubs(argc, argv, rcsid);
            /* get arguments */
    nnargs_combine_argtabs(gen_tab, GEN_TABSIZE, gauss_tab, GAUSS_TABSIZE,
               &argtab, &tabsize);
    argc = nnargs(argc, argv, argtab, tabsize);

    print_log_header();
            
    /* read the defaults file */
    if(strcmp(desc_fname, "") == 0){ /* desc_fname is unset */
    fatal_error("%s: -fdescribe must be set", argv[0]);
    }
    /* set the number of inputs and outputs and read the labels,
       allocating the space */
    df_args = (DEFAULTS_FILE_STRUCT *)
      nn_malloc(sizeof(DEFAULTS_FILE_STRUCT), "for the defaults file struct");
    read_defaults_file(path_build(data_path, desc_fname), df_args);
    if(df_args->noutputs == 0)
      fatal_error("noutputs in %s must be greater than 0", desc_fname);
    if(df_args->ninputs == 0)
      fatal_error("ninputs in %s must be greater than 0", desc_fname);

    /* set the number of sequences to no more than are in the data file */
    if(strcmp(data_fname, "") == 0){ /* data_fname is unset */
    fatal_error("%s: -flist must be set", argv[0]);
    }
    check_file_length(path_build(data_path, data_fname), &nsequences);
    if(nsequences == 0)
      fatal_error("%s: Empty data file %s", argv[0],
          path_build(data_path, data_fname));
    /* check the data base type */
    if (df_args->db_type != SEQ_DATA)
      fatal_error("sequence data required, check defaults file");

    /* fix feature selection arguemnts */
    if (ninputs > df_args->ninputs)
    ninputs = df_args->ninputs;
    if(features == NULL){
    /* no feature selection is specified */
    features = allocate_ivector(1);
    features[0] = 0;
    }
    ninputs = fix_features(df_args->ninputs, ninputs, features);
    nwindow_frames = fix_windows(window_width, window_frames);

    /* check flags for consistancy */
    if(train_flag && cross_valid > 2)
      warning("%s\n%s\n","-train and -cross_valid are both set.",
          " -cross_valid will be ignored");
    if(train_flag && !create_flag){
    warning("%s%s\n%s","Training for Gaussians cannot be continued.",
        "  All data must be presented at once.",
        "  -create flag will be set.");
    create_flag = 1;
    }

    /* set up cross validation arguments */
    use_splits =
      fix_cross_valid_args(&cross_valid, &cv_nfolds, &cv_nsplits, cv_splits,
               cv_train_mask, cv_test_mask, cross_valid_from_file,
               path_build(data_path, cross_valid_fname));

    /* check arguments on generic list that are ignored here */
    if(verbosity > 3){
    if(random_flag)
      warning("-random flag has no effect in %s", argv[0]);
    }

    /* make and print header comment */
    comment = make_comment_string(nnargs_get_flags(argtab, tabsize));
    if (verbosity > 1) {
    printf("%s", comment);
    }
    fflush(stdout); /* ensure log headers appear before error outputs */

    srandom(seed);  /* initialize random numbers */
    
    /* call the appropriate driver */
    if(train_flag){
    gauss_seq_train(df_args->ninputs, ninputs, features, nwindow_frames, window_frames,
            df_args->noutputs,
            normalize, nsequences, debug, verbosity, data_path,
            data_fname, norm_fname, exp_path, param_fname, df_args->output_labels,
            comment, minimum_variance, covariance_type,
            variance_type, max_ratio);
    } else if(cv_nfolds > 1) {
    gauss_seq_cross_valid(df_args->db_type, df_args->ninputs, ninputs, features, nwindow_frames,
                  window_frames, df_args->noutputs,
                  normalize, nsequences, debug, verbosity, verror,
                  cv_nfolds, cv_nsplits, use_splits, cv_splits,
                  &cv_train_mask[1], &cv_test_mask[1], cv_random,
                  data_path,
                  data_fname, norm_fname, exp_path, error_fname,
                  param_fname, df_args->output_labels, comment, covariance_type,
                  variance_type, minimum_variance, max_ratio);
    } else {
    gauss_seq_test(create_flag, df_args->db_type,
               nsequences, debug, verbosity, verror,
               data_path, data_fname, exp_path, error_fname,
               param_fname, df_args->output_labels);
    }

    /* finish up */
    print_log_footer();

    nnargs_free_argtab(argtab, tabsize);
    /* free the vectors allocated when the defaults file was read */
    free_defaults_file_struct(df_args);

    exit_nnsubs();
    exit(0);
}
