/***------------------------------------------------------------------------
 * NAME:        gauss_seq_test.c 
 * PURPOSE: driver testing for Gaussian classifier 
 *
 * SEE ALSO:    gauss_main.c gauss_pack.c
 *
 * BUGS AND LIMITATIONS:
 *
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_seq_test.c,v 11.4 2001/08/23 20:11:44 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_seq_test.c,v $
 * CREATED: 7-May-93 lck            MIT LINCOLN LABORATORY
 ********************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <nnlib.h>
#include "gauss_seq.h"
#include "gauss_seq_pack.h"
#include "gauss_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];  /* static variables */

void
gauss_seq_test(int create_flag, int db_type, int nsequences, int debg,
           int verbosity, int verror, char *data_path, char *data_fname,
           char *exp_path, char *error_fname, char *param_fname,
           char **labels)
{
    int     model, seq_model, err_model, data_model, rv, normalization=0;
    int     *features, nraw, ninputs, noutputs, model_ninputs;
    int     nwindow_frames, *window_frames;
    int     desired_class, class;
    int     n_frames, i, n_seq, n_files;
    float   *inputs, *seq_sums, *priors_list;
    char    *tmp_path;
    FILE    *error_file = NULL, *param_file;
        int     log_outputs = 1,post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());
    
    /* read the parameter file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
          "on normalization parameters in file", tmp_path);
    rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
    if(rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
          "on feature selection parameters in file", tmp_path);
    rv = window_read_paramfile(param_file, &nwindow_frames, &window_frames);
    if(rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
          "on windowing parameters in file", tmp_path);
    model = gauss_read_model(param_file);
    nn_fclose(param_file, tmp_path, "for read of model file");

    noutputs = *(int *)gauss_param(model, "noutputs", 0);
    model_ninputs = *(int *)gauss_param(model, "ninputs", 0);
    seq_model = gauss_seq_build(noutputs);
    gauss_seq_init(seq_model, model);
    seq_sums = *(float **) gauss_seq_param(seq_model, "seq_sums");
    
    if (verror > 0) {   /* open file for classification errors */
    if(create_flag)
      error_file = nn_fopen(path_build(exp_path, error_fname),
                "w", "for classification errors");
    else
      error_file = nn_fopen(path_build(exp_path, error_fname),
                "a", "for classification errors");
    }
    /* prepare error reporting */
    err_model = report_build (verbosity, verror, noutputs);
    /* log_outputs*/
    report_init (err_model, db_type, model_ninputs, LOG_LIKE, log_outputs, post_sum_to_one);
    
    /* read data */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
                   ninputs, noutputs, features, normalization,
                   labels, verbosity, nwindow_frames,
                   window_frames);
    nsequences = seq_data_get_nfiles(data_model);
    
    /******** prepare gauss_model for testing **************/
    /* gauss should adjust for uniform priors for the frames */
    priors_list = allocate_vector(noutputs);
    for(i = 0; i < noutputs; i++) priors_list[i] = 1.;
    gauss_test_build (model);
    /* adjust for uniform priors */
    gauss_test_init (model, 1, priors_list);
    
    for(n_files = 0; n_files < nsequences; n_files++){
    if(seq_data_get_next_file(data_model) == -1) break;
    desired_class = seq_data_get_class(data_model);
    gauss_seq_init(seq_model, model);
    for(n_frames = 0;; n_frames++){
        inputs = seq_data_get_next_frame(data_model);
        if(inputs == NULL) break;
        gauss_seq_batch_classify(seq_model, inputs);
    }

    if(n_frames > 0){
        /* find the class for this sequence */
        class = gauss_seq_update_classify(seq_model);
        n_seq = seq_data_get_file_num(data_model);
        report_seq_trial(err_model, error_file, n_seq, seq_sums,
                 desired_class, class);
    }
    }

    /* finish the error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    if (verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }    

    /* free space and close files */
    gauss_test_free (model);

    seq_data_exit(data_model);
    gauss_free (model);
    gauss_seq_free(seq_model);
    norm_free();
}
