/***---------------------------------------------------------------------------
* NAME:     gauss.c 
* PURPOSE:  driver for Gaussian classifier 
*
* SEE ALSO: gauss_main.c gauss_pack.c
*
* FUNCTIONS:
*   gauss   driver for gauss classifier
*       handles model creation, training and testing of data
*
* BUGS AND LIMITATIONS:
*
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_train.c,v 11.4 2001/08/23 20:11:44 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gauss/gauss_train.c,v $
* CREATED: 26-MAR-91 dan            MIT LINCOLN LABORATORY
****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include "gauss.h"
#include "gauss_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];  /* static variables */

void
gauss_train(int create_flag, int nraw, int ninputs, int *features,
        int noutputs, int normalize, int npatterns, int debg,
        int verbosity, int random_flag, char *data_path, char *data_fname,
        char *norm_fname, char *exp_path, char *param_fname,
        char *comment, int adjust_priors, int priors_npatterns,
        float *priors_list, float minimum_variance, int covariance_type,
        int variance_type, float max_ratio)
{
    int     model, data_model, rv, normalization = 0;
    int     desired_class;
    float   *inputs;
    char    *tmp_path;
    FILE    *param_file;

    strcpy(prog_name, get_program_name());

    if (create_flag) {
    /* allocate and initialize model structure */
    model=gauss_build(ninputs, noutputs, variance_type, covariance_type);
    gauss_init(model, features, minimum_variance);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);
    } else {        /* retrain */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
        fatal_error("%s: read error %d %s %s", prog_name, rv,
            "on normalization parameter file", tmp_path);

    model = gauss_read_model(param_file);
    nn_fclose(param_file, tmp_path, "for read of model file");
    verify_model_params (model, 
        ninputs, *(int *) gauss_param(model, "ninputs", 0),
        features, *(int **) gauss_param(model, "features", 0));
    }
            /* use features from the model */
    features = *(int **) gauss_param(model, "features", 0);
    ninputs = *(int *) gauss_param(model, "ninputs", 0);

    /* print various debugging and information messages before starting
     * actual training or testing 
     */
    display_model_params(model, ninputs, normalization, features, verbosity);

    if(adjust_priors){ /* priors adjustment by sampling */
    data_model = class_data_read(data_path, data_fname, npatterns,
                     nraw, ninputs, noutputs, features,
                     normalization, verbosity);
    if(random_flag) class_data_random_init(data_model);
    class_data_priors_init(data_model, priors_list, priors_npatterns);
    } else {
    data_model = class_data_single_init(data_path, data_fname, npatterns,
                        nraw, ninputs, noutputs, features,
                        normalization, verbosity);
    }
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
    gauss_batch_ctrain(model, inputs, desired_class);
    }
    gauss_batch_update(model, max_ratio);
    class_data_exit(data_model);

    /********** store model in param file **********************/
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "gauss model\n");
    gauss_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- gauss model saved in \"%s\"\n", tmp_path);

    norm_free();
    gauss_free (model);
}
