/***--------------------------------------------------------------------------
* NAME:     generic.c
* PURPOSE:  driver for generic classifier
*
* SEE ALSO:     generic_pack.c
*
* FUNCTIONS:
*   generic driver for generic classifier
*       handles model creation, training and testing of data
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/generic/generic_cv.c,v 11.5 2001/08/23 20:11:48 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/generic/generic_cv.c,v $
* CREATED: 18-JUN-91 lck                             MIT LINCOLN LABORATORY
-----------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "generic.h"
#include "generic_pack.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

static char prog_name[MAXSTR];  /* static variables */

float
generic_cross_valid(int map_flag, int nraw, int ninputs,
            int *features, int noutputs, int normalize,
            int npatterns, int debg, int verbosity, int verror,
            int nepochs, int random_flag, int cv_nfolds, int cv_nsplits,
            int use_cv_splits, int *cv_splits, char **cv_train_mask,
            char **cv_test_mask, char *data_path, char *data_fname,
            char *norm_fname, char *exp_path, char *error_fname,
            char *param_fname, char **labels, char *comment,
            int adjust_train, int adjust_test, int priors_npatterns,
            float *train_priors_list, float *test_priors_list,
            int cv_random, int log_outputs,
            int report_cost, int roc_target, float roc_lower,
            float roc_upper)
{
    int     model, err_model, err_part_model, data_model;
    int     normalization=0, epoch;
    int     desired_class, class;
    int     n_read, n_trials, trial_num;
    int     fold;
    float   *inputs, *outputs, percent_error;
    char    *tmp_path;
    FILE    *error_file = NULL, *param_file;
    int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());

    /* allocate and initialize model structure */
    model = generic_build(ninputs, noutputs);
    generic_init(model, features);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    /* use outputs from the model */
    outputs = *(float **) generic_param(model, "outputs");

    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* prepare results reporting */
    if (verror > 0) {   /* open file for classification errors */
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    }
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, report_cost,
         log_outputs, post_sum_to_one);
    report_roc_init(err_model, roc_target, roc_lower, roc_upper);
    err_part_model = report_build (verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    data_model =
      class_data_read(data_path, data_fname, npatterns, nraw, ninputs,
              noutputs, features, normalization, verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
               cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_train)
      class_data_priors_init(data_model, train_priors_list, priors_npatterns);

    /* print out the normalization once for all cross valid models */
    if(param_fname != NULL){
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    }

    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);
    generic_init(model, features);

    /* training */
    class_data_start_cv_fold(data_model, fold, TRUE);
    n_trials = class_data_get_npatterns(data_model);

    /*****************epoch training*****************************/
    report_init(err_part_model, map_flag, ninputs, report_cost,
            log_outputs, post_sum_to_one);
    report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);
    report_epoch_header(err_part_model, stdout, n_trials);

    for(epoch = 1; epoch <= nepochs; epoch++) {
        class_data_epoch_init(data_model);

        for(n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;

        class = generic_inc_ctrain(model, inputs,
                       desired_class);
        report_class_trial(err_part_model, error_file, inputs,
                   outputs, desired_class, class,
                   class_data_get_input_num(data_model));
        }
        report_epoch(err_part_model, stdout, epoch);
        report_init(err_part_model,map_flag, ninputs, report_cost,
            log_outputs, post_sum_to_one);
        report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);
    }

    /********************** end epoch training **************************/
    /********************* single pass training *************************/
    class_data_set_n_read_zero(data_model);
    /* set_n_read_zero is only needed in this driver,
       do not use in a real driver */

    for(n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;

        generic_inc_ctrain(model, inputs, desired_class);
    }
    /************** end single pass training *******************/

    class_data_end_cv_fold(data_model);

    /* print out the current model */
    if(param_fname != NULL){
        fprintf(param_file, "generic model %d\n", fold);
        generic_write_model(model, param_file);
    }

    /* reinitialize reporting for testing */
    report_init(err_part_model, map_flag, ninputs, report_cost,
            log_outputs, post_sum_to_one);
    report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);

    /* test */
    if(adjust_test)
      generic_test_init(model, adjust_test, test_priors_list);

    class_data_start_cv_fold(data_model, fold, FALSE);
    n_trials = class_data_get_npatterns(data_model);
    for (n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;

        class = generic_classify(model, inputs);
        trial_num = class_data_get_input_num(data_model);
        report_class_trial(err_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
        report_class_trial(err_part_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
    }
    /* report on testing */
    report_overall(err_part_model, stdout, error_fname, labels);

    class_data_end_cv_fold(data_model);
    } /* end fold */

    /* finish up the parameter file output */
    if(param_fname != NULL){
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    }

    /* free space and close files */
    class_data_exit(data_model);

    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);
    
    if (verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
        "for error file");
    }
    generic_free(model);
    norm_free();

    return (percent_error);
}
