/***---------------------------------------------------------------------------
* NAME:         generic_pack.c
* PURPOSE:      Functions specific to generic.c
*
* SEE ALSO: generic.c generic_main.c
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   generic_build       Allocate storage for model parameters.
*   generic_init        Initialize model parameters.
*   generic_free        Deallocates model storage
*   generic_read_model  Initializes model from file
*   generic_write_model Writes model to output file
*   generic_param       Returns pointers to elements in model
*   generic_inc_ctrain  Does training on a vector
*   generic_classify    Classifies a vector using generic algortihm
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/generic/generic_pack.c,v 11.4 2001/08/23 20:11:49 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/generic/generic_pack.c,v $
* CREATED: 12-Dec-90 lck                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */


/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define inc_pack_h_internal_types
#include "generic_pack.h"

                /* constant definitions */
#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))

                /* globals */

                /* Internal Globals */
static GENERIC_PARAMS **genpp;  /* array of parameters for each model */
static int      max_models = 0; /* current length of genpp */


/*****************************************************************************
* NAME: generic_build(ninputs,noutputs)
*   ninputs - number of dimensions in training vectors
*   noutputs - number of classes
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
*****************************************************************************/
int
generic_build(int ninputs, int noutputs)
{
    register int    new_model;
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

                /* find entry in model array for new model */
    new_model = model_get_number((void ***)&genpp,
                 &max_models, sizeof(GENERIC_PARAMS));
    gpp = genpp[new_model];

                    /* set fields which are array sizes */
    gpp->ninputs = ninputs;
    gpp->noutputs = noutputs;

                    /* allocate space for vectors */
    gpp->features = allocate_ivector(ninputs+1);
    gpp->outputs = allocate_vector(noutputs);
    gpp->class_cnt = allocate_vector(noutputs);
    gpp->priors_list = allocate_vector(noutputs);

    return new_model;
}
/*****************************************************************************
* NAME: generic_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
*****************************************************************************/
void
generic_init(int model, int *nn_features)
{
    int             i;
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

    gpp = genpp[model];
                    /* copy feature data into feature list */
    for (i = 0; i <= nn_features[0]; i++)
    gpp->features[i] = nn_features[i];

    for (i = 0; i < gpp->noutputs; i++)
      gpp->class_cnt[i] = 0.;
    gpp->total_npatterns = 0.;
    gpp->adjust_priors = 0;
}
/*****************************************************************************
* NAME: generic_test_init()
* DESCRIPTION:
*   sets model parameters needed for testing, not for training
*
* EXAMPLE:
*****************************************************************************/
void generic_test_init(int model, int adjust_priors, float *priors_list)
{
    int i;
    GENERIC_PARAMS *gpp;

    gpp = genpp[model];
    gpp->adjust_priors = adjust_priors;
    if(adjust_priors)
      for(i = 0; i < gpp->noutputs; i++)
    gpp->priors_list[i] = priors_list[i];
}

/*****************************************************************************
* NAME: generic_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
*****************************************************************************/
void
generic_free(int model)
{
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

    gpp = genpp[model];
                    /* free vectors and matrices */
    deallocate_ivector(gpp->features);
    deallocate_vector(gpp->class_cnt);
    deallocate_vector(gpp->priors_list);

                    /* free structure itself */
    nn_free((char *) gpp, "for generic parameter structure");
    genpp[model] = NULL;
}

/*****************************************************************************
* NAME: generic_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*       allocates space for and loads data into vectors and matricies
*
*       On error from param_getblk, returns the error value
*       (there is already an error check on the allocation, so we don't
*       have to do that here)
*
* EXAMPLE:
*****************************************************************************/
int
generic_read_model(FILE *fp)
{
    int             model;
    int             rv, ninputs, noutputs;
    float   total_npatterns;
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

    param_setvarptr("ninputs", &ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &noutputs, paramtab, TABSIZE);
    param_setvarptr("total_npat",(int *)&total_npatterns, paramtab, TABSIZE);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

                    /* allocate space for model and vectors */
    model = generic_build(ninputs, noutputs);
    gpp = genpp[model];

    param_setvarptr("features", &gpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", gpp->ninputs+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("class_cnt", (int *)&gpp->class_cnt[0],
            paramtab_l, TABSIZE_L);
    param_setdim("class_cnt", gpp->noutputs, 0, paramtab_l, TABSIZE_L);

    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

                    /* set any remaining parameters */
    gpp->total_npatterns = total_npatterns;

    return (model);
}

/*****************************************************************************
* NAME: generic_write_model()
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
*
* EXAMPLE:
*
******************************************************************************/
int
generic_write_model(int model, FILE *fp)
{
    int             rv;
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

    gpp = genpp[model];
                    /* save scalars */
    param_setvarptr("ninputs", &gpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &gpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("total_npat", (int *)&gpp->total_npatterns,
            paramtab, TABSIZE);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

                    /* save vectors and matrices */
    param_setvarptr("features", &gpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", gpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("class_cnt", (int *)&gpp->class_cnt[0],
            paramtab_l, TABSIZE_L);
    param_setdim("class_cnt", gpp->noutputs, 0, paramtab_l, TABSIZE_L);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    return (0);
}


/************************************************************************
* NAME: generic_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/

void *
generic_param(int model, char *param_name)
{
    void           *param_address;
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

    gpp = genpp[model];
    if (PEQ("features"))
    param_address = &gpp->features;
    else if (PEQ("ninputs"))
    param_address = &gpp->ninputs;
    else if (PEQ("noutputs"))
    param_address = &gpp->noutputs;
    else if (PEQ("outputs"))
    param_address = &gpp->outputs;
    else if (PEQ("class_cnt"))
      param_address = &gpp->class_cnt;
    else if (PEQ("total_npatterns"))
      param_address = &gpp->total_npatterns;
    else {
    warning("generic_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}


/*****************************************************************************
* NAME: generic_inc_ctrain()
* DESCRIPTION:
*       incremental training for a single input vector
* EXAMPLE:
*
*****************************************************************************/
int
generic_inc_ctrain(int model, float *inputs, int desired_class)
{
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

    gpp = genpp[model];
    gpp->class_cnt[desired_class] += 1.;
    gpp->total_npatterns += 1.;
    return (rand() % gpp->noutputs);
}
/*****************************************************************************
* NAME: generic_classify()
* DESCRIPTION:
*       Implements the generic classifier to return an integer class.
*
* EXAMPLE:
*
*****************************************************************************/
int
generic_classify(int model, float *inputs)
{
    GENERIC_PARAMS *gpp;    /* temporary pointer to genpp[] */

    gpp = genpp[model];
#ifdef notdef
    /* adjust priors on outputs, continuous output classifiers only */
    if(gpp->adjust_priors){
    int i;
    float total;
    /* adjust priors */
    for(i = 0; i < gpp->noutputs; i++){
        gpp->outputs[i] *= gpp->priors_list[i];
        gpp->outputs[i] /= gpp->class_cnt[i]/gpp->total_npatterns;
    }
    /* scale adjusted outputs to sum to 1 */
    total = 0.;
    for(i = 0; i < gpp->noutputs; i++) total += gpp->outputs[i];
    for(i = 0; i < gpp->noutputs; i++) gpp->outputs[i] /= total;
    }
#endif /* notdef */

    return (rand() % gpp->noutputs);
}
