/***--------------------------------------------------------------------------
* NAME:     generic.c
* PURPOSE:  driver for training generic classifier
*
* SEE ALSO:     generic_pack.c
*
* FUNCTIONS:
*   generic driver for generic classifier
*       handles model creation and training
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/generic/generic_train.c,v 11.4 2001/08/23 20:11:49 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/generic/generic_train.c,v $
* CREATED: 18-JUN-91 lck                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "generic.h"
#include "generic_pack.h"

#define MAXSTR (256)

static char prog_name[MAXSTR];  /* static variables */

void
generic_train(int create_flag, int map_flag, int nraw, int ninputs,
          int *features, int noutputs, int normalize, int npatterns,
          int debg, int verbosity, int verror, int nepochs,
          int random_flag, char *data_path, char *data_fname,
          char *norm_fname, char *exp_path, char *error_fname,
          char *param_fname, char **labels, char *comment, int adjust_priors,
          int priors_npatterns, float *priors_list, int log_outputs)
{
    int     model, err_model, err_part_model, data_model;
    int     rv, normalization=0, epoch;
    int     desired_class, class;
    int     n_read, n_trials, trial_num;
    float   *inputs, *outputs;
    char    *tmp_path, *open_type;
    FILE    *error_file = NULL, *param_file;
    int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());

    if (create_flag) {
    /* allocate and initialize model structure */
    model = generic_build(ninputs, noutputs);
    generic_init(model, features);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);
    } else {        /* retrain */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
              "on normalization parameter file", tmp_path);

    model = generic_read_model(param_file);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s", 
              prog_name, model, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of model file");
    verify_model_params(model,
                ninputs, *(int *) generic_param(model, "ninputs"),
                features,
                *(int **) generic_param(model, "features"));
    }
    /* use features from the model */
    features = *(int **) generic_param(model, "features");
    ninputs = *(int *) generic_param(model, "ninputs");
    outputs = *(float **) generic_param(model, "outputs");

    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* set up for reporting the results */
    if (verror > 0) {   /* open file for classification errors */
    if (create_flag) open_type = "w";
    else open_type = "a";
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  open_type, "for classification errors");
    }
    /* do remaining initializations and allocations */
    /******************* epoch style *****************************/
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, SQ_ERR,
         log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
                 ninputs, noutputs, features, normalization,
                 verbosity);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
    class_data_priors_init(data_model, priors_list, priors_npatterns);

    /* get the number of patterns per epoch */
    n_trials = class_data_get_npatterns(data_model);
    report_epoch_header(err_model, stdout, n_trials);

    for (epoch = 1; epoch <= nepochs; epoch++) {
    class_data_epoch_init(data_model);
    report_init (err_part_model, map_flag, ninputs, SQ_ERR,
             log_outputs, post_sum_to_one);

    for (n_read = 0; n_read < n_trials; n_read++) {
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        trial_num = class_data_get_input_num(data_model);
        class = generic_inc_ctrain(model, inputs, desired_class);
        report_class_trial(err_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
        report_class_trial(err_part_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
    }
    report_epoch(err_part_model, stdout, epoch);
    }
    report_last_epoch(err_part_model, stdout, epoch-1);
    class_data_exit(data_model);
    /****************** end of epoch training *****************************/

    /********************* single pass training ***************************/
    if(adjust_priors){
    /* read data and initialize data handling routines */
    data_model =
      class_data_read(data_path, data_fname, npatterns, nraw, ninputs,
              noutputs, features, normalization, verbosity);
    if(random_flag) class_data_random_init(data_model);
    class_data_priors_init(data_model, priors_list, priors_npatterns);
    } else { /* single pass training, no priors */
    data_model = class_data_single_init(data_path, data_fname, npatterns,
                        nraw, ninputs, noutputs, features,
                        normalization, verbosity);
    }
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
    generic_inc_ctrain(model, inputs, desired_class);
    }
    class_data_exit(data_model);
    /***************** end single pass training *************************/

    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "generic model\n");
    generic_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);

    /********* epoch classifiers **********/
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_part_model);
    report_free(err_model);
    /**************************************/
    
    if (verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }

    /* free space */
    norm_free();
    generic_free(model);
}
