/***--------------------------------------------------------------------------
 * NAME:    gmix_cv.c
 * PURPOSE: cross_validation driver for gaussian mixture classifier
 *
 * SEE ALSO:    gmix_pack.c
 *
 * FUNCTIONS:
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_cv.c,v 11.6 2001/08/23 20:11:52 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_cv.c,v $
 * CREATED: 18-JUN-91 lck                             MIT LINCOLN LABORATORY
 ****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "gmix.h"
#include "gmix_pack.h"

#define ABS(a) ((a) >= 0 ? (a) : -(a))

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

static char prog_name[MAXSTR];  /* static variables */

float
gmix_cross_valid(int map_flag, int nraw, int ninputs, int *pfeatures,
         int noutputs, int npatterns, int debg, int verbosity,
         int verror, int nepochs, int random_flag, int cv_nfolds,
         int cv_nsplits, int use_cv_splits, int *cv_splits,
         char **cv_train_mask, char **cv_test_mask, char *data_path,
         char *data_fname, char *exp_path, char *error_fname,
         char *param_fname, char **labels, char *comment,
         int adjust_train, int adjust_test, int priors_npatterns,
         float *train_priors_list, float *test_priors_list,
         int cv_random, int log_outputs,
         int report_cost, int roc_target, float roc_lower,
         float roc_upper, 
         int tied_flag,
         int grand_flag, int full_flag, float minimum_variance,
         float max_ratio, float var_spread, char *cluster_param_fname,
         float stop_percentage)
{
    int     model, err_model, err_part_model, rv, data_model;
    int     normalization=0, epoch;
    int     desired_class, class;
    int     n_read, n_trials, trial_num;
    int     fold;
    float   *inputs, *outputs, percent_error;
    float   ave_logl, last_ave_logl, percent_change;
    char    data_fullname[MAXSTR], *tmp_path, clus_path[MAXSTR];
    FILE    *error_file, *param_file, *clus_file;
    int     *features, *ngauss_param = NULL;
    int     post_sum_to_one = 0;
    int     i;

    strcpy(prog_name, get_program_name());
    
    /* put together the full path to the data file*/
    strcpy(data_fullname, path_build(data_path, data_fname));
    
    /* read centers from cluster file */
    tmp_path = path_build(exp_path, cluster_param_fname);
    strncpy(clus_path, tmp_path, MAXSTR -1);
    
    clus_file = nn_fopen(clus_path, "r",
             "for read of cluster parameters");
    if(verbosity > 1) {
    printf("Reading cluster parameters from \"%s\"\n", clus_path);
    fflush(stdout);
    }
    model = norm_read_paramfile(clus_file, &normalization);
    if(model < 0)
      fatal_error("gmix: read error %d %s %s",
          model, "on normalization parameter file", clus_path);
    rv = gmix_read_centers(clus_path, clus_file, &ninputs, &noutputs,
               &ngauss_param, pfeatures, tied_flag);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
          "on cluster parameter file", tmp_path);
    /* if we are not doing cross validation, this is all we need from
       the cluster file */
    
    model = gmix_build(ninputs, noutputs, ngauss_param, tied_flag,
               grand_flag, full_flag);
    gmix_init(model, pfeatures, minimum_variance, max_ratio);
    gmix_initialize_cluster_centers(model, rv, var_spread);
    cluster_free(rv);
    deallocate_ivector(ngauss_param);
    /* really want to verify the cluster features */
    verify_model_params (model, ninputs,
             *(int *) gmix_param(model, "ninputs", 0, 0),
             pfeatures,
             *(int **) gmix_param(model, "features", 0, 0));
    
    /* use features from the model */
    features = *(int **) gmix_param(model, "features", 0, 0);
    ninputs = *(int *) gmix_param(model, "ninputs", 0, 0);
    outputs = *(float **) gmix_param(model, "outputs", 0, 0);
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {   /* open file for classification errors */
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, report_cost,
         log_outputs, post_sum_to_one);
    report_roc_init(err_model, roc_target, roc_lower, roc_upper);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
                 ninputs, noutputs, features, normalization,
                 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
               cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_train)
      class_data_priors_init(data_model, train_priors_list, priors_npatterns);
    
    /* print out the normalization parameters */
    if(param_fname != NULL){
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    }

    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);
    class_data_start_cv_fold(data_model, fold, TRUE);
    n_trials = class_data_get_npatterns(data_model);
    
    /* read in the next set of cluster centers */
    if(fold != 0){
        gmix_free(model);
        rv = gmix_read_centers(clus_path, clus_file, &ninputs,
                   &noutputs, &ngauss_param, pfeatures,
                   tied_flag);
        if (rv < 0)
          fatal_error("%s: read error %d %s %s", prog_name, rv,
              "on cluster parameter file", tmp_path);
        model = gmix_build(ninputs, noutputs, ngauss_param,
                   tied_flag, grand_flag, full_flag);
        gmix_init(model, pfeatures, minimum_variance, max_ratio);
        gmix_initialize_cluster_centers(model, rv, var_spread);
        cluster_free(rv);
        deallocate_ivector(ngauss_param);
        /* use newly allocated outputs array */
        outputs = *(float **) gmix_param(model, "outputs", 0, 0);
    }
    
    /* epoch training */
    report_init(err_part_model, map_flag, ninputs, report_cost,
            log_outputs, post_sum_to_one);
    report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);
    report_epoch_header(err_part_model, stdout, n_trials);

    last_ave_logl = 0.;
    for(epoch = 0, percent_change = MAXFLOAT;
        (epoch < nepochs) && (percent_change > stop_percentage); epoch++) {
        class_data_epoch_init(data_model);
        for(n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
            
        class = gmix_batch_ctrain(model, inputs, desired_class);
        report_class_trial(err_part_model, error_file, inputs,
                   outputs, desired_class, class,
                   class_data_get_input_num(data_model));
        }
        gmix_batch_update(model);

        ave_logl = report_get_avg_cost(err_part_model);
        if(last_ave_logl != 0.){
        percent_change = 100.0 * (ave_logl - last_ave_logl)/
          ABS(last_ave_logl);
        if(verbosity > 3){
            if(percent_change < 0.0){
            warning("gmix: diverging\n%s %g\n%s %g\n%s %g",
                "Last ave log likelihood", last_ave_logl,
                "Current log_likelihood", ave_logl,
                "Percentage change", percent_change);
            }
        }
        }
        percent_change = ABS(percent_change);
        last_ave_logl = ave_logl;
        report_epoch(err_part_model, stdout, epoch);
        report_init(err_part_model, map_flag, ninputs, report_cost,
            log_outputs, post_sum_to_one);
        report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);
    }
    
    /* print out the current model */
    if(param_fname != NULL){
        fprintf(param_file, "gmix model\n");
        if(gmix_write_model(model, param_file) < 0)
          warning("did not write whole parameter file");
    }
    
    /* initialize error reporting */
    report_init(err_part_model, map_flag, ninputs, report_cost,
            log_outputs, post_sum_to_one);
    report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);

    /* finish the training fold and start the testing fold */
    class_data_end_cv_fold(data_model);
    class_data_start_cv_fold(data_model, fold, FALSE);
    n_trials = class_data_get_npatterns(data_model);

    /* set up the model for testing */
    gmix_test_init(model, adjust_test, test_priors_list);

    for (n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        class = gmix_classify(model, inputs);
        trial_num = class_data_get_input_num(data_model);

        if(report_cost == ROC_COST){
          /* turn these into linear */
          for(i = 0; i < noutputs; i++){
        outputs[i] = nn_exp10(outputs[i]);
          }
        }
        report_class_trial(err_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
        report_class_trial(err_part_model, error_file, inputs, outputs,
                   desired_class, class, trial_num);
    }
    /* report on testing */
    report_overall(err_part_model, stdout, error_fname, labels);
    
    /* finish the testing fold */
    class_data_end_cv_fold(data_model);
    } /* end fold */
    /* close the parameter file and the cluster file */
    if(param_fname != NULL){
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    }
    nn_fclose(clus_file, clus_path, "for read of clusters");
    
    class_data_exit(data_model);
    
    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);

    if (verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }
    /* free space */
    gmix_free(model);
    norm_free();

    return (percent_error);
}
