/***--------------------------------------------------------------------------
* NAME:         gmix_pack.c
* PURPOSE:      Functions specific to gmix.c
*
* SEE ALSO: gmix.c gmix_main.c
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   gmix_build      Allocate storage for model parameters.
*   gmix_init       Initialize model parameters.
*   gmix_free       Deallocates model storage
*   gmix_read_model     Initializes model from file
*   gmix_write_model    Writes model to output file
*   gmix_param      Returns pointers to elements in model
*   gmix_batch_ctrain   Does training on a vector
*   gmix_batch_update   Updates parameters of model after training
*   gmix_classify       Classifies a vector using gaussian mixtures
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_pack.c,v 11.3 2001/08/23 20:11:53 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_pack.c,v $
* CREATED: 12-Dec-90 lck                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include <math.h>
#define gmix_pack_h_internal_types
#include "gmix_pack.h"

                /* constant definitions */
#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))

                /* Internal Globals */
static GMIX_PARAMS **gmixpp;    /* array of parameters for each model */
static int      max_models = 0; /* current length of gmixpp */

static int check_minimum_var(float *covar_entry, float minimum_variance);

#define SMALLEST 1.e-30

/*****************************************************************************
* NAME: gmix_build(ninputs,noutputs,ngauss_param,
*          tied_flag, grand_flag, full_flag)
*   ninputs - number of dimensions in training vectors
*   noutputs - number of classes
*   ngauss_param - number of components in each gaussian mixture
*           (for number of weights and centers)
*           (when tied, number of centers is ngauss_param[0])
*   tied_flag - 1 if tied mixtures, 0 if per class mixtures
*   grand_flag - 2 if grand overall covariances
*       1 if grand per class covariances
*       0 if per mixture component covariances
*   full_flag - 1 if full covariance matrices, 0 if diagonal
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
*****************************************************************************/
int gmix_build(int ninputs, int noutputs, int *ngauss_param, int tied_flag,
           int grand_flag, int full_flag)
{
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */
    register int    new_model;
    int i, j;

    new_model = model_get_number((void ***)&gmixpp,
                 &max_models, sizeof(GMIX_PARAMS));

    gmpp = gmixpp[new_model];

    /* set mode flags */
    gmpp->tied_flag = tied_flag;
    gmpp->grand_flag = grand_flag;
    gmpp->full_flag = full_flag;
                    /* set fields which are array sizes */
    gmpp->ninputs = ninputs;
    gmpp->noutputs = noutputs;

                    /* allocate space for vectors */
    gmpp->features = allocate_ivector(ninputs+1);
    gmpp->class_cnt = allocate_vector(noutputs);
    gmpp->log_class_priors = allocate_vector(noutputs);
    gmpp->outputs = allocate_vector(noutputs);

    gmpp->class_mix = (struct class_gauss **)allocate_pvector(noutputs);
    for(i = 0; i < noutputs; i++){
    gmpp->class_mix[i] =
      (struct class_gauss *)nn_malloc(sizeof(struct class_gauss),
                      "for class_mix structures");
    gmpp->class_mix[i]->ngauss = ngauss_param[i];
    if(ngauss_param[i] <= 0){
        gmpp->class_mix[i]->gaussian = NULL;
    } else {
        gmpp->class_mix[i]->gaussian =
          (struct gauss_desc **)allocate_pvector(ngauss_param[i]);
        for(j = 0; j < ngauss_param[i]; j++){
        gmpp->class_mix[i]->gaussian[j] =
          (struct gauss_desc *)nn_malloc(sizeof(struct gauss_desc),
                         "for holding weights");
        }
    }
    }
    gmpp->ngauss = ngauss_param[0];

    /* initialize some things to NULL, set them only if needed */
    gmpp->inv_covar = NULL; /* needed for grand==2, full */
    gmpp->new_covar = NULL; /* needed for grand==2, full */
    gmpp->diag_inv_var = NULL; /* needed for grand==2, diag */
    gmpp->diag_new_var = NULL; /* needed for grand==2, diag */
    gmpp->tied_gaussian = NULL;
    for(i = 0; i < noutputs; i++){
        gmpp->class_mix[i]->inv_covar = NULL;/* untied, grand == 1, full*/
        gmpp->class_mix[i]->new_covar = NULL;/* untied, grand == 1, full*/
        gmpp->class_mix[i]->diag_inv_var = NULL;/* untied, grand == 1, diag*/
        gmpp->class_mix[i]->diag_new_var = NULL;/* untied, grand == 1, diag*/
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        gmpp->class_mix[i]->gaussian[j]->means = NULL;/* untied */
        gmpp->class_mix[i]->gaussian[j]->means = NULL;/* untied */
        /* untied, grand==0, full*/
        gmpp->class_mix[i]->gaussian[j]->inv_covar = NULL;
        gmpp->class_mix[i]->gaussian[j]->new_covar = NULL;
        /* untied, grand==0, diag*/
        gmpp->class_mix[i]->gaussian[j]->diag_inv_var = NULL;
        gmpp->class_mix[i]->gaussian[j]->diag_new_var = NULL;
        
    }
    }

    if(tied_flag){
    if(gmpp->ngauss <= 0){
        gmpp->tied_gaussian = NULL;
    } else {
        gmpp->tied_gaussian =
          (struct gauss_desc **)allocate_pvector(gmpp->ngauss);
        /* allocate space for means */
        for(i = 0; i < gmpp->ngauss; i++){
        gmpp->tied_gaussian[i] =
          (struct gauss_desc *)nn_malloc(sizeof(struct gauss_desc),
                         "for holding means");
        gmpp->tied_gaussian[i]->means = allocate_vector(ninputs);
        gmpp->tied_gaussian[i]->new_means = allocate_vector(ninputs);
        }
    }
    /* initialize things that might not be allocated */
    for(i = 0; i < gmpp->ngauss; i++){
        gmpp->tied_gaussian[i]->inv_covar = NULL;
        gmpp->tied_gaussian[i]->new_covar = NULL;
        gmpp->tied_gaussian[i]->diag_inv_var = NULL;
        gmpp->tied_gaussian[i]->diag_new_var = NULL;
    }
    /* allocate space for variances */
    if(grand_flag){
        if(full_flag){
        gmpp->inv_covar = allocate_matrix(ninputs, ninputs);
        gmpp->new_covar = allocate_matrix(ninputs, ninputs);
        } else {
        gmpp->diag_inv_var = allocate_vector(ninputs);
        gmpp->diag_new_var = allocate_vector(ninputs);
        }
    } else {
        for(i = 0; i < gmpp->ngauss; i++){
        if(full_flag){
            gmpp->tied_gaussian[i]->inv_covar =
              allocate_matrix(ninputs, ninputs);
            gmpp->tied_gaussian[i]->new_covar = 
              allocate_matrix(ninputs, ninputs);
        } else {
            gmpp->tied_gaussian[i]->diag_inv_var =
              allocate_vector(ninputs);
            gmpp->tied_gaussian[i]->diag_new_var = 
              allocate_vector(ninputs);
        }
        }
    }
    } else { /* per class mixtures */
    /* allocate space for means */
    for(i = 0; i < noutputs; i++){
        for(j = 0; j < ngauss_param[i]; j++){
        gmpp->class_mix[i]->gaussian[j]->means =
          allocate_vector(ninputs);
        gmpp->class_mix[i]->gaussian[j]->new_means =
          allocate_vector(ninputs);
        }
    }
    if(grand_flag == 2){ /* grand overall */
        if(full_flag){
        gmpp->inv_covar = allocate_matrix(ninputs, ninputs);
        gmpp->new_covar = allocate_matrix(ninputs, ninputs);
        } else {
        gmpp->diag_inv_var = allocate_vector(ninputs);
        gmpp->diag_new_var = allocate_vector(ninputs);
        }
    } else if(grand_flag == 1){ /* grand per class */
        for(i = 0; i < noutputs; i++){
        if(full_flag){
            gmpp->class_mix[i]->inv_covar =
              allocate_matrix(ninputs, ninputs);
            gmpp->class_mix[i]->new_covar =
              allocate_matrix(ninputs, ninputs);
        } else {
            gmpp->class_mix[i]->diag_inv_var =
              allocate_vector(ninputs);
            gmpp->class_mix[i]->diag_new_var =
              allocate_vector(ninputs);
        }
        }
    } else { /* per mixture component */
        for(i = 0; i < noutputs; i++){
        for(j = 0; j < ngauss_param[i]; j++){
            if(full_flag){
            gmpp->class_mix[i]->gaussian[j]->inv_covar =
              allocate_matrix(ninputs, ninputs);
            gmpp->class_mix[i]->gaussian[j]->new_covar =
              allocate_matrix(ninputs, ninputs);
            } else {
            gmpp->class_mix[i]->gaussian[j]->diag_inv_var =
              allocate_vector(ninputs);
            gmpp->class_mix[i]->gaussian[j]->diag_new_var =
              allocate_vector(ninputs);
            }
        }
        }
    }
    }
    return new_model;
}
/*****************************************************************************
* NAME: gmix_init(model, nn_features, minimum_variance)
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
*****************************************************************************/
void gmix_init(int model, int *nn_features, float minimum_variance,
           float eigen_ratio)
{
    int             i, j;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */
    struct class_gauss *cg;

    gmpp = gmixpp[model];

    gmpp->minimum_variance = minimum_variance;
    gmpp->eigen_ratio = eigen_ratio;

                    /* copy feature data into feature list */
    for (i = 0; i <= nn_features[0]; i++)
      gmpp->features[i] = nn_features[i];

    for(i = 0; i < gmpp->noutputs; i++){
    gmpp->log_class_priors[i] = 0.;
    gmpp->class_cnt[i] = 0.;
    }
    gmpp->total_npatterns = 0.;

    gmpp->npatterns = 0;
    for(i = 0; i < gmpp->noutputs; i++){
    cg = gmpp->class_mix[i];
    gmpp->class_mix[i]->npatterns = 0;
    gmpp->class_mix[i]->ave_logl = 0.;
    for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        gmpp->class_mix[i]->gaussian[j]->weighted_gauss_output = 0.;
        cg->gaussian[j]->weight = 1./cg->ngauss;
        cg->gaussian[j]->log_weight =
          (float)log10((double)(1./cg->ngauss));
        cg->gaussian[j]->new_weight = 0.;
    }
    }
    if(gmpp->grand_flag == 2){ /* grand overall covariances */
    gmpp->log_det_covar = 0.;
    } else if(gmpp->grand_flag == 1){ /* grand per class covariances */
    for(i = 0; i < gmpp->noutputs; i++){
        gmpp->class_mix[i]->log_det_covar = 0.;
    }
    } else { /* per mixture component covariances */
    if(gmpp->tied_flag){
        for(i = 0; i < gmpp->ngauss; i++){
        gmpp->tied_gaussian[i]->log_det_covar = 0.;
        }
    } else {
        for(i = 0; i < gmpp->noutputs; i++){
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
            gmpp->class_mix[i]->gaussian[j]->log_det_covar = 0.;
        }
        }
    }
    }
/* re-intializing entries in the covariance arrays and the means? */
}
/*****************************************************************************
* NAME: gmix_test_init()
* DESCRIPTION:
*       Set model parameters needed for testing, not for training
*
* EXAMPLE:
*
*****************************************************************************/
void
gmix_test_init(int model, int adjust_priors, float *priors_list)
{
    int i;
    GMIX_PARAMS *gmpp;

    gmpp = gmixpp[model];

    if(adjust_priors)
      for(i = 0; i < gmpp->noutputs; i++)
    gmpp->log_class_priors[i] = (float)log10((double)priors_list[i]);
    else
      for(i = 0; i < gmpp->noutputs; i++){
      if(gmpp->class_cnt[i] != 0.)
        gmpp->log_class_priors[i] =
          (float)log10((double)gmpp->class_cnt[i]/gmpp->total_npatterns);
      else
        gmpp->log_class_priors[i] = 0.;
      }
}

/*****************************************************************************
* NAME: gmix_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
*****************************************************************************/
void
gmix_free(int model)
{
    int i,j;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */

    gmpp = gmixpp[model];
                    /* free vectors and matrices */
    deallocate_ivector(gmpp->features);
    deallocate_vector(gmpp->class_cnt);
    deallocate_vector(gmpp->log_class_priors);
    deallocate_vector(gmpp->outputs);
    if(gmpp->tied_flag){
    /* free means */
    for(i = 0; i < gmpp->ngauss; i++){
        deallocate_vector(gmpp->tied_gaussian[i]->means);
        deallocate_vector(gmpp->tied_gaussian[i]->new_means);
    }
    /* free variances */
    if(gmpp->grand_flag){
        if(gmpp->full_flag){
        deallocate_matrix(gmpp->inv_covar);
        deallocate_matrix(gmpp->new_covar);
        } else {
        deallocate_vector(gmpp->diag_inv_var);
        deallocate_vector(gmpp->diag_new_var);
        }
    } else {
        for(i = 0; i < gmpp->ngauss; i++){
        if(gmpp->full_flag){
            deallocate_matrix(gmpp->tied_gaussian[i]->inv_covar);
            deallocate_matrix(gmpp->tied_gaussian[i]->new_covar);
        } else {
            deallocate_vector(gmpp->tied_gaussian[i]->diag_inv_var);
            deallocate_vector(gmpp->tied_gaussian[i]->diag_new_var);
        }
        }
    }
    /* free structs for means */
    for(i = 0; i < gmpp->ngauss; i++){
        nn_free((char *)gmpp->tied_gaussian[i], "for holding tied means");
    }
    deallocate_pvector((void **)gmpp->tied_gaussian);
    } else { /* per class means */
    for(i = 0; i < gmpp->noutputs; i++){
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        deallocate_vector(gmpp->class_mix[i]->gaussian[j]->means);
        deallocate_vector(gmpp->class_mix[i]->gaussian[j]->new_means);
        }
    }
    if(gmpp->grand_flag == 2){
        if(gmpp->full_flag){
        deallocate_matrix(gmpp->inv_covar);
        deallocate_matrix(gmpp->new_covar);
        } else {
        deallocate_vector(gmpp->diag_inv_var);
        deallocate_vector(gmpp->diag_new_var);
        }
    } else if(gmpp->grand_flag == 1){
        for(i = 0; i < gmpp->noutputs; i++){
        if(gmpp->full_flag){
            deallocate_matrix(gmpp->class_mix[i]->inv_covar);
            deallocate_matrix(gmpp->class_mix[i]->new_covar);
        } else {
            deallocate_vector(gmpp->class_mix[i]->diag_inv_var);
            deallocate_vector(gmpp->class_mix[i]->diag_new_var);
        }
        }
    } else {
        for(i = 0; i < gmpp->noutputs; i++){
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
            if(gmpp->full_flag){
            deallocate_matrix(
              gmpp->class_mix[i]->gaussian[j]->inv_covar);
            deallocate_matrix(
              gmpp->class_mix[i]->gaussian[j]->new_covar);
            } else {
            deallocate_vector(
              gmpp->class_mix[i]->gaussian[j]->diag_inv_var);
            deallocate_vector(
              gmpp->class_mix[i]->gaussian[j]->diag_new_var);
            }
        }
        }
    }
    }
    for(i = 0; i < gmpp->noutputs; i++){
    for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        nn_free((char *)gmpp->class_mix[i]->gaussian[j],
            "for storing weights");
    }
    deallocate_pvector((void **)gmpp->class_mix[i]->gaussian);
    nn_free((char *)gmpp->class_mix[i], "for storing class weights");
    }
    deallocate_pvector((void **)gmpp->class_mix);

                    /* free structure itself */
    nn_free((char *) gmpp, "for gmix parameter structure");
    gmixpp[model] = NULL;
}

/*****************************************************************************
* NAME: gmix_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*       allocates space for and loads data into vectors and matricies
*
*       On error from param_getblk, returns the error value
*       (there is already an error check on the allocation, so we don't
*       have to do that here)
*
* EXAMPLE:
*****************************************************************************/
int
gmix_read_model(FILE *fp)
{
    int     model, i, j;
    int     rv, ninputs, noutputs, ngauss, tied_flag,
        grand_flag, full_flag, max_ngauss;
    int     *features, *ngauss_param;
    float   minimum_variance = 1e-5, eigen_ratio = 1e-6,
        **inv_covar, *diag_inv_var, *log_weights,
        total_npatterns, *class_cnt;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */

    param_setvarptr("ninputs", &ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &noutputs, paramtab, TABSIZE);
    param_setvarptr("ngauss", &ngauss, paramtab, TABSIZE);
    param_setvarptr("tied_flag", &tied_flag, paramtab, TABSIZE);
    param_setvarptr("grand_flag", &grand_flag, paramtab, TABSIZE);
    param_setvarptr("full_flag", &full_flag, paramtab, TABSIZE);
    param_setvarptr("total_npat",(int *)&total_npatterns, paramtab, TABSIZE);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    features = allocate_ivector(ninputs+1);
    ngauss_param = allocate_ivector(noutputs);
    class_cnt = allocate_vector(noutputs);

    param_setvarptr("features", &features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", ninputs+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("ngauss", &ngauss_param[0], paramtab_l, TABSIZE_L);
    param_setdim("ngauss", noutputs, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("class_cnt",(int *)&class_cnt[0], paramtab_l, TABSIZE_L);
    param_setdim("class_cnt", noutputs, 0, paramtab_l, TABSIZE_L);
    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

                    /* allocate space for model and vectors */
    model = gmix_build(ninputs, noutputs, ngauss_param, tied_flag, grand_flag,
               full_flag);
    gmpp = gmixpp[model];
                    /* set any remaining parameters */
    gmix_init(model, features, minimum_variance, eigen_ratio);
    for(i = 0; i < noutputs; i++){
    gmpp->class_cnt[i] = class_cnt[i];
    }
    gmpp->total_npatterns = total_npatterns;

    max_ngauss = 0;
    for(i = 0; i < noutputs; i++){
    if(gmpp->class_mix[i]->ngauss > max_ngauss)
      max_ngauss = gmpp->class_mix[i]->ngauss;
    }
    log_weights = allocate_vector(max_ngauss);
    param_setvarptr("log_weights",(int *)&log_weights[0],
            paramtab_w, TABSIZE_W);
    for(i = 0; i < noutputs; i++){
    param_setdim("log_weights", gmpp->class_mix[i]->ngauss, 0,
             paramtab_w, TABSIZE_W);
    if((rv = param_getblk(fp, paramtab_w, TABSIZE_W)) < 0)
      return rv;
    for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        gmpp->class_mix[i]->gaussian[j]->log_weight = log_weights[j];
    }
    }
    deallocate_vector(log_weights);

    param_setdim("means", ninputs, 0, paramtab_m, TABSIZE_M);
    if(tied_flag){
    for(i = 0; i < ngauss; i++){
        param_setvarptr("means",(int *)&gmpp->tied_gaussian[i]->means[0],
                paramtab_m, TABSIZE_M);
        if((rv = param_getblk(fp, paramtab_m, TABSIZE_M)) < 0)
          return rv;
    }
    } else {
    for(i = 0; i < noutputs; i++){
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        param_setvarptr("means",
            (int *)&gmpp->class_mix[i]->gaussian[j]->means[0],
                paramtab_m, TABSIZE_M);
        if((rv = param_getblk(fp, paramtab_m, TABSIZE_M)) < 0)
          return rv;
        }
    }
    }
    if(full_flag){
    diag_inv_var = allocate_vector(ninputs);
    param_setvarptr("diag_inv_var",(int *) &diag_inv_var[0],
            paramtab_c, TABSIZE_C);
    inv_covar = NULL;
    } else {
    inv_covar = allocate_matrix(ninputs, ninputs);
    param_setvarptr("inv_covar", (int *)&inv_covar[0][0],
            paramtab_c, TABSIZE_C);
    diag_inv_var = NULL;
    }
    param_setdim("inv_covar", ninputs, ninputs, paramtab_c, TABSIZE_C);
    param_setdim("diag_inv_var", ninputs, 0, paramtab_c, TABSIZE_C);

    if(grand_flag == 2){
    if(full_flag){
        param_setvarptr("inv_covar", (int *)&gmpp->inv_covar[0][0],
                paramtab_c, TABSIZE_C);
    } else {
        param_setvarptr("diag_inv_var",(int *)&gmpp->diag_inv_var[0],
                paramtab_c, TABSIZE_C);
    }
    param_setvarptr("log_det_covar", (int *)&gmpp->log_det_covar,
            paramtab_c, TABSIZE_C);
    if((rv = param_getblk(fp, paramtab_c, TABSIZE_C)) < 0)
      return rv;
    } else if(grand_flag == 1){
    for(i = 0; i < noutputs; i++){
        if(full_flag){
        param_setvarptr("inv_covar", (int *)
                &gmpp->class_mix[i]->inv_covar[0][0],
                paramtab_c, TABSIZE_C);
        } else {
        param_setvarptr("diag_inv_var", (int *)
                &gmpp->class_mix[i]->diag_inv_var[0],
                paramtab_c, TABSIZE_C);
        }
        param_setvarptr("log_det_covar", (int *)
                &gmpp->class_mix[i]->log_det_covar,
                paramtab_c, TABSIZE_C);
        if((rv = param_getblk(fp, paramtab_c, TABSIZE_C)) < 0)
          return rv;
    }
    } else {
    if(tied_flag){
        for(i = 0; i < gmpp->ngauss; i++){
        if(full_flag){
            param_setvarptr("inv_covar", (int *)
                    &gmpp->tied_gaussian[i]->inv_covar[0][0],
                    paramtab_c, TABSIZE_C);
        } else {
            param_setvarptr("diag_inv_var", (int *)
                    &gmpp->tied_gaussian[i]->diag_inv_var[0],
                    paramtab_c, TABSIZE_C);
        }
        param_setvarptr("log_det_covar", (int *)
                &gmpp->tied_gaussian[i]->log_det_covar,
                paramtab_c, TABSIZE_C);
        if((rv = param_getblk(fp, paramtab_c, TABSIZE_C)) < 0)
          return rv;
        }
    } else {
        for(i = 0; i < noutputs; i++){
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
            if(full_flag){
            param_setvarptr("inv_covar", (int *)
            &gmpp->class_mix[i]->gaussian[j]->inv_covar[0][0],
                    paramtab_c, TABSIZE_C);
            } else {
            param_setvarptr("diag_inv_var", (int *)
            &gmpp->class_mix[i]->gaussian[j]->diag_inv_var[0],
                    paramtab_c, TABSIZE_C);
            }
            param_setvarptr("log_det_covar", (int *)
            &gmpp->class_mix[i]->gaussian[j]->log_det_covar,
                    paramtab_c, TABSIZE_C);
            if((rv = param_getblk(fp, paramtab_c, TABSIZE_C)) < 0)
              return rv;
        }
        }
    }
    }

    if(full_flag){
    deallocate_vector(diag_inv_var);
    } else {
    deallocate_matrix(inv_covar);
    }
    deallocate_ivector(features);
    deallocate_vector(class_cnt);
    deallocate_ivector(ngauss_param);

    return (model);
}

/*****************************************************************************
* NAME: gmix_write_model()
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int
gmix_write_model(int model, FILE *fp)
{
    int             rv, i, j, max_ngauss, *ngauss_param;
    float   *log_weights, *diag_inv_var, **inv_covar;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */

    gmpp = gmixpp[model];
                    /* save scalars */
    param_setvarptr("ninputs", &gmpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &gmpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("ngauss", &gmpp->ngauss, paramtab, TABSIZE);
    param_setvarptr("tied_flag", &gmpp->tied_flag, paramtab, TABSIZE);
    param_setvarptr("grand_flag", &gmpp->grand_flag, paramtab, TABSIZE);
    param_setvarptr("full_flag", &gmpp->full_flag, paramtab, TABSIZE);
    param_setvarptr("total_npat",(int *) &gmpp->total_npatterns,
            paramtab, TABSIZE);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    ngauss_param = allocate_ivector(gmpp->noutputs);

    for(i = 0; i < gmpp->noutputs; i++){
    ngauss_param[i] = gmpp->class_mix[i]->ngauss;
    }
    param_setvarptr("features", &gmpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", gmpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("ngauss", &ngauss_param[0], paramtab_l, TABSIZE_L);
    param_setdim("ngauss", gmpp->noutputs, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("class_cnt", (int *)&gmpp->class_cnt[0],
            paramtab_l, TABSIZE_L);
    param_setdim("class_cnt", gmpp->noutputs, 0, paramtab_l, TABSIZE_L);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    max_ngauss = 0;
    for(i = 0; i < gmpp->noutputs; i++){
    if(gmpp->class_mix[i]->ngauss > max_ngauss)
      max_ngauss = gmpp->class_mix[i]->ngauss;
    }
    log_weights = allocate_vector(max_ngauss);
    param_setvarptr("log_weights",(int *)&log_weights[0],
            paramtab_w, TABSIZE_W);
    for(i = 0; i < gmpp->noutputs; i++){
    fprintf(fp, "log weights, class %d\n",i);
    for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        log_weights[j] = gmpp->class_mix[i]->gaussian[j]->log_weight;
    }
    param_setdim("log_weights", gmpp->class_mix[i]->ngauss, 0,
             paramtab_w, TABSIZE_W);
    if((rv = param_putblk(fp, paramtab_w, TABSIZE_W)) < 0)
      return rv;
    }
    deallocate_vector(log_weights);

    param_setdim("means", gmpp->ninputs, 0, paramtab_m, TABSIZE_M);
    if(gmpp->tied_flag){
    for(i = 0; i < gmpp->ngauss; i++){
        fprintf(fp, "tied means. gaussian %d\n", i);
        param_setvarptr("means",(int *)&gmpp->tied_gaussian[i]->means[0],
                paramtab_m, TABSIZE_M);
        if((rv = param_putblk(fp, paramtab_m, TABSIZE_M)) < 0)
          return rv;
    }
    } else {
    for(i = 0; i < gmpp->noutputs; i++){
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
        fprintf(fp, "means, class %d gaussian %d\n", i, j);
        param_setvarptr("means", (int *)
                &gmpp->class_mix[i]->gaussian[j]->means[0],
                paramtab_m, TABSIZE_M);
        if((rv = param_putblk(fp, paramtab_m, TABSIZE_M)) < 0)
          return rv;
        }
    }
    }
    param_setdim("diag_inv_var", gmpp->ninputs, 0, paramtab_c, TABSIZE_C);
    param_setdim("inv_covar", gmpp->ninputs, gmpp->ninputs,
         paramtab_c, TABSIZE_C);
    if(gmpp->full_flag){
    diag_inv_var = allocate_vector(gmpp->ninputs);
    inv_covar = NULL;
    param_setvarptr("diag_inv_var", (int *)&diag_inv_var[0],
            paramtab_c, TABSIZE_C);
    } else {
    inv_covar = allocate_matrix(gmpp->ninputs, gmpp->ninputs);
    diag_inv_var = NULL;
    param_setvarptr("inv_covar", (int *)&inv_covar[0][0],
            paramtab_c, TABSIZE_C);
    }
    if(gmpp->grand_flag == 2){
    fprintf(fp, "grand overall covariances\n");
    if(gmpp->full_flag){
        param_setvarptr("inv_covar", (int *)&gmpp->inv_covar[0][0],
                paramtab_c, TABSIZE_C);
    } else {
        param_setvarptr("diag_inv_var", (int *)&gmpp->diag_inv_var[0],
                paramtab_c, TABSIZE_C);
    }
    param_setvarptr("log_det_covar", (int *)&gmpp->log_det_covar,
            paramtab_c, TABSIZE_C);
    if((rv = param_putblk(fp, paramtab_c, TABSIZE_C)) < 0)
      return rv;
    } else if(gmpp->grand_flag == 1){
    for(i = 0; i < gmpp->noutputs; i++){
        fprintf(fp, "grand covariances. class %d\n", i);
        if(gmpp->full_flag){
        param_setvarptr("inv_covar", (int *)
                &gmpp->class_mix[i]->inv_covar[0][0],
                paramtab_c, TABSIZE_C);
        } else {
        param_setvarptr("diag_inv_var", (int *)
                &gmpp->class_mix[i]->diag_inv_var[0],
                paramtab_c, TABSIZE_C);
        }
        param_setvarptr("log_det_covar", (int *)
                &gmpp->class_mix[i]->log_det_covar,
                paramtab_c, TABSIZE_C);
        if((rv = param_putblk(fp, paramtab_c, TABSIZE_C)) < 0)
          return rv;
    }
    } else {
    if(gmpp->tied_flag){
        for(i = 0; i < gmpp->ngauss; i++){
        fprintf(fp, "tied covariances. gauss %d\n", i);
        if(gmpp->full_flag){
            param_setvarptr("inv_covar", (int *)
                    &gmpp->tied_gaussian[i]->inv_covar[0][0],
                    paramtab_c, TABSIZE_C);
        } else {
            param_setvarptr("diag_inv_var", (int *)
                    &gmpp->tied_gaussian[i]->diag_inv_var[0],
                    paramtab_c, TABSIZE_C);
        }
        param_setvarptr("log_det_covar", (int *)
                &gmpp->tied_gaussian[i]->log_det_covar,
                paramtab_c, TABSIZE_C);
        if((rv = param_putblk(fp, paramtab_c, TABSIZE_C)) < 0)
          return rv;
        }
    } else {
        for(i = 0; i < gmpp->noutputs; i++){
        for(j = 0; j < gmpp->class_mix[i]->ngauss; j++){
            fprintf(fp, "covariances. class %d gauss %d\n",i,j);
            if(gmpp->full_flag){
            param_setvarptr("inv_covar", (int *)
            &gmpp->class_mix[i]->gaussian[j]->inv_covar[0][0],
                    paramtab_c, TABSIZE_C);
            } else {
            param_setvarptr("diag_inv_var", (int *)
            &gmpp->class_mix[i]->gaussian[j]->diag_inv_var[0],
                    paramtab_c, TABSIZE_C);
            }
            param_setvarptr("log_det_covar", (int *)
            &gmpp->class_mix[i]->gaussian[j]->log_det_covar,
                    paramtab_c, TABSIZE_C);
            if((rv = param_putblk(fp, paramtab_c, TABSIZE_C)) < 0)
              return rv;
        }
        }
    }
    }

    if(gmpp->full_flag){
    deallocate_vector(diag_inv_var);
    } else {
    deallocate_matrix(inv_covar);
    }
    deallocate_ivector(ngauss_param);

    return (0);
}


/************************************************************************
* NAME: gmix_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/

void *
gmix_param(int model, char *param_name, int class, int g)
{
    void           *param_address;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */

    gmpp = gmixpp[model];
    if (PEQ("features"))
      param_address = &gmpp->features;
    else if (PEQ("ninputs"))
      param_address = &gmpp->ninputs;
    else if (PEQ("noutputs"))
      param_address = &gmpp->noutputs;
    else if (PEQ("outputs"))
      param_address = &gmpp->outputs;
    else if (PEQ("total_trials"))
      param_address = &gmpp->total_npatterns;
    else if (PEQ("class_trials"))
      param_address = &gmpp->class_cnt[class];
    else if (PEQ("minimum_variance"))
      param_address = &gmpp->minimum_variance;
    else if (PEQ("eigen_ratio"))
      param_address = &gmpp->eigen_ratio;
    else if (PEQ("tied_flag"))
      param_address = &gmpp->tied_flag;
    else if (PEQ("grand_flag"))
      param_address = &gmpp->grand_flag;
    else if (PEQ("full_flag"))
      param_address = &gmpp->full_flag;
    else if (PEQ("class_ngauss"))
      param_address = &gmpp->class_mix[class]->ngauss;
    else if (PEQ("ngauss"))
      param_address = &gmpp->ngauss;
    else if (PEQ("class_means"))
      param_address = &gmpp->class_mix[class]->gaussian[g]->means;
    else if (PEQ("tied_means"))
      param_address = &gmpp->tied_gaussian[g]->means;
    else if (PEQ("full_var"))
      param_address = &gmpp->inv_covar;
    else if (PEQ("diag_var"))
      param_address = &gmpp->diag_inv_var;
    else if (PEQ("determinant"))
      param_address = &gmpp->log_det_covar;
    else if (PEQ("class_full_var"))
      param_address = &gmpp->class_mix[class]->inv_covar;
    else if (PEQ("class_diag_var"))
      param_address = &gmpp->class_mix[class]->diag_inv_var;
    else if (PEQ("class_determinant"))
      param_address = &gmpp->class_mix[class]->log_det_covar;
    else if (PEQ("per_class_full_var"))
      param_address = &gmpp->class_mix[class]->gaussian[g]->inv_covar;
    else if (PEQ("per_class_diag_var"))
      param_address = &gmpp->class_mix[class]->gaussian[g]->diag_inv_var;
    else if (PEQ("per_class_determinant"))
      param_address = &gmpp->class_mix[class]->gaussian[g]->log_det_covar;
    else if (PEQ("tied_full_var"))
      param_address = &gmpp->tied_gaussian[g]->inv_covar;
    else if (PEQ("tied_diag_var"))
      param_address = &gmpp->tied_gaussian[g]->diag_inv_var;
    else if (PEQ("tied_determinant"))
      param_address = &gmpp->tied_gaussian[g]->log_det_covar;
    else if (PEQ("ave_logl"))
      param_address = &gmpp->ave_logl;
    else if (PEQ("class_ave_logl"))
      param_address = &gmpp->class_mix[class]->ave_logl;
    else if (PEQ("log_weight"))
      param_address = &gmpp->class_mix[class]->gaussian[g]->log_weight;
    else {
    warning("gmix_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}


/*****************************************************************************
* NAME: gmix_batch_ctrain()
* DESCRIPTION:
*       incremental training for a single input vector
* EXAMPLE:
*
*****************************************************************************/
int
gmix_batch_ctrain(int model, float *inputs, int desired_class)
{
    float **inv_covar, **new_covar, *diag_inv_var, *diag_new_var;
    float log_det_covar, *means, *new_means;
    float log_output, log_total_pdf, tau;
    struct class_gauss *cg;
    struct gauss_desc *cur_gauss, *gptr;
    int g, i, j, class;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */
    
    gmpp = gmixpp[model];

    if(gmpp->grand_flag == 2){
    if(gmpp->full_flag){
        inv_covar = gmpp->inv_covar;
        new_covar = gmpp->new_covar;
    } else {
        diag_inv_var = gmpp->diag_inv_var;
        diag_new_var = gmpp->diag_new_var;
    }
    log_det_covar = gmpp->log_det_covar;
    } else if (gmpp->grand_flag == 1){
    if(gmpp->full_flag){
        inv_covar = gmpp->class_mix[desired_class]->inv_covar;
        new_covar = gmpp->class_mix[desired_class]->new_covar;
    } else {
        diag_inv_var = gmpp->class_mix[desired_class]->diag_inv_var;
        diag_new_var = gmpp->class_mix[desired_class]->diag_new_var;
    }
    log_det_covar = gmpp->class_mix[desired_class]->log_det_covar;
    }
    
    /* update the number of patterns seen */
    cg = gmpp->class_mix[desired_class];
    cg->npatterns++;
    gmpp->npatterns++;
    gmpp->total_npatterns += 1.;
    gmpp->class_cnt[desired_class] += 1.;

    log_total_pdf = -(MAXFLOAT);
    /* find the output of each gaussian in this class */
    for(g = 0; g < cg->ngauss; g++){
    cur_gauss = cg->gaussian[g];
    
    if(gmpp->tied_flag) gptr = gmpp->tied_gaussian[g];
    else gptr = cur_gauss;

    means = gptr->means;

    if(!gmpp->grand_flag){  /* per mixture component covariances */
        if(gmpp->full_flag) inv_covar = gptr->inv_covar;
        else diag_inv_var = gptr->diag_inv_var;
        log_det_covar = gptr->log_det_covar;
    }
    
    if(gmpp->full_flag){
        log_output = get_log_gauss_output_full(gmpp->ninputs, inputs,
        means, inv_covar, log_det_covar);
    } else {
        log_output = get_log_gauss_output_diag(gmpp->ninputs, inputs,
        means, diag_inv_var, log_det_covar);
    }
    cur_gauss->weighted_gauss_output =
      cur_gauss->log_weight + log_output;
    /* log_total_pdf is the log of the output of the mixture for this
       class for this input vector (log likelihood for this class)*/
    log_total_pdf = logadd(log_total_pdf,
                   cur_gauss->weighted_gauss_output);
    }
    cg->ave_logl += log_total_pdf;

    /* update the weight, mean, and variance structures */
    for(g = 0; g < cg->ngauss; g++){
    cur_gauss = cg->gaussian[g];
    
    tau = nn_exp10(cur_gauss->weighted_gauss_output-log_total_pdf);
    cur_gauss->new_weight += tau;
    
    if(gmpp->tied_flag) gptr = gmpp->tied_gaussian[g];
    else gptr = cur_gauss;
    
    new_means = gptr->new_means;
    
    for(i = 0; i < gmpp->ninputs; i++){
        new_means[i] += tau * inputs[i];
    }
    if(!gmpp->grand_flag){  /* per mixture component covariances */
        for(i = 0; i < gmpp->ninputs; i++){
        if(gmpp->full_flag){
            for(j = 0; j < gmpp->ninputs; j++){
            gptr->new_covar[i][j] += inputs[i]*inputs[j]*tau;
            }
        } else {
            gptr->diag_new_var[i] += inputs[i]*inputs[i]*tau;
        }
        }
    }
    }
    /* if granding in any way, we don't include tau in new_covar
     * this means that we are continually recal'cing the same new_covar
     */
    if(gmpp->grand_flag){
    for(i = 0; i < gmpp->ninputs; i++){
        if(gmpp->full_flag){
        for(j = 0; j < gmpp->ninputs; j++){
            new_covar[i][j] += inputs[i]*inputs[j];
        }
        } else {
        diag_new_var[i] += inputs[i]*inputs[i];
        }
    }
    }
    class = gmix_classify(model, inputs);
    return(class);
}

/*****************************************************************************
* NAME: gmix_batch_update()
* DESCRIPTION:
*   updates gaussian mixtures, using information accumulated during
*   gmix_batch_ctrain()
*
* EXAMPLE:
*
*****************************************************************************/
void
gmix_batch_update(int model)
{
    int class, g, i, j;
    struct class_gauss *cg;
    struct gauss_desc *cur_gauss, *gptr;
    float **inv_covar, **new_covar, *diag_inv_var, *diag_new_var;
    float *eigenvalues;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */

    gmpp = gmixpp[model];
    if(gmpp->npatterns <= 0) return;

    if(gmpp->full_flag){
    eigenvalues = allocate_vector(gmpp->ninputs);
    }
    /* update the weights */
    gmpp->ave_logl = 0.;
    for(class = 0; class < gmpp->noutputs; class++){
    cg = gmpp->class_mix[class];
    for(g = 0; g < cg->ngauss; g++){
        cur_gauss = cg->gaussian[g];
        if(cg->npatterns > 0)
          cur_gauss->weight = cur_gauss->new_weight / cg->npatterns;
        if(cur_gauss->weight < SMALLEST) cur_gauss->weight = SMALLEST;
        if(cur_gauss->new_weight < SMALLEST)
          cur_gauss->new_weight = SMALLEST;
        /* new_weight cannot be 0 */
        cur_gauss->log_weight =
          (float)log10((double)cur_gauss->weight);
    }
    gmpp->ave_logl += cg->ave_logl;
    cg->ave_logl = 0.;
    }
    gmpp->ave_logl /= gmpp->npatterns;

    /* update the means */
    if(gmpp->tied_flag){
    for(g = 0; g < gmpp->ngauss; g++){
        gptr = gmpp->tied_gaussian[g];
        
        /* sum the weights for each gaussian to make new_weight[g] */
        gptr->new_weight = 0.;
        for(class = 0; class < gmpp->noutputs; class++){
        gptr->new_weight +=
          gmpp->class_mix[class]->gaussian[g]->new_weight;
        }

        for(i = 0; i < gmpp->ninputs; i++){
        gptr->means[i] = 
          gptr->new_means[i] / gptr->new_weight;
        gptr->new_means[i] = 0.;
        }
    }
    } else {
    for(class = 0; class < gmpp->noutputs; class++){
        cg = gmpp->class_mix[class];
        for(g = 0; g < cg->ngauss; g++){
        cur_gauss = cg->gaussian[g];
        for(i = 0; i < gmpp->ninputs; i++){
            cur_gauss->means[i] =
              cur_gauss->new_means[i] / cur_gauss->new_weight;
            cur_gauss->new_means[i] = 0.;
        }
        }
    }
    }
    /* update the variances */
    if(gmpp->grand_flag == 2){  /* grand overall */
    if(gmpp->full_flag){
        inv_covar = gmpp->inv_covar;
        new_covar = gmpp->new_covar;
        for(i = 0; i < gmpp->ninputs; i++){
        for(j = 0; j < gmpp->ninputs; j++){
            inv_covar[i][j] = new_covar[i][j];
            new_covar[i][j] = 0.;
            if(gmpp->tied_flag){
            for(g = 0; g < gmpp->ngauss; g++){
                gptr = gmpp->tied_gaussian[g];
                inv_covar[i][j] -=
                  gptr->means[i]*gptr->means[j]*gptr->new_weight;
                /* gptr->new_weight = sum(g)cur_gauss->new_weight */
            }
            } else {
            for(class = 0; class < gmpp->noutputs; class++){
                cg = gmpp->class_mix[class];
                for(g = 0; g < cg->ngauss; g++){
                cur_gauss = cg->gaussian[g];
                inv_covar[i][j] -=
                  cur_gauss->means[i]*cur_gauss->means[j]*
                    cur_gauss->new_weight;
                }
            }
            }
            inv_covar[i][j] /= gmpp->npatterns;
        }
        check_minimum_var(&inv_covar[i][i],
                  gmpp->minimum_variance);
        }
        if(invert_and_calc_det(gmpp->ninputs, inv_covar, eigenvalues,
                   gmpp->eigen_ratio, &gmpp->log_det_covar)
           == -1){
        warning("ill-conditioned matrix");
        }
    } else { /* grand overall - diagonal covariances */
        diag_inv_var = gmpp->diag_inv_var;
        diag_new_var = gmpp->diag_new_var;
        gmpp->log_det_covar = 0.;
        for(i = 0; i < gmpp->ninputs; i++){
        diag_inv_var[i] = diag_new_var[i];
        diag_new_var[i] = 0.;
        if(gmpp->tied_flag){
            for(g = 0; g < gmpp->ngauss; g++){
            gptr = gmpp->tied_gaussian[g];
            diag_inv_var[i] -=
              gptr->means[i]*gptr->means[i]*gptr->new_weight;
            }
        } else {
            for(class = 0; class < gmpp->noutputs; class++){
            cg = gmpp->class_mix[class];
            for(g = 0; g < cg->ngauss; g++){
                cur_gauss = cg->gaussian[g];
                diag_inv_var[i] -= 
                  cur_gauss->means[i]*cur_gauss->means[i]*
                cur_gauss->new_weight;
            }
            }
        }
        diag_inv_var[i] /= gmpp->npatterns;
        check_minimum_var(&diag_inv_var[i], gmpp->minimum_variance);
        gmpp->log_det_covar += (float)log10((double)diag_inv_var[i]);
        diag_inv_var[i] = 1./diag_inv_var[i];
        }
    }
    } else if(gmpp->grand_flag == 1){ /* un-tied only - per_class variances */
    for(class = 0; class < gmpp->noutputs; class++){
        cg = gmpp->class_mix[class];
        if(gmpp->full_flag){
        inv_covar = cg->inv_covar;
        new_covar = cg->new_covar;
        for(i = 0; i < gmpp->ninputs; i++){
            for(j = 0; j < gmpp->ninputs; j++){
            if(cg->npatterns > 0)
              inv_covar[i][j] = new_covar[i][j] / cg->npatterns;
            new_covar[i][j] = 0.;
            for(g = 0; g < cg->ngauss; g++){
                cur_gauss = cg->gaussian[g];
                inv_covar[i][j] -=
                  cur_gauss->means[i]*cur_gauss->means[j]*
                cur_gauss->weight;
            }
            }
            check_minimum_var(&inv_covar[i][i],
                      gmpp->minimum_variance);
        }
        if(invert_and_calc_det(gmpp->ninputs, inv_covar,
                       eigenvalues, gmpp->eigen_ratio,
                       &cg->log_det_covar) == -1){
            warning("ill-conditioned matrix: class %d", class);
        }
        } else {
        cg->log_det_covar = 0.;
        diag_inv_var = cg->diag_inv_var;
        diag_new_var = cg->diag_new_var;
        for(i = 0; i < gmpp->ninputs; i++){
            if(cg->npatterns > 0)
              diag_inv_var[i] = diag_new_var[i] / cg->npatterns;
            diag_new_var[i] = 0.;
            for(g = 0; g < cg->ngauss; g++){
            cur_gauss = cg->gaussian[g];
            diag_inv_var[i] -=
              cur_gauss->means[i]*cur_gauss->means[i]*
                cur_gauss->weight;
            }
            check_minimum_var(&diag_inv_var[i],
                      gmpp->minimum_variance);
            cg->log_det_covar +=
              (float)log10((double)diag_inv_var[i]);
            diag_inv_var[i] = 1./diag_inv_var[i];
        }
        }
    }
    } else { /* per mixture component */
    if(gmpp->tied_flag){
        for(g = 0; g < gmpp->ngauss; g++){
        gptr = gmpp->tied_gaussian[g];
        if(gmpp->full_flag){
            inv_covar = gptr->inv_covar;
            new_covar = gptr->new_covar;
            for(i = 0; i < gmpp->ninputs; i++){
            for(j = 0; j < gmpp->ninputs; j++){
                inv_covar[i][j] =
                  new_covar[i][j]/gptr->new_weight;
                new_covar[i][j] = 0.;
                inv_covar[i][j] -= gptr->means[i]*gptr->means[j];
            }
            check_minimum_var(&inv_covar[i][i],
                      gmpp->minimum_variance);
            }
            if(invert_and_calc_det(gmpp->ninputs, inv_covar,
                       eigenvalues, gmpp->eigen_ratio,
                       &gptr->log_det_covar) == -1){
            warning("ill-conditioned matrix: matrix %d", g);
            }
        } else {
            diag_inv_var = gptr->diag_inv_var;
            diag_new_var = gptr->diag_new_var;
            gptr->log_det_covar = 0.;
            for(i = 0; i < gmpp->ninputs; i++){
            diag_inv_var[i] = diag_new_var[i]/gptr->new_weight;
            diag_new_var[i] = 0.;
            diag_inv_var[i] -= gptr->means[i]*gptr->means[i];
            check_minimum_var(&diag_inv_var[i],
                      gmpp->minimum_variance);
            gptr->log_det_covar +=
              (float)log10((double)diag_inv_var[i]);
            diag_inv_var[i] = 1./diag_inv_var[i];
            }
        }
        } /* for each tied gaussian */
    } else { /* not tied mixtures */
        for(class = 0; class < gmpp->noutputs; class++){
        cg = gmpp->class_mix[class];
        for(g = 0; g < cg->ngauss; g++){
            cur_gauss = cg->gaussian[g];
            if(gmpp->full_flag){
            inv_covar = cur_gauss->inv_covar;
            new_covar = cur_gauss->new_covar;
            for(i = 0; i < gmpp->ninputs; i++){
                for(j = 0; j < gmpp->ninputs; j++){
                inv_covar[i][j] = 
                  new_covar[i][j] / cur_gauss->new_weight;
                new_covar[i][j] = 0.;
                inv_covar[i][j] -=
                  cur_gauss->means[i]*cur_gauss->means[j];
                }
                check_minimum_var(&inv_covar[i][i],
                          gmpp->minimum_variance);
            }
            if(invert_and_calc_det(gmpp->ninputs, inv_covar,
                           eigenvalues, gmpp->eigen_ratio,
                           &cur_gauss->log_det_covar)
               == -1){
                warning("%s: class %d matrix %d",
                    "ill-conditioned matrix", class, g);
            }
            } else {
            cur_gauss->log_det_covar = 0.;
            diag_inv_var = cur_gauss->diag_inv_var;
            diag_new_var = cur_gauss->diag_new_var;
            for(i = 0; i < gmpp->ninputs; i++){
                diag_inv_var[i] = 
                  diag_new_var[i] / cur_gauss->new_weight;
                diag_new_var[i] = 0.;
                diag_inv_var[i] -=
                  cur_gauss->means[i]*cur_gauss->means[i];
                check_minimum_var(&diag_inv_var[i],
                          gmpp->minimum_variance);
                cur_gauss->log_det_covar +=
                  (float)log10((double)diag_inv_var[i]);
                diag_inv_var[i] = 1./diag_inv_var[i];
            }
            }
        } /* for each mixture component */
        } /* for each class */
    } /* not tied */
    } /* per_mixture component */

    /* clear the new_weight and npatterns fields */
    /* update the log_class_priors field */
    for(class = 0; class < gmpp->noutputs; class++){
    cg = gmpp->class_mix[class];
    for(g = 0; g < cg->ngauss; g++){
        cg->gaussian[g]->new_weight = 0.;
    }
    if(cg->npatterns != 0)
      gmpp->log_class_priors[class] =
        (float)log10((double)cg->npatterns/(double)gmpp->npatterns);
    else gmpp->log_class_priors[class] = SMALLEST;
    cg->npatterns = 0;
    }

    gmpp->npatterns = 0;
    if(gmpp->full_flag) deallocate_vector(eigenvalues);
}

/*****************************************************************************
* NAME: gmix_eval()
* DESCRIPTION:
*   calculates outputs for gmix classifier
*
*****************************************************************************/
void
gmix_eval(int model, float *inputs)
{
    float **inv_covar, *diag_inv_var, log_det_covar, log_output, *means;
    int g, class;
    struct class_gauss *cg;
    struct gauss_desc *cur_gauss, *gptr;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */

    gmpp = gmixpp[model];

    if(gmpp->grand_flag == 2){
    if(gmpp->full_flag) inv_covar = gmpp->inv_covar;
    else diag_inv_var = gmpp->diag_inv_var;
    log_det_covar = gmpp->log_det_covar;
    }

    for(class = 0; class < gmpp->noutputs; class++){
    cg = gmpp->class_mix[class];
    if (gmpp->grand_flag == 1){
        if(gmpp->full_flag) inv_covar = cg->inv_covar;
        else diag_inv_var = cg->diag_inv_var;
        log_det_covar = cg->log_det_covar;
    }
    gmpp->outputs[class] = -(MAXFLOAT);
    for(g = 0; g < cg->ngauss; g++){
        cur_gauss = cg->gaussian[g];
        if(gmpp->tied_flag) gptr = gmpp->tied_gaussian[g];
        else gptr = cur_gauss;
        means = gptr->means;
        if(!gmpp->grand_flag){  /*per mixture component covariances*/
        if(gmpp->full_flag) inv_covar = gptr->inv_covar;
        else diag_inv_var = gptr->diag_inv_var;
        log_det_covar = gptr->log_det_covar;
        }
        if(gmpp->full_flag){
        log_output = get_log_gauss_output_full(gmpp->ninputs, inputs,
                means, inv_covar, log_det_covar);
        } else {
        log_output = get_log_gauss_output_diag(gmpp->ninputs, inputs,
                means, diag_inv_var, log_det_covar);
        }
        log_output += cur_gauss->log_weight;
        gmpp->outputs[class] = logadd(gmpp->outputs[class], log_output);
    }
    gmpp->outputs[class] += gmpp->log_class_priors[class];
    }
}

/*****************************************************************************
* NAME: gmix_classify()
* DESCRIPTION:
*       Implements the gmix classifier to return an integer class.
*
* EXAMPLE:
*
*****************************************************************************/
int gmix_classify(int model, float *inputs)
{
    int class;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */

    gmix_eval(model, inputs);

    gmpp = gmixpp[model];
    class = max_pick(gmpp->outputs, gmpp->noutputs);

    return class;
}

/*****************************************************************************
* NAME: gmix_initialize_cluster_centers()
* DESCRIPTION:
*   sets initial centers and variances to those in the cluster model
*   the gaussian mixture is assumed to have been built the same
*   size as the cluster centers
*
* EXAMPLE:
*
*****************************************************************************/
void gmix_initialize_cluster_centers(int model, int clus_model,
                     float var_spread)
{
    int i, g, j, class;
    float **dim_class_var, *dim_var, var_temp;
    float **pcenter_part, **pvariance_part;
    int cluster_by_class, *ncenters_list;
    struct class_gauss *cg;
    struct gauss_desc *cur_gauss;
    GMIX_PARAMS *gmpp;  /* temporary pointer to gmixpp[] */
    
    gmpp = gmixpp[model];

    cluster_by_class = *(int *)cluster_param(clus_model,"cluster_by_class",0);

    dim_var = allocate_vector(gmpp->ninputs);
    dim_class_var = allocate_matrix(gmpp->noutputs, gmpp->ninputs);

    if(gmpp->tied_flag){
    /* initialize the variances */
    if(gmpp->full_flag){
        if(gmpp->grand_flag){
        for(i = 0; i < gmpp->ninputs; i++){
            for(j = 0; j < gmpp->ninputs; j++){
            gmpp->inv_covar[i][j] = 0.;
            }
        }
        } else {
        for(g = 0; g < gmpp->ngauss; g++){
            for(i = 0; i < gmpp->ninputs; i++){
            for(j = 0; j < gmpp->ninputs; j++){
                gmpp->tied_gaussian[g]->inv_covar[i][j] = 0.;
            }
            }
        }
        }
    }

    if(cluster_by_class){
        /* copy the centers.  There should be room for them all */
        ncenters_list =
          *(int **)cluster_param(clus_model, "ncenters_list", 0);
        g = 0;
        for(class = 0; class < gmpp->noutputs; class++){
        pcenter_part =
          *(float ***)cluster_param(clus_model, "pcenter", class);
        pvariance_part =
          *(float ***)cluster_param(clus_model, "pvariance", class);

        for(j = 0; j < ncenters_list[class] && g < gmpp->ngauss;
            j++, g++){
            if(!gmpp->grand_flag)
              gmpp->tied_gaussian[g]->log_det_covar = 0.;
            for(i = 0; i < gmpp->ninputs; i++){
            /* check to see if there are zero variances */
            if(check_minimum_var(&pvariance_part[j][i],
                         gmpp->minimum_variance));
              /*
              warning("%s: %s, changed to %lf\n",
                  "gmix_initialize_cluster_center",
                  "zero variance found in cluster data",
                  gmpp->minimum_variance);
                  */
            gmpp->tied_gaussian[g]->means[i] = pcenter_part[j][i];
            var_temp = pvariance_part[j][i]*var_spread;
            if(gmpp->grand_flag){
                dim_var[i] += var_temp;
            } else {
                gmpp->tied_gaussian[g]->log_det_covar +=
                  (float)log10((double)var_temp);
                if(gmpp->full_flag){
                gmpp->tied_gaussian[g]->inv_covar[i][i] = 
                  1./var_temp;
                } else {
                gmpp->tied_gaussian[g]->diag_inv_var[i] =
                  1./var_temp;
                }
            }
            }
        }
        }
    } else {    /* centers not stored by class */
        pcenter_part = *(float ***)cluster_param(clus_model, "pcenter", 0);
        pvariance_part =
          *(float ***)cluster_param(clus_model, "pvariance", 0);

        for(g = 0; g < gmpp->ngauss; g++){
        if(!gmpp->grand_flag)
          gmpp->tied_gaussian[g]->log_det_covar = 0.;
        for(i = 0; i < gmpp->ninputs; i++){
            if(check_minimum_var(&pvariance_part[g][i],
                     gmpp->minimum_variance));
              /*
              warning("%s: %s, changed to %lf\n",
                  "gmix_initialize_cluster_center",
                  "zero variance found in cluster data",
                  gmpp->minimum_variance);
                  */
            gmpp->tied_gaussian[g]->means[i] = pcenter_part[g][i];
            var_temp = pvariance_part[g][i]*var_spread;
            if(gmpp->grand_flag){
            dim_var[i] += var_temp;
            } else {
            gmpp->tied_gaussian[g]->log_det_covar +=
              (float)log10((double)var_temp);
            if(gmpp->full_flag){
                gmpp->tied_gaussian[g]->inv_covar[i][i] = 
                  1./var_temp;
            } else {
                gmpp->tied_gaussian[g]->diag_inv_var[i] =
                  1./var_temp;
            }
            }
        }
        }
    }
    if(gmpp->grand_flag){
        gmpp->log_det_covar = 0.;
        for(i = 0; i < gmpp->ninputs; i++){
        dim_var[i] /= gmpp->ngauss;
        gmpp->log_det_covar +=
          (float)log10((double)dim_var[i]);
        if(gmpp->full_flag){
            gmpp->inv_covar[i][i] = 1./dim_var[i];
        } else {
            gmpp->diag_inv_var[i] = 1./dim_var[i];
        }
        }
    }
    } else { /* not tied */
    /* initialize the variances */
    if(gmpp->full_flag){
        if(gmpp->grand_flag == 2){
        for(i = 0; i < gmpp->ninputs; i++){
            for(j = 0; j < gmpp->ninputs; j++){
            gmpp->inv_covar[i][j] = 0.;
            }
        }
        } else if(gmpp->grand_flag == 1){
        for(class = 0; class < gmpp->noutputs; class++){
            for(i = 0; i < gmpp->ninputs; i++){
            for(j = 0; j < gmpp->ninputs; j++){
                gmpp->class_mix[class]->inv_covar[i][j] = 0.;
            }
            }
        }
        } else {
        for(class = 0; class < gmpp->noutputs; class++){
            for(g = 0; g < gmpp->class_mix[class]->ngauss; g++){
            for(i = 0; i < gmpp->ninputs; i++){
                for(j = 0; j < gmpp->ninputs; j++){
                gmpp->class_mix[class]->
                  gaussian[g]->inv_covar[i][j] = 0.;
                }
            }
            }
        }
        }
    }
    /* copy the centers */
    pcenter_part = *(float ***)cluster_param(clus_model, "pcenter", 0);
    pvariance_part = *(float ***)cluster_param(clus_model, "pvariance", 0);
    for(class = 0; class < gmpp->noutputs; class++){
        if(cluster_by_class){
        pcenter_part =
          *(float ***)cluster_param(clus_model, "pcenter", class);
        pvariance_part =
          *(float ***)cluster_param(clus_model, "pvariance", class);
        }
        cg = gmpp->class_mix[class];
        for(g = 0; g < cg->ngauss; g++){
        cur_gauss = cg->gaussian[g];
        if(!gmpp->grand_flag) cur_gauss->log_det_covar = 0.;
        for(i = 0; i < gmpp->ninputs; i++){
            if(check_minimum_var(&pvariance_part[g][i],
                     gmpp->minimum_variance));
              /*
              warning("%s: %s, changed to %lf\n",
                  "gmix_initialize_cluster_center",
                  "zero variance found in cluster data",
                  gmpp->minimum_variance);
                  */
            cur_gauss->means[i] = pcenter_part[g][i];
            var_temp = pvariance_part[g][i]*var_spread;
            if(gmpp->grand_flag){
            /* average the variances by class, then overall */
            dim_class_var[class][i] += var_temp;
            } else {
            cur_gauss->log_det_covar +=
              (float)log10((double)var_temp);
            if(gmpp->full_flag){
                cur_gauss->inv_covar[i][i] =
                  1./var_temp;
            } else {
                cur_gauss->diag_inv_var[i] =
                  1./var_temp;
            }
            }
        }
        }
        if(gmpp->grand_flag){
        if(gmpp->grand_flag == 1)
          cg->log_det_covar = 0.;
        for(i = 0; i < gmpp->ninputs; i++){
            dim_class_var[class][i] /= cg->ngauss;
            if(gmpp->grand_flag == 1){
            cg->log_det_covar +=
              (float)log10((double)dim_class_var[class][i]);
            if(gmpp->full_flag){
                cg->inv_covar[i][i] = 1./dim_class_var[class][i];
            } else {
                cg->diag_inv_var[i] = 1./dim_class_var[class][i];
            }
            } else {
            dim_var[i] += dim_class_var[class][i];
            }
        }
        }
    }
    /* fill in the variances */
    if(gmpp->grand_flag == 2){
        gmpp->log_det_covar = 0.;
        for(i = 0; i < gmpp->ninputs; i++){
        dim_var[i] /= gmpp->noutputs;
        gmpp->log_det_covar += (float)log10((double)dim_var[i]);
        if(gmpp->full_flag){
            gmpp->inv_covar[i][i] = 1./dim_var[i];
        } else {
            gmpp->diag_inv_var[i] = 1./dim_var[i];
        }
        }
    }
    }
    deallocate_vector(dim_var);
    deallocate_matrix(dim_class_var);
}

/*****************************************************************************
* NAME: gmix_read_centers()
* DESCRIPTION:
*   reads the cluster parameter file and sets up the ngauss_param list
*   properly for use in gmix_build()
*
* EXAMPLE:
*
*****************************************************************************/
int gmix_read_centers(char *param_path, FILE *param_file, int *pninputs,
              int *pnoutputs, int **ngauss_param, int *features,
              int tied_flag)
{
    int rv, ngauss, *ncenters_list, i, cluster_by_class, noutputs;

    rv = cluster_read_paramfile(param_file, pninputs, &noutputs, &ngauss,
                features);
    if(rv < 0)
      fatal_error("gmix: read error %d on cluster parameter file %s",
          rv, param_path);

    /* this ngauss is the total number of centers stored */
    ngauss = *(int *)cluster_param(rv, "ncenters", 0);
    if(ngauss <= 0)
      fatal_error("gmix_read_cluster: No centers from clustering");

    cluster_by_class = *(int *)cluster_param(rv, "cluster_by_class", 0);

    *ngauss_param = allocate_ivector(noutputs);
    if(cluster_by_class){
    *pnoutputs = noutputs;
    ncenters_list = *(int **)cluster_param(rv, "ncenters_list", 0);
    if(tied_flag){
        for(i = 0; i < noutputs; i++){
        (*ngauss_param)[i] = ngauss;
        }
    } else {
        for(i = 0; i < noutputs; i++){
        (*ngauss_param)[i] = ncenters_list[i];
        }
    }
    } else {
    for(i = 0; i < noutputs; i++){
        (*ngauss_param)[i] = ngauss;
    }
    }
    return(rv);
}

/* should be used for on-diagonal entries in the covariance matrix, only */
/* returns -1 if the variance had to be changed, 0 otherwise */
static int check_minimum_var(float *covar_entry, float minimum_variance)
{
    if(*covar_entry < minimum_variance){
    *covar_entry = minimum_variance;
    return -1;
    } else return 0;
}
    
