/***--------------------------------------------------------------------------
 * NAME:    gmix_seq_cv.c
 * PURPOSE: cross_validation driver for gmix sequence classifier
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_seq_cv.c,v 11.4 2001/08/23 20:11:53 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_seq_cv.c,v $
 * CREATED: 16-Aug-93 lck           MIT LINCOLN LABORATORY
 ***************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "gmix_seq.h"
#include "gmix_pack.h"
#include "gmix_seq_pack.h"

#define ABS(a) ((a) >= 0 ? (a) : -(a))

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

static char prog_name[MAXSTR];  /* static variables */

void
gmix_seq_cross_valid(int db_type, int noutputs, int nsequences, int debg,
             int verbosity, int verror, int nepochs, int cv_nfolds,
             int cv_nsplits, int use_cv_splits, int *cv_splits,
             char **cv_train_mask, char **cv_test_mask, int cv_random,
             char *data_path, char *data_fname, char *exp_path,
             char *error_fname, char *param_fname, char **labels,
             char *comment, int log_outputs, int tied_flag,
             int grand_flag, int full_flag, float minimum_variance,
             float max_ratio, float var_spread,
             char *cluster_param_fname, float stop_percentage)
{
    int     model, seq_model, cluster_model, rv;
    int     data_model, err_model, err_part_model;
    int     normalization=0, epoch;
    int     *features, nraw, ninputs, model_ninputs;
    int     nwindow_frames, *window_frames;
    int     desired_class, class;
    int     n_read, n_frames, fold, nfiles, seq_num;
    float   *inputs, *seq_sums;
    float   ave_logl, last_ave_logl, percent_change;
    char    *tmp_path, clus_path[MAXSTR];
    FILE    *error_file, *param_file, *clus_file;
    int     *ngauss_param = NULL;
    int     post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());

    /* read centers from cluster file */
    tmp_path = path_build(exp_path, cluster_param_fname);
    strncpy(clus_path, tmp_path, MAXSTR -1);
    
    clus_file = nn_fopen(clus_path, "r",
             "for read of cluster parameters");
    if(verbosity > 1) {
    printf("Reading cluster parameters from \"%s\"\n", clus_path);
    fflush(stdout);
    }
    model = norm_read_paramfile(clus_file, &normalization);
    if(model < 0)
      fatal_error("gmix: read error %d %s %s",
          model, "on normalization parameter file", clus_path);
    rv = feature_read_paramfile(clus_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
          "on feature selection parameters in file", clus_path);
    rv = window_read_paramfile(clus_file, &nwindow_frames, &window_frames);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
          "on windowing parameters in file", clus_path);
    cluster_model = 
      gmix_read_centers(clus_path, clus_file, &model_ninputs, &noutputs,
            &ngauss_param, features, tied_flag);
    if (cluster_model < 0)
      fatal_error("%s: read error %d %s %s", prog_name, cluster_model,
          "on cluster parameter file", tmp_path);
    
    /* build the gmix model */
    model = gmix_build(model_ninputs, noutputs, ngauss_param, tied_flag,
               grand_flag, full_flag);
    gmix_init(model, features, minimum_variance, max_ratio);
    gmix_initialize_cluster_centers(model, cluster_model, var_spread);
    /* uniform priors ??? */

    /* build the sequence model */
    seq_model = gmix_seq_build(noutputs);
    gmix_seq_init(seq_model, model);
    seq_sums = *(float **) gmix_seq_param(seq_model, "seq_sums");
    cluster_free(cluster_model);
    
    /* prepare for error reporting */
    if (verror > 0) {   /* open file for classification errors */
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    } else error_file = NULL;
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, db_type, model_ninputs, LOG_LIKE,
         log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* load data */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
                   ninputs, noutputs, features, normalization,
                   labels, verbosity, nwindow_frames,
                   window_frames);
    nfiles = seq_data_get_nfiles(data_model);
    
    seq_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
             cv_nfolds, cv_train_mask, cv_test_mask, cv_random);

    /* print out the normalization parameters */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);
    
    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);
    /* read in the next set of cluster centers */
    if(fold != 0){
        gmix_free(model);
        cluster_model =
          gmix_read_centers(clus_path, clus_file, &model_ninputs,
                &noutputs, &ngauss_param, features,
                tied_flag);
        if (cluster_model < 0)
          fatal_error("%s: read error %d %s %s", prog_name, cluster_model,
              "on cluster parameter file", tmp_path);
        model = gmix_build(model_ninputs, noutputs, ngauss_param,
                   tied_flag, grand_flag, full_flag);
        gmix_init(model, features, minimum_variance, max_ratio);
        gmix_initialize_cluster_centers(model, cluster_model, var_spread);
        cluster_free(cluster_model);
    }
    
    /* set up the data routines for this fold */
    seq_data_start_cv_fold(data_model, fold, TRUE);
    nfiles = seq_data_get_nfiles(data_model);

    /* epoch training */
    report_init(err_part_model, db_type, model_ninputs,
            LOG_LIKE, log_outputs, post_sum_to_one);
    report_epoch_header(err_part_model, stdout, nfiles);

    last_ave_logl = 0.;
    for(epoch = 0, percent_change = MAXFLOAT;
        (epoch < nepochs) && (percent_change > stop_percentage); epoch++) {
        seq_data_epoch_init(data_model);
        for(n_read = 0; n_read < nfiles; n_read++){
        if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
        desired_class = seq_data_get_class(data_model);
        gmix_seq_init(seq_model, model);
        for(n_frames = 0; ; n_frames++){
            inputs = seq_data_get_next_frame(data_model);
            if(inputs == NULL) break;
            gmix_batch_ctrain(model, inputs, desired_class);
            gmix_seq_batch_classify(seq_model, inputs);
        }
        if(n_frames > 0){
            class = gmix_seq_update_classify(seq_model);
            report_seq_trial(err_part_model, error_file, seq_num,
                     seq_sums, desired_class, class);
        }
        }
        gmix_batch_update(model);

        ave_logl = report_get_avg_cost(err_part_model);
        if(last_ave_logl != 0.){
        percent_change = 100.0 * (ave_logl - last_ave_logl)/
          ABS(last_ave_logl);
        if(verbosity > 3){
            if(percent_change < 0.0){
            warning("gmix: diverging\n%s %g\n%s %g\n%s %g",
                "Last ave log likelihood", last_ave_logl,
                "Current log_likelihood", ave_logl,
                "Percentage change", percent_change);
            }
        }
        }
        percent_change = ABS(percent_change);
        last_ave_logl = ave_logl;
        report_epoch(err_part_model, stdout, epoch);
        report_init(err_part_model, db_type, model_ninputs, LOG_LIKE,
            log_outputs, post_sum_to_one);
    }
    
    /* print out the current model */
    fprintf(param_file, "gmix model\n");
    if(gmix_write_model(model, param_file) < 0)
      warning("did not write whole parameter file");
    
    /* initialize error reporting */
    gmix_seq_init(seq_model, model);
    /* test_init?*/
    report_init(err_part_model, db_type, model_ninputs,
            LOG_LIKE, log_outputs, post_sum_to_one);
    
    /* finish the training fold, start the testing fold */
    seq_data_end_cv_fold(data_model);
    seq_data_start_cv_fold(data_model, fold, FALSE);
    nfiles = seq_data_get_nfiles(data_model);

    for (n_read = 0; n_read < nfiles; n_read++){
        if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
        desired_class = seq_data_get_class(data_model);
        gmix_seq_init(seq_model, model);
        for(n_frames = 0;  ; n_frames++){
        inputs = seq_data_get_next_frame(data_model);
        if(inputs == NULL) break;
        gmix_seq_batch_classify(seq_model, inputs);
        }
        if(n_frames > 0){
        class = gmix_seq_update_classify(seq_model);
        
        report_seq_trial(err_model, error_file, seq_num, seq_sums,
                 desired_class, class);
        report_seq_trial(err_part_model, error_file, seq_num, seq_sums,
                 desired_class, class);
        }
    }
    /* report on testing */
    report_overall(err_part_model, stdout, error_fname, labels);
    
    /* finish the testing fold */
    seq_data_end_cv_fold(data_model);
    } /* end fold */
    /* close the parameter file and the cluster file */
    nn_fclose(param_file, tmp_path, "for write of model file");
    nn_fclose(clus_file, clus_path, "for read of clusters");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);

    if (verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }
    /* free space */
    gmix_free(model);
    gmix_seq_free(seq_model);
    norm_free();
    seq_data_exit(data_model);
}
