/***--------------------------------------------------------------------------
 * NAME:		gmix_seq_test.c
 * PURPOSE:	testing driver for gaussian mixture sequence classifier
 *
 * SEE ALSO: 	gmix_pack.c gmix_seq_pack.c
 * GLOBALS:	int debug;
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_seq_test.c,v 11.4 2001/08/23 20:11:53 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_seq_test.c,v $
 * CREATED: 18-JUN-91 lck                             MIT LINCOLN LABORATORY
 *----------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "gmix_seq.h"
#include "gmix_seq_pack.h"
#include "gmix_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];	/* static variables */

void
gmix_seq_test(int create_flag, int db_type, int nsequences, int debg,
	      int verbosity, int verror, char *data_path, char *data_fname,
	      char *exp_path, char *error_fname, char *param_fname,
	      char **labels)
{
    int		model, seq_model, err_model, data_model, rv, normalization=0;
    int		*features, ninputs, nraw, model_ninputs, noutputs;
    int		nwindow_frames, *window_frames;
    int		class, desired_class;
    int		n_files, n_frames, seq_num;
    float	*inputs, *seq_sums;
    char	*tmp_path;
    FILE	*error_file, *param_file;
    int		log_outputs = 1, post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s",
		  prog_name, rv, "on normalization parameter file",tmp_path);
    rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on feature selection parameters in file", tmp_path);
    rv = window_read_paramfile(param_file, &nwindow_frames, &window_frames);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on windowing parameters in file", tmp_path);
    
    model = gmix_read_model(param_file);
    /* take features from the model */
    noutputs = *(int *)gmix_param(model, "noutputs", 0, 0);
    model_ninputs = *(int *)gmix_param(model, "ninputs", 0, 0);

    nn_fclose(param_file, tmp_path, "for read of model file");

    seq_model = gmix_seq_build(noutputs);
    gmix_seq_init(seq_model, model);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s", 
		  prog_name, model, tmp_path);

    seq_sums = *(float **) gmix_seq_param(seq_model, "seq_sums");

    /* prepare results reporting */
    if (verror > 0) {	/* open file for classification errors */
	if(create_flag)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    } else error_file = NULL;
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, db_type, model_ninputs, LOG_LIKE, log_outputs, post_sum_to_one);

    /* read data */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    nsequences = seq_data_get_nfiles(data_model);

    /*************** prepare model for testing **************/
    /* priors??? */
    gmix_test_init(model, 0, NULL);/* no adjust priors, use log outputs */

    /* read and classify input from data file */
    for(n_frames = n_files = 0; n_files < nsequences; n_files++){
	if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
	desired_class = seq_data_get_class(data_model);
	gmix_seq_init(seq_model, model);
	for(n_frames = 0; ; n_frames++){
	    inputs = seq_data_get_next_frame(data_model);
	    if(inputs == NULL) break;

	    gmix_seq_batch_classify(seq_model, inputs);
	}
	if(n_frames > 0){
	    class = gmix_seq_update_classify(seq_model);
	    report_seq_trial(err_model, error_file, seq_num, seq_sums,
			     desired_class,class);
	}
    }
    report_overall(err_model, stdout, error_fname, labels);
    
    /* finish error reporting and close error file */
    report_free(err_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    gmix_free(model);
    gmix_seq_free(seq_model);
    seq_data_exit(data_model);
    norm_free();
}
