/***--------------------------------------------------------------------------
 * NAME:		gmix.c
 * PURPOSE:	driver for gaussian mixture classifier
 *
 * SEE ALSO: 	gmix_pack.c
 *
 * FUNCTIONS:
 *	driver for gaussian mixture classifier
 *		handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_test.c,v 11.4 2001/08/23 20:11:53 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/gmix/gmix_test.c,v $
 * CREATED: 18-JUN-91 lck                             MIT LINCOLN LABORATORY
 *----------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "gmix.h"
#include "gmix_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];	/* static variables */

void
gmix_test(int create_flag, int map_flag, int nraw, int ninputs, int noutputs,
	  int npatterns, int debg, int verbosity, int verror, char *data_path,
	  char *data_fname, char *exp_path, char *error_fname,
	  char *param_fname, char **labels, int adjust_priors,
	  float *priors_list, int log_outputs)
{
    int		model, err_model, rv, normalization=0, data_model;
    int		desired_class, class;
    float	*inputs, *outputs;
    char	*tmp_path;
    FILE	*error_file, *param_file;
    int		*features;
    int		post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());
    
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s",
		  prog_name, rv, "on normalization parameter file", tmp_path);
    
    model = gmix_read_model(param_file);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s", 
		  prog_name, model, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of model file");
    
    /* use features from the model */
    features = *(int **) gmix_param(model, "features", 0, 0);
    ninputs = *(int *) gmix_param(model, "ninputs", 0, 0);
    outputs = *(float **) gmix_param(model, "outputs", 0, 0);

    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {	/* open file for classification errors */
	if(create_flag)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, LOG_LIKE,
		 log_outputs, post_sum_to_one);

    /* set up for reading data */
    data_model =
      class_data_single_init(data_path, data_fname, npatterns,
			     nraw, ninputs, noutputs, features,
			     normalization, verbosity);
    
    gmix_test_init(model, adjust_priors, priors_list);
    
    /* read and classify input from data file */
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
	class = gmix_classify(model, inputs);	/* classify pattern and
						 * report errors */
	report_class_trial(err_model, error_file, inputs, outputs,
			   desired_class, class,
			   class_data_get_input_num(data_model));
    }
    
    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_test_result(err_model, stdout, data_fname);
    report_free(err_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    class_data_exit(data_model);
    gmix_free(model);
    norm_free();
}
