/***---------------------------------------------------------------------------
* NAME:         histogram_pack.c
* PURPOSE:      Functions specific to histogram.c
*
* SEE ALSO:	histogram.c histogram_main.c
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*	histogram_build		Allocate storage for model parameters.
*	histogram_init		Initialize model parameters.
*	histogram_free		Deallocates model storage
*	histogram_read_model	Initializes model from file
*	histogram_write_model	Writes model to output file
*	histogram_param		Returns pointers to elements in model
*	histogram_inc_ctrain	Does training on a vector
*	histogram_classify	Classifies a vector using histogram algortihm
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/histogram/histogram_pack.c,v 11.4 2001/08/23 20:11:57 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/histogram/histogram_pack.c,v $
* CREATED: 24-Aug-94 lck				MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#define histogram_pack_h_internal_types
#define histogram_pack_h_static_decl
#include "histogram_pack.h"

				/* Internal Globals */
static HISTOGRAM_PARAMS **histpp;/* array of parameters for each model */
static int      max_models = 0;	/* current length of histpp */
#define BIN_STORE_SIZE 128
#define MIN_PROB (1e-9)

/*****************************************************************************
* NAME: histogram_build(ninputs,noutputs)
*	ninputs - number of dimensions in training vectors
*	noutputs - number of classes
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
*****************************************************************************/
int
histogram_build(int ninputs, int noutputs, int bin_type, int *nbins,
		int grand_bins)
{
    register int    new_model;
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    int c, i, b;
    			/* find entry in model array for new model */
    new_model =
      model_get_number((void ***)&histpp,
		       &max_models, sizeof(HISTOGRAM_PARAMS));
    hpp = histpp[new_model];

    /* set fields which are constants */
    hpp->ninputs = ninputs;
    hpp->noutputs = noutputs;
    hpp->grand_bins = grand_bins;
    hpp->bin_type = bin_type;

    /* allocate space for vectors */
    hpp->class_cnt = allocate_vector(noutputs);
    hpp->nbins = allocate_ivector(noutputs);
    for(c = 0; c < noutputs; c++) hpp->nbins[c] = nbins[c];

    /* set vectors that will not be allocated here to NULL */
    hpp->class_priors = hpp->outputs = hpp->gauss_outputs = NULL;

    /* allocate space for prob structures */
    hpp->prob_list = (HISTOGRAM_PROB **)
      allocate_structmatrix(noutputs, ninputs, sizeof(HISTOGRAM_PROB));
    for(c = 0; c < noutputs; c++){
	b = nbins[c];
	for(i = 0; i < ninputs; i++){
	    if(b == 0) hpp->prob_list[c][i].bin_prob = NULL;
	    else hpp->prob_list[c][i].bin_prob = allocate_vector(b);
	    hpp->prob_list[c][i].bin_npat = NULL;
	}
    }
    /* allocate space for bin structures */
    if(hpp->grand_bins == PER_CLASS){
	hpp->bin_list = (HISTOGRAM_BIN **)
	  allocate_structmatrix(noutputs, ninputs, sizeof(HISTOGRAM_BIN));
	hpp->grand_bin_list = NULL;
	hpp->user_bin = NULL;
    } else if(hpp->grand_bins == GRAND_INPUT){
	hpp->grand_bin_list = (HISTOGRAM_BIN *)
	  allocate_structvector(ninputs, sizeof(HISTOGRAM_BIN));
	hpp->bin_list = NULL;
	hpp->user_bin = NULL;
    } else {
	hpp->user_bin = (HISTOGRAM_BIN *)
	  nn_malloc(sizeof(HISTOGRAM_BIN), "for user bin");
	hpp->bin_list = NULL;
	hpp->grand_bin_list = NULL;
    }
    /* allocate vectors in the bins */
    for(c = 0; c < noutputs; c++){
	if(hpp->grand_bins != PER_CLASS && c > 0) break;
	b = nbins[c];
	for(i = 0; i < ninputs; i++){
	    if(hpp->grand_bins == USER_BIN && i > 0) break;
	    switch(hpp->grand_bins){
	      case PER_CLASS:
		bin = &hpp->bin_list[c][i];
		break;
	      case GRAND_INPUT:
		bin = &hpp->grand_bin_list[i];
		break;
	      case USER_BIN:
		bin = hpp->user_bin;
		break;
	    }
	    bin->bin_ends = allocate_vector(b+1);
	    bin->sorted_inputs = allocate_vector(BIN_STORE_SIZE);
	    bin->sorted_length = BIN_STORE_SIZE;
	}
    }
    return new_model;
}
/*****************************************************************************
* NAME: histogram_train_init()
* DESCRIPTION:
*       Initialize model for training
*
* EXAMPLE:
*
*****************************************************************************/
void
histogram_train_init(int model)
{
    int c, i, b;
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    HISTOGRAM_PROB *prob;

    hpp = histpp[model];

    for (c = 0; c < hpp->noutputs; c++)
      hpp->class_cnt[c] = 0.;
    hpp->total_npatterns = 0.;
    for(c = 0; c < hpp->noutputs; c++){
	b = hpp->nbins[c];
	for(i = 0; i < hpp->ninputs; i++){
	    /* allocate space for training arrays */
	    prob = &hpp->prob_list[c][i];
	    if(prob->bin_npat == NULL && b != 0){
		prob->bin_npat = allocate_ivector(b);
	    }
	    prob->var = 0.;
	    prob->mean = 0.;
	}
    }
    for(c=0; c < hpp->noutputs; c++){
	if(hpp->grand_bins != PER_CLASS && c > 0) break;
	for(i = 0; i < hpp->ninputs; i++){
	    if(hpp->grand_bins == USER_BIN && i > 0) break;
	    switch(hpp->grand_bins){
	      case PER_CLASS:
		bin = &hpp->bin_list[c][i];
		break;
	      case GRAND_INPUT:
		bin = &hpp->grand_bin_list[i];
		break;
	      case USER_BIN:
		bin = hpp->user_bin;
		break;
	    }

	    bin->sorted_inputs =
	      reallocate_vector(bin->sorted_inputs, bin->sorted_length,
				BIN_STORE_SIZE);
	    bin->sorted_length = BIN_STORE_SIZE;
	    
	    bin->max = -MAXFLOAT;
	    bin->min = MAXFLOAT;
	}
    }
}
/*****************************************************************************
* NAME: histogram_test_init()
* DESCRIPTION:
*	initialize model for testing
*
* EXAMPLE:
*****************************************************************************/
void
histogram_test_init(int model, int adjust_priors, float *priors_list,
		    int use_gauss)
{
    int i, c;
    HISTOGRAM_PARAMS *hpp;

    hpp = histpp[model];

    if(adjust_priors){
	for(c = 0; c < hpp->noutputs; c++)
	  hpp->class_priors[c] = priors_list[c];
    } else {
	for(c = 0; c < hpp->noutputs; c++)
	  hpp->class_priors[c] = hpp->class_cnt[c]/hpp->total_npatterns;
    }
    /* initialize natural logs of variances */
    for(c = 0; c < hpp->noutputs; c++)
      for(i = 0; i < hpp->ninputs; i++)
	hpp->prob_list[c][i].LNvar = log((double)hpp->prob_list[c][i].var);
    hpp->use_gauss = use_gauss;
}
/*****************************************************************************
* NAME: histogram_test_alloc()
* DESCRIPTION:
*	initialize model for testing
*
* EXAMPLE:
*****************************************************************************/
void
histogram_test_alloc(int model)
{
    HISTOGRAM_PARAMS *hpp;

    hpp = histpp[model];

    if(hpp->outputs == NULL)
      hpp->outputs = allocate_vector(hpp->noutputs);
    if(hpp->gauss_outputs == NULL)
      hpp->gauss_outputs = allocate_vector(hpp->noutputs);
    if(hpp->class_priors == NULL)
      hpp->class_priors = allocate_vector(hpp->noutputs);
}

/*****************************************************************************
* NAME: histogram_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
*****************************************************************************/
void
histogram_free(int model)
{
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    HISTOGRAM_PROB *prob;
    int c, i;

    hpp = histpp[model];

    /* free vectors allocated in build */
    deallocate_vector(hpp->class_cnt);
    deallocate_ivector(hpp->nbins);

    /* free test vectors */
    if(hpp->outputs != NULL) deallocate_vector(hpp->outputs);
    if(hpp->gauss_outputs != NULL) deallocate_vector(hpp->gauss_outputs);
    if(hpp->class_priors != NULL) deallocate_vector(hpp->class_priors);

    /* free prob structures */
    for(c = 0; c < hpp->noutputs; c++){
	for(i = 0; i < hpp->ninputs; i++){
	    prob = &hpp->prob_list[c][i];
	    if(prob->bin_npat != NULL) deallocate_ivector(prob->bin_npat);
	    deallocate_vector(prob->bin_prob);
	}
    }
    deallocate_structmatrix((void **)hpp->prob_list);

    /* free bin structures */
    for(c = 0; c < hpp->noutputs; c++){
	if(hpp->grand_bins != PER_CLASS && c > 0) break;
	for(i = 0; i < hpp->ninputs; i++){
	    if(hpp->grand_bins == USER_BIN && i > 0) break;
	    switch(hpp->grand_bins){
	      case PER_CLASS:
		bin = &hpp->bin_list[c][i];
		break;
	      case GRAND_INPUT:
		bin = &hpp->grand_bin_list[i];
		break;
	      case USER_BIN:
		bin = hpp->user_bin;
		break;
	    }
	    deallocate_vector(bin->bin_ends);
	    deallocate_vector(bin->sorted_inputs);
	}
    }
    if(hpp->user_bin != NULL) nn_free((char *)hpp->user_bin, "for user bin");
    if(hpp->grand_bin_list != NULL)
      deallocate_structvector(hpp->grand_bin_list);
    if(hpp->bin_list != NULL) deallocate_structmatrix((void **)hpp->bin_list);

    				/* free structure itself */
    nn_free((char *) hpp, "for histogram parameter structure");
    histpp[model] = NULL;
}

/* Parameter Tables */
static PARAM paramtab[] = {
    {"ninputs", P_INT},
    {"noutputs", P_INT},
    {"total_npat", P_FLOAT},
    {"bin_type", P_INT},
    {"grand_type", P_INT},
};
#define TABSIZE (sizeof(paramtab)/sizeof(PARAM))

static PARAM paramtab_l[] = {
    {"class_cnt", P_FLOAT_VECTOR},
    {"nbins", P_INT_VECTOR},
};
#define TABSIZE_L (sizeof(paramtab_l)/sizeof(PARAM))

static PARAM paramtab_bin[] = {
    {"starts", P_FLOAT_VECTOR},
};
#define TABSIZE_BIN (sizeof(paramtab_bin)/sizeof(PARAM))

static PARAM paramtab_prob[] = {
    {"mean", P_FLOAT},
    {"var", P_FLOAT},
    {"probs", P_FLOAT_VECTOR},
};
#define TABSIZE_PROB (sizeof(paramtab_prob)/sizeof(PARAM))


/******************************
 * routines to set param variables for reading and writing a parameter file
 *****************************/
static void
histogram_set_paramtab(HISTOGRAM_PARAMS *hpp)
{
    param_setvarptr("ninputs", &hpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &hpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("total_npat",(int *)&hpp->total_npatterns,
		    paramtab, TABSIZE);
    param_setvarptr("bin_type", &hpp->bin_type, paramtab, TABSIZE);
    param_setvarptr("grand_type", &hpp->grand_bins, paramtab, TABSIZE);
}
static void
histogram_set_paramtab_l(HISTOGRAM_PARAMS *hpp)
{
    param_setvarptr("class_cnt", (int *)&hpp->class_cnt[0],
		    paramtab_l, TABSIZE_L);
    param_setdim("class_cnt", hpp->noutputs, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("nbins", (int *)&hpp->nbins[0],
		    paramtab_l, TABSIZE_L);
    param_setdim("nbins", hpp->noutputs, 0, paramtab_l, TABSIZE_L);
}
static void
histogram_set_paramtab_bin(HISTOGRAM_BIN *bin, int b)
{
    param_setvarptr("starts", (int *)&bin->bin_ends[0],
		    paramtab_bin, TABSIZE_BIN);
    param_setdim("starts", b+1, 0, paramtab_bin, TABSIZE_BIN);
}

static void
histogram_set_paramtab_prob(HISTOGRAM_PROB *prob, int b)
{
    param_setvarptr("mean", (int *)&prob->mean, paramtab_prob, TABSIZE_PROB);
    param_setvarptr("var", (int *)&prob->var, paramtab_prob, TABSIZE_PROB);
    param_setvarptr("probs", (int *)&prob->bin_prob[0],
		    paramtab_prob, TABSIZE_PROB);
    param_setdim("probs", b, 0, paramtab_prob, TABSIZE_PROB);
}

/*****************************************************************************
* NAME: histogram_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*       allocates space for and loads data into vectors and matricies
*
*       On error from param_getblk, returns the error value
*       (there is already an error check on the allocation, so we don't
*       have to do that here)
*
* EXAMPLE:
*****************************************************************************/
int
histogram_read_model(FILE *fp)
{
    int             model;
    int             rv, nbins;
    HISTOGRAM_PARAMS temp_hp, *hpp;
    HISTOGRAM_BIN *bin;
    int c, i;

    histogram_set_paramtab(&temp_hp);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
	return rv;

    temp_hp.class_cnt = allocate_vector(temp_hp.noutputs);
    temp_hp.nbins = allocate_ivector(temp_hp.noutputs);
    histogram_set_paramtab_l(&temp_hp);

    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
	return rv;

    /* allocate space for model and vectors */
    model = histogram_build(temp_hp.ninputs, temp_hp.noutputs,
			    temp_hp.bin_type, temp_hp.nbins,
			    temp_hp.grand_bins);
    hpp = histpp[model];
    hpp->total_npatterns = temp_hp.total_npatterns;
    for(c = 0; c < hpp->noutputs; c++){
	hpp->class_cnt[c] = temp_hp.class_cnt[c];
    }

    /* read the probabilities */
    for(c = 0; c < hpp->noutputs; c++){
	nbins = (int)hpp->nbins[c];
	for(i = 0; i < hpp->ninputs; i++){
	    histogram_set_paramtab_prob(&hpp->prob_list[c][i], nbins);
	    if((rv = param_getblk(fp, paramtab_prob, TABSIZE_PROB)) < 0)
	      return rv;
	}
    }
    /* read the bin starts */
    for(c = 0; c < hpp->noutputs; c++){
	if(hpp->grand_bins != PER_CLASS && c > 0) break;
	nbins = (int)hpp->nbins[c];
	for(i = 0; i < hpp->ninputs; i++){
	    if(hpp->grand_bins == USER_BIN && i > 0) break;
	    switch(hpp->grand_bins){
	      case PER_CLASS:
		bin = &hpp->bin_list[c][i];
		break;
	      case GRAND_INPUT:
		bin = &hpp->grand_bin_list[i];
		break;
	      case USER_BIN:
		bin = hpp->user_bin;
		break;
	    }
	    histogram_set_paramtab_bin(bin, nbins);
	    if((rv = param_getblk(fp, paramtab_bin, TABSIZE_BIN)) < 0)
	      return rv;
	}
    }
	
    deallocate_vector(temp_hp.class_cnt);
    deallocate_ivector(temp_hp.nbins);

    return (model);
}

/*****************************************************************************
* NAME: histogram_write_model()
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
*
* EXAMPLE:
*
******************************************************************************/
int
histogram_write_model(int model, FILE *fp)
{
    int             rv;
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    int c, i, nbins;

    hpp = histpp[model];
    				/* save scalars */
    histogram_set_paramtab(hpp);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
	return rv;

    				/* save vectors and matrices */
    histogram_set_paramtab_l(hpp);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
	return rv;

    /* save probabilities */
    for(c = 0; c < hpp->noutputs; c++){
	nbins = (int)hpp->nbins[c];
	for(i = 0; i < hpp->ninputs; i++){
	    fprintf(fp, "PROBABILITIES FOR CLASS %d, INPUT FEATURE %d\n",
		    c, i);
	    histogram_set_paramtab_prob(&hpp->prob_list[c][i], nbins);
	    if((rv = param_putblk(fp, paramtab_prob, TABSIZE_PROB)) < 0)
	      return rv;
	}
    }
    /* save bin starts */
    for(c = 0; c < hpp->noutputs; c++){
	if(hpp->grand_bins != PER_CLASS && c > 0) break;
	nbins = (int)hpp->nbins[c];
	for(i = 0; i < hpp->ninputs; i++){
	    if(hpp->grand_bins == USER_BIN && i > 0) break;
	    switch(hpp->grand_bins){
	      case PER_CLASS:
		fprintf(fp, "BIN STARTS FOR CLASS %d, INPUT FEATURE %d\n",
			c, i);
		bin = &hpp->bin_list[c][i];
		break;
	      case GRAND_INPUT:
		fprintf(fp, "BIN STARTS FOR INPUT FEATURE %d\n", i);
		bin = &hpp->grand_bin_list[i];
		break;
	      case USER_BIN:
		fprintf(fp, "BIN STARTS\n");
		bin = hpp->user_bin;
		break;
	    }
	    histogram_set_paramtab_bin(bin, nbins);
	    if((rv = param_putblk(fp, paramtab_bin, TABSIZE_BIN)) < 0)
	      return rv;
	}
    }
    return (0);
}


/************************************************************************
* NAME: histogram_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/

void *
histogram_param(int model, char *param_name, int class, int input)
{
    void           *param_address;
    HISTOGRAM_PARAMS *hpp;

    hpp = histpp[model];

    if (PEQ("ninputs"))
	param_address = &hpp->ninputs;
    else if (PEQ("noutputs") || PEQ("nclasses"))
	param_address = &hpp->noutputs;
    else if (PEQ("outputs"))
	param_address = &hpp->outputs;
    else if (PEQ("class_cnt"))
      param_address = &hpp->class_cnt;
    else if (PEQ("total_npatterns"))
      param_address = &hpp->total_npatterns;
    else if (PEQ("nbins"))
      param_address = &hpp->nbins;
    else if (PEQ("prob"))
      param_address = &hpp->prob_list[class][input];
    else if (PEQ("bin")) {
      if(hpp->grand_bins == PER_CLASS)
	param_address = &hpp->bin_list[class][input];
      else if(hpp->grand_bins == GRAND_INPUT)
	param_address = &hpp->grand_bin_list[input];
      else 
	param_address = hpp->user_bin;
    } else if (PEQ("grand"))
      param_address = &hpp->grand_bins;
    else {
	warning("histogram_param: Parameter %s not defined", param_name);
	param_address = NULL;
    }
    return param_address;
}


/*****************************************************************************
* NAME: histogram_inc_train1()
* DESCRIPTION:
*	Count the number of patterns per class,
*	find the width of each class,
*	sum up the means and variances for each Gaussian
* EXAMPLE:
*
*****************************************************************************/
void
histogram_inc_train1(int model, float *inputs, int desired_class)
{
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    int i, npat;

    hpp = histpp[model];
    if(hpp->grand_bins == USER_BIN) return;

    hpp->class_cnt[desired_class] += 1.;
    hpp->total_npatterns += 1.;
    /* find the range of the data */
    for(i = 0; i < hpp->ninputs; i++){
	if(hpp->grand_bins == GRAND_INPUT)
	  bin = &hpp->grand_bin_list[i];
	else
	  bin = &hpp->bin_list[desired_class][i];

	if(bin->max < inputs[i]) bin->max = inputs[i];
	if(bin->min > inputs[i]) bin->min = inputs[i];
    }
    /* if this is U_NPAT, sort the patterns for calculating bin starts */
    if(hpp->bin_type == U_NPAT){
	for(i = 0; i < hpp->ninputs; i++){
	    if(hpp->grand_bins){
		bin = &hpp->grand_bin_list[i];
		npat = (int)hpp->total_npatterns;
	    } else {
		bin = &hpp->bin_list[desired_class][i];
		npat = (int)hpp->class_cnt[desired_class];
	    }

	    /* make sure there is room for the new pattern */
	    if(bin->sorted_length < npat){
		bin->sorted_inputs =
		  reallocate_vector(bin->sorted_inputs, bin->sorted_length,
				    bin->sorted_length+BIN_STORE_SIZE);
		bin->sorted_length += BIN_STORE_SIZE;
	    }
	    hist_insert_heap(inputs[i], bin->sorted_inputs, npat);
	}
    }
    return;
}

/*****************************************************************************
* NAME: histogram_finish_train1()
* DESCRIPTION:
*	Calculate means and variances for the Gaussian part of out answer
*	Find the constant number of patterns or constant width for each bin
* EXAMPLE:
*
*****************************************************************************/
void
histogram_finish_train1(int model, float range_factor)
{
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    HISTOGRAM_PROB *prob;
    int c, i, b;
    int nPerClass, nRem, nbins, nPerBin;
    int bstart;
    float bends, total_width, bin_width, middle;

    hpp = histpp[model];
    if(hpp->grand_bins == USER_BIN) return;
    for(c = 0; c < hpp->noutputs; c++){
	/* get some numbers for use below */
	if(hpp->grand_bins){
	    if(c > 0) break;
	    nPerClass = (int)hpp->total_npatterns;
	} else {
	    nPerClass = (int)hpp->class_cnt[c];
	}
	nbins = hpp->nbins[c];
	if(nbins > 0){
	    if(hpp->bin_type == U_NPAT){
		nPerBin = nPerClass/nbins;
		nRem = nPerClass%nbins;
	    }
	} else {
	    nPerBin = 0;
	    nRem = 0;
	}
	/* reset class_cnt for the next round */
	hpp->class_cnt[c] = 0.;
	
	for(i = 0; i < hpp->ninputs; i++){
	    if(hpp->grand_bins)
	      bin = &hpp->grand_bin_list[i];
	    else
	      bin = &hpp->bin_list[c][i];
	    /* finish finding the range of the data */
	    total_width = bin->max - bin->min;
	    middle = bin->min + total_width/2.;
	    total_width *= range_factor;
	    bin->min = middle - total_width/2.;
	    bin->max = middle + total_width/2.;
	    
	    if(hpp->bin_type == U_NPAT){
		/* finish the sorting */
		hist_sort_heap(bin->sorted_inputs, nPerClass);
		/* calculate bin widths and probabilities */
		bin->bin_ends[0] = bin->min;
		for(bstart = 0, b = 1; b < nbins; b++){
		    /* save the start of this bin */
		    bstart += nPerBin;
		    if(b < nRem+1) bstart++;

		    if(bstart >= nPerClass) break;
		    bin->bin_ends[b] = (bin->sorted_inputs[bstart] +
					 bin->sorted_inputs[bstart -1])/2.;
		}
		bin->bin_ends[nbins] = bin->max;

		/* If this is per class uniform npat,
		   we already know npatterns per bin.
		   If we could skip the gaussian calculation,
		   we could finish up now. Let's at least save ourselves
		   finding which bin these patterns go into by saving the
		   counts per bin now */
		if(hpp->grand_bins == PER_CLASS){
		    prob = &hpp->prob_list[c][i];
		    for(b = 0; b < nbins; b++) prob->bin_npat[b] = nPerBin;
		    for(b = 0; b < nRem; b++) prob->bin_npat[b]++;
		}

	    } else {
		/* OR now that we know the total width of our data,
		   figure out the width of each bin */
		bin_width = total_width/nbins;
		bin_width = MAX(bin_width, MIN_PROB);
		for(b = 0, bends = bin->min; b < nbins;
		    b++, bends += bin_width)
		  bin->bin_ends[b] = bends;
		bin->bin_ends[nbins] = bin->max;
	    }
	}
    }
    /* reset total_npatterns for the next round */
    hpp->total_npatterns = 0;
    return;
}

/* this routine can be used in place of the first round to give
   each class and input the same bin-ing */
void
histogram_set_bins(int model, float min, float max)
{
    int b;
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    float bin_width, bends;
    hpp = histpp[model];

    bin = hpp->user_bin;
    bin->min = min;
    bin->max = max;
    bin_width = (max - min)/hpp->nbins[0];
    for(b = 0, bends = min; b < hpp->nbins[0]; b++, bends += bin_width){
	bin->bin_ends[b] = bends;
    }
    bin->bin_ends[hpp->nbins[0]] = max;
}

/*****************************************************************************
* NAME: histogram_inc_train2()
* DESCRIPTION:
*	Now that we know half of out probability equation, get the other half
* EXAMPLE:
*
*****************************************************************************/
void
histogram_inc_train2(int model, float *inputs, int desired_class)
{
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    HISTOGRAM_PROB *prob;
    int i, bin_num, nbins;
    float tmp, n;

    hpp = histpp[model];

    hpp->class_cnt[desired_class]++;
    hpp->total_npatterns += 1.;

    /* update the means and variances */
    n = hpp->class_cnt[desired_class];
    for(i = 0; i < hpp->ninputs; i++){
	prob = &hpp->prob_list[desired_class][i];
	tmp = inputs[i] - prob->mean;
	tmp *= tmp/n;
	prob->var += tmp;
	prob->var *= (n-1.)/n;
	prob->mean *= (n-1.)/n;
	prob->mean += inputs[i]/n;
    }

    /* we saved the counts per bin for this case in histogram_finish_train1 */
    if(hpp->grand_bins == PER_CLASS && hpp->bin_type == U_NPAT) return;

    if(hpp->grand_bins != PER_CLASS) nbins = hpp->nbins[0];
    else nbins = hpp->nbins[desired_class];
    /* keep count of the patterns that fall in each bin */
    for(i = 0; i < hpp->ninputs; i++){
	switch(hpp->grand_bins){
	  case PER_CLASS:
	    bin = &hpp->bin_list[desired_class][i];
	    break;
	  case GRAND_INPUT:
	    bin = &hpp->grand_bin_list[i];
	    break;
	  case USER_BIN:
	    bin = hpp->user_bin;
	    break;
	}
	/* find the bin, putting out of range patterns in the end bins */
	if(inputs[i] <= bin->min) bin_num = 0;
	else if(inputs[i] >= bin->max) bin_num = nbins-1;
	else bin_num = hist_get_bin(bin, inputs[i], hpp->nbins[desired_class]);
	
	/* increment the number of patterns for the probability calculation */
	if(bin_num >= 0)
	  hpp->prob_list[desired_class][i].bin_npat[bin_num]++;
    }
    return;
}

/*****************************************************************************
* NAME: histogram_finish_train2()
* DESCRIPTION:
*	Finish up the mean and variance calculations
*	Figure out the bin probabilities
*	P(x[i]|C) = N[j]/(N * w[j])
*		where N[j] is the number of patterns that fall in the bin
*		and w[j] is the width of the bin.
*	The sum of the bin probabilities over the whole width should
*		be 1.
* EXAMPLE:
*
*****************************************************************************/
void
histogram_finish_train2(int model, float minimum_variance)
{
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_PROB *prob;
    HISTOGRAM_BIN *bin;
    int c, i, b, nbins, nPerClass;
    float bin_width;

    hpp = histpp[model];

    for(c = 0; c < hpp->noutputs; c++){
	nbins = hpp->nbins[c];
	nPerClass = hpp->class_cnt[c];
	for(i = 0; i < hpp->ninputs; i++){
	    prob = &hpp->prob_list[c][i];
	    
	    switch(hpp->grand_bins){
	      case PER_CLASS:
		bin = &hpp->bin_list[c][i];
		break;
	      case GRAND_INPUT:
		bin = &hpp->grand_bin_list[i];
		break;
	      case USER_BIN:
		bin = hpp->user_bin;
		break;
	    }
	    
	    for(b = 0; b < nbins; b++){
		bin_width = bin->bin_ends[b+1] - bin->bin_ends[b];
		bin_width = MAX(bin_width, MIN_PROB);
		prob->bin_prob[b] = (prob->bin_npat[b]+1/hpp->total_npatterns)/
		  ((nPerClass+(nbins/hpp->total_npatterns))*bin_width);
	    }
	}
    }

    /* finish the variances */
    for(c = 0; c < hpp->noutputs; c++){
	for(i = 0; i < hpp->ninputs; i++){
	    prob = &hpp->prob_list[c][i];
	    if(prob->var < minimum_variance)
	      prob->var = minimum_variance;
	}
    }

    return;
}
/*****************************************************************************
* NAME: histogram_classify()
* DESCRIPTION:
*       Implements the histogram classifier to return an integer class.
*
* EXAMPLE:
*
*****************************************************************************/
int
histogram_classify(int model, float *inputs)
{
    HISTOGRAM_PARAMS *hpp;
    int class, known;

    hpp = histpp[model];

    known = histogram_eval(model, inputs);
    if(known){
	class = max_pick(hpp->outputs, hpp->noutputs);
    } else class = UNKNOWN_CLASS;
    return class;
}

/*********************************************************************
 *
 * the outputs of this classifiers are likelihoods,
 * 	y(C) = P(x|C)*P(C)
 *
 * P(x|C) = Product( P(x[i]|C)*(N-1)/N + G(i,C)/N ) for all input features i
 *	where:
 *	N is the number of patterns in Class C
 *	G(i,C) is the output of a one dimensional per class Gaussian
 *	and P(x[i]|C) is the probability estimate from the bin this
 *		input feature falls into
 *
 *	for a given bin:
 *		   (number of patterns int the class in the bin)
 *  P(x[i]|C) =     --------------------------------------------
 *		(number of patterns in the class) * Width of the bin
 *
 *
 * P(C) is an estimate of the probability of class C.
 **********************************************************************/
int
histogram_eval(int model, float *inputs)
{
    HISTOGRAM_PARAMS *hpp;
    HISTOGRAM_BIN *bin;
    HISTOGRAM_PROB *prob;
    int c, i, in_bin, missed_bin;
    int nbins, bin_num;
    float gauss;

    hpp = histpp[model];
    
    /* find the bin this pattern falls into */
    /* multiply up the probabilities */
    in_bin = 0;
    for(c = 0; c < hpp->noutputs; c++){
	hpp->outputs[c] = 0.;
	hpp->gauss_outputs[c] = 0.;
	nbins = hpp->nbins[c];
	missed_bin = 0;
	for(i = 0; i < hpp->ninputs; i++){
	    prob = &hpp->prob_list[c][i];
	    switch(hpp->grand_bins){
	      case PER_CLASS:
		bin = &hpp->bin_list[c][i];
		break;
	      case GRAND_INPUT:
		bin = &hpp->grand_bin_list[i];
		break;
	      case USER_BIN:
		bin = hpp->user_bin;
		break;
	    }
	    if(!missed_bin){
		bin_num = hist_get_bin(bin, inputs[i], nbins);
		if(bin_num >= 0){
		    hpp->outputs[c] += log(prob->bin_prob[bin_num]);
		} else {
		    missed_bin = 1;
		    hpp->outputs[c] = 0.;
		}
	    }
	    /* calculate gauss outputs */
	    if(!in_bin && hpp->use_gauss){
		gauss = inputs[i] - prob->mean;
		gauss *= gauss;
		gauss /= prob->var;
		gauss += prob->LNvar + LN_2PI;
		gauss *= -.5;
		hpp->gauss_outputs[c] += gauss;
	    }
	}
	if(!missed_bin) in_bin = 1;
    }
    if(in_bin){
	for(c = 0; c < hpp->noutputs; c++){
	    if(hpp->outputs[c] != 0){
		hpp->outputs[c] = exp(hpp->outputs[c]);
		hpp->outputs[c] *= hpp->class_priors[c];
	    }
	}
    } else {
	if(hpp->use_gauss){
	    for(c = 0; c < hpp->noutputs; c++){
		hpp->outputs[c] =
		  exp(hpp->gauss_outputs[c]);
		hpp->outputs[c] *= hpp->class_priors[c];
	    }
	} else {
	    for(c = 0; c < hpp->noutputs; c++)
	      hpp->outputs[c] = 0.;
	}
    }
    if(in_bin || hpp->use_gauss) return 1;
    else return 0;
}

/* insert the new number, x, into the heap, array */
static void
hist_insert_heap(float x, float *array, int n)
{
    int i, j;

    j = n;
    i = (n >> 1);
    while(i > 0){
	if(array[i-1] < x){
	    array[j-1] = array[i-1];
	    j = i;
	    i = (i >> 1);
	} else break;
    }
    array[j-1] = x;
}

/* do a sort on the heap array */
static void
hist_sort_heap(float *array, int n)
{
    int ir, i, j;
    float x;

    for(ir = n; ir > 1; ir--){
	/* swap the biggest number of the heap with the last current
	   place number, then redo the heap */
	x = array[ir-1];
	i = 1; j = 2;
	array[ir-1] = array[0];
	while(j < ir){
	    /* swap with the larger of the two children */
	    if(array[j-1] < array[j]) j++;
	    if(x < array[j-1]){
		array[i-1] = array[j-1];
		i = j;
		j = j << 1;
	    } else break;
	}
	array[i-1] = x;
    }
}

static int
hist_get_bin(HISTOGRAM_BIN *bin, float x, int nbins)
{
    int b;

    for(b = 0; b <= nbins && x > bin->bin_ends[b]; b++);
    if(b > nbins) return -1;
    else return b-1;
}
