/***--------------------------------------------------------------------------
* NAME:		histogram.c
* PURPOSE:	driver for training histogram classifier
*
* SEE ALSO: 	histogram_pack.c
* GLOBALS:	int debug;
*
* FUNCTIONS:
*	histogram	driver for histogram classifier
*		handles model creation and training
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/histogram/histogram_train.c,v 11.4 2001/08/23 20:11:57 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/histogram/histogram_train.c,v $
* CREATED: 24-Aug-94 lck				MIT LINCOLN LABORATORY
--------------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "histogram_pack.h"
#include "histogram.h"

#define MAXSTR (256)

static char prog_name[MAXSTR];	/* static variables */

void
histogram_train(int nraw, int ninputs, int *features, int noutputs,
		int normalize, int npatterns, int verbosity, char *data_path,
		char *data_fname, char *norm_fname, char *exp_path,
		char *param_fname, char *comment, int grand_bins,
		int bin_type, int *nBinsPerClass, float minimum_var,
		float range_factor, float user_min, float user_max)
{
    int		model, data_model;
    int		normalization=0;
    int		desired_class;
    int		n_read;
    float	*inputs;
    char	*tmp_path;
    FILE	*param_file;

    strcpy(prog_name, get_program_name());

    /* allocate and initialize model structure */
    model = histogram_build(ninputs, noutputs, bin_type, nBinsPerClass,
			    grand_bins);
    histogram_train_init(model);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    /* read data and initialize data handling routines */
    data_model =
      class_data_read(data_path, data_fname, npatterns, nraw, ninputs,
		      noutputs, features, normalization, verbosity);

    /* get class patterns from data handlers and train */
    if(grand_bins != USER_BIN){
	for(n_read = 0; n_read < npatterns; n_read++){
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    histogram_inc_train1(model, inputs, desired_class);
	}
	histogram_finish_train1(model, range_factor);
	class_data_epoch_init(data_model);
    } else {
	histogram_set_bins(model, user_min, user_max);
    }
    for(n_read = 0; n_read < npatterns; n_read++){
	if((inputs = class_data_get_next(data_model, &desired_class))
	   == NULL) break;
	histogram_inc_train2(model, inputs, desired_class);
    }
    histogram_finish_train2(model, minimum_var);
    class_data_exit(data_model);

    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "histogram model\n");
    histogram_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);

    /**************************************/

    /* free space */
    norm_free();
    histogram_free(model);
}
