/***---------------------------------------------------------------------------
* NAME:     hyper_pack.c
* PURPOSE:  Functions specific to hyper.c
*
* SEE ALSO: hyper.c
*
* GLOBALS:
*
* DESCRIPTION:
*   Functions in this file
*       hyper_build()       Allocate storage for model parameters.
*       hyper_init()        Initialize model parameters.
*       hyper_free()        Deallocates model storage
*       hyper_read_model()  Initializes a model from file
*       hyper_write_model() Writes a model to output file
*       hyper_param()       Returns pointers to elements in model
*       hyper_inc_ctrain()  Does training on a vector 
*       hyper_classify()    uses hyper_eval to find vector class
*       hyper_clear_cnts()  clears correct_cnt array
*       hyper_eval()        finds hyperspheres closest to a vector
*       hyper_prune()       removes uneeded points from training set
*       hyper_store_inputs()    stores a vector in training set
*       hyper_swap()        swaps two spheres in a model
*
* ALGORITHM:
*   Training: Training vectors are first normalized and randomized, if
*       requested.  
*       Each training vector is first classified.  If the vector
*       is correctly classified, it is ignored.  Otherwise, 
*       If the training vector was covered by a sphere of the wrong
*       class, that sphere is shrunk until it does not cover the
*       training vector
*       If the training vector is not covered by any sphere
*       of the correct class, a new one is added.
*
*       There is also pruning available.  The spheres are sorted
*       according to either size of how often they are used to
*       give a correct classification.  Then, spheres are deleted
*       either until all are larger or used more often than some
*       limit, or until there are N left
*
*   Testing:
*       The distance of the input vector to each center is calculated.
*       Favoring those spheres which the input vector falls inside of,
*       the class of the sphere whose center the input vector is
*       closest to is returned.
*       As an option, no class is returned if the vector does not
*       fall inside any sphere or falls inside spheres of differing
*       classes.
*       
* REFERENCE: Yuchun Lee, "Classifiers: Adaptive Modules in Pattern Recognition
*   Systems", May 1989
*   Bachelor's and Master's Thesis for the MIT Department of Electrical
*   Engineering and Computer Science
* 
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/hyper/hyper_pack.c,v 11.4 2001/08/23 20:12:01 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/hyper/hyper_pack.c,v $
* CREATED: 7-Mar-91 lck                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define hyper_pack_h_internal_types
#include "hyper_pack.h"
                /* contant definitions */
#define SPACE 1024
#define MODEL_STORE_SIZE 10

#define LARGE 10000.0
#define TINY 0.0000001
#define PEQ(y) (0==strcmp(param_name,y))

/* Internal Globals */
static HYPER_PARAMS **hyperpp;  /* array of parameters for each model */
static int      max_models = 0; /* number of hyper models allocated */

/*****************************************************************************
* NAME: hyper_build(ninputs, nclasses)
*   ninputs - number of dimensions in training vectors
*   nclasses - number of classes
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
******************************************************************************/
int
hyper_build(int ninputs, int nclasses, int npatterns)
{
    register int    new_model;
    HYPER_PARAMS *hpp;

    new_model = model_get_number((void ***)&hyperpp,
                 &max_models, sizeof(HYPER_PARAMS));
    hpp = hyperpp[new_model];

    if(npatterns <= 0) npatterns = SPACE;
      
            /* set fields which are array sizes */
    hpp->ninputs = ninputs;
    hpp->nclasses = nclasses;
    hpp->max_spheres = npatterns - 1;
    hpp->accum_trials = 0;

    /* allocate space for vectors */
    hpp->centroids = allocate_matrix(npatterns, ninputs);
    hpp->radii = allocate_vector(npatterns);
    hpp->classes = allocate_ivector(npatterns);
    hpp->correct_cnt = allocate_ivector(npatterns);
    hpp->distance = allocate_vector(npatterns);
    hpp->features = allocate_ivector(ninputs+1);
    hpp->outputs = allocate_vector(nclasses);

    return new_model;
}

/*****************************************************************************
* NAME: hyper_init(model, accum_trials, r_init, force_class, nn_features)
*   model       number of model to initialize
*   accum_trials    number of spheres stored in model
*   r_init      initial radius of new spheres
*   force_class use nearest neighbor to find class of points outside
*           of spheres or inside multiple spheres
*   nn_features features of normalized data to use
*
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
******************************************************************************/
void
hyper_init(int model, int accum_trials, float r_init, int force_class,
       int *nn_features)
{
    int             i;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];

    hpp->r_init = r_init;
    hpp->force_class = force_class;
    hpp->accum_trials = accum_trials;

    /* copy feature data into feature list */
    for(i = 0; i <= nn_features[0]; i++)
    hpp->features[i] = nn_features[i];
}

/*****************************************************************************
* NAME: hyper_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
******************************************************************************/
void
hyper_free(int model)
{
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];
            /* free vectors and matrices */
    deallocate_matrix(hpp->centroids);
    deallocate_vector(hpp->radii);
    deallocate_ivector(hpp->classes);
    deallocate_ivector(hpp->correct_cnt);
    deallocate_vector(hpp->distance);
    deallocate_ivector(hpp->features);
    deallocate_vector(hpp->outputs);

    /* free structure itself and clear pointer */
    nn_free((char *) hpp, "for hyper parameter structure");
    hyperpp[model] = NULL;
}

/*****************************************************************************
* NAME: hyper_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*   allocates space for and loads data into vectors and matricies
*
*   On error from param_getblk, returns the error value
*   (there is already an error check on the allocation, so we don't
*   have to do that here)
*
* EXAMPLE:
******************************************************************************/
int
hyper_read_model(FILE *fp, int *accum_trials)
{
    int             model;
    int             rv, ninputs, nclasses, force_classify;
    float       r_init;
    HYPER_PARAMS *hpp;

    /* read in scalars from model file */
    param_setvarptr("accum_trials", accum_trials, paramtab, TABSIZE);
    param_setvarptr("ninputs", &ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &nclasses, paramtab, TABSIZE);
    param_setvarptr("force_classify", &force_classify, paramtab, TABSIZE);
    param_setvarptr("r_init", (int *)&r_init, paramtab, TABSIZE);

    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    /*
     * allocate space for model and vectors. 
     */
    model = hyper_build(ninputs, nclasses, *accum_trials);
    hpp = hyperpp[model];

    /* read in vectors from model file */
    param_setvarptr("classes", &hpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim("classes", hpp->max_spheres + 1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("centroids", (int *)&hpp->centroids[0][0], paramtab_l, TABSIZE_L);
    param_setdim("centroids", hpp->max_spheres + 1, hpp->ninputs,
         paramtab_l, TABSIZE_L);
    param_setvarptr("radii", (int *)&hpp->radii[0], paramtab_l, TABSIZE_L);
    param_setdim("radii", hpp->max_spheres + 1, hpp->ninputs,
        paramtab_l,TABSIZE_L);
    param_setvarptr("features", &hpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", hpp->ninputs+1, 0, paramtab_l, TABSIZE_L);

    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

            /* set any remaining parameters */
    hpp->r_init = r_init;
    hpp->force_class = force_classify;
    hpp->accum_trials = *accum_trials;

    return (model);
}

/*****************************************************************************
* NAME: hyper_write_model()
* DESCRIPTION:
*   writes parameters and centroids for a model to a file
*   in ASII format
*
* EXAMPLE:
*
******************************************************************************/
int
hyper_write_model(int model, FILE *fp)
{
    int             rv;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];

    /* save scalars */
    param_setvarptr("accum_trials", &hpp->accum_trials, paramtab, TABSIZE);
    param_setvarptr("ninputs", &hpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &hpp->nclasses, paramtab, TABSIZE);
    param_setvarptr("force_classify", &hpp->force_class, paramtab, TABSIZE);
    param_setvarptr("r_init", (int *)&hpp->r_init, paramtab, TABSIZE);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    /* save vectors and matrices */
    param_setvarptr("classes", &hpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim("classes", hpp->accum_trials, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("centroids", (int *)&hpp->centroids[0][0], paramtab_l, TABSIZE_L);
    param_setdim("centroids", hpp->accum_trials, hpp->ninputs,
         paramtab_l, TABSIZE_L);
    param_setvarptr("radii", (int *)&hpp->radii[0], paramtab_l, TABSIZE_L);
    param_setdim("radii", hpp->accum_trials, hpp->ninputs,
         paramtab_l, TABSIZE_L);
    param_setvarptr("features", hpp->features, paramtab_l, TABSIZE_L);
    param_setdim("features", hpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    return (0);
}

/************************************************************************
* NAME: hyper_param(model, param_name)
* DESCRIPTION:
*   Takes model number and parameter name.
*   Returns a pointer to the named parameter.
*
*************************************************************************/

void           *
hyper_param(int model, char *param_name)
{
    void           *param_address;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];
    if (PEQ("accum_trials"))
    param_address = &hpp->accum_trials;
    else if (PEQ("centroids"))
    param_address = &hpp->centroids;
    else if (PEQ("radii"))
    param_address = &hpp->radii;
    else if (PEQ("classes"))
    param_address = &hpp->classes;
    else if (PEQ("ninputs"))
    param_address = &hpp->ninputs;
    else if (PEQ("nclasses"))
    param_address = &hpp->nclasses;
    else if (PEQ("r_init"))
    param_address = &hpp->r_init;
    else if (PEQ("features"))
    param_address = &hpp->features;
    else if (PEQ("outputs"))
    param_address = &hpp->outputs;
    else if (PEQ("force"))
      param_address = &hpp->force_class;
    else {
    warning("Parameter %s not defined", param_name);
    param_address = 0;
    }
    return param_address;
}               /* END hyper_param() */

/*****************************************************************************
* NAME: hyper_inc_ctrain()
* DESCRIPTION:
*   classifies the input data,
*   if the classification was incorrect, decides whether to add a new
*   sphere or to change the exist spheres to get a better result
*   returns the class selected in the classification
* EXAMPLE:
*
******************************************************************************/
int
hyper_inc_ctrain(int model, float *inputs, int desired_class)
{
    register int    class, i;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];

    /* search for closest match and closest sphere */
    hyper_clear_cnts(model);

    class = hyper_eval(model, inputs, desired_class);

    /* if there are no matches add a new sphere */
    if (hpp->cor_i == -1) {
    hyper_store_inputs(model, inputs, desired_class);
    return (class);
    }
    /* only modify on error */
    if (class == desired_class) {
    return (class);
    }
    /* set cor_dist to the real distance */
    hpp->cor_dist = hpp->distance[hpp->cor_i];

    /* if closest match doesn't cover inputs, make a new sphere */
    if (hpp->cor_dist > hpp->radii[hpp->cor_i]) {
    hyper_store_inputs(model, inputs, desired_class);
    return (class);
    }
    /* shrink covering spheres */
    for (i = 0; i < hpp->ninputs; i++)
    if (hpp->classes[i] != desired_class)
        if (hpp->radii[i] >= hpp->distance[i])
        hpp->radii[i] = hpp->distance[i] - TINY;

    return (class);
}

/*****************************************************************************
* NAME: hyper_classify()
* DESCRIPTION:
*   Returns the class of the closest sphere the input vector is inside of
*
* EXAMPLE:
*
******************************************************************************/
int
hyper_classify(int model, float *inputs)
{
    /* use fastdist ? */
    hyper_clear_cnts(model);
    return (hyper_eval(model, inputs, -1));

}

/*****************************************************************************
* NAME: hyper_clear_cnts()
* DESCRIPTION:
*   sets all entries in correct_cnt to 0
*
*****************************************************************************/
void
hyper_clear_cnts(int model)
{
    int             i;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];

    for (i = 0; i < hpp->accum_trials; i++) {
    hpp->correct_cnt[i] = 0;
    }
}

/*****************************************************************************
* NAME: hyper_eval()
* DESCRIPTION:
*   Returns the class of the closest sphere the input vector is inside of
*
* EXAMPLE:
*
******************************************************************************/
int
hyper_eval(int model, float *inputs, int desired_class)
{
    register int    i, class;
    register float  dist;
    int             ambiguous;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];


    hpp->min_dist = hpp->cor_dist = LARGE * 2;
    hpp->min_i = hpp->cor_i = -1;

    class = UNKNOWN_CLASS;
    ambiguous = 0;

    /* calculate distances to all centroids */
    for (i = 0; i < hpp->accum_trials; i++) {
    dist = l2dist(hpp->ninputs, inputs, &hpp->centroids[i][0]);
    hpp->distance[i] = dist;
    if (hpp->distance[i] > hpp->radii[i])
        dist += LARGE;
    else {
        if (!hpp->force_class)
        /*
         * if allowing UNKNOWN_CLASS answer and inside a sphere:
         * if there are other spheres which have covered our point
         * we need to check the class If it is different, we should
         * return UNKNOWN_CLASS
         */
        if (class != UNKNOWN_CLASS)
            if (class != hpp->classes[i]) {
            ambiguous = 1;
            class = UNKNOWN_CLASS;
            }
    }
    if (dist < hpp->min_dist) {
        /* if forcing classification, use nearest neighbor */
        if (hpp->force_class) {
        class = hpp->classes[i];
        } else if (hpp->distance[i] <= hpp->radii[i]) {
        if (!ambiguous) {
            class = hpp->classes[i];
        }
        }
        hpp->min_dist = dist;
        hpp->min_i = i;
        /* else outside. no match, so don't change things */
    }
    if (hpp->classes[i] == desired_class) {
        if (dist < hpp->cor_dist) {
        hpp->cor_dist = dist;
        hpp->cor_i = i;
        }
    }
    }
    if (hpp->accum_trials != 0) {
    if (class == desired_class)
        hpp->correct_cnt[hpp->min_i]++;
    }
    return (class);

}

/**************************************************************************
* NAME: hyper_store_inputs()
* DESCRIPTION:
*       assigns a class to the input vector.
*   If the class assigned is wrong, the vector is added to the
*       exemplar set
*   returns the class selected
*
*   The distances in the distance array must be up to date
* EXAMPLE:
*
******************************************************************************/
void
hyper_store_inputs(int model, float *inputs, int desired_class)
{
    register int    i, newlength;
    float           r_init;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];

    /* check to see if we have the room in centroids for new vector */
    if (hpp->accum_trials > hpp->max_spheres) {
    /* figure out how many vectors we can add to exemplar */
    hpp->max_spheres++; /* now the number of vectors in centroids */
    newlength = hpp->max_spheres + SPACE / (sizeof(float) * hpp->ninputs);
    hpp->centroids = reallocate_matrix(hpp->centroids, hpp->max_spheres,
                       hpp->ninputs, newlength);
    hpp->classes = reallocate_ivector(hpp->classes,
                      hpp->max_spheres, newlength);
    hpp->radii = reallocate_vector(hpp->radii,
                       hpp->max_spheres, newlength);
    hpp->distance = reallocate_vector(hpp->distance,
                      hpp->max_spheres, newlength);
    hpp->correct_cnt = reallocate_ivector(hpp->correct_cnt,
                          hpp->max_spheres, newlength);
    hpp->max_spheres = newlength - 1;
    }
    /* store input in centroids, class in classes, and increment counter */
    for (i = 0; i < hpp->ninputs; i++) {
    hpp->centroids[hpp->accum_trials][i] = inputs[i];
    }
    hpp->classes[hpp->accum_trials] = desired_class;

    /* if no overlap, check for overlap with centers of other classes */
    /* also check that we ourselves are not being overlapped */
    r_init = hpp->r_init;
    for (i = 0; i < hpp->accum_trials; i++) {
    if (hpp->classes[i] != desired_class){
        if (hpp->distance[i] <= r_init)
        /* we will overlap someone. Shrink */
        r_init = hpp->distance[i] - TINY;
        if (hpp->distance[i] <= hpp->radii[i])
        /* Someone overlaps us. Shrink him */
        hpp->radii[i] = hpp->distance[i] - TINY;
    }
    }
    hpp->radii[hpp->accum_trials] = r_init;

    hpp->accum_trials++;

}

/************************************************************************
* int hyper_prune(model, best_n, correct_sort, min)
* int n;        ;model number
* best_n;   if(best_n)
*           spheres are deleted until the remaining set
*           contains the "min" largest or best spheres.
*       else
*           spheres smaller than min or used less often
*           than min are pruned.
*
* correct_sort; if(correct_sort)
*           sort spheres by the number of correct classifications
*           each is responsible for
*       else
*           sort spheres by size
*
* For the second two sorting methods, the counts are collected by
* hyper_eval before hyper_prune is called
*
* float min;        ;interpreted as described above
*
* returns the number of spheres left in the model
************************************************************************/
int 
hyper_prune(int model, int best_n, int correct_sort, float min)
{
    int i, bw;
    int *index, *shift;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];

    if(best_n && hpp->accum_trials <= (int) min){
    return (hpp->accum_trials);
    }
    index = allocate_ivector(hpp->accum_trials);
    shift = allocate_ivector(hpp->accum_trials);
    for (i = 0; i < hpp->accum_trials; i++) 
      index[i] = i;

    if (correct_sort)
      /* sort by the number of correct classifications */
      rev_heap_isort(hpp->accum_trials, hpp->correct_cnt, index);
    else
      /* sort by size */
      rev_heap_sort(hpp->accum_trials, hpp->radii, index);

    /* make an array showing where points are in sorted order */
    for (i = 0; i < hpp->accum_trials; i++)
      shift[index[i]] = i;

    for(bw = hpp->accum_trials - 1; bw >= 0; bw--){
    if(best_n){
        if(bw < (int) min) break;
    } else {
        if(correct_sort){
        if(hpp->correct_cnt[index[bw]] > min) break;
        } else {
        if(hpp->radii[index[bw]] > min) break;
        }
    }
    /* swap it for some other entry and fix the heap */
    hyper_swap(model, index[bw], bw);
    shift[index[bw]] = shift[bw];
    index[shift[bw]] = index[bw];
    shift[bw] = index[bw] = bw;
    }

    deallocate_ivector(index);
    deallocate_ivector(shift);

    hpp->accum_trials = bw + 1;
    return (bw + 1);
}

/*****************************************************************************
* NAME: hyper_swap()
* DESCRIPTION:
*     swaps the spheres stored in fw and bw by swapping all the data for each
*****************************************************************************/
void
hyper_swap(int model, int fw, int bw)
{
    float           temp_f;
    int             temp_i, i;
    HYPER_PARAMS *hpp;

    hpp = hyperpp[model];

    for (i = 0; i < hpp->ninputs; i++) {
    temp_f = hpp->centroids[fw][i];
    hpp->centroids[fw][i] = hpp->centroids[bw][i];
    hpp->centroids[bw][i] = temp_f;
    }
    temp_i = hpp->classes[fw];
    hpp->classes[fw] = hpp->classes[bw];
    hpp->classes[bw] = temp_i;

    temp_f = hpp->radii[fw];
    hpp->radii[fw] = hpp->radii[bw];
    hpp->radii[bw] = temp_f;

    temp_i = hpp->correct_cnt[fw];
    hpp->correct_cnt[fw] = hpp->correct_cnt[bw];
    hpp->correct_cnt[bw] = temp_i;
}
