/***--------------------------------------------------------------------------
 * NAME:		hyper.c
 * PURPOSE:	Hypersphere classifier program
 *
 * SEE ALSO: hyper_pack.c hyper_subs.c
 *
 * DESCRIPTION:
 *	Functions in this file:
 *		hyper()	-- driver for hypersphere classifier
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/hyper/hyper_train.c,v 11.4 2001/08/23 20:12:01 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/hyper/hyper_train.c,v $
 * CREATED: 7-Mar-91 lck			MIT LINCOLN LABORATORY
 *************************************************************************/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include "hyper.h"
#include "hyper_pack.h"

#define MAXSTR (256)

/* static variables */
static char prog_name[MAXSTR];

void
hyper_train(int create_flag, int nraw, int ninputs, int *features,
	    int noutputs, int normalize, int npatterns, int debg,
	    int verbosity, int verror, int nepochs, int random_flag,
	    char *data_path, char *data_fname, char *norm_fname,
	    char *exp_path, char *error_fname, char *param_fname,
	    char **labels, char *comment, int adjust_priors,
	    int priors_npatterns, float *priors_list, int force_class,
	    int add_flag, int prune_flag, int correct_sort, int best_n,
	    float r_init, float prune_val)
{
    int		model, err_model, err_part_model, data_model;
    int		rv, normalization = 0, epoch;
    int		desired_class, class;
    int		n_read, i, n_trials;
    int		accum_trials = 0;
    int		map_flag = 0;	/* needed for error reporting routines */
    float	*inputs, *outputs;
    char	*tmp_path, *open_type;
    FILE	*error_file, *param_file;
    int		post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    if (create_flag){
	/* allocate and initialize
	 * model structure */
	model = hyper_build(ninputs, noutputs, npatterns);
	hyper_init(model, accum_trials, r_init, force_class, features);
	if (normalize) 
	  normalization = read_norm_file(path_build(data_path,norm_fname));
	else
	  norm_setup_none(nraw, noutputs);
    } else {		/* retrain  or test */
	tmp_path = path_build( exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	rv = norm_read_paramfile(param_file, &normalization);
	if(rv < 0)
	  fatal_error ("%s: read error %d %s %s", prog_name, rv,
		       "on normalization parameter file", tmp_path);
	model = hyper_read_model(param_file, &accum_trials);
	if (model < 0)
	  fatal_error("%s: read error %d on model file %s",
		      prog_name, model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of hyper model file");
	verify_model_params (model,
			     ninputs, *(int *) hyper_param(model, "ninputs"),
			     features,*(int **)hyper_param(model,"features"));
    }
    
    /* use features from the model */
    features = *(int **)hyper_param(model, "features");
    ninputs = *(int *)hyper_param(model, "ninputs");
    outputs = *(float **) hyper_param(model, "outputs");
    
    /*
     * print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {	/* open file for classification errors */
	if (create_flag) open_type = "w";
	else open_type = "a";
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      open_type, "for classification errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, CENTERS, 0, post_sum_to_one);
    err_part_model = report_build (verbosity, 0, noutputs);
    
    
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
	class_data_priors_init(data_model, priors_list, priors_npatterns);
    n_trials = class_data_get_npatterns(data_model);
    
    if (add_flag) {
	if(verbosity > 1) printf("adding and resizing spheres\n");
	report_epoch_header(err_model, stdout, n_trials);

	for (epoch = 0; epoch < nepochs; epoch++) {
	  class_data_epoch_init(data_model);
	  report_init (err_part_model, map_flag, ninputs, CENTERS,
		       0, post_sum_to_one);
	    
	    /* read and train on input from data file */
	    for (n_read = 0; n_read < n_trials; n_read++) {
		/* "read" input vector */
	      if((inputs = class_data_get_next(data_model, &desired_class))
		 == NULL) break;
		
		class = hyper_inc_ctrain(model, inputs, desired_class);
		report_class_trial(err_model, error_file,
				   inputs, outputs, desired_class, class,
				   class_data_get_input_num(data_model));
		report_class_trial(err_part_model, error_file,
				   inputs, outputs, desired_class, class,
				   class_data_get_input_num(data_model));
	    }
	    accum_trials = *(int *)hyper_param(model, "accum_trials");
	    report_centers_epoch(err_model, stdout, epoch, accum_trials);
      }
      report_last_epoch(err_part_model, stdout, epoch-1);
    }
    if(prune_flag){
	if(verbosity > 1) printf("pruning spheres\n");
	if (correct_sort){
	    hyper_clear_cnts(model);
	    class_data_epoch_init(data_model);
	    for (i = 0; i < n_trials; i++){
	      if((inputs = class_data_get_next(data_model, &desired_class))
		 == NULL) break;
		hyper_eval(model, inputs, desired_class);
	    }
	}
	accum_trials = hyper_prune(model, best_n, correct_sort, prune_val);
	if(verbosity > 1) printf("Number of spheres: %d\n", accum_trials);
    }
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of hyper model file");
    fprintf(param_file,"%s\n",comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "hyper model\n");
    hyper_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of hyper model file");
    if (verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    
    /* free space */
    class_data_exit(data_model);
    norm_free();
    hyper_free(model);
}
