/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/***-------------------------------------------------------------------------
* NAME: irbf_pack.c
* PURPOSE:  incrementally trained radial basis function classification
*       and prediction routines
* SEE ALSO:     irbf_train.c irbf_test.c irbf_cv.c
* DESCRIPTION:  Radial basis function classification and prediction routines.
*       including gassian and bias nodes.
*
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_pack.c,v 11.3 2001/08/23 20:12:05 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_pack.c,v $
* CREATED: 27-Nov-92 lck                MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/



/* Include necessary header files */
/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#include <nnlib.h>
#define irbf_pack_h_internal_types
#define irbf_pack_h_statics
#include "irbf_pack.h"

#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))


/* static global variables (for this file use only) */
static IRBF_PARAMS **irbfpp;    /* irbf_model set */
static int      max_models = 0; /* number of irbf models */


/************************************************************************
* NAME: irbf_build(ninputs, ngauss_hidden, bias, nclasses)
* DESCRIPTION:
*   allocates space for an irbf model with parameter values set from
*       function arguments.
*   returns the created model number (0,1, to ...).
*   ngauss_hidden is the number of
*       normal gaussian hidden nodes. bias is whether to add a
*       hidden node with constant output.
*       the extra added node has the highest index and a
*       constant output of 1. note that hspread is set in irbf_init
*************************************************************************/
int irbf_build(int ninputs, int ngauss_hidden, int bias, int nclasses,
           int grandvar)
{
    int new_model;
    int nhidden;
    IRBF_PARAMS *irpp;

    nhidden = ngauss_hidden;
    if(bias) nhidden += 1;

    /* check for bad parameters */
    if (ninputs <= 0)
      fatal_error("irbf_build: ninputs is negative = %d", ninputs);
    if (ngauss_hidden <= 0)
      fatal_error("irbf_build: ngauss_hidden is negative = %d", ngauss_hidden);
    if (nclasses <= 0)
      fatal_error("irbf_build: nclasses is negative = %d", nclasses);

    new_model = model_get_number((void ***)&irbfpp, &max_models,
                 sizeof(IRBF_PARAMS));
    irpp = irbfpp[new_model];

    /* set parameter from subroutine arguments */
    irpp->ninputs = ninputs;
    irpp->bias = bias;
    irpp->ngauss_hidden = ngauss_hidden;
    irpp->nhidden = nhidden;
    irpp->nclasses = nclasses;
    irpp->grandvar = grandvar;

    /* allocate the space for the vectors and matrices in the model */
    irpp->hidden_out = allocate_vector(nhidden);
    irpp->desired_outputs = allocate_vector(nclasses);
    irpp->outputs = allocate_vector(nclasses);
    irpp->means = allocate_matrix(ngauss_hidden, ninputs);
    if(grandvar){
    irpp->inv_var = allocate_vector(ninputs);
    irpp->new_inv_var = allocate_vector(ninputs);
    } else
      irpp->inv_var_m = allocate_matrix(ngauss_hidden, ninputs);
    irpp->weights = allocate_matrix(nhidden, nclasses);

    /* allocate space for feature selection array */
    irpp->features = allocate_ivector(ninputs+1);

    /* allocate space for priors adjustment arrays */
    irpp->class_cnt = allocate_vector(nclasses);
    irpp->priors_list = allocate_vector(nclasses);

    /* allocate space for temporary values for training */
    irpp->cluster_g = allocate_vector(nhidden);
    irpp->delta = allocate_vector(nhidden);

    /* initialize some things that I shouldn't in general need */
    irpp->adjust_priors = 0;
    /* return the model number */
    return (new_model);
}

/************************************************************************
* NAME: irbf_init()
*************************************************************************/
void irbf_init(int model, float hspread, int *features, float minimum_inv_var,
           int norm_outputs, float weight_eta, float mean_eta,
           float inv_var_eta, float max_magnitude, int cost_func,
           float cost_param)
{
    int i;
    IRBF_PARAMS *irpp;
    
    irpp = irbfpp[model];
    
    if (hspread < 0.0)
      fatal_error("irbf_init: hspread is negative = %f", hspread);
    irpp->hspread = hspread;
    irpp->extra_hspread = 1.; /* set this in test_init */
    if(weight_eta < 0.0)
      fatal_error("irbf_init: weight_eta %f is out of range", weight_eta);
    irpp->weight_eta = weight_eta;
    if(mean_eta < 0.0)
      fatal_error("irbf_init: mean_eta %f is out of range", mean_eta);
    irpp->mean_eta = mean_eta;
    if(inv_var_eta < 0.0)
      fatal_error("irbf_init: inv_var_eta %f is out of range", inv_var_eta);
    irpp->inv_var_eta = inv_var_eta;

    /* copy features into feature array */
    for (i = 0; i <= features[0]; i++) {
    irpp->features[i] = features[i];
    }
    for(i = 0; i < irpp->nclasses; i++)
      irpp->class_cnt[i] = 0.;
    irpp->total_npatterns = 0.;
    irpp->adjust_priors = 0;
    irbf_init_weights(irpp, max_magnitude);
    irpp->minimum_inv_var = minimum_inv_var;
    irpp->norm_outputs = norm_outputs;
    irpp->cost_func = cost_func;
    irpp->cost_param = cost_param;
    init_cost(cost_func, cost_param);
}

/***---------------------------------------------------------------------------
 * NAME:    irbf_init_weights
 * PURPOSE:initialize weights to small random numbers from +small to -small.
 --------------------------------------------------------------------------***/
static void irbf_init_weights(IRBF_PARAMS *irpp, float max_magnitude)
{
    float random_value;
    int i,j;
    
    for (i = 0; i < irpp->nhidden; i++)
      for (j = 0; j < irpp->nclasses; j++) {
      random_value = random_uniform();
      irpp->weights[i][j] = (2.0*random_value - 1.0) * max_magnitude;
      }
} /* irbf_init_weights() */


/***---------------------------------------------------------------------------
 * NAME:    irbf_initialize_cluster_centers
 * PURPOSE:initialize weights to small random numbers from +small to -small.
 --------------------------------------------------------------------------***/
void irbf_initialize_cluster_centers(int model, int clus_model)
{
    IRBF_PARAMS *irpp;
    int i,j;
    float **means, **spreads;
    
    irpp = irbfpp[model];

    /* get the pointers to the data */
    means = *(float ***)cluster_param(clus_model, "pcenter", 0);
    spreads = *(float ***)cluster_param(clus_model, "pvariance", 0);

    /*
     * copy the means and spreads to the model memory
     * -----------------------------------------------------
     */
    for (i = 0; i < irpp->ngauss_hidden; i++)
      for (j = 0; j < irpp->ninputs; j++)
    irpp->means[i][j] = means[i][j];
    if(irpp->grandvar){
    for(j = 0; j < irpp->ninputs; j++){
        irpp->inv_var[j] = 0.;
    }
    }
    for(i = 0; i < irpp->ngauss_hidden; i++){
    for(j = 0; j < irpp->ninputs; j++){
        if(irpp->grandvar){
        irpp->inv_var[j] += spreads[i][j];
        } else {
        if(spreads[i][j] > 0)
          irpp->inv_var_m[i][j] = 1/spreads[i][j];
        else
          irpp->inv_var_m[i][j] = 1/irpp->minimum_inv_var;
        if (irpp->inv_var_m[i][j] < irpp->minimum_inv_var)
          irpp->inv_var_m[i][j] = irpp->minimum_inv_var;
        }
    }
    }
    if(irpp->grandvar){
    for(j = 0; j < irpp->ninputs; j++){
        irpp->inv_var[j] /= (float)irpp->ngauss_hidden;
        if(irpp->inv_var[j] > 0)
          irpp->inv_var[j] = 1./irpp->inv_var[j];
        else
          irpp->inv_var[j] = 1./irpp->minimum_inv_var;
        if(irpp->inv_var[j] < irpp->minimum_inv_var){
        irpp->inv_var[j] = irpp->minimum_inv_var;
        }
    }
    }
} /* irbf_initialize_cluster_centers() */
/************************************************************************
* NAME: irbf_test_init()
* DESCRIPTION:
*   set model parameters needed for testing, not for training
*
* example:
*
*************************************************************************/
void irbf_test_init(int model, int adjust_priors, float *priors_list,
            float extra_hspread)
{
    int i;
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];
    irpp->extra_hspread = extra_hspread;
    irpp->adjust_priors = adjust_priors;
    if(adjust_priors)
      for(i = 0; i < irpp->nclasses; i++)
    irpp->priors_list[i] = priors_list[i];
}

/************************************************************************
* NAME: irbf_free()
* DESCRIPTION:
*   deallocate storage used in a model
*
* example:
*
*************************************************************************/
void irbf_free(int model)
{
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];

    deallocate_vector(irpp->hidden_out);
    deallocate_vector(irpp->desired_outputs);
    deallocate_vector(irpp->outputs);
    deallocate_matrix(irpp->means);
    if(irpp->grandvar){
    deallocate_vector(irpp->inv_var);
    deallocate_vector(irpp->new_inv_var);
    } else deallocate_matrix(irpp->inv_var_m);
    deallocate_matrix(irpp->weights);
    deallocate_ivector(irpp->features);
    deallocate_vector(irpp->class_cnt);
    deallocate_vector(irpp->priors_list);
    deallocate_vector(irpp->cluster_g);
    deallocate_vector(irpp->delta);

    nn_free((char *)irbfpp[model], "freeing irbf model");
    irbfpp[model] = NULL;
}

/***---------------------------------------------------------------------------
* NAME:  irbf_read_model()
* DESCRIPTION:
*   takes a file pointer as input.
*   reads irbf model from parameter file.
*   builds irbf model using parameters.
*   returns irbf model number.
* notes:
*   see irbf_write_paramfile().
--------------------------------------------------------------------------***/

static void
irbf_set_scalar_params(IRBF_PARAMS *irpp)
{
    param_setvarptr("ninputs", &irpp->ninputs, irbfscalars, NIRBFSCALARS);
    param_setvarptr("nclasses", &irpp->nclasses, irbfscalars, NIRBFSCALARS);
    param_setvarptr("ngauss_hidden", &irpp->ngauss_hidden,
            irbfscalars, NIRBFSCALARS);
    param_setvarptr("bias", &irpp->bias, irbfscalars, NIRBFSCALARS);
    param_setvarptr("nhidden", &irpp->nhidden, irbfscalars, NIRBFSCALARS);
    param_setvarptr("hspread", (int *)&irpp->hspread, irbfscalars, NIRBFSCALARS);
    param_setvarptr("total_npat", (int *)&irpp->total_npatterns,
            irbfscalars, NIRBFSCALARS);
    param_setvarptr("norm_outs", &irpp->norm_outputs,irbfscalars,NIRBFSCALARS);
    param_setvarptr("grandvar", &irpp->grandvar, irbfscalars, NIRBFSCALARS);
    param_setvarptr("cost_func", &irpp->cost_func, irbfscalars, NIRBFSCALARS);
    param_setvarptr("cost_param", (int *)&irpp->cost_param,
             irbfscalars, NIRBFSCALARS);
}

static void
irbf_set_array_params(IRBF_PARAMS *irpp)
{
    param_setvarptr("centroids", (int *) &irpp->means[0][0],
            irbfarrays, NIRBFARRAYS);
    param_setdim("centroids", irpp->ngauss_hidden,
         irpp->ninputs, irbfarrays, NIRBFARRAYS);
    param_setvarptr("weights", (int *) &irpp->weights[0][0],
            irbfarrays, NIRBFARRAYS);
    param_setdim("weights", irpp->nhidden, irpp->nclasses,
         irbfarrays, NIRBFARRAYS);
    param_setvarptr("features", &irpp->features[0], irbfarrays, NIRBFARRAYS);
    param_setdim("features", irpp->ninputs+1, 0, irbfarrays, NIRBFARRAYS);
    param_setvarptr("class_cnt", (int *)&irpp->class_cnt[0],
            irbfarrays, NIRBFARRAYS);
    param_setdim("class_cnt", irpp->nclasses, 0, irbfarrays, NIRBFARRAYS);
}

static void
irbf_set_var_array_params(IRBF_PARAMS *irpp)
{
    param_setvarptr("inv_var", (int *) &irpp->inv_var[0],
            irbfvar_array, NIRBFVAR_ARRAY);
    param_setdim("inv_var", irpp->ninputs, 0,
         irbfvar_array, NIRBFVAR_ARRAY);
}

static void
irbf_set_var_matrix_params(IRBF_PARAMS *irpp)
{
    param_setvarptr("inv_var", (int *) &irpp->inv_var_m[0][0],
            irbfvar_matrix, NIRBFVAR_MATRIX);
    param_setdim("inv_var", irpp->ngauss_hidden, irpp->ninputs,
         irbfvar_matrix, NIRBFVAR_MATRIX);
}

int
irbf_read_model(FILE *fp, float weight_eta, float mean_eta,
        float inv_var_eta, float minimum_inv_var)
{
    int     model;
    int     rv;
    IRBF_PARAMS temp, *irpp;

    /* get the irbf model scalars */
    irbf_set_scalar_params(&temp);
    if ((rv = param_getblk(fp, irbfscalars, NIRBFSCALARS)) < 0) return rv;

    /* allocate the space needed to store the new model */
    model = irbf_build(temp.ninputs, temp.ngauss_hidden, temp.bias, 
               temp.nclasses, temp.grandvar);
    irpp = irbfpp[model];

    /* fill in parameters set during training or by irbf_init() */
    /*irpp->nhidden is ngauss_hidden + bias */
    irpp->hspread = temp.hspread;
    irpp->extra_hspread = 1.; /* set this in test_init */
    irpp->total_npatterns = temp.total_npatterns;
    irpp->norm_outputs = temp.norm_outputs;
    irpp->cost_func = temp.cost_func;
    irpp->cost_param = temp.cost_param;
    init_cost(irpp->cost_func, irpp->cost_param);

    irpp->weight_eta = weight_eta;
    irpp->mean_eta = mean_eta;
    irpp->inv_var_eta = inv_var_eta;
    irpp->minimum_inv_var = minimum_inv_var;

    /* set ptrs and dimensions for allocated matrices */
    irbf_set_array_params(irpp);

    /* get the irbf model arrays */
    if ((rv = param_getblk(fp, irbfarrays, NIRBFARRAYS)) < 0) return rv;

    if(irpp->grandvar){
    irbf_set_var_array_params(irpp);
    if ((rv = param_getblk(fp, irbfvar_array, NIRBFVAR_ARRAY)) < 0)
      return rv;
    } else {
    irbf_set_var_matrix_params(irpp);
    if ((rv = param_getblk(fp, irbfvar_matrix, NIRBFVAR_MATRIX)) < 0)
      return rv;
    }

    return model;

}               /* irbf_read_model() */

/***---------------------------------------------------------------------------
* NAME:  irbf_write_model()
* DESCRIPTION:
*   takes a model number and file pointer as input.
*   writes irbf model to parameter file.
* notes:
*   irbf model is written to parameter file in two blocks.
*   the first block contains the scalars:
*       ninputs
*       ngauss_hidden
*       bias
*       nhidden
*       nclasses
*       hspread
*   the second block contains the arrays:
*       means
*       spreads
*       weights
--------------------------------------------------------------------------***/
void irbf_write_model(int model, FILE *fp)
{
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];

    /* get the irbf model scalars */
    irbf_set_scalar_params(irpp);
    param_putblk(fp, irbfscalars, NIRBFSCALARS);    /* scalars */

    /* set ptrs and dimensions for allocated matrices */
    irbf_set_array_params(irpp);
    /* reset the length for the features array to be sure it is right */
    param_setdim("features", irpp->features[0]+1, 0, irbfarrays, NIRBFARRAYS);

    param_putblk(fp, irbfarrays, NIRBFARRAYS);  /* arrays */

    if(irpp->grandvar){
    irbf_set_var_array_params(irpp);
    param_putblk(fp, irbfvar_array, NIRBFVAR_ARRAY);
    } else {
    irbf_set_var_matrix_params(irpp);
    param_putblk(fp, irbfvar_matrix, NIRBFVAR_MATRIX);
    }

}               /* irbf_write_model() */


/************************************************************************
* NAME: irbf_param(model, param_name)
* DESCRIPTION:
*   takes model number and parameter name.
*   returns a pointer to the parameter.
*   function really not necessary if not interested in accessing
*   parameters (reading) from main()
*************************************************************************/

void * irbf_param(int model, char *param_name)
{
    void *param_address;
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];

    /* return the requested parameter */
    if (PEQ("ninputs"))     param_address = &irpp->ninputs;
    else if (PEQ("ngauss_hidden")) param_address = &irpp->ngauss_hidden;
    else if (PEQ("bias"))   param_address = &irpp->bias;
    else if (PEQ("nhidden"))    param_address = &irpp->nhidden;
    else if (PEQ("nclasses"))   param_address = &irpp->nclasses;
    else if (PEQ("hspread"))    param_address = &irpp->hspread;
    else if (PEQ("features"))   param_address = &irpp->features;
    else if (PEQ("means"))  param_address = &irpp->means;
    else if (PEQ("inv_var"))    param_address = &irpp->inv_var;
    else if (PEQ("inv_var_m"))  param_address = &irpp->inv_var_m;
    else if (PEQ("norm_outputs")) param_address = &irpp->norm_outputs;
    else if (PEQ("outputs"))    param_address = &irpp->outputs;
    else if (PEQ("grandvar"))   param_address = &irpp->grandvar;
    else if (PEQ("weights"))    param_address = &irpp->weights;
    else if (PEQ("class_cnt"))  param_address = &irpp->class_cnt;
    else if (PEQ("total_npatterns")) param_address = &irpp->total_npatterns;
    else {
    warning("parameter %s not defined", param_name);
    param_address = NULL;
    }
    return (param_address);

}

/************************************************************************
* NAME: irbf_inc_ctrain(model,invec,class);
* DESCRIPTION:
*   takes irbf model number, one input vector, and its class.
*   trains the particular irbf model with that input.
*   must call irbf_build() and irbf_init() before this!
*************************************************************************/
int irbf_inc_ctrain(int model, float *invec, int class)
{
    int i;
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];

    irpp->class_cnt[class] += 1.;
    irpp->total_npatterns += 1.;
    for(i = 0; i < irpp->nclasses; i++){
    irpp->desired_outputs[i] = 0.;
    }
    irpp->desired_outputs[class] = 1.;

    irbf_inc_train(model, invec, irpp->desired_outputs);

    return(max_pick(irpp->outputs, irpp->nclasses));
}

/************************************************************************
* NAME: irbf_inc_train(model,invec,desired);
* DESCRIPTION:
*   takes irbf model number, one input vector, and vector of desired
*   outputs.
*   trains the particular irbf model with that input.
*   must call irbf_build() and irbf_init() before this!
*************************************************************************/
void irbf_inc_train(int model, float *invec, float *desired)
{
    /* generate the irbf outputs for one pattern */
    irbf_eval(model, invec);

    /* do back propagation on each of the variables */
    irbf_backprop(model, invec, desired);
}

/************************************************************************
* NAME: irbf_backprop(model, desired, outputs);
* DESCRIPTION:
*   updates weights, means, and variances of model, based on
*   hidden and final outputs calculated by irbf_eval
*************************************************************************/
static void irbf_backprop(int model, float *inputs, float *desired)
{
    int i, j, k;
    float gout, weight_g, temp, diff;
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];
    if(irpp->grandvar)
      for(k = 0; k < irpp->ninputs; k++) irpp->new_inv_var[k] = 0.;

    for(i = 0; i < irpp->nhidden; i++){
    irpp->cluster_g[i] = gout = 0.;
    /* update the weights */
    calc_output_deriv(desired, irpp->outputs, 0., irpp->delta,
              irpp->nclasses, LINEAR, 1., irpp->cost_func);
    for(j = 0; j < irpp->nclasses; j++){
        weight_g = irpp->delta[j] * irpp->hidden_out[i];
        irpp->cluster_g[i] += weight_g*irpp->weights[i][j];
        gout += weight_g * irpp->outputs[j];
        irpp->weights[i][j] -= irpp->weight_eta * weight_g;
    }
    /* don't update the center for the bias node, it doesn't have one */
    if(i == irpp->ngauss_hidden) break;
    /* update the means and non-grand variances */
    for(k = 0; k < irpp->ninputs; k++){
        temp = diff = inputs[k] - irpp->means[i][k];
        if(irpp->grandvar)
          temp *= irpp->inv_var[k]/irpp->hspread;
        else
          temp *= irpp->inv_var_m[i][k]/irpp->hspread;
        if(irpp->norm_outputs)
          temp *= irpp->cluster_g[i]-gout;
        else
          temp *= irpp->cluster_g[i];
        irpp->means[i][k] -= irpp->mean_eta*temp;

        /* update variances */
        temp = diff * diff;
        if(irpp->norm_outputs)
          temp *= irpp->cluster_g[i] - gout;
        else
          temp *= irpp->cluster_g[i];
        if(irpp->grandvar){
        irpp->new_inv_var[k] += temp;
        } else {
        temp /= 2*irpp->hspread;
        /* a plus, because we are updating the inverse variance */
        irpp->inv_var_m[i][k] += irpp->inv_var_eta*temp;
        if(irpp->inv_var_m[i][k] < irpp->minimum_inv_var)
          irpp->inv_var_m[i][k] = irpp->minimum_inv_var;
        }
    }
    }
    /* update grand variances */
    if(irpp->grandvar){
    for(k = 0; k < irpp->ninputs; k++){
        irpp->new_inv_var[k] /= 2*irpp->hspread;
        /*  a plus, because we are updating the inverse variance */
        irpp->inv_var[k] += irpp->inv_var_eta * irpp->new_inv_var[k];
        if(irpp->inv_var[k] < irpp->minimum_inv_var)
          irpp->inv_var[k] = irpp->minimum_inv_var;
    }
    }
}
/************************************************************************
* NAME: irbf_classify(model, invec)
* DESCRIPTION:
*   takes irbf model number and a pointer to an input vector.
*   returns the class assigned to the input vector and fills in the
*       output vector.
*************************************************************************/
int irbf_classify(int model, float *invec)
{
    int j;
    int outclass;   /* calculated class of the invec */
    IRBF_PARAMS *irpp;
    float total;


    irpp = irbfpp[model];

    irbf_eval(model, invec);

    if(irpp->adjust_priors){
    /* adjust priors */
    for(j = 0; j < irpp->nclasses; j++){
        irpp->outputs[j] *= irpp->priors_list[j];
        irpp->outputs[j] /= irpp->class_cnt[j]/irpp->total_npatterns;
    }
    /* scale adjusted outputs to sum to 1 */
    total = 0.;
    for(j = 0; j < irpp->nclasses; j++) total += irpp->outputs[j];
    for(j = 0; j < irpp->nclasses; j++) irpp->outputs[j] /= total;
    }

    /* pick the maximum out of the z array list */
    outclass = max_pick(irpp->outputs, irpp->nclasses);

    /* return the class */
    return (outclass);
}

/************************************************************************
* NAME: irbf_eval(model, invec)
* DESCRIPTION:
*   takes irbf model number and a pointer to an input vector
*   and fill the model's output vector with the classifier outputs
*   for the input vector
*************************************************************************/
void irbf_eval(int model, float *invec)
{
    int j, k;
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];

    /* generate the irbf outputs */
    irbf_hidden_eval(model, invec, irpp->extra_hspread);

    /* calculate the output vector */
    for (j = 0; j < irpp->nclasses; j++) {
    irpp->outputs[j] = 0.0;
    for (k = 0; k < irpp->nhidden; k++)
      irpp->outputs[j] += irpp->weights[k][j] * irpp->hidden_out[k];
    }
}
/************************************************************************
* NAME: irbf_hidden_eval(model, invec, extra_hspread)
* DESCRIPTION:
*   takes as input a pointer to the input vector and a pointer to
*       an output vector that will contain the hidden node outputs on
*       return.
*       extra_hspread multiplies the variances of all
*       basis functions uniformly.
*       it only does this for this evaluation and is not permanent.
**************************************************************************/
static void irbf_hidden_eval(int model, float *invec, float extra_hspread)
{
    int node, i;
    float diff, factor;
    float total_log;
    IRBF_PARAMS *irpp;

    irpp = irbfpp[model];

    /*
     * calculate log10() outputs of all gaussian nodes first
     * ------------------------------------------------
     */
    factor = -M_LOG10E / (2.0 * irpp->hspread * extra_hspread);
    for (node = 0; node < irpp->ngauss_hidden; node++) {
    irpp->hidden_out[node] = 0.;
    for(i = 0; i < irpp->ninputs; i++){
        diff = invec[i] - irpp->means[node][i];
        diff *= diff;
        if(irpp->grandvar)
          irpp->hidden_out[node] += diff * irpp->inv_var[i];
        else 
          irpp->hidden_out[node] += diff * irpp->inv_var_m[node][i];
    }
    irpp->hidden_out[node] *= factor;
    }

    if(irpp->norm_outputs){
    /* compute average node output */
    total_log = irpp->hidden_out[0];
    for (node = 1; node < irpp->ngauss_hidden; node++)
      total_log = logadd(total_log, irpp->hidden_out[node]);

    /* normalize node outputs and convert to linear */
    for (node = 0; node < irpp->ngauss_hidden; node++){
        irpp->hidden_out[node] -= total_log;
        irpp->hidden_out[node] =
          nn_exp10((double) (irpp->hidden_out[node]));
    }
    } else {
    /* just convert to linear */
    for (node = 0; node < irpp->ngauss_hidden; node++)
      irpp->hidden_out[node] = nn_exp10((double) (irpp->hidden_out[node]));
    }

    /* add output of bias node with constant output */
    if (irpp->bias) {
    node = irpp->nhidden - 1;
    irpp->hidden_out[node] = 1.0;
    }
}


/*****************************************
irbf_read_cluster()
*****************************************/
int irbf_read_cluster(char *param_path, FILE *param_file, int *pninputs,
              int *pncenters, int *features)
{
    int model, nclasses;

    model = cluster_read_paramfile(param_file, pninputs, &nclasses, pncenters,
                   features);
    if(model < 0)
      fatal_error("irbf: read error %d on parameter file %s", model,
          param_path);

    if(*pncenters <= 0)
      fatal_error("irbf_read_cluster: No centers from clustering");

    /* insure that the centers are not stored by class */
    cluster_merge_centers(model);
    return model;
}
