/***--------------------------------------------------------------------------
 * NAME:	irbf_seq_cv.c
 * PURPOSE:	cross valid driver for Radial Basis Function sequence
 *		classifier
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_cv.c,v 11.4 2001/08/23 20:12:05 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_cv.c,v $
 * CREATED: 17-Sep-93 lck			MIT LINCOLN LABORATORY
 *************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "irbf_seq.h"
#include "irbf_seq_pack.h"
#include "irbf_pack.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

static char     prog_name[MAXSTR];	/* static variables */

void
irbf_seq_cross_valid(int db_type, int noutputs, int nsequences, int debg,
		     int verbosity, int verror, int random_flag,
		     int cv_nfolds, int cv_nsplits, int use_cv_splits,
		     int *cv_splits, char **cv_train_mask,
		     char **cv_test_mask, int cv_random, char *data_path,
		     char *data_fname, char *exp_path, char *error_fname,
		     char *param_fname, char **labels, char *comment,
		     int nepochs, int bias, int grand_flag,
		     float extra_hspread, float hspread, float max_magnitude,
		     char *cluster_param_fname, float minimum_variance,
		     float weight_eta, float mean_eta, float inv_var_eta,
		     int cost_func, float cost_param, int norm_outputs,
		     int log_outputs, float clip_floor)
{
    /* local variables */
    int		model, seq_model, data_model, err_model, err_part_model;
    int		clus_model, rv;
    int		*features, nraw, ninputs, model_ninputs;
    int		nwindow_frames, *window_frames;
    int		normalization = 0;
    int		desired_class, class;
    int		i, n_frames, seq_num, n_files, epoch, n_read;
    int		ngauss_hidden;
    int		fold;
    float	*inputs, *seq_sums;
    float	*priors_list;
    char	clus_path[MAXSTR];
    FILE	*error_file, *param_file, *clus_file;
    int		post_sum_to_one = 0;

    strcpy(prog_name, get_program_name());
    
    /* read the cluster file */
    strncpy(clus_path,path_build(exp_path, cluster_param_fname), MAXSTR-1);    
    clus_file = nn_fopen(clus_path, "r", "for read of cluster parameters");
    if (verbosity > 1) {
	printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	fflush(stdout);
    }
    /* read norm model from cluster file to get feature selection */
    rv = norm_read_paramfile(clus_file, &normalization);
    if (rv < 0)
      fatal_error("irbf: read error %d %s %s",
		  rv, "on normalization parameter file", clus_path);
    rv = feature_read_paramfile(clus_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on feature selection parameters in file", clus_path);
    rv = window_read_paramfile(clus_file, &nwindow_frames, &window_frames);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on windowing parameters in file", clus_path);
    
    /* read first clusters from cluster file */
    /* features was allocated in features_read_paramfile, so it is ok here */
    clus_model = irbf_read_cluster(clus_path, clus_file, &model_ninputs,
				   &ngauss_hidden, features);
    if (clus_model < 0)
      fatal_error("%s: read error %d %s %s", prog_name, clus_model,
		  "on cluster parameter file", clus_path);
    
    /* build the irbf model */
    if (verbosity > 1){
	printf ("\nBuilding irbf model\n");
	fflush(stdout);
    }
    model = irbf_build (model_ninputs, ngauss_hidden, bias, noutputs,
			grand_flag);
    irbf_init (model, hspread, features, minimum_variance, norm_outputs,
	       weight_eta, mean_eta, inv_var_eta, max_magnitude, cost_func,
	       cost_param);
    irbf_initialize_cluster_centers(model, clus_model);
    cluster_free(clus_model);
    
    /* build the sequence model */
    seq_model = irbf_seq_build(noutputs);
    irbf_seq_init(seq_model, model, clip_floor);
    seq_sums = *(float **) irbf_seq_param(seq_model, "seq_sums");
    /* irbf should adjust for uniform priors for the frames */
    priors_list = allocate_vector(noutputs);
    for(i = 0; i < noutputs; i++) priors_list[i] = 1.;

    /* prepare error handling */
    if (verror > 0) {		/* open file for classification errors */
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      "w", "for classification errors");
    } else error_file = NULL;
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, db_type, model_ninputs, SQ_ERR, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize seq_data routines */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    nsequences = seq_data_get_nfiles(data_model);
    
    seq_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		     cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) seq_data_random_init(data_model);

    /* store the normalization */
    param_file = nn_fopen(path_build(exp_path, param_fname),
			  "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);
    
    for(fold = 0; fold < cv_nfolds; fold++){
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	/* read in the next set of cluster centers */
	if(fold != 0){
	    irbf_free(model);
	    /* the number of centers may have been changed */
	    clus_model = irbf_read_cluster(clus_path, clus_file,
					   &model_ninputs, &ngauss_hidden,
					   features);
	    model = irbf_build(model_ninputs, ngauss_hidden, bias,
			       noutputs, grand_flag);
	    irbf_init (model, hspread, features, minimum_variance,
		       norm_outputs, weight_eta, mean_eta, inv_var_eta,
		       max_magnitude, cost_func, cost_param);
	    irbf_initialize_cluster_centers(model, clus_model);
	    cluster_free(clus_model);
	}
	/* set up data routines for this fold */
	seq_data_start_cv_fold(data_model, fold, TRUE);
	n_files = seq_data_get_nfiles(data_model);
	
	/* verror for training is 0 */
	report_init(err_part_model, db_type, model_ninputs, cost_func,
		    log_outputs, post_sum_to_one);
	report_epoch_header(err_part_model, stdout, n_files);

	for(epoch = 1; epoch <= nepochs; epoch++){
	    seq_data_epoch_init(data_model);
	    for(n_read = 0; n_read < nsequences; n_read++){
		if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
		desired_class = seq_data_get_class(data_model);
		irbf_seq_init(seq_model, model, clip_floor);
		for(n_frames = 0;; n_frames++){
		    inputs = seq_data_get_next_frame(data_model);
		    if(inputs == NULL) break;
		    irbf_seq_batch_classify(seq_model, inputs);
		    irbf_inc_ctrain(model, inputs, desired_class);
		}
		if(n_frames > 0){
		    class = irbf_seq_update_classify(seq_model);
		    report_seq_trial(err_part_model, error_file, seq_num,
				     seq_sums, desired_class, class);
		}
	    }
	    report_epoch(err_part_model, stdout, epoch);
	    report_init(err_part_model, db_type, model_ninputs, cost_func,
			log_outputs, post_sum_to_one);
	}
	/* print out the irbf model */
	fprintf(param_file, "irbf model %d\n", fold);
	irbf_write_model(model, param_file);
	
	/* reinitialize model for testing, if needed */
	/* uniform priors  by frame */
	irbf_test_init(model, 1, priors_list, extra_hspread);
	irbf_seq_init(seq_model, model, clip_floor);
	report_init(err_part_model, db_type, model_ninputs,
		    SQ_ERR, log_outputs, post_sum_to_one);

	/* finish the training fold, start the testing fold */
	seq_data_end_cv_fold(data_model);
	seq_data_start_cv_fold(data_model, fold, FALSE);
	nsequences = seq_data_get_nfiles(data_model);

	for(n_files = 0; n_files < nsequences; n_files++){
	    if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
	    desired_class = seq_data_get_class(data_model);
	    irbf_seq_init(seq_model, model, clip_floor);
	    for(n_frames = 0; ; n_frames++){
		inputs = seq_data_get_next_frame(data_model);
		if(inputs == NULL) break;
		irbf_seq_batch_classify(seq_model, inputs);
	    }
	    if(n_frames > 0){
		class = irbf_seq_update_classify(seq_model);
		report_seq_trial(err_model, error_file, seq_num, seq_sums,
				 desired_class, class);
		report_seq_trial(err_part_model, error_file, seq_num, seq_sums,
				 desired_class, class);
	    }
	}
	/* report on testing */
	report_overall(err_part_model, stdout, error_fname, labels);
	
	/* finish the testing fold */
	seq_data_end_cv_fold(data_model);
    } /* end fold */
    
    /* close the parameter file and the cluster file */
    nn_fclose(clus_file, clus_path, "for read of clusters");
    nn_fclose(param_file, param_fname, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- irbf model saved in \"%s\"\n",
	     path_build(exp_path, param_fname));
    
    /* finish error reporting */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }

    /* free space */
    irbf_free(model);
    irbf_seq_free(seq_model);
    seq_data_exit(data_model);
    norm_free();
}
