/***--------------------------------------------------------------------------
* NAME:     irbf_seq_pack.c
* PURPOSE:  functions for sequence classification using irbf classifier
*
* SEE ALSO: irbf_pack.c
*
* GLOBALS:
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_pack.c,v 11.3 2001/08/23 20:12:05 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_pack.c,v $
* CREATED: 17-Sep-93 lck                MIT LINCOLN LABORATORY
*--------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include <math.h>
#include "irbf_seq_pack.h"
#include "irbf_pack.h"

typedef struct {
    int noutputs;   /* number of elements in patterns */
    float *seq_sums;
    float clip_floor;   /* lowest permitted output for seq_sums */
    int irbf_model;
} IRBF_SEQ_PARAMS;

                /* Internal Globals */
static IRBF_SEQ_PARAMS **irbfpp;/* array of parameters for each model */
static int      max_models = 0; /* current length of irbfpp */

#define SMALLEST 1.e-30

/*****************************************************************************
* NAME: irbf_seq_build(noutputs)
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
*****************************************************************************/
int irbf_seq_build(int noutputs)
{
    IRBF_SEQ_PARAMS *rpp;   /* temporary pointer to irbfpp[] */
    register int    new_model;

    new_model =
      model_get_number((void ***)&irbfpp,
               &max_models, sizeof(IRBF_SEQ_PARAMS));

    rpp = irbfpp[new_model];

    /* set mode flags */
    rpp->noutputs = noutputs;

    /* allocate space for vectors */
    rpp->seq_sums = allocate_vector(noutputs);

    return new_model;
}
/*****************************************************************************
* NAME: irbf_seq_init(model)
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
*****************************************************************************/
void irbf_seq_init(int model, int irbf_model, float clip_floor)
{
    int             i;
    IRBF_SEQ_PARAMS *rpp;   /* temporary pointer to irbfpp[] */

    rpp = irbfpp[model];

    rpp->irbf_model = irbf_model;
    rpp->clip_floor = clip_floor;
    for(i = 0; i < rpp->noutputs; i++) rpp->seq_sums[i] = 0.;
}

/*****************************************************************************
* NAME: irbf_seq_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
*****************************************************************************/
void
irbf_seq_free(int model)
{
    IRBF_SEQ_PARAMS *rpp;   /* temporary pointer to irbfpp[] */

    rpp = irbfpp[model];
    /* free vectors and matrices */
    deallocate_vector(rpp->seq_sums);

    /* free structure itself */
    nn_free((char *) rpp, "for irbf parameter structure");
    irbfpp[model] = NULL;
}

/************************************************************************
* NAME: irbf_seq_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/
#define PEQ(y) (0==strcmp(param_name,y))

void * irbf_seq_param(int model, char *param_name)
{
    void           *param_address;
    IRBF_SEQ_PARAMS *rpp;   /* temporary pointer to irbfpp[] */

    rpp = irbfpp[model];
    if (PEQ("seq_sums"))
      param_address = &rpp->seq_sums;
    else {
    warning("irbf_seq_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}

/* collect outputs of model for series of input frames.
 * summing the log outputs */
void
irbf_seq_batch_classify(int model, float *inputs)
{
    int i;
    IRBF_SEQ_PARAMS *rpp;   /* temporary pointer to irbfpp[] */
    float *outputs;

    rpp = irbfpp[model];
    irbf_eval(rpp->irbf_model, inputs);
    outputs = *(float **)irbf_param(rpp->irbf_model, "outputs");

    /* irbf cannot give log outputs yet */
    for(i = 0; i < rpp->noutputs; i++){
    /* clip the outputs */
    if(outputs[i] < rpp->clip_floor) outputs[i] = rpp->clip_floor;
    if(outputs[i] > 1.) outputs[i] = 1.;
    rpp->seq_sums[i] += log10((double)outputs[i]);
    }
}

int
irbf_seq_update_classify(int model)
{
    IRBF_SEQ_PARAMS *rpp;
    rpp = irbfpp[model];
    return(max_pick(rpp->seq_sums, rpp->noutputs));
}
    
