/***-------------------------------------------------------------------------
 * NAME:	irbf_seq_test.c
 * PURPOSE:	testing driver for Radial Basis Function sequence classifier
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_test.c,v 11.4 2001/08/23 20:12:06 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_test.c,v $
 * CREATED: 17-Sep-93 lck			MIT LINCOLN LABORATORY
 **************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "irbf_seq.h"
#include "irbf_pack.h"
#include "irbf_seq_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];	/* static variables */

void
irbf_seq_test(int create_flag, int db_type, int nsequences, int debg,
	      int verbosity, int verror, char *data_path, char *data_fname,
	      char *exp_path, char *error_fname, char *param_fname,
	      char **labels, float extra_hspread, int cost_func,
	      float cost_param, int log_outputs, float clip_floor)
{
    /* local variables */
    int		model, seq_model, err_model, data_model, rv, normalization = 0;
    int		*features, ninputs, nraw, model_ninputs, noutputs;
    int		nwindow_frames, *window_frames;
    int		n_seq, n_frames, class, n_files;
    char	*tmp_path;
    FILE	*error_file, *param_file;
    int		i, desired_class;
    float	*inputs, *priors_list, *seq_sums;
    int		post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    /* read the parameter file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    if (verbosity > 1)
      printf("\nReading irbf parameters from \"%s\"\n", tmp_path);
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s",
		  prog_name, rv, "on normalization parameter file",
		  tmp_path);
    rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on feature selection parameters in file", tmp_path);
    rv = window_read_paramfile(param_file, &nwindow_frames, &window_frames);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on windowing parameters in file", tmp_path);
    
    /* weight_eta = mean_eta = inv_var_eta = minimum_inv_var = 0.; */
    model = irbf_read_model(param_file, 0., 0., 0., 0.);

    if (model < 0)
      fatal_error("irbf: read error %d on model file %s", model, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of irbf parameters");
    noutputs = *(int *)irbf_param(model, "nclasses");
    model_ninputs = *(int *)irbf_param(model, "ninputs");

    /* build seq model */
    seq_model = irbf_seq_build(noutputs);
    irbf_seq_init(seq_model, model, clip_floor);
    
    seq_sums = *(float **) irbf_seq_param(seq_model, "seq_sums");
    
    if (verror > 0) {		/* open file for classification errors */
	if(create_flag){
	    error_file = nn_fopen(path_build(exp_path, error_fname),
				  "w", "for classification errors");
	} else {
	    error_file = nn_fopen(path_build(exp_path, error_fname),
				  "a", "for classification errors");
	}
    } else error_file = NULL;
    /* do remaining initializations and allocations */
    init_cost(cost_func, cost_param);
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, db_type, model_ninputs, cost_func, log_outputs, post_sum_to_one);

    /* read data */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    nsequences = seq_data_get_nfiles(data_model);
    
    /*************** prepare model for testing **************/
    /* irbf must adjust for uniform priors for irbf_seq */
    priors_list = allocate_vector(noutputs);
    for(i = 0; i < noutputs; i++) priors_list[i] = 1.;
    irbf_test_init(model, 1, priors_list, extra_hspread);
    
    /* read and classify input from data file */
    for(n_files = 0; n_files < nsequences; n_files++){
	if((n_seq = seq_data_get_next_file(data_model)) < 0) break;
	desired_class = seq_data_get_class(data_model);
	irbf_seq_init(seq_model, model, clip_floor);
	for(n_frames = 0; ; n_frames++){
	    inputs = seq_data_get_next_frame(data_model);
	    if(inputs == NULL) break;

	    /* classify frame */
	    irbf_seq_batch_classify(seq_model, inputs);
	}
	if(n_frames > 0){
	    /* find the class and report */
	    class = irbf_seq_update_classify(seq_model);
	    report_seq_trial(err_model, error_file, n_seq,
			     seq_sums, desired_class, class);
	}
    }
    /* finish the error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }

    /* free space and close files */
    irbf_free(model);
    irbf_seq_free(seq_model);
    seq_data_exit(data_model);
    norm_free();
}
