/***--------------------------------------------------------------------------
 * NAME:	irbf_seq_train.c
 * PURPOSE:	training driver for Radial Basis Function sequence classifier
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_train.c,v 11.4 2001/08/23 20:12:06 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_seq_train.c,v $
 * CREATED: 17-Sep-93 lck				MIT LINCOLN LABORATORY
 ****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "irbf_seq.h"
#include "irbf_seq_pack.h"
#include "irbf_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];	/* static variables */

void
irbf_seq_train(int create_flag, int db_type, int noutputs, int nsequences,
	       int debg, int verbosity, int verror, int random_flag,
	       char *data_path, char *data_fname, char *exp_path,
	       char *error_fname, char *param_fname, char **labels,
	       char *comment, int nepochs, int bias, int grand_flag,
	       float hspread, float max_magnitude, char *cluster_param_fname,
	       float minimum_variance, float weight_eta, float mean_eta,
	       float inv_var_eta, int cost_func, float cost_param,
	       int norm_outputs, int log_outputs, float clip_floor)
{
    /* local variables */
    int		model, data_model, err_model, err_part_model, seq_model;
    int		rv, normalization = 0;
    int		*features, nraw, ninputs, model_ninputs, class;
    int		nwindow_frames, *window_frames;
    int		epoch, n_read, n_frames, ncenters;
    int		desired_class, seq_num;
    float	*inputs, *seq_sums;
    char	*tmp_path, *clus_path;
    FILE	*param_file, *clus_file, *error_file;
    int		post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    if(create_flag){
	/***************** read the cluster file *****************/
	clus_path = path_build(exp_path, cluster_param_fname);
    
	clus_file = nn_fopen(clus_path, "r", "for read of cluster parameters");
	if (verbosity > 1) {
	    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	    fflush(stdout);
	}
	/* get normalization from cluster file */
	model = norm_read_paramfile(clus_file, &normalization);
	if (model < 0)
	  fatal_error("irbf: read error %d %s %s",
		      model, "on normalization parameters in file", clus_path);
	rv = feature_read_paramfile(clus_file, &features, &nraw, &ninputs);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on feature selection parameters in file", clus_path);
	rv = window_read_paramfile(clus_file, &nwindow_frames, &window_frames);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on windowing parameters in file", clus_path);
	
	/* read clusters */
	/* features was allocated in feature_read_paramfile,
	   so it is ok here */
	rv = irbf_read_cluster(clus_path, clus_file, &model_ninputs,
			       &ncenters, features);
	if(rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on cluster parameter file", clus_path);
	
	nn_fclose(clus_file, clus_path, "for read of cluster parameters");
    
	/********** build the irbf model ***************/
	if (verbosity > 1){
	    printf ("\nBuilding irbf model\n");
	    fflush(stdout);
	}
	model = irbf_build (model_ninputs, ncenters, bias, noutputs,
			    grand_flag);
	irbf_init (model, hspread, features, minimum_variance,
		   norm_outputs, weight_eta, mean_eta, inv_var_eta,
		   max_magnitude, cost_func, cost_param);
	irbf_initialize_cluster_centers(model, rv);
	cluster_free(rv);
    } else { /* continue training */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	if (verbosity > 1)
	  printf("\nReading irbf parameters from \"%s\"\n", tmp_path);
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s",
		      prog_name, rv, "on normalization parameter file",
		      tmp_path);
	rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on feature selection parameters in file", tmp_path);
	rv = window_read_paramfile(param_file, &nwindow_frames,
				   &window_frames);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on windowing parameters in file", tmp_path);
	
	model = irbf_read_model(param_file, weight_eta, mean_eta, inv_var_eta,
				minimum_variance);
	if (model < 0)
	  fatal_error("%s: read error %d on model file %s", prog_name,
		      model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of irbf parameters");
	model_ninputs = *(int *)irbf_param(model, "ninputs");
    }
    /* set up irbf_pack for testing */
    /* no priors adjustment, extra_hspread = 1. for training */
    irbf_test_init(model, 0, NULL, 1.);

    /* build seq_model */
    seq_model = irbf_seq_build(noutputs);
    irbf_seq_init(seq_model, model, clip_floor);
    seq_sums = *(float **)irbf_seq_param(seq_model, "seq_sums");

    /* set up error reporting */
    if(verror > 0) { /* open file for classification errors */
	if(create_flag) 
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else 
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    } else error_file = NULL;
    err_model = report_build (verbosity, verror, noutputs);
    report_init(err_model, db_type, model_ninputs, cost_func, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);
    report_init(err_part_model, db_type, model_ninputs, cost_func, log_outputs, post_sum_to_one);

    /* read data */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    if(random_flag) seq_data_random_init(data_model);
    
    nsequences = seq_data_get_nfiles(data_model);
    report_epoch_header(err_model, stdout, nsequences);

    for(epoch = 0; epoch < nepochs; epoch++){
	seq_data_epoch_init(data_model);

	for(n_read = 0; n_read < nsequences; n_read++){
	    if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
	    desired_class = seq_data_get_class(data_model);
	    irbf_seq_init(seq_model, model, clip_floor);
	    for(n_frames = 0; ; n_frames++){
		inputs = seq_data_get_next_frame(data_model);
		if(inputs == NULL) break;
		irbf_seq_batch_classify(seq_model, inputs);
		irbf_inc_ctrain(model, inputs, desired_class);
	    }
	    if(n_frames > 0){
		class = irbf_seq_update_classify(seq_model);
		report_seq_trial(err_model, error_file, seq_num, seq_sums,
				 desired_class, class);
		report_seq_trial(err_part_model, error_file, seq_num,
				 seq_sums, desired_class, class);
	    }
	}
	report_epoch(err_part_model, stdout, epoch);
	report_init(err_part_model, db_type, model_ninputs, cost_func,
		    log_outputs, post_sum_to_one);
    }
    /******************** store model in param file ********************/
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);
    fprintf(param_file, "irbf model\n");
    irbf_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0){
	printf("\nFinished -- irbf model saved in \"%s\"\n", tmp_path);
	fflush(stdout);
    }
    /* finish error reporting and close error file */
    /* only show the last epoch not the overall errors */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0){
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space */
    irbf_free(model);
    norm_free();
    seq_data_exit(data_model);
}
