/***--------------------------------------------------------------------------
 * NAME:		irbf.c
 * PURPOSE:	driver for Radial Basis Function classifier
 *
 * SEE ALSO: 	irbf_pack.c
 *
 * FUNCTIONS:
 *	driver for Radial Basis Function classifier
 *		handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_test.c,v 11.4 2001/08/23 20:12:06 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_test.c,v $
 * CREATED: 27-NOV-92				MIT LINCOLN LABORATORY
 *------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "irbf.h"
#include "irbf_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];	/* static variables */

void
irbf_test(int create, int map_flag, int nraw, int ninputs, int noutputs,
	  int npatterns, int debg, int verbosity, int verror, char *data_path,
	  char *data_fname, char *exp_path, char *error_fname,
	  char *param_fname, char **labels, int adjust_priors,
	  float *priors_list, int log_outputs, float extra_hspread,
	  int cost_func, float cost_param)
{
    /* local variables */
    int		model, err_model, rv, normalization = 0, data_model;
    int		desired_class, class;
    float	*inputs, *outputs;
    char	*tmp_path;
    FILE	*error_file = NULL, *param_file;
    int		*features;
    int		post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    if (verbosity > 1)
      printf("\nReading irbf parameters from \"%s\"\n", tmp_path);
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s",
		  prog_name, rv, "on normalization parameter file",
		  tmp_path);

    /* weight_eta = mean_eta = inv_var_eta = minimum_inv_var = 0; */
    model = irbf_read_model(param_file, 0., 0., 0., 0.);
    if (model < 0)
      fatal_error("irbf: read error %d on model file %s", model, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of irbf parameters");
    
    /* use features from the model */
    features = *(int **) irbf_param(model, "features");
    ninputs = *(int *) irbf_param(model, "ninputs");
    outputs = *(float **) irbf_param(model, "outputs");
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {		/* open file for classification errors */
	if(create)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, cost_func,
		 log_outputs, post_sum_to_one);
    init_cost(cost_func, cost_param);

    /* open data file */
    data_model =
      class_data_single_init(data_path, data_fname, npatterns,
			     nraw, ninputs, noutputs, features,
			     normalization, verbosity);

    irbf_test_init(model, adjust_priors, priors_list, extra_hspread);
    
    /* read and classify input from data file */
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
	/* classify pattern and report errors */
	class = irbf_classify(model, inputs);
	report_class_trial(err_model, error_file, inputs, outputs,
			   desired_class, class,
			   class_data_get_input_num(data_model));
    }
    
    /* free space and close files */
    report_overall(err_model, stdout, error_fname, labels);
    report_test_result(err_model, stdout, data_fname);
    report_free(err_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    class_data_exit(data_model);
    irbf_free(model);
    norm_free();
}
