/***--------------------------------------------------------------------------
 * NAME:		irbf.c
 * PURPOSE:	driver for Radial Basis Function classifier
 *
 * SEE ALSO: 	irbf_pack.c
 *
 * FUNCTIONS:
 *	driver for Radial Basis Function classifier
 *		handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_train.c,v 11.4 2001/08/23 20:12:06 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbf_train.c,v $
 * CREATED: 27-Nov-92 lck			MIT LINCOLN LABORATORY
 *-------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "irbf.h"
#include "irbf_pack.h"

#define MAXSTR (256)

static char     prog_name[MAXSTR];	/* static variables */

void
irbf_train(int create_flag, int nraw, int ninputs, int *pfeatures,
	   int noutputs, int npatterns, int debg, int verbosity, int verror, 
	   int random_flag, char *data_path, char *data_fname, char *exp_path,
	   char *error_fname, char *param_fname, char **labels, char *comment,
	   int adjust_priors, int priors_npatterns, float *priors_list,
	   int log_outputs, int nepochs, int bias, int grand_flag,
	   float hspread, float max_magnitude, char *cluster_param_fname,
	   float minimum_variance, float weight_eta, float mean_eta,
	   float inv_var_eta, int cost_func, float cost_param,
	   int norm_outputs)
{
    /* local variables */
    int		model, err_model, err_part_model, data_model;
    int		rv, normalization = 0, map_flag = 0, epoch;
    int		desired_class, class;
    int		n_read, n_trials, ncenters, trial_num;
    float	*inputs, *outputs;
    char	*tmp_path, clus_path[MAXSTR];
    char	*open_type;
    FILE	*param_file, *clus_file, *error_file = NULL;
    int		*features;
    int		post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    if (create_flag){
	/* read the cluster file */
	tmp_path = path_build(exp_path, cluster_param_fname);
	strncpy(clus_path, tmp_path, MAXSTR - 1);
	
	clus_file = nn_fopen(clus_path, "r",
			     "for read of cluster parameters");
	if (verbosity > 1) {
	    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	    fflush(stdout);
	}
	model = norm_read_paramfile(clus_file, &normalization);
	
	if (model < 0)
	  fatal_error("irbf: read error %d %s %s",
		      model, "on normalization parameter file", clus_path);
	
	rv = irbf_read_cluster(clus_path, clus_file, &ninputs,
			       &ncenters, pfeatures);
	if(rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on cluster parameter file", clus_path);
	
	nn_fclose(clus_file, clus_path, "for read of cluster parameters");
	
	model = irbf_build (ninputs, ncenters, bias, noutputs, grand_flag);
	irbf_init (model, hspread, pfeatures, minimum_variance, norm_outputs,
		   weight_eta, mean_eta, inv_var_eta, max_magnitude,
		   cost_func, cost_param);
	irbf_initialize_cluster_centers(model, rv);
	cluster_free(rv);
    } else {	/* retrain */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	if (verbosity > 1)
	  printf("\nReading irbf parameters from \"%s\"\n", tmp_path);
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s",
		      prog_name, rv, "on normalization parameter file",
		      tmp_path);
	
	model = irbf_read_model(param_file, weight_eta, mean_eta, inv_var_eta,
				minimum_variance);
	if (model < 0)
	  fatal_error("irbf: read error %d on model file %s", model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of irbf parameters");
	
	verify_model_params (model, ninputs,
			     *(int *) irbf_param(model, "ninputs"), pfeatures,
			     *(int **) irbf_param(model, "features"));
    }

    /* use features from the model */
    features = *(int **) irbf_param(model, "features");
    ninputs = *(int *) irbf_param(model, "ninputs");
    outputs = *(float **) irbf_param(model, "outputs");
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);

    if(verror > 0) { /* open file for classification errors */
	if(create_flag) open_type = "w";
	else open_type = "a";
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      open_type, "for classification errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init(err_model, map_flag, ninputs, cost_func, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
      class_data_priors_init(data_model, priors_list, priors_npatterns);
    
    /* get the number of patterns per epoch */
    n_trials = class_data_get_npatterns(data_model);
    report_epoch_header(err_model, stdout, n_trials);

    for(epoch = 1; epoch <= nepochs; epoch++){
	class_data_epoch_init(data_model);
	report_init(err_part_model, map_flag, ninputs, cost_func,
		    log_outputs, post_sum_to_one);

	for(n_read = 0; n_read < n_trials; n_read++){
	    /* get the next pattern */
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    trial_num = class_data_get_input_num(data_model);
	    class = irbf_inc_ctrain(model, inputs, desired_class);
	    report_class_trial(err_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	    report_class_trial(err_part_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	}
	report_epoch(err_part_model, stdout, epoch);
    }
    report_last_epoch(err_part_model, stdout, epoch-1);
    class_data_exit(data_model);
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "irbf model\n");
    irbf_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- irbf model saved in \"%s\"\n", tmp_path);

    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error_file");
    }
    
    norm_free();
    irbf_free(model);
}
