/***--------------------------------------------------------------------------
 * NAME:	irbfm_train.c
 * PURPOSE:	training driver for Radial Basis Function mapper
 *
 * SEE ALSO: 	../irbf/irbf_pack.c
 *
 * FUNCTIONS:	irbf_train()
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbfm_train.c,v 11.4 2001/08/23 20:12:06 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/irbf/irbfm_train.c,v $
 * CREATED: 15-Oct-92 lck				MIT LINCOLN LABORATORY
 *--------------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "irbfm.h"
#include "irbf_pack.h"

#define MAXSTR (256)

#define NO_NORM_OUTPUTS (0)

static char     prog_name[MAXSTR];	/* static variables */

void
irbfm_train(int create_flag, int map_flag, int nraw, int ninputs,
	    int *pfeatures, int noutputs, int npatterns, int debg,
	    int verbosity, int verror, int random_flag, char *data_path,
	    char *data_fname, char *exp_path, char *error_fname,
	    char *param_fname, char **labels, char *comment, int nepochs,
	    int bias, int grand_flag, float hspread, float max_magnitude,
	    char *cluster_param_fname, float minimum_variance,
	    float weight_eta, float mean_eta, float var_eta)
{
    /* local variables */
    int		model, err_model, err_part_model;
    int		rv, normalization = 0, epoch;
    int		n_read, i, n_trials, t_index, *trial_index, ncenters;
    float	*inputs, *outputs, *desired_outputs;
    float	**input_matrix, **output_matrix;
    char	data_fullname[MAXSTR], *tmp_path, *open_type;
    char	clus_path[MAXSTR];
    FILE	*data_file, *error_file = NULL, *param_file, *clus_file;
    int		*features;
    int		post_sum_to_one = 0;
    
    strcpy(prog_name, get_program_name());
    
    /* put together the full path to the data file*/
    strcpy(data_fullname, path_build(data_path, data_fname));
    
    if (create_flag){
	/* read the cluster file */
	tmp_path = path_build(exp_path, cluster_param_fname);
	strncpy(clus_path, tmp_path, MAXSTR - 1);
	
	clus_file = nn_fopen(clus_path, "r",
			     "for read of cluster parameters");
	if (verbosity > 1) {
	    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	    fflush(stdout);
	}
	model = norm_read_paramfile(clus_file, &normalization);
	
	if (model < 0)
	  fatal_error("irbf: read error %d %s %s",
		      model, "on normalization parameter file", clus_path);
	
	rv = irbf_read_cluster(clus_path, clus_file, &ninputs,
			       &ncenters, pfeatures);
	if(rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on cluster parameter file", clus_path);
	
	nn_fclose(clus_file, clus_path, "for read of cluster parameters");
	
	if (verbosity > 1)
	  printf ("\nBuilding irbf model\n");
	model = irbf_build (ninputs, ncenters, bias, noutputs, grand_flag);
	irbf_init (model, hspread, pfeatures, minimum_variance,
		   NO_NORM_OUTPUTS, weight_eta, mean_eta, var_eta,
		   max_magnitude, SQ_ERR, 1.);
	irbf_initialize_cluster_centers(model, rv);
	cluster_free(rv);
    } else {	/* retrain */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	if (verbosity > 1)
	  printf("\nReading irbf parameters from \"%s\"\n", tmp_path);
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s",
		      prog_name, rv, "on normalization parameter file",
		      tmp_path);
	
	model = irbf_read_model(param_file, weight_eta, mean_eta, var_eta,
				minimum_variance);
	if (model < 0)
	  fatal_error("irbf: read error %d on model file %s", model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of irbf parameters");
	verify_model_params (model, ninputs,
			     *(int *) irbf_param(model, "ninputs"), pfeatures,
			     *(int **) irbf_param(model, "features"));
    }
    /* use features from the model */
    features = *(int **) irbf_param(model, "features");
    ninputs = *(int *) irbf_param(model, "ninputs");
    outputs = *(float **) irbf_param(model, "outputs");
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);

    if(verror > 0) { /* open file for errors */
	if(create_flag) open_type = "w";
	else open_type = "a";
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      open_type, "for mapping errors");
    }
    /* do remaining intializations and allocations */
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, map_flag, ninputs, SQ_ERR, 0, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);
    report_init(err_part_model, map_flag, ninputs, SQ_ERR, 0, post_sum_to_one);

    /* open data file */
    data_file = nn_fopen(data_fullname, "r", "for training or test data");
    if (verbosity > 1)
      printf("\nReading %s\n", data_fullname);
    
    input_matrix = allocate_matrix(npatterns, ninputs);
    output_matrix = allocate_matrix(npatterns, noutputs);
    n_trials = load_map_data(data_file, npatterns, nraw, noutputs, ninputs,
			     output_matrix, input_matrix,
			     normalization, data_fname, features);
    trial_index = allocate_ivector(n_trials);
    for(i = 0; i < n_trials; i++) trial_index[i] = i;

    report_epoch_header(err_model, stdout, n_trials);

    for(epoch = 1; epoch <= nepochs; epoch++) {
	if(random_flag)
	  randomize_index_list(trial_index, n_trials);
	for (n_read = 0; n_read < n_trials; n_read++) {
	    /* read input vector */
	    t_index = trial_index[n_read];
	    inputs = &input_matrix[t_index][0];
	    desired_outputs = &output_matrix[t_index][0];
	    irbf_inc_train(model, inputs, desired_outputs);
	    report_map_trial(err_model, error_file, inputs, outputs,
			     desired_outputs, t_index);
	    report_map_trial(err_part_model, error_file, inputs, outputs,
			     desired_outputs, t_index);
	}
	report_epoch(err_part_model, stdout, epoch);
	report_init(err_part_model, map_flag, ninputs, SQ_ERR, 0, post_sum_to_one);
    }
    deallocate_ivector(trial_index);
    deallocate_matrix(input_matrix);
    deallocate_matrix(output_matrix);
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "irbf model\n");
    irbf_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- irbf model saved in \"%s\"\n", tmp_path);

    /* free space and close files */
    nn_fclose(data_file, data_fullname, "for training or test data");

    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);

    if(verror > 0){
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    norm_free();
    irbf_free(model);
}
