/***--------------------------------------------------------------------------
 * NAME:		kmeans.c
 * PURPOSE:	driver for kmeans clustering algorithm
 *
 * SEE ALSO: 	kmeans_pack.c
 * GLOBALS:	int debug;
 *
 * FUNCTIONS:
 *	kmeans	driver for kmeans clustering algorithm
 *		handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeans_cv.c,v 11.4 2001/08/23 20:12:12 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeans_cv.c,v $
 * CREATED: 20-Mar-91 dan				MIT LINCOLN LABORATORY
 ****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "kmeans.h"
#include "kmeans_pack.h"

#ifndef TRUE
#define TRUE (1)
#endif
#ifndef FALSE
#define FALSE (0)
#endif
#define MAXSTR (256)
/* globals */
int             debug;		/* enables debug messages if debug > 0*/

void
kmeans_cross_valid(int nraw, int ninputs, int *features, int noutputs,
		   int normalize, int npatterns, int debg, int verbosity,
		   int random_flag, int cv_nfolds, int cv_nsplits,
		   int use_cv_splits, int *cv_splits, char **cv_train_mask,
		   char **cv_test_mask, char *data_path, char *data_fname,
		   char *norm_fname, char *exp_path, char *param_fname,
		   char *comment, int adjust_priors, int priors_npatterns,
		   float *priors_list, int cv_random, int max_iterations,
		   float split_percentage, float stop_percentage,
		   int cluster_by_class, int add_random_offset,
		   int reduce_step, int *desired_ncenter)
{
    int		model, err_model, normalization=0, data_model;
    int		*model_list;
    int		current_class, num_in_class, class;
    int		n_read, i, n_trials;
    int		fold;
    float	*inputs;
    char	*tmp_path;
    FILE	*param_file;
    int		iteration, actual_ncenters, *ncenters;
    int 	report_stop_flag, report_nvectors;
    float	cost;
    float	*pnorm_variance;	/* normalizing variance */
    int		verror = 0;
    int		post_sum_to_one = 0;
#ifdef DEBUG
    float *outputs;
    int desired_class;
    FILE *error_file;
    char *error_fname;
    error_fname = "kmeans.err.cv";
    error_file = nn_fopen(path_build(exp_path, error_fname),
			  "w", "for classification errors");
    outputs = allocate_vector(noutputs);
    verror = 2;
#endif
    
    debug = debg;		/* set globals */
    
    model_list = allocate_ivector(noutputs);
    
    /* allocate and initialize model structure */
    pnorm_variance = allocate_vector(ninputs);
    for (i=0; i<ninputs; i++) pnorm_variance[i] = 1.;
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    current_class = 0;
    do{
	model_list[current_class] =
	  kmeans_build(ninputs, desired_ncenter[current_class]);
	model = model_list[current_class];
	kmeans_init(model, features,
		    pnorm_variance, stop_percentage, 
		    split_percentage, max_iterations, reduce_step,
		    add_random_offset);
	current_class++;
    } while (cluster_by_class && current_class < noutputs);
    
    model = model_list[0];
    
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, 0, ninputs, CLUS_SQ_ERR, 0, post_sum_to_one);

    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		       cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
	class_data_priors_init(data_model, priors_list, priors_npatterns);
    
    /* store normalization once */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of knn model file");
    /* write comment */
    fprintf(param_file, "%s", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    
    for(fold = 0; fold < cv_nfolds; fold++){
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	class_data_start_cv_fold(data_model, fold, TRUE);
	n_trials = class_data_get_npatterns(data_model);
	
	current_class = 0;
	do{
	    model = model_list[current_class];
	    kmeans_init(model, features,
			pnorm_variance, stop_percentage, 
			split_percentage, max_iterations, reduce_step,
			add_random_offset);
	    
	    if(cluster_by_class){
		if (verbosity > 1)
		  printf("training centers for class %d", current_class);
		num_in_class =
		  class_data_get_npatterns_class(data_model, current_class);
		report_epoch_header(err_model, stdout, num_in_class);
	    } else report_epoch_header(err_model, stdout, n_trials);
	    
	    report_stop_flag = FALSE;
	    while (!report_stop_flag) {/* for as many epochs as necessary */
		num_in_class = 0;
		class_data_epoch_init(data_model);
		for(n_read = 0; n_read < n_trials; n_read++){
		    if((inputs =
			class_data_get_next(data_model, &class))
		       == NULL) break;
		    
		    if(!cluster_by_class ||
		       (cluster_by_class && class == current_class)){
			kmeans_batch_train (model, inputs);
			num_in_class++;
		    }
		}
		if(num_in_class != 0){
		    report_stop_flag =
		      kmeans_batch_update(model, &iteration,
					  &actual_ncenters, &cost,
					  &report_nvectors);
		} else {
		    warning("%s %d.\n%s", "There was no data in class",
			    current_class, "Eliminating those centers");
		    actual_ncenters = iteration = 0;
		    ncenters = (int *)kmeans_param(model, "ncenters");
		    *ncenters = 0;
		    report_stop_flag = 1;
		    cost = 0.;
		}
		
		report_cluster_epoch(err_model, stdout, iteration,
				     actual_ncenters, cost);
	    }
	    current_class++;
	} while(current_class < noutputs && cluster_by_class);
	
	/* write out model or models */
	tmp_path = path_build(exp_path, param_fname);
	/* store model in param file */
	fprintf(param_file, "kmeans model %d\n", fold);
	if(cluster_by_class){
	    kmeans_write_model_list(model_list, noutputs, param_file);
	} else {
	    kmeans_write_model(model_list[0], noutputs, param_file);
	}
	/* as there is no testing, we don't have to do any */
	class_data_end_cv_fold(data_model);
#ifdef DEBUG
	/* check the fold assignments for matching with other programs */
	class_data_start_cv_fold(data_model, fold, FALSE);
	n_trials = class_data_get_npatterns(data_model);

	for (n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    report_cluster_trial(err_model, error_file, inputs, outputs,
				 desired_class, desired_class,
				 class_data_get_input_num(data_model));
	}
	class_data_end_cv_fold(data_model);
#endif
    } /* end fold */
    
#ifdef DEBUG
    deallocate_vector(outputs);
    nn_fclose(error_file, error_fname, "for write of errors");
#endif

    /* close parameter file */
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    /* free model structures */
    deallocate_vector(pnorm_variance);
    current_class = 0;
    if(cluster_by_class){
	for(current_class = 0; current_class < noutputs; current_class++)
	  kmeans_free(model_list[current_class]);
    } else {
	kmeans_free(model_list[0]);
    }
    deallocate_ivector(model_list);
    report_free(err_model);
    norm_free();
    class_data_exit(data_model);
}
