/***--------------------------------------------------------------------------
* NAME:         kmeans_pack.c
* PURPOSE:      Functions specific to kmeans.c
*
* SEE ALSO: kmeans.c kmeans_main.c
*
* USAGE:
*   model = kmeans_build(...)
*   kmeans_init(model,...)
*       report_stop_flag=0;
*   while (report_stop_flag<=0)
*       {
*       for(i=0; i<all_patterns; i++)
*           kmeans_batch_train(model, input[i]);
*           report_stop_flag = kmeans_batch_update(model,
*           &iterations, &actual_ncenters, &ave_squared_error);
*       }
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   kmeans_build        Allocate storage for model parameters.
*   kmeans_init     Initialize model parameters.
*   kmeans_free     Deallocates model storage
*   kmeans_read_model   Initializes model from file
*   kmeans_write_model  Writes model to output file
*   kmeans_param        Returns pointers to elements in model
*   kmeans_batch_train  find nearest cluster center to one input
*               pattern and update that clusters counts
*   kmeans_batch_update calculate new cluster centers from counts,
*               check for convergence, split centers to
*               double the number of centers after convergence
*               with a given number of centers or terminate
*               by setting the output to the number of centers
*               used (a power of 2)
*   kmeans_cluster      returns the cluster an input pattern falls in
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeans_pack.c,v 11.3 2001/08/23 20:12:14 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeans_pack.c,v $
* CREATED: 16-Dec-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#define kmeans_pack_h_statics
#define kmeans_pack_h_internal_types
#include "kmeans_pack.h"

                /* constant definitions */
#define ABS(a) ((a) >= 0 ? (a) : -(a))
#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))

static KMEANS_PARAMS **genpp;   /* array of parameters for each model */
static int      max_models = 0; /* current length of genpp */

/*****************************************************************************
* NAME: kmeans_build(ninputs, ncenters)
*   ninputs - number of dimensions in training vectors
*   ncenters - number of cluster centers
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
******************************************************************************/
int
kmeans_build(int ninputs, int ncenters)
{
    register int new_model;
    KMEANS_PARAMS *gpp;
    int i;

    new_model = model_get_number((void ***)&genpp,
                 &max_models, sizeof(KMEANS_PARAMS));
    gpp = genpp[new_model];

    for (i = 1; i < ncenters;)  /* determine maximum number of centers */
    i *= 2;         /* (must be power of 2 ) */
    gpp->max_centers = i;

                    /* set fields which are array sizes */
    gpp->ninputs = ninputs;
    gpp->ncenters = ncenters;
                    /* allocate space */
    gpp->features = allocate_ivector (ninputs+1);
    gpp->outputs = allocate_vector (gpp->max_centers);

    gpp->pcenter = allocate_matrix (gpp->max_centers, ninputs);
    gpp->pvariance = allocate_matrix (gpp->max_centers, ninputs);
    gpp->current_center = allocate_matrix (gpp->max_centers, ninputs);
    gpp->current_variance = allocate_matrix (gpp->max_centers, ninputs);
    gpp->current_sum = allocate_dmatrix (gpp->max_centers, ninputs);
    gpp->current_sum_squared = allocate_dmatrix (gpp->max_centers, ninputs);
    gpp->current_count = allocate_ivector (gpp->max_centers);
    gpp->weight = allocate_vector (ninputs);
    gpp->minimum_variance = 1e-6;

    return new_model;
}

/*****************************************************************************
* NAME: kmeans_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
******************************************************************************/
void
kmeans_init (int model, int *nn_features,
         float *pnorm_variance,/* variances to normalize input vector */
         float stop_percentage,/* stop when rms change less than this */
         float split_percentage,/* +/- sdev this percentage on split */
         int max_iteration,/* max iterations after any split */
         int reduce_step, int add_random_offset)
{
    int i, c;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];
                    /* copy feature data into feature list */
    for (i = 0; i <= nn_features[0]; i++)
    gpp->features[i] = nn_features[i];

                    /* set global variables */
    gpp->stop_percentage = stop_percentage;
    gpp->split_percentage = split_percentage;
    gpp->max_iteration = max_iteration;
    gpp->reduce_step = reduce_step;
    gpp->add_random_offset = add_random_offset;

                    /* special startup initialization */
    gpp->last_ave_squared_error = MAXFLOAT; /* ave squared error on last
                     * iteration */
    gpp->ave_squared_error = MAXFLOAT;
    gpp->splits = 0;            /* number of splits so far */
    gpp->total_iteration = 0;   /* total iterations so far */
    gpp->current_iteration = 0; /* number of iterations for this split */
    if(gpp->ncenters != 0)
      gpp->current_ncenters = 1;/* current number of cluster centers */
    else gpp->current_ncenters = 0;
    gpp->stop_flag = 0;

            /* common initialization before all iterations*/
    kmeans_local_init(model);

    /* initializations needed by cross valid to zero variances */
    for(c = 0; c < gpp->ncenters; c++){
    for(i = 0; i < gpp->ninputs; i++){
        gpp->current_variance[c][i] = 0.;
        gpp->current_center[c][i] = 0.;
    }
    }

            /* initialize weighting function */
    for (i = 0; i < gpp->ninputs; i++) {
    if (pnorm_variance[i] == 0.0)
        fatal_error("%s %d is %f\n%s",
          "kmeans_init: Variance of input vector", i, pnorm_variance[i],
          "This would cause a division by zero during kmeans training");
    gpp->weight[i] = 1.0 / pnorm_variance[i];
    }
}

/*****************************************************************************
* NAME: kmeans_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
******************************************************************************/
void kmeans_free(int model)
{
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];
                        /* free vectors and matrices */
    deallocate_ivector (gpp->features);
    deallocate_vector (gpp->outputs);
    deallocate_matrix (gpp->pcenter);
    deallocate_matrix (gpp->pvariance);
    deallocate_matrix (gpp->current_center);
    deallocate_matrix (gpp->current_variance);
    deallocate_dmatrix (gpp->current_sum);
    deallocate_dmatrix (gpp->current_sum_squared);
    deallocate_ivector (gpp->current_count);
    deallocate_vector (gpp->weight);

                        /* free structure itself */
    nn_free((char *) gpp, "for kmeans parameter structure");
    genpp[model] = NULL;
}

/*****************************************************************************
* NAME: kmeans_read_model()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*   (currently used only by kmeans_plot_bound)
*
* EXAMPLE:
*
******************************************************************************/
int
kmeans_read_model(FILE *fp)
{
    int model, clus_model, noutputs, i, j;
    float **pcenter, **pvariance;
    KMEANS_PARAMS *kpp;
    int ninputs, ncenters, *features;

    clus_model = cluster_read_model(fp);
    if(clus_model < 0) return(clus_model);

    cluster_merge_centers(clus_model);
    ninputs = *(int *)cluster_param(clus_model, "ninputs", 0);
    noutputs = *(int *)cluster_param(clus_model, "noutputs", 0);
    ncenters = *(int *)cluster_param(clus_model, "ncenters", 0);

    model = kmeans_build(ninputs, ncenters);

    features = *(int **)cluster_param(clus_model, "features", 0);

    kpp = genpp[model];

    pcenter = *(float ***)cluster_param(clus_model, "pcenter", 0);
    pvariance = *(float ***)cluster_param(clus_model, "pvariance", 0);
    
    for(j = 0; j < ncenters; j++){
    for(i = 0; i < ninputs; i++){
        kpp->pcenter[j][i] = pcenter[j][i];
        kpp->pvariance[j][i] = pvariance[j][i];
    }
    }

    /* copy vectors and matrices into kpp */
    for(i = 0; i <= features[0]; i++){
    kpp->features[i] = features[i];
    }

    cluster_free(clus_model);
    return(model);
}

/*****************************************************************************
* NAME: kmeans_write_model()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int kmeans_write_model(int model, int noutputs, FILE *fp)
{
    int rv, cluster_by_class = 0, *ncenter_list = NULL;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];

    rv = cluster_write_paramfile(fp, gpp->ninputs, gpp->ncenters,
                 noutputs, &(gpp->pcenter), &(gpp->pvariance),
                 gpp->features, cluster_by_class,
                 ncenter_list);
    return(rv);
}

/*****************************************************************************
* NAME: kmeans_write_model_list()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int kmeans_write_model_list(int *model_list, int noutputs, FILE *fp)
{
    int rv, cluster_by_class = 1, i, *ncenter_list, ncenters;
    float ***center_list, ***variance_list;
    KMEANS_PARAMS *gpp;

    ncenter_list = allocate_ivector(noutputs);
    center_list = (float ***)allocate_pvector(noutputs);
    variance_list = (float ***)allocate_pvector(noutputs);

    for(ncenters = 0, i = 0; i < noutputs; i++){
    gpp = genpp[model_list[i]];
    ncenter_list[i] = gpp->ncenters;
    ncenters += gpp->ncenters;
    center_list[i] = gpp->pcenter;
    variance_list[i] = gpp->pvariance;
    }

    rv = cluster_write_paramfile(fp, gpp->ninputs, gpp->ncenters,
                 noutputs, center_list, variance_list,
                 gpp->features, cluster_by_class,
                 ncenter_list);
    deallocate_ivector(ncenter_list);
    deallocate_pvector((void **)center_list);
    deallocate_pvector((void **)variance_list);

    return(rv);
}

/************************************************************************
* NAME: kmeans_param(model, param_name)
* DESCRIPTION:
*       Takes parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/
void * kmeans_param(int model, char *param_name)
{
    void *param_address;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];
    if (PEQ("features"))
    param_address = &gpp->features;
    else if (PEQ("ninputs"))
    param_address = &gpp->ninputs;
    else if (PEQ("ncenters"))
    param_address = &gpp->ncenters;
    else if (PEQ("current_ncenters"))
      param_address = &gpp->current_ncenters;
    else if (PEQ("outputs"))
    param_address = &gpp->outputs;
    else if (PEQ("pcenter"))
    param_address = &gpp->pcenter;
    else if (PEQ("pvariance"))
    param_address = &gpp->pvariance;
    else {
    warning("kmeans_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}


/*****************************************************************************
* NAME:     kmeans_batch_train()
* DESCRIPTION:
*           provide another vector for kmeans training
* EXAMPLE:
*
******************************************************************************/
void kmeans_batch_train(int model, float *input/* input vector */)
{
    float min_dist;
    int min_i, j;
    float temp;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];
                        /* find centroid of all data  */
    if (gpp->total_iteration == 0) {
    for (j = 0; j < gpp->ninputs; j++) {
        temp = input[j];
        gpp->current_sum[0][j] += temp;
        gpp->current_sum_squared[0][j] += temp * temp;
    }
    gpp->current_count[0]++;
    gpp->total_vectors++;
    } else {
    /* increment count of nearest center */
    /* use this routine so we can get the distance */
    find_k_nearest_neighbors(input, gpp->current_center,
                 &min_dist, &min_i, gpp->ninputs,
                 gpp->current_ncenters, 1);

    for (j = 0; j < gpp->ninputs; j++) {
        temp = input[j];
        gpp->current_sum[min_i][j] += temp;
        gpp->current_sum_squared[min_i][j] += temp * temp;
    }
    gpp->current_count[min_i]++;
    gpp->total_vectors++;
    gpp->total_squared_error += min_dist;
    }
}

/*****************************************************************************
* NAME:     kmeans_batch_update()
* DESCRIPTION:
*           compute new centers, split if necessary, determine if done
*               Return 0 if not done, and number of centers if done.
* EXAMPLE:
*
******************************************************************************/
int
kmeans_batch_update(int model, int *report_iterations,
            int *report_actual_ncenters,
            float *report_ave_squared_error, int *report_nvectors)
{
    float percent_change;
    int i, j;
    double mean, power;
    KMEANS_PARAMS *gpp;
    int *index = NULL, orphan_cnt = 0;
    float *sort_nums;

    gpp = genpp[model];
                /* The iteration is done now label it. First is 1 */
    gpp->current_iteration++;
    gpp->total_iteration++;
                        /* compute new centers  */
    for (i = 0; i < gpp->current_ncenters; i++) {
    if (gpp->current_count[i] == 0) {
        /* if there are no patterns assigned to a cluster, we want
           to move it to near the cluster which accounts for the largest
           portion of the mean squared error.
           */
        /* sort the centers by their error (ncenters * sum(variance)) */
        if(index == NULL){
        int k;

        index = allocate_iindex(gpp->current_ncenters);
        sort_nums = allocate_vector(gpp->current_ncenters);
        for(j = 0; j < gpp->current_ncenters; j++){
            for(sort_nums[j] = 0., k = 0; k < gpp->ninputs; k++){
            sort_nums[j] += gpp->current_variance[j][k];
            }
            sort_nums[j] *= gpp->current_count[j];
        }
        rev_heap_sort(gpp->current_ncenters, sort_nums, index);
        orphan_cnt = 0;
        }
        /* place this orphan near the "best" remaining center.
           we only want to put one orphan with each center each round */
        if(orphan_cnt < gpp->current_ncenters &&
           gpp->current_count[index[orphan_cnt]] > 1){
        float sdev, delta;
        /* place the orphan a random offset away from the "big" one
           as if this were a split. */
        for(j = 0; j < gpp->ninputs; j++){
            gpp->current_center[i][j] =
              gpp->current_center[index[orphan_cnt]][j];
            gpp->current_variance[i][j] =
              gpp->current_variance[index[orphan_cnt]][j];
            sdev = sqrt(gpp->current_variance[i][j]);
            delta = gpp->split_percentage * sdev *
              ((9.5 + random_uniform())/1000.);
            if(random_event(.5)) delta *= -1.;
            gpp->current_center[i][j] += delta;
        }
        } else {
        warning("kmeans_batch_update: There is a cluster with no patterns");
        }
        /* set up so the next orphan goes to another center */
        orphan_cnt++;
    } else {
        for (j = 0; j < gpp->ninputs; j++) {
        mean = gpp->current_sum[i][j] / gpp->current_count[i];
        power = gpp->current_sum_squared[i][j] / gpp->current_count[i];
        gpp->current_center[i][j] = mean;
        gpp->current_variance[i][j] = power - mean * mean;

        /* make sure there is some variance so splits really
           split things */
        if (gpp->current_variance[i][j] < gpp->minimum_variance)
          gpp->current_variance[i][j] = gpp->minimum_variance;
        }
    }
    }
    if(index != NULL){
    deallocate_ivector(index);
    index = NULL;
    deallocate_vector(sort_nums);
    }

                        /* compute average squared error */
    if (gpp->total_vectors > 0)
    gpp->ave_squared_error = gpp->total_squared_error / gpp->total_vectors;
    else
    fatal_error("%s %s",
      "kmeans_batch_update: calling this routine without first calling",
            "kmeans_batch_train");

    /* set return values to current iteration information */
    *report_iterations = gpp->total_iteration;
    *report_actual_ncenters = gpp->current_ncenters;
    *report_ave_squared_error = gpp->ave_squared_error;
    *report_nvectors = gpp->total_vectors;

    if(gpp->ncenters == 0){ /* we don't want any centers */
    return 1;   /* they look for a non-zero return for stopping */
    }
    if (gpp->total_iteration == 1) {
    /* first overall iteration force a split or quit */
    if(gpp->ncenters > 1)
      kmeans_split(model);
    else
      gpp->current_iteration = 0;

    } else if (gpp->current_iteration == 1) {
    /* first iteration after a split, force another iteration */
    } else {
    /* more than one iteration without a split */
    percent_change = 100.0 * (gpp->ave_squared_error - 
        gpp->last_ave_squared_error) / gpp->last_ave_squared_error;
        
    /* error, squared error is increasing */
    if (percent_change > 1e-3){
        /* really small positive changes can be common */
        warning("%s\n%s %g\n%s %g\n%s %g",
            "kmeans_batch_update diverging",
            "Last ave squared error", gpp->last_ave_squared_error,
            "Current squared error", gpp->ave_squared_error,
            "Percentage change", percent_change);
    }
                    /* convergence criterion satisfied */
    if (ABS(percent_change) <= gpp->stop_percentage ||
           gpp->current_iteration >= gpp->max_iteration) {
        if (gpp->current_ncenters == gpp->ncenters) {
        /* all finished */
        gpp->stop_flag = gpp->current_ncenters;
        } else if (gpp->current_ncenters > gpp->ncenters) {
        /* reduce centers */
        if((gpp->current_ncenters - gpp->ncenters) > gpp->reduce_step){
            i = kmeans_reduce_ncenter(model,
                (gpp->current_ncenters - gpp->reduce_step));
        } else {
            i = kmeans_reduce_ncenter(model, gpp->ncenters);
        }
        } else {            /* split centers */
        kmeans_split(model);
        }
    }
    }

                        /* all done, clean up  */
    if (gpp->stop_flag > 0) {
    for (i = 0; i < gpp->current_ncenters; i++)
        for (j = 0; j < gpp->ninputs; j++) {
        gpp->pcenter[i][j] = gpp->current_center[i][j];
        gpp->pvariance[i][j] = gpp->current_variance[i][j];
        }
    return (gpp->current_ncenters);
    /* get ready for another iteration  */
    } else {
    kmeans_local_init(model);
    return (0);
    }
}

/*****************************************************************************
* NAME:     kmeans_split()
* DESCRIPTION:
*           split centers to create twice as many
* EXAMPLE:
*
******************************************************************************/
static void
kmeans_split(int model)
{
    int i, j;
    float mean, sdev, delta;
    static int split_count = 0;
    float frandom;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];
    split_count++;

    for (i = 0; i < gpp->current_ncenters; i++){
    for (j = 0; j < gpp->ninputs; j++) {
        mean = gpp->current_center[i][j];
        sdev = (float) sqrt((double) gpp->current_variance[i][j]);
        /* randomness added to break symetry when two centers have
           exactly the same distance to their nearest neighbor */
        if(gpp->add_random_offset){
        /* get a random number between 0 and 1 */
        frandom = random_uniform();
        delta = gpp->split_percentage * sdev * ((9.5 + frandom)/1000.);
        } else {
        delta = gpp->split_percentage * sdev /100.0;
        }

        gpp->current_center[i + gpp->current_ncenters][j] = mean + delta;
        gpp->current_center[i][j] = mean - delta;
    }
    }
    gpp->current_ncenters *= 2;
    gpp->current_iteration = 0;
    gpp->last_ave_squared_error = gpp->ave_squared_error;
    gpp->splits++;
}


/*****************************************************************************
* NAME:     kmeans_local_init()
* DESCRIPTION:
*           initialize variables at beginning of every iteration
* EXAMPLE:
*
******************************************************************************/
static void
kmeans_local_init(int model)
{
    int i, j;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];
    gpp->total_vectors = 0;     /* total number of vectors */
    gpp->total_squared_error = 0.0; /* total vq error */
    gpp->last_ave_squared_error = gpp->ave_squared_error;

                        /* initialize sums and counters  */
    for (i = 0; i < gpp->current_ncenters; i++) {
    gpp->current_count[i] = 0;
    for (j = 0; j < gpp->ninputs; j++) {
        gpp->current_sum[i][j] = 0.0;
        gpp->current_sum_squared[i][j] = 0.0;
    }
    }
}

/*****************************************************************************
 * NAME:    kmeans_remove_ncenter()
 * DESCRIPTION:
 *      remove the designated center from the model, give its
 *      counter to its nearest center
 * EXAMPLE:
 *
 ****************************************************************************/
void kmeans_remove_ncenter(int model, int center, int nearest_center)
{
    int i, j;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];

    gpp->current_count[nearest_center] += gpp->current_count[center];
    for (i = center; i < (gpp->current_ncenters - 1); i++) {
      
      gpp->current_count[i] = gpp->current_count[i+1];
      for (j = 0; j < gpp->ninputs; j++) {
    gpp->current_center[i][j] = gpp->current_center[i+1][j];
      }
    }
    gpp->current_ncenters--;
    gpp->current_iteration = 0;
}

/*****************************************************************************
* NAME:     kmeans_reduce_ncenter()
* DESCRIPTION:
*           calculate the distance between centers, depending on the
*               center and the number of points in the center, determine
*               the center that can be removed with least increase to overall
*               error.
*               reduce the number of center to the goal specified
* EXAMPLE:
*
*****************************************************************************/
int kmeans_reduce_ncenter(int model, int goal)
{
    double **dist_matrix;
    float *weighted_dist; 
    double dist, min_dist, min_weighted_dist;
    int i, j, min_i, reduce_goal, nearest_center;
    KMEANS_PARAMS *gpp;
    
    gpp = genpp[model];
    
    reduce_goal = gpp->current_ncenters - goal;
    if (reduce_goal <= 0) {
    printf("kmeans_reduce_centers: the number of goals >= ncenter\n");
    }
    
    /* allocate space for a matrix to store distances */
    dist_matrix = allocate_dmatrix(gpp->current_ncenters,
                   gpp->current_ncenters);
    weighted_dist = allocate_vector(gpp->current_ncenters);
    
    while(reduce_goal > 0) {
    /* calculate distances & fill in counters */
    for (i = 0; i < gpp->current_ncenters; i++) {
        for (j = i+1; j < gpp->current_ncenters; j++) {
        dist = l2dist_weight(gpp->ninputs,
                      gpp->current_center[i],
                      gpp->current_center[j],
                      gpp->weight);
        dist_matrix[i][j] = dist;
        dist_matrix[j][i] = dist;
        }
    }
    
    /* now estimate the relative increase in overall error when each
       center is removed by calulating the distance to the nearest
       center multipllied by the number of samples in the cluster 
       Remove the center with lowest number, give its counter value
       to its closest neighbor, and continue */    

    min_weighted_dist = MAXFLOAT;
    for (i = 0; i < gpp->current_ncenters; i++) {
        min_dist = MAXFLOAT;
        for (j = 0; j < gpp-> current_ncenters; j++) {
        if (i != j) {
            if (dist_matrix[i][j] < min_dist) {
            min_dist = dist_matrix[i][j];
            }
        }
        }
        weighted_dist[i] = min_dist * gpp->current_count[i];
        if (weighted_dist[i] < min_weighted_dist) {
        min_weighted_dist = weighted_dist[i];
        min_i = i;
        }
    }

    /* found the center with the minimum weight in min_i
       call remove_ncenter to remove it from model */
    min_dist = MAXFLOAT;
    for (i = 0; i < gpp->current_ncenters; i++) {
        if (i != min_i) {
        if (dist_matrix[min_i][i] < min_dist) {
            min_dist = dist_matrix[min_i][i];
            nearest_center = i;
        }
        }
    }

    kmeans_remove_ncenter(model, min_i, nearest_center);
    reduce_goal--;
    } /* end while */
    deallocate_dmatrix(dist_matrix);
    deallocate_vector(weighted_dist);
    return(min_i);
}
/*****************************************************************************
* NAME: kmeans_cluster()
* DESCRIPTION:
*       Implements the kmeans to return an integer class.
*   does not consider closest exemplar if cvtest_flag is set
*
* EXAMPLE:
*       
******************************************************************************/
int
kmeans_cluster (int model, float *input)
{
    int min_i, i;
    float min_dist;
    KMEANS_PARAMS *gpp;

    gpp = genpp[model];

    min_dist = MAXFLOAT;
    min_i = 0;
    for(i = 0; i < gpp->ncenters; i++){
    gpp->outputs[i] = l2dist(gpp->ninputs, input, &gpp->pcenter[i][0]);
    gpp->outputs[i] = (float)sqrt((double)gpp->outputs[i]);
    if(gpp->outputs[i] < min_dist){
        min_dist = gpp->outputs[i];
        min_i = i;
    }
    }

    return(min_i);
}

/*****************************************************************************
* NAME:     kmeans_combine_models(model1, model2)
* DESCRIPTION:
*       takes two kmeans models and combines them
*       putting the resulting combination in the structure
*       for model1
*       model1 is assumed to have space enough to hold model2,
*       or the combination is not made
* EXAMPLE:
*
*****************************************************************************/
void kmeans_combine_models(int model1, int model2)
{
    KMEANS_PARAMS *gpp1, *gpp2;
    int i, j, k, new_ncenters;

    gpp1 = genpp[model1];
    gpp2 = genpp[model2];

    new_ncenters = gpp1->current_ncenters + gpp2->current_ncenters;
    if(new_ncenters > gpp1->max_centers){
    warning("%s: %s %d (centers: max %d. now %d) %s. %d (%d centers). %s",
        "kmeans_combine_models", "Not enough room in model", model1,
        new_ncenters, gpp1->max_centers, "to hold model", model2,
        gpp2->current_ncenters, "\nnot combining models");
    return;
    }

    for(i = 0, k = gpp1->current_ncenters; i < gpp2->current_ncenters; i++){
    for(j = 0; j < gpp1->ninputs; j++){
        gpp1->pcenter[k][j] = gpp2->pcenter[i][j];
        gpp1->pvariance[k][j] = gpp2->pvariance[i][j];
    }
    }
    gpp1->current_ncenters = new_ncenters;
}

/*****************************************
merge_kmeans_files()
       accept three file names, read the kmeans model from
       first and second file, and then write them to the third
       file
*****************************************/
/* ignores whether the normalization is the same for both files,
   a thing which cannot be checked because there is only one set of
   normalization parameters stored at a time. This also leaks memory
   by not freeing the normalization parameterss */
#ifdef notdef
void merge_kmeans_files(param_path1, param_path2, result_file)
     char *param_path1, *param_path2, *result_file;
{
    int model1, model2, model3, total_ncenter;
    int ninputs1, ninputs2, ncenters1, ncenters2;
    int normalization;
    int i, j, k;
    FILE *param_file;
    float **pcenter, **pvariance;
    float **pcenter1, **pvariance1;
    float **pcenter2, **pvariance2;
    int *pfeatures, features[1024];

    param_file = nn_fopen(param_path1, "r", "for read of cluster parameters");

    model1 = norm_read_paramfile(param_file, &normalization);

    if (model1 < 0) fatal_error ("merge_kmeans: read error %d %s",
                 model1, "on normalization parameter file");

    model1 = kmeans_read_model (param_file, &ninputs1, &ncenters1,
                &pcenter1, &pvariance1, features);
    nn_fclose(param_file, param_path1, "for read of cluster parameters");

    param_file = nn_fopen(param_path2, "r", "for read of cluster parameters");

    model2 = norm_read_paramfile(param_file, &normalization);

    if (model2 < 0) fatal_error ("merge_kmeans: read error %d %s",
                 model2, "on normalization parameter file");
    
    model2 = kmeans_read_model (param_file, &ninputs2, &ncenters2,
                &pcenter2, &pvariance2, features);
    nn_fclose(param_file, param_path2, "for read of cluster parameters");

    total_ncenter =  ncenters1 + ncenters2;

    model3 = kmeans_build(ninputs1, total_ncenter);

    pfeatures = *(int **) kmeans_param(model3, "features");
    for(j = 0; j <= features[0]; j++){
    pfeatures[j] = features[j];
    }

    pcenter = *(float ***) kmeans_param(model3, "pcenter");
    pvariance = *(float ***) kmeans_param(model3, "pvariance");
    for (i = 0; i < ncenters1; i++)
      for (j = 0; j < ninputs1; j++) {
      pcenter[i][j] = pcenter1[i][j];
      pvariance[i][j] = pvariance1[i][j];
      }

    pcenter = *(float ***) kmeans_param(model3, "pcenter");
    pvariance = *(float ***) kmeans_param(model3, "pvariance");
    for (k = 0; i < total_ncenter; k++, i++)
      for (j = 0; j < ninputs1; j++) {
      pcenter[i][j] = pcenter2[k][j];
      pvariance[i][j] = pvariance2[k][j];
      }

    param_file = nn_fopen(result_file, "w", "for read of cluster parameters");
    norm_write_paramfile(param_file);
    kmeans_write_model(model3, 0, param_file);
    nn_fclose(param_file, result_file, "finish writing to result_file");

    kmeans_free(model1);
    kmeans_free(model2);
    kmeans_free(model3);
}
#endif
