/***--------------------------------------------------------------------------
 * NAME:		kmeans_seq_cv.c
 * PURPOSE:	cross validation driver for kmeans clustering algorithm
 *		for sequence data
 *
 * SEE ALSO: 	kmeans_pack.c
 * GLOBALS:	int debug;
 *
 * FUNCTIONS:
 *	kmeans	driver for kmeans clustering algorithm
 *		handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeans_seq_cv.c,v 11.3 2001/08/23 20:12:15 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeans_seq_cv.c,v $
 * CREATED: 16-Aug-93 lck				MIT LINCOLN LABORATORY
 **************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include "kmeans_seq.h"
#include "kmeans_pack.h"

#ifndef TRUE
#define TRUE (1)
#endif
#define MAXSTR (256)
/* globals */
int             debug;		/* enables debug messages if debug > 0*/

/* static variables */
static int *model_list;

void
kmeans_seq_cross_valid(int nraw, int ninputs, int *features, int noutputs,
		       int nwindow_frames, int *window_frames, int normalize,
		       int nsequences, int debg, int verbosity, int cv_nfolds,
		       int cv_nsplits, int use_cv_splits, int *cv_splits,
		       char **cv_train_mask, int cv_random, char *data_path,
		       char *data_fname, char *norm_fname, char *exp_path,
		       char *param_fname, char **labels, char *comment,
		       int max_iterations, float split_percentage,
		       float stop_percentage, int cluster_by_class,
		       int add_random_offset, int reduce_step,
		       int *desired_ncenter)
{
    int		model, err_model, data_model, normalization=0;
    int		current_class, num_in_class, class;
    int		n_read, i, n_files, model_ninputs, n_frames;
    int		fold;
    float	*inputs;
    char	*tmp_path;
    FILE	*param_file;
    int		iteration, actual_ncenters;
    int 	report_stop_flag, report_nvectors;
    float	cost;
    float	*pnorm_variance;	/* normalizing variance */
    int		post_sum_to_one = 0;
    
    debug = debg;		/* set globals */
    
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else norm_setup_none(nraw, noutputs);

    /* build all the models needed */
    model_list = allocate_ivector(noutputs);
    
    /* set normalizing variance to 1 (no weighting of inputs) */
    model_ninputs = ninputs*nwindow_frames;
    pnorm_variance = allocate_vector(model_ninputs);
    for (i=0; i<model_ninputs; i++) pnorm_variance[i] = 1.;
    current_class = 0;
    do{
	model_list[current_class] =
	  kmeans_build(model_ninputs, desired_ncenter[current_class]);
	model = model_list[current_class];
	kmeans_init(model, features,
		    pnorm_variance, stop_percentage, 
		    split_percentage, max_iterations, reduce_step,
		    add_random_offset);
	current_class++;
    } while (cluster_by_class && current_class < noutputs);
    
    model = model_list[0];
    
    /* set up results reporting */
    err_model = report_build(verbosity, 0, noutputs);
    report_init(err_model, 0, model_ninputs, CLUS_SQ_ERR, 0, post_sum_to_one);

    /* read data file */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    seq_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		     cv_nfolds, cv_train_mask, cv_train_mask, cv_random);

    /* store normalization once */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of knn model file");
    /* write comment */
    fprintf(param_file, "%s", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);
    
    for(fold = 0; fold < cv_nfolds; fold++){
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	seq_data_start_cv_fold(data_model, fold, TRUE);
	n_files = seq_data_get_nfiles(data_model);

	current_class = 0;
	do{
	    model = model_list[current_class];
	    kmeans_init(model, features,
			pnorm_variance, stop_percentage, 
			split_percentage, max_iterations, reduce_step,
			add_random_offset);
	    
	    if (verbosity > 1){
		if(cluster_by_class){
		    printf("training centers for class %d", current_class);
		    num_in_class =
		      seq_data_get_nfiles_class(data_model, current_class);
		    report_epoch_header(err_model, stdout,
					num_in_class);
		} else {
		    report_epoch_header(err_model, stdout, n_files);
		}
	    }
	    
	    report_stop_flag = 0;
	    num_in_class = 0;
	    while (!report_stop_flag) {/* for as many epochs as necessary */
		/* initialize data handling routines */
		seq_data_epoch_init(data_model);
		/* read all the data */
		for(n_read = 0; n_read < n_files; n_read++){
		    if(seq_data_get_next_file(data_model) == -1)
		      break;
		    class = seq_data_get_class(data_model);

		    /* train on all data if not clustering by class.
		       train only on the data from this class
		       otherwise */
		    if(!cluster_by_class ||
		       (cluster_by_class && class == current_class)){
			for(n_frames = 0; ; n_frames++){
			    inputs = seq_data_get_next_frame(data_model);
			    if(inputs == NULL) break;
			    kmeans_batch_train (model, inputs);
			}
			if(n_frames > 0) num_in_class++;
		    }
		}
		if(num_in_class != 0){
		    report_stop_flag =
		      kmeans_batch_update(model, &iteration,
					  &actual_ncenters, &cost,
					  &report_nvectors);
		} else {
		    warning("There was no data in class %d\n",
			    current_class);
		    report_stop_flag = 1;
		}
		
		report_cluster_epoch(err_model, stdout, iteration,
				     actual_ncenters, cost);
	    }
	    current_class++;
	} while(current_class < noutputs && cluster_by_class);
	
	/* write out model or models */
	tmp_path = path_build(exp_path, param_fname);
	/* store model in param file */
	fprintf(param_file, "kmeans model %d\n", fold);
	if(cluster_by_class){
	    kmeans_write_model_list(model_list, noutputs, param_file);
	} else {
	    kmeans_write_model(model_list[0], noutputs, param_file);
	}
	/* as there is no testing, we don't have to do any */
	/* finish the training fold */
	seq_data_end_cv_fold(data_model);
    } /* end fold */
    
    /* close parameter file */
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    
    /* free model structures */
    deallocate_vector(pnorm_variance);
    current_class = 0;
    if(cluster_by_class){
	for(current_class = 0; current_class < noutputs; current_class++)
	  kmeans_free(model_list[current_class]);
    } else {
	kmeans_free(model_list[0]);
    }
    report_free(err_model);
    seq_data_exit(data_model);
    norm_free();
}
