/***--------------------------------------------------------------------------
 * NAME:		kmeans.c
 * PURPOSE:	driver for kmeans clustering algorithm
 *
 * SEE ALSO: 	kmeans_pack.c
 * GLOBALS:	int debug;
 *
 * FUNCTIONS:
 *	kmeans	driver for kmeans clustering algorithm
 *		handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeansm_train.c,v 11.4 2001/08/23 20:12:16 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/kmeans/kmeansm_train.c,v $
 * CREATED: 15-Oct-92 lck				MIT LINCOLN LABORATORY
 *------------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "kmeansm.h"
#include "kmeans_pack.h"

#define MAXSTR (256)
/* globals */
int             debug;		/* enables debug messages if debug > 0*/

void
kmeansm_train(int nraw, int ninputs, int *features, int noutputs,
	      int normalize, int npatterns, int debg, int verbosity,
	      char *data_path, char *data_fname, char *norm_fname,
	      char *exp_path, char *param_fname, char *comment,
	      int max_iterations, float split_percentage,
	      float stop_percentage, int add_random_offset, int reduce_step,
	      int desired_ncenters)
{
    int		model, normalization=0;
    int		n_read, i, n_trials;
    float	*inputs, rms_err = 0.0;
    float	**input_matrix, **output_matrix;
    char	data_fullname[MAXSTR], *tmp_path;
    FILE	*data_file, *param_file;
    int		iteration, actual_ncenters;
    int 	report_stop_flag, report_nvectors;
    float	ave_squared_error;
    float	*pnorm_variance;	/* normalizing variance */
    
    debug = debg;		/* set globals */
    
    /* put together the full path to the data file*/
    strcpy(data_fullname, path_build(data_path, data_fname));
    
    /* allocate and initialize model structure */
    pnorm_variance = allocate_vector(ninputs);
    for (i=0; i<ninputs; i++) pnorm_variance[i] = 1.;
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));

    model = kmeans_build(ninputs, desired_ncenters);
    kmeans_init(model, features,
		pnorm_variance, stop_percentage, 
		split_percentage, max_iterations, reduce_step,
		add_random_offset);
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* read data file */
    data_file = nn_fopen(data_fullname, "r", "for training or test data");
    if (verbosity > 1)		
      printf("\nReading %s\n", data_fullname);
    input_matrix = allocate_matrix(npatterns, ninputs);
    output_matrix = allocate_matrix(npatterns, noutputs);
    n_trials = load_map_data(data_file, npatterns, nraw, noutputs,
			       ninputs, output_matrix, input_matrix,
			       normalization, data_fname, features);
    nn_fclose(data_file, data_fullname, "for training or test data");
    
    if (verbosity > 1){
	printf("\n EPOCH ave_sq_err ncenters (%d patterns/epoch)\n",
	       n_trials);
    }
	
    report_stop_flag = 0;
    while (!report_stop_flag) { /* for as many epochs as necessary */
	for (n_read = 0; n_read < n_trials; n_read++) {
	    /* train on input from data file */
	    inputs = &input_matrix[n_read][0];

	    kmeans_batch_train (model, inputs);
	}
	report_stop_flag =
	  kmeans_batch_update (model, &iteration, &actual_ncenters,
			       &ave_squared_error, &report_nvectors);
	if (verbosity>1) {
	    printf("%4d %10.3f %7d\n", iteration, ave_squared_error,
		   actual_ncenters);
	    fflush(stdout);	/* ensures timely results */
	}
    }
    
    deallocate_matrix(input_matrix);
    
    /* write out model or models */
    tmp_path = path_build(exp_path, param_fname);
    /* store model in param file */
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    
    kmeans_write_model(model, noutputs, param_file);
    
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    rms_err = sqrt(ave_squared_error);
    printf("\nClustering results:  ncenters = %d  rmserr = %f\n\n", 
	   actual_ncenters, rms_err);
    
    /* free model structures */
    deallocate_vector(pnorm_variance);
    norm_free();
    kmeans_free(model);
}
