/***---------------------------------------------------------------------------
* NAME:     knn_pack.c
* PURPOSE:  Functions specific to knn.c
* 
* SEE ALSO: knn.c knn_main.c
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   knn_build()     Allocate storage for model parameters.
*   knn_init()      Initialize model parameters.
*   knn_free()      Deallocates model storage
*   knn_read_model()    Initializes a knn model from file
*   knn_write_model()   Writes a knn model to output file
*   knn_param()     Returns pointers to elements in model
*   knn_inc_ctrain()    does training on a vector
*   knn_classify()      Implements knn to return an integer class
*   knn_copy_exemplar() returns class of exemplar copied
* 
* ALGORITHM:
*   Training:  All training vectors are stored for use as exemplars
*       the vectors are normalized to have 0 mean 1 standard deviation
*       if requested
*
*   Testing:  For each test vector, the distance to each training vector
*       is calculated in order to find the K closest training vectors.
*       The majority class of those training vectors is returned as the
*       class of the test vector
*       
* REFERENCE: Kenney Ng, "A Comparative Study of the Practical Characteristics
*   of Neural Network and Conventional Pattern Classifiers", May 1990
*   Master's Thesis for the MIT Department of Electrical Engineering and
*   Computer Science
* 
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/knn/knn_pack.c,v 11.3 2001/08/23 20:12:19 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/knn/knn_pack.c,v $
* CREATED: 12-Dec-90 lck                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define inc_pack_h_internal_types
#include "knn_pack.h"

/* constant definitions */
#define MODEL_STORE_SIZE 10
#define SPACE 1024
#define PEQ(y) (0==strcmp(param_name,y))
                /* Internal Globals */
static KNN_PARAMS **knnpp;  /* array of parameters for each model */
static int max_models = 0;  /* current length of knnpp */


/*****************************************************************************
* NAME: knn_build(ninputs, nclasses, npatterns, k)
*       ninputs        - number of dimensions in training vectors
*       nclasses       - number of classes the data falls into
*       npatterns      - number of vectors to be stored
*       k               - number of nearest neighbors used in classification
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*       knn_build(2,10,338,5)
*               build a model for a 5 nearest neighbor classifier with
*               338 stored data points.
*               The data has 2 dimensions and falls into 10 classes
*
******************************************************************************/
int
knn_build(int ninputs, int nclasses, int npatterns, int k)
{
    register int new_model;
    KNN_PARAMS *kpp;

            /* find entry in model array for new model */
    new_model = model_get_number((void ***)&knnpp,
                 &max_models, sizeof(KNN_PARAMS));
    kpp = knnpp[new_model];

    /* set fields which are array sizes */
    kpp->ninputs = ninputs;
    kpp->nclasses = nclasses;
    kpp->max_exemplars = npatterns-1;
    kpp->k = k;
    kpp->accum_trials = 0;

    /* allocate space for vectors */
    kpp->classes = allocate_ivector(npatterns);
    kpp->exemplars = allocate_matrix(npatterns,ninputs);
    kpp->features = allocate_ivector(ninputs+1);
    kpp->outputs = allocate_vector(nclasses);
    kpp->class_vote = allocate_ivector(nclasses);
    kpp->near_dist = allocate_vector(npatterns);
    kpp->near_index = allocate_ivector(npatterns);
    kpp->vars = allocate_vector(ninputs);

    return new_model;
}
/*****************************************************************************
* NAME: knn_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
******************************************************************************/
void
knn_init(int model, int accum_trials, int *nn_features)
{
    int i;
    KNN_PARAMS *kpp;

    kpp = knnpp[model];
    kpp ->accum_trials = accum_trials;

    /* copy feature data into feature list */
    for(i = 0; i <= nn_features[0]; i++)
    kpp->features[i] = nn_features[i];
}

/*****************************************************************************
* NAME: knn_free()
* DESCRIPTION:
*       Deallocate storage used in a model 
*
* EXAMPLE:
*
******************************************************************************/
void
knn_free(int model)
{
    KNN_PARAMS *kpp;
    kpp = knnpp[model];
            /* free vectors and matrices */
    deallocate_ivector (kpp->classes);
    deallocate_ivector (kpp->features);
    deallocate_matrix (kpp->exemplars);
    deallocate_ivector (kpp->class_vote);
    deallocate_vector (kpp->near_dist);
    deallocate_ivector (kpp->near_index);
    deallocate_vector (kpp->vars);
    deallocate_vector (kpp->outputs);
            /* free structure itself */
    nn_free((char *)kpp, "for knn parameter structure");
    knnpp[model] = NULL;
}
/*****************************************************************************
* NAME: knn_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*   allocates space for and loads data into vectors and matricies
*
*   On error from param_getblk, returns the error value
*   (there is already an error check on the allocation, so we don't
*   have to do that here)
*
* EXAMPLE:
******************************************************************************/
int
knn_read_model(FILE *fp, int *accum_trials)
{
    KNN_PARAMS *kpp;
    int model;
    int rv,k,ninputs,nclasses;

    /* read in scalars from model file */
    param_setvarptr("k", &k, paramtab, TABSIZE);
    param_setvarptr("accum_trials", accum_trials, paramtab, TABSIZE);
    param_setvarptr("ninputs", &ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &nclasses, paramtab, TABSIZE);

    if((rv = param_getblk(fp, paramtab, TABSIZE)) < 0) return rv;

    /* allocate space for model and vectors*/
    model = knn_build(ninputs, nclasses, *accum_trials, k);
    kpp = knnpp[model];
  
    /* read in vectors from model file */
    param_setvarptr ("classes", &kpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim ("classes", kpp->max_exemplars+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr ("exemplars", (int *)&kpp->exemplars[0][0],
             paramtab_l, TABSIZE_L);
    param_setdim ("exemplars", kpp->max_exemplars+1, kpp->ninputs,
          paramtab_l, TABSIZE_L);
    param_setvarptr ("features", &kpp->features[0], paramtab_l, TABSIZE_L);
    /* set features length to the largest it can be */
    param_setdim("features", kpp->ninputs+1, 0, paramtab_l, TABSIZE_L);

    if((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0) return rv;

    /* set remaining parameters */
    kpp->accum_trials = *accum_trials;

    return(model);
}

/*****************************************************************************
* NAME: knn_write_model()
* DESCRIPTION:
*   writes parameters and exemplars for a model to a file
*   in ASCII format
*
* EXAMPLE:
*
******************************************************************************/
int
knn_write_model(int model, FILE *fp)
{
    int rv;
    KNN_PARAMS *kpp;

    kpp = knnpp[model];

    /* save scalars */
    param_setvarptr("k", &kpp->k, paramtab, TABSIZE);
    param_setvarptr("accum_trials", &kpp->accum_trials, paramtab, TABSIZE);
    param_setvarptr("ninputs", &kpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &kpp->nclasses, paramtab, TABSIZE);
    if((rv = param_putblk(fp, paramtab, TABSIZE)) < 0) return rv;

    /* save vectors and matrices */
    param_setvarptr ("classes", &kpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim ("classes", kpp->accum_trials, 0, paramtab_l, TABSIZE_L);
    param_setvarptr ("exemplars", (int *)&kpp->exemplars[0][0],
             paramtab_l, TABSIZE_L);
    param_setdim ("exemplars", kpp->accum_trials, kpp->ninputs, 
        paramtab_l, TABSIZE_L);
    param_setvarptr("features", &kpp->features[0], paramtab_l,TABSIZE_L);
    param_setdim("features", kpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    if((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0) return rv;

    return(0);
}
/************************************************************************
* NAME: knn_param(model, param_name)
* DESCRIPTION:
*   Takes model number and parameter name.
*   Returns a pointer to the named parameter.
*
*************************************************************************/
void *
knn_param(int model, char *param_name)
{
    void *param_address;
    KNN_PARAMS *kpp;
                
    kpp = knnpp[model];
    if (PEQ("k")) param_address = &kpp->k;
    else if (PEQ("accum_trials")) param_address = &kpp->accum_trials;
    else if (PEQ("exemplars")) param_address = &kpp->exemplars;
    else if (PEQ("classes")) param_address = &kpp->classes;
    else if (PEQ("features")) param_address = &kpp->features;
    else if (PEQ("ninputs")) param_address = &kpp->ninputs;
    else if (PEQ("nclasses")) param_address = &kpp->nclasses;
    else if (PEQ("vars")) param_address = &kpp->vars;
    else if (PEQ("outputs")) param_address = &kpp->outputs;
    else {
    warning("Parameter %s not defined", param_name);
    param_address = 0;
    }
    return param_address;
} /* END knn_param() */

/*****************************************************************************
* NAME: knn_inc_ctrain()
* DESCRIPTION:
*       copies input vector into training set.
* EXAMPLE:
*       
******************************************************************************/
int
knn_inc_ctrain(int model, float *inputs, int desired_class)
{
    register int i,  newlength;
    KNN_PARAMS *kpp;

    kpp = knnpp[model];

    /* check to see if we have the room in exemplars for new vector */
    if(kpp->accum_trials > kpp->max_exemplars){
    /* figure out how many vectors we can add to exemplar */
    kpp->max_exemplars++; /* now the number of vectors in exemplars */
    newlength = kpp->max_exemplars + SPACE/(sizeof(float)*kpp->ninputs);
    kpp->exemplars = reallocate_matrix(kpp->exemplars, kpp->max_exemplars,
        kpp->ninputs, newlength);
    kpp->classes = reallocate_ivector(kpp->classes,
        kpp->max_exemplars, newlength);
    kpp->near_index = reallocate_ivector(kpp->near_index,
        kpp->max_exemplars, newlength);
    kpp->near_dist = reallocate_vector(kpp->near_dist,
        kpp->max_exemplars, newlength);
    kpp->max_exemplars = newlength - 1;
    }

    /* store input in exemplars, class in classes, and increment counter */
    for(i = 0; i < kpp->ninputs; i++){
    kpp->exemplars[kpp->accum_trials][i] = inputs[i];
    }
    kpp->classes[kpp->accum_trials] = desired_class;
    kpp->accum_trials++;

    return desired_class;
}
/*****************************************************************************
* NAME: knn_classify()
* DESCRIPTION:
*       Implements the knn to return an integer class.
*   does not consider closest exemplar if cvtest_flag is set
*
* EXAMPLE:
*       
******************************************************************************/
int
knn_classify (int model, float *inputs,
          int k,/* overrides the k in the model file */ int cvtest_flag)
{
    int winner;
    int i;
    KNN_PARAMS *kpp;

    kpp = knnpp[model];

    if(k > kpp->accum_trials) k = kpp->accum_trials;
    /* find the k+1 nearest neighbors */
    find_k_nearest_neighbors(inputs, kpp->exemplars, kpp->near_dist,
                 kpp->near_index, kpp->ninputs, kpp->accum_trials,
                 k + 1);

    /* check for -1 in any of these near_indicies */
    /* make a histogram of the classes of the k nearest neighbors */
    for(i = 0; i < kpp->nclasses; i++) kpp->class_vote[i] = 0;
    for(i = 0; i < k; i++){
    if(kpp->near_index[i] < 0 ||
       kpp->near_index[i] >= kpp->accum_trials)
      continue;
    kpp->class_vote[kpp->classes[kpp->near_index[i]]]++;
    }
    if(cvtest_flag){
    if(k+1 < kpp->accum_trials){
        /* take nearest neighbor out of vote.  add k+1'th */
        kpp->class_vote[kpp->classes[kpp->near_index[0]]]--;
        kpp->class_vote[kpp->classes[kpp->near_index[k]]]++;
    }
    }
    /* find most common class */
    winner = max_pick_int(kpp->class_vote, kpp->nclasses);

    /* set outputs to the normalized number of patterns in each class */
    for(i = 0; i < kpp->nclasses; i++)
      kpp->outputs[i] = (float)kpp->class_vote[i]/(float)k;

    return(winner);
}
/************************************************************************
* NAME: knn_copy_exemplar(model, pattern, inputs)
* DESCRIPTION:
*   copies exemplar[pattern] into inputs
*   returns classes[pattern]
*
*************************************************************************/
int
knn_copy_exemplar(int model, int pattern, float *inputs)
{
    int i;
    KNN_PARAMS *kpp;

    kpp = knnpp[model];
    for(i = 0; i < kpp->ninputs; i++){
    inputs[i] = kpp->exemplars[pattern][i];
    }
    return(kpp->classes[pattern]);
} /* END knn_copy_exemplar() */
