/***--------------------------------------------------------------------------
 * NAME:		knn_test.c
 * PURPOSE:	testing driver for knn classifier
 *
 * SEE ALSO: knn_pack.c
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/knn/knn_test.c,v 11.4 2001/08/23 20:12:19 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/knn/knn_test.c,v $
 * CREATED: 12-Dec-90 lck                             MIT LINCOLN LABORATORY
 ****************************************************************************/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "knn.h"
#include "knn_pack.h"

#define MAXSTR (256)

static char prog_name[MAXSTR];	/* static variables */

void
knn_test(int create, int nraw, int ninputs, int *features, int noutputs,
	 int npatterns, int debg, int verbosity, int verror, char *data_path,
	 char *data_fname, char *exp_path, char *error_fname,
	 char *param_fname, char **labels, int k, int cvtest_flag)
{
    int model, err_model, rv, normalization = 0, data_model = -1;
    int desired_class, class;
    int n_read, trial_num;
    int accum_trials;
    int map_flag = 0;	/* needed for error reporting routines */
    float *inputs = NULL, *outputs;
    char *tmp_path;
    FILE *error_file = NULL, *param_file;
    int	post_sum_to_one = 0;
   
    strcpy(prog_name, get_program_name());
    
    /* do the initialization of the training data
     * structure, reading the previously stored training
     * vectors in the case of retrain and test also
     * initialize the normalization routines, if needed
     */
    
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of knn model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error ("%s: read error %d %s %s", prog_name, rv,
		   "on normalization parameter file", tmp_path);
    model = knn_read_model(param_file, &accum_trials);
    if(model < 0)
      fatal_error("%s: error reading model file %s", prog_name, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of knn model file");
    
    if(k > accum_trials){
	fatal_error("knn -- only %d %s %s\n%s %d %s",
		    accum_trials, "exemplars in model", tmp_path,
		    "need at least", k, "to do classification");
    }
    if(cvtest_flag) npatterns = accum_trials;
    
    /* use the features from the model */
    features = *(int **) knn_param(model, "features");
    ninputs = *(int *) knn_param(model, "ninputs");
    outputs = *(float **) knn_param(model, "outputs");
    
    if (verror>0) {		/* open file for classification errors */
	if(create)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    }
    
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init(err_model, map_flag, ninputs, SQ_ERR, 0, post_sum_to_one);
    
    /* open data file */
    if(cvtest_flag){
    	inputs = allocate_vector(ninputs);
    } else {
	data_model =
	  class_data_single_init(data_path, data_fname, npatterns,
				 nraw, ninputs, noutputs, features,
				 normalization, verbosity);
    }
    /* read and classify input from data file */
    for(n_read = 0; n_read < npatterns; n_read++){
	/* read input vector */
	if(cvtest_flag){
	    desired_class = knn_copy_exemplar(model, n_read, inputs);
	    trial_num = n_read;
	}else{
	    inputs = class_data_get_next(data_model, &desired_class);
	    trial_num = class_data_get_input_num(data_model);
	    if(inputs == NULL) break;
	}
	
	/* classify pattern and report errors */
	class = knn_classify(model, inputs, k, cvtest_flag);
	report_class_trial(err_model, error_file, inputs, outputs,
			   desired_class, class, trial_num);
    }
    if(cvtest_flag){
	deallocate_vector(inputs);
    } else {
	class_data_exit(data_model);
    }
    
    report_overall(err_model, stdout, error_fname, labels);
    report_test_result(err_model, stdout, data_fname);
    report_free(err_model);

    if(verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    
    /*free space and close files, for neatness sake */
    knn_free(model);
    norm_free();
}
