/***--------------------------------------------------------------------------
 * NAME:		lead_clus.c
 * PURPOSE:	driver for lead_clus clustering algorithm
 *
 * SEE ALSO: 	lead_clus_pack.c
 *
 * FUNCTIONS:
 *	lead_clus	driver for lead_clus clustering algorithm
 *		handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lead_clus/lead_clus_cv.c,v 11.4 2001/08/23 20:12:22 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lead_clus/lead_clus_cv.c,v $
 * CREATED: 6-Apr-92 lck			MIT LINCOLN LABORATORY
 ***************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "lead_clus.h"
#include "lead_clus_pack.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define MAXSTR (256)

/* static variables */
static char     prog_name[MAXSTR];

void
lead_clus_cross_valid(int nraw, int ninputs, int *features, int noutputs,
		      int normalize, int npatterns, int debg, int verbosity,
		      int random_flag, int cv_nfolds, int cv_nsplits,
		      int use_cv_splits, int *cv_splits, char **cv_train_mask,
		      char **cv_test_mask, char *data_path, char *data_fname,
		      char *norm_fname, char *exp_path, char *param_fname,
		      char *comment, int adjust_train, int priors_npatterns,
		      float *train_priors_list, int cv_random, float *deltas,
		      int cluster_by_class)
{
    int		model, normalization=0, data_model;
    int		current_class, class;
    int		n_read, n_trials;
    int		fold;
    float	*inputs;
    char	data_fullname[MAXSTR], *tmp_path;
    FILE	*param_file;
    int		total_ncenters;
    int *model_list;
#ifdef DEBUG
    int		err_model;
    FILE *error_file;
    char *error_fname;
    float *outputs;
    int desired_class;
    int		post_sum_to_one = 0;
    error_fname = "lead_clus.err.cv";
    verror = 2;
    outputs = allocate_vector(noutputs);
    if (verror > 0) {
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      "w", "for classification errors");
    } else error_file = NULL;
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, 0, ninputs, CLUS_SQ_ERR, 0, post_sum_to_one);
#endif
    
    strcpy(prog_name, get_program_name());
    
    /* put together the full path to the data file*/
    strcpy(data_fullname, path_build(data_path, data_fname));
    
    model_list = allocate_ivector(noutputs);
    /* allocate and initialize model structure */
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    current_class = 0;
    do{
	model_list[current_class] = lead_clus_build(ninputs, 0);
	model = model_list[current_class];
	lead_clus_init(model, features, deltas[current_class]);
	current_class++;
    } while (cluster_by_class && current_class < noutputs);
    
    model = model_list[0];
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /************ read data and initialize class_data routines ************/
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		       cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_train)
      class_data_priors_init(data_model, train_priors_list, priors_npatterns);

    /* store normalization once */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of knn model file");
    /* write comment */
    fprintf(param_file, "%s", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    
    for(fold = 0; fold < cv_nfolds; fold++){
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	if(cluster_by_class){
	    for(class = 0; class < noutputs; class++)
	      lead_clus_init(model_list[class], features, deltas[class]);
	} else lead_clus_init(model_list[0], features, deltas[0]);

	class_data_start_cv_fold(data_model, fold, TRUE);
	n_trials = class_data_get_npatterns(data_model);
	
	for(n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &class))
	       == NULL) break;
	    if(cluster_by_class){
		lead_clus_inc_train (model_list[class], inputs);
	    } else {
		lead_clus_inc_train (model_list[0], inputs);
	    }
	}
	class_data_end_cv_fold(data_model);
	
	if(verbosity > 0){
	    printf("CLASS NCENTERS\n");
	    if(cluster_by_class){
		total_ncenters = 0;
		for(class = 0; class < noutputs; class++){
		    printf("%d\t%d\n", class,
			   *(int *)lead_clus_param(model_list[class],
						   "ncenters"));
		    total_ncenters +=
		      *(int *)lead_clus_param(model_list[class],
					      "ncenters");
		}
		printf("--------------\n");
		printf("TOTAL\t%d\n", total_ncenters);
	    } else {
		printf("NCENTERS STORED: %d\n",
		       *(int *)lead_clus_param(model_list[0], "ncenters"));
	    }
	} 
	
	/* store model in param file */
	fprintf(param_file, "lead_clus model %d\n", fold);
	if(cluster_by_class){
	    lead_clus_write_model_list(model_list, noutputs, param_file);
	} else {
	    lead_clus_write_model(model_list[0], noutputs, param_file);
	}
	/* as there is no testing, we don't have to do any */
#ifdef DEBUG
	/* check the fold assignments for matching with other programs */
	class_data_start_cv_fold(data_model, fold, FALSE);
	n_trials = class_data_get_npatterns(data_model);

	for (n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    report_cluster_trial(err_model, error_file, inputs, outputs,
				 desired_class, desired_class,
				 class_data_get_input_num(data_model));
	}
	class_data_end_cv_fold(data_model);
#endif
    } /* end fold */
#ifdef DEBUG
    deallocate_vector(outputs);
    nn_fclose(error_file, error_fname, "for write of errors");
#endif

    /* close parameter file */
    tmp_path = path_build(exp_path, param_fname);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);

    class_data_exit(data_model);

    /* free model structures */
    current_class = 0;
    if(cluster_by_class){
	for(current_class = 0; current_class < noutputs; current_class++)
	  lead_clus_free(model_list[current_class]);
    } else {
	lead_clus_free(model_list[0]);
    }
    deallocate_ivector(model_list);
    norm_free();
}
