/***--------------------------------------------------------------------------
* NAME:     lead_clus_pack.c
* PURPOSE:  Functions specific to lead_clus.c
*
* SEE ALSO: lead_clus.c lead_clus_main.c
*
* USAGE:
*   model = lead_clus_build(...)
*   lead_clus_init(model,...)
*   for(i=0; i<all_patterns; i++)
*       lead_clus_inc_train(model, input[i]);
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   lead_clus_build     Allocate storage for model parameters.
*   lead_clus_init      Initialize model parameters.
*   lead_clus_free      Deallocates model storage
*   lead_clus_read_model    Initializes model from file
*   lead_clus_write_model   Writes model to output file
*   lead_clus_param     Returns pointers to elements in model
*   lead_clus_inc_train find distance of a point to all the stored
*               points, save it if it is too far away
*   lead_clus_cluster   returns the cluster an input pattern falls in
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lead_clus/lead_clus_pack.c,v 11.3 2001/08/23 20:12:22 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lead_clus/lead_clus_pack.c,v $
* CREATED: 7-Apr-92 lck             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define lead_clus_pack_h_internal_types
#define lead_clus_pack_h_statics
#include "lead_clus_pack.h"

                /* constant definitions */
#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))
#define MATRIX_STORE_SIZE 128

                /* globals */
extern int   debug;     /* enables debug messages if debug > 0 */

                /* Internal Globals */
static LEAD_CLUS_PARAMS **leadpp;/* array of parameters for each model */
static int      max_models = 0; /* current length of leadpp */

/*****************************************************************************
* NAME: lead_clus_build(ninputs, ncenters)
*   ninputs - number of dimensions in training vectors
*   ncenters - number of cluster centers
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
******************************************************************************/
int
lead_clus_build(int ninputs, int ncenters)
{
    register int    new_model;
    LEAD_CLUS_PARAMS    *lpp;

    new_model = model_get_number((void ***)&leadpp,
                 &max_models, sizeof(LEAD_CLUS_PARAMS));
    lpp = leadpp[new_model];

                    /* set fields which are array sizes */
    lpp->ninputs = ninputs;
    lpp->ncenters = ncenters;

    if(ncenters == 0) lpp->max_centers = MATRIX_STORE_SIZE;
    else lpp->max_centers = ncenters;

                    /* allocate space */
    lpp->features = allocate_ivector (ninputs+1);
    lpp->outputs = allocate_vector(lpp->max_centers);

    lpp->pcenter = allocate_matrix (lpp->max_centers, ninputs);
    lpp->pvariance = allocate_matrix (lpp->max_centers, ninputs);

    return new_model;
}

/*****************************************************************************
* NAME: lead_clus_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
******************************************************************************/
void
lead_clus_init (int model, int *nn_features, float delta)
{
    int     i;
    LEAD_CLUS_PARAMS *lpp;

    lpp = leadpp[model];
                    /* copy feature data into feature list */
    for (i = 0; i <= nn_features[0]; i++)
    lpp->features[i] = nn_features[i];

                    /* set global variables */
    lpp->delta = delta;
    lpp->ncenters = 0;
    lpp->total_vectors = 0;
}

/*****************************************************************************
* NAME: lead_clus_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
******************************************************************************/
void
lead_clus_free(int model)
{
    LEAD_CLUS_PARAMS *lpp;

    lpp = leadpp[model];
                        /* free vectors and matrices */
    deallocate_ivector (lpp->features);
    deallocate_vector (lpp->outputs);
    deallocate_matrix (lpp->pcenter);
    deallocate_matrix (lpp->pvariance);
                        /* free structure itself */
    nn_free((char *) lpp, "for lead_clus parameter structure");
    leadpp[model] = NULL;
}


/*****************************************************************************
* NAME: lead_clus_read_model()
* DESCRIPTION:
*       reads parameters for a model from a file in ASCII format
*
* EXAMPLE:
*
******************************************************************************/
int
lead_clus_read_model(FILE *fp)
{
    int model, noutputs, i, j, clus_model;
    float delta, **pcenter_part, **pvariance_part;
    LEAD_CLUS_PARAMS *lpp;
    int ninputs, ncenters, *features;

    clus_model = cluster_read_model(fp);
    if(clus_model < 0) return(clus_model);
    /* can only read in data as a single model */
    cluster_merge_centers(clus_model);

    ninputs = *(int *)cluster_param(clus_model, "ninputs", 0);
    noutputs = *(int *)cluster_param(clus_model, "noutputs", 0);
    ncenters = *(int *)cluster_param(clus_model, "ncenters", 0);

    model = lead_clus_build(ninputs, ncenters);
    lpp = leadpp[model];

    features = *(int **)cluster_param(clus_model, "features", 0);

    pcenter_part = *(float ***)cluster_param(clus_model, "pcenter", 0);
    pvariance_part = *(float ***)cluster_param(clus_model, "pvariance", 0);
    for(i = 0; i < ncenters; i++){
    for(j = 0; j < ninputs; j++){
        lpp->pcenter[i][j] = pcenter_part[i][j];
        lpp->pvariance[i][j] = pvariance_part[i][j];
    }
    }

    if(ninputs > 0 && ncenters > 0)
      delta = 4.*lpp->pvariance[0][0];
    else delta = 0.;
    lead_clus_init(model, features, delta);

    /* ncenters gets reset to 0 in lead_clus_init */
    lpp->ncenters = ncenters;

    cluster_free(clus_model);
    return(model);
}

/*****************************************************************************
* NAME: lead_clus_write_model()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int
lead_clus_write_model(int model, int noutputs, FILE *fp)
{
    int rv, cluster_by_class = 0, *ncenters_list = NULL;
    LEAD_CLUS_PARAMS *lpp;

    lpp = leadpp[model];

    /* use the version of cluster write that doesn't use model numbers */
    rv = cluster_write_paramfile(fp, lpp->ninputs, lpp->ncenters,
                 noutputs, &(lpp->pcenter), &(lpp->pvariance),
                 lpp->features, cluster_by_class,
                 ncenters_list);
    return(rv);
}

/*****************************************************************************
* NAME: lead_clus_write_model_list()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int
lead_clus_write_model_list(int *model_list, int noutputs, FILE *fp)
{
    int rv, cluster_by_class = 1, i, *ncenter_list, ncenters;
    float ***center_list, ***variance_list;
    LEAD_CLUS_PARAMS *lpp;

    if(noutputs <= 0){
      return(-1);
    }
    /* list of centers per class */
    ncenter_list = allocate_ivector(noutputs);
    /* list of pointers to matrices of centers and variances */
    center_list = (float ***)allocate_pvector(noutputs);
    variance_list = (float ***)allocate_pvector(noutputs);

    for(ncenters = 0, i = 0; i < noutputs; i++){
    lpp = leadpp[model_list[i]];
    ncenter_list[i] = lpp->ncenters;
    ncenters += lpp->ncenters;
    center_list[i] = lpp->pcenter;
    variance_list[i] = lpp->pvariance;
    }

    lpp = leadpp[model_list[0]];

    /* use the version of cluster write that doesn't use model numbers */
    rv = cluster_write_paramfile(fp, lpp->ninputs, ncenters,
                 noutputs, center_list, variance_list,
                 lpp->features, cluster_by_class,
                 ncenter_list);
    deallocate_ivector(ncenter_list);
    deallocate_pvector((void **)center_list);
    deallocate_pvector((void **)variance_list);

    return(rv);
}

/************************************************************************
* NAME: lead_clus_param(model, param_name)
* DESCRIPTION:
*       Takes parameter name.
*       Returns a pointer to the named parameter.
*************************************************************************/
void    *
lead_clus_param(int model, char *param_name)
{
    void    *param_address;
    LEAD_CLUS_PARAMS *lpp;

    lpp = leadpp[model];
    if (PEQ("features"))
      param_address = &lpp->features;
    else if (PEQ("outputs"))
      param_address = &lpp->outputs;
    else if (PEQ("ninputs"))
      param_address = &lpp->ninputs;
    else if (PEQ("ncenters"))
      param_address = &lpp->ncenters;
    else if (PEQ("pcenter"))
      param_address = &lpp->pcenter;
    else if (PEQ("pvariance"))
      param_address = &lpp->pvariance;
    else if (PEQ("delta"))
      param_address = &lpp->delta;
    else {
    warning("lead_clus_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}

/*****************************************************************************
* NAME:     lead_clus_inc_train()
* DESCRIPTION:
*           provide another vector for lead_clus training
* EXAMPLE:
*
******************************************************************************/
void
lead_clus_inc_train(int model, float *input)
{
    float   dist, min_dist;
    int     i;
    LEAD_CLUS_PARAMS *lpp;

    lpp = leadpp[model];
                    /* find nearest center */
    min_dist = MAXFLOAT;
    for (i = 0; i < lpp->ncenters; i++) {
    dist = l2dist(lpp->ninputs, input, lpp->pcenter[i]);
    if (dist < min_dist) min_dist = dist;
    }
    if(min_dist > lpp->delta) {
    /* new point is outside radius of all stored points
       store it */
    if(lpp->ncenters == lpp->max_centers){
        lpp->max_centers += MATRIX_STORE_SIZE;
        /* need to lengthen matrix */
        lpp->pcenter =
          reallocate_matrix(lpp->pcenter, lpp->ncenters,
                lpp->ninputs, lpp->max_centers);
        lpp->pvariance =
          reallocate_matrix(lpp->pvariance, lpp->ncenters,
                lpp->ninputs, lpp->max_centers);
    }
    for(i = 0; i < lpp->ninputs; i++){
        lpp->pcenter[lpp->ncenters][i] = input[i];
        lpp->pvariance[lpp->ncenters][i] = lpp->delta/4.;
    }
    lpp->ncenters++;
    }
    lpp->total_vectors++;
}
/*****************************************************************************
* NAME: lead_clus_cluster()
* DESCRIPTION:
*   uses Euclidean distances to determine which cluster an input
*   pattern belongs to
*   As an option, this could be done using the radii assigned to
*   the clusters, giving UNKNOWN_CLASS for patterns that do
*   not fall inside any cluster
*
* EXAMPLE:
*       
******************************************************************************/
int
lead_clus_cluster (int model, float *input)
{
    int min_i, i;
    float min_dist;
    LEAD_CLUS_PARAMS *gpp;

    gpp = leadpp[model];

    min_dist = MAXFLOAT;
    min_i = 0;
    for(i = 0; i < gpp->ncenters; i++){
    gpp->outputs[i] = l2dist(gpp->ninputs, input, &gpp->pcenter[i][0]);
    gpp->outputs[i] = (float)sqrt((double)gpp->outputs[i]);
    if(gpp->outputs[i] < min_dist){
        min_dist = gpp->outputs[i];
        min_i = i;
    }
    }

    return(min_i);
}
