/***--------------------------------------------------------------------------
 * NAME:		lead_clus_train.c
 * PURPOSE:	traininc driver for lead_clus clustering algorithm
 *
 * SEE ALSO: 	lead_clus_pack.c
 * GLOBALS:	int debug;
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lead_clus/lead_clus_train.c,v 11.4 2001/08/23 20:12:23 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lead_clus/lead_clus_train.c,v $
 * CREATED: 6-Apr-92 lck                             MIT LINCOLN LABORATORY
**************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "lead_clus.h"
#include "lead_clus_pack.h"

#define MAXSTR (256)
/* globals */
int             debug;		/* enables debug messages if debug > 0*/
extern int      stop_flag;	/* defined in main_subs.c */

/* static variables */
static char     prog_name[MAXSTR];
static int *model_list;

void
lead_clus_train(int nraw, int ninputs, int *features, int noutputs,
		int normalize, int npatterns, int debg, int verbosity,
		int random_flag, char *data_path, char *data_fname,
		char *norm_fname, char *exp_path, char *param_fname,
		char *comment, int adjust_priors, int priors_npatterns,
		float *priors_list, float *deltas, int cluster_by_class)
{
    int		model, normalization=0, data_model;
    int		current_class, class;
    int		n_trials;
    float	*inputs;
    char	*tmp_path;
    FILE	*param_file;
    int		total_ncenters;
    
    debug = debg;		/* set globals */
    strcpy(prog_name, get_program_name());
    
    model_list = allocate_ivector(noutputs);
    /* allocate and initialize model structure */
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    current_class = 0;
    do{
	model_list[current_class] = lead_clus_build(ninputs, 0);
	model = model_list[current_class];
	lead_clus_init(model, features, deltas[current_class]);
	current_class++;
    } while (cluster_by_class && current_class < noutputs);

    model = model_list[0];

    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    

    model = model_list[0];
    if(adjust_priors || random_flag){
	data_model = class_data_read(data_path, data_fname, npatterns,
				     nraw, ninputs, noutputs, features,
				     normalization, verbosity);
	if(random_flag) class_data_random_init(data_model);
	if(adjust_priors)
	  class_data_priors_init(data_model, priors_list, priors_npatterns);
	n_trials = class_data_get_npatterns(data_model);
    } else {
	data_model = class_data_single_init(data_path, data_fname, npatterns,
					    nraw, ninputs, noutputs, features,
					    normalization, verbosity);
    }
    while((inputs = class_data_get_next(data_model, &class)) != NULL){
	/* train the model for this class, single pass over the sampled
	   or random data */
	if(cluster_by_class){
	    lead_clus_inc_train(model_list[class], inputs);
	} else {
	    lead_clus_inc_train(model, inputs);
	}
    }
    class_data_exit(data_model);

    if(verbosity > 0){
	if(cluster_by_class){
	    printf("CLASS NCENTERS\n");
	    total_ncenters = 0;
	    for(class = 0; class < noutputs; class++){
		printf("%d\t%d\n", class,
		       *(int *)lead_clus_param(model_list[class],
					       "ncenters"));
		total_ncenters +=
		  *(int *)lead_clus_param(model_list[class], "ncenters");
	    }
	    printf("--------------\n");
	    printf("TOTAL\t%d\n", total_ncenters);
	} else {
	    printf("NCENTERS STORED: %d\n",
		   *(int *)lead_clus_param(model_list[0], "ncenters"));
	}
    } 

    /* write out model or models */
    tmp_path = path_build(exp_path, param_fname);
    /* store model in param file */
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    
    if(cluster_by_class){
	lead_clus_write_model_list(model_list, noutputs, param_file);
    } else {
	lead_clus_write_model(model_list[0], noutputs, param_file);
    }
    
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    norm_free();
    /* free model structures */
    current_class = 0;
    if(cluster_by_class){
	for(current_class = 0; current_class < noutputs; current_class++)
	  lead_clus_free(model_list[current_class]);
    } else {
	lead_clus_free(model_list[0]);
    }
}
