/***--------------------------------------------------------------------------
 * NAME:		lvq_cross_valid.c
 * PURPOSE:	driver cross validation using Learning Vector Quantizer
 *		Classifier
 *
 * SEE ALSO: lvq_pack.c
 *
 * DESCRIPTION: 
 *	-- description of algorithm in lvq_pack.c
 * 
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lvq/lvq_cv.c,v 11.3 2001/08/23 20:12:26 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lvq/lvq_cv.c,v $
 * CREATED: 6-May-92 lck				MIT LINCOLN LABORATORY
 ***************************************************************************/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include <nnlib.h>
#include "lvq.h"
#include "lvq_pack.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

float
lvq_cross_valid(int nraw, int ninputs, int *pfeatures, int noutputs,
		int npatterns, int debg, int verbosity, int verror,
		int nepochs, int random_flag, int cv_nfolds, int cv_nsplits,
		int use_cv_splits, int *cv_splits, char **cv_train_mask,
		char **cv_test_mask, char *data_path, char *data_fname,
		char *exp_path, char *error_fname, char *param_fname,
		char **labels, char *comment, int adjust_train,
		int priors_npatterns, float *train_priors_list,
		int cv_random, char *cluster_param_fname, int lvq_type,
		float alpha, float window, float epsilon)
{
    int		model, err_model, err_part_model, data_model;
    int		normalization = 0, epoch;
    int		desired_class, class;
    int		n_read, n_seen, n_trials, trial_num;
    int		map_flag = 0;	/* needed for error reporting routines */
    int		fold;
    int		ncenters, total_trials, *features;
    float	**pcenters, talpha;
    float	*inputs, *outputs, percent_error;
    char	*tmp_path;
    char	clus_path[MAXSTR];
    FILE	*error_file, *param_file, *clus_file;
    int		post_sum_to_one = 0;
    
    if(cv_nfolds <= 1) return 101.;
    
    /* read the normalization from the cluster file */
    tmp_path = path_build(exp_path, cluster_param_fname);
    strncpy(clus_path, tmp_path, MAXSTR -1);
    
    clus_file = nn_fopen(clus_path, "r",
			 "for read of cluster parameters");
    
    if(verbosity > 1) {
	printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	fflush(stdout);
    }
    model = norm_read_paramfile(clus_file, &normalization);
    
    if(model < 0)
      fatal_error("lvq: read error %d %s %s",
		  model, "on normalization parameter file", clus_path);
    
    /* read first cluster to get features */
    lvq_read_cluster(clus_path, clus_file, &ninputs, noutputs, &ncenters,
		     &pcenters, pfeatures);
    model = lvq_build(ninputs, noutputs, ncenters);
    lvq_init(model, pfeatures, lvq_type, alpha, epsilon, window);
    lvq_copy_centers(model, pcenters);
    
    /* want to verify the parameters from the cluster with the ones from
       main */
    verify_model_params(model,
			ninputs, *(int *) lvq_param(model, "ninputs"),
			pfeatures, *(int **) lvq_param(model, "features"));
    
    features = *(int **) lvq_param(model, "features");
    ninputs = *(int *) lvq_param(model, "ninputs");
    
    if (verror > 0) {		/* open file for classification errors */
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      "w", "for classification errors");
    } else {
      error_file = NULL;
    }
    
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, NO_COST, 0, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);
    
    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		       cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_train)
      class_data_priors_init(data_model, train_priors_list, priors_npatterns);
    
    /* write normalization data */
    if(param_fname != NULL){
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "w", "for write of lvq model file");
	fprintf(param_file, "%s\n", comment);
	fprintf(param_file, "normalization data\n");
	norm_write_paramfile(param_file);
    }
    
    for(fold = 0; fold < cv_nfolds; fold++){
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	class_data_start_cv_fold(data_model, fold, TRUE);
	n_trials = class_data_get_npatterns(data_model);
	
	/* read in the next set of cluster centers */
	if(fold != 0){
	    lvq_free(model);
	    lvq_read_cluster(clus_path, clus_file, &ninputs, noutputs,
			     &ncenters, &pcenters, pfeatures);
	    model = lvq_build(ninputs, noutputs, ncenters);
	    lvq_init(model, pfeatures, lvq_type, alpha, epsilon, window);
	    lvq_copy_centers(model, pcenters);
	}
	outputs = *(float **) lvq_param(model, "outputs");

	
	/* label the centers */
	for(n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    lvq_inc_label(model, inputs, desired_class);
	}
	lvq_label_update(model);

	/* set total_trials for talpha */
	n_trials = class_data_get_npatterns(data_model);
	total_trials = nepochs*n_trials;

	report_init(err_part_model, map_flag, ninputs, NO_COST,
		    0, post_sum_to_one);
	report_epoch_header(err_part_model, stdout, n_trials);
	
	/* do actual training for this fold */
	for(epoch = 0; epoch < nepochs; epoch++) {
	    class_data_epoch_init(data_model);
	    /* train */
	    /* go over data once for this epoch */
	    for(n_read = n_seen = 0; n_read < n_trials; n_read++){
		if((inputs = class_data_get_next(data_model, &desired_class))
		   == NULL) break;
		talpha = (alpha*(total_trials - (epoch*n_trials + n_seen))
			  )/ total_trials;
		n_seen++;
		class = lvq_inc_ctrain(model, inputs, desired_class, talpha);
		report_class_trial(err_part_model, error_file, inputs,
				   outputs, desired_class, class,
				   class_data_get_input_num(data_model));
	    }
	    report_epoch(err_part_model, stdout, epoch);
	}
	
	/* finish the training fold and start the testing fold */
	class_data_end_cv_fold(data_model);
	class_data_start_cv_fold(data_model, fold, FALSE);
	n_trials = class_data_get_npatterns(data_model);
	
	/* print out lvq model */
	if(param_fname != NULL){
	    fprintf(param_file, "lvq model %d\n", fold);
	    lvq_write_model(model, param_file);
	}
	
	report_init(err_part_model, map_flag, ninputs, NO_COST,
		    0, post_sum_to_one);
	
	for (n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    class = lvq_classify(model, inputs);
	    trial_num = class_data_get_input_num(data_model);
	    report_class_trial(err_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	    report_class_trial(err_part_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	}
	/* report on this fold */
	report_overall(err_part_model, stdout, error_fname, labels);

	/* finish the testing fold */
	class_data_end_cv_fold(data_model);
    } /* end fold */
    
    /* close files */
    if(param_fname != NULL){
	nn_fclose(param_file, tmp_path, "for write of lvq model file");
	if (verbosity > 0)
	  printf("\nFinished -- lvq model saved in \"%s\"\n",tmp_path);
    }
    nn_fclose(clus_file, clus_path, "for read of cluster parameters");

    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    
    class_data_exit(data_model);
    lvq_free(model);
    norm_free();

    return (percent_error);
}
