/***---------------------------------------------------------------------------
 * NAME:	lvq_pack.c
 * PURPOSE:	Functions specific to lvq.c
 *
 * SEE ALSO: lvq.c lvq_subs.c
 * GLOBALS:
 *
 * DESCRIPTION:
 *	Functions in this file
 *	    lvq_build()	Allocate storage for model parameters.
 *	    lvq_init()		Initialize model parameters.
 *	    lvq_free()		Deallocates model storage
 *	    lvq_read_model()	Initializes a lvq model from file
 *	    lvq_write_model()	Writes a lvq model to output file
 *	    lvq_param()		Returns pointers to elements in model
 *	    lvq_inc_ctrain()	does training on a vector
 *	    lvq_classify()	Implements lvq to return an integer class
 *
 * ALGORITHM:
 *	Training:
 *	    an initial set of centers is read in
 *	    on each call of lvq_inc_ctrain the 1 or 2 nearest centers (m)
 *		are found and there positions are updated:
 *		m(t+1) = m(t) + s*alpha(t)*(inputs - m(t))
 *		where s = 1 if m is of inputs class and s = -1 otherwise
 *		alpha(t) is usually alpha(0)*(ntrials - t)/ntrials
 *		(linearly decreasing from alpha(0) to 0 over ntrials)
 *		(this must be sent set by the caller to lvq_inc_ctrain)
 *
 *	    lvq1: update only nearest neighbor
 *	    olvq1: as lvq1 except that there is an alpha for each center.
 *		the alpha for the nearest center to inputs is updated as:
 *		alpha(t+1) = alpha(t)/(1+s*alpha(t))
 *		but alpha may never go above alpha(0)
 *	    lvq2: update 2 nearest neighbors IF
 *		one is the correct class and the other isn't AND
 *		the point falls in a certain window between them:
 *		  distance(inputs, nearest)/distance(inputs, 2nd nearest)
 *			< (1 - window)/(1 + window)
 *	    lvq3: update 2 nearest neighbors as you would in lvq2
 *		if 2 nearest are both the correct class, update both:
 *		m(t+1) = m(t) + epsilon*alpha(t)*(inputs - m(t))
 *	
 *	Testing:  For any test vector: find the nearest center.
 *		return the class of that center
 *		
 * REFERENCE: 
 *	Kohonen, Kangas, Laaksonen, Torkkola
 *	"LVQ_PAK The Learning Vector Quantization Program Package V2.0"
 *	Copyright 1992.
 *	A document available by ftp on chochlea.hut.fi
 *		(130.233.168.48) on the internet
 *	ftp as anonymous.  cd /pub/lvq_pak. get document.ps
 *
 *	There may be a better reference from IEEE somewhere.  This is the
 *	one I used.
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lvq/lvq_pack.c,v 11.3 2001/08/23 20:12:27 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lvq/lvq_pack.c,v $
 * CREATED: 6-May-92 lck				MIT LINCOLN LABORATORY
 --------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define lvq_pack_h_internal_types
#include "lvq_pack.h"

#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))
#define SPACE 1024

/* lvq_types */
#define OLVQ 0
#define LVQ1 1
#define LVQ2 2
#define LVQ3 3

/* Internal Globals */
static LVQ_PARAMS **lvqpp;	/* array of parameters for each model */
static int max_models = 0;	/* current length of lvqpp */

/*****************************************************************************
 * NAME: lvq_build(ninputs, nclasses, ncenters)
 *	ninputs	- number of dimensions in training vectors
 *	nclasses	- number of classes the data falls into
 *	ncenters	- number of vectors to be stored
 *
 * DESCRIPTION:
 *	  Allocate storage for model parameters.
 *
 * EXAMPLE:
 *
 *****************************************************************************/
int
lvq_build(int ninputs, int nclasses, int ncenters)
{
    register int new_model;
    LVQ_PARAMS *lpp;
    
    /* find entry in model array for new model */
    for(new_model = 0; new_model < max_models; new_model++)
      if(lvqpp[new_model] == NULL) break;
    
    /* if there were no empty entries in the model array */
    if(new_model == max_models){
	/* if the model array is of 0 length */
	if(max_models == 0){
	    lvqpp = (LVQ_PARAMS **)allocate_pvector(MODEL_STORE_SIZE);
	    max_models = MODEL_STORE_SIZE;
	} else {
	    lvqpp = (LVQ_PARAMS **)
	      reallocate_pvector((void **)lvqpp, max_models,
				 max_models+MODEL_STORE_SIZE);
	    max_models += MODEL_STORE_SIZE;
	}
    }
    
    /* allocate space for model */
    lvqpp[new_model] = (LVQ_PARAMS *)
      nn_malloc(sizeof(LVQ_PARAMS), "for lvq parameter structure");
    lpp = lvqpp[new_model];
    
    /* set fields which are array sizes */
    lpp->ninputs = ninputs;
    lpp->nclasses = nclasses;
    lpp->ncenters = ncenters;    
    
    /* allocate space for vectors */
    lpp->features = allocate_ivector(ninputs+1);
    lpp->outputs = allocate_vector(nclasses);
    lpp->classes = allocate_ivector(ncenters);
    lpp->centers = allocate_matrix(ncenters, ninputs);
    lpp->class_vote = allocate_imatrix(ncenters, nclasses);
    lpp->alphas = allocate_vector(ncenters);
    
    return new_model;
}
/*****************************************************************************
 * NAME: lvq_init()
 * DESCRIPTION:
 *	  Initialize model parameters not set in build
 *
 * EXAMPLE:
 *
 *****************************************************************************/
void
lvq_init(int model, int *nn_features, int lvq_type, float alpha0,
	      float epsilon, float window)
{
    int		i;
    LVQ_PARAMS *lpp;
    
    lpp = lvqpp[model];

    /* copy feature data into feature list */
    for(i = 0; i <= nn_features[0]; i++)
      lpp->features[i] = nn_features[i];

    lpp->lvq_type = lvq_type;
    switch(lvq_type){
      case OLVQ:
      case LVQ1:
      case LVQ2:
      case LVQ3:
	break;
      default:
	fatal_error("lvq_init: bad lvq_type (%d)", lvq_type);
    }

    if(alpha0 > 1.){
	fatal_error("lvq_init: initial stepsize, alpha (%f) greater than 1.");
    } else lpp->init_alpha = alpha0;
    if(lvq_type == LVQ2 || lvq_type == LVQ3){
	if(window >= 1. || window <= -1.){
	    fatal_error("lvq_init: window (%f) not in range [-1.,1.]");
	}
	lpp->window = (1.-window)/(1.+window);
    }
    if(lvq_type == LVQ3)
      lpp->epsilon = epsilon;	/* I guess I'll just trust these */

    if(lvq_type == OLVQ){
	for(i = 0; i < lpp->ncenters; i++){
	    lpp->alphas[i] = alpha0;
	}
    }
}
/*****************************************************************************
 * NAME: lvq_free()
 * DESCRIPTION:
 *	  Deallocate storage used in a model
 *
 * EXAMPLE:
 *
 *****************************************************************************/
void
lvq_free(int model)
{
    LVQ_PARAMS *lpp;
    
    lpp = lvqpp[model];
    /* free vectors and matrices */
    deallocate_ivector(lpp->classes);
    deallocate_ivector(lpp->features);
    deallocate_vector(lpp->outputs);
    deallocate_matrix(lpp->centers);
    deallocate_imatrix(lpp->class_vote);
    deallocate_vector(lpp->alphas);
    /* free structure itself */
    nn_free((char *) lpp, "for lvq parameter structure");
    lvqpp[model] = NULL;
}


/*****************************************************************************
 * NAME: lvq_read_model()
 * DESCRIPTION:
 *	  Loads parameters from ascii input file
 *	allocates space for and loads data into vectors and matricies
 *
 *	On error from param_getblk, returns the error value
 *	(there is already an error check on the allocation, so we don't
 *	have to do that here)
 *
 * EXAMPLE:
 *****************************************************************************/
int
lvq_read_model(FILE *fp)
{
    int		model;
    int		rv, ninputs, nclasses, ncenters;
    float	alpha0;
    LVQ_PARAMS *lpp;
    
    /* read in scalars from model file */
    param_setvarptr("ncenters", &ncenters, paramtab, TABSIZE);
    param_setvarptr("ninputs", &ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &nclasses, paramtab, TABSIZE);
    param_setvarptr("alpha0", (int *)&alpha0, paramtab, TABSIZE);

    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
      return rv;

    /* allocate space for model and vectors */
    model = lvq_build(ninputs, nclasses, ncenters);
    lpp = lvqpp[model];

    /* read in vectors from model file */
    param_setvarptr("classes", &lpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim("classes", lpp->ncenters, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("centers", (int *)&lpp->centers[0][0], paramtab_l, TABSIZE_L);
    param_setdim("centers", lpp->ncenters, lpp->ninputs,
		 paramtab_l, TABSIZE_L);
    param_setvarptr("features", &lpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", lpp->ninputs+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("alphas",(int *) &lpp->alphas[0], paramtab_l, TABSIZE_L);
    param_setdim("alphas", lpp->ncenters, 0, paramtab_l, TABSIZE_L);
    
    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
      return rv;
    
    /* set init variables to something reasonable */
    lpp->lvq_type = LVQ1;
    lpp->init_alpha = .3;
    lpp->epsilon = .2;
    lpp->window = .8; /* (1 - .25)/(1 + .25) */
    return (model);
}

/*****************************************************************************
 * NAME: lvq_write_model()
 * DESCRIPTION:
 *	writes parameters and centers for a model to a file
 *	in ASCII format
 *
 * EXAMPLE:
 *
 *****************************************************************************/
int
lvq_write_model(int model, FILE *fp)
{
    int		rv;
    LVQ_PARAMS *lpp;	/* temporary pointer to lvqpp[] */
    
    lpp = lvqpp[model];
    
    /* save scalars */
    param_setvarptr("ncenters", &lpp->ncenters, paramtab, TABSIZE);
    param_setvarptr("ninputs", &lpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("nclasses", &lpp->nclasses, paramtab, TABSIZE);
    param_setvarptr("alpha0", (int *)&lpp->init_alpha, paramtab, TABSIZE);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
      return rv;
    
    /* save vectors and matrices */
    param_setvarptr("classes", &lpp->classes[0], paramtab_l, TABSIZE_L);
    param_setdim("classes", lpp->ncenters, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("centers", (int *)&lpp->centers[0][0], paramtab_l, TABSIZE_L);
    param_setdim("centers", lpp->ncenters, lpp->ninputs,
		 paramtab_l, TABSIZE_L);
    param_setvarptr("features", &lpp->features[0], paramtab_l, TABSIZE_L);
    param_setdim("features", lpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("alphas", (int *)&lpp->alphas[0], paramtab_l, TABSIZE_L);
    param_setdim("alphas", lpp->ncenters, 0, paramtab_l, TABSIZE_L);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
      return rv;
    
    return (0);
}

/************************************************************************
 * NAME: lvq_param(model, param_name)
 * DESCRIPTION:
 *	Takes model number and parameter name.
 *	Returns a pointer to the named parameter.
 *
 *************************************************************************/

void *
lvq_param(int model, char *param_name)
{
    void *param_address;
    LVQ_PARAMS *lpp;
    
    lpp = lvqpp[model];
    if (PEQ("ninputs"))
      param_address = &lpp->ninputs;
    else if (PEQ("nclasses"))
      param_address = &lpp->nclasses;
    else if (PEQ("ncenters"))
      param_address = &lpp->ncenters;
    else if (PEQ("alpha"))
      param_address = &lpp->init_alpha;
    else if (PEQ("epsilon"))
      param_address = &lpp->epsilon;
    else if (PEQ("window"))
      param_address = &lpp->window;
    else if (PEQ("centers"))
      param_address = &lpp->centers;
    else if (PEQ("classes"))
      param_address = &lpp->classes;
    else if (PEQ("class_vote"))
      param_address = &lpp->class_vote;
    else if (PEQ("features"))
      param_address = &lpp->features;
    else if (PEQ("outputs"))
      param_address = &lpp->outputs;
    else if (PEQ("alphas"))
      param_address = &lpp->alphas;
    else {
	warning("lvq_param: Parameter %s not defined", param_name);
	param_address = NULL;
    }
    return param_address;
} /* END lvq_param() */
/*****************************************************************************
 * NAME: lvq_inc_ctrain()
 * DESCRIPTION:
 *	  assigns a class to the input vector.
 *	If the class assigned is wrong, the vector is added to the
 *		exemplar set
 *	returns the class selected
 * EXAMPLE:
 *
 *****************************************************************************/
int lvq_inc_ctrain(int model, float *inputs, int desired_class, float talpha)
{
    int i, two_near[2];
    int class;
    LVQ_PARAMS *lpp;
    float two_dists[2], diff, sign;

    lpp = lvqpp[model];

    switch(lpp->lvq_type){
      case OLVQ:
      case LVQ1:
	if(lpp->ncenters < 1) return -1;
	break;
      case LVQ2:
      case LVQ3:
	if(lpp->ncenters < 2) return -1;
	break;
      default:
	return -1;
    }

    /* find two nearest neighbors */
    find_k_nearest_neighbors(inputs, lpp->centers, two_dists, two_near,
			     lpp->ninputs, lpp->ncenters, 2);

    /* class is the class of the nearest neighbor */
    if(two_near[0] == -1){
	/* centers have been swept off the map */
	return -1;
    }
    class = lpp->classes[two_near[0]];

    if(lpp->lvq_type == OLVQ || lpp->lvq_type == LVQ1){
	if(class == desired_class) sign = 1.;
	else sign = -1.;

	if (lpp->lvq_type == OLVQ){
	    talpha = lpp->alphas[two_near[0]];
	    lpp->alphas[two_near[0]] /= 1. + sign * lpp->alphas[two_near[0]];
	    if(lpp->alphas[two_near[0]] > lpp->init_alpha)
	      lpp->alphas[two_near[0]] = lpp->init_alpha;
	}

	for(i = 0; i < lpp->ninputs; i++){
	    diff = inputs[i] - lpp->centers[two_near[0]][i];
	    lpp->centers[two_near[0]][i] += sign * talpha * diff;
	}
    } else { /* LVQ2 or LVQ3 */
	if(two_near[1] == -1) return -1;
	if(lpp->classes[two_near[0]] != lpp->classes[two_near[1]]){
	    if(lpp->classes[two_near[0]] == desired_class ||
	       lpp->classes[two_near[1]] == desired_class){
		if(two_dists[0]/two_dists[1] > lpp->window){
		    /* make it so one in "nearest" is correct one */
		    if(lpp->classes[two_near[1]] == desired_class){
			i = two_near[1]; two_near[1] = two_near[0];
			two_near[0] = i;
		    }
		    /* shift two nearest neighbors */
		    for(i = 0; i < lpp->ninputs; i++){
			lpp->centers[two_near[0]][i] =
			  lpp->centers[two_near[0]][i] +
			  talpha*(inputs[i] - lpp->centers[two_near[0]][i]);
			lpp->centers[two_near[1]][i] =
			  lpp->centers[two_near[1]][i] -
			  talpha*(inputs[i] - lpp->centers[two_near[1]][i]);
		    }
		}
	    }
	} else if(lpp->lvq_type == LVQ3){
	    if(lpp->classes[two_near[0]] == desired_class){
		talpha *= lpp->epsilon;
		for(i = 0; i < lpp->ninputs; i++){
		    lpp->centers[two_near[0]][i] =
		      lpp->centers[two_near[0]][i] +
		      talpha*(inputs[i] - lpp->centers[two_near[0]][i]);
		    lpp->centers[two_near[1]][i] =
		      lpp->centers[two_near[1]][i] +
		      talpha*(inputs[i] - lpp->centers[two_near[1]][i]);
		}
	    }
	}
    }
    return class;
}

/*****************************************************************************
 * NAME: lvq_classify()
 * DESCRIPTION:
 *	  Implements the lvq to return an integer class.
 *
 * EXAMPLE:
 *
 *****************************************************************************/
int
lvq_classify(int model, float *inputs)
{
    int		min_i;
    LVQ_PARAMS *lpp;
    
    lpp = lvqpp[model];
    
    if(lpp->ncenters < 1) return -1;
    min_i = find_nearest_neighbor(inputs, lpp->centers,
				  lpp->ninputs, lpp->ncenters);
    if(min_i < 0) return -1;
    return lpp->classes[min_i];
}

/*****************************************************************************
 * NAME: lvq_copy_centers()
 * DESCRIPTION:
 *	copy centers into model center list
 * EXAMPLE:
 *
 *****************************************************************************/
void
lvq_copy_centers(int model, float **centers)
{
    int i, j;
    LVQ_PARAMS *lpp;
    
    if(centers == NULL) return;
    lpp = lvqpp[model];
    for(i=0; i < lpp->ncenters; i++)
      for(j = 0; j < lpp->ninputs; j++)
	lpp->centers[i][j] = centers[i][j];
}

/*****************************************************************************
 * NAME: lvq_inc_label(model, inputs, class)
 * DESCRIPTION:
 *	collect data for labeling centers
 * EXAMPLE:
 *
 *****************************************************************************/
void
lvq_inc_label(int model, float *inputs, int class)
{
    int min_i;
    LVQ_PARAMS *lpp;
    
    lpp = lvqpp[model];
    
    if(lpp->ncenters < 1) return;

    min_i = find_nearest_neighbor(inputs, lpp->centers, lpp->ninputs,
				  lpp->ncenters);
    if(min_i < 0 || min_i >= lpp->ncenters) return;
    lpp->class_vote[min_i][class]++;
}

/*****************************************************************************
 * NAME: lvq_label_update(model)
 * DESCRIPTION:
 *	select labels for centers
 * EXAMPLE:
 *
 *****************************************************************************/
void
lvq_label_update(int model)
{
    int i, class;
    LVQ_PARAMS *lpp;

    lpp = lvqpp[model];

    for(i = 0; i < lpp->ncenters; i++){
	class = int_max_pick(lpp->class_vote[i], lpp->nclasses);
	lpp->classes[i] = class;
    }
}

/*****************************************
lvq_read_cluster()
*****************************************/
void
lvq_read_cluster(char *param_path, FILE *param_file, int *pninputs,
		 int noutputs, int *pncenters, float ***centroids,
		 int *features)
{
    int model;

    model = cluster_read_paramfile(param_file, pninputs, &noutputs, pncenters,
				   features);
    if(model < 0)
      fatal_error("lvq: read error %d on parameter file %s", model,
		  param_path);

    if(*pncenters <= 0)
      fatal_error("lvq_read_cluster: No centers from clustering");

    /* merge the centers into a single list */
    cluster_merge_centers(model);

    *centroids = *(float ***)cluster_param(model, "pcenter", 0);
}
