/***--------------------------------------------------------------------------
 * NAME:		lvq_train.c
 * PURPOSE:	driver for Training Learning Vector Quantizer Classifier
 *
 * SEE ALSO: lvq_pack.c
 *
 * DESCRIPTION: 
 *	-- description of algorithm in lvq_pack.c
 * 
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lvq/lvq_train.c,v 11.3 2001/08/23 20:12:27 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/lvq/lvq_train.c,v $
 * CREATED: 6-May-92 lck				MIT LINCOLN LABORATORY
 *************************************************************************/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include <nnlib.h>
#include "lvq.h"
#include "lvq_pack.h"

#define MAXSTR (256)

void
lvq_train(int create_flag, int nraw, int ninputs, int *pfeatures,
	  int noutputs, int npatterns, int debg, int verbosity, int verror,
	  int nepochs, int random_flag, char *data_path, char *data_fname,
	  char *exp_path, char *error_fname, char *param_fname, char **labels,
	  char *comment, int adjust_priors, int priors_npatterns,
	  float *priors_list, char *cluster_param_fname, int lvq_type,
	  float alpha, float window, float epsilon)
{
    int		model, err_model, err_part_model, data_model;
    int		rv, normalization = 0;
    int		epoch;
    int		desired_class, class;
    int		n_read, n_trials;
    int		map_flag = 0;	/* needed for error reporting routines */
    int		total_trials, ncenters;
    int		*features;
    float	talpha, **pcenters;
    float	*inputs, *outputs;
    char	*tmp_path, *open_type;
    char	clus_path[MAXSTR];
    int		post_sum_to_one = 0;
    FILE	*error_file, *param_file, *clus_file;
    
    if(create_flag){
	/* allocate and initialize model structure */
	/* read the cluster file */
	tmp_path = path_build(exp_path, cluster_param_fname);
	strncpy(clus_path, tmp_path, MAXSTR -1);
	
	clus_file = nn_fopen(clus_path, "r",
			     "for read of cluster parameters");
	
	if(verbosity > 1) {
	    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	    fflush(stdout);
	}
	model = norm_read_paramfile(clus_file, &normalization);
	
	if(model < 0)
	  fatal_error("lvq: read error %d %s %s",
		      model, "on normalization parameter file", clus_path);
	
	lvq_read_cluster(clus_path, clus_file, &ninputs, noutputs, &ncenters,
			 &pcenters, pfeatures);
	nn_fclose(clus_file, clus_path, "for read of cluster parameters");
	
	model = lvq_build(ninputs, noutputs, ncenters);
	lvq_init(model, pfeatures, lvq_type, alpha, epsilon, window);
	lvq_copy_centers(model, pcenters);
    } else {			/* continue training */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of lvq model file");
	rv = norm_read_paramfile(param_file, &normalization);
	if(rv < 0)
	  fatal_error ("%s: read error %d %s %s", "lvq", rv,
		       "on normalization parameter file", tmp_path);
	model = lvq_read_model(param_file);
	if(model < 0)
	  fatal_error("%s: error reading model file %s", "lvq", tmp_path);
	nn_fclose(param_file, tmp_path, "for read of lvq model file");
	
	verify_model_params(model,
			    ninputs, *(int *) lvq_param(model, "ninputs"),
			    pfeatures, *(int **) lvq_param(model, "features"));
    }
    
    features = *(int **) lvq_param(model, "features");
    ninputs = *(int *) lvq_param(model, "ninputs");
    outputs = *(float **) lvq_param(model, "outputs");
    
    if (verror > 0) {		/* open file for classification errors */
	if (create_flag) open_type = "w";
	else open_type = "a";
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      open_type, "for classification errors");
    }
    
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, NO_COST, 0, post_sum_to_one);
    err_part_model = report_build (verbosity, 0, noutputs);
    
    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
      class_data_priors_init(data_model, priors_list, priors_npatterns);
    
    /* get the number of patterns per epoch */
    n_trials = class_data_get_npatterns(data_model);
    report_epoch_header(err_model, stdout, n_trials);

    if(create_flag){
	/* label the centers */
	for(n_read = 0; n_read < n_trials; n_read++){
	    /* get the next pattern */
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    lvq_inc_label(model, inputs, desired_class);
	}
	lvq_label_update(model);
    }

    total_trials = nepochs * n_trials;
    for (epoch = 0; epoch < nepochs; epoch++) {
	class_data_epoch_init(data_model);
	report_init (err_part_model, map_flag, ninputs, NO_COST, 0,
		     post_sum_to_one);

	for(n_read = 0; n_read < n_trials; n_read++){
	    talpha = (alpha*(total_trials -
			     (epoch*n_trials + n_read)))/ total_trials;
	    /* get the next pattern */
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    
	    class = lvq_inc_ctrain(model, inputs, desired_class, talpha);
	    report_class_trial(err_model, error_file,
			       inputs, outputs, desired_class, class,
			       class_data_get_input_num(data_model));
	    report_class_trial(err_part_model, error_file,
			       inputs, outputs, desired_class, class,
			       class_data_get_input_num(data_model));
	}
	report_epoch(err_model, stdout, epoch);
    }
    report_last_epoch(err_part_model, stdout, epoch-1);
    class_data_exit(data_model);
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of lvq model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "lvq model\n");
    lvq_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of lvq model file");
    if (verbosity > 0)
      printf("\nFinished -- lvq model saved in \"%s\"\n",tmp_path);
    
    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if(verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    
    /* free space and close files */
    norm_free();
    lvq_free(model);
}
