/***------------------------------------------------------------------------
 * NAME:		mlp_cv.c
 * PURPOSE:	cross validation driver for MultiLayer Perceptron
 *
 * SEE ALSO: 	mlp_pack.c
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_cv.c,v 11.6 2001/08/23 20:12:31 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_cv.c,v $
 * CREATED: 21-SEP-90 dan                             MIT LINCOLN LABORATORY
 ***************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlp.h"

static char prog_name[MAXSTR];	/* static variables */

float
mlp_cross_valid(int map_flag, int nraw, int ninputs, int *features,
		int noutputs, int normalize, int npatterns, int verbosity,
		int verror, int nepochs, int random_flag,
		int cv_nfolds, int cv_nsplits, int use_cv_splits,
		int *cv_splits, char **cv_train_mask, char **cv_test_mask,
		char *data_path, char *data_fname, char *norm_fname,
		char *exp_path, char *error_fname, char *param_fname,
		char **labels, char *comment, int report_cost,
		int roc_target, float roc_lower, float roc_upper,
		int adjust_train, int priors_npatterns,
		float *train_priors_list, int adjust_test,
		float *test_priors_list, int cv_random,
		float *etta_list, float alpha, float tolerance, float decay,
		float kappa, float epsilon, float *sigmoid_param_list,
		int *nnodes, int *batch, int cost_type, float cost_param,
		int etta_type, int hfunction, int ofunction,
		float max_magnitude, int log_outputs, float etta_nepochs)
{
    /* Local variables */
    int		model, err_model, err_part_model, data_model;
    int		normalization=0, epoch, etta_tau;
    int		desired_class, class;
    int		n_read, i;
    int		fold, n_trials, trial_num;
    float	*inputs, *outputs, percent_error, *desired_outputs;
    char	*tmp_path;
    FILE	*error_file, *param_file;
    int		post_sum_to_one = 0;
    
    int		nlayers;	/* number of layers */
    int		layers[20];	/* number of nodes in each layer */
    int		w_update_cnt = 0;
    int		w_update_int, w_update_max, w_update_increase;
    
    strcpy(prog_name, get_program_name());
    
    w_update_int = batch[1];
    w_update_max = batch[2];
    w_update_increase = batch[3];
    nlayers = nnodes[0];	/* Put the number of nodes in each layer in
				 * nlayers */
    for (i = 0; i < nlayers; i++)
      layers[i] = nnodes[i + 1];
    
    /* allocate and initialize model structure */
    model = mlp_build(nlayers, layers, etta_type);

    etta_tau = (int)((float)npatterns * etta_nepochs);
    mlp_init(model, max_magnitude, hfunction, ofunction,
	     etta_list, features, etta_tau);
    mlp_set_train_params(model, cost_type, cost_param, etta_list, alpha, decay,
			 kappa, epsilon, tolerance, sigmoid_param_list);

    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    outputs = *(float **) mlp_param(model, "output", 0);
    desired_outputs = allocate_vector(noutputs);
    
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {	/* open file for classification errors */
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      "w", "for classification errors");
    } else error_file = NULL;
    /* do remaining initializations and allocations */
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, map_flag, ninputs, report_cost,
		log_outputs, post_sum_to_one);
    report_roc_init(err_model, roc_target, roc_lower, roc_upper);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    /* if cv_random is set, this should be done before the mlp random
       weights are set */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		       cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_train)
      class_data_priors_init(data_model, train_priors_list, priors_npatterns);

    /* store the normalization once only */
    if(param_fname != NULL){
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "w", "for write of model file");
	fprintf(param_file, "%s\n", comment);
	fprintf(param_file, "normalization data\n");
	norm_write_paramfile(param_file);
    }
    
    for(fold = 0; fold < cv_nfolds; fold++) {
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	  
	class_data_start_cv_fold(data_model, fold, TRUE);
	n_trials = class_data_get_npatterns(data_model);

	etta_tau = (int)((float)n_trials * etta_nepochs);
	mlp_init(model, max_magnitude, hfunction,
		 ofunction, etta_list, features, etta_tau);
	mlp_set_train_params(model, cost_type, cost_param, etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
	
	/* verror for training is 0 */
	report_init(err_part_model, map_flag, ninputs, report_cost,
		    log_outputs, post_sum_to_one);
	report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);
	report_epoch_header(err_part_model, stdout, n_trials);

	for (epoch = 1; epoch <= nepochs; epoch++) {
	    class_data_epoch_init(data_model);

	    /* go through the data for this fold */
	    w_update_cnt = 0;
	    for (n_read = 0; n_read < n_trials; n_read++) {
		if((inputs = class_data_get_next(data_model, &desired_class))
		   == NULL) break;
		if (w_update_int <= 1)
		  class = mlp_inc_ctrain(model, inputs, desired_class);
		else {
		    class = mlp_batch_ctrain(model, inputs, desired_class);
		    if (++w_update_cnt > w_update_int) {
			mlp_batch_update_weights(model);
			w_update_cnt = 0;
		    }
		}
		report_class_trial(err_part_model, error_file, inputs,
				   outputs, desired_class, class,
				   class_data_get_input_num(data_model));
		if(cost_type == ROC_COST){
		    desired_outputs[desired_class] = 1.;
		    roc_cost_train(outputs, desired_outputs);
		    desired_outputs[desired_class] = 0.;
		}
	    } /* end reading data */
	    if (w_update_cnt > 0)
	      mlp_batch_update_weights(model);
	    if (w_update_int < w_update_max)
	      w_update_int += w_update_increase;
	    report_epoch(err_part_model, stdout, epoch);
	    report_init(err_part_model, map_flag, ninputs, report_cost,
			log_outputs, post_sum_to_one);
	    report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);
	    if(cost_type == ROC_COST)
	      roc_cost_finish_epoch();
	} /* end epoch */
	
	/* print out the mlp model */
	if(param_fname != NULL){
	    fprintf(param_file, "mlp model %d\n", fold);
	    mlp_write_model(model, param_file);
	    if ( verbosity > 0)
	      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
	}
	
	/* reinitialize model for testing */
	if(adjust_test)
	  mlp_test_init(model, adjust_test, test_priors_list);
	/* this hack is normally done in mlp_read_model,
	   which we are not calling */
	if(cost_type == PC_PROC){
	  int nlayers = *(int *)mlp_param(model, "nlayers", 0);
	  int *sigm = (int *)mlp_param(model, "sigm_type", nlayers-1);
	  *sigm = LINEAR;
	}
	report_init(err_part_model, map_flag, ninputs, report_cost,
		    log_outputs, post_sum_to_one);
	report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);
	
	/* finish the training fold and start the testing fold */
	class_data_end_cv_fold(data_model);
	class_data_start_cv_fold(data_model, fold, FALSE);
	n_trials = class_data_get_npatterns(data_model);

	for (n_read = 0; n_read < n_trials; n_read++) {
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    class = mlp_classify(model, inputs);
	    trial_num = class_data_get_input_num(data_model);
	    report_class_trial(err_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);

	    report_class_trial(err_part_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	}
	report_overall(err_part_model, stdout, error_fname, labels);
	
	/* finish the testing fold */
	class_data_end_cv_fold(data_model);
    } /* end fold */
    
    /* finish up writing the models */
    if(param_fname != NULL){
	nn_fclose(param_file, tmp_path, "for write of model file");
	if ( verbosity > 0)
	  printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    }
    class_data_exit(data_model);
    
    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);
    
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    mlp_free(model);
    norm_free();
    deallocate_vector(desired_outputs);
    return(percent_error); /* needed for feature search */
}
