/***---------------------------------------------------------------------------
* NAME:     mlp_feat_sel_main.c
* PURPOSE:  argument parsing and preparation for mlp based
*       feature search driver
*
* USAGE:    mlp_feat_sel -<flags>
* EXAMPLE:  mlp_feat_sel --     * shows all available flags
*
* SEE ALSO: mlp_feat_sel.c mlp_pack.c
*
* GLOBALS: <flags> defined in mlp_feat_sel_args.h
*
* DESCRIPTION:  entry routine for Feature Selection using
*       mlp classifier
*
* RETURNS:
*   creates a parameter file with an ordered list of features
*       and the error rate when that feature was introduced
*   depending on the verbosity, information is printed on stdout:
*   verbosity 0) best feature set found in search and its error rate
*         1) all level 0 printouts plus best feature set found for
*           each number of features and their error rates
*           also, the command line used
*         2) all level 1 printouts plus each feature set tried
*           and their error rates
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_feat_sel_main.c,v 11.6 2001/08/23 20:12:31 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_feat_sel_main.c,v $
* CREATED: 21-Mar-94 lck             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define NO_CV_TABLE
#include <cv_args.h>
#include <gen_class_args.h>
#include "mlp_args.h"
#include "mlp_feat_sel.h"
#include <gen_search_args.h>

int
main(int argc, char **argv)
{
    char *comment;
    int use_splits, i;
    /* arguments for reading defaults file */
    DEFAULTS_FILE_STRUCT *df_args;
    /* argument table with all entries */
    ARG *argtab, *argtab2;
    int tabsize, tabsize2;

    /******************* read argument list **********************/
    init_nnsubs(argc, argv, NULL);
    nnargs_combine_argtabs(gen_tab, GEN_TABSIZE, mlp_tab, MLP_TABSIZE,
               &argtab, &tabsize);
    nnargs_combine_argtabs(argtab, tabsize, search_tab, SEARCH_TABSIZE,
               &argtab2, &tabsize2);
    argc = nnargs(argc, argv, argtab2, tabsize2);

    print_log_header();

    /************* read defaults and set input data parameters ********/
    if(strcmp(desc_fname, "") == 0){
    fatal_error("%s: -fdescribe must be set", argv[0]);
    }
    df_args = (DEFAULTS_FILE_STRUCT *)
      nn_malloc(sizeof(DEFAULTS_FILE_STRUCT), "for the defaults file struct");
    read_defaults_file(path_build(data_path, desc_fname), df_args);
    if(df_args->noutputs == 0)
      fatal_error("noutputs in %s must be greater than 0", desc_fname);
    if(df_args->ninputs == 0)
      fatal_error("ninputs in %s must be greater than 0", desc_fname);

    /* set the number of patterns to no more than the number of lines
       in the data file */
    if(strcmp(data_fname, "") == 0){ /* data_fname is unset */
    fatal_error("%s: -finput must be set", argv[0]);
    }
    check_file_length(path_build(data_path, data_fname), &npatterns);
    if(npatterns == 0)
      fatal_error("%s: Empty data file %s", argv[0],
          path_build(data_path, data_fname));

    if (df_args->db_type != CLASS_DATA)
      fatal_error("class data required, check defaults file");

    /*************** zero out feature selection arguments ***********/
    features[0] = 0;
    nnodes[1] = df_args->ninputs;
    nnodes[nnodes[0]] = df_args->noutputs;
    nnodes[1] = ninputs = fix_features(df_args->ninputs, 0, features);

    /******** zero out other output arguments for cv driver ***********/
    param_fname = NULL;
    verror = 0;
    verbosity = -1;

    /************** set up cross validation arguments ***************/
    if(cross_valid < 2)
      fatal_error("%s: -cross_valid must be >= 2", argv[0]);
    cross_valid_from_file = 0; /* no cv_from_file here */
    use_splits =
      fix_cross_valid_args(&cross_valid, &cv_nfolds, &cv_nsplits, cv_splits,
               cv_train_mask, cv_test_mask, cross_valid_from_file,
               path_build(data_path, cross_valid_fname));

    /********* fix up priors adjustment arguments *********************/
    fix_priors(adjust_train, train_uniform_priors, &train_priors_list,
           adjust_test, test_uniform_priors, &test_priors_list,
           train_flag, cross_valid, df_args->noutputs);

    /****************** check search arguments ********************/
    if(search_n > df_args->ninputs){
    warning("%s: %s\nDoing full search", argv[0],
        "Searching for more features than there are.");
    stop_early = 0;
    }
    /* if search_n is 0, feat_sel does the whole search */
    if(!stop_early) search_n = 0;

    /********** check MLP args *****************/
    /* check the ofunction and cost_func flags */
    if(cost_func == MAX_LIKE || cost_func == CROSS_ENTROPY)
      if(ofunction != 0){
      warning("%s %d %s %d.\n%s %d",
          "mlp_main: sigmoid function type", ofunction,
          "is not allowed with cost_func", cost_func,
          "changing the output sigmoid to type", SQ_ERR);
      ofunction = SQ_ERR;
      }
    
    /* work out the etta_list flags.  by the end, etta_list should be
       filled either with what the user asked for or with etta on each level
       */
    /* I would prefer to do this without a goto,
       but this seemed a little cleaner for now */
    if(etta_list != NULL){
    if(use_etta_list){
        if(((int)etta_list[0]) < nnodes[0]-1){
        warning("etta_list is too short. Using etta for all layers");
        } else goto etta_list_ok;  /* etta list is all right */
    }
    /* we need to free the old etta_list so we can build the new one */
    deallocate_vector(etta_list);
    } else {
    if(use_etta_list)
      warning("etta_list has not been set.  Using etta for all layers");
    }
    /* if etta list is ok, we should have jumped ahead a little bit */
    etta_list = allocate_vector(nnodes[0]);
    for(i = 1; i < nnodes[0]; i++) etta_list[i] = etta;
    etta_list[0] = (float)nnodes[0]-1;

  etta_list_ok:

    /* work out the sigmoid_list flags.  by the end, sigmoid list should be
       filled either with what the user asked for or with sigmoid_param
       on each level */
    if(sigmoid_param_list != NULL){
    if(use_sigmoid_param_list){
        if((int)sigmoid_param_list[0] < nnodes[0]-1){
        warning("sigmoid_param_list is too short.  Using sigmoid_param for all layers");
        } else goto sigmoid_list_ok;
    }
    /* free the old list so we can build a new one */
    deallocate_vector(sigmoid_param_list);
    } else {
    if(use_sigmoid_param_list)
      warning("sigmoid_param_list has not been set.  Using sigmoid_param for all layers");
    }
    sigmoid_param_list = allocate_vector(nnodes[0]);
    for(i = 1; i < nnodes[0]; i++) sigmoid_param_list[i] = sigmoid_param;
    sigmoid_param_list[0] = (float)nnodes[0]-1;

  sigmoid_list_ok:

    /************** make and print header comment *******************/
    comment = make_comment_string(nnargs_get_flags(argtab2, tabsize2));
    if (search_verbosity > FLAGS_VERBOSITY) {
    printf("%s", comment);
    fflush(stdout);
    }

    srandom(seed);      /* initialize random numbers */

    /* call the feature selection driver */
    mlp_feat_sel(search_direction, search_n, search_verbosity,
         search_param_fname, roc_cost, roc_target, roc_lower,
         roc_upper, df_args->categories,
         /* mlp_cross_valid_arguments */
         df_args->db_type, df_args->ninputs, ninputs, features, df_args->noutputs,
         normalize, npatterns, debug, verbosity, verror, nepochs,
         random_flag, cv_nfolds, cv_nsplits, use_splits,
         cv_splits, &cv_train_mask[1], &cv_test_mask[1],
         data_path, data_fname, norm_fname, exp_path, error_fname,
         param_fname, df_args->output_labels, comment, adjust_train,
         priors_npatterns, &train_priors_list[1], adjust_test,
         &test_priors_list[1], cv_random, etta_list, alpha,
         tolerance, decay, kappa, epsilon, sigmoid_param_list,
         nnodes,
         batch, cost_func,  cost_param, step_change, hfunction,
         ofunction, max_magnitude, log_outputs, etta_nepochs);

    print_log_footer();

    nnargs_free_argtab(argtab, tabsize);
    nnargs_free_argtab(argtab2, tabsize2);
    /* free the vectors allocated when the defaults file was read */
    free_defaults_file_struct(df_args);

    exit_nnsubs();
    exit(0);
}
