/***--------------------------------------------------------------------------
* NAME:         mlp_pack.c
* PURPOSE:      Functions specific to mlp.c
*
* SEE ALSO: mlp.c mlp_main.c
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   initialize_storage  Allocate storage, initialize some values
*   mlp_build       Allocate storage for model parameters.
*   mlp_init        Initialize model parameters.
*   mlp_free        Deallocates model storage
*   mlp_read_model      Initializes model from file
*   mlp_write_model     Writes model to output file
*   mlp_param       Returns pointers to elements in model
*   mlp_inc_train       Implements trial by trial back-propagation
*   mlp_inc_ctrain      Trial by trial back-propagation, class output
*   mlp_batch_train     Implements batch model back-propagation
*   mlp_batch_ctrain    batch back-propagation, class input
*   mlp_batch_update_weights Implement weight change for batch mode
*   mlp_classify        Implements mlp to return an integer class
*   mlp_eval        Propagates the input through the network
*
*    Internal processing functions
*   feed_fwd_one        Propagates input through one layer of network
*   back_prop       Update weights and checks for underflow
*   calc_hidden_deltas  Calculates deltas for hidden nodes
*   normal_back_prop    Back propagate at hidden layers
*
*
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* Load include files  */
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#include <nnlib.h>

/* to include internal declerations and definitions from `mlp_pack.h' */
#define   mlp_pack_h_internal_types
#define   mlp_pack_h_internal_macros
#define   mlp_pack_h_internal_functions
#define   mlp_pack_h_statics
#include "mlp_pack.h"

/* Global definitions */
#define PEQ(y) (0==strcmp(param_name,y))
#define ABS(X) (((X)>0)? (X) : -(X))
#define SMALLEST 1.e-30
#define TRUE 1
#define FALSE 0

/* Globals */
static MLP_PARAMS **mlpp;   /* Array of parameters for each model */
static int max_models = 0;  /* next new mlp model number to be assigned */
#define MODEL_STORE_SIZE 10

/*****************************************************************************
* NAME: initialize_storage(model_n)
* DESCRIPTION:
*       Allocate storage for mlp and initialize some values.
*
*
*****************************************************************************/
static void
initialize_storage(int model)
{
    int             i, j, k;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    /* First layer is input layer -- Same inputs and outputs */
    mpp->mlp_model[0].nodes = mpp->layers[0];
    mpp->mlp_model[0].node_sum = allocate_vector(mpp->layers[0]);
    mpp->mlp_model[0].node_output = allocate_vector(mpp->layers[0]);

    /* Now load data for hidden and output layers */
    for (i = 1; i < mpp->nlayers; i++) {
    /* Set up storage */
    mpp->mlp_model[i].nodes = mpp->layers[i];
    j = mpp->layers[i];
    k = mpp->layers[i - 1];
    mpp->mlp_model[i].node_sum = allocate_vector(j);
    mpp->mlp_model[i].node_output = allocate_vector(j);
    mpp->mlp_model[i].error_delta = allocate_vector(j);
    mpp->mlp_model[i].weights = allocate_matrix(j, k);
    mpp->mlp_model[i].weight_delta = allocate_matrix(j, k);
    mpp->mlp_model[i].batch_gradient = allocate_matrix(j, k);
    if (mpp->etta_type == MAS) {
        mpp->mlp_model[i].old_gradient = allocate_matrix(j, k);
        mpp->mlp_model[i].m_etta = allocate_matrix(j, k);
    }
    for (j = 0; j < mpp->layers[i]; j++)
      for (k = 0; k < mpp->layers[i - 1]; k++) {
          mpp->mlp_model[i].batch_gradient[j][k] = 0.;
      }
    }
    mpp->outputs = mpp->mlp_model[mpp->nlayers - 1].node_output;
    mpp->desired_outputs = allocate_vector(mpp->noutputs);
    mpp->check_underflow = FALSE;
    mpp->batch_on = FALSE;
    mpp->batch_cnt = 0;
}
/*****************************************************************************
* NAME: mlp_build()
* DESCRIPTION:
*       Allocate storage for model parameters.
*
*
*****************************************************************************/
int
mlp_build(int nlayers_param, /* number of layers, including input,
                output and hidden layers (must be > 1) */
      int *layers_param, /* pointer to array with nodes/layer */
      int etta_type)
{
    int             i, n_inputs;
    int             new_model;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    new_model = model_get_number((void ***)&mlpp, &max_models,
                 sizeof(MLP_PARAMS));
    mpp = mlpp[new_model];

    mpp->etta_type = etta_type;
    mpp->nlayers = nlayers_param;
    mpp->layers = allocate_ivector(nlayers_param);
    for (i = 0; i < nlayers_param; i++) {
    mpp->layers[i] = layers_param[i];
    }
    n_inputs = mpp->layers[0];
    mpp->ninputs = n_inputs;    /* don't include offset node */
    mpp->noutputs = mpp->layers[nlayers_param - 1];

    mpp->features = allocate_ivector(n_inputs+1);
    mpp->class_cnt = allocate_vector(mpp->noutputs);
    mpp->priors_list = allocate_vector(mpp->noutputs);

    mpp->mlp_model = (MLP_MODEL *)
      nn_malloc(sizeof(MLP_MODEL) * mpp->nlayers, "for mlp model structure");

    /*********** 1st layer is assumed to be all inputs ********/
    for(i = 0; i < nlayers_param; i++)
      mpp->layers[i]++;     /* Add a constant node of 1 for offset */

    initialize_storage(new_model);

    return new_model;
}
/*****************************************************************************
* NAME: mlp_init()
* DESCRIPTION:
*       Initialize model parameters.
*
*
****************************************************************************/
void
mlp_init(int model,
     float max_magnitude, /* all weights initialized to
                 -max_magnitude < weight < max_magnitude */
     int hfunction, int ofunction,
     float *etta_list, int *nn_features, int etta_tau)
{
    int             i, j, k, n_layers;
    float           random_value;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];

    /* copy feature list */
    for (i = 0; i <= nn_features[0]; i++) {
    mpp->features[i] = nn_features[i];
    }
    for(i = 0; i < mpp->noutputs; i++){
    mpp->class_cnt[i] = 0.;
    }
    mpp->total_npatterns = 0.;
    mpp->adjust_priors = 0;

    mpp->mlp_model[0].sigm_type = LINEAR;
    mpp->mlp_model[0].etta = mpp->mlp_model[0].etta0 = 0.;
    mpp->etta_tau = etta_tau;

    /* Set up weight structures and storage for other layers */
    n_layers = mpp->nlayers;
    for (i = 1; i < n_layers; i++) {
    mpp->mlp_model[i].sigm_type = hfunction;

    /*
     * Initialize all weights to a uniform random number between -1
     * and 1
     */
    for (j = 0; j < mpp->layers[i]; j++)
      for (k = 0; k < mpp->layers[i - 1]; k++) {
          random_value = random_uniform();
          mpp->mlp_model[i].weights[j][k] =
        (2. * random_value - 1.) * max_magnitude;
          mpp->mlp_model[i].weight_delta[j][k] = 0.;
          if (mpp->etta_type == MAS) {
          mpp->mlp_model[i].old_gradient[j][k] = 0.;
          mpp->mlp_model[i].m_etta[j][k] = etta_list[i];
          /* etta_list[0] is a length.
             etta_list[1..N] are stepsizes for each layer */
          /* set etta0 so something real is put in the parameter
             file eventually */
          mpp->mlp_model[i].etta0 = etta_list[i];
          } else {
          mpp->mlp_model[i].etta =
            mpp->mlp_model[i].etta0 = etta_list[i];
          } /* if constant, set etta in mlp_set_train_params() */
      }
    }/* End of setting up weight structures and storage */

    /* output function type can be different from hidden layer functions */
    mpp->mlp_model[n_layers -1].sigm_type = ofunction;

    mpp->accum_trials = 0;
}
/*****************************************************************************
* NAME: mlp_set_train_params()
* DESCRIPTION:
*    set model parameters that can be changed when training is continued
****************************************************************************/
void
mlp_set_train_params(int model, int cost_func, float cost_param,
             float *etta_list, float alpha, float decay, float kappa,
             float epsilon, float tolerance, float *sigmoid_param_list)
{
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */
    int i;

    mpp = mlpp[model];

    mpp->mlp_model[0].sigm_param = sigmoid_param_list[0];
    for(i = 1; i < mpp->nlayers; i++){
    mpp->mlp_model[i].etta = etta_list[i];
    mpp->mlp_model[i].sigm_param = sigmoid_param_list[i];
    }
    mpp->alpha = alpha;
    mpp->decay = decay;
    mpp->kappa = kappa;
    mpp->epsilon = epsilon;
    mpp->tolerance = tolerance;
    mpp->cost_func = cost_func;
    init_cost(cost_func, cost_param);
}

void
mlp_test_init(int model, int adjust_priors, float *priors_list)
{
    int i;
    MLP_PARAMS *mpp;

    mpp = mlpp[model];
    mpp->adjust_priors = adjust_priors;
    if(adjust_priors)
      for(i = 0; i < mpp->noutputs; i++)
    mpp->priors_list[i] = priors_list[i];
}
/****************************************************************************
* NAME: mlp_free()
* DESCRIPTION:
*       Deallocate storage for model parameters.
*
*
****************************************************************************/
void
mlp_free(int model)
{
    int             i;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    deallocate_ivector(mpp->layers);
    deallocate_ivector(mpp->features);
    deallocate_vector(mpp->class_cnt);
    deallocate_vector(mpp->priors_list);

    deallocate_vector(mpp->mlp_model[0].node_sum);
    deallocate_vector(mpp->mlp_model[0].node_output);
    deallocate_vector(mpp->desired_outputs);

    for (i = 1; i < mpp->nlayers; i++) {
    deallocate_vector(mpp->mlp_model[i].node_sum);
    deallocate_vector(mpp->mlp_model[i].node_output);
    deallocate_vector(mpp->mlp_model[i].error_delta);
    deallocate_matrix(mpp->mlp_model[i].weights);
    deallocate_matrix(mpp->mlp_model[i].weight_delta);
    deallocate_matrix(mpp->mlp_model[i].batch_gradient);
    if (mpp->etta_type == MAS) {
        deallocate_matrix(mpp->mlp_model[i].old_gradient);
        deallocate_matrix(mpp->mlp_model[i].m_etta);
    }
    }
    nn_free((char *)mpp->mlp_model, "for freeing mlp layer");
    nn_free((char *)mpp, "for freeing mlp model");
    mlpp[model] = NULL;
}

/*** Parameter Table ***/
static PARAM    paramtab[] = {
    {"nlayers", P_INT},
    {"accum_trials", P_INT},
    {"net_inputs", P_INT},
    {"net_outputs", P_INT},
    {"criterion", P_INT},
    {"etta_type", P_INT},
    {"etta_tau", P_INT},
    {"total_npat", P_FLOAT},
};
#define TABSIZE (sizeof(paramtab)/sizeof(PARAM))

/*** Paramters for number of nodes in each layer, etc. ***/
static PARAM    paramtab_l[] = {
    {"type0", P_INT},
    {"parm0", P_FLOAT},
    {"layers", P_INT_VECTOR},
    {"features", P_INT_VECTOR},
    {"class_cnt", P_FLOAT_VECTOR},
};
#define TABSIZE_L (sizeof(paramtab_l)/sizeof(PARAM))

/*** Parameter Table of weights for each layer ***/
static PARAM    paramtab_w[] = {
    {"type", P_INT},
    {"parm", P_FLOAT},
    {"etta0", P_FLOAT},
    {"weights", P_FLOAT_MATRIX},
    {"weight_delta", P_FLOAT_MATRIX},
};
#define TABSIZE_W (sizeof(paramtab_w)/sizeof(PARAM))

/*** Parameter Table for multiple adaptive stepsize if on ***/
static PARAM    paramtab_mas[] = {
    {"old_gradient", P_FLOAT_MATRIX},
    {"m_etta", P_FLOAT_MATRIX},
};
#define TABSIZE_MAS (sizeof(paramtab_mas)/sizeof(PARAM))

static void
mlp_set_paramtab(MLP_PARAMS *mpp)
{
    param_setvarptr("nlayers", &mpp->nlayers, paramtab, TABSIZE);
    param_setvarptr("accum_trials", &mpp->accum_trials, paramtab, TABSIZE);
    param_setvarptr("net_inputs", &mpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("net_outputs", &mpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("criterion", &mpp->cost_func, paramtab, TABSIZE);
    param_setvarptr("etta_type", &mpp->etta_type, paramtab, TABSIZE);
    param_setvarptr("etta_tau", &mpp->etta_tau, paramtab, TABSIZE);
    param_setvarptr("total_npat", (int *)&mpp->total_npatterns,
            paramtab, TABSIZE);
}
static void
mlp_set_paramtab_l(MLP_PARAMS *mpp)
{
    param_setvarptr("layers", &mpp->layers[0], paramtab_l, TABSIZE_L);
    param_setdim("layers", mpp->nlayers, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("type0", &mpp->mlp_model[0].sigm_type,
            paramtab_l, TABSIZE_L);
    param_setvarptr("parm0", (int *)&mpp->mlp_model[0].sigm_param,
            paramtab_l, TABSIZE_L);
    param_setvarptr("features", mpp->features, paramtab_l, TABSIZE_L);
    param_setdim("features", mpp->ninputs+1, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("class_cnt",(int *)mpp->class_cnt, paramtab_l,TABSIZE_L);
    param_setdim("class_cnt", mpp->noutputs, 0, paramtab_l, TABSIZE_L);
}

static void
mlp_set_paramtab_w(MLP_PARAMS *mpp, int i)
{
    param_setvarptr("type", &mpp->mlp_model[i].sigm_type,
            paramtab_w, TABSIZE_W);
    param_setvarptr("parm", (int *)&mpp->mlp_model[i].sigm_param,
            paramtab_w, TABSIZE_W);
    param_setvarptr("etta0", (int *)&mpp->mlp_model[i].etta0, paramtab_w,
            TABSIZE_W);
    param_setvarptr("weights",
            (int *)&mpp->mlp_model[i].weights[0][0],
            paramtab_w, TABSIZE_W);
    param_setdim("weights", mpp->layers[i], mpp->layers[i - 1],
         paramtab_w, TABSIZE_W);
    param_setvarptr("weight_delta",
            (int *) &mpp->mlp_model[i].weight_delta[0][0],
            paramtab_w, TABSIZE_W);
    param_setdim("weight_delta", mpp->layers[i], mpp->layers[i - 1],
         paramtab_w, TABSIZE_W);
}

static void
mlp_set_paramtab_mas(MLP_PARAMS *mpp, int i)
{
    param_setvarptr("old_gradient",
            (int *)&mpp->mlp_model[i].old_gradient[0][0],
            paramtab_mas, TABSIZE_MAS);
    param_setdim
      ("old_gradient", mpp->layers[i], mpp->layers[i - 1],
       paramtab_mas, TABSIZE_MAS);
    param_setvarptr("m_etta", (int *)&mpp->mlp_model[i].m_etta[0][0],
            paramtab_mas, TABSIZE_MAS);
    param_setdim("m_etta", mpp->layers[i], mpp->layers[i - 1],
         paramtab_mas, TABSIZE_MAS);
}
/****************************************************************************
* NAME: mlp_read_model(fp)
* DESCRIPTION:
*       Loads weight structures from ascii input file
*
****************************************************************************/
int
mlp_read_model(FILE *fp)
{
    int             i, rv;
    int             new_model;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    new_model = model_get_number((void ***)&mlpp, &max_models,
                 sizeof(MLP_PARAMS));
    mpp = mlpp[new_model];

    mlp_set_paramtab(mpp);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    mpp->layers = allocate_ivector(mpp->nlayers);
    mpp->features = allocate_ivector(mpp->ninputs+1);
    mpp->class_cnt = allocate_vector(mpp->noutputs);
    mpp->priors_list = allocate_vector(mpp->noutputs);

    mpp->mlp_model = (MLP_MODEL *)
    nn_malloc(sizeof(MLP_MODEL) * mpp->nlayers,
          "for mlp model structure during load");

    mlp_set_paramtab_l(mpp);
    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    initialize_storage(new_model);

    for (i = 1; i < mpp->nlayers; i++) {
    mlp_set_paramtab_w(mpp, i);
    if ((rv = param_getblk(fp, paramtab_w, TABSIZE_W)) < 0)
        return rv;
    if (mpp->etta_type == MAS) {
        mlp_set_paramtab_mas(mpp, i);
        if ((rv = param_getblk(fp, paramtab_mas, TABSIZE_MAS)) < 0)
        return rv;
    }
    }

    if(mpp->cost_func == PC_PROC){
      /* force a LINEAR cost function, rather than STEP on the final
     layer of PC_PROC */
      /* this will mess up retraining the PC_PROC */
      mpp->mlp_model[mpp->nlayers-1].sigm_type = LINEAR;
    }
    mpp->adjust_priors = 0;
    return new_model;
}

/*****************************************************************************
* NAME:  mlp_write_model()
* DESCRIPTION:
*       Writes weights and other parameters in ASCII format
*
*
******************************************************************************/
int
mlp_write_model(int model, FILE *fp)
{
    int             i, rv;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */


    /* For each layer, the function type and parameter is first written. */
    /* Next the weights for each layer are written, starting with the */
    /* first layer.  For each layer the weights for each node are then */
    /* written, followed by the momentum terms */
    mpp = mlpp[model];
    mlp_set_paramtab(mpp);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
    return rv;

    mlp_set_paramtab_l(mpp);
    param_setdim("features", mpp->features[0]+1, 0, paramtab_l, TABSIZE_L);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
    return rv;

    for (i = 1; i < mpp->nlayers; i++) {
    mlp_set_paramtab_w(mpp, i);
    if ((rv = param_putblk(fp, paramtab_w, TABSIZE_W)) < 0)
        return rv;
    if (mpp->etta_type == MAS) {
        mlp_set_paramtab_mas(mpp, i);
        if ((rv = param_putblk(fp, paramtab_mas, TABSIZE_MAS)) < 0)
        return rv;
    }
    }
    return(0);
}
/************************************************************************
* NAME: mlp_param(model, param_name, layer)
* DESCRIPTION:
*   Takes model number and parameter name.
*   Returns a pointer to the parameter.
*
*************************************************************************/
void *
mlp_param(int model, char *param_name, int layer)
{
    void           *param_address;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    if (PEQ("etta"))
    param_address = &mpp->mlp_model[layer].etta;
    else if (PEQ("alpha"))
    param_address = &mpp->alpha;
    else if (PEQ("decay"))
    param_address = &mpp->decay;
    else if (PEQ("tolerance"))
    param_address = &mpp->tolerance;
    else if (PEQ("kappa"))
    param_address = &mpp->kappa;
    else if (PEQ("epsilon"))
    param_address = &mpp->epsilon;
    else if (PEQ("ninputs"))
    param_address = &mpp->ninputs;
    else if (PEQ("noutputs"))
    param_address = &mpp->noutputs;
    else if (PEQ("trial"))
    param_address = &mpp->accum_trials;
    else if (PEQ("output"))
    param_address = &mpp->outputs;
    else if (PEQ("features"))
    param_address = &mpp->features;
    else if (PEQ("nlayers"))
    param_address = &mpp->nlayers;
    else if (PEQ("layers"))
    param_address = &mpp->layers;
    else if (PEQ("weights"))
    param_address = &mpp->mlp_model[layer].weights;
    else if (PEQ("sigm_type"))
      param_address = &mpp->mlp_model[layer].sigm_type;
    else if (PEQ("sigm_param"))
      param_address = &mpp->mlp_model[layer].sigm_param;
    else if (PEQ("class_cnt"))
      param_address = &mpp->class_cnt;
    else if (PEQ("cost"))
      param_address = &mpp->cost;
    else if (PEQ("cost_func"))
      param_address = &mpp->cost_func;
    else if (PEQ("total_npatterns"))
      param_address = &mpp->total_npatterns;
    else {
    warning("Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}               /* END mlp_param() */

/*****************************************************************************
* NAME: mlp_inc_train (model, inputs, desired_outputs)
* DESCRIPTION:
*       Implements trial by trial back-propagation
*
*****************************************************************************/
void
mlp_inc_train(int model, float *inputs, float *desired_outputs)
{
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    mpp->accum_trials++;    /* accumulate total number of trials for
                 * weights */
    /* ----------- feed forward through network ----- */
    mlp_eval(model, inputs);

    /* ------- back propagate error corrections  ---- */
    back_prop(model, desired_outputs, mpp->outputs, mpp->tolerance,
          mpp->check_underflow);

    /* update the weights, if necessary */
    if(mpp->etta_type == DECREASE_ETTA) set_etta_values(mpp);

    mpp->cost = calc_cost(mpp->outputs, desired_outputs, mpp->noutputs,
              mpp->cost_func);
    /*
     * If cost is low, start checking for underflow -- When the error is
     * small, underflow problems can affect the momentum and weight update
     * and slow the program
     */
    if (!mpp->check_underflow && (mpp->cost) < mpp->tolerance)
    mpp->check_underflow = TRUE;
}
/*****************************************************************************
* NAME: mlp_inc_ctrain (model, inputs, desired_class)
* DESCRIPTION:
*       Implements trial by trial back-propagation
*   returns computed class
*
*****************************************************************************/
int
mlp_inc_ctrain(int model, float *inputs, int desired_class)
{
    int             i;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];

    mpp->class_cnt[desired_class] += 1.;
    mpp->total_npatterns += 1.;

    for (i = 0; i < mpp->noutputs; i++)
      mpp->desired_outputs[i] = 0;
    mpp->desired_outputs[desired_class] = 1;

    mlp_inc_train(model, inputs, mpp->desired_outputs);

    return (max_pick(mpp->outputs, mpp->noutputs));
}

/*****************************************************************************
* NAME: mlp_batch_train (model, inputs, desired_outputs)
* DESCRIPTION:
*       Implements batch model back-propagation
*
*****************************************************************************/
void
mlp_batch_train(int model, float *inputs, float *desired_outputs)
{
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    if (mpp->batch_on)
    mpp->batch_cnt++;
    else {
    mpp->batch_on = TRUE;
    mpp->batch_cnt = 1;
    }

    mlp_inc_train(model, inputs, desired_outputs);
}


/*****************************************************************************
* NAME: mlp_batch_ctrain (model, inputs, desired_class)
* DESCRIPTION:
*       Implements batch model back-propagation
*   returns computed class
*
*****************************************************************************/
int
mlp_batch_ctrain(int model, float *inputs, int desired_class)
{
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    if (mpp->batch_on)
    mpp->batch_cnt++;
    else {
    mpp->batch_on = TRUE;
    mpp->batch_cnt = 1;
    }

    return (mlp_inc_ctrain(model, inputs, desired_class));
}

/*****************************************************************************
* NAME:   mlp_batch_update_weights ()
* DESCRIPTION:
*    Implement weight change for batch mode. Initializes batch variables to 0
*
*****************************************************************************/
void
mlp_batch_update_weights(int model)
{
    register int    i, j, k;
    register float  new_weight_inc, old_weight_inc;
    float           gradient_product, delta_etta, etta_tmp;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */
    MLP_MODEL *wpt, *prev_wpt;

    mpp = mlpp[model];

    wpt = &(mpp->mlp_model[0]);
    for (i = 1; i < mpp->nlayers; i++) {
    prev_wpt = wpt;
    wpt = &(mpp->mlp_model[i]);
    for (j = 0; j < wpt->nodes; j++) {
        for (k = 0; k < prev_wpt->nodes; k++) {
        /* use etta for this weight, or etta for the layer */
        if (mpp->etta_type == MAS)
          etta_tmp = wpt->m_etta[j][k];
        else
          etta_tmp = wpt->etta;
        new_weight_inc = wpt->batch_gradient[j][k] * etta_tmp;
        old_weight_inc = wpt->weight_delta[j][k];
        if (!(mpp->check_underflow &&
              (ABS((double) old_weight_inc) < SMALLEST)))
            new_weight_inc += mpp->alpha * old_weight_inc;
        if (mpp->etta_type == MAS) {
            if (wpt->old_gradient[j][k] != 0.) {
            gradient_product =
              wpt->batch_gradient[j][k] * wpt->old_gradient[j][k];
            /* Multiple adaptive stepsize algorithm update */
            if (gradient_product > SMALLEST)
              delta_etta = mpp->kappa;
            else if (gradient_product < -SMALLEST)
              delta_etta = -(mpp->epsilon * etta_tmp);
            else /*between*/
              delta_etta = 0.;
            wpt->m_etta[j][k] += delta_etta;
            if(wpt->m_etta[j][k] > 1.) wpt->m_etta[j][k] = 1.;
            }
            wpt->old_gradient[j][k] = wpt->batch_gradient[j][k];
        }
        wpt->weights[j][k] += new_weight_inc;
        wpt->weight_delta[j][k] = new_weight_inc;
        wpt->weights[j][k] -= wpt->weights[j][k] * mpp->decay;

        /* reset batch_gradient */
        wpt->batch_gradient[j][k] = 0.;
        }
    }
    }
    mpp->batch_cnt = 0;
}
/* SuperSAB algorithm */
/***    if(gradient_product < 0) {
         *m_etta_p = etta_tmp / epsilon;
         *weights_p = *weights_p - *weight_delta_p;
         new_weight_inc = 0;
} else {
         *m_etta_p = etta_tmp * kappa;
         *m_etta_p += kappa;
         *old_gradient_p = *batch_gradient_p;
}
*/

/*****************************************************************************
* NAME: mlp_classify()
* DESCRIPTION:
*       Implements the mlp to return an integer class.
*
*
*****************************************************************************/
int
mlp_classify(int model, float *inputs)
{
    int i;
    float total;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];

    mlp_eval(model, inputs);
    if(mpp->adjust_priors){
    /* adjust priors */
    for(i = 0; i < mpp->noutputs; i++){
        mpp->outputs[i] *= mpp->priors_list[i];
        mpp->outputs[i] /= mpp->class_cnt[i]/mpp->total_npatterns;
    }
    /* scale adjusted outputs to sum to 1 */
    total = 0.;
    for(i = 0; i < mpp->noutputs; i++) total += mpp->outputs[i];
    for(i = 0; i < mpp->noutputs; i++) mpp->outputs[i] /= total;
    }
    return max_pick(mpp->outputs, mpp->noutputs);
}

/*****************************************************************************
* NAME: mlp_eval()
* DESCRIPTION:
*       Propagates the input through the network.  Calculates the
*       network values for each layer and stores the network
*   output in the outputs array for the model
*****************************************************************************/
void
mlp_eval(int model, float *inputs)
{
    register int    i, j, output_layer;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */
    MLP_MODEL *wpt, *next_wpt;

    mpp = mlpp[model];

    /* Load input vector into network */
    for (j = 0; j < mpp->ninputs; j++) {
    mpp->mlp_model[0].node_sum[j] = inputs[j];
    }

    /* feed vector through all of network, except for last sigmoid */
    output_layer = mpp->nlayers - 1;
    next_wpt = &(mpp->mlp_model[0]);
    for (i = 0; i < output_layer; i++) {
    wpt = next_wpt;
    next_wpt = &(mpp->mlp_model[i+1]);
    feed_fwd_one(wpt->nodes, next_wpt->nodes,
             wpt->node_sum, next_wpt->node_sum, next_wpt->weights,
             wpt->sigm_type, wpt->sigm_param, wpt->node_output);
    }/* end of calculating layer inputs and outputs */

    /* Calculate the network outputs */
    calc_final_sigm(mpp->mlp_model[output_layer].node_sum,
            mpp->outputs, mpp->noutputs,
            mpp->mlp_model[output_layer].sigm_type,
            mpp->mlp_model[output_layer].sigm_param,
            mpp->cost_func);
}

/*****************************************************************************
* NAME: feed_fwd_one()
* DESCRIPTION:
*       Propagates the input through one layer of the network.
*   Called by feed_forward.
*
*   weight->node_output is the signal from the previous layer
*   weight->node_sum is the summed activity for the current node
*
******************************************************************************/
static void
feed_fwd_one(int n_curr, int n_next, float *curr_node_sum,
         float *next_node_sum, float **w_array, int type,
         float parm, float *curr_node_output)
{
    int j;

    /* send outputs of previous layer through sigmoid
       to be this layer's inputs*/
    for (j = 0; j < n_curr - 1; j++)
    curr_node_output[j] = sigm(curr_node_sum[j], type, parm);
    curr_node_output[n_curr - 1] = 1.;  /* add threshold value  */

    /* take weighted sum of inputs to be this layers outputs */
    matrix_vector_multiply(w_array, curr_node_output, next_node_sum,
               n_next, n_curr);
}

/*****************************************************************************
* NAME: back_prop()
* DESCRIPTION:
*       Update weights and checks for underflow
*
*   The formula for weight update is
*       wij(tau + 1) = wij(tau) + etta * deltaj * xi
*       where deltaj = yj * (1 - yj) * (dj - yj) for the output layer
*       and   deltaj = xj * (1 - xj) * Sum__over_k (deltak * wjk)
*       for the hidden layer where k are nodes in the layer above
*           node j
*
*       when momentum is added, the additional term
*       alpha * (wij(tau) - wij(tau - 1))
*       is added to the first equation
*
*       wij(tau) is the weight array mlp_model->weights
*       the deltaj values are stored in mlp_model->error_delta
*       the xj * (1 - xj) term (yj also) is computed by the dsigm func.
*       wij(tau - 1) is stored in the mlp_model->weight_delta array
*       (dj - yj) is the desired value (desired_outputs) - outputs
*
****************************************************************************/
static void
back_prop(int model, float *desired_outputs, float *outputs, float tolerance,
      int check_underflow)
{
    MLP_MODEL *wpt;
    int in_tolerance;   /* whether outp err is too small to change
             * weights */
    MLP_PARAMS *mpp;    /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    /* Calculate delta for output nodes */
    wpt = &(mpp->mlp_model[mpp->nlayers - 1]);
    in_tolerance =
      calc_output_deriv(desired_outputs, outputs, tolerance, wpt->error_delta,
            wpt->nodes - 1, wpt->sigm_type, wpt->sigm_param,
            mpp->cost_func);

    if(!in_tolerance){
    /* only propogate changes if errors were large enough */
    calc_hidden_deltas(model, mpp->nlayers);
    normal_back_prop(model, mpp->nlayers, mpp->alpha, check_underflow);
    }
}

/*****************************************************************************
* NAME:   calc_hidden_deltas ()
* DESCRIPTION:
*       If at least one error is big enough, update the normal way
*   First calculate the delta's for the nodes in all layers but the  last
*
*****************************************************************************/
static void
calc_hidden_deltas(int model, int n_layers)
{
    register MLP_MODEL *wpt;
    MLP_MODEL      *wpt_next;
    register int    typ, j;
    register float  par;
    MLP_PARAMS *mpp;    /* temporary pointer to mlpp[] */

    mpp = mlpp[model];
    wpt = &(mpp->mlp_model[--n_layers]);
    while(--n_layers) {
    wpt_next = wpt;     /* ptr. to next layer */
    wpt = &(mpp->mlp_model[n_layers]);

    /* Do a transposed multiplication -- For each node in the present
     * layer, each weight from that node to the next layer is multiplied
     * by the delta of the appropriate node in the next layer.  The
     * products are summed and stored in the delta for that node. */
    transposed_matrix_vector_multiply(wpt_next->weights,
                      wpt_next->error_delta,
                      wpt->error_delta, wpt->nodes,
                      wpt_next->nodes - 1);

    /* The deltas are then multiplied by the derivative of the node
     * function */
    typ = wpt->sigm_type;
    par = wpt->sigm_param;
    for (j = 0; j < wpt->nodes - 1; j++)
        wpt->error_delta[j] *= dsigm(wpt->node_output[j], typ, par);
    }
}

/*****************************************************************************
* NAME:   normal_back_prop ()
* DESCRIPTION:
*       Back propagate at hidden layers
*
****************************************************************************/
static void
normal_back_prop(int model, int n_layers, float alpha, int check_underflow)
{
    register MLP_MODEL *wpt, *prev_wpt;
    register int i, j, k;
    register float new_weight_inc, old_weight_inc;
    MLP_PARAMS *mpp;        /* temporary pointer to mlpp[] */

    mpp = mlpp[model];

    /* Since input layer (layer 0) has no weights, start with second layer */
    wpt = &(mpp->mlp_model[0]);
    if(alpha == 0. && mpp->decay == 0. ){
    /* this loop is much shorter if there is no momentum and
       no decay */
    for (i = 1; i < n_layers; i++) {
        prev_wpt = wpt;
        wpt = &(mpp->mlp_model[i]);
        
        for (j = 0; j < wpt->nodes - 1; j++) {
        for (k = 0; k < prev_wpt->nodes; k++) {
            new_weight_inc =
              (-1) * wpt->error_delta[j] * prev_wpt->node_output[k];
            if (mpp->batch_on)
              wpt->batch_gradient[j][k] += new_weight_inc;
            else {
            new_weight_inc *= wpt->etta;
            wpt->weights[j][k] += new_weight_inc;
            }
        }
        }
    }
    } else {
    for (i = 1; i < n_layers; i++) {
        prev_wpt = wpt;
        wpt = &(mpp->mlp_model[i]);
        
        for (j = 0; j < wpt->nodes - 1; j++) {
        /*nodes-1 is upper limit because extra node has no
          input weights*/
        
        for (k = 0; k < prev_wpt->nodes; k++) {
            /* Weight increment equals 
             * (etta(stepsize)*delta*corresponding input)
             * + weight_delta*alpha
             * if underflow, ignore momentum */
            new_weight_inc =
              (-1) * wpt->error_delta[j] * prev_wpt->node_output[k];
            if (mpp->batch_on)
              wpt->batch_gradient[j][k] += new_weight_inc;
            else {
            new_weight_inc *= wpt->etta;
            old_weight_inc = wpt->weight_delta[j][k];
            if (!(check_underflow && (ABS((double) old_weight_inc) <
                          SMALLEST)))
              new_weight_inc += alpha * old_weight_inc;
            wpt->weight_delta[j][k] = new_weight_inc;
            wpt->weights[j][k] += new_weight_inc;
            wpt->weights[j][k] -= wpt->weights[j][k] * mpp->decay;
            }
        }
        }
    }
    }
}

/* calculate the new step size values for the next training trial
   this should be called only when mpp->etta_type == DECREASE_ETTA

   etta for each level should be approximately etta0 for etta_tau
   trials.  etta then should decrease sharply such that the rate of
   decrease, when plotted on a log/log scale should have a slope of
   1/(number of trials)
*/
static void
set_etta_values(MLP_PARAMS *mpp)
{
    int i;
    MLP_MODEL *wpt;

    for(i = 1; i < mpp->nlayers; i++){
    wpt = &(mpp->mlp_model[i]);
    wpt->etta = wpt->etta0;
    wpt->etta *= 1. + (float)mpp->accum_trials;
    wpt->etta /= 1. + (float)mpp->accum_trials +
      (float)mpp->accum_trials*(float)mpp->accum_trials/(float)mpp->etta_tau;
    }
}
