/***--------------------------------------------------------------------------
 * NAME:		mlp_seq_cv.c
 * PURPOSE:	cross validation driver for MultiLayer Perceptron
 *
 * SEE ALSO: 	mlp_seq_pack.c
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_seq_cv.c,v 11.4 2001/08/23 20:12:32 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_seq_cv.c,v $
 * CREATED: 27-Sep-93 lck			MIT LINCOLN LABORATORY
 **************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlp_seq_pack.h"
#include "mlp_seq.h"

static char prog_name[MAXSTR];	/* static variables */

void
mlp_seq_cross_valid(int db_type, int nraw, int ninputs, int *features,
		    int nwindow_frames, int *window_frames, int noutputs,
		    int normalize, int nsequences, int debg, int verbosity,
		    int verror, int nepochs, int random_flag, int cv_nfolds,
		    int cv_nsplits, int use_cv_splits, int *cv_splits,
		    char **cv_train_mask, char **cv_test_mask, int cv_random,
		    char *data_path, char *data_fname, char *norm_fname,
		    char *exp_path, char *error_fname, char *param_fname,
		    char **labels, char *comment,
		    int adjust_priors, int priors_nsequences,
		    float *priors_list, float *etta_list,
		    float alpha, float tolerance, float decay, float kappa,
		    float epsilon, float *sigmoid_param_list,
		    int *nnodes, int *batch, int cost_type, float cost_param,
		    int etta_type, int hfunction, int ofunction,
		    float max_magnitude, int log_outputs, float etta_nepochs,
		    float clip_floor)
{
    /* Local variables */
    int		model, err_model, err_part_model, data_model, seq_model;
    int		normalization=0, epoch, etta_tau;
    int		model_ninputs;
    int		desired_class, class;
    int		n_read, seq_num, n_frames;
    int		fold, nfiles;
    int		post_sum_to_one = 0;
    float	*inputs, *seq_sums;
    char	*tmp_path;
    FILE	*error_file = NULL, *param_file;
    int		w_update_cnt = 0;
    int		w_update_int, w_update_max, w_update_increase;
    
    strcpy(prog_name, get_program_name());
    
    w_update_int = batch[1];
    w_update_max = batch[2];
    w_update_increase = batch[3];

    /* allocate and initialize model structure */
    model = mlp_build(nnodes[0], &nnodes[1], etta_type);
    /* initialization must be done later, after we know how many
       frames of data will be used each fold */
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else norm_setup_none(nraw, noutputs);
    model_ninputs = nnodes[1];

    /* build seq model */
    seq_model = mlp_seq_build(noutputs);
    mlp_seq_init(seq_model, model, clip_floor);
    seq_sums = *(float **)mlp_seq_param(seq_model, "seq_sums");
    
    /* prepare for error reporting */
    if (verror > 0) {	/* open file for classification errors */
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      "w", "for classification errors");
    }
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, db_type, model_ninputs, cost_type, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    nfiles = seq_data_get_nfiles(data_model);
    seq_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		     cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) seq_data_random_init(data_model);
    if(adjust_priors)
      seq_data_priors_init(data_model, priors_list, priors_nsequences);

    /* store the normalization, feature selection, and windowing once only */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);

    for(fold = 0; fold < cv_nfolds; fold++) {
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	etta_tau = (int)(etta_nepochs *
			 (float)seq_data_get_nframes(data_model));
	mlp_init(model, max_magnitude, hfunction, ofunction,
		 etta_list, features, etta_tau);
	mlp_set_train_params(model, cost_type, cost_param, etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
	seq_data_start_cv_fold(data_model, fold, TRUE);
	nfiles = seq_data_get_nfiles(data_model);
	
	/* verror for training is 0 */
	report_init(err_part_model, db_type, model_ninputs, cost_type,
		    log_outputs, post_sum_to_one);
	report_epoch_header(err_part_model, stdout, nfiles);

	for (epoch = 1; epoch <= nepochs; epoch++) {
	    seq_data_epoch_init(data_model);

	    /* go through the data for this fold */
	    w_update_cnt = 0;
	    for (n_read = 0; n_read < nfiles; n_read++) {
		if(seq_data_get_next_file(data_model) == -1) break;
		desired_class = seq_data_get_class(data_model);
		mlp_seq_init(seq_model, model, clip_floor);
		for(n_frames = 0; ; n_frames++){
		    inputs =
		      seq_data_get_next_frame(data_model);
		    if(inputs == NULL) break;

		    mlp_seq_batch_classify(seq_model, inputs);
		    if (w_update_int <= 1)
		      mlp_inc_ctrain(model, inputs, desired_class);
		    else {
			mlp_batch_ctrain(model, inputs, desired_class);
			if (++w_update_cnt > w_update_int) {
			    mlp_batch_update_weights(model);
			    w_update_cnt = 0;
			}
		    }
		}
		if(n_frames > 0){
		    class = mlp_seq_update_classify(seq_model);
		    seq_num = seq_data_get_file_num(data_model);
		    report_seq_trial(err_part_model, error_file, seq_num,
				     seq_sums, desired_class, class);
		}
	    } /* end reading data */
	    if (w_update_cnt > 0) /* do a final update, if needed */
	      mlp_batch_update_weights(model);
	    if (w_update_int < w_update_max)
	      w_update_int += w_update_increase;
	    report_epoch(err_part_model, stdout, epoch);
	    report_init(err_part_model, db_type, model_ninputs, cost_type,
			log_outputs, post_sum_to_one);
	} /* end epoch */
	
	/* print out the mlp model */
	fprintf(param_file, "mlp model %d\n", fold);
	mlp_write_model(model, param_file);
	if ( verbosity > 0)
	  printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
	
	/* reinitialize model for testing */
	mlp_seq_init(seq_model, model, clip_floor);
	report_init(err_part_model, db_type, model_ninputs, cost_type,
		    log_outputs, post_sum_to_one);
	
	/* finish the training fold and start the testing fold */
	seq_data_end_cv_fold(data_model);
	seq_data_start_cv_fold(data_model, fold, FALSE);
	nfiles = seq_data_get_nfiles(data_model);

	for (n_read = 0; n_read < nfiles; n_read++) {
	    if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
	    desired_class = seq_data_get_class(data_model);
	    mlp_seq_init(seq_model, model, clip_floor);
	    for(n_frames = 0; 1; n_frames++){
		inputs =
		  seq_data_get_next_frame(data_model);
		if(inputs == NULL) break;

		mlp_seq_batch_classify(seq_model, inputs);
	    }
	    if(n_frames > 0){
		class = mlp_seq_update_classify(seq_model);
		report_seq_trial(err_model, error_file, seq_num, seq_sums,
				 desired_class, class);
		report_seq_trial(err_part_model, error_file, seq_num, seq_sums,
				 desired_class, class);
	    }
	}
	report_overall(err_part_model, stdout, error_fname, labels);
	
	/* finish the testing fold */
	seq_data_end_cv_fold(data_model);
    } /* end fold */
    
    /* finish up writing the models */
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    seq_data_exit(data_model);
    mlp_seq_free(seq_model);
    mlp_free(model);
    norm_free();
}
