/***--------------------------------------------------------------------------
 * NAME:		mlp_seq_test.c
 * PURPOSE:	testing driver for MultiLayer Perceptron sequence classifier
 *
 * SEE ALSO: 	mlp_pack.c mlp_seq_pack.c
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_seq_test.c,v 11.4 2001/08/23 20:12:32 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_seq_test.c,v $
 * CREATED: 27-Sep-93 lck			MIT LINCOLN LABORATORY
 *********************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlp_seq_pack.h"
#include "mlp_seq.h"

static char     prog_name[MAXSTR];	/* static variables */

void
mlp_seq_test(int create_flag, int db_type, int nsequences, int debg,
	     int verbosity, int verror, char *data_path, char *data_fname,
	     char *exp_path, char *error_fname, char *param_fname,
	     char **labels, int adjust_priors, float *priors_list,
	     int *nnodes, int cost_type, float cost_param, int log_outputs,
	     float clip_floor)
{
    /* Local variables */
    int		model, err_model, seq_model, data_model, rv, normalization=0;
    int		*features, nraw, ninputs, noutputs, model_ninputs;
    int		nwindow_frames, *window_frames;
    int		class, desired_class;
    int		i, n_files, seq_num, n_frames;
    float	*inputs, *seq_sums;
    char	*tmp_path;
    FILE	*error_file = NULL, *param_file;
    int		post_sum_to_one = 0;
    
    int		nlayers;	/* number of layers */
    int		layers[20];	/* number of nodes in each layer */
    
    strcpy(prog_name, get_program_name());
    
    nlayers = nnodes[0];	/* Put the number of nodes in each layer in
				 * nlayers */
    for (i = 0; i < nlayers; i++)
      layers[i] = nnodes[i + 1];
    
    /* open parameter file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");

    /* read normalization from parameter file */
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on normalization parameters in file", tmp_path);
    rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on feature selection parameters in file", tmp_path);
    rv = window_read_paramfile(param_file, &nwindow_frames, &window_frames);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on windowing parameters in file", tmp_path);

    /* read mlp model from parameter file */
    model = mlp_read_model(param_file);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s",
		  prog_name, model, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of model file");

    /* build sequence model */
    noutputs = *(int *)mlp_param(model, "noutputs", 0);
    model_ninputs = *(int *)mlp_param(model, "ninputs", 0);
    seq_model = mlp_seq_build(noutputs);
    mlp_seq_init(seq_model, model, clip_floor);
    
    seq_sums = *(float **) mlp_seq_param(seq_model, "seq_sums");
    
    /* prepare results reporting */
    if (verror > 0) {	/* open file for classification errors */
	if(create_flag)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    }
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, db_type, model_ninputs, cost_type, log_outputs, post_sum_to_one);
    init_cost(cost_type, cost_param);

    /* read data */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    nsequences = seq_data_get_nfiles(data_model);
    
    /* read and classify input from data file */
    for(n_files = 0; n_files < nsequences; n_files++){
	if(seq_data_get_next_file(data_model) == -1) break;
	mlp_seq_init(seq_model, model, clip_floor);
	desired_class = seq_data_get_class(data_model);
	for(n_frames = 0; ; n_frames++){
	    inputs = seq_data_get_next_frame(data_model);
	    if(inputs == NULL) break;
	    mlp_seq_batch_classify(seq_model, inputs);
	}
	if(n_frames > 0){
	    class = mlp_seq_update_classify(seq_model);
	    seq_num = seq_data_get_file_num(data_model);
	    report_seq_trial(err_model, error_file, seq_num, seq_sums,
			     desired_class, class);
	}
    }

    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);    
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    seq_data_exit(data_model);
    mlp_seq_free(seq_model);
    mlp_free(model);
    norm_free();
}
