/***--------------------------------------------------------------------------
 * NAME:	mlp_seq_train.c
 * PURPOSE:	training driver for MultiLayer Perceptron sequence classifier
 *
 * SEE ALSO: 	mlp_pack.c 
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_seq_train.c,v 11.4 2001/08/23 20:12:32 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_seq_train.c,v $
 * CREATED: 24-Sep-93					MIT LINCOLN LABORATORY
 *****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlp_seq.h"
#include "mlp_seq_pack.h"

static char     prog_name[MAXSTR];	/* static variables */

void
mlp_seq_train(int create_flag, int db_type, int nraw, int ninputs,
	      int *features, int nwindow_frames, int *window_frames,
	      int noutputs, int normalize, int nsequences, int debg,
	      int verbosity, int verror, int nepochs, int random_flag,
	      char *data_path, char *data_fname, char *norm_fname,
	      char *exp_path, char *error_fname, char *param_fname,
	      char **labels, char *comment, int adjust_priors,
	      int priors_nsequences, float *priors_list, float *etta_list,
	      float alpha, float tolerance, float decay, float kappa,
	      float epsilon, float *sigmoid_param_list, int *nnodes,
	      int *batch, int cost_type, float cost_param, int etta_type,
	      int hfunction, int ofunction, float max_magnitude,
	      int log_outputs, float etta_nepochs, float clip_floor)
{
    
    /* Local variables */
    int		model, err_model, err_part_model, data_model, seq_model;
    int		rv, normalization=0, epoch, etta_tau;
    int		model_ninputs;
    int		desired_class, class;
    int		post_sum_to_one = 0;
    int		n_read, nfiles, seq_num, n_frames;
    float	*inputs, *seq_sums;
    char	*tmp_path, *open_type;
    FILE	*error_file = NULL, *param_file;
    
    int		w_update_cnt = 0;
    int		w_update_int, w_update_max, w_update_increase;
    
    strcpy(prog_name, get_program_name());
    
    w_update_int = batch[1];
    w_update_max = batch[2];
    w_update_increase = batch[3];
    
    if (create_flag) {
	/* allocate and initialize model structure */
	model = mlp_build(nnodes[0], &nnodes[1], etta_type);
	/* initialization must be done later, after we know how many
	   frames of data will be used each epoch */

	if (normalize)
	  normalization = read_norm_file(path_build(data_path, norm_fname));
	else
	  norm_setup_none(nraw, noutputs);
	model_ninputs = nnodes[1];
    } else {		/* retrain */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on normalization parameters in file", tmp_path);
	rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on feature selection parameters in file", tmp_path);
	rv = window_read_paramfile(param_file, &nwindow_frames, &window_frames);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s", prog_name, rv,
		      "on windowing parameters in file", tmp_path);
	model = mlp_read_model(param_file);
	if (model < 0)
	  fatal_error("%s: read error %d on model file %s", 
		      prog_name, model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of model file");
	mlp_set_train_params(model, cost_type, cost_param, etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
	model_ninputs = *(int *)mlp_param(model, "ninputs", 0);
    }

    /* build seq model */
    seq_model = mlp_seq_build(noutputs);
    mlp_seq_init(seq_model, model, clip_floor);
    seq_sums = *(float **)mlp_seq_param(seq_model, "seq_sums");

    /* set up for reporting the results */
    if (verror > 0) {	/* open file for classification errors */
	if (create_flag) open_type = "w";
	else open_type = "a";
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      open_type, "for classification errors");
    }
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, db_type, model_ninputs, cost_type, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);
    report_init(err_part_model, db_type, model_ninputs, cost_type,
		log_outputs, post_sum_to_one);

    /* read data and initialize data handling routines */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames, window_frames);

    if(random_flag) seq_data_random_init(data_model);
    if(adjust_priors)
      seq_data_priors_init(data_model, priors_list, priors_nsequences);
    
    /* get the number of patterns per epoch */
    nfiles = seq_data_get_nfiles(data_model);
    report_epoch_header(err_model, stdout, nfiles);
    
    if(create_flag){
	/* we still need to set up the step size parameters */
	etta_tau = (int)(etta_nepochs *
			 (float) seq_data_get_nframes(data_model));
	mlp_init(model, max_magnitude, hfunction, ofunction,
		 etta_list, features, etta_tau);
	mlp_set_train_params(model, cost_type, cost_param, etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
    }
    for (epoch = 0; epoch < nepochs; epoch++) {
	seq_data_epoch_init(data_model);

	w_update_cnt = 0;
	for (n_read = 0; n_read < nfiles; n_read++) {
	    /* get the next pattern */
	    if((seq_num = seq_data_get_next_file(data_model)) < 0) break;
	    desired_class = seq_data_get_class(data_model);
	    mlp_seq_init(seq_model, model, clip_floor);
	    for(n_frames = 0; ; n_frames++){
		inputs = seq_data_get_next_frame(data_model);
		if(inputs == NULL) break;

		mlp_seq_batch_classify(seq_model, inputs);
		/* train in batch mode or updating weights immediately */
		if (w_update_int <= 1)
		  mlp_inc_ctrain(model, inputs, desired_class);
		else {
		    mlp_batch_ctrain(model, inputs, desired_class);
		    if (++w_update_cnt > w_update_int) {
			mlp_batch_update_weights(model);
			w_update_cnt = 0;
		    }
		}
	    }
	    if(n_frames > 0){
		class = mlp_seq_update_classify(seq_model);
		report_seq_trial(err_model, error_file, seq_num, seq_sums,
				 desired_class, class);
		report_seq_trial(err_part_model, error_file, seq_num, seq_sums,
				 desired_class, class);
	    }
	}
	if (w_update_cnt > 0)
	  mlp_batch_update_weights(model);
	if (w_update_int < w_update_max)
	  w_update_int += w_update_increase;

	report_epoch(err_part_model, stdout, epoch);
	report_init(err_part_model, db_type, model_ninputs, cost_type,
		    log_outputs, post_sum_to_one);
    }
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);
    fprintf(param_file, "mlp model\n");
    mlp_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting and close error file */
    /* only show the last epoch, not the overall errors */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    seq_data_exit(data_model);
    norm_free();
    mlp_free(model);
}
