/***--------------------------------------------------------------------------
 * NAME:		mlp_test.c
 * PURPOSE:	testing driver for MultiLayer Perceptron
 *
 * SEE ALSO: 	mlp_pack.c
 *
 * FUNCTIONS:
 *	mlp	driver for MultiLayer Perceptron
 *		handles model creation, testing of data
 *		Constant node of 1 (bias) appended to each layer
 *		First layer is linear, other layers are user definable
 *		For classification, error file contains sample number,
 *			correct class, class chosen, 
 *			correct/incorrect decision, and rmserr
 *		For mapping, error file contains sample number, rmserr
 *		Network inputs followed by outputs for each sample
 *			optionally sent to error file
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_test.c,v 11.4 2001/08/23 20:12:32 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_test.c,v $
 * CREATED: 21-SEP-90 dan                             MIT LINCOLN LABORATORY
 *********************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlp.h"

static char     prog_name[MAXSTR];	/* static variables */

void
mlp_test(int create_flag, int map_flag, int nraw, int ninputs, int *features,
	 int noutputs, int npatterns, int verbosity, int verror,
	 char *data_path, char *data_fname, char *exp_path,
	 char *error_fname, char *param_fname, char **labels,
	 int adjust_priors, float *priors_list,
	 int *nnodes, int cost_type, float cost_param, int log_outputs)
{
    
    /* Local variables */
    int		model, err_model, data_model, rv, normalization=0;
    int		desired_class, class, i;
    int		post_sum_to_one = 0;
    float	*inputs, *outputs;
    char	*tmp_path;
    FILE	*error_file, *param_file;
    
    int		nlayers;	/* number of layers */
    int		layers[20];	/* number of nodes in each layer */
    
    strcpy(prog_name, get_program_name());
    
    
    nlayers = nnodes[0];	/* Put the number of nodes in each layer in
				 * nlayers */
    for (i = 0; i < nlayers; i++)
      layers[i] = nnodes[i + 1];
    
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on normalization parameter file", tmp_path);
    
    model = mlp_read_model(param_file);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s",
		  prog_name, model, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of model file");
    
    /* use features from the model */
    features = *(int **) mlp_param(model, "features", 0);
    ninputs = *(int *) mlp_param(model, "ninputs", 0);

    display_model_params (model, ninputs, normalization, features, verbosity);
    
    outputs = *(float **) mlp_param(model, "output", 0);
    
    if (verror > 0) {	/* open file for classification errors */
	if(create_flag)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    } else error_file = NULL;
    /* do remaining initializations and allocations */
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, map_flag, ninputs, cost_type, log_outputs, post_sum_to_one);
    init_cost(cost_type, cost_param);

    /* set up for reading data */
    data_model =
      class_data_single_init(data_path, data_fname, npatterns,
			     nraw, ninputs, noutputs, features,
			     normalization, verbosity);
    
    if(adjust_priors)
      mlp_test_init(model, adjust_priors, priors_list);
    
    /* read and classify input from data file */
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
	/* classify pattern and report errors */
	class = mlp_classify(model, inputs);
	report_class_trial(err_model, error_file, inputs, outputs,
			   desired_class, class,
			   class_data_get_input_num(data_model));
    }

    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_test_result(err_model, stdout, data_fname);
    report_free(err_model);
    
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    class_data_exit(data_model);
    mlp_free(model);
    norm_free();
}
