/***--------------------------------------------------------------------------
 * NAME:		mlp_train.c
 * PURPOSE:	training driver for MultiLayer Perceptron
 *
 * SEE ALSO: 	mlp_pack.c 
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_train.c,v 11.5 2001/08/23 20:12:32 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlp_train.c,v $
 * CREATED: 21-SEP-90 dan                             MIT LINCOLN LABORATORY
 *****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlp.h"

static char     prog_name[MAXSTR];	/* static variables */

void mlp_train(int create_flag, int map_flag, int nraw, int ninputs,
	       int *features, int noutputs, int normalize, int npatterns,
	       int verbosity, int verror, int nepochs, int random_flag,
	       char *data_path, char *data_fname, char *norm_fname,
	       char *exp_path, char *error_fname, char *param_fname,
	       char **labels, char *comment, int adjust_priors,
	       int priors_npatterns, float *priors_list, float *etta_list,
	       float alpha, float tolerance, float decay, float kappa,
	       float epsilon, float *sigmoid_param_list, int *nnodes,
	       int *batch, int cost_type, float cost_param, int etta_type,
	       int hfunction, int ofunction,
	       float max_magnitude, int log_outputs, float etta_nepochs)
{
    
    /* Local variables */
    int		model, err_model, err_part_model, data_model;
    int		rv, normalization=0, epoch;
    int		desired_class, class, etta_tau;
    int		n_read, i, n_trials, trial_num;
    float	*inputs, *outputs, *desired_outputs;
    char	*tmp_path, *open_type;
    int		post_sum_to_one = 0;
    FILE	*error_file = NULL, *param_file;
    
    int		nlayers;	/* number of layers */
    int		layers[20];	/* number of nodes in each layer */
    int		w_update_cnt = 0;
    int		w_update_int, w_update_max, w_update_increase;
    
    strcpy(prog_name, get_program_name());
    
    w_update_int = batch[1];
    w_update_max = batch[2];
    w_update_increase = batch[3];
    nlayers = nnodes[0];	/* Put the number of nodes in each layer in
				 * nlayers */
    for (i = 0; i < nlayers; i++)
      layers[i] = nnodes[i + 1];
    
    if (create_flag) {
	/* allocate and initialize model structure */
	model = mlp_build(nlayers, layers, etta_type);

	etta_tau = (int)((float)npatterns * etta_nepochs);
	mlp_init(model, max_magnitude, hfunction,
		 ofunction, etta_list, features, etta_tau);
	mlp_set_train_params(model, cost_type, cost_param, etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);

	if (normalize)
	  normalization = read_norm_file(path_build(data_path, norm_fname));
	else
	  norm_setup_none(nraw, noutputs);
    } else {		/* retrain */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s",
		      prog_name, rv,
		      "on normalization parameter file", tmp_path);
	
	model = mlp_read_model(param_file);
	if (model < 0)
	  fatal_error("%s: read error %d on model file %s", 
		      prog_name, model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of model file");
	verify_model_params (model, ninputs,
			     *(int *) mlp_param(model, "ninputs", 0),
			     features,
			     *(int **) mlp_param(model, "features", 0));
	/* if this is PC_PROC and we are retraining, we need to re-set the 
	   sigmoid for the final layer to STEP. It is currently LINEAR,
	   due to a hack we needed to pull !!!!*/
	if(cost_type == PC_PROC){
	  int nlayers = *(int *)mlp_param(model, "nlayers", 0);
	  int *sigm = (int *)mlp_param(model, "sigm_type", nlayers-1);
	  *sigm = STEP;
	}
	mlp_set_train_params(model, cost_type, cost_param, etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
    }
    /* use features from the model */
    features = *(int **) mlp_param(model, "features", 0);
    ninputs = *(int *) mlp_param(model, "ninputs", 0);
    outputs = *(float **) mlp_param(model, "output", 0);
    desired_outputs = allocate_vector(noutputs);

    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* set up for reporting the results */
    if (verror > 0) {	/* open file for classification errors */
	if (create_flag) open_type = "w";
	else open_type = "a";
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      open_type, "for classification errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, map_flag, ninputs, cost_type,
		log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				 ninputs, noutputs, features, normalization,
				 verbosity);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_priors)
      class_data_priors_init(data_model, priors_list, priors_npatterns);
    
    /* get the number of patterns per epoch */
    n_trials = class_data_get_npatterns(data_model);
    report_epoch_header(err_model, stdout, n_trials);
    
    for (epoch = 1; epoch <= nepochs; epoch++) {
	class_data_epoch_init(data_model);
	report_init(err_part_model, map_flag, ninputs,
		    cost_type, log_outputs, post_sum_to_one);

	w_update_cnt = 0;
	for (n_read = 0; n_read < n_trials; n_read++) {
	    /* get the next pattern */
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    trial_num = class_data_get_input_num(data_model);
	    /* train in batch mode or updating weights immediately */
	    if (w_update_int <= 1)
	      class = mlp_inc_ctrain(model, inputs, desired_class);
	    else {
		class = mlp_batch_ctrain(model, inputs, desired_class);
		if (++w_update_cnt > w_update_int) {
		    mlp_batch_update_weights(model);
		    w_update_cnt = 0;
		}
	    }
	    report_class_trial(err_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	    report_class_trial(err_part_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	    if(cost_type == ROC_COST){
		desired_outputs[desired_class] = 1.;
		roc_cost_train(outputs, desired_outputs);
		desired_outputs[desired_class] = 0.;
	    }
	}
	if (w_update_cnt > 0)
	  mlp_batch_update_weights(model);
	if (w_update_int < w_update_max)
	  w_update_int += w_update_increase;

	report_epoch(err_part_model, stdout, epoch);
	if(cost_type == ROC_COST){
	    roc_cost_finish_epoch();
	}
    }
    report_last_epoch(err_part_model, stdout, epoch-1);
    class_data_exit(data_model);

    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "mlp model\n");
    mlp_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting and close error file */
    /* only show the last epoch, not the overall errors */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    mlp_free(model);
    norm_free();
    deallocate_vector(desired_outputs);
}
