/***--------------------------------------------------------------------------
 * NAME:		mlpm_cv.c
 * PURPOSE:	cross validation driver for MultiLayer Perceptron Mapper
 *
 * SEE ALSO: 	mlpm_pack.c
 * GLOBALS:	int debug;
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlpm_cv.c,v 11.4 2001/08/23 20:12:33 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlpm_cv.c,v $
 * CREATED: 4-Mar-93 lck				MIT LINCOLN LABORATORY
 *****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlpm.h"

static char prog_name[MAXSTR];	/* static variables */

void
mlpm_cross_valid(int map_flag, int nraw, int ninputs, int *features,
		 int noutputs, int normalize, int npatterns, int verbosity,
		 int verror, int nepochs, int random_flag, int cv_nfolds,
		 int cv_nsplits, int use_cv_splits, int *cv_splits,
		 char **cv_train_mask, char **cv_test_mask, char *data_path,
		 char *data_fname, char *norm_fname, char *exp_path,
		 char *error_fname, char *param_fname, char **labels,
		 char *comment, float *etta_list, float alpha, float tolerance,
		 float decay, float kappa, float epsilon,
		 float *sigmoid_param_list, int *nnodes, int *batch,
		 int etta_type, int hfunction, int ofunction,
		 float max_magnitude, float etta_nepochs)
{
    /* Local variables */
    int		model, err_model, err_part_model, data_model;
    int		normalization=0, epoch, log_outputs = 0, post_sum_to_one = 0;
    int		etta_tau;
    int		n_read, i, n_trials, input_num;
    int		fold;
    float	*inputs, *outputs, *desired_outputs;
    char	*tmp_path;
    FILE	*error_file = NULL, *param_file;
    
    int		nlayers;	/* number of layers */
    int		layers[20];	/* number of nodes in each layer */
    int		w_update_cnt = 0;
    int		w_update_int, w_update_max, w_update_increase;
    int		cost_type = SQ_ERR;
    
    strcpy(prog_name, get_program_name());
    
    w_update_int = batch[1];
    w_update_max = batch[2];
    w_update_increase = batch[3];
    nlayers = nnodes[0];	/* Put the number of nodes in each layer in
				 * nlayers */
    for (i = 0; i < nlayers; i++)
      layers[i] = nnodes[i + 1];
    
    /* allocate and initialize model structure */
    model = mlp_build(nlayers, layers, etta_type);
    etta_tau = (int)((float)npatterns * etta_nepochs);
    mlp_init(model, max_magnitude, hfunction,
	     ofunction, etta_list, features, etta_tau);
    mlp_set_train_params(model, cost_type, 0., etta_list, alpha,
			 decay, kappa, epsilon, tolerance,
			 sigmoid_param_list);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    outputs = *(float **) mlp_param(model, "output", 0);
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);

    
    if (verror > 0) {	/* open file for classification errors */
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      "w", "for classification errors");
    } else error_file = NULL;
    /* do remaining initializations and allocations */
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, map_flag, ninputs, SQ_ERR, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);

    /* read data and initialize data handling routines */
    data_model = map_data_read(data_path, data_fname, npatterns,
			       nraw, ninputs, noutputs, features,
			       normalization, verbosity);
    map_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		     cv_nfolds, cv_train_mask, cv_test_mask);
    if(random_flag) map_data_random_init(data_model);

    /* store the normalization once only */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    
    for(fold = 0; fold < cv_nfolds; fold++) {
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	map_data_start_cv_fold(data_model, fold, TRUE);
	n_trials = map_data_get_npatterns(data_model);

	etta_tau = (int)((float)n_trials * etta_nepochs);
	mlp_init(model, max_magnitude, hfunction,
		 ofunction, etta_list, features, etta_tau);
	mlp_set_train_params(model, cost_type, 0., etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
	
	/* verror for training is 0 */
	report_init(err_part_model, map_flag, ninputs, cost_type,
		    log_outputs, post_sum_to_one);
	report_epoch_header(err_part_model, stdout, n_trials);

	for (epoch = 1; epoch <= nepochs; epoch++) {
	    map_data_epoch_init(data_model);

	    w_update_cnt = 0;
	    for (n_read = 0; n_read < n_trials; n_read++) {
		if((inputs = map_data_get_next(data_model, &desired_outputs))
		   == NULL) break;
		input_num = map_data_get_input_num(data_model);
		if (w_update_int <= 1)
		  mlp_inc_train(model, inputs, desired_outputs);
		else {
		    mlp_batch_train(model, inputs, desired_outputs);
		    if (++w_update_cnt > w_update_int) {
			mlp_batch_update_weights(model);
			w_update_cnt = 0;
		    }
		}
		report_map_trial(err_part_model, error_file, inputs,
				 outputs, desired_outputs, input_num);
	    } /* end reading data */
	    if (w_update_cnt > 0)
	      mlp_batch_update_weights(model);
	    if (w_update_int < w_update_max)
	      w_update_int += w_update_increase;
	    report_epoch(err_part_model, stdout, epoch);
	    report_init(err_part_model, map_flag, ninputs, cost_type,
			log_outputs, post_sum_to_one);
	} /* end epoch */
	
	/* print out the mlpm model */
	fprintf(param_file, "mlpm model %d\n", fold);
	mlp_write_model(model, param_file);
	if ( verbosity > 0)
	  printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
	
	/* reinitialize model for testing */
	report_init(err_part_model, map_flag, ninputs, cost_type,
		    log_outputs, post_sum_to_one);
	
	map_data_start_cv_fold(data_model, fold, FALSE);
	n_trials = map_data_get_npatterns(data_model);
	/* test */

	for (n_read = 0; n_read < n_trials; n_read++) {
	    if((inputs = map_data_get_next(data_model, &desired_outputs))
	       == NULL) break;
	    input_num = map_data_get_input_num(data_model);
	    mlp_eval(model, inputs);
	    report_map_trial(err_model, error_file, inputs, outputs,
			     desired_outputs, input_num);
	    report_map_trial(err_part_model, error_file, inputs, outputs,
			     desired_outputs, input_num);
	}
	report_overall(err_part_model, stdout, error_fname, labels);
    } /* end fold */
    
    /* finish up writing the models */
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);

    map_data_exit(data_model);
    
    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);

    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    mlp_free(model);
    norm_free();
}
