/***--------------------------------------------------------------------------
 * NAME:		mlpm_train.c
 * PURPOSE:	training driver for Mapping MultiLayer Perceptron
 *
 * SEE ALSO: 	mlp/mlp_pack.c
 *
 * FUNCTIONS:
 *	mlp	driver for MultiLayer Perceptron
 *		handles model creation, training and testing of data
 *		Assumes inputs in form of desired output followed by input
 *		Constant node of 1 (bias) appended to each layer
 *		First layer is linear, other layers are user definable
 *		error file contains sample number, rmserr
 *		Network inputs followed by outputs for each sample
 *			optionally sent to error file
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlpm_train.c,v 11.4 2001/08/23 20:12:33 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/mlp/mlpm_train.c,v $
 * CREATED: 13-Jul-92				MIT LINCOLN LABORATORY
 *****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#define MAXSTR (256)

#include "mlp_pack.h"
#include "mlpm.h"

static char     prog_name[MAXSTR];	/* static variables */

void
mlpm_train(int create_flag, int db_type, int nraw, int ninputs, int *features,
	   int noutputs, int normalize, int npatterns, int verbosity,
	   int verror, int nepochs, int random_flag, char *data_path,
	   char *data_fname, char *norm_fname, char *exp_path,
	   char *error_fname, char *param_fname, char **labels, char *comment,
	   float *etta_list, float alpha, float tolerance, float decay,
	   float kappa, float epsilon, float *sigmoid_param_list, int *nnodes,
	   int *batch, int etta_type, int hfunction, int ofunction,
	   float max_magnitude, float etta_nepochs)
{
    /* Local variables */
    int		model, err_model, err_part_model, data_model, rv;
    int		normalization=0, epoch, log_outputs = 0, post_sum_to_one = 0;
    int		n_read, i, n_trials, input_num, etta_tau;
    float	*inputs, *outputs, *desired_outputs;
    char	*tmp_path, *open_type;
    FILE	*error_file = NULL, *param_file;
    
    int		nlayers;	/* number of layers */
    int		layers[20];	/* number of nodes in each layer */
    int		w_update_cnt = 0;
    int		w_update_int, w_update_max, w_update_increase;
    int		cost_type = SQ_ERR;
    
    strcpy(prog_name, get_program_name());
    
    w_update_int = batch[1];
    w_update_max = batch[2];
    w_update_increase = batch[3];
    nlayers = nnodes[0];	/* Put the number of nodes in each layer in
				 * nlayers */
    for (i = 0; i < nlayers; i++)
      layers[i] = nnodes[i + 1];
    
    if (create_flag) {
	/* allocate and initialize model structure */
	model = mlp_build(nlayers, layers, etta_type);
	etta_tau = (int)((float)npatterns * etta_nepochs);
	mlp_init(model, max_magnitude, hfunction,
		 ofunction, etta_list, features, etta_tau);
	mlp_set_train_params(model, cost_type, 0., etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
	if (normalize)
	  normalization = read_norm_file(path_build(data_path, norm_fname));
	else
	  norm_setup_none(nraw, noutputs);
    } else {		/* retrain */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s",
		      prog_name, rv,
		      "on normalization parameter file", tmp_path);

	model = mlp_read_model(param_file);
	if (model < 0)
	  fatal_error("%s: read error %d on model file %s", 
		      prog_name, model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of model file");
	verify_model_params (model, ninputs,
			     *(int *) mlp_param(model, "ninputs", 0),
			     features,
			     *(int **) mlp_param(model, "features", 0));
	mlp_set_train_params(model, cost_type, 0., etta_list, alpha,
			     decay, kappa, epsilon, tolerance,
			     sigmoid_param_list);
    }
    outputs = *(float **) mlp_param(model, "output", 0);

    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);

    if (verror > 0) {	/* open file for errors */
	if (create_flag) open_type = "w";
	else open_type = "a";
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      open_type, "for mapping errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build(verbosity, verror, noutputs);
    report_init(err_model, db_type, ninputs, cost_type, log_outputs, post_sum_to_one);
    err_part_model = report_build(verbosity, 0, noutputs);
    report_init(err_part_model, db_type, ninputs, cost_type, log_outputs, post_sum_to_one);

    /* read data and initialize data handling routines */
    data_model = map_data_read(data_path, data_fname, npatterns,
			       nraw, ninputs, noutputs, features,
			       normalization, verbosity);
    if(random_flag) map_data_random_init(data_model);

    /* get the number of patterns per epoch */
    n_trials = map_data_get_npatterns(data_model);
    report_epoch_header(err_model, stdout, n_trials);

    for (epoch = 1; epoch <= nepochs; epoch++) {
	map_data_epoch_init(data_model);

	w_update_cnt = 0;
	for (n_read = 0; n_read < n_trials; n_read++) {
	    if((inputs = map_data_get_next(data_model, &desired_outputs))
	       == NULL) break;
	    input_num = map_data_get_input_num(data_model);
	    if (w_update_int <= 1)
	      mlp_inc_train(model, inputs, desired_outputs);
	    else {
		mlp_batch_train(model, inputs, desired_outputs);
		if (++w_update_cnt > w_update_int) {
		    mlp_batch_update_weights(model);
		    w_update_cnt = 0;
		}
	    }
	    report_map_trial(err_model, error_file,
			     inputs, outputs, desired_outputs, input_num);
	    report_map_trial(err_part_model, error_file,
			     inputs, outputs, desired_outputs, input_num);
	}
	if (w_update_cnt > 0)
	  mlp_batch_update_weights(model);
	if (w_update_int < w_update_max)
	  w_update_int += w_update_increase;

	report_epoch(err_part_model, stdout, epoch);
	report_init(err_part_model, db_type, ninputs, cost_type, log_outputs, post_sum_to_one);
    }
    map_data_exit(data_model);
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "mlp model\n");
    mlp_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    report_free(err_part_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    norm_free();
    mlp_free(model);
}
