/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
                                /* nbayes_augment.c */
/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#define nbayes_pack_h_internal_types
#include "nbayes_pack.h"

static char *orig_param_fname = "";
static char *new_param_fname = "";
static char *aug_param_fname = "";
static float weight = -1.;
int ignore_unseen = 0;
int do_unseen = 0;
int unity_laplace = 0;

static ARG argtab[] = 
{
    {"original_model", STRING, (int *)&orig_param_fname, "model to be augmented"},
    {"new_model", STRING, (int *)&new_param_fname, "model of new data"},
    {"augmented_model", STRING, (int *)&aug_param_fname, "resulting model to create"},
    {"weight", FLOAT, (int *)&weight, "counts are (1-weight)*old + weight*new"},
    {"ignore_unseen", FLAG, &ignore_unseen, "Ignore test inputs with unknown values. Default is use laplace correction"},
    {"unity_laplace", FLAG, &unity_laplace, "Laplace correction is 1. Default is 1/Npat"},
};
#define TABSIZE (sizeof(argtab)/sizeof(ARG))

struct shared_params {
  int normalization;
  int *features;
  int nraw;
  int ninputs;
  int noutputs;
};

void check_norm(struct shared_params shared_orig,
		struct shared_params shared_new,
		char *prog_name, char *orig_name, char *new_name);
void combine_probs(int orig_model, int new_model, int aug_model, float weight);
int read_model_file(char *param_fname, struct shared_params *param_struct,
		    char *prog_name);
void write_model_file(char *orig_name, char *new_name, char *aug_fname,
		      float weight, struct shared_params shared_data,
		      int model);
static int build_aug_bins(int orig_model, int new_model);

int
main(int argc, char **argv)
{
    int orig_model, new_model, aug_model;
    struct shared_params shared_orig, shared_new;

    /* process arguments */
    argc = nnargs(argc, argv, argtab, TABSIZE);

    if(weight >= 0.){
      if(weight > 1.){
	fatal_error("%s: weight (%f) must be between 0 and 1",
		    argv[0], weight);
      }
    }
    do_unseen = ignore_unseen ? 0 : 1;

    /* read in parameters */
    orig_model = read_model_file(orig_param_fname, &shared_orig, argv[0]);
    norm_free();
    new_model = read_model_file(new_param_fname, &shared_new, argv[0]);

    /* check that the normalization and feature selection is the same for
       both models */
    check_norm(shared_orig, shared_new, argv[0], orig_param_fname,
	       new_param_fname);
    
    /* set up the bin ends and allocate the augmented model */
    aug_model = build_aug_bins(orig_model, new_model);

    /* figure out combined likelihoods */
    combine_probs(orig_model, new_model, aug_model, weight);

    /* write augmented model */
    write_model_file(orig_param_fname, new_param_fname, aug_param_fname,
		     weight, shared_orig, aug_model);

    /* free memory */
    norm_free();
    nbayes_free(new_model);
    nbayes_free(orig_model);
    nbayes_free(aug_model);

    exit(0);
}

int
read_model_file(char *param_fname, struct shared_params *param_struct, char *prog_name)
{
    int   model, rv, normalization;
    int	  nraw;
    FILE *param_file;
    int *features, ninputs;
    int i;

    param_file = nn_fopen(param_fname, "r", "for read of model file");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on normalization parameter file", param_fname);
    param_struct->normalization = normalization;

    rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on feature selection parameters in file", param_fname);
    param_struct->nraw = nraw;
    param_struct->ninputs = ninputs;
    param_struct->features = allocate_ivector(features[0]+1);
    for(i = 0; i <= features[0]; i++){
      param_struct->features[i] = features[i];
    }
    
    model = nbayes_read_model(param_file);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s",
		  prog_name, model, param_fname);
    nn_fclose(param_file, param_fname, "for read of model file");

    param_struct->noutputs = *(int *)nbayes_param(model, "nclasses", 0, 0);

    return((int) model);
}

void
write_model_file(char *orig_name, char *new_name, char *aug_fname,
		 float weight, struct shared_params shared_data, int model)
{
  FILE *param_file;
  int *features, nraw, ninputs;

  features = shared_data.features;
  nraw = shared_data.nraw;
  ninputs = shared_data.ninputs;
  
  param_file = nn_fopen(aug_fname, "w", "for write of augmented model");
  if(weight >= 0.){
    fprintf(param_file, "model is %f*%s + %f*%s\n",
	    (1.-weight), orig_name, weight, new_name);
  } else {
    fprintf(param_file, "model is %s + %s\n", new_name, orig_name);
  }
  fprintf(param_file, "normalization data\n");
  norm_write_paramfile(param_file);
  fprintf(param_file, "feature selection\n");
  feature_write_paramfile(param_file, features, nraw, ninputs);
  fprintf(param_file, "nbayes model\n");
  nbayes_write_model(model, param_file);
  nn_fclose(param_file, aug_fname, "for write of model file");
}

void check_norm(struct shared_params shared_orig,
		struct shared_params shared_new,
		char *prog_name, char *orig_name, char *new_name)
{
  int i;

  if(shared_orig.normalization != shared_new.normalization){
    fatal_error("%s: Normalization parameters in %s and %s do not match",
		prog_name, orig_name, new_name);
  }
  if(shared_orig.nraw != shared_new.nraw){
    fatal_error("%s: Number of raw input features in %s and %s do not match",
		prog_name, orig_name, new_name);
  }
  if(shared_orig.ninputs != shared_new.ninputs){
    fatal_error("%s: Number of ninputs in %s and %s do not match",
		prog_name, orig_name, new_name);
  }
  if(shared_orig.features[0] != shared_new.features[0]){
    fatal_error("%s: Feature selection parameters in %s and %s do not match",
		prog_name, orig_name, new_name);
  } else {
    for(i = 0; i <= shared_orig.features[0]; i++){
      if(shared_orig.features[i] != shared_new.features[i]){
	fatal_error("%s: Feature selection parameters in %s and %s do not match",
		    prog_name, orig_name, new_name);
      }
    }
  }
}

static int build_aug_bins(int orig_model, int new_model)
{
  int *orig_bins, *new_bins, *aug_bins;
  int i, j, top;
  float *orig_ends, *new_ends, *aug_ends;
  int aug_model;
  int ninputs, noutputs;
  
  ninputs = *(int *)nbayes_param(orig_model, "ninputs", 0, 0);
  noutputs = *(int *)nbayes_param(orig_model, "noutputs", 0, 0);

  /* check that the bin ends match at the lower end of the scales */
  orig_bins = *(int **)nbayes_param(orig_model, "nbins", 0, 0);
  new_bins = *(int **)nbayes_param(new_model, "nbins", 0, 0);
  aug_bins = allocate_ivector(ninputs);
  for(i = 0; i < ninputs; i++){
    if(orig_bins[i] < new_bins[i]){
      top = orig_bins[i];
      aug_bins[i] = new_bins[i];
    } else {
      top = new_bins[i];
      aug_bins[i] = orig_bins[i];
    }
    orig_ends = *(float **)nbayes_param(orig_model, "bin", i, 0);
    new_ends = *(float **)nbayes_param(new_model, "bin", i, 0);
    for(j = 0; j < top; j++){
      if(orig_ends[j] != new_ends[j]){
	fatal_error("Bin ends do not match for input feature %d.",i);
      }
    }
  }
  aug_model = nbayes_build(ninputs, noutputs, aug_bins);

  /* copy the bin ends we want into the augmented model we are building */
  for(i = 0; i < ninputs; i++){
    /* figure out bin ends for the augmented model */
    aug_ends = *(float **)nbayes_param(aug_model, "bin", i, 0);
    if(orig_bins[i] < new_bins[i]){
      new_ends = *(float **)nbayes_param(new_model, "bin", i, 0);
    } else {
      new_ends = *(float **)nbayes_param(orig_model, "bin", i, 0);
    }
    for(j = 0; j <= aug_bins[i]; j++){
      aug_ends[j] = new_ends[j];
    }
  }
  return aug_model;
}

void combine_probs(int orig_model, int new_model, int aug_model, float weight)
{
  /* retrieve count from models */
  /*  get N, Nc, nbins, and probs.
      undo algebra to get "count". This might not be an integer if
      we have applied weights during a previous augmentation */
  /* p = (Ncij + 1/N)/(Nc + (NBINSi+1)/N) */
  /* Ncij = p*(Nc + (NBINSi+1)/N) - 1/N */
  int i, j, c;
  float *n_class_orig, n_total_orig, *n_class_new, n_total_new;
  float *n_class_aug, n_total_aug, *pn_total_aug;
  int *orig_bins, *new_bins, *aug_bins;
  float *orig_probs, *new_probs, *aug_probs;
  float norig, nnew, orig_count, new_count;
  int ninputs, noutputs;
  float orig_weight;
  
  ninputs = *(int *)nbayes_param(aug_model, "ninputs", 0, 0);
  noutputs = *(int *)nbayes_param(aug_model, "noutputs", 0, 0);

  orig_bins = *(int **)nbayes_param(orig_model, "nbins", 0, 0);
  new_bins = *(int **)nbayes_param(new_model, "nbins", 0, 0);
  aug_bins = *(int **)nbayes_param(aug_model, "nbins", 0, 0);

  n_class_orig = *(float **)nbayes_param(orig_model, "class_cnt", 0, 0);
  n_total_orig = *(float *)nbayes_param(orig_model, "total_npatterns", 0, 0);
  n_class_new = *(float **)nbayes_param(new_model, "class_cnt", 0, 0);
  n_total_new = *(float *)nbayes_param(new_model, "total_npatterns", 0, 0);
  n_class_aug = *(float **)nbayes_param(aug_model, "class_cnt", 0, 0);
  pn_total_aug = (float *)nbayes_param(aug_model, "total_npatterns", 0, 0);

  if(weight >= 0.){
    orig_weight = 1.-weight;
  } else {
    orig_weight = weight = 1;
  }
  for(c = 0; c < noutputs; c++){
    n_class_aug[c] = orig_weight*n_class_orig[c] + weight*n_class_new[c];
  }
  *pn_total_aug = n_total_aug = orig_weight*n_total_orig + weight*n_total_new;

  for(i = 0; i < ninputs; i++){
    for(j = 0; j <= aug_bins[i]; j++){
      /* we have run out of original data */
      if(j > orig_bins[i]){
	orig_probs = *(float **)nbayes_param(orig_model, "prob", i,
					     orig_bins[i]);
      } else {
	orig_probs = *(float **)nbayes_param(orig_model, "prob", i, j);
      }
      if(j > new_bins[i]){
	new_probs = *(float **)nbayes_param(new_model, "prob", i,
					    new_bins[i]);
      } else {
	new_probs = *(float **)nbayes_param(new_model, "prob", i, j);
      }
      aug_probs = *(float **)nbayes_param(aug_model, "prob", i, j);
      for(c = 0; c < noutputs; c++){
	orig_count = nn_exp10(orig_probs[c]);
	new_count = nn_exp10(new_probs[c]);
	if(unity_laplace){
	  norig = orig_count* (n_class_orig[c] + orig_bins[i]+do_unseen) - 1.;
	  nnew = new_count*(n_class_new[c] + new_bins[i]+do_unseen) - 1.;
	} else {
	  norig = orig_count*(n_class_orig[c] +
			      (orig_bins[i]+do_unseen)/n_total_orig)
	    - 1./n_total_orig;
	  nnew = new_count*(n_class_new[c] + 
			    (new_bins[i]+do_unseen)/n_total_new)
	    - 1./n_total_new;
	}
	/* scale new counts by weights */
	norig *= orig_weight;
	nnew *= weight;
	/* add models to combine them */
	aug_probs[c] = norig+nnew;
	/* calculate new floor and apply it */
	if(unity_laplace){
	  aug_probs[c] = (aug_probs[c] + do_unseen)/
	    (n_class_aug[c] + aug_bins[i]+do_unseen);
	} else {
	  aug_probs[c] = (aug_probs[c] + do_unseen/n_total_aug)/
	    (n_class_aug[c] + (aug_bins[i]+do_unseen)/n_total_aug);
	}
	aug_probs[c] = log10(aug_probs[c]);
      }
    }
  }
}
