/***---------------------------------------------------------------------------
* NAME:         nbayes_pack.c
* PURPOSE:      Functions specific to nbayes.c
*
* SEE ALSO:	nbayes.c nbayes_main.c
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*	nbayes_build		Allocate storage for model parameters.
*	nbayes_init		Initialize model parameters.
*	nbayes_free		Deallocates model storage
*	nbayes_read_model	Initializes model from file
*	nbayes_write_model	Writes model to output file
*	nbayes_param		Returns pointers to elements in model
*	nbayes_inc_ctrain	Does training on a vector
*	nbayes_classify	Classifies a vector using nbayes algortihm
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/nbayes/nbayes_pack.c,v 1.5 2001/08/23 20:12:36 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/nbayes/nbayes_pack.c,v $
* CREATED: 16-Oct-00 lck				MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 2000 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#define nbayes_pack_h_internal_types
#define nbayes_pack_h_static_decl
#include "nbayes_pack.h"

				/* Internal Globals */
static NBAYES_PARAMS **nbpp;/* array of parameters for each model */
static int      max_models = 0;	/* current length of nbpp */
#define MIN_PROB (1e-9)

/*****************************************************************************
* NAME: nbayes_build(ninputs,noutputs)
*	ninputs - number of dimensions in training vectors
*	noutputs - number of classes
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
*****************************************************************************/
int
nbayes_build(int ninputs, int noutputs, int *nbins)
{
    register int    new_model;
    int i, b;
    NBAYES_PARAMS *hpp;
    			/* find entry in model array for new model */
    new_model =
      model_get_number((void ***)&nbpp,
		       &max_models, sizeof(NBAYES_PARAMS));
    hpp = nbpp[new_model];

    /* set fields which are constants */
    hpp->ninputs = ninputs;
    hpp->noutputs = noutputs;

    /* allocate space for vectors */
    hpp->class_cnt = allocate_vector(noutputs);
    hpp->nbins = allocate_ivector(ninputs);
    for(i = 0; i < ninputs; i++) hpp->nbins[i] = nbins[i];

    /* set vectors that will not be allocated here to NULL */
    hpp->bin_width = NULL;
    hpp->bin_min = hpp->bin_max = NULL;
    hpp->log_class_priors = hpp->outputs = NULL;

    /* allocate space for probs and bin ends */
    hpp->prob_list = (float ***)allocate_pvector(ninputs);
    hpp->bin_ends = (float **)allocate_pvector(ninputs);
    for(i = 0; i < ninputs; i++){
      hpp->bin_ends[i] = allocate_vector(nbins[i]+1);
      hpp->prob_list[i] = (float **)allocate_pvector(nbins[i]+1);
      for(b = 0; b <= nbins[i]; b++){
	hpp->prob_list[i][b] = allocate_vector(noutputs);
      }
    }

    return new_model;
}
/*****************************************************************************
* NAME: nbayes_train_init()
* DESCRIPTION:
*       Initialize model for training
*
* EXAMPLE:
*
*****************************************************************************/
void
nbayes_train_init(int model, int ignore_unseen, int unity_laplace)
{
    int c, i, b;
    NBAYES_PARAMS *hpp;

    hpp = nbpp[model];

    if(!hpp->bin_min) hpp->bin_min = allocate_vector(hpp->ninputs);
    if(!hpp->bin_max) hpp->bin_max = allocate_vector(hpp->ninputs);
    if(!hpp->bin_width) hpp->bin_width = allocate_vector(hpp->ninputs);

    for (c = 0; c < hpp->noutputs; c++)
      hpp->class_cnt[c] = 0.;
    hpp->total_npatterns = 0.;

    for(i = 0; i < hpp->ninputs; i++){
      for(b = 0; b <= hpp->nbins[i]; b++){
	for(c = 0; c < hpp->noutputs; c++){
	  hpp->prob_list[i][b][c] = 0.;
	}
      }
      hpp->bin_min[i] = MAXFLOAT;
      hpp->bin_max[i] = -(MAXFLOAT);
    }
    hpp->ignore_unseen = ignore_unseen;
    hpp->unity_laplace = unity_laplace;
}
/*****************************************************************************
* NAME: nbayes_test_init()
* DESCRIPTION:
*	initialize model for testing
*
* EXAMPLE:
*****************************************************************************/
void
nbayes_test_init(int model, int adjust_priors, float *priors_list)
{
    int c;
    NBAYES_PARAMS *hpp;

    hpp = nbpp[model];

    if(adjust_priors){
	for(c = 0; c < hpp->noutputs; c++)
	  hpp->log_class_priors[c] = log10(priors_list[c]);
    } else {
      for(c = 0; c < hpp->noutputs; c++){
	if(hpp->class_cnt[c] != 0.){
	  hpp->log_class_priors[c] = log10(hpp->class_cnt[c]/
				       (double)hpp->total_npatterns);
	} else {
	  hpp->log_class_priors[c] = 0.;
	}
      }
    }
}
/*****************************************************************************
* NAME: nbayes_test_alloc()
* DESCRIPTION:
*	initialize model for testing
*
* EXAMPLE:
*****************************************************************************/
void
nbayes_test_alloc(int model)
{
    NBAYES_PARAMS *hpp;

    hpp = nbpp[model];

    if(hpp->outputs == NULL)
      hpp->outputs = allocate_vector(hpp->noutputs);
    if(hpp->log_class_priors == NULL)
      hpp->log_class_priors = allocate_vector(hpp->noutputs);
}

/*****************************************************************************
* NAME: nbayes_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
*****************************************************************************/
void
nbayes_free(int model)
{
    NBAYES_PARAMS *hpp;
    int i, b;

    hpp = nbpp[model];

    /* free vectors allocated in build */
    deallocate_vector(hpp->class_cnt);

    /* free train vectors */
    if(hpp->bin_min) deallocate_vector(hpp->bin_min);
    if(hpp->bin_max) deallocate_vector(hpp->bin_max);

    /* free test vectors */
    if(hpp->outputs != NULL) deallocate_vector(hpp->outputs);
    if(hpp->log_class_priors != NULL) deallocate_vector(hpp->log_class_priors);

    /* free prob/bin structures */
    for(i = 0; i < hpp->ninputs; i++){
      for(b = 0; b <= hpp->nbins[i]; b++){
	deallocate_vector(hpp->prob_list[i][b]);
      }
      deallocate_pvector((void *)hpp->prob_list[i]);
      deallocate_vector(hpp->bin_ends[i]);
    }
    deallocate_pvector((void *)hpp->prob_list);
    deallocate_pvector((void *)hpp->bin_ends);
    deallocate_ivector(hpp->nbins);

    /* free structure itself */
    nn_free((char *) hpp, "for nbayes parameter structure");
    nbpp[model] = NULL;
}

/* Parameter Tables */
static PARAM paramtab[] = {
    {"ninputs", P_INT},
    {"noutputs", P_INT},
    {"total_npat", P_FLOAT},
};
#define TABSIZE (sizeof(paramtab)/sizeof(PARAM))

static PARAM paramtab_l[] = {
    {"class_cnt", P_FLOAT_VECTOR},
    {"nbins", P_INT_VECTOR},
};
#define TABSIZE_L (sizeof(paramtab_l)/sizeof(PARAM))

static PARAM paramtab_m[] = {
    {"bin_ends", P_FLOAT_VECTOR},
};
#define TABSIZE_M (sizeof(paramtab_m)/sizeof(PARAM))

static PARAM paramtab_mm[] = {
    {"prob_list", P_FLOAT_VECTOR},
};
#define TABSIZE_MM (sizeof(paramtab_mm)/sizeof(PARAM))

/******************************
 * routines to set param variables for reading and writing a parameter file
 *****************************/
static void
nbayes_set_paramtab(NBAYES_PARAMS *hpp)
{
    param_setvarptr("ninputs", &hpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &hpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("total_npat",(int *)&hpp->total_npatterns,
		    paramtab, TABSIZE);
}
static void
nbayes_set_paramtab_l(NBAYES_PARAMS *hpp)
{
    param_setvarptr("class_cnt", (int *)&hpp->class_cnt[0],
		    paramtab_l, TABSIZE_L);
    param_setdim("class_cnt", hpp->noutputs, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("nbins", (int *)&hpp->nbins[0],
		    paramtab_l, TABSIZE_L);
    param_setdim("nbins", hpp->ninputs, 0, paramtab_l, TABSIZE_L);
}

static void
nbayes_set_paramtab_m(float *bin_ends, int len)
{
  param_setvarptr("bin_ends", (int *)bin_ends, paramtab_m, TABSIZE_M);
  param_setdim("bin_ends", len, 0, paramtab_m, TABSIZE_M);
}

static void
nbayes_set_paramtab_mm(float *prob_list, int len)
{
  param_setvarptr("prob_list", (int *)prob_list, paramtab_mm, TABSIZE_MM);
  param_setdim("prob_list", len, 0, paramtab_mm, TABSIZE_MM);
}

/*****************************************************************************
* NAME: nbayes_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*       allocates space for and loads data into vectors and matricies
*
*       On error from param_getblk, returns the error value
*       (there is already an error check on the allocation, so we don't
*       have to do that here)
*
* EXAMPLE:
*****************************************************************************/
int
nbayes_read_model(FILE *fp)
{
    int             model;
    int             rv;
    NBAYES_PARAMS temp_hp, *hpp;
    int c, i, b;

    nbayes_set_paramtab(&temp_hp);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0)
	return rv;

    temp_hp.class_cnt = allocate_vector(temp_hp.noutputs);
    temp_hp.nbins = allocate_ivector(temp_hp.ninputs);
    nbayes_set_paramtab_l(&temp_hp);

    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0)
	return rv;

    /* allocate space for model and vectors */
    model = nbayes_build(temp_hp.ninputs, temp_hp.noutputs, temp_hp.nbins);
    hpp = nbpp[model];
    hpp->total_npatterns = temp_hp.total_npatterns;
    for(c = 0; c < hpp->noutputs; c++){
	hpp->class_cnt[c] = temp_hp.class_cnt[c];
    }

    /* read the probabilities */
    for(i = 0; i < hpp->ninputs; i++){
      for(b = 0; b <= hpp->nbins[i]; b++){
	nbayes_set_paramtab_mm(hpp->prob_list[i][b], hpp->noutputs);
	if((rv = param_getblk(fp, paramtab_mm, TABSIZE_MM)) < 0)
	  return rv;
      }
    }
    /* read the bin ends */
    for(i = 0; i < hpp->ninputs; i++){
      nbayes_set_paramtab_m(hpp->bin_ends[i], hpp->nbins[i]+1);
      if((rv = param_getblk(fp, paramtab_m, TABSIZE_M)) < 0)
	return rv;
    }
	
    deallocate_vector(temp_hp.class_cnt);
    deallocate_ivector(temp_hp.nbins);

    return (model);
}

/*****************************************************************************
* NAME: nbayes_write_model()
* DESCRIPTION:
*       writes parameters and exemplars for a model to a file
*       in ASCII format
*
* EXAMPLE:
*
******************************************************************************/
int
nbayes_write_model(int model, FILE *fp)
{
    int             rv;
    NBAYES_PARAMS *hpp;
    int i, b;

    hpp = nbpp[model];
    				/* save scalars */
    nbayes_set_paramtab(hpp);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0)
	return rv;

    				/* save vectors and matrices */
    nbayes_set_paramtab_l(hpp);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0)
	return rv;

    /* save probabilities */
    for(i = 0; i < hpp->ninputs; i++){
      for(b = 0; b <= hpp->nbins[i]; b++){
	fprintf(fp, "LOG LIKELIHOODS FOR INPUT FEATURE %d, BIN %d\n",
		i, b);
	nbayes_set_paramtab_mm(hpp->prob_list[i][b], hpp->noutputs);
	if((rv = param_putblk(fp, paramtab_mm, TABSIZE_MM)) < 0)
	  return rv;
      }
    }
    /* save bin starts */
    for(i = 0; i < hpp->ninputs; i++){
      fprintf(fp, "BIN_ENDS FOR INPUT %d\n", i);
      nbayes_set_paramtab_m(hpp->bin_ends[i], hpp->nbins[i]+1);
      if((rv = param_putblk(fp, paramtab_m, TABSIZE_M)) < 0)
	return rv;
    }
    return (0);
}


/************************************************************************
* NAME: nbayes_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/

void *
nbayes_param(int model, char *param_name, int input, int bin)
{
    void           *param_address;
    NBAYES_PARAMS *hpp;

    hpp = nbpp[model];

    if (PEQ("ninputs"))
	param_address = &hpp->ninputs;
    else if (PEQ("noutputs") || PEQ("nclasses"))
	param_address = &hpp->noutputs;
    else if (PEQ("outputs"))
	param_address = &hpp->outputs;
    else if (PEQ("class_cnt"))
      param_address = &hpp->class_cnt;
    else if (PEQ("total_npatterns"))
      param_address = &hpp->total_npatterns;
    else if (PEQ("nbins"))
      param_address = &hpp->nbins;
    else if (PEQ("prob"))
      param_address = &hpp->prob_list[input][bin];
    else if (PEQ("bin"))
      param_address = &hpp->bin_ends[input];
    else {
	warning("nbayes_param: Parameter %s not defined", param_name);
	param_address = NULL;
    }
    return param_address;
}


/*****************************************************************************
* NAME: nbayes_inc_train1()
* DESCRIPTION:
*	Count the number of patterns per class,
*	find the width of each class,
*	sum up the means and variances for each Gaussian
* EXAMPLE:
*
*****************************************************************************/
void
nbayes_inc_train1(int model, float *inputs, int desired_class)
{
    NBAYES_PARAMS *hpp;
    int i;

    hpp = nbpp[model];

    hpp->class_cnt[desired_class] += 1.;
    hpp->total_npatterns += 1.;
    /* find the range of the data */
    for(i = 0; i < hpp->ninputs; i++){
      if(hpp->bin_max[i] < inputs[i]) hpp->bin_max[i] = inputs[i];
      if(hpp->bin_min[i] > inputs[i]) hpp->bin_min[i] = inputs[i];
    }
    return;
}

/*****************************************************************************
* NAME: nbayes_finish_train1()
* DESCRIPTION:
*	Calculate means and variances for the Gaussian part of out answer
*	Find the constant number of patterns or constant width for each bin
* EXAMPLE:
*
*****************************************************************************/
void
nbayes_finish_train1(int model, int discreteData, float rangeFactor)
{
    NBAYES_PARAMS *hpp;
    int c, i, b;
    float bends, bin_width, half_width, middle;

    hpp = nbpp[model];

    for(c = 0; c < hpp->noutputs; c++){
      /* reset class_cnt for the next round */
      hpp->class_cnt[c] = 0.;
    }
    hpp->discreteData = discreteData;
    if(discreteData){
      bin_width = 1;
      for(i = 0; i < hpp->ninputs; i++){
	hpp->bin_width[i] = 1.;
	if(hpp->bin_max[i] > hpp->nbins[i]){
	  warning("There were more discrete values than bins to hold them.\nValues over %d will be ignored", hpp->nbins[i]);
	  hpp->bin_max[i] = hpp->nbins[i];
	}
	hpp->bin_min[i] = -.5;	/* always leave room at the bottom */
	for(b = 0, bends = -.5; b <= hpp->nbins[i];
	    b++, bends += 1){
	  hpp->bin_ends[i][b] = bends;
	}
      }
    } else {
      for(i = 0; i < hpp->ninputs; i++){
	/* adjust the bins sizes for continuous data, to get the stuff
	   on the outsides */
	half_width = (hpp->bin_max[i] - hpp->bin_min[i])/2.;
	middle = hpp->bin_min[i] + half_width;
	half_width *= rangeFactor;
	hpp->bin_min[i] = middle - half_width;
	hpp->bin_max[i] = middle + half_width;

	/* figure out the width of each bin */
	bin_width = (hpp->bin_max[i] - hpp->bin_min[i])/hpp->nbins[i];
	hpp->bin_width[i] = MAX(bin_width, MIN_PROB);
	for(b = 0, bends = hpp->bin_min[i]; b < hpp->nbins[i];
	    b++, bends += bin_width)
	  hpp->bin_ends[i][b] = bends;
	hpp->bin_ends[i][hpp->nbins[i]] = hpp->bin_max[i];
      }
    }

    /* reset total_npatterns for the next round */
    hpp->total_npatterns = 0;
    return;
}


/*****************************************************************************
* NAME: nbayes_inc_train2()
* DESCRIPTION:
*	Now that we know half of out probability equation, get the other half
* EXAMPLE:
*
*****************************************************************************/
void
nbayes_inc_train2(int model, float *inputs, int desired_class)
{
    NBAYES_PARAMS *hpp;
    int i, bin_num, nbins;

    hpp = nbpp[model];

    hpp->class_cnt[desired_class]++;
    hpp->total_npatterns += 1.;

    /* keep count of the patterns that fall in each bin */
    /* increment the number of patterns for the probability calculation */
    for(i = 0; i < hpp->ninputs; i++){
      nbins = hpp->nbins[i];
      if(hpp->discreteData){
	bin_num = nbayes_get_bin(inputs[i], hpp->bin_ends[i], nbins);
      } else {
	if(inputs[i] <= hpp->bin_ends[i][0]) bin_num = 0;
	else if (inputs[i] > hpp->bin_ends[i][nbins]) bin_num = nbins-1;
	else 
	  bin_num = nbayes_get_bin(inputs[i], hpp->bin_ends[i], nbins);
      }
      hpp->prob_list[i][bin_num][desired_class]++;
    }
    return;
}

/*****************************************************************************
* NAME: nbayes_finish_train2()
* DESCRIPTION:
*	Finish up the mean and variance calculations
*	Figure out the bin probabilities

if ignore_unseen is set, likelihood for bin Nbins[i] is 1 and nbins+1 below
	is nbins
if unity_laplace is set, L = 1
L = 1/N (if unity_laplace is set, L = 1)

       count[i,b,c] + L
       ---------------------
       number_in_c + (nbins+1)*L

*	P(x[i]|C) = (NC[j]+L)/((NC+nbins*L)*width)
*		where NC is the number of patterns in the class
*		and NC[j] is the number from class C that fall in the bin
*	The 1/N and nbins/N are correction terms in case a bin is empty
* EXAMPLE:
*
*****************************************************************************/
void
nbayes_finish_train2(int model)
{
    NBAYES_PARAMS *hpp;
    int c, i, b;
    float min_prob;
    int do_unseen;
    /* at least one pattern could fall in a bin and we somehow missed it */

    hpp = nbpp[model];
    do_unseen = (hpp->ignore_unseen) ? 0 : 1;

    min_prob = (hpp->unity_laplace)? 1. : 1./hpp->total_npatterns;

    for(i = 0; i < hpp->ninputs; i++){
      for(b = 0; b < hpp->nbins[i]+do_unseen; b++){
	for(c = 0; c < hpp->noutputs; c++){
	  hpp->prob_list[i][b][c] += min_prob;
	  hpp->prob_list[i][b][c] /= hpp->class_cnt[c] +
	    (hpp->nbins[i]+do_unseen)*min_prob;
	  hpp->prob_list[i][b][c] /= hpp->bin_width[i];
	  /* go to log domain */
	  hpp->prob_list[i][b][c] = log10(hpp->prob_list[i][b][c]);
	}
      }
      if(hpp->ignore_unseen){
	for(c = 0; c < hpp->noutputs; c++){
	  hpp->prob_list[i][hpp->nbins[i]][c] = 0;
	}
      }
    }
    return;
}
/*****************************************************************************
* NAME: nbayes_classify()
* DESCRIPTION:
*       Implements the nbayes classifier to return an integer class.
*
* EXAMPLE:
*
*****************************************************************************/
int
nbayes_classify(int model, float *inputs)
{
    NBAYES_PARAMS *hpp;
    int class, known;

    hpp = nbpp[model];

    known = nbayes_eval(model, inputs);
    if(known){
	class = max_pick(hpp->outputs, hpp->noutputs);
    } else class = UNKNOWN_CLASS;
    return class;
}

/*********************************************************************
 *
 * the outputs of this classifiers are posterior probabilities,
 * 	y(C) = P(C|x)
 **********************************************************************/
int
nbayes_eval(int model, float *inputs)
{
    NBAYES_PARAMS *hpp;
    int c, i, in_bin;
    int nbins, bin_num;

    hpp = nbpp[model];
    
    /* find the bin this pattern falls into */
    /* multiply up the probabilities */
    for(c = 0; c < hpp->noutputs; c++){
      hpp->outputs[c] = 0.;
    }
    in_bin = 0;
    for(i = 0; i < hpp->ninputs; i++){
      nbins = hpp->nbins[i];
      bin_num = nbayes_get_bin(inputs[i], hpp->bin_ends[i], nbins);
      if(bin_num != nbins){
	in_bin = 1;
      }
      for(c = 0; c < hpp->noutputs; c++){
	hpp->outputs[c] += hpp->prob_list[i][bin_num][c];
      }
    }
    for(c = 0; c < hpp->noutputs; c++){
      if(hpp->outputs[c] != 0){
	hpp->outputs[c] += hpp->log_class_priors[c];
      }
    }
    return in_bin;
}

static int
nbayes_get_bin(float x, float *bin_ends, int nbins)
{
    int b;

    if(x < bin_ends[0]) return nbins;
    if(x == bin_ends[0]) return 0;
    for(b = 0; b <= nbins && x > bin_ends[b]; b++);
    if(b > nbins) return nbins;
    else return b-1;
}
