/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/***-------------------------------------------------------------------------
* NAME: nc_class_pack.c
* PURPOSE:	radial basis function classification and prediction routines
* USAGE:
*
* SEE ALSO: 	nc_class.c
* DESCRIPTION:  Kmeans classifier.
*	nc_class_build()	constructs a nc_class model
*	nc_class_init()		initialize model for training
*	nc_class_free()		Deallocates model storage
*	nc_class_write_model()	writes a model to parameter file
*	nc_class_read_model()	reads a model from parameter file
*	nc_class_param()	returns value of parameter
*	nc_class_batch_ctrain()	train given class
*	nc_class_batch_update()	update weights of model
*	nc_class_classify()	evaluates one input, returns class
*
* ALGORITHM:
*	for training: assign each training point to a center
*		the center chosen is either the nearest one or
*		the one which gives the max gaussian output for the point
*
*		the class assigned to the center is the majority class of
*		the points assigned to that center.
*
*	to classify: find which center this point falls in.
*		the class of the point is the class of that center
*
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/nc_class/nc_class_pack.c,v 11.3 2001/08/23 20:12:39 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/nc_class/nc_class_pack.c,v $
* CREATED: 15-Apr-92				MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/



/* Include necessary header files */
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#include <nnlib.h>
#define nc_class_pack_h_internal_types
#define nc_class_pack_h_statics
#include "nc_class_pack.h"

#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))


/* Static Global Variables (for this file use only) */
static NC_CLASS_PARAMS **nc_classpp;	/* nc_class_model set */
static int      max_models = 0;	/* number of nc_class models */

/* macros for dist_type */
#define NEAR_NEIGHBOR 0
#define GAUSS 1

/************************************************************************
* NAME: nc_class_build(ninputs, nclasses, ncenters)
* DESCRIPTION:
*	Allocates space for an nc_class model with parameter values set from
*       function arguments.
*************************************************************************/
int
nc_class_build(int ninputs, int nclasses, int ncenters)
{
    int	     new_model;
    NC_CLASS_PARAMS *ncpp;

    /* check for bad parameters */
    if (ninputs <= 0)
	fatal_error("nc_class_build: ninputs is negative = %d", ninputs);
    if (ncenters <= 0)
	fatal_error("nc_class_build: ncenters is negative = %d", ncenters);
    if (nclasses <= 0)
	fatal_error("nc_class_build: nclasses is negative = %d", nclasses);

    /* find entry in model array for new model */
    new_model = model_get_number((void ***)&nc_classpp, &max_models,
				 sizeof(NC_CLASS_PARAMS));
    ncpp = nc_classpp[new_model];

    /* Set parameter from subroutine arguments */
    ncpp->ninputs = ninputs;
    ncpp->ncenters = ncenters;
    ncpp->nclasses = nclasses;

    /* Allocate the space for the vectors and matrices in the model */
    ncpp->pcenter = allocate_matrix(ncenters, ninputs);
    ncpp->inv_variance = allocate_matrix(ncenters, ninputs);
    ncpp->log_det_var = allocate_vector(ncenters);
    ncpp->class_vote = allocate_imatrix(ncenters, nclasses);
    ncpp->class = allocate_ivector(ncenters);

    /* Allocate space for feature selection array */
    ncpp->features = allocate_ivector(ninputs+1);
    ncpp->class_cnt = allocate_vector(nclasses);

    /* Return the model number */
    return (new_model);
}

/************************************************************************
* NAME: nc_class_init(model_num, pcentroids, pvariaces, features)
*
* DESCRIPTION:
*	Takes as input model number and pointers to the centroid and
*	spread (variance) vectors.
*	This function copies the centroid and spread vectors into the model.
*	if also copies the features list into the model for later storage
*	It initializes the class count array
*************************************************************************/
void
nc_class_init(int model_num, float **pcentroids, float **pspreads,
	      int *features, int dist_type, float minimum_variance)
{
    int	     i, j;
    NC_CLASS_PARAMS *ncpp;

    ncpp = nc_classpp[model_num];

    ncpp->dist_type = dist_type;

    /* copy features into feature array */
    for (i = 0; i <= ncpp->features[0]; i++) {
	ncpp->features[i] = features[i];
    }
    for(i = 0; i < ncpp->nclasses; i++)
      ncpp->class_cnt[i] = 0.;
    ncpp->total_npatterns = 0.;

    /*
     * Copy the centroids and spreads to the model memory
     * -----------------------------------------------------
     */
    for (i = 0; i < ncpp->ncenters; i++){
	ncpp->log_det_var[i] = 0.;
	for (j = 0; j < ncpp->ninputs; j++) {
	    if (pspreads[i][j] <= 0.) {
		pspreads[i][j] = minimum_variance;
		/*
		warning("%s: %s, changed to minvar %lf\n", "nc_class_init",
			"zero variance found in cluster data",
			minimum_variance);
			*/
	    }
	    ncpp->inv_variance[i][j] = 1./pspreads[i][j];
	    ncpp->log_det_var[i] += (float)log10((double)pspreads[i][j]);
	    ncpp->pcenter[i][j] = pcentroids[i][j];
	}
	for(j = 0; j < ncpp->nclasses; j++){
	    ncpp->class_vote[i][j] = 0;
	}
    }
}

/************************************************************************
* NAME: nc_class_free()
* DESCRIPTION:
*	Deallocate storage used in a model
*
* EXAMPLE:
*
*************************************************************************/
void
nc_class_free(int model_num)
{
    NC_CLASS_PARAMS *ncpp;

    ncpp = nc_classpp[model_num];

    deallocate_matrix(ncpp->pcenter);
    deallocate_matrix(ncpp->inv_variance);
    deallocate_vector(ncpp->log_det_var);
    deallocate_ivector(ncpp->features);
    deallocate_vector(ncpp->class_cnt);
    deallocate_imatrix(ncpp->class_vote);
    deallocate_ivector(ncpp->class);

    nn_free((char *)nc_classpp[model_num], "freeing nc_class model");
    nc_classpp[model_num] = NULL;
}

/***---------------------------------------------------------------------------
* NAME:	 nc_class_read_model()
* DESCRIPTION:
*	Takes a file pointer as input.
*	Reads nc_class model from parameter file.
*	Builds nc_class model using parameters.
*	Returns nc_class model number.
* NOTES:
*	See nc_class_write_paramfile().
--------------------------------------------------------------------------***/
int
nc_class_read_model(FILE *fp)
{
    int		model_num;
    int		rv, ninputs, nclasses, ncenters, dist_type;
    float	total_npatterns;
    NC_CLASS_PARAMS *ncpp;

    /* Get the nc_class model scalars */
    param_setvarptr("ninputs", &ninputs, nc_scalars, NNC_SCALARS);
    param_setvarptr("nclasses", &nclasses, nc_scalars, NNC_SCALARS);
    param_setvarptr("ncenters", &ncenters, nc_scalars, NNC_SCALARS);
    param_setvarptr("total_npat", (int *)&total_npatterns,
		    nc_scalars, NNC_SCALARS);
    param_setvarptr("dist_type", &dist_type, nc_scalars, NNC_SCALARS);
    if ((rv = param_getblk(fp, nc_scalars, NNC_SCALARS)) < 0)
	return rv;

    /* Allocate the space needed to store the new model */
    model_num = nc_class_build(ninputs, nclasses, ncenters);
    ncpp = nc_classpp[model_num];

    /* Fill in parameters set during training or by nc_class_init() */
    ncpp->dist_type = dist_type;
    ncpp->total_npatterns = total_npatterns;

    /* Set ptrs and dimensions for allocated matrices */
    param_setvarptr("class_cnt", (int *)&ncpp->class_cnt[0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("class_cnt", ncpp->nclasses, 0, nc_arrays, NNC_ARRAYS);
    param_setvarptr("features", &ncpp->features[0], nc_arrays, NNC_ARRAYS);
    param_setdim("features", ncpp->ninputs+1, 0, nc_arrays, NNC_ARRAYS);
    param_setvarptr("class", &ncpp->class[0], nc_arrays, NNC_ARRAYS);
    param_setdim("class", ncpp->ncenters, 0, nc_arrays, NNC_ARRAYS);
    param_setvarptr("centers", (int *) &ncpp->pcenter[0][0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("centers", ncpp->ncenters, ncpp->ninputs,
		 nc_arrays, NNC_ARRAYS);
    param_setvarptr("inv_vars", (int *) &ncpp->inv_variance[0][0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("inv_vars", ncpp->ncenters, ncpp->ninputs,
		 nc_arrays, NNC_ARRAYS);
    param_setvarptr("log_det_vars", (int *) &ncpp->log_det_var[0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("log_det_vars", ncpp->ncenters, 0, nc_arrays, NNC_ARRAYS);

    /* Get the nc_class model arrays */
    if ((rv = param_getblk(fp, nc_arrays, NNC_ARRAYS)) < 0)
	return rv;

    return model_num;

} /* nc_class_read_model() */

/***---------------------------------------------------------------------------
* NAME:	 nc_class_write_model()
* DESCRIPTION:
*	Takes a model number and file pointer as input.
*	Writes nc_class model to parameter file.
--------------------------------------------------------------------------***/
void
nc_class_write_model(int model_num, FILE *fp)
{
    NC_CLASS_PARAMS *ncpp;

    ncpp = nc_classpp[model_num];


    /* Get the nc_class model scalars */
    param_setvarptr("ninputs", &ncpp->ninputs, nc_scalars, NNC_SCALARS);
    param_setvarptr("nclasses", &ncpp->nclasses, nc_scalars, NNC_SCALARS);
    param_setvarptr("ncenters", &ncpp->ncenters, nc_scalars, NNC_SCALARS);
    param_setvarptr("total_npat", (int *)&ncpp->total_npatterns,
		    nc_scalars, NNC_SCALARS);
    param_setvarptr("dist_type", &ncpp->dist_type, nc_scalars, NNC_SCALARS);

    /* Set ptrs and dimensions for allocated matrices */
    param_setvarptr("class_cnt", (int *)&ncpp->class_cnt[0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("class_cnt", ncpp->nclasses, 0, nc_arrays, NNC_ARRAYS);
    param_setvarptr("features", &ncpp->features[0], nc_arrays, NNC_ARRAYS);
    param_setdim("features", ncpp->features[0]+1, 0, nc_arrays, NNC_ARRAYS);
    param_setvarptr("class", &ncpp->class[0], nc_arrays, NNC_ARRAYS);
    param_setdim("class", ncpp->ncenters, 0, nc_arrays, NNC_ARRAYS);
    param_setvarptr("centers", (int *) &ncpp->pcenter[0][0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("centers", ncpp->ncenters, ncpp->ninputs,
		 nc_arrays, NNC_ARRAYS);
    param_setvarptr("inv_vars", (int *) &ncpp->inv_variance[0][0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("inv_vars", ncpp->ncenters, ncpp->ninputs,
		 nc_arrays, NNC_ARRAYS);
    param_setvarptr("log_det_vars", (int *) &ncpp->log_det_var[0],
		    nc_arrays, NNC_ARRAYS);
    param_setdim("log_det_vars", ncpp->ncenters, 0, nc_arrays, NNC_ARRAYS);

    param_putblk(fp, nc_scalars, NNC_SCALARS);	/* scalars */
    param_putblk(fp, nc_arrays, NNC_ARRAYS);	/* arrays */

}				/* nc_class_write_model() */


/************************************************************************
* NAME: nc_class_param(model_num, param_name)
* DESCRIPTION:
*	Takes model number and parameter name.
*	Returns a pointer to the parameter.
*	Function really not necessary if not interested in accessing
*	parameters (reading) from main()
*************************************************************************/

void *
nc_class_param(int model_num, char *param_name)
{
    void *param_address;
    NC_CLASS_PARAMS *ncpp;

    ncpp = nc_classpp[model_num];

    /* Return the requested parameter */
    if (PEQ("ninputs"))
      param_address = &ncpp->ninputs;
    else if (PEQ("ncenters"))
      param_address = &ncpp->ncenters;
    else if (PEQ("nclasses"))
      param_address = &ncpp->nclasses;
    else if (PEQ("features"))
      param_address = &ncpp->features;
    else if (PEQ("centers"))
      param_address = &ncpp->pcenter;
    else if (PEQ("inv_variances"))
      param_address = &ncpp->inv_variance;
    else if (PEQ("log_det_var"))
      param_address = &ncpp->log_det_var;
    else if (PEQ("classes"))
      param_address = &ncpp->class;
    else if (PEQ("dist_type"))
      param_address = &ncpp->dist_type;
    else {
	warning("Parameter %s not defined", param_name);
	param_address = NULL;
    }
    return (param_address);

}

/************************************************************************
* NAME: nc_class_batch_ctrain(model_num, invec, class);
* DESCRIPTION:
*	Takes nc_class model number, ONE input vector, and its class.
*	Trains the particular nc_class model with that input.
*************************************************************************/
void
nc_class_batch_ctrain(int model_num, float *invec, int class)
{
    int best_i;
    NC_CLASS_PARAMS *ncpp;

    ncpp = nc_classpp[model_num];

    ncpp->class_cnt[class] += 1.;
    ncpp->total_npatterns += 1.;

    /* find the nearest center.  Update its class counts */
    best_i = get_best_center(ncpp, invec);
    ncpp->class_vote[best_i][class]++;
}

/************************************************************************
* NAME: nc_class_batch_update(model_num)
* DESCRIPTION:
*	Takes the nc_class model number and finds the classes for the centers
*************************************************************************/
void
nc_class_batch_update(int model_num)
{
    NC_CLASS_PARAMS *ncpp;
    int i;

    ncpp = nc_classpp[model_num];

    for(i = 0; i < ncpp->ncenters; i++){
	ncpp->class[i] = int_max_pick(ncpp->class_vote[i], ncpp->nclasses);
    }
}

/************************************************************************
* NAME: nc_class_classify(model_num, invec)
* DESCRIPTION:
*	Takes nc_class model number and a pointer to an input vector.
*	Returns the class assigned to the input vector
*************************************************************************/
int
nc_class_classify(int model_num, float *invec)
{
    NC_CLASS_PARAMS *ncpp;
    int best_i;

    ncpp = nc_classpp[model_num];

    best_i = get_best_center(ncpp, invec);
    return (ncpp->class[best_i]);
}

static int
get_best_center(NC_CLASS_PARAMS *ncpp, float *invec)
{
    int best_i, i;
    float best_dist, dist;

    if(ncpp->ncenters < 1) return -1;
    if(ncpp->dist_type == NEAR_NEIGHBOR){
	best_i = find_nearest_neighbor(invec, ncpp->pcenter,
				       ncpp->ninputs, ncpp->ncenters);
    } else /* must be GAUSS */ {
	best_i = 0; 
	best_dist = -MAXFLOAT;
	/* find the center with the greatest gauss output for invec */
	for(i = 0; i < ncpp->ncenters; i++){
	    dist = get_log_gauss_output_diag(ncpp->ninputs, invec,
					     ncpp->pcenter[i],
					     ncpp->inv_variance[i],
					     ncpp->log_det_var[i]);
	    if(dist > best_dist) {
		best_i = i;
		best_dist = dist;
	    }
	}
    }
    return(best_i);
}
/*****************************************
nc_class_read_cluster()
*****************************************/
void
nc_class_read_cluster(char *param_path, FILE *param_file, int *pninputs,
		      int *pncenters, float ***centroids, float ***spreads,
		      int *features)
{
    int model, noutputs;

    model = cluster_read_paramfile(param_file, pninputs, &noutputs, pncenters,
				   features);
    if(model < 0)
	fatal_error("nc_class: read error %d on parameter file %s", model,
		  param_path);

    if(*pncenters <= 0)
      fatal_error("nc_class_read_cluster: No centers from clustering");

    /* insure that the centers are not stored by class */
    cluster_merge_centers(model);

    *centroids = *(float ***)cluster_param(model, "pcenter", 0);
    *spreads = *(float ***)cluster_param(model, "pvariance", 0);
}

