/***--------------------------------------------------------------------------
 * NAME:		nc_class_train.c
 * PURPOSE:	training driver for nearest cluster classifier
 *
 * SEE ALSO:	nc_class_pack.c
 *
 * FUNCTIONS:	void nc_class_train();
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/nc_class/nc_class_train.c,v 11.3 2001/08/23 20:12:40 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/nc_class/nc_class_train.c,v $
 * CREATED: 15-Apr-92 lck			     MIT LINCOLN LABORATORY
 */

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include <nnlib.h>
#include "nc_class.h"
#include "nc_class_pack.h"

#define MAXSTR (256)

static char	prog_name[MAXSTR];	/* static variables */

void
nc_class_train(int create_flag, int train_flag, int nraw, int ninputs,
	       int *pfeatures, int noutputs, int npatterns, int debg,
	       int verbosity, int verror, int random_flag, char *data_path,
	       char *data_fname, char *exp_path, char *error_fname,
	       char *param_fname, char *comment, int adjust_priors,
	       int priors_npatterns, float *priors_list, int dist_type,
	       char *cluster_param_fname, float min_var)
{
    /* local variables */
    int		model, rv, normalization = 0, data_model;
    int		desired_class;
    int		ncenters;
    float	*inputs, **pcentroids, **pspreads;
    char	*tmp_path, clus_path[MAXSTR];
    FILE	*error_file, *param_file, *clus_file;
    int		*features;
    
    strcpy(prog_name, get_program_name());
    
    if (train_flag && create_flag) {
	/* read the cluster file */
	tmp_path = path_build(exp_path, cluster_param_fname);
	strncpy(clus_path, tmp_path, MAXSTR - 1);
	
	clus_file = nn_fopen(clus_path, "r",
			     "for read of cluster parameters");
	if (verbosity > 1) {
	    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	    fflush(stdout);
	}
	model = norm_read_paramfile(clus_file, &normalization);
	
	if (model < 0)
	  fatal_error("nc_class: read error %d %s %s",
		      model, "on normalization parameter file", clus_path);
	
	nc_class_read_cluster(clus_path, clus_file, &ninputs, &ncenters,
			      &pcentroids, &pspreads, pfeatures);
	
	/* if we are not doing cross validation, this is all we need from
	   the cluster file */
	if(train_flag)/* cannot do a cross valid run when train is set */
	  nn_fclose(clus_file, clus_path, "for read of cluster parameters");
	
	if (verbosity > 1)
	  printf ("\nBuilding nc_class model\n");
	model = nc_class_build (ninputs, noutputs, ncenters);
	nc_class_init (model, pcentroids, pspreads, pfeatures, dist_type,
		       min_var);
    } else {		/* retrain or test */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	if (verbosity > 1)
	  printf("\nReading nc_class parameters from \"%s\"\n", tmp_path);
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s",
		      prog_name, rv, "on normalization parameter file",
		      tmp_path);
	
	model = nc_class_read_model(param_file);
	if (model < 0)
	  fatal_error("nc_class: read error %d on model file %s", model,
		      tmp_path);
	nn_fclose(param_file, tmp_path, "for read of nc_class parameters");
	verify_model_params (model, ninputs,
			     *(int *) nc_class_param(model, "ninputs"),
			     pfeatures,
			     *(int **) nc_class_param(model, "features"));
    }
    /* use features from the model */
    features = *(int **) nc_class_param(model, "features");
    ninputs = *(int *) nc_class_param(model, "ninputs");
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {		/* open file for classification errors */
	if (create_flag)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    } else error_file = NULL;

    if(adjust_priors){
	data_model = class_data_read(data_path, data_fname, npatterns,
				     nraw, ninputs, noutputs, features,
				     normalization, verbosity);
	if(random_flag) class_data_random_init(data_model);
	class_data_priors_init(data_model, priors_list, priors_npatterns);
    } else {
	data_model = class_data_single_init(data_path, data_fname, npatterns,
					    nraw, ninputs, noutputs, features,
					    normalization, verbosity);
    }
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
	nc_class_batch_ctrain(model, inputs, desired_class);
    }
    nc_class_batch_update(model);
    class_data_exit(data_model);
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "nc_class model\n");
    nc_class_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- nc_class model saved in \"%s\"\n", tmp_path);
    
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space and close files */
    norm_free();
    nc_class_free(model);
}
