/***---------------------------------------------------------------------------
* NAME:		parzen.c 
* PURPOSE:	cross validation driver for Parzen window classifier 
*
* SEE ALSO:	parzen_main.c parzen_pack.c
*
*
* FUNCTIONS:
*	parzen	driver for parzen classifier
*		handles model creation, training and testing of data
*
* BUGS AND LIMITATIONS:
*
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_cv.c,v 11.6 2001/08/23 20:12:45 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_cv.c,v $
* CREATED: 05-Oct-94 lck				MIT LINCOLN LABORATORY
***************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "parzen_pack.h"
#include "parzen.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)
		/* globals */  

static char     prog_name[MAXSTR];	/* static variables */

float
parzen_cross_valid(DEFAULTS_FILE_STRUCT *df_args, int ninputs, int *features,
		   int normalize, int npatterns, int verbosity, int verror,
		   int random_flag, int cv_nfolds, int cv_nsplits,
		   int use_cv_splits, int *cv_splits, char **cv_train_mask,
		   char **cv_test_mask, char *data_path, char *data_fname,
		   char *norm_fname, char *exp_path, char *error_fname,
		   char *param_fname, char *comment, int adjust_train,
		   int adjust_test, int priors_npatterns,
		   float *train_priors_list, float *test_priors_list,
		   int cv_random,
		   int report_cost, int roc_target, float roc_lower,
		   float roc_upper, 
		   int log_outputs, float minimum_variance,
		   int per_input, int per_class, float spread, int linear)
{
    int		model, err_model, err_part_model, data_model, normalization=0;
    int		desired_class, class;
    int		n_read, n_trials, trial_num;
    int		fold;
    float	*inputs, *outputs, percent_error;
    char	*tmp_path;
    FILE	*error_file = NULL, *param_file;
    int		post_sum_to_one = 0;
    int		i;

    strcpy(prog_name, get_program_name());

    /* allocate and initialize model structure */
    model = parzen_build(ninputs, df_args->noutputs, per_input, per_class,
			 &df_args->input_types[1],
			 &df_args->input_type_values[1]);
			 
    parzen_init(model, minimum_variance);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(df_args->ninputs, df_args->noutputs);

    /* print various debugging and information messages before starting
     * actual training or testing 
     */
    display_model_params (model, ninputs, normalization, features, verbosity);

    /******** prepare results reporting *************/
    if (verror > 0) {
	error_file = nn_fopen(path_build(exp_path, error_fname),
			      "w", "for classification errors");
    }
    err_model = report_build (verbosity, verror, df_args->noutputs);
    report_init (err_model, df_args->db_type, ninputs, report_cost,
		 log_outputs, post_sum_to_one);
    report_roc_init(err_model, roc_target, roc_lower, roc_upper);
    err_part_model = report_build(verbosity, 0, df_args->noutputs);

    /************ read data and initialize class_data routines ************/
    data_model = class_data_read(data_path, data_fname, npatterns, df_args->ninputs,
				 ninputs, df_args->noutputs, features, normalization,
				 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
		       cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_train)
      class_data_priors_init(data_model, train_priors_list, priors_npatterns);

    /* write normalization and feature selection parameters to model file */
    if(param_fname != NULL){
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "w", "for write of model file");
	fprintf(param_file, "%s\n", comment);
	fprintf(param_file, "normalization data\n");
	norm_write_paramfile(param_file);
	fprintf(param_file, "feature selection\n");
	feature_write_paramfile(param_file, features, df_args->ninputs, ninputs);
    }
	

    for(fold = 0; fold < cv_nfolds; fold++){
	if(verbosity > OVERALL_VERBOSITY)
	  printf(">>>>>> FOLD %d <<<<<<\n", fold);
	parzen_init(model, minimum_variance);

	class_data_start_cv_fold(data_model, fold, TRUE);
	n_trials = class_data_get_npatterns(data_model);

	for(n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    parzen_batch_ctrain (model, inputs, desired_class);
	}
	parzen_batch_update (model);
	class_data_end_cv_fold(data_model);

	/* reinitialize model for testing */
	parzen_test_init (model, adjust_test, test_priors_list,
			  spread, linear);

	report_init(err_part_model, df_args->db_type, ninputs, report_cost,
		    log_outputs, post_sum_to_one);
	report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);

	/* write parzen model to parameter file */
	if(param_fname != NULL){
	    fprintf(param_file, "parzen model %d\n", fold);
	    parzen_write_model(model, param_file);
	}

	/* get the current pointer to the outputs for error reporting */
	outputs = *(float **) parzen_param(model, "outputs");

	/* test */
	/* prepare for testing fold */
	class_data_start_cv_fold(data_model, fold, FALSE);
	n_trials = class_data_get_npatterns(data_model);

	for (n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &desired_class))
	       == NULL) break;
	    class = parzen_classify(model, inputs);
	    trial_num = class_data_get_input_num(data_model);
	    if(report_cost == ROC_COST){
	      /* turn these into linear */
	      for(i = 0; i < df_args->noutputs; i++){
		outputs[i] = nn_exp10(outputs[i]);
	      }
	    }
	    report_class_trial(err_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	    report_class_trial(err_part_model, error_file, inputs, outputs,
			       desired_class, class, trial_num);
	}    
	/* report on testing */
	report_overall(err_part_model, stdout, error_fname, df_args->output_labels);
	class_data_end_cv_fold(data_model);
    } /* end fold */

    /************* close files and free space *********************/
   /* close model file */
    if(param_fname != NULL){
	nn_fclose(param_file, tmp_path, "for write of model file");
	if ( verbosity > 0)
	  printf("\nFinished -- parzen model saved in \"%s\"\n", tmp_path);
    }

    /* finish error reporting and close error file */
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, df_args->output_labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }
    /* free space */
    class_data_exit(data_model);
    parzen_free (model);
    norm_free();

    return (percent_error);
}
