/***--------------------------------------------------------------------------
* NAME:		parzen_main.c
* PURPOSE:	Entry routine for Parzen window classifier
*
* USAGE:	parzen -<flags>
*		
* EXAMPLE: 	parzen --                shows all available flags
*
* SEE ALSO:	parzen_train.c parzen_cv.c parzen_test.c parzen_pack.c
*
* GLOBALS:	<flags> defined in parzen_args.h
*
* DESCRIPTION: 
*		main	Processes command line parameters and calls parzen
*
* BUGS AND LIMITATIONS:
*
* Created: 5-Oct-94 lck				MIT Lincoln Laboratory
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_main.c,v 11.5 2001/08/23 20:12:46 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_main.c,v $
*-------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_main.c,v 11.5 2001/08/23 20:12:46 craig Exp $";



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define NO_CV_TABLE
#include <cv_args.h>
#include <gen_class_args.h>
#include "parzen_args.h"			/* parameter definitions */
#include "parzen.h"

int
main(int argc, char **argv)
{
    char *comment;
    int use_splits;
    /* arguments for reading defaults file */
    DEFAULTS_FILE_STRUCT *df_args;
    /* argument table with all entries */
    ARG *argtab;
    int tabsize;
    int roc_cost = 0, roc_target = 0;
    float roc_lower = 0, roc_upper = 1;

    init_nnsubs(argc, argv, rcsid);
    /******************* read argument list **********************/
    nnargs_combine_argtabs(gen_tab, GEN_TABSIZE, parzen_tab, PARZEN_TABSIZE,
			   &argtab, &tabsize);
    argc = nnargs(argc, argv, argtab, tabsize);

    print_log_header();

    /************* read defaults and set input data parameters ********/
    /* read the defaults file */
    if(strcmp(desc_fname, "") == 0){ /* desc_fname is unset */
	fatal_error("%s: -fdescribe must be set", argv[0]);
    }
    /* set the number of inputs and outputs and read the labels,
       allocating the space */
    df_args = (DEFAULTS_FILE_STRUCT *)
      nn_malloc(sizeof(DEFAULTS_FILE_STRUCT), "for the defaults file struct");
    read_defaults_file(path_build(data_path, desc_fname), df_args);
    if(df_args->noutputs == 0)
      fatal_error("noutputs in %s must be greater than 0", desc_fname);
    if(df_args->ninputs == 0)
      fatal_error("ninputs in %s must be greater than 0", desc_fname);

    /* set the number of patterns to no more than the number of lines
       in the data file */
    if(strcmp(data_fname, "") == 0){ /* data_fname is unset */
	fatal_error("%s: -finput must be set", argv[0]);
    }
    check_file_length(path_build(data_path, data_fname), &npatterns);
    if(npatterns == 0)
      fatal_error("%s: Empty data file %s", argv[0],
		  path_build(data_path, data_fname));
    /* check the data base type */
    if (df_args->db_type != CLASS_DATA)
      fatal_error("class data required, check defaults file");

    /***************** check feature selection flags *******************/
    if (ninputs > df_args->ninputs)
      ninputs = df_args->ninputs;
    ninputs = fix_features(df_args->ninputs, ninputs, features);

    /************ play with the input_type lists *****************/
    /* after this call, the input_types array is at least ninputs long 
       and exists and has been initialized to match the feature list */
    fix_input_types(df_args, normalize, ninputs, features);

    /*************** set up the cross validation arguments ***********/
    if(train_flag && cross_valid > 2){
	warning("%s\n%s\n","-train and -cross_valid are both set.",
		" -cross_valid will be ignored");
	cross_valid = 0;
    }
    use_splits =
      fix_cross_valid_args(&cross_valid, &cv_nfolds, &cv_nsplits, cv_splits,
			   cv_train_mask,cv_test_mask,cross_valid_from_file,
			   path_build(data_path, cross_valid_fname));
        
    /********* fix up priors adjustment arguments *********************/
    fix_priors(adjust_train, train_uniform_priors, &train_priors_list,
	       adjust_test, test_uniform_priors, &test_priors_list,
	       train_flag, cross_valid, df_args->noutputs);
    
    /********** check PARZEN args ***************/
    if(spread <= 0.)
      fatal_error("parzen_main: spread must be greater than 0");
    /* we cannot do granding across input features when there is mixed
       circular and continuous data. check for this */
    /* granding across circular variables only makes sense in cases
       where the units of the variables is similar. the granding will
       take place in units of radians. if the original units are not
       similar, the results will not be what the user expects, since there
       is no normalization with circular variables (although, if you ask
       me, translating it all to radians is a normalization). */
    if(!per_input){
	int circ, i;

	if(df_args->input_types[1] == CIRC) circ = 1;
	else circ = 0;
	for(i = 1; i <= ninputs; i++){
	    if((circ && df_args->input_types[i] != CIRC) ||
	       (!circ && df_args->input_types[i] == CIRC)) {
		fatal_error("parzen_main: -per_input must be set when there is a mix of circular and non-circular features");
	    }
	}
    }

    /************** make and print header comment *******************/
    comment = make_comment_string(nnargs_get_flags(argtab, tabsize));
    if (verbosity > FLAGS_VERBOSITY) {
	printf("%s", comment);
	fflush(stdout);	/* ensure log headers appear before error outputs */
    }

    srandom(seed);		/* initialize random numbers */
    
    /************* call the selected driver ***************/
    if(train_flag){
	parzen_train(df_args, ninputs, features,
		    normalize, npatterns, verbosity, random_flag,
		    data_path, data_fname, norm_fname, exp_path, param_fname,
		    comment, adjust_train, priors_npatterns,
		    &train_priors_list[1],
		    minimum_variance, per_input, per_class);
    } else if (cv_nfolds > 1){
	parzen_cross_valid(df_args, ninputs, features,
			   normalize, npatterns, verbosity, verror,
			   random_flag, cv_nfolds, cv_nsplits, use_splits,
			   cv_splits, &cv_train_mask[1], &cv_test_mask[1],
			   data_path, data_fname, norm_fname, exp_path,
			   error_fname, param_fname, comment,
			   adjust_train, adjust_test, priors_npatterns,
			   &train_priors_list[1], &test_priors_list[1],
			   cv_random,
			   (roc_cost?ROC_COST:(linear?SQ_ERR:LOG_LIKE)),
			   roc_target, roc_lower, roc_upper,
			   log_outputs,
			   minimum_variance, per_input, per_class, 
			   spread, linear);
    } else {
	parzen_test(df_args, create_flag, ninputs, features,
		    npatterns, verbosity, verror, data_path,
		    data_fname, exp_path, error_fname, param_fname,
		    adjust_test, &test_priors_list[1], log_outputs,
		    spread, linear);
    }

    print_log_footer();

    nnargs_free_argtab(argtab, tabsize);
    /* free the vectors allocated when the defaults file was read */
    free_defaults_file_struct(df_args);

    exit_nnsubs();
    exit(0);
}
