/***-------------------------------------------------------------------------
 * NAME:		parzen_pack.c
 * PURPOSE:	Functions specific to parzen window classification
 *			
 * SEE ALSO:	parzen.c parzen_main.c
 *
 *
 * DESCRIPTION:
 *    Functions in this file
 *	parzen_build		Allocate storage for model parameters
 *	parzen_init		Initialize model parameters
 *	parzen_free		Deallocates model storage
 *	parzen_read_model	Initializes a parzen model from file
 *	parzen_write_model	Writes a parzen model to output file
 *	parzen_param		Returns pointers to elements in model
 *	parzen_batch_ctrain	does training on a vector
 *	parzen_batch_update	updates parameters of model after training
 *	parzen_test_build	Allocate storage for test mode parameters
 *	parzen_test_free		Deallocates test mode storage
 *	parzen_test_init		Initialize test mode parameters
 *	parzen_classify		classifies a vector using the parzen algortihm
 *
 * ALGORITHM:
 * REFERENCE:
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_pack.c,v 11.3 2001/08/23 20:12:46 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_pack.c,v $
 * CREATED: 5-Oct-94 lck			MIT LINCOLN LABORATORY
 ------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#define parzen_pack_h_internal_types
#define parzen_pack_h_param_types
#define parzen_pack_h_static_decl
#include "parzen_pack.h"

/* constant definitions */
#define PEQ(y) (0==strcmp(param_name,y))
#define ALLOC_SIZE 128
#define ABS(x) (((x) >= 0) ? (x) : -(x))

/* Internal Globals */
static PARZEN_PARAMS **parzenpp;/* array of parameters for each model */
static int max_models = 0;	/* number of entries in parzenpp */

/************************************************************************
 * NAME: parzen_build(ninputs, noutputs, per_class, full,
 *                    input_types, input_type_values)
 * DESCRIPTION:
 *	Allocates space for a parzen model 
 *
 *************************************************************************/
int
parzen_build(int ninputs, int noutputs, int per_input, int per_class,
	     INPUT_FEATURE_TYPE *input_types, float *input_type_values)
{
    register int    new_model;
    int circ, i;
    PARZEN_PARAMS *gpp;
    
    new_model = model_get_number((void ***)&parzenpp,
				 &max_models, sizeof(PARZEN_PARAMS));
    gpp = parzenpp[new_model];
    
    /* set fields which are array sizes */
    gpp->ninputs = ninputs;
    gpp->noutputs = noutputs;
    gpp->per_input = per_input;
    gpp->per_class = per_class;

    /* allocate space */
    gpp->centers = allocate_matrix(ALLOC_SIZE, gpp->ninputs);
    gpp->classes = allocate_ivector(ALLOC_SIZE);
    gpp->trials_allocd = ALLOC_SIZE;
    gpp->total_trials = 0;
    gpp->class_trials = allocate_ivector(gpp->noutputs);

    gpp->train_gauss_mean = allocate_matrix(gpp->noutputs, gpp->ninputs);
    gpp->train_gauss_var = allocate_matrix(gpp->noutputs, gpp->ninputs);
    gpp->train_cos_sum = allocate_matrix(gpp->noutputs, gpp->ninputs);
    gpp->train_sin_sum = allocate_matrix(gpp->noutputs, gpp->ninputs);

    if(gpp->per_class){
	gpp->log_det.per_class = allocate_vector(gpp->noutputs);
	if(gpp->per_input){
	    gpp->inv_var.per_class_input =
	      allocate_matrix(gpp->noutputs, gpp->ninputs);
	} else {
	    gpp->inv_var.per_class = allocate_vector(gpp->noutputs);
	}
    } else {
	if(gpp->per_input){
	    gpp->inv_var.per_input = allocate_vector(gpp->ninputs);
	}
    }

    gpp->outputs = allocate_vector(gpp->noutputs);
    gpp->log_class_priors = allocate_vector(gpp->noutputs);
    
    gpp->input_types = allocate_ivector(ninputs);
    gpp->input_type_values = allocate_vector(ninputs);
    if(input_types != NULL && input_type_values != NULL){
	int i;
	for(i = 0; i < ninputs; i++){
	    gpp->input_types[i] = input_types[i];
	    gpp->input_type_values[i] = input_type_values[i];
	}
    } else {
	for(i = 0; i < ninputs; i++){
	    gpp->input_types[i] = CONT;
	    gpp->input_type_values[i] = 0.;
	}
    }
    /* we cannot do granding across input features when there is mixed
       circular and continuous data. check for this */
    if(!per_input){
	if(gpp->input_types[0] == CIRC) circ = 1;
	else circ = 0;
	for(i = 0; i < ninputs; i++){
	    if((circ && gpp->input_types[i] != CIRC) ||
	       (!circ && gpp->input_types[i] == CIRC)) {
		fatal_error("parzen_build: -per_input must be set when there is a mix of circular and non-circular features");
	    }
	}
    }

    return new_model;
}

/*****************************************************************************
 * NAME: parzen_init ()
 * DESCRIPTION:		
 *		Initialize parzenian classifier variables
 *
 *****************************************************************************/
void
parzen_init (int model, float minimum_variance)
{
    int i, j;
    PARZEN_PARAMS *gpp;
    
    gpp = parzenpp[model];
    
    gpp->minimum_variance = minimum_variance;
    gpp->total_trials = 0;
    for(i = 0; i < gpp->noutputs; i++){
	gpp->class_trials[i] = 0.;
	for(j = 0; j < gpp->ninputs; j++){
	    gpp->train_gauss_mean[i][j] = 0.;
	    gpp->train_gauss_var[i][j] = 0.;
	    gpp->train_cos_sum[i][j] = 0.;
	    gpp->train_sin_sum[i][j] = 0.;
	}
    }
}      

/************************************************************************
 * NAME: parzen_free()
 * DESCRIPTION:
 *		Deallocates space used in a parzen model 
 *
 *************************************************************************/
void
parzen_free(int model)
{
    PARZEN_PARAMS *gpp;
    
    /* does not free those things allocated in parzen_test_build */
    gpp = parzenpp[model];

    deallocate_matrix(gpp->centers);
    deallocate_ivector(gpp->classes);
    deallocate_ivector(gpp->class_trials);
    deallocate_matrix(gpp->train_gauss_mean);
    deallocate_matrix(gpp->train_gauss_var);
    deallocate_matrix(gpp->train_cos_sum);
    deallocate_matrix(gpp->train_sin_sum);
    if(gpp->per_class){
	deallocate_vector(gpp->log_det.per_class);
	if(gpp->per_input){
	    deallocate_matrix(gpp->inv_var.per_class_input);
	} else {
	    deallocate_vector(gpp->inv_var.per_class);
	}
    } else {
	if(gpp->per_input){
	    deallocate_vector(gpp->inv_var.per_input);
	}
    }
    deallocate_vector(gpp->outputs);
    deallocate_vector(gpp->log_class_priors);

    deallocate_ivector(gpp->input_types);
    deallocate_vector(gpp->input_type_values);

    nn_free((char *)gpp, "parzen model");
    parzenpp[model] = NULL;
}

static void
parzen_set_scalar_params(PARZEN_PARAMS *gpp)
{
    param_setvarptr("ninputs", &gpp->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &gpp->noutputs, paramtab, TABSIZE);
    param_setvarptr("total_trials", &gpp->total_trials, paramtab, TABSIZE);
    param_setvarptr("per_input", &gpp->per_input, paramtab, TABSIZE);
    param_setvarptr("per_class", &gpp->per_class, paramtab, TABSIZE);
}

static void
parzen_set_vector_params(PARZEN_PARAMS *gpp)
{
    param_setvarptr("class_trials", (int *)&gpp->class_trials[0],
		    paramtab_l, TABSIZE_L);
    param_setdim("class_trials",gpp->noutputs,0, paramtab_l, TABSIZE_L);
    param_setvarptr("means", (int *)&gpp->centers[0][0],
		    paramtab_l, TABSIZE_L);
    param_setdim("means",gpp->total_trials, gpp->ninputs,
		 paramtab_l, TABSIZE_L);
    param_setvarptr("classes", (int *)&gpp->classes[0],
		    paramtab_l, TABSIZE_L);
    param_setdim("classes",gpp->total_trials,0, paramtab_l, TABSIZE_L);
    param_setvarptr("input_types", (int *)&gpp->input_types[0],
		     paramtab_l, TABSIZE_L);
    param_setdim("input_types", gpp->ninputs, 0, paramtab_l, TABSIZE_L);
    param_setvarptr("type_values", (int *)&gpp->input_type_values[0],
		     paramtab_l, TABSIZE_L);
    param_setdim("type_values", gpp->ninputs, 0, paramtab_l, TABSIZE_L);
    
}

static void
parzen_set_var_params(PARZEN_PARAMS *gpp)
{
    if(gpp->per_class){
	param_setvarptr("log_det", (int *)&gpp->log_det.per_class[0],
			det_vector, NVAR);
	param_setdim("log_det", gpp->noutputs, 0, det_vector, NVAR);
	if(gpp->per_input){
	    param_setvarptr("inv_var",
			    (int *)&gpp->inv_var.per_class_input[0][0],
			    var_matrix, NVAR);
	    param_setdim("inv_var", gpp->noutputs, gpp->ninputs,
			 var_matrix, NVAR);
	} else {
	    param_setvarptr("inv_var", (int *)&gpp->inv_var.per_class[0],
			    var_vector, NVAR);
	    param_setdim("inv_var", gpp->noutputs, 0, var_vector, NVAR);
	}
    } else {
	param_setvarptr("log_det", (int *)&gpp->log_det.grand_class,
			det_scalar, NVAR);
	if(gpp->per_input){
	    param_setvarptr("inv_var", (int *)&gpp->inv_var.per_input[0],
			    var_vector, NVAR);
	    param_setdim("inv_var", gpp->ninputs, 0, var_vector, NVAR);
	} else {
	    param_setvarptr("inv_var", (int *)&gpp->inv_var.grand,
			    var_scalar, NVAR);
	}
    }
}

/***---------------------------------------------------------------------------
 * NAME:	 parzen_read_model()
 * DESCRIPTION:
 *       Loads parameters from ascii input file
 *       allocates space for and loads data into vectors and matricies
 *
 *       on error from param_getblk, returns the error value
 *       (there is already an error check on the allocation, so we don't
 *       have to do that here)
 *
 * example:
 --------------------------------------------------------------------------***/
int
parzen_read_model (FILE *fp)
{
    int rv, model;
    int ninputs, noutputs, total_trials, per_input, per_class;
    PARZEN_PARAMS *gpp, temp;
    
    parzen_set_scalar_params(&temp);
    if ((rv = param_getblk(fp, paramtab, TABSIZE)) < 0) return rv;
    
    ninputs = temp.ninputs;
    noutputs = temp.noutputs;
    total_trials = temp.total_trials;
    per_class = temp.per_class;
    per_input = temp.per_input;

    /* allocate space for model and vectors */
    model = parzen_build(ninputs, noutputs, per_input, per_class, NULL, NULL);
    gpp = parzenpp[model];
    gpp->total_trials = total_trials;
    
    /* reallocate centers and classes to be sure there is enough room */
    gpp->centers = reallocate_matrix(gpp->centers, ALLOC_SIZE,
				     gpp->ninputs, gpp->total_trials);
    gpp->classes = reallocate_ivector(gpp->classes, ALLOC_SIZE,
				      gpp->total_trials);

    /* get arrays */
    /* set feature list length to the max we should need to read */
    parzen_set_vector_params(gpp);
    if ((rv = param_getblk(fp, paramtab_l, TABSIZE_L)) < 0) return rv;

    parzen_set_var_params(gpp);
    if (gpp->per_class){
	if ((rv = param_getblk(fp, det_vector, NVAR)) < 0) return rv;
	if(gpp->per_input){
	    if ((rv = param_getblk(fp, var_matrix, NVAR)) < 0) return rv;
	} else {
	    if ((rv = param_getblk(fp, var_vector, NVAR)) < 0) return rv;
	}
    } else {
	if ((rv = param_getblk(fp, det_scalar, NVAR)) < 0) return rv;
	if (gpp->per_input){
	    if ((rv = param_getblk(fp, var_vector, NVAR)) < 0) return rv;
	} else {
	    if ((rv = param_getblk(fp, var_scalar, NVAR)) < 0) return rv;
	}
    }
    return (model);
} 				/* parzen_read_model() */
/***--------------------------------------------------------------------------
 * NAME:	 parzen_write_model()
 * DESCRIPTION:
 *       Writes parameters and exemplars for a model to a file
 *       in ascii format
 *
 * example:
 *
 --------------------------------------------------------------------------***/
int
parzen_write_model (int model, FILE *fp)
{
    int rv;
    PARZEN_PARAMS *gpp;
    
    gpp = parzenpp[model];
    /* save scalars */
    /* get the parzen model scalars */
    parzen_set_scalar_params(gpp);
    if ((rv = param_putblk(fp, paramtab, TABSIZE)) < 0) return rv;
    
    /* save vectors */
    parzen_set_vector_params(gpp);
    if ((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0) return rv;

    parzen_set_var_params(gpp);
    /* save variances */
    if(gpp->per_class){
	if ((rv = param_putblk(fp, det_vector, NVAR)) < 0) return rv;
	if(gpp->per_input){
	    if ((rv = param_putblk(fp, var_matrix, NVAR)) < 0) return rv;
	} else {
	    if ((rv = param_putblk(fp, var_vector, NVAR)) < 0) return rv;
	}
    } else {
	if ((rv = param_putblk(fp, det_scalar, NVAR)) < 0) return rv;
	if (gpp->per_input){
	    if ((rv = param_putblk(fp, var_vector, NVAR)) < 0) return rv;
	} else {
	    if ((rv = param_putblk(fp, var_scalar, NVAR)) < 0) return rv;
	}
    }
    return (0);
}/* parzen_write_model() */

/************************************************************************
 * NAME: parzen_param(param_name)
 * DESCRIPTION:
 *	takes parameter name.
 *	returns a floating point pointer to the parameter.
 *	single value parameters are read only.
 *	function really not necessary if not interested in accessing
 *	parameters (reading) from main()
 *************************************************************************/
void *
parzen_param(int model, char *param_name)
{
    void *param_address;
    PARZEN_PARAMS *gpp;
    
    gpp = parzenpp[model];
    if (PEQ("ninputs"))
      param_address = &gpp->ninputs;
    else if (PEQ("noutputs"))
      param_address = &gpp->noutputs;
    else if (PEQ("nclasses"))
      param_address = &gpp->noutputs;
    else if (PEQ("outputs"))
      param_address = &gpp->outputs;
    else if (PEQ("total_trials"))
      param_address = &gpp->total_trials;
    else if (PEQ("class_trials"))
      param_address = &gpp->class_trials;
    else if (PEQ("per_input"))
      param_address = &gpp->per_input;
    else if (PEQ("per_class"))
      param_address = &gpp->per_class;
    else if (PEQ("classes"))
      param_address = &gpp->classes;
    else if (PEQ("means"))
      param_address = &gpp->centers;
    else if (PEQ("centers"))
      param_address = &gpp->centers;
    else if(PEQ("inv_var"))
      param_address = &gpp->inv_var;
    else if(PEQ("inv_var_grand"))
      param_address = &gpp->inv_var.grand;
    else if(PEQ("inv_var_input"))
      param_address = &gpp->inv_var.per_input;
    else if(PEQ("inv_var_class"))
      param_address = &gpp->inv_var.per_class;
    else if(PEQ("inv_var_class_input"))
      param_address = &gpp->inv_var.per_class_input;
    else if(PEQ("log_det"))
      param_address = &gpp->log_det;
    else if(PEQ("log_det_per_class"))
      param_address = &gpp->log_det.per_class;
    else if(PEQ("log_det_grand"))
      param_address = &gpp->log_det.grand_class;
    else if(PEQ("input_types"))
      param_address = &gpp->input_types;
    else if(PEQ("input_type_values"))
      param_address = &gpp->input_type_values;
    else {
	warning("parzen_param: parameter %s not defined", param_name);
	param_address = NULL;
    }
    return param_address;
} 

/************************************************************************
 * NAME: parzen_batch_ctrain(model, inputs, desired_class);
 * DESCRIPTION:
 *	stores the center, making sure there is enough room
 *	updates the means and variances using inputs
 *
 *************************************************************************/
void
parzen_batch_ctrain(int model, float *inputs, int desired_class)
{
    register float tmp, radian;
    register int i, cur_center;
    PARZEN_PARAMS *gpp;
    float n;
    
    gpp = parzenpp[model];

    /* keep track of the priors info */
    cur_center = gpp->total_trials++;
    gpp->class_trials[desired_class]++;

    /* check that there is space enough for the new center */
    if(gpp->total_trials > gpp->trials_allocd){
	gpp->centers = reallocate_matrix(gpp->centers, gpp->trials_allocd,
					 gpp->ninputs,
					 gpp->trials_allocd+ALLOC_SIZE);
	gpp->classes = reallocate_ivector(gpp->classes, gpp->trials_allocd,
					  gpp->trials_allocd+ALLOC_SIZE);
	gpp->trials_allocd += ALLOC_SIZE;
    }

    /* store center */
    gpp->classes[cur_center] = desired_class;
    for(i = 0; i < gpp->ninputs; i++){
	gpp->centers[cur_center][i] = inputs[i];
    }

    /* update variances */
    n = (float)gpp->class_trials[desired_class];
    for(i = 0; i < gpp->ninputs; i++){
	tmp = inputs[i] - gpp->train_gauss_mean[desired_class][i];
	tmp *= tmp/n;
	gpp->train_gauss_var[desired_class][i] += tmp;
	gpp->train_gauss_var[desired_class][i] *= (n-1.)/n;
    }
    /* update means */
    for (i = 0; i < gpp->ninputs; i++) {
	n = (float)gpp->class_trials[desired_class];
	gpp->train_gauss_mean[desired_class][i] *= (n-1.)/n;
    	gpp->train_gauss_mean[desired_class][i] += inputs[i]/n;
    }
    /* update vonmises numbers */
    for(i = 0; i < gpp->ninputs; i++){
	if(gpp->input_types[i] == CIRC){
	    radian = get_radians(inputs[i], gpp->input_type_values[i]);
	    gpp->train_cos_sum[desired_class][i] += cos(radian);
	    gpp->train_sin_sum[desired_class][i] += sin(radian);
	}
    }
}

/************************************************************************
 * NAME: parzen_batch_update()
 * DESCRIPTION:
 *    the variances are calculated all at once, so this is just
 *    a final check that the minimum variances are ok and taking averages
 *    for the grand versions.  It's too hard to do the overall average
 *    variance from the per class incremental means all at once, so per
 *    class variances are calculated then averaged
 *    
 *    inversion can be done here, now, so it is.
 *
 *************************************************************************/
void
parzen_batch_update(int model)
{
    register int i, k;
    float min_var, var;
    PARZEN_PARAMS *gpp;
    
    gpp = parzenpp[model];

    min_var = gpp->minimum_variance;

    /* get the variances from the circular variables */
    for(i = 0; i < gpp->ninputs; i++){
	if (gpp->input_types[i] == CIRC){
	    for(k = 0; k < gpp->noutputs; k++){
		gpp->train_gauss_var[k][i] =
		  1./vonmises_inv_var(gpp->train_cos_sum[k][i],
				      gpp->train_sin_sum[k][i],
				      gpp->class_trials[k]);
	    }
	}
    }
    if(gpp->per_input){
	if (gpp->per_class){
	    /* check on minimum variances */
	    for (k = 0; k < gpp->noutputs; k++) {
		gpp->log_det.per_class[k] = 0.;
		for(i = 0; i < gpp->ninputs; i++){
		    var = gpp->train_gauss_var[k][i];
		    if(var < min_var) var = min_var;
		    if (gpp->input_types[i] != CIRC){
			gpp->log_det.per_class[k] += log10(var);
		    }
		    gpp->inv_var.per_class_input[k][i] = 1./var;
		}
	    }
	} else {
	    /* average over outputs */
	    gpp->log_det.grand_class = 0.;
	    for(i = 0; i < gpp->ninputs; i++){
		var = 0.;
		for(k = 0; k < gpp->noutputs; k++){
		    var += gpp->train_gauss_var[k][i];
		}
		var /= gpp->noutputs;
		if(var < min_var) var = min_var;
		if (gpp->input_types[i] != CIRC){
		    gpp->log_det.grand_class += log10(var);
		}
		gpp->inv_var.per_input[i] = 1./var;
	    }
	}
    } else { /* not per_input */
	/* I don't use log_det for these in testing,
	   but I do write it to the parameter file */
	/* I checked in build to be sure that these variables are either
	   all circular or none of the are circular */
	if(gpp->per_class){
	    /* average over inputs */
	    for(k = 0; k < gpp->noutputs; k++){
		var = 0.;
		for(i = 0; i < gpp->ninputs; i++){
		    var += gpp->train_gauss_var[k][i];
		}
		var /= gpp->noutputs;
		if(var < min_var) var = min_var;
		/* this determinant is ignored */
		gpp->log_det.per_class[k] = gpp->ninputs*log10(var);
		gpp->inv_var.per_class[k] = 1./var;
	    }
	} else {
	    /* average over inputs and outputs */
	    var = 0.;
	    for( k = 0; k < gpp->noutputs; k++){
		for(i = 0; i < gpp->ninputs; i++){
		    var += gpp->train_gauss_var[k][i];
		}
	    }
	    var /= gpp->noutputs * gpp->ninputs;
	    if (var < min_var) var = min_var;
	    gpp->inv_var.grand = 1./var;
	    /* this determinant is ignored */
	    gpp->log_det.grand_class = gpp->ninputs*log10(var);
	}
    }
}

/*****************************************************************************
 * NAME: parzen_test_init ()
 * DESCRIPTION:		
 *	invert the covariance matrices, if full covariance
 *	store the inverse of the diagonal variance if diagonal
 *	store the log_class_priors
 *	
 *	Side effect:  If full, grand, the grand covariance matrix is inverted
 *		If full, per class, the per class covarainces are inverted
 *		otherwise, these matrices are left alone.
 *****************************************************************************/
void
parzen_test_init (int model, int adjust_priors, float *priors_list,
		  float spread, int linear)
{
    int i, k;
    PARZEN_PARAMS *gpp;
    float log_spread;
    
    gpp = parzenpp[model];
    gpp->linear = linear;
    
    /* We will be dividing the sums by class_trials to make
       each likelihood go between 0 and 1. It saves time to do that
       ahead here */
    if(adjust_priors) {
      for (i = 0; i < gpp->noutputs; i++) 
	if(priors_list[i] > 0.)
	  gpp->log_class_priors[i] =
	    log10(priors_list[i]/gpp->class_trials[i]);
	else
	  gpp->log_class_priors[i] = 0.;
    } else {
      for (i = 0; i < gpp->noutputs; i++){
	  if(gpp->class_trials[i] > 0.)
	    gpp->log_class_priors[i] =
	      log10(1. / (double)gpp->class_trials[i]);
	  else
	    gpp->log_class_priors[i] = 0.;
      }
    }
    log_spread = gpp->ninputs*log10(spread);
    if(gpp->per_input){
	if(gpp->per_class){
	    for(k = 0; k < gpp->noutputs; k++){
		for (i = 0; i < gpp->ninputs; i++){
		    gpp->inv_var.per_class_input[k][i] /= spread;
		}
		gpp->log_det.per_class[k] += log_spread;
	    }
	} else {
	    for(i = 0; i < gpp->ninputs; i++){
		gpp->inv_var.per_input[i] /= spread;
	    }
	    gpp->log_det.grand_class += log_spread;
	}
    } else {
	if(gpp->per_class){
	    for(k = 0; k < gpp->noutputs; k++){
		gpp->inv_var.per_class[k] /= spread;
		gpp->log_det.per_class[k] += log_spread;
	    }
	} else {
	    gpp->inv_var.grand /= spread;
	    gpp->log_det.grand_class += log_spread;
	}
    }
    /* I will let this stand for the circular variables, too */
    if(gpp->linear){
	double temp;
	/* make the inv_var really be 2*stddev */
	if(gpp->per_input){
	    if(gpp->per_class){
		for(k = 0; k < gpp->noutputs; k++){
		    for(i = 0; i < gpp->ninputs; i++){
			temp = gpp->inv_var.per_class_input[k][i];
			temp = 2.*(1./sqrt(temp));
			gpp->inv_var.per_class_input[k][i] = temp;
		    }
		}
	    } else {
		for(i = 0; i < gpp->ninputs; i++){
		    temp = gpp->inv_var.per_input[i];
		    temp = 2*(1./sqrt(temp));
		    gpp->inv_var.per_input[i] = temp;
		}
	    }
	} else {
	    if(gpp->per_class){
		for(k = 0; k < gpp->noutputs; k++){
		    temp = gpp->inv_var.per_class[k];
		    temp = 2.*(1./sqrt(temp));
		    gpp->inv_var.per_class[k] = temp;
		}
	    } else {
		gpp->inv_var.grand = 2.*(1./sqrt(gpp->inv_var.grand));
	    }
	}
    }
}

/************************************************************************
 * NAME: parzen_eval(model, inputs)
 * DESCRIPTION:
 *	Takes parzen model number and a pointer to an input vector.
 *	fills in the output vector.
 *
 *************************************************************************/
void
parzen_eval(int model, float *inputs)
{
    int i, j, p, c;
    PARZEN_PARAMS *gpp;
    float output, dist, inv_var, x, m;
    int ok;
    
    gpp = parzenpp[model];
    if(gpp->linear){/* step functions */
	/* zero out outputs for linear summation */
	for(i = 0; i < gpp->noutputs; i++){
	    gpp->outputs[i] = 0.;
	}
	for(p = 0; p < gpp->total_trials; p++){
	    c = gpp->classes[p];
	    ok = 1;
	    /* check if we are inside the box around this center */
	    for(j = 0; j < gpp->ninputs && ok; j++){
		if(gpp->input_types[j] == CIRC){
		    /* as a sanity check, if this range is bigger than
		       2PI, everythiing is inside this box... */
		    if(gpp->per_input){
			if(gpp->per_class){
			    ok =
			      inside_circ(inputs[j], gpp->centers[p][j],
					  gpp->inv_var.per_class_input[c][j],
					  gpp->input_type_values[j]);
			} else {
			    ok = inside_circ(inputs[j], gpp->centers[p][j],
					     gpp->inv_var.per_input[j],
					     gpp->input_type_values[j]);
			}
		    } else {
			if(gpp->per_class){
			    ok = inside_circ(inputs[j], gpp->centers[p][j],
					     gpp->inv_var.per_class[c],
					     gpp->input_type_values[j]);
			} else {
			    ok = inside_circ(inputs[j], gpp->centers[p][j],
					     gpp->inv_var.grand,
					     gpp->input_type_values[j]);
			}
		    }
		} else { /* not circular feature */
		    dist = inputs[j] - gpp->centers[p][j];
		    dist = ABS(dist);
		    /* inv_var has been converted to 2*stddev for linear */
		    if(gpp->per_input){
			if(gpp->per_class){
			    if(dist > gpp->inv_var.per_class_input[c][j])
			      ok = 0;
			} else {
			    if(dist > gpp->inv_var.per_input[j]) ok = 0;
			}
		    } else {
			if(gpp->per_class){
			    if(dist > gpp->inv_var.per_class[c]) ok = 0;
			} else {
			    if(dist > gpp->inv_var.grand) ok = 0;
			}
		    }
		}
	    }
	    gpp->outputs[c] += ok;
	}
	/* put in priors and divide by the number of patterns per class */
	for(i = 0; i < gpp->noutputs; i++){
	    gpp->outputs[i] *= nn_exp10(gpp->log_class_priors[i]);
	}
    } else { /* gauss functions */
	/* initalize outputs for summing in log domain */
	for(c = 0; c < gpp->noutputs; c++){
	    gpp->outputs[c] = -MAXFLOAT;
	}
	/* get Gauss output for each stored center */
	if(!gpp->per_input && !gpp->per_class) inv_var = gpp->inv_var.grand;
	for(p = 0; p < gpp->total_trials; p++){
	    c = gpp->classes[p];
	    if(!gpp->per_input && gpp->per_class)
	      inv_var = gpp->inv_var.per_class[c];

	    output = 0;
	    for(j = 0; j < gpp->ninputs; j++){
		if(gpp->per_input){
		    if(gpp->per_class)
		      inv_var = gpp->inv_var.per_class_input[c][j];
		    else
		      inv_var = gpp->inv_var.per_input[j];
		}
		if(gpp->input_types[j] == CIRC){
		    x = get_radians(inputs[j], gpp->input_type_values[j]);
		    m = get_radians(gpp->centers[p][j],
				    gpp->input_type_values[j]);
		    output +=
		      vonmises_log_output_scalar(x, m, inv_var);
		} else {
		    output +=
		      get_log_gauss_output_scalar(inputs[j],
						  gpp->centers[p][j], inv_var);
		}
	    }
	    gpp->outputs[c] = logadd(gpp->outputs[c], output);
	}
	/* adjust sum by priors and scale output by number
	   of contributing training patterns */
	for(c = 0; c < gpp->noutputs; c++){
	    gpp->outputs[c] += gpp->log_class_priors[c];
	}
    }
}
/************************************************************************
 * NAME: parzen_classify(model, inputs)
 * DESCRIPTION:
 *	Takes parzen model number and a pointer to an input vector.
 *	Returns the class assigned to the input vector and fills in the
 *      output vector.
 *
 *************************************************************************/
int
parzen_classify(int model, float *inputs)
{
    int class;
    PARZEN_PARAMS *gpp;
    
    gpp = parzenpp[model];
    parzen_eval(model, inputs);
    if(gpp->linear){
	int i, ok = 0;
	for(i = 0; i < gpp->noutputs && !ok; i++){
	    if(gpp->outputs[i] > 0) ok = 1;
	}
	if(!ok) return UNKNOWN_CLASS;
    }
    /* these outputs are linear for Step function kernals,
       Logarithmic for Gaussian Kernals */
    class = (max_pick (gpp->outputs, gpp->noutputs));
    
    return class;
}

/* take a variable and convert it to radians given it falls on
   a circle with values [0,range] */
static float
get_radians(float x, float range)
{
    return (x * 2. * M_PI)/range;
}

/* take a value in radians and force it to be in the range [0,2PI] */
static float
make_0_2PI(float x)
{
    double twoPi = 2.*M_PI;

    while (x < 0) x += twoPi;
    while (x >= twoPi) x -= twoPi;
    return x;
}
     
static int
inside_circ(float x, float mean, float var, float range)
{
    float rad_mean, radian, lower, upper;

    if(var >= 2.*M_PI) return 1;
    rad_mean = get_radians(mean, range);
    lower = rad_mean - var;
    lower = make_0_2PI(lower);
    upper = rad_mean + var;
    upper = make_0_2PI(upper);
    radian = get_radians(x, range);
    radian = make_0_2PI(radian);
    if(lower <= upper){
	if( radian < lower || radian > upper ) return 0;
	else return 1;
	/* if lower == upper, radian must be == also to pass */
    } else { /* lower > upper, range spans the break */
	if(radian < lower && radian > upper) return 0;
	else return 1;
    }
}
