/***---------------------------------------------------------------------------
* NAME:		parzen.c 
* PURPOSE:	training driver for parzen window classifier 
*
* SEE ALSO:	parzen_main.c parzen_pack.c
*
* GLOBALS:	int debug;
*
* FUNCTIONS:
*	parzen	driver for parzen classifier
*		handles model creation, training and testing of data
*
* BUGS AND LIMITATIONS:
*
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_train.c,v 11.4 2001/08/23 20:12:46 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/parzen/parzen_train.c,v $
* CREATED: 5-Oct-94 lck			MIT LINCOLN LABORATORY
****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "parzen_pack.h"
#include "parzen.h"

#define MAXSTR (256)
		/* globals */  

static char     prog_name[MAXSTR];	/* static variables */

void
parzen_train(DEFAULTS_FILE_STRUCT *df_args, int ninputs, int *features,
	     int normalize, int npatterns, int verbosity, int random_flag,
	     char *data_path, char *data_fname, char *norm_fname,
	     char *exp_path, char *param_fname, char *comment,
	     int adjust_priors, int priors_npatterns, float *priors_list,
	     float minimum_variance, int per_input, int per_class)
{
    int		model, data_model, normalization = 0;
    int		desired_class;
    float	*inputs;
    char	*tmp_path;
    FILE	*param_file;

    strcpy(prog_name, get_program_name());

    /* allocate and initialize model structure */
    /* only create mode is allowed */
    if(df_args->input_types == NULL || df_args->input_type_values == NULL ||
	df_args->input_types[0] <= 0 || df_args->input_type_values <= 0) {
	model = parzen_build(ninputs, df_args->noutputs, per_input, per_class,
			     NULL, NULL);
    } else {
	model =
	  parzen_build(ninputs, df_args->noutputs, per_input, per_class,
		       &df_args->input_types[1],
		       &df_args->input_type_values[1]);
    }
    parzen_init(model, minimum_variance);
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(df_args->ninputs, df_args->noutputs);

    if(adjust_priors){ /* priors adjustment by sampling */
	data_model = class_data_read(data_path, data_fname, npatterns,
				     df_args->ninputs, ninputs, df_args->noutputs, features,
				     normalization, verbosity);
	if(random_flag) class_data_random_init(data_model);
	class_data_priors_init(data_model, priors_list, priors_npatterns);
    } else { /* normal training */
	/* put together the path to the data file and open it */
	data_model =
	  class_data_single_init(data_path, data_fname, npatterns, df_args->ninputs,
				 ninputs, df_args->noutputs, features,
				 normalization, verbosity);
    }
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
	parzen_batch_ctrain (model, inputs, desired_class);
    }
    parzen_batch_update (model);
    class_data_exit(data_model);
    /********** store model in param file **********************/
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, df_args->ninputs, ninputs);
    fprintf(param_file, "parzen model\n");
    parzen_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- parzen model saved in \"%s\"\n", tmp_path);

    norm_free();
    parzen_free (model);
}
