/***-------------------------------------------------------------------------
* NAME:     ran_clus_pack.c
* PURPOSE:  Functions specific to ran_clus.c
*
* SEE ALSO: ran_clus_train.c ran_clus_cv.c ran_clus_main.c
*
* USAGE:
*   model = ran_clus_build(...);
*   ran_clus_init(model,...);
*   <store centers>
*   for(i=0; i<all_patterns; i++)
*       ran_clus_inc_train(model, input[i], class[i]);
*   <calculate variances>
*   for(i =0; i<all_patterns; i++)
*       ran_clus_batch_train(model, input[i], class[i]);
*   ran_clus_batch_update(model);
*   ran_clus_write_model(model,...);
*   ran_clus_free(model);
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*   ran_clus_build      Allocate storage for model parameters.
*   ran_clus_init       Initialize model parameters.
*   ran_clus_free       Deallocates model storage
*   ran_clus_read_model Initializes model from file
*   ran_clus_write_model    Writes model to output file
*   ran_clus_param      Returns pointers to elements in model
*   ran_clus_inc_train  Store point as center, if centers still needed
*   ran_clus_batch_train    Update sums for variances of stored centers
*   ran_clus_batch_update   Calculate variances for stored centers
*   ran_clus_cluster    returns the cluster an input pattern falls in
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/ran_clus/ran_clus_pack.c,v 11.3 2001/08/23 20:12:49 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/ran_clus/ran_clus_pack.c,v $
* CREATED: 7-Apr-92 lck             MIT LINCOLN LABORATORY
*--------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */




/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#define ran_clus_pack_h_internal_types
#include "ran_clus_pack.h"

                /* constant definitions */
#define PEQ(y) (0==strcmp(param_name,y))
#define MATRIX_STORE_SIZE 128


                /* Internal Globals */
static RAN_CLUS_PARAMS **ranpp;/* array of parameters for each model */
static int      max_models = 0; /* current length of ranpp */

/*****************************************************************************
* NAME: ran_clus_build(ninputs, noutputs, ncenters_list, cluster_by_class)
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
* EXAMPLE:
*
******************************************************************************/
int
ran_clus_build(int ninputs, int noutputs, int *ncenters_list,
           int cluster_by_class)
{
    register int    new_model, i;
    RAN_CLUS_PARAMS *rpp;

    new_model = model_get_number((void ***)&ranpp,
                 &max_models, sizeof(RAN_CLUS_PARAMS));

    rpp = ranpp[new_model];

    rpp->cluster_by_class = cluster_by_class;
                    /* set fields which are array sizes */
    rpp->ninputs = ninputs;
    rpp->noutputs = noutputs;
    if(cluster_by_class){
    rpp->ncenters = 0;
    for(i = 0; i < noutputs; i++){
        rpp->ncenters += ncenters_list[i];
    }
    } else rpp->ncenters = ncenters_list[0];
                    /* allocate space */
    rpp->features = allocate_ivector (ninputs+1);

    if(rpp->ncenters == 0) rpp->outputs = NULL;
    else rpp->outputs = allocate_vector (rpp->ncenters);

    if(cluster_by_class){
    rpp->ncenters_list = allocate_ivector(noutputs);
    rpp->nstored_list = allocate_ivector(noutputs);
    rpp->pcenter_list = (float ***)allocate_pvector(noutputs);
    rpp->pvariance_list = (float ***)allocate_pvector(noutputs);
    rpp->count_list = (int **)allocate_pvector(noutputs);
    for(i = 0; i < noutputs; i++){
        rpp->ncenters_list[i] = ncenters_list[i];
        if(rpp->ncenters_list[i] == 0){
        rpp->pcenter_list[i] = NULL;
        rpp->pvariance_list[i] = NULL;
        rpp->count_list[i] = NULL;
        } else {
        rpp->pcenter_list[i] =
          allocate_matrix(ncenters_list[i], ninputs);
        rpp->pvariance_list[i] =
          allocate_matrix(ncenters_list[i], ninputs);
        rpp->count_list[i] = allocate_ivector(ncenters_list[i]);
        }
    }
    } else {
    rpp->ncenters_list = NULL;
    rpp->nstored_list = NULL;
    if(rpp->ncenters == 0){
        rpp->pcenter = NULL;
        rpp->pvariance = NULL;
        rpp->counts = NULL;
    } else {
        rpp->pcenter = allocate_matrix (rpp->ncenters, ninputs);
        rpp->pvariance = allocate_matrix (rpp->ncenters, ninputs);
        rpp->counts = allocate_ivector(rpp->ncenters);
    }
    rpp->pcenter_list = &rpp->pcenter;
    rpp->pvariance_list = &rpp->pvariance;
    rpp->count_list = NULL;
    }

    return new_model;
}

/*****************************************************************************
* NAME: ran_clus_init()
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
******************************************************************************/
void
ran_clus_init (int model, int *nn_features)
{
    int     i, j;
    RAN_CLUS_PARAMS *rpp;

    rpp = ranpp[model];
    /* copy features (array length is included in array) */
    for (i = 0; i <= nn_features[0]; i++)
    rpp->features[i] = nn_features[i];

    rpp->nstored = 0;
    if(rpp->cluster_by_class){
    for(i = 0; i < rpp->noutputs; i++){
        rpp->nstored_list[i] = 0;
        for(j = 0; j < rpp->ncenters_list[i]; j++){
        rpp->count_list[i][j] = 0;
        }
    }
    } else {
    for(j = 0; j < rpp->ncenters; j++){
        rpp->counts[j] = 0;
    }
    }
}

/*****************************************************************************
* NAME: ran_clus_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
******************************************************************************/
void
ran_clus_free(int model)
{
    int i;
    RAN_CLUS_PARAMS *rpp;

    rpp = ranpp[model];
                        /* free vectors and matrices */
    deallocate_ivector (rpp->features);
    deallocate_vector (rpp->outputs);
    if(rpp->cluster_by_class){
    for(i = 0; i < rpp->noutputs; i++){
        deallocate_matrix (rpp->pcenter_list[i]);
        deallocate_matrix (rpp->pvariance_list[i]);
        deallocate_ivector (rpp->count_list[i]);
    }
    deallocate_pvector ((void **)rpp->pcenter_list);
    deallocate_pvector ((void **)rpp->pvariance_list);
    deallocate_ivector (rpp->ncenters_list);
    deallocate_ivector (rpp->nstored_list);
    deallocate_pvector ((void **)rpp->count_list);
    } else {
    deallocate_matrix (rpp->pcenter);
    deallocate_matrix (rpp->pvariance);
    deallocate_ivector (rpp->counts);
    }
                        /* free structure itself */
    nn_free((char *) rpp, "for ran_clus parameter structure");
    ranpp[model] = NULL;
}


/****************************************************************************
* NAME: ran_clus_read_model()
* DESCRIPTION:
*       reads parameters for a model from a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int
ran_clus_read_model(FILE *fp)
{
    int model, ninputs, noutputs, ncenters, i, j, clus_model;
    int cluster_by_class, *features;
    float **pcenter_part, **pvariance_part;
    RAN_CLUS_PARAMS *rpp;

    clus_model = cluster_read_model(fp);
    if(clus_model < 0) return(clus_model);

    cluster_merge_centers(clus_model);
    ninputs = *(int *)cluster_param(clus_model, "ninputs", 0);
    noutputs = *(int *)cluster_param(clus_model, "noutputs", 0);
    ncenters = *(int *)cluster_param(clus_model, "ncenters", 0);
    cluster_by_class = *(int *)cluster_param(clus_model, "cluster_by_class", 0);

    model = ran_clus_build(ninputs, noutputs, &ncenters, cluster_by_class);

    features = *(int **)cluster_param(clus_model, "features", 0);
    ran_clus_init(model, features);
    rpp = ranpp[model];

    rpp->nstored = ncenters;

    pcenter_part = *(float ***)cluster_param(clus_model, "pcenter", 0);
    pvariance_part = *(float ***)cluster_param(clus_model, "pvariance", 0);
    for(i = 0; i < ncenters; i++){
    for(j = 0; j < ninputs; j++){
        rpp->pcenter[i][j] = pcenter_part[i][j];
        rpp->pvariance[i][j] = pvariance_part[i][j];
    }
    }
    cluster_free(clus_model);
    return(model);
}

/*****************************************************************************
* NAME: ran_clus_write_model()
* DESCRIPTION:
*       writes parameters for a model to a file in ASCII format
*
* EXAMPLE:
*
*****************************************************************************/
int
ran_clus_write_model(int model, FILE *fp)
{
    int rv;
    RAN_CLUS_PARAMS *rpp;

    rpp = ranpp[model];

    /* use the version of cluster write that doesn't use model numbers */
    rv = cluster_write_paramfile(fp, rpp->ninputs, rpp->nstored,
                 rpp->noutputs, rpp->pcenter_list,
                 rpp->pvariance_list,
                 rpp->features, rpp->cluster_by_class,
                 rpp->nstored_list);
    return(rv);
}

/************************************************************************
* NAME: ran_clus_param(model, param_name)
* DESCRIPTION:
*       Takes parameter name.
*       Returns a pointer to the named parameter.
*************************************************************************/
void    *
ran_clus_param(int model, char *param_name)
{
    void    *param_address;
    RAN_CLUS_PARAMS *rpp;

    rpp = ranpp[model];
    if (PEQ("features"))
      param_address = &rpp->features;
    else if (PEQ("outputs"))
      param_address = &rpp->outputs;
    else if (PEQ("ninputs"))
      param_address = &rpp->ninputs;
    else if (PEQ("ncenters"))
      param_address = &rpp->ncenters;
    else if (PEQ("nstored"))
      param_address = &rpp->nstored;
    else if (PEQ("ncenters_list"))
      param_address = &rpp->ncenters_list;
    else if (PEQ("nstored_list"))
      param_address = &rpp->nstored_list;
    else if (PEQ("noutputs"))
      param_address = &rpp->noutputs;
    else if (PEQ("pcenter"))
      param_address = &rpp->pcenter;
    else if (PEQ("pvariance"))
      param_address = &rpp->pvariance;
    else if (PEQ("pcenter_list"))
      param_address = &rpp->pcenter_list;
    else if (PEQ("pvariance_list"))
      param_address = &rpp->pvariance_list;
    else if (PEQ("cluster_by_class"))
      param_address = &rpp->cluster_by_class;
    else {
    warning("ran_clus_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}

/*****************************************************************************
* NAME:     ran_clus_inc_train()
* DESCRIPTION:
*           copy given vector into random clus model
*       use ran_clus_batch_train() after to do variances
* EXAMPLE:
*
******************************************************************************/
void
ran_clus_inc_train(int model, float *input, int class)
{
    int     i;
    RAN_CLUS_PARAMS *rpp;

    rpp = ranpp[model];

    if(rpp->cluster_by_class){
    if(rpp->nstored_list[class] < rpp->ncenters_list[class]){
        for(i = 0; i < rpp->ninputs; i++)
          rpp->pcenter_list[class][rpp->nstored_list[class]][i] = input[i];
        rpp->nstored_list[class]++;
        rpp->nstored++;
    }
    } else {
    if(rpp->nstored < rpp->ncenters){
        for(i = 0; i < rpp->ninputs; i++)
          rpp->pcenter[rpp->nstored][i] = input[i];
        rpp->nstored++;
    }
    }
}

void
ran_clus_batch_train(int model, float *input, int class)
{
    int min_i, i;
    float temp;
    RAN_CLUS_PARAMS *rpp;

    rpp = ranpp[model];

    if(rpp->cluster_by_class){
    min_i = find_nearest_neighbor(input, rpp->pcenter_list[class],
                      rpp->ninputs, rpp->nstored_list[class]);

    /* check that the selected point is valid */
    if(min_i < 0 || min_i > rpp->nstored_list[class]) return;

    for(i = 0; i < rpp->ninputs; i++){
        temp = input[i] - rpp->pcenter_list[class][min_i][i];
        rpp->pvariance_list[class][min_i][i] += temp*temp;
    }
    rpp->count_list[class][min_i]++;
    } else {
    min_i = find_nearest_neighbor(input, rpp->pcenter,
                      rpp->ninputs, rpp->nstored);
    /* check that the selected point is valid */
    if(min_i < 0 || min_i > rpp->nstored) return;

    for(i = 0; i < rpp->ninputs; i++){
        temp = input[i] - rpp->pcenter[min_i][i];
        rpp->pvariance[min_i][i] += temp * temp;
    }
    rpp->counts[min_i]++;
    }
}

void
ran_clus_batch_update(int model)
{
    int i, class, c;
    RAN_CLUS_PARAMS *rpp;

    rpp = ranpp[model];

    if(rpp->cluster_by_class){
    for(class = 0; class < rpp->noutputs; class++){
        for(c = 0; c < rpp->nstored_list[class]; c++){
        if(rpp->count_list[class][c] != 0){
            for(i = 0; i < rpp->ninputs; i++){
            rpp->pvariance_list[class][c][i] /=
              rpp->count_list[class][c];
            }
        }
        }
    }
    } else {
    for(c = 0; c < rpp->nstored; c++){
        if(rpp->counts[c] != 0){
        for(i = 0; i < rpp->ninputs; i++){
            rpp->pvariance[c][i] /= rpp->counts[c];
        }
        }
    }
    }
}
/*****************************************************************************
* NAME: ran_clus_cluster()
* DESCRIPTION:
*   Uses Euclidean distances to determine which cluster an input
*   pattern belongs to
*
* EXAMPLE:
*       
******************************************************************************/
int
ran_clus_cluster (int model, float *input)
{
    int min_i, i;
    float min_dist;
    RAN_CLUS_PARAMS *gpp;

    gpp = ranpp[model];

    min_dist = MAXFLOAT;
    min_i = 0;
    for(i = 0; i < gpp->ncenters; i++){
    gpp->outputs[i] = l2dist(gpp->ninputs, input, &gpp->pcenter[i][0]);
    gpp->outputs[i] = (float)sqrt((double)gpp->outputs[i]);
    if(gpp->outputs[i] < min_dist){
        min_dist = gpp->outputs[i];
        min_i = i;
    }
    }

    return(min_i);
}
