/***--------------------------------------------------------------------------
 * NAME:	ran_clus.c
 * PURPOSE:	driver for random clustering algorithm
 *
 * SEE ALSO: 	ran_clus_pack.c
 * GLOBALS:	int debug;
 *
 * FUNCTIONS:
 *	ran_clus	driver for ran_clus clustering algorithm
 *		handles model creation, training, and storage
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/ran_clus/ran_clus_train.c,v 11.4 2001/08/23 20:12:49 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/ran_clus/ran_clus_train.c,v $
 * CREATED: 11-May-92 lck			MIT LINCOLN LABORATORY
 ***************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "ran_clus.h"
#include "ran_clus_pack.h"

#define MAXSTR (256)

/* static variables */
static char     prog_name[MAXSTR];

void
ran_clus_train(int nraw, int ninputs, int *features, int noutputs,
	       int normalize, int npatterns, int debg, int verbosity,
	       int random_flag, char *data_path, char *data_fname,
	       char *norm_fname, char *exp_path, char *param_fname,
	       char *comment, int adjust_train, int priors_npatterns,
	       float *train_priors_list, int *ncenters_list,
	       int cluster_by_class)
{
    int		model, normalization=0, data_model;
    int		class, n_read, n_trials;
    float	*inputs;
    char	data_fullname[MAXSTR], *tmp_path;
    FILE	*data_file, *param_file;
    int		total_ncenters, *nstored_list;
    
    strcpy(prog_name, get_program_name());
    
    /* allocate and initialize model structure */
    if (normalize)
      normalization = read_norm_file(path_build(data_path, norm_fname));
    else
      norm_setup_none(nraw, noutputs);

    model = ran_clus_build(ninputs, noutputs, ncenters_list, cluster_by_class);
    ran_clus_init(model, features);

    if (verbosity > 1) 
      display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */

    if(adjust_train || random_flag){
	/********* read data and initialize class_data routines **********/
	data_model = class_data_read(data_path, data_fname, npatterns, nraw,
				     ninputs, noutputs, features,
				     normalization, verbosity);
	if(random_flag) class_data_random_init(data_model);
	if(adjust_train)
	  class_data_priors_init(data_model, train_priors_list,
				 priors_npatterns);
	n_trials = class_data_get_npatterns(data_model);

	for(n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &class))
	       == NULL) break;
	    ran_clus_inc_train(model, inputs, class);
	}
	class_data_epoch_init(data_model);
	for(n_read = 0; n_read < n_trials; n_read++){
	    if((inputs = class_data_get_next(data_model, &class))
	       == NULL) break;
	    ran_clus_batch_train(model, inputs, class);
	}
	ran_clus_batch_update(model);
	class_data_exit(data_model);
    } else {
	/* put together the full path to the data file*/
	strcpy(data_fullname, path_build(data_path, data_fname));
    
	data_file = nn_fopen(data_fullname, "r", "for training data");
	if (verbosity > 1) printf("\nReading %s\n", data_fullname);
	inputs = allocate_vector(ninputs);
	    
	/* select patterns to be centers */
	for(n_read = 0; n_read < npatterns; n_read++){
	    class =
	      get_class_line(data_file, data_fullname, inputs, nraw,
			     normalization, features, ninputs);
	    if(class == EOF) break;
	    ran_clus_inc_train(model, inputs, class);
	}
	rewind(data_file);
	/* calculate variances given centers */
	for(n_read = 0; n_read < npatterns; n_read++){
	    class =
	      get_class_line(data_file, data_fullname, inputs, nraw,
			     normalization, features, ninputs);
	    if(class == EOF) break;
	    ran_clus_batch_train(model, inputs, class);
	}
	ran_clus_batch_update(model);
	deallocate_vector(inputs);
	nn_fclose(data_file, data_fullname, "for training or test data");
    }

    if(verbosity > 0){
	if(cluster_by_class){
	    printf("CLASS NCENTERS\n");
	    total_ncenters = 0;
	    nstored_list = *(int **)ran_clus_param(model, "nstored_list");
	    for(class = 0; class < noutputs; class++){
		printf("%d\t%d\n", class, nstored_list[class]);
		total_ncenters += nstored_list[class];
	    }
	    printf("--------------\n");
	    printf("TOTAL\t%d\n", total_ncenters);
	} else {
	    printf("NCENTERS STORED: %d\n",
		   *(int *)ran_clus_param(model, "nstored"));
	}
    } 

    /* write out model or models */
    tmp_path = path_build(exp_path, param_fname);
    /* store model in param file */
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);

    ran_clus_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- model saved in \"%s\"\n", tmp_path);
    
    norm_free();
    ran_clus_free(model);
}
