/***--------------------------------------------------------------------------
 * NAME:        rbf.c
 * PURPOSE: driver for Radial Basis Function classifier
 *
 * SEE ALSO:    rbf_pack.c
 *
 * FUNCTIONS:
 *  driver for Radial Basis Function classifier
 *      handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_cv.c,v 11.4 2001/08/23 20:12:52 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_cv.c,v $
 * CREATED: 20-Mar-91 dan               MIT LINCOLN LABORATORY
 *************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include "rbf.h"
#include "rbf_pack.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define MAXSTR (256)


static char     prog_name[MAXSTR];  /* static variables */
static int  verbose;

float
rbf_cross_valid(int map_flag, int nraw, int ninputs, int *pfeatures,
        int noutputs, int npatterns, int debg, int verbosity,
        int verror, int random_flag, int cv_nfolds, int cv_nsplits,
        int use_cv_splits, int *cv_splits, char **cv_train_mask,
        char **cv_test_mask, char *data_path, char *data_fname,
        char *exp_path, char *error_fname, char *param_fname,
        char **labels, char *comment, int adjust_train,
        int adjust_test, int priors_npatterns,
        float *train_priors_list, float *test_priors_list,
        int cv_random,
        int report_cost, int roc_target, float roc_lower,
        float roc_upper, 
        int bias, float extra_hspread, float hspread,
        float max_ratio, char *cluster_param_fname,
        float minimum_variance, int norm_outputs, int fast_train,
        int fast_nhidden)
{
    /* local variables */
    int     model, err_model, err_part_model, data_model;
    int     log_outputs = 0, post_sum_to_one = 0, normalization = 0;
    int     desired_class, class;
    int     n_read, n_trials, trial_num;
    int     ngauss_hidden;
    int     fold;
    float   *inputs, *outputs, **pcentroids, **pspreads, percent_error;
    char    data_fullname[MAXSTR], *tmp_path, clus_path[MAXSTR];
    FILE    *error_file, *param_file, *clus_file;
    int     *features;
    
    strcpy(prog_name, get_program_name());
    verbose = verbosity;
    
    /* put together the full path to the data file*/
    strcpy(data_fullname, path_build(data_path, data_fname));
    
    /* read the cluster file */
    /* doesn't deal with clustered cross valid yet */
    tmp_path = path_build(exp_path, cluster_param_fname);
    strncpy(clus_path, tmp_path, MAXSTR - 1);
    
    clus_file = nn_fopen(clus_path, "r",
             "for read of cluster parameters");
    if (verbose > 1) {
    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
    fflush(stdout);
    }
    model = norm_read_paramfile(clus_file, &normalization);
    
    if (model < 0)
      fatal_error("rbf: read error %d %s %s",
          model, "on normalization parameter file", clus_path);
    
    rbf_read_cluster(clus_path, clus_file, &ninputs,
             &ngauss_hidden, &pcentroids, &pspreads, pfeatures);
    
    if (verbosity > 1)
      printf ("\nBuilding rbf model\n");
    model = rbf_build (ninputs, ngauss_hidden, bias, noutputs);
    rbf_init (model, pcentroids, pspreads, hspread, debg, pfeatures,
          minimum_variance, norm_outputs);
    if(fast_train)
      rbf_init_fast_train(model, fast_nhidden);

    /* use features from the model */
    features = *(int **) rbf_param(model, "features");
    ninputs = *(int *) rbf_param(model, "ninputs");
    outputs = *(float **) rbf_param(model, "outputs");

    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {       /* open file for classification errors */
    error_file = nn_fopen(path_build(exp_path, error_fname),
                  "w", "for classification errors");
    } else {
      error_file = NULL;
    }
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, report_cost,
         log_outputs, post_sum_to_one);
    report_roc_init(err_model, roc_target, roc_lower, roc_upper);
    err_part_model = report_build(verbosity, 0, noutputs);

    /************ read data and initialize class_data routines ************/
    data_model = class_data_read(data_path, data_fname, npatterns, nraw,
                 ninputs, noutputs, features, normalization,
                 verbosity);
    class_data_cv_init(data_model, cv_nsplits, use_cv_splits, cv_splits,
               cv_nfolds, cv_train_mask, cv_test_mask, cv_random);
    if(random_flag) class_data_random_init(data_model);
    if(adjust_train)
      class_data_priors_init(data_model, train_priors_list, priors_npatterns);

    /* store the normalization */
    if(param_fname != NULL){
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    }
    
    for(fold = 0; fold < cv_nfolds; fold++){
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> FOLD %d <<<<<<\n", fold);
    /* read the next set of centers from the cluster file */
    /* and initialize rbf accordingly */
    
    if(fold != 0){
        rbf_free(model);
        /* the number of centers may have been changed */
        rbf_read_cluster(clus_path, clus_file, &ninputs,
                 &ngauss_hidden, &pcentroids, &pspreads,
                 pfeatures);
        model = rbf_build(ninputs, ngauss_hidden, bias,
                  noutputs);
        rbf_init (model, pcentroids, pspreads, hspread, debg,
              pfeatures, minimum_variance, norm_outputs);
        if(fast_train)
          rbf_init_fast_train(model, fast_nhidden);
        outputs = *(float **) rbf_param(model, "outputs");
    }

    class_data_start_cv_fold(data_model, fold, TRUE);
    n_trials = class_data_get_npatterns(data_model);
    
    /* classify data for this fold */
    for(n_read = 0; n_read < n_trials; n_read++){
        
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        rbf_batch_ctrain(model, inputs, desired_class);
    }
    if (verbosity > 1) {
        printf("\nInverting rbf matrix\n");
        fflush(stdout);
    }
    rbf_batch_update(model, max_ratio);
    
    class_data_end_cv_fold(data_model);
    
    /* print out the rbf model */
    if(param_fname != NULL){
        fprintf(param_file, "rbf model %d\n", fold);
        rbf_write_model(model, param_file);
    }
    
    /* reinitialize model and errors for testing */
    rbf_test_init(model, adjust_test, test_priors_list, extra_hspread);
    report_init(err_part_model, map_flag, ninputs, report_cost,
            log_outputs, post_sum_to_one);
    report_roc_init(err_part_model, roc_target, roc_lower, roc_upper);

    /* prepare for testing fold */
    class_data_start_cv_fold(data_model, fold, FALSE);
    n_trials = class_data_get_npatterns(data_model);

    for (n_read = 0; n_read < n_trials; n_read++){
        if((inputs = class_data_get_next(data_model, &desired_class))
           == NULL) break;
        class = rbf_classify(model, inputs);
        trial_num = class_data_get_input_num(data_model);
        report_class_trial(err_model, error_file,
                   inputs, outputs, desired_class, class,
                   trial_num);
        report_class_trial(err_part_model, error_file,
                   inputs, outputs, desired_class, class,
                   trial_num);
    }
    /* report on testing */
    report_overall(err_part_model, stdout, error_fname, labels);
    
    class_data_end_cv_fold(data_model);
    } /* end fold */
    
    /* close the parameter file and the cluster file */
    nn_fclose(clus_file, clus_path, "for read of clusters");
    if(param_fname != NULL){
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- rbf model saved in \"%s\"\n", tmp_path);
    }
    
    class_data_exit(data_model);
    
    if(verbosity > OVERALL_VERBOSITY)
      printf(">>>>>> OVERALL <<<<<<\n");
    report_overall(err_model, stdout, error_fname, labels);
    report_cv_result(err_model, stdout, data_fname, cv_nfolds);
    percent_error = report_get_percent_error(err_model);
    report_free(err_model);
    report_free(err_part_model);
    if (verror > 0) {
    nn_fclose(error_file, path_build(exp_path, error_fname),
          "for error file");
    }
    rbf_free(model);
    norm_free();

    return (percent_error);
}
