/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/***-------------------------------------------------------------------------
* NAME: rbf_pack.c
* PURPOSE:  radial basis function classification and prediction routines
* USAGE:
*
* SEE ALSO:     rbf.c
* DESCRIPTION:  Radial basis function classification and prediction routines.
*   An example of their use is in rbf.c. These routines were
*       modified by rpl to use out new nnutil.c utility routines
*       instead of the older numerical recipes nrutil.c routines.
*   The routine svdfunc was also added. An extra bias node
*       with a constant output is added if desired. When it is added,
*       the number of hidden nodes is increased by 1 and the bias node is
*       added as the highest number hidden node. When reading in centers,
*       centers are read in only for the active gaussian hidden nodes
*       but not for the last bias node. In all routines, ngauss_hidden is the
*       number of active gaussian hidden nodes, nbias_hidden (0-1) is the
*       number of bias nodes and nhidden is the total number of hidden nodes
*       including gassian and bias nodes.
*
*   rbf_build()     constructs a rbf model
*   rbf_init()      initialize model for training
*   rbf_free()      Deallocates model storage
*   rbf_write_model()   writes a model to parameter file
*   rbf_read_model()    reads a model from parameter file
*   rbf_param()     returns value of parameter
*   rbf_batch_ctrain()  train given class
*   rbf_batch_train()   train given desired output vector
*   rbf_batch_update()  update weights of model
*   rbf_eval()      evaluates one input, return outvec
*   rbf_classify()      evaluates one input, returns class
*   rbf_write_paramfile()   old name, calls rbf_write_model
*   rbf_read_paramfile()    old name, calls rbf_read_model
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*   Assumes number of classes < number of hidden nodes.
*   Assumes nodes are diagonal Gaussian and spreads are
*       variances in each dimension spread value per node.
*       Could speed up by precomputing .5/spread in rbf_eval.
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_pack.c,v 11.3 2001/08/23 20:12:52 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_pack.c,v $
* CREATED: 26-DEC-89 kng                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/



/* Include necessary header files */
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#include <nnlib.h>
#define   rbf_pack_h_internal_types
#include "rbf_pack.h"

#define MODEL_STORE_SIZE 10
#define PEQ(y) (0==strcmp(param_name,y))

/* Static Global Variables (for this file use only) */
static RBF_PARAMS **rbfpp;  /* rbf_model set */
static int      max_models = 0; /* number of rbf models */

static float **inverse_correl_matrix;   /* inverse correlation matrix */
static int verbose = 0;

/*** Parameter Table ***/
static PARAM    rbfscalars[] = {
    {"ninputs", P_INT},
    {"ngauss_hidden", P_INT},
    {"bias", P_INT},
    {"nhidden", P_INT},
    {"nclasses", P_INT},
    {"hspread", P_FLOAT},
    {"total_npat", P_FLOAT},
    {"norm_outs", P_INT},
};
static PARAM    rbfarrays[] = {
    {"centroids", P_FLOAT_MATRIX, NULL, 0, 0},
    {"spreads", P_FLOAT_MATRIX, NULL, 0, 0},
    {"weights", P_FLOAT_MATRIX, NULL, 0, 0},
    {"features", P_INT_VECTOR, NULL, 0, 0},
    {"class_cnt", P_FLOAT_VECTOR, NULL, 0, 0},
};
#define NRBFSCALARS (sizeof(rbfscalars)/sizeof(PARAM))
#define NRBFARRAYS (sizeof(rbfarrays)/sizeof(PARAM))

/************************************************************************
* NAME: rbf_build(ninputs, ngauss_hidden, bias, nclasses)
* DESCRIPTION:
*   Allocates space for an rbf model with parameter values set from
*       function arguments.
*   Returns the created model number (0,1, to ...).
*   ngauss_hidden is the number of
*       normal gaussian hidden nodes. bias is whether to add a
*       hidden node with constant output.
*       The extra added node has the highest index and a
*       constant output of 1. Note that hspread and center spreads and
*       variance are sent in rbf_init().
*************************************************************************/
int 
rbf_build(int ninputs, int ngauss_hidden, int bias, int nclasses)
{
    int             new_model;
    int             nhidden;
    RBF_PARAMS *rpp;

    if(bias != 0) nhidden = ngauss_hidden + 1;
    else nhidden = ngauss_hidden;

    /* check for bad parameters */
    if (ninputs <= 0)
      fatal_error("rbf_build: ninputs is negative = %d", ninputs);
    if (ngauss_hidden <= 0)
      fatal_error("rbf_build: ngauss_hidden is negative = %d", ngauss_hidden);
    if (nclasses <= 0)
      fatal_error("rbf_build: nclasses is negative = %d", nclasses);

    new_model = model_get_number((void ***)&rbfpp, &max_models, sizeof(RBF_PARAMS));
    rpp = rbfpp[new_model];

    /* Set parameter from subroutine arguments */
    rpp->ninputs = ninputs;
    rpp->bias = bias;
    rpp->ngauss_hidden = ngauss_hidden;
    rpp->nhidden = nhidden;
    rpp->nclasses = nclasses;

    /* Allocate the space for the vectors and matrices in the model */
    rpp->hidden_out = allocate_vector(nhidden);
    rpp->outputs = allocate_vector(nclasses);
    rpp->centroids = allocate_matrix(ngauss_hidden, ninputs);
    rpp->spreads = allocate_matrix(ngauss_hidden, ninputs);
    rpp->weights = allocate_matrix(nhidden, nclasses);

    /* Allocate memory for correlation and partial weight matrices */
    rpp->correl_matrix = allocate_matrix(nhidden, nhidden);
    rpp->partial_weight_matrix = allocate_matrix(nhidden, nclasses);

    /* Allocate space for feature selection array */
    rpp->features = allocate_ivector(ninputs+1);

    /* Allocate space for priors adjustment arrays */
    rpp->class_cnt = allocate_vector(nclasses);
    rpp->priors_list = allocate_vector(nclasses);

    rpp->fast_train = 0;
    
    /* Return the model number */
    return (new_model);
}

/************************************************************************
* NAME: rbf_init(model_num, pcentroids, pspreads, hspread, verbose, features,
*                minimum_variance)
* DESCRIPTION:
*   Takes as input model number and pointers to the centroid and
*   spread (variance) vectors.
*   This function allocates space for the M correlation matrix and
*   the partial weight matrix.  In addition, it copies the centroid and
*   spread vectors into the model.
*   This function zeros the correllation, partial weight,
*       and weight matrices.
*       Hspread is a factor that multiplies all spreads. It is used
*       to widen or narrow all basis functions uniformly.
*       minimum_variance is the value to be used to replace a spread value
*       of 0.
*************************************************************************/
void
rbf_init(int model_num, float **pcentroids, float **pspreads, float hspread,
     int xverbose, int *features, float minimum_variance, int norm_outputs)
{
    int i, j;
    RBF_PARAMS *rpp;
    
    rpp = rbfpp[model_num];
    /*
     * get verbosity -----------------
     */
    verbose = xverbose;
    
    /*
     * set hspread in rbf structure and in global variables
     * -------------------------------------------------------
     */
    if (hspread < 0.0)
      fatal_error("rbf_init: hspread is negative = %f", hspread);
    rpp->hspread = hspread;
    rpp->extra_hspread = 1.; /* set this in rbf_test_init */
    
    /* copy features (length of array is included in array) */
    for (i = 0; i <= features[0]; i++) {
    rpp->features[i] = features[i];
    }
    for(i = 0; i < rpp->nclasses; i++)
      rpp->class_cnt[i] = 0.;
    rpp->total_npatterns = 0.;
    rpp->adjust_priors = 0;
    
    /*
     * Initialize correlation, partial weight, and weight matrices
     * -------------------------------------------------------------------
     */
    /* Initialize the matrices to zero */
    for (i = 0; i < rpp->nhidden; i++)
      for (j = 0; j < rpp->nhidden; j++)
    rpp->correl_matrix[i][j] = 0.0;
    for (i = 0; i < rpp->nhidden; i++)
      for (j = 0; j < rpp->nclasses; j++) {
      rpp->weights[i][j] = 0.0;
      rpp->partial_weight_matrix[i][j] = 0.0;
      }
    
    /*
     * Copy the centroids and spreads to the model memory
     * -----------------------------------------------------
     */
    for (i = 0; i < rpp->ngauss_hidden; i++)
      for (j = 0; j < rpp->ninputs; j++) {
      rpp->spreads[i][j] = pspreads[i][j];
      if (rpp->spreads[i][j] < minimum_variance) {
          rpp->spreads[i][j] = minimum_variance;
          /*
          warning("rbf_init: %s, changed to minvar %lf\n",
              "zero variance found in cluster data", minimum_variance);
              */
      }
      rpp->centroids[i][j] = pcentroids[i][j];
      }
    
    /*
     * print out centers and spreads if desired
     * --------------------------------------------
     */
    if (verbose > 0)
      print_centers_spreads(model_num);

    /* Somewhere we should free up the pcentroid and pspread memories */
    /* Do it in main() after the init call */

    rpp->norm_outputs = norm_outputs;
}

/************************************************************************
* NAME: rbf_test_init()
* DESCRIPTION:
*   Set model parameters needed for testing, not for training
*
* EXAMPLE:
*
*************************************************************************/
void
rbf_test_init(int model_num, int adjust_priors, float *priors_list,
          float extra_hspread)
{
    int i;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];
    rpp->extra_hspread = extra_hspread;
    rpp->adjust_priors = adjust_priors;
    if(adjust_priors)
      for(i = 0; i < rpp->nclasses; i++)
    rpp->priors_list[i] = priors_list[i];
}

/************************************************************************
* NAME: rbf_test_init()
* DESCRIPTION:
*   Set model parameters needed for testing, not for training
*
* EXAMPLE:
*
*************************************************************************/
void rbf_init_fast_train(int model_num, int fast_nhidden)
{
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];
    if(0 < fast_nhidden && fast_nhidden < rpp->nhidden){
    rpp->fast_train = 1;
    rpp->fast_nhidden = fast_nhidden;
    rpp->fast_index = allocate_ivector(fast_nhidden);
    rpp->fast_outputs = allocate_vector(fast_nhidden);
    } else {
    rpp->fast_train = 0;
    }
}

/************************************************************************
* NAME: rbf_free()
* DESCRIPTION:
*   Deallocate storage used in a model
*
* EXAMPLE:
*
*************************************************************************/
void rbf_free(int model_num)
{
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    deallocate_vector(rpp->hidden_out);
    deallocate_vector(rpp->outputs);
    deallocate_matrix(rpp->centroids);
    deallocate_matrix(rpp->spreads);
    deallocate_matrix(rpp->weights);
    deallocate_matrix(rpp->correl_matrix);
    deallocate_matrix(rpp->partial_weight_matrix);
    deallocate_ivector(rpp->features);
    deallocate_vector(rpp->class_cnt);
    deallocate_vector(rpp->priors_list);
    if(rpp->fast_train){
    deallocate_ivector(rpp->fast_index);
    deallocate_vector(rpp->fast_outputs);
    }

    nn_free((char *)rbfpp[model_num], "freeing rbf model");
    rbfpp[model_num] = NULL;
}

/***---------------------------------------------------------------------------
* NAME:  rbf_read_model()
* DESCRIPTION:
*   Takes a file pointer as input.
*   Reads rbf model from parameter file.
*   Builds rbf model using parameters.
*   Returns rbf model number.
* NOTES:
*   See rbf_write_paramfile().
--------------------------------------------------------------------------***/
int rbf_read_model(FILE *fp)
{
    int     model_num, norm_outputs;
    int     rv, ninputs, nclasses, ngauss_hidden, bias, nhidden;
    float   hspread, total_npatterns;
    RBF_PARAMS *rpp;

    /* Get the rbf model scalars */
    param_setvarptr("ninputs", &ninputs, rbfscalars, NRBFSCALARS);
    param_setvarptr("nclasses", &nclasses, rbfscalars, NRBFSCALARS);
    param_setvarptr("ngauss_hidden", &ngauss_hidden, rbfscalars, NRBFSCALARS);
    param_setvarptr("bias", &bias, rbfscalars, NRBFSCALARS);
    param_setvarptr("nhidden", &nhidden, rbfscalars, NRBFSCALARS);
    param_setvarptr("hspread", (int *)&hspread, rbfscalars, NRBFSCALARS);
    param_setvarptr("total_npat", (int *)&total_npatterns,
            rbfscalars, NRBFSCALARS);
    param_setvarptr("norm_outs", &norm_outputs, rbfscalars, NRBFSCALARS);
    if ((rv = param_getblk(fp, rbfscalars, NRBFSCALARS)) < 0)
        return rv;

    /* Allocate the space needed to store the new model */
    model_num = rbf_build(ninputs, ngauss_hidden, bias, nclasses);
    rpp = rbfpp[model_num];

    /* Fill in parameters set during training or by rbf_init() */
    rpp->hspread = hspread;
    rpp->total_npatterns = total_npatterns;
    rpp->norm_outputs = norm_outputs;

    /* Set ptrs and dimensions for allocated matrices */
    param_setvarptr("centroids", (int *) &rpp->centroids[0][0],
            rbfarrays, NRBFARRAYS);
    param_setvarptr("spreads", (int *) &rpp->spreads[0][0],
        rbfarrays, NRBFARRAYS);
    param_setvarptr("weights", (int *) &rpp->weights[0][0],
            rbfarrays, NRBFARRAYS);
    param_setvarptr("features", &rpp->features[0], rbfarrays, NRBFARRAYS);
    param_setvarptr("class_cnt", (int *)&rpp->class_cnt[0],
            rbfarrays, NRBFARRAYS);
    param_setdim("centroids", rpp->ngauss_hidden,
         rpp->ninputs, rbfarrays, NRBFARRAYS);
    param_setdim("spreads", rpp->ngauss_hidden,
         rpp->ninputs, rbfarrays, NRBFARRAYS);
    param_setdim("weights", rpp->nhidden,
         rpp->nclasses, rbfarrays, NRBFARRAYS);
    param_setdim("features", rpp->ninputs+1, 0, rbfarrays, NRBFARRAYS);
    param_setdim("class_cnt", rpp->nclasses, 0, rbfarrays, NRBFARRAYS);

    /* Get the rbf model arrays */
    if ((rv = param_getblk(fp, rbfarrays, NRBFARRAYS)) < 0)
        return rv;

    return model_num;

}               /* rbf_read_model() */

/***---------------------------------------------------------------------------
* NAME:  rbf_write_model()
* DESCRIPTION:
*   Takes a model number and file pointer as input.
*   Writes rbf model to parameter file.
* NOTES:
*   Rbf model is written to parameter file in two blocks.
*   The first block contains the scalars:
*       ninputs
*       ngauss_hidden
*       bias
*       nhidden
*       nclasses
*       hspread
*   The second block contains the arrays:
*       centroids
*       spreads
*       weights
--------------------------------------------------------------------------***/
void rbf_write_model(int model_num, FILE *fp)
{
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    if (verbose > 0)
    printf("rbf_write_model: %d %d %d %d\n",
    model_num, rpp->ngauss_hidden, rpp->nhidden, rpp->ninputs);

    /* Get the rbf model scalars */
    param_setvarptr("ninputs", &rpp->ninputs, rbfscalars, NRBFSCALARS);
    param_setvarptr("nclasses", &rpp->nclasses, rbfscalars, NRBFSCALARS);
    param_setvarptr("ngauss_hidden", &rpp->ngauss_hidden,
            rbfscalars, NRBFSCALARS);
    param_setvarptr("bias", &rpp->bias, rbfscalars, NRBFSCALARS);
    param_setvarptr("nhidden", &rpp->nhidden, rbfscalars, NRBFSCALARS);
    param_setvarptr("hspread", (int *)&rpp->hspread,
            rbfscalars, NRBFSCALARS);
    param_setvarptr("total_npat", (int *)&rpp->total_npatterns,
            rbfscalars, NRBFSCALARS);
    param_setvarptr("norm_outs", &rpp->norm_outputs, rbfscalars, NRBFSCALARS);

    /* Set ptrs and dimensions for allocated matrices */
    param_setvarptr("centroids",
            (int *) &rpp->centroids[0][0], rbfarrays, NRBFARRAYS);
    param_setvarptr("spreads",
            (int *) &rpp->spreads[0][0], rbfarrays, NRBFARRAYS);
    param_setvarptr("weights",
            (int *) &rpp->weights[0][0], rbfarrays, NRBFARRAYS);
    param_setvarptr("features", &rpp->features[0], rbfarrays, NRBFARRAYS);
    param_setvarptr("class_cnt", (int *)&rpp->class_cnt[0],
            rbfarrays, NRBFARRAYS);
    param_setdim("centroids", rpp->ngauss_hidden,
         rpp->ninputs, rbfarrays, NRBFARRAYS);
    param_setdim("spreads", rpp->ngauss_hidden,
         rpp->ninputs, rbfarrays, NRBFARRAYS);
    param_setdim("weights", rpp->nhidden,
         rpp->nclasses, rbfarrays, NRBFARRAYS);
    param_setdim("features", rpp->features[0]+1, 0, rbfarrays, NRBFARRAYS);
    param_setdim("class_cnt", rpp->nclasses, 0, rbfarrays, NRBFARRAYS);

    param_putblk(fp, rbfscalars, NRBFSCALARS);  /* scalars */
    param_putblk(fp, rbfarrays, NRBFARRAYS);    /* arrays */

}               /* rbf_write_model() */


/************************************************************************
* NAME: rbf_param(model_num, param_name)
* DESCRIPTION:
*   Takes model number and parameter name.
*   Returns a pointer to the parameter.
*   Function really not necessary if not interested in accessing
*   parameters (reading) from main()
*************************************************************************/

void * rbf_param(int model_num, char *param_name)
{
    void *param_address;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    /* Return the requested parameter */
    if (PEQ("ninputs"))
    param_address = &rpp->ninputs;
    else if (PEQ("ngauss_hidden"))
    param_address = &rpp->ngauss_hidden;
    else if (PEQ("bias"))
    param_address = &rpp->bias;
    else if (PEQ("nhidden"))
    param_address = &rpp->nhidden;
    else if (PEQ("nclasses"))
    param_address = &rpp->nclasses;
    else if (PEQ("hspread"))
    param_address = &rpp->hspread;
    else if (PEQ("features"))
    param_address = &rpp->features;
    else if (PEQ("centroids"))
    param_address = &rpp->centroids;
    else if (PEQ("spreads"))
    param_address = &rpp->spreads;
    else if (PEQ("norm_outputs"))
      param_address = &rpp->norm_outputs;
    else if (PEQ("weights"))
      param_address = &rpp->weights;
    else if (PEQ("outputs"))
      param_address = &rpp->outputs;
    else if (PEQ("class_cnt"))
      param_address = &rpp->class_cnt;
    else if (PEQ("total_npatterns"))
      param_address = &rpp->total_npatterns;
    else {
    warning("Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return (param_address);

}

/************************************************************************
* NAME: rbf_batch_ctrain(model_num,invec,class);
* DESCRIPTION:
*   Takes rbf model number, ONE input vector, and its class.
*   Trains the particular rbf model with that input.
*   Updates the M correlation and partial_weight_matrix matrices.
*   Must call rbf_build() and rbf_init() before this!
*   Note that final calculation of the weights (completely trained)
*   is not done here but is done with a call to rbf_batch_update()
*   after all the input training vectors have been run through this.
*************************************************************************/
void rbf_batch_ctrain(int model_num, float *invec, int class)
{
    int i, k, nodei, nodek;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    rpp->class_cnt[class] += 1.;
    rpp->total_npatterns += 1.;

    if(rpp->fast_train){
    rbf_fast_hidden_eval(model_num, invec);
    for(i = 0; i < rpp->fast_nhidden; i++){
        nodei = rpp->fast_index[i];
        for(k = i; k < rpp->fast_nhidden; k++){
        nodek = rpp->fast_index[k];
        rpp->correl_matrix[nodei][nodek] +=
          rpp->hidden_out[nodei]* rpp->hidden_out[nodek];
        }
    }
    } else {
    /* Generate the correlation matrix with the invec pattern */
    /* Generate the rbf outputs for one pattern */
    rbf_hidden_eval(model_num, invec);
    
    /*
     * Generate one sum of the Correlation and Partial matrices Only
     * calculates the upper half, but then fills in the complete matrix
     * before the matrix inversion
     * -------------------------------------------
     */
    for (i = 0; i < rpp->nhidden; i++)
      for (k = i; k < rpp->nhidden; k++)
        rpp->correl_matrix[i][k] +=
          rpp->hidden_out[i] * rpp->hidden_out[k];
    }
    /* Calculate the Partial Weight Matrix */
    /* This is used later to calculate the weights */
    for (k = 0; k < rpp->nhidden; k++)
    rpp->partial_weight_matrix[k][class] += rpp->hidden_out[k];
}

/************************************************************************
* NAME: rbf_batch_train(model_num,invec,desired);
* DESCRIPTION:
*   Takes rbf model number, ONE input vector, and vector of desired
*   outputs.
*   Trains the particular rbf model with that input.
*   Updates the M correlation and partial_weight_matrix matrices.
*   Must call rbf_build() and rbf_init() before this!
*   Note that final calculation of the weights (completely trained)
*   is not done here but is done with a call to rbf_batch_update()
*   after all the input training vectors have been run through this.
*************************************************************************/
void rbf_batch_train(int model_num, float *invec, float *desired)
{
    int i, k, nodei, nodek;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];


    if(rpp->fast_train){
    rbf_fast_hidden_eval(model_num, invec);
    for(i = 0; i < rpp->fast_nhidden; i++){
        nodei = rpp->fast_index[i];
        for(k = i; k < rpp->fast_nhidden; k++){
        nodek = rpp->fast_index[k];
        rpp->correl_matrix[nodei][nodek] +=
          rpp->hidden_out[nodei]* rpp->hidden_out[nodek];
        }
    }
    } else {
    /* Generate the rbf outputs for one pattern */
    rbf_hidden_eval(model_num, invec);

    /*
     * Generate sum of the Correlation and Partial matrices Only calculates
     * the upper half, but then fills in the complete matrix before the
     * matrix inversion -------------------------------------------
     */
    for (i = 0; i < rpp->nhidden; i++)
      for (k = i; k < rpp->nhidden; k++)
        rpp->correl_matrix[i][k] +=
          rpp->hidden_out[i] * rpp->hidden_out[k];
    }

    /* Calculate the Partial Weight Matrix */
    /* This is used later to calculate the weights */
    for (i = 0; i < rpp->nclasses; i++)
    for (k = 0; k < rpp->nhidden; k++)
        rpp->partial_weight_matrix[k][i] +=
          rpp->hidden_out[k] * desired[i];
}

/************************************************************************
* NAME: rbf_batch_update(model_num, max_ratio)
* DESCRIPTION:
*   Takes the rbf model number and maximum eigenvalue ratio
*       and updates the weights.
*   Inversion of M, calculation of weights.
*   This is called after all the training data has been processed
*   with the rbf_batch_ctrain() routine.
*   Deallocates the correlation and partial_weight_matrix matrix
*       after it is done.
*************************************************************************/
void rbf_batch_update(int model_num, float max_ratio)
{
    int i, j, k;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    /*
     * Fill in the lower half of the correlation matrix
     * ---------------------------------------------------
     */
    for (i = 1; i < rpp->nhidden; i++)
    for (k = 0; k < i; k++)
        rpp->correl_matrix[i][k] = rpp->correl_matrix[k][i];

    /*
     * print out correlation matrix if desired
     * --------------------------------------------
     */
    if (verbose > 0) {
    printf("\ncorrelation matrix  ( ... hidden node ... )\n");
    for (i = 0; i < rpp->nhidden; i++) {
        printf("node %2d: ", i);
        for (k = 0; k <= i; k++)
        printf("%g ", rpp->correl_matrix[i][k]);
        printf("\n");
    }
    }
    /*
     * print out partial weight matrix if desired
     * --------------------------------------------
     */
    if (verbose > 0) {
    printf("\npartial weight matrix  ( ... hidden node ...)\n");
    for (i = 0; i < rpp->nclasses; i++) {
        printf("class %2d: ", i);
        for (k = 0; k < rpp->nhidden; k++)
        printf("%g ", rpp->partial_weight_matrix[k][i]);
        printf("\n");
    }
    }
    /*
     * Invert the correlation Matrix ----------------------------------
     */
    inverse_correl_matrix =
      invertM(rpp->nhidden, rpp->correl_matrix, max_ratio);

    /*
     * Generate the weight matrix from accumulated matrices
     * -------------------------------------------------------
     */
    for (i = 0; i < rpp->nhidden; i++)
      for (j = 0; j < rpp->nclasses; j++)
    for (k = 0; k < rpp->nhidden; k++)
      rpp->weights[i][j] +=
        inverse_correl_matrix[i][k] * rpp->partial_weight_matrix[k][j];

    /*
     * print out weight matrix ----------------------------
     */
    if (verbose > 0) {
    printf("\nweight matrix  ( ... class ... )");
    for (i = 0; i < rpp->nhidden; i++) {
        printf("node %d: ", i);
        for (j = 0; j < rpp->nclasses; j++)
        printf("%g ", rpp->weights[i][j]);
        printf("\n");
    }
    }
}

/************************************************************************
* NAME: rbf_eval(model_num, invec)
* DESCRIPTION:
*   Takes rbf model number and a pointer to an input and output
*   vector. Fills in output vector.
*       Extra_hspread multiplies the spreads (variances) of all
*       basis functions uniformly.
*       It only does this for this evaluation and is not permanent.
*************************************************************************/
void rbf_eval(int model_num, float *invec)
{
    int j, k;
    RBF_PARAMS *rpp;
    float total;

    rpp = rbfpp[model_num];

    /* generate the rbf outputs */
    rbf_hidden_eval(model_num, invec);

    /* Calculate the output vector */
    for (j = 0; j < rpp->nclasses; j++) {
    rpp->outputs[j] = 0.0;
    for (k = 0; k < rpp->nhidden; k++)
      rpp->outputs[j] += rpp->weights[k][j] * rpp->hidden_out[k];
    }

    if(rpp->adjust_priors){
    /* adjust priors */
    for(j = 0; j < rpp->nclasses; j++){
        rpp->outputs[j] *= rpp->priors_list[j];
        rpp->outputs[j] /= rpp->class_cnt[j]/rpp->total_npatterns;
    }
    /* scale adjusted outputs to sum to 1 */
    total = 0.;
    for(j = 0; j < rpp->nclasses; j++) total += rpp->outputs[j];
    for(j = 0; j < rpp->nclasses; j++) rpp->outputs[j] /= total;
    }
}

/************************************************************************
* NAME: rbf_classify(model_num, invec)
* DESCRIPTION:
*   Takes rbf model number and a pointer to an input vector.
*   Returns the class assigned to the input vector and fills in the
*       output vector.
*       Extra_hspread multiplies the spreads (variances) of all
*       basis functions uniformly.
*       It only does this for this evaluation and is not permanent.
*************************************************************************/
int rbf_classify(int model_num, float *invec)
{
    int outclass;   /* calculated class of the invec */
    RBF_PARAMS *rpp;


    rpp = rbfpp[model_num];

    rbf_eval(model_num, invec);

    /* Pick the maximum out of the Z array list */
    outclass = max_pick(rpp->outputs, rpp->nclasses);

    /* Return the class */
    return (outclass);
}

/************************************************************************
* NAME: rbf_hidden_eval(model_num, invec)
* DESCRIPTION:
*   Takes as input a pointer to the input vector and a pointer to
*       an output vector that will contain the hidden node outputs on
*       return.
*       Extra_hspread multiplies the spreads (variances) of all
*       basis functions uniformly.
*       It only does this for this evaluation and is not permanent.
**************************************************************************/
void
rbf_hidden_eval(int model_num, float *invec)
{
    int node, i;
    float diff, factor;
    float total_log;
    float ave_log;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    /*
     * calculate log10() outputs of all gaussian nodes first
     * ------------------------------------------------
     */
    factor = -M_LOG10E / (2.0 * rpp->hspread * rpp->extra_hspread);
    for (node = 0; node < rpp->ngauss_hidden; node++) {
    rpp->hidden_out[node] = 0.;
    for(i = 0; i < rpp->ninputs; i++){
        diff = invec[i] - rpp->centroids[node][i];
        diff *= diff;
        rpp->hidden_out[node] += diff / rpp->spreads[node][i];
    }
    rpp->hidden_out[node] *= factor;
    }

    if(rpp->norm_outputs){
    /* compute average node output */
    total_log = rpp->hidden_out[0];
    for (node = 1; node < rpp->ngauss_hidden; node++)
      total_log = logadd(total_log, rpp->hidden_out[node]);
    ave_log = total_log - log10((double) rpp->ngauss_hidden);

    /* normalize node outputs and convert to linear */
    for (node = 0; node < rpp->ngauss_hidden; node++)
      rpp->hidden_out[node] =
        nn_exp10((double) (rpp->hidden_out[node] - ave_log));
    } else {
    /* just convert to linear */
    for (node = 0; node < rpp->ngauss_hidden; node++)
      rpp->hidden_out[node] = nn_exp10((double) (rpp->hidden_out[node]));
    }

    /* add output of bias node with constant output */
    if (rpp->bias) {
    node = rpp->nhidden - 1;
    rpp->hidden_out[node] = 1.0;
    }
}

/************************************************************************
* NAME: rbf_fast_hidden_eval(model_num, invec)
* DESCRIPTION:
*   Just keep the top 10 outputs in the hidden layer
*   This speeds up the correlation matrix calculations
**************************************************************************/
void
rbf_fast_hidden_eval(int model_num, float *invec)
{
    int node, i;
    float diff, factor, output;
    float total_log;
    float ave_log;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    /*
     * calculate log10() outputs of all gaussian nodes first
     * ------------------------------------------------
     */
    factor = -M_LOG10E / (2.0 * rpp->hspread * rpp->extra_hspread);
    for (i = 0; i < rpp->fast_nhidden; i++){
    rpp->fast_outputs[i] = -MAXFLOAT;
    rpp->fast_index[i] = 0;
    }
    for (node = 0; node < rpp->ngauss_hidden; node++) {
    rpp->hidden_out[node] = 0.;
    output = 0.;
    for(i = 0; i < rpp->ninputs; i++){
        diff = invec[i] - rpp->centroids[node][i];
        diff *= diff;
        output += diff / rpp->spreads[node][i];
    }
    output *= factor;

    /* if this new node is one of the biggest */
    if(output > rpp->fast_outputs[rpp->fast_nhidden -1]){
        /* find the right place for it */
        for (i = rpp->fast_nhidden-1;
         i > 0 && output > rpp->fast_outputs[i-1]; i--){
        rpp->fast_outputs[i] = rpp->fast_outputs[i-1];
        rpp->fast_index[i] = rpp->fast_index[i-1];
        }
        rpp->fast_outputs[i] = output;
        rpp->fast_index[i] = node;
    }
    }
    /* copy in the surviving fast_nhidden biggest outputs */
    for(node = 0; node < rpp->fast_nhidden; node++){
    rpp->hidden_out[rpp->fast_index[node]] = rpp->fast_outputs[node];
    }

    if(rpp->norm_outputs){
    /* compute average node output, just using the biggest ones */
    total_log = rpp->hidden_out[rpp->fast_index[0]];
    for (node = 1; node < rpp->fast_nhidden; node++)
      total_log = logadd(total_log, rpp->hidden_out[rpp->fast_index[node]]);
    ave_log = total_log - log10((double) rpp->ngauss_hidden);

    /* normalize node outputs and convert to linear */
    for (i = 0; i < rpp->fast_nhidden; i++){
        node = rpp->fast_index[i];
        rpp->hidden_out[node] =
          nn_exp10((double) (rpp->hidden_out[node] - ave_log));
    }
    } else {
    /* just convert to linear */
    for (i = 0; i < rpp->fast_nhidden; i++){
        node = rpp->fast_index[i];
        rpp->hidden_out[node] =
          nn_exp10((double) (rpp->hidden_out[node]));
    }
    }

    /* add output of bias node with constant output */
    if (rpp->bias) {
    node = rpp->nhidden - 1;
    rpp->hidden_out[node] = 1.0;
    }
}

/************************************************************************
* NAME: rbf_gen_probability(model_num, invec, extra_hspread)
* DESCRIPTION:
*   Takes as input a pointer to the input vector and print out the
*       log probability of each hidden node when the hidden node is
*       used as the Gaussian function in a Gaussian classifier
*       Extra_hspread multiplies the spreads (variances) of all
*       basis functions uniformly.
#
**************************************************************************/
void
rbf_gen_probability(int model_num, float *invec, float extra_hspread)
{
    int node, i;
    float diff, num, node_sum, factor;
    float determinant;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    /*
     * calculate log10() outputs of all Gaussian nodes and print them
     * --------------------------------------------------------------
     */
    factor = -M_LOG10E / (2.0 * rpp->hspread * extra_hspread);
    for (node = 0; node < rpp->ngauss_hidden; node++) {
    node_sum = 0.0;
    determinant = 0.0;
    for (i = 0; i < rpp->ninputs; i++) {
        diff = invec[i] - rpp->centroids[node][i];  /* difference */
        diff *= diff;   /* differenc squared */
        num = diff / rpp->spreads[node][i]; /* numerator */
        determinant += (rpp->spreads[node][i] * rpp->spreads[node][i]);
        node_sum += num;    /* accumulate sum */
    }
    node_sum *= factor;
    node_sum -= 0.5 * rpp->ninputs * LOG_2PI;
    node_sum -= 0.5 * log10((double) determinant);
    printf("%g\t", node_sum);
    }
}

void
rbf_write_paramfile(int model_num, FILE *fp)
{
    rbf_write_model(model_num, fp);
}

void
rbf_read_paramfile(FILE *fp)
{
    rbf_read_model(fp);
}

void
print_centers_spreads(int model_num)
{
    int i, j;
    RBF_PARAMS *rpp;

    rpp = rbfpp[model_num];

    printf("\n rbf centers ( ... input index ... )\n");
    for (i = 0; i < rpp->ngauss_hidden; i++) {
    printf("node %2d: ", i);
    for (j = 0; j < rpp->ninputs; j++)
        printf("%g ", rpp->centroids[i][j]);
    printf("\n");
    }

    printf("\n rbf spreads,variances ( ... input index ... )\n");
    for (i = 0; i < rpp->ngauss_hidden; i++) {
    printf("node %2d: ", i);
    for (j = 0; j < rpp->ninputs; j++)
        printf("%g ", rpp->spreads[i][j]);
    printf("\n");
    }
}

/*****************************************
rbf_read_cluster()
*****************************************/
void
rbf_read_cluster(char *param_path, FILE *param_file, int *pninputs,
         int *pncenters, float ***centroids, float ***spreads,
         int *features)
{
    int model, noutputs;

    model = cluster_read_paramfile(param_file, pninputs, &noutputs, pncenters,
                   features);
    if(model < 0)
      fatal_error("rbf: read error %d on parameter file %s", model,
          param_path);

    if(*pncenters <= 0)
      fatal_error("rbf_read_cluster: No centers from clustering");

    /* insure that the centers are not stored by class */
    cluster_merge_centers(model);

    *centroids = *(float ***)cluster_param(model, "pcenter", 0);
    *spreads = *(float ***)cluster_param(model, "pvariance", 0);
}
