/***--------------------------------------------------------------------------
* NAME:     rbf_seq_pack.c
* PURPOSE:  functions for sequence classification using rbf classifier
*
* SEE ALSO: rbf_pack.c
*
* GLOBALS:
*
* ALGORITHM:
* REFERENCE:
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_seq_pack.c,v 11.3 2001/08/23 20:12:52 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_seq_pack.c,v $
* CREATED: 17-Sep-93 lck                MIT LINCOLN LABORATORY
*--------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include <math.h>
#include "rbf_seq_pack.h"

typedef struct {
    int noutputs;   /* number of elements in patterns */
    float *seq_sums;
    float clip_floor;   /* lowest permitted output for seq_sums */
    int rbf_model;
} RBF_SEQ_PARAMS;

                /* Internal Globals */
static RBF_SEQ_PARAMS **rbfpp;/* array of parameters for each model */
static int      max_models = 0; /* current length of rbfpp */

#define SMALLEST 1.e-30

/*****************************************************************************
* NAME: rbf_seq_build(noutputs)
*
* DESCRIPTION:
*       Allocate storage for model parameters.
*
*****************************************************************************/
int rbf_seq_build(int noutputs)
{
    RBF_SEQ_PARAMS *rpp;    /* temporary pointer to rbfpp[] */
    register int    new_model;

    new_model = model_get_number((void ***)&rbfpp,
                 &max_models, sizeof(RBF_SEQ_PARAMS));

    rpp = rbfpp[new_model];

    /* set mode flags */
    rpp->noutputs = noutputs;

    /* allocate space for vectors */
    rpp->seq_sums = allocate_vector(noutputs);

    return new_model;
}
/*****************************************************************************
* NAME: rbf_seq_init(model)
* DESCRIPTION:
*       Initialize model parameters not set in build
*
* EXAMPLE:
*
*****************************************************************************/
void rbf_seq_init(int model, int rbf_model, float clip_floor)
{
    int             i;
    RBF_SEQ_PARAMS *rpp;    /* temporary pointer to rbfpp[] */

    rpp = rbfpp[model];

    rpp->rbf_model = rbf_model;
    rpp->clip_floor = clip_floor;
    for(i = 0; i < rpp->noutputs; i++) rpp->seq_sums[i] = 0.;
}

/*****************************************************************************
* NAME: rbf_seq_free()
* DESCRIPTION:
*       Deallocate storage used in a model
*
* EXAMPLE:
*
*****************************************************************************/
void
rbf_seq_free(int model)
{
    RBF_SEQ_PARAMS *rpp;    /* temporary pointer to rbfpp[] */

    rpp = rbfpp[model];
    /* free vectors and matrices */
    deallocate_vector(rpp->seq_sums);

    /* free structure itself */
    nn_free((char *) rpp, "for rbf parameter structure");
    rbfpp[model] = NULL;
}

/************************************************************************
* NAME: rbf_seq_param(model, param_name)
* DESCRIPTION:
*       Takes model number and parameter name.
*       Returns a pointer to the named parameter.
*
*************************************************************************/
#define PEQ(y) (0==strcmp(param_name,y))

void * rbf_seq_param(int model, char *param_name)
{
    void           *param_address;
    RBF_SEQ_PARAMS *rpp;    /* temporary pointer to rbfpp[] */

    rpp = rbfpp[model];
    if (PEQ("seq_sums"))
      param_address = &rpp->seq_sums;
    else {
    warning("rbf_seq_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}

/* collect outputs of model for series of input frames.
 * summing the log outputs */
void
rbf_seq_batch_classify(int model, float *inputs)
{
    int i;
    RBF_SEQ_PARAMS *rpp;    /* temporary pointer to rbfpp[] */
    void *rbf_param(), rbf_eval();
    float *outputs;

    rpp = rbfpp[model];
    rbf_eval(rpp->rbf_model, inputs);
    outputs = *(float **)rbf_param(rpp->rbf_model, "outputs");

    /* rbf cannot give log outputs yet */
    for(i = 0; i < rpp->noutputs; i++){
    /* clip the outputs */
    if(outputs[i] < rpp->clip_floor) outputs[i] = rpp->clip_floor;
    if(outputs[i] > 1.) outputs[i] = 1.;
    rpp->seq_sums[i] += log10((double)outputs[i]);
    }
}

int
rbf_seq_update_classify(int model)
{
    RBF_SEQ_PARAMS *rpp;
    rpp = rbfpp[model];
    return(max_pick(rpp->seq_sums, rpp->noutputs));
}
    
