/***--------------------------------------------------------------------------
 * NAME:	rbf_seq_train.c
 * PURPOSE:	training driver for Radial Basis Function sequence classifier
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_seq_train.c,v 11.4 2001/08/23 20:12:53 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_seq_train.c,v $
 * CREATED: 17-Sep-93 lck				MIT LINCOLN LABORATORY
 ****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "rbf_seq.h"
#include "rbf_pack.h"
#include "rbf_seq_pack.h"

#define MAXSTR (256)

/* globals */
int debug;			/* enables debug messages if debug > 0*/

static char     prog_name[MAXSTR];	/* static variables */

void
rbf_seq_train(int noutputs, int nsequences, int debg, int verbosity,
	      char *data_path, char *data_fname, char *exp_path,
	      char *param_fname, char **labels, char *comment, int bias,
	      float hspread, float max_ratio, char *cluster_param_fname,
	      float minimum_variance, int norm_outputs)
{
    /* local variables */
    int		model, data_model, rv, normalization = 0;
    int		*features, nraw, ninputs, model_ninputs;
    int		nwindow_frames, *window_frames;
    int		n_frames, ncenters, n_files, desired_class;
    float	*inputs, **pcentroids, **pspreads;
    char	*tmp_path, *clus_path;
    FILE	*param_file, *clus_file;
    
    debug = debg;		/* set globals */
    strcpy(prog_name, get_program_name());
    
    /***************** read the cluster file *****************/
    clus_path = path_build(exp_path, cluster_param_fname);
    
    clus_file = nn_fopen(clus_path, "r", "for read of cluster parameters");
    if (verbosity > 1) {
	printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	fflush(stdout);
    }
    /* get normalization from cluster file */
    model = norm_read_paramfile(clus_file, &normalization);
    if (model < 0)
      fatal_error("rbf: read error %d %s %s",
		      model, "on normalization parameters in file", clus_path);
    rv = feature_read_paramfile(clus_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on feature selection parameters in file", clus_path);
    rv = window_read_paramfile(clus_file, &nwindow_frames, &window_frames);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s", prog_name, rv,
		  "on windowing parameters in file", clus_path);
    
    /* read clusters */
    /* features was allocated in feature_read_paramfile, so it is ok here */
    rbf_read_cluster(clus_path, clus_file, &model_ninputs,
		     &ncenters, &pcentroids, &pspreads, features);
    
    nn_fclose(clus_file, clus_path, "for read of cluster parameters");
    
    /********** build the rbf model ***************/
    if (verbosity > 1){
	printf ("\nBuilding rbf model\n");
	fflush(stdout);
    }
    model = rbf_build (model_ninputs, ncenters, bias, noutputs);
    rbf_init (model, pcentroids, pspreads, hspread, debug, features,
	      minimum_variance, norm_outputs);

    /* read data */
    data_model = seq_data_read(data_path, data_fname, nsequences, nraw,
			       ninputs, noutputs, features, normalization,
			       labels, verbosity, nwindow_frames,
			       window_frames);
    nsequences = seq_data_get_nfiles(data_model);

    for(n_frames = n_files = 0; n_files < nsequences; n_files++){
	if(seq_data_get_next_file(data_model) == -1) break;
	desired_class = seq_data_get_class(data_model);
	for(; ; n_frames++){
	    inputs = seq_data_get_next_frame(data_model);
	    if(inputs == NULL) break;
	    rbf_batch_ctrain(model, inputs, desired_class);
	}
    }
    if(n_frames > 0){
	if (verbosity > 1) {
	    printf("\nInverting rbf matrix\n");
	    fflush(stdout);
	}
	rbf_batch_update(model, max_ratio);
    }
    
    /******************** store model in param file ********************/
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "feature selection\n");
    feature_write_paramfile(param_file, features, nraw, ninputs);
    fprintf(param_file, "windowing parameters\n");
    window_write_paramfile(param_file, window_frames, nwindow_frames);
    fprintf(param_file, "rbf model\n");
    rbf_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0){
	printf("\nFinished -- rbf model saved in \"%s\"\n", tmp_path);
	fflush(stdout);
    }
    /* free space */
    norm_free();
    rbf_free(model);
    seq_data_exit(data_model);
}
