/***--------------------------------------------------------------------------
 * NAME:        rbf.c
 * PURPOSE: driver for Radial Basis Function classifier
 *
 * SEE ALSO:    rbf_pack.c
 *
 * FUNCTIONS:
 *  driver for Radial Basis Function classifier
 *      handles model creation, training and testing of data
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_train.c,v 11.3 2001/08/23 20:12:53 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbf_train.c,v $
 * CREATED: 20-Mar-91 dan               MIT LINCOLN LABORATORY
 ****************************************************************************/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include <nnlib.h>
#include "rbf.h"
#include "rbf_pack.h"

#define MAXSTR (256)
static char     prog_name[MAXSTR];  /* static variables */
static int  verbose;

void
rbf_train(int nraw, int ninputs, int *pfeatures, int noutputs, int npatterns,
      int debg, int verbosity, int random_flag, char *data_path,
      char *data_fname, char *exp_path, char *param_fname, char *comment,
      int adjust_priors, int priors_npatterns, float *priors_list,
      int bias, float hspread, float max_ratio, char *cluster_param_fname,
      float minimum_variance, int norm_outputs, int fast_train,
      int fast_nhidden)
{
    /* local variables */
    int     model, normalization = 0, data_model;
    int     desired_class;
    int     ncenters;
    float   *inputs, **pcentroids, **pspreads;
    char    *tmp_path, clus_path[MAXSTR];
    FILE    *param_file, *clus_file;
    int     *features;
    
    strcpy(prog_name, get_program_name());
    verbose = verbosity;
    
    /* read the cluster file */
    tmp_path = path_build(exp_path, cluster_param_fname);
    strncpy(clus_path, tmp_path, MAXSTR - 1);
    
    clus_file = nn_fopen(clus_path, "r",
             "for read of cluster parameters");
    if (verbose > 1) {
    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
    fflush(stdout);
    }
    model = norm_read_paramfile(clus_file, &normalization);
    
    if (model < 0)
      fatal_error("rbf: read error %d %s %s",
          model, "on normalization parameter file", clus_path);
    
    rbf_read_cluster(clus_path, clus_file, &ninputs,
             &ncenters, &pcentroids, &pspreads, pfeatures);
    
    nn_fclose(clus_file, clus_path, "for read of cluster parameters");
    
    if (verbosity > 1)
      printf ("\nBuilding rbf model\n");
    model = rbf_build (ninputs, ncenters, bias, noutputs);
    rbf_init (model, pcentroids, pspreads, hspread, debg, pfeatures,
          minimum_variance, norm_outputs);
    if(fast_train)
      rbf_init_fast_train(model, fast_nhidden);

    /* use features from the model */
    features = *(int **) rbf_param(model, "features");
    ninputs = *(int *) rbf_param(model, "ninputs");
    
    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if(adjust_priors){
    data_model = class_data_read(data_path, data_fname, npatterns,
                     nraw, ninputs, noutputs, features,
                     normalization, verbosity);
    if(random_flag) class_data_random_init(data_model);
    class_data_priors_init(data_model, priors_list, priors_npatterns);
    } else {
    data_model = class_data_single_init(data_path, data_fname, npatterns,
                        nraw, ninputs, noutputs, features,
                        normalization, verbosity);
    }
    while((inputs = class_data_get_next(data_model, &desired_class)) != NULL){
    rbf_batch_ctrain(model, inputs, desired_class);
    }
    if (verbosity > 1) {
    printf("\nInverting rbf matrix\n");
    fflush(stdout);
    }
    rbf_batch_update(model, max_ratio);
    
    class_data_exit(data_model);
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "rbf model\n");
    rbf_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- rbf model saved in \"%s\"\n", tmp_path);
    
    
    /* free space and close files */
    norm_free();
    rbf_free(model);
}
