/***--------------------------------------------------------------------------
 * NAME:	rbfm_test.c
 * PURPOSE:	testing driver for Radial Basis Function mapper
 *
 * SEE ALSO: 	rbf_pack.c
 * GLOBALS:	int debug;
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbfm_test.c,v 11.4 2001/08/23 20:12:53 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbfm_test.c,v $
 * CREATED: 15-Oct-92 lck			MIT LINCOLN LABORATORY
 *--------------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "rbfm.h"
#include "rbf_pack.h"

#define MAXSTR (256)
/* globals */
int debug;			/* enables debug messages if debug > 0*/

static char     prog_name[MAXSTR];	/* static variables */

void
rbfm_test(int create_flag, int map_flag, int nraw, int ninputs, int noutputs,
	  int npatterns, int debg, int verbosity, int verror, char *data_path,
	  char *data_fname, char *exp_path, char *error_fname,
	  char *param_fname, char **labels, float extra_hspread)
{
    /* local variables */
    int		model, err_model, rv;
    int		normalization = 0, n_read, log_outputs = 0, post_sum_to_one = 0;
    float	*inputs, *outputs, *desired_outputs;
    char	data_fullname[MAXSTR], *tmp_path;
    FILE	*data_file, *error_file, *param_file;
    int		*features;
    
    debug = debg;		/* set globals */
    strcpy(prog_name, get_program_name());
    
    /* put together the full path to the data file*/
    strcpy(data_fullname, path_build(data_path, data_fname));
    
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "r", "for read of model file");
    if (verbosity > 1)
      printf("\nReading rbf parameters from \"%s\"\n", tmp_path);
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0)
      fatal_error("%s: read error %d %s %s",
		  prog_name, rv, "on normalization parameter file",
		  tmp_path);
    
    model = rbf_read_model(param_file);
    if (model < 0)
      fatal_error("rbf: read error %d on model file %s", model, tmp_path);
    nn_fclose(param_file, tmp_path, "for read of rbf parameters");
    
    if (debug > 0)
      print_centers_spreads(model);
    
    /* use features from the model */
    features = *(int **) rbf_param(model, "features");
    ninputs = *(int *) rbf_param(model, "ninputs");
    outputs = *(float **) rbf_param(model, "outputs");

    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    if (verror > 0) {		/* open file for classification errors */
	if(create_flag)
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"w", "for classification errors");
	else
	  error_file = nn_fopen(path_build(exp_path, error_fname),
				"a", "for classification errors");
    }
    /* do remaining initializations and allocations */
    err_model = report_build (verbosity, verror, noutputs);
    report_init (err_model, map_flag, ninputs, SQ_ERR, log_outputs, post_sum_to_one);

    /* open data file */
    data_file = nn_fopen(data_fullname, "r", "for training or test data");
    if (verbosity > 1) printf("\nReading %s\n", data_fullname);
    
    inputs = allocate_vector(ninputs);
    desired_outputs = allocate_vector(noutputs);
    rbf_test_init(model, 0, NULL, extra_hspread);
    
    /* read and classify input from data file */
    for (n_read = 0; n_read < npatterns; n_read++) {
	if(get_map_line(data_file, data_fullname, inputs, desired_outputs,
			nraw, ninputs, noutputs, normalization, features)
	   == EOF) break;
	
	/* classify pattern and report errors */
	rbf_eval(model, inputs);
	
	report_map_trial(err_model, error_file,
			 inputs, outputs, desired_outputs, n_read);
    }
    /* free space and close files */
    deallocate_vector(inputs);
    deallocate_vector(desired_outputs);
    
    /* finish error reporting and close error file */
    report_overall(err_model, stdout, error_fname, labels);
    report_free(err_model);
    if (verror > 0) {
	nn_fclose(error_file, path_build(exp_path, error_fname),
		  "for error file");
    }

    nn_fclose(data_file, data_fullname, "for training or test data");
    rbf_free(model);
    norm_free();
}
