/***--------------------------------------------------------------------------
 * NAME:	rbfm_train.c
 * PURPOSE:	training driver for Radial Basis Function mapper
 *
 * SEE ALSO: 	../rbf/rbf_pack.c
 * GLOBALS:	int debug;
 *
 * FUNCTIONS:	rbf_train()
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbfm_train.c,v 11.4 2001/08/23 20:12:53 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/rbf/rbfm_train.c,v $
 * CREATED: 15-Oct-92 lck				MIT LINCOLN LABORATORY
 *--------------------------------------------------------------------------**/

/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "rbfm.h"
#include "rbf_pack.h"

#define MAXSTR (256)

#define NO_NORM_OUTPUTS (0)

/* globals */
int debug;			/* enables debug messages if debug > 0*/

static char     prog_name[MAXSTR];	/* static variables */

void
rbfm_train(int create_flag, int nraw, int ninputs, int *pfeatures,
	   int noutputs, int npatterns, int debg, int verbosity,
	   char *data_path, char *data_fname, char *exp_path,
	   char *param_fname, char *comment, int bias, float hspread,
	   float max_ratio, char *cluster_param_fname, float minimum_variance)
{
    /* local variables */
    int		model, rv, normalization = 0;
    int		n_read, ncenters;
    float	*inputs, *outputs, **pcentroids, **pspreads;
    char	data_fullname[MAXSTR], *tmp_path, clus_path[MAXSTR];
    FILE	*data_file, *param_file, *clus_file;
    int		*features;
    
    debug = debg;		/* set globals */
    strcpy(prog_name, get_program_name());
    
    /* put together the full path to the data file*/
    strcpy(data_fullname, path_build(data_path, data_fname));
    
    if (create_flag){
	/* read the cluster file */
	tmp_path = path_build(exp_path, cluster_param_fname);
	strncpy(clus_path, tmp_path, MAXSTR - 1);
	
	clus_file = nn_fopen(clus_path, "r",
			     "for read of cluster parameters");
	if (verbosity > 1) {
	    printf("\nReading cluster parameters from \"%s\"\n", clus_path);
	    fflush(stdout);
	}
	model = norm_read_paramfile(clus_file, &normalization);
	
	if (model < 0)
	  fatal_error("rbf: read error %d %s %s",
		      model, "on normalization parameter file", clus_path);
	
	rbf_read_cluster(clus_path, clus_file, &ninputs,
			 &ncenters, &pcentroids, &pspreads, pfeatures);
	
	nn_fclose(clus_file, clus_path, "for read of cluster parameters");
	
	if (verbosity > 1)
	  printf ("\nBuilding rbf model\n");
	model = rbf_build (ninputs, ncenters, bias, noutputs);
	rbf_init (model, pcentroids, pspreads, hspread, debug, pfeatures,
		  minimum_variance, NO_NORM_OUTPUTS);
    } else {	/* retrain */
	tmp_path = path_build(exp_path, param_fname);
	param_file = nn_fopen(tmp_path, "r", "for read of model file");
	if (verbosity > 1)
	  printf("\nReading rbf parameters from \"%s\"\n", tmp_path);
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error("%s: read error %d %s %s",
		      prog_name, rv, "on normalization parameter file",
		      tmp_path);
	
	model = rbf_read_model(param_file);
	if (model < 0)
	  fatal_error("rbf: read error %d on model file %s", model, tmp_path);
	nn_fclose(param_file, tmp_path, "for read of rbf parameters");
	
	verify_model_params (model, ninputs,
			     *(int *) rbf_param(model, "ninputs"), pfeatures,
			     *(int **) rbf_param(model, "features"));
	if (debug > 0)
	  print_centers_spreads(model);
    }
    /* use features from the model */
    features = *(int **) rbf_param(model, "features");
    ninputs = *(int *) rbf_param(model, "ninputs");

    /* print various debugging and information messages before starting
     * actual training or testing
     */
    display_model_params (model, ninputs, normalization, features, verbosity);
    
    /* open data file */
    data_file = nn_fopen(data_fullname, "r", "for training or test data");
    if (verbosity > 1)
      printf("\nReading %s\n", data_fullname);
    
    inputs = allocate_vector(ninputs);
    outputs = allocate_vector(noutputs);
    for (n_read = 0; n_read < npatterns; n_read++) {
	/* read input vector */
	if(get_map_line(data_file, data_fullname, inputs, outputs, nraw,
			ninputs, noutputs, normalization, features)
	   == EOF) break;
	rbf_batch_train(model, inputs, outputs);
    }
    if (verbosity > 1) {
	printf("\nInverting rbf matrix\n");
	fflush(stdout);
    }
    rbf_batch_update(model, max_ratio);
    deallocate_vector(inputs);
    deallocate_vector(outputs);
    
    /* store model in param file */
    tmp_path = path_build(exp_path, param_fname);
    param_file = nn_fopen(tmp_path, "w", "for write of model file");
    fprintf(param_file, "%s\n", comment);
    fprintf(param_file, "normalization data\n");
    norm_write_paramfile(param_file);
    fprintf(param_file, "rbf model\n");
    rbf_write_model(model, param_file);
    nn_fclose(param_file, tmp_path, "for write of model file");
    if ( verbosity > 0)
      printf("\nFinished -- rbf model saved in \"%s\"\n", tmp_path);
    
    
    nn_fclose(data_file, data_fullname, "for training or test data");
    /* free space and close files */
    norm_free();
    rbf_free(model);
}
