/***-------------------------------------------------------------------------
* NAME: svm_pack.c
* PURPOSE:     support vector machine classification routines
* USAGE:
*
* SEE ALSO:     
* DESCRIPTION:  support vector machine classification routines.
*       An example of their use is in svm.c. 
*       Implements John Klatt's Sequential Mininal Optimization (SMO)
*       algorithm as improved by Keerthi, Shevade, Bhattacharyya and
*       Murthy. 
*            Routines ending in _multi are for one high-level multi-class controller 
*       classifier that controls many binary low-level component classifiers. 
*       One high-level controller is built for every problem. There are one or more low-level
*       classifiers. If there are M classes, then there are either
*       M low level classifiers (each class versus all other classes) or there are
*       M*(M-1)/2 low level classifiers (all combinations of classes two at a time).
*            The output of the low level classifier is an unbounded linear or 
*       nonlinear function of the inputs. A sigmoid is always fit to this output
*       to provide a good estimate of the output posterior probability for
*       the training data (weighting points near +/- 1 much less in the fit).
*       The sigmoid is or is not used according to the flag "sigmoid_fit".
*       The sigmoid has an output of 0.5 when the input is 0.0 and the slope
*       is adjusted based on a squared error fit to desired outputs of 0 and 1
*       for the two classes.
*            A decision from all classes is made by scanning across all contributing
*       binary classifiers. For M low-level classifiers, the class with the highest
*       output is selected. For M*(M-1)/2 pairs, the minimum output for each class is
*       found across all binary classifiers. The final class is that with the highest
*       minimum output. This assumes that the outputs are roughly posteriors because
*       then, they are upper bounds on the desired overall class prior probability.
*       In practice, the result is almost identical to voting for each class.
*            For linear svm classifiers, the support vectors are not typicallly 
*       stored and only the weights are saved. Internals plots show the lines formed
*       by these weights as in a multi-layer perceptron.
*       For nonlinear classifiers, internals plots both 
*       support vectors at the upper bound (circles) and support vectors that are
*       unbounded (circles with x's). A special type "LINEAR_USE_SVS" of classifier is
*       provided that is a linear classifier that uses and stores support vectors for
*       classification. When plotted, internals shows where the support vectors are.
*       This classifier is less efficient than the straight linear classifier.
*       It is selected using -kernel_type 4 from the main svm driver.
*            Training for svms is different from other classifiers because multiple
*       classifiers must be trained. The svm_train_multi routine needs to be passed
*       the identity of a data model to retrieve data from for all sub classifiers.
*       It also has a flag named "cv_flag". If this is set then the structures required
*       for testing are left intact and can be used for subsequent testing. If it is
*       not set, all data is written to a parameter file and structures are freed.
*
*       Some modifications can be made to reduce training and test times if you have
*       some time for programming changes.
*       Training: (1) For sparse long vectors compute the kernel by storing the vectors
*                     not as long float vectors by as integers indicating where components of
*                     the vectors are non-zero.
*                 (2) A lot of memory is allocated for training that is not always needed.
                      The memory use can be made more efficient with more work.
*   Testing:  (1) I already keep track of the support vectors that are unique across
*                     all the component classifiers. For each component classifier, the 
*                     kernel output for the non-unique support vectors could be stored once
*                     in a cache and then re-used for the other classifiers that use the
*                     same support vector.
*                 (2) The combination of classifier outputs can stop if a majority already
*                     has been reached.
*                     
*
* HIGH-LEVEL CONTROLLER CLASSIFIER :
*   svm_build_multi - allocate space for controller to keep track of component classifiers 
*   svm_train_multi - train multiple binary support vector machines and store to param file
*       svm_test_init_multi - initialize multi model classifier for testing 
*   svm_classify_multi - classify one pattern for high-level multi class classifier 
*   svm_set_component_model_multi - set component model number in high-level classifier 
*   svm_set_class_cnt_multi - set component class count in high-level classifier 
*   svm_get_outputs_multi - get pointer to outputs for multi model 
*   svm_free_multi - deallocate memory used for testing high-level classifier
*   svm_write_model_multi - writes multi-classifier parameters to file 
*   svm_read_model_multi - reads parameters for multi-classifier from a file 
*   svm_read_model_multi_end - reset sv count parameters for multi-classifier stored at end of file 
*   svm_print_multi - writes multi-classifier information to std out 
*   svm_set_sv_multi - set support vector counts for multi model 
*   svm_param_multi - return a pointer to the named parameter 
*
* COMPONENT MODELS :
*   svm_build()     - allocate space for model
*   svm_init()  - initialize model prior to training 
*       svm_eval()      - provide output for one input pattern
*       svm_store_inputs() - get input exemplars into local matrix for fast access
*       svm_free()  - free up memory grabbed by svm_build
*       svm_read_model()    - read stored model from file
*       svm_write_model()   - save model in file
*       svm_param()         - access a parameter in the svm model
*   svm_get_nsv_pos_below_bound() - return number of support vectors that
*                                       are positive but below the c bound
*       svm_get_nsv_at_bound()  - return the number of support vectors that
*                                 are at the upper bound
*       svm_get_sv_pos_below_bound() - return jth component of the ith 
*                                      positive below bound support vector
*   svm_get_sv_at_bound()   - return jth component of the ith support
*                                 vector that is at the upper bound
* LOCAL FUNCTIONS:
*   svm_compute_output()        - compute raw svm output
*   svm_compute_no_bias_output()    - dont subtract bias term 
*   svm_compute_kernel()        - compute kernel for two patterns
*   svm_examine_example()       - are kkt conds satisfied for pattern?
*   svm_take_step()     - find new lagrange multipiers for two patterns
*   svm_compute_bounds()          - compute current alpha bounds
*   svm_print_positive_alphas()   - print lagrange multiplers  > 0
*   svm_update_alpha()        - update two lagrange multiplers
*   svm_sv_add()              - add new support vector to list
*   svm_sv_update()           - update existing support vector
*   svm_sv_remove()           - remove support vector from list
*s  svm_low_update()          - update lowest bias 
*   svm_high_update()         - update highest bias
*   svm_update_pair_corrections()     - increment counter for corrections
*   svm_print_model_parameters()      - print parameters
*   svm_set_fcache()              - put new output-bias in cache
*   svm_get_fcache()          - get output-bias from cache
*   svm_invalidate_fcache()       - mark fcache entry as bad
*   svm_init_fcache()         - initialize cache for output-bias
*   svm_fpattern2sv()         - convert pattern to sv index
*   svm_fsv2pattern()         - convert sv to pattern index
*   svm_update_fcache_positive_below_bound() - update all fcache entries
*                                                  for pbb support vectors
*   svm_update_bias_bounds()      - update low and high bias values
*   svm_print_summary_header()    - header for summary_values 
*   svm_print_summary_values()    - print summary during training
*   svm_fit_posterior_sigmoid()       - fit sigmoid to produce output 
*                                           posterior probabilities
*   svm_check_final_kkt_conditions()  - final check on kkt conds
*   svm_save_final_support_vectors()  -copy exemplars from training matrix to support vector matrix 
*
* ALGORITHM: see svm man page
* REFERENCE: see svm man page
*
* BUGS AND LIMITATIONS: see svm man page
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/svm/svm_pack.c,v 1.6 2004/01/15 19:14:03 kukolich Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/algorithm/svm/svm_pack.c,v $
* CREATED: 17-Dec-2000 rpl                           MIT LINCOLN LABORATORY
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory 

   Permission is hereby granted, without written agreement or royalty fee, 
   to use, copy, modify, and distribute this software and its 
   documentation for any purpose, provided that the above copyright notice 
   and the following three paragraphs appear in all copies of this 
   software. 
     
     

     IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR 
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF 
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN 
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  M.I.T. LINCOLN 
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT. 

        THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN 
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, 
ENHANCEMENTS, OR MODIFICATIONS. 
*/


/* Include necessary header files */
/* config.h must be first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>

/* to include internal declerations and definitions from `bintree_pack.h' */
#define   svm_pack_h_internal_types
#define   svm_pack_h_statics
/*
#define   svm_pack_h
*/
#include "svm_pack.h"

#define PEQ(y) (0 == strcmp(param_name,y))
#define ABS(x) (((x) < 0)? -(x) : (x))
#define TRUE 1
#define FALSE 0

/* Static Variables for component classifiers (for this file use only) */
static SVM_PARAMS **svmpp;  /* svm_model set */

/* Static Variables for top-level multi-component classifier */
static SVM_MULTI **svmpp_multi; /* svm_model set */


/* edit this locally to get more debug printout */
static int  verbose_flag=0;


/**********************************************************************
 **********************************************************************
 *****           DEFINITIONS OF EXTERNAL FUNCTIONS                *****
 **********************************************************************
 **********************************************************************/

/************************************************************************
* NAME:svm_build(ninputs, nclasses, npatterns)
* DESCRIPTION:
*   Allocates memory for the first node (root) of the tree.
*       Initializes the elements and reads in the data file.
*************************************************************************/
int
svm_build(int ninputs, int noutputs, int npatterns)
{
    int new_model;
    static int max_models=0;
    SVM_PARAMS *svm; /* temporary pointer to svmpp[] */

    /* find entry in model array for new model */
    new_model = model_get_number((void ***)&svmpp,
                 &max_models, sizeof(SVM_PARAMS));
    svm = svmpp[new_model];

    /* check for bad parameters */
    if (ninputs <= 0)
      fatal_error("svm_build: ninputs not  positive  negative = %d", ninputs);
    if (noutputs <= 0)
      fatal_error("svm_build: noutputs not positive = %d", noutputs);
    if (npatterns <= 0)
      fatal_error("svm_build: npatterns not positive = %d", npatterns);


    /* constants */
    svm->ninputs = ninputs;
    svm->noutputs = noutputs;
    svm->npatterns = npatterns;

    /* store all training data internally */
    svm->exemplars = allocate_matrix(npatterns, ninputs);
    svm->y = allocate_ivector(npatterns);

    /* standard vectors */
    svm->outputs = allocate_vector(noutputs);
    svm->class_cnt = allocate_vector(noutputs);

    /* vectors kept for each training pattern*/
    svm->alpha_type = nn_malloc(npatterns,"for storing sv type");
    svm->fcache_value = allocate_vector(npatterns);
    svm->fcache_flag = allocate_ivector(npatterns);
    svm->pattern2sv = allocate_ivector(npatterns);
    svm->alpha = allocate_vector(npatterns);

    /* vectors as long as the number of inputs */
    svm->i_low_vector= allocate_vector(ninputs);
    svm->i_high_vector = allocate_vector(ninputs);
    svm->v1_current = allocate_vector(ninputs);
    svm->v2_current = allocate_vector(ninputs);
    svm->weight = allocate_vector(ninputs);

    /* vectors as long as the number of support vectors.
       To simplify logic, these are all set to the number of
       inputs except for the support vector storage. 
       Storage for support vectors is initially small, but
       then increased to the correct size after training when
       it is increased to the correct size and 
       support vectors are copied in.
       The memory required could be reduced somewhat
       when the number of inputs is small and there are
       many patterns by allocating this as needed.
       This will only matter when the input vector is
       short because it only takes as much space as roughly
       three more vector components.
    ---------------------------------------------------*/
    svm->sv_alpha = allocate_vector(npatterns);
    svm->sv_y = allocate_ivector(npatterns);
    svm->sv_alpha_type = nn_malloc(npatterns, "for storing sv alpha type");
    svm->sv_unique = allocate_ivector(npatterns);
    svm->sv2pattern = allocate_ivector(npatterns);
    svm->support_vectors = allocate_matrix(1, ninputs);

    return(new_model);
}
/************************************************************************
*   Allocates memory for a multi-class model that keeps track
*       of many component models
*************************************************************************/
int
svm_build_multi(int ninputs, int noutputs, int type, int n_classifiers_to_build, 
                int n_classifiers_that_will_be_trained)
{
    int new_model;
    static int max_models=0;
    SVM_MULTI *svm_multi; /* temporary pointer to svmpp_multi[] */

    /* check for bad parameters */
    if (ninputs <= 0)
      fatal_error("svm_build_multi: ninputs is negative = %d", ninputs);
    if (noutputs <= 0)
      fatal_error("svm_build_multi: noutputs is negative = %d", noutputs);
    if (n_classifiers_to_build <= 0)
      fatal_error("svm_build_multi: build classifiers is negative = %d", 
                   n_classifiers_to_build);
    if (n_classifiers_that_will_be_trained <= 0)
      fatal_error("svm_build_multi: train classifiers is negative = %d", 
                   n_classifiers_that_will_be_trained);
    if (type < 0)
      fatal_error("svm_build_multi: type is negative = %d", type);


    /* find entry in model array for new model */
    new_model = model_get_number((void ***)&svmpp_multi,
                 &max_models, sizeof(SVM_MULTI));
    svm_multi = svmpp_multi[new_model];

    /* fill in paramters */
    svm_multi->ninputs = ninputs;
    svm_multi->noutputs = noutputs;
    svm_multi->type = type;
    svm_multi->n = n_classifiers_that_will_be_trained;
    svm_multi->sv_n = 0;
    svm_multi->n_positive_below_bound=0;
    svm_multi->n_at_cbound=0;
    svm_multi->sv_unique=0;


    /* standard vectors */
    svm_multi->outputs = allocate_vector(noutputs);
    svm_multi->component_model = allocate_ivector(n_classifiers_to_build);
    svm_multi->component_outputs = allocate_matrix(n_classifiers_to_build,2);
    svm_multi->component_decisions = allocate_ivector(n_classifiers_to_build);
    svm_multi->votes = allocate_ivector(noutputs);
    svm_multi->class_min = allocate_vector(noutputs);
    svm_multi->priors_list = allocate_vector(noutputs);
    svm_multi->class_cnt = allocate_vector(noutputs);


    return(new_model);
}
/************************************************************************
*   Set model number in high-level multi classifier for a component classifier 
*************************************************************************/
void
svm_set_component_model_multi(int multi_model, int classifier_model, int i)
{
    SVM_MULTI *svm_multi; /* temporary pointer for model */

    svm_multi = svmpp_multi[multi_model];

    if ( i < 0 || i >= svm_multi->n)
    {
    fatal_error("set_component_model_multi illegal index %d not >=0 and <=%d\n",
                     svm_multi->n);
    }

    svm_multi->component_model[i] = classifier_model;
}
/************************************************************************
*   Set class count in a high-level multi classifier 
*************************************************************************/
void
svm_set_class_cnt_multi(int multi_model, float count, int i)
{
    SVM_MULTI *svm_multi; /* temporary pointer for model */

    svm_multi = svmpp_multi[multi_model];

    if ( i < 0 || i >= svm_multi->noutputs)
    {
    fatal_error("set_class_cnt_multi illegal index %d not >=0 and <=%d\n",
                     svm_multi->noutputs);
    }

    svm_multi->class_cnt[i] = count;
}

#define EPSILON_ROUNDING 1e-06
/************************************************************************
* DESCRIPTION:
*   Initializes smo algorithm, before training
************************************************************************/
void svm_init(int model, float cbound, int kernel_type, float kernel_sigma,
              int kernel_order, float kernel_poly_divisor,
              int sigmoid_fit,
              float kkt_tolerance, float alpha_tolerance, 
              int verbosity,
              int multi_class_type, int multi_class_a, int multi_class_b,
              int multi_class_i)
    {
    int i,j, ntrain, ninputs;
    SVM_PARAMS *svm; /* temporary pointer for model */

    svm = svmpp[model];
    ntrain = svm->npatterns;
    ninputs = svm->ninputs;

    /* copy parameters and set others to default values */
    svm->cbound = cbound;
    svm->kernel_type = kernel_type;
    svm->gaussian_sigma = kernel_sigma;
    svm->polynomial_order = kernel_order;
    svm->polynomial_divisor = kernel_poly_divisor;
    if ( kernel_poly_divisor > 0.0)
        svm->inv_poly_divisor = 1.0/kernel_poly_divisor;
    svm->sigmoid_fit = sigmoid_fit;
    svm->posterior_aexp = .001;
    svm->verbose_flag = 0;
    svm->warning_flag = 0;
    svm->kkt_tolerance = kkt_tolerance;
    svm->alpha_tolerance = alpha_tolerance;
    svm->final_check_tolerance = 4.0*kkt_tolerance;
    svm->epsilon_rounding = EPSILON_ROUNDING;
    svm->verbosity = verbosity;
    svm->multi_class_type = multi_class_type;
    svm->multi_class_a = multi_class_a;
    svm->multi_class_b = multi_class_b;
    svm->multi_class_i = multi_class_i;

    /* initialize support vector counts and flags */
    svm->sv_n = 0;
    svm->n_positive_below_bound=0;
    svm->n_zero=ntrain;     
    svm->n_at_cbound = 0;
    svm->threshold = 0.0;
    svm->support_vectors_valid = 0;

    /* zero patterns per class */
    for(i=0; i< svm->noutputs; i++)
    {
        svm->class_cnt[i]=0;
    }

    /* initialize alpha and index for all patterns */
    for( i=0; i<ntrain; i++)
    {
    svm->alpha_type[i] = ZERO;
    svm->pattern2sv[i] = NOT_A_SUPPORT_VECTOR;
    svm->alpha[i]=0.0;
    }

    /* initialize support vector information */
    for( i=0; i<ntrain; i++)
    {
    svm->sv_alpha[i] = 0.0;
    svm->sv_alpha_type[i] = ZERO;
    svm->sv_unique[i] = ZERO;
    svm->sv2pattern[i] = NOT_A_SUPPORT_VECTOR;
    }

    /* cache starts empty */
    svm_init_fcache(svm, ntrain);

    /* initialize weights to zero */
    if ( kernel_type == SMO_LINEAR)
        {
        for(j=0; j<ninputs; j++)
            svm->weight[j] = 0.0;
        }
    else if ( kernel_type == SMO_GAUSSIAN)
        {
        svm->gaussian_variance_times_two = 
                          (svm->gaussian_sigma)*(svm->gaussian_sigma)*2.0;
        }
    else if ( kernel_type == SMO_POLYNOMIAL)
        {
        svm->inv_poly_divisor = 1.0/(svm->polynomial_divisor);
        if ( svm->polynomial_order <= 1 )
            {
            warning("svm: polynomial of order <= 1, set to LINEAR kernel\n");
            svm->kernel_type = SMO_LINEAR;
            }
        }
    else if ( kernel_type == SMO_INHOMOGENEOUS_POLYNOMIAL)
        {
        svm->inv_poly_divisor = 1.0/(svm->polynomial_divisor);
        if ( svm->polynomial_order <= 1 )
            {
        warning("svm: inhomogeneous polynomial of order <= 1, set to LINEAR kernel\n");
            svm->kernel_type = SMO_LINEAR;
            }
        }
    else if ( kernel_type == SMO_LINEAR_USE_SVS)
        {
        /* nothing here */
        }
    else
        {
        fatal_error("svm: invalid kernel type %d\n", kernel_type);
        }

    }
/************************************************************************
* NAME: svm_free(model)
* DESCRIPTION:
*   Deallocates all memory used for a svm classifier
************************************************************************/
void
svm_free(int model)
{
    SVM_PARAMS *svm; /* temporary pointer for model */

    svm = svmpp[model];

    /* ----------------------------------------------
       used only during training, prevents too much
       deallocation after svm_free_train is called 
    -------------------------------------------------*/
    if ( svm->exemplars != NULL)
    {
    /* raw training data */
    deallocate_matrix (svm->exemplars);
    deallocate_ivector(svm->y);

    /* vectors kept for each training pattern*/
    nn_free(svm->alpha_type, "for storing pattern sv type");
    deallocate_vector(svm->fcache_value);
    deallocate_ivector(svm->fcache_flag);
    deallocate_ivector(svm->pattern2sv);
    deallocate_vector(svm->alpha);

        /* vectors as long as the number of inputs */
        deallocate_vector(svm->i_low_vector);
        deallocate_vector(svm->i_high_vector);
        deallocate_vector(svm->v1_current);
        deallocate_vector(svm->v2_current);

        /* vectors as long as the number of support vectors */
        deallocate_ivector(svm->sv2pattern);
    }

    /* -----------------------------
       used during training and testing 
    -------------------------------*/

    deallocate_vector(svm->outputs);
    deallocate_vector(svm->class_cnt);


    /* vectors as long as the number of inputs */
    deallocate_vector(svm->weight);

    /* vectors as long as the number of support vectors.
    ---------------------------------------------------*/
    deallocate_vector(svm->sv_alpha);
    deallocate_ivector(svm->sv_y);
    deallocate_ivector(svm->sv_unique);
    nn_free(svm->sv_alpha_type, "for storing sv alpha type");
    deallocate_matrix(svm->support_vectors);

    /* free main model pointer */
    nn_free((char*) svmpp[model], "svm model");
    svmpp[model] = NULL;
    
}
/************************************************************************
*   get pointer to outputs for multi model 
************************************************************************/
float *svm_get_outputs_multi(int model)
{
    SVM_MULTI *svm_multi; /* temporary pointer for model */

    svm_multi = svmpp_multi[model];
    return (&(svm_multi->outputs[0]));
}
/************************************************************************
*   Deallocates all memory used for testing high-level classifier
************************************************************************/
void
svm_free_multi(int model)
{
    SVM_MULTI *svm_multi; /* temporary pointer for model */
    SVM_PARAMS *svm;
    int i,j;

    svm_multi = svmpp_multi[model];

    /* free memory for component classifiers */
    for(i=0; i<(svm_multi->n); i++)
    {
    j = svm_multi->component_model[i];
    svm = svmpp[j];
        svm->exemplars = NULL;
    svm_free(j);
    }

    /* free memory for main high-level classifier */
    svm_free_multi_train(model);
}

/************************************************************************
*   Deallocates all memory used for high-level classifier
************************************************************************/
void
svm_free_multi_train(int model)
{
    SVM_MULTI *svm_multi; /* temporary pointer for model */
    svm_multi = svmpp_multi[model];

    deallocate_ivector(svm_multi->component_model);
    deallocate_matrix(svm_multi->component_outputs);
    deallocate_ivector(svm_multi->component_decisions);
    deallocate_ivector(svm_multi->votes);
    deallocate_vector(svm_multi->class_min);
    deallocate_vector(svm_multi->priors_list);
    deallocate_vector(svm_multi->outputs);
    deallocate_vector(svm_multi->class_cnt);

    /* free main model pointer */
    nn_free((char*) svmpp_multi[model], "svm multi model");
    svmpp_multi[model] = NULL;
}
/************************************************************************
*   Deallocates memory used during training a svm classifier
************************************************************************/
void svm_free_train(int model)
{

    SVM_PARAMS *svm;

    svm = svmpp[model];
    /* normal vectors and matrices */
    deallocate_matrix (svm->exemplars);
    svm->exemplars = NULL;
    deallocate_ivector(svm->y);

    /* vectors kept for each training pattern*/
    nn_free(svm->alpha_type, "for storing pattern sv type");
    deallocate_vector(svm->fcache_value);
    deallocate_ivector(svm->fcache_flag);
    deallocate_ivector(svm->pattern2sv);
    deallocate_vector(svm->alpha);

    /* vectors as long as the number of inputs */
    deallocate_vector(svm->i_low_vector);
    deallocate_vector(svm->i_high_vector);
    deallocate_vector(svm->v1_current);
    deallocate_vector(svm->v2_current);


    deallocate_ivector(svm->sv2pattern);
}

/* Parameter Tables */
static PARAM paramtab[] = {
    {"ninputs", P_INT},
    {"noutputs", P_INT},
    {"npatterns", P_INT},
    {"cbound", P_FLOAT},
    {"kernel_type", P_INT},
    {"kernel_sigma", P_FLOAT},
    {"kernel_order", P_INT},
    {"kpoly_divisor", P_FLOAT},
    {"sigmoid_fit", P_INT},
    {"posterior_aexp", P_FLOAT},
    {"kkt_tolerance", P_FLOAT},
    {"sv_n", P_INT},
    {"threshold", P_FLOAT},
    {"sv_valid", P_INT},
    {"mclass_type", P_INT},
    {"mclass_a", P_INT},
    {"mclass_b", P_INT},
    {"mclass_i", P_INT},
};
#define TABSIZE (sizeof(paramtab)/sizeof(PARAM))

static PARAM ptab_linear_arrays[] = {
    {"class_cnt", P_FLOAT_VECTOR, NULL, 0, 0},
    {"weight", P_FLOAT_VECTOR, NULL, 0, 0},
};
#define TABSIZE_LINEAR_ARRAYS (sizeof(ptab_linear_arrays)/sizeof(PARAM))

static PARAM ptab_nonlinear_arrays[] = {
    {"class_cnt", P_FLOAT_VECTOR, NULL, 0, 0},
    {"sv_alpha", P_FLOAT_VECTOR, NULL, 0, 0},
    {"sv_y", P_INT_VECTOR, NULL, 0, 0},
    {"sv_unique", P_INT_VECTOR, NULL, 0, 0},
    {"support_vectors", P_FLOAT_MATRIX, NULL, 0, 0},
};
#define TABSIZE_NONLINEAR_ARRAYS (sizeof(ptab_nonlinear_arrays)/sizeof(PARAM))

/* for multi class data  */
static PARAM multi_paramtab[] = {
    {"ninputs", P_INT},
    {"noutputs", P_INT},
    {"type", P_INT},
    {"n", P_INT},
    {"sv_n", P_INT},
    {"n_pbb", P_INT},
    {"n_atb", P_INT},
    {"sv_unique", P_INT},
    {"npatterns", P_FLOAT},
};
#define TABSIZE_MULTI (sizeof(multi_paramtab)/sizeof(PARAM))


/******************************
 * routines to set param variables for reading and writing a parameter file
 *****************************/
static void
svm_set_paramtab(SVM_PARAMS *svm)
{
    param_setvarptr("ninputs", &svm->ninputs, paramtab, TABSIZE);
    param_setvarptr("noutputs", &svm->noutputs, paramtab, TABSIZE);
    param_setvarptr("npatterns", &svm->npatterns,  paramtab, TABSIZE);
    param_setvarptr("cbound", (int *)&svm->cbound,  paramtab, TABSIZE);
    param_setvarptr("kernel_type", &svm->kernel_type,  paramtab, TABSIZE);
    param_setvarptr("kernel_sigma", (int *)&svm->gaussian_sigma, paramtab, TABSIZE);
    param_setvarptr("kernel_order", &svm->polynomial_order, paramtab, TABSIZE);
    param_setvarptr("kpoly_divisor", (int *)&svm->polynomial_divisor, 
                                                                    paramtab, TABSIZE);
    param_setvarptr("sigmoid_fit", 
                                  &svm->sigmoid_fit, paramtab, TABSIZE);
    param_setvarptr("posterior_aexp", (int *)&svm->posterior_aexp, paramtab, TABSIZE);
    param_setvarptr("kkt_tolerance", (int *)&svm->kkt_tolerance, paramtab, TABSIZE);
    param_setvarptr("sv_n", &svm->sv_n, paramtab, TABSIZE);
    param_setvarptr("threshold", (int *)&svm->threshold, paramtab, TABSIZE);
    param_setvarptr("sv_valid", &svm->support_vectors_valid, paramtab, TABSIZE);
    param_setvarptr("mclass_type", &svm->multi_class_type, paramtab, TABSIZE);
    param_setvarptr("mclass_a", &svm->multi_class_a, paramtab, TABSIZE);
    param_setvarptr("mclass_b", &svm->multi_class_b, paramtab, TABSIZE);
    param_setvarptr("mclass_i", &svm->multi_class_i, paramtab, TABSIZE);

}
static void
svm_set_ptab_linear_arrays(SVM_PARAMS *svm)
{
    param_setvarptr("class_cnt", (int *)&(svm->class_cnt[0]),
            ptab_linear_arrays, TABSIZE_LINEAR_ARRAYS);
    param_setdim("class_cnt", svm->noutputs, 0,
                    ptab_linear_arrays, TABSIZE_LINEAR_ARRAYS);
    param_setvarptr("weight", (int *)&(svm->weight[0]),
            ptab_linear_arrays, TABSIZE_LINEAR_ARRAYS);
    param_setdim("weight", svm->ninputs, 0, 
                    ptab_linear_arrays, TABSIZE_LINEAR_ARRAYS);
}

static void
svm_set_ptab_nonlinear_arrays(SVM_PARAMS *svm)
{
    param_setvarptr("class_cnt", (int *)&(svm->class_cnt[0]),
            ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setdim("class_cnt", svm->noutputs,0,
                    ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setvarptr("sv_alpha", (int *)&(svm->sv_alpha[0]),
            ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setdim("sv_alpha", svm->sv_n, 0, 
                    ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setvarptr("sv_y", (int *)&(svm->sv_y[0]),
            ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setdim("sv_y", svm->sv_n, 0, 
                    ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setvarptr("sv_unique", (int *)&(svm->sv_unique[0]),
            ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setdim("sv_unique", svm->sv_n, 0, 
                    ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setvarptr("support_vectors", (int *)&(svm->support_vectors[0][0]),
            ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
    param_setdim("support_vectors", svm->sv_n, svm->ninputs, 
                    ptab_nonlinear_arrays, TABSIZE_NONLINEAR_ARRAYS);
}
static void
svm_set_multi_paramtab(SVM_MULTI *svm)
{
    param_setvarptr("ninputs", &svm->ninputs, 
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("noutputs",&svm->noutputs, 
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("type",    &svm->type,  
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("n",       &svm->n,  
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("sv_n",       &svm->sv_n,  
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("n_pbb",       &svm->n_positive_below_bound,  
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("n_atb",       &svm->n_at_cbound,  
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("sv_unique",       &svm->sv_unique,  
                               multi_paramtab, TABSIZE_MULTI);
    param_setvarptr("npatterns", (int *)&svm->npatterns,
            multi_paramtab, TABSIZE_MULTI);
}


/*****************************************************************************
* NAME: svm_write_model()
* DESCRIPTION:
*       writes parameters and support vectors or weights to a file
*       in ASCII format
*
******************************************************************************/
int
svm_write_model(int model, FILE *fp)
{
    int return_value;
    SVM_PARAMS *svm;

    svm = svmpp[model];

    /* save scalars */
    svm_set_paramtab(svm);
    if ((return_value = param_putblk(fp, paramtab, TABSIZE)) < 0)
    return return_value;

    /* save vectors and matrices */
    /* linear support vector machine, only need weights */
    if ( svm->kernel_type == SMO_LINEAR)
    {
    svm_set_ptab_linear_arrays(svm);
    if ((return_value = param_putblk(fp, ptab_linear_arrays, 
                                             TABSIZE_LINEAR_ARRAYS)) < 0)
        return return_value;
    }
     /* nonlinear support vector machine, need alpha and support vectors */
     else
    {
    svm_set_ptab_nonlinear_arrays(svm);
    if ((return_value = param_putblk(fp, ptab_nonlinear_arrays, 
                                             TABSIZE_NONLINEAR_ARRAYS)) < 0)
        return return_value;

    }

    return (0);
}
/*****************************************************************************
*       writes multi-classifier parameters to a file in ASCII format
******************************************************************************/
int
svm_write_model_multi(int model, FILE *fp)
{
    int return_value;
    SVM_MULTI *svm_multi;

    svm_multi = svmpp_multi[model];


    /* save scalars */
    svm_set_multi_paramtab(svm_multi);
    if ((return_value = param_putblk(fp, multi_paramtab, TABSIZE_MULTI)) < 0)
    return (return_value);

    return(0);

}
/****************************************************************************
*       Loads parameters for multi-classifier 
****************************************************************************/
int 
svm_read_model_multi(FILE *fp, int verbosity)
{
    int model, i, j, k, total_unique,return_value;
    int ia, ib, n;
    SVM_MULTI *temp, *svm_multi;
    SVM_PARAMS *svm;

    temp = (SVM_MULTI *)nn_malloc(sizeof(SVM_MULTI), "for temp multi stcture");

    /* read in scalars for multiple models */
    svm_set_multi_paramtab(temp);
    if ((return_value = param_getblk(fp, multi_paramtab, TABSIZE_MULTI)) < 0)
    return return_value;
    model = svm_build_multi(temp->ninputs, temp->noutputs, 
                            temp->type, temp->n, temp->n);
    svm_multi = svmpp_multi[model];

    /* read in all the rest of the models */
    for(i=0; i<(svm_multi->n); i++)
    {
        (svm_multi->component_model[i]) = svm_read_model(fp);
    }

    /* fill in multi class_counts from component classifiers */
    if ( svm_multi->type == 0)
    {
     for(i=0; i<(svm_multi->n); i++)
        {
        j = svm_multi->component_model[i];
        svm = svmpp[j];
        ia = svm->multi_class_a;
        n = svm->class_cnt[0];
        svm_multi->class_cnt[ia] = n;
        }
    }
     else
    {
     for(i=0; i<(svm_multi->n); i++)
        {
        j = svm_multi->component_model[i];
        svm = svmpp[j];
        ia = svm->multi_class_a;
        n = svm->class_cnt[0];
        svm_multi->class_cnt[ia] = n;
        ib = svm->multi_class_b;
        n = svm->class_cnt[1];
        svm_multi->class_cnt[ib] = n;
        }
    }
    svm_read_model_multi_end(model,fp );

    if ( verbosity >= 1)
    printf("classifiers %d  unique support vectors %d  total %d  UnBound %d  AtBound %d\n",
                svm_multi->n, svm_multi->sv_unique, svm_multi->sv_n,
                svm_multi->n_positive_below_bound, svm_multi->n_at_cbound);

    j = svm_multi->component_model[0];
    svm = svmpp[j];
    /* verify unique count */
    if ( svm->kernel_type != SMO_LINEAR)
        {
        total_unique = 0;
        for(i=0; i<(svm_multi->n); i++)
        {
        j = svm_multi->component_model[i];
        svm = svmpp[j];
            for(k=0; k<svm->sv_n; k++)
        {
                total_unique += svm->sv_unique[k];
        }
        }
        if ( total_unique != svm_multi->sv_unique)
        fatal_error("Number of unique sv's in component classifiers inconsistent with multi model.\nParameter file may be corrupted.");
    }

           
    nn_free((char *)temp, "for temp multi stcture");
    return(model);
}
/****************************************************************************
*       reset sv count parameters for multi-classifier stored at end of file 
****************************************************************************/
void 
svm_read_model_multi_end(int model, FILE *fp)
{
    SVM_MULTI *svm_multi;

    svm_multi = svmpp_multi[model];

    /* read in scalars for multiple models */
    svm_set_multi_paramtab(svm_multi);
    if (param_getblk(fp, multi_paramtab, TABSIZE_MULTI) < 0)
        {
    fatal_error("couldnt read multi-classifier data at end of param file");
    }
}
/****************************************************************************
* NAME: svm_read_model(fp)
* DESCRIPTION:
*       Loads parameters and weights (if linear) or
*       support vectors and values (if nonlinear)
*       Allocates memory as in svm_build, so copies much of svm_build 
*       Only allocates enough memory to perform classification
****************************************************************************/
int
svm_read_model(FILE *fp)
{
    int new_model, return_value;
    static int max_models=0;
    SVM_PARAMS *svm; /* temporary pointer to svmpp[] */

    /* find entry in model array for new model */
    new_model = model_get_number((void ***)&svmpp,
                 &max_models, sizeof(SVM_PARAMS));
    svm = svmpp[new_model];

    /* read in scalars */
    svm_set_paramtab(svm);
    if ((return_value = param_getblk(fp, paramtab, TABSIZE)) < 0)
    return return_value;

    /* initialize scalars required for kernel evaluation */
    if ( svm->kernel_type == SMO_LINEAR)
        {
        /* nothing here */
        }
    else if ( svm->kernel_type == SMO_GAUSSIAN)
        {
        svm->gaussian_variance_times_two = 
                          (svm->gaussian_sigma)*(svm->gaussian_sigma)*2.0;
        }
    else if ( svm->kernel_type == SMO_POLYNOMIAL)
        {
        svm->inv_poly_divisor = 1.0/(svm->polynomial_divisor);
        if ( svm->polynomial_order <= 1 )
            {
            warning("svm: polynomial of order <= 1, set to LINEAR kernel\n");
            svm->kernel_type = SMO_LINEAR;
            }
        }
    else if ( svm->kernel_type == SMO_INHOMOGENEOUS_POLYNOMIAL)
        {
        svm->inv_poly_divisor = 1.0/(svm->polynomial_divisor);
        if ( svm->polynomial_order <= 1 )
            {
        warning("svm: inhomogeneous polynomial of order <= 1, set to LINEAR kernel\n");
            svm->kernel_type = SMO_LINEAR;
            }
        }
    else if ( svm->kernel_type == SMO_LINEAR_USE_SVS)
        {
        ;
        /* nothing here */
        }
    else
        {
        fatal_error("invalid kernel type in parameter file %d\n", svm->kernel_type);
        }


    /* allocate space required by eval routines */
    svm->outputs = allocate_vector(svm->noutputs);

    /* read in vectors and matrices */
    /* linear support vector machine, only need weights */
    if ( svm->kernel_type == SMO_LINEAR)
    {
    svm->class_cnt = allocate_vector(svm->noutputs);
        svm->weight = allocate_vector(svm->ninputs);
    svm_set_ptab_linear_arrays(svm);
    if ((return_value = param_getblk(fp, ptab_linear_arrays, 
                                             TABSIZE_LINEAR_ARRAYS)) < 0)
        return return_value;
    /* set nonlinear vectors to NULL */
    svm->sv_alpha = NULL;
    svm->sv_alpha_type = NULL;
    svm->sv_y = NULL;
    svm->sv_unique = NULL;
    svm->support_vectors = NULL;
    }
     /* nonlinear support vector machine, need alpha and support vectors */
     else
    {
    svm->class_cnt = allocate_vector(svm->noutputs);
        svm->sv_alpha = allocate_vector(svm->sv_n);
    svm->sv_alpha_type = NULL;
        svm->sv_y = allocate_ivector(svm->sv_n);
        svm->sv_unique = allocate_ivector(svm->sv_n);
        svm->support_vectors = allocate_matrix(svm->sv_n, svm->ninputs);
    svm->weight = NULL;
    svm_set_ptab_nonlinear_arrays(svm);
    if ((return_value = param_getblk(fp, ptab_nonlinear_arrays, 
                                             TABSIZE_NONLINEAR_ARRAYS)) < 0)
        return return_value;

    }
    return new_model;
}


/*****************************************************************************
* NAME: svm_store_inputs 
* DESCRIPTION:
*       Copy exemplars into local array in structure for fast access.
*       This means you don't want to train two or more at a time but
*       do it successively. It also means that twice the memory required to
*       store the data is required to train. This could be fixed by not using
*       the fancy data handling routines.
******************************************************************************/
void
svm_store_inputs(int model, float *inputs, int desired_class, int i)
{
    SVM_PARAMS *svm;
    int j;
    svm = svmpp[model];
    
    if ( i >= svm->npatterns || i < 0)
    {
    fatal_error("svm_store_inputs: pattern index bad = %d, allowed = 0 to %d\n",
                    i, (svm->npatterns) - 1);
    }
    else
    {
    for(j=0; j< svm->ninputs; j++)
        {
        svm->exemplars[i][j]=inputs[j];
        }
    /* svm is a binary classifier, make sure classes are only 0 and 1 */
    if ( desired_class == 0)
        {
        svm->y[i]= 1;
        svm->class_cnt[0] +=1;
        }
    else if ( desired_class == 1)
        {
        svm->y[i]= -1;
        svm->class_cnt[1] +=1;
        }
    else    
        {
        warning("Class of pattern %d = %d (Should be 0 or 1) Coerced into class 1",
                         i, desired_class);
        svm->y[i] = -1;
        svm->class_cnt[1] +=1;
        }
    }
}
/*****************************************************************************
* NAME: svm_run_training 
* DESCRIPTION:
*       perform complete training run 
******************************************************************************/
void svm_run_training(int model, int verbosity, char *run_name, char **labels)
    {
    SVM_PARAMS *svm;
    int i, j, examine_all_flag;
    int inner_loop_success;
    int n_changed, nsweeps;
    int y1, y2;
    int ntrain, ninputs;
    int bias_different_flag;

    svm = svmpp[model];
    ntrain = svm->npatterns;
    ninputs = svm->ninputs;

    /* initialize counters */
    examine_all_flag = 1;
    nsweeps = 0;
    svm->total_pair_corrections=0;
    svm->nsweeps=0;
    svm->total_kernel_evaluations=0;

#ifdef print_exemplars_to_output
    /* debug print exemplars to output */
    for(i=0; i< ntrain; i++)
        {
        if ( svm->y[i] > 0)
            printf("0 ");
        else
            printf("1 ");
        for(j=0; j<ninputs; j++)
            printf("%6.4f ", svm->exemplars[i][j]);
        printf("\n");
        }
#endif

    if ( verbosity >= 1)
         {
         svm_print_model_parameters(model);
         }

    /* get one example from class A (+1) and B (-1) to
       set initial values of b_low and b_high 
    -------------------------------------------*/
    svm_find_ab_patterns(svm, ntrain);

    svm_print_summary_header(verbosity);
    /* loop over data 
    --------------------*/
    while ( nsweeps < MAX_SWEEPS )
    {
    n_changed = 0;
    /* loop over all examples */
    if ( examine_all_flag > 0 )
        {
        for(i=0; i<ntrain; i++)
            {
            n_changed += svm_examine_example(svm, i, svm->y[i], svm->exemplars[i]);
            }
        nsweeps++;
        svm_print_summary_values(svm, nsweeps, n_changed, 0, verbosity);
#ifdef april_debug
        /* april debug */
        if ( nsweeps >= 7)
            {
            verbose_flag = 5;
            }
#endif
        }
    /* loop over unbound examples */
    else
        {
        inner_loop_success=1;
        while (1)
            {
            y1 = svm->i_high_desired_output;
            y2 = svm->i_low_desired_output;
            for(j=0; j<ninputs;j++)
                {
                svm->v1_current[j] = svm->i_high_vector[j];
                svm->v2_current[j] = svm->i_low_vector[j];
                    }
            inner_loop_success = svm_take_step(svm, svm->i_high, svm->i_low, 
                                                    y1, y2, svm->v1_current, svm->v2_current);
            if ( verbose_flag >= 1)
                printf("inner loop success %d ihigh %d ilow %d yhigh %d ylow %d\n",
                                                inner_loop_success, svm->i_high, svm->i_low, 
                                                y1, y2);
            n_changed += inner_loop_success;
            /* stop if thresholds close or no success */
            if ( (svm->b_high > svm->b_low - (2.0*svm->kkt_tolerance)) || 
                             (inner_loop_success == 0)  )
                    break;
            }
        svm_print_summary_values(svm, nsweeps, n_changed, 0, verbosity);
        n_changed = 0;
        }

    /* stop if looped over all examples and didn't change any */
    if ( examine_all_flag > 0 && n_changed == 0 )
        break;
    /* after looping over all examples, focus on unbound examples 
       after looping over unbound examples with no change, loop over all examples*/
    if ( examine_all_flag >  0)
        examine_all_flag = 0;
    else if ( n_changed == 0 )
        examine_all_flag = 1;
    }
    /* see if upper and low bias are close enough */
    if ( svm->b_high <= (svm->b_low) -2.0*(svm->kkt_tolerance) )
    {
        bias_different_flag = 1;
        if ( verbose_flag >= 2)
        {
        warning("%s\n%s\n%s",
            "Bad solution, high and low bias are different after convergence",
            "Try decreasing  cbound, or a different kernel type",
            "Also try increasing kkt_tolerance, kernel_poly_divisor, or alpha_tolerance");
        }
    }
       else
        bias_different_flag=0;

    svm->threshold = (svm->b_low + svm->b_high)/2.0;
    svm->nsweeps = nsweeps;
    if ( verbose_flag >= 1)
        svm_print_summary_values(svm, nsweeps, n_changed, 0, 4);


    if ( verbose_flag >= 1)
        printf("Total support vectors %d  UnBound %d  AtUpper %d\n", 
                svm->sv_n, svm->n_positive_below_bound, svm->n_at_cbound);

    /* always find sigmoid to approximate posterior probability */
    svm_fit_posterior_sigmoid(svm);
    svm_check_final_kkt_conditions(svm, verbosity, run_name, labels, bias_different_flag);
    svm_save_final_support_vectors(svm);
}
/*****************************************************************************
*       find one pattern from class A and one from B for initialization
******************************************************************************/
static void svm_find_ab_patterns(SVM_PARAMS *svm, int ntrain)
    {
    int i, found_high, found_low;
    int desired_output;
    float *pinput;


    /* get one example from class A (+1) and B (-1) to
       set initial values of b_low and b_high 
    -------------------------------------------*/
    found_high = 0;
    found_low = 0;
    for(i=0; i<ntrain; i++)
     {
         desired_output = svm->y[i];
     pinput = svm->exemplars[i];

    /* Class A (+1) */
    if ( found_high == 0 && desired_output == 1 )
        {
        found_high =1;
        svm->i_high = i; 
        svm->b_high = -1.0;
        svm_high_update(svm, -1.0, i, 1, pinput);
        svm_set_fcache( svm, i, -1.0);
        }
    /* Class B (-1) */
    else if ( found_low == 0 && desired_output == -1 )
        {
        found_low = 1;
        svm->i_low = i;
        svm->b_low = 1.0;
        svm_low_update(svm, 1.0, i, -1, pinput);
        svm_set_fcache( svm, i, 1.0);
        }
    }
      
    if ( found_low == 0 || found_high == 0)
    {
    fatal_error("Trying to train with only patterns from one class\n");
    }

    }
/*****************************************************************************
*       print parameters for this run 
******************************************************************************/
static void svm_print_model_parameters(int model)
    {
        SVM_PARAMS *svm;
        svm = svmpp[model];

    printf("\n---------------------------------------------------------------------------------\n");
    printf("svm params ntrain %d  ninputs %d  c %.2f", svm->npatterns, svm->ninputs, svm->cbound);
    /* print kernel type */
        if ( svm->kernel_type == SMO_LINEAR)
        {
        printf("  linear ");
        }
    else if ( svm->kernel_type == SMO_GAUSSIAN)
        {
        printf("  gaussian sigma %.2f ", svm->gaussian_sigma);
        }
    else if ( svm->kernel_type == SMO_POLYNOMIAL)
        {
        printf("  polynomial  order %d divisor %f", 
                          svm->polynomial_order, svm->polynomial_divisor);
        }
    else if ( svm->kernel_type == SMO_INHOMOGENEOUS_POLYNOMIAL)
        {
        printf("  inhomogeneous polynomial  order %d divisor %f", 
                          svm->polynomial_order, svm->polynomial_divisor);
        }
        else if ( svm->kernel_type == SMO_LINEAR_USE_SVS)
        {
        printf("  linear using sv's ");
        }
    else
        {
        fatal_error("invalid kernel type %d\n", svm->kernel_type);
        }

    printf("\n---------------------------------------------------------------------------------\n");
    }



/************************************************************************
* NAME: fcache routines 
* DESCRIPTION:
*  These routines keep cached values of f which is the actual minus desired
*  outputs without using the bias term.
*  These values are updated for all patterns with 
*  unbounded alpha (0<alpha<cbound) whenever alpha is changed. 
*  They are also computed from scratch in the main loop when
*  looking for ktt condition violations. 
*  They are invalidated for a pattern when the pattern changes class from
*  being unbounded to being bounded in the  update_alpha() routine.
************************************************************************/

/* set cached value of F for pattern i 
--------------------------------------------------------*/
void svm_set_fcache(SVM_PARAMS *svm, int i, float x)
    {
    if ( verbose_flag >4 )
        printf("        set_fcache i %d new value %f old value %f\n", 
                                i, x, svm->fcache_value[i]);
    svm->fcache_value[i] = x;
    svm->fcache_flag[i] = 1;
    }

/* update fcache for positive below bound support vectors 
--------------------------------------------------------------------------*/
void svm_update_fcache_positive_below_bound(SVM_PARAMS *svm, int i1, int i2, 
           float a1old, float a1new, float a2old, float a2new,
           int y1, int y2, float *p1, float *p2)
    {
    int i_sv, i;
    float *psv;
    /* this goes over all non-bound support vectors */
    if ( svm->n_positive_below_bound > 0 )
        {
        for(i_sv=0; i_sv<svm->sv_n; i_sv++)
            if ( svm->sv_alpha_type[i_sv] == UNBOUNDED )
                {
                i = svm_fsv2pattern(svm, i_sv);
                psv = svm->exemplars[i];
        svm->fcache_value[i] = svm->fcache_value[i] + 
                  y1*(a1new - a1old)*svm_compute_kernel(svm, p1, psv)+
                  y2*(a2new - a2old)*svm_compute_kernel(svm, p2, psv);
                svm->fcache_flag[i] = 1;
            if (verbose_flag > 4)
                printf("        update_fcache_pbb i %d value %f\n", i, svm->fcache_value[i]); 
                }
        }
    }

/* get one fcache value 
-------------------------------*/
float svm_get_fcache(SVM_PARAMS *svm, int i, int desired_output, float *p1)
        {
        if ( svm->fcache_flag[i] == 0 )
                {
                svm->fcache_value[i] = 
                    svm_compute_no_bias_output(svm, p1) - desired_output;
                svm->fcache_flag[i] = 1;
                }
        return (svm->fcache_value[i]);
        }

/* invalidate fcache value 
-------------------------------*/
void svm_invalidate_fcache(SVM_PARAMS *svm, int i)
        {
        svm->fcache_flag[i] = 0;
        }

/* initialize fcache flags 
-------------------------------*/
void svm_init_fcache(SVM_PARAMS *svm, int ntrain)
        {
        int i;
        for( i=0; i<ntrain; i++)
                svm->fcache_flag[i] = 0;
        }

/************************************************************************
* NAME: routines keep track of upper and lower thresholds 
************************************************************************/
/* update values for current high threshold
-----------------------------------------------*/
static void svm_high_update(SVM_PARAMS *svm, float f2, int i2, int y2, float *pinput)
    {
    int j;
    if ( verbose_flag >= 2 && i2 == svm->i_high)
        printf("WARNING: b:high_update trying to update same index %d\n",i2);
    if ( verbose_flag >= 4)
        printf("        High update: new i %d old i %d y %d new bias %6.4f old bias %6.4f\n",
            i2, svm->i_high, y2, f2, svm->b_high);
    svm->b_high = f2;
    svm->i_high = i2;
    svm->i_high_desired_output = y2;
    for(j=0; j<svm->ninputs; j++)
        svm->i_high_vector[j] = pinput[j];
    }
/* update values for current low threshold
-----------------------------------------------*/
static void svm_low_update(SVM_PARAMS *svm, float f2, int i2, int y2, float *pinput)
    {
    int j;
    if ( verbose_flag >= 2 && i2 == svm->i_high)
        printf("WARNING: b:low_update trying to update same index %d\n",i2);
    if ( verbose_flag >= 4)
        printf("        low_update: new i %d old i %d y %d new bias %6.4f old bias %6.4f\n",
            i2, svm->i_low, y2, f2, svm->b_low);
    svm->b_low = f2;
    svm->i_low = i2;
    svm->i_low_desired_output = y2;
    for(j=0; j<svm->ninputs; j++)
        svm->i_low_vector[j] = pinput[j];
    }

/************************************************************************
* NAME: routines to translate indices between sv and patterns 
* DESCRIPTION:
************************************************************************/

/* translate a support vector index to a global pattern index
------------------------------------------------------------*/
int svm_fsv2pattern(SVM_PARAMS *svm, int i_sv)
    {
    int i_pattern;
    if ( i_sv < 0 || i_sv >= svm->sv_n)
         {
         fatal_error("cant provide entry %d of sv2pattern min 0 max %d\n",
            i_sv, svm->sv_n);
         }
        
    i_pattern = svm->sv2pattern[i_sv];
    if ( i_pattern < 0 || i_pattern >= svm->npatterns)
        {
        fatal_error("index %d in sv2pattern is bad %d min 0 max %d\n",
               i_sv, i_pattern, svm->npatterns); 
        }
    return(i_pattern);
    }

/* translate a global index for a pattern to the
   index for a support vector
------------------------------------------------------------*/
int svm_fpattern2sv(SVM_PARAMS *svm, int i_pattern)
    {
    int i_sv;
    if ( i_pattern < 0 || i_pattern > svm->npatterns)
        {
        fatal_error("cant provide entry %d of pattern2sv min 0 max %d\n",
            i_pattern, svm->npatterns);
        }
        
    i_sv = svm->pattern2sv[i_pattern];
    if ( i_sv < 0 || i_sv >= svm->sv_n)
        {
        fatal_error("index %d in pattern2sv is bad %d min 0 max %d\n",
               i_pattern, i_sv, (svm->sv_n)-1); 
        }
    return(i_sv);
    }

/************************************************************************
* NAME: compute kernel between two patterns 
* DESCRIPTION:
************************************************************************/
float svm_compute_kernel(SVM_PARAMS *svm, float *p1, float *p2 )
    {
    float total;
    double diff, sum_squared, dot_product;
    int j,ninputs, kernel_type;

    ninputs = svm->ninputs;
    kernel_type = svm->kernel_type;
    svm->total_kernel_evaluations++;

    if ( kernel_type == SMO_LINEAR || kernel_type == SMO_LINEAR_USE_SVS)
        {
        dot_product = 0.0;
        for(j=0; j<ninputs; j++)
            dot_product += p1[j]*p2[j];
        total = dot_product;
        }
    else if ( kernel_type == SMO_GAUSSIAN )
        {
        sum_squared = 0.0;
        for(j=0; j<ninputs; j++)
            { 
            diff = (p1[j] - p2[j]);
            sum_squared -= diff*diff;
            }
        sum_squared /= svm->gaussian_variance_times_two;
        total = exp(sum_squared);
        }
    else if ( kernel_type == SMO_POLYNOMIAL)
        {
        dot_product = 0.0;
        for(j=0; j<ninputs; j++)
            dot_product += p1[j]*p2[j];
        total = dot_product*(svm->inv_poly_divisor);
        for(j=1; j<(svm->polynomial_order); j++)
            total *= dot_product; 
        }
    else if ( kernel_type == SMO_INHOMOGENEOUS_POLYNOMIAL)
        {
        dot_product = 0.0;
        for(j=0; j<ninputs; j++)
            dot_product += p1[j]*p2[j];
        dot_product = (1.0 + dot_product)*(svm->inv_poly_divisor);
        total = dot_product;
        for(j=1; j<svm->polynomial_order; j++)
            total *= dot_product; 
        }
    else
        {
        fatal_error("invalid kernel_type %d\n", kernel_type);
        }
        
    return(total);
    }

/* compute output without the bias term
   for the improved algorithm
----------------------------------------------*/
float svm_compute_no_bias_output(SVM_PARAMS *svm, float *pinput)
        {
        return (svm_compute_output(svm, pinput) + svm->threshold);
        }


/* compute output for a giving point
   This is only valid after convergence, if b_high = blow
   It is used to compute the current output when threshold=0.0
------------------------------------------------------------*/
float svm_compute_output(SVM_PARAMS *svm, float *pinput)
    {
    float total, xkernel;
    int i,j;

    if ( svm->kernel_type == SMO_LINEAR )
        {
                total = 0.0;
        for(j=0; j<(svm->ninputs); j++)
            total += ((svm->weight[j])*pinput[j]);
        total -= (svm->threshold);
        }
    else if ( (svm->kernel_type == SMO_GAUSSIAN) || 
                  (svm->kernel_type == SMO_POLYNOMIAL) || 
                  (svm->kernel_type == SMO_INHOMOGENEOUS_POLYNOMIAL) ||
                  (svm->kernel_type == SMO_LINEAR_USE_SVS) )
        {
        /* sum up all patterns across all support vectors with 0 < alpha <= C */
        /* get sv patterns from exemplar matrix during training */
                if ( svm->support_vectors_valid == 0)
            {
                        total = 0.0;
            for(i=0; i<(svm->sv_n); i++)
                {
                total += (svm->sv_y[i])*(svm->sv_alpha[i])*
                      svm_compute_kernel(svm, pinput, 
                      svm->exemplars[svm_fsv2pattern(svm, i)]);
                }
            }
        /* get sv patterns from support_vector matrix after training */
                else 
            {
                        total = 0.0;
            for(i=0; i<(svm->sv_n); i++)
                {
                            xkernel = svm_compute_kernel(svm, pinput, 
                      svm->support_vectors[i]);
                total += (svm->sv_y[i])*(svm->sv_alpha[i])*xkernel;
                }
            }
        total -= svm->threshold;
        }
    else
        {
        fatal_error("invalid kernel type %d\n", svm->kernel_type);
        }
    return (total);
    }

/************************************************************************
* summary printouts 
************************************************************************/
/* print header for progress summary */
void svm_print_summary_header(int verbosity) 
    {
    if ( verbosity >= 1)
     printf("NSweeps Changed TChanged KernelEvals UnBounded AtUpper HiBias LoBias DeltaBias \n");
    }

/* print progress summary */
void svm_print_summary_values(SVM_PARAMS *svm, int nsweeps, int n_changed, int alpha_flag, int verbosity) 
    {
        if ( verbosity >= 1)
        {
            printf("%-7d %-7d %-8d %-11d %-9d %-7d %-6.2f %-6.2f %-8.6f\n", 
          nsweeps, n_changed, svm->total_pair_corrections, svm->total_kernel_evaluations,
          svm->n_positive_below_bound, svm->n_at_cbound, 
          svm->b_low, svm->b_high, svm->b_low - svm->b_high );

        /* print alpha values for patterns with 0<a<=C */
        if ( verbosity >=4 || alpha_flag > 0)
        {
        printf("\n");
        svm_print_positive_alphas(svm, 1);
        printf("\n");
        }
        }
    }

/* print index and alpha's for on non-zero training patterns 
always prints counts 
verbose = 1,  add index and alphas and x and actual and desired outputy 
---------------------------------------------------------------------------*/
void svm_print_positive_alphas(SVM_PARAMS *svm, int verbose)
{
int j,k,i_sv;
int ninputs_print=4;
int header_flag=0;
float *pinput;

if (verbose >= 1)
    {
    /* print input, output, alpha for sv's with unbounded alpha's 
         ------------------------------------------------------------*/
    if (svm->n_positive_below_bound  > 0)
        {
        printf ("i             ");
        if ( svm->ninputs >4)
            ninputs_print = 4;
        else
            ninputs_print = svm->ninputs;
        for(j=1; j<ninputs_print; j++)
            printf("        ");
        printf("x  out(desired)    out(actual)   0<alpha<C\n");
        header_flag = 1;
        for(i_sv=0; i_sv<svm->sv_n; i_sv++)
            {
            if ( svm->sv_alpha_type[i_sv] == UNBOUNDED )
                {
                j = svm_fsv2pattern(svm, i_sv);
                pinput = svm->exemplars[j];
                printf ("%-6d  ",j);
                for(k=0; k<ninputs_print; k++)
                    printf("%7.4f ", pinput[k]);
                printf("           %2d        %7.4f  %f\n", 
                svm->sv_y[i_sv], svm_compute_output(svm, pinput), 
                                                 svm->sv_alpha[i_sv]);
                }
            }
        }
    /* print input, output, alpha for sv's with alpha = C
         ------------------------------------------------------------*/
    if (svm->n_at_cbound  > 0)
        {
        printf ("\ni             ");
        if ( svm->ninputs >4)
            ninputs_print = 4;
        else
            ninputs_print = svm->ninputs;
        for(j=1; j<ninputs_print; j++)
            printf("        ");
        printf("x  out(desired)    out(actual)   alpha=C\n");
        for(i_sv=0; i_sv<svm->sv_n; i_sv++)
            {
            if ( svm->sv_alpha_type[i_sv] == CBOUND )
                {
                j = svm_fsv2pattern(svm, i_sv);
                pinput = svm->exemplars[j];
                printf ("%-6d  ",j);
                for(k=0; k<ninputs_print; k++)
                    printf("%7.4f ", pinput[k]);
                printf("           %2d        %7.4f  %g\n", 
                       svm->sv_y[i_sv], svm_compute_output(svm, pinput), 
                                                        svm->sv_alpha[i_sv]);
                }
            }

        }
    }
}

/* determine whether to update alpha for this point, and update if necessary 
   return 1 if kkt conditions satisfied, or for successful update
----------------------------------------------------------------------------*/
int svm_examine_example(SVM_PARAMS *svm, int i2, int y2, float *pinput )
    {
    float alpha2;
    float f2, *p2, *p1;
    int i1, j, y1;
    int atype;
    int optimality_flag;

    p2 = pinput;

    alpha2 = svm->alpha[i2];
    atype = (int)(svm->alpha_type[i2]);
        if ( verbose_flag >= 1)
        printf("examine_example %d y %d alpha %f type %d\n",
            i2, y2, alpha2, atype);

    /* alpha is positive and not bounded, work already done */
    if ( atype == UNBOUNDED )
        {
        f2 = svm_get_fcache(svm, i2, y2, pinput);
        }
    /* alpha at bound, need to update bias bounds and compute f */
    else
        {
        /* force fcache computation because its only correct for
                   unbounded patterns */
        svm_invalidate_fcache(svm, i2);
        f2 = svm_get_fcache(svm, i2, y2, pinput);

        /* update bias bounds */
        /* in set I1 or I2 , (y=1,alpha=0)  or (y= -1 , alpha=C) */
        if ( (y2 == 1 && atype == ZERO ) || (y2 == -1 && atype == CBOUND) )
            {
            if ( f2 < svm->b_high)
                {
                svm_high_update(svm, f2, i2, y2, pinput);
                }
            } 
        /* in set I3 or I4 , (y=1,alpha=C)  or (y= -1 , alpha=0) */
        else if ( (y2 == 1 && atype == CBOUND ) || (y2 == -1 && atype == ZERO) )
            {
            if ( f2 > svm->b_low)
                {
                svm_low_update(svm, f2, i2, y2, pinput);
                }
            } 
        }

    /*--------------------------------------------
       check optimality, whether violates kkt conditions
           using current b_low and b_high, if violated, find an index
           i1 to perform joint optimization with i2 
        -----------------------------------------------*/
    optimality_flag = 1;
    /* unbounded or in set I1 or I2 , (y=1,alpha=0)  or (y= -1 , alpha=C) */
    if ( atype == UNBOUNDED || (y2 ==1 && atype == ZERO) || ( y2 == -1 && atype == CBOUND))
        {
        if ( (svm->b_low - f2) > (2.0 * svm->kkt_tolerance) )
            {
            optimality_flag = 0;
            i1 = svm->i_low;
            y1 = svm->i_low_desired_output;
            for(j=0; j<svm->ninputs; j++)
                svm->v1_current[j] = svm->i_low_vector[j];
            p1 = svm->v1_current;
            }
        }

    /* unbounded or in set I3 or I4 , (y=1,alpha=C)  or (y= -1 , alpha=0) */
    if ( atype == UNBOUNDED || (y2 ==1 && atype == CBOUND) || ( y2 == -1 && atype == ZERO))
        {
        if ( (f2 - svm->b_high ) > (2.0 * svm->kkt_tolerance) )
            {
            optimality_flag = 0;
            i1 = svm->i_high;
            y1 = svm->i_high_desired_output;
            for(j=0; j<svm->ninputs; j++)
                svm->v1_current[j] = svm->i_high_vector[j];
            p1 = svm->v1_current;
            }
        }
    /* this example satisfies the kkt conditions, return */
    if ( optimality_flag == 1)
        {
        if ( verbose_flag >= 2)
            printf("    %d kkt satisfied, in examine_example\n", i2); 
        return ( 0 );
        }

    /* if i2 is unbounded, choose the better i1 */
    if ( atype == UNBOUNDED)
        {
        if ( (svm->b_low - f2) > (f2 - svm->b_high))
            {
            i1 = svm->i_low;
            y1 = svm->i_low_desired_output;
            for(j=0; j<svm->ninputs; j++)
                svm->v1_current[j] = svm->i_low_vector[j];
            p1 = svm->v1_current;
            }
        else
            {
            i1 = svm->i_high;
            y1 = svm->i_high_desired_output;
            for(j=0; j<svm->ninputs; j++)
                svm->v1_current[j] = svm->i_high_vector[j];
            p1 = svm->v1_current;
            }
        }
    
    
    if ( svm_take_step(svm, i1, i2, y1, y2, p1, p2) > 0)
        return(1);
    else
        {
        if ( verbose_flag >= 2)
        printf("    %d take step didnt in examine_example\n", i2); 
       return(0);
        }
    }


/* calcuate new apha's for two training patterns
   using an analytic solution
-----------------------------------------------------*/
int svm_take_step (SVM_PARAMS *svm, int i1, int i2, int y1, int y2, float *p1, float *p2)
    {
    float alpha1, alpha2;
    float a1, a2;   /* new alpha values */
    float sratio, gamma, diff, v1, v2, lbound, hbound;
    float l1bound, h1bound;
    float lobj, hobj;
    float k11, k12, k22, etta;
    int j;
    float f1,f2, f1new, f2new; 

    if ( verbose_flag >= 2)
        {
        printf("  take_step i1 %d i2 %d \n", i1, i2);
        }
    /* both points are the same */
    if ( i1 == i2)
        {
        if ( verbose_flag >= 3)
            printf("      patterns are same %d %d\n", i1, i2);
        return(0);
        }

    alpha1 = svm->alpha[i1];
    alpha2 = svm->alpha[i2];
    f1 = svm_get_fcache(svm, i1, y1, p1);
    f2 = svm_get_fcache(svm, i2, y2, p2);
    svm_compute_bounds(y1, y2, alpha1, alpha2, &lbound, &hbound, 
                                                   &l1bound, &h1bound, svm->cbound);
    if ( verbose_flag >= 3)
        printf("      lbound %f hbound %f l1bound %f h1bound%f\n", 
                       lbound, hbound, l1bound, h1bound);
    /* both bounds are equal */
    if (lbound == hbound)
        {
        if ( verbose_flag >= 2)
            printf("    exit take_step, lbound=hbound\n");
        return(0);
        }

    sratio = y1*y2;
    gamma = alpha1 + sratio*alpha2;
    k11 = svm_compute_kernel(svm, p1, p1);
    k12 = svm_compute_kernel(svm, p1, p2);
    k22 = svm_compute_kernel(svm, p2, p2);
    /* second derivative along diagonal */
    etta = 2*k12 - k11 -k22;
    if ( verbose_flag > 4)
        printf("        etta %f k11 %f k12 %f k22 %f y2 %d f1 %f f2 %f\n", 
                                etta, k11, k12, k22, y2, f1, f2);
    /* normal case with negative second derivative for maximum */
    if ( etta < 0 )
        {
        a2 = alpha2 - y2*(f1 - f2)/etta;
        if ( a2 < lbound)
            {
            a2 = lbound;
            }
            else if ( a2 > hbound)
            {
            a2 = hbound;
            }
        if ( verbose_flag >= 3 )
            {
            printf("      alpha2 update  delta: %f  2nd deriv: %f   old: %f   new: %f  l %f h %f\n", 
                                a2-alpha2, etta, alpha2, a2, lbound, hbound);
            }
        }
        /* unusual case, with non-negative second derivative */
    else
        {
        /* calculate partial objective function at end points where
                   a2 = lbound, or hbound and choose maximum 
                   v1 and v2 are missing the threshold term, but this
                   still seems to work */
        v1 = svm_compute_output(svm, p1) + svm->threshold -y1*alpha1*k11 - y2*alpha2*k12;
        v2 = svm_compute_output(svm, p2) + svm->threshold -y1*alpha1*k12 - y2*alpha2*k22;
        a2 = lbound;
        diff = gamma - sratio*a2;
        lobj = -sratio*a2 + a2 -.5*k11*diff*diff -.5*k22*a2*a2 -sratio*k12*diff*a2 
                        - y1*diff*v1 -y2*a2*v2;
        a2 = hbound;
        diff = gamma - sratio*a2;
        hobj = -sratio*a2 + a2 -.5*k11*diff*diff -.5*k22*a2*a2 -sratio*k12*diff*a2 
                        - y1*diff*v1 -y2*a2*v2;
        /* select alpha at endpoint with highest objective function */
        if (lobj > hobj + svm->epsilon_rounding )
            a2 = lbound;
        else if ( lobj < hobj - svm->epsilon_rounding )
            a2 = hbound;
        else
            a2 = alpha2;
        if ( verbose_flag >= 3 )
            {
            printf("      choose alpha using obj functiona\nlow obj: %f  lbound: %f  high obj: %f  hbound: %f  alpha: %6.2f\n",
                               lobj, lbound, hobj, hbound, a2);
            }
        }
    /* new modification to make sure round off error doesn't force a point
           to be a support vector */
        if ( a2 < svm->alpha_tolerance)
            a2 = 0.0;
        if ( a2 > (svm->cbound) - (svm->alpha_tolerance))
            a2 = svm->cbound;
    
    /* making too small progress */
    if ( fabs(a2 - alpha2) < (svm->epsilon_rounding)*(a2 + alpha2 + svm->epsilon_rounding))
        {
        if ( verbose_flag >= 2)
            printf("    exit take_step, too small progress %d %d new %f old %f diff %f\n", 
                i1, i2, a2, alpha2, a2-alpha2);
        return(0);
        }

    a1 = alpha1 + sratio*(alpha2 - a2);
    /* new modification to make sure round off error doesn't force a point
           to be a support vector */
        if ( a1 < svm->alpha_tolerance)
            a1 = 0.0;
        if ( a1 > (svm->cbound) - (svm->alpha_tolerance))
            a1 = svm->cbound;
    if ( verbose_flag >= 3 )
        printf("      alpha1 update  delta: %f  sratio: %f   old: %f   new: %f\n", 
            sratio*(alpha2-a2), sratio, alpha1, a1);

    /* -----------------------------------------------------------------
           update weight vector (linear machine only) 
         -------------------------------------------------------------------*/ 
    if (svm->kernel_type == SMO_LINEAR)
        {
        for(j=0; j<svm->ninputs; j++)
            {
            svm->weight[j] +=  y1*p1[j]*(a1 - alpha1)  + y2*p2[j]*(a2 - alpha2);
            }
        }


    /* -------------------------
           update fcache for positive alphas below bound
        -----------------------------------------*/
    svm_update_fcache_positive_below_bound(svm, i1, i2, alpha1, a1, alpha2, 
                                                            a2, y1, y2, p1, p2);

    /* -----------------------------------------------------------------
           update alpha's and set alpha belongs to
         -------------------------------------------------------------------*/ 
    svm_update_alpha(svm, i1, a1, alpha1, y1, p1);
        svm_update_alpha(svm, i2, a2, alpha2, y2, p2);
    
    /* -------------------------
           update f values for i1 and i2 
        -----------------------------------------*/
    f1new = f1 + y1*(a1 - alpha1)*k11 + y2*(a2 - alpha2)*k12; 
    f2new = f2 + y1*(a1 - alpha1)*k12 + y2*(a2 - alpha2)*k22; 
    /*
    if ( verbose_flag > 4)
        {
        printf("        i1 expensive fcache i %d f %f\n", 
                                i1, svm_compute_no_bias_output(p1) - y1); 
        printf("        i2 expensive fcache i %d f %f\n", 
                                i2, svm_compute_no_bias_output(p2) - y2); 
        }
    */
    svm_set_fcache(svm, i1, f1new);
    svm_set_fcache(svm, i2, f2new);

    /* -----------------------------------
           compute new lower bounds on bias using search only
           through i1, i2, and positive below bound alphas
        ------------------------------------------------------*/
    svm_update_bias_bounds (svm, i1, i2, y1, y2, p1, p2);
    

    svm_update_pair_corrections(svm);

    
    return(1);
    
    }

/* compute upper and lower alpha bounds for analytic optimization */
void svm_compute_bounds(float y1, float y2, float alpha1, float alpha2, 
           float *lbound, float *hbound, float *l1bound, float *h1bound, float cbound)
    {
    float diff, sum;

    /* two targets not equal 
    ------------------------*/
    if ( y1 != y2 )
        {
        /* alpha2 bounds
           L = max(0, alpha2 -alpha1) 
           H = min(C, C + alpha2 -alpha1) */
        /* alpha1 bounds
           L = max(0, alpha1 -alpha2) 
           H = min(C, C + alpha1 - alpha2) */

            diff = alpha2 - alpha1;
            if ( diff < 0.0 )
            {
            *lbound = 0.0;
            *hbound = cbound + diff;
            *l1bound = -diff;
            *h1bound = cbound;
            }
        else
            { 
            *lbound = diff;
            *hbound = cbound;
            *l1bound = 0.0;
            *h1bound = cbound -diff;
            }
        }
    /* two targets equal 
    ------------------------*/
    else 
        {
        /* alpha2 bounds
           L = max(0, alpha1 +alpha2 -C) 
           H = min(C,  alpha1 +alpha2) */
        /* alpha1 bounds
           L = max(0, alpha1 +alpha2 -C) 
           H = min(C,  alpha1 +alpha2) */
        sum = alpha1 + alpha2 -cbound;
        if ( sum < 0.0 )
            {
            *lbound = 0.0;
            *hbound = cbound + sum;
            *l1bound = 0.0;
            *hbound = cbound + sum;
            }
        else
            { 
            *lbound = sum;
            *hbound = cbound;
            *l1bound = sum;
            *h1bound = cbound;
            }
        }
    }

/* store new updated alpha, and adjust arrays and their counts as required 
---------------------------------------------------------------------------*/
void svm_update_alpha(SVM_PARAMS *svm, int i1, float new, float old, 
                      int desired_output, float *pinput)
    {
    int i_sv;
    int atype;


    atype = (int)(svm->alpha_type[i1]);

    if ( new < 0.0 )
        new = 0.0;
    if ( new > svm->cbound )
        new = svm->cbound;

    /* old is zero 
    ------------------------------------------------*/
    if ( atype == ZERO )
        {
        /* old =0 and new positive, not at bound */
        if ( ( new > 0.0 ) && (new < svm->cbound ) )
            {
            /* printf("        update_alpha old zero new unbound\n");*/
            i_sv = svm_sv_add(svm, i1, new, desired_output, pinput, (int)UNBOUNDED);
            svm->n_positive_below_bound += 1;
            svm->n_zero -= 1;
            svm->alpha_type[i1] = UNBOUNDED;
            svm->alpha[i1] = new;
            }
        /* old =0 and new positive at bound */
        else if ( new >= svm->cbound )
            {
            /*printf("debug: old zero new cbound\n");*/
            i_sv = svm_sv_add(svm, i1, new, desired_output, pinput, (int)CBOUND);
            svm->n_at_cbound += 1;
            svm->n_zero -= 1;
            svm->alpha_type[i1] = CBOUND;
            svm->alpha[i1] = new;
            }
        /* old = 0 and new =0 doesn't change anything, so skip */
        }
    /* old is positive, not at bound 
    ------------------------------------------------*/
    else if ( atype == UNBOUNDED )
        {
        /* old positive, not at bound new zero */
        if ( new <=  0.0)  
            {
            /* printf("debug: old unbound new zero\n"); */
            if (verbose_flag > 3)
                printf("      old alpha unbounded new zero i %d %f\n", i1,new);
            i_sv = svm_fpattern2sv(svm, i1);
            svm->n_positive_below_bound -= 1;
            svm_sv_remove(svm, i_sv);
            svm->n_zero += 1;
            svm_invalidate_fcache(svm, i1);
            svm->alpha_type[i1] = ZERO;
            svm->alpha[i1] = new;
            }
        /* old positive, not at bound and new positive not at bound */
        else if ( ( new > 0.0 ) && (new < svm->cbound ) )
            {
            if (verbose_flag > 3)
                printf("      old alpha unbounded new unbounded i %d %f\n", i1,new);
            i_sv = svm_fpattern2sv(svm, i1);
            svm_sv_update(svm, i_sv, new, (int)UNBOUNDED);
            svm->alpha[i1] = new;
            }
        /* old positive, not at bound new at bound */
        else if ( new >= svm->cbound)  
            {
            /*printf("debug: old unbound new at cbound\n");*/
            if (verbose_flag > 3)
                printf("      old alpha unbounded new at cbound  i %d %f\n", i1,new);
            i_sv = svm_fpattern2sv(svm, i1);
            svm_sv_update(svm, i_sv, new, (int)CBOUND);
            svm->n_positive_below_bound -= 1;
            svm->n_at_cbound += 1;
            svm_invalidate_fcache(svm, i1);
            svm->alpha_type[i1] = CBOUND;
            svm->alpha[i1] = new;
            }
        }
    /* old is at bound 
    ------------------------------------------------*/
    else if ( atype == CBOUND )
        {
        /* old at bound, new positive */
        if ( ( new > 0.0) && (new < svm->cbound) )
            {
            /*printf("debug: old cbound new unbounded\n");*/
            if (verbose_flag > 3)
                printf("       old at cbound new positive i %d alpha %f\n", i1, new);
            i_sv = svm_fpattern2sv(svm, i1);
            svm_sv_update(svm, i_sv, new, (int)UNBOUNDED);
            svm->n_at_cbound -= 1;
            svm->n_positive_below_bound += 1;
            svm->alpha_type[i1] = UNBOUNDED;
            svm->alpha[i1] = new;
            }
        /* old at bound, new at bound doesn't change anything so ignore */
        /* old at bound, new at zero */
        else if ( new <= 0.0 )
            {
            /*printf("debug: old cbound new zero\n");*/
            if (verbose_flag > 3)
                printf("      old at cbound new zero i %d alpha %f\n", i1, new);
            i_sv = svm_fpattern2sv(svm, i1);
            svm->n_at_cbound -= 1;
            svm_sv_remove(svm, i_sv);
            svm->n_zero += 1;
            svm->alpha_type[i1] = ZERO;
            svm->alpha[i1] = new;
            }
        }

    if ( svm->n_zero + svm->n_positive_below_bound + svm->n_at_cbound != svm->npatterns)
        {
        warning("sums of separate counts %d not equal to %d after updating\n",
            svm->n_zero + svm->n_positive_below_bound + svm->n_at_cbound, svm->npatterns);
        svm_print_positive_alphas(svm, 1);
        }
    if ( verbose_flag >= 2)
        {
        printf( "    updated alpha %d new %f old %f new type %d\n", 
                         i1, svm->alpha[i1], old, svm->alpha_type[i1]);
        }
    }

/************************************************************************
* support vector handling routines 
************************************************************************/
/* update support vector alpha and alpha type
-------------------------------------------------*/
void svm_sv_update(SVM_PARAMS *svm, int i_sv, float alpha, int atype)
    {
    svm->sv_alpha[i_sv] = alpha;
    svm->sv_alpha_type[i_sv] = atype;
    }

/* add a support vector when alpha for an input becomes positive
---------------------------------------------------------------*/
int svm_sv_add(SVM_PARAMS *svm, int i_pattern, float a, 
               int desired_output, float *pinput, int atype)
{
int i_sv;

i_sv = svm->sv_n;
if (i_sv >= svm->npatterns )
    {
    fatal_error("Too many support vectors to store %d\n", i_sv);
    }
/* add a new support vector if this pattern isn't already on list */
if ( svm->pattern2sv[i_pattern] ==  NOT_A_SUPPORT_VECTOR)
    { 
    svm->sv_alpha[i_sv] = a;
    svm->sv_alpha_type[i_sv] = atype;
    svm->sv_y[i_sv] = desired_output;
    svm->sv2pattern[i_sv] = i_pattern;
    svm->pattern2sv[i_pattern] = i_sv;
    svm->sv_n += 1;
    if ( verbose_flag >= 3)
        printf("      sv_add add pattern %d in entry %d sv_n %d pat2sv %d\n", 
            i_pattern, i_sv, svm->sv_n, svm->pattern2sv[i_pattern]);
    }
/* pattern is on list, just update */
else    
    {
    i_sv = svm->pattern2sv[i_pattern];
    warning("sv_add updated pattern %d in entry %d \n", i_pattern, i_sv);
    svm->sv_alpha[i_sv] = a;
    svm->sv_alpha_type[i_sv] = atype;
    }
return(i_sv);
}
/* remove a support vector when alpha for an input becomes zero
---------------------------------------------------------------*/
void svm_sv_remove(SVM_PARAMS *svm, int i_old_sv)
{
int i_old_pattern, i_last_sv, i_last_pattern;

if ( verbose_flag >= 3)
    printf("      sv_remove %d pattern %d sv_n %d old alpha %f\n", 
        i_old_sv, svm_fsv2pattern(svm, i_old_sv), svm->sv_n, svm->sv_alpha[i_old_sv]); 

/* swap last element down into free slot on support vector list */
i_last_sv = (svm->sv_n) -1;
i_old_pattern = svm_fsv2pattern(svm, i_old_sv);


if ( i_last_sv <0 )
    {
    fatal_error("trying to remove too many support vectors\n");
    }

/* fix forward and backward pointers */
if ( svm->pattern2sv[i_old_pattern] == NOT_A_SUPPORT_VECTOR)
    {
    fatal_error("sv_remove pattern2sv[%d] is not valid\n", i_old_pattern);  
    }
/* remove the only remaining item on the list without swapping */
if ( svm->sv_n == 1)
    {
    if (i_old_sv != 0)
        {
        fatal_error("sv_remove, can only remove index 0, not %d\n", i_old_sv);
        }
    svm->pattern2sv[i_old_pattern] = NOT_A_SUPPORT_VECTOR; 
    svm->sv2pattern[i_old_sv] = NOT_A_SUPPORT_VECTOR; 
    }
/* remove the last item on the list without swapping */
else if ( i_old_sv == i_last_sv)
    {
    svm->pattern2sv[i_old_pattern] = NOT_A_SUPPORT_VECTOR; 
    svm->sv2pattern[i_old_sv] = NOT_A_SUPPORT_VECTOR; 
    }
/* remove an intermediate item on the list */
else    
    {
    svm->pattern2sv[i_old_pattern] = NOT_A_SUPPORT_VECTOR; 
    i_last_pattern = svm_fsv2pattern(svm, i_last_sv);
    svm->pattern2sv[i_last_pattern] = i_old_sv;
    svm->sv2pattern[i_old_sv] = i_last_pattern;
    svm->sv_alpha[i_old_sv] = svm->sv_alpha[i_last_sv];
    svm->sv_alpha_type[i_old_sv] = svm->sv_alpha_type[i_last_sv];
    svm->sv_y[i_old_sv] = svm->sv_y[i_last_sv];
    }
svm->sv_n -= 1;
}

/* -----------------------------------
   Compute new upper and lower bounds on bias using only partial search 
   including i1, i2, and positive below bound alphas.
   Use intersection with these i's and those in formula 11a and 11b in Keerthi et al. paper 
------------------------------------------------------*/
void svm_update_bias_bounds(SVM_PARAMS *svm, int i1, int i2, int y1, int y2, float *p1, float *p2)
    {
    int i, i_sv;
    float xmax, xmin;
    int imax, imin;
        int max_desired_output, min_desired_output;
        float *pmax_input, *pmin_input;
    int atype1, atype2;

    if ( verbose_flag >= 2)
        printf("    update_bias_bounds %d %d\n", i1, i2);

    atype1 = (int)(svm->alpha_type[i1]);
    atype2 = (int)(svm->alpha_type[i2]);

    /* find new upper and lower bound on bias 
       ------------------------------------------*/
    xmax = -MAXFLOAT;
    xmin = MAXFLOAT;

    /* -----------------------------
           max search for i_low  
    --------------------------------*/
    /* search for max check against i1 if i1 in set I0 I3 or I4*/
    if ( (svm->fcache_value[i1] > xmax)
       && ((atype1==UNBOUNDED) || (y1==1 && atype1==CBOUND) || (y1== -1 &&atype1==ZERO)))
        {
        xmax = svm->fcache_value[i1];
        imax = i1;
        pmax_input = p1;
        max_desired_output = y1;
        if ( verbose_flag >= 4)
            printf("        new i_low max i1 I0||I3||I4  max %f i %d y %d atype1 %d\n", 
                                        xmax, imax, y1, atype1);
        }

    /* search for max check against i2  if i2 in set I0 I3 or I4 */
    if ( (svm->fcache_value[i2] > xmax) 
         && ((atype2 == UNBOUNDED) || (y2==1 && atype2==CBOUND) || (y2== -1 && atype2==ZERO)))
        {
        xmax = svm->fcache_value[i2];
        imax = i2;
        pmax_input = p2;
        max_desired_output = y2;
        if ( verbose_flag >= 4)
            printf("        new i_low max i2 I0||I3||I4  max %f i %d y %d atype2 %d\n", 
                                        xmax, imax, y2, atype2);
        }

    /* combined search for i_low and i_high
        -----------------------------------------*/
    /* check against all unbounded points */
    for(i_sv=0; i_sv<svm->sv_n; i_sv++)
        if ( svm->sv_alpha_type[i_sv] == UNBOUNDED )
            {
            i = svm_fsv2pattern(svm, i_sv);
            if ( svm->fcache_value[i] > xmax)
                {
                xmax = svm->fcache_value[i];
                imax = i;
                pmax_input = svm->exemplars[i];
                max_desired_output = svm->sv_y[i_sv];
                }
            if ( svm->fcache_value[i] < xmin)
                {
                xmin = svm->fcache_value[i];
                imin = i;
                pmin_input = svm->exemplars[i];
                min_desired_output = svm->sv_y[i_sv];
                }
            }

    /* special search for i_high
        --------------------------------*/
    /* search for min check against i1  if in set I0 I1 or I2*/
    if ( (svm->fcache_value[i1] < xmin)
         && ((atype1 == UNBOUNDED) || (y1==1 && atype1 == ZERO) || (y1== -1 && atype1==CBOUND)))
        {
        xmin = svm->fcache_value[i1];
        imin = i1;
        pmin_input = p1;
        min_desired_output = y1;
        if ( verbose_flag >= 4)
            printf("        new i_high min i1 I0||I1||I2  min %f i %d y %d atype1 %d\n", 
                                        xmin, imin, y1, atype1);
        }
    /* search for min check against i2 if in set I0, I1, or I2*/
    if ( (svm->fcache_value[i2] < xmin)
        && ((atype2 == UNBOUNDED) || (y2==1 && atype2 == ZERO) || (y2== -1 && atype2==CBOUND)))
        {
        xmin = svm->fcache_value[i2];
        imin = i2;
        pmin_input = p2;
        min_desired_output = y2;
        if ( verbose_flag >= 4)
            printf("        new i_high min i2 I0||I1||I2  min %f i %d y %d atype2 %d\n", 
                                        xmin, imin, y2, atype2);
        }

    /* check and update i_high and i_low
        -------------------------------------*/
    if ( xmax == -MAXFLOAT)
        warning("didnt find new i_low, ilow %d f %f\n", imax, xmax);
    if ( xmin == MAXFLOAT)
        warning("didnt find new i_high, ihigh %d f %f\n", imin, xmin);;
    /*
    printf("debug: update_bias low update %d %f\n", imax, xmax);
    printf("debug: update_bias high update %d %f\n", imin, xmin);
    */
    svm_low_update( svm, xmax, imax, max_desired_output, pmax_input);
    svm_high_update( svm, xmin, imin, min_desired_output, pmin_input);

    /*
    This occurs at times and is not necessarily bad
    if ( b_high > b_low )
        printf("WARNING b_high %f > b_low %f %d %d\n", b_high, b_low, i_high, i_low);
    */
        
    }

/* increment number of times training pairs have been corrected */
void svm_update_pair_corrections(SVM_PARAMS *svm)
        {
        svm->total_pair_corrections += 1;
        }


/*
fit a sigmmoid passing through unclipped svm output = 0, posterior = .5
with one of nine slopes to the training data.
This is a simple hack seems to work well.
Sigmoid form is 1/(1+exp(ax)) where x is the un clipped output of the
svm classifier. A is selected from the nine values in posterior_achoices
using only the training patterns that not support vectors. 
Actually using training patters with outputs not near +/- 1.
If there are too few training patterns, the flattest posterior curve
is selected. Fitting this sigmoid to the posterior does not affect the 
classification error. It does allow different priors to be assigned to
classes and class costs to be varied. This fit may not be accurate under
all conditions, but seems to work well for a wide range of conditions. 

A more complicated fit would be to fit on test or evalution data, but
this would require major changes to the way lnknet currently works.
It would be reasonable to add another program that maps network outputs
to posterior probabilities using a sigmoid and test data.
----------------------------------------------------------------------*/
#define NPOSTERIOR_ACHOICES  34
void svm_fit_posterior_sigmoid(SVM_PARAMS *svm)
    {
    int i, j, i_min, atotal, btotal, ntotal;
    static float posterior_achoices[NPOSTERIOR_ACHOICES]=
             {-10.0,  -4.0, -2.0, -1.0,
                -.6,   -.4,  -.3,  -.2, -.15, -.1,
               -.06,  -.04, -.03, -.02, -.01,
              -.005, -.001, .001, .005,
                .01,   .02,  .03,  .04, .06,
                 .1,   .15,   .2,   .3, .4, .6,
                1.0,   2.0,  4.0, 10.0};

    double perr[NPOSTERIOR_ACHOICES];
    double a, err_min, error, yexp;
        float  a_min, output, desired_output;
    int ntrain;

    ntrain = svm->npatterns;

    /* initialize error accumulation */
    for(j=0; j<NPOSTERIOR_ACHOICES; j++)
        {
        perr[j]=0.0;
        }
    btotal = 0;
        atotal = 0;
        ntotal = 0;

    /* accumulate error for training patterns with output not near +/-1 
           this reduces bias from bound and unbound support vectors 
           eliminates points within +/- 5 kkt tolernance */
    for(i=0; i<ntrain; i++)
        {
                /* Computes p(A|x), desired output for class A is 1*/
        output = svm_compute_output(svm, svm->exemplars[i]);
        if ( svm->y[i] == 1)
            {
            desired_output = 1;
                        atotal++;
            }
        else
            {
            desired_output = 0;
                        btotal++;
            }
        /* update accumulated errors */
        for(j=0; j<NPOSTERIOR_ACHOICES; j++)
            {
            a = posterior_achoices[j];
                        yexp = 1.0/(1.0+exp(a*output));
            error = desired_output - yexp;
            error *= error;
            /* weight patterns near +/- 1 less
               because they have been affected more by training */
            if ( fabs(output -1.0) < (svm->kkt_tolerance)*5 || 
                 fabs(output + 1) < (svm->kkt_tolerance)*5)
                {
                perr[j] += .01*error;
                }
            else
                {
                perr[j] += error;
                }
            ntotal += 1;
                }
        }
    /* find a that minimizes the total squared error 
    ------------------------------------------------*/
    err_min = MAXFLOAT;
    for(j=0; j<NPOSTERIOR_ACHOICES; j++)
        {
        if ( perr[j]<err_min )
            {
            i_min = j;
            err_min = perr[j];
            a_min = posterior_achoices[j];
            }
        }
    if ( verbose_flag >= 1)
        {
        printf("posterior probs: amin %6.3f imin %d err %f n %d\n",
                        a_min, i_min, err_min, ntotal);
        }
    svm->posterior_aexp = a_min;
    }

#define MAX_WARNING 400
#define MAX_COMPONENT 100

/* check that KKT conditions are satisfied 
   This routine also combines all errors together, formats, and prints
   one error message per training run. If the verbosity is <=2 then you
   only get the one message. It it is >=3 then individual errors are provided
   with more detail. Error handling was added because many errors often occur
   in one run and details aren't necessary. Summary errors are only reported
   if KKT conditions aren't satisfied, even if the upper and lower biases are
   not within tolerance.
---------------------------------------------------------------------------*/
void svm_check_final_kkt_conditions( SVM_PARAMS *svm, int verbosity, char *run_name, char **labels, int bias_different_flag)
    {
    int bad_output_count; /* output not at correct value or range */
    int bad_alpha_count;    /* alpha out of range */
    int aytotal_flag;   /* sum of alpha*output not = 0 */
    int  misclassified_count;   /* misclassifed if use  y>= 0 as threshold */
    int bad_flag;
    float diff,yactual, aytotal, alpha_max;
    int ntrain;
    int i,  desired_output;
    float *pinput;
    int err_model, desired_class, class;
    FILE *fnull=NULL;
    float outputs[2];
    int bad_unbounded_alpha_count, bad_bounded_alpha_count;
    int bad_big_alpha_count, bad_negative_alpha_count;
    static char warning_string[MAX_WARNING];
    static char warning_component[MAX_COMPONENT];
    int all_alpha_zero_flag, zero_alpha_count;
    
    ntrain = svm->npatterns;
    /* initialize */
    bad_output_count = 0;
    bad_alpha_count = 0;
    aytotal_flag= 0;
    misclassified_count = 0;
    aytotal = 0.0;
    alpha_max = 0.0;
    bad_flag=0; 
    bad_unbounded_alpha_count=0;
    bad_bounded_alpha_count=0;
    bad_big_alpha_count=0;
    bad_negative_alpha_count=0;
    all_alpha_zero_flag=0;
    zero_alpha_count=0;

    err_model = report_build (verbosity, 0, 2);
    report_init(err_model, 0, svm->ninputs, SQ_ERR, 0, 0);

    for(i=0; i<ntrain; i++)
       {
       desired_output = svm->y[i];
       pinput = svm->exemplars[i];

        /* count number of alpha's at zero */
        if ( svm->alpha[i] == 0)
        zero_alpha_count++;

    /* check alpha, 0< a<= C */
    if ( svm->alpha[i] < 0.0)
        {
        bad_alpha_count++;
        bad_flag++;
                bad_negative_alpha_count++;
                if ( verbosity >= 3)
                    {
            warning("NEGATIVE ALPHA pattern %d alpha %f\n",
                i, svm->alpha[i]);
            }
        }
        /* alpha above cbound */
    if ( svm->alpha[i] > svm->cbound)
        {
        bad_alpha_count++;
        bad_flag++;
                bad_big_alpha_count++;
                if ( verbosity >= 3)
                    {
            warning("ALPHA > C pattern %d alpha %f cbound %f\n",
                i, svm->alpha[i], svm->cbound);
            }
        }
    if ( svm->alpha[i] > alpha_max)
        alpha_max = svm->alpha[i];

    /* check that sum of output and alpha is zero */
    aytotal += (svm->alpha[i])*(svm->y[i]);
    
        /* check that the output for unbound support vectors is +/- 1 */

    yactual = svm_compute_output(svm, pinput);
    diff = yactual - desired_output;
    if ( svm->alpha[i] > 0 && svm->alpha[i] < svm->cbound )
        {
        if ( fabs(diff) > svm->final_check_tolerance) 
            {
            bad_flag++;
            bad_output_count++;
            bad_unbounded_alpha_count++;
            if ( verbosity >= 3)
                    {
                        /* correct */
                        if ( (desired_output == 1 && yactual > 0) ||
                             (desired_output == -1 && yactual < 0) )
                {
             warning("%s\nalpha %f pattern %d desired = %d actual %f diff %f\n%s%s",
                         "BAD OUTPUT not exactly +/1 for an unbound support vector",
              svm->alpha[i], i, desired_output, yactual, diff,
     "This may be OK for classification, because the sv is classified correctly.",
                         "See svm man page for suggestions.");
                }
       
                        /* incorrect */
            else
                {
             warning("%s\nalpha %f pattern %d desired = %d actual %f diff %f\n%s%s",
                         "BAD OUTPUT not exactly +/1 for an unbound support vector",
              svm->alpha[i], i, desired_output, yactual, diff,
     "This is BAD for classification, because the sv is classified wrong.",
                         "See svm man page for suggestions.");
                }
                            }
            }

        }
    /* patterns at bound shouldn't satisfiy kkt conditions */
    if ( svm->alpha[i] == svm->cbound )
        {
        if ( ((desired_output == 1) && (yactual>1+(svm->final_check_tolerance))) ||
             ((desired_output == -1 ) && (yactual <-1-(svm->final_check_tolerance))))
            {
            bad_output_count++;
            bad_flag++;
            bad_bounded_alpha_count++;
            if ( verbosity >= 3)
                {
            warning("%s\nalpha %f pattern %d desired %d actual %f diff %f\n%s",
                "BAD OUTPUT alpha at cbound but satisfies KKT conditions",
              svm->alpha[i], i, desired_output, yactual, diff,
                        "This is OK for classification, but unusual");
                            }
            }
        }
    /* tally all incorrectly classified points */
        if ( ((yactual > 0.0 +(svm->kkt_tolerance) ) && (desired_output == -1)) ||
             ((yactual < 0.0 -(svm->kkt_tolerance) ) && (desired_output == 1)))
            {
            misclassified_count++;
            if ( verbose_flag >= 1)
                printf("MISCLASSIFIED pattern %d  alpha %f desired %d actual %f\n",
                i, svm->alpha[i], desired_output, yactual);
            }
    /* get ready to report confusion matrix and other stats */
    outputs[0]= yactual;
        outputs[1]= -1.0*yactual;
        if (desired_output == 1)
        {
        desired_class =0;
        }
    else
        {
        desired_class = 1;
        }
        if ( yactual >= 0)
        {
        class = 0;
        }
    else
        {
        class = 1;
        }
    report_class_trial(err_model, fnull, pinput,  outputs,
                           desired_class, class, i);

      }
    report_overall(err_model, stdout, run_name, labels);
    report_free(err_model);


    /* all apha's can be zero if data is not normalized and alpha step
       is too small compared to alpha_tolerance */
    if ( zero_alpha_count == ntrain)
        {
        all_alpha_zero_flag=1;
        bad_flag++;
        }
            
    /* final checks and printout */
    if ( aytotal > (svm->kkt_tolerance) || 
             aytotal < -(svm->kkt_tolerance))
        {
                if ( verbosity >= 3)
            warning("SUM OF DESIRED OUTPUTS * ALPHA's is NOT ZERO %f\n%s\n%s %f.\n", aytotal,
                "    This may be OK for classification",
                "    if the difference is small relative to kkt_tolerance which is", svm->kkt_tolerance);
        aytotal_flag = 1;
        bad_flag++;
        }
    if ( bad_flag == 0 )
        {
        if ( verbose_flag >= 1)
            {
            printf("GOOD SOLUTION: Karush-Kuhn-Tucker (KKT) Conditions Satisfied\n");
            }
        }
    else 
        {
            strncpy(warning_string,
                      "BAD SOLUTION: Karush-Kuhn-Tucker (KKT) Conditions NOT Satisfied\n",
                       MAX_WARNING);
                if ( all_alpha_zero_flag)
                   strncat(warning_string, "All alpha's are zero.\n    Try normalizing input patterns to zero mean unit variance.\n    Also try reducing alpha_tolerance because the alpha step size may very small.\n",MAX_WARNING);

                if ( bias_different_flag)
                   strncat(warning_string, "Hi and low bias differ.\n",MAX_WARNING);

        if ( aytotal_flag )
           strncat(warning_string,"Y*alpha sum is not zero.\n",MAX_WARNING);

        if (bad_unbounded_alpha_count == 1)
           strncat(warning_string,"Unbounded SV output not +/- 1.\n    This may be ok if the difference is not too large.\n",MAX_WARNING);
        if (bad_unbounded_alpha_count > 1)
                   {
           snprintf(warning_component, MAX_COMPONENT, 
            "Unbounded SV output not +/- 1 (%d times).\n    This may be ok if the difference is not too large.\n", 
                             bad_unbounded_alpha_count);
                   strncat(warning_string, warning_component, MAX_WARNING);
           }
        if (bad_bounded_alpha_count == 1)
           strncat(warning_string,"At bound SV output satisfies kkt conds.\n    This doesn't affect performance, but is unusual.\n", MAX_WARNING);
        if (bad_bounded_alpha_count > 1)
                   {
           snprintf(warning_component, MAX_COMPONENT, 
                    "At bound SV output satisfies kkt conds (%d times).\n    This doesn't affect performance but is unusual.\n", 
                             bad_bounded_alpha_count);
                   strncat(warning_string, warning_component, MAX_WARNING);
           }
        
                if (bad_big_alpha_count == 1) 
           strncat(warning_string,"Alpha > cbound.\n", MAX_WARNING);
                if (bad_big_alpha_count > 1)
                   {
           snprintf(warning_component, MAX_COMPONENT, 
                    "Alpha > cbound (%d times).\n", bad_big_alpha_count);
                   strncat(warning_string, warning_component, MAX_WARNING);
           }

                if (bad_negative_alpha_count == 1)
           strncat(warning_string,"Alpha negative.\n", MAX_WARNING);
                if (bad_negative_alpha_count > 1)
                   {
           snprintf(warning_component, MAX_COMPONENT, 
                    "Alpha negative (%d times).\n", bad_negative_alpha_count);
                   strncat(warning_string, warning_component, MAX_WARNING);
           }
        warning("%s", warning_string);

        }
        
    }

/* copy exemplar vectors from training matrix to support vector matrix 
---------------------------------------------------------------------------*/
void svm_save_final_support_vectors(SVM_PARAMS *svm)
    {
    int i_sv, i_exemplar, j; 
    
    /* make support vector matrix big enough */
    deallocate_matrix (svm->support_vectors);
    svm->support_vectors = allocate_matrix(svm->sv_n, svm->ninputs);

    /* copy support vectors from training matrix into sv matrix */
    for(i_sv=0; i_sv<svm->sv_n; i_sv++)
    {
    i_exemplar = svm_fsv2pattern(svm, i_sv);
        for(j=0; j<svm->ninputs; j++)
        {
        svm->support_vectors[i_sv][j] = svm->exemplars[i_exemplar][j];
        }
    }
    svm->support_vectors_valid=1;
   }

/*****************************************************************************
*       classify one pattern for high-level multi class classifier 
******************************************************************************/
int svm_classify_multi(int model, float *pinput)
    {
    SVM_MULTI *svm_multi;
    SVM_PARAMS *svm;
    int i, j, ia, ib, imax, nmax, i_cmin;
    float total, ya, yb, y_cmin;

    svm_multi = svmpp_multi[model];

    /* ----------------------------------
        One vs all other component classifiers. 
        Picking the maximum or voting gives the
        same result. This code picks the maximum
        and returns the class A output of each
        component classifier.
       ----------------------------------------- */
    if ( svm_multi->type == 0 )
    {
        for(j=0; j<svm_multi->noutputs; j++)
            svm_multi->outputs[j] = 0.0;
    
        /* get results from all component classifiers */
    for(i=0; i<(svm_multi->n); i++)
        {
        /* get outputs of each component classifier */
        j = svm_multi->component_model[i];
        svm = svmpp[j];
        svm_eval(j,  pinput);
        ya = svm->outputs[0];
        yb = svm->outputs[1];
        svm_multi->component_outputs[i][0] = ya;
        svm_multi->component_outputs[i][1] = yb;
        ia = svm->multi_class_a;
        ib = svm->multi_class_b;
        svm_multi->outputs[ia] = ya;
        /*
            printf("%d model %d  ia %d ib %d ya %6.5f  yb %6.5f input %f %f\n", 
                    i, j, ia, ib, ya, yb, pinput[0], pinput[1]); 
        
        for(k=0; k<svm_multi->noutputs; k++)
        {
        if ( svm_multi->outputs[k] > 0.00001 )
            printf("%6.5f ", svm_multi->outputs[k]);
        else    
            printf("  -    ");
        }
             printf("\n");
         */
         }
        /* special case if only one class because the
           other symmetric classifier was not been built */
        if ( svm_multi->n == 1)
         {
         svm_multi->outputs[ib] = yb;
             }
    }

    /* ----------------------------------
        Combinatorial (n*n-1/2) binary classifiers.
        Select class by largest vote.
        Break vote based on largest output for those that are tied.
        Can stop if one class gets M-1 votes. 
        If fit using sigmoid, output is minimum over A vs other
        classifiers because if outputs are posteriors, the desired
        posterior is less than this minimum (due to added terms in
        the denominator).
       ----------------------------------------- */
    else 
    {
        /* initialize votes and class minima */
        for(i=0; i<svm_multi->noutputs; i++)
        {
        svm_multi->votes[i] = 0;
            svm_multi->class_min[i] = MAXFLOAT;
        }
        /* sum over all component classifiers */
    for(i=0; i<svm_multi->n; i++)
        {
        /* get outputs of each component classifier */
        j = svm_multi->component_model[i];
        svm = svmpp[j];
        svm_eval(j,  pinput);
        ya = svm->outputs[0];
        yb = svm->outputs[1];
        svm_multi->component_outputs[i][0] = ya;
        svm_multi->component_outputs[i][1] = yb;
        ia = svm->multi_class_a;
        ib = svm->multi_class_b;

            /* adjust class minimum outputs */
            if ( ya < svm_multi->class_min[ia] )
        svm_multi->class_min[ia] = ya;
            if ( yb < svm_multi->class_min[ib] )
        svm_multi->class_min[ib] = yb;

            /* increment votes */
        /* class A is greater */
            if ( ya > yb )
        {
                svm_multi->component_decisions[i] = ia;
        svm_multi->votes[ia] += 1;
                svm_multi->votes[ib] -= 1;
        }
        /* class B is greater */
        else if ( yb > ya)
        {
                svm_multi->component_decisions[i] = ib;
        svm_multi->votes[ia] -= 1;
                svm_multi->votes[ib] += 1;
        }
        /* outputs are equal, no class gets a vote */
        else
        {
        }

       }
    /* find maximum votes, and max of min outputs */
        nmax = -MAXINT;
        imax = -1;
        y_cmin = -1.0*MAXFLOAT;
        i_cmin = -1;
        for(i=0; i<svm_multi->noutputs; i++)
        {
            if ( svm_multi->class_min[i] > y_cmin)
        {
                y_cmin = svm_multi->class_min[i];
                i_cmin = i;
                }
            if ( svm_multi->votes[i] > nmax)
        {
                nmax = svm_multi->votes[i];
                imax = i;
        }
        }
         /* all minimum class outputs underflow 
        should pick class with highest prior prob, but instead
        default to class 0 */
     if ( i_cmin == -1 )
        {
        i_cmin = 0;
        }
     /* no class received any votes */
         if ( imax == -MAXINT )
        {
        warning("no class has the largest vote");
        y_cmin = -1.0*MAXFLOAT + 1e-12;
        i_cmin = 0;
        nmax = svm_multi->noutputs-1;
        imax = 0;
        }
         /* break ties only if one class doesn't dominate */
        if ( nmax < svm_multi->noutputs-1)
           {
       for(i=0; i<svm_multi->noutputs; i++)
            {
        /* pairwise comparison of two tied classes */
        if ( svm_multi->votes[i] == nmax)
            {
            if ( svm_multi->class_min[i] > svm_multi->class_min[imax])
                 {
#ifdef debug
                printf("breaking tie nmax %d i old %d new %d min old %f new %f\n", 
                         nmax, imax, i, 
                         svm_multi->class_min[imax], svm_multi->class_min[i]);
#endif
             imax = i;
             }
            }
                 }
        }
#ifdef debug
        if ( imax != i_cmin )
        {
        warning("vote selection not equal to min out selection");
                printf("votes %d class %d  ymaxmin %7.5f class %d\n", 
                nmax, imax, y_cmin, i_cmin);
        }
#endif
        for(i=0; i<svm_multi->noutputs; i++)
        {
        svm_multi->outputs[i] = svm_multi->class_min[i];
        }
        /*
        svm_print_multi( model, 1);
        */
    }

    /* adjust priors only if sigmoid fit to posterior */
    if(svm_multi->adjust_priors && svm_multi->sigmoid_fit)
        {
        /* adjust priors */
        for(i = 0; i < svm_multi->noutputs; i++)
            {
            svm_multi->outputs[i] *= svm_multi->priors_list[i];
            svm_multi->outputs[i] /= svm_multi->class_cnt[i]/svm_multi->npatterns;
        }
    /* scale adjusted outputs to sum to 1 */
    total = 0.0;
    for(i = 0; i < svm_multi->noutputs; i++) 
            total += svm_multi->outputs[i];
    for(i = 0; i < svm_multi->noutputs; i++) 
            svm_multi->outputs[i] /= total;
    }
    
    i = max_pick(svm_multi->outputs, svm_multi->noutputs);
    return  i ;
    }

/*****************************************************************************
*       Evaluate two-class outputs and store in output array after
*       training is complete. 
******************************************************************************/
void svm_eval(int model, float *inputs)
    {
    SVM_PARAMS *svm;

    svm = svmpp[model];
    
    /* Raw outputs when not using sigmoid. 
       Class B output is negative class A
       Class A if out > 0  */
    if ( svm->sigmoid_fit == 0) 
    {
    (svm->outputs[0]) = svm_compute_output( svm, inputs); 
    (svm->outputs[1]) = -1.0*svm->outputs[0]; 
    }
    /* Use sigmoid to approximate posterior prob.
       Class B output is 1-class A
       Class A if out > 0.5  */
    else
    {
    (svm->outputs[0]) = 1.0/(1.0+exp((svm->posterior_aexp)*svm_compute_output(svm, inputs)));
    (svm->outputs[1]) = 1.0 - svm->outputs[0]; 
    }
    /*
    printf("svm_eval model %d inputs %f %f outputs %f %f\n", 
            model, inputs[0], inputs[1], svm->outputs[0], svm->outputs[1]);
    */
    }

/* initialize multi model classifier for testing */
void
svm_test_init_multi(int model, int adjust_priors, 
                    float *priors_list, int sigmoid_fit)
{
    int i, j;
    SVM_MULTI *svm_multi;
    SVM_PARAMS *svm;

    svm_multi = svmpp_multi[model];
    svm_multi->sigmoid_fit = sigmoid_fit;
    svm_multi->adjust_priors = adjust_priors;
    if(adjust_priors )
        {
        for(i = 0; i < svm_multi->noutputs; i++)
            svm_multi->priors_list[i] = priors_list[i];
        }
    else
    {
        for(i = 0; i < svm_multi->noutputs; i++)
            svm_multi->priors_list[i] = 1.0/(svm_multi->noutputs);
    }

    /* force component classifiers to use or not use sigmoid */
    for(i=0; i< svm_multi->n; i++)
    {
        j = svm_multi->component_model[i];
        svm = svmpp[j];
        svm->sigmoid_fit = sigmoid_fit;
        svm->total_kernel_evaluations=0;
    }
}
/* return number of support vectors
-------------------------------------------------------------*/
int svm_get_nsv(int model)
    {
    SVM_PARAMS *svm;
    svm = svmpp[model];
    return svm->sv_n;
    }
/* return alpha_type flag on training pattern 
-------------------------------------------------------------*/
int svm_get_alpha_type(int model, int i)
    {
    SVM_PARAMS *svm;
    svm = svmpp[model];
    if ( i<0 || i>svm->npatterns)
    fatal_error("get_alpha_type: bad i = %d should be >0 and < %d",
                     i, svm->sv_n);
    return (int)(svm->alpha_type[i]);
    }
/* return the number of positive below bound support vectors
-------------------------------------------------------------*/
int svm_get_nsv_pos_below_bound(int model)
    {
    SVM_PARAMS *svm;
    svm = svmpp[model];
    return svm->n_positive_below_bound;
    }
/* return the number of at bound support vectors
-------------------------------------------------------------*/
int svm_get_nsv_at_bound(int model)
    {
    SVM_PARAMS *svm;
    svm = svmpp[model];
    return svm->n_at_cbound;
    }
/* Return the jth feature of the ith below bound support vector.
   Have to pull these out from the patterns with positive alphas.
   Good only after training is complete and support_vector
   matrix is filled in.
---------------------------------------------------------------*/
float svm_get_sv_pos_below_bound(int model, int idesired, int jdesired)
    {
    int i,ibelow;
    float y;
        SVM_PARAMS *svm;

        svm = svmpp[model];
    ibelow = -1;
    /* find the ith support vector not at bound */
    for(i=0; i<svm->sv_n; i++)
          {
          if ( (svm->sv_alpha_type[i]) == UNBOUNDED )
            {
            ibelow++;
            }
          if ( idesired == ibelow)
            {
            y = svm->support_vectors[i][jdesired];
            break;
            }
        }
    return(y);
    }
/* Return the jth feature of the ith at bound support vector.
   Have to pull these out from the patterns with positive alphas.
   Good only after training when the support_vector matrix is
   filled in.
---------------------------------------------------------------*/
float svm_get_sv_at_bound(int model, int idesired, int jdesired)
    {
    int i,ibelow;
    float y;
        SVM_PARAMS *svm;

        svm = svmpp[model];
    ibelow = -1;
    /* find the ith support vector at bound */
    for(i=0; i<svm->sv_n; i++)
          {
          if ( (svm->sv_alpha_type[i]) == CBOUND )
            {
            ibelow++;
            }
          if ( idesired == ibelow)
            {
            y = svm->support_vectors[i][jdesired];
            break;
            }
        }
    return(y);
    }


/***--------------------------------------------------------------------------
* train multiple binary support vector machines and store to param file
* cv_flag determines whether this is used as part of cross-validation testing
* cv_flag = 0 build all, train, write to parameter file, free all 
* cv_flag = 1 build all, train, free train but keep test data
--------------------------------------------------------------------------**/
#define OTHERS (-1)

int
svm_train_multi(int nraw, int ninputs, int noutputs,
        int normalize, int npatterns, int verbosity, 
                char *data_path, char *data_fname, char *norm_fname, 
                char *exp_path, char *param_fname, char *comment, char **labels, 
                float cbound, int kernel_type, float kernel_sigma,
                int kernel_order, float kernel_poly_divisor,
                int sigmoid_fit,
                float kkt_tolerance, float alpha_tolerance,
                int multi_class_type,
                int data_model, int param_file_flag, FILE *param_file, 
                char *tmp_path, int cv_flag, int cv_fold)
{
    int     model, multi_model; 
    int     desired_class;
    int     n_read, n_used, i_orig, i_used,i, na, nb;
    float   *inputs;
    char    *run_name;
    int         multi_class_n_train, multi_class_n_build; 
    int     multi_class_i, multi_class_npatterns; 
    int     first_class, second_class;
    char    *first_class_name, *second_class_name;
    char    *first_class_number_name, *second_class_number_name;
    char    *multi_labels[3];
    int     max_length;
    int     total_nsv, total_nsv_pos_below_bound, total_nsv_at_bound;
    int     unique_nsv;
    int     *train_pattern_is_a_sv;
    int     *two_class_to_orig;


    /* keep track of which patterns are used as sv's */
    train_pattern_is_a_sv = allocate_ivector(npatterns);
    two_class_to_orig = allocate_ivector(npatterns);
    for(i=0; i<npatterns; i++)
        train_pattern_is_a_sv[i]=0;

    /* determine maximum length for class labels and allocate space 
       for run name and labels */
    max_length =10;
    for(i=0; i<noutputs; i++)
    {
    if ( strlen(labels[i+1]) > max_length)
        max_length = strlen(labels[i+1]);
    }

    max_length += 4;
    run_name = nn_malloc(max_length*2+60,"for storing run name");
    multi_labels[0] = nn_malloc(max_length,"for multi labels");
    multi_labels[1] = nn_malloc(max_length,"for multi labels");
    multi_labels[2] = nn_malloc(max_length,"for multi labels");
    first_class_name = nn_malloc(max_length,"for first class name");
    second_class_name = nn_malloc(max_length,"for second class name");
    first_class_number_name = nn_malloc(8,"for first class name");
    second_class_number_name = nn_malloc(8,"for second class name");

    /* determine the number of binary classifiers required */
    /* one versus all others */
    if ( multi_class_type == 0)
    {
    multi_class_n_train = noutputs;
    multi_class_n_build = multi_class_n_train;
    first_class = noutputs-1;
    second_class = OTHERS;
    }
    /* all pairs two at a time */
    else
    {
    multi_class_n_train = (noutputs*(noutputs-1))/2;
    multi_class_n_build = multi_class_n_train;
        first_class = noutputs-1;
        second_class = noutputs-2;
    }

    /* with two classifiers only build one */
    if ( noutputs == 2)
        {
        multi_class_n_train = 1;
        multi_class_n_build = 2;
        first_class = 1;
        second_class = 0;
        }

    multi_model = svm_build_multi(ninputs, noutputs, multi_class_type, 
                  multi_class_n_build, multi_class_n_train);

    svmpp_multi[multi_model]->npatterns = npatterns;
    /* store multi model information at top of param file */
    if ( param_file_flag )
    {
        if (cv_flag)
        fprintf(param_file, "multi model information cv fold %d\n", cv_fold);
    else
        fprintf(param_file, "multi model information\n");
    svm_write_model_multi(multi_model, param_file);
    }
    total_nsv=0;
    total_nsv_pos_below_bound=0;
    total_nsv_at_bound=0;

    /* ---------------------------------------------------------------
    create all the binary classifiers required
    ----------------------------------------------------------------*/
    for(multi_class_i=0; multi_class_i < multi_class_n_train; multi_class_i++)
    {
        /* determine classes a and b and number of patterns used*/
        if ( multi_class_type == 0)
        {
            if ( multi_class_i > 0)
        {
        first_class -= 1;
        second_class = OTHERS;
        }
            multi_class_npatterns = npatterns;
        svm_set_class_cnt_multi(multi_model, (float)multi_class_npatterns, first_class);
        }
        /* all pairs two at a time */
        else
        {
            if ( multi_class_i > 0)
        {
        second_class -= 1;
        if ( second_class < 0)
             {
                 first_class -= 1;
             second_class = first_class -1;
             }
        }
            na = class_data_get_npatterns_class(data_model, first_class);
        nb = class_data_get_npatterns_class(data_model, second_class);
        multi_class_npatterns = na + nb; 
        svm_set_class_cnt_multi(multi_model, (float)na, first_class);
        svm_set_class_cnt_multi(multi_model, (float)nb, second_class);
        }

        /* allocate and initialize model structure */
        model = svm_build(ninputs, 2, multi_class_npatterns);
        svm_init(model, cbound, kernel_type, kernel_sigma,
                 kernel_order, kernel_poly_divisor,
                 sigmoid_fit,
                 kkt_tolerance, alpha_tolerance, verbosity,
                 multi_class_type, first_class, second_class, multi_class_i );
    svm_set_component_model_multi(multi_model, model, multi_class_i);


    /* form class names */
        snprintf(first_class_name, max_length,"%s",labels[first_class+1]);
        snprintf(first_class_number_name, 8,"%d",first_class);
        if (second_class < 0)
         {
             snprintf(second_class_name, max_length,"others");
             snprintf(second_class_number_name, 8,"others");
         }
    else
         {
             snprintf(second_class_name, max_length,"%s",labels[second_class+1]);
             snprintf(second_class_number_name, 8,"%d",second_class);
             }
        
        /* pass exemplars to one binary classifier */
        n_used = 0;
        class_data_epoch_init(data_model);
        for(n_read = 0; n_read < npatterns; n_read++)
        {
        if((inputs = class_data_get_next(data_model, &desired_class)) == NULL) 
        break;
            /* print out data for cv folds */
            /*
            if ( multi_class_i == 0)
        {
            printf("DATA %d %d %f %f\n", n_read, 
            desired_class, inputs[0], inputs[1]); 
        }
            */
            /* turn this into a two class problem and only allow desired
               patterns through */
        /* one versus others */
            if ( multi_class_type == 0) 
        {
                if ( desired_class == first_class)
                        {
            desired_class = 0;
            }
                else
            {
            desired_class = 1;
            }
            svm_store_inputs(model, inputs, desired_class, n_used);
                two_class_to_orig[n_used] = n_read;
        n_used++;
            }
            /* all combinations */
        else
        {
        if ( (desired_class == first_class) || (desired_class == second_class))
            {
            if ( desired_class == first_class)
                {
                desired_class = 0;
                }
            else
                {
                desired_class = 1;
                }
            svm_store_inputs(model, inputs, desired_class, n_used);
                        two_class_to_orig[n_used] = n_read;
            n_used++;
            }
        }
        }
        /* train the binary classifier */
        snprintf(run_name, max_length*2+60,"class %s vs %s    %s vs %s", 
                 first_class_number_name, second_class_number_name, 
                 first_class_name, second_class_name);
        snprintf(multi_labels[0], 4,"2");
        snprintf(multi_labels[1], max_length, "%s", first_class_name);
        snprintf(multi_labels[2], max_length, "%s", second_class_name);

        if ( verbosity >= 1)
        printf("%s\n", run_name);
        svm_run_training(model, verbosity, run_name, multi_labels);

    /* update unique support vectors */
        for(i_used=0; i_used<n_used; i_used++)
        {
                i_orig = two_class_to_orig[i_used];
                /* training pattern is a support vector */
                if ( svm_get_alpha_type(model, i_used) > 0 )
            {
                        /* unique support vector */
                        if ( train_pattern_is_a_sv[i_orig] == 0)
                {
                                svm_set_sv_unique(model, i_used);
                                }
                train_pattern_is_a_sv[i_orig] += svm_get_alpha_type(model, i_used);
            }
        }
        svm_free_train(model);

        /* store component model in param file */
        if ( param_file_flag )
        {
        if (cv_flag)
        fprintf(param_file, "svm model cv fold %d\n", cv_fold);
        else
        fprintf(param_file, "svm model\n");
        svm_write_model(model, param_file);
        if ( verbosity > 1)
            printf("component model saved in \"%s\"\n\n", tmp_path);
        }
        total_nsv += svm_get_nsv(model);
        total_nsv_pos_below_bound += svm_get_nsv_pos_below_bound(model);
        total_nsv_at_bound += svm_get_nsv_at_bound(model);
        /* free space */
    if ( cv_flag == 0)
        svm_free(model);


    }
        unique_nsv = 0;
        for(i=0; i<npatterns; i++)
        {
                if ( train_pattern_is_a_sv[i] )
            unique_nsv++;
        }
    if ( verbosity >= 1)
        printf("All classifiers  unique support vectors %d  total %d  UnBound %d  AtUpper %d\n",
            unique_nsv, total_nsv, total_nsv_pos_below_bound, total_nsv_at_bound);
    svm_set_sv_multi(multi_model, total_nsv, total_nsv_pos_below_bound, total_nsv_at_bound,
                      unique_nsv);

    /* update support vector information by duplicating multi info at the end of
       the parameter file */
    if ( param_file_flag )
    svm_write_model_multi(multi_model, param_file);

    nn_free(run_name,"for storing run name");
    nn_free(multi_labels[0],"for storing run multi labels");
    nn_free(multi_labels[1],"for multi labels");
    nn_free(multi_labels[2],"for multi labels");
    nn_free(first_class_name,"for first class name");
    nn_free(second_class_name,"for second class name");
    nn_free(first_class_number_name,"for first class name");
    nn_free(second_class_number_name,"for second class name");
    deallocate_ivector(train_pattern_is_a_sv);
    deallocate_ivector(two_class_to_orig);
    if ( cv_flag == 0)
        {
    svm_free_multi_train(multi_model);
        multi_model = -1;
        }
    return(multi_model);
}
/*****************************************************************************
*       writes multi-classifier information to std out 
******************************************************************************/
void
svm_print_multi(int model, int vote_flag)
{
    SVM_PARAMS *svm;
    SVM_MULTI *svm_multi;
    int i, j;


    svm_multi = svmpp_multi[model];
    printf("multi params inputs %d  outputs %d  type %d  n %d\n",
        svm_multi->ninputs , svm_multi->noutputs,
        svm_multi->type , svm_multi->n );
   for(i=0; i< svm_multi->n; i++)
    {
        j = svm_multi->component_model[i];
        svm = svmpp[j];
        printf("%d model %d a %d  b %d  ya %7.5f  yb %7.5f  decide %d", 
                i, j, svm->multi_class_a, svm->multi_class_b,
                svm_multi->component_outputs[i][0],
                svm_multi->component_outputs[i][1],
                svm_multi->component_decisions[i]);
        /* print votes and min output */
        if ( i < svm_multi->noutputs && vote_flag == 1)
            {
            printf("     class %d  votes %d  min_out %7.5f\n",
                    i, svm_multi->votes[i], svm_multi->class_min[i]); 
        }
    else
        printf("\n");
    }
}
/************************************************************************
*   set unique flag for one support vector in a classifier 
************************************************************************/
void
svm_set_sv_unique(int model, int i)
{
    SVM_PARAMS *svm;
    int j;

    svm = svmpp[model];
    j =  svm_fpattern2sv(svm, i);
    if ( j<0 || j>svm->sv_n)
        fatal_error("set_sv_unique i was %d led to bad sv index %d should be >0 and <%d",
                    i, j, svm->sv_n);
                    
    svm->sv_unique[j] = 1;
}
/************************************************************************
*   set support vector counts for multi model 
************************************************************************/
void
svm_set_sv_multi(int model, int sv_n, int n_positive_below_bound, 
                       int n_at_cbound, int sv_unique)

{
    SVM_MULTI *svm_multi;

    svm_multi = svmpp_multi[model];

    svm_multi->sv_n = sv_n;
    svm_multi->n_positive_below_bound = n_positive_below_bound;
    svm_multi->n_at_cbound = n_at_cbound;
    svm_multi->sv_unique = sv_unique;
    
}
/************************************************************************
*   get unique flag for one support vector in a classifier 
************************************************************************/
int
svm_get_sv_unique(int model, int i)
{
    SVM_PARAMS *svm;

    svm = svmpp[model];

    return (svm->sv_unique[i]);
}

/************************************************************************
*       Returns a pointer to the named parameter for multi classifier.
*************************************************************************/
void *
svm_param_multi(int multi_model, char *param_name )
{
    void      *param_address;
    SVM_MULTI *svm_multi;

    svm_multi = svmpp_multi[multi_model];

    if (PEQ("ninputs"))
    param_address = &svm_multi->ninputs;
    else if (PEQ("noutputs") || PEQ("nclasses"))
    param_address = &svm_multi->noutputs;
    else if (PEQ("type") )
    param_address = &svm_multi->type;
    else if (PEQ("n") )
    param_address = &svm_multi->n;
    else if (PEQ("outputs"))
    param_address = &svm_multi->outputs;
    else if (PEQ("class_cnt"))
      param_address = &svm_multi->class_cnt;
    else if (PEQ("component_model"))
      param_address = &svm_multi->component_model;
    else if (PEQ("sv_n"))
      param_address = &svm_multi->sv_n;
    else if (PEQ("n_positive_below_bound"))
      param_address = &svm_multi->n_positive_below_bound;
    else if (PEQ("n_at_cbound"))
      param_address = &svm_multi->n_at_cbound;
    else if (PEQ("sv_unique"))
      param_address = &svm_multi->sv_unique;
    else if (PEQ("total_npatterns"))
      param_address = &svm_multi->npatterns;
    else {
    fatal_error("svm_param_multi: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}

/************************************************************************
*       Returns a pointer to the named parameter for multi classifier.
*************************************************************************/

void *
svm_param(int model, char *param_name )
{
    void      *param_address;
    SVM_PARAMS *svm;

    svm = svmpp[model];

    if (PEQ("ninputs"))
    param_address = &svm->ninputs;
    else if (PEQ("noutputs") || PEQ("nclasses"))
    param_address = &svm->noutputs;
    else if (PEQ("kernel_type") )
    param_address = &svm->kernel_type;
    else if (PEQ("gaussian_sigma") )
    param_address = &svm->gaussian_sigma;
    else if (PEQ("polynomial_order"))
    param_address = &svm->polynomial_order;
    else if (PEQ("sigmoid_fit"))
      param_address = &svm->sigmoid_fit;
    else if (PEQ("multi_class_a"))
      param_address = &svm->multi_class_a;
    else if (PEQ("multi_class_b"))
      param_address = &svm->multi_class_b;
    else if (PEQ("sv_n"))
      param_address = &svm->sv_n;
    else if (PEQ("n_positive_below_bound"))
      param_address = &svm->n_positive_below_bound;
    else if (PEQ("n_at_cbound"))
      param_address = &svm->n_at_cbound;
    else if (PEQ("sv_unique"))
      param_address = &svm->sv_unique;
    else if (PEQ("threshold"))
      param_address = &svm->threshold;
    else if (PEQ("weight"))
      param_address = &svm->weight;
    else if (PEQ("cbound"))
      param_address = &svm->cbound;
    else if (PEQ("sv_alpha"))
      param_address = &svm->sv_alpha;
    else if (PEQ("sv_y"))
      param_address = &svm->sv_y;
    else if (PEQ("support_vectors"))
      param_address = &svm->support_vectors;
    else if (PEQ("posterior_aexp"))
      param_address = &svm->posterior_aexp;
    else if (PEQ("polynomial_divisor"))
      param_address = &svm->polynomial_divisor;
    else {
    fatal_error("svm_param: Parameter %s not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}
