/***---------------------------------------------------------------------------
* NAME:     vonmises
* PURPOSE:  estimate location and spread parameters for samples
*           from a von mises distribution
* DESCRIPTION:
*           The von mises pdf function is defined by 
*           p(theta) = C*exp(kappa*cos(theta-alpha))
*                       theta   = circular input variable ranging from 0 to 2pi
*                                 (like x, 0 is 3 o'clock, 
*                                  increases counter clockwise,)
*                       kappa   = spread parameter 
*                                 (like 1/variance)
*                       alpha   = location parameter ranging from 0 to 2pi
*                                 (like mean)
*                       C       = 1/[2pi*I0(kappa)], where I0(k) is the
*                                 modified Bessel function 
*
* USAGE: 
* RETURN VALUE:
*   reads in numbers like lipstat and prints out location and spread
*   parameter of the vonmises distribution.
*   all flags: mean sdev min max
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/datagen/vonmises.c,v 1.3 2001/08/23 20:13:01 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/datagen/vonmises.c,v $
* CREATED: 30-Jul-86 rpl                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/datagen/vonmises.c,v 1.3 2001/08/23 20:13:01 craig Exp $";

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <math.h>
#include "/u/kukolich/lnknet/src/include/nnlib.h"
#include "vonmises_lib.h"

static float circle_max=M_PI;
static int verbosity=1;
static char *fileName=NULL;

static ARG      argtab[]= 
{
    {"circle_max", FLOAT, (int *)&circle_max,   "max input value"   },
    {"verbosity", INT,      &verbosity,         "amount to print"   },
    {"file", STRING, (int *)&fileName, "filename" },
};
#define TABSIZE (sizeof(argtab)/sizeof(ARG))

main (argc, argv)
int argc;
char **argv;
{
    double xsum=0.0, ysum=0.0;
    double x, y;
    int n=0;
    float radian;
    int class;
    float kappa, alpha;
    FILE *fp;

    init_nnsubs(argc, argv, rcsid);
    argc = nnargs(argc, argv, argtab, TABSIZE);

    if(fileName == NULL) fp = stdin;
    else fp = fopen(fileName, "r");
    /* accumulate sums from all data */
    while (fscanf (fp, "%d %f", &class, &radian) == 2)
        {
        radian = 2.0*M_PI*radian/circle_max;
        x = cos(radian);
        y = sin(radian);
        xsum += x;
        ysum += y;
        n++;
        if ( verbosity >= 2)
            printf("%d %g %g\n", n, x, y); 
        }
    /* calculate mean and spread and print out results 
    ---------------------------------------------------*/
    if ( n )
        {
        alpha = alpha_vonmises(xsum, ysum);
        kappa = k_vonmises(xsum, ysum, n);
        printf("mean %f   k %f (%f)\n", alpha*circle_max/(2.0*M_PI),
	       kappa*2.0*M_PI/circle_max,
	       circle_max/(kappa*2.0*M_PI));
        printf("[mean %f   k %f (%f)]\n", alpha, kappa, 1./kappa);
        }
}
