                          /* bintree_pack.h */

/* DEFINITIONS AND DECLARATIONS FOR bintree_pack.c */

#ifndef bintree_pack_h
#define bintree_pack_h


typedef struct{
    int class;        /* pattern class (desired or correct class of pattern) */
    float *inputs;    /* pattern vector coordinates */
} PATTERN; /* pattern input and output variables */

/* DEFINITIONS OF INTERNAL STRUCTURES AND TYPES */
#ifdef bintree_pack_h_internal_types
#undef bintree_pack_h_internal_types
/* bintree_pack_h_internal_types is defined in `bintree_pack.c' */

/* Define structure of the bintree_model */
typedef struct {
    int     ninputs;	    /* dimension of each input vector */
    int     nclasses;	    /* number of output classes or nodes */
    int     verbosity;
    int    *features;       /* feature selections */
    float  *outputs;        /* to store outputs of a given patterns */ 

    int     linear;         /* to use linear combination, set linear = 1 */
    float   leaf_min_npatterns;/* do not split a node if npatterns <= this */

    int	    adjust_priors;  /* whether to adjust testing priors */
    float  *priors_list;    /* when adjusting priors in testing */ 

    float  *class_count;    /* number of samples seen in each class */
    int     npatterns;      /* number of total patterns */
    int     pattern_size;   /* allocated length of patterns vector */
} BINTREE_PARAMS;

struct bintree_node_struct{
    int motherno; 
    int selfno;		/* address ID numbers */ 
    int yesno;
    int nono;

    int npatterns;	/* number of patterns in the node */
    PATTERN **patterns;		/* pointers to pattern vectors of the node */

    int class;		/* class type of node highest class_distribution */
    int level;		/* level of node: root node has level 0 */
    int nnodes;		/* number of nodes under this node */
    int nleaves;	/* number of leaves of under this node */

    float max_distrib;		/* highest class_distribution rate */
    float *class_distribs;	/* class distribution rate */
    int split_input;	/* if binary this gives the split coordinate */
    float *beta;		/* linear combination parameters */

    int prune;		/* to prune tree under node, set prune = 1 */
    int topflag;	/* to define node as top of subtree set topflag */
    int visit;		/* for traveling along the tree */  
    int yesvisit;
    int novisit;

    float node_misclass;	/* node misclassification rate */
    float sub_misclass;		/* node sub misclassification rate */
    float split_constant;	/* used for linear combination */
    float probability;
    float strength;		/* the strength of a splitting */

    int x;	/* added for plot_bintree for plot of tree structure */

    /* pointers to the following yes and no nodes and to the mother node */
    struct bintree_node_struct *yeschild;
    struct bintree_node_struct *nochild;
    struct bintree_node_struct *mother;
};
typedef struct bintree_node_struct BTNODE;
typedef BTNODE *BTLINK;

typedef struct{
    int     *leaves_index; /* queue of leaves waiting for splitting */
    int      leaves_size;  /* size of leaves_index array */
    int      nleaves;      /* number of leaves in growing queue OR
			      number of leaves in the tree */

    int      nlevels;     /* total number of levels in the tree */
    int      nnodes;      /* number of splitting nodes in the tree */
    int      nbins;       /* total  number of nodes (nnodes + nleaves)*/

    BTLINK  *bins;        /* array to store the addresses of nodes */
    int     *bin_index;   /* index to nodes array will be sorted for pruning */
    int      bins_size;   /* size of bins and bins_index arrays */

    float    tree_error; /* error rate of the tree
			    sum over all leaves of misclass*probability */
    /* temporary vectors for growing */
    int	*classes;
    float *search_array;
    int *yes_class_count;
    int *no_class_count;
    /* temporaries for finding linear combinations for split values */
    int *linear_classes;
    float *linear_search_array;
    float *linear_beta;
    float linear_split_constant;
    /* temporaries for normalizing data for linear combo searches */
    float **linear_node_data;
    float *median;
    float *range;
} BTBODY;

typedef struct{
    BTBODY           body;     /* for size of tree and addresseses of nodes */
    BTNODE           root;     /* root node */
    PATTERN         *patterns; /* array to store the patterns */
    BINTREE_PARAMS   params;           /* model parameters */
} BTREE;
 
typedef BTREE *BTREEP;

/* DECLERATION OF INTERNAL FUNCTIONS */

#ifdef bintree_pack_h_internal_functions
#undef bintree_pack_h_internal_functions
/* bintree_pack_h_internal_functions is defined in `bintree_pack.c' */
static void bintree_free_node (BTLINK node);
static void bintree_free_tree (BTREEP tree);
static void bintree_allocate_tree_patterns (BTREEP tree);
static void bintree_initialize_node (BTREEP tree, BTLINK node);
static void bintree_add_new_node (BTREEP tree, BTLINK node, char *child_kind);
static void bintree_set_node_class (BTREEP tree, BTLINK node);
static int bintree_grow_node (BTREEP tree, BTLINK node);
static BTLINK bintree_get_split_node (BTREEP tree);
static int bintree_need_to_split (BTREEP tree, BTLINK node);
static int bintree_split_node (BTREEP tree, BTLINK node);
static int bintree_set_univariate_split (BTREEP tree, BTLINK node);
static void bintree_copy_and_normalize_data (BTREEP tree, BTLINK node);
static int bintree_find_best_split (int npatterns, int nclasses, int *class,
				    float *max_phi, float *split_values,
				    BTREEP tree);
static float bintree_split_value (float *split_values, int npatterns,
				  int patternnum);
static void bintree_set_linear_combination_parameters (BTREEP tree,
						       BTLINK node);
static void bintree_set_linear_combination_values (BTREEP tree, BTLINK node);
static float bintree_find_best_delta_and_gamma (int coordinate,
						float *best_delta,
						float *best_gamma,
						BTREEP tree, BTLINK node);
static BTLINK bintree_make_child (BTREEP tree, BTLINK node, char *child_kind);
static int bintree_bad_division (BTREEP tree, BTLINK node);
static void bintree_divide_patterns (BTREEP tree, BTLINK node);
static void bintree_update_tree_error (BTREEP tree, BTLINK node);
static void bintree_add_new_leaf (BTREEP tree, BTLINK node);
static void bintree_calculate_nleaves_and_strength (BTREEP tree);
static void bintree_calculate_tree_error (BTREEP tree);
static int bintree_find_weakest_node (BTREEP tree);
static int bintree_prune_children (BTREEP tree, BTLINK node, int prune_num);
static void bintree_link_tree (BTREEP tree);
static int bintree_set_pruning (BTREEP tree, int max_nodes);
static BTLINK bintree_find_pattern_node (BTREEP tree, float *inputs,
					 float *outputs, int *output_class);
static float bintree_node_linear_combination_value (BTREEP tree, BTLINK node,
						    float *inputs);
static int bintree_write_model_params (BTREEP tree, FILE *fp);
static int bintree_write_tree_model (BTREEP tree, FILE *fp);
static int bintree_write_tree_params (BTREEP tree, FILE *fp);
static int bintree_write_node_params (BTREEP tree, FILE *fp);
static int bintree_write_pruning_order (BTREEP tree, FILE *fp);
static int bintree_read_model_params (FILE *fp);
static int bintree_read_tree_model (BTREEP tree, FILE *fp);
static int bintree_read_tree_params (BTREEP tree, FILE *fp);
static int bintree_read_node_params (BTREEP tree, FILE *fp);
static int bintree_read_pruning_order (BTREEP tree, FILE *fp);
static void bintree_make_bins (BTREEP tree);
static void bintree_initialize_read_node (BTREEP tree, BTLINK node);
static void bintree_set_node_addresses (BTREEP tree, BTLINK node);
static void bintree_set_model_scalar_params (BTREEP tree);
static void bintree_set_model_array_params (BTREEP tree);
static void bintree_set_tree_params (BTREEP tree);
static void bintree_set_node_params (BTREEP tree, BTLINK node);
static void bintree_set_pruning_pointer (BTREEP tree);
static void bintree_display_node_head (BTREEP tree, int grow);
static void bintree_display_node (BTREEP tree, BTLINK node, int grow);

#endif /* INTERNAL FUNCTIONS */
#endif /* INTERNAL TYPES */



/* INTERNAL MACRO DEFINITIONS */

#ifdef bintree_pack_h_internal_macros
#undef bintree_pack_h_internal_macros
/* bintree_pack_h_internal_macros is defined in `bintree_pack.c' */

/* macro function definitions */
#define ROOT_NODE(node)  ((node)->mother == NULL)
#define TRMNL_NODE(node) ((node)->yeschild == NULL && (node)->nochild == NULL)
#define WHICH_CHILD(str) (!strcmp((str), "yes") + 2 * !strcmp((str), "no") + 4*!strcmp((str),"root"))
#define VISITED_NODE(node) VSTND(node) /* to fit the definition on one line */
#define VSTND(node) (((node)->yesvisit && (node)->novisit) || TRMNL_NODE(node))
#define RESET_NODE(node)   RSTND(node)
#define RSTND(node)  (((node)->yesvisit || node->novisit) && !TRMNL_NODE(node))
#define PRUNED_NODE(node)  ((node)->prune || TRMNL_NODE(node))
#define PRUNE_SET(node) (node)->prune
#define TOP_NODE(node)  ((node)->topflag || ROOT_NODE(node))

/* macro constant definitions */
#define YES 1
#define NO 2
#define ROOT 4
#define EOTREE -1  /* End of Tree */
#define MEMORY_BLOCK 128

#endif  /* INTERNAL MACROS */


/* DECLERATION OF EXTERNAL FUNCTIONS */

int bintree_build (int ninputs, int nclasses, int verbosity);
void bintree_init (int model, int linear, int full_tree,
		   int leaf_min_npatterns, int *features);
void bintree_free (int model);
void bintree_store_pattern (int model, int desired_class, float *inputs);
void bintree_init_tree (int model);
int bintree_grow_tree (int model);
void bintree_prune (int model);
int bintree_test_init (int model, int full_tree, int max_nodes,
		       int adjust_priors, float *priors_list);
int bintree_classify (int model, float *inputs);
void bintree_eval (int model, float *inputs, float *outputs);
int bintree_write_model (int model, FILE *fp);
int bintree_read_model (FILE *fp);
void *bintree_param (int model, char *param_name);
void bintree_reset_visit_flags (/*BTLINK node*/);

#endif /* EXTERNAL (FILE bintree_pack.h)  */
