/***-------------------------------------------------------------------------
 * NAME:        allocate.c 
 * PURPOSE: allocate/deallocate/reallocate utility subroutines 
 *
 * DESCRIPTION:
 *   char *nn_malloc(size, for_message)
 *  unsigned size; char *for_message
 *  - allocate size bytes in memory with error checking.
 *      uses for_message in fatal error message
 *
 *   void nn_free(p, for_message)
 *  char *p; char *for_message;
 *  - free memory allocated with nn_malloc() or nn_realloc(),
 *      do error checking and use for_message in fatal error
 *
 *   char *nn_realloc(p, size, for_message)
 *  char *p; unsigned size; char *for_message;
 *  - allocate size byte in memory with error checking,
 *      copy data stored at p to new memory block,
 *      in case of error, use for_message in fatal error message
 *
 *   floats:
 *   float *allocate_vector(n)
 *  int n;
 *  - allocate space for an array of n floats, initialize the entries to
 *     0.0 and return a pointer to it
 *      v = allocate_vector(n);
 *      v[0]..v[n-1] = 0.0;
 *      fatal error if unsuccessful
 *
 *   float *reallocate_vector(v, n, newn)
 *  float *v; int n, newn;
 *  - change the length of the floating point array v (previously
 *      allocated using allocate_vector) from n to newn, copy the data
 *      from v to the new array, initialize the new entries to 0.0,
 *       and return a pointer to the new array
 *      p = reallocate_vector(v, n, newn);
 *      p[0]..p[n-1] = v[0]..v[n-1];
 *      p[n]..p[newn-1] = 0.0;
 *
 *   void deallocate_vector(v)
 *  float *v;
 *  - deallocate space for a floating point array v allocated by
 *      allocate_vector()
 *
 *   float **allocate_matrix(nrow, ncol)
 *  int nrow, ncol;
 *  - allocate space for a floating point matrix nrow by ncol using
 *      two calls to malloc, initialize the matrix data to 0.0,
 *      initialize the pointers to the matrix rows, and return a pointer
 *      to the vector of matrix row pointers
 *  - a contiguous block of memory nrow*ncol*sizeof(float) is allocated
 *      data stored in it are stored sequentially as if they were
 *      stored using a C declared matrix with constant row and
 *      column sizes
 *  - a vector of nrow pointers to floating point arrays is allocated and
 *      filled with the addresses of the rows in the matrix block.
 *  float **m, *v;
 *  m = allocate_matrix(nrow, ncol);
 *  v = m[i]; (0 <= i < nrow.  0 <= j < ncol)
 *  (v[j] == m[i][j])
 *  (&m[i][j] == &m[i][j-1] + sizeof(float))
 *  (&m[i][j] == &m[i-1][j] + nrow*sizeof(float))
 *
 *   float **reallocate_matrix(m, nrow, ncol, newrow)
 *  float **m; int nrow, ncol, newrow;
 *  - change the number of rows in a 2D float matrix allocated using
 *      allocate_matrix() using two calls to realloc, copy the data
 *      from m to the new matrix, initialize the new entries to 0.0,
 *      initialize the pointers to the matrix rows and return a
 *      pointer to the vector of matrix row pointers
 *  - a contiguous block of memory newrow*ncol*sizeof(float) bytes long is
 *      allocated
 *  - another contiguous block of memory newrow*sizeof(float *) is
 *      allocated
 *  - the data is initialized as follows:
 *      m2 = reallocate_matrix(m, nrow, ncol, newrow);
 *      m2[0][0]..m2[nrow-1][ncol-1] = m[0][0]..m[nrow-1][ncol-1];
 *      m2[nrow][0]..m2[newrow-1][ncol-1] = 0.0;
 *  - the new matrix has the same properties as the original, namely
 *      v = m2[i]; (0 <= i < newrow; 0 <= j < ncol)
 *      (v[j] == m2[i][j])
 *      (&m2[i][j] == &m2[i][j-1] + sizeof(float))
 *      (&m2[i][j] == &m2[i-1][j] + newrow*sizeof(float))
 *  - if m was set by a previous call to reallocate_matrix_noncontig(),
 *      a fatal error will result
 *
 *   float **reallocate_matrix_noncontig(m, nrow, ncol, newrow)
 *  float **m; int nrow, ncol, newrow;
 *  - increase the number of rows in a 2D float matrix allocated using
 *      allocate_matrix(), initializing the added rows to 0.0
 *      This routine is similar to reallocate_matrix().  The difference
 *      being that the space for the new data is allocated as a new
 *      data block, not attached to the rest of the matrix data.
 *      The old matrix data is not moved in memory, as it would be
 *      in reallocate_matrix().
 *      A new array of pointers (m[i]) is allocated, as before.
 *  - note: cannot be used to decrease the number of rows in a matrix,
 *      as reallocate_matrix can be.
 *
 *   void deallocate_matrix(m)
 *  float **m;
 *  - free space for a floating point matrix previously allocated by
 *      allocate_matrix().  Will free all blocks of data allocated
 *      during calls to reallocate_matrix_noncontig()
 *
 *   There are also routines for allocating vectors and matrices of doubles,
 *   integers, pointers, and structs.  Each works the same as its
 *   corresponding floating point version.
 *   The routine names are listed below.
 *
 *   doubles:
 *   double *allocate_dvector(), *reallocate_dvector();
 *   double **allocate_dmatrix(), **reallocate_dmatrix(),
 *  **reallocate_dmatrix_noncontig();
 *   void deallocate_dvector(), deallocate_dmatrix();
 *
 *   ints:
 *   int *allocate_ivector(), *reallocate_ivector();
 *   int **allocate_imatrix(), **reallocate_imatrix(),
 *  **reallocate_imatrix_noncontig();
 *   void deallocate_ivector(), deallocate_imatrix();
 *
 *   pointers:
 *   void **allocate_pvector(), **reallocate_pvector();
 *   void ***allocate_pmatrix(), ***reallocate_pmatrix(),
 *  ***reallocate_pmatrix_noncontig();
 *   void deallocate_pvector(), deallocate_pmatrix();
 *
 *   structs:
 *   void *allocate_structvector(), *reallocate_structvector();
 *   void **allocate_structmatrix(), **reallocate_structmatrix();
 *  **reallocate_structmatrix_noncontig();
 *   void deallocate_structvector(), deallocate_structmatrix();
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/allocate.c,v 11.5 2001/08/23 20:14:42 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/allocate.c,v $
 * CREATED: 26-July-90 rpl                MIT LINCOLN LABORATORY
 ------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#if HAVE_MALLOC_H
#include <malloc.h>
#endif
/* this is a Sun only file and it compiles fine without it... */
/*#include <alloca.h>*/
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include "nnlib.h"

#define NONCONTIG_INC (10)

/***------------------------------------------------------------------------
 * NAME:        nn_malloc
 * PURPOSE: malloc that checks return and prints error message if problem
 * EXAMPLE:
 *    char *p;
 *    p=nn_malloc((unsigned)100,"for word names in subroutine add_wordname");
 * NOTE:
 *     The second argument is used in an error message if the malloc() fails.
 *     It should describe what the space is for and begin with "for ...".
 ------------------------------------------------------------------------***/
char *nn_malloc(unsigned int n, char *for_message/* what storage is for */)
{
    char *p;
    
    p = malloc((unsigned)n);
    if ( p == NULL)
      fatal_error( "nn_malloc: Can't allocate %u bytes %s", 
          n, for_message);
    return(p);
}

/***-------------------------------------------------------------------------
 * NAME:        nn_free
 * PURPOSE: free that checks return and prints error message if problem
 * EXAMPLE:
 *    char *p;
 *    p=nn_malloc((unsigned)100,"for word names in subroutine add_wordname");
 *    nn_free(p, "for word word names in subroutine add_wordname");
 * NOTE:
 *     The second argument is used in an error message if the free() fails.
 *     It should describe what the space was for and begin with "for ...".
 ------------------------------------------------------------------------***/
void nn_free(char *p, char *for_message/* what storage was for */)
{
    free(p);
}

/***-------------------------------------------------------------------------
 * NAME:   nn_realloc
 * PURPOSE:realloc that checks return and prints error message if problem
 * EXAMPLE:
 *  char *p;
 *  p=nn_realloc((unsigned)100, "for word names in subroutine add_wordname");
 * NOTE:
 *   The second argument is used in an error message if the realloc() fails.
 *   It should describe what the space is for and begin with "for ...".
 ------------------------------------------------------------------------***/
char *nn_realloc(char *ptr,/* from previous  malloc or realloc call */
         unsigned int n,/* number of bytes of storage desired */
         char *for_message/* storage is for "for ..." */)
{
    char *p;
    
    p = realloc(ptr, n);
    if ( p == NULL)
      fatal_error("nn_realloc: Can't reallocate %u bytes %s", 
          n, for_message);
    return(p);
}

/***------------------------------------------------------------------------
 * NAME:        nn_strdup
 * PURPOSE: allocate space for a copy of the given string, giving a
 *      fatal error if the allocation fails
 * EXAMPLE:
 *    char *p;
 *    p=nn_strdup("this string");
 ------------------------------------------------------------------------***/
char *nn_strdup(char *str)
{
    char *p;
    
#if HAVE_STRDUP
    p = strdup(str);
#else
    p = malloc(sizeof(str)+1);
    strcpy(p, str);
#endif
    if ( p == NULL)
      fatal_error( "nn_strdup: Can't allocate space for string %s", str);
    return(p);
}

/***-------------------------------------------------------------------------
 * NAME:        allocate_vector
 * PURPOSE: allocate space for a float vector with n elements
 * EXAMPLE:
 *  float *pfloat;
 *  pfloat = allocate_vector(10);
 ------------------------------------------------------------------------***/
float *allocate_vector(int n)
{
    float *v;
    
    v=(float *)calloc((unsigned)n, sizeof(float));
    if (!v) fatal_error( "%s %s %u elements", "allocate_vector:",
            "Can't allocate memory for float vector with", n);
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        reallocate_vector
 * PURPOSE: changes length of vector of floats previously
 *  allocated using allocate_vector
 * EXAMPLE:
 *  int long *pfloat;
 *  pfloat = reallocate_vector(pfloat, 10, 20);
 ------------------------------------------------------------------------***/
float *reallocate_vector(float *v, unsigned int n, unsigned int newn)
{
    unsigned i;
    
    v = (float *)realloc((char *)v, (unsigned)newn * sizeof(float));
    if (!v)
      fatal_error( "%s %s %u elements",
          "reallocate_vector:",
          "Can't allocate memory for float vector with", n);
    /* zero out new entries */
    for(i = n; i < newn; i++){ v[i] = 0.; }
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        deallocate_vector
 * PURPOSE: free up space previously allocated using allocate_vector()
 * EXAMPLE:
 *  float *pvector;
 *  pvector = allocate_vector(10);
 *  ...
 *  deallocate_vector(pvector);
 ------------------------------------------------------------------------***/
void deallocate_vector(float *v)
{
    free((char*) v);
}

/***------------------------------------------------------------------------
 * NAME:        allocate_matrix
 * PURPOSE: allocate space for a float 2D matrix 
 * EXAMPLE:
 *          float **pfloat;
 *          pfloat = allocate_matrix(10, 100);
 ------------------------------------------------------------------------***/
float **allocate_matrix(int nrow, int ncol)
{
    int i;
    float **row_pointers;
    float *data_pointer;
    int nitems;
    
    nitems = nrow*ncol;
    
    /* allocate array of pointers to point to beginning of rows 
       -----------------------------------------------------------*/
    if(nrow == 0)
      /* allocate at least one entry to hold matrix */
      row_pointers = (float **)malloc((unsigned)sizeof(float*));
    else
      row_pointers= (float **) malloc((unsigned)nrow*sizeof(float*));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_matrix:",
            "Can't allocate row pointer memory for float matrix",
            nrow, ncol, nitems);
    }
    
    /* allocate zeroed out contiguous memory for all data 
       ------------------------------------------------------*/
    data_pointer = (float *)calloc((unsigned)nitems,sizeof(float));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_matrix:",
            "Can't allocate data memory for float matrix",
            nrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space 
       -------------------------------------------------------------*/
    if(nrow == 0) row_pointers[0] = data_pointer;
    for(i=0; i<nrow; i++) {
    row_pointers[i]= data_pointer;
    data_pointer += ncol;
    }
    
    return row_pointers;
}

static float **noncontig_float = NULL;
static int *noncontig_float_sizes = NULL;
static int noncontig_float_length = 0;

/***-------------------------------------------------------------------------
 * NAME:        is_noncontig_float
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_float of entry which matches m
 *      if m is not in noncontig_float, returns -1
 ------------------------------------------------------------------------***/
static int is_noncontig_float(float **m)
{
    int i;
    for(i = 0; i < noncontig_float_length; i++)
      if((float *)m == noncontig_float[i]) return(i);
    return(-1);
}

/***-------------------------------------------------------------------------
 * NAME:        get_noncontig_float_entry
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_float of entry which is not
 *      currently in use.
 *      if noncontig_float is full, expands it to create unused entries
 ------------------------------------------------------------------------***/
static int get_noncontig_float_entry(void)
{
    int i, rv;
    for(i = 0; i < noncontig_float_length; i++)
      if(noncontig_float[i] == NULL) return(i);
    
    /* didn't find empty */
    if(i == 0){
    noncontig_float = (float **)allocate_pvector(NONCONTIG_INC);
    noncontig_float_sizes = (int *)allocate_ivector(NONCONTIG_INC);
    noncontig_float_length = NONCONTIG_INC;
    rv = 0;
    } else {
    noncontig_float = (float **)
      reallocate_pvector((void **)noncontig_float,noncontig_float_length,
                 noncontig_float_length + NONCONTIG_INC);
    noncontig_float_sizes = (int *)
      reallocate_ivector(noncontig_float_sizes, noncontig_float_length,
                 noncontig_float_length + NONCONTIG_INC);
    rv = noncontig_float_length;
    noncontig_float_length += NONCONTIG_INC;
    }
    return(rv);
}

/***-------------------------------------------------------------------------
 * NAME:         reallocate_matrix
 * PURPOSE:      change the number of rows in a float matrix previously
 *  allocated using allocate_matrix.
 *  warning: the data moves, so old pointers into the data become invalid
 * EXAMPLE:
 *  float **pfloat;
 *  pfloat = allocate_matrix(10, 100);
 *  ...
 *  pfloat = reallocate_matrix(pfloat, 10, 100, 20);
 ------------------------------------------------------------------------***/
float **reallocate_matrix(float **row_pointers, int nrow, int ncol, int newrow)
{
    int i,j;
    float *data_pointer;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    nitems = newrow*ncol;
    if(is_noncontig_float(row_pointers) >= 0)
      fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n%s",
          "reallocate_matrix:",
          "Can't allocate memory for float matrix",
          newrow, ncol, nitems, "Matrix is not contiguous");
    
    /* if this won't change the size, just return */
    if(nrow == newrow) return row_pointers;

    /* allocate contiguous memory for all data, copying current data
       ----------------------------------------------------------------*/
    data_pointer = (float *)realloc((char *)row_pointers[0],
                    (unsigned)nitems*sizeof(float));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_matrix:",
            "Can't allocate data memory for float matrix",
            newrow, ncol, nitems);
    }
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    if(newrow == 0)
      row_pointers= (float **) realloc((char *)row_pointers,
                       (unsigned) sizeof(float*));
    else
      row_pointers= (float **) realloc((char *)row_pointers,
                       (unsigned) newrow*sizeof(float*));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_matrix:",
            "Can't allocate row pointer memory for float matrix",
            newrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space
       -----------------------------------------------------------*/
    if(newrow == 0) row_pointers[0] = data_pointer;
    for(i=0; i<newrow; i++) {
    row_pointers[i]= data_pointer;
    data_pointer +=  ncol;
    }
    
    /* zero the space added to the data matrix
       -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
    for(j = 0; j < ncol; j++){
        row_pointers[i][j] = 0.0;
    }
    }
    return row_pointers;
}

/***-------------------------------------------------------------------------
 * NAME:         reallocate_matrix_noncontig()
 * PURPOSE:      increase the number of rows in a float matrix previously
 *  allocated using allocate_matrix, but add the new rows as a separate,
 *  non-contiguous block of memory
 * NOTES:
 *  reallocate_matrix() may not be called after reallocate_matrix_noncontig()
 *  reallocate_matrix_noncontig() will produce a warning if called to shrink
 *  a matrix.  The matrix will be untouched and no associated memory
 *  will be freed
 * EXAMPLE:
 *  float **pfloat;
 *  pfloat = allocate_matrix(10, 100);
 *  ...
 *  pfloat = reallocate_matrix_noncontig(pfloat, 10, 100, 20);
 ------------------------------------------------------------------------***/
float **reallocate_matrix_noncontig(float **row_pointers, int nrow,
                    int ncol, int newrow)
{
    int i;
    float *data_pointer, **new_row_ptrs;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    if(nrow > newrow){
    warning("%s %s\nNew rows %u, old rows %u, %u columns\n%s",
        "reallocate_matrix_noncontig:",
        "Can't decrease number of rows in a non-contiguous matrix.",
        newrow, nrow, ncol, "Will not alter matrix.");
    return(row_pointers);
    }
    if(nrow == newrow)
      return row_pointers;
    /* 0 <= nrow < newrow */

    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    new_row_ptrs =
      (float **) realloc((char *)row_pointers,
             (unsigned) newrow*sizeof(float*));
    if (!new_row_ptrs) {
    fatal_error( "%s %s\nwith %u rows, %u columns\n",
            "reallocate_matrix_noncontig:",
            "Can't allocate row pointer memory for float matrix",
            newrow, ncol);
    }
    if((i = is_noncontig_float(row_pointers)) < 0){ /* newly noncontig */
    /* put the row pointer into noncontig_float */
    i = get_noncontig_float_entry();
    noncontig_float[i] = (float *)new_row_ptrs;
    noncontig_float_sizes[i] = newrow;
    /* put the first block pointer into noncontig_float */
    if(nrow != 0){
        i = get_noncontig_float_entry();
        noncontig_float[i] = new_row_ptrs[0];
        noncontig_float_sizes[i] = 0;
    } else {/* if nrow is 0, we need to free the "empty" space */
        free((char *)new_row_ptrs[0]);
    }
    } else {
    noncontig_float[i] = (float *)new_row_ptrs;
    noncontig_float_sizes[i] = newrow;
    }
    
    /* allocate contiguous memory for extra data
       ---------------------------------------------*/
    nitems = (newrow - nrow)*ncol;
    
    data_pointer = (float *)calloc((unsigned)nitems, sizeof(float));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_matrix_noncontig:",
            "Can't allocate data memory for float matrix",
            newrow, ncol, nitems);
    }
    i = get_noncontig_float_entry();
    noncontig_float[i] = data_pointer;
    noncontig_float_sizes[i] = 0;

    /* make row pointers point to the right place in the data space
       ---------------------------------------------------------------*/
    for(i = nrow; i < newrow; i++) {
    new_row_ptrs[i] = data_pointer;
    data_pointer += ncol;
    }
    return new_row_ptrs;
}

/***------------------------------------------------------------------------
 * NAME:        deallocate_matrix
 * PURPOSE: free space previously allocated using allocate_matrix()
 *      for a float 2D matrix 
 * EXAMPLE:
 *  float **pmatrix;
 *  pmatrix = allocate_matrix(10, 100);
 *      ...
 *  deallocate_matrix(pmatrix);
 ------------------------------------------------------------------------***/
void deallocate_matrix(float **m)
{
    int i, j;
    int row_i;
    
    if ( m == NULL ) return;

    /* free up linear data memory 
       -----------------------------*/
    if((row_i = is_noncontig_float(m)) >= 0){
    /* free all noncontiguous blocks */
    for(i = 0; i < noncontig_float_sizes[row_i]; i++)
      for(j = 0; j < noncontig_float_length; j++)
        if(m[i] == noncontig_float[j]){
        free((char *)m[i]);
        noncontig_float[j] = NULL;
        }
    noncontig_float[row_i] = NULL;
    noncontig_float_sizes[row_i] = 0;
    } else {
    free((char*) (m[0]));
    }
    
    /* free up row pointer memory
       -----------------------------*/
    free((char*) (m));
}

/***------------------------------------------------------------------------
 * NAME:        allocate_dvector
 * PURPOSE: allocate space for a double vector with n elements
 * EXAMPLE:
 *  double *pdouble;
 *  pdouble = allocate_dvector(10);
 ------------------------------------------------------------------------***/
double *allocate_dvector(int n)
{
    double *v;
    
    v=(double *)calloc((unsigned)n, sizeof(double));
    if (!v) 
      fatal_error( "%s %s %u elements",
          "allocate_dvector:",
          "Can't allocate memory for double vector with", n);
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:     reallocate_dvector
 * PURPOSE:      changes length of a double vector previously
 *  allocated using allocate_dvector
 * EXAMPLE:
 *  double *pdouble;
 *  pdouble = reallocate_dvector(pdouble, 10, 20);
 ------------------------------------------------------------------------***/
double *reallocate_dvector(double *v, int n, int newn)
{
    int i;
    
    v=(double *)realloc((char *)v,(unsigned)newn * sizeof(double));
    if (!v)
      fatal_error( "%s %s %u elements",
          "reallocate_dvector:",
          "Can't allocate memory for double vector with", n);
    /* zero out new entries */
    for(i = n; i < newn; i++){ v[i] = 0.; }
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        deallocate_dvector
 * PURPOSE: free up space previously allocated using allocate_dvector()
 * EXAMPLE:
 *  double *pvector;
 *  pvector = allocate_dvector(10);
 *  ...
 *  deallocate_dvector(pvector);
 ------------------------------------------------------------------------***/
void deallocate_dvector(double *v)
{
    free((char*) v);
}

/***-------------------------------------------------------------------------
 * NAME:        allocate_dmatrix
 * PURPOSE: allocate space for a double 2D matrix 
 * EXAMPLE:
 *  double **pdouble;
 *  pdouble = allocate_dmatrix(10, 100);
 ------------------------------------------------------------------------***/
double **allocate_dmatrix(int nrow, int ncol)
{
    int i;
    double **row_pointers;
    double *data_pointer;
    int nitems;
    
    nitems = nrow*ncol;
    
    /* allocate array of pointers to point to beginning of rows 
       -----------------------------------------------------------*/
    if(nrow == 0)
      /* allocate at least one entry to hold matrix */
      row_pointers= (double **) malloc((unsigned) sizeof(double*));
    else
      row_pointers= (double **) malloc((unsigned) nrow*sizeof(double*));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_dmatrix:",
            "Can't allocate row pointer memory for double matrix",
            nrow, ncol, nitems);
    }
    
    /* allocate zeroed out contiguous memory for all data 
       ------------------------------------------------------*/
    data_pointer = (double *)calloc((unsigned)nitems,sizeof(double));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_dmatrix:",
            "Can't allocate data memory for double matrix",
            nrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space 
       -------------------------------------------------------------*/
    if(nrow == 0) row_pointers[0] = data_pointer;
    for(i=0; i<nrow; i++) {
    row_pointers[i]= data_pointer;
    data_pointer +=  ncol;
    }
    return row_pointers;
}

static double **noncontig_double = NULL;
static int *noncontig_double_sizes = NULL;
static int noncontig_double_length = 0;

/***-------------------------------------------------------------------------
 * NAME:        is_noncontig_double
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_double of entry which matches m
 *      if m is not in noncontig_double, returns -1
 ------------------------------------------------------------------------***/
static int is_noncontig_double(double **m)
{
    int i;
    for(i = 0; i < noncontig_double_length; i++)
      if((double *)m == noncontig_double[i]) return(i);
    return(-1);
}

/***-------------------------------------------------------------------------
 * NAME:        get_noncontig_double_entry
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_double of entry which is not
 *      currently in use. if noncontig_double is full, expands it to
 *      create unused entries
 ------------------------------------------------------------------------***/
static int get_noncontig_double_entry(void)
{
    int i, rv;
    for(i = 0; i < noncontig_double_length; i++)
      if(noncontig_double[i] == NULL) return(i);
    
    /* didn't find empty */
    if(i == 0){
    noncontig_double = (double **)allocate_pvector(NONCONTIG_INC);
    noncontig_double_sizes = (int *)allocate_ivector(NONCONTIG_INC);
    noncontig_double_length = NONCONTIG_INC;
    rv = 0;
    } else {
    noncontig_double = (double **)
      reallocate_pvector((void **)noncontig_double,
                 noncontig_double_length,
                 noncontig_double_length + NONCONTIG_INC);
    noncontig_double_sizes =
      reallocate_ivector(noncontig_double_sizes, noncontig_double_length,
                 noncontig_double_length + NONCONTIG_INC);
    rv = noncontig_double_length;
    noncontig_double_length += NONCONTIG_INC;
    }
    return(rv);
}

/***-------------------------------------------------------------------------
 * NAME:        reallocate_dmatrix
 * PURPOSE: change the number of rows in a double matrix previously
 *  allocated using allocate_dmatrix.
 *  warning: the data moves, so old pointers into the data become invalid
 * EXAMPLE:
 *  double **pdouble;
 *  pdouble = allocate_matrix(10, 100);
 *  ...
 *  pdouble = reallocate_dmatrix(pdouble, 10, 100, 20);
 ------------------------------------------------------------------------***/
double **reallocate_dmatrix(double **row_pointers,
                int nrow, int ncol, int newrow)
{
    int i,j;
    double *data_pointer;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    nitems = newrow*ncol;
    
    if(is_noncontig_double(row_pointers) >= 0)
      fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n%s",
          "reallocate_dmatrix:",
          "Can't allocate memory for double matrix",
          newrow, ncol, nitems, "Matrix is not contiguous");
    
    /* if this won't change the size, just return */
    if(nrow == newrow) return row_pointers;

    /* allocate contiguous memory for all data, copying current data
       ----------------------------------------------------------------*/
    data_pointer = (double *)realloc((char *)row_pointers[0],
                     (unsigned)nitems*sizeof(double));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_dmatrix:",
            "Can't allocate data memory for double matrix",
            newrow, ncol, nitems);
    }
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    if(newrow == 0)
      row_pointers= (double **) realloc((char *)row_pointers,
                    (unsigned) sizeof(double*));
    else
      row_pointers= (double **) realloc((char *)row_pointers,
                    (unsigned) newrow*sizeof(double*));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_dmatrix:",
            "Can't allocate row pointer memory for double matrix",
            newrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space
       -----------------------------------------------------------*/
    if(newrow == 0)  row_pointers[0] = data_pointer;
    for(i=0; i<newrow; i++) {
    row_pointers[i]= data_pointer;
    data_pointer +=  ncol;
    }
    
    /* zero the space added to the data matrix
       -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
    for(j = 0; j < ncol; j++){
        row_pointers[i][j] = 0.0;
    }
    }
    return row_pointers;
}

/***-------------------------------------------------------------------------
 * NAME:         reallocate_dmatrix_noncontig()
 * PURPOSE:      increase the number of rows in a double matrix previously
 *  allocated using allocate_dmatrix, but add the new rows as a separate,
 *  non-contiguous block of memory
 * NOTES:
 *reallocate_dmatrix() may not be called after reallocate_dmatrix_noncontig()
 *reallocate_dmatrix_noncontig() will produce a warning if called to shrink
 *  a matrix.  The matrix will be untouched and no associated memory
 *  will be freed
 * EXAMPLE:
 *  double **pdouble;
 *  pdouble = allocate_dmatrix(10, 100);
 *  ...
 *  pdouble = reallocate_dmatrix_noncontig(pdouble, 10, 100, 20);
 ------------------------------------------------------------------------***/
double **reallocate_dmatrix_noncontig(double **row_pointers,
                      int nrow, int ncol, int newrow)
{
    int i;
    double *data_pointer, **new_row_ptrs;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    if(nrow > newrow){
    warning("%s %s\nNew rows %u, old rows %u, %u columns\n%s",
        "reallocate_dmatrix_noncontig:",
        "Can't decrease number of rows in a non-contiguous matrix.",
        newrow, nrow, ncol, "Will not alter matrix.");
    return(row_pointers);
    }
    if(nrow == newrow)
      return row_pointers;
    /* 0 <= nrow < newrow */
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    new_row_ptrs =
      (double **) realloc((char *)row_pointers,
              (unsigned) newrow*sizeof(double*));
    if (!new_row_ptrs) {
    fatal_error( "%s %s\nwith %u rows, %u columns\n",
            "reallocate_dmatrix_noncontig:",
            "Can't allocate row pointer memory for double matrix",
            newrow, ncol);
    }
    if((i = is_noncontig_double(row_pointers)) < 0){ /* newly noncontig */
    /* put the row pointer into noncontig_double */
    i = get_noncontig_double_entry();
    noncontig_double[i] = (double *)new_row_ptrs;
    noncontig_double_sizes[i] = newrow;
    /* put the first block pointer into noncontig_double */
    if(nrow != 0){
        i = get_noncontig_double_entry();
        noncontig_double[i] = new_row_ptrs[0];
        noncontig_double_sizes[i] = 0;
    } else {/* if nrow is 0, we need to free the "empty" space */
        free((char *)new_row_ptrs[0]);
    }
    } else {
    noncontig_double[i] = (double *)new_row_ptrs;
    noncontig_double_sizes[i] = newrow;
    }
    
    /* allocate contiguous memory for extra data
       ---------------------------------------------*/
    nitems = (newrow - nrow)*ncol;
    
    data_pointer =
      (double *)calloc((unsigned)nitems, sizeof(double));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_dmatrix_noncontig:",
            "Can't allocate data memory for double matrix",
            newrow, ncol, nitems);
    }
    i = get_noncontig_double_entry();
    noncontig_double[i] = data_pointer;
    noncontig_double_sizes[i] = 0;
    
    /* make row pointers point to the right place in the data space
       ---------------------------------------------------------------*/
    for(i = nrow; i < newrow; i++) {
    new_row_ptrs[i] = data_pointer;
    data_pointer += ncol;
    }
    return new_row_ptrs;
}

/***------------------------------------------------------------------------
 * NAME:        deallocate_dmatrix
 * PURPOSE: free space previously allocated using allocate_dmatrix()
 *      for a double 2D matrix 
 * EXAMPLE:
 *  double **pmatrix;
 *  pmatrix = allocate_dmatrix(10, 100);
 *      ...
 *  deallocate_dmatrix(pmatrix);
 ------------------------------------------------------------------------***/
void deallocate_dmatrix(double **m)
{
    int i, j;
    int row_i;
    
    if ( m == NULL ) return;
    /* free up linear data memory 
       -----------------------------*/
    if((row_i = is_noncontig_double(m)) >= 0){
    /* free all noncontiguous blocks */
    for(i = 0; i < noncontig_double_sizes[row_i]; i++)
      for(j = 0; j < noncontig_double_length; j++)
        if(m[i] == noncontig_double[j]){
            free((char *)m[i]);
        noncontig_double[j] = NULL;
        }
    noncontig_double[row_i] = NULL;
    noncontig_double_sizes[row_i] = 0;
    } else {
    free((char*) (m[0]));
    }
    
    /* free up row pointer memory
       -----------------------------*/
    free((char*) (m));
}

/***------------------------------------------------------------------------
 * NAME:        allocate_ivector
 * PURPOSE: allocate space for an integer vector with n elements
 * EXAMPLE:
 *  int *pinteger;
 *  pinteger = allocate_ivector(10);
 ------------------------------------------------------------------------***/
int *
allocate_ivector(int n)
{
    int *v;
    
    v=(int *)calloc((unsigned)n, sizeof(int));
    if (!v) 
      fatal_error( "%s %s %u elements",
          "allocate_ivector:",
          "Can't allocate memory for int vector with", n);
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        reallocate_ivector
 * PURPOSE: changes length of an integer vector previously
 *      allocated using allocate_ivector
 *
 * EXAMPLE:
 *  int *pint;
 *  pint = reallocate_ivector(pint,10,20);
 ------------------------------------------------------------------------***/
int *reallocate_ivector(int *v, int n, int newn)
{
    int i;
    
    v=(int *)realloc((char *) v,(unsigned)newn * sizeof(int));
    if (!v)
      fatal_error( "%s %s %u elements",
          "reallocate_ivector:",
          "Can't allocate memory for integer vector with", n);
    /* zero out new entries */
    for(i = n; i < newn; i++){ v[i] = 0; }
    return v;
}

/***------------------------------------------------------------------------
 * NAME:        deallocate_ivector
 * PURPOSE: free up space previously allocated using allocate_ivector()
 * EXAMPLE:
 *  int *pvector;
 *  pvector = allocate_ivector(10);
 *  ...
 *  deallocate_ivector(pvector);
 ------------------------------------------------------------------------***/
void deallocate_ivector(int *v)
{
    free((char*) v);
}
/***-------------------------------------------------------------------------
 * NAME:        allocate_iindex
 * PURPOSE: allocate space for an index and initialize the index values
 *      such that index[i] = i;
 * EXAMPLE:
 *  int *pinteger;
 *  pinteger = allocate_iindex(10);
 ------------------------------------------------------------------------***/
int *
allocate_iindex(int n)
{
    int *v, i;
    v = allocate_ivector(n);
    for(i = 0; i < n; i++) v[i] = i;
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        allocate_imatrix
 * PURPOSE: allocate space for an integer 2D matrix 
 * EXAMPLE:
 *  int **pinteger;
 *  pinteger = allocate_imatrix(10, 100);
 ------------------------------------------------------------------------***/
int **allocate_imatrix(int nrow, int ncol)
{
    int i;
    int **row_pointers;
    int *data_pointer;
    int nitems;
    
    nitems = nrow*ncol;
    
    /* allocate array of pointers to point to beginning of rows 
       -----------------------------------------------------------*/
    if(nrow == 0)
      /* allocate at least one entry to hold matrix */
      row_pointers= (int **) malloc((unsigned)sizeof(int*));
    else
      row_pointers= (int **) malloc((unsigned)nrow*sizeof(int*));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_imatrix:",
            "Can't allocate row pointer memory for int matrix",
            nrow, ncol, nitems);
    }
    
    /* allocate zeroed out contiguous memory for all data 
       ------------------------------------------------------*/
    data_pointer = (int *)calloc((unsigned)nitems, sizeof(int));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_imatrix:",
            "Can't allocate data memory for int matrix",
            nrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space 
       -------------------------------------------------------------*/
    if(nrow == 0) row_pointers[0] = data_pointer;
    for(i=0; i<nrow; i++) {
    row_pointers[i] = data_pointer;
    data_pointer +=  ncol;
    }
    return row_pointers;
}

static int **noncontig_int = NULL;
static int *noncontig_int_sizes = NULL;
static int noncontig_int_length = 0;

/***-------------------------------------------------------------------------
 * NAME:        is_noncontig_int
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_int of entry which matches m
 *      if m is not in noncontig_int, returns -1
 ------------------------------------------------------------------------***/
static int is_noncontig_int(int **m)
{
    int i;
    for(i = 0; i < noncontig_int_length; i++)
      if((int *)m == noncontig_int[i]) return(i);
    return(-1);
}

/***-------------------------------------------------------------------------
 * NAME:        get_noncontig_int_entry
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_int of entry which is not
 *      currently in use.
 *      if noncontig_int is full, expands it to create unused entries
 ------------------------------------------------------------------------***/
static int get_noncontig_int_entry(void)
{
    int i, rv;
    for(i = 0; i < noncontig_int_length; i++)
      if(noncontig_int[i] == NULL) return(i);
    
    /* didn't find empty */
    if(i == 0){
    noncontig_int = (int **)allocate_pvector(NONCONTIG_INC);
    noncontig_int_sizes = allocate_ivector(NONCONTIG_INC);
    noncontig_int_length = NONCONTIG_INC;
    rv = 0;
    } else {
    noncontig_int = (int **)
      reallocate_pvector((void **)noncontig_int, noncontig_int_length,
                 noncontig_int_length + NONCONTIG_INC);
    noncontig_int_sizes =
      reallocate_ivector(noncontig_int_sizes, noncontig_int_length,
                 noncontig_int_length + NONCONTIG_INC);
    rv = noncontig_int_length;
    noncontig_int_length += NONCONTIG_INC;
    }
    return(rv);
}

/***-------------------------------------------------------------------------
 * NAME:     reallocate_imatrix
 * PURPOSE:      change the number of rows in an integer matrix previously
 *  allocated using allocate_imatrix.
 *  warning: the data moves, so old pointers into the data become invalid
 * EXAMPLE:
 *       int **pint;
 *  pint = allocate_imatrix(10, 100);
 *  ...
 *       pint = reallocate_imatrix(pint, 10, 100, 20);
 ------------------------------------------------------------------------***/
int **reallocate_imatrix(int **row_pointers, int nrow, int ncol, int newrow)
{
    int i,j;
    int *data_pointer;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    nitems = newrow*ncol;
    
    if(is_noncontig_int(row_pointers) >= 0)
      fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n%s",
          "reallocate_imatrix:",
          "Can't allocate memory for int matrix",
          newrow, ncol, nitems, "Matrix is not contiguous");
    
    /* if this won't change the size, just return */
    if(nrow == newrow) return row_pointers;

    /* allocate contiguous memory for all data, copying current data
       -----------------------------------------------------------*/
    data_pointer = (int *)realloc((char *) row_pointers[0],
                  (unsigned)nitems*sizeof(int));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_imatrix:",
            "Can't allocate data memory for int matrix",
            newrow, ncol, nitems);
    }  
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    if(newrow == 0)
      row_pointers= (int **) realloc((char *)row_pointers,
                     (unsigned) sizeof(int*));
    else
      row_pointers= (int **) realloc((char *)row_pointers,
                     (unsigned) newrow*sizeof(int*));
    if (!row_pointers) {  
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_imatrix:",
            "Can't allocate row pointer memory for int matrix",
            newrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space
       -----------------------------------------------------------*/
    if(newrow == 0) row_pointers[0] = data_pointer;
    for(i=0; i<newrow; i++) {
    row_pointers[i]= data_pointer;
    data_pointer +=  ncol;
    }
    
    /* zero the space added to the data matrix
       -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
    for(j = 0; j < ncol; j++){
        row_pointers[i][j] = 0;
    }
    }
    return row_pointers;
}

/***-------------------------------------------------------------------------
 * NAME:         reallocate_imatrix_noncontig()
 * PURPOSE:      increase the number of rows in an int matrix previously
 *  allocated using allocate_imatrix, but add the new rows as a separate,
 *  non-contiguous block of memory
 * NOTES:
 *reallocate_imatrix() may not be called after reallocate_imatrix_noncontig()
 *reallocate_imatrix_noncontig() will produce a warning if called to shrink
 *  a matrix.  The matrix will be untouched and no associated memory
 *  will be freed
 * EXAMPLE:
 *  int **pint;
 *  pint = allocate_imatrix(10, 100);
 *  ...
 *  pint = reallocate_imatrix_noncontig(pint, 10, 100, 20);
 ------------------------------------------------------------------------***/
int **reallocate_imatrix_noncontig(int **row_pointers,
                   int nrow, int ncol, int newrow)
{
    int i;
    int *data_pointer, **new_row_ptrs;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    if(nrow > newrow){
    warning("%s %s\nNew rows %u, old rows %u, %u columns\n%s",
        "reallocate_imatrix_noncontig:",
        "Can't decrease number of rows in a non-contiguous matrix.",
        newrow, nrow, ncol, "Will not alter matrix.");
    return(row_pointers);
    }
    if(nrow == newrow)
      return row_pointers;
    /* 0 <= nrow < newrow */
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    new_row_ptrs =
      (int **) realloc((char *) row_pointers, (unsigned) newrow*sizeof(int*));
    if (!new_row_ptrs) {
    fatal_error( "%s %s\nwith %u rows, %u columns\n",
            "reallocate_imatrix_noncontig:",
            "Can't allocate row pointer memory for int matrix",
            newrow, ncol);
    }
    if((i = is_noncontig_int(row_pointers)) < 0){ /* newly noncontig */
    /* put the row pointer into noncontig_int */
    i = get_noncontig_int_entry();
    noncontig_int[i] = (int *)new_row_ptrs;
    noncontig_int_sizes[i] = newrow;
    /* put the first block pointer into noncontig_int */
    if(nrow != 0){
        i = get_noncontig_int_entry();
        noncontig_int[i] = new_row_ptrs[0];
        noncontig_int_sizes[i] = 0;
    } else {/* if nrow is 0, we need to free the "empty" space */
        free((char *)new_row_ptrs[0]);
    }
    } else {
    noncontig_int[i] = (int *)new_row_ptrs;
    noncontig_int_sizes[i] = newrow;
    }
    
    /* allocate contiguous memory for extra data
       ---------------------------------------------*/
    nitems = (newrow - nrow)*ncol;
    
    data_pointer =
      (int *)calloc((unsigned)nitems, sizeof(int));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_imatrix_noncontig:",
            "Can't allocate data memory for int matrix",
            newrow, ncol, nitems);
    }
    i = get_noncontig_int_entry();
    noncontig_int[i] = data_pointer;
    noncontig_int_sizes[i] = 0;

    /* make row pointers point to the right place in the data space
       ---------------------------------------------------------------*/
    for(i = nrow; i < newrow; i++) {
    new_row_ptrs[i] = data_pointer;
    data_pointer += ncol;
    }
    return new_row_ptrs;
}

/***-------------------------------------------------------------------------
 * NAME:        deallocate_imatrix
 * PURPOSE: free space previously allocated using allocate_imatrix()
 *      for an integer 2D matrix 
 * EXAMPLE:
 *  int **pmatrix;
 *  pmatrix = allocate_imatrix(10, 100);
 *      ...
 *  deallocate_imatrix(pmatrix);
 ------------------------------------------------------------------------***/
void deallocate_imatrix(int **m)
{
    int i, j, row_i;
    
    if ( m == NULL ) return;

    /* free up linear data memory 
       --------------------------*/
    if((row_i = is_noncontig_int(m)) >= 0){
    /* free all noncontiguous blocks */
    for(i = 0; i < noncontig_int_sizes[row_i]; i++)
      for(j = 0; j < noncontig_int_length; j++)
        if(m[i] == noncontig_int[j]){
        free((char *)m[i]);
        noncontig_int[j] = NULL;
        }
    noncontig_int[row_i] = NULL;
    noncontig_int_sizes[row_i] = 0;
    } else {
    free((char*) (m[0]));
    }
    
    /* free up row pointer memory
       -----------------------------*/
    free((char*) (m));
}

/***-------------------------------------------------------------------------
 * NAME:        allocate_pvector
 * PURPOSE: allocate space for a vector of pointers with n elements
 * EXAMPLE:
 *  void **pinteger;
 *  pinteger = allocate_pvector(10);
 ------------------------------------------------------------------------***/
void **allocate_pvector(int n)
{
    void **v;
    
    v=(void **)calloc((unsigned)n, sizeof(void *));
    if (v == NULL)
      fatal_error( "%s %s %u elements",
          "allocate_pvector:",
          "Can't allocate memory for pointer vector with", n);
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        reallocate_pvector
 * PURPOSE: changes length of a pointer vector previously
 *      allocated using allocate_pvector
 *
 * EXAMPLE:
 *  void **pvoid;
 *  pvoid = reallocate_ivector(pvoid,10,20);
 ------------------------------------------------------------------------***/
void **reallocate_pvector(void **v, int n, int newn)
{
    int i;
    
    v=(void **)realloc((char *) v,(unsigned)newn * sizeof(void *));
    if (!v)
      fatal_error( "%s %s %u elements",
          "reallocate_pvector:",
          "Can't allocate memory for pointer vector with", n);
    /* zero out new entries */
    for(i = n; i < newn; i++){ v[i] = NULL; }
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        deallocate_pvector
 * PURPOSE: free up space previously allocated using allocate_pvector()
 * EXAMPLE:
 *  void **pvector;
 *  pvector = allocate_pvector(10);
 *  ...
 *  deallocate_pvector(pvector);
 ------------------------------------------------------------------------***/
void deallocate_pvector(void **v)
{
    free((char*) v);
}

/***-------------------------------------------------------------------------
 * NAME:        allocate_pmatrix
 * PURPOSE: allocate space for a 2D matrix of pointers
 * EXAMPLE:
 *  void ***ppointer;
 *  ppointer = allocate_imatrix(10, 100);
 ------------------------------------------------------------------------***/
void ***allocate_pmatrix(int nrow, int ncol)
{
    int i;
    void ***row_pointers;
    void **data_pointer;
    int nitems;
    
    nitems = nrow*ncol;
    
    /* allocate array of pointers to point to beginning of rows 
       -----------------------------------------------------------*/
    if(nrow == 0)
      /* allocate at least one entry to hold matrix */
      row_pointers= (void ***) malloc((unsigned) sizeof(void**));
    else
      row_pointers= (void ***) malloc((unsigned) nrow*sizeof(void**));
    if (!row_pointers){
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_pmatrix:",
            "Can't allocate row pointer memory for pointer matrix",
            nrow, ncol, nitems);
    }
    
    /* allocate zeroed out contiguous memory for all data 
       ------------------------------------------------------*/
    data_pointer = (void **)calloc((unsigned)nitems,sizeof(void*));
    if (!data_pointer){
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_pmatrix:",
            "Can't allocate data memory for pointer matrix",
            nrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space 
       -------------------------------------------------------------*/
    if(nrow == 0) row_pointers[0] = data_pointer;
    for(i=0; i<nrow; i++){
    row_pointers[i] = data_pointer;
    data_pointer +=  ncol;
    }
    return row_pointers;
}

static void ***noncontig_pointer = NULL;
static int *noncontig_pointer_sizes = NULL;
static int noncontig_pointer_length = 0;

/***------------------------------------------------------------------------
 * NAME:        is_noncontig_pointer
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_pointer of entry which matches m
 *      if m is not in noncontig_pointer, returns -1
 ------------------------------------------------------------------------***/
static int is_noncontig_pointer(void ***m)
{
    int i;
    for(i = 0; i < noncontig_pointer_length; i++)
      if((void **)m == noncontig_pointer[i]) return(i);
    return(-1);
}

/***------------------------------------------------------------------------
 * NAME:        get_noncontig_pointer_entry
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_pointer of entry which is not
 *      currently in use. if noncontig_pointer is full, expands it
 *      to create unused entries
 ------------------------------------------------------------------------***/
static int get_noncontig_pointer_entry(void)
{
    int i, rv;
    for(i = 0; i < noncontig_pointer_length; i++)
      if(noncontig_pointer[i] == NULL) return(i);
    
    /* didn't find empty */
    if(i == 0){
    noncontig_pointer = (void ***)allocate_pvector(NONCONTIG_INC);
    noncontig_pointer_sizes = allocate_ivector(NONCONTIG_INC);
    noncontig_pointer_length = NONCONTIG_INC;
    rv = 0;
    } else {
    noncontig_pointer = (void ***)
      reallocate_pvector((void **)noncontig_pointer,
                 noncontig_pointer_length,
                 noncontig_pointer_length + NONCONTIG_INC);
    noncontig_pointer_sizes =
      reallocate_ivector(noncontig_pointer_sizes,
                 noncontig_pointer_length,
                 noncontig_pointer_length + NONCONTIG_INC);
    rv = noncontig_pointer_length;
    noncontig_pointer_length += NONCONTIG_INC;
    }
    return(rv);
}

/***-------------------------------------------------------------------------
 * NAME:     reallocate_pmatrix
 * PURPOSE:      change the number of rows in a pointer matrix previously
 *  allocated using allocate_pmatrix.
 *  warning: the data moves, so old pointers into the data become invalid
 * EXAMPLE:
 *  void ***pvoid;
 *  pvoid = allocate_pmatrix(10, 100);
 *  ...
 *  pvoid = reallocate_pmatrix(pvoid, 10, 100, 20);
 ------------------------------------------------------------------------***/
void ***reallocate_pmatrix(void ***row_pointers, int nrow, int ncol,
               int newrow)
{
    int i, j;
    void **data_pointer;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    nitems = newrow*ncol;
    
    if(is_noncontig_pointer(row_pointers) >= 0)
      fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n%s",
          "reallocate_pmatrix:",
          "Can't allocate memory for pointer matrix",
          newrow, ncol, nitems, "Matrix is not contiguous");
    
    /* if this won't change the size, just return */
    if(nrow == newrow) return row_pointers;

    /* allocate contiguous memory for all data, copying current data
       -----------------------------------------------------------*/
    data_pointer = (void **)realloc((char *) row_pointers[0],
                    (unsigned)nitems*sizeof(void **));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_pmatrix:",
            "Can't allocate data memory for pointer matrix",
            newrow, ncol, nitems);
    }
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    if(newrow == 0)
      row_pointers = (void ***) realloc((char *)row_pointers,
                    (unsigned) sizeof(void**));
    else
      row_pointers = (void ***) realloc((char *)row_pointers,
                    (unsigned) newrow*sizeof(void**));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_pmatrix:",
            "Can't allocate row pointer memory for pointer matrix",
            newrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space
       -----------------------------------------------------------*/
    if(newrow == 0) row_pointers[0] = data_pointer;
    for(i=0; i<newrow; i++) {
    row_pointers[i]= data_pointer;
    data_pointer +=  ncol;
    }
    
    /* zero the space added to the data matrix
       -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
    for(j = 0; j < ncol; j++){
        row_pointers[i][j] = NULL;
    }
    }
    return row_pointers;
}

/***-------------------------------------------------------------------------
 * NAME:         reallocate_pmatrix_noncontig()
 * PURPOSE:      increase the number of rows in a pointer matrix previously
 *  allocated using allocate_pmatrix, but add the new rows as a separate,
 *  non-contiguous block of memory
 * NOTES:
 *reallocate_pmatrix() may not be called after reallocate_pmatrix_noncontig()
 *reallocate_pmatrix_noncontig() will produce a warning if called to shrink
 *  a matrix.  The matrix will be untouched and no associated memory
 *  will be freed
 * EXAMPLE:
 *  void ***pointer;
 *  ppointer = allocate_pmatrix(10, 100);
 *  ...
 *  ppointer = reallocate_pmatrix_noncontig(ppointer, 10, 100, 20);
 ------------------------------------------------------------------------***/
void ***reallocate_pmatrix_noncontig(void ***row_pointers,
                     int nrow, int ncol, int newrow)
{
    int i;
    void **data_pointer, ***new_row_ptrs;
    int nitems;
    
    if(row_pointers == NULL) return NULL;
    
    if(nrow > newrow){
    warning("%s %s\nNew rows %u, old rows %u, %u columns\n%s",
        "reallocate_pmatrix_noncontig:",
        "Can't decrease number of rows in a non-contiguous matrix.",
        newrow, nrow, ncol, "Will not alter matrix.");
    return(row_pointers);
    }
    if(nrow == newrow)
      return row_pointers;
    /* 0 <= nrow < newrow */
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    new_row_ptrs =
      (void ***) realloc((char *) row_pointers,
             (unsigned) newrow*sizeof(void **));
    if (!new_row_ptrs) {
    fatal_error( "%s %s\nwith %u rows, %u columns\n",
            "reallocate_pmatrix_noncontig:",
            "Can't allocate row pointer memory for pointer matrix",
            newrow, ncol);
    }
    if((i = is_noncontig_pointer(row_pointers)) < 0){ /* newly noncontig */
    /* put the row pointer into noncontig_pointer */
    i = get_noncontig_pointer_entry();
    noncontig_pointer[i] = (void **)new_row_ptrs;
    noncontig_pointer_sizes[i] = newrow;
    /* put the first block pointer into noncontig_pointer */
    if(nrow != 0){
        i = get_noncontig_pointer_entry();
        noncontig_pointer[i] = new_row_ptrs[0];
        noncontig_pointer_sizes[i] = 0;
    } else {/* if nrow is 0, we need to free the "empty" space */
        free((char *)new_row_ptrs[0]);
    }
    } else {
    noncontig_pointer[i] = (void **)new_row_ptrs;
    noncontig_pointer_sizes[i] = newrow;
    }
    
    /* allocate contiguous memory for extra data
       ---------------------------------------------*/
    nitems = (newrow - nrow)*ncol;
    
    data_pointer =
      (void **)calloc((unsigned)nitems, sizeof(void *));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_pmatrix_noncontig:",
            "Can't allocate data memory for pointer matrix",
            newrow, ncol, nitems);
    }
    i = get_noncontig_pointer_entry();
    noncontig_pointer[i] = data_pointer;
    noncontig_pointer_sizes[i] = 0;
    
    /* make row pointers point to the right place in the data space
       ---------------------------------------------------------------*/
    for(i = nrow; i < newrow; i++) {
    new_row_ptrs[i] = data_pointer;
    data_pointer += ncol;
    }
    return new_row_ptrs;
}

/***------------------------------------------------------------------------
 * NAME:        deallocate_pmatrix
 * PURPOSE: free space previously allocated using allocate_pmatrix()
 *      for a 2D matrix of pointers
 * EXAMPLE:
 *  void ***pmatrix;
 *  pmatrix = allocate_pmatrix(10, 100);
 *      ...
 *  deallocate_pmatrix(pmatrix);
 ------------------------------------------------------------------------***/
void deallocate_pmatrix(void ***m)
{
    int i, j, row_i;
    
    if (m == NULL) return;
    /* free up linear data memory 
       --------------------------*/
    if((row_i = is_noncontig_pointer(m)) >= 0){
    /* free all noncontiguous blocks */
    for(i = 0; i < noncontig_pointer_sizes[row_i]; i++)
      for(j = 0; j < noncontig_pointer_length; j++)
        if(m[i] == noncontig_pointer[j]){
        free((char *)m[i]);
        noncontig_pointer[j] = NULL;
        }
    noncontig_pointer[row_i] = NULL;
    noncontig_pointer_sizes[row_i] = 0;
    } else {
    free((char*) (m[0]));
    }
    
    /* free up row pointer memory
       -----------------------------*/
    free((char*) (m));
}

/***------------------------------------------------------------------------
 * NAME:        allocate_structvector
 * PURPOSE: allocate space for a vector of structures with n elements
 * EXAMPLE:
 *  struct struct_type  *vector;
 *  vector = (struct struct_type *)
 *            allocate_vector(10, sizeof(struct struct_type));
 *   vector[3].field = 123.0;
 ------------------------------------------------------------------------***/
void *allocate_structvector(int n, int size)
{
    void *v;
    
    v=(void *)calloc((unsigned)n, (unsigned) size);
    if (!v) 
      fatal_error( "%s %s %u elements",
          "allocate_structvector:",
          "Can't allocate memory for struct vector with", n);
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:        reallocate_structvector
 * PURPOSE: changes length of vector of structures previously
 *  allocated using allocate_structvector
 * EXAMPLE:
 *  struct struct_type  *vector;
 *  vector = (struct struct_type *)
 *            allocate_vector(10, sizeof(struct struct_type));
 *  vector[3].field = 123.0;
 *  vector = reallocate_structvector(vector, 10, 20,
 *                   sizeof(struct struct_type));
 *  vector[13].field = 123.0;
 ------------------------------------------------------------------------***/
void *reallocate_structvector(void *v, int n, int newn, int size)
{
    v = (void *)realloc((char *) v, (unsigned)(newn * size));
    if (!v)
      fatal_error( "%s %s %u elements",
          "reallocate_structvector:",
          "Can't allocate memory for struct vector with", n);
    return v;
}

/***-------------------------------------------------------------------------
 * NAME:   deallocate_structvector
 * PURPOSE:free up space previously allocated using allocate_structvector()
 * EXAMPLE:
 *  struct struct_type  *vector;
 *  vector = (struct struct_type *)
 *            allocate_vector(10, sizeof(struct struct_type));
 *   vector[3].field = 123.0;
 *  ...
 *  deallocate_vector(vector);
 ------------------------------------------------------------------------***/
void deallocate_structvector(void *v)
{
    free((char*) v);
}

/***-------------------------------------------------------------------------
 * NAME:        allocate_structmatrix
 * PURPOSE: allocate space for a float 2D matrix of structures 
 * EXAMPLE:
 *  struct struct_type **matrix;
 *  matrix = (struct struct_type **)
 *      allocate_structmatrix(10,100,sizeof(struct struct_type));
 *  matrix[3][67].field = 123.0;
 ------------------------------------------------------------------------***/
void **allocate_structmatrix(int nrow, int ncol, int size)
{
    int i;
    void **row_pointers;
    void *data_pointer;
    int nitems;
    char *byte_pointer;
    
    nitems = nrow*ncol;
    
    /* allocate array of pointers to point to beginning of rows 
       -----------------------------------------------------------*/
    if(nrow == 0)
      /* allocate at least one entry to hold matrix */
      row_pointers= (void **) malloc((unsigned) sizeof(void *));
    else
      row_pointers= (void **) malloc((unsigned) nrow*sizeof(void *));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_structmatrix:",
            "Can't allocate row pointer memory for struct matrix",
            nrow, ncol, nitems);
    }
    
    /* allocate zeroed out contiguous memory for all data 
       ------------------------------------------------------*/
    data_pointer = (void *)calloc((unsigned)nitems, (unsigned)size);
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "allocate_structmatrix:",
            "Can't allocate data memory for struct matrix",
            nrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space 
       -------------------------------------------------------------*/
    byte_pointer = (char *)data_pointer;
    if(nrow == 0) row_pointers[0] = byte_pointer;
    for(i=0; i<nrow; i++) {
    row_pointers[i]= (void *)byte_pointer;
    byte_pointer += ncol*size;
    }
    
    return row_pointers;
}

static void **noncontig_struct = NULL;
static int *noncontig_struct_sizes = NULL;
static int noncontig_struct_length = 0;

/***-------------------------------------------------------------------------
 * NAME:        is_noncontig_struct
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_struct of entry which matches m
 *      if m is not in noncontig_struct, returns -1
 ------------------------------------------------------------------------***/
static int is_noncontig_struct(void **m)
{
    int i;
    for(i = 0; i < noncontig_struct_length; i++)
      if((void *)m == noncontig_struct[i]) return(i);
    return(-1);
}

/***-------------------------------------------------------------------------
 * NAME:        get_noncontig_struct_entry
 * PURPOSE: internal to allocate.c
 *      returns index in noncontig_struct of entry which is not
 *      currently in use. if noncontig_struct is full, expands it to
 *      create unused entries
 ------------------------------------------------------------------------***/
static int get_noncontig_struct_entry(void)
{
    int i, rv;
    for(i = 0; i < noncontig_struct_length; i++)
      if(noncontig_struct[i] == NULL) return(i);
    
    /* didn't find empty */
    if(i == 0){
    noncontig_struct = (void **)allocate_pvector(NONCONTIG_INC);
    noncontig_struct_sizes = allocate_ivector(NONCONTIG_INC);
    noncontig_struct_length = NONCONTIG_INC;
    rv = 0;
    } else {
    noncontig_struct = (void **)
      reallocate_pvector(noncontig_struct, noncontig_struct_length,
                 noncontig_struct_length + NONCONTIG_INC);
    noncontig_struct_sizes =
      reallocate_ivector(noncontig_struct_sizes, noncontig_struct_length,
                 noncontig_struct_length + NONCONTIG_INC);
    rv = noncontig_struct_length;
    noncontig_struct_length += NONCONTIG_INC;
    }
    return(rv);
}

/***-------------------------------------------------------------------------
 * NAME:         reallocate_structmatrix
 * PURPOSE:      change the number of rows in a struct matrix previously
 *  allocated using allocate_structmatrix.
 *  warning: the data moves, so old pointers into the data become invalid
 * EXAMPLE:
 *  struct struct_type **matrix;
 *  matrix = (struct struct_type **)
 *      allocate_structmatrix(10,100,sizeof(struct struct_type));
 *  matrix[3][67].field = 123.0;
 *  matrix = reallocate_structmatrix(matrix, 10, 100, 20,
 *                   sizeof (struct struct_type));
 *  matrix[13][67].field = 123.0;
 ------------------------------------------------------------------------***/
void **reallocate_structmatrix(void **row_pointers, int nrow, int ncol,
                   int newrow, int size)
{
    int i;
    void *data_pointer;
    int nitems;
    char *byte_pointer;
    
    if(row_pointers == NULL) return NULL;
    
    nitems = newrow*ncol;
    
    if(is_noncontig_struct(row_pointers) >= 0)
      fatal_error("%s %s\nwith %u rows, %u columns, and %u elements\n%s",
          "reallocate_structmatrix:",
          "Can't allocate memory for struct matrix",
          newrow, ncol, nitems, "Matrix is not contiguous");
    
    /* if this won't change the size, just return */
    if(nrow == newrow) return row_pointers;

    /* allocate contiguous memory for all data, copying current data
       ----------------------------------------------------------------*/
    data_pointer = (void *)realloc((char *) row_pointers[0],
                   (unsigned)nitems*size);
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_structmatrix:",
            "Can't allocate data memory for struct matrix",
            newrow, ncol, nitems);
    }
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    if(newrow == 0)
      row_pointers= (void **) realloc((char *) row_pointers,
                      (unsigned) sizeof(void *));
    else
      row_pointers= (void **) realloc((char *) row_pointers,
                      (unsigned) newrow*sizeof(void *));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_structmatrix:",
            "Can't allocate row pointer memory for struct matrix",
            newrow, ncol, nitems);
    }
    
    /* make row pointers point to the right place in the data space
       -----------------------------------------------------------*/
    byte_pointer = (char *)data_pointer;
    if(newrow == 0) row_pointers[0] = (void *)byte_pointer;
    for(i=0; i<newrow; i++) {
    row_pointers[i]= (void *)byte_pointer;
    byte_pointer +=  ncol*size;
    }
    /* can't zero new space, don't know the sizes of things */
    
    return row_pointers;
}

/***-------------------------------------------------------------------------
 * NAME:         reallocate_structmatrix_noncontig()
 * PURPOSE:      increase the number of rows in a struct matrix previously
 *  allocated using allocate_structmatrix, but add the new rows as a
 *  separate, non-contiguous block of memory
 * NOTES:
 *  reallocate_structmatrix() may not be called after
 *  reallocate_structmatrix_noncontig()
 *  reallocate_structmatrix_noncontig() will produce a warning if called to
 *  shrink a matrix.  The matrix will be untouched and no associated
 *  memory will be freed
 * EXAMPLE:
 *  struct struct_type **matrix;
 *  matrix = (struct struct_type **)
 *      allocate_structmatrix(10, 100, sizeof(struct struct_type));
 *  matrix[3][67].field = 123.0;
 *  matrix = (struct struct_type **)
 *      reallocate_structmatrix_noncontig(matrix, 10, 100, 20,
 *          sizeof(struct struct_type));
 *  matrix[13][67].field = 123.0;
 ------------------------------------------------------------------------***/
void **reallocate_structmatrix_noncontig(void **row_pointers, int nrow,
                     int ncol, int newrow, int size)
{
    int i;
    void *data_pointer, **new_row_ptrs;
    int nitems;
    char *byte_pointer;
    
    if(row_pointers == NULL) return NULL;
    
    if(nrow > newrow){
    warning("%s %s\nNew rows %u, old rows %u, %u columns\n%s",
        "reallocate_structmatrix_noncontig:",
        "Can't decrease number of rows in a non-contiguous matrix.",
        newrow, nrow, ncol, "Will not alter matrix.");
    return(row_pointers);
    }
    if(nrow == newrow)
      return row_pointers;
    /* 0 <= nrow < newrow */
    
    /* resize array of pointers to beginning of rows
       -----------------------------------------------------------*/
    new_row_ptrs =
      (void **) realloc((char *) row_pointers,
            (unsigned) newrow*sizeof(void *));
    if (!new_row_ptrs) {
    fatal_error( "%s %s\nwith %u rows, %u columns\n",
            "reallocate_structmatrix_noncontig:",
            "Can't allocate row pointer memory for struct matrix",
            newrow, ncol);
    }
    if((i = is_noncontig_struct(row_pointers)) < 0){ /* newly noncontig */
    /* put the row pointer into noncontig_struct */
    i = get_noncontig_struct_entry();
    noncontig_struct[i] = (void *)new_row_ptrs;
    noncontig_struct_sizes[i] = newrow;
    /* put the first block pointer into noncontig_struct */
    if(nrow != 0){
        i = get_noncontig_struct_entry();
        noncontig_struct[i] = new_row_ptrs[0];
        noncontig_struct_sizes[i] = 0;
    } else {/* if nrow is 0, we need to free the "empty" space */
        free((char *)new_row_ptrs[0]);
    }
    } else {
    noncontig_struct[i] = (void *)new_row_ptrs;
    noncontig_struct_sizes[i] = newrow;
    }
    
    /* allocate contiguous memory for extra data
       ---------------------------------------------*/
    nitems = (newrow - nrow)*ncol;
    
    data_pointer = (void *)calloc((unsigned)nitems, (unsigned)size);
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "reallocate_structmatrix_noncontig:",
            "Can't allocate data memory for struct matrix",
            newrow, ncol, nitems);
    }
    i = get_noncontig_struct_entry();
    noncontig_struct[i] = data_pointer;
    noncontig_struct_sizes[i] = 0;
    
    /* make row pointers point to the right place in the data space
       ---------------------------------------------------------------*/
    byte_pointer = (char *)data_pointer;
    for(i = nrow; i < newrow; i++) {
    new_row_ptrs[i] = (void *)byte_pointer;
    byte_pointer += ncol*size;
    }
    return new_row_ptrs;
}

/***-------------------------------------------------------------------------
 * NAME:        deallocate_structmatrix
 * PURPOSE: free space previously allocated using allocate_structmatrix()
 *      for a struct 2D matrix 
 * EXAMPLE:
 *  struct struct_type **matrix;
 *  matrix = (struct struct_type **)
 *      allocate_structmatrix(10,100,sizeof(struct struct_type));
 *  matrix[3][67].field = 123.0;
 *  ...
 *  deallocate_structmatrix(matrix);
 ------------------------------------------------------------------------***/
void deallocate_structmatrix(void **m)
{
    int i, j, row_i;
    
    if (m == NULL) return;
    /* free up linear data memory 
       -----------------------------*/
    if((row_i = is_noncontig_struct(m)) >= 0){
    /* free all noncontiguous blocks */
    for(i = 0; i < noncontig_struct_sizes[row_i]; i++)
      for(j = 0; j < noncontig_struct_length; j++)
        if(m[i] == noncontig_struct[j]){
            free((char *)m[i]);
        noncontig_struct[j] = NULL;
        }
    noncontig_struct[row_i] = NULL;
    noncontig_struct_sizes[row_i] = 0;
    } else {
    free((char*) (m[0]));
    }
    
    /* free up row pointer memory
       -----------------------------*/
    free((char*) (m));
}

/***-------------------------------------------------------------------------
 * NAME:        nr_allocate_vector
 * PURPOSE: allocate space for a float vector with indices 
 *      from nlow to nhigh that is initialized to zero 
 *      DO NOT USE, OBSOLETE, KEPT ONLY FOR numerical recipes
 *      compatibility.
 * EXAMPLE:
 *  float *pfloat;
 *  pfloat = nr_allocate_vector(0, 10);
 ------------------------------------------------------------------------***/
float *nr_allocate_vector(int nlow, int nhigh)
{
    float *v;
    int nitems;
    
    nitems = 1 + nhigh - nlow;
    
    v=(float *)calloc((unsigned)nitems, sizeof(float));
    if (!v) 
      fatal_error( "%s %s %u elements",
          "nr_allocate_vector:",
          "Can't allocate memory for float vector with", nitems);
    return v-nlow;
}

/***-------------------------------------------------------------------------
 * NAME:        nr_allocate_matrix
 * PURPOSE: allocate space for a float 2D matrix 
 * DESCRIPTION:  Allocates space for a float 2D matrix. The first row index
 *      is to a vector of pointers to the columns. The second column index
 *      is to a large linear array where all matrix items are stored
 *      sequentially as if they were stored using a C declared matrix
 *      with static row and column sizes.
 *      This program performs only two calls to malloc,
 *      zeroes out all data, and guarantees that the data is stored
 *      in one large linear area with the 
 *      second (j, x[i][j],or column) index varying fastest. 
 *      DO NOT USE, OBSOLETE, KEPT ONLY FOR numerical recipes
 *      compatibility.
 * EXAMPLE:
 *          float **pfloat;
 *          pfloat = nr_allocate_matrix(0, 10, 0 100);
 ------------------------------------------------------------------------***/
float **nr_allocate_matrix(int row_low, int row_high, int col_low,
               int col_high)
{
    int i;
    float **row_pointers;
    float *data_pointer;
    int ncols, nrows, nitems;
    
    nrows = 1 + row_high - row_low;
    ncols = 1 + col_high - col_low;
    nitems = nrows*ncols;
    
    /* allocate array of pointers to point to beginning of rows 
       -----------------------------------------------------------*/
    row_pointers= (float **) malloc((unsigned) nrows*sizeof(float*));
    if (!row_pointers) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "nr_allocate_matrix:",
            "Can't allocate row pointer memory for float matrix",
            nrows, ncols, nitems);
    }
    row_pointers -= row_low;
    
    /* allocate zeroed out contiguous memory for all data 
       ------------------------------------------------------*/
    data_pointer = (float *)calloc((unsigned)nitems,sizeof(float));
    if (!data_pointer) {
    fatal_error( "%s %s\nwith %u rows, %u columns, and %u elements\n",
            "nr_allocate_matrix:",
            "Can't allocate data memory for float matrix",
            nrows, ncols, nitems);
    }
    
    /* make row pointers point to the right place in the data space 
       -------------------------------------------------------------*/
    for(i=row_low;i<=row_high;i++) {
    row_pointers[i]= data_pointer - col_low;
    data_pointer +=  ncols;
    }
    
    return row_pointers;
}

/***-------------------------------------------------------------------------
 * NAME:        nr_deallocate_vector
 * PURPOSE: free space previously allocated using nr_allocate_vector()
 *     for a float vector with indices from nlow to nhigh 
 * EXAMPLE:
 *      float *pvector;
 *      pvector = nr_allocate_vector(0, 10);
 *     ...
 *     nr_deallocate_vector(pvector, 0, 10);
 ------------------------------------------------------------------------***/
void nr_deallocate_vector(float *v, int nlow, int nhigh)
{
    free((char*) (v+nlow));
}

/***-------------------------------------------------------------------------
 * NAME:   nr_deallocate_matrix
 * PURPOSE:free space was previously allocated using nr_allocate_matrix()
 *     for a float 2D matrix 
 *     with row indices from row_low to row_high
 *     and column indices from col_low to col_high
 * EXAMPLE:
 *  float **pmatrix;
 *  pmatrix = nr_allocate_matrix(0, 10, 0, 100);
 *  ...
 *  nr_deallocate_matrix(pmatrix, 0, 10, 0, 100);
 ------------------------------------------------------------------------***/
void nr_deallocate_matrix(float **m, int row_low, int row_high,
              int col_low, int col_high)
{
    
    /* free up linear data memory 
       --------------------------*/
    free((char*) (m[row_low]+col_low));
    
    /* free up column pointer memory
       --------------------------------*/
    free((char*) (m + row_low));
}

/*
   void
   deallocate_string_vector(array, length)
   char **array;
   int length

   array is an allocated array of allocated strings
   the array is assumed to have allocated with allocate_pvector.
   the strings are assumed to have been allocated with nn_malloc

   to free array, first free all the strings it points to, then free the
   array itself.
   */
void
deallocate_string_vector(char **array, int length)
{
    for(length--; length >= 0; length--)
      nn_free(array[length], "for deallocating string vector string");
    deallocate_pvector((void **)array);
}
