
/***--------------------------------------------------------------------------
* NAME:         bootstrap_lnk.c
* PURPOSE:      creates bootstrap samples of a lnknet data file 
*
* USAGE: 
*	bootstrap_lnk -<flags>
* 
* RETURN VALUE:  NONE
*
* EXAMPLE: bootstrap_lnk --		 * shows all available flags
*
* GLOBALS:
*
* DESCRIPTION:
*    Bootstrap sampling takes a data set with n patterns and generates
*    many new versions of this data set, each with n patterns
*    by sampling with replacement. If desired, a test set will also be
*    created for each bootstrap training file. The test file creates the
*    patterns left over after sampling from the original file.
*
* HISTORY:
*  9/21/94        Yuchun Lee        creation
*  17 April 2002  Richard Lippmann  updated for general use
*
* SEE ALSO:
*
------------------------------------------------------------------------***/
/* (c) Copyright 2004 M.I.T. Lincoln Laboratory 

   Permission is hereby granted, without written agreement or royalty fee, 
   to use, copy, modify, and distribute this software and its 
   documentation for any purpose, provided that the above copyright notice 
   and the following three paragraphs appear in all copies of this 
   software. 

     IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR 
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF 
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN 
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  M.I.T. LINCOLN 
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT. 

        THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN 
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, 
ENHANCEMENTS, OR MODIFICATIONS. 
*/

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/bootstrap_lnk.c,v 11.1 2004/02/13 19:39:21 kukolich Exp $";


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <nnlib.h>
#include <get_subs.h>

#define MYLINE 4095
#define MYFILE 512

/*  Arguments */
static char *in_prefix= "vowel";
static char *out_prefix= "boot";
static int nreps = 10;
static int seed= 1234;

static ARG detect_argtab[]=
{
  {"in_prefix",STRING,(int *)&in_prefix,"prefix for input training and defaults file"},
  {"out_prefix",STRING, (int *)&out_prefix,"prefix for output train, test, and defaults file"},
  {"nreps",INT, (int *)&nreps,"number of bootstrap files desired"},
  {"seed", INT, &seed, "seed for random number"},
};

#define DETECT_TABSIZE (sizeof(detect_argtab)/sizeof(ARG))

int main(argc, argv)
     int	argc;
     char	**argv;
{
  FILE *input, *out_train, *out_test;
  int i,j,n, *intrain;
  int match, ntrain, ntest, nunique_train;
  float percent;
  char *s, *fin_train, *fout_train, *fout_test, *fout_defaults, *fin_defaults;
  int irep=0;

  /* initialization */
  init_nnsubs(argc, argv, rcsid);

  /* read args */
  argc = nnargs(argc, argv, detect_argtab, DETECT_TABSIZE);

  /* count number of patterns in training file 
  ---------------------------------------------*/
  /* build space for input lines */
  s= nn_malloc(MYLINE, "to store lines of input file");

  /* build space for file names */
  fin_train= nn_malloc(MYFILE, "to store name of input training file");
  fin_defaults= nn_malloc(MYFILE, "to store name of input defaults file");
  fout_train= nn_malloc(MYFILE, "to store name of output training file");
  fout_test= nn_malloc(MYFILE, "to store name of output test file");
  fout_defaults= nn_malloc(MYFILE, "to store name of output defaults file");

  for (irep=0; irep<nreps; irep++)
  {
  printf("\ncreating %s%d bootstrap files\n", out_prefix, irep); 
  srandom(irep+seed);
  /* create file names */
  snprintf(fin_train, MYFILE, "%s.train", in_prefix);
  snprintf(fin_defaults, MYFILE, "%s.defaults", in_prefix);
  snprintf(fout_train, MYFILE, "%s%d.train", out_prefix, irep);
  snprintf(fout_test, MYFILE, "%s%d.test", out_prefix, irep);
  snprintf(fout_defaults, MYFILE, "%s%d.defaults", out_prefix, irep);

  /* count the number of training patterns 
  ---------------------------------------------*/
  input= nn_fopen(fin_train,"r","for reading in training patterns");
  n=0;
  while (fscanf(input,"%d",&i)!=EOF)
    {
      fgets(s,MYLINE,input);
      n++;
    }
  nn_fclose(input, fout_train, "for reading in training patterns");

  /* fill array with random numbers from 0 to n-1 
	 where n is the number of training exemplars
  -------------------------------------------------*/
  intrain= allocate_ivector(n);
  for (i=0;i<n;i++)
    intrain[i]= (int)(random_uniform()*(float)n);


  /* read patterns in again and assign to training or test files
  ---------------------------------------------------------------*/
  ntrain=0;
  ntest=0;
  nunique_train=0;
  input= nn_fopen(fin_train,"r","to read in training patterns");
  out_train= nn_fopen(fout_train,"w","to write bootstrap training patterns");
  out_test= nn_fopen(fout_test,"w","to write bootstrap test patterns");
  for (i=0;i<n;i++)
    {
      fgets(s,MYLINE,input);

      /* count number of times this pattern should appear in training data */
      match=0;
      for (j=0;j<n;j++)
	{
	if (intrain[j]==i) 
	    match++;
	}

      if (!match)
	{
	/* in test set */
	fprintf(out_test,"%s",s);
	ntest++;
	}
      else
	{
	/* in train, output multiple times */
	for (j=0;j<match;j++)
	    {
	    fprintf(out_train,"%s",s);
	    }
	ntrain+= match;
        nunique_train++;
	}
    }

  nn_fclose(input, fout_train,"for reading input patterns");
  nn_fclose(out_train, fout_train,"for writing training patters");
  nn_fclose(out_test, fout_test,"for writing test patterns");

  /* now copy the defaults file
  -----------------------------------*/
  snprintf(s, MYLINE, "cp %s %s", fin_defaults, fout_defaults);
  system (s);

  printf("Number of input patterns:           %d\n",n);
  percent = (100.0*nunique_train)/ntrain;
  printf("Number of patterns in training set: %d unique %d %.2f%%\n",ntrain, nunique_train, percent);
  percent = 100.0*ntest/n;
  printf("Number of patterns in test set:     %d %.2f%%\n",ntest, percent);
  }
  return 0;
}
