/***------------------------------------------------------------------------
 * NAME:        class_data.c
 * PURPOSE: package for giving training or cross validation patterns to
 *      classifiers on request.  Takes into account random
 *      presentation order and priors adjustment by sampling
 *
 * DESCRIPTION:
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/class_data.c,v 11.3 2001/08/23 20:14:42 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/class_data.c,v $
 * CREATED: 15-June-lck             MIT LINCOLN LABORATORY
 *--------------------------------------------------------------------------
 * Copyright 1993 MIT Lincoln Laboratory.
 * This software is not for release and should neither be copied
 * nor distributed without explicit permission of MIT Lincoln Laboratory.
 -----------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include "nnlib.h"

/* GLOBAL */
#define TRUE 1
#define FALSE 0

typedef struct{
    /****************** data ********************/
    int noutputs, npatterns, ninputs, input_num, nraw, normalization;
    int *features;
    float **input_matrix;
    int *classes;
    /****** state variables during reading ******/
    int n_read, train_flag;
    /************ random variables **************/
    int *random_index, random_flag;
    /************ priors variables **************/
    int priors_flag, priors_npatterns;
    float *priors_list;
    int *fold_classes, *fold_indices;
    /********** cross valid variables ***********/
    int cross_valid_flag, *cross_index;
    int fold, nfolds, *n_in_train_fold, *n_in_test_fold;
    char **train_masks, **test_masks, *cv_mask;
    /*********** single pass variables **********/
    int single_flag;
    FILE *data_file;
    float *inputs;
    char *data_fname;
} CLASS_PATTERNS;

CLASS_PATTERNS **cp_list;
static int max_models = 0;

/**************************
 * Read npatterns from the data file given.
 * build a model to store the data in
 ***************************/
int
class_data_read(char *data_path, char *data_fname, int npatterns,
        int nraw, int ninputs, int nclasses, int *features,
        int normalization, int verbosity)
{
    int new_model;
    char *data_full_name;
    CLASS_PATTERNS *cpp;
    FILE *data_file;

    /* allocate structure */
    new_model = model_get_number((void ***)&cp_list, &max_models,
                 sizeof(CLASS_PATTERNS));
    /* store what we know so far */
    cpp = cp_list[new_model];
    cpp->noutputs = nclasses;
    cpp->ninputs = ninputs;
    cpp->nraw = nraw;
    cpp->normalization = normalization;
    cpp->features = features;
    cpp->train_flag = TRUE;
    cpp->n_read = 0;
    cpp->random_flag = cpp->priors_flag = cpp->cross_valid_flag = FALSE;
    cpp->single_flag = FALSE;
    cpp->data_file = NULL;
    cpp->inputs = NULL;

    /* allocate space for input matrix and class vector */
    cpp->input_matrix = allocate_matrix(npatterns, ninputs);
    cpp->classes = allocate_ivector(npatterns);

    /* open file, load data, close file */
    data_full_name = path_build(data_path, data_fname);
    data_file = nn_fopen(data_full_name, "r", "for training or test data");
    if(verbosity > 1) printf("\nReading %s\n", data_full_name);
    npatterns = load_class_data(data_file, npatterns, nraw, ninputs,
                cpp->classes, cpp->input_matrix,
                normalization, data_full_name,
                features);
    nn_fclose(data_file, data_full_name, "for training or test data");

    /* store the real number of patterns */
    cpp->npatterns = npatterns;
    cpp->input_num = 0;

    /* return a model number */
    return new_model;
}

/* initialize for single pass operation */
int
class_data_single_init(char *data_path, char *data_fname, int npatterns,
               int nraw, int ninputs, int noutputs, int *features,
               int normalization, int verbosity)
{
    char *data_full_name;
    int new_model;
    CLASS_PATTERNS *cpp;

    /* allocate structure */
    new_model = model_get_number((void ***)&cp_list, &max_models,
                 sizeof(CLASS_PATTERNS));
    /* store what we know so far */
    cpp = cp_list[new_model];
    cpp->noutputs = noutputs;
    cpp->ninputs = ninputs;
    cpp->nraw = nraw;
    cpp->normalization = normalization;
    cpp->features = features;
    cpp->train_flag = TRUE;
    cpp->n_read = 0;
    cpp->random_flag = cpp->priors_flag = cpp->cross_valid_flag = FALSE;
    cpp->single_flag = TRUE;
    cpp->inputs = allocate_vector(ninputs);
    cpp->input_matrix = NULL;
    cpp->classes = NULL;
    cpp->npatterns = npatterns;
    cpp->input_num = -1;/* so the first one we fill is number 0 */

    /* open data file for later use */
    data_full_name = path_build(data_path, data_fname);
    cpp->data_fname = data_fname;
    cpp->data_file =
      nn_fopen(data_full_name, "r", "for training or test data");
    if(verbosity > 1) printf("\nReading %s\n", data_full_name);

    /* return the model number */
    return new_model;
}

/************************************
 * return the number of patterns that would be presented in one epoch
 * using this model.
 *******************************************/
int
class_data_get_npatterns(int data_model)
{
    int npat;
    CLASS_PATTERNS *cpp;
    cpp = cp_list[data_model];

    if(cpp->single_flag) return cpp->npatterns;
    if(cpp->priors_flag){
    if(cpp->cross_valid_flag){
        /* this is incorrect for hand split cross valid */
        npat = cpp->priors_npatterns/cpp->nfolds;
        if(!cpp->train_flag)
          /* we don't do sampling in cross valid testing */
          return cpp->n_in_test_fold[cpp->fold];
        else return(cpp->priors_npatterns - npat);
    } else return cpp->priors_npatterns;
    } else {
      if(cpp->cross_valid_flag){
    if(cpp->train_flag) return cpp->n_in_train_fold[cpp->fold];
    else return cpp->n_in_test_fold[cpp->fold];
      } else return cpp->npatterns;
    }
}
/*****************************************************
 * count the patterns in this class
 *
 * if cross valid has been initialized, cpp->cv_mask must be set in
 *   class_data_start_cv_fold() before
 *   class_data_get_npatterns_class() is called
 *
 * if using priors: calculate the number of patterns in this class as
 * probability(class)*priors_npatterns
 *******************************************************/
int
class_data_get_npatterns_class(int data_model, int class)
{
    int npat = 0, n;
    CLASS_PATTERNS *cpp;
    cpp = cp_list[data_model];

    if(cpp->single_flag) return 0;
    if(cpp->cross_valid_flag){
    if(cpp->priors_flag && cpp->train_flag){
        /* we sample cross valid train, use the standard approximation */
        npat = class_data_get_npatterns(data_model);
        npat *= cpp->priors_list[class];
    } else {
        for(n = 0; n < cpp->npatterns; n++){
        if(cpp->cv_mask[cpp->cross_index[n]] == '1')
          if(cpp->classes[n] == class) npat++;
        }
    }
    } else if(cpp->priors_flag){
    npat = class_data_get_npatterns(data_model);
    npat *= cpp->priors_list[class];
    } else {
    for(n = 0; n < cpp->npatterns; n++)
      if(cpp->classes[n] == class) npat++;
    }
    return npat;
}

/*************************
 * initialize for priors adjustment by sampling data
 **************************/
void
class_data_priors_init(int data_model, float *priors_list,
               int priors_npatterns)
{
    CLASS_PATTERNS *cpp;
    int i;
    float priors_total;

    cpp = cp_list[data_model];

    if(cpp->single_flag) return;

    cpp->priors_flag = TRUE;
    cpp->priors_npatterns = priors_npatterns;

    if(cpp->cross_valid_flag){
    cpp->fold_classes = allocate_ivector(cpp->npatterns);
    cpp->fold_indices = allocate_ivector(cpp->npatterns);
#if 0
    } else if (cpp->boot_strap_flag){
#endif
    } else {
    init_priors_data(cpp->classes, cpp->noutputs, cpp->npatterns,
             cpp->random_flag, priors_list);
    }
    if(cpp->random_flag)
      deallocate_ivector(cpp->random_index);

    /* allocate space and copy priors */
    cpp->priors_list = allocate_vector(cpp->noutputs);
    for(priors_total = 0., i = 0; i < cpp->noutputs; i++){
    priors_total +=  priors_list[i];
    cpp->priors_list[i] = priors_list[i];
    }
    /* make sure that the priors sum to 1 */
    if(priors_total  != 1.){
    for(i = 0; i < cpp->noutputs; i++)
      cpp->priors_list[i] /= priors_total;
    }
}

/*************************
 * initialize for random presentation order for each training epoch
 ************************/
void
class_data_random_init(int data_model)
{
    CLASS_PATTERNS *cpp;
    int i;

    cpp = cp_list[data_model];

    if(cpp->single_flag) return;
    cpp->random_flag = TRUE;
    if(cpp->priors_flag){
    /* reinitialize the priors with random turned on */
    if(!cpp->cross_valid_flag){
        free_priors_data();
        init_priors_data(cpp->classes, cpp->noutputs, cpp->npatterns,
                 cpp->random_flag, cpp->priors_list);
    }
#if 0
    } else if(cpp->boot_flag){
#endif
    } else {
    /* allocate space for randomized list */
    cpp->random_index = allocate_ivector(cpp->npatterns);
    for(i = 0; i < cpp->npatterns; i++) cpp->random_index[i] = i;
    }
}

/*****************
 * initialize for cross validation mode
 ******************/
void
class_data_cv_init(int data_model, int cv_nsplits, int use_cvsplits,
           int *cv_splits, int cv_nfolds, char **cv_train_mask,
           char **cv_test_mask, int cv_random)
{
    CLASS_PATTERNS *cpp;
    int i, f;
    int *index, *temp_array;

    cpp = cp_list[data_model];

    if(cpp->single_flag) return;

    cpp->cross_valid_flag = TRUE;

#if 0
    if(cpp->boot_flag){
    warning("class_data: boot strap initialization cannot be used at the same time as automatic cross validation, will only do cross validation");
    /* free the boot strap stuff */
    }
#endif

    if(cpp->priors_flag){
    /* free priors info, it must be re-done each fold */
    free_priors_data();
    cpp->fold_classes = allocate_ivector(cpp->npatterns);
    cpp->fold_indices = allocate_ivector(cpp->npatterns);
    }

    cpp->nfolds = cv_nfolds;

    if(cv_random){
    index = allocate_iindex(cpp->npatterns);
    randomize_index_list(index, cpp->npatterns);
    /* make a copy of the class list that has the random
       ordering */
    temp_array = allocate_ivector(cpp->npatterns);
    for(i = 0; i < cpp->npatterns; i++)
      temp_array[i] = cpp->classes[index[i]];
    /* for build_cross_index, replace cpp->classes with temp_array */
    /* if use_cvsplits, I ignore temp_array anyway */
    cpp->cross_index = 
      build_cross_index(cpp->noutputs, cpp->npatterns, cv_nsplits,
                temp_array, use_cvsplits, cv_splits);
    /* unsort the split assignments by first copying them into
       temp_array, then unsorting them back into cross_index */
    for(i = 0; i < cpp->npatterns; i++)
      temp_array[i] = cpp->cross_index[i];
    for(i = 0; i < cpp->npatterns; i++)
      cpp->cross_index[index[i]] = temp_array[i];
    deallocate_ivector(index);
    deallocate_ivector(temp_array);
    } else {
    /* set up and store cross index */
    cpp->cross_index =
      build_cross_index(cpp->noutputs, cpp->npatterns, cv_nsplits, 
                cpp->classes, use_cvsplits, cv_splits);
    }
    /* count patterns per fold */
    cpp->n_in_train_fold = allocate_ivector(cv_nfolds);
    cpp->n_in_test_fold = allocate_ivector(cv_nfolds);
    for(i = 0; i < cpp->npatterns; i++)
      for(f = 0; f < cv_nfolds; f++){
      if(cv_train_mask[f][cpp->cross_index[i]] == '1')
        cpp->n_in_train_fold[f]++;
      if(cv_test_mask[f][cpp->cross_index[i]] == '1')
        cpp->n_in_test_fold[f]++;
      }
    cpp->train_masks = cv_train_mask;
    cpp->test_masks = cv_test_mask;
#ifdef DEBUG
    printf("cross validation fold assignments\n");
    printf("index class split");
    for(i = 0; i < cpp->npatterns; i++){
    printf("%s\t%d\t%d\n", i, cpp->classes[i], cpp->cross_index[i]);
    }
#endif /* DEBUG */
}
#if 0
void
class_data_boot_init(data_model, nboot_patterns)
     int data_model;
     int nboot_patterns;
{
    CLASS_PATTERNS *cpp;
    int i;
    cpp = cp_list[data_model];

    if(cpp->single_flag) return;

    if(cpp->cross_valid_flag){
    warning("class_data: boot strap initialization cannot be used at the same time as automatic cross validation, will only do cross validation");
    return;
    }
    if(cpp->priors_flag){
    free_priors_data();
    cpp->fold_classes = allocate_ivector(nboot_patterns);
    }
    cpp->nboot_pat = nboot_patterns;
    cpp->train_list = allocate_ivector(nboot_patterns);
    cpp->test_list = allocate_ivector(cpp->npatterns);
    cpp->in_train = allocate_ivector(cpp->npatterns);
}

int
class_data_start_boot_fold(data_model, is_train)
     int data model,is_train;
{
    int i,j;

    if(is_train){
    for(i = 0; i < cpp->npatterns; i++) cpp->test_list[i] = 0;
    for(i = 0; i < cpp->nboot_pat; i++){
        j = random_int(0, cpp->npatterns-1);
        cpp->train_list[i] = j;
        cpp->test_list[j] = 1;
    }
    for(j = i = 0; i < cpp->npatterns; i++)
      if(cpp->test_list[i] != 0) test_list[j++] = i;
    cpp->boot_ntest = j;
    if(cpp->priors_flag){
        for(i = 0; i < cpp->npatterns; i++)
          cpp->fold_classes[i] = cpp->classes[cpp->train_list[i]];
        init_priors_data(cpp->fold_classes, cpp->noutputs, cpp->nboot_pat,
                 cpp->random_flag, cpp->priors_list);
    }
    return cpp->nboot_pat;
    } else {/* testing fold */
    return cpp->boot_ntest;
    }
}
#endif
/*************************
 * start a new training or testing fold in cross validation
 *************************/
int
class_data_start_cv_fold(int data_model, int fold, int is_train)
{
    CLASS_PATTERNS *cpp;
    int nf, n;

    cpp = cp_list[data_model];

    if(cpp->single_flag) return 0;

    cpp->n_read = 0;
/*
    if(fold >= cpp->nfolds) return 0;
*/
    cpp->fold = fold;

    /* set up current mask */
    if(is_train)
      cpp->cv_mask = cpp->train_masks[fold];
    else
      cpp->cv_mask = cpp->test_masks[fold];
    cpp->train_flag = is_train;
#ifdef DEBUG
    printf("cross validation train mask: %s\n", cpp->cv_mask);
#endif /* DEBUG */

    if(cpp->priors_flag && is_train){
    /* make up lists of the patterns in this fold and the real
       indices for those patterns */
    nf = 0;
    for(n = 0; n < cpp->npatterns; n++){
        if(cpp->cv_mask[cpp->cross_index[n]] == '1'){
        cpp->fold_classes[nf] = cpp->classes[n];
        cpp->fold_indices[nf] = n;
        nf++;
        }
    }
    /* set up priors adjustment for this fold */
    init_priors_data(cpp->fold_classes, cpp->noutputs, nf,
             cpp->random_flag, cpp->priors_list);
    }
    /* return the number of patterns this epoch */
    if(is_train)
      return cpp->n_in_train_fold[fold];
    else
      return cpp->n_in_test_fold[fold];
}

/*************************
 * Finish a cross validation training or testing fold
 ************************/
void
class_data_end_cv_fold(int data_model)
{
    if(cp_list[data_model]->single_flag) return;

    if(cp_list[data_model]->priors_flag && cp_list[data_model]->train_flag)
      free_priors_data();
}

/************************
 * Start a new epoch of training
 ************************/
void
class_data_epoch_init(int data_model)
{
    CLASS_PATTERNS *cpp;
    cpp = cp_list[data_model];

    cpp->n_read = 0;
    if(cpp->single_flag) return;
    if(cpp->random_flag && !cpp->priors_flag){
    randomize_index_list(cpp->random_index, cpp->npatterns);
    }
}

/***********************
 * get the next pattern, taking into consideration priors adjustment,
 * random presentation order, and training or testing for cross validation
 *
 * If cross validation has been initialized, class_data_start_cv_fold
 * should have been called
 *
 * If these patterns are being requested by an epoch training algorithm,
 * class_data_epoch init should have been called
 ***********************/
float *
class_data_get_next(int data_model, int *class)
{
    CLASS_PATTERNS *cpp;
    int npat, t_index, desired_class;
    cpp = cp_list[data_model];
    
    if(cpp->single_flag){
    if(cpp->n_read >= cpp->npatterns) return NULL;
    desired_class =
      get_class_line(cpp->data_file, cpp->data_fname, cpp->inputs,
             cpp->nraw, cpp->normalization, cpp->features,
             cpp->ninputs);
    if(desired_class == EOF) return NULL;
    if(desired_class < 0 || desired_class >= cpp->noutputs)
      fatal_error("%s: class of pattern (%d) is out of range (0,%d)",
              "class_data_get_next", desired_class, cpp->noutputs);
    *class = desired_class;
    cpp->input_num++;
    cpp->n_read++;
    return(&cpp->inputs[0]);
    }

    if(cpp->priors_flag){
    if(cpp->cross_valid_flag && !cpp->train_flag)
      npat = cpp->npatterns;
    else npat = class_data_get_npatterns(data_model);
    } else npat = cpp->npatterns;

    /* if this is cross valid, loop until we pick one, else do once */
    while(1){
    /* if we have read all of the data, return NULL; */
    if(cpp->n_read >= npat) return NULL;

    /* select a pattern, paying attention to random, priors, cv */
    /* if cross valid testing, do not do priors or random */
    if(cpp->cross_valid_flag && !cpp->train_flag){
        t_index = cpp->n_read;
    }
    else if(cpp->priors_flag){
        /* select a pattern from the current fold by selecting the
           class first.  If there is no data in the class
           selected, keep trying until we select a class that has
           data in it */
        /* this has the potential for being an infinite loop */
        do{
        *class = get_priors_class();
        t_index = get_priors_index(*class);
        } while (t_index == -1);
        if(cpp->cross_valid_flag)
          /* the t_index above is relative to fold_classes. we
         need an index into desired classes */
          t_index = cpp->fold_indices[t_index];
    } else if(cpp->random_flag && cpp->train_flag){
        t_index = cpp->random_index[cpp->n_read];
    } else t_index = cpp->n_read;
    cpp->n_read++;
    if(cpp->cross_valid_flag)
      if(cpp->cv_mask[cpp->cross_index[t_index]] != '1') continue;

    /* copy the class into class, return inputs */
    cpp->input_num = t_index;
    *class = cpp->classes[t_index];
    return(&cpp->input_matrix[t_index][0]);
    }
}

/***********************
 * free space allocated in class_data_read and in the initialization
 * routines
 **********************/
void
class_data_exit(int data_model)
{
    CLASS_PATTERNS *cpp;
    cpp = cp_list[data_model];

    if(cpp->single_flag){
    deallocate_vector(cpp->inputs);
    if(cpp->data_file != NULL)
      nn_fclose(cpp->data_file, "single pass data file",
            "for training or test data");
    nn_free((char *)cpp, "for data model");
    cp_list[data_model] = NULL;
    return;
    }
    deallocate_matrix(cpp->input_matrix);
    deallocate_ivector(cpp->classes);
    if(cpp->cross_valid_flag){
    deallocate_ivector(cpp->cross_index);
    deallocate_ivector(cpp->n_in_train_fold);
    deallocate_ivector(cpp->n_in_test_fold);
    }
    if(cpp->priors_flag){
    if(cpp->cross_valid_flag){
        deallocate_ivector(cpp->fold_classes);
        deallocate_ivector(cpp->fold_indices);
    } else /* priors should be freed after every cv fold */
      free_priors_data();
    deallocate_vector(cpp->priors_list);
    } else
      if(cpp->random_flag) deallocate_ivector(cpp->random_index);
    nn_free((char *)cpp, "for data model");
    cp_list[data_model] = NULL;
}

/********************
 * a routine that was added for testing purposes.
 * class_data_epoch_init should be used instead
 *********************/
void
class_data_set_n_read_zero(int dm)
{
    cp_list[dm]->n_read = 0;
}

float **
class_data_get_input_matrix(int data_model)
{
    if(cp_list[data_model]->single_flag) return NULL;
    return (cp_list[data_model]->input_matrix);
}

int
class_data_get_input_num(int data_model)
{
    return(cp_list[data_model]->input_num);
}
