/***--------------------------------------------------------------------------
* NAME:  clust_param.c
* PURPOSE:  Functions called by cluster algorithms to read and write parameters
* USAGE
*
* SEE ALSO: kmeans.c
* GLOBALS:
* DESCRIPTION:  
*
*   cluster_write_paramfile
*   cluster_read_paramfile
*   cluster_build
*   cluster_param
*
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/clust_param.c,v 11.3 2001/08/23 20:14:42 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/clust_param.c,v $
 * CREATED: 20-Mar-91 dan               MIT LINCOLN LABORATORY
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>            
#include <math.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"

typedef struct {
    int ninputs, ncenters, noutputs, cluster_by_class;
    int *ncenters_list;
    int *features;
    float ***centers_list, ***variances_list;
    float **centers, **variances;
} CLUS_PARAMS;

static CLUS_PARAMS **cluspp;
static int max_models = 0;

#define MODEL_STORE_SIZE 10

/*** Parameter Table ***/
static PARAM clusterscalars[] = {
    {"ninputs",P_INT},
    {"noutputs", P_INT},
    {"ncenters",P_INT},
    {"clus_by_class", P_FLAG},
};
#define NSCALARS (sizeof(clusterscalars)/sizeof(PARAM))
static PARAM clusterfeatures[] = {
    {"features",P_INT_VECTOR},
};
#define NFEATURES (sizeof(clusterfeatures)/sizeof(PARAM))
static PARAM cluster_ncenters_list[] = {
    {"ncenters_list",P_INT_VECTOR},
};
#define NCENTERS_LIST (sizeof(cluster_ncenters_list)/sizeof(PARAM))
static PARAM clusterarrays[] = {
    {"pcenter",P_FLOAT_MATRIX},
    {"pvariance",P_FLOAT_MATRIX},
};
#define NARRAYS (sizeof(clusterarrays)/sizeof(PARAM))


int
cluster_build(int ninputs, int ncenters, int noutputs, int cluster_by_class,
          int *ncenters_list)
{
    int i, new_model;
    CLUS_PARAMS *cpp;

    new_model = model_get_number((void ***)&cluspp,
                 &max_models, sizeof(CLUS_PARAMS));
    cpp = cluspp[new_model];

    cpp->ninputs = ninputs;
    cpp->ncenters = ncenters;
    cpp->noutputs = noutputs;
    cpp->cluster_by_class = cluster_by_class;

    if(ninputs < 0) ninputs = 0;
    cpp->features = allocate_ivector(ninputs+1);

    if(cluster_by_class){
    if(noutputs <= 0){
        cpp->ncenters_list = NULL;
        cpp->centers_list = NULL;
        cpp->variances_list = NULL;
    } else {
        cpp->ncenters_list = allocate_ivector(noutputs);
        cpp->centers_list = (float ***)allocate_pvector(noutputs);
        cpp->variances_list = (float ***)allocate_pvector(noutputs);
        for(i = 0; i < noutputs; i++){
        cpp->ncenters_list[i] = ncenters_list[i];
        if(ncenters_list[i] <= 0){
            cpp->centers_list[i] = NULL;
            cpp->variances_list[i] = NULL;
        } else {
            cpp->centers_list[i] = allocate_matrix(ncenters_list[i],
                               ninputs);
            cpp->variances_list[i] =
              allocate_matrix(ncenters_list[i], ninputs);
        }
        }
        cpp->centers = cpp->variances = NULL;
    }
    } else {
    if(ncenters <= 0){
        cpp->centers = NULL;
        cpp->variances = NULL;
    } else {
        cpp->centers = allocate_matrix (ncenters, ninputs);
        cpp->variances = allocate_matrix (ncenters, ninputs);
    }
    cpp->centers_list = cpp->variances_list = NULL;
    cpp->ncenters_list = NULL;
    }
    return(new_model);
}

void
cluster_init(int model, int *features)
{
    CLUS_PARAMS *cpp;
    int i;
    cpp = cluspp[model];
    /* copy faetures (array length is included in array) */
    for(i = 0; i <= features[0]; i++){
    cpp->features[i] = features[i];
    }
}

void
cluster_free(int model)
{
    CLUS_PARAMS *cpp;
    int i;

    cpp = cluspp[model];

    deallocate_ivector(cpp->features);
    if(cpp->cluster_by_class){
    deallocate_ivector(cpp->ncenters_list);
    for(i = 0; i < cpp->noutputs; i++){
        deallocate_matrix(cpp->centers_list[i]);
        deallocate_matrix(cpp->variances_list[i]);
    }
    deallocate_pvector((void **)cpp->centers_list);
    deallocate_pvector((void **)cpp->variances_list);
    } else {
    deallocate_matrix(cpp->centers);
    deallocate_matrix(cpp->variances);
    }
    nn_free((char *)cpp, "for cluster model");
    cluspp[model] = NULL;
}

/***--------------------------------------------------------------------------
* NAME:  cluster_write_paramfile()
* DESCRIPTION:
*   Takes a file pointer as input.
*   Writes cluster model to parameter file.
*   DOES NOT USE MODEL NUMBERS!!
*   IT IS NOT NECESSARY TO CALL cluster_build BEFORE USING
* NOTES:
*   cluster model is written to parameter file in three or more blocks.
*   The first block contains the scalars:
*       ninputs
*       ncenters
*       noutputs
*       cluster_by_class
*   The second block contains the vector:
*       features
*   If we are clustering by class,
*        the third block contains the vector:
*           ncenters_list
*       The next noutputs blocks contains the centers for each class:
*           pcenter
*           pvariance
*   otherwise (not clustering by class)
*       The third block contains the centers and variances:
*           pcenter
*           pvariance
--------------------------------------------------------------------------***/
void
cluster_write_model(int model, FILE *fp)
{
    CLUS_PARAMS *cpp;

    cpp = cluspp[model];

    if(cpp->cluster_by_class){
    cluster_write_paramfile(fp, cpp->ninputs, cpp->ncenters, cpp->noutputs,
                cpp->centers_list, cpp->variances_list,
                cpp->features, cpp->cluster_by_class,
                cpp->ncenters_list);
    } else {
    cluster_write_paramfile(fp, cpp->ninputs, cpp->ncenters, cpp->noutputs,
                &cpp->centers, &cpp->variances,
                cpp->features, cpp->cluster_by_class,
                cpp->ncenters_list);
    }
}

int
cluster_write_paramfile (FILE *fp, int ninputs, int ncenters, int noutputs,
             float ***pcenter, float ***pvariance, int *features,
             int cluster_by_class, int *ncenters_list)
{
    int rv, i;

    param_setvarptr("ninputs", &ninputs, clusterscalars, NSCALARS);
    param_setvarptr("ncenters", &ncenters, clusterscalars, NSCALARS);
    param_setvarptr("noutputs", &noutputs, clusterscalars, NSCALARS);
    param_setvarptr("clus_by_class", &cluster_by_class,
            clusterscalars, NSCALARS);

    if(cluster_by_class){
    /* we never trust ncenters when by class */
    ncenters = 0;
    for(i = 0; i < noutputs; i++){
        ncenters += ncenters_list[i];
    }
    }
    if ((rv = param_putblk(fp, clusterscalars, NSCALARS)) < 0) return rv;

    param_setvarptr("features", features, clusterfeatures, NFEATURES);
    if(features != NULL)
      param_setdim("features", features[0]+1, 0, clusterfeatures, NFEATURES);
    else
      param_setdim("features", 0, 0, clusterfeatures, NFEATURES);
    if ((rv = param_putblk(fp, clusterfeatures, NFEATURES)) < 0) return rv;

    if(cluster_by_class){
    param_setvarptr("ncenters_list", ncenters_list, cluster_ncenters_list,
            NCENTERS_LIST);
    param_setdim("ncenters_list", noutputs, 0, cluster_ncenters_list,
             NCENTERS_LIST);
    if((rv = param_putblk(fp, cluster_ncenters_list, NCENTERS_LIST)) < 0)
      return rv;

    for(i = 0; i < noutputs; i++){
        fprintf(fp, "clusters for class %d\n", i);
        /* Set ptrs and dimensions for allocated matrices */
        if(pcenter[i] != NULL)
          param_setvarptr("pcenter",
                  (int *)&pcenter[i][0][0],clusterarrays,NARRAYS);
        else
          param_setvarptr("pcenter",
                  (int *)NULL,clusterarrays,NARRAYS);
        if(pvariance[i] != NULL)
          param_setvarptr("pvariance",
                  (int *)&pvariance[i][0][0],clusterarrays,NARRAYS);
        else
          param_setvarptr("pvariance",
                  (int *)NULL,clusterarrays,NARRAYS);
        param_setdim("pcenter",
            ncenters_list[i], ninputs, clusterarrays, NARRAYS);
        param_setdim("pvariance",
             ncenters_list[i], ninputs, clusterarrays, NARRAYS);
        if ((rv = param_putblk(fp,clusterarrays,NARRAYS)) < 0) return rv;
    }
    } else {
    if(pcenter[0] != NULL)
        param_setvarptr("pcenter",
                (int *)&pcenter[0][0][0],clusterarrays,NARRAYS);
    else
        param_setvarptr("pcenter",
                (int *)NULL,clusterarrays,NARRAYS);
    if(pvariance[0] != NULL)
        param_setvarptr("pvariance",
                (int *)&pvariance[0][0][0],clusterarrays,NARRAYS);
    else
        param_setvarptr("pvariance",
                (int *)NULL,clusterarrays,NARRAYS);
    param_setdim("pcenter",
             ncenters, ninputs, clusterarrays, NARRAYS);
    param_setdim("pvariance",
             ncenters, ninputs, clusterarrays, NARRAYS);
    if ((rv = param_putblk(fp, clusterarrays, NARRAYS)) < 0) return rv;
    }

    return 0;
    
} /* cluster_write_paramfile() */


/***---------------------------------------------------------------------------
* NAME:  cluster_read_paramfile()
* DESCRIPTION:
*   Takes a file pointer as input.
*   Reads cluster model from parameter file.
*   Builds cluster model using parameters.
*   Returns cluster model number.
* NOTES:
*   See cluster_write_paramfile().
--------------------------------------------------------------------------***/
int
cluster_read_paramfile (FILE *fp, int *ninputs, int *noutputs, int *ncenters,
            int *features)
{
    int rv, i, cluster_by_class, *pncenters_list, model;
    CLUS_PARAMS *cpp;

    /* Get the cluster model scalars */
    param_setvarptr("ninputs", ninputs, clusterscalars, NSCALARS);
    param_setvarptr("ncenters", ncenters, clusterscalars, NSCALARS);
    param_setvarptr("noutputs", noutputs, clusterscalars, NSCALARS);
    param_setvarptr("clus_by_class", &cluster_by_class,
            clusterscalars, NSCALARS);
    if ((rv = param_getblk(fp,clusterscalars,NSCALARS)) < 0) return rv;

    /* read the features */
    /* the space is already allocated for them */
    /* copy ninputs into i.  I don't like dereferencing pointers */
    i = *ninputs;
    param_setvarptr("features", features, clusterfeatures, NFEATURES);
    param_setdim("features", i+1, 0, clusterfeatures, NFEATURES);
    if ((rv = param_getblk(fp,clusterfeatures,NFEATURES)) < 0) return rv;

    /* read ncenters_list */
    pncenters_list = NULL;
    if(cluster_by_class){
        pncenters_list = allocate_ivector(*noutputs);
        param_setvarptr("ncenters_list", pncenters_list,
                cluster_ncenters_list, NCENTERS_LIST);
        param_setdim("ncenters_list", *noutputs, 0, cluster_ncenters_list,
             NCENTERS_LIST);
        if((rv = param_getblk(fp,cluster_ncenters_list, NCENTERS_LIST))
           < 0) return rv;
    }

    /* Allocate the space needed to store the new model */
    model = cluster_build (*ninputs, *ncenters, *noutputs,
                   cluster_by_class, pncenters_list);
    cluster_init(model, features);

    cpp = cluspp[model];

    if(cluster_by_class){
        for(i = 0; i < *noutputs; i++){
        /* Set ptrs and dimensions for allocated matrices */
        if(cpp->centers_list[i] != NULL){
            param_setvarptr("pcenter",
                    (int *)&(cpp->centers_list[i][0][0]),
                    clusterarrays,NARRAYS);
        } else {
            param_setvarptr("pcenter",
                    (int *)NULL,
                    clusterarrays,NARRAYS);
        }
        if(cpp->variances_list[i] != NULL){
            param_setvarptr("pvariance",
                    (int *)&(cpp->variances_list[i][0][0]),
                    clusterarrays, NARRAYS);
        } else {
            param_setvarptr("pvariance",
                    (int *)NULL,
                    clusterarrays, NARRAYS);
        }
        param_setdim("pcenter", pncenters_list[i], *ninputs,
                 clusterarrays, NARRAYS);
        param_setdim("pvariance", pncenters_list[i], *ninputs,
                 clusterarrays, NARRAYS);

        /* Get the cluster model arrays */
        if ((rv = param_getblk(fp,clusterarrays,NARRAYS)) < 0)
          return rv;
        }
        deallocate_ivector(pncenters_list);
    } else {
        /* Set ptrs and dimensions for allocated matrices */
        if(cpp->centers != NULL){
        param_setvarptr("pcenter",
                (int *)&(cpp->centers[0][0]),
                clusterarrays, NARRAYS);
        } else {
        param_setvarptr("pcenter",
                (int *)NULL,
                clusterarrays, NARRAYS);
        }
        if(cpp->variances != NULL){
        param_setvarptr("pvariance",
                (int *)&(cpp->variances[0][0]),
                clusterarrays, NARRAYS);
        } else {
        param_setvarptr("pvariance",
                (int *)NULL,
                clusterarrays, NARRAYS);
        }
        param_setdim("pcenter",
             *ncenters, *ninputs, clusterarrays, NARRAYS);
        param_setdim("pvariance",
             *ncenters, *ninputs, clusterarrays, NARRAYS);

        /* Get the cluster model arrays */
        if ((rv = param_getblk(fp, clusterarrays, NARRAYS)) < 0) return rv;
    }
    return model;
} /* cluster_read_paramfile() */

/***---------------------------------------------------------------------------
* NAME:  cluster_read_model()
* DESCRIPTION:
*   Takes a file pointer as input.
*   Reads cluster model from parameter file.
*   Builds cluster model using parameters.
*   Returns cluster model number.
* NOTES:
*   See cluster_write_model().
--------------------------------------------------------------------------***/
int cluster_read_model (FILE *fp)
{
    int ninputs, ncenters, noutputs, cluster_by_class;
    int *features;
    int rv, i, *pncenters_list, model;
    CLUS_PARAMS *cpp;

    /* Get the cluster model scalars */
    param_setvarptr("ninputs", &ninputs, clusterscalars, NSCALARS);
    param_setvarptr("ncenters", &ncenters, clusterscalars, NSCALARS);
    param_setvarptr("noutputs", &noutputs, clusterscalars, NSCALARS);
    param_setvarptr("clus_by_class", &cluster_by_class,
            clusterscalars, NSCALARS);
    if ((rv = param_getblk(fp, clusterscalars, NSCALARS)) < 0) return rv;

    /* read the features */
    features = allocate_ivector(ninputs+1);
    param_setvarptr("features", features, clusterfeatures, NFEATURES);
    param_setdim("features", ninputs+1, 0, clusterfeatures, NFEATURES);
    if ((rv = param_getblk(fp, clusterfeatures, NFEATURES)) < 0) return rv;

    /* read ncenters_list */
    pncenters_list = NULL;
    if(cluster_by_class){
    pncenters_list = allocate_ivector(noutputs);
    param_setvarptr("ncenters_list", pncenters_list,
            cluster_ncenters_list, NCENTERS_LIST);
    param_setdim("ncenters_list", noutputs, 0, cluster_ncenters_list,
             NCENTERS_LIST);
    if((rv = param_getblk(fp, cluster_ncenters_list, NCENTERS_LIST))
       < 0) return rv;
    }

    /* Allocate the space needed to store the new model */
    model = cluster_build (ninputs, ncenters, noutputs,
               cluster_by_class, pncenters_list);
    cluster_init(model, features);

    cpp = cluspp[model];

    if(cluster_by_class){
    for(i = 0; i < noutputs; i++){
        /* Set ptrs and dimensions for allocated matrices */
        param_setvarptr("pcenter",
                (int *)&(cpp->centers_list[i][0][0]),
                clusterarrays,NARRAYS);
        param_setvarptr("pvariance",
                (int *)&(cpp->variances_list[i][0][0]),
                clusterarrays, NARRAYS);
        param_setdim("pcenter", pncenters_list[i], ninputs,
             clusterarrays, NARRAYS);
        param_setdim("pvariance", pncenters_list[i], ninputs,
             clusterarrays, NARRAYS);

        /* Get the cluster model arrays */
        if ((rv = param_getblk(fp,clusterarrays,NARRAYS)) < 0)
          return rv;
    }
    deallocate_ivector(pncenters_list);
    } else {
    /* Set ptrs and dimensions for allocated matrices */
    param_setvarptr("pcenter",
            (int *)&(cpp->centers[0][0]),
            clusterarrays, NARRAYS);
    param_setvarptr("pvariance",
            (int *)&(cpp->variances[0][0]),
            clusterarrays, NARRAYS);
    param_setdim("pcenter",
             ncenters, ninputs, clusterarrays, NARRAYS);
    param_setdim("pvariance",
             ncenters, ninputs, clusterarrays, NARRAYS);

    /* Get the cluster model arrays */
    if ((rv = param_getblk(fp, clusterarrays, NARRAYS)) < 0) return rv;
    }
    return model;
} /* cluster_read_model() */

#define PEQ(y) (0==strcmp(param_name,y))

void *
cluster_param(int model, char *param_name, int class_num)
{
    void *param_address;
    CLUS_PARAMS *cpp;

    cpp = cluspp[model];
    if(PEQ("ninputs"))
      param_address = &cpp->ninputs;
    else if(PEQ("noutputs"))
      param_address = &cpp->noutputs;
    else if(PEQ("ncenters"))
      param_address = &cpp->ncenters;
    else if(PEQ("cluster_by_class"))
      param_address = &cpp->cluster_by_class;
    else if(PEQ("ncenters_list"))
      param_address = &cpp->ncenters_list;
    else if(PEQ("pcenter")) {
      if(cpp->cluster_by_class) param_address = &cpp->centers_list[class_num];
      else param_address = &cpp->centers;
    } else if(PEQ("pvariance")) {
      if(cpp->cluster_by_class)
    param_address = &cpp->variances_list[class_num];
      else param_address = &cpp->variances;
    } else if(PEQ("features"))
      param_address = &cpp->features;
    else {
    warning("cluster_param: Parameter %s is not defined", param_name);
    param_address = NULL;
    }
    return param_address;
}

/* change a model with clusters by class to one with all centers in
   one matrix */
void
cluster_merge_centers(int model)
{
    CLUS_PARAMS *cpp;
    int i, j, k, l;

    cpp = cluspp[model];

    if(cpp->cluster_by_class){
    cpp->variances = allocate_matrix(cpp->ncenters, cpp->ninputs);
    cpp->centers = allocate_matrix(cpp->ncenters, cpp->ninputs);
    for (i = 0, l = 0; i < cpp->noutputs; i++){
        for(j = 0; j < cpp->ncenters_list[i] && l < cpp->ncenters;
        j++, l++){
        for(k = 0; k < cpp->ninputs; k++){
            cpp->centers[l][k] = cpp->centers_list[i][j][k];
            cpp->variances[l][k] = cpp->variances_list[i][j][k];
        }
        }
        deallocate_matrix(cpp->centers_list[i]);
        deallocate_matrix(cpp->variances_list[i]);
    }
    deallocate_pvector((void **)cpp->centers_list);
    deallocate_pvector((void **)cpp->variances_list);
    deallocate_ivector(cpp->ncenters_list);
    cpp->centers_list = cpp->variances_list = NULL;
    cpp->ncenters_list = NULL;
    cpp->cluster_by_class = 0;
    }
}
