/***---------------------------------------------------------------------------
* NAME:     cross.c
* PURPOSE:  subroutines for handling cross validation
*
* USAGE:
*
* DESCRIPTION: 
*       build_cross_index
*       build_cross_valid_masks
*       cross_valid_check_args
*       
* BUGS AND LIMITATIONS:
*
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/cross.c,v 11.5 2001/08/23 20:14:43 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/cross.c,v $
 * CREATED: 13-May-92 lck               MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include "nnlib.h"

#define MAXSTR (256)

/***********************
  build_cross_index(nclasses, n_trials, n_folds, desired_classes, use_splits
            splits)
  Variables for assigning points automatically
     nclasses -- number of classes in the data
     n_trials -- length of desired_classes array
     n_folds -- number of folds to split things into
     desired_classes -- classes of points to be divided
  Variables for user specified fold assignment
    use_splits -- flag telling us to use the user's divisions
    splits -- array containing the user's divisions
    splits[0] must equal 2*nfolds.  if it does not, the points
        will be assigned automatically instead
    splits[2*i+1] and splits[2*i+2] contain the start and end
        indices of fold i

  there are now two versions of this routine.  One uses xvalid splits
  supplied and simply assignes the folds as requested.
  the other, described below, assigns points to folds

  Automatic assignment of points to folds:
    the idea in this routine is that we have nclasses piles of beans
    and n_folds bins to put them in.

  Starting with the first bean pile, put one bean in each bin, cycling
  through bins, until we run out of beans.  Then start with the next
  pile of beans, starting with the bin where we left off.
  a a   a   a   4 bins, 5 beans of each type, 3 types
  a b   b   b       ==
  b b   c   c   4 folds, 5 vectors per class, 3 classes
  c c   c
  Since doing the assignment would require going through all of the data
  nclasses times, (because the beans don't come pre-sorted in their piles)
  I've added some complications that let us run through the data only twice.
  First you calculate how many vectors there are in each class.
  Then you figure out where the first vector from each class would fall
  if we did it the easy/slow way. Then you go through the data, putting
  vectors in folds and updating an array that tells you where the next
  vector of each class is going
  Easy, right?
  ***********************/
int *
build_cross_index(int nclasses, int n_trials, int n_splits,
          int *desired_classes, int use_splits, int *splits)
{
    register int i, j, class, rem, total;
    int *class_cnt, *cross_index;

    cross_index = allocate_ivector(n_trials);
    
    if(use_splits){
    if(splits[0] == 2*n_splits){
        for(i = 0; i < n_splits; i++){
        /* if start > end, nothing goes in that fold */
        if(splits[2*i+1] == -1 && splits[2*i+2] == -1)
          /* empty split */
          continue;
        for(j = splits[2*i+1];
            j <= splits[2*i+2] && j < n_trials; j++){
            cross_index[j] = i;
        }
        }
        return(cross_index);
    }
    }
    class_cnt = allocate_ivector(nclasses);
    
    /* find the number of patterns in each class */
    for (i = 0; i < n_trials; i++) class_cnt[desired_classes[i]]++;
    
    /* find the starting position for each class */
    for (i = 0, total = 0; i < nclasses; i++){
    rem = total % n_splits;
    total += class_cnt[i];
    class_cnt[i] = rem;
    }
    /* build an index for splitting patterns into folds */
    for(i = 0; i < n_trials; i++){
    class = desired_classes[i];
    if(class >= nclasses)
      fatal_error("build_cross_index: class number %d bad, nclasses $d",
              class, nclasses);
    cross_index[i] = class_cnt[class];
    class_cnt[class] = (class_cnt[class] + 1) % n_splits;
    }
    
    deallocate_ivector(class_cnt);

    return(cross_index);
}
/***********************
  build_map_cross_index(nclasses, n_trials, n_folds, use_splits splits)
  Variables for assigning points automatically
     nclasses -- number of classes in the data
     n_trials -- length of desired_classes array
     n_folds -- number of folds to split things into
  Variables for user specified fold assignment
    use_splits -- flag telling us to use the user's divisions
    splits -- array containing the user's divisions
    splits[0] must equal 2*nfolds.  if it does not, the points
        will be assigned automatically instead
    splits[2*i+1] and splits[2*i+2] contain the start and end
        indices of fold i

  there are now two versions of this routine.  One uses xvalid splits
  supplied and simply assignes the folds as requested.
  the other assigns points to folds evenly

  ***********************/
int *
build_map_cross_index(int nclasses, int n_trials, int n_splits,
              int use_splits, int *splits)
{
    register int i, j;
    int *cross_index;

    cross_index = allocate_ivector(n_trials);
    
    if(use_splits){
    if(splits[0] == 2*n_splits){
        for(i = 0; i < n_splits; i++){
        /* if start > end, nothing goes in that fold */
        if(splits[2*i+1] == -1 && splits[2*i+2] == -1)
          /* empty split */
          continue;
        for(j = splits[2*i+1];
            j <= splits[2*i+2] && j < n_trials; j++){
            cross_index[j] = i;
        }
        }
        return(cross_index);
    }
    }
    /* build an index for splitting patterns into folds */
    for(j = i = 0; i < n_trials; i++){
    cross_index[i] = j;
    j = (j + 1) % n_splits;
    }
    return(cross_index);
}

/* build_cross_valid_masks(nfolds, train_mask, test_mask)
 *  nfolds - number of folds to create masks for
 *      also the length of the strings used as masks
 *  train_mask, test_mask - nnargs style array of pointers to char
 *      the strings reference by that array need to be allocated
 *
 * these masks must have their first entry be the array length, as with
 * nnargs lists
 *
 * PURPOSE: produces masks for cross validation testing.
 * Example:
 * build_cross_valid_masks(4, train_mask, test_mask)
 * ==> train_mask = {"4", "0111", "1011", "1101", "1110", NULL, NULL, ...};
 *     test_mask = {"4", "1000", "0100", "0010", "0001", NULL, NULL, ...};
 *
 * BUGS AND WARNINGS:
 *  the mask strings are not deallocated by anything yet
 */
void
build_cross_valid_masks(int nfolds, char **train_mask, char **test_mask)
{
    int i, j;

    train_mask[0] = nn_malloc(nfolds/10 + 2, "for xvalid train_mask");
    sprintf(train_mask[0], "%d", nfolds);
    test_mask[0] = nn_malloc(nfolds/10 + 2, "for xvalid test_mask");
    sprintf(test_mask[0], "%d", nfolds);

    for(i = 1; i <= nfolds; i++){
    train_mask[i] = nn_malloc(nfolds+1, "for xvalid train_mask");
    test_mask[i] = nn_malloc(nfolds+1, "for xvalid test_mask");
    for(j = 0; j < nfolds; j++){
        train_mask[i][j] = '1';
        test_mask[i][j] = '0';
    }
    train_mask[i][nfolds] = test_mask[i][nfolds] = '\0';
    train_mask[i][i-1] = '0';
    test_mask[i][i-1] = '1';
#if DEBUG
      printf("train_mask: %s, test_mask: %s\n",
         train_mask[i], test_mask[i]);
#endif
    }
}

void
cross_valid_check_args(int *pnfolds, int *pnsplits, int *cv_splits,
               char **cv_train_mask, char **cv_test_mask)
{
    char *sub_name = "cross_valid_check_args";
    int train_length, test_length, i;

    /* check the number of folds in the masks */
    train_length = atoi(cv_train_mask[0]);
    test_length = atoi(cv_test_mask[0]);
    if(train_length != test_length)
      fatal_error("%s: %s. train %d test %d",
          sub_name, "different number of train and test folds",
          train_length, test_length);
    /* check the statied number of folds */
    if(*pnfolds == 0) *pnfolds = train_length;
    if(*pnfolds != train_length)
      fatal_error("%s: %s. train_folds %d nfolds %d",
          sub_name,
          "nfolds does not agree with train and test mask folds",
          train_length, *pnfolds);
    
    /* check the number of splits */
    if(cv_splits[0]%2)
      fatal_error("%s: error in cv_split list length: length %d",
          sub_name, cv_splits[0]);
    /* check the stated number of splits */
    if(*pnsplits == 0) *pnsplits = cv_splits[0]/2;
    if(*pnsplits != cv_splits[0]/2)
      fatal_error("%s: %s. nsplits %d cv_split_length %d/2",
          sub_name, "nsplits and cv_splits list length do not agree",
          *pnsplits, cv_splits[0]);
    /* check the split endpoints for negative numbers
       and negative split lengths */
    for(i = 1; i <= *pnsplits; i+=2){
    if(cv_splits[i] > cv_splits[i+1])
      fatal_error("%s: bad pair in cv_splits: %d:%d", sub_name,
              cv_splits[i], cv_splits[i+1]);
    if(cv_splits[i] < 0 || cv_splits[i+1] < 0){
        /* if both entries are equal to -1, this is an empty fold and ok
           otherwise, this is an error */
        if(cv_splits[i] != cv_splits[i+1] || cv_splits[i] != -1){
        fatal_error("%s: negative array indices in cv_splits: %d:%d",
                sub_name, cv_splits[i], cv_splits[i+1]);
        }
    }
    }
        
    /* check the length of the strings in the masks */
    for(i = 1; i <= *pnfolds; i++){
    if(strlen(cv_train_mask[i]) != *pnsplits)
       fatal_error("%s: incorrect mask length in train_mask[%d]. %d",
         sub_name, i, strlen(cv_train_mask[i]));
    if(strlen(cv_test_mask[i]) != *pnsplits)
       fatal_error("%s: incorrect mask length in test_mask[%d]. %d",
         sub_name, i, strlen(cv_test_mask[i]));
    }
}
