/***---------------------------------------------------------------------------
* NAME:		distance.c 
* PURPOSE:	distance calculating routines 
*
* DESCRIPTION:
*   l2dist() 	    - compute squared Euclidean distance between two vectors 
*   l2dist_quit()   - same as l2dist but stop accumulating if sum exceeds thresh
*   l2dist_weight() - compute weighted squared Euclidean distance
*   l2dist_quit_weight() - same as l2dist_weight but stop if sum exceeds thresh 
*   logadd()		- return log10(x+y) given logx and logy
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/distance.c,v 11.3 2001/08/23 20:14:43 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/distance.c,v $
* CREATED: 26-July-90 rpl                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#include "nnlib.h"

/***---------------------------------------------------------------------------
* NAME:		l2dist
* PURPOSE:	Calculate squared L2 distance between two float vectors 
--------------------------------------------------------------------------***/
float
l2dist(register int n, register float *vector1, register float *vector2)
{
  register float diff,total_diff;

  total_diff=0.0;

  /* compute total squared distance */
  while(n)
	  {
	  diff = *vector1 - *vector2;
	  diff *= diff; 
	  total_diff += diff;
	  vector1++;
	  vector2++;
	  n--;
	  }
  return(total_diff);
}

/***---------------------------------------------------------------------------* NAME:         l2dist_quit
* PURPOSE:      Calculate squared L2 distance between two float vectors
*                       for speed:  only sum while dist < max
--------------------------------------------------------------------------***/
float
l2dist_quit(register int n, register float *vector1, register float *vector2,
	    float max)
{
    register float diff,total_diff;

    total_diff=0.0;

    /* compute total squared distance */
    while(n) 
	{
	diff = *vector1 - *vector2;
	diff *= diff;
	total_diff += diff;
	if(total_diff > max) 
		break;
	vector1++;
	vector2++;
	n--;
	}
    return(total_diff);
}

/***---------------------------------------------------------------------------
* NAME:	     l2dist_weight
* PURPOSE:   Calculate weighted squared L2 distance between two float vectors 
--------------------------------------------------------------------------***/
float
l2dist_weight(register int n, register float *vector1,
	      register float *vector2, register float *weight)
{
  register float diff,total_diff;

  total_diff=0.0;

  /* compute total squared distance */
  while(n)
	  {
	  diff = *vector1 - *vector2;
	  diff *= diff; 
	  total_diff += *weight * diff;
	  vector1++;
	  vector2++;
	  weight++;
	  n--;
	  }
  return(total_diff);
}

/***---------------------------------------------------------------------------
* NAME:         l2dist_quit_weight
* PURPOSE:      Calculate weighted squared L2 distance between two float vectors
*                       for speed:  only sum while dist < max
--------------------------------------------------------------------------***/
float
l2dist_quit_weight(register int n, register float *vector1,
		   register float *vector2, register float *weight,
		   float max)
{
    register float diff,total_diff;

    total_diff=0.0;

    /* compute total squared distance */
    while(n) 
	{
	diff = *vector1 - *vector2;
	diff *= diff;
	total_diff += *weight * diff;
	if(total_diff > max) 
		break;
	vector1++;
	vector2++;
	weight++;
	n--;
	}
    return(total_diff);
}

/***---------------------------------------------------------------------------
* NAME:		logadd
* PURPOSE:	return log10(x+y) given log10(x) and log10(y)
* NOTES:	uses C library log and not a lookup table
--------------------------------------------------------------------------***/
float
logadd(register float x,register float y)
{
	register double b;

	if (x < y)         /*** exchange if x<y ***/
		b=x, x=y, y=b;

	b = x-y;
	if (b > 16.0)     /** keep up to 16 decimal places of precision **/
		return(x);
	else
		return(y + (float)log10(1.0 + nn_exp10(b)));
} 

/***---------------------------------------------------------------------------
* NAME:		lnadd
* PURPOSE:	return log(x+y) given log(x) and log(y)
* NOTES:	uses C library log and not a lookup table
--------------------------------------------------------------------------***/
float lnadd(register float x,register float y)
{
    register double b;

    if (x < y)         /*** exchange if x<y ***/
      b=x, x=y, y=b;

    b = x-y;
    if (b > 16.0)     /** keep up to 16 decimal places of precision **/
      return(x);
    else
      return(y + (float)log(1.0 + exp(b)));
} 
