/***--------------------------------------------------------------------------
* NAME:  report.c
* PURPOSE:  Error accumulation, write to file and reporting functions
* USAGE
*
* SEE ALSO: mlp.c
* GLOBALS:
* DESCRIPTION:
*
*	error_report_build()	   Allocate storage for error reporting
*	error_report_free()	   Free storage for error reporting
*	error_report_init()	   Initialize error reporting variables
*	error_report_part_build()  Allocate storage for partial report
*	error_report_part_free()   Free storage for partial report
*	error_report_part_init()   Initialize variables for partial report
*	error_report_calc_rmserr() Calculate rms_err for use in reports
*	error_report_class_trial() Update errors for a single trial,
*					for classification problem (!map_flag)
*	error_report_map_trial()   Update errors for a single trial,
*					for mapping problem (map_flag)
*	error_report_overall()	   Report overall and class by class statistics
*	error_report_part()	   Report partial class by class statistics
*
------------------------------------------------------------------------***/
static char     copyright[] =
"***************************************************************************\n\
                Copyright 1990 MIT Lincoln Laboratory.\n\
This software is not for release and should neither be copied\n\
nor distributed without explicit permission of MIT Lincoln Laboratory\n\
***************************************************************************\n";

static char     rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/ereport.c,v 8.1 2001/08/23 20:14:43 craig Exp $";


#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <strings.h>
#include "nnlib.h"

#define TRUE 1
#define FALSE 0
#define TABLE_VERBOSITY 0
#define CONFUSION_VERBOSITY 1
#define SMALLEST 1.e-30

static int      error_cnt, *class_error_cnt, **class_confusion_matrix;
static int      part_total_error_cnt, *part_error_cnt, **part_confusion_matrix;
static int      trial_cnt, *class_trial_cnt;
static int      part_total_trial_cnt, *part_trial_cnt;
static float    rmserr_sum, *class_rmserr_sum;
static float    part_total_rmserr_sum, *part_rmserr_sum;
static float   *desired_outputs;/* desired output vector */
static int      map_flag, verbosity, n_inputs, n_outputs = -1, part_verbosity;

/*****************************************************************************
* NAME: error_report_build ()
* DESCRIPTION:
*       Allocate storage for error reporting
*
******************************************************************************/
void
error_report_build (verbosity_p, n_outputs_p)
    int             verbosity_p, n_outputs_p;
{
				/* set static global variables */
    n_outputs = n_outputs_p;
    verbosity = verbosity_p;
    
    /* allocate space for error tables and mapping outputs*/
			
    desired_outputs = allocate_vector(n_outputs);
    
    if (verbosity > TABLE_VERBOSITY && n_outputs > 0) {
	class_error_cnt = allocate_ivector(n_outputs);
	class_trial_cnt = allocate_ivector(n_outputs);
	class_rmserr_sum = allocate_vector(n_outputs);
	if (verbosity > CONFUSION_VERBOSITY)
	    class_confusion_matrix =
		    allocate_imatrix(n_outputs + 1, n_outputs);
    }

#if DEBUG
	printf("error_report_build -- %d %d\n", verbosity, n_outputs);
#endif
}

/*****************************************************************************
* NAME: error_report_free ()
* DESCRIPTION:
*       Free storage for error reporting
*
******************************************************************************/
void
error_report_free()
{
    deallocate_vector(desired_outputs);
    
    if (verbosity > TABLE_VERBOSITY && n_outputs > 0) {
	deallocate_ivector(class_error_cnt);
	deallocate_ivector(class_trial_cnt);
	deallocate_vector(class_rmserr_sum);
	if (verbosity > CONFUSION_VERBOSITY)
	    deallocate_imatrix(class_confusion_matrix);
    }
}

/*****************************************************************************
* NAME: error_report_init ()
* DESCRIPTION:
*       Initialize error reporting variables
*
******************************************************************************/
void
error_report_init (map_flag_p, n_inputs_p)
    int            map_flag_p, n_inputs_p;
{
    int		   i, j;
				/* set static global variables */
    trial_cnt = 0;
    error_cnt = 0;
    rmserr_sum = 0.0;
    n_inputs = n_inputs_p;
    map_flag = map_flag_p;
    if (verbosity > TABLE_VERBOSITY && n_outputs > 0) {
        for (i = 0; i < n_outputs; i++)
	    class_error_cnt[i] = class_trial_cnt[i] = class_rmserr_sum[i] = 0;
	if (verbosity > CONFUSION_VERBOSITY)
            for (i = 0; i < n_outputs + 1; i++)
                for (j = 0; j < n_outputs; j++)
		    class_confusion_matrix[i][j] = 0;
    }

#if DEBUG
	printf("error_report_init -- %d %d\n", map_flag, n_inputs);
#endif
}

/*****************************************************************************
* NAME: error_report_part_build ()
* DESCRIPTION:
*       Allocate storage for error reporting for partial results
*	Must be done after error_report_init()
*
******************************************************************************/
void
error_report_part_build (verbosity_p)
    int                  verbosity_p;
{
				/* set static global variables */
    part_verbosity = verbosity_p;
    
    if (part_verbosity > TABLE_VERBOSITY && n_outputs > 0) {
	part_error_cnt = allocate_ivector(n_outputs);
	part_trial_cnt = allocate_ivector(n_outputs);
	part_rmserr_sum = allocate_vector(n_outputs);
	if (part_verbosity > CONFUSION_VERBOSITY)
	    part_confusion_matrix =
		    allocate_imatrix(n_outputs + 1, n_outputs);
    }
}

/*****************************************************************************
* NAME: error_report_part_free ()
* DESCRIPTION:
*       Free storage for error reporting for partial results
*
******************************************************************************/
void
error_report_part_free()
{
    if (part_verbosity > TABLE_VERBOSITY && n_outputs > 0) {
	deallocate_ivector(part_error_cnt);
	deallocate_ivector(part_trial_cnt);
	deallocate_vector(part_rmserr_sum);
	if (part_verbosity > CONFUSION_VERBOSITY)
	    deallocate_imatrix(part_confusion_matrix);
    }
}

/*****************************************************************************
* NAME: error_report_part_init ()
* DESCRIPTION:
*       Initialize error reporting variables
*
******************************************************************************/
void
error_report_part_init ()
{
    int		   i, j;
				/* set static global variables */
    part_total_trial_cnt = 0;
    part_total_error_cnt = 0;
    part_total_rmserr_sum = 0.0;
    if (verbosity > TABLE_VERBOSITY && n_outputs > 0) {
        for (i = 0; i < n_outputs; i++)
	    part_error_cnt[i] = part_trial_cnt[i] = part_rmserr_sum[i] = 0;
	if (verbosity > CONFUSION_VERBOSITY)
            for (i = 0; i < n_outputs + 1; i++)
                for (j = 0; j < n_outputs; j++)
		    part_confusion_matrix[i][j] = 0;
    }
}

/*****************************************************************************
* NAME: error_report_calc_rmserr ()
* DESCRIPTION:
*       Calculate rms_err for reports
*
******************************************************************************/
float
error_report_calc_rmserr(class, desired_class, outputs)
    int             class, desired_class;
    float          *outputs;
{
    float           mean_squared_error, rms_err;
    int             i;

    for (i = 0; i < n_outputs; i++)
	if (i == desired_class)
	    desired_outputs[i] = 1;
	else
	    desired_outputs[i] = 0;

    mean_squared_error = l2dist(n_outputs, outputs, desired_outputs) /
	    (float) n_outputs;

    if (mean_squared_error < SMALLEST)
	rms_err = 0;
    else
	rms_err = sqrt(mean_squared_error);

#if DEBUG
	printf("%s%d desired_class=%d noutputs=%d rms_err=%g outputs=",
	           "error_report_calc_rmserr: class=", class,
	           desired_class, n_outputs, rms_err);
	for (i = 0; i < n_outputs; i++)
	    printf(" %g", outputs[i]);
	printf("\n");
#endif
    return rms_err;
}

/*****************************************************************************
* NAME: error_report_class_trial ()
* DESCRIPTION:
*       Update errors for a single trial, write to file for classification
*		problem (!map_flag)
*
******************************************************************************/
void
error_report_class_trial(error_file, trial, verror, rms_err,
			 inputs, outputs, desired_class, class)
    FILE           *error_file;
    int             trial, verror, desired_class, class;
    float           rms_err, *inputs, *outputs;
{
    int             i;
    float mean_squared_error, tmp;

    /* rms_err given, at least by mlp, is wrong */
    mean_squared_error = 0;
    for(i = 0; i < n_outputs; i++){
	if(i == desired_class)
	  tmp = outputs[i] - 1.;
	else tmp = outputs[i];
	mean_squared_error += tmp*tmp;
    }
    mean_squared_error /= (float)n_outputs;
    if(mean_squared_error < SMALLEST) rms_err = 0;
    else rms_err = sqrt(mean_squared_error);

    if (verror > 0) {
	fprintf(error_file, "%d %d %d %d ",
	     trial, desired_class, class, (class == desired_class) ? 0 : 1);
	fprintf(error_file, "%7.4e ", rms_err);
	if (verror > 1) {
	    for (i = 0; i < n_inputs; i++)	/* print inputs */
		fprintf(error_file, "%7.4e ", inputs[i]);
	    for (i = 0; i < n_outputs; i++)	/* print outputs */
		fprintf(error_file, "%7.4e ", outputs[i]);
	}
	fprintf(error_file, "\n");
    }
    trial_cnt++;
    part_total_trial_cnt++;
    rmserr_sum += mean_squared_error;
    part_total_rmserr_sum += mean_squared_error;
    if (class != desired_class) {
	error_cnt++;
	part_total_error_cnt++;
	if (verbosity > TABLE_VERBOSITY)
	    class_error_cnt[desired_class]++;
	if (part_verbosity > TABLE_VERBOSITY)
	    part_error_cnt[desired_class]++;
    }
    if (verbosity > TABLE_VERBOSITY) {
	class_trial_cnt[desired_class]++;
	class_rmserr_sum[desired_class] += mean_squared_error;
	if (verbosity > CONFUSION_VERBOSITY)
	    (class_confusion_matrix[desired_class][class])++;
    }
    if (part_verbosity > TABLE_VERBOSITY) {
	part_trial_cnt[desired_class]++;
	part_rmserr_sum[desired_class] += mean_squared_error;
	if (part_verbosity > CONFUSION_VERBOSITY)
	    (part_confusion_matrix[desired_class][class])++;
    }
}

/*****************************************************************************
* NAME: error_report_map_trial ()
* DESCRIPTION:
*       Update errors for a single trial, write to file for mapping problem
*		(map_flag)
*
******************************************************************************/
void error_report_map_trial(error_file, trial, verror,
			    inputs, outputs, desired_outputs)
     FILE	*error_file;
     int	trial, verror;
     float	*inputs, *outputs, *desired_outputs;
{
    int		i;
    float	rms_err, mean_squared_error, tmp;

    /* calculate rms_err */
    mean_squared_error = l2dist(n_outputs, outputs, desired_outputs);
    mean_squared_error /= (float)n_outputs;
    if(mean_squared_error < SMALLEST) rms_err = 0;
    else rms_err = sqrt(mean_squared_error);

    if (verror > 0) {
	fprintf(error_file, "%d ", trial);
	for (i = 0; i < n_outputs; i++)	/* print desired output */
	    fprintf(error_file, "%7.4e ", desired_outputs[i]);
	for (i = 0; i < n_outputs; i++)	/* print actual output */
	    fprintf(error_file, "%7.4e ", outputs[i]);
	fprintf(error_file, "%7.4e ", rms_err);
	if (verror > 1) {
	    for (i = 0; i < n_inputs; i++)	/* print inputs */
		fprintf(error_file, "%7.4e ", inputs[i]);
	}
	fprintf(error_file, "\n");
    }
    part_total_trial_cnt++;
    trial_cnt++;
    rmserr_sum += mean_squared_error;
    part_total_rmserr_sum += mean_squared_error;
    if(verbosity > TABLE_VERBOSITY) {
	for(i = 0; i < n_outputs; i++){
	    class_trial_cnt[i]++;
	    tmp = desired_outputs[i] - outputs[i];
	    class_rmserr_sum[i] += tmp*tmp;
	}
    }
    if(part_verbosity > TABLE_VERBOSITY) {
	for(i = 0; i < n_outputs; i++){
	    part_trial_cnt[i]++;
	    tmp = desired_outputs[i] - outputs[i];
	    part_rmserr_sum[i] += tmp*tmp;
	}
    }
}

/*****************************************************************************
* NAME: error_report_overall ()
* DESCRIPTION:
*       Report overall and class by class statistics
*
******************************************************************************/
void
error_report_overall(log_file, error_fname, labels)
    FILE           *log_file;
    char           *error_fname;/* file name to be printed in log file */
    char          **labels;	/* set of labels describing outputs */
{
    int             i, j, itmp, tmp_cnt;
    float           rmserr_tmp, percent_error, stddev, tmp_err;

    if (trial_cnt <= 0) return;
    if(verbosity > CONFUSION_VERBOSITY){
	if(!map_flag){
	    fprintf(log_file, "\n\n%s %s", "Classification Confusion Matrix -",
		    error_fname);
	    fprintf(log_file, "\n%s%s", "--------------------------------",
		    "-----------------------------------------------");
	    fprintf(log_file, "\n\nDesired%41s", "Computed Class\n");
	    fprintf(log_file, " Class ");
	    for (j = 0; j < n_outputs; j++)
		fprintf(log_file, "%5d ", j);
	    fprintf(log_file, "  Total\n ----- ");
	    for (j = 0; j < n_outputs; j++)
		fprintf(log_file, "  ----");
	    fprintf(log_file, "  -----\n");
	    for (i = 0; i < n_outputs; i++) {
		fprintf(log_file, "%4d  ", i);
		for (j = 0; j < n_outputs; j++) {
		    itmp = class_confusion_matrix[i][j];
		    if (itmp > 0)
			fprintf(log_file, " %5d", itmp);
		    else
			fprintf(log_file, "      ");
		    class_confusion_matrix[n_outputs][j] += itmp;
		}
		fprintf(log_file, " %6d\n", class_trial_cnt[i]);
	    }
	    fprintf(log_file, " ----- ");
	    for (j = 0; j < n_outputs; j++)
		fprintf(log_file, "  ----");
	    fprintf(log_file, "  -----\n Total");
	    for (j = 0; j < n_outputs; j++) {
		itmp = class_confusion_matrix[n_outputs][j];
		fprintf(log_file, " %5d", itmp);
	    }
	    fprintf(log_file, " %6d\n\n\n", trial_cnt);
	}
    }
    if(verbosity > TABLE_VERBOSITY) {
	fprintf(log_file, "\n%s %s\n", "Error Report -",
		error_fname);
	fprintf(log_file, "%s%s\n", "--------------------------------",
		"-----------------------------------------------\n");
	if(!map_flag){
	    fprintf(log_file, "%s %12s %10s %12s %s    %s     %s\n",
		    "Class", "Patterns", "# Errors", "%% Errors", "StdDev",
		    "Rmserr", "Label");
	    for (i = 0; i < n_outputs; i++) {
		tmp_cnt = class_trial_cnt[i];
		if (tmp_cnt <= 0)
		  fprintf(log_file, "%3d %12d %49s\n",
			  i, tmp_cnt, labels[i + 1]);
		else {
		    tmp_err = ((float) class_error_cnt[i]/((float) tmp_cnt));
		    percent_error = tmp_err * 100.;
		    stddev = sqrt((tmp_err*(1-tmp_err))/(float)tmp_cnt)*100.;
		    /* mean of squared error for patterns of class i */
		    rmserr_tmp = class_rmserr_sum[i] / (float) tmp_cnt;
		    rmserr_tmp = sqrt(rmserr_tmp);
		    fprintf(log_file,
			    "%3d %12d %10d %12.2f  (%4.1f) %9.3f     %s\n",
			    i, tmp_cnt, class_error_cnt[i], percent_error,
			    stddev, rmserr_tmp, labels[i + 1]);
		}
	    }
	} else {
	    fprintf(log_file, "Output    Rmserr   Label\n");
	    for(i = 0; i < n_outputs; i++){
		tmp_cnt = class_trial_cnt[i];
		if(tmp_cnt <= 0)
		  fprintf(log_file, "%3d %20s\n", i, labels[i+1]);
		else {
		    /* mean of squared error in ith outputs */
		    rmserr_tmp = class_rmserr_sum[i] / (float) tmp_cnt;
		    rmserr_tmp = sqrt(rmserr_tmp);
		    fprintf(log_file, "%3d %9.3f %s\n",
			    i, rmserr_tmp, labels[i+1]);
		}
	    }
	}
    }
    /* mean of squared error for all patterns */
    rmserr_tmp = rmserr_sum / trial_cnt;
    rmserr_tmp = sqrt(rmserr_tmp);
    if (map_flag) {
	fprintf(log_file, "\nOverall %8d patterns, %7.3f rmserr\n\n",
		trial_cnt, rmserr_tmp);
    } else {
	if (verbosity <= TABLE_VERBOSITY)
	  fprintf(log_file,
		  "\n          Patterns   # Errors   %% Errors  StdDev    Rmserr\n");
	fprintf(log_file,
	        "          --------     ------     ------  ------    ------\n");
	tmp_err = ((float) error_cnt / ((float) trial_cnt));
	percent_error = tmp_err * 100.;
	stddev = sqrt ((tmp_err * (1-tmp_err)) / (float) trial_cnt) * 100.;
	fprintf(log_file, "Overall %8d %10d %12.2f  (%4.1f) %9.3f\n\n",
		trial_cnt, error_cnt, percent_error, stddev, rmserr_tmp);
    }
}

/*****************************************************************************
* NAME: error_report_part ()
* DESCRIPTION:
*       Report summary and class by class statistics
*
******************************************************************************/
void
error_report_part(log_file, error_fname, labels)
    FILE           *log_file;
    char           *error_fname;/* file name to be printed in log file */
    char          **labels;	/* set of labels describing outputs */
{
    int             i, j, itmp, tmp_cnt;
    float           rmserr_tmp, percent_error, stddev, tmp_err;

    if(part_total_trial_cnt <= 0) return;
    if (part_verbosity > CONFUSION_VERBOSITY) {
	if (!map_flag){
	    fprintf(log_file, "\n\n%s %s", "Classification Confusion Matrix -",
		    error_fname);
	    fprintf(log_file, "\n%s%s", "--------------------------------",
		    "-----------------------------------------------");
	    fprintf(log_file, "\n\nDesired%41s", "Computed Class\n");
	    fprintf(log_file, " Class ");
	    for (j = 0; j < n_outputs; j++)
		fprintf(log_file, "%5d ", j);
	    fprintf(log_file, "  Total\n ----- ");
	    for (j = 0; j < n_outputs; j++)
		fprintf(log_file, "  ----");
	    fprintf(log_file, "  -----\n");
	    for (i = 0; i < n_outputs; i++) {
		fprintf(log_file, "%4d  ", i);
		for (j = 0; j < n_outputs; j++) {
		    itmp = part_confusion_matrix[i][j];
		    if (itmp > 0)
			fprintf(log_file, " %5d", itmp);
		    else
			fprintf(log_file, "      ");
		    part_confusion_matrix[n_outputs][j] += itmp;
		}
		fprintf(log_file, " %6d\n", part_trial_cnt[i]);
	    }
	    fprintf(log_file, " ----- ");
	    for (j = 0; j < n_outputs; j++)
		fprintf(log_file, "  ----");
	    fprintf(log_file, "  -----\n Total");
	    for (j = 0; j < n_outputs; j++) {
		itmp = part_confusion_matrix[n_outputs][j];
		fprintf(log_file, " %5d", itmp);
	    }
	    fprintf(log_file, " %6d\n\n\n", part_total_trial_cnt);
	}
    }
    if (part_verbosity > TABLE_VERBOSITY){
	fprintf(log_file, "\n%s %s\n", "Error Report -",
		error_fname);
	fprintf(log_file, "%s%s\n", "--------------------------------",
		"-----------------------------------------------\n");
	if(!map_flag){
	    fprintf(log_file,
    "Class     Patterns   # Errors   %% Errors  StdDev    Rmserr     Label\n");
	    for (i = 0; i < n_outputs; i++) {
		tmp_cnt = part_trial_cnt[i];
		if (tmp_cnt <= 0)
		  fprintf(log_file, "%3d %12d %49s\n",
			  i, tmp_cnt, labels[i + 1]);
		else {
		    tmp_err = ((float) part_error_cnt[i] / ((float) tmp_cnt));
		    percent_error = tmp_err * 100.;
		    stddev = sqrt((tmp_err*(1-tmp_err))/(float)tmp_cnt)*100.;
		    rmserr_tmp = part_rmserr_sum[i] / (float) tmp_cnt;
		    fprintf(log_file,
			    "%3d %12d %10d %12.2f  (%4.1f) %9.3f     %s\n",
			    i, tmp_cnt, part_error_cnt[i], percent_error,
			    stddev, rmserr_tmp, labels[i + 1]);
		}
	    }
	} else {
	    fprintf(log_file, "Output    Rmserr   Label\n");
	    for(i = 0; i < n_outputs; i++){
		tmp_cnt = part_trial_cnt[i];
		if(tmp_cnt <= 0)
		  fprintf(log_file, "%3d %20s\n", i, labels[i+1]);
		else {
		    /* mean of squared error in ith outputs */
		    rmserr_tmp = part_rmserr_sum[i] / (float) tmp_cnt;
		    rmserr_tmp = sqrt(rmserr_tmp);
		    fprintf(log_file, "%3d %9.3f %s\n",
			    i, rmserr_tmp, labels[i+1]);
		}
	    }
	}
    }
    rmserr_tmp = part_total_rmserr_sum / part_total_trial_cnt;
    rmserr_tmp = sqrt(rmserr_tmp);
    if (map_flag) {
	fprintf(log_file, "\nOverall %8d patterns, %7.3f rmserr\n\n",
		part_total_trial_cnt, rmserr_tmp);
    } else {
	if (part_verbosity <= TABLE_VERBOSITY)
	  fprintf(log_file,
	"\n          Patterns   # Errors   %% Errors  StdDev    Rmserr\n");
	fprintf(log_file,
	"          --------     ------     ------  ------    ------\n");
	tmp_err = ((float) part_total_error_cnt / 
		   ((float) part_total_trial_cnt));
	percent_error = tmp_err * 100.;
	stddev = sqrt ((tmp_err * (1-tmp_err)) / 
		       (float) part_total_trial_cnt) * 100.;
	fprintf(log_file, "Overall %8d %10d %12.2f  (%4.1f) %9.3f\n\n",
		part_total_trial_cnt, part_total_error_cnt, percent_error, 
		stddev, rmserr_tmp);
    }
    fflush(stdout);
}

/*****************************************************************************
* NAME: error_report_between_classes ()
* DESCRIPTION:
*       Report on the error between one pair of classes
*
******************************************************************************/
int  error_report_between_classes(log_file, class_1, class_2, verbosity_p)
     FILE           *log_file;
     int            class_1, class_2, verbosity_p;
{
  int itmp;
  
  itmp = class_confusion_matrix[class_1][class_2] + 
    class_confusion_matrix[class_2][class_1];
  if ((itmp > 0) && (verbosity_p > 2))
    fprintf(log_file, "Between classes %d and %d, %d errors \n",
	    class_1, class_2, itmp);
  return(itmp);
  }

/*****************************************************************************
* NAME: error_report_overall_error_rate ()
* DESCRIPTION:
*       Return the error rate overall all patterns
*
******************************************************************************/

float error_report_overall_error_rate()
{
    return(((float) part_total_error_cnt /  
		   ((float) part_total_trial_cnt)));
}

/*****************************************************************************
* NAME: error_report_overall_rms ()
* DESCRIPTION:
*       Return the normalized sum of squared errors over all patterns
*
******************************************************************************/

float error_report_overall_rms()
{
    return(((float) part_total_rmserr_sum /  
		   ((float) part_total_trial_cnt)));
}
