/***-------------------------------------------------------------------------
* NAME:     errors.c 
* PURPOSE:  error reporting and debugging utilities 
* DESCRIPTION:
*   coredump_error - print error message , dump core,  and terminate 
*   exit_nnsubs - print number of non-fatal errors so far
*       if there were any
*   fatal_error - print error message using printf format and terminate 
*   get_program_name - return string pointer to program name 
*   get_command_line - return string pointer to command line 
*   init_nnsubs     - store program name, command line, and initialize
*       warning count
*   warning     - print warning message using printf format 
* 
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/errors.c,v 11.8 2003/05/16 18:15:08 kukolich Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/errors.c,v $
* CREATED: 26-July-90 rpl                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#endif
/* sys/types and signal are for kill */
#include <sys/types.h>
#include <signal.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "nnlib.h"

static int init_error_message_flag=0;
static char error_program_name[100];
static char error_command_line[1000];
static int total_warnings=0;
static char last_warning[512] = {'\0'};
static int same_warning_cnt = 0;

/***---------------------------------------------------------------------------
* NAME:     init_nnsubs 
* PURPOSE:  store program name, command line, and initialize warning count
* EXAMPLE:
*       static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/errors.c,v 11.8 2003/05/16 18:15:08 kukolich Exp $";
*       main(argc, argv)
*       int argc;
*       char **argv;
*           {
*           init_nnsubs(argc, argv, rcsid);
*           
* NOTE:     Must be called before argc and argv are touched and does not
*           alter argc or argv. Assumes program name and rcsid are shorter
*           than 100 characters long.
--------------------------------------------------------------------------***/
void init_nnsubs(int argc, char **argv, char *rcsid)
{
    int i;

    init_error_message_flag++;
    total_warnings=0;

    /* store program name */
    strcpy(error_program_name, argv[0]);

    /* store command line */
    strcpy(error_command_line, argv[0]);
    for(i=1;
    i<argc && (2 + strlen(error_command_line) + strlen(argv[i]) < 1000)
    ;i++) {
      strcat(error_command_line, " ");
      strcat(error_command_line, argv[i]);
      }
}
/***---------------------------------------------------------------------------
* NAME:     exit_nnsubs 
* PURPOSE:  print number of non-fatal errors so far if there were any
* EXAMPLE:
*       main(argc, argv)
*       int argc;
*       char **argv;
*           {
*           init_nnsubs(argc, argv, rcsid);
*           .....
*           exit_nnsubs();
* NOTES:
*        Prints nothing if there were no warnings.
*        Prints number of warnings if there were some.
*        Counter used to record warnings is reset to zero
*        using init_nnsubs().
--------------------------------------------------------------------------***/
void exit_nnsubs(void)
{
    if( same_warning_cnt > 0)
      fprintf(stderr, "\nPrevious warning repeated %d times\n",
          same_warning_cnt);
    if ( total_warnings > 0)
      fprintf(stderr, "\nWARNING\n-------\n%d %s\n", total_warnings,
         "non-fatal errors occurred during program execution!");
    init_error_message_flag = same_warning_cnt = total_warnings = 0;
    last_warning[0] = '\0';
}

/***---------------------------------------------------------------------------
* NAME:     fatal_error 
* PURPOSE:  print error message using printf format and terminate 
* EXAMPLE:
*   fatal_error("nnargs: Can't open file %s after %d tries", filename, n);
*
* NOTE:     Should initialize with init_nnsubs, adds final carriage return
--------------------------------------------------------------------------***/
void fatal_error (char *fmt, ...)
{
    va_list args;

    if( same_warning_cnt > 0)
      fprintf(stderr, "\nPrevious warning repeated %d times\n",
          same_warning_cnt);

    if ( init_error_message_flag )
        fprintf(stderr,"\nERROR: %s\n----------------\n",
            error_program_name);
    else
        fprintf(stderr,"\nERROR: \n----------------\n");

    /* print error message using variable number of arguments */
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);
    fprintf(stderr, "\n");

    if ( init_error_message_flag )
        {
        fprintf(stderr, 
          "\nCOMMAND LINE WAS:\n------------------\n%s\n",
            error_command_line);
         }
    exit(-1);
}

/***---------------------------------------------------------------------------
* nAME:     coredump_error 
* PURPOSE:  print error message using printf format, dump core,
*       and terminate program
* EXAMPLE:
*   coredump_error("Could not open file %s after %d tries", filename, n);
*
* NOTE:     Should initialize with init_nnsubs
--------------------------------------------------------------------------***/
void coredump_error (char *fmt, ...)
{
    va_list args;

    if ( init_error_message_flag )
        fprintf(stderr,
         "\nERROR CORE BEING DUMPED: %s\n------------------------\n",
         error_program_name);
    else
        fprintf(stderr,
         "\nERROR CORE BEING DUMPED: \n------------------------\n");

    /* print error message using variable number of arguments */
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);
    fprintf(stderr, "\n");

    if ( init_error_message_flag )
        {
        fprintf(stderr, 
            "\nCOMMAND LINE WAS:\n------------------\n%s\n",
            error_command_line);
         }

     /* dump core using bus error */
     kill(getpid(),10);  

     /* just in case kill signal is caught */
     exit(-1);
}

/***---------------------------------------------------------------------------
* NAME:     warning 
* PURPOSE:  print warning message using printf format 
* EXAMPLE:
*   warning("Could not open file %s after %d tries", filename, n);
*
* NOTE:     Should be initialized with init_nnsubs, adds final carriage return
--------------------------------------------------------------------------***/
void warning (char *fmt, ...)
{
    va_list args;
    char current_warning[512];

    total_warnings += 1;

    /* copy out warning message using the variable length argument lit */
    va_start(args, fmt);
    vsprintf(current_warning, fmt, args);
    va_end(args);

    if(strncmp(last_warning, current_warning, 512) != 0){
    /* if this warning is different from the lat one */
    if ( init_error_message_flag )
      fprintf(stderr,"\nWARNING: %s\n----------------\n",
          error_program_name);
    else
      fprintf(stderr,"\nWARNING: \n----------------\n");
    if(same_warning_cnt != 0)
      fprintf(stderr, "Previous warning repeated %d times\n",
          same_warning_cnt);
    same_warning_cnt = 0;
    strncpy(last_warning, current_warning, 512);
    fprintf(stderr, "%s\n", current_warning);
    } else {
    same_warning_cnt++;
    }
    va_end(args);
}

/***---------------------------------------------------------------------------
* NAME:     get_program_name 
* PURPOSE:  return string pointer to program name 
* EXAMPLE:
*           printf("Program name is %s\n", get_program_name());
*
* NOTE:     Must be initialized with init_nnsubs
--------------------------------------------------------------------------***/
char *get_program_name(void)
{
    return error_program_name;
}

/***---------------------------------------------------------------------------
* NAME:     get_command_line
* PURPOSE:  return string pointer to command line 
* EXAMPLE:
*           printf("Command line is %s\n", get_program_name());
*
* NOTE:     Must be initialized with init_nnsubs
--------------------------------------------------------------------------***/
char *get_command_line(void)
{
    return error_command_line;
}

