/***---------------------------------------------------------------------------
* NAME:     dist_subs.c
* PURPOSE:  fastdist routines
*
* GLOBALS: 
*
* DESCRIPTION:
*   fastdist_build() -- allocates and initializes space for fastdist
*   fastdist_init() -- initializes distances for fastdist_eval
*   fastdist_eval() -- calculates distances from given vector to vectors in
*       a matrix given to fastdist_init()
*   fastdist_exit() -- deallocates space for fastdist_eval
*   fastdist_param() -- return pointer to specified parameter
* 
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/fastdist.c,v 11.3 2001/08/23 20:14:43 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/fastdist.c,v $
* CREATED: 1-Feb-91 lck                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------
* Copyright 1991 MIT Lincoln Laboratory.
* This software is not for release and should neither be copied
* nor distributed without explicit permission of MIT Lincoln Laboratory.
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"

typedef struct {
    float **centers;
    float *inv_vars;
    int ncenters, ndim;
    float *centers_sumsq;
    float *input_norm;
} FASTDIST;

static FASTDIST **fdistpp;
static FASTDIST *fdist;
static int max_models = 0;
#define MODEL_STORE_SIZE 10

/*****************************************************************************
* NAME: fastdist_build(n_inputs,n_centers)
* DESCRIPTION:
*   allocates space for vectors mentioned in the comment for fastdist_eval
*
******************************************************************************/
int fastdist_build(int n_inputs, int n_centers)
{
    register int new_model;

    new_model = model_get_number((void ***)&fdistpp, &max_models,
                 sizeof(FASTDIST));
    fdist = fdistpp[new_model];

    /* set pointers and lengths */
    fdist->ncenters = n_centers;
    fdist->ndim = n_inputs;

    /* allocate vectors */
    fdist->input_norm = allocate_vector(n_inputs);
    fdist->inv_vars = allocate_vector(n_inputs);
    fdist->centers_sumsq = allocate_vector(n_centers);

    return(new_model);

}
/*****************************************************************************
* NAME: fastdist_init(model,pcenters,pvariances)
* DESCRIPTION:
*   pre-calculates the third summation in fastdist_eval
*
******************************************************************************/
void fastdist_init(int model, float **pcenters, float *pvariances)
{
    register int i,j;

    fdist = fdistpp[model];

    for(i = 0; i < fdist->ndim; i++){
    if(pvariances[i] == 0.0){
        fatal_error("fastdist_init: pvariance[%d] = 0.0",i);
    }
    fdist->inv_vars[i] = 1.0/pvariances[i];
    }

    fdist->centers = pcenters;

    for(j = 0; j < fdist->ncenters; j++){
    for(i = 0; i < fdist->ndim; i++){
        fdist->centers_sumsq[j] += fdist->centers[j][i] *
                       fdist->centers[j][i] *
                       fdist->inv_vars[i];
    }
    }
}
/*****************************************************************************
* NAME: fastdist_eval(model,inputs,distances)
* DESCRIPTION:
*   Calculates the square of the distances between the vectors in centers
*   and the input vector using the following formula:
*
*        ndim
*         ---               2
*         \  (inputs[i] - centers[j][i])
*   dist[j] =  > ---------------------------- 
*         /__   variances[i]
*         i=0
*
*   for speed, this formula is rewritten as:
*
*       ndim            ndim
*        ---        2    ---
*        \   (inputs[i])     \  inputs[i]*centers[j][i]
*   dist[j] = >  -----------   - 2 *  > -----------------------  +
*        /__ variances[i]    /__    variances[i]
*        i=0             i=0
*
*            ndim
*             ---           2
*             \  (centers[j][i])
*           +  > --------------- 
*             /__  variances[i]
*             i=0
*
*   The third sum is calculated during fastdist_init()
*   The first sum is calculated at the start of fastdist_eval
* RETURNS:
*   index in pcenter of closest vector in centers
*   -1 if no distances are calculated
*
******************************************************************************/
int fastdist_eval(int model, float *inputs, float *distances)
{
    register int i,j;
    register float input_sumsq = 0.0,dist;
    register float min_dist;
    register int min_index;

    fdist = fdistpp[model];

    for(i = 0; i < fdist->ndim; i++){
    fdist->input_norm[i] = inputs[i]*fdist->inv_vars[i];
    input_sumsq += fdist->input_norm[i] * inputs[i];
    }

    min_dist = MAXFLOAT;
    min_index = -1;
    for(j = 0; j < fdist->ncenters; j++){
    dist = 0.0;
    for(i = 0; i < fdist->ndim; i++){
        dist += fdist->input_norm[i] * fdist->centers[j][i];
    }
    distances[j] = input_sumsq - 2.0 * dist + fdist->centers_sumsq[j];

    /* do check for min */
    if(distances[j] < min_dist){
        min_dist = distances[j];
        min_index = j;
    }
    }
    return(min_index);
}
/*********************************************************************
* NAME: fastdist_exit(model)
* DESCRIPTION:
*   deallocates space allocated in fastdist_init()
*************************************************************************/
void fastdist_exit(int model)
{

    fdist = fdistpp[model];

    deallocate_vector(fdist->input_norm);
    deallocate_vector(fdist->inv_vars);
    deallocate_vector(fdist->centers_sumsq);
    nn_free((char *)fdist,"model for fastdist");
    fdistpp[model] = NULL;
}

void * fastdist_param(int model, char *param_name)
{
    void *param_address;

    fdist = fdistpp[model];
    if(0 == strcmp(param_name,"n_inputs")) param_address = &fdist->ndim;
    else if (0 == strcmp(param_name,"n_centers"))
    param_address = &fdist->ncenters;
    else{
    warning("Parameter %s is not defined",param_name);
    param_address = NULL;
    }
    return(param_address);
} /* END fastdist_param */
