/***---------------------------------------------------------------------------
* NAME:		feat_sel.c
* PURPOSE:	feature list creation, results printing, and parameter
*		file routines for feature searches, including the possibility
*		that there is categorical
*		data that has been reinterpretted as sets of binary features
*
*	State: MA,NY,NH,ME,VT -> One field with 5 values becomes 5 binary
*	fields, MA?, NY? NH? ME? VT?, which are correlated since only one
*	is on at a time.
*
* USAGE:	see mlp_feat_sel.c
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/feat_sel.c,v 11.6 2001/08/23 20:14:43 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/feat_sel.c,v $
* CREATED: 21-Mar-94 lck			 MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include "nnlib.h"

#define FORWARD 0
#define BACKWARD 1
#define FOR_AND_BACK 2
#define FLOATING 3

static int total_segs;
static int *selected_segs; /* list of selected features */
static int cur_cnt;
static int swap_seg; /* counters for search */
static int cur_min;
static int cur_feat_min;
static int *segments;
static char **feat_names;
static int *features;
static int total_feat;
static float *perr_features;
static float *perr_segs;
static float perr_min;
static int nfeatures;
static int nsegs;

static int backward; /* for FOR_AND_BACK, current direction */
static int sel_cnt;  /* counter to keep track of direction change for F&B */
static int search_high, search_low; /* search limits */
static int search_type;
static int verbosity;

static void cat_sel_fill_features(int nsegs);
static void cat_sel_fill_perr(int nsegs);
static void feat_sel_print_result(int *feature_list, float percent_error,
				  int nfeatures, int nsegs);
/**************
  feat_sel_build()
  Call the new routine
  ***************/
void
feat_sel_build(int zinputs, int zsearch_type, int zsearch_n, int zverbosity){
  cat_sel_build(zinputs, zsearch_type, zsearch_n, zverbosity, NULL);
}

/**************
  cat_sel_build()
  allocates space for a feature search, initializes search limits
  according to search direction
  ***************/
void
cat_sel_build(int zninputs, int zsearch_type, int zsearch_n, int zverbosity,
	      int *zseg_list)
{
    int i;
    char seg_name[64]; /* space enough to write 2 base 10 numbers */

    if(zseg_list && zseg_list[0] == 0){
      zseg_list = NULL;
    }

    /* save important values */
    if(zseg_list){
      total_segs = zseg_list[0];
      total_segs /= 2; /* this gives the first and last, so chop it in half */
    } else {
      total_segs = zninputs;
    }
    total_feat = zninputs;
    search_type = zsearch_type;
    verbosity = zverbosity;

    /* allocate and initialize features and perr_features */
    selected_segs = allocate_ivector(total_segs+1);
    perr_segs = allocate_vector(total_segs);
    feat_names = (char **)allocate_pvector(total_segs);
    if(zseg_list){
      features = allocate_ivector(total_feat+1);
      segments = allocate_ivector(2*total_segs);
      perr_features = allocate_vector(total_feat);
    } else {
      segments = NULL;
      features = selected_segs;
      perr_features = perr_segs;
    }

    selected_segs[0] = total_segs;
    for(i = 0; i < total_segs; i++){
      selected_segs[i+1] = i;
      perr_segs[i] = 101.0;
    }

    if(zseg_list){
      for(i = 0; i < total_segs*2; i++){
	segments[i] = zseg_list[i+1];
      }
      for(i = 0; i < total_segs; i++){
	sprintf(seg_name, "%d:%d",segments[2*i],segments[2*i+1]);
	feat_names[i] = nn_strdup(seg_name);
      }
    } else {
      for(i = 0; i < total_segs; i++){
	sprintf(seg_name, "%d", i);
	feat_names[i] = nn_strdup(seg_name);
      }
    }
    if (search_type == BACKWARD){
	backward = 1;
	nsegs = total_segs;
	nfeatures = total_feat;
    } else {
	backward = 0;
	nsegs = 1;
	nfeatures = 1;
    }
    sel_cnt = 1;

    /* initialize search limits */
    search_high = total_segs;
    search_low = 1;
    if(zsearch_n != 0){
	if(search_type == BACKWARD) search_low = zsearch_n;
	else search_high = zsearch_n;
    }
}

/* return 1 if the feature search over all numbers of features is done */
int
feat_sel_search_done(void)
{
    /* if the next round would be backward below search_low or
       forward above search_high, return 1 */
    if(backward){
	if(nsegs < search_low) return 1;
    } else {
	if(nsegs > search_high) return 1;
    }
    return 0;
}

/* initialize nsegs, cur_cnt, cur_min, and perr_min for a
   search for the best nsegs features */
void
feat_sel_init_nfeatures(void)
{
    if(backward && nsegs != total_segs){
	swap_seg = cur_cnt = nsegs + 1;
    } else {
	swap_seg = cur_cnt = nsegs;
    }
    cur_min = cur_cnt;
    perr_min = 101.;
    selected_segs[0] = nsegs;
}

/* if all the features available have been tried this round, return NULL
   otherwise,
   pick the next feature to include or exclude, swap it into the
   active section (the first nsegs entries) of selected_segs
*/
int *
feat_sel_next_features(void)
{
    /* check if we are done */
    if(backward){
	if(nsegs == total_segs && cur_cnt < total_segs) return NULL;
	if(cur_cnt <= 0) return NULL;
    } else {
	if(cur_cnt > total_segs) return NULL;
    }
    /* swap the next feature in */
    swap_int(&selected_segs[swap_seg], &selected_segs[cur_cnt]);
    cat_sel_fill_features(nsegs);
    return features;
}

void
feat_sel_check_best(float percent_error)
{
    /* print result */
    if (verbosity > 1) {
      printf("    ");
      feat_sel_print_result(selected_segs, percent_error, nfeatures, nsegs);
    }

    /* check the percent error against the min */
    if (percent_error < perr_min) {
	perr_min = percent_error;
	cur_min = cur_cnt;
    }
    /* swap the current feature back to its place */
    swap_int(&selected_segs[swap_seg], &selected_segs[cur_cnt]);
    if(backward) cur_cnt--;
    else cur_cnt++;
}

void
feat_sel_pick_best(void)
{
    int cur_feature;

    /* move the features between the selected one and its
       new place, to preserve the order of the remaining ones */
    cur_feature = selected_segs[cur_min];
    if(!backward){
      /* shift other features up 1 to make space at the start */
      for(; cur_min > nsegs; cur_min--)
	selected_segs[cur_min] = selected_segs[cur_min-1];
      selected_segs[nsegs] = cur_feature;
      /* record the error rate */
      perr_segs[nsegs - 1] = perr_min;
      /* print best result */
      cat_sel_fill_features(nsegs);
      if (verbosity > 0) {
	printf(" ** ");
	feat_sel_print_result(selected_segs, perr_min, nfeatures, nsegs);
      }
      if((search_type == FORWARD) || (nsegs <= 2) || 
	 ((search_type == FOR_AND_BACK) && (sel_cnt <= 2))){
	nsegs++;
	sel_cnt++;
      } else { /* FLOATING or FOR_AND_BACK */
	/* only search for best single once */
	sel_cnt = 1;
	backward = 1;
	nsegs--;
      }
    } else { /* backward */
      if(((search_type == FOR_AND_BACK) && (cur_min == nsegs+1)) ||
	 ((search_type == FLOATING) && (sel_cnt == 1) && (cur_min == nsegs+1))
	 /* taking out the feature just added in. Don't bother */ ||
	 ((search_type == FLOATING) && (perr_min >= perr_segs[nsegs-1]))) {
	/* taking out this feature doesn't improve the system over
	   what we used to have with this number of features */
	/* print best result */
	cat_sel_fill_features(nsegs);
	if (verbosity > 0) {
	  printf(" NC ");
	  feat_sel_print_result(selected_segs, perr_min, nfeatures, nsegs);
	}
	backward = 0;
	sel_cnt = 3;
	nsegs += 2;
      } else { /* actually remove the feature */
	/* shift other features down 1 to make space at the end */
	if(nsegs != total_segs){
	  for(;cur_min <= nsegs; cur_min++)
	    selected_segs[cur_min] = selected_segs[cur_min+1];
	  selected_segs[nsegs+1] = cur_feature;
	}
	/* record the error rate */
	perr_segs[nsegs - 1] = perr_min;
	/* print best result */
	cat_sel_fill_features(nsegs);
	if (verbosity > 0) {
	  printf(" ** ");
	  feat_sel_print_result(selected_segs, perr_min, nfeatures, nsegs);
	}
	sel_cnt ++;
	if((search_type == FOR_AND_BACK) ||
	   ((search_type != BACKWARD) && (nsegs <= 2))){
	  backward = 0;
	  nsegs++;
	} else {
	  nsegs--;
	}
      }
    }
    nfeatures = nsegs;
}

void
feat_sel_print_head(void)
{
  if(segments){
    printf("\n#Segments  %%error  Segments List\n");
  } else {
    printf("\n#Features  %%error  Features List\n");
  }
  printf("---------  ------  -------------\n");
}

void
feat_sel_print_final_result(void)
{
    int i, j;

    /* print out results */
    cur_min = 0;
    if(segments){
      cur_feat_min = segments[1]-segments[0]+1;
    }
    for (i = 0, j = 0; i < total_segs; i++){
      if(segments){
	j += segments[selected_segs[i+1]*2+1]-segments[selected_segs[i+1]*2]+1;
      } else {
	j++;
      }
      if (perr_segs[i] < perr_segs[cur_min]){
	cur_min = i;
	cur_feat_min = j;
      }
    }

    cat_sel_fill_features(total_segs);
    cat_sel_fill_perr(total_segs);

    /* print the whole list */
    printf("ALL ");
    feat_sel_print_result(selected_segs, perr_segs[total_segs - 1], total_feat, total_segs);
    printf("BEST");
    /* print best */
    feat_sel_print_result(selected_segs, perr_segs[cur_min], cur_feat_min, cur_min + 1);
}

static void
feat_sel_print_result(int *feature_list, float percent_error, int nfeatures, int nsegs)
{
    int i;

    printf("%5d", nsegs);

    printf("  %7.1f   ", percent_error);

    for (i = 1; i <= nsegs; i++) {
	printf("%s ", feat_names[feature_list[i]]);
    }

    printf("\n");
    fflush(stdout);
}

void
feat_sel_save_list(char *file_path, char *comment)
{
    FILE *param_file;

    cat_sel_fill_features(total_segs);
    cat_sel_fill_perr(total_segs);

    /* save results in parameter file */
    param_file = nn_fopen(file_path, "w", 
			  "for write of feat_sel param file");
    /* write comment */
    fprintf(param_file, "%s\n", comment);
    if(segments){
      feat_sel_write_model(param_file, &features[1]/* take off length entry */,
			  perr_features, cur_feat_min + 1, total_feat);
    } else {
      feat_sel_write_model(param_file, &selected_segs[1],
			   perr_segs, cur_min +1, total_segs);
    }
    nn_fclose(param_file, file_path, "for write of feat_sel param file");
    if ( verbosity > 0)
      printf("\nFinished -- features saved in \"%s\"\n", file_path);
    
}

void
feat_sel_free(void)
{
    /* free allocated space */
    deallocate_ivector(selected_segs);
    deallocate_vector(perr_segs);
    if(segments){
      deallocate_ivector(features);
      deallocate_ivector(segments); segments = NULL;
      deallocate_vector(perr_features);
    }
}

		/*** Parameter Tables ***/
static PARAM paramtab[] = {
    {"n_inputs",P_INT},
    {"best", P_INT},
};
#define TABSIZE (sizeof(paramtab)/sizeof(PARAM))

static PARAM paramtab_l[] = {
    {"selected", P_INT_VECTOR},
    {"errors", P_FLOAT_VECTOR},
};
#define TABSIZE_L (sizeof(paramtab_l)/sizeof(PARAM))


/****************************************************************************
* NAME: feat_sel_read_model()
* DESCRIPTION:
*       Loads parameters from ascii input file
*	allocates space for and loads data into vectors and matricies
*
*	On error from param_getblk, returns the error value
*	(there is already an error check on the allocation, so we don't
*	have to do that here)
*
* EXAMPLE:
****************************************************************************/
int
feat_sel_read_model(FILE *fp, int *n_inputs, int *best,int **selected,
		    float **errors)
{
    int rv;

    /* read in scalars from model file */
    param_setvarptr("best",best,paramtab,TABSIZE);
    param_setvarptr("n_inputs",n_inputs,paramtab,TABSIZE);

    if((rv = param_getblk(fp,paramtab,TABSIZE)) < 0) return rv;

    *selected = allocate_ivector(*n_inputs);
    *errors = allocate_vector(*n_inputs);

    /* allocate space for model and vectors*/
    param_setvarptr ("selected", *selected, paramtab_l, TABSIZE_L);
    param_setdim ("selected", *n_inputs, 0, paramtab_l, TABSIZE_L);
    param_setvarptr ("errors", (int *)*errors, paramtab_l, TABSIZE_L);
    param_setdim ("errors", *n_inputs,0, paramtab_l, TABSIZE_L);

    if((rv = param_getblk(fp,paramtab_l,TABSIZE_L)) < 0) return rv;

    return(*n_inputs);
}

/****************************************************************************
* NAME: feat_sel_write_model()
* DESCRIPTION:
*	writes parameters and exemplars for a model to a file
*	in ASII format
*
* EXAMPLE:
*
****************************************************************************/
int
feat_sel_write_model(FILE *fp, int *selected, float *errors, int best,
		     int n_inputs)
{
    int rv;

    /* save scalars */
    param_setvarptr("best",&best,paramtab,TABSIZE);
    param_setvarptr("n_inputs",&n_inputs,paramtab,TABSIZE);
    if((rv = param_putblk(fp,paramtab,TABSIZE)) < 0) return rv;

    /* save vectors and matrices */
    param_setvarptr ("selected", selected, paramtab_l, TABSIZE_L);
    param_setdim ("selected", n_inputs, 0, paramtab_l, TABSIZE_L);
    param_setvarptr ("errors",(int *)errors, paramtab_l, TABSIZE_L);
    param_setdim ("errors", n_inputs, 0, paramtab_l, TABSIZE_L);
    if((rv = param_putblk(fp, paramtab_l, TABSIZE_L)) < 0) return rv;

    return(0);
}

static void cat_sel_fill_features(int nsegs){
  int i, j, k;

  if(!segments) return;

  for(i = 1, k=1; i <= nsegs; i++){
    for(j = segments[selected_segs[i]*2];
	j <= segments[selected_segs[i]*2+1];
	j++,k++){
      features[k] = j;
    }
  }
  features[0] = nfeatures = --k;
}

static void cat_sel_fill_perr(int nsegs){
  int i, j, k;

  if(!segments) return;

  for(i = 1, k=0; i <= nsegs; i++){
    for(j = segments[selected_segs[i]*2];
	j <= segments[selected_segs[i]*2+1];
	j++,k++){
      perr_features[k] = perr_segs[i-1];
    }
  }
}
