/***--------------------------------------------------------------------------
* NAME:		features.c
* PURPOSE:	routines for checking feature selection and frame selection
*	parameters for LNKnet and LNKsequence classifiers. also, routines
*	for reading and writing those parameters in parameter files.
*
* SEE ALSO: 
* GLOBALS:
*
* DESCRIPTION:
*	General purpose functions
*
*	fix_features()           fixes features array for future use in
*			         get_class_line()
*	fix_windows()
*	features_read_paramfile()
*	features_write_paramfile()
*	windows_read_paramfile()
*	windows_write_paramfile()
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/features.c,v 11.4 2001/08/23 20:14:43 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/features.c,v $
* CREATED: 25-Jan-94 lck			MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------
* Copyright 1990 MIT Lincoln Laboratory.
* This software is not for release and should neither be copied
* nor distributed without explicit permission of MIT Lincoln Laboratory.
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"

/****************************************************************************
* NAME: fix_features(nraw, ninputs, features)
* DESCRIPTION:
**	checks the features array to determine the number of features
*	requested by the caller
*
*	The format of the features array here has a length in features[0].
*	the data reading routines all take a feature array without the length
*	( the length is stored in ninputs in the end )
*
*	One way to do this is to store all the information in the features
*	array.  The current method is to leave the features array alone,
*	but to make sure that ninputs is correct
*
*	Rules are as follows:
*	The length of the features array is always stored in features[0].
*
*	If ninputs and features[0] are both 0, you should use all nraw
*	points in order.  Ninputs is set to nraw
*		fix_features(12, 0, NULL) or
*		fix_features(12, 0, {0}) -> 12 (0,1,2,3,4,5,6,7,8,9,10,11)
*
*	if ninputs is 0 and features[0] is not, use features
*	as is. ninputs becomes features[0].
*		fix_features(12, 0, {3, 2, 4, 5}) -> 3 (2,4,5)
*
*	if ninputs is not 0, but features[0] is, you should use the first
*	ninputs points in order.
*		fix_features(12, 3, NULL) or
*		fix_features(12, 3, {0}) -> 3 (0,1,2)
*
*	if ninputs is not 0, and features[0] is not 0, you should use
*	all of features or just the first ninputs points, whichever is fewer.
*	set ninputs and features[0] to the proper lengths
*		fix_features(12, 3, {4, 2, 4, 5, 0}) -> 3 (2,4,5)
*				and features[0] is set to 3
*		fix_features(12, 3, {2, 2, 4}) -> 2 (2,4)
*
* RETURNS:
*	the number of points that are going to be used
*	also sets features[0] if necessary
*
****************************************************************************/
int fix_features(int nraw, int ninputs, int *features)
{
    int i;

    if(features == NULL || features[0] <= 0){
	/* take all inputs or first ninputs of them */
	if(ninputs == 0 || ninputs > nraw) ninputs = nraw;
    } else {
	if(ninputs == 0){
	    /* take all of the features from array */
	    ninputs = features[0];
	} else {
	    /* take first ninputs features */
	    if(features[0] > ninputs) features[0] = ninputs;
	    else ninputs = features[0];
	}
	/* don't go over nraw in any case */
	if(features[0] > nraw){
	    features[0] = ninputs = nraw;
	}
	/* check that all the features in the list are valid */
	for(i = 1; i <= features[0]; i++){
	    if(features[i] < 0 || features[i] >= nraw){
		fatal_error("entry in feature list out of range [0,%d-1]",
			    nraw);
	    }
	}
    }
    /* if we have managed to select 0 inputs, we have a problem */
    return(ninputs);
}

/****************************************************************************
  fix up the windowing paramters.
   window_width is the length of the window we will select frames from.
   window_frames, if it is not NULL is a list of frames to select
   if window_frames does not have a minimum value of 0 and a maximum
   value of window_width-1, there is an error
   the return value is the number of frames that will actually be used
   by the classifier.
****************************************************************************/

int fix_windows(int window_width, int *window_frames)
{
    int nwindow_frames, min_frame, max_frame, i;

    if(window_width < 1){
	warning("fix_windows: window_width must be >= 1.  Setting it to 1");
	window_width = 1;
    }
    if(window_frames == NULL || window_frames[0] <= 0){
	nwindow_frames = window_width;
    } else {
	/* find the max and min in the window_frames list */
	min_frame = max_frame = window_frames[1];
	for(i = 2; i <= window_frames[0]; i++){
	    if(window_frames[i] < min_frame) min_frame = window_frames[i];
	    if(window_frames[i] > max_frame) max_frame = window_frames[i];
	}
	/* check the min and max entries in window_frames */
	if(min_frame != 0)
	  warning("fix_windows: Frame 0 should be selected in each window.");
	if(max_frame < window_width -1){
	    warning("fix_windows: %s %d.\n%s %d %s",
		    "The highest selected frame is", max_frame,
		    "Shortening window_width to", max_frame+1,
		    "to match window_frames selections.");
	    window_width = max_frame+1;
	}
	if(max_frame > window_width - 1){
	    fatal_error("fix_windows: %s  Window_width = %d",
			"window_frames selects frames outside of window.",
			window_width);
	}
	nwindow_frames = window_frames[0];
    }
    return nwindow_frames;
}

int count_nwindow_frames(int window_width, int *window_frames)
{
    /* not doing the fancy checks for min and max */
    
    if(window_width < 1){
	return 1;
    }
    if(window_frames == NULL || window_frames[0] <= 0){
	return window_width;
    } else {
	return window_frames[0];
    }
}

/*** Parameter Table ***/
static PARAM featurescalars[] = {
    {"nraw",P_INT},
    {"ninputs",P_INT},
};
#define NFEATURESCALARS (sizeof(featurescalars)/sizeof(PARAM))

static PARAM featurearrays[] = {
    {"features", P_INT_VECTOR},
};
#define NFEATUREARRAYS (sizeof(featurearrays)/sizeof(PARAM))

/***---------------------------------------------------------------------------
* NAME:	 feature_read_paramfile()
* DESCRIPTION:
--------------------------------------------------------------------------***/
int feature_read_paramfile (FILE *fp, int **pfeatures, int *pnraw,
			    int *pninputs)
{
    int rv;
    
    /* assumes space has been allocated for features array already */
    param_setvarptr("ninputs", pninputs, featurescalars, NFEATURESCALARS);
    param_setvarptr("nraw", pnraw, featurescalars, NFEATURESCALARS);
    
    /* Get the feature model scalars */
    if ((rv = param_getblk(fp,featurescalars,NFEATURESCALARS)) < 0)
      return rv;
    
    *pfeatures = allocate_ivector(*pninputs+1);
    param_setvarptr("features", *pfeatures,
		    featurearrays, NFEATUREARRAYS);
    param_setdim("features", *pninputs+1, 0, featurearrays, NFEATUREARRAYS);
    if((rv = param_getblk(fp, featurearrays, NFEATUREARRAYS)) < 0)
      return rv;
    return 0;
} /* feature_read_paramfile() */

/***---------------------------------------------------------------------------
* NAME:	 feature_write_paramfile()
* DESCRIPTION:
--------------------------------------------------------------------------***/
int feature_write_paramfile (FILE *fp, int *features, int nraw, int ninputs)
{
    int rv;
    
    /* assumes space has been allocated for features array alwritey */
    param_setvarptr("features", features,
		    featurearrays, NFEATUREARRAYS);
    param_setvarptr("ninputs", &ninputs, featurescalars, NFEATURESCALARS);
    param_setvarptr("nraw", &nraw, featurescalars, NFEATURESCALARS);
    param_setdim("features", features[0]+1, 0,
		 featurearrays, NFEATUREARRAYS);
    
    /* Put the feature model scalars */
    if ((rv = param_putblk(fp,featurescalars,NFEATURESCALARS)) < 0)
      return rv;
    
    if((rv = param_putblk(fp, featurearrays, NFEATUREARRAYS)) < 0)
      return rv;
    return 0;
} /* feature_write_paramfile() */

/*** Parameter Table ***/
static PARAM windowscalars[] = {
    {"nwindow_frames", P_INT},
};
#define NWINDOWSCALARS (sizeof(windowscalars)/sizeof(PARAM))

static PARAM windowarrays[] = {
    {"window_frames", P_INT_VECTOR},
};
#define NWINDOWARRAYS (sizeof(windowarrays)/sizeof(PARAM))

/***---------------------------------------------------------------------------
* NAME:	 window_read_paramfile()
* DESCRIPTION:
--------------------------------------------------------------------------***/
int window_read_paramfile (FILE *fp, int *pnwindow_frames,
			   int **pwindow_frames)
{
    int rv;
    
    /* assumes space has been allocated for windows array already */
    param_setvarptr("nwindow_frames", pnwindow_frames,
		    windowscalars, NWINDOWSCALARS);
    /* Get the window model scalars */
    if ((rv = param_getblk(fp,windowscalars,NWINDOWSCALARS)) < 0)
      return rv;

    *pwindow_frames = allocate_ivector(*pnwindow_frames+1);
    param_setvarptr("window_frames", *pwindow_frames,
		    windowarrays, NWINDOWARRAYS);
    param_setdim("window_frames", *pnwindow_frames+1, 0,
		 windowarrays, NWINDOWARRAYS);
    
    if((rv = param_getblk(fp, windowarrays, NWINDOWARRAYS)) < 0)
      return rv;
    return 0;
} /* window_read_paramfile() */

/***---------------------------------------------------------------------------
* NAME:	 window_write_paramfile()
* DESCRIPTION:
--------------------------------------------------------------------------***/
int
window_write_paramfile (FILE *fp, int *window_frames, int nwindow_frames)
{
    int rv;

    /* assumes space has been allocated for window_frames array alwritey */
    param_setvarptr("window_frames", window_frames,
		    windowarrays, NWINDOWARRAYS);
    param_setvarptr("nwindow_frames", &nwindow_frames,
		    windowscalars, NWINDOWSCALARS);
    param_setdim("window_frames", window_frames[0]+1, 0,
		 windowarrays, NWINDOWARRAYS);
    
    /* Put the window model scalars */
    if ((rv = param_putblk(fp,windowscalars,NWINDOWSCALARS)) < 0)
      return rv;
    
    if((rv = param_putblk(fp, windowarrays, NWINDOWARRAYS)) < 0)
      return rv;
    return 0;
} /* window_write_paramfile() */
