/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include "nnlib.h"

/*****************************************************************************
* NAME: get_log_gauss_output_full()
* DESCRIPTION:
*	return the log10 gaussian output for a full covariance matrix
*
* EXAMPLE:
*
*****************************************************************************/

float
get_log_gauss_output_full(int ninputs, float *inputs, float *means,
			  float **inv_covar, float log_det_covar)
{
    float log_output;
    float p, q;
    int i, j;

    q = 0.;
    for(i = 0; i < ninputs; i++){
	p = 0.;
	for(j = 0; j < ninputs; j++){
	    p += (inputs[j] - means[j])*inv_covar[i][j];
	}
	q += (inputs[i] - means[i])*p;
    }
    log_output = q * M_LOG10E;
    log_output += log_det_covar + ninputs*LOG_2PI;
    log_output *= -0.5;
    return(log_output);
}

/*****************************************************************************
* NAME: get_log_gauss_output_diag()
* DESCRIPTION:
*	returns the log10 gaussian output for a diagonal covariance matrix
*
*		            T  -1
*	 /    - 1/2 ((x - m)  C  (x - m))  \
*	 |   e				   |
*    log |   ---------------------	   |
*	 |	  1/2	       (d/2)	   |
*	 \	|C|    * (2*PI)		   /
*
* EXAMPLE:
*
*****************************************************************************/

float
get_log_gauss_output_diag(int ninputs, float *inputs, float *means,
			  float *inv_var, float log_det_covar)
{
    float log_output, dist;

    dist = l2dist_weight(ninputs, inputs, means, inv_var);
    log_output = dist * M_LOG10E;
    log_output += log_det_covar + ninputs * LOG_2PI;
    log_output *= -0.5;
    return(log_output);
}

/*****************************************************************************
* NAME: get_log_gauss_output_sphere()
* DESCRIPTION:
*	returns the log10 gaussian output for a diagonal covariance matrix
*
*		          T   
*	 /  - 1/2 ((x - m) * 1/C * (x - m) \
*	 | e                               |
*  log10 | ---------------------           |
*	 |       d/2       (d/2)           |
*	 \      C    * (2*PI)              /
*
* EXAMPLE:
*
*****************************************************************************/

float
get_log_gauss_output_sphere(int ninputs, float *inputs, float *means,
			    float inv_var)
{
    float log_output, dist;

    dist = l2dist(ninputs, inputs, means);
    log_output = dist * M_LOG10E * inv_var;
    log_output += ninputs *(LOG_2PI - log10(inv_var));
    log_output *= -0.5;
    return(log_output);
}


/*****************************************************************************
* NAME: get_log_gauss_output_scalar()
* DESCRIPTION:
*	returns the log10 gaussian output for a one dimensional gaussian
*	To get the equivalent of the sphere or diag answers, sum up
*	the scalar answers for each value of input[i], mean[i], and inv_var[i]
*
*		          2   
*	 /  - 1/2 ((x - m) * 1/C \
*	 | e                     |
*  log10 | --------------------- |
*	 |       1/2       (1/2) |
*	 \      C    * (2*PI)    /
*
* EXAMPLE:
*
*****************************************************************************/
float
get_log_gauss_output_scalar(float input, float mean, float inv_var)
{
    float log_output, dist;

    dist = input - mean;
    dist *= dist;
    log_output = dist * M_LOG10E * inv_var;
    log_output += LOG_2PI - log10(inv_var);
    log_output *= -0.5;
    return(log_output);
}

/* returns -1 if it had to fix an eigenvalue.
   return 0 otherwise
*/
int
invert_and_calc_det(int ninputs, float **inv_covar, float *eigenvalues,
		    float max_ratio, float *p_log_det)
{
    int i, rv;

    rv = 0;
    square_matrix_invert(ninputs, inv_covar, inv_covar,
			 eigenvalues, max_ratio);
    *p_log_det = 0.;
    /* calculate the determinant */
    for(i = 0; i < ninputs; i++){
	/* check the eigenvalues to see that they are
	   all within max_ratio of each other.
	   Fix the ones that aren't */
	if((eigenvalues[i] < (eigenvalues[0]/max_ratio)) && (i != 0)){
	    warning("invert_and_calc_det:%s \n%s.",
		    " ill-conditioned matrix,",
		    "ignoring eigenvector with low eigenvalue");
	    eigenvalues[i] = eigenvalues[0]/max_ratio;
	    rv = -1;
	}
	*p_log_det += (float)log10((double)eigenvalues[i]);
    }
    return rv;
}

float
vector_vector_multiply(float *v1, float *v2, int n)
{
    int i;
    float tmp;

    tmp = 0.;
    for(i = 0; i < n; i++)
      tmp += v1[i] * v2[i];
    return tmp;
}
