/***--------------------------------------------------------------------------
* NAME:		get_subs.c
* PURPOSE:	routines for reading arrays of floats from files
*
* SEE ALSO: norm_subs.c
* GLOBALS:
*
* DESCRIPTION:
*	General purpose functions
*
*	load_trial_data()        loads trial data into structures and
*                                normalizes.
*       load_error_file_data()   loads error file data into an array of
*                                error line structure.
*       reallocate_error_lines() (re)allocates memory for an array of error
*                                line structure.
*	get_class_line()         loads a specified number of floats from
*			         a data file into an array.
*	get_map_line()           loads a specified number of floats from
*			         a data file into two arrays.
*	get_err_line()           loads a specified number of floats from
*		        	 an error file into an array
*
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/get_subs.c,v 11.4 2001/08/23 20:14:44 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/get_subs.c,v $
* CREATED: 1-Feb-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------
* Copyright 1990 MIT Lincoln Laboratory.
* This software is not for release and should neither be copied
* nor distributed without explicit permission of MIT Lincoln Laboratory.
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"

/* GLOBAL */
#define TRUE 1
#define FALSE 0



/*****************************************************************************
* NAME: load_trial_data()
* DESCRIPTION:
*       Load trial data into given structures and normalize
*		data_file:	pointer to trial data file
*		map_flag:	classification if FALSE, mapping if TRUE
*		n_trials:	number of trials (lines) in data set
*		nraw:		number of actual input values in data file
*		noutputs:	number of output values in each trial for
*				mapping
*		ninputs:	number of input values in each trial in data
*		trial_class:	set of desired classifications for this trial
*		desire:		array for desired map output data
*				(n_trials x noutputs)
*		trial_data:	array for trial data (n_trials x ninputs)
*		normalize:	type of normalization to use
*		data_fname:	name of data file
*		features:	features to use from whole set
*
* RETURNS:
*	number of trials actually loaded
****************************************************************************/
int
load_trial_data(FILE *data_file, int map_flag, int n_trials, int nraw,
		int noutputs, int ninputs, int *trial_class, float **desire,
		float **trial_data, int normalize, char *data_fname,
		int *features)
{
    int             j, return_value;

#if DEBUG
	printf("Enter load n_trials=%d normalize=%d\n", n_trials, normalize);
	fflush(stdout);
#endif
    for (j = 0; j < n_trials; j++) {
	if (map_flag) {
	    return_value =
	      get_map_line(data_file, data_fname, &trial_data[j][0],
			   &desire[j][0], nraw, ninputs, noutputs,
			   normalize, features); 
	} else {
	    return_value =
	      get_class_line(data_file, data_fname, &trial_data[j][0],
			     nraw, normalize, features, ninputs);
	    trial_class[j] = return_value;
	}
	if (return_value == EOF) break;
    }
    return(j);
}

/*****************************************************************************
* NAME: load_class_data()
* DESCRIPTION:
*       Load trial data into given structures and normalize
*		data_file:	pointer to trial data file
*		map_flag:	classification if FALSE, mapping if TRUE
*		n_trials:	number of trials (lines) in data set
*		nraw:		number of actual input values in data file
*		ninputs:	number of input values in each trial in data
*		trial_class:	set of desired classifications for this trial
*		trial_data:	array for trial data (n_trials x ninputs)
*		normalize:	type of normalization to use
*		data_fname:	name of data file
*		features:	features to use from whole set
*
* RETURNS:
*	number of trials actually loaded
*****************************************************************************/
int
load_class_data(FILE *data_file, int n_trials, int nraw, int ninputs,
		int *trial_class, float **trial_data, int normalize,
		char *data_fname, int *features)
{
    int             j, return_value;

#if DEBUG
	printf("Enter load n_trials=%d normalize=%d\n", n_trials, normalize);
	fflush(stdout);
#endif
    for (j = 0; j < n_trials; j++) {
	return_value = get_class_line(data_file, data_fname,
				      &trial_data[j][0], nraw, normalize,
				      features, ninputs);
	trial_class[j] = return_value;
	if (return_value == EOF) break;
    }
    return(j);
}

/*****************************************************************************
* NAME: load_map_data()
* DESCRIPTION:
*       Load trial data into given structures and normalize
*		data_file:	pointer to trial data file
*		map_flag:	classification if FALSE, mapping if TRUE
*		n_trials:	number of trials (lines) in data set
*		nraw:		number of actual input values in data file
*		noutputs:	number of output values in each trial for
*				mapping
*		ninputs:	number of input values in each trial in data
*		trial_class:	set of desired classifications for this trial
*		desire:		array for desired map output data
*				(n_trials x noutputs)
*		trial_data:	array for trial data (n_trials x ninputs)
*		normalize:	type of normalization to use
*		data_fname:	name of data file
*		features:	features to use from whole set
*
* RETURNS:
*	number of trials actually loaded
*****************************************************************************/
int
load_map_data(FILE *data_file, int n_trials, int nraw, int noutputs,
	      int ninputs, float **desire, float **trial_data, int normalize,
	      char *data_fname, int *features)
{
    int             j, return_value;

#if DEBUG
	printf("Enter load n_trials=%d normalize=%d\n", n_trials, normalize);
	fflush(stdout);
#endif
    for (j = 0; j < n_trials; j++) {
	return_value = get_map_line(data_file, data_fname, &trial_data[j][0],
				    &desire[j][0], nraw, ninputs, noutputs,
				    normalize, features);
	if (return_value == EOF) break;
    }
    return(j);
}

#define TLEN 128

int
load_seq_files(char *path, char *filename, DATA_BLOCK **block_list,
	       char **labels, int normalize, int *features, int ninputs,
	       int nraw, int nsequences)
{
    int nfiles, j, n, t_rtn, start = 0;
    FILE *sfile, *data_file;
    DATA_BLOCK *bl;
    static char prog_name[] = "load_seq_files", *fullname;

    /* open the file and count the number of sequences */
    fullname = path_build(path, filename);
    data_file = nn_fopen(fullname, "r", "for reading list of data files");
    nfiles = get_n_lines(data_file);
    if(nsequences < nfiles) nfiles = nsequences; /* stop early if asked */

    /* allocate the block list */
    bl = allocate_structvector(nfiles, sizeof(DATA_BLOCK));
    *block_list = bl;

    /* fill the block structures for the sequences */
    for(j = 0; j < nfiles; j++){
	/* read and parse the class */
	sfile = get_seq_file(data_file, path, filename, &(bl[j]), labels);
	if(sfile == NULL) break;
	/* check the number of inputs */
	if(bl[j].ninputs < nraw){
	    fatal_error("%s: %s %s.\nFile has %d inputs.  Needs %d",
			prog_name, "Not enough inputs in", bl[j].fname,
			bl[j].ninputs, nraw);
	}
	if(nraw == 0) nraw = bl[j].ninputs;
	if(nraw < bl[j].ninputs){
	    warning("%s: %s %s.\nFile has %d inputs.  Only needed %d",
		    prog_name, "Too many inputs in", bl[j].fname,
		    bl[j].ninputs, nraw);
	}

	/* fill the data matrix for this block */
	bl[j].data_matrix = allocate_matrix(bl[j].npatterns, bl[j].ninputs);
	for(n = 0; n < bl[j].npatterns; n++){
	    /* read line */
	    t_rtn = get_seq_line(sfile, bl[j].fname, &bl[j].data_matrix[n][0],
				 nraw, normalize, features, ninputs);
	    if(t_rtn == EOF) break;
	}
	/* fix up the start field */
	bl[j].start = start;
	start += bl[j].npatterns;
	/* close the sequence file */
	nn_fclose (sfile, bl[j].fname, "for reading sequence data");
    }
    return nfiles;
}


/*****************************************************************************
* NAME: load_error_file_data(fptr, filename, error_lines, error_file_verbosity,
*                            ninputs, noutputs)
* DESCRIPTION:
*	Gets a line from the classification error file, advancing fptr.
*
* RETURNS:
*	npatterns: the number of patterns stored in the file
*                 = the number of lines read from the file.
*
*	reports fatal error if there are not enough floats on the lines
*	or if the tokens read from the lines are not correct types.
*
*****************************************************************************/
int
load_error_file_data(FILE *fptr, char *fname, ERROR_LINE **err_lines,
		     int error_file_verbosity, int ninputs, int noutputs)
{
    ERROR_LINE *lines;
    int npatterns=0;    /* Start counting patterns */
    int old_size=0, new_size=BLOCK_SIZE;

    /* pointer substitution for clarity */
    lines = *err_lines;

    /* initial memory allocation: old_size = 0, new_size = BLOCK_SIZE */
    lines = (ERROR_LINE *)
      reallocate_error_lines(lines,
			     old_size, new_size, ninputs, noutputs);

    /* reading patterns from the error file */ 
    while(get_err_line(fptr, fname, &(lines[npatterns]),
		       error_file_verbosity) != EOF){

	/* counting patterns read from the error file */
	npatterns++;

	/* Expanding the err_line vector if neccessary */
	if(npatterns >= new_size){
	    old_size = new_size;
	    new_size += BLOCK_SIZE;
	    lines = (ERROR_LINE *)
	      reallocate_error_lines
		(lines, old_size, new_size, ninputs, noutputs); 
	}
    }

    /* back pointer-substitution */
    *err_lines = lines;

    return(npatterns);
}



/*****************************************************************************
* NAME: reallocate_error_lines
*          (error_lines, old_size, new_size, ninputs, noutputs)
* DESCRIPTION:
*       Reallocates memory for error_lines of type ERROR_LINE.
*       If the old_size = 0, the function allocates new memory
*       for a vector with new_size elements. 
*
* RETURNS:
*      The pointer to error_lines with the new memory allocation.
*
*****************************************************************************/
ERROR_LINE *
reallocate_error_lines (ERROR_LINE *error_lines, int old_size, int new_size,
			int ninputs, int noutputs)
{
    int i;
    static int line_size = sizeof(ERROR_LINE);

    if(old_size == 0){
	if(error_lines != NULL) deallocate_structvector((void *) error_lines);

	error_lines = (ERROR_LINE *)
	  allocate_structvector(new_size, line_size);
	for(i = 0; i < new_size; i++){
	    error_lines[i].ninputs = ninputs;
	    error_lines[i].inputs = allocate_vector(ninputs);
	    error_lines[i].noutputs = noutputs;
	    error_lines[i].outputs = allocate_vector(noutputs);
	    error_lines[i].desired_outputs = allocate_vector(noutputs);
	}
    }
    else{
	error_lines = (ERROR_LINE *)
	  reallocate_structvector((void *) error_lines,
				  old_size, new_size, line_size);
	for(i = old_size; i < new_size; i++){
	    error_lines[i].ninputs = ninputs;
	    error_lines[i].inputs = allocate_vector(ninputs);
	    error_lines[i].noutputs = noutputs;
	    error_lines[i].outputs = allocate_vector(noutputs);
	    error_lines[i].desired_outputs = allocate_vector(noutputs);
	}
    }

    return((ERROR_LINE *) error_lines);
}


/*****************************************************************************
* NAME: get_class_line(fptr,filename,input,nraw,norm_flag,features,ninputs)
* DESCRIPTION:
*	Gets a line from the file, advancing fptr.
*	Read the first token from the line, which is the integer class number.
*	Starting from the second number on the line, load nraw floats
*	into the array input.
*	If norm_flag is set, normalizes input vector
*       Normalization must be precalculated using the norm_init(),
*       and norm_update().
*
*	features controls which entries in the full array
*	actually get returned in input, when not all are wanted
*
* RETURNS:
*	class: the first integer on the line read from fptr
*	EOF on end of file
*
*	reports fatal error if there are not enough floats on the line
*	or if the tokens read from the line are not interpretable as floats
*
*****************************************************************************/
int
get_class_line(FILE *fptr, char *filename,/* only used in fatal error */
	       float *input,/* assumed to be long enough */ int nraw,
	       int norm_flag, int *features, int ninputs)
     /* get the line, then strip off tokens from it, parsing them as floats */
{
    static float   *all_inputs = NULL;
    double temp_input; /* for the actual read in scanf */
    static int	nraw_store = 0;
    char            token[64], c;
    int             i, class, end_of_line = 0, rv;

    if(nraw_store != nraw){
	if(all_inputs != NULL) deallocate_vector(all_inputs);
	all_inputs = allocate_vector(nraw);
	nraw_store = nraw;
    }

    /* cannot use just fscanf, it ignores the ends of lines */
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if(rv == 0) end_of_line = 1;
    if(sscanf(token, "%d", &class) == 0){
	fatal_error("get_class_line: %s %s\n%s (%s) %s",
		    "Error reading from file", filename,
		    "First token", token, "was not integer");
    }
    /* check class against noutputs, which we need to be given */
#ifdef notdef
    if(class < 0 || class > noutputs-1){
	fatal_error("get_class_line: Class from file (%d) %s (%d)",
		    class, " >= number of classes", noutputs);
    }
#endif
    for (i = 0; i < nraw; i++) {
	if(end_of_line){
	    /* we were expecting another token, the line stopped short */
	    fatal_error("get_class_line: %s %s\nOnly %d %s %d",
			"Error reading from file", filename, i,
			"elements in line, wanted", nraw);
	}
	if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	if(rv == 0) end_of_line = 1;
	if(sscanf(token, "%lg", &temp_input) == 0){
	    fatal_error("get_class_line: %s %s\nToken %s %s",
			"Error reading from file", filename,
			token, "was not a float");
	}
	all_inputs[i] = (float) temp_input;
    }
    /* read off any of the rest of the line */
    if(!end_of_line) do c = getc(fptr); while(c != '\n');

    /* normalize input data, if requested */
    if (norm_flag)
	norm_apply(&all_inputs[0]);

    /* take out entries, if requested */
    if(ninputs != 0){
	if(features != NULL && features[0] != 0)
	  for(i = 0; i < ninputs; i++) input[i] = all_inputs[features[i+1]];
	else
	  for(i = 0; i < ninputs; i++) input[i] = all_inputs[i];
    } else {
	for(i = 0; i < nraw; i++) input[i] = all_inputs[i];
    }

    return (class);
}
/**********************************************
 * NAME: write_class_line(fptr, filename, inputs, class, ninputs)
 * DESCRIPTION: writes a line to the file with the input vector and class
 *
 **************************************************/
void
write_class_line(FILE *fptr, char *filename, float *inputs, int class,
		 int ninputs)
{
    int i;
    if(fprintf(fptr, "%d", class) == EOF)
      fatal_error("write_class_line: Error writing to file %s",
		  filename);
    for(i = 0; i < ninputs; i++)
      if(fprintf(fptr, " %g", inputs[i]) == EOF)
      fatal_error("write_class_line: Error writing to file %s",
		  filename);
    if(fprintf(fptr, "\n") == EOF)
      fatal_error("write_class_line: Error writing to file %s",
		  filename);
}

/*****************************************************************************
* NAME: get_map_line(fptr,filename,input,output,nraw,
*		ninputs,noutputs,norm_flag, features)
* DESCRIPTION:
*	Gets a line from the file, advancing fptr.
*	Reads the first nraw tokens, loading the entries specified in
*	features into the array input.
*	Loads the next noutputs floats into output.
*	If norm_flag is set, normalizes input vector.
*	All tokens are assumed to be strings which represent floating point
*	numbers
*       Normalization must be precalculated using the norm_init(),
*       and norm_update().
*
* RETURNS:
*	0 on successful completion
*	EOF on end of file
*
*	Reports fatal error if there are not enough floats on the line,
*	or if the tokens read from the line are not interpretable as floats.
*
*****************************************************************************/
int
get_map_line(FILE *fptr, char *filename,/* only used in fatal error */
	     float *input, float *output,/* arrays of floats,
					    assumed to be long enough */
	     int nraw, int ninputs, int noutputs, int norm_flag, int *features)
{
    static float	*all_inputs = NULL;
    double	temp_input;
    static int		nraw_store = 0;
    char	token[64];
    int		i, rv, end_of_line;

    if(nraw_store != nraw){
	if(all_inputs != NULL) deallocate_vector(all_inputs);
	all_inputs = allocate_vector(nraw);
	nraw_store = nraw;
    }

    end_of_line = 0;
    for(i = 0; i < nraw; i++){
	if(end_of_line){
	    /* we were expecting another token, the line stopped short */
	    fatal_error("get_map_line: %s %s\nOnly %d %s %d",
			"Error reading from file", filename, i,
			"elements in line, wanted", nraw + noutputs);
	}
	if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	if (sscanf(token, "%lg", &temp_input) == 0) {
	    fatal_error("get_map_line: %s %s\nToken %s %s",
			"Error reading from file", filename,
			token, "was not a float");
	}
	all_inputs[i] = (float)temp_input;
	if(rv == 0) end_of_line = 1;
    }
    for (i = 0; i < noutputs; i++) {
	if(end_of_line){
	    /* we were expecting another token, the line stopped short */
	    fatal_error("get_map_line: %s %s\nOnly %d %s %d",
			"Error reading from file", filename, i,
			"elements in line, wanted", nraw + noutputs);
	}
	if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	if (sscanf(token, "%lg", &temp_input) == 0) {
	    fatal_error("get_map_line: %s %s\nToken %s %s",
			"Error reading from file", filename,
			token, "was not a float");
	}
	output[i] = (float) temp_input;
	if(rv == 0) end_of_line = 1;
    }
    /* read off any of the rest of the line */
    if(!end_of_line) skip_to_eol(fptr);

    /* normalize input data, if requested */
    if (norm_flag) norm_apply(&all_inputs[0]);

    /* take out entries, if requested */
    if(ninputs != 0){
	if(features != NULL && features[0] != 0)
	  for(i = 0; i < ninputs; i++) input[i] = all_inputs[features[i+1]];
	else
	  for(i = 0; i < ninputs; i++) input[i] = all_inputs[i];
    } else
      for(i = 0; i < nraw; i++) input[i] = all_inputs[i];

    return (0);
}

/*****************************************************************************
* NAME: get_err_line(fptr, filename, err_line_p, error_file_verbosity)
* DESCRIPTION:
*       Gets a line from the file, advancing fptr.
*       Reads the first token from the line, which is the integer err number.
*       Starting from the second number on the line, load ninputs floats
*       into the array input.
*       Starting from the next number on the line, load noutputs floats
*       into the array output.
*       
*       error_file_verbosity = 
*                       [1] short error file: 4 integers, 1 float;
*                       [2] mid error file: 4 integers, 1 float,
*                          noutputs floats.
*			[3] long error file: 4 integers, 1 float,
*			   noutputs floats, ninputs floats.
*
* RETURNS:
*       err: the first integer on the line read from fptr
*       EOF on end of file
*
*       reports fatal error if there are not enough floats on the line
*       or if the tokens read from the line are not interpretable as floats
*
*****************************************************************************/
#define NONE 0
#define SHORT 1
#define OUTPUTS 2
#define INS_AND_OUTS 3
int
get_err_line(FILE *fptr, char *filename,/* only used in fatal error */
	     ERROR_LINE *err_line_p, int error_file_verbosity)
{
    char	token[64];
    double temp_input;
    int		i, end_of_line, rv;

    end_of_line = 0;

    /* if we don't want any data, return */
    if(error_file_verbosity < SHORT){
	return !EOF;
    }

    /* file is at least SHORT, interpret first 4 tokens */
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->trial) == 0) {
	fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "First token not an integer");
    }
    if (rv == 0)
      fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Not enough tokens on line");
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->desired_class) == 0) {
	fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Second token was not integer");
    }
    if (rv == 0)
	fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Not enough tokens on line");
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->class) == 0) {
	fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Third token was not integer");
    }
    if (rv == 0)
	fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Not enough tokens on line");
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->err_f) == 0) {
	fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Fourth token was not integer");
    }
    if (rv == 0)
      fatal_error("get_err_line: %s %s\n%s",
		  "Error reading from file", filename,
		  "Not enough tokens on line");
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if(rv == 0) end_of_line = 1;
    if (sscanf(token, "%lf", &temp_input) == 0) {
	fatal_error("get_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Fifth token was not float");
    }
    err_line_p->rmserr = (float) temp_input;

    if (error_file_verbosity >= OUTPUTS){
	for (i = 0; i < err_line_p->noutputs; i++) {
	    if(end_of_line)
	      fatal_error("get_err_line: %s %s\nOnly %d %s %d",
			  "Error reading from file", filename, i,
			  "elements in line, wanted", err_line_p->noutputs);
	    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	    if(rv == 0) end_of_line = 1;
	    if (sscanf(token, "%lf", &temp_input) == 0) {
		fatal_error("get_err_line: %s %s\nToken %s %s",
			    "Error reading from file", filename,
			    token, "was not a float");
	    }
	    err_line_p->outputs[i] = (float) temp_input;
	}
    }
    if(error_file_verbosity >= INS_AND_OUTS){
	for (i = 0; i < err_line_p->ninputs; i++) {
	    if(end_of_line)
	      fatal_error("get_err_line: %s %s\nOnly %d %s %d",
			  "Error reading from file", filename, i,
			  "elements in line, wanted", err_line_p->ninputs);
	    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	    if(rv == 0) end_of_line = 1;
	    if (sscanf(token, "%lf", &temp_input) == 0) {
		fatal_error("get_err_line: %s %s\nToken %s %s",
			    "Error reading from file", filename,
			    token, "was not a float");
	    }
	    err_line_p->inputs[i] = (float) temp_input;
	}
    }

    /* read off the rest of the line */
    if(!end_of_line){
	skip_to_eol(fptr);
	end_of_line = 1;
    }

    return (!EOF);
}

/*****************************************************************************
* NAME: get_map_err_line(fptr, filename, err_line_p, error_file_verbosity)
* DESCRIPTION:
*       Gets a line from the file, advancing fptr.
*       Reads the first token from the line, which is the integer err number.
*	Reads the cost, which is a float
*       Starting from the next number on the line, load noutputs floats
*       into the array output.
*       Starting from the next number on the line, load ninputs floats
*       into the array input.

*       error_file_verbosity = 
*                       [1] short error file: 1 integer, 1 float
*                       [2] mid error file: 1 integer, 1 float,
*                          noutputs+noutputs floats.
*			[3] long error file: 1 integer, 1 float,
*			   noutputs+noutputs floats, ninputs floats.
*
*       
* RETURNS:
*       err: the first integer on the line read from fptr
*       EOF on end of file
*
*       reports fatal error if there are not enough floats on the line
*       or if the tokens read from the line are not interpretable as floats
*
*****************************************************************************/
int
get_map_err_line(FILE *fptr, char *filename,/* only used in fatal error */
		 ERROR_LINE *err_line_p, int error_file_verbosity)
{
    char	token[64];
    double temp_input;
    int		i, end_of_line, rv;

    end_of_line = 0;

    /* if we don't want any data, return */
    if(error_file_verbosity < SHORT){
	return !EOF;
    }

    /* file is at least SHORT, interpret first 2 tokens */
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->trial) == 0) {
	fatal_error("get_map_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "First token not an integer");
    }
    if (rv == 0)
      fatal_error("get_map_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Not enough tokens on line");
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if(rv == 0) end_of_line = 1;
    if (sscanf(token, "%lf", &temp_input) == 0) {
	fatal_error("get_map_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Fifth token was not float");
    }
    err_line_p->rmserr = (float) temp_input;

    if (error_file_verbosity >= OUTPUTS){
	for (i = 0; i < err_line_p->noutputs; i++) {
	    if(end_of_line)
	      fatal_error("get_map_err_line: %s %s\nOnly %d %s %d",
			  "Error reading from file", filename, i,
			  "elements in line, wanted", err_line_p->noutputs);
	    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	    if(rv == 0) end_of_line = 1;
	    if (sscanf(token, "%lf", &temp_input) == 0) {
		fatal_error("get_map_err_line: %s %s\nToken %s %s",
			    "Error reading from file", filename,
			    token, "was not a float");
	    }
	    err_line_p->desired_outputs[i] = (float) temp_input;
	}
	for (i = 0; i < err_line_p->noutputs; i++) {
	    if(end_of_line)
	      fatal_error("get_map_err_line: %s %s\nOnly %d %s %d",
			  "Error reading from file", filename, i,
			  "elements in line, wanted", err_line_p->noutputs);
	    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	    if(rv == 0) end_of_line = 1;
	    if (sscanf(token, "%lf", &temp_input) == 0) {
		fatal_error("get_map_err_line: %s %s\nToken %s %s",
			    "Error reading from file", filename,
			    token, "was not a float");
	    }
	    err_line_p->outputs[i] = (float) temp_input;
	}
    }
    if(error_file_verbosity >= INS_AND_OUTS){
	for (i = 0; i < err_line_p->ninputs; i++) {
	    if(end_of_line)
	      fatal_error("get_map_err_line: %s %s\nOnly %d %s %d",
			  "Error reading from file", filename, i,
			  "elements in line, wanted", err_line_p->ninputs);
	    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	    if(rv == 0) end_of_line = 1;
	    if (sscanf(token, "%lf", &temp_input) == 0) {
		fatal_error("get_map_err_line: %s %s\nToken %s %s",
			    "Error reading from file", filename,
			    token, "was not a float");
	    }
	    err_line_p->inputs[i] = (float) temp_input;
	}
    }

    /* read off the rest of the line */
    if(!end_of_line){
	skip_to_eol(fptr);
	end_of_line = 1;
    }

    return (!EOF);
}

/*****************************************************************************
* NAME: get_seq_err_line(fptr, filename, err_line_p, error_file_verbosity)
* DESCRIPTION:
*       Gets a line from the file, advancing fptr.
*       Reads the first token from the line, which is the integer err number.
*	reads the desired class, class, whether they match,
*	and the total cost
*       Starting from the next number on the line, load noutputs floats
*       into the array output.
*       
*       error_file_verbosity = 
*                       [1] short error file: 4 integers, 1 float;
*                       [2] mid error file: 4 integers, 1 float,
*                          noutputs floats.
*
* RETURNS:
*       err: the first integer on the line read from fptr
*       EOF on end of file
*
*       reports fatal error if there are not enough floats on the line
*       or if the tokens read from the line are not interpretable as floats
*
*****************************************************************************/
int
get_seq_err_line(FILE *fptr, char *filename,/* only used in fatal error */
		 ERROR_LINE *err_line_p, int error_file_verbosity)
{
    char	token[64];
    double temp_input;
    int		i, end_of_line, rv;

    end_of_line = 0;

    /* if we don't want any data, return */
    if(error_file_verbosity < SHORT){
	return !EOF;
    }

    /* file is at least SHORT, interpret first 4 tokens */
    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->trial) == 0) {
	fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "First token not an integer");
    }
    if (rv == 0)
      fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Not enough tokens on line");

    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->desired_class) == 0) {
	fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Second token was not integer");
    }
    if (rv == 0)
	fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Not enough tokens on line");

    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->class) == 0) {
	fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Third token was not integer");
    }
    if (rv == 0)
	fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Not enough tokens on line");

    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if (sscanf(token, "%d", &err_line_p->err_f) == 0) {
	fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Fourth token was not integer");
    }
    if (rv == 0)
      fatal_error("get_seq_err_line: %s %s\n%s",
		  "Error reading from file", filename,
		  "Not enough tokens on line");

    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
    if(rv == 0) end_of_line = 1;
    if (sscanf(token, "%lf", &temp_input) == 0) {
	fatal_error("get_seq_err_line: %s %s\n%s",
		    "Error reading from file", filename,
		    "Fifth token was not float");
    }
    err_line_p->rmserr = (float) temp_input;

    if (error_file_verbosity >= OUTPUTS){
	for (i = 0; i < err_line_p->noutputs; i++) {
	    if(end_of_line)
	      fatal_error("get_seq_err_line: %s %s\nOnly %d %s %d",
			  "Error reading from file", filename, i,
			  "elements in line, wanted", err_line_p->noutputs);
	    if((rv = get_line_token(fptr, token, 64)) == EOF) return EOF;
	    if(rv == 0) end_of_line = 1;
	    if (sscanf(token, "%lf", &temp_input) == 0) {
		fatal_error("get_seq_err_line: %s %s\nToken %s %s",
			    "Error reading from file", filename,
			    token, "was not a float");
	    }
	    err_line_p->outputs[i] = (float) temp_input;
	}
    }

    /* read off the rest of the line */
    if(!end_of_line){
	skip_to_eol(fptr);
	end_of_line = 1;
    }

    return (!EOF);
}

/*****************************************************************************
* NAME: get_seq_line(fptr,filename,input,nraw,norm_flag,features,ninputs)
* DESCRIPTION:
*	Gets a line from the file, advancing fptr.
*	load nraw floats into the array input.
*	If norm_flag is set, normalizes input vector
*       Normalization must be precalculated using the norm_init(),
*       and norm_update().
*
*	features controls which entries in the full array
*	actually get returned in input, when not all are wanted
*
* RETURNS:
*	0 on success
*	EOF on end of file
*	if there was more on the line, returns > 0
*
*	reports fatal error if there are not enough floats on the line
*	or if the tokens read from the line are not interpretable as floats
*
*****************************************************************************/
int
get_seq_line(FILE *fptr, char *filename,/* only used in fatal error */
	     float *input,/* assumed to be long enough */ int nraw,
	     int norm_flag, int *features, int ninputs)
     /* get the line, then strip off tokens from it, parsing them as floats */
{
    static float   *all_inputs = NULL;
    double temp_input;
    static int	nraw_store = 0;
    char            token[64];
    int             i, end_of_line, rv = EOF;

    if(nraw_store != nraw){
	if(all_inputs != NULL) deallocate_vector(all_inputs);
	all_inputs = allocate_vector(nraw);
	nraw_store = nraw;
    }

    end_of_line = 0;
    for (i = 0; i < nraw; i++) {
	if(end_of_line){
	    /* we were expecting another token, the line stopped short */
	    fatal_error("get_seq_line: %s %s\nOnly %d %s %d",
			"Error reading from file", filename, i,
			"elements in line, wanted", nraw);
	}
	rv = get_line_token(fptr, token, 64);
	if(rv == EOF) return EOF;
	if(rv == 0) end_of_line = 1;
	if(sscanf(token, "%lg", &temp_input) == 0){
	    fatal_error("get_seq_line: %s %s\nToken %s %s",
			"Error reading from file", filename,
			token, "was not a float");
	}
	all_inputs[i] = (float) temp_input;
    }
    /* read off any of the rest of the line */
    if(!end_of_line) skip_to_eol(fptr);

    /* normalize input data, if requested */
    if (norm_flag)
	norm_apply(&all_inputs[0]);

    /* take out entries, if requested */
    if(ninputs != 0){
	if(features != NULL && features[0] != 0)
	  for(i = 0; i < ninputs; i++) input[i] = all_inputs[features[i+1]];
	else
	  for(i = 0; i < ninputs; i++) input[i] = all_inputs[i];
    } else {
	for(i = 0; i < nraw; i++) input[i] = all_inputs[i];
    }

    return (rv);
}

/* fills string with a token read from fptr.
   length is the maximum number of charaters to read
   returns: 0 if the line ends at the end of the token
	    EOF if the file ends at the end of the token
	    1 if there is more white space at the end of the token
	    2 if the token was longer than length
   NOTE: If a line ends in white space, the last token returned from
   the line will be empty (the string "");
   There are two ways I deal with this. Usually, I know how many tokens
   to expect. I read that many and then use skip_to_eol() to finish
   the line. When I don't know the number of tokens (cf get_n_tokens() )
   I test for and ignore empty strings. Logically, such empty strings
   only come up as the last token on lines that end with whitespace.
*/
int
get_line_token(FILE *fptr, char *string, int length)
{
    int i;
    char c;
    i = 0;
    while(1){ /* skip initial white space */
	c = getc(fptr);
	if(c != ' ' && c != '\t') break;
    }
    if(c == '\n'){ string[i] = '\0'; return 0;}
    if(c == (char)EOF){ string[i] = '\0'; return EOF;}
    string[i] = c;
    i++;
    for(; i < length; i++){
	c = getc(fptr);
	if(c == ' ' || c == '\t'){ string[i] = '\0'; return 1;}
	if(c == '\n'){ string[i] = '\0'; return 0;}
	if(c == (char)EOF){ string[i] = '\0'; return EOF;}
	string[i] = c;
    }
    string[length-1] = '\0';
    /* the rest of the string is left on the stream */
    return 2;
}


void
skip_to_eol(FILE *fp)
{
    char c;
    do c = getc(fp);
    while(c != '\n' && c != (char)EOF);
}

int
label2class(char *token, char **labels)
{
    int rv;
    int nlabels;

    if(labels == NULL) {
      if(sscanf(token, "%d", &rv) == 0) return -1;
      else return rv;
    }

    nlabels = atoi(labels[0]);
    if(nlabels == 0) {
      if(sscanf(token, "%d", &rv) == 0) return -1;
      else return rv;
    }

    for(rv = 0; rv < nlabels; rv++){
	if(strcmp(labels[rv+1], token) == 0)
	  return rv;
    }
    /* allowing numbered labels, for backward compatibility */
    if(sscanf(token, "%d", &rv) == 1)
      if(rv >= 0 && rv < nlabels) return rv;
    return -1;
}
    
int
get_n_lines(FILE *fp)
{
    int i = 0;
    char c;
    do{
	c = fgetc(fp);
	if(c == '\n') i++;
    } while(c != (char)EOF);
    rewind(fp);
    return i;
}

int
get_n_tokens(FILE *fp)
{
    int i;
    char token[128];
    int rv;

    i = 0;
    do{
	rv = get_line_token(fp, token, 128);
	if(token[0] != '\0') i++;
    } while(rv != EOF && rv != 0);
    rewind(fp);
    return i;
}

FILE *
get_seq_file(FILE *list_file, char *data_path, char *filename,
	     DATA_BLOCK *blockp, char **labels)
{
    int t_rtn, start = 0, duration = 0, n;
    char token[TLEN];
    FILE *sfile;

    if((t_rtn = get_line_token(list_file, token, TLEN)) != 1) return NULL;
    blockp->class = label2class(token, labels);
    if(blockp->class < 0){
	fatal_error("get_seq_file: %s %s/%s\n%s (%s) %s",
		    "Error reading from file", data_path, filename,
		    "First token", token, "is not a class label");
    }
    /* read and parse the file name */
    if((t_rtn = get_line_token(list_file, token, TLEN)) == EOF) return NULL;
    strcpy(blockp->fname, path_build(data_path, token));
    
    blockp->start = 0;
    /* get the start and duration */
    /* if they are not both present, start is 0, duration is file length */
    if(t_rtn != 0 && t_rtn != EOF){
	t_rtn = get_line_token(list_file, token, TLEN);
	if(t_rtn != 0 && t_rtn != EOF){
	    if(sscanf(token, "%d", &start) == 1){
		t_rtn = get_line_token(list_file, token, TLEN);
		if(sscanf(token, "%d", &duration) != 1){
		    duration = 0;
		}
		if(t_rtn != 0 && t_rtn != EOF) skip_to_eol(list_file);
	    } else {
		start = 0;
		duration = 0;
	    }
	}
    }
    /* open the file and count the lines and the number of inputs */
    sfile = nn_fopen(blockp->fname, "r", "for read of sequence file");
    blockp->npatterns = get_n_lines(sfile);
    blockp->ninputs = get_n_tokens(sfile);
    /* skip the lines before start */
    if(start != 0 || duration != 0){
	if(start + duration <= blockp->npatterns){
	    for(n = 0; n < start; n++) skip_to_eol(sfile);
	    blockp->npatterns = duration;
	} else {
	    warning("get_seq_file: File %s/%s %s(%d lines) %s(%d) %s(%d)",
		    data_path, filename, "is not long enough ",
		    blockp->npatterns, "for start ", start,
		    "and duration ", duration);
	}
    }
    /* debug stuff */
    return(sfile);
}

void
find_class_priors(FILE *fp, int npatterns, int noutputs,
		  float *priors_list/* a vector noutputs+1 long */)
{
    int i, rv, class, npat;
    char token[64];

    /* initialize priors_list */
    for(i = 1; i <= noutputs; i++) priors_list[i] = 0.;
    for(npat = 0; npat < npatterns; npat++){
	/* read the class token from each line, skipping to the end */
	rv = get_line_token(fp, token, 64);
	if(rv == EOF) break;
	if(sscanf(token, "%d", &class) == 0) break;
	/* update priors_list */
	if(class < noutputs)
	  priors_list[class+1] += 1.;
	else{
	    warning("find_seq_priors: %s (%d) doesn't match data file",
		    "noutputs from defaults file", noutputs);
	    break;
	}
	if(rv != 0) skip_to_eol(fp); /* if not already at a line end */
    }
    /* I could normalize the priors here, but it makes the displayed strings
       in LNKnet too long */
    rewind(fp);
}

void
find_seq_priors(FILE *fp, int npatterns, int noutputs,
		float *priors_list/* a vector noutputs+1 long */)
{
    int i, rv, class, npat;
    char token[64];

    /* initialize priors_list */
    for(i = 1; i <= noutputs; i++) priors_list[i] = 0.;
    for(npat = 0; npat < npatterns; npat++){
	/* read the class token from each line, skipping to the end */
	rv = get_line_token(fp, token, 64);
	if(rv == EOF) break;
	if(sscanf(token, "%d", &class) == 0) break;
	/* update priors_list */
	if(class < noutputs)
	  priors_list[class+1] += 1.;
	else{
	    warning("find_seq_priors: %s (%d) doesn't match data file",
		    "noutputs from defaults file", noutputs);
	    break;
	}
	if(rv != 0) skip_to_eol(fp); /* if not already at a line end */
    }
    /* I could normalize the priors here, but it makes the displayed strings
       in LNKnet too long */
    rewind(fp);
}

int
find_frame_priors(FILE *fp, int npatterns, int noutputs,
		  float *priors_list,/* a vector noutputs+1 long */
		  char *data_path, char *filename, char **labels)
{
    int i, npat, total;
    DATA_BLOCK tmp_block;
    FILE *tmpfp;

    total = 0;
    /* initialize priors_list */
    for(i = 1; i <= noutputs; i++) priors_list[i] = 0.;
    for(npat = 0; npat < npatterns; npat++){
	tmpfp = get_seq_file(fp, data_path, filename, &tmp_block,
			     labels);
	if(tmpfp == NULL) break;
	priors_list[tmp_block.class+1] += tmp_block.npatterns;
	total += tmp_block.npatterns;
	nn_fclose(tmpfp, tmp_block.fname, "for read of sequence file");
    }
    rewind(fp);
    return total;
}
