/***---------------------------------------------------------------------------
* NAME:		heap_subs.c
* PURPOSE:	routines for sorting using a heap
*
* GLOBALS: 
*
* DESCRIPTION:
*	General purpose functions
*
*	Sorts to give smallest to largest order
*		heap_init() -- sorts float array in heap order
*			(largest value will be on top)
*		reheap() -- moves first element in float array to return
*			array to heap order
*		heap_sort() -- sorts float array using a heap
*
*	Sorts to give largest to smallest order
*		rev_heap_init() -- sorts float array in heap order
*			(smallest value will be on top)
*		rev_reheap() -- moves first element in float array to return
*			array to heap order
*		rev_heap_sort() -- sorts float array using a heap
* 
*	also, these same for integers:
*		heap_iinit(), ireheap(), heap_isort(),
*		rev_heap_iinit(), rev_ireheap(), rev_heap_isort()
*	and for doubles:
*		heap_dinit(), dreheap(), heap_dsort(),
*		rev_heap_dinit(), rev_dreheap(), rev_heap_dsort()
* 
* To sort an array:
*   allocate and fill xarray, the float (int, double) array of values.
*   allocate an array the same length (n) as an index.
*   heap_sort(n, xarray, index);
*   then use xarray[index[i]] to access the ith entry in the sorted array.
*   xarray is unaltered, only the index is changed.
*   To resort xarray and one other array, call heap_sort2()
*   OR:
*   
*   allocate another index iindex, which is a permutation index
*   #initialize the permutation index
*   for(i = 0; i < n; i++) iindex[index[i]] = i;
*   #sort the array and keep the indicies up to date.
*   for(i = 0; i < n; i++){
*     #swap the ith entry in the unsorted and sorted xarray
*     temp = xarray[index[i]];
*     xarray[index[i]] = xarray[i];
*     xarray[i] = temp;
*     #update the index and permutation index
*     iindex[index[i]] = iindex[i];
*     index[iindex[i]] = index[i];
*     iindex[i] = index[i] = i;
*   }
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/heap.c,v 11.3 2001/08/23 20:14:44 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/heap.c,v $
* CREATED: 1-Feb-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------
* Copyright 1990 MIT Lincoln Laboratory.
* This software is not for release and should neither be copied
* nor distributed without explicit permission of MIT Lincoln Laboratory.
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include "nnlib.h"

/*********************************************************************
* NAME: heap_init(n,xarray,index)
* DESCRIPTION:
*       Inputs a float array xarray[0...n-1] and provides output indexes
*       in an array index[0...n-1] such such that
*	for all i xarray[index[i]] >= xarray[index[2*i+1]] 
*       and xarray[index[2*i+2]].
*	The largest value in the array is at xarray[index[0]].
*************************************************************************/
void heap_init(int n, float *xarray, int *index)
{
  register int l,j,ir,temp,i;
  register float q;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  for (j=0;j<n;j++) index[j]=j;
  l= (n >> 1);
  ir=n-1;
  for (;;) {
      if (l > 0)
	q=xarray[(temp=index[--l])];
      else return;
      i=l;
      j=(l << 1)+1;
      while (j <= ir) {
	  if (j < ir && xarray[index[j]] < xarray[index[j+1]]) j++;
	  if (q < xarray[index[j]]) {
	      index[i]=index[j];
	      j += (i=j)+1;
	  } else j=ir+1;
      }
      index[i]=temp;
  }
} /* END heap_init() */

/*********************************************************************
* NAME: reheap(n,xarray,index)
* DESCRIPTION:
*	Reorganizes heap after index[0] or xarray[index[0]] is modified.
*       xarray[index[]] is in heap order except for xarray[index[0]]
*	Modifies index so that the xarray[index[]] is a heap again.
*************************************************************************/
void reheap(int n, float *xarray, int *index)
{
  register int j,temp,i;
  register float q;

    q = xarray[(temp = index[0])];
    i=0;
    j=1;
    while (j <= n-1){
	if (j < n-1 && xarray[index[j]] < xarray[index[j+1]]) j++;
	if (q < xarray[index[j]]) {
	    index[i]=index[j];
	    j += (i=j)+1;
	}
	else j=n;
    }
    index[i]=temp;
} /* END reheap() */

/*********************************************************************
* NAME: heap_sort(n,xarray,index)
* DESCRIPTION:
*        Inputs a float array xarray[0...n-1] and provides output indexes
*        in an array index[0...n-1] such such that
*        xarray[index[i]] goes from smallest to largest as i goes from 0 to n-1
*************************************************************************/
void heap_sort(int n, float *xarray, int *index)
{
  register int ir,temp;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  heap_init(n,xarray,index);
  for(ir = n - 1; ir > 0; ir--){
	temp=index[ir];
	index[ir]=index[0];
	index[0] = temp;
	reheap(ir,xarray,index);
    }
} /* END heap_sort() */

/*********************************************************************
* NAME: heap_sort2(n,xarray,iarray)
* DESCRIPTION:
*	sorts the array xarray and performs the corresponding rearrangement
*	of iarray at the same time. 
*       If more than two arrays need to be sorted in parallel, use heap_sort
*	if only xarray needs sorting, set iarray to NULL
*	after sorting, xarray[i] <= xarray[i+1]
*************************************************************************/
void heap_sort2(int n, float *xarray, int *iarray)
{
  register int ir, l, i, j, iq = -1;
  register float xq;

  if(n <= 0) return;
  if(n==1)return;

  l = (n >> 1); /* find the parent of the last node in the heap */
  ir = n-1;
  for(;;){
      /* build a heap by making heaps of the subtrees starting at the bottom,
	 adding a node and re-heaping for each subtree pair going up
	 until you have added the topmost node */
      if(l > 0){
	  l--;
	  xq = xarray[l];
	  if(iarray != NULL)iq = iarray[l];
      } else {
	  /* now that we have a heap, do the actual sorting of the array
	     by swapping the first element for the last and reheaping */
	  xq = xarray[ir];
	  if(iarray != NULL)iq = iarray[ir];
	  xarray[ir] = xarray[0];
	  if(iarray != NULL) iarray[ir] = iarray[0];
	  ir--;
	  if(ir == 0){
	      xarray[0] = xq;
	      if(iarray != NULL)iarray[0] = iq;
	      return;
	  }
      }
      /* make this subtree a heap.  The only element out of heap order
	 is the top entry in the subtree, so we just need to swap it
	 with the larger of its children until this "top" element is
	 larger than both its children or hits bottom */
      i = l;
      j = (l << 1)+1;
      while(j <= ir){
	  if(j < ir && xarray[j] < xarray[j+1]) j++;
	  if(xq < xarray[j]){
	      xarray[i] = xarray[j];
	      if(iarray != NULL)iarray[i] = iarray[j];
	      j += (i=j)+1;
	  } else j = ir+1;
      }
      xarray[i] = xq;
      if(iarray != NULL)iarray[i] = iq;
  }
} /* END heap_sort2() */

/*********************************************************************
* NAME: rev_heap_init(n,xarray,index)
* DESCRIPTION:
*       Inputs a float array xarray[0...n-1] and provides output indexes
*       in an array index[0...n-1] such such that
*	for all i xarray[index[i]] <= xarray[index[2*i+1]] 
*       and xarray[index[2*i+2]].
*	The smallest value in the array is at xarray[index[0]].
*************************************************************************/
void rev_heap_init(int n, float *xarray, int *index)
{
  register int l,j,ir,temp,i;
  register float q;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  for (j=0;j<n;j++) index[j]=j;
  l= (n >> 1);
  ir=n-1;
  for (;;) {
      if (l > 0)
	q=xarray[(temp=index[--l])];
      else return;
      i=l;
      j=(l << 1)+1;
      while (j <= ir) {
	  if (j < ir && xarray[index[j]] > xarray[index[j+1]]) j++;
	  if (q > xarray[index[j]]) {
	      index[i]=index[j];
	      j += (i=j)+1;
	  } else j=ir+1;
      }
      index[i]=temp;
  }
} /* END rev_heap_init() */

/*********************************************************************
* NAME: rev_reheap(n,xarray,index)
* DESCRIPTION:
*	Reorganizes heap after index[0] or xarray[index[0]] is modified.
*       xarray[index[]] is in heap order except for xarray[index[0]]
*	Modifies index so that the xarray[index[]] is a heap again.
*************************************************************************/
void rev_reheap(int n, float *xarray, int *index)
{
  register int j,temp,i;
  register float q;

    q = xarray[(temp = index[0])];
    i=0;
    j=1;
    while (j <= n-1){
	if (j < n-1 && xarray[index[j]] > xarray[index[j+1]]) j++;
	if (q > xarray[index[j]]) {
	    index[i]=index[j];
	    j += (i=j)+1;
	}
	else j=n;
    }
    index[i]=temp;
} /* END rev_reheap() */

/*********************************************************************
* NAME: rev_heap_sort(n,xarray,index)
* DESCRIPTION:
*        Inputs a float array xarray[0...n-1] and provides output indexes
*        in an array index[0...n-1] such such that
*        xarray[index[i]] goes from largest to smallest as i goes from 0 to n-1
*************************************************************************/
void rev_heap_sort(int n, float *xarray, int *index)
{
  register int ir,temp;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  rev_heap_init(n,xarray,index);
  for(ir = n - 1; ir > 0; ir--){
	temp=index[ir];
	index[ir]=index[0];
	index[0] = temp;
	rev_reheap(ir,xarray,index);
    }
} /* END rev_heap_sort() */


/*********************************************************************
* NAME: rev_heap_sort2(n,xarray,iarray)
* DESCRIPTION:
*	sorts the array xarray and performs the corresponding rearrangement
*	of iarray at the same time. 
*       If more than two arrays need to be sorted in parallel, use heap_sort
*	if only xarray needs sorting, set iarray to NULL
*	after sorting, xarray[i] >= xarray[i+1]
*************************************************************************/
void
rev_heap_sort2(int n, float *xarray, int *iarray)
{
  register int ir,l,i,j,iq = -1;
  register float xq;

  if(n <= 0) return;
  if(n==1)return;

  l = (n >> 1); /* find the parent of the last node in the heap */
  ir = n-1;
  for(;;){
      /* build a heap by making heaps of the subtrees starting at the bottom,
	 adding a node and re-heaping for each subtree pair going up
	 until you have added the topmost node */
      if(l > 0){
	  l--;
	  xq = xarray[l];
	  if(iarray != NULL)iq = iarray[l];
      } else {
	  /* now that we have a heap, do the actual sorting of the array
	     by swapping the first element for the last and reheaping */
	  xq = xarray[ir];
	  if(iarray != NULL)iq = iarray[ir];
	  xarray[ir] = xarray[0];
	  if(iarray != NULL) iarray[ir] = iarray[0];
	  ir--;
	  if(ir == 0){
	      xarray[0] = xq;
	      if(iarray != NULL)iarray[0] = iq;
	      return;
	  }
      }
      /* make this subtree a heap.  The only element out of heap order
	 is the top entry in the subtree, so we just need to swap it
	 with the larger of its children until this "top" element is
	 larger than both its children or hits bottom */
      i = l;
      j = (l << 1)+1;
      while(j <= ir){
	  if(j < ir && xarray[j] > xarray[j+1]) j++;
	  if(xq > xarray[j]){
	      xarray[i] = xarray[j];
	      if(iarray != NULL)iarray[i] = iarray[j];
	      j += (i=j)+1;
	  } else j = ir+1;
      }
      xarray[i] = xq;
      if(iarray != NULL)iarray[i] = iq;
  }
} /* END rev_heap_sort2() */

/*********************************************************************
* NAME: heap_iinit(n,xarray,index)
* DESCRIPTION:
*       Inputs an integer array xarray[0...n-1] and provides output indexes
*       in an array index[0...n-1] such such that
*	for all i xarray[index[i]] >= xarray[index[2*i+1]] 
*       and xarray[index[2*i+2]].
*	The largest value in the array is at xarray[index[0]].
*************************************************************************/
void heap_iinit(int n, int *xarray, int *index)
{
  register int l,j,ir,temp,i;
  register int q;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  for (j=0;j<n;j++) index[j]=j;
  l= (n >> 1);
  ir=n-1;
  for (;;) 
    {
      if (l > 0)
	q=xarray[(temp=index[--l])];
      else return;
      i=l;
      j=(l << 1)+1;
      while (j <= ir) 
	{
	  if (j < ir && xarray[index[j]] < xarray[index[j+1]]) j++;
	  if (q < xarray[index[j]]) 
	    {
	      index[i]=index[j];
	      j += (i=j)+1;
	    }
	  else j=ir+1;
	}
      index[i]=temp;
    }
} /* END heap_iinit() */

/*********************************************************************
* NAME: ireheap(n,xarray,index)
* DESCRIPTION:
*	Reorganizes heap after index[0] or xarray[index[0]] is modified.
*       xarray[index[]] is in heap order except for xarray[index[0]]
*	Modifies index so that the xarray[index[]] is a heap again.
*************************************************************************/
void ireheap(int n, int *xarray, int *index)
{
  register int j,temp,i;
  register int q;

    q = xarray[(temp = index[0])];
    i=0;
    j=1;
    while (j <= n-1){
	if (j < n-1 && xarray[index[j]] < xarray[index[j+1]]) j++;
	if (q < xarray[index[j]]) {
	    index[i]=index[j];
	    j += (i=j)+1;
	}
	else j=n;
    }
    index[i]=temp;
} /* END ireheap() */

/*********************************************************************
* NAME: heap_isort(n,xarray,index)
* DESCRIPTION:
*        Inputs an int array xarray[0...n-1] and provides output indexes
*        in an array index[0...n-1] such such that
*        xarray[index[i]] goes from smallest to largest as i goes from 0 to n-1
*************************************************************************/
void heap_isort(int n, int *xarray, int *index)
{
  register int ir,temp;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  heap_iinit(n,xarray,index);
  for(ir = n - 1; ir > 0; ir--){
	temp=index[ir];
	index[ir]=index[0];
	index[0] = temp;
	ireheap(ir,xarray,index);
    }
} /* END heap_isort() */


/*********************************************************************
* NAME: rev_heap_iinit(n,xarray,index)
* DESCRIPTION:
*       Inputs an integer array xarray[0...n-1] and provides output indexes
*       in an array index[0...n-1] such such that
*	for all i xarray[index[i]] <= xarray[index[2*i+1]] 
*       and xarray[index[2*i+2]].
*	The smallest value in the array is at xarray[index[0]].
*************************************************************************/
void rev_heap_iinit(int n, int *xarray, int *index)
{
  register int l,j,ir,temp,i;
  register int q;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  for (j=0;j<n;j++) index[j]=j;
  l= (n >> 1);
  ir=n-1;
  for (;;) 
    {
      if (l > 0)
	q=xarray[(temp=index[--l])];
      else return;
      i=l;
      j=(l << 1)+1;
      while (j <= ir) 
	{
	  if (j < ir && xarray[index[j]] > xarray[index[j+1]]) j++;
	  if (q > xarray[index[j]]) 
	    {
	      index[i]=index[j];
	      j += (i=j)+1;
	    }
	  else j=ir+1;
	}
      index[i]=temp;
    }
} /* END rev_heap_iinit() */

/*********************************************************************
* NAME: rev_ireheap(n,xarray,index)
* DESCRIPTION:
*	Reorganizes heap after index[0] or xarray[index[0]] is modified.
*       xarray[index[]] is in heap order except for xarray[index[0]]
*	Modifies index so that the xarray[index[]] is a heap again.
*************************************************************************/
void rev_ireheap(int n, int *xarray, int *index)
{
  register int j,temp,i;
  register int q;

    q = xarray[(temp = index[0])];
    i=0;
    j=1;
    while (j <= n-1){
	if (j < n-1 && xarray[index[j]] > xarray[index[j+1]]) j++;
	if (q > xarray[index[j]]) {
	    index[i]=index[j];
	    j += (i=j)+1;
	}
	else j=n;
    }
    index[i]=temp;
} /* END rev_ireheap() */

/*********************************************************************
* NAME: rev_heap_isort(n,xarray,index)
* DESCRIPTION:
*        Inputs an integer array xarray[0...n-1] and provides output indexes
*        in an array index[0...n-1] such such that
*        xarray[index[i]] goes from largest to smallest as i goes from 0 to n-1
*************************************************************************/
void rev_heap_isort(int n, int *xarray, int *index)
{
  register int ir,temp;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  rev_heap_iinit(n,xarray,index);
  for(ir = n - 1; ir > 0; ir--){
	temp=index[ir];
	index[ir]=index[0];
	index[0] = temp;
	rev_ireheap(ir,xarray,index);
    }
} /* END rev_heap_isort() */

/*********************************************************************
* NAME: heap_dinit(n,xarray,index)
* DESCRIPTION:
*       Inputs a double array xarray[0...n-1] and provides output indexes
*       in an array index[0...n-1] such such that
*	for all i xarray[index[i]] >= xarray[index[2*i+1]] 
*       and xarray[index[2*i+2]].
*	The largest value in the array is at xarray[index[0]].
*************************************************************************/
void heap_dinit(int n, double *xarray, int *index)
{
  register int l,j,ir,temp,i;
  register double q;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  for (j=0;j<n;j++) index[j]=j;
  l= (n >> 1);
  ir=n-1;
  for (;;) 
    {
      if (l > 0)
	q=xarray[(temp=index[--l])];
      else return;
      i=l;
      j=(l << 1)+1;
      while (j <= ir) 
	{
	  if (j < ir && xarray[index[j]] < xarray[index[j+1]]) j++;
	  if (q < xarray[index[j]]) 
	    {
	      index[i]=index[j];
	      j += (i=j)+1;
	    }
	  else j=ir+1;
	}
      index[i]=temp;
    }
} /* END heap_dinit() */

/*********************************************************************
* NAME: dreheap(n,xarray,index)
* DESCRIPTION:
*	Reorganizes heap after index[0] or xarray[index[0]] is modified.
*       xarray[index[]] is in heap order except for xarray[index[0]]
*	Modifies index so that the xarray[index[]] is a heap again.
*************************************************************************/
void dreheap(int n, double *xarray, int *index)
{
  register int j,temp,i;
  register double q;

    q = xarray[(temp = index[0])];
    i=0;
    j=1;
    while (j <= n-1){
	if (j < n-1 && xarray[index[j]] < xarray[index[j+1]]) j++;
	if (q < xarray[index[j]]) {
	    index[i]=index[j];
	    j += (i=j)+1;
	}
	else j=n;
    }
    index[i]=temp;
} /* END dreheap() */

/*********************************************************************
* NAME: heap_dsort(n,xarray,index)
* DESCRIPTION:
*        Inputs a double array xarray[0...n-1] and provides output indexes
*        in an array index[0...n-1] such such that
*        xarray[index[i]] goes from smallest to largest as i goes from 0 to n-1
*************************************************************************/
void heap_dsort(int n, double *xarray, int *index)
{
  register int ir,temp;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  heap_dinit(n,xarray,index);
  for(ir = n - 1; ir > 0; ir--){
	temp=index[ir];
	index[ir]=index[0];
	index[0] = temp;
	dreheap(ir,xarray,index);
    }
} /* END heap_dsort() */


/*********************************************************************
* NAME: rev_heap_dinit(n,xarray,index)
* DESCRIPTION:
*       Inputs a double array xarray[0...n-1] and provides output indexes
*       in an array index[0...n-1] such such that
*	for all i xarray[index[i]] <= xarray[index[2*i+1]] 
*       and xarray[index[2*i+2]].
*	The smallest value in the array is at xarray[index[0]].
*************************************************************************/
void rev_heap_dinit(int n, double *xarray, int *index)
{
  register int l,j,ir,temp,i;
  register double q;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  for (j=0;j<n;j++) index[j]=j;
  l= (n >> 1);
  ir=n-1;
  for (;;) 
    {
      if (l > 0)
	q=xarray[(temp=index[--l])];
      else return;
      i=l;
      j=(l << 1)+1;
      while (j <= ir) 
	{
	  if (j < ir && xarray[index[j]] > xarray[index[j+1]]) j++;
	  if (q > xarray[index[j]]) 
	    {
	      index[i]=index[j];
	      j += (i=j)+1;
	    }
	  else j=ir+1;
	}
      index[i]=temp;
    }
} /* END rev_heap_dinit() */

/*********************************************************************
* NAME: rev_dreheap(n,xarray,index)
* DESCRIPTION:
*	Reorganizes heap after index[0] or xarray[index[0]] is modified.
*       xarray[index[]] is in heap order except for xarray[index[0]]
*	Modifies index so that the xarray[index[]] is a heap again.
*************************************************************************/
void rev_dreheap(int n, double *xarray, int *index)
{
  register int j,temp,i;
  register double q;

    q = xarray[(temp = index[0])];
    i=0;
    j=1;
    while (j <= n-1){
	if (j < n-1 && xarray[index[j]] > xarray[index[j+1]]) j++;
	if (q > xarray[index[j]]) {
	    index[i]=index[j];
	    j += (i=j)+1;
	}
	else j=n;
    }
    index[i]=temp;
} /* END rev_dreheap() */

/*********************************************************************
* NAME: rev_heap_dsort(n,xarray,index)
* DESCRIPTION:
*        Inputs a float array xarray[0...n-1] and provides output indexes
*        in an array index[0...n-1] such such that
*        xarray[index[i]] goes from largest to smallest as i goes from 0 to n-1
*************************************************************************/
void rev_heap_dsort(int n, double *xarray, int *index)
{
  register int ir,temp;

  if(n <= 0) return;
  if(n==1) {index[0]=0; return;}

  rev_heap_dinit(n,xarray,index);
  for(ir = n - 1; ir > 0; ir--){
	temp=index[ir];
	index[ir]=index[0];
	index[0] = temp;
	rev_dreheap(ir,xarray,index);
    }
} /* END rev_heap_dsort() */

