/***--------------------------------------------------------------------------
 * NAME:        map_data.c
 * PURPOSE: package for giving training or cross validation patterns to
 *      mappter on request.  Takes into account random
 *      presentation order.
 *
 * DESCRIPTION:
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/map_data.c,v 11.3 2001/08/23 20:14:44 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/map_data.c,v $
 * CREATED: 15-Aug-93 lck           MIT LINCOLN LABORATORY
 *---------------------------------------------------------------------------
 * Copyright 1993 MIT Lincoln Laboratory.
 * This software is not for release and should neither be copied
 * nor distributed without explicit permission of MIT Lincoln Laboratory.
 --------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include "nnlib.h"

/* GLOBAL */
#define TRUE 1
#define FALSE 0

typedef struct{
    /****************** data ********************/
    int noutputs, npatterns, ninputs, input_num;
    float **input_matrix, **output_matrix;
    /****** state variables during reading ******/
    int n_read, train_flag;
    /************ random variables **************/
    int *random_index, random_flag;
    /********** cross valid variables ***********/
    int cross_valid_flag, *cross_index;
    int fold, nfolds, *n_in_train_fold, *n_in_test_fold;
    char **train_masks, **test_masks, *cv_mask;
} MAP_PATTERNS;

MAP_PATTERNS **mp_list;
static int max_models = 0;

int
map_data_read(char *data_path, char *data_fname, int npatterns,
          int nraw, int ninputs, int noutputs, int *features,
          int normalization, int verbosity)
{
    int new_model;
    char *data_full_name;
    MAP_PATTERNS *mpp;
    FILE *data_file;

    /* allocate structure */
    new_model = model_get_number((void ***)&mp_list, &max_models,
                 sizeof(MAP_PATTERNS));
    /* store what we know so far */
    mpp = mp_list[new_model];
    mpp->noutputs = noutputs;
    mpp->ninputs = ninputs;
    mpp->train_flag = TRUE;
    mpp->n_read = 0;
    mpp->random_flag = mpp->cross_valid_flag = FALSE;

    /* allocate space for input and output matricies */
    mpp->input_matrix = allocate_matrix(npatterns, ninputs);
    mpp->output_matrix = allocate_matrix(npatterns, noutputs);
                
    /* open file, load data, close file */
    data_full_name = path_build(data_path, data_fname);
    data_file = nn_fopen(data_full_name, "r", "for training or test data");
    if(verbosity > 1) printf("\nReading %s\n", data_full_name);

    npatterns = load_map_data(data_file, npatterns, nraw, noutputs, ninputs,
                mpp->output_matrix, mpp->input_matrix,
                normalization, data_full_name, features);
    nn_fclose(data_file, data_full_name, "for training or test data");

    /* store the real number of patterns */
    mpp->npatterns = npatterns;
    mpp->input_num = 0;

    /* return a model number */
    return new_model;
}

int
map_data_get_npatterns(int data_model)
{
    MAP_PATTERNS *mpp;
    mpp = mp_list[data_model];

    if(mpp->cross_valid_flag) {
      if(mpp->train_flag) return mpp->n_in_train_fold[mpp->fold];
      else return mpp->n_in_test_fold[mpp->fold];
    } else return mpp->npatterns;
}

void
map_data_random_init(int data_model)
{
    MAP_PATTERNS *mpp;
    int i;

    mpp = mp_list[data_model];
    mpp->random_flag = TRUE;

    /* allocate space for randomized list */
    mpp->random_index = allocate_ivector(mpp->npatterns);
    for(i = 0; i < mpp->npatterns; i++) mpp->random_index[i] = i;
}

void
map_data_cv_init(int data_model, int cv_nsplits, int use_cvsplits,
         int *cv_splits, int cv_nfolds,
         char **cv_train_mask, char **cv_test_mask)
{
    MAP_PATTERNS *mpp;
    int i, f;

    mpp = mp_list[data_model];

    mpp->cross_valid_flag = TRUE;
    mpp->nfolds = cv_nfolds;

    /* set up and store cross index */
    mpp->cross_index =
      build_map_cross_index(mpp->noutputs, mpp->npatterns, cv_nsplits, 
                use_cvsplits, cv_splits);

    /* count patterns per fold */
    mpp->n_in_train_fold = allocate_ivector(cv_nfolds);
    mpp->n_in_test_fold = allocate_ivector(cv_nfolds);
    for(i = 0; i < mpp->npatterns; i++)
      for(f = 0; f < cv_nfolds; f++){
      if(cv_train_mask[f][mpp->cross_index[i]] == '1')
        mpp->n_in_train_fold[f]++;
      if(cv_test_mask[f][mpp->cross_index[i]] == '1')
        mpp->n_in_test_fold[f]++;
      }
    mpp->train_masks = cv_train_mask;
    mpp->test_masks = cv_test_mask;
#ifdef DEBUG
    printf("cross validation fold assignments\n");
    printf("index split");
    for(i = 0; i < mpp->npatterns; i++){
    printf("%s\t%d\t%d\n", i, mpp->cross_index[i]);
    }
#endif /* DEBUG */
}

int
map_data_start_cv_fold(int data_model, int fold, int is_train)
{
    MAP_PATTERNS *mpp;

    mpp = mp_list[data_model];

    mpp->n_read = 0;
/*
    if(fold >= mpp->nfolds) return 0;
*/
    mpp->fold = fold;

    /* set up current mask */
    if(is_train)
      mpp->cv_mask = mpp->train_masks[fold];
    else
      mpp->cv_mask = mpp->test_masks[fold];
    mpp->train_flag = is_train;
#ifdef DEBUG
    printf("cross validation train mask: %s\n", mpp->cv_mask);
#endif /* DEBUG */

    /* return the number of patterns this epoch */
    if(is_train)
      return mpp->n_in_train_fold[fold];
    else
      return mpp->n_in_test_fold[fold];
}

void
map_data_epoch_init(int data_model)
{
    MAP_PATTERNS *mpp;
    mpp = mp_list[data_model];
    mpp->n_read = 0;
    if(mpp->random_flag){
    randomize_index_list(mpp->random_index, mpp->npatterns);
    }
}

float *
map_data_get_next(int data_model, float **outputs)
{
    MAP_PATTERNS *mpp;
    int t_index;
    float *inputs;
    mpp = mp_list[data_model];
    
    /* if this is cross valid, loop until we pick one, else do once */
    while(1){
    /* if we have read all of the data, return NULL; */
    if(mpp->n_read >= mpp->npatterns) return NULL;

    /* select a pattern, paying attention to random and cv */
    if(mpp->random_flag && mpp->train_flag){
        t_index = mpp->random_index[mpp->n_read];
    } else t_index = mpp->n_read;
    mpp->n_read++;
    if(mpp->cross_valid_flag)
      if(mpp->cv_mask[mpp->cross_index[t_index]] != '1') continue;

    /* return pointers to the input and output vectors */
    inputs = &(mpp->input_matrix[t_index][0]);
    *outputs = &(mpp->output_matrix[t_index][0]);
    mpp->input_num = t_index;
    return inputs;
    }
}

void
map_data_exit(int data_model)
{
    MAP_PATTERNS *mpp;
    mpp = mp_list[data_model];
    deallocate_matrix(mpp->input_matrix);
    deallocate_matrix(mpp->output_matrix);
    if(mpp->cross_valid_flag){
    deallocate_ivector(mpp->cross_index);
    deallocate_ivector(mpp->n_in_train_fold);
    deallocate_ivector(mpp->n_in_test_fold);
    }
    if(mpp->random_flag) deallocate_ivector(mpp->random_index);
    nn_free((char *)mpp, "for data model");
    mp_list[data_model] = NULL;
}

int
map_data_get_input_num(int data_model)
{
    return(mp_list[data_model]->input_num);
}

/* don't use this, it is for debugging only */
void
map_data_set_n_read_zero(int dm)
{
    mp_list[dm]->n_read = 0;
}
