/***--------------------------------------------------------------------------
* NAME:  matrix.c
* DESCRIPTION: 
*    General purpose functions
*	invertM		invert the M matrix
*	invertM_ev	invert the M matrix saving eigenvalues
*	matrix_vector_multiply	Matrix multiply, matrix * vector	
*	square_matrix_invert
*	square_matrix_multiply
*	square_matrix_transpose
*       square_matrix_normalize_and_sort      Normalizes and sorts eigenvectors
*	square_singular_value_decomposition 	Calls svdcmp function
*	transposed_matrix_vector_multiply	result = (matrix)' * vector
*	
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include "nnlib.h"
/************************************************************************
* NAME: invertM(nhidden,M_matrix,max_ratio)
* DESCRIPTION:
*	Takes as input the number of nodes, pointer to the M matrix, and
*       the maximum eigenvalue ratio.
*	Generates the inverse of correlation matrix,
*	returning a pointer to the new matrix.
*       Includes automatic screening of the nodes.  number of nodes 
*       can be changed by this routine.
*       Eigenvalues (w) from svdcmp routine < min_eigenvalue are "zeroed."
*	Note that there is some matrix, array, pointer massaging
*	done because of the conventions used in the svdcmp() function.
*************************************************************************/
float **
invertM(int nhidden, float **M_matrix, float max_ratio)
{
  int i,j,k;
  float *w,**u,**v;
  float min_eigenvalue;

  /* Allocate memory for the w, and v matrices */
  w = allocate_vector(nhidden+1);
  v = allocate_matrix(nhidden+1,nhidden+1);
  u = allocate_matrix(nhidden+1,nhidden+1);

  /* Copy M_matrix (normal indexing) to u (1 to N indexing)  */
  for (i=1;i<=nhidden;i++)
      for (j=1;j<=nhidden;j++)
	  u[i][j]=M_matrix[i-1][j-1];

  /* perform decomposition */
  svdcmp(u,nhidden,nhidden,w,v);
  
  /* Calculate the minimum allowable eigenvalue  */
  min_eigenvalue = w[1]/max_ratio;

  /* Printout the w matrix */
#if DEBUG
      printf("\neigenvalues: (cutoff = %g)\n",min_eigenvalue);
      for (i=1,j=1;i<=nhidden;i++)
	{
	  printf(" %f ",w[i]);
	  if (j==5) {printf("\n"); j=0;}
	  j++;
	}
      printf("\n");
#endif

  /* generate the inverse */
  for (i=1;i<=nhidden;i++)
      for (j=1;j<=nhidden;j++)
	{
	  M_matrix[i-1][j-1]=0.0;
	  for (k=1;k<=nhidden;k++)
	    {
	      if (w[k]>=min_eigenvalue)
		  M_matrix[i-1][j-1] += (v[i][k]/w[k])*u[j][k];
	    }
	} 

  /* free the memory used for the matrices */
  deallocate_vector(w);
  deallocate_matrix(u);
  deallocate_matrix(v);
  return(M_matrix);
} 

/************************************************************************
* NAME: invertM_ev (nhidden,M_matrix,max_ratio,eigenvalues)
* DESCRIPTION:
*	Takes as input the number of nodes, pointer to the M matrix, and
*       the maximum eigenvalue ratio.
*	Generates the inverse of correlation matrix,
*	returning a pointer to the new matrix.
*       Includes automatic screening of the nodes.  number of nodes 
*       can be changed by this routine.
*       Eigenvalues (w) from svdcmp routine < min_eigenvalue are "zeroed."
*	the vector of eigenvalues is returned in the eigenvalues parameter 
*	    the eigenvalues vector is converted from numerical recipes 
*	    format of 1:nhidden to 0:nhidden-1 in returned vector
*	Note that there is some matrix, array, pointer massaging
*	done because of the conventions used in the svdcmp() function.
*************************************************************************/
float **
invertM_ev (int nhidden, float **M_matrix, float max_ratio,
	    float **eigenvalues)
{
    int i,j,k;
    float *w,**u,**v;
    float min_eigenvalue;

    /* Allocate memory for the w, and v matrices */
    w = allocate_vector(nhidden+1);
    v = allocate_matrix(nhidden+1, nhidden+1);
    u = allocate_matrix(nhidden+1, nhidden+1);

    /* Copy M_matrix (normal indexing) to u (1 to N indexing)  */
    for (i=1;i<=nhidden;i++)
        for (j=1;j<=nhidden;j++)
	    u[i][j]=M_matrix[i-1][j-1];

    /* perform decomposition */
    svdcmp(u,nhidden,nhidden,w,v);
  
    /* Calculate the minimum allowable eigenvalue  */
    min_eigenvalue = w[1]/max_ratio;

    /* Printout the w matrix */
#if DEBUG
        printf("eigenvalues: (cutoff = %g)",min_eigenvalue);
        for (i=1,j=1;i<=nhidden;i++) {
	    printf(" %f ",w[i]);
	    if (j==5) {printf("\n"); j=0;}
	    j++;
        }
        printf("\n");
        fflush(stdout);
#endif
    			/* return actual eigenvalues without the cutoff */
    
    for (i=1; i<=nhidden; i++) (*eigenvalues)[i-1] = w[i];

    /* generate the inverse */
    for (i=1;i<=nhidden;i++)
        for (j=1;j<=nhidden;j++) {
	    M_matrix[i-1][j-1]=0.0;
	    for (k=1;k<=nhidden;k++) {
	        if (w[k]>=min_eigenvalue)
		    M_matrix[i-1][j-1] += (v[i][k]/w[k])*u[j][k];
	    }
	} 

    /* free the memory used for the matrices */
    deallocate_vector(w);
    deallocate_matrix(u);
    deallocate_matrix(v);
    return(M_matrix);
} 

/*****************************************************************************
* NAME: matrix_vector_multiply()
* DESCRIPTION:
*       Matrix multiply:   result = matrix * vector, floating point
*	matrix is a float matrix, rows by columns
*	vector is a float matrix, columns by 1
*	result is rows by 1
*
*****************************************************************************/
void
matrix_vector_multiply (float **matrix, float *vector, float *result,
			int rows, register int columns)
{
    register int k,i;
    register float *vp, *mp, tmp;

    for (i=0; i<rows; i++) {
	vp = vector;
	mp = &matrix[i][0];
	tmp = 0.;
	for (k=0; k<columns; k++)
	  tmp += *mp++ * *vp++;
	*result++ = tmp;
    }
}
/*****************************************************************************
* NAME: square_matrix_invert ()
* DESCRIPTION:
*       Invert matrix:   output_matrix = input_matrix(-1)
*	(one matrix can be used for both variables)
*
*****************************************************************************/
void
square_matrix_invert (int ninputs, float **input_matrix,
		      float **output_matrix, float *eigenvalues,
		      float max_ratio)
{
    register int i, j, k;
    float **eigenvectors, min_eigenvalue;
    
    eigenvectors = allocate_matrix(ninputs, ninputs);

    square_singular_value_decomposition (ninputs, input_matrix, eigenvectors, 
					 eigenvalues);

    /* Calculate the minimum allowable eigenvalue  */
    min_eigenvalue = eigenvalues[0] / max_ratio;

    /* generate the inverse */
    /* inverse = eigenvectors * 1/eigenvalues * eigenvectors(T) */
    for (i=0;i<ninputs;i++)
        for (j=0;j<ninputs;j++) {
	    output_matrix[i][j]=0.0;
	    for (k=0;k<ninputs;k++)
	        if (eigenvalues[k]>=min_eigenvalue)
		    output_matrix[i][j] += 
		        (eigenvectors[k][i]/eigenvalues[k])*eigenvectors[k][j];
	}
	
    deallocate_matrix(eigenvectors);
}

/*****************************************************************************
* NAME: square_matrix_multiply ()
* DESCRIPTION:
*       Matrix multiply:   output_matrix = input_matrix_1 * input_matrix_2
*	(the output matrix must be different from the input matrices)
*
*****************************************************************************/
void
square_matrix_multiply (int ninputs, float **input_matrix_1,
			float **input_matrix_2, float **output_matrix)
{
    register int i, j, k;

    for (i=0;i<ninputs;i++)
        for (j=0;j<ninputs;j++) {
	    output_matrix[i][j]=0.0;
	    for (k=0;k<ninputs;k++)
		output_matrix[i][j] +=input_matrix_1[i][k]*input_matrix_2[k][j];
	}
}


/*****************************************************************************
* NAME: square_matrix_transpose ()
* DESCRIPTION:
*       Transpose matrix:   output_matrix =(input_matrix)'
*	(one matrix can be used for both variables)
*
*****************************************************************************/
void
square_matrix_transpose (int ninputs, float **input_matrix,
			 float **output_matrix)
{
    register int i,k;
    register float tmp;

    for (k=0; k<ninputs-1; k++) {
	for (i=k+1; i<ninputs; i++) {
	    tmp = output_matrix[k][i];
	    output_matrix[k][i] = input_matrix[i][k];
	    output_matrix[i][k] = tmp;
	}
    }
}

/*****************************************************************************
* NAME: square_normalize_and_sort (ninputs, eigenvectors, eigenvalues)
*
* DESCRIPTION:
*         1) Normalizes (sets magnitude = 1) all eigenvectors (ROWS of matrix)
*         2) Does full sort on eigenvalues and corresponding eigenvectors
*
*****************************************************************************/
void
square_normalize_and_sort (int ninputs, float **eigenvectors,
			   float *eigenvalues)
{
    float magnitude, tmp;
    int i, j, k;

    for (i=0; i<ninputs; i++) {
        magnitude = 0;
        for(j=0;j<ninputs;j++)
	    magnitude += eigenvectors[i][j] * eigenvectors[i][j];
        magnitude = sqrt(magnitude);
        for(j=0;j<ninputs;j++)
	    eigenvectors[i][j] /= magnitude;
    }

    for (i=0; i<ninputs - 1; i++)
        for (j=0; j<ninputs - 1; j++)
            if (eigenvalues[j+1] > eigenvalues[j]) {
	        tmp = eigenvalues[j];
	        eigenvalues[j] = eigenvalues[j+1];
	        eigenvalues[j+1] = tmp;
	        for (k=0; k<ninputs; k++) {
	            tmp=eigenvectors[j][k];
	            eigenvectors[j][k] = eigenvectors[j+1][k];
	            eigenvectors[j+1][k] = tmp;
	        }
	    }
}

/************************************************************************
* NAME: square_singular_value_decomposition (input_matrix, output_matrix,
*		ninputs, eigenvalues)
*
* DESCRIPTION:
*		uses Singular Value Decomposition function svdcmp
*		eigenvalues	vector of eigenvalues
*
*************************************************************************/
void
square_singular_value_decomposition (int ninputs, float **input_matrix, 
				     float **eigenvectors, float *eigenvalues)
{
    int i,j;
    float *w,**u,**v;
    float sum=0;

    /* Allocate memory for the w, and v matrices */
    w = allocate_vector(ninputs+1);
    v = allocate_matrix(ninputs+1, ninputs+1);
    u = allocate_matrix(ninputs+1, ninputs+1);

    /* Copy M_matrix (normal indexing) to u (1 to N indexing)  */
    for (i=1;i<=ninputs;i++)
      for (j=1;j<=ninputs;j++)
	u[i][j]=input_matrix[i-1][j-1];

    /* perform decomposition */
    svdcmp(u,ninputs,ninputs,w,v);
  
	/* copy u transpose into eigenvector array (rows are eigenvectors) */
    for (i=1; i<=ninputs; i++) {
        eigenvalues[i-1] = w[i];
        for (j=1;j<=ninputs;j++) eigenvectors[j-1][i-1]=u[i][j];
    }

#if DEBUG
    printf("square_singular_value_decomposition: %f %f %d\n", 
	   eigenvalues[0], eigenvalues[1], ninputs);
#endif

       /* if sum of diagonal of eigenvector matrix is negative, reverse sign */
    for (i=0; i<ninputs; i++) {
        sum += eigenvectors[i][i];
    }
    if (sum < -.00001) {
#if DEBUG
        printf("  Changing sign %d\n",ninputs);
#endif
        for (i=0; i<ninputs; i++) {
            for (j=0; j<ninputs; j++) {
                eigenvectors[i][j] = -eigenvectors[i][j];
            }
        }
    }

    square_normalize_and_sort(ninputs, eigenvectors, eigenvalues);
#if DEBUG
    for (i=0; i<ninputs; i++) {
        printf("eigenvalue %d = %f\n",i,eigenvalues[i]);
    }
#endif
    /* free the memory used for the matrices */
    deallocate_vector(w);
    deallocate_matrix(u);
    deallocate_matrix(v);
} 

/*****************************************************************************
* NAME: transposed_matrix_vector_multiply ()
* DESCRIPTION:
*       Transposed matrix multiply:   result =(matrix)' * vector
*	matrix is a matrix, columns by rows
*	vector is a matrix, columns by 1
*	result is the result, rows by 1
*
*****************************************************************************/
void
transposed_matrix_vector_multiply (float **matrix, float *vector,
				   float *result, int rows, int columns)
{
    register int i,k;
    register float *vp, *mp;
    register float tmp_v;

    vp = result;
    for (i=0; i<rows; i++) *vp++ = 0.;

    for (k=0; k<columns; k++) {
	vp = result;
	mp = &matrix[k][0];
	tmp_v = *vector++;
	for (i=0; i<rows; i++)
	    *vp++ += *mp++ * tmp_v;
    }
}
