/***--------------------------------------------------------------------------
* NAME:  mlp_subs.c
* PURPOSE:  Functions called by mlp.c and by other functions in mlp_pack.c
* USAGE
*
* SEE ALSO: mlp.c
* GLOBALS:
* DESCRIPTION: 
*
*    General purpose functions
*       dsigm		Calculates derivatives used for updating weights
*	invertM		invert the M matrix
*	invertM_ev	invert the M matrix saving eigenvalues
*	matrix_vector_multiply	Matrix multiply, matrix * vector	
*	max_pick	Returns integer corresp. to maximum value of vec
*	pick_max	calls max_pick (old function)
*	randomize_index_list	Rearranges index list randomly
*	sigm		Calculates function used for upadating weights
*	square_matrix_invert
*	square_matrix_multiply
*	square_matrix_transpose
*       square_matrix_normalize_and_sort      Normalizes and sorts eigenvectors
*	square_singular_value_decomposition 	Calls svdcmp function
*	transposed_matrix_vector_multiply	result = (matrix)' * vector
*	
------------------------------------------------------------------------***/
static char copyright[]=
"**************************************************************************\n\
                Copyright 1990 MIT Lincoln Laboratory.\n\
This software is not for release and should neither be copied\n\
nor distributed without explicit permission of MIT Lincoln Laboratory\n\
**************************************************************************\n";

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/mlp_subs.c,v 6.2 2001/08/23 20:14:44 craig Exp $";



/* Load include files  */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <stdio.h>
#include <fcntl.h>
#include <strings.h>
#include "nnlib.h"
#define TRUE 1
#define FALSE 0
extern int debug;

/*****************************************************************************
* NAME: dsigm(val,type,parm)
* DESCRIPTION:
*        Calculates derivatives used for updating weights using output value
*
*        Non-linear functions for nodes
     
     type = 0:   sigmoid, (Rummelhart)  asymptotes 0 and 1
     type = 1:   symmetric sigmoid      asymptotes -1 and 1
     type = 2:   not used
     type = 3:   step function for Perceptron Convergence Procedure
     type = 4:   linear, for input layer   slope = 1
*****************************************************************************/

float dsigm (val,type,parm)   /* the derivative of the function */
     float val;/*Weighted sum of inputs already passed through sigmoid func.*/
     int type; /* Type of above function used in sigmoid */
     float parm; /* Parameter used in sigmoid (cases 0 and 1) */
{
  switch(type) {
    case 0: 
      return((float) parm*val*(1.-val));
    case 1:
      return((float) parm*(1.+val)*(1.-val)/2.);
    case 3:
      return((float) 1.); /* this makes pcp work in mlp framework */
    case 4:
      return((float) 1.);
    default:
      fatal_error("dsigm:  function of type %d not allowed\n",type);
    }
}

/************************************************************************
* NAME: invertM(nhidden,M_matrix,max_ratio)
* DESCRIPTION:
*	Takes as input the number of nodes, pointer to the M matrix, and
*       the maximum eigenvalue ratio.
*	Generates the inverse of correlation matrix,
*	returning a pointer to the new matrix.
*       Includes automatic screening of the nodes.  number of nodes 
*       can be changed by this routine.
*       Eigenvalues (w) from svdcmp routine < min_eigenvalue are "zeroed."
*	Note that there is some matrix, array, pointer massaging
*	done because of the conventions used in the svdcmp() function.
*************************************************************************/
float **invertM(nhidden,M_matrix,max_ratio)
     int nhidden;
     float **M_matrix,max_ratio;
{
  int i,j,k;
  float *w,**u,**v;
  float min_eigenvalue;
  void svdcmp();
  /* these subroutines are kept only for use with numerical recipes routines */ 
  /* Allocate memory for the w, and v matrices */
  w = allocate_vector(nhidden+1);
  v = allocate_matrix(nhidden+1,nhidden+1);
  u = allocate_matrix(nhidden+1,nhidden+1);

  /* Copy M_matrix (normal indexing) to u (1 to N indexing)  */
  for (i=1;i<=nhidden;i++)
      for (j=1;j<=nhidden;j++)
	  u[i][j]=M_matrix[i-1][j-1];

  /* perform decomposition */
  svdcmp(u,nhidden,nhidden,w,v);
  
  /* Calculate the minimum allowable eigenvalue  */
  min_eigenvalue = w[1]/max_ratio;

  /* Printout the w matrix */
  if (debug > 0)
    {
      printf("\neigenvalues: (cutoff = %g)\n",min_eigenvalue);
      for (i=1,j=1;i<=nhidden;i++)
	{
	  printf(" %f ",w[i]);
	  if (j==5) {printf("\n"); j=0;}
	  j++;
	}
      printf("\n");
    }

  /* generate the inverse */
  for (i=1;i<=nhidden;i++)
      for (j=1;j<=nhidden;j++)
	{
	  M_matrix[i-1][j-1]=0.0;
	  for (k=1;k<=nhidden;k++)
	    {
	      if (w[k]>=min_eigenvalue)
		  M_matrix[i-1][j-1] += (v[i][k]/w[k])*u[j][k];
	    }
	} 

  /* free the memory used for the matrices */
  deallocate_vector(w);
  deallocate_matrix(u);
  deallocate_matrix(v);
  return(M_matrix);
} 

/************************************************************************
* NAME: invertM_ev (nhidden,M_matrix,max_ratio,eigenvalues)
* DESCRIPTION:
*	Takes as input the number of nodes, pointer to the M matrix, and
*       the maximum eigenvalue ratio.
*	Generates the inverse of correlation matrix,
*	returning a pointer to the new matrix.
*       Includes automatic screening of the nodes.  number of nodes 
*       can be changed by this routine.
*       Eigenvalues (w) from svdcmp routine < min_eigenvalue are "zeroed."
*	the vector of eigenvalues is returned in the eigenvalues parameter 
*	    the eigenvalues vector is converted from numerical recipes 
*	    format of 1:nhidden to 0:nhidden-1 in returned vector
*	Note that there is some matrix, array, pointer massaging
*	done because of the conventions used in the svdcmp() function.
*************************************************************************/
float **invertM_ev (nhidden,M_matrix,max_ratio,eigenvalues)
    int nhidden;
    float **M_matrix,max_ratio,**eigenvalues;
{
    int i,j,k;
    float *w,**u,**v;
    float min_eigenvalue;
    void svdcmp();
    /* these subroutines are kept only for use with numerical recipes routines*/
    /* Allocate memory for the w, and v matrices */
    w = allocate_vector(nhidden+1);
    v = allocate_matrix(nhidden+1, nhidden+1);
    u = allocate_matrix(nhidden+1, nhidden+1);

    /* Copy M_matrix (normal indexing) to u (1 to N indexing)  */
    for (i=1;i<=nhidden;i++)
        for (j=1;j<=nhidden;j++)
	    u[i][j]=M_matrix[i-1][j-1];

    /* perform decomposition */
    svdcmp(u,nhidden,nhidden,w,v);
  
    /* Calculate the minimum allowable eigenvalue  */
    min_eigenvalue = w[1]/max_ratio;

    /* Printout the w matrix */
    if (debug > 0) {
        printf("eigenvalues: (cutoff = %g)",min_eigenvalue);
        for (i=1,j=1;i<=nhidden;i++) {
	    printf(" %f ",w[i]);
	    if (j==5) {printf("\n"); j=0;}
	    j++;
        }
        printf("\n");
        fflush(stdout);
    }
    			/* return actual eigenvalues without the cutoff */
    
    for (i=1; i<=nhidden; i++) (*eigenvalues)[i-1] = w[i];

    /* generate the inverse */
    for (i=1;i<=nhidden;i++)
        for (j=1;j<=nhidden;j++) {
	    M_matrix[i-1][j-1]=0.0;
	    for (k=1;k<=nhidden;k++) {
	        if (w[k]>=min_eigenvalue)
		    M_matrix[i-1][j-1] += (v[i][k]/w[k])*u[j][k];
	    }
	} 

    /* free the memory used for the matrices */
    deallocate_vector(w);
    deallocate_matrix(u);
    deallocate_matrix(v);
    return(M_matrix);
} 

/*****************************************************************************
* NAME: matrix_vector_multiply()
* DESCRIPTION:
*       Matrix multiply:   result = matrix * vector, floating point
*	matrix is a float matrix, rows by columns
*	vector is a float matrix, columns by 1
*	result is rows by 1
*
*****************************************************************************/
void
matrix_vector_multiply (matrix, vector, result, rows, columns)
    register int columns;
    int rows;
    float **matrix;
    float *result, *vector;
{
    register int k,i;
    register float *vp, *mp, tmp;

    for (i=0; i<rows; i++) {
	vp = vector;
	mp = &matrix[i][0];
	tmp = 0.;
	for (k=0; k<columns; k++)
	  tmp += *mp++ * *vp++;
	*result++ = tmp;
    }
}

/*****************************************************************************
* NAME: max_pick()
* DESCRIPTION:
*       Returns integer corresponding to maximum value of vec
*
*       
*****************************************************************************/

int
max_pick (vec, size)
    int size;
    float *vec;
{
    float max_value, tmp;
    register int max_class=0, i;

    max_value = *vec++;
    if (size>1) for (i=1; i<size; i++) {
        tmp = *vec++;
	if (tmp > max_value) {
	    max_class=i;
	    max_value=tmp;
	}
    }
    return max_class;
}

/*********************************************************************
* NAME: pick_max(size,vec)
* DESCRIPTION:
*	Takes as input the size of vec
*       and a pointer to the output vector vec.
*	Returns the index (class) of the largest component of vec.
*************************************************************************/
int pick_max(size,vec)
     int size;
     float *vec;
{
    return max_pick (vec, size);
} 

/*****************************************************************************
* NAME: randomize_index_list()
* DESCRIPTION:
*       Rearranges index list randomly by swapping each entry with a randomly
*		picked other entry
*
*	index_list = int array list_length long
*	list_length = length of index_list
*       
*****************************************************************************/

void
randomize_index_list (index_list, list_length)
	int *index_list, list_length;
{
	int i, tmp, t_index;
	int random_int(); 	/* Uniform random number from n1 to n2 */

	for (i=0; i<list_length; i++) {
		tmp = index_list[i];
		t_index = random_int(0,list_length-1);
		if (t_index >= list_length) {
			warning ("%s%s %d %d",
			     "In random_index, random_int returned index ",
			     "larger than array_size",t_index,list_length-1);
			t_index = list_length-1;
		}
		index_list[i] = index_list[t_index];
		index_list[t_index] = tmp;
	} 
}

/*****************************************************************************
* NAME: sigm(val,type,parm)
* DESCRIPTION:
*        Calculates function used for upadating weights
*
*        Non-linear functions for nodes
     
     type = 0:   sigmoid, (Rummelhart)  asymptotes 0 and 1
     type = 1:   symmetric sigmoid      asymptotes -1 and 1
     type = 2:   not used
     type = 3:   step function
     type = 4:   linear, for input layer   slope = 1
*****************************************************************************/

float sigm(val, type, parm)
     float val; /* val is the weighted sum of inputs */
     int type; /* Type is one of the above non-linear functions */
     float parm; /* parm is the scale used in the sigmoid (cases 0 and 1) */
{
  switch(type) 
    {
    case 0:
      return((float) 1./(1.+exp(-parm*val)));
    case 1:
      return((float)  2./(1.+exp(-parm*val))-1.);
    case 3:
      return( (val < 0.) ? 0. : 1.);
    case 4:
      return(val);
    default:
      fatal_error("sigm:  function type %d not allowed\n",type);
    }
}

/*****************************************************************************
* NAME: square_matrix_invert ()
* DESCRIPTION:
*       Invert matrix:   output_matrix = input_matrix(-1)
*	(one matrix can be used for both variables)
*
*****************************************************************************/
void
square_matrix_invert (ninputs, input_matrix, output_matrix, eigenvalues,
    max_ratio)
    
    int ninputs;
    float **input_matrix, **output_matrix, *eigenvalues, max_ratio;
{
    register int i, j, k;
    float **eigenvectors, min_eigenvalue;
    void square_singular_value_decomposition();
    
    eigenvectors = allocate_matrix(ninputs, ninputs);

    square_singular_value_decomposition (ninputs, input_matrix, eigenvectors, 
        eigenvalues);

    			    /* Calculate the minimum allowable eigenvalue  */
    min_eigenvalue = eigenvalues[0] / max_ratio;

    					/* generate the inverse */
    		  /* inverse = eigenvectors * 1/eigenvalues * eigenvectors(T) */
    for (i=0;i<ninputs;i++)
        for (j=0;j<ninputs;j++) {
	    output_matrix[i][j]=0.0;
	    for (k=0;k<ninputs;k++)
	        if (eigenvalues[k]>=min_eigenvalue)
		    output_matrix[i][j] += 
		        (eigenvectors[k][i]/eigenvalues[k])*eigenvectors[k][j];
	}
	
    deallocate_matrix(eigenvectors);
}

/*****************************************************************************
* NAME: square_matrix_multiply ()
* DESCRIPTION:
*       Matrix multiply:   output_matrix = input_matrix_1 * input_matrix_2
*	(the output matrix must be different from the input matrices)
*
*****************************************************************************/
void
square_matrix_multiply (ninputs, input_matrix_1, input_matrix_2, output_matrix)
    int ninputs;
    float **input_matrix_1, **input_matrix_2, **output_matrix;
{
    register int i, j, k;

    for (i=0;i<ninputs;i++)
        for (j=0;j<ninputs;j++) {
	    output_matrix[i][j]=0.0;
	    for (k=0;k<ninputs;k++)
		output_matrix[i][j] +=input_matrix_1[i][k]*input_matrix_2[k][j];
	}
}


/*****************************************************************************
* NAME: square_matrix_transpose ()
* DESCRIPTION:
*       Transpose matrix:   output_matrix =(input_matrix)'
*	(one matrix can be used for both variables)
*
*****************************************************************************/
void
square_matrix_transpose (ninputs, input_matrix, output_matrix)
    int ninputs;
    float **input_matrix, **output_matrix;
{
    register int i,k;
    register float tmp;

    for (k=0; k<ninputs-1; k++) {
	for (i=k+1; i<ninputs; i++) {
	    tmp = output_matrix[k][i];
	    output_matrix[k][i] = input_matrix[i][k];
	    output_matrix[i][k] = tmp;
	}
    }
}

/*****************************************************************************
* NAME: square_normalize_and_sort (ninputs, eigenvectors, eigenvalues)
*
* DESCRIPTION:
*         1) Normalizes (sets magnitude = 1) all eigenvectors (ROWS of matrix)
*         2) Does full sort on eigenvalues and corresponding eigenvectors
*
*****************************************************************************/
void
square_normalize_and_sort (ninputs, eigenvectors, eigenvalues)
    int ninputs;
    float **eigenvectors, *eigenvalues;
{
    float magnitude, tmp;
    int i, j, k;

    for (i=0; i<ninputs; i++) {
        magnitude = 0;
        for(j=0;j<ninputs;j++)
	    magnitude += eigenvectors[i][j] * eigenvectors[i][j];
        magnitude = sqrt(magnitude);
        for(j=0;j<ninputs;j++)
	    eigenvectors[i][j] /= magnitude;
    }

    for (i=0; i<ninputs - 1; i++)
        for (j=0; j<ninputs - 1; j++)
            if (eigenvalues[j+1] > eigenvalues[j]) {
	        tmp = eigenvalues[j];
	        eigenvalues[j] = eigenvalues[j+1];
	        eigenvalues[j+1] = tmp;
	        for (k=0; k<ninputs; k++) {
	            tmp=eigenvectors[j][k];
	            eigenvectors[j][k] = eigenvectors[j+1][k];
	            eigenvectors[j+1][k] = tmp;
	        }
	    }
}

/************************************************************************
* NAME: square_singular_value_decomposition (input_matrix, output_matrix,
*		ninputs, eigenvalues)
*
* DESCRIPTION:
*		uses Singular Value Decomposition function svdcmp
*		eigenvalues	vector of eigenvalues
*
*************************************************************************/
void
square_singular_value_decomposition (ninputs, input_matrix, 
    eigenvectors, eigenvalues)
    
    int ninputs;
    float **input_matrix, **eigenvectors, *eigenvalues;
{
    int i,j;
    float *w,**u,**v;
    float sum=0;
    void svdcmp();
    /* these subroutines are kept only for use with numerical recipes routines*/
    /* Allocate memory for the w, and v matrices */
    w = allocate_vector(ninputs+1);
    v = allocate_matrix(ninputs+1, ninputs+1);
    u = allocate_matrix(ninputs+1, ninputs+1);

    /* Copy M_matrix (normal indexing) to u (1 to N indexing)  */
    for (i=1;i<=ninputs;i++)
      for (j=1;j<=ninputs;j++)
	u[i][j]=input_matrix[i-1][j-1];

    /* perform decomposition */
    svdcmp(u,ninputs,ninputs,w,v);
  
	/* copy u transpose into eigenvector array (rows are eigenvectors) */
    for (i=1; i<=ninputs; i++) {
        eigenvalues[i-1] = w[i];
        for (j=1;j<=ninputs;j++) eigenvectors[j-1][i-1]=u[i][j];
    }

    if (debug>0) printf("square_singular_value_decomposition: %f %f %d\n", 
        eigenvalues[0], eigenvalues[1], ninputs);

       /* if sum of diagonal of eigenvector matrix is negative, reverse sign */
    for (i=0; i<ninputs; i++) {
        sum += eigenvectors[i][i];
    }
    if (sum < -.00001) {
        if (debug>0) printf("  Changing sign %d\n",ninputs);
        for (i=0; i<ninputs; i++) {
            for (j=0; j<ninputs; j++) {
                eigenvectors[i][j] = -eigenvectors[i][j];
            }
        }
    }

    square_normalize_and_sort(ninputs, eigenvectors, eigenvalues);

    if (debug>0) for (i=0; i<ninputs; i++) {
        printf("eigenvalue %d = %f\n",i,eigenvalues[i]);
    }

    /* free the memory used for the matrices */
    deallocate_vector(w);
    deallocate_matrix(u);
    deallocate_matrix(v);
} 

/*****************************************************************************
* NAME: transposed_matrix_vector_multiply ()
* DESCRIPTION:
*       Transposed matrix multiply:   result =(matrix)' * vector
*	matrix is a matrix, columns by rows
*	vector is a matrix, columns by 1
*	result is the result, rows by 1
*
*****************************************************************************/
void
transposed_matrix_vector_multiply (matrix, vector, result, rows, columns)
    int rows, columns;
    float **matrix;
    float *result, *vector;
{
    register int i,k;
    register float *vp, *mp;
    register float tmp_v;

    vp = result;
    for (i=0; i<rows; i++) *vp++ = 0.;

    for (k=0; k<columns; k++) {
	vp = result;
	mp = &matrix[k][0];
	tmp_v = *vector++;
	for (i=0; i<rows; i++)
	    *vp++ += *mp++ * tmp_v;
    }
}



