/***---------------------------------------------------------------------------
* NAME:		nnargs.c 
* PURPOSE:	argument parser subroutines
*
* DESCRIPTION:
*   nnargs()		 - process command line 
*   nnargs_part()	 - process command line, allow args not in table
*   nnargs_part_from_file() - process command line arguments from file 
*   nnargs_from_file()	 - process command line arguments from file 
*   nnargs_print_table() - print table with flags, descriptions, and values
*   nnargs_print_flags() - print all command line flags and values
*   nnargs_get_flags()   - return string with all flags and values 
*   nnargs_get_part_flags()
*   nnargs_get_double_list() - parse double list string
*   nnargs_get_float_list() - parse float list string
*   nnargs_get_int_list() - parse integer list string
*   nnargs_get_string_list() - parse string list string
*   nnargs_get_double_alloc_list() - parse double list string, allocating array
*   nnargs_get_float_alloc_list() - parse float list string, allocating array
*   nnargs_get_int_alloc_list() - parse integer list string, allocating array
*   nnargs_get_string_alloc_list() - parse string list string, allocating array
* 
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/nnargs.c,v 11.6 2004/02/12 14:16:30 kukolich Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/nnargs.c,v $
* CREATED: 26-July-90 rpl                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"

typedef int (*PFI)();
static int nnargs_parse_table(int argc, char **argv, ARG *tabp,
			      int tabsize, int part_flag);
static void 	setarg(ARG *argp, char *value);
static ARG	*findarg(char *name, ARG *tabp, int tabsize);
static char *nnargs_local_get_flags(ARG *tabp, int tabsize, int csh_flag);
static void nnargs_add_string(char *head_string, char *addition, int csh_flag);
static int stoi(register char *str);
static void set_allocated_lists(ARG *tabp, int tabsize);
/* macros for nnargs_parse_table */
#define NOT_PART 0
#define PART 1

/***---------------------------------------------------------------------------
* NAME:		nnargs 
* PURPOSE:	command line argument processor
* DESCRIPTION: 
*   Processes command line arguments. Strips all recognized command
*   line flags from argv and sets values for each flag.
*   Rpl got the original idea from
*   Dr. Dobbs Toolbook of C, QA76.73.C15 D7 1986 pp 465-477.
*   This code was first entered, debugged and modified by rpl 
*   to add more argument types.
*   It was then modified by maz to allow string argument names 
*   with spaces between the flag and the value. 
*   It was most recently modified by rpl to add a network descriptor
*   flag, allow longer argument names, allow more room 
*   on each line for flag description, provide new functions 
*   nnargs_get_flags(), nnargs_print_flags(), nnargs_print_table(),
*   and eliminate the function flag type.
--------------------------------------------------------------------------***/
int
nnargs(int argc, char **argv, ARG *tabp,
       int tabsize/* tabsize is number of entries */)
{
    return nnargs_parse_table(argc, argv, tabp, tabsize, NOT_PART);
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_part
* PURPOSE:	nnargs for multiple argument tables
* DESCRIPTION: 
*	parses arguments as nnargs() would, except that it allows
*	arguments not in its table to pass
--------------------------------------------------------------------------***/
int nnargs_part(int argc, char **argv, ARG *tabp, int tabsize)
{
    return nnargs_parse_table(argc, argv, tabp, tabsize, PART);
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_parse_table
* PURPOSE:	nnargs call that handles either nnargs or nnargs_part calls
* DESCRIPTION: 
*	parses arguments on the command line
--------------------------------------------------------------------------***/
static int nnargs_parse_table(int argc, char **argv, ARG *tabp,
			      int tabsize, int part_flag)
{
    register int	nargc;
    register char	**nargv;
    register ARG	*argp;
    char	*switch_name;
    char	*switch_val;
    char	*prog_name;

    prog_name = argv[0];
    nargc = 1;
    for (nargv = ++argv; --argc > 0; argv++) {
	/* pass this argument on to the main driver program 
	   ------------------------------------*/
        if ( **argv != '-' ) {
	    *nargv++ = *argv;
	    nargc++;
	    continue;
	}
	/* print out all arguments 
	   ------------------------------*/
        if ((strcmp(*argv, "--") == 0) || (strcmp(*argv, "-help") == 0)) {
	    /* this is a way for the user to print the possible args */
	    fprintf(stderr, "Arguments are:\n");
	    nnargs_print_table (tabp, tabsize, stderr);
	    if(part_flag){
		/* put -- back on the argument list.  all arguments after --
		   are lost */
		*nargv++ = *argv;
		nargc++;
		return(nargc);
	    } else {
		exit(1);
	    }
	}
	/* decode argument
	   ----------------------*/
	switch_name=(*argv)+1; /* move past the minus */
	argp = findarg(switch_name, tabp, tabsize);
	/* invalid argument */
	if (argp == NULL) {
	    if(part_flag){
		/* pass this argument to the next table */
		*nargv++ = *argv;
		nargc++;
		continue;
	    } else {
		/* if argument is a negative number, let it pass on */
		if('0' <= switch_name[0] && switch_name[0] <= '9'){
		    /* could have trouble with some floats */
		    *nargv++ = *argv;
		    nargc++;
		    continue;
		}
		/* else, this is a bad flag */
		fprintf(stderr, 
			"nnargs: bad flag \"-%s\".\nAllowable flags are:\n", 
			switch_name);
		nnargs_print_table (tabp, tabsize, stderr);
		fprintf(stderr, "type %s -- %s\n",
			prog_name, "to see all flags");
		exit(1);
	    }
	}
	if (argp->type == BOOLEAN){
	    /* BOOLEAN type has no following value */
	    setarg(argp, NULL);
	    continue;
	}
	/* pick up value after flag */
	--argc;
	/* value missing */
	if (argc <= 0) {
	    warning( "nnargs: flag \"%s\" must have an argument\n",
		    switch_name);
	} else {
	    /* get the value and set the argument to it */
	    ++argv;
	    switch_val = *argv;
	    setarg(argp, switch_val);
	}
    }
    /* fix up NULL allocated lists */
    set_allocated_lists(tabp, tabsize);
    return nargc;
}

/* setarg - Set an argument
   argp points at the argument table entry to be set.
   value points at the desired value
--------------------------------------------------------------------------*/
static void
setarg(ARG *argp, char *value)
{
    switch(argp->type) {
      case INT:
	*(argp->variable) = stoi(value);
	if(argp->lower_limit != argp->upper_limit)
	  if(*(argp->variable) < argp->lower_limit ||
	     *(argp->variable) > argp->upper_limit)
	    fatal_error("%s -%s %s (%d) %s %d to %d\n",
			"nnargs: Argument",argp->arg,"has value",
			*argp->variable, "out of allowed range:",
			(int)(argp->lower_limit), 
			(int)(argp->upper_limit));
	break;
      case INT_LIST:
	nnargs_get_int_list(value, (int *)(argp->variable),
			    (int)argp->lower_limit);
	break;
      case INT_ALLOC_LIST:
	*((int **)argp->variable) = nnargs_get_int_alloc_list(value);
	argp->lower_limit = (float)((*(int **)argp->variable)[0]);
	break;
      case FLOAT:
	*((float *)(argp->variable)) = atof(value);
	if(argp->lower_limit != argp->upper_limit)
	  if(*((float *)(argp->variable))<argp->lower_limit ||
	     *((float *)(argp->variable))>argp->upper_limit)
	    fatal_error("%s -%s %s (%f) %s: %f to %f\n",
			"nnargs: Argument",argp->arg,"has value",
			*((float *)argp->variable), 
			"out of allowed range",
			(float)(argp->lower_limit), 
			(float)(argp->upper_limit));
	break;
      case FLOAT_LIST:
	nnargs_get_float_list(value, (float *)(argp->variable),
			      (int)argp->lower_limit);
	break;
      case FLOAT_ALLOC_LIST:
	*((float **)argp->variable) = nnargs_get_float_alloc_list(value);
	argp->lower_limit = argp->variable[0];
	break;
      case DOUBLE:
	*((double *)(argp->variable)) = atof(value);
	if(argp->lower_limit != argp->upper_limit)
	  if(*((double *)(argp->variable))<argp->lower_limit||
	     *((double *)(argp->variable))>argp->upper_limit)
	    fatal_error("%s -%s %s (%f) %s: %d to %d\n",
			"nnargs: Argument",argp->arg,"has value",
			*((double *)argp->variable), 
			"out of allowed range",
			argp->lower_limit, argp->upper_limit);
	break;
      case DOUBLE_LIST:
	nnargs_get_double_list(value,(double*)(argp->variable),
			       (int)argp->lower_limit);
	break;
      case DOUBLE_ALLOC_LIST:
	*((double **)argp->variable) = nnargs_get_double_alloc_list(value);
	argp->lower_limit = (float)(((double *)argp->variable)[0]);
	break;
      case FLAG:
	*(argp->variable) = 1;
	break;
      case CHAR:
	*(argp->variable) = value[0];
	break;
      case STRING:
	*((char **)(argp->variable)) = value;
	break;
      case STRING_LIST:
	nnargs_get_string_list(value,(char **)(argp->variable),
			       (int)argp->lower_limit);
	break;
      case STRING_ALLOC_LIST:
	*((char ***)argp->variable) = nnargs_get_string_alloc_list(value);
	argp->lower_limit = atof(((char **)argp->variable)[0]);
	break;
      default:
	fatal_error("%s %d\n",
		    "nnargs: Bad Argument Type in Table:",
		    argp->type);
    }
}
/* findarg - find argument
   Return pointer to argument table entry corresponding to name
   (or NULL if name isn't in table)
------------------------------------------------------*/
static 	ARG *
findarg(char *name, ARG *tabp, int tabsize)
{
    for (; --tabsize >= 0; tabp++)
      if (strcmp(tabp->arg,name) == 0)
	return tabp;
    return (NULL);
}

/*---------------------------------------------
  fill_argv()
   Fill an argv style string with tokens taken from describe_file.
   The string returned is a static.
   Each time fill_argv is called, the string's contents are overwritten.

   Function returns the number of tokens read (argc)
   -------------------------------------------------*/
int fill_argv(char **argv, FILE *describe_file, char *dfname)
{
    int	argc, t, stop, s, e, eof = 0;
    static int command_len = 0;
    static char *command = NULL, c;
#ifdef notdef
    int i;
    char 	buffer[NNMAX_LINE];
    static	char command[NNMAX_LINE];
#endif
    /* fill up command with the data */
    /* we need to make command an allocated (reallocated)
       array.  this makes not strdup'ing strings from a second
       argv call VERY bad, since they would no longer point into
       allocated space */
    if(command == NULL){
	/* allocate NNMAX_LINE's worth to start */
	command = nn_malloc(NNMAX_LINE, "for fill_argv string");
	command_len = NNMAX_LINE;
    }
    command[0] = '\0';
    stop = 0;
    t = 0;
    while(stop == 0){
	c = getc(describe_file);
	/* check if command is long enough, if not, make it bigger */
	if(t >= command_len){
	    command = nn_realloc(command, command_len+NNMAX_LINE,
				 "for fill_argv string");
	    command_len += NNMAX_LINE;
	}
	if(c == (char)EOF){
	    /* if at the end of a file, stop reading */
	    command[t] = '\0';
	    stop = 1;
	    eof = 1;
	} else if(c == '\n'){
	    /* if at the end of a line, check for a backslash to
	       continue or stop reading */
	    if(t > 0 && command[t-1] == '\\'){
		command[t-1] = '\0';
		t -= 2; /* overwrite \\ and \n */
	    } else if (t > 1 && command[t-1] == '\r'&& command[t-2] == '\\'){
	      command[t-2] = '\0';
	      t -= 3; /* same thing, DOS format */
	    } else {
		command[t] = '\0';
		stop = 1;
	    }
	} else {
	    /* add the character to the command line */
	    command[t] = c;
	}
	t++;
    }
#ifdef notdef
    while (stop == 0 && 
	   fgets(buffer,NNMAX_LINE,describe_file) != NULL) {
	i = strlen(buffer);
	if(i > 1 && buffer[i-2] == '\\'){
	    buffer[i-2] = NULL;  /* delete '\' */
	} else {
	    /* last line */
	    stop = 1;
	}
	if(buffer[i-1] == '\n') buffer[i-1] = NULL;/* delete new line */

	strncat(&command[0],&buffer[0], NNMAX_LINE-t);
	t += i;
    }
    command[NNMAX_LINE-1] = '\0';
#endif
    /* fill argv with pointers to the entries in command */
    argc = 0;
    s = 0;
    while(1){
	/* eat initial white space */
	for(;s < command_len-1; s++)
	  if(command[s] != ' ' && command[s] != '\t') break;
	if(command[s] == '\0') {
	  if(argc == 0 && eof) return EOF;
	  else return argc;
	}
	argv[argc] = &command[s];
	if(argv[argc++][0] == '\"'){ /* inside a quote " (for hilit) */
	    for(e = s+1; e < command_len-1; e++){
		if(command[e] == '\"'){ /* " (for emacs hilit)*/
		    e++;
		    /* if the quote was the last thing on the line */
		    if(command[e] == '\0') {
		      if(argc == 0 && eof) return EOF;
		      else return argc;
		    }
		    if(e < command_len) command[e] = '\0';
		    break;
		} else if(command[e] == '\0'){
		    if(argc == 0 && eof) return EOF;
		    else return argc;
		}
	    }
	} else {
	    /* regular token */
	    for(e = s+1; e < command_len; e++){
		if(command[e] == ' ' || command[e] == '\t'){
		    command[e] = '\0';
		    break;
		} else if(command[e] == '\0'){
		    if(argc == 0 && eof) return EOF;
		    else return argc;
		}
	   }
	}
	s = e+1;
	if(command[s] == '\0') break;
    }
    if(argc == 0 && eof) return EOF;
    else return argc;
}    
/***---------------------------------------------------------------------------
* NAME:     nnargs_from_file
* PURPOSE:  get command line arguments from file  
* USAGE: 
*  describe_file = nn_fopen(describe_fname,"r","for reading data description");
*  n = nnargs_from_file(describe_file, describe_fname, argtab, TABSIZE);
*  nn_fclose(describe_file, describe_fname, "for reading data description");
*
* DESCRIPTION:
*	Used primarily to read a descriptor file associated with a data file
*	and process arguments by passing to nnargs.  
*	The first word in the file (command) is irrelevant and is ignored.
*	Multiple lines are allowed.  
*	A line can be continued with a \ which will be ignored.
*	The file will be read until a line without a \ on the end is found
* RETURNS:
*	nargc:	the number of arguments remaining on the list after
*	nnargs is called
*
* CREATED: 11-Feb-91 dan                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
int
nnargs_from_file(FILE *describe_file, char *describe_fname,
		 ARG *tabp, int tabsize)          	               
#define NNMAX_ARGS 100
{
    char	*argv[NNMAX_ARGS];
    int		argc;

    argc = fill_argv(argv, describe_file, describe_fname);
    return nnargs (argc, argv, tabp, tabsize);
}
/***---------------------------------------------------------------------------
* NAME:     nnargs_print_table
* PURPOSE:  print tabular information on all flags and values 
* USAGE: 
*		nnargs_print_table(tabp, tabsize, fp);
* DESCRIPTION:
*	Prints to the desired output a table with one line per flag
*       containing the flag name, type, description, and value.
*       This is printed with a bad flag (--) or it can be used to
*       print the values of all flags. It is less verbose (less print space) to use
*       nnargs_print_flags() to get the value of all flags.
* CREATED: 8-Nov-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
void
nnargs_print_table(ARG *tabp, int tabsize, FILE *fp)
{
    int i, max_length;
    int *array;
    float *f_array;
    double *d_array;
    char **c_array;

    for(; --tabsize >= 0; tabp++) {
	switch(tabp->type) {
	  case INT:
	    fprintf(fp,"-%-10s <int>         %-35s = ",
		    tabp->arg,tabp->errmsg);
	    fprintf(fp,"%-5d", *(tabp->variable));
	    if((int)tabp->lower_limit != (int)tabp->upper_limit)
	      fprintf(fp,":%d to %d", 
		      (int)(tabp->lower_limit),
		      (int)(tabp->upper_limit));
	    break;
	  case INT_LIST:
	    if((int)tabp->lower_limit <= 0) max_length = MAX_INT_LIST;
	    else max_length = (int)tabp->lower_limit;

	    fprintf(fp,"-%-10s <int_list>    %-35s = ",
		    tabp->arg,tabp->errmsg);
	    array = tabp->variable;
	    if (array[0] > 0) {
		fprintf(fp, "%d", array[1]);
	    }
	    if (array[0] > 1) {
		for(i=2; i<= array[0] && i <= max_length; i++)
		  fprintf(fp,",%d", array[i]);
	    }
	    fprintf(fp,":%d long", max_length-1);
	    break;
	  case INT_ALLOC_LIST:
	    fprintf(fp,"-%-10s <int_list>    %-35s = ",
		    tabp->arg,tabp->errmsg);
	    array = *(int **)tabp->variable;
	    if(array == NULL || array[0] <= 0) fprintf(fp, "<empty>");
	    else {
		if (array[0] > 0) {
		    fprintf(fp, "%d", array[1]);
		}
		if (array[0] > 1) {
		    for(i=2; i<= array[0]; i++)
		      fprintf(fp,",%d", array[i]);
		}
	    }
	    break;
	  case FLOAT:
	    fprintf(fp,"-%-10s <float>       %-35s = ",
		    tabp->arg,tabp->errmsg);
	    fprintf(fp,"%-5g", *(float *)(tabp->variable));
	    if(tabp->lower_limit != tabp->upper_limit)
	      fprintf(fp,":%g to %g", 
		      (float)(tabp->lower_limit),
		      (float)(tabp->upper_limit));
	    break;
	  case FLOAT_LIST:
	    if((int)tabp->lower_limit <= 0) max_length = MAX_FLOAT_LIST;
	    else max_length = (int)tabp->lower_limit;

	    fprintf(fp,"-%-10s <float_list>  %-35s = ",
		    tabp->arg,tabp->errmsg);
	    f_array = (float *)tabp->variable;

	    if(f_array[0] > 0) {
		fprintf(fp, "%g", f_array[1]);
	    }
	    if (f_array[0] > 1) {
		for(i=2; i<= f_array[0] && i <= max_length; i++)
		  fprintf(fp,",%g", f_array[i]);
	    }
	    fprintf(fp,":%d long", max_length-1);
	    break;
	  case FLOAT_ALLOC_LIST:
	    fprintf(fp,"-%-10s <float_list>  %-35s = ",
		    tabp->arg,tabp->errmsg);
	    f_array = *(float **)tabp->variable;
	    if(f_array == NULL || f_array[0] <= 0) fprintf(fp, "<empty>");
	    else {
		if(f_array[0] > 0) {
		    fprintf(fp, "%g", f_array[1]);
		}
		if (f_array[0] > 1) {
		    for(i=2; i<= f_array[0]; i++)
		      fprintf(fp,",%g", f_array[i]);
		}
	    }
	    break;
	  case DOUBLE:
	    fprintf(fp,"-%-10s <double>      %-35s = ",
		    tabp->arg,tabp->errmsg);
	    fprintf(fp,"%-5g", *(double *)(tabp->variable));
	    if(tabp->lower_limit != tabp->upper_limit)
	      fprintf(fp,":%g to %g", 
		      (double)(tabp->lower_limit),
		      (double)(tabp->upper_limit));
	    break;
	  case DOUBLE_LIST:
	    if((int)tabp->lower_limit <= 0) max_length = MAX_DOUBLE_LIST;
	    else max_length = (int)tabp->lower_limit;

	    fprintf(fp,"-%-10s <double_list> %-35s = ",
		    tabp->arg,tabp->errmsg);
	    d_array = (double *)tabp->variable;

	    if (d_array[0] > 0) {
		fprintf(fp, "%g", d_array[1]);
	    }
	    if (d_array[0] > 1) {
		for(i=2; i<= d_array[0] && i <= max_length; i++)
		  fprintf(fp,",%g", d_array[i]);
	    }
	    fprintf(fp,":%d long", max_length-1);
	    break;
	  case DOUBLE_ALLOC_LIST:
	    fprintf(fp,"-%-10s <double_list> %-35s = ",
		    tabp->arg,tabp->errmsg);
	    d_array = *(double **)tabp->variable;
	    if(d_array == NULL || d_array[0] <= 0) fprintf(fp, "<empty>");
	    else {
		if (d_array[0] > 0) {
		    fprintf(fp, "%g", d_array[1]);
		}
		if (d_array[0] > 1) {
		    for(i=2; i<= d_array[0]; i++)
		      fprintf(fp,",%g", d_array[i]);
		}
	    }
	    break;
	  case FLAG:
	    fprintf(fp,"-%-10s <flag>        %-35s = ",
		    tabp->arg,tabp->errmsg);
	    fprintf(fp,"%-5s", *(tabp->variable) ? "TRUE":"FALSE");
	    break;
	  case CHAR:
	    fprintf(fp,"-%-10s <char>        %-35s = ",
		    tabp->arg,tabp->errmsg);
	    fprintf(fp,"%c", *(tabp->variable) );
	    break;
	  case STRING:
	    fprintf(fp,"-%-10s <str>         %-35s = ",
		    tabp->arg,tabp->errmsg);
	    fprintf(fp,"\"%s\"", *(char **)tabp->variable );
	    break;
	  case STRING_LIST:
	    fprintf(fp,"-%-10s <string_list> %-35s = ",
		    tabp->arg,tabp->errmsg);
	    c_array = (char **)tabp->variable;

	    if((int)tabp->lower_limit <= 0) max_length = MAX_STRING_LIST;
	    else max_length = (int)tabp->lower_limit;

	    if (c_array[0] != NULL && atoi(c_array[0]) > 0){
		fprintf(fp, "%s", c_array[1]);
	    }
	    if (c_array[0] != NULL && atoi(c_array[0]) > 1) {
		for(i=2; i<= atoi(c_array[0]) && i <= max_length; i++)
		  fprintf(fp,",%s", c_array[i]);
	      }
	    fprintf(fp,":%d long", max_length-1);
	    break;
	  case STRING_ALLOC_LIST:
	    fprintf(fp,"-%-10s <string_list> %-35s = ",
		    tabp->arg,tabp->errmsg);
	    c_array = *(char ***)tabp->variable;
	    if(c_array == NULL) fprintf(fp, "<empty>");
	    else {
		if(c_array[0] == NULL) fprintf(fp, "<empty>");
		else {
		    i = atoi(c_array[0]);
		    if(i <= 0) fprintf(fp, "<empty>");
		    else {
			if(i > 0) fprintf(fp, "%s", c_array[1]);
			if(i > 1)
			  for(i=2; i<= atoi(c_array[0]); i++)
			    fprintf(fp,",%s", c_array[i]);
		    }
		}
	    }
	    break;
	}
	fprintf(fp, "\n");
    }
}

/***---------------------------------------------------------------------------
* NAME:     nnargs_get_flags
* PURPOSE:  return string with command line flags and values
* USAGE: 
*	char command_line[1024];
*       strcat(command_line, "mlp ");
*       strcat(command_line, nnargs_get_flags(tabp, tabsize));
*       system(command_line);
* DESCRIPTION:
*	Returns a string which contains every single flag and the value
*       of the flag. This can be used to create a shell
*       script or command line to start a program.
*       Breaks lines with backslash carriage return to make the string
*       readable but still useful as a cshell command 
* SEE ALSO: nnargs_local_get_flags()
* CREATED: 8-Nov-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
char *
nnargs_get_flags(ARG *tabp, int tabsize)
{
    char *tmp_string;
    
    tmp_string = nnargs_local_get_flags(tabp, tabsize, 1);
    strcat(tmp_string, "\n");
    return (tmp_string);
}

/***---------------------------------------------------------------------------
* NAME:     nnargs_get_part_flags
* PURPOSE:  return string with command line flags and values
*	like nnargs_get_flags, but with a \ on the end of the last line
* USAGE: 
*	char command_line[1024];
*       strcat(command_line, "mlp ");
*       strcat(command_line, nnargs_get_part_flags(tabp, tabsize));
*	strcat(command_line, nanrgs_get_flags(tabp2, tabsize2));
*       system(command_line);
* SEE ALSO: nnargs_local_get_flags()
* CREATED: 10-Feb-93 lck			MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
char *
nnargs_get_part_flags(ARG *tabp, int tabsize)
{
    char *tmp_string;
    
    tmp_string = nnargs_local_get_flags(tabp, tabsize, 1);
    strcat(tmp_string, " \\");
    strcat(tmp_string, "\n");
    return (tmp_string);
}

/***---------------------------------------------------------------------------
* NAME:     nnargs_get_flags_one_line
* PURPOSE:  return string with command line flags and values
*	like nnargs_get_flags, but with no carriage returns and no backslashes
* USAGE: 
*	char command_line[1024];
*       strcat(command_line, nnargs_get_flags_one_line(tabp, tabsize));
* SEE ALSO: nnargs_local_get_flags()
* CREATED: 23-Feb-95 lck			MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
char *
nnargs_get_flags_one_line(ARG *tabp, int tabsize)
{
    return nnargs_local_get_flags(tabp, tabsize, 2);
}

/***---------------------------------------------------------------------------
* NAME:     nnargs_print_flags
* PURPOSE:  print all (even default) command line flags and values
* USAGE: 
*		nnargs_print_flags(tabp, tabsize, stderr);
* DESCRIPTION:
*	Prints to the desired output a string which contains every single flag 
*       with its values including default values. This is a short alternative
*       to prusage.
* SEE ALSO: nnargs_get_flags()
* CREATED: 8-Nov-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
void
nnargs_print_flags(ARG *tabp, int tabsize, FILE *fp)
{
    fprintf(fp, "ALL FLAGS:\n%s\n", nnargs_local_get_flags(tabp, tabsize, 0));
}

/***---------------------------------------------------------------------------
* NAME:     nnargs_local_get_flags
* PURPOSE:  build string with all (even default) command line flags and values
* USAGE: 
*	Used by nnargs_get_flags and nnargs_print_flags
* DESCRIPTION:
*	Returns a string which contains every single flag and the value
*       of the flag. Depending on the value of the csh_flag flag, 
*       lines will be broken either by carriage returns (csh_flag=0) for 
*       easy reading or by backslash carriage returns (csh_flag=1) for 
*       creating a cshell command line.
* BUGS AND LIMITATIONS:
* 	Maximum command line string length is 
*       NNMAX_LINE (currently 2048) characters.
*       Doesn't check length of each argument and value string when 
*       building it up using the dummy array but assumes the 
*       argument and value fit in 256 characters.
*
* CREATED: 8-Nov-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
static char *
nnargs_local_get_flags(ARG *tabp, int tabsize, int csh_flag)
{
    int i, max_length;
    int *array;
    float *f_array;
    double *d_array;
    char **c_array;
    static char all_flags[NNMAX_LINE];
    static char dummy[256];
    
    all_flags[0] = '\0';
    for(; --tabsize >= 0; tabp++) {
	switch(tabp->type) {
	  case INT:
	    sprintf(dummy, "-%s %d", tabp->arg, *(tabp->variable));
	    nnargs_add_string(all_flags, dummy, csh_flag);
	    break;
	  case INT_LIST:
	    if((int)tabp->lower_limit >= 0) max_length = MAX_INT_LIST;
	    else max_length = (int)tabp->lower_limit;
	    array = tabp->variable;
	    if (array[0] > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%d", array[1]);
		if (array[0] > 1) {
		    for(i=2; i<= array[0] && i <= max_length; i++)
		      sprintf( dummy + strlen(dummy), ",%d", array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	  case INT_ALLOC_LIST:
	    array = *(int **)tabp->variable;
	    if (array != NULL && array[0] > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%d", array[1]);
		if (array[0] > 1) {
		    for(i=2; i<= array[0]; i++)
		      sprintf( dummy + strlen(dummy), ",%d", array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	  case FLOAT:
	    sprintf(dummy, "-%s %g", tabp->arg, *(float *)(tabp->variable));
	    nnargs_add_string(all_flags, dummy, csh_flag);
	    break;
	  case FLOAT_LIST:
	    if((int)tabp->lower_limit >= 0) max_length = MAX_FLOAT_LIST;
	    else max_length = (int)tabp->lower_limit;
	    f_array = (float *)tabp->variable;
	    if (f_array[0] > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%g", f_array[1]);
		if (f_array[0] > 1) {
		    for(i=2; i<= f_array[0] && i <= max_length; i++)
		      sprintf( dummy + strlen(dummy), ",%g", f_array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	  case FLOAT_ALLOC_LIST:
	    f_array = *(float **)tabp->variable;
	    if (f_array != NULL && f_array[0] > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%g", f_array[1]);
		if (f_array[0] > 1) {
		    for(i=2; i<= f_array[0]; i++)
		      sprintf( dummy + strlen(dummy), ",%g", f_array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	  case DOUBLE:
	    sprintf(dummy, "-%s %g", tabp->arg, *(double *)(tabp->variable));
	    nnargs_add_string(all_flags, dummy, csh_flag);
	    break;
	  case DOUBLE_LIST:
	    if((int)tabp->lower_limit >= 0) max_length = MAX_DOUBLE_LIST;
	    else max_length = (int)tabp->lower_limit;
	    d_array = (double *)tabp->variable;
	    if (d_array[0] > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%g", d_array[1]);
		if (d_array[0] > 1) {
		    for(i=2; i<= d_array[0] && i <= max_length; i++)
		      sprintf( dummy + strlen(dummy), ",%g", d_array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	  case DOUBLE_ALLOC_LIST:
	    d_array = *(double **)tabp->variable;
	    if (d_array != NULL && d_array[0] > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%g", d_array[1]);
		if (d_array[0] > 1) {
		    for(i=2; i<= d_array[0]; i++)
		      sprintf( dummy + strlen(dummy), ",%g", d_array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	  case FLAG:
	    /* if flag set, print it in all cases */
	    if ( *(tabp->variable) > 0) {
		sprintf(dummy, "-%s", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else {
		/* if flag not set,  print it with brackets when not csh */
		if ( csh_flag <= 0) {
		    sprintf(dummy, "[-%s NOT SET]", tabp->arg);
		    nnargs_add_string(all_flags, dummy, csh_flag);
		}
	    }
	    break;
	  case CHAR:
	    sprintf(dummy, "-%s %c", tabp->arg, *(tabp->variable));
	    nnargs_add_string(all_flags, dummy, csh_flag);
	    break;
	  case STRING:
	    if(*(char **)tabp->variable != NULL &&
	       strcmp(*(char **)tabp->variable, "") != 0){
		sprintf(dummy, "-%-s %s", tabp->arg,
			*(char **)tabp->variable );
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else {
		if( csh_flag <= 0) {
		    sprintf(dummy, "[-%-s NOT SET]", tabp->arg);
		    nnargs_add_string(all_flags, dummy, csh_flag);
		}
	    }
	    break;
	  case STRING_LIST:
	    if((int)tabp->lower_limit >= 0) max_length = MAX_INT_LIST;
	    else max_length = (int)tabp->lower_limit;
	    c_array = (char **)tabp->variable;
	    if (c_array[0] != NULL && atoi(c_array[0]) > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%s", c_array[1]);
		if (atoi(c_array[0]) > 1) {
		    for(i=2; i<= atoi(c_array[0]) && i <= max_length; i++)
		      sprintf( dummy + strlen(dummy), ",%s", c_array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	  case STRING_ALLOC_LIST:
	    c_array = *(char ***)tabp->variable;
	    if (c_array != NULL && c_array[0] != NULL && atoi(c_array[0]) > 0){
		sprintf(dummy, "-%s ", tabp->arg);
		sprintf( dummy + strlen(dummy), "%s", c_array[1]);
		if (atoi(c_array[0]) > 1) {
		    for(i=2; i<= atoi(c_array[0]); i++)
		      sprintf( dummy + strlen(dummy), ",%s", c_array[i]);
		}
		nnargs_add_string(all_flags, dummy, csh_flag);
	    } else if(csh_flag <= 0){
		sprintf(dummy, "[-%s EMPTY]", tabp->arg);
		nnargs_add_string(all_flags, dummy, csh_flag);
	    }
	    break;
	}
    }
    return(all_flags);
    
}
/***---------------------------------------------------------------------------
* NAME:     nnargs_add_string
* PURPOSE:  cat flag/value strings together to build up a long command line
* DESCRIPTION:
*	Cats string, checks for too long string, adds cr when necessary.
*       Separates flag/value pairs by two spaces. Adds a space at
*       the beginning of all but the first line to allow the backslash
*       cr to be gobbled up by the Bourne shell.
*
* CREATED: 8-Nov-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
static void
nnargs_add_string(char *head_string, char *addition,
		  int csh_flag/* if 1 adds backslash at end of line */)
{
    int head_length;
    int addition_length;
    static int last_cr;
    int new_line_flag=0;

    addition_length = strlen(addition) + 2;
    head_length = strlen(head_string);

    /* initialize at beginning of new string */
    if ( head_length == 0) {
	new_line_flag=1;
	last_cr=0;
	}

    /* add line break if string for this flag and value are too long 
       for csh_flag >= 2, don't put in ANY line breaks */
    if (csh_flag < 2){
	if ( head_length - last_cr + addition_length > 78) {
	    if ( csh_flag)
	      strcat(head_string, "\\\n");     
	    else
	      strcat(head_string, "\n");     
	    
	    last_cr = strlen(head_string);
	    new_line_flag=1;
	}
    }

    /* error if try to create a string that is too long */
    if ( head_length + addition_length > NNMAX_LINE)
      fatal_error("%s: %s = %d %s\n%s\n%s:\n%s",
		  "nnargs_get_flags or nnargs_print_flags",
		  "Flags and values take more than NNMAX_LINE",
		  NNMAX_LINE, "characters to store.",
		  "To fix increase NNMAX_LINE.",
		  "Flags and values that fit in this space are:",
		  head_string);

    /* add addition between flag/value pairs */
    if (new_line_flag != 1) strcat( head_string, "  ");
    /* add single space at the beginning of all but the first line */
    else {
	if ( head_length > 0)
	    strcat( head_string, " ");
    }
    strcat(head_string, addition);
}

/* stoi - convert ascii string to integer ( enhances atoi)
   If string starts with 0x it is interpreted as hex, else
   if it starts with 0 it is octal, else it is decimal.
   Conversion stops on first character not a digit in the indicated radix.
------------------------------------------------------*/
#define islower(c) ('a'<= (c) && (c) <= 'z')
#define toupper(c) ( islower(c) ? (c) - ('a' - 'A') : (c) )

static int
stoi(register char *str)
{
    register int num=0;
    int sign=1;

    while(*str == ' ' || *str == '\t' || *str == '\n') str++;

    if (*str == '-') {
	sign = -1;
	str++;
    }
    if (*str == '0') {
	++str;
	if (*str == 'x' || *str == 'X') {
	    str++;
	    while( ('0' <= *str && *str <= '9') ||
		  ('a' <= *str && *str <= 'f') ||
		  ('A' <= *str && *str <= 'F')   ) {
		num *= 16;
		num += ('0' <= *str && *str <= '9') ?
		  *str - '0': toupper(*str) - 'A' + 10;
		str++;
	    }
	} else {
	    while ('0' <= *str && *str <= '7') {
		num *= 8;
		num += *str++ - '0';
	    }
	}
    } else {
	while ('0' <= *str && *str <= '9') {
	    num *= 10;
	    num += *str++ - '0';
	}
    }
    return (num*sign);
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_get_double_list
* PURPOSE:	extract double array from string like 1.3,-50.,.50,1.0E3 
* USAGE: 
* 	nnargs_get_double_list(string, array, length)
*   	char	*string;		- string list of doubles
*   	double	*array;			- double array for result
*	int	length;			- maximum length for array
* DESCRIPTION:
*	When done, array[0] will have the number of layers, array[1] will
*   have the number of nodes in the first layer, array[2] in the second,
*   array[3] in the third ....
*   For example, if the input string is 1.3,-50.,.50,1.0E3
*   Then after the function call array[0]=4, array[1]=1.3, array[2]=-50,
*   array[3]=.5
* 
* BUGS AND LIMITATIONS:
* 	Destroys the string.  If length is 0, MAX_NNLIST is used.
*
* CREATED: 13-Feb-91 dan                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
void nnargs_get_double_list(char *string,/* string from command line */
			   double *array,/* double array to store values */
			   int length/* maximum length for the array */)
{
    char *slevel;
    int nlevel=0;

    if(length == 0) length = MAX_NNLIST;

    slevel = nn_strtok(string, ",:");
    /* first separator exists read in nodes for all other levels
    -----------------------------------------------------------*/
    if(slevel != NULL){
	array[nlevel+1] = atof(slevel);
	nlevel++;
	while ( (slevel = nn_strtok(NULL, ",:")) != NULL) {
	    array[nlevel+1] = atof(slevel);
	    nlevel++;
	    if (nlevel > length-1) {
		warning("%s %d %s\n",
			"Number of double list entries has exceeded",
			length-1,"The list has been truncated.");
		nlevel--;
		break;
	    }
	}
    }
    array[0] = nlevel;
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_get_float_list
* PURPOSE:	extract float array from string like 1.3,-50.,.50,1.0E3 
* USAGE:
* 	nnargs_get_float_list(string, array,length)
*   	char	*string;		- string list of floats
*   	float	*array;			- double array for result
*	int	length;			- maximum length for array
* DESCRIPTION:
*	nnargs_get_float_list is the same as nnargs_get_double_list
*	except that the array being filled is a vector of floats
*
* CREATED: 13-Feb-91 dan                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
void nnargs_get_float_list(char *string,/* string from command line */
			  float *array,/* float array to store values */
			  int length/* maximum length for the array */)
{
    char *slevel;
    int nlevel=0;

    if(length == 0) length = MAX_FLOAT_LIST;

    slevel = nn_strtok(string, ",:");

    /* first separator exists read in nodes for all other levels
    -----------------------------------------------------------*/
    if(slevel != NULL){
	array[nlevel+1] = (float)atof(slevel);
	nlevel++;
	while ( (slevel = nn_strtok(NULL, ",:")) != NULL) {
	    array[nlevel+1] = (float) atof(slevel);
	    nlevel++;
	    if (nlevel > length-1) {
		warning("%s %d %s\n",
			"Number of float list entries has exceeded",
			length-1, ".  The list has been truncated.");
		nlevel--;
		break;	
	    }
	}
    }
    array[0] = nlevel;
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_get_int_list
* PURPOSE:	extract integer array from string like 13,50,50,10 
* USAGE: 
* 	nnargs_get_int_list(string, array, length)
*   	char	*string;		- string list of integers
*   	int	*array;			- integer array for result
*	int	length;			- maximum length for array
* DESCRIPTION:
*	nnargs_get_int_list is the same as nnargs_get_double_list except
*	that the array being filled is a vector of ints
*
* CREATED: 7-Sept-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
void
nnargs_get_int_list(char *string,/* string from command line */
		    int *array,/* integer array to store values */
		    int length/* maximum length for the array */)
{
    char *slevel;
    int nlevel=0;

    if (length == 0) length = MAX_INT_LIST;

    slevel = nn_strtok(string, ",:");

    /* first separator exists read in nodes for all other levels
    -----------------------------------------------------------*/
    if( slevel != NULL){
	/*printf("int_list: slevel: %s\n",slevel);*/
	array[nlevel+1] = atoi(slevel);
	nlevel++;
	while ( (slevel = nn_strtok(NULL, ",:")) != NULL) {
	    array[nlevel+1] = atoi(slevel);
	    nlevel++;
	    if (nlevel > length-1) {
		warning("%s %d %s\n",
			"Number of integer list entries has exceeded",
			length-1, "The list has been truncated.");
		nlevel--;
		break;	
	    }
	}
    }
    array[0] = nlevel;
}
/*--------------------------------------------------------------------------*\
* NAME:		nnargs_get_string_list
* PURPOSE:	extract array of strings from string like E,AH,OH,U,EH 
* USAGE: 
* 	nnargs_get_string_list(string, array, length)
*   	char	*string;	- string list of strings
*   	char	*array;		- string array for result
*	int	length		- maximum length for array
* DESCRIPTION:
*	When done, array[0] will have the number of layers in ascii, array[1]
*   will have the first label, array[2] the second, array[3] the third ....
*   For example, if the input string is E,AH,OH
*   Then after the function call array[0]="3", array[1]="E", array[2]="AH",
*   array[3]="OH".
* 
* BUGS AND LIMITATIONS:
* 	Replaces the delimiters in the string with NULLs
*	If length is 0, MAX_NNLIST is used.
*	the strings returned in array are still stored in the original
*	string, they are not copied and are frequently overwritten later
*
* CREATED: 31-Jan-91 dan                             MIT LINCOLN LABORATORY
*--------------------------------------------------------------------------*/
void
nnargs_get_string_list(char *string,/* string from command line */
		       char **array,/* string array to store labels */
		       int length/* maximum array length */)
{
    char *s_label;
    int n_label=0;
    /* must be allocated to allow multiple string lists */
    char *ascii_length;
    
    if (length == 0) length = MAX_NNLIST;

    /* !!! because the pointers in array are pointers into string,
       we need to use the real strtok, not nn_strtok. nn_strtok returns
       a static string which is constantly overwritten, not a pointer
       into string !!! */
    s_label = strtok(string, ",:-");
    
    if ( s_label != NULL) {
	/* first separator exists read in nodes for all other levels
	   -----------------------------------------------------------*/
	array[n_label+1] = s_label;
	n_label++;
	while ( (s_label = strtok(NULL, ",:-")) != NULL) {
	    array[n_label+1] = s_label;
	    n_label++;
	    if (n_label > length-1){
		warning("%s %d %s\n",
			"Number of labels has exceeded",
			length-1,"The list has been truncated.");
		n_label--;
		break;
	    }
	}
    }
    ascii_length = nn_malloc(8, "for string length allocation\n");
    sprintf(ascii_length, "%d", n_label);
    array[0] = ascii_length;
}

void
nnargs_set_variable(char *name, int *variable, ARG *tabp, int tabsize)
{
    ARG *argp;

    argp = findarg(name, tabp, tabsize);
    if(argp == NULL) return;
    argp->variable = variable;
}

void
nnargs_set_list_length(char *name, int length, ARG *tabp, int tabsize)
{
    ARG *argp;

    argp = findarg(name, tabp, tabsize);
    if(argp == NULL) return;
    argp->lower_limit = length;
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_get_double_alloc_list
* PURPOSE:	return a double array with entries parsed from the ASCII
*		list of doubles in string.  The length of the array is
*		stored in array[0]
* USAGE: 
* 	double *nnargs_get_double_alloc_list(string)
*   	char	*string;	- string list of doubles
* DESCRIPTION:
*
* CREATED: 31-Aug-93 lck			MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
double *
nnargs_get_double_alloc_list(char *string/* string from command line */)
{
    double *array;
    char *stoken;
    int ntokens, n;

    /* first count the number of separators in the string */
    stoken = string;
    ntokens = 0;
    while(*stoken != '\0'){
	if(*stoken == ',' || *stoken == ':') ntokens++;
	stoken++;
    }
    /* allocate array to hold doubles */
    /* add one for storing the length and one for fence-posting */
    array = allocate_dvector(ntokens+2);
    
    /* parse the tokens */
    stoken = nn_strtok(string, ",:"); /* find first and initialize strtok */
    n = 1;
    if(stoken != NULL){
	/* parse first token, get and parse remaining tokens */
	array[n++] = atof(stoken);
	while ( (stoken = nn_strtok(NULL, ",:")) != NULL) {
	    array[n++] = atof(stoken);
	}
    }
    /* set array[0] for the number of tokens parsed */
    array[0] = n-1;
    return array;
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_get_float_alloc_list
* PURPOSE:	return a float array with entries parsed from the ASCII
*		list of floats in string.  The length of the array is
*		stored in array[0]
* USAGE: 
* 	float *nnargs_get_float_alloc_list(string)
*   	char	*string;	- string list of floats
* DESCRIPTION:
*
* CREATED: 31-Aug-93 lck			MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
float *
nnargs_get_float_alloc_list(char *string/* string from command line */)
{
    float *array;
    char *stoken;
    int ntokens, n;

    /* first count the number of separators in the string */
    stoken = string;
    ntokens = 0;
    while(*stoken != '\0'){
	if(*stoken == ',' || *stoken == ':') ntokens++;
	stoken++;
    }
    /* allocate array to hold floats */
    /* add one for storing the length and one for fence-posting */
    array = allocate_vector(ntokens+2);
    
    /* parse the tokens */
    stoken = nn_strtok(string, ",:"); /* find first and initialize strtok */
    n = 1;
    if(stoken != NULL){
	/* parse first token, get and parse remaining tokens */
	array[n++] = atof(stoken);
	while ( (stoken = nn_strtok(NULL, ",:")) != NULL) {
	    array[n++] = atof(stoken);
	}
    }
    /* set array[0] for the number of tokens parsed */
    array[0] = n-1;
    return array;
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_get_int_alloc_list
* PURPOSE:	return an int array with entries parsed from the ASCII
*		list of ints in string.  The length of the array is
*		stored in array[0]
* USAGE: 
* 	int *nnargs_get_int_alloc_list(string)
*   	char	*string;	- string list of ints
* DESCRIPTION:
*
* CREATED: 31-Aug-93 lck			MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
int *
nnargs_get_int_alloc_list(char *string/* string from command line */)
{
    int *array;
    char *stoken;
    int ntokens, n;

    /* first count the number of separators in the string */
    stoken = string;
    ntokens = 0;
    while(*stoken != '\0'){
	if(*stoken == ',' || *stoken == ':') ntokens++;
	stoken++;
    }
    /* allocate array to hold ints */
    /* add one for storing the length and one for fence-posting */
    array = allocate_ivector(ntokens+2);
    
    /* parse the tokens */
    stoken = nn_strtok(string, ",:"); /* find first and initialize strtok */
    n = 1;
    if(stoken != NULL){
	/* parse first token, get and parse remaining tokens */
	array[n++] = atoi(stoken);
	while ( (stoken = nn_strtok(NULL, ",:")) != NULL) {
	    array[n++] = atoi(stoken);
	}
    }
    /* set array[0] for the number of tokens parsed */
    array[0] = n-1;
    return array;
}
/***---------------------------------------------------------------------------
* NAME:		nnargs_get_string_alloc_list
* PURPOSE:	return a string array with entries parsed from the ascii
*		list of strings in string.  The length of the array is
*		stored in array[0]
*		Unlike nnargs_get_string_list, the strings in the returned
*		array are newly allocated copies of the sub-strings of string.
* USAGE: 
* 	char **nnargs_get_string_alloc_list(string)
*   	char	*string;	- string list of strings
* DESCRIPTION:
*
* CREATED: 31-Aug-93 lck			MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
char **
nnargs_get_string_alloc_list(char *string/* string from command line */)
{
    char **array;
    char *stoken;
    int ntokens, n;

    /* first count the number of separators in the string */
    stoken = string;
    ntokens = 0;
    while(*stoken != '\0'){
	if(*stoken == ',' || *stoken == ':' || *stoken == '-') ntokens++;
	stoken++;
    }
    /* allocate array to hold strings */
    /* add one for storing the length and one for fence-posting */
    array = (char **)allocate_pvector(ntokens+2);
    
    /* parse the tokens */
    /* because we are allocating new space for the tokens once parsed,
       it is ok to use nn_strtok rather than strtok */
    stoken = nn_strtok(string, ",:-"); /* find first and initialize strtok */
    n = 1;
    if(stoken != NULL){
	/* copy first token */
	array[n] = nn_malloc(strlen(stoken)+1, "for string list string");
	strcpy(array[n], stoken);
	n++;
	/* get and copy remaining tokens */
	while ( (stoken = nn_strtok(NULL, ",:-")) != NULL) {
	    array[n] = nn_malloc(strlen(stoken)+1, "for string list string");
	    strcpy(array[n], stoken);
	    n++;
	}
    }
    /* set array[0] for the number of tokens copied */
    array[0] = nn_malloc(8, "for string length allocation\n");
    sprintf(array[0], "%d", n-1);
    return array;
}

/* Combine two argument tables, copying them into a newly allocated
   third table.  return the new table and new size through the pointers
   provided in the argument list */
void
nnargs_combine_argtabs(ARG *first_tab, int first_size,
		       ARG *second_tab, int second_size,
		       ARG **pnew_tab, int *pnew_size)
{
    int oldN, newN, new_nargs;
    ARG *new_tab;

    /* find the length of the new table */
    new_nargs = first_size + second_size;
    *pnew_size = new_nargs;

    /* allocate space for the new table */
    new_tab = allocate_structvector(new_nargs, sizeof(ARG));
    *pnew_tab = new_tab;

    /* fill the new table */
    for(oldN = newN = 0; oldN < first_size; oldN++, newN++){
	/* copy the argument entry */
	new_tab[newN].arg = nn_malloc(strlen(first_tab[oldN].arg)+1,
				      "for argument table entries");
	strcpy(new_tab[newN].arg, first_tab[oldN].arg);
	new_tab[newN].type = first_tab[oldN].type;
	new_tab[newN].variable = first_tab[oldN].variable;
	new_tab[newN].errmsg = nn_malloc(strlen(first_tab[oldN].errmsg)+1,
					 "for argument table entries");
	strcpy(new_tab[newN].errmsg, first_tab[oldN].errmsg);
	new_tab[newN].lower_limit = first_tab[oldN].lower_limit;
	new_tab[newN].upper_limit = first_tab[oldN].upper_limit;
    }
    for(oldN = 0; oldN < second_size; oldN++, newN++){
	/* copy the argument entry */
	new_tab[newN].arg = nn_malloc(strlen(second_tab[oldN].arg)+1,
				      "for argument table entries");
	strcpy(new_tab[newN].arg, second_tab[oldN].arg);
	new_tab[newN].type = second_tab[oldN].type;
	new_tab[newN].variable = second_tab[oldN].variable;
	new_tab[newN].errmsg = nn_malloc(strlen(second_tab[oldN].errmsg)+1,
					 "for argument table entries");
	strcpy(new_tab[newN].errmsg, second_tab[oldN].errmsg);
	new_tab[newN].lower_limit = second_tab[oldN].lower_limit;
	new_tab[newN].upper_limit = second_tab[oldN].upper_limit;
    }
}

/* free an allocated argument table.  it is assumed that this
   table was made using nnargs_combine_argtabs */
void
nnargs_free_argtab(ARG *tab, int size)
{
    int N;

    /* free strings from the table */
    for(N = 0; N < size; N++){
	nn_free(tab[N].arg, "for name in argtab");
	nn_free(tab[N].errmsg, "for help in argtab");
    }
    deallocate_structvector(tab);
}

/* check the allocated lists.  if they point to NULL, allocate enough
   space for them to point to a zero for the list length */
static void
set_allocated_lists(ARG *tabp, int tabsize)
{
    int *i_array;
    float *f_array;
    double *d_array;
    char **c_array;

    for(; --tabsize >= 0; tabp++)
      switch(tabp->type) {
	case INT_ALLOC_LIST:
	  i_array = *(int **)tabp->variable;
	  if(i_array == NULL){
	      i_array = allocate_ivector(1);
	      i_array[0] = 0;
	      *(int **)tabp->variable = i_array;
	  }
	  break;
	case FLOAT_ALLOC_LIST:
	  f_array = *(float **)tabp->variable;
	  if(f_array == NULL){
	      f_array = allocate_vector(1);
	      f_array[0] = 0;
	      *(float **)tabp->variable = f_array;
	  }
	  break;
	case DOUBLE_ALLOC_LIST:
	  d_array = *(double **)tabp->variable;
	  if(d_array == NULL){
	      d_array = allocate_dvector(1);
	      d_array[0] = 0;
	      *(double **)tabp->variable = d_array;
	  }
	  break;
	case STRING_ALLOC_LIST:
	  c_array = *(char ***)tabp->variable;
	  if(c_array == NULL){
	      c_array = (char **)allocate_pvector(1);
	      c_array[0] = nn_malloc(2, "for empty string length allocation");
	      strcpy(c_array[0], "0");
	      *(char ***)tabp->variable = c_array;
	  }
	  break;
      }
}

/*******************************************************************
  char *nn_strtok(char *str, char*tokEnd);

  This is a reimplementation of strtok.

  The major difference is that rather than insert NIL characters in the
  user's string, this routine maintains a static string. The token is
  copied into this static string.

  NOTE: because the string which is returned is a static, it needs to be
  copied becore nn_strtok is called again, or it will be lost.

  Another difference is that if tokEnd is a NULL, nn_strtok will treat
  it as an empty string. strtok has a segmentation fault in that case.
********************************************************************/
char *
nn_strtok(char *str, char *tokEnd)
{
    static char *cur_str = NULL;
    static char *token = NULL;
    static int token_len = 0;
    register char *pstr, *cpy_str1, *cpy_str2;
    register int nEnds, i;

    if(str != NULL){
	cur_str = str;
    }
    if(*cur_str == '\0') return NULL;
    pstr = cur_str;
    if(tokEnd == NULL){
	while(1){
	    if(*pstr == '\0') goto GotToken;
	    pstr++;
	}
    } else {
	nEnds = strlen(tokEnd);
	while(1){
	    if(*pstr == '\0') goto GotToken;
	    for(i = 0; i < nEnds; i++) if(*pstr == tokEnd[i]) goto GotToken;
	    pstr++;
	}
    }
  GotToken:
    i = (int)(pstr - cur_str)+1;
    if(i > token_len){
	if(token_len == 0){
	    token = nn_malloc(token_len = i, "for nn_strtok");
	} else {
	    token = nn_realloc(token, token_len = i, "for nn_strtok");
	}
    }
    cpy_str1 = cur_str;
    cpy_str2 = token;
    while (cpy_str1 < pstr){
	*cpy_str2++ = *cpy_str1++;
    }
    *cpy_str2 = '\0';
    if(*pstr == '\0') cur_str = pstr;
    else cur_str = pstr+1;

    return token;
}
