/***--------------------------------------------------------------------------
* NAME:  norm_subs.c
* PURPOSE:  Normalization functions
* USAGE
*
* SEE ALSO: mlp.c
* GLOBALS:
* DESCRIPTION:
*		norm_build	allocates storage for means, sdevs (optional)
*		norm_init	initialize normalization, allocate storage
*		norm_inc	add vector to normalization total
*		norm_apply	normalize vector using means and sdevs
*		norm_mean	return pointer to mean vector
*		norm_sdev	return pointer to sdev vector
*		norm_param	returns addresses of various parameters
*		norm_update	calculates final parameters as needed
*		norm_calc_needed	return pointer to calc_needed flag
*		norm_normalization	return current normalization type
*		norm_read_paramfile	read normalization parameters
*		norm_write_paramfile	write normalization parameters
*		norm_description	make character string description
*		read_norm_file		opens, reads, closes normalization file
*
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"
#define SMALLEST 1.e-30	/* variance is not allowed to be less than SMALLEST */
#define TRUE 1
#define FALSE 0
#define MAXSTR 256
			/* definitions for normalization parameter */
#define SIMPLE 1		/* simple normalization (mean=0, variance=1) */
#define PCA 2			/* Principle Components Analysis */
#define LDA 3			/* Linear Discriminant Analysis */

		/* globals */
typedef struct {
  float **per_class_covariances;
} PCCOV;

static PCCOV *cv_class;
static float max_ratio=1.0e6;

static int total_trials;	/* current count of input vectors */
static int total_trials2;	/* count for norm_inc2 */
static int calculation_needed;	/* TRUE if normalization is not up to date */
static int calculation_needed2;	/* TRUE if normalization is not up to date */
static int ninputs;		/* size of input vector */
static int nclasses;		/* number of output classes */
static int normalization;	/* type of normalization */
		/* simple normalization parameters */
static float *means;		/* pointer to a vector of means */
static float *sdevs;		/* pointer to a vector of standard deviations */
static double *input_total;	/* sum of input values so far */
static double *input_sq_total;	/* sum of input values**2 so far */
		/* principle components parameters */
static int max_features;
static float *feature_means;
static float **grand_covariances;
static float **transform_matrix;
static float *eigenvalues;
		/* Linear Discriminant parameters */
static int *class_trials; /* current count of the number of input vectors */
static float **class_means;
static float **within_scatter_matrix;
static float **between_scatter_matrix;
static float **input_total_per_class;	/* sum of input values so far */
static float **atranspose;   /* temporay arrays for LDA calculation */
static float **lminushalf;
static float **tmp_array1;
static float **tmp_array2;
static float *tmp_inputs;

/*****************************************************************************
* NAME: norm_build ()
* DESCRIPTION:		allocates storage for normalization
*
*			ninputs = number of vector items 
*       
******************************************************************************/
void
norm_build (int ninputs_param, int nclasses_param, int normalization_param)
{
    int i;
    
    ninputs = ninputs_param;
    nclasses = nclasses_param;
    normalization = normalization_param;
    
    switch(normalization) {
        case LDA: 
	    class_trials = allocate_ivector (nclasses);
	    class_means = allocate_matrix (nclasses, ninputs);
	    input_total_per_class = allocate_matrix (nclasses, ninputs);
	    within_scatter_matrix = allocate_matrix (ninputs, ninputs);
	    between_scatter_matrix = allocate_matrix (ninputs, ninputs);
	    atranspose = allocate_matrix (ninputs, ninputs);
	    lminushalf = allocate_matrix (ninputs, ninputs);
	    tmp_array1 = allocate_matrix (ninputs, ninputs);
	    tmp_array2 = allocate_matrix (ninputs, ninputs);
	    cv_class = (PCCOV *) 
	        nn_malloc(sizeof(PCCOV)*nclasses,"for covariance structure");
	    for (i=0; i<nclasses; i++)
	        cv_class[i].per_class_covariances = 
	            allocate_matrix (ninputs, ninputs);
	    /*** fall through and allocate PCA variables also */
        case PCA: 
	    eigenvalues = allocate_vector (ninputs);
	    feature_means = allocate_vector (ninputs);
	    transform_matrix = allocate_matrix (ninputs, ninputs);
	    grand_covariances = allocate_matrix (ninputs, ninputs);
	    tmp_inputs = allocate_vector (ninputs);
	    /*** fall through and allocate SIMPLE variables also ****/
        case SIMPLE: 
            means = allocate_vector(ninputs);
            sdevs = allocate_vector(ninputs);
	    input_total = allocate_dvector(ninputs);
	    input_sq_total = allocate_dvector(ninputs);
            break;
        default:
            break;
    }
}
void norm_free(void)
{
    int i;

    switch(normalization){
      case LDA:
	deallocate_ivector(class_trials);
	deallocate_matrix(class_means);
	deallocate_matrix(input_total_per_class);
	deallocate_matrix(within_scatter_matrix);
	deallocate_matrix(between_scatter_matrix);
	deallocate_matrix(atranspose);
	deallocate_matrix(lminushalf);
	deallocate_matrix(tmp_array1);
	deallocate_matrix(tmp_array2);
	for(i = 0; i < nclasses; i++)
	  deallocate_matrix(cv_class[i].per_class_covariances);
	nn_free((char *)cv_class, "for covariance structure");
	/*** fall through and free PCA variables also ***/
      case PCA:
	deallocate_vector(eigenvalues);
	deallocate_vector(feature_means);
	deallocate_matrix(transform_matrix);
	deallocate_matrix(grand_covariances);
	deallocate_vector(tmp_inputs);
	/**** fall through and free SIMPLE variables also ***/
      case SIMPLE:
	deallocate_vector(means);
	deallocate_vector(sdevs);
	deallocate_dvector(input_total);
	deallocate_dvector(input_sq_total);
	break;
    }
}
/*****************************************************************************
* NAME: norm_setup_none(ninputs, noutputs)
* DESCRIPTION:	set global flags for norm=NONE
*       This routine should be called when no normalization file
*	will be read
******************************************************************************/
void
norm_setup_none(int n_inputs, int n_outputs)
{
    ninputs = n_inputs;
    nclasses = n_outputs;
}

/*****************************************************************************
* NAME: norm_init (total_trials_param)
* DESCRIPTION:		initialize normalization software
*
*		total_trials = total number of trials accumulated before
*       
******************************************************************************/
void 
norm_init (int total_trials_param)
{
    register int i;
#if DEBUG
    printf("norm_init: Begin\n");
    fflush(stdout);
#endif
    total_trials = total_trials_param;
    switch(normalization) {
        case LDA:
        case PCA: 
	    calculation_needed = FALSE;
	    calculation_needed2 = FALSE;
	    /* fall through and initialize the SIMPLE stuff */
        case SIMPLE: 
	    if (total_trials<1) {
	        for (i=0; i<ninputs; i++){
		    means[i] = 0.;
		    sdevs[i] = 1.;
		    input_total[i] =input_sq_total[i]= 0.;
		}
	        total_trials = 0;
	        calculation_needed = TRUE;
	    } else {
	        for (i=0; i<ninputs; i++) {
		    input_total[i] = means[i] * total_trials;
		    input_sq_total[i] = means[i] * input_total[i];
	        }
	        calculation_needed = FALSE;
	    }
            break;
        default:
            break;
    }
#if DEBUG
    printf("norm_init: End\n");fflush(stdout);
#endif
}

/*****************************************************************************
* NAME: norm_inc(inputs, class)
* DESCRIPTION:		add vector to normalization total
*
*			vector = input vector
*       
******************************************************************************/
void 
norm_inc (float *inputs, int class)
{
    register double dtmp;
    register float tmp;
    register int i, j;

    calculation_needed = TRUE;
    total_trials++;
    switch(normalization) {
        case SIMPLE: 
	    for (i=0; i<ninputs; i++) {
		dtmp = inputs[i];
		input_total[i] += dtmp;
		input_sq_total[i] += dtmp * dtmp;
	    }
            break;
        case PCA: 
	    for (i=0; i<ninputs; i++) {
		tmp = inputs[i];
		feature_means[i] += tmp;
		for (j=0; j<ninputs; j++) {
		    grand_covariances[i][j] += tmp * inputs[j];
		}
	    }
            break;
        case LDA: 
	    class_trials[class]++;
	    for (i=0; i<ninputs; i++) {
		tmp = inputs[i];
		feature_means[i] += tmp;
		input_total_per_class [class][i] += tmp;
		for (j=0; j<ninputs; j++) {
		    cv_class[class].per_class_covariances[i][j] += 
		        tmp * inputs[j];
		}
	    }
            break;
        default:
            break;
    }
}
/*****************************************************************************
* NAME: norm_inc2(inputs, class)
* DESCRIPTION:	do a second pass, collecting simple normalization
*		parameters for scaling PCA and LDA rotated vectors
*       
******************************************************************************/
void 
norm_inc2 (float *inputs, int class)
{
    register double dtmp;
    register int i;

    calculation_needed2 = TRUE;
    total_trials2++;
    /* we only need to find for the max_features valid ones */
    for (i=0; i<ninputs; i++) {
	dtmp = inputs[i];
	input_total[i] += dtmp;
	input_sq_total[i] += dtmp * dtmp;
    }
}

/************************************************************************
* NAME: norm_param(param_name)
* DESCRIPTION:
*	Takes parameter name.
*	Returns a pointer to the parameter.
*	Single value parameters are read only.
*	Function really not necessary if not interested in accessing
*	parameters (reading) from main()
*************************************************************************/
#define PNEQ(y) (0==strcmp(param_name,y))

void *
norm_param(char *param_name)
{
    void *param_address;
#if DEBUG
    printf("norm_param: param_name=%d normalization=%d\n",
	   param_name,normalization);
#endif
    /* Return the requested parameter */
    if (PNEQ("ninputs"))		param_address = (void *)&ninputs;
    else if (PNEQ("nclasses"))		param_address = (void *)&nclasses;
    else if (PNEQ("normalization"))	param_address = (void *)&normalization;
    else if (PNEQ("max_features"))	param_address = (void *)&max_features;
    else if (PNEQ("eigenvalues"))	param_address = (void *)&eigenvalues;
    else if (PNEQ("transform"))
      param_address = (void *)&transform_matrix;
    else if (PNEQ("means"))		param_address = (void *)&means;
    else if (PNEQ("sdevs"))		param_address = (void *)&sdevs;
    else {
        warning("norm_param: Requested parameter %s not defined", param_name);
	param_address = NULL;
    }
    return param_address;
} 

/*****************************************************************************
* NAME: norm_update()
* DESCRIPTION:		update normalization (calculate parameters)
*
*       
******************************************************************************/
void 
norm_update (void)
{
    register double m_tmp, var_tmp;
    register int i, j, k;
#if DEBUG
    printf("norm_update: ninputs=%d total_trials=%d\n", 
	   ninputs, total_trials);
    fflush(stdout);
#endif
    calculation_needed = FALSE;
    if (total_trials<1) {
        warning("%s%s , total_trials=%d",
	    "norm_update: norm_inc must be called at least once before",
	    " norm_update can be called",total_trials);
	total_trials = 1;
    }
    switch(normalization) {
        case SIMPLE: 
	    for (i=0; i<ninputs; i++) {
		m_tmp = input_total[i] / total_trials;
		means[i] = (float)m_tmp;
		var_tmp = input_sq_total[i] / total_trials - 
			(m_tmp * m_tmp);
		if(var_tmp < SMALLEST) {
		    warning("norm_update: variance %g for feature %d %s %g%s %g",
		        var_tmp, i, "smaller than", SMALLEST,
		        "; is being set to", SMALLEST);
		    var_tmp = SMALLEST;
		}
		sdevs[i] = (float)sqrt(var_tmp);
	    }
	    max_features = ninputs;
            break;
        case PCA: 
	    for (j=0; j<ninputs; j++)
	        feature_means[j] = feature_means[j] / total_trials;
	        
	    for (j=0; j<ninputs; j++)
	        for (i=0; i<ninputs; i++)
		    grand_covariances[i][j] = grand_covariances[i][j] / 
		        total_trials - feature_means[j] * feature_means[i];

	    make_transform_matrix (ninputs, grand_covariances,
				   transform_matrix,
				   eigenvalues, max_ratio, &max_features);
#if DEBUG
	    printf("norm_update: PCA, max_features = %d\n",max_features);
	    for (i=0; i<ninputs; i++)
	      printf("eigenvalue: %d = %f\n",i,eigenvalues[i]);
#endif
            break;
        case LDA: 
	    for (j=0; j<ninputs; j++) {
	        feature_means[j] = feature_means[j] / total_trials;
	        for (i=0; i<nclasses; i++) {
	            class_means[i][j] = 
	                input_total_per_class[i][j] / class_trials[i]; 
		}
	    }
	    for (j=0; j<ninputs; j++) {
	        for (i=0; i<ninputs; i++) {
	            for (k=0; k<nclasses; k++) {
		        within_scatter_matrix[i][j] += 
		            cv_class[k].per_class_covariances[i][j]/
			      class_trials[k] -
			    (class_means[k][i] * class_means[k][j]);
		        between_scatter_matrix[i][j] += class_trials[k] * 
		            (class_means[k][i] - feature_means[i]) *
		            (class_means[k][j] - feature_means[j]);
		    }
		    between_scatter_matrix[i][j] /= total_trials;
	        }
	    }

	    make_transform_matrix (ninputs, within_scatter_matrix, atranspose,
				  eigenvalues, max_ratio, &max_features);
	    for(i=0;i<max_features;i++)
	      lminushalf[i][i] = 1.0 / (float)sqrt(eigenvalues[i]);

	    square_matrix_multiply (ninputs, lminushalf, atranspose,
				    tmp_array1);
	    square_matrix_multiply (ninputs, tmp_array1,between_scatter_matrix,
				    tmp_array2);
	    square_matrix_transpose (ninputs, atranspose, atranspose);
	    square_matrix_multiply (ninputs, tmp_array2, atranspose,
				    tmp_array1);
	    square_matrix_multiply (ninputs, tmp_array1, lminushalf,
				    tmp_array2);
	    make_transform_matrix (ninputs, tmp_array2, tmp_array1,eigenvalues,
				   max_ratio, &max_features);
	    square_matrix_transpose (ninputs, tmp_array1, tmp_array1);
	    square_matrix_multiply (ninputs, atranspose, lminushalf,
				    tmp_array2);
	    square_matrix_multiply (ninputs, tmp_array2, tmp_array1,
				    transform_matrix);
	    square_matrix_transpose (ninputs, transform_matrix,
				     transform_matrix);
    	    square_normalize_and_sort(ninputs, transform_matrix, eigenvalues);
	    for(i=max_features;i<ninputs;i++)
	      for(j=0;j<ninputs;j++)
		transform_matrix[i][j] = 0.0;

#if DEBUG
	        printf("norm_update: LDA, max_features = %d\n",max_features);
	        for (i=0; i<ninputs; i++)
	            printf("eigenvalue: %d = %f\n",i,eigenvalues[i]);
#endif
            break; 
        default:
            break;
    }
}
/*****************************************************************************
* NAME: norm_update2()
* DESCRIPTION:	update normalization parameters for second pass, simple
*		normalization on PCA and LDA rotated vectors
*       
******************************************************************************/
void 
norm_update2(void)
{
    register double m_tmp, var_tmp;
    register int i;

    calculation_needed2 = FALSE;
    if (total_trials2<1) {
        warning("%s%s , total_trials=%d",
	    "norm_update2: norm_inc must be called at least once before",
	    " norm_update2 can be called",total_trials);
	total_trials = 1;
    }

    for (i=0; i < max_features; i++) {
	/* when LDA is used, the numbers for extra features have
	   variance 0 (the values are all 0's */
	m_tmp = input_total[i] / total_trials2;
	means[i] = (float)m_tmp;
	var_tmp = input_sq_total[i] / total_trials - 
	  (m_tmp * m_tmp);
	if(var_tmp < SMALLEST) {
	    warning("norm_update2: variance %g for feature %d %s %g%s %g",
		    var_tmp, i, "smaller than", SMALLEST,
		    "; is being set to", SMALLEST);
	    var_tmp = SMALLEST;
	}
	sdevs[i] = (float)sqrt(var_tmp);
    }
}

/*****************************************************************************
* NAME: norm_apply(inputs)
* DESCRIPTION:		normalize vector using means and sdevs
*
*			inputs = input vector
*       
******************************************************************************/
void 
norm_apply (float *inputs)
{
    register int i;

    if (calculation_needed)	norm_update();
    switch(normalization) {
        case SIMPLE: 
	    for (i=0; i<ninputs; i++)
		inputs[i] = (inputs[i] - means[i]) / sdevs[i];	
            break;
        case PCA:
        case LDA: 
#if DEBUG
	    printf("Eigenvectors:\n");
	    for (i=0; i<ninputs; i++) {
		for (j=0; j<ninputs; j++) {
		    printf(" %f",transform_matrix[i][j]);
		}
		printf("\n");
	    }
	    printf("Input: ");
	    for (i=0; i<ninputs; i++) printf(" %5.2f",inputs[i]);
	    printf("\n");
#endif
	    for (i=0; i<ninputs; i++) tmp_inputs[i] = inputs[i];
	    matrix_vector_multiply (transform_matrix, tmp_inputs, inputs, 
				    ninputs, ninputs); 
#if DEBUG
	    printf("Output:");
	    for (i=0; i<ninputs; i++) printf(" %5.2f",inputs[i]);
	    printf("\n");
#endif
	    /* also do simple on resulting vector */
	    /* should this ninputs be a smaller number? */
	    for (i=0; i < ninputs; i++)
	      inputs[i] = (inputs[i] - means[i]) / sdevs[i];
            break;
        default:
            break;
    }
}

static int pninputs, pnclasses, ptotal_trials, pnormalization, pmax_features;

/*** Parameter Table ***/
static PARAM normscalars[] = {
	{"normalization",P_INT,&pnormalization},
	{"ninputs",P_INT,&pninputs},
	{"nclasses",P_INT,&pnclasses},
	{"total_trials",P_INT,&ptotal_trials},
	{"max_features",P_INT,&pmax_features},
};
#define NSCALARS (sizeof(normscalars)/sizeof(PARAM))

static PARAM normarrays[] = {
	{"means",P_FLOAT_VECTOR,NULL,0,0},
	{"sdevs",P_FLOAT_VECTOR,NULL,0,0},
};
#define NARRAYS (sizeof(normarrays)/sizeof(PARAM))

static PARAM pcaarrays[] = {
	{"eigenvalues",P_FLOAT_VECTOR,NULL,0,0},
	{"transform_matrix",P_FLOAT_MATRIX,NULL,0,0},
};
#define NPARRAYS (sizeof(pcaarrays)/sizeof(PARAM))

/***---------------------------------------------------------------------------
* NAME:	 norm_set_normarrays
* DESCRIPTION: sets variables in normarrays
*
* BE SURE TO SET pninputs AHEAD OF TIME!!!
--------------------------------------------------------------------------***/
static void
norm_set_normarrays(void)
{
    param_setvarptr("means",
	            (int *)&(means[0]),normarrays,NARRAYS);
    param_setvarptr("sdevs",
	            (int *)&(sdevs[0]),normarrays,NARRAYS);
    param_setdim("means",pninputs,0,normarrays,NARRAYS);
    param_setdim("sdevs",pninputs,0,normarrays,NARRAYS);
}
/***---------------------------------------------------------------------------
* NAME:	 norm_set_pcaarrays()
* DESCRIPTION: sets variables in pcaarrays
*
* BE SURE TO SET pninputs AHEAD OF TIME!!!
--------------------------------------------------------------------------***/
static void
norm_set_pcaarrays(void)
{
    param_setvarptr("eigenvalues",
		    (int *)&(eigenvalues[0]),pcaarrays,NPARRAYS);
    param_setvarptr("transform_matrix",
		    (int *)&(transform_matrix[0][0]),pcaarrays,NPARRAYS);
    param_setdim("eigenvalues",pninputs,0,pcaarrays,NPARRAYS);
    param_setdim("transform_matrix",pninputs,pninputs,
		 pcaarrays,NPARRAYS);
}
/***---------------------------------------------------------------------------
* NAME:	 norm_read_paramfile()
* DESCRIPTION:
*	Takes a file pointer as input.
*	Reads norm model from parameter file.
*	Builds norm model using parameters.
*	Returns normalization indicator.
* NOTES:
*	See norm_write_paramfile().
--------------------------------------------------------------------------***/
int
norm_read_paramfile (FILE *fp, int *normalization_param)
{
	int rv;

	/* Get the norm model scalars */
	if ((rv = param_getblk(fp,normscalars,NSCALARS)) < 0) return rv;

	*normalization_param = pnormalization;
	
	/* copy scalar parameters */
	ninputs = pninputs;
	nclasses = pnclasses;
	total_trials = ptotal_trials;
	normalization = pnormalization;
	max_features = pmax_features;

	if (pnormalization > 0) {
	    /* Allocate the space needed to store the new model */
	    norm_build (pninputs, pnclasses, pnormalization);

	    /* Set ptrs and dimensions for allocated matrices */
    	    switch(normalization) {
              case PCA: 
              case LDA: 
		norm_set_pcaarrays();
	        if ((rv = param_getblk(fp,pcaarrays,NPARRAYS)) < 0) return rv;
		/* fall through and do the SIMPLE norm stuff, too */
              case SIMPLE: 
		norm_set_normarrays();
	        /* Get the norm model arrays */
	        if ((rv = param_getblk(fp,normarrays,NARRAYS)) < 0) return rv;
	        break;
              default:
                break;
	    }
	    norm_init (ptotal_trials);
	}

	return 0;
	
} /* norm_read_paramfile() */

/***---------------------------------------------------------------------------
* NAME:	 norm_write_paramfile()
* DESCRIPTION:
*	Takes a file pointer as input.
*	Writes normalization model to parameter file.
*		means, sdevs if data is normalized
--------------------------------------------------------------------------***/
int
norm_write_paramfile (FILE *fp)
{
	int rv;

	pninputs = ninputs;
	pnclasses = nclasses;
	ptotal_trials = total_trials;
	pnormalization = normalization;
	pmax_features = max_features;
	
	if ((rv = param_putblk(fp,normscalars,NSCALARS)) < 0) return rv;
    	switch(normalization) {
          case PCA: 
          case LDA: 
	    norm_set_pcaarrays();
	    if ((rv = param_putblk(fp,pcaarrays,NPARRAYS)) < 0) return rv;
	    /* fall through and do SIMPLE norm stuff */
          case SIMPLE: 
	    /* Set ptrs and dimensions for allocated matrices */
	    norm_set_normarrays();
	    if ((rv = param_putblk(fp,normarrays,NARRAYS)) < 0) return rv;
	    break;
          default:
            break;
	}

	return 0;
	
} /* norm_write_paramfile() */


/*****************************************************************************
* NAME: norm_normalization()
* DESCRIPTION:		return pointer to normalization
*
*       
******************************************************************************/
int 
norm_normalization (void)
{
    return normalization;
}


/*****************************************************************************
* NAME: norm_mean()
* DESCRIPTION:		return pointer to mean vector
*
*       
******************************************************************************/
float *
norm_mean (void)
{
    if (calculation_needed)	norm_update();
    return means;
}


/*****************************************************************************
* NAME: norm_sdev()
* DESCRIPTION:		return pointer to sdev vector
*
*       
******************************************************************************/
float *
norm_sdev (void)
{
    if (calculation_needed)	norm_update();
    return sdevs;
}


/*****************************************************************************
* NAME: norm_calc_needed()
* DESCRIPTION:		return pointer to calculation_needed flag
*
*       
******************************************************************************/
int *
norm_calc_needed (void)
{
    return &calculation_needed;
}

/************************************************************************
* NAME: make_transform_matrix (ninputs, input_matrix, output_matrix, 
*		eigenvalues, max_ratio, max_features)
* DESCRIPTION:
*		makes Principle Components normalization transformation matrix
*			uses Singular Value Decomposition function svdcmp
*		ninputs		size of square input (output) matrix
*		input_matrix	
*		max_ratio	maximum eigenvalue ratio
*	returns
*		output_matrix	Transposed eigenvector matrix
*		max_features 	(address) Number of eigenvalues above max_ratio
*		eigenvalues	vector of eigenvalues
*
*************************************************************************/
void
make_transform_matrix (int ninputs, float **input_matrix,
		       float **output_matrix, float *eigenvalues,
		       float max_ratio, int *max_features)
{
    int i;
    float min_eigenvalue;
    
#if DEBUG
    printf("make_transform_matrix:\nOriginal matrix (A):\n");
    for (i=0; i<ninputs; i++) {
	for (j=0; j<ninputs; j++) {
	    printf(" %f",input_matrix[i][j]);
	}
	printf("\n");
    }
#endif
    square_singular_value_decomposition (ninputs, input_matrix, output_matrix, 
					 eigenvalues);
    /* Calculate the minimum allowable eigenvalue  */
    min_eigenvalue = eigenvalues[0] / max_ratio;
    *max_features = ninputs;
    for (i=0; i<ninputs; i++)
      if (eigenvalues[i] < min_eigenvalue) {*max_features = i; break;}

#if DEBUG
        printf("Eigenvectors:\n");
        for (i=0; i<ninputs; i++) {
            for (j=0; j<ninputs; j++) {
                printf(" %f",output_matrix[i][j]);
            }
            printf("\n");
        }
        printf("A*ev, Eigenvalue*ev:\n");
        result = allocate_vector(ninputs);
        for (i=0; i<ninputs; i++) {
            matrix_vector_multiply (input_matrix, &output_matrix[i][0], result,
                ninputs, ninputs);
            for (j=0; j<ninputs; j++) {
                printf(" %f %f",
                    result[j],eigenvalues[i]*output_matrix[i][j]);
            }
            printf("\n");
        }
        deallocate_vector(result);
#endif
}

/*****************************************************************************
* NAME: norm_description ()
* DESCRIPTION:	Returns string describing current normalization method
*       
******************************************************************************/
char *
norm_description (int normalization_param)
{
    static char description_string[MAXSTR];
    
    strcpy(description_string, "");
    switch(normalization_param) {
        case SIMPLE: 
            strcat (description_string, "Simple");
            break;
        case PCA: 
            strcat (description_string, "PCA");
            break;
        case LDA: 
            strcat (description_string, "LDA");
            break;
        default: 
            strcat (description_string, "None");
            break;
    }
    return description_string;
}

/*****************************************************************************
* NAME: read_norm_file ()
* DESCRIPTION:	opens normalization file, reads parameters, closes it
*       
******************************************************************************/
int
read_norm_file(char *norm_filename)
{
    FILE *param_file;
    int normalization, rv;
    
    param_file = nn_fopen (norm_filename, "r",
        "for read of normalization information");
    rv = norm_read_paramfile(param_file, &normalization);
    if (rv < 0) fatal_error ("read_norm_file: read error %d %s %s",
        rv,"on normalization parameter file",norm_filename);
    nn_fclose(param_file,norm_filename, 
        "for read of normalization information");
        
    return normalization;
}

    
