/***--------------------------------------------------------------------------
* NAME:		param_util.c
*
* PURPOSE:	parameter table utility functions
*
*	Routines for general use:
*		param_putblk() -- put a block of parameters into the file
*		param_getblk() -- get a block of parameters from the file
*		param_setvarptr() -- set the value of the variable-ptr
*		param_setdim() -- set dimensions in parameter table
*		param_inc_allptrs() -- increment all variable-ptrs
*
*	Low level routines:
*		param_put() -- put a parameter into the parameter file
*		param_get() -- get a parameter from the parameter file
*		param_put_begblkstr() -- put the begin block string into the file
*		param_put_endblkstr() -- put the end block string into the file
*		param_find_begblk() -- find the beginning of the next param block
*		param_find_endblk() -- find the end of the current param block
*		param_find() -- find the named parameter in the table
*	Internal routines
*		param_setscalar() -- set a scalar parameter from a file
*		param_setarray() -- set a scalar parameter from a file
*		match_nextentry() -- match the next parameter with the table
*		stoi() -- enhanced atoi
*
* DESCRIPTION:
*	Writes and reads ascii parameter files using standard format.
*	Style based on zargs.
*
* PROBLEMS:
* 
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/param_util.c,v 11.4 2001/08/23 20:14:46 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/param_util.c,v $
* CREATED:	13-Aug-90 rpl & es                      MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------*/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"

static void param_setscalar(PARAM *paramp, char *value);
static void param_setarray(PARAM *paramp, int dim1, int dim2, FILE *fileptr);
static int match_nextentry(char *line, char *str1, char *estr, FILE *filep);
static int stoi(char *str);

static char *begblkstr = BEG_BLK_STR;
static char *endblkstr = END_BLK_STR;
static char *begvecstr = BEG_VEC_STR;
static char *endvecstr = END_VEC_STR;
static char *begmatstr = BEG_MAT_STR;
static char *endmatstr = END_MAT_STR;



/***---------------------------------------------------------------------------
* NAME:		param_putblk
* PURPOSE:	Put all parameters into a parameter block beginning
*		with <BEGIN_PARAM> and ending with <END_PARAM>.
* NOTES:	
---------------------------------------------------------------------------***/
int param_putblk(FILE *fileptr,/* file pointer */
		 PARAM *tabp,/* ptr to parameter table */
		 int nparam/* number of parameters */)
{
	PARAM *paramp;
	int i;

	param_put_begblkstr(fileptr);
	paramp = tabp;
	for (i=0; i<nparam; i++) {
		param_put(fileptr,paramp);
		paramp++;
	}
	param_put_endblkstr(fileptr);
	return i;

} /* param_putblk() */


/***---------------------------------------------------------------------------
* NAME:		param_getblk
* PURPOSE:	Get parameters from a parameter block.
* NOTES:	A parameter block begins with <BEGIN_PARAM> and ends 
*		with <END_PARAM>.
*		Return the number of parameters in the parameter block if it
*		exists or EOF on end-of-file.
*		Leave file ptr at line following <END_PARAM>.
--------------------------------------------------------------------------***/
int param_getblk(FILE *fileptr,/* file pointer */
		 PARAM *tabp,/* pointer to parameter table */
		 int nparam/* number of parameters */)
{
	int np;		/* number of parameters read from block */
	static char routine[] = "param_getblk";

	if (param_find_begblk(fileptr) == EOF) {
		return EOF;
	}

	np = 0;
	for(;;) {
		switch(param_get(fileptr,tabp,nparam)) {
			case -1:	/* error */
				printf("%s: Can't finish\n",routine);
				exit(1);
			case 0:		/* end of block */
				return np;
			case 1:		/* empty object or param not in table */
				break;
			case 2:		/* parameter read from file */
				np++;
				break;
			default:
				printf("%s: param_get returned unknowned value\n",
					routine);
				exit(1);
		} /* switch */
	}

} /* param_getblk() */


/***---------------------------------------------------------------------------
* NAME:		param_setvarptr
* PURPOSE:	Set variable-ptr field in parameter table.
* NOTES:
---------------------------------------------------------------------------***/
void
param_setvarptr(char *name,/* identifying name in parameter structure */
		int *varp,/* value to be inserted in variable pointer field */
		PARAM *tabp,/* pointer to parameter table */
		int nparam/* number of parameters in table */)
{
	PARAM *pp;
	static char routine[] = "param_setvarptr";

        pp = param_find(name,tabp,nparam);
	if (pp == NULL) {
		printf("%s: Can't find parameter %s\n",routine,name);
		exit(1);
	}
	else {
        	pp->variable = varp;
	}

} /* param_setvarptr() */


/***---------------------------------------------------------------------------
* NAME:		param_setdim
* PURPOSE:	Set dimensions in parameter table.
* NOTES:	Use 0 for dim2 if vector.
---------------------------------------------------------------------------***/
void
param_setdim(char *name,/* identifying name in parameter structure */
	     int dim1,/* value to be inserted in dimension #1 field */
	     int dim2,/* value to be inserted in dimension #2 field */
	     PARAM *tabp,/* pointer to parameter table */
	     int nparam/* number of parameters in table */)
{
	PARAM *pp;
	static char routine[] = "param_setdim";

        pp = param_find(name,tabp,nparam);
	if (pp == NULL) {
		printf("%s: Can't find parameter %s\n",routine,name);
		exit(1);
	}
	else {
        	pp->n_dim1 = dim1;
        	pp->n_dim2 = dim2;
	}

} /* param_setdim() */


/***---------------------------------------------------------------------------
* NAME:		param_inc_allptrs
* PURPOSE:	Increment all variable pointer fields in parameter table.
* NOTES:
---------------------------------------------------------------------------***/
void param_inc_allptrs(int pincr,/* increment for all variable pointer fields */
		       PARAM *tabp,/* pointer to parameter table */
		       int nparam/* number of parameters in table */)
{
	int i;

	for (i=0; i<nparam; i++) {
        	tabp->variable += pincr;
		tabp++;
	}

} /* param_inc_allptrs() */


/***--------------------------------------------------------------------------
* NAME:		param_put
* PURPOSE:	Put a parameter into a parameter file.
* NOTES:	
--------------------------------------------------------------------------***/
void param_put(FILE *fileptr,/* fileptr */
	      PARAM *paramp/* ptr to parameter in table */)
{
	int i,j,k;
	int dim1,dim2;		/* array dimensions */
	int noprint;		/* noprint flag */
	int nitems;
	char *cptr;
	int *iptr;
	float *fptr;
	double *dptr;

	switch(paramp->type) {
		case P_INT:
			iptr = paramp->variable;
			fprintf(fileptr,"%-15s%d\n",paramp->label,*iptr);
			break;
		case P_INT_VECTOR:
			iptr = paramp->variable;
			dim1 = paramp->n_dim1;
			nitems = dim1;
			fprintf(fileptr,"%s\n",begvecstr);
			fprintf(fileptr,"%s %d\n",paramp->label,dim1);
			noprint = 0;
			for (j=0; j<2; j++) {
				if (j < dim1)
					fprintf(fileptr,"%d    %d\n",j,*iptr++);
			}
			for (j=2; j < nitems-1; j++) {
				if ((*iptr == *(iptr-1)) && (*iptr == *(iptr-2))) {
					if (noprint == 0) {
						fprintf(fileptr,"...\n");
						noprint = 1;
					}
				} else {
					fprintf(fileptr,"%d    %d\n",j,*iptr);
					noprint = 0;
				}
				iptr++;
			}
			if (j < nitems)
				fprintf(fileptr,"%d    %d\n",j,*iptr);
			fprintf(fileptr,"%s\n",endvecstr);
			break;
		case P_INT_MATRIX:
			iptr = paramp->variable;
			dim1 = paramp->n_dim1;
			dim2 = paramp->n_dim2;
			nitems = dim1*dim2;
			fprintf(fileptr,"%s\n",begmatstr);
			fprintf(fileptr,"%s %d %d\n",paramp->label,dim1,dim2);
			noprint = 0;
			j = k = 0;
			for (i=0; i<2; i++) {
				if (i < nitems){
					fprintf(fileptr,"%d %d    %d\n",
						j,k,*iptr++);
					if ((k = ++k % dim2) == 0) j++;
				}
			}
			for (i=2; i<nitems-1; i++) {
				if ((*iptr == *(iptr-1)) && (*iptr == *(iptr-2))) {
					if (noprint == 0) {
						fprintf(fileptr,"...\n");
						noprint = 1;
					}
				} else {
					fprintf(fileptr,"%d %d    %d\n",
								j,k,*iptr);
					noprint = 0;
				}
				if ((k = ++k % dim2) == 0) j++;
				iptr++;
			}
			if (i < nitems)
				fprintf(fileptr,"%d %d    %d\n",j,k,*iptr);
			fprintf(fileptr,"%s\n",endmatstr);
			break;
		case P_FLOAT:
			fptr = (float *)paramp->variable;
			fprintf(fileptr,"%-15s%-.6g\n",paramp->label,*fptr);
			break;
		case P_FLOAT_VECTOR:
			fptr = (float *)paramp->variable;
			dim1 = paramp->n_dim1;
			nitems = dim1;
			fprintf(fileptr,"%s\n",begvecstr);
			fprintf(fileptr,"%s %d\n",paramp->label,dim1);
			noprint = 0;
			for (j=0; j<2; j++) {
				if (j < dim1)
					fprintf(fileptr,"%d    %-.6g\n",j,*fptr++);
			}
			for (j=2; j < nitems-1; j++) {
				if ((*fptr == *(fptr-1)) && (*fptr == *(fptr-2))) {
					if (noprint == 0) {
						fprintf(fileptr,"...\n");
						noprint = 1;
					}
				} else {
					fprintf(fileptr,"%d    %-.6g\n",j,*fptr);
					noprint = 0;
				}
				fptr++;
			}
			if (j < nitems)
				fprintf(fileptr,"%d    %-.6g\n",j,*fptr);
			fprintf(fileptr,"%s\n",endvecstr);
			break;
		case P_FLOAT_MATRIX:
			fptr = (float *)paramp->variable;
			dim1 = paramp->n_dim1;
			dim2 = paramp->n_dim2;
			nitems = dim1*dim2;
			fprintf(fileptr,"%s\n",begmatstr);
			fprintf(fileptr,"%s %d %d\n",paramp->label,dim1,dim2);
			noprint = 0;
			j = k = 0;
			for (i=0; i<2; i++) {
				if (i < nitems){
					fprintf(fileptr,"%d %d    %-.6g\n",
						j,k,*fptr++);
					if ((k = ++k % dim2) == 0) j++;
				}
			}
			for (i=2; i<nitems-1; i++) {
				if ((*fptr == *(fptr-1)) && (*fptr == *(fptr-2))) {
					if (noprint == 0) {
						fprintf(fileptr,"...\n");
						noprint = 1;
					}
				} else {
					fprintf(fileptr,"%d %d    %-.6g\n",
						j,k,*fptr);
					noprint = 0;
				}
				if ((k = ++k % dim2) == 0) j++;
				fptr++;
			}
			if (i < nitems)
				fprintf(fileptr,"%d %d    %-.6g\n",j,k,*fptr++);
			fprintf(fileptr,"%s\n",endmatstr);
			break;
		case P_DOUBLE:
			dptr = (double *)paramp->variable;
			fprintf(fileptr,"%-15s%-.16g\n",paramp->label,*dptr);
			break;
		case P_DOUBLE_VECTOR:
			dptr = (double *)paramp->variable;
			dim1 = paramp->n_dim1;
			nitems = dim1;
			fprintf(fileptr,"%s\n",begvecstr);
			fprintf(fileptr,"%s %d\n",paramp->label,dim1);
			noprint = 0;
			for (j=0; j<2; j++) {
				if (j < dim1)
					fprintf(fileptr,"%d    %-.16g\n",j,*dptr++);
			}
			for (j=2; j < nitems-1; j++) {
				if ((*dptr == *(dptr-1)) && (*dptr == *(dptr-2))) {
					if (noprint == 0) {
						fprintf(fileptr,"...\n");
						noprint = 1;
					}
				} else {
					fprintf(fileptr,"%d    %-.16g\n",j,*dptr);
					noprint = 0;
				}
				fptr++;
			}
			if (j < nitems)
				fprintf(fileptr,"%d    %-.16g\n",j,*dptr);
			fprintf(fileptr,"%s\n",endvecstr);
			break;
		case P_DOUBLE_MATRIX:
			dptr = (double *)paramp->variable;
			dim1 = paramp->n_dim1;
			dim2 = paramp->n_dim2;
			nitems = dim1*dim2;
			fprintf(fileptr,"%s\n",begmatstr);
			fprintf(fileptr,"%s %d %d\n",paramp->label,dim1,dim2);
			noprint = 0;
			j = k = 0;
			for (i=0; i<2; i++) {
				if (i < nitems){
					fprintf(fileptr,"%d %d    %-.16g\n",
						j,k,*dptr++);
					if ((k = ++k % dim2) == 0) j++;
				}
			}
			for (i=2; i<nitems-1; i++) {
				if ((*dptr == *(dptr-1)) && (*dptr == *(dptr-2))) {
					if (noprint == 0) {
						fprintf(fileptr,"...\n");
						noprint = 1;
					}
				} else {
					fprintf(fileptr,"%d %d    %-.16g\n",
						j,k,*dptr);
					noprint = 0;
				}
				if ((k = ++k % dim2) == 0) j++;
				dptr++;
			}
			if (i < nitems)
				fprintf(fileptr,"%d %d    %-.16g\n",j,k,*dptr++);
			fprintf(fileptr,"%s\n",endmatstr);
			break;
		case P_STRING:
			cptr = (char *)paramp->variable;
			fprintf(fileptr,"%-15s%s\n",paramp->label,cptr);
			break;
		case P_STRING_VECTOR:
			cptr = (char *)paramp->variable;
			fprintf(fileptr,"%s\n",begvecstr);
			fprintf(fileptr,"%s %d %d\n",paramp->label,
				paramp->n_dim1,paramp->n_dim2);
			for (j=0; j < paramp->n_dim1; j++) {
				fprintf(fileptr,"%d    %s\n",j,cptr);
				cptr += paramp->n_dim2;
			}
			fprintf(fileptr,"%s\n",endvecstr);
			break;
		case P_FLAG:
			fprintf(fileptr,"%-15s%-5s\n",paramp->label,
				*(paramp->variable) ? FLAG_TRUE:FLAG_FALSE);
			break;
		default:
			printf("Type %d unknown, parameter %s\n",
				paramp->type,paramp->label);
	} /* switch */

} /* param_put() */


/***---------------------------------------------------------------------------
* NAME:		param_get
* PURPOSE:	Get a parameter from a file and fill its values.
* NOTES:	Expects file pointer positioned at {BEGIN_VECTOR,
*		{BEGIN_MATRIX, or scalar.
*		Scan the table for the corresponding parameter label.
*		If there is a match, set the parameter values from the file.
*		If not, skip to the next object.
*		Return 1 for a match, 0 for a non-match, and -1 for an error.
---------------------------------------------------------------------------***/
int param_get(FILE *fileptr,/* file pointer */
	      PARAM *tabp,/* pointer to parameter table */
	      int nparam/* number of parameters in table */)
{
	int nf;
	int dim1, dim2;
	PARAM *paramp;		/* ptr to parameter in table */
	char line[MAXLINE];
	char *lptr;
	char label[MAXFIELD], lval1[MAXFIELD];
	static char routine[] = "param_get";


	/* get next line */
	lptr = fgets(line,MAXLINE,fileptr);

	/* unexpected EOF */
	if (lptr == NULL) {
		printf("%s: Unexpected EOF, expecting %s\n",routine,endblkstr);
		return -1;
	}

	/* decode line */
	/* allowed: {BEGIN*, scalar_name scalar_value, or <END_PARAM> */
	nf = sscanf(line,"%s%s",label,lval1);
	if ((nf < 1) || (nf > 2)) { 		/* error */
		printf("%s: Bad line, %s\n",routine,line);
		return -1;
	}
	if (strcmp(label,endblkstr) == 0) {	/* end of block */
		return 0;
	}

	if (nf == 2) {	/* scalar */
		paramp = param_find(label,tabp,nparam);
		if (paramp == NULL) {	/* scalar not in table */
			return 1;
		}
		else {	/* get scalar */
			param_setscalar(paramp,lval1);
			return 2;	/* successful read */
		}
	} /* scalar */

	if (nf == 1) {	/* vector, matrix, or error */
		/* error: not "{BEGIN*" */
		if ((strcmp(label,begvecstr) != 0) && 
			(strcmp(label,begmatstr) != 0)) {
			printf("%s: Can't find array begin string\n",routine);
			printf("\tLine: %s\n",line);
			return -1;
		}	
		/* vector */
		if (strcmp(label,begvecstr) == 0) {
			/* get next line: "vec_name dim", end vector, or error */
			lptr = fgets(line,MAXLINE,fileptr);
			nf = sscanf(line,"%s%d%d",label,&dim1,&dim2);
			/* unexpected EOF */
			if (line == NULL) {
				printf("%s: Unexpected EOF, ",routine);
				printf("expecting %s\n",endvecstr);
				return -1;
			}
			/* unexpected end of block */
			if ((strcmp(label,endblkstr) == 0)) {
				printf("%s: Unexpected %s, ",routine,endblkstr);
				printf("expecting %s\n",endvecstr);
				return -1;
			}
			/* empty vector */
			if (strcmp(label,endvecstr) == 0) {
				return 1;
			}
			/* find vector in table */
			if ((paramp=param_find(label,tabp,nparam)) == NULL) {
				/* skip to next object */
				while(match_nextentry(line,lval1,endvecstr,fileptr)
					!= 0);
				return 1;	/* vector not in table */
			}
			/* check syntax of label line */
			if (((paramp->type == P_STRING_VECTOR) && (nf != 3)) ||
				((paramp->type != P_STRING_VECTOR) && (nf != 2)))
			{
				printf("%s: Bad line, %s\n",routine,line);
				return -1;
			}
			/* check vector dimension */
			if (dim1 > paramp->n_dim1) {
				printf("%s: Dimension of \"%s\": ",routine,label);
				printf("file (%d) > table (%d)\n",
					dim1,paramp->n_dim1);
				return -1;
			}
			if ((paramp->type == P_STRING_VECTOR) &&
					(dim2 != paramp->n_dim2))
			{
				printf("%s: Dimension 2 of \"%s\": ",
					routine,label);
				printf("file (%d) != table (%d)\n",
					dim2,paramp->n_dim2);
				return -1;
			}
			/* copy data */
			if (paramp->type != P_STRING_VECTOR)
				dim2 = 1;
			param_setarray(paramp,dim1,dim2,fileptr);
			return 2;	/* successful read */
		} /* vector */

		/* matrix */
		if (strcmp(label,begmatstr) == 0) {
			/* get next line: "mat_name dim1 dim2",
				end matrix, or error */
			lptr = fgets(line,MAXLINE,fileptr);
			nf = sscanf(line,"%s%d%d",label,&dim1,&dim2);
			/* unexpected EOF */
			if (line == NULL) {
				printf("%s: Unexpected EOF, ",routine);
				printf("expecting %s\n",endvecstr);
				return -1;
			}
			/* unexpected end of block */
			if (strcmp(label,endblkstr) == 0) {
				printf("%s: Unexpected %s, ",routine,endblkstr);
				printf("expecting %s\n",endmatstr);
				return -1;
			}
			/* empty matrix */
			if (strcmp(label,endmatstr) == 0) {
				return 1;
			}
			/* check syntax of label line */
			if (nf != 3) {
				printf("%s: Bad line, %s\n",routine,line);
				return -1;
			}
			/* find matrix in table */
			if ((paramp=param_find(label,tabp,nparam)) == NULL) {
				/* skip to next object */
				while(match_nextentry(line,lval1,endmatstr,fileptr)
					!= 0);
				return 1;	/* matrix not in table */
			}
			/* check matrix dimensions */
			if (dim1 > paramp->n_dim1) {
				printf("%s: Dimension 1 of \"%s\": ",
							routine,label);
				printf("file (%d) > table (%d)\n",
					dim1,paramp->n_dim1);
				return -1;
			}
			if (dim2 != paramp->n_dim2) {
				printf("%s: Dimension 2 of \"%s\": ",
							routine,label);
				printf("file (%d) != table (%d)\n",
					dim2,paramp->n_dim2);
				return -1;
			}
			/* copy data */
			param_setarray(paramp,dim1,dim2,fileptr);
			return 2;	/* successful read */
		} /* matrix */
	} /* array */
    return -1;  /* if we get here, it is a mistake */
} /* param_get() */


/***---------------------------------------------------------------------------
* NAME:		param_put_begblkstr
* PURPOSE:	Label the beginning of a parameter file.
* NOTES:	
---------------------------------------------------------------------------***/
void param_put_begblkstr(FILE *fp/* fileptr */)
{
	fprintf(fp,"%s\n",begblkstr);

} /* param_put_begblkstr() */


/***---------------------------------------------------------------------------
* NAME:		param_put_endblkstr
* PURPOSE:	Label the end of a parameter file.
* NOTES:	
---------------------------------------------------------------------------***/
void param_put_endblkstr(FILE *fp/* fileptr */)
{
	fprintf(fp,"%s\n",endblkstr);

} /* param_put_endblkstr() */


/***---------------------------------------------------------------------------
* NAME:		param_find_begblk
* PURPOSE:	Find the beginning of a parameter block.
* NOTES:	Leave file ptr positioned at line following <BEGIN_PARAM>.
*		Return 0 if <BEGIN_PARAM> was found and	EOF at end-of-file.
*		Unfortunately, EOF is also returned if there is a file read
*		error and there is no way of distinguishing the two.
--------------------------------------------------------------------------***/
int param_find_begblk(FILE *fileptr/* file pointer */)
{
	int rval;
	char line[MAXLINE];
	char label[MAXFIELD];

	/* find beginning of parameter block */
	for(;;) {
		if (fgets(line,MAXLINE,fileptr) == NULL) {
			rval = EOF;
			break;
		}
		/* match line to beginning of parameter file */
		else {
			/* ignore blank line (learned this the hard way) */
			if (sscanf(line,"%s",label) == EOF)
				continue;
			if ((rval=strcmp(label,begblkstr)) == 0) {
				break;	/* found beginning */
			}
		}
	} /* next line */
	return rval;

} /* param_find_begblk() */


/***---------------------------------------------------------------------------
* NAME:		param_find_endblk
* PURPOSE:	Find the end of the current parameter block.
* NOTES:	
---------------------------------------------------------------------------***/
int param_find_endblk(FILE *fileptr/* file pointer */)
{
	int rval;
	char line[MAXLINE];
	char label[MAXFIELD];
	static char routine[] = "param_find_endblk";

	/* find end of parameter block */
		if (fgets(line,MAXLINE,fileptr) == NULL) {
			printf("%s: Can't find %s line\n",routine,endblkstr);
			rval = -1;
		}
		/* match line to end of parameter block */
		else {
			sscanf(line,"%s",label);
			if (strcmp(label,endblkstr) == 0) {
				rval = 0;	/* found end */
			}
			else {
				printf("%s: Next line not %s\n",routine,endblkstr);
				rval = -1;
			}
		}
	return rval;

} /* param_find_endblk() */


/***---------------------------------------------------------------------------
* NAME:		param_find
* PURPOSE:	Find a parameter in the parameter table.
* NOTES:	Return pointer to parameter table entry corresponding to name
*		or NULL if name isn't in table.
---------------------------------------------------------------------------***/
PARAM *param_find(char *name,/* name of parameter (label) */
		  PARAM *tabp,/* pointer to parameter table */
		  int nparam/* number of parameters in table */)
{
	for (; --nparam >= 0; tabp++)
		if (strcmp(tabp->label,name) == 0)
			return tabp;
	return (NULL);
} /* param_find() */


/***---------------------------------------------------------------------------
* NAME:		param_setscalar
* PURPOSE:	Set a scalar parameter with a value.
* NOTES:	
---------------------------------------------------------------------------***/
static void param_setscalar(PARAM *paramp,/* ptr to parameter in table */
			   char *value/* value points at the desired value */)
{
	static char routine[] = "param_setscalar";

	switch(paramp->type)
		{
		case P_INT:
			*(paramp->variable) = stoi(value);
			break;
		case P_FLOAT:
			*((float *)(paramp->variable)) = atof(value);
			break;
		case P_DOUBLE:
			*((double *)(paramp->variable)) = atof(value);
			break;
		case P_FLAG:
			if (strcmp(value,FLAG_TRUE) == 0)
				*(paramp->variable) = 1;
			else if (strcmp(value,FLAG_FALSE) == 0)
				*(paramp->variable) = 0;
			else {
				printf("%s: Unknown flag value %s for %s\n",
						routine,value,paramp->label);
				exit(1);
			}
			break;
		case P_CHAR:
			*(paramp->variable) = value[0];
			break;
		case P_STRING:
			strcpy((char *)(paramp->variable),value);
			break;
		default:
			printf("%s: Bad argument type %d\n",routine,paramp->type);
			exit(1);
		}
} /* param_setscalar() */


/***---------------------------------------------------------------------------
* NAME:		param_setarray
* PURPOSE:	Set an array parameter with values.
* NOTES:	
---------------------------------------------------------------------------***/
static void param_setarray(PARAM *paramp,/* ptr to parameter in table */
			  int dim1,/* size of dimension 1 */
			  int dim2,/* size of dimension 2 */
			  FILE *fileptr/* data file ptr */)
{
	char *cptr;
	int *iptr,*ip_fill = NULL,*ip_final;
	float *fptr,*fp_fill = NULL,*fp_final;
	double *dptr;
	int offset = 0;
	int x1,x2;
	char ystr[MAXFIELD];
	int yi,yyi = 0;
	float yf,yyf = 0.;
	double ydouble;
	int nitems;
	int range;
	char *estr;
	char line[MAXLINE], label[MAXFIELD], str1[MAXFIELD];
	int partype;
	int fillin;
	int j;
	static char routine[] = "param_setarray";

	partype = paramp->type;
	switch(partype)
	{
		case P_STRING_VECTOR:
			cptr = (char *)paramp->variable;
			nitems = dim1;
			for (j=0; j<nitems; j++) {
				fgets(line,MAXLINE,fileptr);
				sscanf(line,"%d%s",&x1,ystr);
				strcpy(cptr+x1*dim2,ystr);
			}
			fgets(line,MAXLINE,fileptr);
			sscanf(line,"%s",label);
			if (strcmp(label,endvecstr) != 0) {
				printf("%s: Expecting %s, line=%s\n",
					routine,endvecstr,line);
				exit(1);
			}
			break;

		case P_INT_VECTOR:
		case P_INT_MATRIX:
			iptr = paramp->variable;
			if (partype == P_INT_VECTOR) {
				dim2 = 1;
				x2 = 0;
				estr = endvecstr;
			}
			else	estr = endmatstr;
			range = dim1*dim2 - 1;
			fillin = 0;
			/* these are used for continuations */
			yi = yf = offset = 0;
			while(match_nextentry(line,str1,estr,fileptr) != 0) {
				if (strcmp(str1,"...") == 0) {
					yyi = yi;
					ip_fill = &iptr[offset+1];
					fillin = 1;
					continue;
				} else {
					if (partype == P_INT_VECTOR)
						sscanf(line,"%d%d",&x1,&yi);
					else
						sscanf(line,"%d%d%d",&x1,&x2,&yi);
					if ((offset=x1*dim2+x2) > range) {
						printf("%s: %s indices ",
							routine,paramp->label);
						printf("(%d,%d) out of range\n",
							x1,x2);
						exit(1);
					}
					iptr[offset] = yi;
				}
				if (fillin == 1) {
					ip_final = &iptr[offset];
					while (ip_fill != ip_final)
						*ip_fill++ = yyi;
					fillin = 0;
				}
			}
			break;
	
		case P_FLOAT_VECTOR:
		case P_FLOAT_MATRIX:
			fptr = (float *)paramp->variable;
			if (partype == P_FLOAT_VECTOR) {
				dim2 = 1;
				x2 = 0;
				estr = endvecstr;
			}
			else	estr = endmatstr;
			range = dim1*dim2 - 1;
			fillin = 0;
			while(match_nextentry(line,str1,estr,fileptr) != 0) {
				if (strcmp(str1,"...") == 0) {	/* fillin symbol */					yyf = yf;
					fp_fill = &fptr[offset+1];
					fillin = 1;
					continue;
				} else {		/* new value */
					if (partype == P_FLOAT_VECTOR)
						sscanf(line,"%d%f",&x1,&yf);
					else
						sscanf(line,"%d%d%f",&x1,&x2,&yf);
					if ((offset=x1*dim2+x2) > range) {
						printf("%s: %s indices ",
							routine,paramp->label);
						printf("(%d,%d) out of range\n",
							x1,x2);
						exit(1);
					}
					fptr[offset] = yf;
				}
				if (fillin == 1) {	/* fill in repeats */
					fp_final = &fptr[offset];
					while (fp_fill != fp_final)
						*fp_fill++ = yyf;
					fillin = 0;
				}
			}
			break;

		case P_DOUBLE_VECTOR:
		case P_DOUBLE_MATRIX:
			if (partype == P_DOUBLE_VECTOR) estr = endvecstr;
			else	estr = endmatstr;
			dptr = (double *)paramp->variable;
			nitems = dim1*dim2;
			for (j=0; j<nitems; j++) {
				fgets(line,MAXLINE,fileptr);
				if (partype == P_DOUBLE_VECTOR) {
					sscanf(line,"%d%lf",&x1,&ydouble);
					dim2 = 1;
					x2 = 0;
				}
				else
					sscanf(line,"%d%d%lf",&x1,&x2,&ydouble);
				dptr[x1*dim2+x2] = ydouble;
			}
			fgets(line,MAXLINE,fileptr);
			sscanf(line,"%s",label);
			if (strcmp(label,estr) != 0) {
				printf("%s: Expecting %s, line=%s\n",
					routine,estr,line);
				exit(1);
			}
			break;

		default:
			printf("%s: No parameter type %d\n",routine,paramp->type);
		}
} /* param_setarray() */


/***---------------------------------------------------------------------------
* NAME:		match_nextentry
* PURPOSE:	Match first string of next line of parameter file.
* NOTES:	Return 0 if strings match, -1 otherwise.
*		Line returned in line.
*		First string returned in str1.
*		File ptr points to next line in file.
---------------------------------------------------------------------------***/
static int match_nextentry(char *line, char *str1, char *estr,
			   FILE *filep/* data file ptr */)
{
	fgets(line,MAXLINE,filep);
	sscanf(line,"%s",str1);
	if (strcmp(str1,estr) == 0)
		return 0;
	else	return -1;
} /* match_nextentry() */


/***---------------------------------------------------------------------------
* NAME:		stoi
* PURPOSE:	Convert ascii string to integer (enhances atoi)
* NOTES:	If string starts with 0x it is interpreted as hex, else
*		if it starts with 0 it is octal, else it is decimal.
*		Conversion stops on first character not a digit in the
*		indicated radix.
---------------------------------------------------------------------------***/
#define islower(c) ('a'<= (c) && (c) <= 'z')
#define toupper(c) ( islower(c) ? (c) - ('a' - 'A') : (c) )

static int stoi(char *str)
{
	int	num=0;
	int	sign=1;

	while(*str == ' ' || *str == '\t' || *str == '\n')
		str++;

	if (*str == '-')
		{
		sign = -1;
		str++;
		}
	if (*str == '0')
		{
		++str;
		if (*str == 'x' || *str == 'X')
			{
			str++;
			while( ('0' <= *str && *str <= '9') ||
				   ('a' <= *str && *str <= 'f') ||
				   ('A' <= *str && *str <= 'F')   )
				{
				num *= 16;
				num += ('0' <= *str && *str <= '9') ?
					*str - '0': toupper(*str) - 'A' + 10;
				str++;
				}
			}
		else
			{
			while ('0' <= *str && *str <= '7')
				{
				num *= 8;
				num += *str++ - '0';
				}
			}
		}
	else
		{
		while ('0' <= *str && *str <= '9')
			{
			num *= 10;
			num += *str++ - '0';
			}
		}
	return (num*sign);
} /* stoi() */
