/***--------------------------------------------------------------------------
* NAME:		priors.c
* PURPOSE:	routines for selecting training data according to requested
*		apriori probabilities, rather than using it as is
*
* SEE ALSO:
* GLOBALS:
*
* DESCRIPTION:
*	General purpose functions
*		init_priors_classes()
*		get_priors_class()
*		get_priors_index()
*		free_priors_classes()
*
* BUGS AND LIMITATIONS:
*	only one set of priors can be active at a time
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/priors.c,v 11.3 2001/08/23 20:14:47 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/priors.c,v $
* CREATED: 5-Mar-92 lck                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------
* Copyright 1992 MIT Lincoln Laboratory.
* This software is not for release and should neither be copied
* nor distributed without explicit permission of MIT Lincoln Laboratory.
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include "nnlib.h"



static int *positions = NULL;
static int **indices = NULL;
static int *n_per_class = NULL;
static int priors_randomize = 0;
static int priors_nclasses = 0;
static float *norm_priors = NULL;

/************************************************************************
 * init_priors_data(classes, nclasses, npatterns, random_flag, priors_list)
 *	classes -- pointer to an array npatterns long
 *		array holds class number for npatterns input data points
 *		class numbers must be in the range [0..nclasses-1]
 *	nclasses -- number of classes represented in classes
 *	random_flag -- whether we should randomize the inputs when we later
 *		present them or show them in the order given (the order is
 *		relative to points in the particular class)
 *	priors_list -- pointer to an array of probabilities for each class
 ***********************************************************************/
void
init_priors_data(int *classes, int nclasses, int npatterns,
		 int random_flag, float *priors_list)
{
    int i, class;
    double priors_total;

    priors_randomize = random_flag;
    priors_nclasses = nclasses;
    positions = allocate_ivector(nclasses);
    indices = (int **)allocate_pvector(nclasses);
    n_per_class = allocate_ivector(nclasses);
    norm_priors = allocate_vector(nclasses);

    /* count up the number of patterns per class */
    for(i = 0; i < npatterns; i++){
      if (classes[i] >= 0 && classes[i] < nclasses)
	n_per_class[classes[i]]++;
    }

    /* allocate and intialize indices for each class */
    for(class = 0; class < nclasses; class++){
	indices[class] = allocate_ivector(n_per_class[class]);
    }
    for(i = 0; i < npatterns; i++){
	/* for each class, record where in the overall class list each point
	   in that class falls */
	class = classes[i];
	if(class >= nclasses)
	  fatal_error("init_priors_data: pattern with class %d, nclasses = %d",
		      class, nclasses);
	indices[class][positions[class]] = i;
	positions[class]++;
    }
    /* reset the positions array and randomize the indices */
    for(class = 0; class < nclasses; class++){
	positions[class] = 0;
	if(priors_randomize){
	    if(n_per_class[class] == 0)
	      warning("init_priors_data: there is no data for class %d",
		      class);
	    else
	      randomize_index_list(indices[class], n_per_class[class]);
	}
    }

    /* make priors sum to one.  store a cumulative sum of the priors
       for use with a floating point random number routine */
    priors_total = 0.;
    for(class = 0; class < nclasses; class++)
      priors_total += priors_list[class];
    for(class = 0; class < nclasses; class++)
      norm_priors[class] = priors_list[class]/priors_total;
    for(class = 1; class < nclasses; class++)
      norm_priors[class] = norm_priors[class] + norm_priors[class-1];
}

/* returns a class number using the probabilities given to init_priors_data
   in the array priors_list
   the number returned is in the range [0..nclasses] where nclasses is
   the value given to init_priors_data
*/
int
get_priors_class(void)
{
    int class;
    float r;

    /* select the class */
    r = random_uniform();
    for(class = 0; class < priors_nclasses; class++)
      if(r < norm_priors[class]) return(class);

    /* fell off the end.  Inaccuracy in the sums in norm_priors */
    return(priors_nclasses - 1);
}

/* returns an index into the array of classes given in init_priors_data for
   a data point in the class class.
   returns -1 if there are no points for that class
*/
int
get_priors_index(int class)
{
    int rv;

    /* check if we have data form this class */
    if(n_per_class[class] == 0) return -1;

    /* check if we have data left in this pass through this class */
    if(positions[class] >= n_per_class[class]){
	/* at the end of this classes data */
	if(priors_randomize)
	  /* n_per_class[class] != 0 */
	  randomize_index_list(indices[class], n_per_class[class]);
	positions[class] = 0;
    }

    /* get the index of the next point for this class */
    rv = indices[class][positions[class]];
    /* update our position in the index array */
    positions[class]++;
    return rv;
}


/************************************************************************
 * free_priors_data()
 *	free arrays associated with priors data selection
 ***********************************************************************/
void
free_priors_data(void)
{
    int class;

    deallocate_ivector(positions);
    positions = NULL;
    deallocate_ivector(n_per_class);
    n_per_class = NULL;
    deallocate_vector(norm_priors);
    norm_priors = NULL;
    for(class = 0; class < priors_nclasses; class++){
	deallocate_ivector(indices[class]);
    }
    deallocate_pvector((void **)indices);
    indices = NULL;
}
