/***---------------------------------------------------------------------------
* NAME:     random.c
* PURPOSE:  routines to calculate random numbers of various types
* DESCRIPTION:
*   random_event   - generate 1 with probability p and 0 with prob 1-p
*   random_gauss   - generate gaussian random var with given mean and variance
*   random_gauss_height - calculate height of gaussian pdf given x,mean,and var
*   random_int     - generate uniform random integer from n1 to n2 
*   random_select  - select one of n events with specified probabilities 
*   random_uniform - generate uniform random number from 0.0 to 1.0
*   randomize_index_list - rearrange array entries by swapping random pairs
* 
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/random.c,v 11.3 2001/08/23 20:14:47 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/random.c,v $
* CREATED: 17-Dec-82 rpl                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#include "nnlib.h"
/* these declarations taken from the random 3C man page */
#ifndef STDC_HEADERS
long random();
char *initstate(unsigned seed, char *state, int n);
char *setstate(char *state);
#endif

#define MAX_RANDOM  2147483647.0     /* maximum value of random number */

/* random_uniform - generate uniform random variable from 0 to 1
---------------------------------------------------------------*/
float
random_uniform(void)
{
  return ((float)random()/(MAX_RANDOM));
}

/* random_gauss - gaussian random number generator based on Box_Muller Algorithm
----------------------------------------------------------------------------*/
float
random_gauss( float mean, float sdev)
{
  double u1,u2,x,chi,angle;

  u1=random()/(MAX_RANDOM);
  u2=random()/(MAX_RANDOM);
  chi=sqrt(-2*log(u1));
  angle=2*3.141593*u2;
  x=chi*cos(angle);
  x = mean + x*sdev;
  return((float)x);
}

/* random_event - generate 1 with probability p
---------------------------------------------------------------*/
int
random_event(float p)
{
  if ( (double)random()/MAX_RANDOM  <= p)
    return (1);
  else
    return (0);
}

/* random_gauss_height - calculate height of gaussian density function 
---------------------------------------------------------------*/
float
random_gauss_height(float m, float sigma, float x)
{
  double pi = 3.14159;
  double y;

  y = 1.0/(sqrt(2*pi)*sigma);
  y=y*exp(-(x-m)*(x-m)/(2.0*sigma*sigma));
  return((float)y);
}

/***---------------------------------------------------------------------------
*   random_select - select one of n events (0 to n-1) with specified probabilities 
--------------------------------------------------------------------------***/
int
random_select(float *pprobs, /* event probability array */
          int n /* number of events */)
{
  int i;
  float x;
  float ptotal;

  /* x is uniform from 0 to 1 */
  x = (float)random()/MAX_RANDOM;

  ptotal = 0.0;
  for(i=0;i<n;i++) {
    ptotal += *pprobs;
    ++pprobs;
    if ( x <= ptotal)
      return(i);
  }
  warning("random_select: probabilities don't sum to 1.0\n");
  return(-1);
}

/***---------------------------------------------------------------------------
*   random_int - generate uniform random integer from n1 to n2 
*       float random_int(n1, n2)
*       int n1, n2 lower and upper limits for integer
--------------------------------------------------------------------------***/
int
random_int(int n1, int n2)
{
  double x;

  x =  random()/(MAX_RANDOM+1.0);
  return(n1 + (int)(x*(n2-n1+1.0)));
}
/*****************************************************************************
* NAME: randomize_index_list()
* DESCRIPTION:
*       Rearranges index list randomly by swapping each entry with a randomly
*       picked other entry
*
*   index_list = int array list_length long
*   list_length = length of index_list
*       
*****************************************************************************/

void
randomize_index_list (int *index_list, int list_length)
{
    int i, tmp, t_index;
    
    for (i=0; i<list_length; i++) {
    tmp = index_list[i];
    t_index = random_int(0,list_length-1);
    if (t_index >= list_length) {
        warning ("%s%s %d %d",
             "In random_index, random_int returned index ",
             "larger than array_size",t_index,list_length-1);
        t_index = list_length-1;
    }
    index_list[i] = index_list[t_index];
    index_list[t_index] = tmp;
    } 
}

/*These routines were written by Linda Kukolich on September 15,
1993 at the request of Dr. Richard Lippmann.  We decided to allow 
random splitting  of cross validation data which would allow the
user to change the splits by    changing the seed.  This could cause
different splits to be used by a clusterer and a classifier (say
Gmix and Kmeans) which would be incorrect.  This routine should not
be used often, as the order of random numbers will be the same after
each time it is called.
*/
/* USE: set_random_seed(seed);
   set_nn_random();
   */
static int nn_seed = 0;
void
set_random_seed(int seed)
{
    nn_seed = seed;
}

/* call set_random_seed first to control the starting state */
void
set_nn_random(void)
{
    static long state[64] = {
    0x7be9c1bd, 0x088aa102, 0x3d38509b, 0x746b9fbe,
    0x2d04417f, 0x775d4351, 0x53c48d96, 0x02b26e0b,
    0x418fedcf, 0x19dbc19e, 0x78512adb, 0x1a1f5e2b,
    0x307d7761, 0x6584c1f0, 0x24e3c36f, 0x2232310f,

    0x2dac5ceb, 0x106e8b5a, 0x5a05a938, 0x5e6392b6,
    0x66b90348, 0x75264901, 0x4174f402, 0x618b18a4,
    0x3a6ab4bf, 0x3c4bc289, 0x2657a9a9, 0x4e68589b,
    0x09648aa6, 0x3fc489bb, 0x1c1b715c, 0x054e4c63,

    0x484f2abd, 0x5953c1f8, 0x79b9ec22, 0x75536c3d,
    0x50b10549, 0x4d7e79b8, 0x7805da48, 0x1240f318,
    0x675a3b56, 0x70570523, 0x2c605143, 0x17d7b2b7,
    0x55dbc713, 0x514414b2, 0x3a09e3c7, 0x038823ff,

    0x61b2a00c, 0x140f8cff, 0x61ebb6b5, 0x486ba354,
    0x935d600 , 0x2360aab8, 0x29f6bbf8, 0x43a08abf,
    0x5fac6d41, 0x504e65a2, 0x1208e35b, 0x6910f7e7,
    0x1012ef5d, 0x2e2454b7, 0x6e5f444b, 0x58621a1b,
    };
    initstate(nn_seed, (char *)state, 256);
    setstate((char *)state);
}

