/***---------------------------------------------------------------------------
* NAME:		realloc.c
* PURPOSE:	routines for resizing allocated matrices and vectors
*
* SEE ALSO: allocate.c
* GLOBALS: 
*
* DESCRIPTION:
*	General purpose functions
*
*	reallocate_matrix() -- changes size of 2D float matrix
*	reallocate_dmatrix() -- changes size of 2D double matrix
*	reallocate_imatrix() -- changes size of 2D integer matrix
*	reallocate_pmatrix() -- changes size of 2d matrix of pointers
*	reallocate_vector() -- changes size of 1D float vector
*	reallocate_dvector() -- changes size of 1D double vector 
*	reallocate_ivector() -- changes size of 1D integer vector 
*	reallocate_pvector() -- changes size of 1D vector of pointers
* 
* 
* BUGS AND LIMITATIONS:
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/realloc.c,v 5.2 2001/08/23 20:14:48 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/realloc.c,v $
* CREATED: 1-Feb-90 rpl                             MIT LINCOLN LABORATORY
*---------------------------------------------------------------------------
* Copyright 1990 MIT Lincoln Laboratory.
* This software is not for release and should neither be copied
* nor distributed without explicit permission of MIT Lincoln Laboratory.
--------------------------------------------------------------------------***/
static char copyright[]=
"***************************************************************************\n\
                Copyright 1990 MIT Lincoln Laboratory.\n\
This software is not for release and should neither be copied\n\
nor distributed without explicit permission of MIT Lincoln Laboratory\n\
***************************************************************************\n";

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/realloc.c,v 5.2 2001/08/23 20:14:48 craig Exp $";



/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <strings.h>
#include "nnlib.h"

/***---------------------------------------------------------------------------
* NAME:         reallocate_matrix
* PURPOSE:      allocate more space for a float 2D matrix
* DESCRIPTION:  Allocates space to be added to a float 2D matrix previously
*	allocated using allocate_matrix.
*	the space to be added is given in number of rows
*	a matrix which was accessed using
*		m[0][0] through m[nrow][ncol] 
*	if now accessed using
*		m[0][0] through m[newrow][ncol]
*
*	warning: the data moves, so old pointers into the data become invalid
*
* EXAMPLE:
*                       float **pfloat;
*                       pfloat = reallocate_matrix(pfloat, 10, 100, 20);
--------------------------------------------------------------------------***/
float **reallocate_matrix(row_pointers, nrow, ncol, newrow)
float **row_pointers;
int nrow, ncol,newrow;
{
    int i,j;
    float *data_pointer;
    int nitems;

    nitems = newrow*ncol;

    /* allocate contiguous memory for all data, copying current data
    -----------------------------------------------------------*/
    if(nrow != 0){
	data_pointer = (float *)realloc( row_pointers[0],
		(unsigned)nitems*sizeof(float));
    } else {
	data_pointer = (float *)calloc( (unsigned)nitems,
		(unsigned)sizeof(float));
    }
    if (!data_pointer) {
	fatal_error( "%s Can't allocate data memory for floating matrix\n\
	       with %d rows, %d columns, and %d elements\n",
	       "reallocate_matrix:",newrow, ncol, nitems);
    }  

    /* resize array of pointers to begining of rows
    -----------------------------------------------------------*/
    row_pointers= (float **) realloc(row_pointers,
	    (unsigned) newrow*sizeof(float*));
    if (!row_pointers) {  
	fatal_error( "%s Can't allocate row pointer memory for %s\n\
		with %d rows, %d columns, and %d elements\n",
		"reallocate_matrix:","floating matrix",newrow, ncol,
		nitems);
    }

    /* make row pointers point to the right place in the data space
    -----------------------------------------------------------*/
    for(i=0; i<newrow; i++) {
	row_pointers[i]= data_pointer;
	data_pointer +=  ncol;
    }

    /* zero the space added to the data matrix
    -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
	for(j = 0; j < ncol; j++){
	    row_pointers[i][j] = 0.0;
	}
    }


    return row_pointers;
}
/***---------------------------------------------------------------------------
* NAME:         reallocate_dmatrix
* PURPOSE:      allocate more space for a double 2D matrix
* DESCRIPTION:  Allocates space to be added to a double 2D matrix previously
*	allocated using allocate_dmatrix.
*	the space to be added is given in number of rows
*	a matrix which was accessed using
*		m[0][0] through m[nrow][ncol] 
*	if now accessed using
*		m[0][0] through m[newrow][ncol]
*
*	warning: the data moves, so old pointers into the data become invalid
*
* EXAMPLE:
*                       double **pdouble;
*                       pdouble = reallocate_dmatrix(pdouble, 10, 100, 20);
--------------------------------------------------------------------------***/
double **reallocate_dmatrix(row_pointers, nrow, ncol, newrow)
double **row_pointers;
int nrow, ncol,newrow;
{
    int i,j;
    double *data_pointer;
    int nitems;

    nitems = newrow*ncol;

    /* allocate contiguous memory for all data, copying current data
    -----------------------------------------------------------*/
    if(nrow != 0){
	data_pointer = (double *)realloc( row_pointers[0],
		(unsigned)nitems*sizeof(double));
    } else {
	data_pointer = (double *)calloc( (unsigned)nitems,
		(unsigned)sizeof(double));
    }
    if (!data_pointer) {
	fatal_error( "%s Can't allocate data memory for double matrix\n\
	       with %d rows, %d columns, and %d elements\n",
	       "reallocate_dmatrix:",newrow, ncol, nitems);
    }  

    /* resize array of pointers to begining of rows
    -----------------------------------------------------------*/
    row_pointers= (double **) realloc(row_pointers,
	    (unsigned) newrow*sizeof(double*));
    if (!row_pointers) {  
	fatal_error( "%s Can't allocate row pointer memory for %s\n\
		with %d rows, %d columns, and %d elements\n",
		"reallocate_dmatrix:","double matrix",newrow, ncol,
		nitems);
    }

    /* make row pointers point to the right place in the data space
    -----------------------------------------------------------*/
    for(i=0; i<newrow; i++) {
	row_pointers[i]= data_pointer;
	data_pointer +=  ncol;
    }

    /* zero the space added to the data matrix
    -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
	for(j = 0; j < ncol; j++){
	    row_pointers[i][j] = 0.0;
	}
    }


    return row_pointers;
}
/***--------------------------------------------------------------------------
* NAME:         reallocate_imatrix
* PURPOSE:      allocate more space for an integer 2D matrix
* DESCRIPTION:  Allocates space to be added to an integer 2D matrix previously
*	allocated using allocate_imatrix.
*	the space to be added is given in number of rows
*	a matrix which was accessed using
*		m[0][0] through m[nrow][ncol] 
*	if now accessed using
*		m[0][0] through m[newrow][ncol]
*
*	warning: the data moves, so old pointers into the data become invalid
*
* EXAMPLE:
*                       int **pint;
*                       pint = reallocate_imatrix(pint, 10, 100, 20);
--------------------------------------------------------------------------***/
int **reallocate_imatrix(row_pointers, nrow, ncol, newrow)
int **row_pointers;
int nrow, ncol,newrow;
{
    int i,j;
    int *data_pointer;
    int nitems;

    nitems = newrow*ncol;

    /* allocate contiguous memory for all data, copying current data
    -----------------------------------------------------------*/
    if(nrow != 0){
	data_pointer = (int *)realloc( row_pointers[0],
		(unsigned)nitems*sizeof(int));
    } else {
	data_pointer = (int *)calloc( (unsigned)nitems,
		(unsigned)sizeof(int));
    }
    if (!data_pointer) {
	fatal_error( "%s Can't allocate data memory for integer matrix\n\
	       with %d rows, %d columns, and %d elements\n",
	       "reallocate_imatrix:",newrow, ncol, nitems);
    }  

    /* resize array of pointers to begining of rows
    -----------------------------------------------------------*/
    row_pointers= (int **) realloc(row_pointers,
	    (unsigned) newrow*sizeof(int*));
    if (!row_pointers) {  
	fatal_error( "%s Can't allocate row pointer memory for %s\n\
		with %d rows, %d columns, and %d elements\n",
		"reallocate_imatrix:","integer matrix",newrow, ncol,
		nitems);
    }

    /* make row pointers point to the right place in the data space
    -----------------------------------------------------------*/
    for(i=0; i<newrow; i++) {
	row_pointers[i]= data_pointer;
	data_pointer +=  ncol;
    }

    /* zero the space added to the data matrix
    -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
	for(j = 0; j < ncol; j++){
	    row_pointers[i][j] = 0.0;
	}
    }


    return row_pointers;
}
/***--------------------------------------------------------------------------
* NAME:         reallocate_pmatrix
* PURPOSE:      allocate more space for an 2D matrix of pointers
* DESCRIPTION:  Allocates space to be added to an integer 2D matrix previously
*	allocated using allocate_pmatrix.
*	the space to be added is given in number of rows
*	a matrix which was accessed using
*		m[0][0] through m[nrow][ncol] 
*	if now accessed using
*		m[0][0] through m[newrow][ncol]
*
*	warning: the data moves, so old pointers into the data become invalid
*
* EXAMPLE:
*                       void ***pvoid;
*                       pvoid = reallocate_imatrix(pvoid, 10, 100, 20);
--------------------------------------------------------------------------***/
void ***reallocate_pmatrix(row_pointers, nrow, ncol, newrow)
void ***row_pointers;
int nrow, ncol,newrow;
{
    int i,j;
    void **data_pointer;
    int nitems;

    nitems = newrow*ncol;

    /* allocate contiguous memory for all data, copying current data
    -----------------------------------------------------------*/
    if(nrow != 0){
	data_pointer = (void **)realloc( row_pointers[0],
		(unsigned)nitems*sizeof(void **));
    } else {
	data_pointer = (void **)calloc( (unsigned)nitems,
		(unsigned)sizeof(void *));
    }
    if (!data_pointer) {
	fatal_error( "%s Can't allocate data memory for pointer matrix\n\
with %d rows, %d columns, and %d elements\n",
	       "reallocate_imatrix:", newrow, ncol, nitems);
    }

    /* resize array of pointers to begining of rows
    -----------------------------------------------------------*/
    row_pointers = (void ***) realloc(row_pointers,
	    (unsigned) newrow*sizeof(void**));
    if (!row_pointers) {
	fatal_error( "%s Can't allocate row pointer memory for %s\n\
with %d rows, %d columns, and %d elements\n",
		"reallocate_imatrix:","pointer matrix",newrow, ncol,
		nitems);
    }

    /* make row pointers point to the right place in the data space
    -----------------------------------------------------------*/
    for(i=0; i<newrow; i++) {
	row_pointers[i]= data_pointer;
	data_pointer +=  ncol;
    }

    /* zero the space added to the data matrix
    -----------------------------------------------------------*/
    for(i = nrow; i < newrow; i++){
	for(j = 0; j < ncol; j++){
	    row_pointers[i][j] = NULL;
	}
    }


    return row_pointers;
}
/***---------------------------------------------------------------------------
* NAME:         reallocate_vector
* PURPOSE:      allocates more space for an integer vector previously
*	allocated using allocate_vector
*
* EXAMPLE:
*                       int *pint;
*                       pint = reallocate_vector(pint,10,20);
--------------------------------------------------------------------------***/
float *reallocate_vector(v,n,newn)
float *v;
int n,newn;
{
	int i;

        v=(float *)realloc(v,(unsigned)newn * sizeof(int));
        if (!v)
                fatal_error( "%s %s %d elements",
                     "reallocate_vector:",
                     "Can't allocate memory for integer vector with", n);
	/* zero out new entries */
	for(i = n; i < newn; i++){ v[i] = 0; }
        return v;
}
 
/***---------------------------------------------------------------------------
* NAME:         reallocate_dvector
* PURPOSE:      allocates more space for an integer vector previously
*	allocated using allocate_dvector
*
* EXAMPLE:
*                       double *pdouble;
*                       pdouble = reallocate_dvector(pdouble,10,20);
--------------------------------------------------------------------------***/
double *reallocate_dvector(v,n,newn)
double *v;
int n,newn;
{
	int i;

        v=(double *)realloc(v,(unsigned)newn * sizeof(double));
        if (!v)
                fatal_error( "%s %s %d elements",
                     "reallocate_dvector:",
                     "Can't allocate memory for double vector with", n);
	/* zero out new entries */
	for(i = n; i < newn; i++){ v[i] = 0; }
        return v;
}
 
 
/***---------------------------------------------------------------------------
* NAME:         reallocate_ivector
* PURPOSE:      allocates more space for an integer vector previously
*	allocated using allocate_ivector
*
* EXAMPLE:
*                       int *pint;
*                       pint = reallocate_ivector(pint,10,20);
--------------------------------------------------------------------------***/
int *reallocate_ivector(v,n,newn)
int *v;
int n,newn;
{
	int i;

        v=(int *)realloc(v,(unsigned)newn * sizeof(int));
        if (!v)
                fatal_error( "%s %s %d elements",
                     "reallocate_ivector:",
                     "Can't allocate memory for integer vector with", n);
	/* zero out new entries */
	for(i = n; i < newn; i++){ v[i] = 0; }
        return v;
}
 
/***--------------------------------------------------------------------------
* NAME:         reallocate_pvector
* PURPOSE:      allocates more space for an pointer vector previously
*	allocated using allocate_pvector
*
* EXAMPLE:
*                       void **pvoid;
*                       pvoid = reallocate_ivector(pvoid,10,20);
--------------------------------------------------------------------------***/
void **reallocate_pvector(v,n,newn)
void **v;
int n,newn;
{
	int i;

        v=(void **)realloc(v,(unsigned)newn * sizeof(void *));
        if (!v)
                fatal_error( "%s %s %d elements",
                     "reallocate_pvector:",
                     "Can't allocate memory for pointer vector with", n);
	/* zero out new entries */
	for(i = n; i < newn; i++){ v[i] = NULL; }
        return v;
}
 
