/***------------------------------------------------------------------------
* NAME:  report.c
* PURPOSE:  Error accumulation, write to file and reporting functions
* USAGE
*
* SEE ALSO:
* GLOBALS:
* DESCRIPTION:
*
*	report_build()	   Allocate storage for error reporting
*	report_free()	   Free storage for error reporting
*	report_init()	   Initialize error reporting variables
*	report_class_trial() Update errors for a single classification trial
*	report_map_trial()   Update errors for a single mapping trial
*	report_overall()   Report overall and class by class statistics
*	report_epoch()
*	report_epoch_header()
*
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
/* needed by str commands */
#if HAVE_STRINGS_H
#include <strings.h>
#endif
/* needed by sqrt, exp, and other things */
#include <math.h>
/* for getrusage calls for notebook */
#include <sys/resource.h>
#include "nnlib.h"

#define TRUE 1
#define FALSE 0
#define SMALLEST 1.e-30
#define NONE 0
#define SHORT 1
#define OUTPUTS 2
#define INS_AND_OUTS 3
/* keep this one for mapping only */
#define LONG 2

typedef struct {
    int error_cnt;
    int *class_error_cnt;
    int **class_confusion_matrix;
    int trial_cnt;
    int *class_trial_cnt;
    int cost_func;
    float cost_sum;
    float *class_cost;
    float *class_cost_sum;
    float *desired_outputs;
    int db_type;
    int verbosity;
    int verror;
    int n_inputs;
    int n_outputs;
    int log_outputs;/* print outputs as logs */
    int sum_to_one;/* normalize outputs to sum to one before printing */
  float roc_lower;
  float roc_upper;
  int roc_model;
  int roc_target;
} ERROR_MODEL;

static ERROR_MODEL **errpp;
static int max_models = 0;
#define MODEL_STORE_SIZE 10

/*****************************************************************************
* NAME: report_build ()
* DESCRIPTION:
*       Allocate storage for error reporting
*
******************************************************************************/
int
report_build (int verbosity_p, int verror_p, int n_outputs_p)
{
    int new_model;
    ERROR_MODEL *epp;

    new_model = model_get_number((void ***)&errpp, &max_models, sizeof(ERROR_MODEL));
    epp = errpp[new_model];

    epp->n_outputs = n_outputs_p;
    epp->verbosity = verbosity_p;
    epp->verror = verror_p;
    
    /* allocate space for error tables and mapping outputs*/
    epp->desired_outputs = allocate_vector(n_outputs_p);
    epp->class_cost = allocate_vector(n_outputs_p);
    
    if (verbosity_p > TABLE_VERBOSITY && n_outputs_p > 0) {
	epp->class_error_cnt = allocate_ivector(n_outputs_p);
	epp->class_trial_cnt = allocate_ivector(n_outputs_p);
	epp->class_cost_sum = allocate_vector(n_outputs_p);
	if (verbosity_p > CONFUSION_VERBOSITY)
	    epp->class_confusion_matrix =
		    allocate_imatrix(n_outputs_p + 1, n_outputs_p);
    }
    epp->roc_model = -1;
    return new_model;
}

/*****************************************************************************
* NAME: report_free ()
* DESCRIPTION:
*       Free storage for error reporting
*
******************************************************************************/
void
report_free(int model)
{
    ERROR_MODEL *epp;

    epp = errpp[model];
    deallocate_vector(epp->desired_outputs);
    deallocate_vector(epp->class_cost);
    
    if (epp->verbosity > TABLE_VERBOSITY && epp->n_outputs > 0) {
	deallocate_ivector(epp->class_error_cnt);
	deallocate_ivector(epp->class_trial_cnt);
	deallocate_vector(epp->class_cost_sum);
	if (epp->verbosity > CONFUSION_VERBOSITY)
	    deallocate_imatrix(epp->class_confusion_matrix);
    }
    if(epp->roc_model != -1){
      roc_free(epp->roc_model);
    }
    nn_free((char *)epp, "for report model");
    errpp[model] = NULL;
}

/*****************************************************************************
* NAME: report_init ()
* DESCRIPTION:
*       Initialize error reporting variables
*
******************************************************************************/
void
report_init (int model, int db_type_p, int n_inputs_p, int cost_func,
	     int log_outputs, int sum_to_one)
{
    int		i, j;
    ERROR_MODEL *epp;

    epp = errpp[model];

    epp->trial_cnt = 0;
    epp->error_cnt = 0;
    epp->cost_sum = 0.0;
    epp->cost_func = cost_func;
    epp->n_inputs = n_inputs_p;
    epp->db_type = db_type_p;
    epp->log_outputs = log_outputs;
    epp->sum_to_one = sum_to_one;
    if(epp->cost_func == ROC_COST){
      if(epp->roc_model != -1){
	roc_init(epp->roc_model);
      } else {
	epp->roc_model = roc_build(epp->n_outputs, 0);
      }
      roc_cost_init(epp->n_outputs, 10);
    }

    if (epp->verbosity > TABLE_VERBOSITY && epp->n_outputs > 0) {
        for (i = 0; i < epp->n_outputs; i++){
	    epp->class_error_cnt[i] = epp->class_trial_cnt[i] = 0;
	    epp->class_cost_sum[i] = 0.;
	}
	if (epp->verbosity > CONFUSION_VERBOSITY)
            for (i = 0; i <= epp->n_outputs; i++)
                for (j = 0; j < epp->n_outputs; j++)
		    epp->class_confusion_matrix[i][j] = 0;
    }
}

void
report_roc_init(int model, int roc_target, float roc_lower, float roc_upper)
{
    ERROR_MODEL *epp;

    epp = errpp[model];

    if(epp->cost_func == ROC_COST){
      epp->roc_lower = roc_lower;
      epp->roc_upper = roc_upper;
      epp->roc_target = roc_target;
    }
}

/*****************************************************************************
* NAME: report_class_trial ()
* DESCRIPTION:
*       Update errors for a single trial, write to file for classification
*		problem (not map data)
*
******************************************************************************/
void
report_class_trial(int model, FILE *error_file, float *inputs,
		   float *outputs, int desired_class, int class, int trial_num)
{
    int		i;
    ERROR_MODEL *epp;
    float cost, total;

    epp = errpp[model];

    for(i = 0; i < epp->n_outputs; i++) epp->desired_outputs[i] = 0.;
    epp->desired_outputs[desired_class] = 1.;
    if(epp->cost_func == ROC_COST){
      roc_batch_train(epp->roc_model, desired_class, outputs);
      roc_cost_train(outputs, epp->desired_outputs);
    }
    cost = calc_cost_vector(outputs, epp->desired_outputs, epp->class_cost,
			    epp->n_outputs, epp->cost_func);

    if (epp->verror >= SHORT) {
	fprintf(error_file, "%d %d %d %d ",
		trial_num, desired_class, class,
		(class == desired_class) ? 0 : 1);
	fprintf(error_file, "%9.4e ", cost);
	if (epp->verror >= OUTPUTS){
	    switch(epp->cost_func){
	      case LOG_LIKE:
		if(epp->sum_to_one){
		    total = -MAXFLOAT;
		    for(i = 0; i < epp->n_outputs; i++)
		      total = logadd(total, outputs[i]);
		    for(i = 0; i < epp->n_outputs; i++) outputs[i] -= total;
		}
		if(epp->log_outputs)
		  for(i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", outputs[i]);
		else
		  for (i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", nn_exp10(outputs[i]));
		break;
	      default:
		if(epp->sum_to_one){
		    total = 0.;
		    for(i = 0; i < epp->n_outputs; i++) total += outputs[i];
		    for(i = 0; i < epp->n_outputs; i++) outputs[i] /= total;
		}
		if(epp->log_outputs)
		  for (i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", log10(outputs[i]));
		else
		  for(i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", outputs[i]);
	    }
	    if (epp->verror >= INS_AND_OUTS) {
		for (i = 0; i < epp->n_inputs; i++)	/* print inputs */
		  fprintf(error_file, "%9.4e ", inputs[i]);
	    }
	}
	fprintf(error_file, "\n");
    }
    epp->trial_cnt++;
    epp->cost_sum += cost;
    if (class != desired_class) {
	epp->error_cnt++;
	if (epp->verbosity > TABLE_VERBOSITY)
	    epp->class_error_cnt[desired_class]++;
    }
    if( class >= epp->n_outputs){
	fatal_error("report_class_trial: class from classifier(%d) is out of range (0 - %d)", class, epp->n_outputs-1);
    }
    if (epp->verbosity > TABLE_VERBOSITY) {
	epp->class_trial_cnt[desired_class]++;
	epp->class_cost_sum[desired_class] += cost;
	if (epp->verbosity > CONFUSION_VERBOSITY){
	    if(class != UNKNOWN_CLASS){
		epp->class_confusion_matrix[desired_class][class] += 1;
	    }
	}
    }
}

/*****************************************************************************
* NAME: report_cluster_trial ()
* DESCRIPTION:
*       Update errors for a single trial, write to file for classification
*		problem (not map data)
*
******************************************************************************/
void
report_cluster_trial(int model, FILE *error_file, float *inputs,
		     float *outputs, int class, int trial_num)
{
    int		i;
    ERROR_MODEL *epp;
    float cost, total;

    epp = errpp[model];

    /* cost for clustering is the distance to the selected center */
    switch(epp->cost_func){
      case CLUS_LOG_LIKE:
      case LOG_LIKE:
#ifdef notdef
	/* if I include these lines, the output isn't a LOG likelihood...*/
	cost = nn_exp10(outputs[class]);
	break;
#endif
      default:
	cost = outputs[class];
	break;
    }

    if (epp->verror >= SHORT) {
	fprintf(error_file, "%d %d %d %d ",
		trial_num, class, class, 0);
	fprintf(error_file, "%9.4e ", cost);
	if (epp->verror >= OUTPUTS) {
	    switch(epp->cost_func){
	      case CLUS_LOG_LIKE:
	      case LOG_LIKE:
		if(epp->sum_to_one){
		    total = -MAXFLOAT;
		    for(i = 0; i < epp->n_outputs; i++)
		      total = logadd(total, outputs[i]);
		    for(i = 0; i < epp->n_outputs; i++) outputs[i] -= total;
		}
		if(epp->log_outputs)
		  for(i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", outputs[i]);
		else
		  for (i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", nn_exp10(outputs[i]));
		break;
	      default:
		if(epp->sum_to_one){
		    total = 0.;
		    for(i = 0; i < epp->n_outputs; i++) total += outputs[i];
		    for(i = 0; i < epp->n_outputs; i++) outputs[i] /= total;
		}
		if(epp->log_outputs)
		  for (i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", log10(outputs[i]));
		else
		  for(i = 0; i < epp->n_outputs; i++)
		    fprintf(error_file, "%9.4e ", outputs[i]);
	    }
	    if(epp->verror >= INS_AND_OUTS){
		for (i = 0; i < epp->n_inputs; i++)	/* print inputs */
		  fprintf(error_file, "%9.4e ", inputs[i]);
	    }
	}
	fprintf(error_file, "\n");
    }
    epp->trial_cnt++;
    epp->cost_sum += cost;

    if (epp->verbosity > TABLE_VERBOSITY) {
	epp->class_trial_cnt[class]++;
	epp->class_cost_sum[class] += cost;
	if (epp->verbosity > CONFUSION_VERBOSITY){
	    if(class != UNKNOWN_CLASS)
	      epp->class_confusion_matrix[class][class] += 1;
	}
    }
}

/*****************************************************************************
* NAME: report_map_trial ()
* DESCRIPTION:
*       Update errors for a single trial, write to file for mapping problem
*		(map data)
*
******************************************************************************/
void
report_map_trial(int model, FILE *error_file, float *inputs, float *outputs,
		 float *desired_outputs, int trial_num)
{
    int		i;
    ERROR_MODEL *epp;
    float cost;

    epp = errpp[model];

    cost = calc_cost_vector(outputs, desired_outputs, epp->class_cost, 
			    epp->n_outputs, epp->cost_func);
    if (epp->verror >= SHORT) {
	fprintf(error_file, "%d ", trial_num);
	fprintf(error_file, "%9.4e ", cost);
    }
    if(epp->verror >= OUTPUTS){
	for (i = 0; i < epp->n_outputs; i++)	/* print desired output */
	    fprintf(error_file, "%9.4e ", desired_outputs[i]);
	for (i = 0; i < epp->n_outputs; i++)	/* print actual output */
	    fprintf(error_file, "%9.4e ", outputs[i]);
    }
    if (epp->verror >= LONG) {
	for (i = 0; i < epp->n_inputs; i++)	/* print inputs */
	  fprintf(error_file, "%9.4e ", inputs[i]);
    }
    if(epp->verror >= SHORT) fprintf(error_file, "\n");
    epp->trial_cnt++;
    epp->cost_sum += cost;
    if(epp->verbosity > TABLE_VERBOSITY) {
	for(i = 0; i < epp->n_outputs; i++){
	    epp->class_trial_cnt[i]++;
	    epp->class_cost_sum[i] += epp->class_cost[i];
	}
    }
}

/*****************************************************************************
* NAME: report_overall ()
* DESCRIPTION:
*       Report overall and class by class statistics
*
*  There is a problem if labels is NULL!!! or if labels is length 0
******************************************************************************/
void
report_overall(int model, FILE *log_file, char *error_fname, char **labels)
     /* error_fname is the file name to be printed in log file */
     /* labels is the set of labels describing outputs */
{
    int		i, j, itmp, tmp_cnt, length;
    float	cost_tmp, percent_error, stddev, tmp_err;
    ERROR_MODEL *epp;

    epp = errpp[model];

    /* check for labels list length 0 */
    if(labels != NULL){
	length = atoi(labels[0]);
	if(length == 0 || length < epp->n_outputs) labels = NULL;
    }

    if (epp->trial_cnt <= 0) return;
    /* print out the confusion matrix */
    if(epp->verbosity > CONFUSION_VERBOSITY){
	/* confusion matrix for sequence or classification data.
	   there is no confusion matrix for mapping data */
	if(epp->db_type != MAP_DATA){
	    fprintf(log_file, "\n\n%s %s", "Classification Confusion Matrix -",
		    error_fname);
	    fprintf(log_file, "\n%s%s", "--------------------------------",
		    "-----------------------------------------------");
	    fprintf(log_file, "\n%s%s", "--------------------------------",
		    "-----------------------------------------------");
	    fprintf(log_file, "\n\nDesired%41s", "Computed Class\n");
	    fprintf(log_file, " Class ");
	    for (j = 0; j < epp->n_outputs; j++)
		fprintf(log_file, "%5d ", j);
	    fprintf(log_file, "  Total\n ----- ");
	    for (j = 0; j < epp->n_outputs; j++)
		fprintf(log_file, "  ----");
	    fprintf(log_file, "  -----\n");
	    for (i = 0; i < epp->n_outputs; i++) {
		fprintf(log_file, "%4d  ", i);
		for (j = 0; j < epp->n_outputs; j++) {
		    itmp = epp->class_confusion_matrix[i][j];
		    if (itmp > 0)
			fprintf(log_file, " %5d", itmp);
		    else
			fprintf(log_file, "      ");
		    epp->class_confusion_matrix[epp->n_outputs][j] += itmp;
		}
		fprintf(log_file, " %6d\n", epp->class_trial_cnt[i]);
	    }
	    fprintf(log_file, " ----- ");
	    for (j = 0; j < epp->n_outputs; j++)
		fprintf(log_file, "  ----");
	    fprintf(log_file, "  -----\n Total");
	    for (j = 0; j < epp->n_outputs; j++) {
		itmp = epp->class_confusion_matrix[epp->n_outputs][j];
		fprintf(log_file, " %5d", itmp);
	    }
	    fprintf(log_file, " %6d\n\n", epp->trial_cnt);
	    fprintf(log_file, "%s%s\n", "--------------------------------",
		    "-----------------------------------------------");
	    fprintf(log_file, "%s%s\n", "--------------------------------",
		    "-----------------------------------------------");
	    fprintf(log_file, "\n\n\n");
	}
    }
    /* print out error rate for each class */
    if(epp->verbosity > TABLE_VERBOSITY) {
	fprintf(log_file, "\n%s %s\n", "Error Report -",
		error_fname);
	fprintf(log_file, "%s%s\n", "--------------------------------",
		"-----------------------------------------------");
	fprintf(log_file, "%s%s\n", "--------------------------------",
		"-----------------------------------------------");
	if(epp->db_type == MAP_DATA){
	    /* error table for mapping data */
	    fprintf(log_file, "Output%8s   Label\n",
		    cost_string(epp->cost_func));
	    for(i = 0; i < epp->n_outputs; i++){
		tmp_cnt = epp->class_trial_cnt[i];
		if(tmp_cnt <= 0){
		    fprintf(log_file, "%3d", i);
		    if(labels != NULL)
		      fprintf(log_file, " %20s", labels[i+1]);
		}
		else {
		    /* mean of squared error in ith outputs */
		    fprintf(log_file, "%3d %9.3f\n",
			    i, calc_avg_cost(epp->class_cost_sum[i],
					     tmp_cnt, epp->cost_func));
		    if(labels != NULL)
		      fprintf(log_file, " %s", labels[i+1]);
		}
		fprintf(log_file, "\n");
	    }
	} else {
	    /* error table for sequence or classification data */
	    fprintf(log_file, "%s %12s %10s %12s %6s ",
		    "Class", "Patterns", "# Errors", "% Errors", "StdDev");
	    if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
	      fprintf(log_file, "%7s", cost_string(epp->cost_func));
	    fprintf(log_file, "%s\n", "   Label");
	    for (i = 0; i < epp->n_outputs; i++) {
		tmp_cnt = epp->class_trial_cnt[i];
		if (tmp_cnt <= 0){
		    fprintf(log_file, "%3d %12d ", i, tmp_cnt);
		    if(labels != NULL){
			if(epp->cost_func != CENTERS &&
			   epp->cost_func != NO_COST){
			    fprintf(log_file, "%49s", labels[i + 1]);
			} else {
			    fprintf(log_file, "%40s", labels[i+1]);
			}
		    }
		    fprintf(log_file, "\n");
		} else {
		    tmp_err = ((float) epp->class_error_cnt[i]/
			       ((float) tmp_cnt));
		    percent_error = tmp_err * 100.;
		    stddev = sqrt((tmp_err*(1-tmp_err))/(float)tmp_cnt)*100.;
		    
		    fprintf(log_file,
			    "%3d %12d %10d %12.2f   (%4.1f) ",
			    i, tmp_cnt, epp->class_error_cnt[i],
			    percent_error, stddev);
		    if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
		      fprintf(log_file, "%9.3f",
			      calc_avg_cost(epp->class_cost_sum[i],
					    tmp_cnt, epp->cost_func));
		    if(labels != NULL)
		      fprintf(log_file,"  %s", labels[i+1]);
		    fprintf(log_file, "\n");
		}
	    }
	}
    }
    /* print out overall error */
    cost_tmp = calc_avg_cost(epp->cost_sum, epp->trial_cnt, epp->cost_func);
    if(epp->verbosity > OVERALL_VERBOSITY){
	if (epp->db_type == MAP_DATA) {
	    fprintf(log_file, "\nOverall %8d patterns, %7.3f cost\n\n",
		    epp->trial_cnt, cost_tmp);
	} else {
	    /* print the labels for the columns if necessary */
	    if (epp->verbosity < CONFUSION_VERBOSITY){
		fprintf(log_file, "\n%18s %12s %10s %9s ",
			"Patterns", "# Errors", "% Errors", "StdDev");
		if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
		  fprintf(log_file, "%7s\n", cost_string(epp->cost_func));
		fprintf(log_file, "\n");
	    }
	    fprintf(log_file, "%18s %12s %10s %6s",
		    "--------", "--------", "-------", "------");
	    if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
	      fprintf(log_file, "%7s\n", "-------");
	    fprintf(log_file, "\n");
	    tmp_err = ((float) epp->error_cnt / ((float) epp->trial_cnt));
	    percent_error = tmp_err * 100.;
	    stddev = sqrt ((tmp_err * (1-tmp_err)) / (float) epp->trial_cnt)
	      * 100.;
	    fprintf(log_file, "Overall %8d %10d %12.2f   (%4.1f)",
		    epp->trial_cnt, epp->error_cnt, percent_error, stddev);
	    if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
	      fprintf(log_file, "%9.3f", cost_tmp);
	    fprintf(log_file, "\n\n");
	}
    }
}

/*****************************************************************************
* NAME: report_cluster_overall ()
* DESCRIPTION:
*       Report overall and cluster by cluster statistics
*
*  There is a problem if labels is NULL!!! or if labels is length 0
******************************************************************************/
void
report_cluster_overall(int model, FILE *log_file,
		       char *error_fname,/* file name to be printed in log */
		       char **labels/* set of labels describing outputs */)
{
    int		i, tmp_cnt;
    float	cost_tmp;
    ERROR_MODEL *epp;

    epp = errpp[model];

    /* check for labels list length 0 */
#ifdef notdef
    if(labels != NULL){
	length = atoi(labels[0]);
	if(length == 0 || length < epp->n_outputs) labels = NULL;
    }
#else
    /* turn labels off, they will only confuse us */
    labels = NULL;
#endif

    if (epp->trial_cnt <= 0) return;
    /* print out error rate for each cluster */
    if(epp->verbosity > TABLE_VERBOSITY) {
	fprintf(log_file, "\n%s %s\n", "Error Report -",
		error_fname);
	fprintf(log_file, "%s%s\n", "--------------------------------",
		"-----------------------------------------------");
	fprintf(log_file, "%s%s\n", "--------------------------------",
		"-----------------------------------------------");
	/* error table for sequence or classification data */
	fprintf(log_file, "%s %12s ",
		"Cluster", "Patterns");
	if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
	  fprintf(log_file, "%7s\n", cost_string(epp->cost_func));
	else fprintf(log_file, "\n");
	for (i = 0; i < epp->n_outputs; i++) {
	    tmp_cnt = epp->class_trial_cnt[i];
	    if (tmp_cnt <= 0){
		fprintf(log_file, "%3d %12d ", i, tmp_cnt);
		fprintf(log_file, "\n");
	    } else {
		fprintf(log_file, "%3d %12d ", i, tmp_cnt);
		if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
		  fprintf(log_file, "%9.3f",
			  calc_avg_cost(epp->class_cost_sum[i],
					tmp_cnt, epp->cost_func));
		fprintf(log_file, "\n");
	    }
	}
    }
    /* print out overall error */
    cost_tmp = calc_avg_cost(epp->cost_sum, epp->trial_cnt, epp->cost_func);
    if(epp->verbosity > OVERALL_VERBOSITY){
	if (epp->db_type == MAP_DATA) {
	    fprintf(log_file, "\nOverall %8d patterns, %7.3f cost\n\n",
		    epp->trial_cnt, cost_tmp);
	} else {
	    /* print the labels for the columns if necessary */
	    if (epp->verbosity < CONFUSION_VERBOSITY){
		fprintf(log_file, "\n%18s ", "Patterns");
		if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
		  fprintf(log_file, "%7s\n", cost_string(epp->cost_func));
		fprintf(log_file, "\n");
	    }
	    fprintf(log_file, "%18s ", "--------");
	    if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
	      fprintf(log_file, "%7s\n", "-------");
	    fprintf(log_file, "\n");
	    fprintf(log_file, "Overall %8d", epp->trial_cnt);
	    if(epp->cost_func != CENTERS && epp->cost_func != NO_COST)
	      fprintf(log_file, "%9.3f", cost_tmp);
	    fprintf(log_file, "\n\n");
	}
    }
}

/* print header information before epoch training */
void
report_epoch_header(int model, FILE *report_file, int n_trials)
{
    ERROR_MODEL *epp;

    epp = errpp[model];
    if(epp->cost_func == ROC_COST){
      roc_cost_finish_epoch();
    }
    if(epp->verbosity > EPOCH_VERBOSITY){
	switch(epp->db_type){
	  case MAP_DATA:
	    fprintf(report_file, "\n EPOCH %7s ",
		    cost_string(epp->cost_func));
	    break;
	  default:/* CLASS_DATA or SEQ_DATA */
	    switch(epp->cost_func){
	      case NO_COST: fprintf(report_file, "\n EPOCH %7s ",
			"%error");
		break;
	      case CLUS_SQ_ERR:
	      case CLUS_LOG_LIKE:
		fprintf(report_file, "\n EPOCH %7s ncenters ",
			cost_string(epp->cost_func));
		break;
	      case ARTMAP_COST:
		fprintf(report_file,"\n EPOCH%7s%6s%9s%9s ",
			"%serr", "%err", "num_f2a", "num_f2b");
		break;
	      default:
		fprintf(report_file, "\n EPOCH %7s %10s",
			"%error", cost_string(epp->cost_func));
		break;
	    }
	}
	fprintf(report_file, "(%d patterns/epoch)\n", n_trials);
	fflush(report_file);
    }
}

/* do a report_epoch_header that uses the data model to pull the number of
 patterns per epoch */
void
report_data_epoch_header(int model, FILE *report_file, int data_model)
{
    ERROR_MODEL *epp;
    int n_frames = 0, n_patterns;

    epp = errpp[model];

    switch(epp->db_type){
      case SEQ_DATA:
	n_patterns = seq_data_get_nfiles(data_model);
	n_frames = seq_data_get_nframes(data_model);
	break;
      case MAP_DATA:
	n_patterns = map_data_get_npatterns(data_model);
	break;
      case CLASS_DATA:
	n_patterns = class_data_get_npatterns(data_model);
	break;
      default:
	n_patterns = 0;
    }
    if(epp->verbosity > EPOCH_VERBOSITY){
	switch(epp->db_type){
	  case MAP_DATA:
	    fprintf(report_file, "\n EPOCH %7s (%d patterns/epoch)\n",
		    cost_string(epp->cost_func), n_patterns);
	    break;
	  case CLASS_DATA:
	  case SEQ_DATA:
	    switch(epp->cost_func){
	      case NO_COST:
		fprintf(report_file,
			"\n EPOCH %7s", "%error");
		break;
	      case CLUS_SQ_ERR:
	      case CLUS_LOG_LIKE:
		fprintf(report_file, "\n EPOCH %7s ncenters",
			cost_string(epp->cost_func));
		break;
	      case ARTMAP_COST:
		fprintf(report_file,"\n EPOCH%7s%7s%9s%9s",
			"%serr", "%err", "num_f2a", "num_f2b");
		break;
	      default:
		fprintf(report_file, "\n EPOCH %7s %10s",
			"%error", cost_string(epp->cost_func));
	    }
	    if(epp->db_type == CLASS_DATA)
	      fprintf(report_file, " (%d patterns/epoch)\n", n_patterns);
	    else
	      fprintf(report_file, " (%d sequences, %d frames/epoch)\n",
		      n_patterns, n_frames);
	    break;
	}
	fflush(report_file);
    }
}

void
report_epoch(int model, FILE *report_file, int epoch)
{
    ERROR_MODEL *epp;
    float percent_error, cost;

    epp = errpp[model];

    if (epp->verbosity > EPOCH_VERBOSITY) {
	cost = calc_avg_cost(epp->cost_sum, epp->trial_cnt, epp->cost_func);
	if(epp->db_type == MAP_DATA) {
	    fprintf(report_file, "%4d%10.4g\n", epoch, cost);
	} else {
	    percent_error =
	      ((float) epp->error_cnt / ((float) epp->trial_cnt)) * 100.;
	    fprintf(report_file, "%4d%10.1f", epoch, percent_error);
	    if(epp->cost_func == NO_COST)
	      fprintf(report_file, "\n");
	    else
	      fprintf(report_file, "%10.4g\n", cost);
	}
	fflush(report_file);	/* ensures timely results */
    }
}    

void
report_centers_epoch(int model, FILE *report_file, int epoch, int centers)
{
    ERROR_MODEL *epp;
    float percent_error;

    epp = errpp[model];

    if (epp->verbosity > CONFUSION_VERBOSITY) {
	if(epp->db_type == MAP_DATA) {
	    fprintf(report_file, "%4d%10d\n", epoch, centers);
	} else {
	    percent_error =
	      ((float) epp->error_cnt / ((float) epp->trial_cnt)) * 100.;
	    fprintf(report_file, "%4d%10.1f%10d\n",
		    epoch, percent_error, centers);
	}
	fflush(report_file);	/* ensures timely results */
    }
}    

void
report_artmap_epoch(int model, FILE *report_file, int epoch, int err_cnt,
		    int num_a, int num_b)
{
    ERROR_MODEL *epp;
    float percent_error;

    epp = errpp[model];

    if (epp->verbosity > CONFUSION_VERBOSITY) {
	/* print percent error counting returns of Novel class */
	percent_error =
	  ((float) epp->error_cnt/((float)epp->trial_cnt)) *100.;
	printf("%4d%7.0f", epoch, percent_error);
	/* print percent error counting Novel's as correct */
	percent_error =
	  ((float) err_cnt / ((float) epp->trial_cnt)) * 100.;
	fprintf(report_file, "%7.0f", percent_error);
	/* print number of ArtA and ArtB nodes */
	fprintf(report_file, "%9d%9d\n", num_a, num_b);
	fflush(report_file);
    }
}    

void
report_cluster_epoch(int model, FILE *report_file, int epoch, int centers,
		     float cost)
{
    ERROR_MODEL *epp;

    epp = errpp[model];

    if (epp->verbosity > CONFUSION_VERBOSITY) {
	fprintf(report_file, "%4d %10.3f %10d\n",
		epoch, cost, centers);
	fflush(report_file);	/* ensures timely results */
    }
}    

/* these three routines, report_test_result, report_cv_result, and
   report_last_epoch are intended for use with the LNKnet notebook */
void
report_test_result(int model, FILE *report_file, char *dataname)
{
    ERROR_MODEL *epp;
    float percent_error, cost, seconds;
    struct rusage rusage_s;

    epp = errpp[model];

    if(epp->verbosity > OVERALL_VERBOSITY){
	/* get cost, percent error, and cpu time info */
	cost = calc_avg_cost(epp->cost_sum, epp->trial_cnt, epp->cost_func);
	percent_error = 100. *((float) epp->error_cnt /
			       ((float) epp->trial_cnt));
	getrusage(RUSAGE_SELF, &rusage_s);
	seconds = rusage_s.ru_utime.tv_sec +
	  (float)rusage_s.ru_utime.tv_usec*1.e-6;
	seconds += rusage_s.ru_stime.tv_sec +
	  (float)rusage_s.ru_stime.tv_usec*1.e-6;
	
	fprintf(report_file, " TEST:%s    Time: %.2f secs",
		dataname,seconds);
	if(epp->cost_func != NO_COST && epp->cost_func != CENTERS)
	  fprintf(report_file, "  %s: %.3g",
		  cost_string(epp->cost_func), cost);
	fprintf(report_file, "  Err: %.2f%%", percent_error);
	fprintf(report_file, "\n");
    }
}

void
report_cv_result(int model, FILE *report_file, char *dataname, int nfolds)
{
    ERROR_MODEL *epp;
    float percent_error, cost, seconds;
    struct rusage rusage_s;

    epp = errpp[model];

    if(epp->verbosity > OVERALL_VERBOSITY){
	cost = calc_avg_cost(epp->cost_sum, epp->trial_cnt, epp->cost_func);
	percent_error = 100. *((float) epp->error_cnt / ((float) epp->trial_cnt));
	getrusage(RUSAGE_SELF, &rusage_s);
	seconds = rusage_s.ru_utime.tv_sec +
	  (float)rusage_s.ru_utime.tv_usec*1.e-6;
	seconds += rusage_s.ru_stime.tv_sec +
	  (float)rusage_s.ru_stime.tv_usec*1.e-6;

	fprintf(report_file, " CV%2d:%s    Time: %.2f secs",
		nfolds, dataname, seconds);
	if(epp->cost_func != NO_COST && epp->cost_func != CENTERS)
	  fprintf(report_file, "  %s: %.3g",
		  cost_string(epp->cost_func), cost);
	fprintf(report_file, "  Err: %.2f%%", percent_error);
	fprintf(report_file, "\n");
    }
}

void
report_last_epoch(int model, FILE *report_file, int epoch)
{
    ERROR_MODEL *epp;
    float percent_error, cost, seconds;
    struct rusage rusage_s;

    epp = errpp[model];

    if (epp->verbosity > OVERALL_VERBOSITY) {
	fprintf(report_file, " LAST TRAIN EPOCH:");
	cost = calc_avg_cost(epp->cost_sum, epp->trial_cnt, epp->cost_func);
	getrusage(RUSAGE_SELF, &rusage_s);
	seconds = rusage_s.ru_utime.tv_sec +
	  (float)rusage_s.ru_utime.tv_usec*1.e-6;
	seconds += rusage_s.ru_stime.tv_sec +
	  (float)rusage_s.ru_stime.tv_usec*1.e-6;

	if(epp->db_type == MAP_DATA) {
	    fprintf(report_file, " %d Time: %.2f secs  %s: %.2g\n",
		    epoch, seconds, cost_string(epp->cost_func), cost);
	} else {
	    percent_error =
	      ((float) epp->error_cnt / ((float) epp->trial_cnt)) * 100.;
	    fprintf(report_file, " %d Time: %.2f secs", epoch, seconds);
	    if(epp->cost_func != NO_COST && epp->cost_func != CENTERS)
	      fprintf(report_file, "  %s: %.3g",
		      cost_string(epp->cost_func), cost);
	    fprintf(report_file, "  Err: %.2f%%", percent_error);
	    fprintf(report_file, "\n");
	}
	fflush(report_file);	/* ensures timely results */
    }
}    

float
report_get_avg_cost(int model)
{
    float cost;
    ERROR_MODEL *epp;

    epp = errpp[model];
    if(epp->cost_func == ROC_COST){
      return roc_get_percent_fom(epp->roc_model, epp->roc_target, 0.0,
				 epp->roc_lower, epp->roc_upper);
    } else {
      cost = calc_avg_cost(epp->cost_sum, epp->trial_cnt, epp->cost_func);
    }
    return cost;
}

void
report_seq_trial(int model, FILE *error_file, int trial_num, float *outputs,
		 int desired_class, int class)
{
    int i;
    ERROR_MODEL *epp;
    float cost;

    epp = errpp[model];

    for(i = 0; i < epp->n_outputs; i++) epp->desired_outputs[i] = 0.;
    epp->desired_outputs[desired_class] = 1.;
    cost = calc_cost_vector(outputs, epp->desired_outputs, epp->class_cost,
			    epp->n_outputs, epp->cost_func);

    if(epp->verror >= SHORT){
	fprintf(error_file, "%d %d %d %d ",
	       trial_num, desired_class, class,
	       (class == desired_class) ? 0 : 1);
	fprintf(error_file, "%9.4e ", cost);
	if(epp->verror >= OUTPUTS){
	    for(i = 0; i < epp->n_outputs; i++){
		fprintf(error_file, "%9.4e ", outputs[i]);
	    }
	}
	fprintf(error_file, "\n");
    }
    epp->trial_cnt++;
    epp->cost_sum += cost;
    if(class != desired_class){
	epp->error_cnt++;
	if(epp->verbosity > TABLE_VERBOSITY)
	  epp->class_error_cnt[desired_class]++;
    }
    if(epp->verbosity > TABLE_VERBOSITY) {
	epp->class_trial_cnt[desired_class]++;
	epp->class_cost_sum[desired_class] += cost;
	if (epp->verbosity > CONFUSION_VERBOSITY)
	  epp->class_confusion_matrix[desired_class][class] += 1;
    }
}

float
report_get_percent_error(int model)
{
    ERROR_MODEL *epp;
    float percent_error;

    epp = errpp[model];
    if(epp->cost_func == ROC_COST){
      return 100 - roc_get_percent_fom(epp->roc_model, epp->roc_target, 0.0,
				       epp->roc_lower, epp->roc_upper);
    } else {
      percent_error = ((float) epp->error_cnt/ ((float) epp->trial_cnt));
      percent_error *= 100.;
      return percent_error;
    }
}
