/***--------------------------------------------------------------------------
* NAME:     roc_cost.c
* PURPOSE:  package for computing roc functions as cost functions
*
* USAGE:
*   roc_cost_init(noutputs, nbins);
*   ...
*   foreach ecpoch of training:
*   foreach training pattern:
*       cost = roc_calc_cost(outputs, desired_outputs);
*       OR
*       cost = roc_calc_cost_vector(outputs, desired_outputs, cost_vector)
*       OR
*       roc_calc_deriv(outputs, desired_outputs);
*
*       roc_cost_train(outputs, desired_outputs);
*   roc_cost_finish_epoch();
*   ...
*
*   roc_cost_free();
*
* GLOBALS: 
*
* FUNCTIONS:
*   roc_cost_init()
*   roc_calc_cost()
*   roc_calc_deriv()
*   roc_calc_cost_vector()
*   roc_cost_train()
*   roc_cost_finish_epoch()
*   roc_cost_free()
*
* DESCRIPTION:
*   The ROC curve is approximated by assigning incoming patterns to bins
*   and calculating and ROC area for the binned version of the curve.
*
*   Each epoch, the user presents patterns to build the next ROC curve
*   while getting cost and derivative results from the previous curve.
*   A squared error cost function is used for the first epoch.
*   The user signals the end of each epoch by calling
*   roc_cost_finish_epoch().
*
*   The cost returned for a pattern is the area of the ROC curve from the
*   previous epoch, with this pattern added to it.
*
*   The derivative of the cost is the change in ROC area caused by the
*   inclusion of the given pattern.
*
* SEE ALSO: report.c sigm.c roc_pack.c
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/roc_cost.c,v 11.4 2001/08/23 20:14:48 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/roc_cost.c,v $
* Created: January 11, 1995 lck     MIT LINCOLN LABORATORY
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory

Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include "nnlib.h"

typedef struct {
    int nfalse;
    int nhit;
    int nfalse_above;
    int nhit_below;
} ROC_BIN;

typedef struct {
    int nfalse;
    int nhit;
    float area;
    float bin_min, bin_max;
    ROC_BIN *bins;
    /* change in area if a pattern is moved down by 1 */
    float *false_slope;
    float *true_slope;
} ROC_OUTPUT;

typedef struct{
    int noutputs;
    int nbins;
    int cur_roc;
    int next_roc;
    float *cost_vector; /* to make cost calculation easy */
    ROC_OUTPUT **roc;
} ROC_STRUCT;
static ROC_STRUCT *rocpp = NULL;

static void roc_init_next_roc(void);
static int roc_calc_change(float *outputs, float *desired_outputs,
               float *change);
static int roc_cost_get_class(float *desired_outputs);
static int roc_cost_get_bin(float output);
static int epoch_num = 0;

#define NROCS 2
/* allocate space for an roc structure
   initialize the calculation parameters */
void
roc_cost_init(int noutputs, int nbins)
{
    int i, j;

    if(rocpp != NULL) roc_cost_free();
    rocpp = (ROC_STRUCT *)nn_malloc(sizeof(ROC_STRUCT),
                    "for roc cost structure");
    epoch_num = 0;
    rocpp->noutputs = noutputs;
    rocpp->nbins = nbins;
    rocpp->cost_vector = allocate_vector(noutputs);
    rocpp->roc = (ROC_OUTPUT **)
      allocate_structmatrix(NROCS, noutputs, sizeof(ROC_OUTPUT));
    /* allocate bins */
    for(i = 0; i < NROCS; i++){
    for(j = 0; j < noutputs; j++){
        rocpp->roc[i][j].bins =
          (ROC_BIN *)allocate_structvector(nbins, sizeof(ROC_BIN));
        rocpp->roc[i][j].false_slope = allocate_vector(nbins);
        rocpp->roc[i][j].true_slope = allocate_vector(nbins);
    }
    }
    rocpp->cur_roc = -1;
    rocpp->next_roc = 0;
    roc_init_next_roc();
}

/* initialize calculation parameters for the next roc curve */
static void
roc_init_next_roc(void)
{
    ROC_OUTPUT *output;
    ROC_BIN *bin;
    int i, j;

    for(i = 0; i < rocpp->noutputs; i++){
    output = &(rocpp->roc[rocpp->next_roc][i]);
    output->nfalse = output->nhit = 0;
    output->bin_min = MAXFLOAT;
    output->bin_max = -(MAXFLOAT);
    output->area = 0.;
    for(j = 0; j < rocpp->nbins; j++){
        bin = &(output->bins[j]);
        bin->nfalse = bin->nhit = 0;
        bin->nhit_below = bin->nfalse_above = 0;
    }
    }
}

/* calculate the roc area, having added this pattern */
float
roc_calc_cost(float *outputs, float *desired_outputs)
{
    if(rocpp->cur_roc < 0 ){
    return calc_cost(outputs, desired_outputs, rocpp->noutputs, SQ_ERR);
    }
    return roc_calc_cost_vector(outputs, desired_outputs, rocpp->cost_vector);
}

/* calculate the roc area for each class, having added this pattern */
float
roc_calc_cost_vector(float *outputs, float *desired_outputs,
             float *cost_vector)
{
    int i;
    float cost;

    if(rocpp->cur_roc < 0 ){
    return calc_cost_vector(outputs, desired_outputs, cost_vector,
                rocpp->noutputs, SQ_ERR);
    }
    /* find the change in the area and then add it to the area */
    roc_calc_change(outputs, desired_outputs, cost_vector);
    for(cost = 0., i = 0; i < rocpp->noutputs; i++){
    cost_vector[i] += rocpp->roc[rocpp->cur_roc][i].area;
    cost += cost_vector[i];
    }
    cost /= rocpp->noutputs;
    return cost;
}

/* return the slope of the roc for each output value */
int
roc_calc_deriv(float *outputs, float *desired_outputs, float *derivative)
{
    int class, i, bin_num;
    ROC_OUTPUT *output;

    class = roc_cost_get_class(desired_outputs);
    if(rocpp->cur_roc < 0){
    /* A nice feature would be to make tolerance, sigmoid type,
       sigmoid_param, and error function match what the user may
       have asked for... */
    /* I can't just call calc_output_deriv because that includes
       the sigmoid function, which we don't want */
    for(i = 0; i < rocpp->noutputs; i++){
        derivative[i] = outputs[i] - desired_outputs[i];
    }
    } else {
    /* should we be doing a tolerance on this? */
    for(i = 0; i < rocpp->noutputs; i++){
        output = &(rocpp->roc[rocpp->cur_roc][i]);
        bin_num = roc_cost_get_bin(outputs[i]);
        if(bin_num < 0) bin_num = 0;
        if(bin_num >= rocpp->nbins) bin_num = rocpp->nbins-1;
        if(i == class){ /* this is a hit */
        derivative[i] = output->true_slope[bin_num];
        } else { /* this is a false alarm */
        derivative[i] = output->false_slope[bin_num];
        }
    }
    }
    return 0;
}
/* calculate the change to the roc area for each class if this pattern
 were added */
static int
roc_calc_change(float *outputs, float *desired_outputs, float *change)
{
    int class, i, bin_num;
    ROC_OUTPUT *output;

    class = roc_cost_get_class(desired_outputs);
    if(rocpp->cur_roc < 0){
    for(i = 0; i < rocpp->noutputs; i++){
        change[i] = 0.;
    }
    return 0;
    }
    /* should we be doing a tolerance on this? */
    for(i = 0; i < rocpp->noutputs; i++){
    output = &(rocpp->roc[rocpp->cur_roc][i]);
    bin_num = roc_cost_get_bin(outputs[i]);
    if(bin_num < 0) bin_num = 0;
    if(bin_num >= rocpp->nbins) bin_num = rocpp->nbins-1;
    if(i == class){ /* this is a hit */
        change[i] = 0.5*output->bins[bin_num].nfalse +
          output->bins[bin_num].nfalse_above;
        change[i] /= output->nfalse;
        change[i] -= output->area;
        change[i] /= output->nhit + 1.;
    } else { /* this is a false alarm */
        change[i] = 0.5*output->bins[bin_num].nhit +
          output->bins[bin_num].nhit_below;
        change[i] /= output->nhit;
        change[i] -= output->area;
        change[i] /= output->nfalse + 1.;
    }
    }
    return 0;
}

/* add a pattern to the roc curves */
/* foreach class:
   find the bin this output falls in
   update the nhit or nfalse entries for the class and bin
   OR
   update the bin min and max values
*/
float
roc_cost_train(float *outputs, float *desired_outputs)
{
    int class, i, bin_num;
    ROC_OUTPUT *output;

    /* find the correct class */
    class = roc_cost_get_class(desired_outputs);

    for(i = 0; i < rocpp->noutputs; i++){
    output = &(rocpp->roc[rocpp->next_roc][i]);

    /* get the bin number */
    bin_num = roc_cost_get_bin(outputs[i]);

    /* update the max and min for this round */
    if(output->bin_max < outputs[i]) output->bin_max = outputs[i];
    if(output->bin_min > outputs[i]) output->bin_min = outputs[i];

    /* if this didn't fall in a bin, skip it */
    if(bin_num < 0 || bin_num >= rocpp->nbins) continue;
    /* if this fell in a bin, update it */
    if(i == class){
        output->nhit++;
        output->bins[bin_num].nhit++;
    } else {
        output->nfalse++;
        output->bins[bin_num].nfalse++;
    }
    }
    return roc_calc_cost(outputs, desired_outputs);
}

static int
roc_cost_get_class(float *desired_outputs)
{
    int class, i;
    for(class = i = 0; i < rocpp->noutputs; i++){
    if(desired_outputs[i] > desired_outputs[class]) class = i;
    }
    return class;
}

static int
roc_cost_get_bin(float output)
{
    return  (int)((1.-output)*rocpp->nbins);
}

/* calculate the new roc curves, initialize the next set for training */
void
roc_cost_finish_epoch(void)
{
    ROC_OUTPUT *output;
    ROC_BIN *bin, *bin2;
    int i, j;
    int false_above, hit_below;
    float denom;

    epoch_num++;
    if(epoch_num <= 1){
    roc_init_next_roc();
    return;
    }
    rocpp->cur_roc = rocpp->next_roc;
    rocpp->next_roc++;
    rocpp->next_roc = rocpp->next_roc%NROCS;
    
    for(i = 0; i < rocpp->noutputs; i++){
    output = &(rocpp->roc[rocpp->cur_roc][i]);
    denom = (float)output->nfalse*output->nhit;
    false_above = output->nfalse;
    hit_below = 0;
    output->area = 0.;
    /* set the number of false alarms above, number of hits below,
       and total up the area */
    for(j = 0; j < rocpp->nbins; j++){
        bin = &(output->bins[j]);
        false_above -= bin->nfalse;
        bin->nfalse_above = false_above;
        bin->nhit_below = hit_below;
        hit_below += bin->nhit;
        output->area += bin->nfalse*(0.5 * bin->nhit + bin->nhit_below);
    }
    /* divide by the total number of false alarms and hits,
       provided there are some of both */
    if(denom != 0){
        output->area /= denom;
    }
    /* precalculate the necessary slope values */
    for(j = 0; j < rocpp->nbins; j++){
        bin = &(output->bins[j]);
        /* find the change in the area 
           if a false alarm or hit were moved down 1 bin */
        /* this is a simplified version of the change equation that is
           possible because the bins involved are neighbors */
        if(j > 0){
        bin2 = &(output->bins[j-1]);
        output->false_slope[j] = .5*(bin2->nhit + bin->nhit);
        output->true_slope[j] = -.5*(bin2->nfalse + bin->nfalse);
        } else {
        output->false_slope[j] = 0.;
        output->true_slope[j] = 0.;
        }
        /* I need to scale the slopes to make them less dependant
           on the number of patterns in the data base */
        if(denom != 0.){
        output->false_slope[j] *= rocpp->nbins/(float)output->nhit;
        output->true_slope[j] *= rocpp->nbins/(float)output->nfalse;
        }
    }
    /* do smoothing on the slopes */
    for(j = 0; j < rocpp->nbins; j++){
        float false_sum, true_sum, n;
        int k;
/*      static float window[10] = {1., 1., 1., 1., 1., 1., 1., 1., 1., 1};*/
        static float window[10] = {10., 9., 8., 7., 6., 5., 4., 3., 2., 1};

        false_sum = output->false_slope[j]*window[0];
        true_sum = output->true_slope[j]*window[0];
        n = window[0];
        for(k = 1; k < 10; k++){
        if(j-k >= 0){
            false_sum += output->false_slope[j-k]*window[k];
            true_sum += output->true_slope[j-k]*window[k];
            n += window[k];
        }
        if(j+k < rocpp->nbins){
            false_sum += output->false_slope[j+k]*window[k];
            true_sum += output->true_slope[j+k]*window[k];
            n += window[k];
        }
        }
        output->false_slope[j] = false_sum/n;
        output->true_slope[j] = true_sum/n;
    }
    }
    roc_init_next_roc();
}

void
roc_print_values(int target, FILE *fp)
{
    ROC_OUTPUT *output;
    ROC_BIN *bin;
    int j;

    if(target < 0 || target >= rocpp->noutputs) return;
    if(rocpp->cur_roc < 0) return;
    output = &(rocpp->roc[rocpp->cur_roc][target]);
    for(j = 0; j < rocpp->nbins; j++){
    bin = &(output->bins[j]);
    if(output->nfalse != 0)
      fprintf(fp, "%f ", 100.*(float)(output->nfalse -
                      (bin->nfalse_above + bin->nfalse))
          /(float)output->nfalse);
    else
      fprintf(fp, "%f ", 0.);
    if(output->nhit != 0)
      fprintf(fp, "%f ", 100.*(float)bin->nhit_below/(float)output->nhit);
    else
      fprintf(fp, "%f ", 0.);
    fprintf(fp, "\n");
    }
}

void
roc_print_slopes(int target, FILE *fp)
{
    ROC_OUTPUT *output;
    int j;

    if(target < 0 || target >= rocpp->noutputs) return;
    if(rocpp->cur_roc < 0) return;
    output = &(rocpp->roc[rocpp->cur_roc][target]);
    for(j = 0; j < rocpp->nbins; j++){
    fprintf(fp, "%d %f %f\n",
        j,
        output->false_slope[j], /* false down */
        output->true_slope[j]); /* true down */
    }
}

/* free the roc structures */
void
roc_cost_free(void)
{
    int i, j;
    for(i = 0; i < NROCS; i++){
    for(j = 0; j < rocpp->noutputs; j++){
        deallocate_structvector(rocpp->roc[i][j].bins);
        deallocate_vector(rocpp->roc[i][j].false_slope);
        deallocate_vector(rocpp->roc[i][j].true_slope);
    }
    }
    deallocate_structmatrix((void **)rocpp->roc);
    deallocate_vector(rocpp->cost_vector);
    nn_free((char *)rocpp, "for roc cost structure");
    rocpp = NULL;
}

float
roc_cost_get_area(int target)
{
    if(target < 0 || target >= rocpp->noutputs) return 0.;
    if(rocpp->cur_roc < 0) return 0.;
    return rocpp->roc[rocpp->cur_roc][target].area;
}
