/*
* NAME:         roc_pack.c
* PURPOSE:      package to compute roc's for a set of data
*
* USAGE: 
* RETURN VALUE:  NONE
*
* EXAMPLE:
*
* GLOBALS: 
*
* DESCRIPTION:
*    Functions in this file
*       roc_build               Creates roc structure that holds the data
*                                 for an roc model
*       roc_free                Destroys the roc module
*       roc_param               Returns a pointer to the roc model
*       roc_batch_train         Accumulates the roc w/ given pattern
*       roc_batch_compute       Performs roc computation on the model for a
*                                 specific class
*       roc_print_table         Prints tables of detections and false alarms
*                                 for the last computed roc
*       roc_get_percent_fom     Computes the figure of merit score for an
*                                 roc model
*
* SEE ALSO: 
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/roc_pack.c,v 11.9 2001/08/23 20:14:48 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/roc_pack.c,v $
* Created: January 11, 1995 lck     MIT LINCOLN LABORATORY
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"
#include "get_subs.h"

/* static roc model pointer array */
static int nrocs=0;
static roc **roc_models=NULL;
#ifndef ABS
#define ABS(x) ((x < 0) ? -1*x : x )
#endif

static void roc_print_table_row(FILE *fp, roc *aroc, int i);
static void roc_print_interpolated_row(FILE *fp, roc *aroc, int i, float fa);
static void roc_print_table_head(FILE *fp, roc *aroc);
/********************************************************************
* NAME: roc_build()
* DESCRIPTION:
*        Allocate storage and initialize object for an roc model
*
***************************************************************************/
int
roc_build(int nclasses, int npatterns)
{
    roc *aroc;
    int i, new_model;

    new_model = model_get_number((void ***)&roc_models, &nrocs, sizeof(roc));
    aroc= roc_models[new_model];

    /* clear memory and initialize attributes for the new roc */
    aroc->nallocpatterns= npatterns;
    aroc->npatterns= 0;
    aroc->nclasses= nclasses;
    aroc->output_val= allocate_matrix(npatterns,nclasses+1);
    aroc->class= -1;
    aroc->reject= 0.0;
    aroc->reject_flag= allocate_ivector(npatterns);
    for (i=0;i<npatterns;i++) aroc->reject_flag[i]=0;
    aroc->ninclass= allocate_ivector(nclasses);
    for (i=0;i<nclasses;i++) aroc->ninclass[i]=0;
    aroc->class_index= allocate_ivector(npatterns);

    /* table data */
    aroc->table_length= 0;
    aroc->npat= allocate_ivector(npatterns+2);
    aroc->ndr= allocate_ivector(npatterns+2);
    aroc->nfa= allocate_ivector(npatterns+2);
    aroc->threshold= allocate_vector(npatterns+2);
    aroc->false_alarm= allocate_vector(npatterns+2);
    aroc->detect_rate= allocate_vector(npatterns+2);

  return(new_model);
}
/********************************************************************
* NAME: roc_init()
* DESCRIPTION:
*        Allocate storage and initialize object for an roc model
*
***************************************************************************/
void
roc_init(int model_number)
{
  roc *aroc;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL)){
      warning("Invalid roc model number %d",model_number);
      return;
  }
  aroc= roc_models[model_number];

  aroc->npatterns = 0;
  aroc->npatterns= 0;
  aroc->class= -1;
  aroc->reject= 0.0;
  aroc->table_length= 0;

}

static void roc_expand_model(roc *aroc){
  int i, length, more;
  length = aroc->nallocpatterns;
  more = aroc->nallocpatterns += 1024;

  /* clear memory and initialize attributes for the new roc */
  aroc->output_val= reallocate_matrix(aroc->output_val, length,
                      aroc->nclasses+1, more);
  aroc->reject_flag= reallocate_ivector(aroc->reject_flag, length, more);
  for (i=length;i<more;i++) aroc->reject_flag[i]=0;
  aroc->class_index= reallocate_ivector(aroc->class_index, length, more);

  /* table data */
  aroc->table_length= 0;
  aroc->npat= reallocate_ivector(aroc->npat, length+2, more+2);
  aroc->ndr= reallocate_ivector(aroc->ndr, length+2, more+2);
  aroc->nfa= reallocate_ivector(aroc->nfa, length+2, more+2);
  aroc->threshold= reallocate_vector(aroc->threshold, length+2, more+2);
  aroc->false_alarm= reallocate_vector(aroc->false_alarm, length+2, more+2);
  aroc->detect_rate= reallocate_vector(aroc->detect_rate, length+2, more+2);
}

/**************************************************************************
* NAME: roc_free()
* DESCRIPTION:
*        Destroy an roc object
*
***************************************************************************/
void
roc_free(int model_number)
{
  roc *aroc;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL)){
      warning("Invalid roc model number %d",model_number);
      return;
  }
  /* free roc individually */
  aroc= roc_models[model_number];

  if (aroc->reject_flag!=NULL) deallocate_ivector(aroc->reject_flag);
  if (aroc->output_val!=NULL) deallocate_matrix(aroc->output_val);
  if (aroc->ninclass!=NULL) deallocate_ivector(aroc->ninclass);
  if (aroc->class_index!=NULL) deallocate_ivector(aroc->class_index);
  if (aroc->npat!=NULL) deallocate_ivector(aroc->npat);
  if (aroc->ndr!=NULL) deallocate_ivector(aroc->ndr);
  if (aroc->nfa!=NULL) deallocate_ivector(aroc->nfa);
  if (aroc->threshold!=NULL) deallocate_vector(aroc->threshold);
  if (aroc->false_alarm!=NULL) deallocate_vector(aroc->false_alarm);
  if (aroc->detect_rate!=NULL) deallocate_vector(aroc->detect_rate);

  nn_free((char *)aroc, "for freeing roc model");
  roc_models[model_number] = NULL;
}


/**************************************************************************
* NAME: roc_param()
* DESCRIPTION:
*        Returns a pointer to the roc model
*
***************************************************************************/
void
roc_param(int model_number, roc **model_pointer)
{
  roc *aroc;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL))
    { warning("Invalid model number %d",model_number); 
      *model_pointer=NULL; return; }
  aroc= roc_models[model_number];

  *model_pointer= aroc;
}


/**************************************************************************
* NAME: roc_log_train()
* DESCRIPTION:
*        Every call accumulates the result for a batch roc curve computation
*
***************************************************************************/
void
roc_log_train(int model_number, int class, float *output_vector)
{
  roc *aroc;
  int i, badInput = 0;
  float sum = 0.;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL))
    { warning("Invalid model number %d",model_number); return; }
  aroc= roc_models[model_number];

  if (aroc->npatterns >= aroc->nallocpatterns){
    roc_expand_model(aroc);
  }

  /* add the output_val to the list */
  for (i=0;i<aroc->nclasses;i++){
    aroc->output_val[aroc->npatterns][i]= output_vector[i];
    sum = logadd(output_vector[i], sum);
  }
  /* normalize the outputs to sum to one, to make sure they are posteriors */
  for (i=0;i<aroc->nclasses;i++){
    aroc->output_val[aroc->npatterns][i] -= sum;
    /*    aroc->output_val[aroc->npatterns][i] =
      nn_exp10(aroc->output_val[aroc->npatterns][i]);*/
    if(isnan((double)aroc->output_val[aroc->npatterns][i])) badInput = 1;
  }

  /* add the class to the list */
  aroc->output_val[aroc->npatterns][aroc->nclasses]= (float)class;

  if(!badInput) aroc->npatterns++;
}

/**************************************************************************
* NAME: roc_batch_train()
* DESCRIPTION:
*        Every call accumulates the result for a batch roc curve computation
*
***************************************************************************/
void
roc_batch_train(int model_number, int class, float *output_vector)
{
  roc *aroc;
  int i, badInput = 0;
  float sum = 0.;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL))
    { warning("Invalid model number %d",model_number); return; }
  aroc= roc_models[model_number];

  if (aroc->npatterns >= aroc->nallocpatterns){
    roc_expand_model(aroc);
  }

  /* add the output_val to the list */
  for (i=0;i<aroc->nclasses;i++){
    aroc->output_val[aroc->npatterns][i]= output_vector[i];
    sum += output_vector[i];
  }
  /* normalize the outputs to sum to one, to make sure they are posteriors */
  for (i=0;i<aroc->nclasses;i++){
    aroc->output_val[aroc->npatterns][i] /= sum;
    if(isnan((double)aroc->output_val[aroc->npatterns][i])) badInput = 1;
  }

  /* add the class to the list */
  aroc->output_val[aroc->npatterns][aroc->nclasses]= (float)class;

  if(!badInput) aroc->npatterns++;
}


/**************************************************************************
* NAME: roc_batch_compute()
* DESCRIPTION:
*        Computes roc curve information based on patterns accumulated by
*          roc_batch_train().  The computation is determined by a target
*          class and the fractional rejection rate [0,1].
*
***************************************************************************/
void
roc_batch_compute(int model_number, int target_class, float reject)
{
#define DETECTION 0
#define FALSE_ALARM 1
  roc *aroc;
  int i,j,npat, left_npat;
  int nfa=0, ndr=0, tot_ndetect, tot_nfalse;
  float threshold, last_threshold;
  int this_type, last_type;
  int mixed_type, mixed_ndr, mixed_nfa;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL))
    { warning("Invalid roc model number %d",model_number); return; }
  aroc= roc_models[model_number];

  if (target_class>= aroc->nclasses)
    { warning("Invalid roc target_class %d: (must be between 0 and %d)",
          target_class,aroc->nclasses-1); return; }

  if (reject < 0.0 || reject > 1.0)
    { warning("Invalid roc rejection value %f (must be between 0 and 1)",
          reject); return; }

  aroc->class= target_class;
  aroc->reject= reject;

  /* find predicted classes from output vectors */
  npat= aroc->npatterns;
  for (i=0;i<npat;i++) {
      float max= aroc->output_val[i][0];
      int max_index=0;
      for (j=1;j<aroc->nclasses;j++)
    if (aroc->output_val[i][j]>max) {
        max= aroc->output_val[i][j];
        max_index= j;
    }
      aroc->class_index[i]= max_index;
  }

  /* sort the output_val matrix by the selected class (descending order) */
  sort_matrix_rows(aroc->npatterns,aroc->nclasses+1,
           target_class,aroc->output_val);

  /* set reject_flag for the nR patterns with the lowest "highest" output */
  if (reject > 0){
    float *max_output= allocate_vector(npat);
    int *index= allocate_iindex(npat);
    int nrejects;

    /* sort the patterns by their highest outputs */
    for (i=0;i<npat;i++)
      max_output[i]= aroc->output_val[i][aroc->class_index[i]];
    heap_sort(npat,max_output,index); /* ascending order */

    /* mark lowest patterns as rejected */
    nrejects= (int)(npat*reject+0.5);
    for (i=0;i<nrejects;i++)
      aroc->reject_flag[index[i]]= 1;
    deallocate_vector(max_output);
    deallocate_ivector(index);
  }

  /* for non-rejected portion of the data: */
  /* compute total number in each class for the remaining data set */
  for (i=0;i<aroc->nclasses;i++)
    aroc->ninclass[i]=0;
  left_npat=0;
  for (i=0;i<npat;i++)
    if (!aroc->reject_flag[i]) {
    aroc->ninclass[(int)(aroc->output_val[i][aroc->nclasses])]++;
    left_npat++;
    }
  tot_ndetect = aroc->ninclass[target_class];
  tot_nfalse = left_npat - tot_ndetect;
  
  aroc->npatterns_left = left_npat;

  /* make the first entry of the table be 100% detect, 100% false alarm,
     threshold of zero */
  aroc->table_length = 0;

  aroc->threshold[aroc->table_length] = 0.;
  aroc->npat[aroc->table_length] = left_npat;
  aroc->ndr[aroc->table_length]= tot_ndetect;
  aroc->nfa[aroc->table_length]= tot_nfalse;
  aroc->false_alarm[aroc->table_length]= 100.;
  aroc->detect_rate[aroc->table_length]= 100.;
  aroc->table_length++;

  /* find other entries in the table */
  /* I have tried several different methods for generating this table.

     The original method creates a table entry for each time the threshold
     changes. This is often too often. If a string of uninterupted
     false alarms comes in, you only need a point at the start and stop
     of the string to describe the straight line.

     The second method addressed this problem by ignoring the threshold
     and looking only at the string of false alarms and detections to
     determine the shape of the curve. That is very easy to do and it
     is very easy to find the area of things later. Problem is, if you
     get a string of mixed detections and false alarms all at the same
     threshold value, you get this jagged line, or worse, all the detections
     followed by all the false alarms and a very warped view of the
     true curve and area. (The degenerate case is a sorted list of patterns
     with all the same output value. The curve should be the diagonal on
     the 50% line. This method of calucating things gives either 100%
     area or 0%, depending on whether the detections or false alarms came
     first.

     The current method tries to have it both ways. I want to generate
     diagonal lines where the threshold doesn't change and there are
     mixed false alarms and detections and I want to only generate points for
     the ends of lines where the threshold changes but the type of data
     doesn't (all false alarms or all hits.) First and last, I output
     a table entry for (0,0) and (100,100). Initialize the type and threshold
     information for the first point that actually is on the chart (not
     rejected.) Add table entries where the threshold changes and we are
     generating a diagonal line, where the threshold changes, we are drawing
     a straight line, and the type changes, and if we discover that we
     should be drawing a diagonal line because the threshold didn't change
     but the type did, add a point at the start of this threshold level.
     This algorithm generates duplicate table entries where a diagonal
     meets a diagonal, so I check for duplicates and eliminate them as
     I go.
     */

  last_type=this_type =  -1;
  last_threshold = 0;
  ndr = tot_ndetect;
  nfa = tot_nfalse;
  mixed_ndr = ndr;
  mixed_nfa = nfa;
  mixed_type = 0;
  for (i=npat-1;i>=0;i--) {
      if(aroc->reject_flag[i]) continue;

      /* limit the number of significant digits */
#ifdef elliot
      threshold = aroc->output_val[i][target_class];
#else
      threshold= (float)
    ((int)(aroc->output_val[i][target_class]*1000))/1000.0;
#endif
      
      /* take the previous point out of the
     detection and false alarm numbers */
      if(aroc->output_val[i][aroc->nclasses] == target_class){
      ndr--;
      this_type = DETECTION;
      } else {
      nfa--;
      this_type = FALSE_ALARM;
      }
      if(last_type == -1){ last_threshold = threshold; last_type = this_type;}

      if(last_threshold != threshold){
      if (mixed_type || (this_type != last_type)) {
          /* put in new value */
          aroc->threshold[aroc->table_length]= threshold;
          aroc->npat[aroc->table_length]= ndr+nfa+1;
          aroc->ndr[aroc->table_length]=
        ndr+((this_type==DETECTION)? 1: 0);
          aroc->nfa[aroc->table_length]=
        nfa+((this_type==FALSE_ALARM)? 1: 0);
          aroc->false_alarm[aroc->table_length]= (float)
        aroc->nfa[aroc->table_length]/
          ((float)(tot_nfalse+ROC_TINY))* 100.0;
          aroc->detect_rate[aroc->table_length]= (float)
        aroc->ndr[aroc->table_length]/
          ((float)(tot_ndetect+ROC_TINY))* 100.0;
      
          /* check that we needed to do this. If we have
         already entered a point here, we don't need to do it again */
          if((aroc->nfa[aroc->table_length] !=
          aroc->nfa[aroc->table_length -1]) ||
         (aroc->ndr[aroc->table_length] !=
          aroc->ndr[aroc->table_length-1]))
        aroc->table_length++;

          last_type= this_type;
          mixed_type = 0;
      }
      /* keep track of the threshold change, in case the line
         ahead is diagonal */
      last_threshold = threshold;
      mixed_ndr = ndr+((this_type==DETECTION)? 1: 0);
      mixed_nfa = nfa+((this_type==FALSE_ALARM)? 1: 0);
      } else {
      if (this_type != last_type){
          mixed_type = 1;
          /* start a diagonal line where the threshold last changed. */
          
          aroc->threshold[aroc->table_length]= threshold;
          aroc->npat[aroc->table_length]= mixed_ndr+mixed_nfa;
          aroc->ndr[aroc->table_length]= mixed_ndr;
          aroc->nfa[aroc->table_length]= mixed_nfa;
          aroc->false_alarm[aroc->table_length]= (float)
        aroc->nfa[aroc->table_length]/
          ((float)(tot_nfalse+ROC_TINY))* 100.0;
          aroc->detect_rate[aroc->table_length]= (float)
        aroc->ndr[aroc->table_length]/
          ((float)(tot_ndetect+ROC_TINY))* 100.0;

          /* check that we needed to do this. If we have
         two diagonals in a row, we already have entered this
         endpoint */
          if((aroc->nfa[aroc->table_length] !=
          aroc->nfa[aroc->table_length -1]) ||
         (aroc->ndr[aroc->table_length] !=
          aroc->ndr[aroc->table_length-1]))
        aroc->table_length++;

          last_type = this_type;
      }
      }
  }
  /* make the last entry of the table be 0% detect, 0% false alarm,
     threshold of 1. */
  aroc->threshold[aroc->table_length] = 1.;
  aroc->npat[aroc->table_length] = 0;
  aroc->ndr[aroc->table_length]= 0;
  aroc->nfa[aroc->table_length]= 0;
  aroc->false_alarm[aroc->table_length]= 0.;
  aroc->detect_rate[aroc->table_length]= 0.;
  aroc->table_length++;
#undef DETECTION
#undef FALSE_ALARM
}


/**************************************************************************
* NAME: roc_print_table()
* DESCRIPTION:
*        Outputs roc curve information based on the target class and the 
*          reject rate.  If the currently stored information in the model 
*          is not of the same class and/or reject rate, this call will 
*          trigger a call to roc_batch_compute() to compute the correct 
*          information.
*        The amount of output information is controlled by verbosity flag:
*          0: no output
*          1: output average fom between 0 and 100% false alarm
*          2: output a column of false alarm and a column of detect rate
*          3: output full table
*        Computes roc curve information based on patterns accumulated by
*          roc_batch_train().  The computation is determined by a target
*          class and the fractional rejection rate [0,1].
*
***************************************************************************/
#define TABLE_WIDTH 6
void
roc_print_table(FILE *fp, int model_number, int target_class, float reject,
        int verbosity)
{
  roc *aroc;
  int i;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL))
    { warning("Invalid model number %d",model_number); return; }
  aroc= roc_models[model_number];

  if (aroc->class!=target_class || aroc->reject!=reject)
    roc_batch_compute(model_number,target_class,reject);
  if(verbosity > 1){
      /* full reporting */
      roc_print_table_head(fp, aroc);
      for (i=aroc->table_length-1;i >=0; i--) {
      roc_print_table_row(fp, aroc, i);
      }
  }
  fprintf(fp," TARGET %d ROC AREA = %f\n",target_class,
      roc_get_percent_fom(model_number, target_class, reject, 0.0, 1.));
  return;
}
static void
roc_print_table_row(FILE *fp, roc *aroc, int i)
{
    fprintf(fp,"%5d",aroc->npat[i]);
    fprintf(fp,"%12d",aroc->ndr[i]);
    fprintf(fp,"%14.3f",aroc->detect_rate[i]);
    fprintf(fp,"%13d",aroc->nfa[i]);
    fprintf(fp,"%12.3f",aroc->false_alarm[i]);
    fprintf(fp,"%12.3f",aroc->threshold[i]);
    fprintf(fp,"\n");
}
#define INTERPOLATE(x1, x2, r)  ((x1) + ((x2) - (x1))*(r))
static void
roc_print_interpolated_row(FILE *fp, roc *aroc, int i, float fa)
{
    float ratio = 1., temp;

    if(i > 0 && i < aroc->table_length){
    if(aroc->false_alarm[i] != aroc->false_alarm[i+1]){
        ratio = fa - aroc->false_alarm[i+1];
        ratio /= aroc->false_alarm[i]-aroc->false_alarm[i+1];
    } else
      /* if the false alarm rates are equal, we are on a straight
         vertical line between two detection values */
      ratio = .5;
    }
    /* interpolation between table entries */
    /* entry is between i and i-1 */
    temp = (i <= 0) ? aroc->npatterns_left :
      INTERPOLATE(aroc->npat[i+1], aroc->npat[i], ratio);
    fprintf(fp,"%5d", (int)temp);
   
    temp = (i <= 0) ? aroc->ndr[0] :
      INTERPOLATE(aroc->ndr[i+1], aroc->ndr[i], ratio);
    fprintf(fp,"%12d",(int)temp);
    
    temp = (i <= 0) ? 100 :
      INTERPOLATE(aroc->detect_rate[i+1], aroc->detect_rate[i], ratio);
    fprintf(fp,"%14.3f",temp);

    temp = (i <= 0) ? aroc->nfa[0] :
      INTERPOLATE(aroc->nfa[i+1], aroc->nfa[i], ratio);
    fprintf(fp,"%13d",(int)temp);
    fprintf(fp,"%12.3f",fa);

    temp = (i <= 0) ? 0. :
      INTERPOLATE(aroc->threshold[i+1], aroc->threshold[i], ratio);
    fprintf(fp,"%12.3f",temp);
    fprintf(fp,"\n");
}
/**************************************************************************
* NAME: roc_print_sampled_table()
* DESCRIPTION:
*        Outputs roc curve information based on the target class and the 
*          reject rate.  If the currently stored information in the model 
*          is not of the same class and/or reject rate, this call will 
*          trigger a call to roc_batch_compute() to compute the correct 
*          information.
*        The amount of output information is controlled by verbosity flag:
*          0: no output
*          1: output average fom between 0 and 100% false alarm
*          2: output a column of false alarm and a column of detect rate
*          3: output full table
*        Computes roc curve information based on patterns accumulated by
*          roc_batch_train().  The computation is determined by a target
*          class and the fractional rejection rate [0,1].
*
***************************************************************************/
#define TABLE_WIDTH 6
void
roc_print_sampled_table(FILE *fp, int model_number, int target_class,
            float reject, int verbosity, float table_begin,
            float table_end, float table_step)
{
  roc *aroc;
  float fa;
  int i, lasti;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL))
    { warning("Invalid model number %d",model_number); return; }
  aroc= roc_models[model_number];

  if (aroc->class!=target_class || aroc->reject!=reject)
    roc_batch_compute(model_number,target_class,reject);
  if(verbosity > 1){
      roc_print_table_head(fp, aroc);
      for(fa = table_begin, i = aroc->table_length-1, lasti = -1;
      fa <= table_end && i >= 0; fa += table_step){
      /* for each false alarm level, find the closest table entry >=
         to the level and print a table row for it */
      for(; i >= 0; i--)
        if(aroc->false_alarm[i] >= fa) break;
      if(lasti != i){
          if(fa == aroc->false_alarm[i]){
          roc_print_table_row(fp, aroc, i);
          } else {
          roc_print_interpolated_row(fp, aroc, i, fa);
          }
          lasti = i;
      }
      }
  }
  /* always print out the ROC AREA */
  fprintf(fp," TARGET %d ROC AREA = %f\n", target_class,
      roc_get_percent_fom(model_number, target_class, reject,
                  0.0, 1.));
  return;
}
static void
roc_print_table_head(FILE *fp, roc *aroc)
{
    int i;
    static  char *columns[TABLE_WIDTH] = {
    "# PATTERNS ","# CORRECT ","% CORRECT"," # FALSE  ",
    "% FALSE_ALARM","THRESHOLD"};
      fprintf(fp, "---------------------------------");
      fprintf(fp, "-------------------------------\n");
      fprintf(fp,"target class = %d          reject = %.2f%%\n",
          aroc->class, aroc->reject*100.);
      fprintf(fp,"number of patterns from target class = %d\n",
          aroc->ninclass[aroc->class]);
      fprintf(fp, "---------------------------------");
      fprintf(fp, "-------------------------------\n");
      for(i = 0; i < 2; i++) fprintf(fp, "%11s", columns[i]);
      fprintf(fp, "%13s", columns[i++]);
      fprintf(fp, "%11s", columns[i++]);
      fprintf(fp, "%13s", columns[i++]);
      for(; i < TABLE_WIDTH; i++) fprintf(fp, "%11s", columns[i]);

      fprintf(fp, "\n");
      for(i = 0; i < TABLE_WIDTH; i++) fprintf(fp, " ---------- ");
      fprintf(fp, "\n"); fflush(fp);
}
#ifdef notdef
/* This is the old way to do fom calculations, kept around because it
   may be useful in the future 
 */
/**************************************************************************
* NAME: roc_get_percent_fom()
* DESCRIPTION:
*        Returns the figure of merit calculation based on averaging the 
*          detection rate for a given number of false_alarm rates.
*        The false_alarm sampling is determined based on <false_alarm_min>,
*          <false_alarm_max> . NOTE: THESE LIMITS 
*          should be between 0.0 and 1.0.
*        If the currently stored information in the model is not of the 
*          same class and/or reject rate, this call will trigger a call to 
*          roc_batch_compute() to compute the correct information.
*
***************************************************************************/

float
roc_get_percent_fom2(int model_number, int target_class, float reject,
             float false_alarm_min, float false_alarm_max,
             float false_alarm_step)
{
  float fa,min,max,f;
  int i,index_min,index_max, index,n;
  float dr;
  int go;
  roc *aroc;
  float *dval;

  if ((model_number>=nrocs || model_number < 0) ||
      (roc_models[model_number] == NULL))
    { warning("Invalid model number %d",model_number); return  0.; }
  aroc= roc_models[model_number];

  if (aroc->class!=target_class || aroc->reject!=reject)
    roc_batch_compute(model_number,target_class,reject);

  /* sanity check for limits */
  if (false_alarm_min > 1.0 || false_alarm_max > 1.0 ||
      false_alarm_min < 0.0 || false_alarm_max < 0.0 ||
      false_alarm_step < 0.0 || false_alarm_step > 1.0)
    { warning("Invalid Limits in ROC_GET_PERCENT_FOM(): %f %f %f",
          false_alarm_min,false_alarm_max,false_alarm_step); return(0.0); }
  if (false_alarm_min > false_alarm_max)
    {
      warning("False Alarm Minimum %f is Greater Than Maximum %f: Switched...",
          false_alarm_min,false_alarm_max);
      f= false_alarm_min;
      false_alarm_min= false_alarm_max; false_alarm_max= f;
    }

  false_alarm_max*=100.;
  false_alarm_min*=100.;
  false_alarm_step*=100.;
  n= (false_alarm_max-false_alarm_min+false_alarm_step)/false_alarm_step;
  dval= allocate_vector(n);
  n=0; /* track number of sampling points */

  /* we don't need to to linear interpolations here, the trapezoidal
     integration does that for us */
  /* linearly interpolate the nearest entry smaller and greater than each point */
  for (fa= false_alarm_min; fa<=false_alarm_max; fa+= false_alarm_step)
    {
      /* find nearest point >= to fa */
      min=100.0; go=1;
      for (i=aroc->table_length-1;i>=0 && go; i--)
    {
      if ((f=ABS(aroc->false_alarm[i]-fa))<min)
        { min= f; index=i; }
      if (aroc->false_alarm[i]>= fa) go=0;
    }
      index_min= index;

      /* find nearest point <= to fa */
      max=100.0; go=1;
      for (i=0;i<aroc->table_length && go; i++)
    {
      if ((f=ABS(aroc->false_alarm[i]-fa))<min)
        { min= f; index=i; }
      if (aroc->false_alarm[i]< fa) go=0;
    }
      index_max= index;

      /* now, interpolate between the two points */
      if (aroc->false_alarm[index_min]==aroc->false_alarm[index_max])
    /* same point for lower and upper limit: assign to be the same */
    f= aroc->detect_rate[index_min];
      else
    {
      f= (aroc->detect_rate[index_min]+
          ((aroc->detect_rate[index_max]-aroc->detect_rate[index_min])*
           (fa - aroc->false_alarm[index_min])/
           (aroc->false_alarm[index_max]-aroc->false_alarm[index_min])));
    }
      dval[n]= f;
      n++;
    }

  /* use trapedzoidal integration */
  dr=0.0;
  for (i=0;i<n-1;i++)
    dr += (dval[i]+dval[i+1]);
  dr*= 0.5*false_alarm_step;
  if (dr==0.0) dr=dval[0];

  /* divide out max-min to make this area the average output */
  dr /= (false_alarm_max - false_alarm_min);

  deallocate_vector(dval);

  return(dr);
}
#endif

/**************************************************************************
* NAME: roc_get_percent_fom()
* DESCRIPTION:
*        Returns the figure of merit calculation based on averaging the 
*          detection rate for a given number of false_alarm rates.
*        The false_alarm sampling is determined based on <false_alarm_min>,
*          <false_alarm_max> . NOTE: THESE LIMITS 
*          should be between 0.0 and 1.0.
*        If the currently stored information in the model is not of the 
*          same class and/or reject rate, this call will trigger a call to 
*          roc_batch_compute() to compute the correct information.
*
*   This is the third version of this routine. The first one
*   sampled the ROC table to do a trapezoidal integration to determine
*   the area. That area was consistantly large because ROC curves are
*   concave up. The second area calculation took an area from the
*   data. This has the problem that it ignores regions where the
*   threshold doesn't change but both the number of false alarms and
*   of detections do. This is discussed above in roc_batch_compute.
*   The third area calcuation does an exact integration, using the
*   the endpoints listed on the ROC table as describing the tops
*   of trapezoids. There is some wasted calculation where those
*   tops really describe rectangles. I've left the old calcuations
*   around, in case I need some inspiration when doing this a fourth
*   time because some other obscure problem arose.
*
***************************************************************************/
float
roc_get_percent_fom(int model_number, int target_class, float reject,
            float false_alarm_min, float false_alarm_max)
{
    roc *aroc;
#ifdef notdef
    int ndr = 0, nfa = 0, ndetect, nfalse, fa_min_num, fa_max_num;
#endif
    int i;
    float area = 0, f, d;
    
    if ((model_number>=nrocs || model_number < 0) ||
    (roc_models[model_number] == NULL))
      { warning("Invalid model number %d",model_number); return 0.; }
    aroc= roc_models[model_number];
    
    /* sanity check for limits */
    if (false_alarm_min > 1.0 || false_alarm_max > 1.0 ||
    false_alarm_min < 0.0 || false_alarm_max < 0.0)
      { warning("Invalid Limits in ROC_GET_PERCENT_FOM(): %f %f",
        false_alarm_min,false_alarm_max); return(0.0); }
    if (false_alarm_min > false_alarm_max) {
    warning("False Alarm Minimum %f is Greater Than Maximum %f: Switched...",
        false_alarm_min,false_alarm_max);
    f= false_alarm_min;
    false_alarm_min= false_alarm_max; false_alarm_max= f;
    }
    
    false_alarm_max*=100.;
    false_alarm_min*=100.;
    
    if (aroc->class!=target_class || aroc->reject!=reject)
      roc_batch_compute(model_number,target_class,reject);
    
    /* pass the entries up to false max */
    for (i = 0; i < aroc->table_length; i++){
    if (false_alarm_max >= aroc->false_alarm[i]) break;
    }
    /* add in the partial for this entry */
    if(i > 0 && false_alarm_max != aroc->false_alarm[i]){
    d = aroc->detect_rate[i]+
      (aroc->detect_rate[i-1] - aroc->detect_rate[i])*
        (false_alarm_max - aroc->false_alarm[i])/
          (aroc->false_alarm[i-1] - aroc->false_alarm[i]);
    area += (aroc->detect_rate[i]+d)/2. *
      (false_alarm_max - aroc->false_alarm[i]);
    }
    /* add in all the entries that are wholly included */
    for (i++; i < aroc->table_length; i++){
    if(false_alarm_min >= aroc->false_alarm[i]) break;
    
    area += (aroc->detect_rate[i] + aroc->detect_rate[i-1])/2. *
      (aroc->false_alarm[i-1] - aroc->false_alarm[i]);
    }
    /* add in the partial for the final entry included */
    if(false_alarm_min == aroc->false_alarm[i]){
    area += (aroc->detect_rate[i] + aroc->detect_rate[i-1])/2. *
      (aroc->false_alarm[i-1] - aroc->false_alarm[i]);
    } else {
    d = aroc->detect_rate[i]+
      (aroc->detect_rate[i-1] - aroc->detect_rate[i])*
        (false_alarm_min - aroc->false_alarm[i])/
          (aroc->false_alarm[i-1] - aroc->false_alarm[i]);
    area += (aroc->detect_rate[i-1]+d)/2. *
      (aroc->false_alarm[i-1] - false_alarm_min);
    }
    
#ifdef notdef
    ndetect = aroc->ninclass[target_class];
    nfalse = aroc->npatterns_left - ndetect;
    fa_min_num = false_alarm_min * nfalse;
    fa_max_num = false_alarm_max * nfalse;

    /* count up the detects and falses
       because each pattern changes ndr or nfa by exactly one, 
       we don't need to use any fancy integrations. just sum up
       ndr (y) each time nfa changes (x) */
    /* skip patterns before min pattern */
    for (i = 0; i < aroc->npatterns && nfa < fa_min_num; i++) {
    if(!aroc->reject_flag[i]){
        if(aroc->output_val[i][aroc->nclasses]==target_class){
        ndr++;
        } else {
        nfa++;
        }
    }
    }
    /* fa_min_num == nfa */
    area = (1. - (nfalse * false_alarm_min - fa_min_num)) * ndr;

    for (; i < aroc->npatterns && nfa < fa_max_num; i++) {
    if(!aroc->reject_flag[i]){
        if(aroc->output_val[i][aroc->nclasses]==target_class){
        ndr++;
        } else {
        nfa++;
        area += (float)ndr;
        }
    }
    }
    /* fa_max_num == nfa */
    area += (nfalse * false_alarm_max - fa_max_num) * ndr;

    /* make this area an area on the percent graph */
    area /= (float)nfalse+ROC_TINY;
    area /= (float)ndetect+ROC_TINY;

    /* get the scale right */
    area *= 100.;

#endif
    /* divide out max-min to make this the average height in
       percent probability */
    area /= (false_alarm_max - false_alarm_min);
    return(area);
}


/**************************************************************************
* NAME: roc_get_area()
* DESCRIPTION:
*        Returns the area under the roc graph normalized to [0, 1].
*        If the currently stored information in the model is not of the 
*          same class and/or reject rate, this call will trigger a call to 
*          roc_batch_compute() to compute the correct information.
*
***************************************************************************/

float
roc_get_area(int model_number, int target_class, float reject)
{
  float f;

  f= roc_get_percent_fom(model_number, target_class, reject, 0.0, 1.0
#ifdef notdef
             ,0.01
#endif
             );
  return(f/100.);
}

float
roc_get_eer(FILE *fp, int model_number, int target_class, float reject)
{
  roc *aroc;
  int tot_ndetect, tot_nfalse, nfa, nmi, i;
  float threshold, probMi, probFa, diff, oldDiff;

  if((model_number >= nrocs || model_number < 0) ||
     (roc_models[model_number] == NULL))
    { warning("Invalid model number %d", model_number); return 0.;}
  aroc = roc_models[model_number];

  /* sort all the data */
  if(aroc->class != target_class || aroc->reject != reject)
    roc_batch_compute(model_number, target_class, reject);

  /* do a threshold through the sorted data, looking for where
     1-Pdetect = Pfalse */
  tot_ndetect = aroc->ninclass[target_class];
  tot_nfalse = aroc->npatterns_left - tot_ndetect;

  nfa = tot_nfalse;
  nmi = 0;
  oldDiff = 1.;

  for(i = aroc->npatterns-1; i >= 0; i--){
    if(aroc->reject_flag[i]) continue;
    threshold = aroc->output_val[i][target_class];

    if(aroc->output_val[i][aroc->nclasses] == target_class){
      nmi++;
    } else {
      nfa--;
    }
    probMi = (float)nmi/(float)tot_ndetect;
    probFa = (float)nfa/(float)tot_nfalse;
    if(probMi >= probFa){
      diff = probMi - probFa;
      /* either this one or the one before is closer to our actual point */
      if(diff > oldDiff){
    /* the point before had been closer to the equal error */
    if(aroc->output_val[i][aroc->nclasses] == target_class){
      nmi--;
    } else {
      nfa++;
    }
    probMi = (float)nmi/(float)tot_ndetect;
    probFa = (float)nfa/(float)tot_nfalse;

      }
      /*      printf("oldDiff %f, diff %f\n", oldDiff, diff);
          printf("probMi= %d/%d =  %f, probFa = %d/%d = %f\n",
          nmi, tot_ndetect, probMi, nfa, tot_nfalse, probFa);*/
      break;
    }
    oldDiff = probFa - probMi;
  }
  /* report the Pmiss this produces and what the threshold is here */
  if(i < 0){
    warning("Prob(miss) was never equal to Prob(false alarm)");
  } else {
    fprintf(fp, "EER = %f threshold = %g\n", probMi, threshold);
  }
  return(probMi);
}
