/***--------------------------------------------------------------------------
 * NAME:        seq_data.c
 * PURPOSE: package for giving training or cross validation patterns to
 *      sequence classifiers on request.  Takes into account random
 *      presentation order.
 *
 * DESCRIPTION:
 *
 * BUGS AND LIMITATIONS:
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/seq_data.c,v 11.3 2001/08/23 20:14:49 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/seq_data.c,v $
 * CREATED: 15-Aug-93 lck               MIT LINCOLN LABORATORY
 *---------------------------------------------------------------------------
 * Copyright 1993 MIT Lincoln Laboratory.
 * This software is not for release and should neither be copied
 * nor distributed without explicit permission of MIT Lincoln Laboratory.
 --------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include "nnlib.h"

/* GLOBAL */
#define TRUE 1
#define FALSE 0

typedef struct{
    /****************** data ********************/
    int noutputs, ninputs, nfiles, nframes;
    int nwindow_frames, *window_frames;
    float *inputs;
    DATA_BLOCK *block_list, *block;
    /****** state variables during reading ******/
    int n_read, file_num, frame_no, train_flag;
    /************ random variables **************/
    int *random_index, random_flag;
    /************ priors variables **************/
    int priors_flag, priors_npatterns;
    float *priors_list;
    int *fold_classes, *fold_indices;
    /********** cross valid variables ***********/
    int cross_valid_flag, *cross_index;
    int fold, nfolds;
    int *n_frm_in_train, *n_frm_in_test, *n_seq_in_train, *n_seq_in_test;
    char **train_masks, **test_masks, *cv_mask;
} SEQ_PATTERNS;

SEQ_PATTERNS **sp_list;
static int max_models = 0;

int seq_data_read(char *data_path, char *data_fname, int nsequences,
          int nraw, int ninputs, int noutputs, int *features,
          int normalization, char **labels, int verbosity,
          int nwindow_frames, int *window_frames)
{
    int new_model;
    SEQ_PATTERNS *spp;
    int i;

    /* allocate structure */
    new_model = model_get_number((void ***)&sp_list, &max_models,
                 sizeof(SEQ_PATTERNS));
    /* store what we know so far */
    spp = sp_list[new_model];
    spp->noutputs = noutputs;
    spp->ninputs = ninputs;
    spp->train_flag = TRUE;
    spp->n_read = 0;
    spp->random_flag = spp->priors_flag = spp->cross_valid_flag = FALSE;

    /* load data */
    if(verbosity > 1)
      printf("\nReading %s\n", path_build(data_path, data_fname));
    spp->nfiles =
      load_seq_files(data_path, data_fname, &spp->block_list, labels,
             normalization, features, ninputs, nraw, nsequences);

    /* figure out windowing */
    spp->nwindow_frames = nwindow_frames;
    if(window_frames == NULL || window_frames[0]  == 0)
      spp->window_frames = NULL;
    else{
    spp->window_frames = allocate_ivector(spp->nwindow_frames);
    for(i = 0; i < spp->nwindow_frames; i++)
      spp->window_frames[i] = window_frames[i+1];
    }
    /* allocate a vector to build the input pattern the classifier sees */
    spp->inputs = allocate_vector(spp->nwindow_frames*ninputs);
    
    /* count the total number of frames */
    spp->nframes = 0;
    for(i = 0; i < spp->nfiles; i++)
      spp->nframes += spp->block_list[i].npatterns;

    if(verbosity > 1)
      printf("Read %d frames from %d sequences -- avg %d frames/sequence\n",
         spp->nframes, spp->nfiles, (spp->nframes/spp->nfiles));

    /* initialize the block pointer for getting frames */
    spp->block = NULL;
    spp->file_num = 0;

    /* return a model number */
    return new_model;
}
int seq_data_get_nfiles(int data_model)
{
    SEQ_PATTERNS *spp;
    int npat;
    spp = sp_list[data_model];

    if(spp->priors_flag){
    if(spp->cross_valid_flag){
        /* this is incorrect for hand split cross valid */
        npat = spp->priors_npatterns/spp->nfolds;
        if(!spp->train_flag) return npat;
        else return(spp->priors_npatterns - npat);
    } else return spp->priors_npatterns;
    } else {
      if(spp->cross_valid_flag) {
      if(spp->train_flag) return spp->n_seq_in_train[spp->fold];
      else return spp->n_seq_in_test[spp->fold];
      } else return spp->nfiles;
    }
}
/* this doesn't account for the number of frames taken by the window
   width */
int seq_data_get_nframes(int data_model)
{
    SEQ_PATTERNS *spp;
    int npat;
    spp = sp_list[data_model];

    if(spp->priors_flag){
    npat = spp->nframes * (spp->priors_npatterns/spp->nfiles);
    if(spp->cross_valid_flag){
        /* this is an estimate */
        if(!spp->train_flag) return npat/spp->nfolds;
        else return((spp->nfolds - 1)*npat)/spp->nfolds;
    } else return npat;
    }
    if(spp->cross_valid_flag) {
      if(spp->train_flag) return spp->n_frm_in_train[spp->fold];
      else return spp->n_frm_in_test[spp->fold];
    } else return spp->nframes;
}
int seq_data_get_nfiles_class(int data_model, int class)
{
    int npat = 0, n;
    SEQ_PATTERNS *spp;
    spp = sp_list[data_model];

    if(spp->priors_flag){
    npat = seq_data_get_nfiles(data_model);
    npat *= spp->priors_list[class];
    } else {
    if(spp->cross_valid_flag){
        for(n = 0; n < spp->nfiles; n++){
        if(spp->cv_mask[spp->cross_index[n]] == '1')
          if(spp->block_list[n].class == class) npat++;
        }
    } else {
        for(n = 0; n < spp->nfiles; n++){
        if(spp->block_list[n].class == class) npat++;
        }
    }
    }
    return npat;
}

/*****************
 * initialize for presenting sequences in a random order each epoch
 **********************/
void seq_data_random_init(int data_model)
{
    SEQ_PATTERNS *spp;
    int i, *desired_classes;

    spp = sp_list[data_model];
    spp->random_flag = TRUE;

    if(spp->priors_flag){
    /* reinitilize priors with random turned on */
    if(!spp->cross_valid_flag){
        free_priors_data();
        desired_classes = allocate_ivector(spp->nfiles);
        for(i = 0; i < spp->nfiles; i++)
          desired_classes[i] = spp->block_list[i].class;
        init_priors_data(desired_classes, spp->noutputs, spp->nfiles,
                 spp->random_flag, spp->priors_list);
        deallocate_ivector(desired_classes);
    }
    } else {
    /* allocate space for randomized list */
    spp->random_index = allocate_ivector(spp->nfiles);
    for(i = 0; i < spp->nfiles; i++) spp->random_index[i] = i;
    }
}

/***********************
 * initialize for priors adjustment by sampling sequences
 **********************/
void seq_data_priors_init(int data_model, float *priors_list,
              int priors_npatterns)
{
    SEQ_PATTERNS *spp;
    int i, *desired_classes;
    float priors_total;

    spp = sp_list[data_model];
    spp->priors_flag = TRUE;
    spp->priors_npatterns = priors_npatterns;


    if(spp->cross_valid_flag){
    spp->fold_classes = allocate_ivector(spp->nfiles);
    spp->fold_indices = allocate_ivector(spp->nfiles);
    } else {
    /* make a list of the class entries */
    desired_classes = allocate_ivector(spp->nfiles);
    for(i = 0; i < spp->nfiles; i++)
      desired_classes[i] = spp->block_list[i].class;
    init_priors_data(desired_classes, spp->noutputs, spp->nfiles,
             spp->random_flag, priors_list);
    deallocate_ivector(desired_classes);
    }
    if(spp->random_flag)
      deallocate_ivector(spp->random_index);

    /* allocate space and copy priors */
    spp->priors_list = allocate_vector(spp->noutputs);
    for(priors_total = 0., i = 0; i < spp->noutputs; i++){
    priors_total +=  priors_list[i];
    spp->priors_list[i] = priors_list[i];
    }
    /* make sure that the priors sum to 1 */
    if(priors_total  != 1.){
    for(i = 0; i < spp->noutputs; i++)
      spp->priors_list[i] /= priors_total;
    }
}

/******************
 * initialize for cross validation mode
 *******************/
void seq_data_cv_init(int data_model, int cv_nsplits, int use_cvsplits,
              int *cv_splits, int cv_nfolds, char **cv_train_mask,
              char **cv_test_mask, int cv_random)
{
    SEQ_PATTERNS *spp;
    int i, f, *desired_classes, *index = NULL;

    spp = sp_list[data_model];

    spp->cross_valid_flag = TRUE;
    spp->nfolds = cv_nfolds;

    if(spp->priors_flag){
    /* free priors info, it must be re-done each fold */
    free_priors_data();
    spp->fold_indices = allocate_ivector(spp->nfiles);
    }

    /* set up and store cross index */
    desired_classes = allocate_ivector(spp->nfiles);

    if(cv_random){ /* randomize the patterns before building the index */
    index = allocate_ivector(spp->nfiles);
    for(i = 0; i < spp->nfiles; i++) index[i] = i;
    randomize_index_list(index, spp->nfiles);
    /* fill the list of classes */
    for(i = 0; i < spp->nfiles; i++)
      desired_classes[i] = spp->block_list[index[i]].class;
    } else {
    /* fill the list of classes */
    for(i = 0; i < spp->nfiles; i++)
      desired_classes[i] = spp->block_list[i].class;
    }
    /* build the index, either using the splits given or splitting
       the data automatically */
    spp->cross_index =
      build_cross_index(spp->noutputs, spp->nfiles, cv_nsplits, 
            desired_classes, use_cvsplits, cv_splits);

    /* undo the randomization */
    if(cv_random){
    /* copy the cross index into desired_classes */
    for(i = 0; i < spp->nfiles; i++)
      desired_classes[i] = spp->cross_index[i];
    /*un-sort the fold assignments, copying them back into cross_index*/
    for(i = 0; i < spp->nfiles; i++)
      spp->cross_index[index[i]] = desired_classes[i];
    deallocate_ivector(index);
    }
    deallocate_ivector(desired_classes);
    
    /* count patterns per fold */
    spp->n_seq_in_train = allocate_ivector(cv_nfolds);
    spp->n_seq_in_test = allocate_ivector(cv_nfolds);
    spp->n_frm_in_train = allocate_ivector(cv_nfolds);
    spp->n_frm_in_test = allocate_ivector(cv_nfolds);
    for(i = 0; i < spp->nfiles; i++)
      for(f = 0; f < cv_nfolds; f++){
      if(cv_train_mask[f][spp->cross_index[i]] == '1'){
          spp->n_seq_in_train[f]++;
          spp->n_frm_in_train[f] += spp->block_list[i].npatterns;
      }
      if(cv_test_mask[f][spp->cross_index[i]] == '1'){
          spp->n_seq_in_test[f]++;
          spp->n_frm_in_train[f] += spp->block_list[i].npatterns;
      }
      }
    spp->train_masks = cv_train_mask;
    spp->test_masks = cv_test_mask;
#ifdef DEBUG
    printf("cross validation fold assignments\n");
    printf("index class split");
    for(i = 0; i < spp->nfiles; i++){
    printf("%s\t%d\t%d\n", i, spp->block_list[i].class,
           spp->cross_index[i]);
    }
#endif /* DEBUG */
}

int seq_data_start_cv_fold(int data_model, int fold, int is_train)
{
    SEQ_PATTERNS *spp;

    spp = sp_list[data_model];

    spp->n_read = 0;
/*
    if(fold >= spp->nfolds) return 0;
*/
    spp->fold = fold;

    /* set up current mask */
    if(is_train)
      spp->cv_mask = spp->train_masks[fold];
    else
      spp->cv_mask = spp->test_masks[fold];
    spp->train_flag = is_train;
#ifdef DEBUG
    printf("cross validation train mask: %s\n", spp->cv_mask);
#endif /* DEBUG */

    if(spp->priors_flag){
    int nf, n;

    /* make up lists of the patterns in this fold and the real
       indices for those patterns */
    nf = 0;
    for(n = 0; n < spp->nfiles; n++){
        if(spp->cv_mask[spp->cross_index[n]] == '1'){
        spp->fold_classes[nf] = spp->block_list[n].class;
        spp->fold_indices[nf] = n;
        nf++;
        }
    }
    /* set up priors adjustment for this fold */
    init_priors_data(spp->fold_classes, spp->noutputs, nf,
             spp->random_flag, spp->priors_list);
    }

    /* return the number of patterns this epoch */
    if(is_train)
      return spp->n_seq_in_train[fold];
    else
      return spp->n_seq_in_test[fold];
}

void seq_data_end_cv_fold(int data_model)
{
    if(sp_list[data_model]->priors_flag) free_priors_data();
}

void seq_data_epoch_init(int data_model)
{
    SEQ_PATTERNS *spp;
    spp = sp_list[data_model];
    spp->n_read = 0;
    spp->frame_no = 0;
    spp->block = NULL;
    if(spp->random_flag && !spp->priors_flag){
    randomize_index_list(spp->random_index, spp->nfiles);
    }
}

/* select the next block, taking into account priors adjustment,
 * random presentation order, and cross validation training or testing
 *
 * If cross validation has been initialized,, seq_data_start_cv_fold
 * should have been called
 *
 * If these patterns are being requested by an epoch training algorithm,
 * seq_data_epoch_init should have been called */
int seq_data_get_next_file(int data_model)
{
    SEQ_PATTERNS *spp;
    int t_index;
    spp = sp_list[data_model];
    
    /* if this is cross valid, loop until we pick one, else do once */
    while(1){
    /* if we have read all of the data, return -1; */
    if(spp->n_read >= spp->nfiles) return -1;

    /* select a pattern, paying attention to random, priors, cv */
    if(spp->priors_flag){
        int class;
        /* select a pattern from the current fold by selecting the
           class first.  If there is no data in the class
           selected, keep trying until we select a class that has
           data in it */
        /* this has the potential for being an infinite loop */
        do{
        class = get_priors_class();
        t_index = get_priors_index(class);
        } while (t_index == -1);
        if(spp->cross_valid_flag)
          /* the t_index above is relative to fold_classes. we
         need an index into desired classes */
          t_index = spp->fold_indices[t_index];
    } else if(spp->random_flag && spp->train_flag){
        t_index = spp->random_index[spp->n_read];
    } else t_index = spp->n_read;
    spp->n_read++;
    if(spp->cross_valid_flag)
      if(spp->cv_mask[spp->cross_index[t_index]] != '1') continue;

    spp->frame_no = 0;
    spp->file_num = t_index;
    spp->block = &(spp->block_list[t_index]);
    return t_index;
    }
}

int seq_data_get_file_num(int data_model)
{
    return sp_list[data_model]->file_num;
}

int seq_data_get_class(int data_model)
{
    return sp_list[data_model]->block->class;
}

/* get frames from the current block */
float *seq_data_get_next_frame(int data_model)
{
    SEQ_PATTERNS *spp;
    int i, j, k, frame_no;
    spp = sp_list[data_model];
    
    /* check for reaching the end of a block */
    if(spp->block == NULL ||
       spp->frame_no+spp->nwindow_frames > spp->block->npatterns){
    return NULL;
    }

    /* fill the input vector and return it */
    for(k = i = 0; i < spp->nwindow_frames; i++){
    if(spp->window_frames != NULL)
      frame_no = spp->frame_no + spp->window_frames[i];
    else frame_no = spp->frame_no+i;
    if(frame_no >= spp->block->npatterns)
      /* reached the end of block prematurely */
      return NULL;
    for(j = 0; j < spp->ninputs; j++, k++)
      spp->inputs[k] = spp->block->data_matrix[frame_no][j];
    }
    spp->frame_no++;
    return spp->inputs;
}

void seq_data_exit(int data_model)
{
    SEQ_PATTERNS *spp;
    spp = sp_list[data_model];

    deallocate_structvector(spp->block_list);
    if(spp->cross_valid_flag){
    deallocate_ivector(spp->cross_index);
    deallocate_ivector(spp->n_seq_in_train);
    deallocate_ivector(spp->n_seq_in_test);
    deallocate_ivector(spp->n_frm_in_train);
    deallocate_ivector(spp->n_frm_in_test);
    }
    if(spp->priors_flag){
    if(spp->cross_valid_flag){
        deallocate_ivector(spp->fold_classes);
        deallocate_ivector(spp->fold_indices);
    } else /* priors should be freed after every cv fold */
      free_priors_data();
    } else
      if(spp->random_flag) deallocate_ivector(spp->random_index);

    deallocate_vector(spp->inputs);
    if(spp->window_frames != NULL) deallocate_ivector(spp->window_frames);

    nn_free((char *)spp, "for data model");
    sp_list[data_model] = NULL;
}

void seq_data_set_n_read_zero(int dm)
{
    sp_list[dm]->n_read = 0;
}
