/***--------------------------------------------------------------------------
* NAME:  sigm.c
* DESCRIPTION: 
*    sigmoid and cost fuction routines
*       dsigm		Calculates derivatives used for updating weights
*	sigm		Calculates function used for upadating weights
*	init_cost
*	calc_cost
*	calc_cost_vector
*	calc_avg_cost
*	calc_final_sigm
*	calc_output_deriv
*	cost_string
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/sigm.c,v 11.3 2001/08/23 20:14:49 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/sigm.c,v $
 * CREATED: 14-Jan-93 lck				MIT LINCOLN LABORATORY
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include "nnlib.h"

/*****************************************************************************
* NAME: sigm(val,type,parm)
* DESCRIPTION:
*        Calculates function used for upadating weights
*
*        Non-linear functions for nodes
     
     type = SIGM:	sigmoid, (Rummelhart)  asymptotes 0 and 1
     type = SYM_SIGM:	symmetric sigmoid      asymptotes -1 and 1
     type = STEP:	step function
     type = LINEAR:	linear, for input layer   slope = 1
*****************************************************************************/

float
sigm(float val, /* val is the weighted sum of inputs */
     int type, /* Type is one of the above non-linear functions */
     float param /* param is the scale used with SIGM and SYM_SIGM */)
{
    switch(type) {
      case SIGM:
	return((float) 1./(1.+exp(-param*val)));
      case SYM_SIGM:
	return((float)  2./(1.+exp(-param*val))-1.);
      case STEP:
	return( (val < 0.) ? 0. : 1.);
      case LINEAR:
	return(val);
      default:
	fatal_error("sigm:  function type %d not allowed\n",type);
	return(0);
    }
}
/*****************************************************************************
* NAME: dsigm(val,type,param)
* DESCRIPTION:
*        Calculates derivatives used for updating weights using output value
*
*        Non-linear functions for nodes
*****************************************************************************/

/* the derivative of the function */
float
dsigm (float val,/*Weighted sum of inputs passed through sigmoid func.*/
       int type, /* Type of above function used in sigmoid */
       float param /* Parameter used in sigmoid (cases 0 and 1) */)
{
    switch(type) {
      case SIGM:
	return((float) param*val*(1.-val));
      case SYM_SIGM:
	return((float) param*(1.+val)*(1.-val)/2.);
      case STEP:
	return((float) 1.); /* this makes pcp work in mlp framework */
      case LINEAR:
	return((float) 1.);
      default:
	fatal_error("dsigm:  function of type %d not allowed\n",type);
	return(0);
    }
}

void
init_cost(int cost_func, float param)
{
    switch(cost_func){
      case CFOM:
	top2diff_init(param);
	break;
      case ROC_COST:
	/* should there be anything here for ROC cost? */
	/* roc_cost_init(noutputs, nbins) */
	roc_cost_init((int)param, 200);
	break;
    }
}

float
calc_cost_vector(float *outputs, float *desired_outputs, float *cost_vector,
		 int noutputs, int type)
{
    float rv;
    int j, correct;

    rv = 0.;
    switch(type){
      case SQ_ERR:
      case PC_PROC:
	for(j = 0; j < noutputs; j++){
	    cost_vector[j] = outputs[j] - desired_outputs[j];
	    cost_vector[j] *= cost_vector[j];
	    rv += cost_vector[j];
	}
	rv /= noutputs;
	break;
      case CROSS_ENTROPY:
	for(j = 0; j < noutputs; j++){
	    cost_vector[j] = (float)log((double)outputs[j])*desired_outputs[j];
	    cost_vector[j] +=
	      (float)log((double)(1-outputs[j]))*(1-desired_outputs[j]);
	    rv += cost_vector[j];
	}
	rv /= noutputs;
	break;
      case MAX_LIKE:
      case CFOM:
      case LOG_LIKE:
	/* intended only for use with classification */
	correct = 0;
	for(j = 0; j < noutputs; j++){
	    cost_vector[j] = 0;
	    if(desired_outputs[j] > desired_outputs[correct])
	      correct = j;
	}
	switch(type){
	  case MAX_LIKE:
	    rv = (float)log((double)outputs[correct]);
	    break;
	  case CFOM:
	    rv = top2diff_cost(outputs, correct, noutputs);
	    break;
	  case LOG_LIKE:
	    rv = outputs[correct];
	    break;
	}
	cost_vector[correct] = rv;
	break;
      case ROC_COST:
	rv = roc_calc_cost_vector(outputs, desired_outputs, cost_vector);
	break;
      default:
	/* CENTERS, NO_COST, any others */
	for(j = 0; j < noutputs; j++){
	    cost_vector[j] = desired_outputs[j] - outputs[j];
	    rv += cost_vector[j];
	}
	rv /= noutputs;
	break;
    }
    return rv;
}

/* a copy of the routine above, with out the cost_vector */
float
calc_cost(float *outputs, float *desired_outputs, int noutputs, int type)
{
    float rv, temp;
    int j, correct;

    rv = 0.;
    switch(type){
      case SQ_ERR:
      case PC_PROC:
	for(j = 0; j < noutputs; j++){
	    temp = outputs[j] - desired_outputs[j];
	    rv += temp*temp;
	}
	rv /= noutputs;
	break;
      case CROSS_ENTROPY:
	for(j = 0; j < noutputs; j++){
	    temp = (float)log((double)outputs[j])*desired_outputs[j];
	    temp +=
	      (float)log((double)(1-outputs[j]))*(1-desired_outputs[j]);
	    rv += temp;
	}
	rv /= noutputs;
	break;
      case MAX_LIKE:
      case CFOM:
      case LOG_LIKE:
	/* intended only for use with classification */
	correct = 0;
	for(j = 0; j < noutputs; j++){
	    if(desired_outputs[j] > desired_outputs[correct])
	      correct = j;
	}
	switch(type){
	  case MAX_LIKE:
	    rv = (float)log((double)outputs[correct]);
	    break;
	  case CFOM:
	    rv = top2diff_cost(outputs, correct, noutputs);
	    break;
	  case LOG_LIKE:
	    rv = outputs[correct];
	    break;
	}
	break;
      case ROC_COST:
	rv = roc_calc_cost(outputs, desired_outputs);
	break;
      default:
	for(j = 0; j < noutputs; j++){
	    temp = outputs[j] - desired_outputs[j];
	    rv += temp*temp;
	}
	rv /= noutputs;
	warning("cost:  function of type %d not allowed\n",type);
    }
    return rv;
}
/* computes and returns average cost and resets counts for it */
float
calc_avg_cost(float cost_sum, int cost_cnt, int cost_func)
{
    float rv;
    if(cost_cnt > 0)
      rv = cost_sum/(float) cost_cnt;
    else rv = 0.;
    switch(cost_func){
      case SQ_ERR:
      case PC_PROC:
	rv = (float)(sqrt((double) rv));
    }
    return rv;
}

/* send the noutputs inputs through the sigmoid function specified,
   storing the results in outputs.  some cost functions require a normalization
   over the outputs, so the cost function is also given */
void
calc_final_sigm(float *inputs, float *outputs, int noutputs, int sigm_type,
		float sigm_param, int cost_func)
{
    int j;
    float out_sum;

    if(cost_func == MAX_LIKE){
	out_sum = 0.;
	for(j = 0; j < noutputs; j++){
	    outputs[j] = exp(inputs[j] * sigm_param);
	    out_sum += outputs[j];
	}
	if(out_sum > 0)
	  for(j = 0; j < noutputs; j++){
	      outputs[j] /= out_sum;
	  }
    } else {
	for(j = 0; j < noutputs; j++){
	    outputs[j] = sigm(inputs[j], sigm_type, sigm_param);
	}
    }
}

#define ABS(X) (((X)>0)? (X) : -(X))
int
calc_output_deriv(float *desired_outputs, float *outputs, float tolerance,
		  float *derivative, int noutputs, int sigm_type,
		  float sigm_param, int cost_func)
{
    int j, in_tolerance = 1, correct;
    float diff;

    switch(cost_func){
      case SQ_ERR:
	for(j = 0; j < noutputs; j++){
	    diff = outputs[j] - desired_outputs[j];
	    if(ABS(diff) < tolerance) derivative[j] = 0;
	    else{
		derivative[j] = diff*dsigm(outputs[j], sigm_type, sigm_param);
		in_tolerance = 0;
	    }
	}
	break;
      case CROSS_ENTROPY:
      case MAX_LIKE:
      case PC_PROC:
	for(j = 0; j < noutputs; j++){
	    diff = outputs[j] - desired_outputs[j];
	    if(ABS(diff) < tolerance) derivative[j] = 0;
	    else{
		derivative[j] = diff;
		in_tolerance = 0;
	    }
	}
	break;
      case CFOM:
	correct = 0;
	for(j = 0; j < noutputs; j++){
	    if(desired_outputs[correct] < desired_outputs[j]) correct = j;
	}
	d_top2diff_cost(outputs, correct, noutputs, derivative);
	in_tolerance = 0;
	break;
      case ROC_COST:
	roc_calc_deriv(outputs, desired_outputs, derivative);
	for(j = 0; j < noutputs; j++){
	    derivative[j] *= dsigm(outputs[j], sigm_type, sigm_param);
	}
	in_tolerance = 0;
	break;
    }
    return in_tolerance;
}
char *
cost_string(int cost_func)
{
    switch(cost_func){
      case SQ_ERR:
	return "RMS Err";
      case CROSS_ENTROPY:
	return "Cross Entropy";
      case MAX_LIKE:
	return "-log(LIKE)";
      case PC_PROC:
	return "RMS Err";
      case CFOM:
	return "Cost FOM";
      case LOG_LIKE:
	return "Avg LogL";
      case CENTERS:
	return "Centers";
      case NO_COST:
	return "";
      case CLUS_SQ_ERR:
	return "ave.sq.err";
      case CLUS_LOG_LIKE:
	return "log_like";
      case ROC_COST:
	return "ROC area";
      default:
	return "unknown";
    }
}

