/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/***---------------------------------------------------------------------------
 *   chi_square - calculate chi square statistic and determine significance
 *
 --------------------------------------------------------------------------***/

/* chi_square - calculate chi-square statistic and determine significance
   Forms new bins, making sure the expected number of patterns per bin is
   greater than 5. Adjusts the dof down if bins are combined and changes the
   dof provided. Returns dof and significance.
   Leaves the two actual and predicted input vectors unchanged.
----------------------------------------------------------------------------*/
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <math.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include "nnlib.h"

static void gcf(float *gammcf, float a, float x, float *gln);
static void gser(float *gamser, float a, float x, float *gln);
static float gammln(float xx);
static float chi_square_significance(int dof, float x);
static float gammq(float a, float x);
static float gammp(float a, float x);

float
chi_square(int n, int *dof,/* original # of bins, degrees of freedom */
	   float *actual, /* actual number of heads per bin */
	   float *predicted,/* predicted number of heads per bin*/
	   float *significance/*probability of value above this chi square */)
{
    int i, j, ismall;
    float xsmall;
    float total,diff; 
    float *sorted_actual, *combined_actual;
    float *sorted_predicted, *combined_predicted;
    float temp;
    int bin, nbins;
    
    /* allocate vectors for sorted list and combined bins */
    sorted_predicted = allocate_vector(n);
    combined_predicted = allocate_vector(n);
    sorted_actual = allocate_vector(n);
    combined_actual = allocate_vector(n);

    /* initialize vectors */
    for(i=0; i<n; i++)
      {
	  sorted_predicted[i] = predicted[i];
	  sorted_actual[i] = actual[i];
	  combined_predicted[i] = 0.0;
	  combined_actual[i] = 0.0;
      }
    
    /* sort estimated number of heads in each bin by smallest numbers first */
    /* bubble sort to order by smallest numbers first */
    for(i=0; i<n; i++)
      {
	  xsmall = sorted_predicted[i];
	  ismall=i;
	  for(j=i+1; j<n; j++)
	    if ( sorted_predicted[j] < xsmall)
	      {
		  xsmall = sorted_predicted[j];
		  ismall=j;
	      }
	  sorted_predicted[ismall] = sorted_predicted[i];
	  sorted_predicted[i] = xsmall;
	  temp = sorted_actual[ismall];
	  sorted_actual[ismall] = sorted_actual[i];
	  sorted_actual[i] = temp;
      }
    
    /* combine bins if estimated number in a bin is <= 5 
       reduce dof by one for every combination  and left out bin */
    bin=0;
    nbins=0;
    for(i=0; i<n; i++) {
	combined_predicted[bin] += sorted_predicted[i];
	combined_actual[bin] += sorted_actual[i];
	if ( combined_predicted[bin] > 5.0) {
	    bin++;
	    nbins++;
	}
    }
    *dof -= n-nbins;
    
    /* calculate total chi square */
    total = 0.0;
    for(bin=0; bin<nbins; bin++) {
	diff = combined_actual[bin] - combined_predicted[bin];
	if ( combined_predicted[bin] > 0.0)
	  total += diff*diff/(combined_predicted[bin]);
    }
	if ( *dof >= 1)
		{
		*significance = chi_square_significance(*dof, total);
		}
	else
		{
		*significance = 1.0;
        warning("chi_square dof=0, significance set to 1.0");
		}
		
    
    /* deallocate vectors for sorted list and combined bins */
    deallocate_vector(sorted_predicted);
    deallocate_vector(combined_predicted);
    deallocate_vector(sorted_actual);
    deallocate_vector(combined_actual);
    
    return(total);
}

static float
chi_square_significance(int dof, float x)
{
    return(gammq(.5*dof, .5*x));
}


static float
gammq(float a, float x)
{
    float gamser,gammcf,gln;
    
    if (x < 0.0 || a <= 0.0) 
	  {
      warning("Bad arguments in chi_square_significance:  %f  chi %f", a, x);
	  return (1.0);
	  }
    if (x < (a+1.0)) {
	gser(&gamser,a,x,&gln);
	return 1.0-gamser;
    } else {
	gcf(&gammcf,a,x,&gln);
	return gammcf;
    }
}

static float
gammp(float a, float x)
{
    float gamser,gammcf,gln;

    if (x < 0.0 || a <= 0.0) 
      warning("Bad arguments in chi_square_significance:  %f  chi %f", 
		  a, x);
    if (x < (a+1.0)) {
	gser(&gamser,a,x,&gln);
	return gamser;
    } else {
	gcf(&gammcf,a,x,&gln);
	return 1.0-gammcf;
    }
}

#define ITMAX 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30

static void
gcf(float *gammcf, float a, float x, float *gln)
{
    int i;
    float an,b,c,d,del,h;
    
    *gln=gammln(a);
    b=x+1.0-a;
    c=1.0/FPMIN;
    d=1.0/b;
    h=d;
    for (i=1;i<=ITMAX;i++) {
	an = -i*(i-a);
	b += 2.0;
	d=an*d+b;
	if (fabs(d) < FPMIN) d=FPMIN;
	c=b+an/c;
	if (fabs(c) < FPMIN) c=FPMIN;
	d=1.0/d;
	del=d*c;
	h *= del;
	if (fabs(del-1.0) < EPS) break;
    }
    if (i > ITMAX) 
      fatal_error("a too large %f, ITMAX too small %d in gcf", a, ITMAX);
    *gammcf=exp(-x+a*log(x)-(*gln))*h;
}
#undef ITMAX
#undef EPS
#undef FPMIN

#define ITMAX 100
#define EPS 3.0e-7

static void
gser(float *gamser, float a, float x, float *gln)
{
    int n;
    float sum,del,ap;

    *gln=gammln(a);
    if (x <= 0.0) {
	if (x < 0.0) 
	  fatal_error("x less than 0 %f in routine gser", x);
	*gamser=0.0;
	return;
    } else {
	ap=a;
	del=sum=1.0/a;
	for (n=1;n<=ITMAX;n++) {
	    ++ap;
	    del *= x/ap;
	    sum += del;
	    if (fabs(del) < fabs(sum)*EPS) {
		*gamser=sum*exp(-x+a*log(x)-(*gln));
		return;
	    }
	}
	fatal_error("a too large %f, ITMAX too small %d in routine gser",
		    a, ITMAX);
	return;
    }
}
#undef ITMAX
#undef EPS

static float
gammln(float xx)
{
    double x,y,tmp,ser;
    static double cof[6]={76.18009172947146,-86.50532032941677,
			    24.01409824083091,-1.231739572450155,
			    0.1208650973866179e-2,-0.5395239384953e-5};
    int j;

    y=x=xx;
    tmp=x+5.5;
    tmp -= (x+0.5)*log(tmp);
    ser=1.000000000190015;
    for (j=0;j<=5;j++) ser += cof[j]/++y;
    return -tmp+log(2.5066282746310005*ser/x);
}
