/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/***---------------------------------------------------------------------------
 * PURPOSE:	compute top2diff cost and its derivative 
 * FUNCTIONS:
 *    top2diff_init(steepness) - initialize top2diff routines
 *    top2diff_cost(out_vector, correct_class, noutputs) - compute total cost 
 *    d_top2diff_cost(out_vector, correct_class, 
 *		      noutputs, d_vector) - compute derivatives
 * DESCRIPTION:
 * The top 2 differential cost is derived from the 
 * classification-figure-of-merit (CFM) cost function developed by John
 * Hampshire at CMU. The cost for a single pattern is the difference between
 * the classifier output corresponding to the correct class minus the next 
 * highest output passes through a sigmoid-like function. This function
 * provides a low output near zero for high positive differences corresponding
 * to correct classification. It also provides high positive outputs for
 * negative differences corresponding to incorrect classification. The
 * steepness of the sigmoid is controlled by a parameter called steepness
 * which can range from 0 where the sigmoid is almost linear to 1.0 where
 * the sigmoid is almost a step function with a transition at the input
 * difference is 0.0. Note that the returned derivatives of all but two
 * outputs will be zero.  The sigmoid curves used are similar to those
 * used by John Hampshire but different to avoid copyright problems.
 * NOTES:
 * The sigmoid is approximated as a linear segment for low difference inputs,
 * followed by a sigmoid, followed by a constant for high inputs.
 * The derivative is computed by sampling the sigmoid curve at two points
 * around the current input to get a smooth curve.
 --------------------------------------------------------------------------***/
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include <math.h>
#include "nnlib.h"

/* static functions */
static float top2diff_sigmoid(float x);
static float d_top2diff_sigmoid(float x);

/* steepness starts by default at .5 if init not called */
static float last_steepness = .5;
static float last_width = 14.706;
static float last_epsilon = 0.0;
static float last_shift = .084344;
static float last_delta = -.04;
static float ydelta = 0.1384;

/*------------------------------------------------------
 * NAME: top2diff_cost
 * PURPOSE: Compute total cost for the top2 distance function from
 *          the correct class and a vector of outputs. Calculates
 *          the difference between the correct class output and the
 *          next highest and runs this through a sigmoid.
 *----------------------------------------------------------------*/
float top2diff_cost(float *out_vector, int correct_class,/* 0 to noutputs-1 */
		    int noutputs)
{
    int i;
    float correct_class_output;
    float next_highest_output;
    float diff;
    float cost;
    
    correct_class_output = out_vector[correct_class];
    next_highest_output = -(MAXFLOAT);
    
    /* find next highest output */
    for(i=0; i<noutputs; i++) {
	if ( i != correct_class && out_vector[i] > next_highest_output ) {
	    next_highest_output = out_vector[i];
	}
    }
    diff = correct_class_output - next_highest_output;
    cost = top2diff_sigmoid(diff);
    return(cost);
}
/*------------------------------------------------------
 * NAME: d_top2diff_cost
 * PURPOSE: Compute derivative of top2 distance function with
 *          respect to all outputs. Only the derivative with 
 *          respect to the correct class output and the next highest
 *          are non zero.
 *----------------------------------------------------------------*/
void d_top2diff_cost(float *out_vector, int correct_class,/* 0 to noutputs-1 */
		     int noutputs, float *d_vector)
{
    int i, inext;
    float correct_class_output;
    float next_highest_output;
    float diff;
    float derivative;
    
    correct_class_output = out_vector[correct_class];
    next_highest_output = -(MAXFLOAT);
    
    /* find next highest output and zero derivative vector */
    for(i=inext=0; i<noutputs; i++) {
	if ( i != correct_class && out_vector[i] > next_highest_output ) {
	    next_highest_output = out_vector[i];
	    inext = i;
	}
	d_vector[i] = 0.0;
    }
    
    diff = correct_class_output - next_highest_output;
    
    /* fill in non-zero derivatives */
    
    derivative = d_top2diff_sigmoid(diff);
    d_vector[correct_class] = derivative;
    d_vector[inext] = -derivative;
    
}

/*------------------------------------------------------
 * NAME: top2diff_sigmoid
 * PURPOSE: Calculate sigmoid function used for top2 difference
 *          cost function.
 *          The output ranges from 0 to positive infinity.
 *----------------------------------------------------------------*/
static float top2diff_sigmoid(float x)
{
    float y;
    
    /* clip output at 1.0 */
    if ( x >= 1.0)
      y=1.0;
    /* sigmoid output for slightly negative and positive */
    else if ( x >= last_delta) {
	y = -last_epsilon/2.0 + 
	  (1.0+last_epsilon)*(1.0/(1.0+exp(-last_width*(x-last_shift))));
	if ( y > 1.0)
	  y=1.0;
    }
    /* linear output for negative which continues negative */
    else 
      y = (x+1.0)*ydelta/(last_delta + 1.0);
    
    /* make into a cost where 0 is good and big positive is bad */
    return(1.0 - y);
}

/*------------------------------------------------------
 * NAME: d_top2diff_sigmoid
 * PURPOSE: Compute first derivation of top2 sigmoid
 * NOTES: It does this numerically instead of analytically to 
 *        provide a continuous smooth derivative and also because
 *        the extra computation is minimal.
 *----------------------------------------------------------------*/
#define DDELTA .05
#define TWICE_DDELTA .1
static float d_top2diff_sigmoid(float x)
{
    float y, y1, y2;
    
    y1 = top2diff_sigmoid(x - DDELTA);
    y2 = top2diff_sigmoid(x + DDELTA);
    y = (y2-y1)/TWICE_DDELTA;
    
    return(y);
}

/*------------------------------------------------------
 * NAME: top2diff_init
 * PURPOSE: Initialize constants required to compute Lincoln CFM cost function
 *----------------------------------------------------------------*/
/* update constants required to compute Lincoln CFM */
void top2diff_init(float steepness)
{
    float confidence;
    /* calculate constants for curve only when steepness changes */
    if ( last_steepness != steepness) {
	/* store steepness */
	last_steepness = steepness;

	/* limit range*/
	if ( steepness > 1.0)
	  steepness = 1.0;
	if ( steepness < 0.0 )
	  steepness = 0.0;
	confidence = .1+ .55*(1-steepness);

	/* calculate constants */
	last_width = 77.18 + 
	  confidence*(-321.87 + confidence*(531.25 -312.5*confidence));
	last_epsilon = 1.0/(exp((double)last_width) -1.0);
	last_shift = -.0005 + confidence*( .32 -.25*confidence);
	ydelta = -last_epsilon/2.0 + (1.0+last_epsilon)*
	  (1.0/(1.0+exp(-last_width*(last_delta-last_shift))));
	/* just in case you notice this again.
	   it is ok that last_delta is not changed here. It doesn't
	   need to */
    }
}
