/***---------------------------------------------------------------------------
* NAME: vonmises
* PURPOSE: utility routines for use with vonmises circular distribution function
* DESCRIPTION:
*           The von mises pdf function is defined by 
*           p(theta) = C*exp(kappa*cos(theta-alpha))
*                       theta   = circular input variable ranging from 0 to 2pi
*                                 (like x)
*                       kappa   = spread parameter 
*                                 (like 1/variance)
*                       alpha   = location parameter ranging from 0 to 2pi
*                                 (like mean)
*                       C       = 1/[2pi*I0(kappa)], where I0(k) is the
*                                 modified Bessel function 
*
* USAGE: 
*  float alpha_vonmises(xsum, ysum) - Compute alpha, the location or mean 
*                                   parameter from sums of the cosines (xsum) 
*                                   and sines (ysum) of sample radian values
*  float k_vonmises(xsum, ysum, n)  - Compute kappa, the spread parameter
*                                   from xsum, ysum, and the number
*                                   of radian values used.
*  float pdf_vonmises(alpha, kappa, theta)  - Compute pdf value for vonmises
*                                     distribution with location alpha,
*                                     spread kappa, and input radian
*                                     value theta (0 to 2pi)
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lib/vonmises.c,v 11.3 2001/08/23 20:14:50 craig Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lib/vonmises.c,v $
* CREATED: 13-Dec-95 rpl                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include "nnlib.h"
static float bessi0(float x);

/* compute mean in radians for a vonmises distribution given
   the xsum = sum cos(theta) and the ysum = sum sin(theta)
   where theta is from 0 to 2pi, zero is 3 o'clock, 
   and theta increases counter clockwise

   this is more properly alpha, in Rich's equations ----- lck
   ----------------------------------------------------*/
float
vonmises_mean(double xsum, double ysum)
{
    float rad_mean;
    /* atan returns values from -pi/2 to pi/2 */
    if ( xsum > 0.0 ){
    if ( ysum >= 0.0 ){
        /* rad_mean from 0 to pi/2 */
        rad_mean = atan(ysum/xsum);
    } else {
        /* rad_mean from 3*pi/2 to 2*pi */
        rad_mean = 2*M_PI + atan(ysum/xsum);
    }
    } else if ( xsum < 0.0 ){
    /* rad_mean from pi/2 to 3*pi/2 */
    rad_mean = M_PI + atan(ysum/xsum);
    } else { /* (xsum == 0.0), degenerate case */
    if ( ysum > 0.0)
      rad_mean = 0.5*M_PI;
    else
      rad_mean = 1.5*M_PI;
    }
    
    return(rad_mean);
}

/* Find k for von mises distribution given R/n.
   -------- Rich Lippman ------
   This subroutine approximates the nonlinear relationship between k and R/n
   using polynomials and the limiting form of the relationship
   which is 1/tanh(k) - 1/k = R/n.
   In this equation, r = sqrt(xsum*xsum + ysum*ysum) and n is the 
   number of samples. 
   I fit polynomials such that the maximum error when estimating k 
   is less than 0.5% of the true value of k for 0<k<inf.
   A final factor of 2.0 is added at the end.

   ------- Linda Kukolich -------
   to prevent confusion for programmers used to the gaussian equivlents
   (namely me) k is refered to as inv_var here.
-------------------------------------------------------------------*/
float
vonmises_inv_var(double xsum, double ysum, int n)
{
    double r;
    float kappa, x;

    r = sqrt(xsum*xsum + ysum*ysum);

    if ( n <= 1)
      kappa = 1.0;
    else {
        x = r/n;
        if (  x > .85)
      kappa = 1/(1-x);
        else if ( x > .75)
      kappa = 869.31*x*x*x -1952.279*x*x + 1479.262*x - 374.045;
        else if ( x > .65)
      kappa = 36.1786*x*x -38.802*x + 12.737;
        else  if ( x > 0.3)
      kappa = 14.8537*x*x*x - 14.5692*x*x +8.6075*x -0.724276;
        else if ( x > 0.1)
      kappa = 1.2592*x*x + 2.74381*x + 0.015784;
        else
      kappa = 3.0*x;
        
    }
    return(kappa/2.0);
}

/* calculate pdf of vonmises function */
/* note that the inputs to this must be in radians */
float
vonmises_log_output_scalar(float radian, float rad_mean, float kappa)
{
    double constant;
    float pdf;

    /* modified bessel function is ok if k is small */
    if ( kappa < 50.0) {
        constant = 1.0/(2.0*M_PI*bessi0(kappa));
        pdf = log10(constant) + M_LOG10E*kappa*cos(radian-rad_mean);
    } else {
    /* when k is large, approximate bessel function by exp(k)/sqrt(2pik) 
       and fold this into the exponential in the pdf numerator */
        constant = 1.0/(2.0*M_PI*sqrt(2.0*M_PI*kappa));
        pdf = log10(constant) + M_LOG10E*kappa*(cos(radian-rad_mean)-1.0);
    }
    return(pdf);
}

/* modified bessel function good if x < 50.0 
---------------------------------------------------*/
static float
bessi0(float x)
{
    float ax,ans;
    double y;

    if ((ax=fabs(x)) < 3.75) {
        y=x/3.75;
        y*=y;
        ans=1.0+y*(3.5156229+y*(3.0899424+y*(1.2067492
            +y*(0.2659732+y*(0.360768e-1+y*0.45813e-2)))));
    } else {
        y=3.75/ax;
        ans=(exp(ax)/sqrt(ax))*(0.39894228+y*(0.1328592e-1
            +y*(0.225319e-2+y*(-0.157565e-2+y*(0.916281e-2
            +y*(-0.2057706e-1+y*(0.2635537e-1+y*(-0.1647633e-1
            +y*0.392377e-2))))))));
    }
    return ans;
}
