/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/***---------------------------------------------------------------------------
* NAME:     xybar
* PURPOSE:  plot bar graph from floating numbers 
* USAGE: 
* xybar(x, y, n,
*         bar_type, bar_fract, 
*         grid_type, graph_label,
*         auto_abscissas, auto_xmin, auto_xstep,
*         height_fract, width_fract, xoffset_fract, yoffset_fract,
*         xlog, auto_xaxis, xmin, xmax, xstep, xlabel,
*         ylog, auto_yaxis, ymin, ymax, ystep, ylabel, color)
*
*   float x[], y[] - arrays with x and y coordinates
*   int n          - number of points to plot
*   int bar_type   -0=unfilled bars , 1=solid fill,  2=dot fill, 3=dot-dash
*                    4=short-dash, 5=long-dashed, 6=rotate between 1 to 6
*   float bar_fract - fractional width of bar relative to 1/2 distance to
*                     next x point on either side (0.0 to 1.0) 
*                     Set to 0.0 for single line instead of bar.
*                     Set to 1.0 for bars that touch.
*   int grid_type  - 0=no grid, 1=frame with ticks, 2=full grid
*   int auto_abscissas - 0=x points are provided in x array
*                          (auto_xmin, auto_xstep ignored)
*                        1=generate x points from auto_xmin, auto_xstep 
*   float auto_xmin, auto_xstep - used if auto_abscissas set to 1 
*   char *label - label for top of graph
*   float height_fract, width_fract - fraction of space for height and width
*   float xoffset_fract, yoffset_fract - fractional x and y offsets
*   int xlog - 0=linear axis, 1=log axis
*               (If a log axis is used, all x values must be greater than zero
*                and if auto_xaxis is 0 then xmin and xmax must also be greater
*                than zero)
*   int auto_xaxis - 0=x axis generated from xmin, xmax, xstep 
*                    1=x axis limits generated automatically
*                      (xmin, xmax, xstep are ignored)
*   float xmin, xmax, xstep - x axis lower limit, upper limit, and
*                                 grid spacing (filled in automatically
*                                 if all three are set to 0.0)
*   char *xlabel - label for x axis
*   int ylog - 0=linear axis, 1=log axis
*               (If a log axis is used, all y values must be greater than zero
*                and if auto_yaxis is 0 then ymin and ymax must also be greater
*                than zero)
*   int auto_yaxis - 0=y axis generated from ymin, ymax, ystep
*                    1=y axis limits generated automatically
*                      (ymin, ymax, ystep are ignored)
*   float ymin, ymax, ystep - y axis lower limit, upper limit, and
*                                 grid spacing (filled in  automatically
*                                 if all three are set to 0.0)
*   char *ylabel - label for y axis
*       char *color - color for bars
* 
* RETURN VALUE:
*   Zero if ok, -1 if error
*
* EXAMPLE:
*   # include "gvt.h"
*   float x[100], y[100];
*
*   ginit(4,1);
*   xybar(x, y, 100,
*       1, 1.0,
*       2, "Graph Example",
*       0, 0.0, 0.0,
*       1.0, 1.0, 0.0, 0.0, 
*       0, 0, 0.0, 100.0, 10.0, "Time(Secs)",
*       1, 0, .001, 100.0, 10.0, "Energy", "red");
*
* SEE ALSO: xygraph(), GRAPH(1G) in UNIV Manual Vol. 1
* 
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lippic/xlippic/xybar.c,v 11.5 2004/01/15 19:28:35 kukolich Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lippic/xlippic/xybar.c,v $
* CREATED: 14-Jun-84 rpl    LAST UPDATE: 14-Jun-84 rpl  MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "g_gvt.h"
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include "xlippic.h"
#if STDC_HEADERS
#include <stdlib.h>
#endif

#define INF 1.e+37
#define DONT_PLOT 1.e+37
#define F   .25
#define GVT 1
#ifdef GVT
#define YOFF_BOT 20
#define YOFF_TOP 10
#define XOFF_LEFT 4
#else
#define YOFF_BOT 80
#define YOFF_TOP 40
#define XOFF_LEFT 15
#endif

static char *backspaces = "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";

static struct xy {
    int xlbf;   /*flag:explicit lower bound*/
    int     xubf;   /*flag:explicit upper bound*/
    int xqf;    /*flag:explicit quantum*/
    double (*xf)(); /*transform function, e.g. log*/
    float   xa,xb;  /*scaling coefficients*/
    float   xlb,xub;    /*lower and upper bound*/
    float   xquant; /*quantum*/
    int xxmin;  /* minimum value of range */
    int xxmax;  /* maximum value of range */
    float   xoff;       /*screen offset fraction*/
    float   xsize;      /*screen fraction*/
    int xbot,xtop;  /*screen coords of border*/ 
    float   xmult;  /*scaling constant*/
} xd,yd;


static int tick;
static int top;
static int bot;

static int  n;
static int  gtype;
static int bartype;
static float barfract;
static int  absf;
static float absbot;
static float    dx;
static int symbf;
static int stype;

static char     *xlab, *ylab;
static char     *titlebuf;

static float *xx, *yy;

/* functions declarations */
static double BAidentity(/* double x */);
static void BAstruct_init(/* struct xy *p */);
static void BAlim_init(/* struct xy *p, int ll, int auto_axis,
             float lb, float ub, float step */);
static float BAxget(/* int i */);
static float BAyget(/* int i */);
static float BAmodceil(/* float f, float t */);
static float BAmodfloor(/* float f, float t */);
static void BAgetlim(/* struct xy *p, float (*getfunct)(void) */);
static void BAsetlim(/* struct xy *p */);
static struct z BAloglim(/*int lbf, int ubf, float lb, float ub*/);
static struct z BAlinlim(/* int lbf int ubf, double xlb_d, double xub_d */);
static void BAscale(/* struct xy *p, float (*getfunct)(void) */);
static void BAaxes(/* void */);
static int BAmark(/* int *xmark, double *xvalue, struct xy *p */);
static void BAsubmark(/* int *xmark, double *xvalue, int *pxn, float x,
            struct xy *p */);
static void BAplot(/* char *color */);
static int BAbarmult(/* int delta */);
static int BAmax(/* int i, int j */);
static int BAmin(/* int i, int j */);
static int BAvisible(/* int i, int *xp, int *yp */);
static int BAconv(/* float xy, struct xy *p, int *ip */);
static void BAtitle(/* void */);
static void BAaxlab(/* void */);
static void BAbadarg(/* void */);
static void BAextinit(/* void */);

static double BAidentity(x)
double x;
{
    return(x);
}

int
xybar(xinput, yinput, npoints,
    bar_type, bar_fract, 
    grid_type, graph_label,
    auto_abscissas, auto_xmin, auto_xstep,
    height_fract, width_fract, xoffset_fract, yoffset_fract,
    xlog, auto_xaxis, xmin, xmax, xstep, xlabel,
    ylog, auto_yaxis, ymin, ymax, ystep, ylabel, color)

float xinput[], yinput[];
int npoints;
int bar_type;
float bar_fract;
int grid_type;  
char *graph_label;
int auto_abscissas;
float auto_xmin, auto_xstep;
float height_fract, width_fract, xoffset_fract, yoffset_fract;
int xlog, auto_xaxis;
float xmin, xmax, xstep;
char *xlabel;
int ylog, auto_yaxis;
float ymin, ymax, ystep; 
char *ylabel;
char *color;
{
    char save_color[MAX_COLOR_STRING];
    strcpy(save_color, GG_.color);

    BAextinit();
    greset();
    BAstruct_init(&xd);
    BAstruct_init(&yd);

    /* assign global variables to values specified in parameters 
    ------------------------------------------------------------*/
    yy = yinput;
    if ( auto_abscissas == 0)
        xx = xinput;
    titlebuf = graph_label;
    xlab = xlabel;
    ylab = ylabel;
    gtype = grid_type;
    bartype = bar_type;
    barfract = bar_fract;
    symbf = 0;

    /* make sure limits are correct for log scale
    ------------------------------------------------*/
    if ( xlog != 0 && auto_xaxis == 0 && (xmin <=0 || xmax <=0))
        {
        lippic_fatal_error(
         "xybar: X limits invalid for log scale; xmin=%f, xmax=%f",                        xmin, xmax);
        }
    if ( ylog != 0 && auto_yaxis == 0 && (ymin <=0 || ymax <=0))
        {
        lippic_fatal_error(
         "xybar: Y limits invalid for log scale; ymin=%f, ymax=%f", 
                 ymin, ymax);
        }

    BAlim_init(&xd, xlog, auto_xaxis, xmin, xmax, xstep);
    BAlim_init(&yd, ylog, auto_yaxis, ymin, ymax, ystep);
    xd.xsize = width_fract;
    yd.xsize = height_fract;
    xd.xoff = xoffset_fract;
    yd.xoff = yoffset_fract;
    n = npoints;
    /* generate x points automatically */
    if (auto_abscissas != 0)
        {
        absf = 2;
        absbot = auto_xmin;
        dx = auto_xstep;
        }
    else
        absf = 0;

    /* scale to fit on screen
    ---------------------------------*/
    BAscale(&xd,BAxget);
    BAscale(&yd,BAyget);

    /* draw axes and title
    -------------------------------*/
    BAaxes();
    BAtitle();
    gcolor(color);
    BAplot(color);
    gcolor(save_color);
    return(0);
}

static void BAstruct_init(p)
struct xy *p;
{
    p->xf = BAidentity;
    p->xmult = 1;
}

static void BAlim_init(p, ll, auto_axis, lb, ub, step)
register struct xy *p;
int ll;
int auto_axis;
float lb, ub, step;
{
    p->xlb = INF;
    p->xub = -INF;

    if ( ll != 0)
        p->xf = log10;

    /* supply automatically */
    if ( auto_axis == 1)
        {
        p->xlbf = 0;
        p->xubf = 0;
        p->xqf = 0;
        }
    else
        {
        p->xlb = lb;
        p->xlbf = 1;
        p->xub = ub;
        p->xubf = 1;
        p->xquant = step;
        p->xqf = 1;
        }
}

static float
BAxget(i)
int i;
{
    float x;

    if( absf !=0 )
        x = i*dx + absbot;
    else 
        x = xx[i];

    return(x);
}
static float
BAyget(i)
int i;
{
    return(yy[i]);
}

static float
BAmodceil(f,t)
float f,t;
{

    t = fabs(t);
    return(ceil(f/t)*t);
}

static float
BAmodfloor(f,t)
float f,t;
{
    t = fabs(t);
    return(floor(f/t)*t);
}

static void BAgetlim(p,getfunct)
register struct xy *p;
float (*getfunct)();
{
    register int i;

    i = 0;
    do {
        if(!p->xlbf && p->xlb>(*getfunct)(i))
            p->xlb = (*getfunct)(i);
        if(!p->xubf && p->xub<(*getfunct)(i))
            p->xub = (*getfunct)(i);
        i++;
    } while(i < n);
}

static struct z {
    float lb,ub,mult,quant;
} BAloglim(), BAlinlim();

#define MAX_MARKS 250
static void BAsetlim(p)
register struct xy *p;
{
    float t,delta,sign;
    struct z z;
    int mark[MAX_MARKS];
    double value[MAX_MARKS];
    float lb,ub;
    int lbf,ubf;

    lb = p->xlb;
    ub = p->xub;
    delta = ub-lb;
    if(p->xqf) {
        if(delta*(p->xquant) <=0.0 )
            BAbadarg();
        return;
    }
    sign = 1;
    lbf = p->xlbf;
    ubf = p->xubf;
    if(delta < 0) {
        sign = -1;
        t = lb;
        lb = ub;
        ub = t;
        t = lbf;
        lbf = ubf;
        ubf = t;
    }
    else if(delta == 0) {
        if(ub > 0) {
            ub = 2*ub;
            lb = 0;
        } 
        else
            if(lb < 0) {
                lb = 2*lb;
                ub = 0;
            } 
            else {
                ub = 1;
                lb = -1;
            }
    }
    if(p->xf==log10 && lb>0 && ub>lb) {
        z = BAloglim(lbf,ubf,lb,ub);
        p->xlb = z.lb;
        p->xub = z.ub;
        p->xmult *= z.mult;
        p->xquant = z.quant;
        if(BAmark(mark,value,p)<2) {
            p->xqf = lbf = ubf = 1;
            lb = z.lb; ub = z.ub;
        } else
            return;
    }
    z = BAlinlim(lbf,ubf,(double) lb,(double) ub);
    if(sign > 0) {
        p->xlb = z.lb;
        p->xub = z.ub;
    } else {
        p->xlb = z.ub;
        p->xub = z.lb;
    }
    p->xmult *= z.mult;
    p->xquant = sign*z.quant;
}

static struct z
BAloglim(lbf,ubf,lb,ub)
int lbf, ubf;
float lb,ub;
{
    float r,s,t;
    struct z z;

    for(s=1; lb*s<1; s*=10) ;
    lb *= s;
    ub *= s;
    for(r=1; 10*r<=lb; r*=10) ;
    for(t=1; t<ub; t*=10) ;
    z.lb = !lbf ? r : lb;
    z.ub = !ubf ? t : ub;
    if(ub/lb<100) {
        if(!lbf) {
            if(lb >= 5*z.lb)
                z.lb *= 5;
            else if(lb >= 2*z.lb)
                z.lb *= 2;
        }
        if(!ubf) {
            if(ub*5 <= z.ub)
                z.ub /= 5;
            else if(ub*2 <= z.ub)
                z.ub /= 2;
        }
    }
    z.mult = s;
    z.quant = r;
    return(z);
}

static struct z
BAlinlim(lbf,ubf,xlb_d,xub_d)
int lbf,ubf;
double xlb_d,xub_d;
{
    struct z z;
    float r,s,delta;
    float ub,lb;
    float xlb, xub;

    xlb = (float) xlb_d;
    xub = (float) xub_d;

loop:
    ub = xub;
    lb = xlb;
    delta = ub - lb;
    /*scale up by s, a power of 10, so range (delta) exceeds 1*/
    /*find power of 10 quantum, r, such that delta/10<=r<delta*/
    r = s = 1;
    while(delta*s < 10)
        s *= 10;
    delta *= s;
    while(10*r < delta)
        r *= 10;
    lb *= s;
    ub *= s;
    /*set r=(1,2,5)*10**n so that 3-5 quanta cover range*/
    if(r>=delta/2)
        r /= 2;
    else if(r<delta/5)
        r *= 2;
    z.ub = ubf? ub: BAmodceil(ub,r);
    z.lb = lbf? lb: BAmodfloor(lb,r);
    if(!lbf && z.lb<=r && z.lb>0) {
        xlb = 0;
        goto loop;
    }
    else if(!ubf && z.ub>=-r && z.ub<0) {
        xub = 0;
        goto loop;
    }
    z.quant = r;
    z.mult = s;
    return(z);
}

static void BAscale(p, getfunct)
register struct xy *p;
float (*getfunct)();
{
    float edge;

    BAgetlim(p,getfunct);
    BAsetlim(p);
    edge = top-bot;
    p->xa = p->xsize*edge/((*p->xf)(p->xub) - (*p->xf)(p->xlb));
    p->xbot = bot + edge*p->xoff;
    p->xtop = p->xbot + (top-bot)*p->xsize + 0.5; /* round-off - dhj 8/83 */
    p->xb = p->xbot - (*p->xf)(p->xlb)*p->xa + .5;

}

static void BAaxes()
{
/* use dashed lines for grid marks - dhj 10/5/81 */
    register int i;
    int mark[MAX_MARKS];
    double value[MAX_MARKS];
    double x;
    char buff[BUFSIZ+BUFSIZ/2], out[BUFSIZ + BUFSIZ/2];
    int xn, yn;

    if(gtype==0)
        return;

    uvline(xd.xbot,yd.xbot,xd.xtop,yd.xbot);
    uvcont(xd.xtop,yd.xtop);
    uvcont(xd.xbot,yd.xtop);
    uvcont(xd.xbot,yd.xbot);

    xn = BAmark(mark,value,&xd);
    if (gtype == 2)
        gtexture(3);
    for(i=0; i<xn; i++) 
        {
        if(gtype==2)
            uvline(mark[i],yd.xbot,mark[i],yd.xtop);
        if(gtype==1) 
            {
            uvline(mark[i],yd.xbot,mark[i],yd.xbot+tick);
            uvline(mark[i],yd.xtop-tick,mark[i],yd.xtop);
            }
        x = value[i];
        if ( x != DONT_PLOT)
            {
            sprintf(buff, "%g", x/xd.xmult);
            out[0] = '\0';
            strncat(out, backspaces, strlen(buff)/2);
            strcat(out, buff);
            uvmove(mark[i], yd.xbot-YOFF_BOT);
            gprintf(1, out);
            }
        }
    yn = BAmark(mark,value,&yd);
    for(i=0; i<yn; i++) 
        {
        if(gtype==2)
            uvline(xd.xbot,mark[i],xd.xtop,mark[i]);
        if(gtype==1) 
            {
            uvline(xd.xbot,mark[i],xd.xbot+tick,mark[i]);
            uvline(xd.xtop-tick,mark[i],xd.xtop,mark[i]);
            }
        x = value[i];
        if ( x != DONT_PLOT)
            {
            sprintf(buff, "%g", x/yd.xmult);
            out[0] = '\0';
            strncat(out, backspaces, strlen(buff));
            strcat(out, buff);
            uvmove(xd.xbot-XOFF_LEFT,mark[i]);
            gprintf(1, out);
            }
        }
    gtexture(0);
}

static int BAmark(xmark,xvalue,p)
int *xmark;
double *xvalue;
register struct xy *p;
{
    int xn = 0;
    float x,xl,xu;
    float q;
    float xdelta;
    /* log and delta not given */
    if(p->xf==log10&&!p->xqf) 
        {
        for(x=p->xquant; x<p->xub; x*=10) 
            {
            BAsubmark(xmark,xvalue,&xn,x,p);
            if ( p->xlb == 0.0)
                {
              lippic_fatal_error(
              "xybar: lower limits invalid for log scale; xmin=%f, ymin=%f", 
                    xd.xlb, yd.xlb); 
                }
            if(p->xub/p->xlb<=100) 
                {
                BAsubmark(xmark,xvalue,&xn,2*x,p);
                BAsubmark(xmark,xvalue,&xn,5*x,p);
                }
            }
        } 
    else 
        {
        xn = 0;
        q = p->xquant;
        if(q>0) 
            {
            xl = BAmodceil(p->xlb+q/6,q);
            xu = BAmodfloor(p->xub-q/6,q)+q/2;
            } 
        else 
            {
            xl = BAmodceil(p->xub-q/6,q);
            xu = BAmodfloor(p->xlb+q/6,q)-q/2;
            }
        xdelta = fabs(p->xquant);
        for(x=xl; x<=xu; x+=xdelta)
            {
            if ( xn >= MAX_MARKS)
                lippic_fatal_error(
                "xybar: To many tick marks desired. Maximum = %d",
                 MAX_MARKS);
            xmark[xn] = (*p->xf)(x)*p->xa + p->xb;
            /* no room if too many labels */
            if ( xn > 2 && (p->xf==log10))
                xvalue[xn] = DONT_PLOT;
            else
                {
                if ( fabs(x) < xdelta/10000.0 )
                    x = 0.0;
                xvalue[xn] = x;
                }
            xn++;
            }
        }
    return(xn);
}
static void
BAsubmark(xmark,xvalue,pxn,x,p)
int *xmark;
double *xvalue;
int *pxn;
float x;
struct xy *p;
{
    if(1.001*p->xlb < x && .999*p->xub > x)
        {
        if ( (*pxn) >= MAX_MARKS)
            lippic_fatal_error(
            "xybar: To many tick marks desired. Maximum = %d",
             MAX_MARKS);
        xmark[(*pxn)] = log10(x)*p->xa + p->xb;
        xvalue[(*pxn)] = x;
        (*pxn)++;
        }
}

#define LAST 0
#define CURRENT 1
#define NEXT 2
static void BAplot(color)
char *color;
{
    
    int i,j;
    int left_bar;
    int right_bar;
    int top_bar;
    int bot_bar;
    int ix[3], iy[3], vis[3];
    int delta;


    /* bottom of bars */
    bot_bar = yd.xbot;

    /* plot all bars 
    ----------------------*/
    /* initialize */
    ix[LAST] = -1;
    iy[LAST] = -1;
    vis[LAST] = 0;
    vis[CURRENT] = BAvisible(0, &ix[CURRENT], &iy[CURRENT]);

    for(i=1; i<n; i++) 
        { 
        /* get next point */
        if ( i< n)
            vis[NEXT] = BAvisible(i, &ix[NEXT], &iy[NEXT]);
        else
            vis[NEXT] = 0;

        /* draw current bar 
        ----------------------------*/
        if (vis[CURRENT] == 1)
            {
            /* determine width of bar
            -------------------------*/
            /* bar goes 1/2 to next point */
            if (vis[LAST] == 0 && vis[NEXT] == 1)
                {
                delta = BAbarmult((ix[NEXT] - ix[CURRENT])/2);
                left_bar = BAmax(ix[CURRENT] - delta, xd.xbot);
                right_bar = ix[CURRENT] + delta;
                }
            /* bar goes 1/2 to last point */
            else if (vis[LAST] == 1 && vis[NEXT] == 0)
                {
                delta = BAbarmult((ix[CURRENT] - ix[LAST])/2);
                left_bar = ix[CURRENT] - delta;
                right_bar = BAmin(ix[CURRENT] + delta, xd.xtop);
                }
            /* bar goes 1/2 to last and next point */
            else if (vis[LAST] == 1 && vis[NEXT] == 1)
                {
                delta = BAbarmult((ix[CURRENT] - ix[LAST])/2);
                left_bar = ix[CURRENT] - delta;
                delta = BAbarmult((ix[NEXT] - ix[CURRENT])/2);
                right_bar = ix[CURRENT] + delta;
                }
            /* bar goes 1/2 to upper and lower x limits */
            else 
                {
                delta = BAbarmult((ix[CURRENT] - xd.xbot)/2);
                left_bar = ix[CURRENT] - delta;
                delta = BAbarmult((xd.xtop - ix[CURRENT])/2);
                right_bar = ix[CURRENT] + delta;
                }
            
            /* draw bar */
            top_bar = iy[CURRENT];

            /* filled bar */
            if (bartype == 1)
                uvblock(color, left_bar, bot_bar, right_bar, top_bar);
            /* others */
            else
                {
                /* bar outline */
                gtexture(0);
                uvrect(left_bar, bot_bar, right_bar, top_bar);
                /* non-solid filling */
                if (bartype > 1)
                    {
                    if (bartype == 6)
                        gtexture(i%5);
                    else
                        gtexture(bartype-1);
                    for(j=left_bar+2; j< right_bar-1; j+=2)
                        uvline(j, bot_bar, j, top_bar);
                    }
                }

            }
        /* update arrays for next bar
        -------------------------------*/
        ix[LAST] = ix[CURRENT];
        iy[LAST] = iy[CURRENT];
        vis[LAST] = vis[CURRENT];
        ix[CURRENT] = ix[NEXT];
        iy[CURRENT] = iy[NEXT];
        vis[CURRENT] = vis[NEXT];
        } 
}
static int BAbarmult(delta)
int delta;
{
    if ( barfract == 1.0) 
        return(abs(delta));
    else
        return((int)(barfract*abs(delta)));
}

static int BAmax(i,j)
int i,j;
{
if ( i > j)
    return(i);
else
    return(j);
}
static int BAmin(i,j)
int i,j;
{
if ( i< j)
    return(i);
else
    return(j);
}

static int BAvisible(i, xp, yp)
int i;              /* array index */
int *xp, *yp;       /* x and y values */
{
        if(!BAconv(BAxget(i),&xd,xp) || !BAconv(BAyget(i),&yd,yp)) 
            return(0);
        else
            return(1);
}
static int BAconv(xv,p,ip)
float xv;
register struct xy *p;
int *ip;
{
    long ix;
    ix = p->xa*(*p->xf)(xv*p->xmult) + p->xb;
    if(ix<p->xbot || ix>p->xtop) 
        {
#ifdef DEBUG
        aprintf("out of bounds: %d (%d-%d)\n", ix, p->xbot, p->xtop);
#endif /* DEBUG */
        return(0);
        }
    *ip = ix;
    return(1);
}




static void BAtitle()
{
char out[BUFSIZ+BUFSIZ/2];
/* put label at top of plot instead of bottom - dhj 10/5/81 */
    if (gtype !=0 && titlebuf[0] != 0) { /* center the label - dhj 7/83 */
        uvmove((xd.xbot+xd.xtop)/2, yd.xtop+YOFF_TOP);
        out[0] = '\0'; strncat(out, backspaces, strlen(titlebuf)/2);
        strcat(out, titlebuf);
        gprintf(1,out);
    }
    if(gtype) {
        BAaxlab();  /* label graphs differently - dhj 2/83 */
    }
}

static void BAaxlab()
{
char buff[BUFSIZ+BUFSIZ/2], out[BUFSIZ+BUFSIZ/2];
int mark[50];
double value[50];
int yn;
int i, ylabel_location;

 /* x axis minimum */
 sprintf(buff, "%g", xd.xlb/xd.xmult);
 uvmove(xd.xbot, yd.xbot-YOFF_BOT); gprintf(1,buff);
 /* x  axis label */
 if (xlab[0] != 0) { /* x-axis label specified */
    uvmove((xd.xbot+xd.xtop)/2, yd.xbot-2*YOFF_BOT);
    out[0] = '\0'; strncat(out, backspaces, strlen(xlab)/2);
    strcat(out, xlab);
    gprintf(1,out);
    }
 /* x axis maximum */
 sprintf(buff, "%g", xd.xub/xd.xmult);
 out[0] = '\0'; strncat(out, backspaces, strlen(buff));
 strcat(out, buff);
 uvmove(xd.xtop, yd.xbot-YOFF_BOT); gprintf(1,out);

 /* y axis minimum */
 sprintf(buff, "%g", yd.xlb/yd.xmult);
 out[0] = '\0'; strncat(out, backspaces, strlen(buff));
 strcat(out, buff);
 uvmove(xd.xbot-XOFF_LEFT, yd.xbot); gprintf(1,out);
 /* y axis label */
 /* make sure label doesnt go on top of a numerical value at a tick */
 yn = BAmark(mark, value, &yd);
 ylabel_location = (yd.xbot+yd.xtop)/2;
 for(i=0; i<yn; i++)
    while( abs(ylabel_location - mark[i]) < (3*YOFF_BOT)/2)
        ylabel_location +=2;
 if (ylab[0] != 0) { /* y-axis label specified */
    uvmove(xd.xbot-XOFF_LEFT, ylabel_location);
#ifdef NEW
    rotate(6, 8, 90);
#endif /* NEW */
    out[0] = '\0'; strncat(out, backspaces, strlen(ylab));
    strcat(out, ylab);
    gprintf(1,out);
#ifdef NEW
    rotate(6, 8, 0);
#endif /* NEW */
    }
 /* y axis maximum */
 sprintf(buff, "%g", yd.xub/yd.xmult);
 out[0] = '\0'; strncat(out, backspaces, strlen(buff));
 strcat(out, buff);
 uvmove(xd.xbot-XOFF_LEFT, yd.xtop); gprintf(1,out);
}


static void BAbadarg()
{
    lippic_fatal_error("xybar: error in arguments");
}
static void BAextinit()
{
#ifdef GVT
tick = 13;
top = 755;
bot = 83;
#else
tick = 50;
top = 3900;
bot = 415;
#endif
absf = 0;
absbot = 0.0;
symbf= 0;
stype= 0;
}
