/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/***---------------------------------------------------------------------------
* NAME:     xygraph
* PURPOSE:  create line or scatter graph from floating numbers 
* USAGE: 
* xygraph(x, y, n,
*         symbol_mode, symbol_string, symbol_type, symbol_radius,
*         line_type, grid_type, graph_label,
*         auto_abscissas, auto_xmin, auto_xstep,
*         height_fract, width_fract, xoffset_fract, yoffset_fract,
*         xlog, auto_xaxis, xmin, xmax, xstep, xlabel,
*         ylog, auto_yaxis, ymin, ymax, ystep, ylabel, color)
*
*   float x[], y[] - arrays with x and y coordinates
*   int n          - number of points to plot
*   int symbol_mode - 0=no symbol, 1=symbol is string stored in symbol_string, 
*                     2=symbol is specified in symbol_type
*   char *symbol_string - string printed at each point if symbol_mode = 1
*   int symbol_type - symbol printed at each point if symbol_mode = 2
*                     0=open square, 1=open triangle, 2=plus , 3=x, 
*                     4=open diamond, 5=open upside down triangle, 
*                     6=filled square, 7=circle, 8=filled circle
*                     9=right arrow head, 10=left arrow head
*             11=up arrow head, 12=down arrow head, 13=upper error bar,
*             14=lower error bar
*   int symbol_radius - radius of symbol if symbol_mode = 1 (0-1023)
*   int line_type   -0=only symbols with no lines, 1=solid, 2=dot, 3=dot-dash
*                    4=short-dash, 5=long-dashed
*   int grid_type  - 0=no grid, 1=frame with ticks, 2=full grid
*   int auto_abscissas - 0=x points are provided in x array
*                          (auto_xmin, auto_xstep ignored)
*                        1=generate x points from auto_xmin, auto_xstep 
*   float auto_xmin, auto_xstep - used if auto_abscissas set to 1 
*   char *label - label for top of graph
*   float height_fract, width_fract - fraction of space for height and width
*   float xoffset_fract, yoffset_fract - fractional x and y offsets
*   int xlog - 0=linear axis, 1=log axis
*               (If a log axis is used, all x values must be greater than zero
*                and if auto_xaxis is 0 then xmin and xmax must also be greater
*                than zero)
*   int auto_xaxis - 0=x axis generated from xmin, xmax, xstep 
*                    1=x axis limits generated automatically
*                      (xmin, xmax, xstep are ignored)
*   float xmin, xmax, xstep - x axis lower limit, upper limit, and
*                                 grid spacing (filled in automatically
*                                 if all three are set to 0.0)
*   char *xlabel - label for x axis
*   int ylog - 0=linear axis, 1=log axis
*               (If a log axis is used, all y values must be greater than zero
*                and if auto_yaxis is 0 then ymin and ymax must also be greater
*                than zero)
*   int auto_yaxis - 0=y axis generated from ymin, ymax, ystep
*                    1=y axis limits generated automatically
*                      (ymin, ymax, ystep are ignored)
*   float ymin, ymax, ystep - y axis lower limit, upper limit, and
*                                 grid spacing (filled in  automatically
*                                 if all three are set to 0.0)
*   char *ylabel - label for y axis
*   char *color - color for lines and symbols
* 
* RETURN VALUE:
*   Zero if ok, -1 if error
*
* EXAMPLE:
*   # include "gvt.h"
*   float x[100], y[100];
*
*   ginit(4,1);
*   xygraph(x, y, 100,
*       2, 'x', 5, 10,
*       0, 2, "Graph Example",
*       0, 0.0, 0.0,
*       1.0, 1.0, 0.0, 0.0, 
*       0, 0, 0.0, 100.0, 10.0, "Time(Secs)",
*       1, 0, .001, 100.0, 10.0, "Energy", "red");
*
* NOTES:
*  Improvements over graph(1G) are the addition of a number of symbols,
*  the ability to blank with symbols and characters, the ability to
*  get hardcopy, and the ability to run with a gvt terminal using the
*  alphanumeric scrolling screen for interactions.
* SEE ALSO: GRPH(1G) in UNIV Manual Vol. 1
* 
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/lippic/xlippic/xygraph.c,v 11.4 2004/01/15 19:28:35 kukolich Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/lippic/xlippic/xygraph.c,v $
* CREATED: 8-Jun-84 rpl    LAST UPDATE: 22-Jun-84 rpl   MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/


/* MAZ 9dec88 */
/*
 * fixed SUN4 bug in passing slinlim
 */

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include "g_gvt.h"
#include "xlippic.h"

#define MAX_MARKS 50
#define INF 1.e+37
#define F   .25
#define DONT_PLOT 1.e+37
#define GVT 1
#ifdef GVT
#define YOFF_BOT 20
#define YOFF_TOP 10
#define XOFF_LEFT 4
#else
#define YOFF_BOT 80
#define YOFF_TOP 40
#define XOFF_LEFT 15
#endif


static struct xy {
    int xlbf;   /*flag:explicit lower bound*/
    int     xubf;   /*flag:explicit upper bound*/
    int xqf;    /*flag:explicit quantum*/
    double (*xf)(); /*transform function, e.g. log*/
    float   xa,xb;  /*scaling coefficients*/
    float   xlb,xub;    /*lower and upper bound*/
    float   xquant; /*quantum*/
    int xmin;   /* minimum value of range */
    int xmax;   /* maximum value of range */
    float   xoff;       /*screen offset fraction*/
    float   xsize;      /*screen fraction*/
    int xbot,xtop;  /*screen coords of border*/ 
    float   xmult;  /*scaling constant*/
} xd,yd;

static char *backspaces = "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";


static int tick;
static int top;
static int bot;
static int  n;
static int  gtype;
static int ltype;
static int  absf;
static float absbot;
static float    dx;
static int symbf;
static int stype;
static int sradius;
static char *plotsymb;

#define BSIZ 80
static char     *xlab, *ylab;
static char     *titlebuf;

static float *xx, *yy;

/* function declarations */
static double GRident(/* double x */);
static void GRstruct_init(/* struct xy *p */);
static void GRlim_init(/* struct xy *p, int ll, int auto_axis,
             float lb, float ub, float step */);
static float GRxget(/* int i */);
static float GRyget(/* int i */);
static float GRmodceil(/* float f, float t */);
static void GRlimget(/* struct xy *p, float (*getfunct)(void) */);
static void GRsetlim(/* struct xy *p */);
static struct z GRsloglim(/* int lbf, int ubf, float lb, float ub */);
static struct z GRslinlim(/* int lbf, int ubf, double xlb_d, double xub_d */);
static void GRscale(/* struct xy *p, float (*getfunct)(void) */);
static void GRaxes(/* void */);
static int GRmarkset(/* int *xmark, double *xvalue, struct xy *p */);
static void GRsubmark(/*int *xmark, double *xvalue, int *pxn, float x,
               struct xy *p */);
static void GRplot(/* void */);
static int GRconv(/* float xv, struct xy *p, int *ip */);
static int GRsymbol(/* int ix, int iy */);
static void GRtitle(/* void */);
static void GRaxlab(/* void */);
static void GRextinit(/* void */);

static double GRident(x)
double x;
{
    return(x);
}
int
xygraph(xinput, yinput, npoints,
    symbol_mode, symbol_string, symbol_type, symbol_radius,
    line_type, grid_type, graph_label,
    auto_abscissas, auto_xmin, auto_xstep,
    height_fract, width_fract, xoffset_fract, yoffset_fract,
    xlog, auto_xaxis, xmin, xmax, xstep, xlabel,
    ylog, auto_yaxis, ymin, ymax, ystep, ylabel, color)

float xinput[], yinput[];
int npoints, symbol_mode, symbol_type, symbol_radius;
char *symbol_string;
int line_type, grid_type;  
char *graph_label;
int auto_abscissas;
float auto_xmin, auto_xstep;
float height_fract, width_fract, xoffset_fract, yoffset_fract;
int xlog, auto_xaxis;
float xmin, xmax, xstep;
char *xlabel;
int ylog, auto_yaxis;
float ymin, ymax, ystep; 
char *ylabel;
char *color;
{
    char save_color[MAX_COLOR_STRING];
    strcpy(save_color, GG_.color);

    GRextinit();
    GRstruct_init(&xd);
    GRstruct_init(&yd);

    /* assign global variables to values specified in parameters 
    ------------------------------------------------------------*/
    yy = yinput;
    if ( auto_abscissas == 0)
        xx = xinput;
    titlebuf = graph_label;
    xlab = xlabel;
    ylab = ylabel;
    gtype = grid_type;
    ltype = line_type;
    /* symbols */
    if ( symbol_mode == 0)
        symbf = 0;
    else if ( symbol_mode == 1)
        {
        symbf = 1;
        plotsymb = symbol_string;
        }
    else if (symbol_mode == 2)
        {
        symbf = 2;
        stype = symbol_type;
        sradius = symbol_radius;
        }
    else
        {
        lippic_fatal_error("xygraph: symbol_mode can be 0 to 2 not %d", 
                      symbol_mode);
        }

    /* make sure limits are correct for log scale
    ------------------------------------------------*/
    if ( xlog != 0 && auto_xaxis == 0 && (xmin <=0 || xmax <=0))
        {
        lippic_fatal_error(
               "xygraph: X limits invalid for log scale; xmin=%f, xmax=%f", 
                 xmin, xmax);
        }
    if ( ylog != 0 && auto_yaxis == 0 && (ymin <=0 || ymax <=0))
        {
        lippic_fatal_error(
                "xygraph: X limits invalid for log scale; ymin=%f, ymax=%f", 
                 ymin, ymax);
        }

    GRlim_init(&xd, xlog, auto_xaxis, xmin, xmax, xstep);
    GRlim_init(&yd, ylog, auto_yaxis, ymin, ymax, ystep);
    xd.xsize = width_fract;
    yd.xsize = height_fract;
    xd.xoff = xoffset_fract;
    yd.xoff = yoffset_fract;
    n = npoints;
    /* generate x points automatically */
    if (auto_abscissas != 0)
        {
        absf = 2;
        absbot = auto_xmin;
        dx = auto_xstep;
        }
    else
        absf = 0;

    /* scale to fit on screen
    ---------------------------------*/
    GRscale(&xd,GRxget);
    GRscale(&yd,GRyget);

    /* draw axes and title
    -------------------------------*/
    uvmove(0,0);
    GRaxes();
    GRtitle();
    gcolor(color);
    GRplot();
    gcolor(save_color);
    uvmove(0,0);
    return(0);
}

static void GRstruct_init(p)
struct xy *p;
{
    p->xf = GRident;
    p->xmult = 1;
}

static void GRlim_init(p, ll, auto_axis, lb, ub, step)
register struct xy *p;
int ll;
int auto_axis;
float lb, ub, step;
{
    p->xlb = INF;
    p->xub = -INF;

    if ( ll != 0)
        p->xf = log10;

    /* supply automatically */
    if ( auto_axis == 1)
        {
        p->xlbf = 0;
        p->xubf = 0;
        p->xqf = 0;
        }
    else
        {
        p->xlb = lb;
        p->xlbf = 1;
        p->xub = ub;
        p->xubf = 1;
        p->xquant = step;
        p->xqf = 1;
        }
}

static float
GRxget(i)
int i;
{
    float x;

    if( absf !=0 )
        x = i*dx + absbot;
    else 
        x = xx[i];

    return(x);
}
static float
GRyget(i)
int i;
{
    return(yy[i]);
}

static float
GRmodceil(f,t)
float f,t;
{

    t = fabs(t);
    return(ceil(f/t)*t);
}

static float
GRmodfloor(f,t)
float f,t;
{
    t = fabs(t);
    return(floor(f/t)*t);
}

static void
GRlimget(p,getfunct)
register struct xy *p;
float (*getfunct)();
{
    register int i;

    i = 0;
    do {
        if(!p->xlbf && p->xlb>(*getfunct)(i))
            p->xlb = (*getfunct)(i);
        if(!p->xubf && p->xub<(*getfunct)(i))
            p->xub = (*getfunct)(i);
        i++;
    } while(i < n);
}

struct z {
    float lb,ub,mult,quant;
} GRsloglim(), GRslinlim();

static void
GRsetlim(p)
register struct xy *p;
{
    float t,delta,sign;
    struct z z;
    int mark[MAX_MARKS];
    double value[MAX_MARKS];
    float lb,ub;
    int lbf,ubf;

    lb = p->xlb;
    ub = p->xub;
    delta = ub-lb;
    if(p->xqf) {
        if(delta*(p->xquant) <=0.0 )
            {
            lippic_fatal_error(
        "xygraph: auto x or y axis is bad; low lim=%f, upper lim=%f, step=%f",
        lb, ub, p->xquant);
            }
        return;
    }
    sign = 1;
    lbf = p->xlbf;
    ubf = p->xubf;
    if(delta < 0) {
        sign = -1;
        t = lb;
        lb = ub;
        ub = t;
        t = lbf;
        lbf = ubf;
        ubf = t;
    }
    else if(delta == 0) {
        if(ub > 0) {
            ub = 2*ub;
            lb = 0;
        } 
        else
            if(lb < 0) {
                lb = 2*lb;
                ub = 0;
            } 
            else {
                ub = 1;
                lb = -1;
            }
    }
    if(p->xf==log10 && lb>0 && ub>lb) {
        z = GRsloglim(lbf,ubf,lb,ub);
        p->xlb = z.lb;
        p->xub = z.ub;
        p->xmult *= z.mult;
        p->xquant = z.quant;
        if(GRmarkset(mark,value,p)<2) {
            p->xqf = lbf = ubf = 1;
            lb = z.lb; ub = z.ub;
        } else
            return;
    }
    z = GRslinlim(lbf,ubf,(double)lb,(double)ub);
    if(sign > 0) {
        p->xlb = z.lb;
        p->xub = z.ub;
    } else {
        p->xlb = z.ub;
        p->xub = z.lb;
    }
    p->xmult *= z.mult;
    p->xquant = sign*z.quant;
}

static struct z
GRsloglim(lbf,ubf,lb,ub)
int lbf, ubf;
float lb,ub;
{
    float r,s,t;
    struct z z;

    for(s=1; lb*s<1; s*=10) ;
    lb *= s;
    ub *= s;
    for(r=1; 10*r<=lb; r*=10) ;
    for(t=1; t<ub; t*=10) ;
    z.lb = !lbf ? r : lb;
    z.ub = !ubf ? t : ub;
    if(ub/lb<100) {
        if(!lbf) {
            if(lb >= 5*z.lb)
                z.lb *= 5;
            else if(lb >= 2*z.lb)
                z.lb *= 2;
        }
        if(!ubf) {
            if(ub*5 <= z.ub)
                z.ub /= 5;
            else if(ub*2 <= z.ub)
                z.ub /= 2;
        }
    }
    z.mult = s;
    z.quant = r;
    return(z);
}

static struct z
GRslinlim(lbf,ubf,xlb_d,xub_d)
int lbf,ubf;
double xlb_d,xub_d;
{
    struct z z;
    float r,s,delta;
    float ub,lb;
    float xub,xlb;

    xub = (float) xub_d;
    xlb = (float) xlb_d;

loop:
    ub = xub;
    lb = xlb;
    delta = ub - lb;
    /*scale up by s, a power of 10, so range (delta) exceeds 1*/
    /*find power of 10 quantum, r, such that delta/10<=r<delta*/
    r = s = 1;
    while(delta*s < 10)
        s *= 10;
    delta *= s;
    while(10*r < delta)
        r *= 10;
    lb *= s;
    ub *= s;
    /*set r=(1,2,5)*10**n so that 3-5 quanta cover range*/
    if(r>=delta/2)
        r /= 2;
    else if(r<delta/5)
        r *= 2;
    z.ub = ubf? ub: GRmodceil(ub,r);
    z.lb = lbf? lb: GRmodfloor(lb,r);
    if(!lbf && z.lb<=r && z.lb>0) {
        xlb = 0;
        goto loop;
    }
    else if(!ubf && z.ub>=-r && z.ub<0) {
        xub = 0;
        goto loop;
    }
    z.quant = r;
    z.mult = s;
    return(z);
}
static void
GRscale(p, getfunct)
register struct xy *p;
float (*getfunct)();
{
    float edge;

    GRlimget(p,getfunct);
    GRsetlim(p);
    edge = top-bot;
    p->xa = p->xsize*edge/((*p->xf)(p->xub) - (*p->xf)(p->xlb));
    p->xbot = bot + edge*p->xoff;
    p->xtop = p->xbot + (top-bot)*p->xsize + 0.5; /* round-off - dhj 8/83 */
    p->xb = p->xbot - (*p->xf)(p->xlb)*p->xa + .5;

}
static void
GRaxes()
{
/* use dashed lines for grid marks - dhj 10/5/81 */
    register int i;
    int mark[MAX_MARKS];
    double value[MAX_MARKS];
    int xn, yn;
    double x;
    char buff[BUFSIZ+BUFSIZ/2], out[BUFSIZ+BUFSIZ/2];

    if(gtype==0)
        return;

    uvmove(0,0);
    uvmove(1,1);
    uvmove(xd.xbot,yd.xbot);
    uvline(xd.xbot,yd.xbot,xd.xtop,yd.xbot);
    uvcont(xd.xtop,yd.xtop);
    uvcont(xd.xbot,yd.xtop);
    uvcont(xd.xbot,yd.xbot);
    uvmove(0,0);
    uvmove(1,1);

    xn = GRmarkset(mark,value,&xd);
    if (gtype == 2)
        gtexture(3);
    for(i=0; i<xn; i++) 
        {
        if(gtype==2)
            uvline(mark[i],yd.xbot,mark[i],yd.xtop);
        if(gtype==1) 
            {
            uvline(mark[i],yd.xbot,mark[i],yd.xbot+tick);
            uvline(mark[i],yd.xtop-tick,mark[i],yd.xtop);
            }
         /* x axis values at ticks */
         x = value[i];
         if ( x != DONT_PLOT)
            {
             sprintf(buff, "%g", x/xd.xmult);
             out[0] = '\0'; 
             strncat(out, backspaces, strlen(buff)/2);
             strcat(out, buff);
             uvmove(mark[i], yd.xbot-YOFF_BOT); 
             gprintf(1,out);
             }
        }
    yn = GRmarkset(mark,value,&yd);
    for(i=0; i<yn; i++) 
        {
        if(gtype==2)
            uvline(xd.xbot,mark[i],xd.xtop,mark[i]);
        if(gtype==1) 
            {
            uvline(xd.xbot,mark[i],xd.xbot+tick,mark[i]);
            uvline(xd.xtop-tick,mark[i],xd.xtop,mark[i]);
            }
         /* y axis values at ticks */
         x = value[i];
         if (x != DONT_PLOT)
            {
             sprintf(buff, "%g", x/yd.xmult);
             out[0] = '\0'; 
             strncat(out, backspaces, strlen(buff));
             strcat(out, buff);
             uvmove(xd.xbot-XOFF_LEFT, mark[i]); 
             gprintf(1,out);
            }
        }
    gtexture(0);
}

static int
GRmarkset(xmark,xvalue,p)
int *xmark;
double *xvalue;
register struct xy *p;
{
    int xn = 0;
    float x,xl,xu;
    float q;
    float xdelta;
    /* log axis and no delta given */
    if(p->xf==log10&&!p->xqf) 
        {
        for(x=p->xquant; x<p->xub; x*=10) 
            {
            GRsubmark(xmark,xvalue,&xn,x,p);
            if ( p->xlb == 0.0)
                {
                lippic_fatal_error(
             "xygraph: lower limits invalid for log scale; xmin=%f, ymin=%f", 
                     xd.xlb, yd.xlb); 
                }
            if(p->xub/p->xlb<=100) 
                {
                GRsubmark(xmark,xvalue,&xn,2*x,p);
                GRsubmark(xmark,xvalue,&xn,5*x,p);
                }
            }
        } 
    else 
        {
        xn = 0;
        q = p->xquant;
        if(q>0) 
            {
            xl = GRmodceil(p->xlb+q/6,q);
            xu = GRmodfloor(p->xub-q/6,q)+q/2;
            } 
        else 
            {
            xl = GRmodceil(p->xub-q/6,q);
            xu = GRmodfloor(p->xlb+q/6,q)-q/2;
            }
        xdelta = fabs(p->xquant);
        for(x=xl; x<=xu; x+=xdelta)
            {
            if ( xn >= MAX_MARKS)
                lippic_fatal_error(
                "xygraph: To many tick marks desired. Maximum = %d",
                 MAX_MARKS);
            xmark[xn] = (*p->xf)(x)*p->xa + p->xb;
            /* no room if too many points */
            if ( xn > 2 && (p->xf==log10))
                xvalue[xn] = DONT_PLOT;
            else
                {
                /* set to zero if small */
                if (fabs(x) < xdelta/10000.0 )
                    x = 0.0;
                xvalue[xn] = x;
                }
            xn++;
            }
        }
    return(xn);
}
static void
GRsubmark(xmark,xvalue,pxn,x,p)
int *xmark;
double *xvalue;
int *pxn;
float x;
struct xy *p;
{
    if(1.001*p->xlb < x && .999*p->xub > x)
        {
        if ( (*pxn) >= MAX_MARKS)
            lippic_fatal_error(
            "xygraph: To many tick marks desired. Maximum = %d",
             MAX_MARKS);
        xmark[(*pxn)] = log10(x)*p->xa + p->xb;
        xvalue[(*pxn)] = x;
        (*pxn)++;
        }
}
static void
GRplot()
{
    int ix,iy, oldix;
    int i;
    int conn;

    conn = 0; oldix = -1;
    for(i=0; i<n; i++) 
        { 
        if(!GRconv(GRxget(i),&xd,&ix) || !GRconv(GRyget(i),&yd,&iy)) 
            {
            conn = 0;
            continue;
            }
        if(ltype!=0) 
            {
            if(conn != 0 )
                uvcont(ix,iy);
            else 
                {
                uvmove(ix, iy);
                gtexture(ltype-1);
                }
            conn = 1; oldix = ix;
            }
        conn &= GRsymbol(ix,iy);
        }
    gtexture(0);
}
static int
GRconv(xv,p,ip)
float xv;
register struct xy *p;
int *ip;
{
    long ix;
    ix = p->xa*(*p->xf)(xv*p->xmult) + p->xb;
    if(ix<p->xbot || ix>p->xtop) 
      {
#ifdef DEBUG
      aprintf("out of bounds: %d (%d-%d)\n", ix, p->xbot, p->xtop);
#endif /* DEBUG */
      return(0);
      }
    *ip = ix;
    return(1);
}

static int
GRsymbol(ix,iy)
     int ix, iy;
{

    if(symbf==0 && ltype==0)
        uvpoint(ix,iy);
    else if ( symbf > 0) 
        {
        uvmove(ix,iy);
        /* character string */
        if ( symbf == 1)
            gprintf(1,plotsymb);
        /* special symbol */
        else 
            gsymbol(stype, sradius);
        uvmove(ix,iy);
        }
    else
        ;

    return(1);
}
static void
GRtitle()
{
char out[BUFSIZ+BUFSIZ/2];
/* put label at top of plot instead of bottom - dhj 10/5/81 */
    if (gtype != 0 && titlebuf[0] != 0) { /* center the label - dhj 7/83 */
        uvmove((xd.xbot+xd.xtop)/2, yd.xtop+YOFF_TOP);
        out[0] = '\0'; strncat(out, backspaces, strlen(titlebuf)/2);
        strcat(out, titlebuf);
        gprintf(1,out);
    }
    if(gtype) {
        GRaxlab();  /* label graphs differently - dhj 2/83 */
    }
}
static void
GRaxlab()
{
char buff[BUFSIZ+BUFSIZ/2], out[BUFSIZ+BUFSIZ/2];
int mark[MAX_MARKS];
double value[MAX_MARKS];
int yn;
int i, ylabel_location;

 /* x axis minimum */
 sprintf(buff, "%g", xd.xlb/xd.xmult);
 uvmove(xd.xbot, yd.xbot-YOFF_BOT); gprintf(1,buff);

 /* x  axis label */
 if (xlab[0] != 0) { /* x-axis label specified */
    uvmove((xd.xbot+xd.xtop)/2, yd.xbot-2*YOFF_BOT);
    out[0] = '\0'; strncat(out, backspaces, strlen(xlab)/2);
    strcat(out, xlab);
    gprintf(1,out);
    }
 /* x axis maximum */
 sprintf(buff, "%g", xd.xub/xd.xmult);
 out[0] = '\0'; strncat(out, backspaces, strlen(buff));
 strcat(out, buff);
 uvmove(xd.xtop, yd.xbot-YOFF_BOT); gprintf(1,out);

 /* y axis minimum */
 sprintf(buff, "%g", yd.xlb/yd.xmult);
 out[0] = '\0'; strncat(out, backspaces, strlen(buff));
 strcat(out, buff);
 uvmove(xd.xbot-XOFF_LEFT, yd.xbot); gprintf(1,out);

 /* y axis label */
 /* first find where marks are and make sure label doesn't
    write over a numerical value */
 yn = GRmarkset(mark,value, &yd);
 ylabel_location = (yd.xbot+yd.xtop)/2;
 for(i=0; i<yn; i++)
    while( abs(ylabel_location -mark[i]) < (3*YOFF_BOT)/2 )
        ylabel_location += 2;
 if (ylab[0] != 0) { /* y-axis label specified */
    uvmove(xd.xbot-XOFF_LEFT, ylabel_location);
#ifdef NEW
    rotate(6, 8, 90);
#endif /* NEW */
    out[0] = '\0'; strncat(out, backspaces, strlen(ylab));
    strcat(out, ylab);
    gprintf(1,out);
#ifdef NEW
    rotate(6, 8, 0);
#endif /* NEW */
    }
 /* y axis maximum */
 sprintf(buff, "%g", yd.xub/yd.xmult);
 out[0] = '\0'; strncat(out, backspaces, strlen(buff));
 strcat(out, buff);
 uvmove(xd.xbot-XOFF_LEFT, yd.xtop); gprintf(1,out);
}

/* initialize external variables */
static void
GRextinit()
{

#ifdef GVT
tick = 13;
top = 755;
bot = 83;
#else
tick = 50;
top = 3900;
bot = 415;
#endif

absf = 0;
absbot = 0.0;
symbf= 0;
stype= 0;
sradius= 0;
}
