/***-------------------------------------------------------------------------
 * NAME:	plot_bintree.c
 * PURPOSE:	Plots structure of a binary tree classifier
 * USAGE:
 *
 * SEE ALSO:	
 *
 * DESCRIPTION: Reads in a bintree parameter file and plots information
 *		about each node and leaf in the tree
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_bintree.c,v 11.4 2001/08/23 20:18:08 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_bintree.c,v $
 * CREATED: 2-Nov-93 lck			MIT LINCOLN LABORATORY
 -------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_bintree.c,v 11.4 2001/08/23 20:18:08 craig Exp $";


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#define bintree_pack_h_internal_types
#define bintree_pack_h_internal_macros
#include <bintree_pack.h>
#include "plot_struct.h"
#include "xlippic.h"

static int linear;

static int prune_tree = 0;
static int max_nodes = 0;
static ARG extra_plot_argtab[] = 
{
    {"max_nodes", INT, &max_nodes,
       "non-terminal nodes in tree after pruning"},
    {"prune_tree", BOOLEAN, &prune_tree, 
       "prune tree to max_nodes nodes. Default, use full tree"},
};
#define EXTRA_PLOT_TABSIZE (sizeof(extra_plot_argtab)/sizeof(ARG))

#define BLOCK_RADIUS 40
#define BLOCK_SEPARATION (3*BLOCK_RADIUS)
#define LOWER_TEXT 20
#define CENTERED 2

static int plot_middle, umiddle, voffset;
static int block_radius, lower_text;
static float  horiz_separation, vert_separation;
static char **labels;
static char **input_labels;
static float struct_height = 1., struct_width = 1.;

static int node2u (int x);
static int level2v (int level);
static void draw_node(BTLINK node);

int
main(int argc, char **argv)
{
    FILE *param_file;
    int model, normalization, i;
    int nbins, nlevels, *max_x, min_x, xwidth;
    BTNODE **bins, *node;
    int ubot = 0, utop  = 1020, vbot = 0, vtop = 780;
    float scale;
    ARG *argtab;
    int tabsize;
    DEFAULTS_FILE_STRUCT df_args;
    
    init_nnsubs(argc, argv, rcsid);
    /********* read command line arguments ************/
    nnargs_combine_argtabs(struct_plot_argtab, STRUCT_PLOT_TABSIZE,
			   extra_plot_argtab, EXTRA_PLOT_TABSIZE,
			   &argtab, &tabsize);
    argc = nnargs(argc, argv, argtab, tabsize);
    nnargs_free_argtab(argtab, tabsize);

    /********** open and read parameter file ***********/
    param_file = nn_fopen(struct_param_fname, "r",
			  "for read of bintree parameters");
    model = norm_read_paramfile(param_file, &normalization);
    model = bintree_read_model(param_file);
    nn_fclose(param_file, struct_param_fname,
	      "for read of bintree parameters");

    if(strcmp(struct_defaults_fname, "") != 0){
	read_defaults_file(struct_defaults_fname, &df_args);
	labels = df_args.output_labels;
	input_labels = df_args.input_labels;
    } else {
	input_labels = labels = NULL;
    }
    /*********** prune tree ***************/
    /* don't adjust priors */
    bintree_test_init(model, !prune_tree, max_nodes, 0, NULL);

    /************* initialize graphics *************/
    if(no_graphics){
	ginit("no graphics");
    } else {
	ginit(" -geometry 510x450 -title struct_plot");
    }
    ghardcopy(1);
    gwtitle("Bintree Structure Plot");
    
    /*************** set x positions of nodes **************/
    bins = *(BTNODE ***)bintree_param(model, "bins");
    nbins = *(int *)bintree_param(model, "nbins");
    nlevels = *(int *)bintree_param(model, "nlevels");
    linear = *(int *)bintree_param(model, "linear");

    max_x = allocate_ivector(nlevels);
    /* start at the top, doing a depth first traversal of the tree */
    node = bins[0];
    node->x = 0;
    for(i = 0; i < nlevels; i++)
      max_x[i] = -1 * ( MAXINT );
    nlevels = 0;
    max_x[node->level] = 0; /* set the max for the root */
    while(!ROOT_NODE(node) || !VISITED_NODE(node)){
	while(!VISITED_NODE(node)){
	    /* terminal nodes are defined as VISITED */
	    if(!node->yesvisit){
		node->yeschild->x = node->x-1;
		node = node->yeschild;
	    } else if(!node->novisit){
		node->nochild->x = node->x+1;
		node = node->nochild;
	    }
	    /* this should be the child, since one of the two has not been
	       visited */
	    if(node->x <= max_x[node->level]) node->x = max_x[node->level]+1;
	    if(nlevels < node->level) nlevels = node->level;
	    max_x[node->level] = node->x;
	}
	node->visit = 1;
	node = node->mother;
	if(node->nochild->visit){
	    node->novisit = 1;
	    node->x = (node->nochild->x + node->yeschild->x)/2;
	    max_x[node->level] = node->x;
	} else if(node->yeschild->visit){
	    node->yesvisit = 1;
	    node->x = node->yeschild->x + 1;
	    max_x[node->level] = node->x;
	}
    }
    bintree_reset_visit_flags(node);
    nlevels++; /* level counting started at 0 */

    /* figure out reference positions for actual plotting */
    /* find width of tree */
    min_x = bins[0]->x;
    for(i = 0; i < nbins; i++)
      if(!PRUNE_SET(bins[i]) && bins[i]->x < min_x){
	  min_x = bins[i]->x;
    }
    xwidth = min_x;
    for(i = 0; i < nlevels; i++) if(max_x[i] > xwidth) xwidth = max_x[i];
    xwidth -= min_x;
    plot_middle = xwidth/2 + min_x;

    if(struct_autoscale){
	struct_height = (vtop-vbot)/(float)((nlevels+1)*BLOCK_SEPARATION);
	struct_width = (utop-ubot-50)/(float)((xwidth+2)*BLOCK_SEPARATION);
    }
    scale = (struct_width < struct_height) ? struct_width : struct_height;
    block_radius = (int)(scale*BLOCK_RADIUS);
    horiz_separation = struct_width*BLOCK_SEPARATION;
    vert_separation = struct_height*BLOCK_SEPARATION;
    lower_text = block_radius+LOWER_TEXT;

    umiddle = (utop - ubot)/2;
    voffset = vtop-lower_text-block_radius-25; /* we need room for the title */

    /* write out the plot title */
    uvprintf(umiddle, vtop - 20, CENTERED, "%s", struct_param_fname);

    /* draw nodes */
    node = bins[0];
    while(!ROOT_NODE(node) || !VISITED_NODE(node)){
	while(!VISITED_NODE(node)){
	    if(!node->yesvisit) node = node->yeschild;
	    else if (!node->novisit) node = node->nochild;
	}
	draw_node(node);
	node->visit = 1;
	node = node->mother;
	if(node->nochild->visit) node->novisit = 1;
	else if(node->yeschild->visit) node->yesvisit = 1;
    }
    draw_node(node);
    gflush();
    gsave(struct_plot_fname, 1);
    gexit(0);/* quit */
    exit_nnsubs();
    exit(0);
}

static int
node2u(int x)
{
    return((x-plot_middle)*horiz_separation + umiddle);
}

static int
level2v(int level)
{
    return(voffset - level*vert_separation);
}

static void
draw_node(BTLINK node)
{
    int nodeu, nodev, childu, childv;

    /* get the uv coordinates for the node */
    nodeu = node2u(node->x);
    nodev = level2v(node->level);
    if(TRMNL_NODE(node)){
	/* draw a square */
	uvrect(nodeu-block_radius, nodev-block_radius,
	       nodeu+block_radius, nodev+block_radius);
	/* give the node error rate */
	if(!struct_no_text){
	    uvprintf(nodeu, nodev, CENTERED, "%.0f%%",
		     100*(1-node->class_distribs[node->class]));
	    /* write the class */
	    if(labels != NULL)
	      uvprintf(nodeu, nodev-lower_text, CENTERED, "%s",
		       labels[node->class+1]); /* labels[0] is a length */
	    else
	      uvprintf(nodeu, nodev-lower_text, CENTERED, "%d",
		       node->class);
	}
    } else {
	/* variables for calculating ends of lines to children */
	int ulength, vlength, r, u1, v1, u2, v2;
	
	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);
	/* write the split constant */
	if(!struct_no_text && !linear){
	    uvprintf(nodeu, nodev, CENTERED, "%.2g", node->split_constant);
	    /* write the split feature */
	    uvprintf(nodeu, nodev-lower_text, CENTERED, "X%d",
		     node->split_input);
	}
	/* get coordinates for yes child */
	childu = node2u(node->yeschild->x);
	childv = level2v(node->yeschild->level);
	/* draw line to yeschild */
	ulength = nodeu-childu;
	vlength = nodev-childv;
	r = (int)sqrt(((double)ulength*ulength)+(double)vlength*vlength);
	u1 = nodeu-(ulength*block_radius)/r;
	v1 = nodev-(vlength*block_radius)/r;
	if(TRMNL_NODE(node->yeschild)){
	    if(ulength > vlength){
		u2 = childu+block_radius;
		v2 = childv+(vlength*block_radius)/ulength;
	    } else {
		u2 = childu+(ulength*block_radius)/vlength;
		v2 = childv+block_radius;
	    }
	} else {
	    u2 = childu+(ulength*block_radius)/r;
	    v2 = childv+(vlength*block_radius)/r;
	}
	uvline(u1, v1, u2, v2);
	/* write the number of yes patterns */
	if(!struct_no_text){
	    uvprintf((nodeu+childu)/2-8, (nodev+childv)/2+5, CENTERED,
		     "%d", node->yeschild->npatterns);
	}
	
	/* get coordinates for no child */
	childu = node2u(node->nochild->x);
	childv = level2v(node->nochild->level);
	/* draw line to nochild */
	ulength = childu-nodeu;
	vlength = nodev-childv;
	r = (int)sqrt(((double)ulength*ulength)+(double)vlength*vlength);
	u1 = nodeu+(ulength*block_radius)/r;
	v1 = nodev-(vlength*block_radius)/r;
	if(TRMNL_NODE(node->nochild)){
	    if(ulength > vlength){
		u2 = childu-block_radius;
		v2 = childv+(vlength*block_radius)/ulength;
	    } else {
		u2 = childu-(ulength*block_radius)/vlength;
		v2 = childv+block_radius;
	    }
	} else {
	    u2 = childu-(ulength*block_radius)/r;
	    v2 = childv+(vlength*block_radius)/r;
	}
	uvline(u1, v1, u2, v2);
	/* write the number of no patterns */
	if(!struct_no_text){
	    uvprintf((nodeu+childu)/2+10, (nodev+childv)/2+5, CENTERED,
		     "%d", node->nochild->npatterns);
	}
    }
}
