/***--------------------------------------------------------------------------
* NAME:         plot_detect.c
* PURPOSE:      to prepare a table and a plot of false vs correct
*               detection rate for a LNKnet classifier
*
* USAGE: 
*   plot_detect -<flags>
* 
* RETURN VALUE:  NONE
*
* EXAMPLE: plot_detect --        * shows all available flags
*
* GLOBALS: <flags> defined in plot_detect_args.h
*
* DESCRIPTION:
*    Functions in this file
*
* HISTORY:
*  1/26/94        Yuchun Lee        creation
*  4/19/94        Yuchun Lee        revised using roc_pack
*
* SEE ALSO:
*
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

/*static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_detect.c,v 11.8 2001/09/11 15:37:59 kukolich Exp $";*/


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include <get_subs.h>
#include "plot_detect_args.h"
#include "xlippic.h"

void roc_plot(int model_number, int class, float reject,
          float xmin, float xmax, float xstep,
          float ymin, float ymax, float ystep, char *exp_file_name);

int
main(int argc, char **argv)
{
  FILE *fp, *err_file;
  int aroc, npoints;
  int error_file_verbosity=2; /* long Error File */
  ERROR_LINE error_line;

  /* initialization */
  init_nnsubs(argc, argv, NULL);

  /* read args */
  argc = nnargs(argc, argv, detect_argtab, DETECT_TABSIZE);
  if(target < 0) target = 0;
  if(target > noutputs-1) target = noutputs - 1;

  if(table_step == 0)
    fatal_error("table_step must be greater than 0");

  /* open error file and count the lines in it */
  err_file = nn_fopen(path_build(plot_exp_path, error_fname),
              "r","error file");

  npoints = get_n_lines(err_file); /* does its own rewind */

  /* create roc model */
  aroc= roc_build(noutputs,npoints);

  /* allocate space for vectors from error file */
  error_line.outputs = allocate_vector(noutputs);
  error_line.inputs = NULL;
  error_line.desired_outputs = allocate_vector(noutputs);
  error_line.ninputs = 0;
  error_line.noutputs = noutputs;

  /* accumulate information for roc plot */
  while(get_err_line(err_file, error_fname,
             &error_line, error_file_verbosity) != EOF){
    if(log_outputs){
      /*
      for(i = 0; i < error_line.noutputs; i++){
      error_line.outputs[i] = nn_exp10(error_line.outputs[i]);
      } */
      roc_log_train(aroc, error_line.desired_class, error_line.outputs);
    }
    roc_batch_train(aroc, error_line.desired_class, error_line.outputs );
  }
  deallocate_vector(error_line.desired_outputs);
  deallocate_vector(error_line.outputs);
  fclose(err_file);

  /* call to output table */
  if (verbose > 0) {
      if (verbose > 1) fprintf(stdout,"ROC-x      ROC-y\n");
      if(verbose > 2)
    roc_print_table(stdout, aroc, target, reject_rate, verbose);
      else 
    roc_print_sampled_table(stdout, aroc, target, reject_rate, verbose,
                table_begin, table_end, table_step);
  }
  fflush(stdout);

  roc_get_eer(stdout, aroc, target, reject_rate);

  /* call to output plot */
  roc_plot(aroc, target, reject_rate, 
       detect_xmin, detect_xmax, detect_xstep,
       detect_ymin, detect_ymax, detect_ystep,
       plot_detect_fname);

  /* call to output mac file */
  if (mac_detect_flag) {
      /* output mac file */
      fp= nn_fopen(mac_detectlines_fname,"w","Mac File");
      roc_print_table(fp,aroc,target,reject_rate,2);
      fclose(fp);
  }
  return 0;
}


/**************************************************************************
* NAME: roc_plot()
* DESCRIPTION:
*        Plots the roc plot identified by model_number, class, and reject.
*        xmin,xmax,xstep,ymin,ymax,ystep, determines the plot 
*          characteristics.
*        exp_file_name is the name of the plot file to output
*
*        Outputs roc curve information based on the target class and the 
*          reject rate.  If the currently stored information in the model 
*          is not of the same class and/or reject rate, this call will 
*          trigger a call to roc_batch_compute() to compute the correct 
*          information.
*        The amount of output information is controlled by verbosity flag:
*          0: no output
*          1: output a column of false alarm and a column of detect rate
*          2: output full table
*        Computes roc curve information based on patterns accumulated by
*          roc_batch_train().  The computation is determined by a target
*          class and the fractional rejection rate [0,1].
*
***************************************************************************/


void
roc_plot(int model_number, int class, float reject,
     float xmin, float xmax, float xstep,
     float ymin, float ymax, float ystep, char *exp_file_name)
{
    roc *aroc;
    int i;
    int detect_auto_scale_flag = 0;
    char title[128];
    int symbol_mode = 0, symbol_type = 0; /* no symbols */
    char *symbol_string = NULL;
    int symbol_radius = 0;
    int grid_type = 1; /* full grid */
    int line_type = 1; /* solid lines */
    int  auto_abscissas = 0; /* x provided by x array */
    float auto_xmin = 0.0, auto_xstep = 1.0; /* ignored */
    float height_fract = 0.9, width_fract = 1.0;
    float xoffset_fract = 0.15, yoffset_fract = 0.1;
    int xlog = 0, ylog = 0; /* linear x and y axes */
    char *xlabel = "%% False Alarm", *ylabel = "%% Detected";
    char *color = "red";
    char *command =
      "-geometry 500x520 -title detect_plot ";
    int quit_flag = 0;
    float axis_xmin, axis_xmax, axis_xstep;
    float axis_ymin, axis_ymax, axis_ystep;
    int ubot, vbot, utop, vtop;
    float area;
    float linex[2], liney[2];

    roc_param(model_number, &aroc);
    if (aroc==NULL)
      { warning("Invalid model number %d",model_number); return; }

    if (aroc->class!=class || aroc->reject!=reject)
      roc_batch_compute(model_number,class,reject);

    /* plot false_alarm versus detect_rate (length=table_length) */
    if (no_detect_graphics)
      ginit("no graphics");
    else
      ginit(command);
    ghardcopy(1);

    /* get area */
    area= roc_get_area(model_number,class,reject);
    
    if(no_borders){
    grid_type = 0;
    title[0] = '\0';
    xlabel = "";
    ylabel = "";
    detect_plot_label = "";
    } else {
    i = 0;
    sprintf(title, "Target=%d; Area=%.3f; #Target=%d; #Total=%d;",
        class, area*100., aroc->ndr[0], aroc->npatterns);
    }
    gwtitle("ROC Plot:");

    /* get size of plot */
    /* no symbols, no lines */
    xygraph_scale(aroc->false_alarm, aroc->detect_rate, aroc->table_length,
          2, symbol_string, symbol_type, symbol_radius,
          0, grid_type, detect_plot_label,
          auto_abscissas, auto_xmin, auto_xstep,
          height_fract, width_fract, xoffset_fract, yoffset_fract,
          xlog, detect_auto_scale_flag,
          xmin, xmax, xstep, "",
          ylog, detect_auto_scale_flag,
          ymin, ymax, ystep, "","foreground",
          &ubot, &vbot, &utop, &vtop,
          &axis_xmin, &axis_xmax, &axis_xstep,
          &axis_ymin, &axis_ymax, &axis_ystep);

    /* display label */
    /* uvprintf((utop+ubot)/2, vbot-75, 2, "ROC Plot");*/
    uvprintf((utop+ubot)/2, vbot-100, 2, title);

    /* do real plots using scale gotten from xygraph_scale() */
    xygraph(aroc->false_alarm, aroc->detect_rate, aroc->table_length, 
        symbol_mode, symbol_string, symbol_type, symbol_radius,
        line_type, grid_type, detect_plot_label,
        auto_abscissas, auto_xmin, auto_xstep,
        height_fract, width_fract, xoffset_fract, yoffset_fract,
        xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
        ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
        color);

    /* draw diagonal line */
    linex[0]=1.0; linex[1]=99.0;
    liney[0]=1.0; liney[1]=99.0;
    if(linex[0] < xmin) linex[0] = liney[0] = xmin;
    if(liney[0] < ymin) liney[0] = linex[0] = ymin;
    if(linex[1] < xmax) linex[1] = liney[1] = xmax;
    if(liney[1] < ymax) linex[1] = linex[1] = ymax;
    xygraph(linex, liney, 2,
        symbol_mode, symbol_string, symbol_type, symbol_radius,
        line_type, grid_type, detect_plot_label,
        auto_abscissas, auto_xmin, auto_xstep,
        height_fract, width_fract, xoffset_fract, yoffset_fract,
        xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
        ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
        "black");

    gsave(exp_file_name,1);
    gexit(quit_flag);

    exit_nnsubs();
    
}  
