/***-------------------------------------------------------------------------
 * NAME:	plot_gauss.c
 * PURPOSE:	Plots structure of a Gaussian classifier
 * USAGE:
 *
 * SEE ALSO:	
 *
 * DESCRIPTION: Reads in a gauss parameter file and plots the structure
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_gauss.c,v 11.4 2001/08/23 20:18:09 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_gauss.c,v $
 * CREATED: 2-Nov-93 lck			MIT LINCOLN LABORATORY
 -------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_gauss.c,v 11.4 2001/08/23 20:18:09 craig Exp $";


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#include "plot_struct.h"
#include <gauss_pack.h>
#include "xlippic.h"

#define BLOCK_RADIUS 20
#define BLOCK_SEPARATION (3*BLOCK_RADIUS)
#define LOWER_TEXT 10
#define CENTERED 2

#ifndef ABS
#define ABS(x) (((x) > 0)? (x) : -(x))
#endif

#define OUTPUT 1
#define INPUT 0

static int umiddle, voffset;
static int block_radius, lower_text;
static float layer_width[2];
static int layers[2];
static char **labels;
static char **input_labels;
static float struct_height = 1., struct_width = 1.;

static int node2u (int x, int layer);
static int level2v (int level);
static void draw_connection (int m, int level, int nodeu, int node);

int
main(int argc, char **argv)
{
    FILE *param_file;
    int model, normalization, n, m, layer;
    int noutputs, ninputs;
    int per_class, full_flag, *features;
    int ubot = 0, utop  = 1020, vbot = 0, vtop = 780;
    float scale;
    int nodeu, nodev;
    int nclass_labels;
    DEFAULTS_FILE_STRUCT df_args;
        
    init_nnsubs(argc, argv, rcsid);
    /********* read command line arguments ************/
    argc = nnargs(argc, argv, struct_plot_argtab, STRUCT_PLOT_TABSIZE);

    /********** open and read parameter file ***********/
    param_file = nn_fopen(struct_param_fname, "r",
			  "for read of gauss parameters");
    model = norm_read_paramfile(param_file, &normalization);
    model = gauss_read_model(param_file);
    nn_fclose(param_file, struct_param_fname, "for read of gauss parameters");

    if(strcmp(struct_defaults_fname, "") != 0){
	read_defaults_file(struct_defaults_fname, &df_args);
	labels = df_args.output_labels;
	input_labels = df_args.input_labels;
    } else {
	input_labels = labels = NULL;
    }

    /* initialize the gauss model for testing */
    gauss_test_build(model);
    /* do not adjust priors, NULL priors list */
    gauss_test_init(model, 0, NULL);

    /************* initialize graphics *************/
    if(no_graphics){
	ginit("no graphics");
    } else {
	ginit(" -geometry 510x450 -title struct_plot");
    }
    ghardcopy(1);
    gwtitle("GAUSS Structure Plot");
    
    /* get data from model */
    noutputs = *(int *)gauss_param(model, "noutputs", 0);
    ninputs = *(int *)gauss_param(model, "ninputs", 0);
    per_class = *(int *) gauss_param(model, "per_class", 0);
    full_flag = *(int *) gauss_param(model, "full_covar", 0);
    features = *(int **) gauss_param(model, "features", 0);

    umiddle = (utop - ubot)/2;

    /* write out the plot title */
    uvprintf(umiddle, vtop - 20, CENTERED, "%s", struct_param_fname);
    
    /* write out the covariance types */
    if(!struct_no_text){
	if(per_class)
	  uvprintf(umiddle, vbot+40, CENTERED, "Per Class");
	else
	  uvprintf(umiddle, vbot+40, CENTERED, "Grand");
	if(full_flag)
	  uvprintf(umiddle, vbot+20, CENTERED, "Full Covariance");
	else
	  uvprintf(umiddle, vbot+20, CENTERED, "Diagonal Covariance");
    }

    vtop -= 20;
    vbot += 40;

    layers[INPUT] = ninputs;
    layers[OUTPUT] = noutputs;
    if(struct_autoscale){
	struct_height = (vtop-vbot-40)/(float)(6*BLOCK_SEPARATION);

	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] =
	      (utop-ubot)/(float)((layers[layer]+2)*BLOCK_SEPARATION);
	}
	struct_width = layer_width[OUTPUT];
    } else {
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] = struct_width;
	}
    }
    /* figure out reference positions for actual plotting */
    scale = (struct_width < struct_height) ? struct_width : struct_height;
    block_radius = BLOCK_RADIUS*scale;
    lower_text = LOWER_TEXT+block_radius;

    voffset = (vtop - vbot)/2;

    /* write out the class labels */
    nodev = level2v(OUTPUT);
    if(labels != NULL) nclass_labels = atoi(labels[0]);
    else nclass_labels = 0;
    for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);
	if(!struct_no_text){
	    if(n < nclass_labels)
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%s",
		       labels[n+1]);
	    else
	      uvprintf(nodeu, nodev+lower_text , CENTERED, "%d", n);
	}
    }

    /* display the output nodes */
    nodev = level2v(OUTPUT);
    for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);

	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);

	/* draw connecting lines */
	for(m = 0; m < ninputs; m++){
	    draw_connection(m, INPUT, nodeu, nodev);
	}
    }

    /* draw input node squares */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	/* draw a filled circle */
	uvdot("foreground", nodeu, nodev, block_radius/2);
    }

    /* write out the input labels */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	if(!struct_no_text){
	    uvprintf(nodeu, nodev-lower_text, CENTERED,
		     "%s", get_input_label(n, input_labels, features,
					   normalization));
	}
    }

    gflush();
    gsave(struct_plot_fname, 1);
    gexit(0);/* quit */
    exit_nnsubs();
    exit(0);
}
static int
node2u(int x, int layer)
{
    /* center the node */
    x -= layers[layer]/2;
    /* scale the position for u coordinates*/
    x *= BLOCK_SEPARATION*layer_width[layer];
    /* center the final position on the screen */
    return(x + umiddle);
}
static int
level2v(int level)
{
    return(voffset + level*2*BLOCK_SEPARATION*struct_height);
}

static void
draw_connection(int m, int level, int nodeu, int nodev)
{
    int v1, v2, u1, u2, ulength, vlength, udir, vdir, r;
    int childu, childv;
    childu = node2u(m, level);
    childv = level2v(level);
    
    ulength = ABS(nodeu-childu);
    udir = (nodeu > childu) ? 1 : -1;
    vlength = ABS(nodev-childv);
    vdir = (nodev > childv) ? 1 : -1;
    r = (int)sqrt(((double)ulength*ulength)
		  +(double)vlength*vlength);
    u1 = nodeu-udir*(ulength*block_radius)/r;
    v1 = nodev-vdir*(vlength*block_radius)/r;
    
    if(level == INPUT){
	u2 = childu;
	v2 = childv;
    } else {
	u2 = childu+udir*(ulength*block_radius)/r;
	v2 = childv+vdir*(vlength*block_radius)/r;
    }
    
    uvline(u1, v1, u2, v2);
}
