/***-------------------------------------------------------------------------
 * NAME:	plot_gmix.c
 * PURPOSE:	Plots structure of a Gaussian mixture classifier
 * USAGE:
 *
 * SEE ALSO:	
 *
 * DESCRIPTION: Reads in a gmix parameter file and plots the structure
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_gmix.c,v 11.4 2001/08/23 20:18:09 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_gmix.c,v $
 * CREATED: 2-Nov-93 lck			MIT LINCOLN LABORATORY
 -------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_gmix.c,v 11.4 2001/08/23 20:18:09 craig Exp $";


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#include "plot_struct.h"
#include <gmix_pack.h>

#include "xlippic.h"

static float threshold = 0.;
static float line_scale = 1.;
static int max_line_width = 10;
static int scale_by_weights = 0;
static ARG extra_plot_argtab[] = 
{
    {"threshold", FLOAT, (int *)&threshold,
       "Do not draw connection if ABS(weight) < threshold"},
    {"show_weight_magnitude", BOOLEAN, &scale_by_weights,
       "display connection weights by changing line widths"},
    {"max_line_width", INT, &max_line_width,
       "width of line for maximum weight"},
};
#define EXTRA_PLOT_TABSIZE (sizeof(extra_plot_argtab)/sizeof(ARG))

#define BLOCK_RADIUS 20
#define BLOCK_SEPARATION (3*BLOCK_RADIUS)
#define LOWER_TEXT 10
#define CENTERED 2

#ifndef ABS
#define ABS(x) (((x) > 0)? (x) : -(x))
#endif

#define OUTPUT 2
#define MIDDLE 1
#define INPUT 0

static int node2u (int x, int layer);
static int level2v (int level);
static void draw_connection (int m, int level, int nodeu, int nodev,
			     float log_weight);

static int umiddle, voffset;
static int block_radius, lower_text;
static float layer_width[3];
static int layers[3];
static char **labels;
static char **input_labels;
static float struct_height = 1., struct_width = 1.;

int
main(int argc, char **argv)
{
    FILE *param_file;
    int model, normalization, n, m, total_m, layer;
    int noutputs, ninputs, *ngauss_list, ngauss;
    int tied_flag, grand_flag, full_flag;
    int ubot = 0, utop  = 1020, vbot = 0, vtop = 780;
    float scale, log_weight;
    int nodeu, nodev, *features;
    ARG *argtab;
    int tabsize;
    DEFAULTS_FILE_STRUCT df_args;
        
    init_nnsubs(argc, argv, rcsid);
    /********* read command line arguments ************/
    nnargs_combine_argtabs(struct_plot_argtab, STRUCT_PLOT_TABSIZE,
			   extra_plot_argtab, EXTRA_PLOT_TABSIZE,
			   &argtab, &tabsize);
    argc = nnargs(argc, argv, argtab, tabsize);
    nnargs_free_argtab(argtab, tabsize);

    /********** open and read parameter file ***********/
    param_file = nn_fopen(struct_param_fname, "r",
			  "for read of gmix parameters");
    model = norm_read_paramfile(param_file, &normalization);
    model = gmix_read_model(param_file);
    nn_fclose(param_file, struct_param_fname, "for read of gmix parameters");

    if(strcmp(struct_defaults_fname, "") != 0){
	read_defaults_file(struct_defaults_fname, &df_args);
	labels = df_args.output_labels;
	input_labels = df_args.input_labels;
    } else {
	input_labels = labels = NULL;
    }

    /* initialize the gmix model for testing */
    /* do not adjust priors, NULL priors list */
    gmix_test_init(model, 0, NULL);

    /************* initialize graphics *************/
    if(no_graphics){
	ginit("no graphics");
    } else {
	ginit(" -geometry 510x450 -title struct_plot");
    }
    ghardcopy(1);
    gwtitle("GMIX Structure Plot");
    
    /* get data from model */
    noutputs = *(int *)gmix_param(model, "noutputs", 0, 0);
    ninputs = *(int *)gmix_param(model, "ninputs", 0, 0);
    tied_flag = *(int *) gmix_param(model, "tied_flag", 0, 0);
    grand_flag = *(int *) gmix_param(model, "grand_flag", 0, 0);
    full_flag = *(int *) gmix_param(model, "full_flag", 0, 0);
    features = *(int **) gmix_param(model, "features", 0, 0);

    /* figure out reference positions for actual plotting */
    ngauss_list = allocate_ivector(noutputs);
    if(tied_flag){
	ngauss = *(int *)gmix_param(model, "ngauss", 0, 0);
    } else {
	/* find the number of gaussians per class and the total */
	ngauss = 0;
	for(n = 0; n < noutputs; n++){
	    ngauss_list[n] = *(int *)gmix_param(model, "class_ngauss", n, 0);
	    ngauss += ngauss_list[n];
	}
    }

    umiddle = (utop - ubot)/2;

    /* write out the plot title */
    uvprintf(umiddle, vtop - 20, CENTERED, "%s", struct_param_fname);

    /* write out the covariance types */
    if(!struct_no_text){
	if(tied_flag && (grand_flag || grand_flag == 2))
	  uvprintf(umiddle, vbot+40, CENTERED, "Grand");
	else if (grand_flag == 0)
	  uvprintf(umiddle, vbot+40, CENTERED, "Per Mixture Component");
	else if(!tied_flag)
	  uvprintf(umiddle, vbot+40, CENTERED, "Grand Per Class");
	if(full_flag)
	  uvprintf(umiddle, vbot+20, CENTERED, "Full Covariance");
	else
	  uvprintf(umiddle, vbot+20, CENTERED, "Diagonal Covariance");
    }

    vtop -= 20;
    vbot += 40;
    voffset = (vtop - vbot)/2;

    layers[INPUT] = ninputs;
    layers[MIDDLE] = ngauss;
    layers[OUTPUT] = noutputs;
    if(struct_autoscale){
	struct_height = (vtop-vbot-40)/(float)(6*BLOCK_SEPARATION);
	struct_width = (utop-ubot)/(float)((ngauss+2)*BLOCK_SEPARATION);
	
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] =
	      (utop-ubot)/(float)((layers[layer]+2)*BLOCK_SEPARATION);
	}
    } else {
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] = struct_width;
	}
    }
    scale = (struct_width < struct_height) ? struct_width : struct_height;
    block_radius = BLOCK_RADIUS*scale;
    lower_text = LOWER_TEXT+block_radius;

    /* write out the class labels */
    nodev = level2v(OUTPUT);
   for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);
	if(!struct_no_text){
	    if(labels != NULL)
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%s",
		       labels[n+1]);
	    else
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%d", n);
	}
    }

    if(scale_by_weights){
	float max_weight = -MAXFLOAT;
	/* find the max log weight */
	if(tied_flag){
	    for(n = 0; n < noutputs; n++){
		for(m = 0; m < ngauss; m++){
		    log_weight = *(float *)gmix_param(model, "log_weight",
						      n, m);
		    if(log_weight > max_weight) max_weight = log_weight;
		}
	    }
	} else {
	    for(n = 0; n < noutputs; n++){
		for(m = 0; m < ngauss_list[n]; m++){
		    log_weight = *(float *)gmix_param(model, "log_weight",
						      n, m);
		    if(log_weight > max_weight) max_weight = log_weight;
		}
	    }
	}
	/* convert to linear */
	max_weight = nn_exp10(max_weight);
	line_scale = (max_line_width-1)/max_weight;
    }

    /* prepare threshold for log domain */
    if(threshold > 0) threshold = (float)log10((double)threshold);
    else threshold = -MAXFLOAT;

    /* display the output nodes */
    total_m = 0;
    nodev = level2v(OUTPUT);
    for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);

	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);

	/* draw connecting lines */
	if(tied_flag){
	    for(m = 0; m < ngauss; m++){
		/* check weight */
		log_weight = *(float *)gmix_param(model, "log_weight", n, m);
		if(log_weight < threshold) continue;

		draw_connection(m, MIDDLE, nodeu, nodev,
				log_weight);
	    }
	} else { /* untied mixtures */
	    for(m = 0; m < ngauss_list[n]; m++, total_m++){
		log_weight = *(float *)gmix_param(model, "log_weight", n, m);
		if(log_weight < threshold) continue;
		
		draw_connection(total_m, MIDDLE, nodeu, nodev,
				log_weight);
	    }
	}
    }


    /* turn off scale by weights for input layer */
    scale_by_weights = 0;

    /* draw hidden node circles and connections */
    nodev = level2v(MIDDLE);
    for(m = 0; m < ngauss; m++){
	nodeu = node2u(m, MIDDLE);
	
	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);
	for(n = 0; n < ninputs; n++){
	    draw_connection(n, INPUT, nodeu, nodev, 0.);
	}
    }
    
    /* draw input node squares */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	/* draw a filled circle */
	uvdot("foreground", nodeu, nodev, block_radius/2);
    }

    /* write out the input labels */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	if(!struct_no_text){
	    uvprintf(nodeu, nodev-lower_text, CENTERED,
		     "%s",  get_input_label(n, input_labels, features,
					    normalization));
	}
    }

    gflush();
    gsave(struct_plot_fname, 1);
    gexit(0);/* quit */
    exit_nnsubs();
    exit(0);
}
static int
node2u(int x, int layer)
{
    /* center the node */
    x -= layers[layer]/2;
    /* scale the position for u coordinates*/
    x *= BLOCK_SEPARATION*layer_width[layer];
    /* center the final position on the screen */
    return(x + umiddle);
}

static int
level2v(int level)
{
    return(voffset + (level - 1)*(2*BLOCK_SEPARATION*struct_height));
}

static void
draw_connection(int m, int level, int nodeu, int nodev, float log_weight)
{
    int v1, v2, u1, u2, ulength, vlength, udir, vdir, r;
    int line_width;
    int childu, childv;
    childu = node2u(m, level);
    childv = level2v(level);
    
    ulength = ABS(nodeu-childu);
    udir = (nodeu > childu) ? 1 : -1;
    vlength = ABS(nodev-childv);
    vdir = (nodev > childv) ? 1 : -1;
    r = (int)sqrt(((double)ulength*ulength)
		  +(double)vlength*vlength);
    u1 = nodeu-udir*(ulength*block_radius)/r;
    v1 = nodev-vdir*(vlength*block_radius)/r;
    
    if(level == INPUT){
	u2 = childu;
	v2 = childv;
    } else {
	u2 = childu+udir*(ulength*block_radius)/r;
	v2 = childv+vdir*(vlength*block_radius)/r;
    }
    
    
    if(scale_by_weights){
	line_width = (int)(line_scale*nn_exp10(log_weight))+1;
	uvbeam(u1, v1, u2, v2, line_width);
    } else
      uvline(u1, v1, u2, v2);
}
