/***-------------------------------------------------------------------------
 * NAME:	plot_irbf.c
 * PURPOSE:	Plots structure of a radial basis function classifier
 * USAGE:
 *
 * SEE ALSO:	
 *
 * DESCRIPTION: Reads in a irbf parameter file and plots the structure
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_irbf.c,v 11.4 2001/08/23 20:18:09 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_irbf.c,v $
 * CREATED: 2-Nov-93 lck			MIT LINCOLN LABORATORY
 -------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_irbf.c,v 11.4 2001/08/23 20:18:09 craig Exp $";


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#include "plot_struct.h"
#include <irbf_pack.h>
#include "xlippic.h"

static float threshold = 0.;
static float line_scale = 1.;
static int max_line_width = 10;
static int scale_by_weights = 0;
static int black_and_white = 0;
static ARG extra_plot_argtab[] = 
{
    {"threshold", FLOAT, (int *)&threshold,
       "Do not draw connection if ABS(weight) < threshold"},
    {"show_weight_magnitude", BOOLEAN, &scale_by_weights,
       "display connection weights by changing line widths"},
    {"max_line_width", INT, &max_line_width,
       "width of line for maximum weight"},
    {"bw", BOOLEAN, &black_and_white,
       "plot negative weights as hollow. Default: plot as orange"},
};
#define EXTRA_PLOT_TABSIZE (sizeof(extra_plot_argtab)/sizeof(ARG))

#define BLOCK_RADIUS 20
#define BLOCK_SEPARATION (3*BLOCK_RADIUS)
#define LOWER_TEXT 10
#define CENTERED 2

#ifndef ABS
#define ABS(x) (((x) > 0)? (x) : -(x))
#endif

#define OUTPUT 2
#define MIDDLE 1
#define INPUT 0
#define LARGE 0
#define SMALL 1

static int umiddle, voffset;
static int block_radius, lower_text;
static float layer_width[3];
static int layers[3];
static char **labels;
static char **input_labels;
static float struct_height = 1., struct_width = 1.;

static int node2u (int x, int layer);
static int level2v (int level);
static void draw_connection(int m, int level, int nodeu, int nodev,
			    float weight, int type);

int
main(int argc, char **argv)
{
    FILE *param_file;
    int model, normalization, n, m, layer;
    int noutputs, ninputs, ngauss, bias;
    int ubot = 0, utop  = 1020, vbot = 0, vtop = 780;
    float scale, **weights;
    int nodeu, nodev, *features;
    ARG *argtab;
    int tabsize;
    float weight_etta = 0, mean_etta = 0, inv_var_etta = 0, min_inv_var = 0;
    DEFAULTS_FILE_STRUCT df_args;
        
    init_nnsubs(argc, argv, rcsid);
    /********* read command line arguments ************/
    nnargs_combine_argtabs(struct_plot_argtab, STRUCT_PLOT_TABSIZE,
			   extra_plot_argtab, EXTRA_PLOT_TABSIZE,
			   &argtab, &tabsize);
    argc = nnargs(argc, argv, argtab, tabsize);
    nnargs_free_argtab(argtab, tabsize);

    /********** open and read parameter file ***********/
    param_file = nn_fopen(struct_param_fname, "r",
			  "for read of irbf parameters");
    model = norm_read_paramfile(param_file, &normalization);
    model = irbf_read_model(param_file, weight_etta, mean_etta, inv_var_etta,
			    min_inv_var);
    nn_fclose(param_file, struct_param_fname, "for read of irbf parameters");

    if(strcmp(struct_defaults_fname, "") != 0){
	read_defaults_file(struct_defaults_fname, &df_args);
	labels = df_args.output_labels;
	input_labels = df_args.input_labels;
    } else {
	input_labels = labels = NULL;
    }

    /* initialize the irbf model for testing */
    /* do not adjust priors, NULL priors list, 1. exspread */
    irbf_test_init(model, 0, NULL, 1.);

    /************* initialize graphics *************/
    if(no_graphics){
	ginit("no graphics");
    } else {
	ginit(" -geometry 510x450 -title struct_plot");
    }
    ghardcopy(1);
    gwtitle("IRBF Structure Plot");
    
    /* get data from model */
    noutputs = *(int *)irbf_param(model, "nclasses");
    ninputs = *(int *)irbf_param(model, "ninputs");
    ngauss = *(int *)irbf_param(model, "ngauss_hidden");
    weights = *(float ***)irbf_param(model, "weights");
    bias = *(int *)irbf_param(model, "bias");
    if(bias) bias = 1;
    features = *(int **) irbf_param(model, "features");

    umiddle = (utop - ubot)/2;

    /* write out the plot title */
    uvprintf(umiddle, vtop - 20, CENTERED, "%s", struct_param_fname);

    /* write out the covariance types */
    if(!struct_no_text){
	if(bias)
	  uvprintf(umiddle, vbot + 20, CENTERED, "Bias Node Included");
    }

    vtop -= 20;
    vbot += 20;
    voffset = (vtop - vbot)/2;

    layers[INPUT] = ninputs;
    layers[MIDDLE] = ngauss+bias;
    layers[OUTPUT] = noutputs;
    if(struct_autoscale){
	struct_height = (vtop-vbot-40)/(float)(6*BLOCK_SEPARATION);

	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] =
	      (utop-ubot)/(float)((layers[layer]+2)*BLOCK_SEPARATION);
	}
	struct_width = layer_width[MIDDLE];
    } else {
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] = struct_width;
	}
    }
    scale = (struct_width < struct_height) ? struct_width : struct_height;
    block_radius = BLOCK_RADIUS*scale;
    lower_text = LOWER_TEXT+block_radius;

    /* write out the class labels */
    nodev = level2v(OUTPUT);
    for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);
	if(!struct_no_text){
	    if(labels != NULL)
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%s",
		       labels[n+1]);
	    else
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%d", n);
	}
    }

    if(scale_by_weights){
	float temp, max_weight = -MAXFLOAT;
	/* find the max ABS weight */
	for(n = 0; n < noutputs; n++){
	    for(m = 0; m < ngauss+bias; m++){
		temp = ABS(weights[m][n]);
		if(temp > max_weight) max_weight = temp;
	    }
	}
	line_scale = (max_line_width-1)/max_weight;
    }

    /* display the output nodes */
    nodev = level2v(OUTPUT);
    for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);

	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);

	/* draw connecting lines */
	for(m = 0; m < ngauss; m++){
	    /* check weight */
	    if(ABS(weights[m][n]) < threshold) continue;
	    draw_connection(m, MIDDLE, nodeu, nodev,
			    weights[m][n], LARGE);
	}
	/* draw connection for bias node */
	if(bias){
	    m = ngauss;
	    if(ABS(weights[m][n]) >= threshold)
	      draw_connection(m, MIDDLE, nodeu, nodev,
			      weights[m][n], SMALL);
	}
    }

    /* turn off scale by weights for input layer */
    scale_by_weights = 0;

    /* draw hidden node circles and connections */
    nodev = level2v(MIDDLE);
    for(m = 0; m < ngauss; m++){
	nodeu = node2u(m, MIDDLE);
	
	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);
	for(n = 0; n < ninputs; n++){
	    draw_connection(n, INPUT, nodeu, nodev, 1., SMALL);
	}
    }
    /* draw bias node */
    if(bias){
	nodeu = node2u(ngauss, MIDDLE);
	uvblock("foreground",
		nodeu-block_radius/2, nodev-block_radius/2,
		nodeu+block_radius/2, nodev+block_radius/2);
    }
    /* draw input node squares */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	/* draw a filled circle */
	uvdot("foreground", nodeu, nodev, block_radius/2);
    }

    /* write out the input labels */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	if(!struct_no_text){
	    uvprintf(nodeu, nodev-lower_text, CENTERED,
		     "%s", get_input_label(n, input_labels, features,
					   normalization));
	}
    }

    gflush();
    gsave(struct_plot_fname, 1);
    gexit(0);/* quit */
    exit_nnsubs();
    exit(0);
}

static int
node2u(int x, int layer)
{
    /* center the node */
    x -= layers[layer]/2;
    /* scale the position for u coordinates*/
    x *= BLOCK_SEPARATION*layer_width[layer];
    /* center the final position on the screen */
    return(x + umiddle);
}
static int
level2v(int level)
{
    return(voffset + (level - 1)*(2*BLOCK_SEPARATION*struct_height));
}

static void
draw_connection(int m, int level, int nodeu, int nodev, float weight, int type)
{
    int v1, v2, u1, u2, ulength, vlength, udir, vdir, r;
    int line_width;
    int childu, childv;
    childu = node2u(m, level);
    childv = level2v(level);
    
    ulength = ABS(nodeu-childu);
    udir = (nodeu > childu) ? 1 : -1;
    vlength = ABS(nodev-childv);
    vdir = (nodev > childv) ? 1 : -1;
    r = (int)sqrt(((double)ulength*ulength)
		  +(double)vlength*vlength);
    u1 = nodeu-udir*(ulength*block_radius)/r;
    v1 = nodev-vdir*(vlength*block_radius)/r;
    
    if(type == SMALL){
	u2 = childu;
	v2 = childv;
    } else {
	u2 = childu+udir*(ulength*block_radius)/r;
	v2 = childv+vdir*(vlength*block_radius)/r;
    }
    
    if(scale_by_weights){
	line_width = (int)(line_scale*ABS(weight))+1;
	if(weight >= 0) uvbeam(u1, v1, u2, v2, line_width);
	else{
	    if(black_and_white)
	      uvbeam_outline(u1, v1, u2, v2, line_width);
	    else{
		gcolor("salmon");
		uvbeam(u1, v1, u2, v2, line_width);
		gcolor("foreground");
	    }
	}
    } else
      uvline(u1, v1, u2, v2);
}
