/***--------------------------------------------------------------------------
 * NAME:		plot_map.c
 * PURPOSE:	Plots 1 dimensional output vs input mapping plots
 * USAGE:
 *
 * SEE ALSO: 	mlp_plot_map.c
 *
 * DESCRIPTION:  Reads in a trained model file and plots the mapping
 *		function for one input and all output dimensions
 *		This program requires 2 classifier dependent subroutines
 *		    (read_model and eval) and the software they call 
 *
 *	main		Initialize, Read model file
 *	find_outputs	Get output values only
 *	x2u		Convert from x to u coordinates
 *	y2v		Convert from y to v coordinates
 *	xlength2u	Convert x length to u coordinates
 *	ylength2v	Convert y length to v coordinates
 *	plot_profile	Draw output values of 1 dimension of input data
 *	plot_histogram	Overlay profile with histogram of data
 *	calc_bound_autoscale
 *			Autoscale bound_ymax, bound_xmax, etc.
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_map.c,v 11.4 2001/08/23 20:18:09 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_map.c,v $
 * CREATED: 13-Jul-92 lck			MIT Lincoln Laboratory
 --------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_map.c,v 11.4 2001/08/23 20:18:09 craig Exp $";


/* Include necessary header files */
/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <math.h>
#include <stdio.h>
#if HAVE_STRINGS_H
#include <strings.h>
#endif
#include <nnlib.h>
#include <xlippic.h>
#include "xlippic.h"
#define PLOT_MAP
#include "plot_map_args.h"
#include "plot_colors.h"
#include "plot_map.h"
#define FALSE 0
#define TRUE 1
#define SQUARE 0
#define SOLID_BLOCK 6
#define CIRCLE 7
#define FILLED_CIRCLE 8
#define NSYMBOLS 12

/*** GLOBALS ***/ 
static float *inputs, *slice_inputs, *outputs, *desired_outputs, *inputs_copy;
static float *graph_y, *graph_x;
static int ubot, vbot, utop, vtop, uwidth, vwidth;
static float xwidth, ywidth;
static int ninputs, noutputs;
static int normalization, norm_temp;
static int features[NNARGS_MAX_LABELS] = {0};
static int *nn_features;

/*** MAIN DRIVER ROUTINE ***/
int
main(register int argc, register char **argv)
{
    FILE *param_file;
    int total_trials, i;
    char *plot_error_fullname, print_filename[100];
    int rv;
    
    init_nnsubs(argc, argv, rcsid);
    /* Process the command line arguments */
    argc = get_extra_map_args(argc, argv);
    argc = nnargs(argc, argv, map_argtab, MAP_TABSIZE);

    fix_features(plot_n_inputs, plot_n_inputs, features);
    nn_features = &features[0];
    
    plot_error_fullname = path_build(plot_exp_path, plot_error_fname);
    /* Open the file, check for errors */
    if(outputs_plot){
	param_file = nn_fopen (plot_error_fullname, "r",
			       "for read of weight structure information");
	rv = norm_read_paramfile(param_file, &normalization);
	if (rv < 0)
	  fatal_error ("plot_map: read error %d %s %s", rv,
		       "on normalization parameter file" ,plot_error_fullname);
	read_model (param_file, &total_trials, &ninputs, &noutputs,
		    &nn_features);
	nn_fclose(param_file,plot_error_fullname,
		  "for read of weight structure information");
	outputs = get_outputs();
    } else {
	ninputs = plot_n_inputs;
	noutputs = plot_n_outputs;
	outputs = allocate_vector(plot_n_outputs);
    }
    inputs = allocate_vector(plot_n_inputs);
    slice_inputs = allocate_vector(plot_n_inputs);
    if((int)map_other_vars[0] == plot_n_inputs){
	for(i = 0; i < plot_n_inputs; i++){
	    slice_inputs[i] = map_other_vars[i+1];
	}
    }
    if(map_no_norm){
	norm_temp = normalization;
	normalization = 0;
	inputs_copy = allocate_vector(plot_n_inputs);
    }
    desired_outputs = allocate_vector(plot_n_outputs);
    
    if (map_in_dim >= ninputs) {
	warning("%s %d %s %d, setting to %d",
		"plot_map main: plot input dimension value", map_in_dim,
		"exceeds maximum number of inputs", ninputs, ninputs);
	map_in_dim = ninputs - 1;
    }
    if (map_out_dim >= noutputs) {
	warning("%s %d %s %d, setting to %d",
		"plot_map main: plot output dimension value", map_out_dim,
		"exceeds maximum number of outputs", noutputs, noutputs);
	map_out_dim = noutputs - 1;
    }
    
    strcpy (print_filename, path_build(plot_exp_path, plot_fname));
    color_table[plot_n_outputs] = "foreground";
    plot_map_in_window (plot_error_fname, print_filename, map_plot_label);

    exit_nnsubs();
    exit(0);
}				/* END main() */

void
plot_map_in_window(char *error_filename, char *print_filename,
		   char *plot_label)
{
    float dummy_x[1], dummy_y[1];
    float axis_xmin, axis_xmax, axis_xstep;
    float axis_ymin, axis_ymax, axis_ystep;
    char xlabel[50], ylabel[50];
    float rms_err;
    int legend_x;
    float height_fract = 1., width_fract = 1.;
    float xoffset_fract = 0.1, yoffset_fract = 0.;
    char plot_heading[100];
    char rms_string[20];
    FILE *data_file;
    
    if(no_map_graphics){
	ginit("no graphics");
    } else {
	ginit("-fg black -bg white -geometry 510x450");
    }
    ghardcopy(1);
    gwtitle("Mapping Plot");
    
    if (map_autoscale) {
	data_file = nn_fopen (map_scat_data_fullname, "r",
			      "for read of train or test data for autoscale");
	calc_map_autoscale(data_file);
	nn_fclose(data_file,map_scat_data_fullname,
		  "for read of train or test data for autoscale");
    }
    /* get limits of graph */
    xygraph_scale (dummy_x, dummy_y, 0,
		   0, 0, 0, 10,
		   1, 2, "",
		   0, 0.0, 0.0,
		   height_fract, width_fract, xoffset_fract, yoffset_fract,
		   0, 0, map_xmin, map_xmax, map_xstep, "xlabel",
		   0, 0, map_ymin, map_ymax, map_ystep, "ylabel", "black",
		   &ubot, &vbot, &utop, &vtop,
		   &axis_xmin, &axis_xmax, &axis_xstep,
		   &axis_ymin, &axis_ymax, &axis_ystep);
    /* set up for x/y-u/v conversion function */
    uwidth = utop - ubot;
    vwidth = vtop - vbot;
    xwidth = map_xmax - map_xmin;
    ywidth = map_ymax - map_ymin;

    strcpy(plot_heading, "");
    if(outputs_plot){
	strcat(plot_heading, error_filename);
	if(outputs_plot){
	    graph_x = allocate_vector(map_npoints+1);
	    graph_y = allocate_vector(map_npoints+1);
	    plot_outputs();
	}
	if(map_scat_plot)
	    strcat(plot_heading, "  ");
    }
    if (map_scat_plot){
	strcat (plot_heading, map_scat_data_fullname);
    }

    gcolor("foreground");
    sprintf(xlabel,"X%d ",map_in_dim);
    sprintf(ylabel,"Y%d", map_out_dim);
    uvprintf(512,vbot/5,2,plot_label);
    /* plot grid */
    xygraph(dummy_x, dummy_y, 0,
	    0, 0, 0, 10,
	    1, 1, plot_heading,
	    0, 0.0, 0.0,
	    height_fract, width_fract, xoffset_fract, yoffset_fract,
	    0, 0, map_xmin, map_xmax, map_xstep, xlabel,
	    0, 0, map_ymin, map_ymax, map_ystep, ylabel, "black");
    /* plot 1-D output values */
    if (outputs_plot)
      xygraph(graph_x, graph_y, map_npoints+1,
	      0, 0, 0, 10,
	      1, 0, "",
	      0, 0.0, 0.0,
	      height_fract, width_fract, xoffset_fract, yoffset_fract,
	      0, 0, map_xmin, map_xmax, map_xstep, "",
	      0, 0, map_ymin, map_ymax, map_ystep, "", "foreground");

    if (map_scat_plot){
	legend_x = utop + 40;
	data_file = nn_fopen (map_scat_data_fullname, "r",
			      "for read of data for scatter plot");
	rms_err = plot_scatter (data_file, map_scat_n_patterns);
	
	if(outputs_plot){
	    gcolor("foreground");
	    uvprintf(legend_x, vbot + 30, 1, cost_string(SQ_ERR));
	    sprintf(rms_string, "%6.2f", rms_err);
	    uvprintf (legend_x, vbot, 1, rms_string);
	    nn_fclose(data_file,map_scat_data_fullname,
		      "for read of train or test data for scatter plot");
	}
    }
    
    gsave (print_filename, 0);
    gexit(0);
}

void
find_outputs (register float x)
{				/* Get outputs for the desired dimension */
    int i;
    
    slice_inputs[map_in_dim]=x;
    if(map_no_norm && norm_temp){
	for(i = 0; i < plot_n_inputs; i++){
	    inputs_copy[i] = slice_inputs[i];
	}
	norm_apply(inputs_copy);
	eval(inputs_copy);
    } else eval (slice_inputs);
}

void plot_outputs (void)
{
    register float xval,del_x;
    int indx;
    
    /* Calculate the increments */
    del_x = xwidth / (float) map_npoints;

    xval = map_xmin;
    for (indx = 0; indx <= map_npoints; indx++, xval += del_x) {
	graph_x [indx] = xval;
	find_outputs(xval);
	graph_y[indx] = outputs[map_out_dim];
    }
}

float plot_scatter(FILE *data_file, int n_patterns)
{
    int i, u1, v1;
    int symbol_radius = 3;
    float rms_err = 0;

    gcolor("foreground");
    for(i = 0; i < n_patterns; i++){
	if( get_map_line (data_file, map_scat_data_fullname, inputs,
			  desired_outputs, plot_n_inputs, ninputs, noutputs,
			  normalization, nn_features) == EOF) break;
	u1 = x2u (inputs[map_in_dim]);
	v1 = y2v (desired_outputs[map_out_dim]);
	if(u1 >= ubot && u1 <= utop && v1 >= vbot && v1 <= vtop){
	    uvmove(u1, v1);
	    gsymbol(SOLID_BLOCK, symbol_radius);
	}
	if(outputs_plot){
	    if(map_no_norm && norm_temp) norm_apply(inputs);
	    eval(inputs);
	    rms_err += calc_cost(outputs, desired_outputs, noutputs, SQ_ERR);
	}
    }
    rms_err = calc_avg_cost(rms_err, i, SQ_ERR);
    return rms_err;
}

void calc_map_autoscale(FILE *data_file)
{
    float tmp;
    
    if( get_map_line (data_file, map_scat_data_fullname, inputs,
		      outputs, plot_n_inputs, ninputs, noutputs,
		      normalization, nn_features) != EOF) {
	map_xmin = map_xmax = inputs[map_in_dim];
	map_ymin = map_ymax = outputs[map_out_dim];
	while (1) {
	    if( get_map_line (data_file, map_scat_data_fullname, inputs,
			      outputs, plot_n_inputs, ninputs, noutputs,
			      normalization, nn_features) == EOF) break;
	    tmp = inputs[map_in_dim];
	    if (tmp < map_xmin) map_xmin = tmp;
	    if (tmp > map_xmax) map_xmax = tmp;
	    tmp = outputs[map_out_dim];
	    if (tmp < map_ymin) map_ymin = tmp;
	    if (tmp > map_ymax) map_ymax = tmp;
	}
	/* add some on each side to make it understandable */
	map_xmin = floor((double) map_xmin);
	map_xmax = ceil((double) map_xmax);
	if (map_xmin == map_xmax) {
	    map_xmin = map_xmin - 1;
	    map_xmax = map_xmax + 1;
	}
	map_xstep = (map_xmax - map_xmin) / 10.;
	map_ymin = floor((double) map_ymin);
	map_ymax = ceil((double) map_ymax);
	if (map_ymin == map_ymax) {
	    map_ymin = map_ymin - 1;
	    map_ymax = map_ymax + 1;
	}
    }
}

int x2u (float raw_x)
{
    return (ubot + (raw_x - map_xmin) * uwidth / xwidth);
}

int y2v (float raw_y)
{
    return (vbot + (raw_y - map_ymin) * vwidth / ywidth);
}

int xlength2u (float raw_x)
{
    return (raw_x * uwidth / xwidth);
}

int ylength2v (float raw_y)
{
    return (raw_y * vwidth / ywidth);
}
