/***-------------------------------------------------------------------------
 * NAME:	plot_mlp.c
 * PURPOSE:	Plots structure of a multi-layer perceptron classifier
 * USAGE:
 *
 * SEE ALSO:	
 *
 * DESCRIPTION: Reads in a mlp parameter file and plots the structure
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_mlp.c,v 11.4 2001/08/23 20:18:09 craig Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_mlp.c,v $
 * CREATED: 2-Nov-93 lck			MIT LINCOLN LABORATORY
 -------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_mlp.c,v 11.4 2001/08/23 20:18:09 craig Exp $";


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#include "plot_struct.h"
#include <mlp_pack.h>
#include "xlippic.h"

static float threshold = 0.;
static float line_scale = 1.;
static int max_line_width = 10;
static int scale_by_weights = 0;
static int show_bias = 0;
static int black_and_white = 0;
static ARG extra_plot_argtab[] = 
{
    {"threshold", FLOAT, (int *)&threshold,
       "Do not draw connection if ABS(weight) < threshold"},
    {"show_weight_magnitude", BOOLEAN, &scale_by_weights,
       "display connection weights by changing line widths"},
    {"max_line_width", INT, &max_line_width,
       "width of line for maximum weight"},
    {"show_bias", BOOLEAN, &show_bias, "show bias nodes"},
    {"bw", BOOLEAN, &black_and_white,
       "plot negative weights as hollow. Default: plot as orange"},
};
#define EXTRA_PLOT_TABSIZE (sizeof(extra_plot_argtab)/sizeof(ARG))

#define BLOCK_RADIUS 20
#define BLOCK_SEPARATION (3*BLOCK_RADIUS)
#define LOWER_TEXT 10
#define CENTERED 2

#ifndef ABS
#define ABS(x) (((x) > 0)? (x) : -(x))
#endif

#define INPUT 0
#define OUTPUT (nlayers-1)
#define LARGE 0
#define SMALL 1

static int umiddle, voffset, nlayers, *layers;
static int block_radius, lower_text;
static float *layer_width;
static char **labels;
static char **input_labels;
static float struct_height = 1., struct_width = 1.;
/* new to keep track of connections */
static int *node_connections_flag;

static int node2u (int x, int layer);
static int level2v (int level);
static void draw_connection(int m, int level, int nodeu, int nodev,
		     float weight, int type);

int
main(int argc, char **argv)
{
    FILE *param_file;
    int model, normalization, layer, n, m;
    int ubot = 0, utop  = 1020, vbot = 0, vtop = 780;
    float scale, **weights;
    int max_nnodes;
    int nodeu, nodev;
    int *features, ninputs, noutputs;
    ARG *argtab;
    int tabsize;
    DEFAULTS_FILE_STRUCT df_args;
        
    init_nnsubs(argc, argv, rcsid);
    /********* read command line arguments ************/
    nnargs_combine_argtabs(struct_plot_argtab, STRUCT_PLOT_TABSIZE,
			   extra_plot_argtab, EXTRA_PLOT_TABSIZE,
			   &argtab, &tabsize);
    argc = nnargs(argc, argv, argtab, tabsize);
    nnargs_free_argtab(argtab, tabsize);

    /********** open and read parameter file ***********/
    param_file = nn_fopen(struct_param_fname, "r",
			  "for read of mlp parameters");
    model = norm_read_paramfile(param_file, &normalization);
    model = mlp_read_model(param_file);
    nn_fclose(param_file, struct_param_fname, "for read of mlp parameters");

    if(strcmp(struct_defaults_fname, "") != 0){
	read_defaults_file(struct_defaults_fname, &df_args);
	labels = df_args.output_labels;
	input_labels = df_args.input_labels;
    } else {
	input_labels = labels = NULL;
    }

    /* initialize the mlp model for testing */
    /* do not adjust priors, NULL priors list */
    mlp_test_init(model, 0, NULL);

    /************* initialize graphics *************/
    if(no_graphics){
	ginit("no graphics");
    } else {
	ginit(" -geometry 510x450 -title struct_plot");
    }
    ghardcopy(1);
    gwtitle("MLP Structure Plot");
    
    /* get structure from model file */
    layers = *(int **)mlp_param(model, "layers", 0);
    nlayers = *(int *)mlp_param(model, "nlayers", 0);
    features = *(int **) mlp_param(model, "features", 0);

    layer_width = allocate_vector(nlayers);
    ninputs = layers[INPUT];
    noutputs = layers[OUTPUT];
    node_connections_flag = allocate_ivector(ninputs);

    /* figure out u and v references */
    umiddle = (utop - ubot)/2;
    vtop -= 20;
    voffset = vtop-LOWER_TEXT-block_radius;

    /* write out the plot title */
    uvprintf(umiddle, vtop, CENTERED, "%s", struct_param_fname);

    /* figure out scaling */
    max_nnodes = layers[INPUT];
    for(layer = INPUT+1; layer <= OUTPUT;  layer++)
      if(max_nnodes < layers[layer]) max_nnodes = layers[layer];

    if(struct_autoscale){
	struct_height = (vtop-vbot)/(float)((nlayers+1)*BLOCK_SEPARATION);
	struct_width = (utop-ubot)/(float)((max_nnodes+2)*BLOCK_SEPARATION);
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] =
	      (utop-ubot)/(float)((layers[layer]+2)*BLOCK_SEPARATION);
	}
    } else {
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] = struct_width;
	}
    }
    scale = (struct_width < struct_height) ? struct_width : struct_height;
    block_radius = BLOCK_RADIUS*scale;
    lower_text = LOWER_TEXT+block_radius;

    /* figure out weight line width scaling */
    if(scale_by_weights){
	float temp, max_weight = -MAXFLOAT;
	/* find the max weight (ABS) value */
	for(layer = nlayers-1; layer > INPUT; layer--){
	    weights = *(float ***)mlp_param(model, "weights", layer);
	    for(n = 0; n < layers[layer]; n++){
		for(m = 0; m < layers[layer-1]; m++){
		    temp = ABS(weights[n][m]);
		    if(temp > max_weight) max_weight = temp;
		}
	    }
	}
	line_scale = (max_line_width-1)/max_weight;
    }

    /* write out the class labels */
    nodev = level2v(OUTPUT);
    for(n = 0; n < noutputs-1; n++){/* one for the bias node? */
	nodeu = node2u(n, OUTPUT);
	if(!struct_no_text){
	    if(labels != NULL)
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%s",
		       labels[n+1]);
	    else
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%d",
		       n);
	}
    }
    /* display the nodes for each layer */
    for (layer = OUTPUT; layer >= INPUT; layer--){/* foreach layer */
	if(layer != 0)
	  weights = *(float ***)mlp_param(model, "weights", layer);
	for(n = 0; n < layers[layer]; n++){/* foreach node in the layer */
	    nodeu = node2u(n, layer);
	    nodev = level2v(layer);
	    if(n == layers[layer]-1){
		if(show_bias && layer != nlayers-1)
		  uvblock("foreground",
			  nodeu-block_radius/2, nodev-block_radius/2,
			  nodeu+block_radius/2, nodev+block_radius/2);
	    } else {
		if(layer == INPUT)
			{
		    /* draw a filled circle */
		    uvdot("foreground", nodeu, nodev, block_radius/2);
			} 
		else {
		    /* draw a circle */
		    uvcircle(nodeu, nodev, block_radius);
		    /* draw connecting lines */
		    for(m = 0; m < layers[layer-1]; m++)
				{
				/* check that the weight is
				   greater than the threshold */
				if( ABS(weights[n][m]) < threshold) 
					continue;
				/* increment count for node connection flag */
				if ( layer == INPUT+1 )
					{
					node_connections_flag[m] = 1;
					}
				if(m == layers[layer-1]-1)
					{
				  if(show_bias)
					draw_connection(m, layer-1,
							nodeu, nodev, weights[n][m],SMALL);
					} 
				else
				  draw_connection(m, layer-1, nodeu, nodev, weights[n][m], LARGE);
				}
		}
	    }
	}
    }

    /* write out the input labels */
    nodev = level2v(INPUT);
    for(n = 0; n <  ninputs-1; n++){
	nodeu = node2u(n, INPUT);
	if(!struct_no_text){
		if ( node_connections_flag[n] > 0.1 )
			{
	    uvprintf(nodeu, nodev-lower_text, CENTERED,
		     "%s", get_input_label(n, input_labels,
					   features, normalization));
			}
	}
    }

    gflush();
    gsave(struct_plot_fname, 1);
    gexit(0);/* quit */
    exit_nnsubs();
    exit(0);
}
static int
node2u(int x, int layer)
{
    /* center the node */
    x -= layers[layer]/2;
    /* scale the position for u coordinates*/
    x *= BLOCK_SEPARATION*layer_width[layer];
    /* center the final position on the screen */
    return(x + umiddle);
}

static int
level2v(int level)
{
    return(voffset -(nlayers-level)*BLOCK_SEPARATION*struct_height);
}

static void
draw_connection(int m, int level, int nodeu, int nodev,
		float weight, int type)
{
    int v1, v2, u1, u2, ulength, vlength, udir, vdir, r;
    int line_width;
    int childu, childv;
    childu = node2u(m, level);
    childv = level2v(level);
    
    ulength = ABS(nodeu-childu);
    udir = (nodeu > childu) ? 1 : -1;
    vlength = ABS(nodev-childv);
    vdir = (nodev > childv) ? 1 : -1;
    r = (int)sqrt(((double)ulength*ulength)
		  +(double)vlength*vlength);
    u1 = nodeu-udir*(ulength*block_radius)/r;
    v1 = nodev-vdir*(vlength*block_radius)/r;
    
    if(level == INPUT || type == SMALL){
	u2 = childu;
	v2 = childv;
    } else {
	u2 = childu+udir*(ulength*block_radius)/r;
	v2 = childv+vdir*(vlength*block_radius)/r;
    }
    
    if(scale_by_weights){
	line_width = (int)(line_scale*ABS(weight))+1;
	if(weight >= 0) uvbeam(u1, v1, u2, v2, line_width);
	else{
	    if(black_and_white)
	      uvbeam_outline(u1, v1, u2, v2, line_width);
	    else {
		gcolor("salmon");
		uvbeam(u1, v1, u2, v2, line_width);
		gcolor("foreground");
	    }
	}
    } else {
	uvline(u1, v1, u2, v2);
    }
}
