/***--------------------------------------------------------------------------
* NAME:         plot_prob.c
* PURPOSE:      to prepare a table and a plot of actual vs. predicted
*               classification probability for a LNKnet classifier
*
* USAGE: 
*	plot_prob -<flags>
* 
* RETURN VALUE:  NONE
*
* EXAMPLE: plot_prob --		 * shows all available flags
*
* DESCRIPTION:
*    Functions in this file
*       read_prob_error_file  Loads the contents of the error file to be
*                               processed into a matrix: data_matrix.
*       sort_prob_data_matrix Sorts the rows of the data_matrix according
*                               to the output variable in descending order
*       make_prob_table       Counts the elements of the prob_table
*                               for each case of a different threshold.
*       display_prob_table    Displays the contents of the table.
*       load_prob_plot_points Loads the data (actual prob vs. classifier prob) into two
*                               vectors x and y.
*       make_macfile            Prepares a ASCII file of data points to enable
*                               the use of a Mac to do the plotting.
*       plot_prob_curve       Displays the plot 
*
*
* SEE ALSO:	plot_detect(1), classifier(1), gauss(1), gmix(1), 
*               irbf(1), mlp(1), plot_perr(1)
*
* spring 1994 yuchun
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_prob.c,v 11.4 2001/08/23 20:18:09 craig Exp $";

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <math.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <nnlib.h>
#include <get_subs.h>
#include "plot_prob_args.h"
#include "xlippic.h"

#define PLOT_PROB
#define PROB_TINY 1.0e-30

/*  Argument Variables for plot_prob */

/* 
    mean is the average output of all points in the bin
    p = nactual_detect / npat = actual probability of detection
    upper_stdv and lower_stdv = 2 binomial stdv from p
*/
typedef struct{
    int bin_num;
    float lower_stdv;
    float upper_stdv;
    int npat;
    float lower_limit;
    float upper_limit;
    int nactual_detect;
    int ndetect;
    int nfalse;
    int ncorrect;
    float mean;
} prob_table_type;

/* I hate globals but is the */
/* simplest change to the original code for working */
static float chi_sq=0.0;
static int dof=0;
static float significance=0;

#define TARGET_OUTPUT 0
#define CLASS_TARGET 1
#define OTHER_OUTPUT 2

/* decleration of (some) functions: */
int read_prob_error_file(float ***data_matrix);
int make_prob_table(int npoints, float **data_matrix,
		    prob_table_type *prob_table);
void display_prob_table(int npoints, prob_table_type *prob_table);
void load_prob_plot_points(int npoints, prob_table_type *prob_table,
			   float *x, float *y, float *uplimit,
			   float *lowlimit, float *npat);
void compute_chi_square(float *x, float *y, prob_table_type *prob_table);
void make_macfile(int npoints, prob_table_type *prob_table);
void plot_prob_curve(int npoints, float *x, float *y, float *uplimit,
		     float *lowlimit, float *npat,
		     prob_table_type *prob_table);
void print_table_row (prob_table_type *dt);

/* function macros */
#define sort_prob_data_matrix(n, matrix)  sort_matrix_rows(n, 2, 0, matrix)

int
main(int argc, char **argv)
{
    int npoints;
    float *x, *y, *uplimit, *lowlimit, *npat;
    float **data_matrix;
    prob_table_type *prob_table;
    int i;

    init_nnsubs(argc, argv, rcsid);

    /* read args */
    argc = nnargs(argc, argv, prob_argtab, PROB_TABSIZE);

    /* validate arguments */
    if (nbins<1) nbins=1;
    if(target < 0) target = 0;
    if(target > noutputs-1) target = noutputs - 1;

#ifdef notdef
/* should be taken care of before running classifier ??? */
    /* determine training set prior from args */
    if (diff_prior) {
	float f;
	if(train_prior == NULL || (int)train_prior[0]!=noutputs)
	  fatal_error("Need to Specify Training Prior for All Classes!");

	for (i=1,f=0.0;i<=noutputs;i++)
	  total += train_prior[i];

	if (total < 1.0-PROB_TINY || total >1.0+PROB_TINY) {
	    warning("Sum of Priors Not Equal to 1. %f Set to 1", total);
	    for (i=1;i<=noutputs;i++) train_prior[i] /= total;
	}
	target_train_prior= train_prior[target+1];
	other_train_prior= 1-target_train_prior;
    }
#endif

    if (var_bin_flag){
	if(var_bin_ends == NULL || var_bin_ends[0] <= 1) {
	    warning("bin_ends must be specified, using constant width bins");
	    var_bin_flag=0;
	} else {
	    nbins = var_bin_ends[0]-1;
	    /* check that the numbers are sorted and go between 0 and 100? */
	    if(var_bin_ends[1] < 0. || var_bin_ends[nbins+1] > 100.){
		warning("var_bin_ends entries must be between 0 and 100");
		var_bin_flag = 0;
	    } else {
		for(i = 1; i <= nbins; i++)
		  if(var_bin_ends[i] > var_bin_ends[i+1]){
		      warning("var_bin_ends entries must be in order");
		      var_bin_flag = 0;
		  }
	    }
	}
    }

    npoints = read_prob_error_file(&data_matrix);

#ifdef notdef
    /* normalize by different prior */
    if(diff_prior) prior_prob_norm(data_matrix, npoints);
#endif

    /* allocate space */
    x = (float *) allocate_vector(nbins);
    y = (float *) allocate_vector(nbins);
    uplimit = (float *) allocate_vector(nbins);
    lowlimit = (float *) allocate_vector(nbins);
    npat = (float *) allocate_vector(nbins);
    prob_table = (prob_table_type *)
      allocate_structvector(nbins, sizeof(prob_table_type));

    sort_prob_data_matrix(npoints, data_matrix);
    make_prob_table(npoints, data_matrix, prob_table);
    compute_chi_square(x,y, prob_table);
    load_prob_plot_points(nbins,prob_table,x,y,uplimit,lowlimit,npat);
    display_prob_table(nbins, prob_table);
    if(mac_prob_flag) make_macfile(nbins, prob_table);
    plot_prob_curve(nbins, x, y, uplimit, lowlimit, npat, prob_table);

    /* deallocate memory */
    deallocate_matrix(data_matrix);
    deallocate_vector(x);
    deallocate_vector(y);
    deallocate_vector(uplimit);
    deallocate_vector(lowlimit);
    deallocate_vector(npat);
    deallocate_structvector(prob_table);

    exit_nnsubs();
    return 0;
}

int
read_prob_error_file(float ***data_matrix)
{
    int npoints, i;
    int error_file_verbosity=2; /* long Error File */
    FILE *err_file;
    float output,other_output;
    float sum;
    int class;
    ERROR_LINE error_line;

    /* open data file */
    err_file = nn_fopen(path_build(plot_exp_path, error_fname),
			"r","error file");

    error_line.outputs = allocate_vector(noutputs);
    error_line.noutputs = noutputs;
    error_line.inputs = NULL;
    error_line.desired_outputs = NULL;

    /* use only the first npatterns points */
    npoints = get_n_lines(err_file);
    if(npatterns != 0 && npatterns < npoints) npoints = npatterns;
    if(npoints <= 0){
	fatal_error("Empty data file: %s", error_fname);
    }

    /* allocating memory and loading the data matrix */
    *data_matrix = (float **) allocate_matrix(npoints, 3);
    i = 0;
    while(get_err_line(err_file, error_fname, &error_line,
		       error_file_verbosity) != EOF){
	output = error_line.outputs[target];
	if(sum_to_one){
	    sum = 0.;
	    for(class = 0; class < noutputs; class++)
	      sum += error_line.outputs[class];
	    output /= sum;
	}
	/* find maximum output of NOT target class */
	other_output= -MAXFLOAT;
	for (class=0; class<noutputs; class++)
	  if(class != target)
	    other_output = MAX(other_output, error_line.outputs[class]);
	if(sum_to_one) other_output /= sum;
	      
	(*data_matrix)[i][TARGET_OUTPUT] = output;
	(*data_matrix)[i][CLASS_TARGET] =
	  (float) (target == error_line.desired_class);
	(*data_matrix)[i][OTHER_OUTPUT] = other_output;

	if(++i >= npoints) break;
    }
    nn_fclose(err_file, path_build(plot_exp_path, error_fname), "error file");
    deallocate_vector(error_line.outputs);
    return(npoints);
}

#ifdef notdef
/* should only be called if diff_prior is set! */
void
prior_prob_norm(data_matrix,npoints)
     float **data_matrix;
     int npoints;
{
    int i;
    float p_target_class;
    float p_other_class;
    float output_val, other_val;
    float f;
    
    /* find prior */
    p_target_class=0.0;
    p_other_class=0.0;
    for (i=0;i<npoints;i++) {
	if (data_matrix[i][CLASS_TARGET]==1)
	  p_target_class+= 1.0;
	else
	  p_other_class+= 1.0;
    }
    p_target_class/= npoints;
    p_other_class/= npoints;

    /* should be taken care of before classification run??? */
    /* normalize by different prior */
    for (i=0;i<npoints;i++) {
	output_val= data_matrix[i][TARGET_OUTPUT];
	other_val= 1.0-output_val;

	output_val= output_val*p_target_class/
	  (target_train_prior+PROB_TINY);
	other_val= other_val*p_other_class/(other_train_prior+PROB_TINY);

	output_val= output_val/(output_val+other_val);
	
	data_matrix[i][TARGET_OUTPUT]=output_val;
    }
}
#endif

int 
make_prob_table(int npoints, float **data_matrix, prob_table_type *prob_table)
{
    int i, j, k;
    float f, p, inc;

    /* initialize probf_table */
    for (i=0;i<nbins;i++) {
	prob_table[i].bin_num= i;
	prob_table[i].nactual_detect= 0;
	prob_table[i].ndetect= 0;
	prob_table[i].nfalse= 0;
	prob_table[i].ncorrect= 0;
	prob_table[i].mean=0.0;
    }

    /* set the upper and lower bin limits */
    if (var_bin_flag) {
	for (i=0,j=1;i<nbins;i++) {
	    prob_table[i].lower_limit= var_bin_ends[j]/100.;
	    j++;
	    prob_table[i].upper_limit= var_bin_ends[j]/100.;
	}
    } else {
	inc= 1.0/(float)nbins;
	for (i=0,f=0.0;i<nbins;i++,f+=inc) {
	    prob_table[i].lower_limit= f;
	    prob_table[i].upper_limit= f+inc;
	}
    }

    /* find ndetect, nfalse, nactual_detect, ncorrect and mean for 
       patterns in each bin */
    i=0;
    for (k=nbins-1;k>=0;k--){
	for (; i<npoints &&
	     data_matrix[i][TARGET_OUTPUT]>=prob_table[k].lower_limit; i++) {
	    /* accumulate bin amounts */
	    if (data_matrix[i][TARGET_OUTPUT] > data_matrix[i][OTHER_OUTPUT])
	      prob_table[k].ndetect++;
	    else
	      prob_table[k].nfalse++;
	    
	    if (data_matrix[i][CLASS_TARGET])
	      prob_table[k].nactual_detect++;
	    
	    if (data_matrix[i][CLASS_TARGET] &&
		(data_matrix[i][TARGET_OUTPUT]>=data_matrix[i][OTHER_OUTPUT]))
	      prob_table[k].ncorrect++;
	    
	    prob_table[k].mean+= data_matrix[i][TARGET_OUTPUT];
	}
    }

    /* combine bins that have too few samples */
    /* combine a bin to a higher bin if insufficient points */
    for (i=0;i<nbins-1;i++) {
	if (prob_table[i].ndetect+prob_table[i].nfalse < min_bin_count){
	    k=i+1;
	    prob_table[k].ndetect+= prob_table[i].ndetect;
	    prob_table[i].ndetect= -1; /* mark as gone */
	    prob_table[k].nfalse+= prob_table[i].nfalse;
	    prob_table[i].nfalse=0;
	    prob_table[k].nactual_detect+= prob_table[i].nactual_detect;
	    prob_table[i].nactual_detect=0;
	    prob_table[k].ncorrect+= prob_table[i].ncorrect;
	    prob_table[i].ncorrect=0;
	    prob_table[k].mean+= prob_table[i].mean;
	    prob_table[i].mean=0;
	    prob_table[k].lower_limit= prob_table[i].lower_limit;
	}
    }
    /* combine a bin to a lower bin if insufficient points */
    /* ??? this should only affect the topmost remaining bin since all the
       remaining lower bins should have the right amount of data in them */
    for (i=nbins-1;i>0;i--) {
	if (prob_table[i].ndetect+prob_table[i].nfalse < min_bin_count &&
	    prob_table[i].ndetect!= -1) {
	    k= i-1;
	    prob_table[k].ndetect+= prob_table[i].ndetect;
	    prob_table[i].ndetect= -1;
	    prob_table[k].nfalse+= prob_table[i].nfalse;
	    prob_table[i].nfalse=0;
	    prob_table[k].nactual_detect+= prob_table[i].nactual_detect;
	    prob_table[i].nactual_detect=0;
	    prob_table[k].ncorrect+= prob_table[i].ncorrect;
	    prob_table[i].ncorrect=0;
	    prob_table[k].mean+= prob_table[i].mean;
	    prob_table[i].mean=0;
	    prob_table[k].upper_limit= prob_table[i].upper_limit;
	  }
      }

    /* compact and remove all the empty bins */
    for (j=k=0; k<nbins && j<nbins; j++) {
	if (prob_table[j].ndetect != -1){
	    if( j != k ){
		/* copy the non-empty bin into this empty one */
		prob_table[k].ndetect= prob_table[j].ndetect;
		prob_table[j].ndetect = -1;
		prob_table[k].nfalse= prob_table[j].nfalse;
		prob_table[k].nactual_detect= prob_table[j].nactual_detect;
		prob_table[k].ncorrect= prob_table[j].ncorrect;
		prob_table[k].mean= prob_table[j].mean;
		prob_table[k].lower_limit= prob_table[j].lower_limit;
		prob_table[k].upper_limit= prob_table[j].upper_limit;
	    }
	    k++;
	}
    }
    nbins= k;
    

    /* for (i=0;i<nbins;i++)
       printf("number of pat: %d   %d\n",i,prob_table[i].npat);
       printf("\n"); */

    for (i=0;i<nbins;i++) {
	/* finish the rest */
	prob_table[i].npat= prob_table[i].ndetect + prob_table[i].nfalse;
	prob_table[i].mean/= (float)prob_table[i].npat+PROB_TINY;

	/* p= prob_table[i].mean; */
	p= prob_table[i].nactual_detect/((float)prob_table[i].npat+PROB_TINY);
	f= (float)sqrt((double)(p*(1-p)/((float)prob_table[i].npat+PROB_TINY)));
	if (prob_table[i].npat==0) f=0.0;

	prob_table[i].lower_stdv= p-2*f;
	prob_table[i].upper_stdv= p+2*f;
    }

    return(nbins);
}

void
compute_chi_square(float *x, float *y, prob_table_type *prob_table)
{
    int i;
#ifdef notdef
    if (chi_square_flag) {
#endif
	if (nbins>1) {
	    dof= nbins;
	    for(i = 0; i < nbins; i++){
		x[i] = (float)prob_table[i].mean * prob_table[i].npat;
		y[i]= prob_table[i].nactual_detect;
	    }
	    chi_sq= chi_square(nbins,&dof, x, y,&significance);
	} else {
	    dof=0; chi_sq=0; significance=0;
	}
#ifdef notdef
    }
#endif
}

void
load_prob_plot_points(int npoints, prob_table_type *prob_table,
		      float *x, float *y, float *uplimit, float *lowlimit,
		      float *npat)
{
  /* x is the mean of n.n. prob estimates
     y is the actual prob. of detection
     uplimit and lowlimit centers on y */
    int i;

    for(i = 0; i < npoints; i++) {
	x[i] = prob_table[i].mean * 100.;
	y[i]= (prob_table[i].nactual_detect/
	       (float)prob_table[i].npat+PROB_TINY)*100.;
	uplimit[i]= (float)(prob_table[i].upper_stdv*100.);
	lowlimit[i]= (float)(prob_table[i].lower_stdv*100.);
	npat[i]= (float)prob_table[i].npat;
    }
}

void
display_prob_table(int npoints, prob_table_type *prob_table)
{
    int i;
    int ntotal;
    static char *table_head[] = {
	"   BIN #   ", " # PATTERNS ", " PREDICTED % ",
	" ACTUAL % ", "   RANGE   ",
    };
#define TABLE_LENGTH (sizeof(table_head)/sizeof(char *))

    if(verbose < 1) return;

    /* find total number of points */
    for (i=0,ntotal=0;i<npoints;i++)
      ntotal+= prob_table[i].npat;
      
    printf("\n ----------------------------------------");
    printf("--------------------------------------\n");
    printf("\n   target class = %2d", target);
    printf("\n   total number of patterns = %4d", ntotal);
    printf("\n ----------------------------------------");
    printf("--------------------------------------\n");
    for(i = 0; i < TABLE_LENGTH; i++) printf("%13s", table_head[i]);
    printf("\n");
    for(i = 0; i < TABLE_LENGTH; i++) printf(" ----------- ");
    printf("\n"); fflush(stdout);

    for (i=0;i<npoints;i++) {
	/* skip to next row */
	print_table_row(&prob_table[i]);
	fflush(stdout);
    }

    if (chi_square_flag) {
	printf("\nChi Square Fit:\n");
	printf("Chi                =   %f\n",chi_sq);
	printf("Degrees of Freedom =   %d\n",dof);
	printf("Significance       =   %f",significance);
	if(significance < 0.05)
	  printf(" (*Poor Posterior Probability Estimates)\n");
	else printf("\n");
    }
    /* print out Chi square values for notebook file */
    printf(" TARGET %d CHI %f DOF %d SIGNIFICANCE %f%s \n",
	   target, chi_sq, dof, significance,
	   (significance < 0.05) ? " (POOR FIT)" : "");

}

void
print_table_row(prob_table_type *dt)
{
    int dof;
    float npred, nactual, chi;

    npred = dt->mean * dt->npat;
    nactual = dt->nactual_detect;
    chi = npred - nactual;
    chi *= chi;
    chi /= npred;
    printf("%10d  ",dt->bin_num);
    printf("%10d  ",dt->npat);
    printf("%10.2f  ",(float)dt->mean*100);
    printf("%10.2f  ",(float)dt->nactual_detect/(float)(dt->npat+PROB_TINY)*100);
    printf("%7.2f - %6.2f",(float)dt->lower_stdv*100,
	   (float)dt->upper_stdv*100);
    dof = 10;
    printf("\n");
}

void
make_macfile(int npoints, prob_table_type *prob_table)
{
    int i;
    FILE *fptr;

    fptr = nn_fopen(path_build(plot_exp_path, mac_problines_fname),
		    "w", "to write a Mac file");

    for(i = 0; i < npoints; i++) {
	fprintf(fptr,"%10f  ",prob_table[i].mean*100.);
	fprintf(fptr,"%10f  ",prob_table[i].nactual_detect/
		(float)(prob_table[i].npat+PROB_TINY)*100.);
	fprintf(fptr,"%10f  ",prob_table[i].upper_stdv*100.);
	fprintf(fptr,"%10f  ",prob_table[i].lower_stdv*100.);
	fprintf(fptr,"%d   ",(int)prob_table[i].npat);
	fprintf(fptr,"%d   ", (int)prob_table[i].nactual_detect);
	fprintf(fptr, "\n");
    }
    /* print out chi sq stuff */
    if(chi_square_flag){
	fprintf(fptr, "Chi %f DOF %d Significance %f%s \n",
		chi_sq, dof, significance,
		(significance < 0.05) ? " (Poor Fit)" : "");
    }
    fclose(fptr);
}

void
plot_prob_curve(int npoints, float *x, float *y,
		float *uplimit, float *lowlimit, float *npat,
		prob_table_type *prob_table)
{
    int i;

    int prob_auto_scale_flag = 0;
    char title[64];
    float linex[2], liney[2];
    float inc;

    int symbol_mode = 2, symbol_type = 8; /* circle */
    char *symbol_string = NULL;
    int symbol_radius = 7;
    int prob_line_type=0;
    int grid_type = 1; /* full grid */
    int  auto_abscissas = 0; /* x provided by x array */
    float auto_xmin = 0.0, auto_xstep = 1.0; /* ignored */
    float height_fract = 0.9, width_fract = 1.;
    float xoffset_fract = 0.1, yoffset_fract = 0.1;
    int xlog = 0, ylog = 0; /* linear x and y axes */
    /* char *xlabel = "AVG % BIN VALUE  ", *ylabel = "PROB";*/
    char *xlabel = "  ", *ylabel = "PROB";
    char *color = "black";
    char *command =
      "-geometry 500x520 -title prob_plot ";
    int quit_flag = 0;
    float axis_xmin, axis_xmax, axis_xstep;
    float axis_ymin, axis_ymax, axis_ystep;
    int ubot, vbot, utop, vtop, u, v;

    int legend_x, legend_y;

    /* diagonal line plot info */
    int diag_line_type = 2;

    /* stdv line plot info */
    int stdv_symbol_mode0= 0, stdv_symbol_mode1= 2;
    int stdv_symbol_type0= 0, stdv_symbol_type1= 13;
    int stdv_symbol_type2= 14, stdv_symbol_radius= 8;
    int stdv_line_type0= 1, stdv_line_type1= 0;
    char *stdv_color = "blue";

    /* bin content plot info */
    int npat_symbol_mode= 1;
    char num_string[32];
    int tot_npat;

    /* plot the curve: y(i) vs x(i) */
    if(no_prob_graphics)
      ginit("no graphics");
    else 
      ginit(command);

    ghardcopy(1);

    tot_npat=0;
    for (i=0;i<nbins;i++) tot_npat+= npat[i];

    sprintf(title, "Target=%d", target);
    if (chi_square_flag)
	sprintf(title, "%s Chi=%.3f DOF=%d Signif=%.3f", title,
		(float)chi_sq, (int)dof, (float)significance);
    sprintf(title, "%s #Patterns=%d", title, (int)tot_npat);

    gwtitle(prob_plot_label);

    if(bin_plot){
	prob_xmin=0.0; prob_xmax=100.0;
	prob_xstep= prob_xmax;
    }

    /* get size of plot */
    xygraph_scale(x, y, npoints,
		  2, symbol_string, symbol_type, symbol_radius,
		  0, grid_type, prob_plot_label,
		  auto_abscissas, auto_xmin, auto_xstep,
		  height_fract, width_fract, xoffset_fract, yoffset_fract,
		  xlog, prob_auto_scale_flag,
		  prob_xmin, prob_xmax, prob_xstep, "",
		  ylog, prob_auto_scale_flag,
		  prob_ymin, prob_ymax, prob_ystep, "","foreground",
		  &ubot, &vbot, &utop, &vtop,
		  &axis_xmin, &axis_xmax, &axis_xstep,
		  &axis_ymin, &axis_ymax, &axis_ystep);

    /* display label */
    /* uvprintf((utop+ubot)/2, vbot-75, 2, prob_plot_label);*/
    uvprintf((utop+ubot)/2, vbot-75, 2,  "BIN VALUE (NETWORK OUTPUT)");
    uvprintf((utop+ubot)/2, vbot-110, 2, title);
    /*uvprintf(ubot, vbot-110, 0, title);*/

    if(!bin_plot){
	/* draw diagonal line */
	linex[0]=0; linex[1]=100;
	liney[0]=0; liney[1]=100;
	if (linex[0]<prob_xmin) linex[0] = liney[0] = prob_xmin;
	if (liney[0]<prob_ymin) liney[0] = linex[0] = prob_ymin;
	if (linex[1]>prob_xmax) linex[1] = liney[1] = prob_xmax;
	if (liney[1]>prob_ymax) liney[1] = linex[1] = prob_ymax;
	xygraph(linex,liney,2,
		stdv_symbol_mode0, symbol_string, stdv_symbol_type0, 
		stdv_symbol_radius,
		diag_line_type, grid_type,
		path_build(plot_exp_path, plot_prob_fname),
		auto_abscissas, auto_xmin, auto_xstep,
		height_fract, width_fract, xoffset_fract, yoffset_fract,
		xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		color);
    }
    
    /* draw stdv lines */
    if(bin_plot) inc= 100.0/((float)nbins+1);
    for (i=0;i<nbins;i++) {
	if(bin_plot){
	    /* number of patterns in each bin */
	    sprintf(num_string,"%d-%d",(int)(prob_table[i].lower_limit*100),
		    (int)(prob_table[i].upper_limit*100));
	    
	    u= ((utop-ubot)/(nbins+1))*(i+1)+ubot-8;
	    v= vbot-35;
	    uvprintf(u,v,1,"%s",num_string);

	    /* draw classifier estimate */
	    symbol_type = 3;
	    linex[0]= (i+1)*inc;
	    liney[0]= x[i];
	    xygraph(linex, liney, 1, 
		    symbol_mode, symbol_string, symbol_type, 
		    symbol_radius,
		    prob_line_type, grid_type,
		    path_build(plot_exp_path, plot_prob_fname),
		    auto_abscissas, auto_xmin, auto_xstep,
		    height_fract, width_fract, xoffset_fract, yoffset_fract,
		    xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		    ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		    color);

	}
	/* draw line */
	if(bin_plot){
	    linex[0]= (i+1)*inc; linex[1]= linex[0];
	} else {
	    linex[0]= x[i]; linex[1]= x[i];
	}

	liney[0]= uplimit[i]; liney[1]= lowlimit[i];
	if (liney[0]>prob_ymax) liney[0]= prob_ymax;
	if (liney[1]>prob_ymax) liney[1]= prob_ymax;
	if (liney[0]<prob_ymin) liney[0]= prob_ymin;
	if (liney[1]<prob_ymin) liney[1]= prob_ymin;

	xygraph(linex,liney,2,
		stdv_symbol_mode0, symbol_string, stdv_symbol_type0, 
		stdv_symbol_radius,
		stdv_line_type0, grid_type,
		path_build(plot_exp_path, plot_prob_fname),
		auto_abscissas, auto_xmin, auto_xstep,
		height_fract, width_fract, xoffset_fract, yoffset_fract,
		xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		stdv_color);
	
	/* draw upper and lower symbols */
	xygraph(linex,liney,1,
		stdv_symbol_mode1, symbol_string, stdv_symbol_type1, 
		stdv_symbol_radius,
		stdv_line_type1, grid_type,
		path_build(plot_exp_path, plot_prob_fname),
		auto_abscissas, auto_xmin, auto_xstep,
		height_fract, width_fract, xoffset_fract, yoffset_fract,
		xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		stdv_color);
	xygraph(linex,&liney[1],1,
		stdv_symbol_mode1, symbol_string, stdv_symbol_type2, 
		stdv_symbol_radius,
		stdv_line_type1, grid_type,
		path_build(plot_exp_path, plot_prob_fname),
		auto_abscissas, auto_xmin, auto_xstep,
		height_fract, width_fract, xoffset_fract, yoffset_fract,
		xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		stdv_color);

	/* number of patterns in each bin */
	if(no_text)
	  strcpy(num_string, "");
	else
	  sprintf(num_string,"%d/%d",
		  (int)prob_table[i].nactual_detect, (int)npat[i]);
	if(bin_plot){
	    linex[0]= (i+1)*inc; liney[0]=uplimit[i]+5;
	} else {
	    linex[0]= x[i]+3; liney[0]=y[i];
	}

	xygraph(linex, liney, 1, 
		npat_symbol_mode, num_string, symbol_type, symbol_radius,
		prob_line_type, grid_type,
		path_build(plot_exp_path, plot_prob_fname),
		auto_abscissas, auto_xmin, auto_xstep,
		height_fract, width_fract, xoffset_fract, yoffset_fract,
		xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		color);

	if(bin_plot){
	    /* draw actual estimate */
	    symbol_type = 8;
	    linex[0]= (i+1)*inc;
	    liney[0]= y[i];
	    xygraph(linex, liney, 1, 
		    symbol_mode, symbol_string, symbol_type, 
		    symbol_radius,
		    prob_line_type, grid_type,
		    path_build(plot_exp_path, plot_prob_fname),
		    auto_abscissas, auto_xmin, auto_xstep,
		    height_fract, width_fract, xoffset_fract, yoffset_fract,
		    xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		    ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		    stdv_color);
	}
    }

    if(!bin_plot){
	/* do real plots using scale gotten from xygraph_scale() */
	xygraph(x, y, npoints, 
		symbol_mode, symbol_string, symbol_type, symbol_radius,
		prob_line_type, grid_type,
		path_build(plot_exp_path, plot_prob_fname),
		auto_abscissas, auto_xmin, auto_xstep,
		height_fract, width_fract, xoffset_fract, yoffset_fract,
		xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
		ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
		stdv_color);
	
    }
    legend_x = ubot+40;
    legend_y = vtop-40;
    if(bin_plot){
	uvmove(legend_x, legend_y);
	gcolor(color);
	gsymbol(3, symbol_radius);
	uvprintf(legend_x+40, legend_y, 1,
		 "Avg Net Output");
	legend_y -= 40;
    }

    uvmove(legend_x, legend_y);
    gcolor(stdv_color);
    gsymbol(8, symbol_radius);
    gcolor(color);
    uvprintf(legend_x+40, legend_y, 1, "%% Target in Bin");
    gflush();

    gsave(path_build(plot_exp_path, plot_prob_fname), 1);
    gexit(quit_flag);
}
