/***--------------------------------------------------------------------------
* NAME:         plot_reject.c
* PURPOSE:      to prepare a table and a plot of error vs rejection rate curve
*             for a LNKnet classifier
*
* USAGE: 
*	plot_reject -<flags>
* 
* RETURN VALUE:  NONE
*
* EXAMPLE: plot_reject --		 * shows all available flags
*
* GLOBALS: <flags> defined in plot_perr_args.h
*
* DESCRIPTION:
*    Functions in this file
*       print_reject_table	Displays the contents of the table.
*       make_macfile		Prepares a ASCII file of data points to enable
*				the use of a Mac to do the plotting.
*       plot_reject		Displays the plot of %error vs %rejection.
*
*
* SEE ALSO:	plot_reject(1), classifier(1), gauss(1), gmix(1), 
*               irbf(1), mlp(1), plot_perr(1)
*
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_reject.c,v 11.4 2001/08/23 20:18:09 craig Exp $";

/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <nnlib.h>
#include <get_subs.h>

#define PLOT_REJECT
#include "plot_reject_args.h"
#include "xlippic.h"

#define TABLE_WIDTH 6

typedef struct{
    int   npatterns;
    int   nrejections;
    float per_rejection;
    int   nerrors;
    float per_error;
    float threshold;
} reject_table_type;

#define CLASS_OUTPUT 0
#define CLASS_ERROR 1

/* decleration of (some) functions: */
int print_reject_table(FILE *fp, int verbose, int npoints,
		       float **data_matrix, reject_table_type *reject_table);
void make_macfile(int table_length, float *x, float *y);
void plot_reject(int table_length, reject_table_type *reject_table);
int reject_read_data(float ***data_matrix);
int reject_compute_table (float **data_matrix, int npoints,
			  reject_table_type *reject_table);

/* function macros */

/* static functions */
static void reject_print_table_head(FILE *fp);
static void reject_print_table_row(FILE *fp, reject_table_type *reject_table,
				  int i);
static void reject_print_interpolated_row(FILE *fp, reject_table_type *rt,
					 int table_length, int i, float g);

int
main(int argc, char **argv)
{
    int npoints, table_length;
    float **data_matrix;
    reject_table_type *reject_table;

    init_nnsubs(argc, argv, rcsid);

    /* read args */
    argc = nnargs(argc, argv, reject_argtab, REJECT_TABSIZE);

    if(table_step == 0)
      fatal_error("table_step must be greater than 0");

    npoints = reject_read_data(&data_matrix);

    reject_table = (reject_table_type *)
      allocate_structvector(npoints, sizeof(reject_table_type));

    /* function calls: make table and plot */
    table_length = print_reject_table(stdout, verbose, npoints,
				      data_matrix, reject_table);
    plot_reject(table_length, reject_table);

    /* free space */
    deallocate_structvector(reject_table);
    deallocate_matrix(data_matrix);
    return 0;
}

int
reject_read_data(float ***data_matrix)
{
    FILE *err_file;
    ERROR_LINE error_line;
    int error_file_verbosity = 2;/* OUTPUTS */
    int i, npoints, class;
    float output, sum;

    /* open data file */
    err_file = nn_fopen(path_build(plot_exp_path, error_fname),
			"r","error file");
    
    npoints = get_n_lines(err_file);
    if(npatterns != 0 && npatterns < npoints) npoints = npatterns;
    if(npoints <= 0){
	fatal_error("Empty data file: %s", error_fname);
    }

    /* allocate space for various arrays we need later */
    *data_matrix = (float **) allocate_matrix(npoints, 2);

    error_line.outputs = allocate_vector(noutputs);
    error_line.noutputs = noutputs;
    error_line.inputs = NULL;
    error_line.desired_outputs = NULL;

    /* load data */
    i = 0;
    while(get_err_line(err_file, error_fname, &error_line,
		       error_file_verbosity) != EOF){
	output = error_line.outputs[error_line.class];
	if(sum_to_one){
	    sum = 0.;
	    for(class = 0; class < noutputs; class++)
	      sum += error_line.outputs[class];
	    output /= sum;
	}
	/* output for selected class */
	(*data_matrix)[i][CLASS_OUTPUT] = output;
	/* whether this pattern was classified wrong */
	(*data_matrix)[i][CLASS_ERROR] = (float) error_line.err_f;
	i++;
	if(i >= npoints) break;
    }

    deallocate_vector(error_line.outputs);
    nn_fclose(err_file, path_build(plot_exp_path, error_fname),
	      "for error file");
    return npoints;
}

int
print_reject_table(FILE *fp, int verbose, int npoints, float **data_matrix, reject_table_type *reject_table)
{
    int i, table_length;
    int lasti;
    float g;

    sort_matrix_rows(npoints, 2, 0, data_matrix);

    /* build the table */
    table_length = reject_compute_table(data_matrix, npoints, reject_table);
    if(verbose >= 1){
	reject_print_table_head(fp);
	if(verbose >= 2){
	    for(i = 0; i < table_length; i++){
		reject_print_table_row(fp, reject_table, i);
	    }
	} else {
	    /* I need to interpolate when I am between rows... */
	    for(g = table_begin, i = 0, lasti = -1;
		g <= table_end && i < table_length;
		g += table_step){
		/* skip to next row */
		for(; i < table_length; i++)
		  if(reject_table[i].per_rejection >= g) break;
		if(i == table_length) break;
		if(lasti != i){
		    if(reject_table[i].per_rejection == g)
		      reject_print_table_row(fp, reject_table, i);
		    else
		      reject_print_interpolated_row(fp, reject_table,
						    table_length, i, g);
		    lasti = i;
		}
	    }
	}
    }
    return table_length;
}

static void
reject_print_table_head(FILE *fp)
{
    static char *table_head[] = {
	" # PATTERNS ", "# REJECTIONS ", "% REJECTED ",
	" # ERRORS   ", "% ERROR   ", " THRESHOLD  ",
    };
    int i;

    fprintf(fp,"\n ---------------------------------------");
    fprintf(fp,"---------------------------------------\n");
    for(i = 0; i < TABLE_WIDTH; i++) fprintf(fp,"%12s", table_head[i]);
    fprintf(fp,"\n");
    for(i = 0; i < TABLE_WIDTH; i++) fprintf(fp," ---------- ");
    fprintf(fp,"\n");
}

static void
reject_print_table_row(FILE *fp, reject_table_type *reject_table, int i)
{
    fprintf(fp,"%10d  ", reject_table[i].npatterns);
    fprintf(fp,"%10d  ", reject_table[i].nrejections);
    fprintf(fp,"%10f  ", reject_table[i].per_rejection);
    fprintf(fp,"%10d  ", reject_table[i].nerrors);
    fprintf(fp,"%10f  ", reject_table[i].per_error);
    fprintf(fp,"%10.3f  ", reject_table[i].threshold);
    fprintf(fp,"\n");
}

#define INTERPOLATE(x1, x2, r)  ((x1) + ((x2) - (x1))*(r))

static void
reject_print_interpolated_row(FILE *fp, reject_table_type *rt, int table_length, int i, float g)
{
    float ratio, temp;
    
    if(i > 0 && i < table_length){
	if(rt[i].per_rejection != rt[i-1].per_rejection){
	    ratio = g - rt[i-1].per_rejection;
	    ratio /= rt[i].per_rejection- rt[i-1].per_rejection;
	} else ratio = .5;
    }
    temp = (i <= 0) ? rt[0].npatterns :
      INTERPOLATE(rt[i-1].npatterns, rt[i].npatterns, ratio);
    fprintf(fp,"%10d  ", (int)temp);
    temp = (i <= 0)? rt[0].nrejections :
      INTERPOLATE(rt[i-1].nrejections, rt[i].nrejections, ratio);
    fprintf(fp,"%10d  ", (int)temp);
    fprintf(fp,"%10f  ", g);
    temp = (i <= 0)?rt[0].nerrors :
      INTERPOLATE(rt[i-1].nerrors, rt[i].nerrors, ratio);
    fprintf(fp,"%10d  ", (int)temp);
    temp = (i <= 0)?rt[0].per_error :
      INTERPOLATE(rt[i-1].per_error, rt[i].per_error, ratio);
    fprintf(fp,"%10f  ", temp);
    temp = (i <= 0)?rt[0].threshold :
      INTERPOLATE(rt[i-1].threshold, rt[i].threshold, ratio);
    fprintf(fp,"%10.3f  ", temp);
    fprintf(fp,"\n");
}

int
reject_compute_table(float **data_matrix, int npoints,
		     reject_table_type *reject_table)
{
    int i, j, table_length, nerrors;
    float threshold, last_threshold = -1;

    for(table_length = i = 0; table_length < npoints && i < npoints; i++){
	/* limit the number of significant figures */
	if(i == 0){
	    threshold = 0.;
	} else if(i == npoints -1){
	    threshold = 1;
	} else {
	    threshold = data_matrix[npoints-1-i][CLASS_OUTPUT];
	    threshold = (float)((int)(threshold*1000))/1000.;
	}
	if(i == 0){
	    nerrors = 0;
	    for(j = 0; j < npoints; j++){
		if((int)data_matrix[j][CLASS_ERROR]) nerrors++;
	    }
	} else {
	    if((int)data_matrix[npoints-i][CLASS_ERROR]) nerrors--;
	}
	if(threshold != last_threshold){
	    reject_table[table_length].threshold = threshold;
	    reject_table[table_length].npatterns = npoints-i;
	    reject_table[table_length].nrejections = i;
	    reject_table[table_length].nerrors = nerrors;

	    reject_table[table_length].per_rejection =
	      (100. * reject_table[table_length].nrejections) /
		(reject_table[table_length].npatterns +
		 reject_table[table_length].nrejections);
	    reject_table[table_length].per_error =
	      (100. * reject_table[table_length].nerrors) /
		reject_table[table_length].npatterns;
	    table_length++;
	    last_threshold = threshold;
	}
    }
    return table_length;
}

void
make_macfile(int table_length, float *x, float *y)
{
    int i;
    FILE *fptr;

    fptr = nn_fopen(path_build(plot_exp_path, mac_rejectlines_fname),
		    "w", "to write a Mac file");

    for(i = 0; i < table_length; i++)
	fprintf(fptr, "%f %f\n", x[i], y[i]);
    nn_fclose(fptr, path_build(plot_exp_path, mac_rejectlines_fname),
	      "to write a Mac file");
}


void
plot_reject(int table_length, reject_table_type *reject_table)
{
    int symbol_mode = 0, symbol_type = 0; /* no symbols */
    char *symbol_string = NULL;
    int symbol_radius = 0;
    int grid_type = 1; /* full grid */
    int reject_line_type = 1; /* solid lines */
    int  auto_abscissas = 0; /* x provided by x array */
    float auto_xmin = 0.0, auto_xstep = 1.0; /* ignored */
    float height_fract = 0.9, width_fract = 1.0;
    float xoffset_fract = 0.1, yoffset_fract = 0.1;
    int xlog = 0, ylog = 0; /* linear x and y axes */
    char *xlabel = "%% Rejection", *ylabel = "%% Error";
    char *color = "red";
    char *command =
      "-geometry 500x520 -title reject_plot ";
    int quit_flag = 0;
    float axis_xmin, axis_xmax, axis_xstep;
    float axis_ymin, axis_ymax, axis_ystep;
    int ubot, vbot, utop, vtop;
    float *x, *y;
    int i;

    /* load percent reject and percent error in arrays */
    x = (float *) allocate_vector(table_length);
    y = (float *) allocate_vector(table_length);
    for(i = 0; i < table_length; i++){
	x[i] = reject_table[i].per_rejection;
	y[i] = reject_table[i].per_error;
    }
    /* write out the mac version of the file */
    if(mac_reject_flag) make_macfile(table_length, x, y);

    /* plot the curve: y(i) vs x(i) */
    if(no_graphics)
      ginit("no graphics");
    else
      ginit(command);

    ghardcopy(1);

    gwtitle("Error Rate Plot");

    /* get size of plot */
    /* no symbols, no lines */
    xygraph_scale(x, y, table_length,
		  2, symbol_string, symbol_type, symbol_radius,
		  0, grid_type, reject_plot_label,
		  auto_abscissas, auto_xmin, auto_xstep,
		  height_fract, width_fract, xoffset_fract, yoffset_fract,
		  xlog, 0, reject_xmin, reject_xmax, reject_xstep, "",
		  ylog, 0, reject_ymin, reject_ymax, reject_ystep, "", color,
		  &ubot, &vbot, &utop, &vtop,
		  &axis_xmin, &axis_xmax, &axis_xstep,
		  &axis_ymin, &axis_ymax, &axis_ystep);

    /* display label */
    uvprintf((utop+ubot)/2, vbot-90, 2, reject_plot_label);

    /* do real plots using scale gotten from xygraph_scale() */
    xygraph(x, y, table_length, 
	    symbol_mode, symbol_string, symbol_type, symbol_radius,
	    reject_line_type, grid_type,
	    path_build(plot_exp_path, plot_reject_fname),
	    auto_abscissas, auto_xmin, auto_xstep,
	    height_fract, width_fract, xoffset_fract, yoffset_fract,
	    xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
	    ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
	    color);

    deallocate_vector(x);
    deallocate_vector(y);

    gsave(path_build(plot_exp_path, plot_reject_fname), 1);
    gexit(quit_flag);

    exit_nnsubs();

}
