/***-------------------------------------------------------------------------
 * NAME:	plot_svm.c
 * PURPOSE:	Plots structure of a Support Vector Machine classifier
 * USAGE:
 *
 * SEE ALSO:	
 *
 * DESCRIPTION: Reads in a svm parameter file and plots the structure
 *
 * $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_svm.c,v 11.2 2001/11/05 19:58:42 kukolich Exp $
 * $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_svm.c,v $
 * CREATED: 2-Nov-93 lck			MIT LINCOLN LABORATORY
 -------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = "$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/plot_svm.c,v 11.2 2001/11/05 19:58:42 kukolich Exp $";


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#include "plot_struct.h"
#include <svm_pack.h>

#include "xlippic.h"


#define BLOCK_RADIUS 20
#define BLOCK_SEPARATION (3*BLOCK_RADIUS)
#define LOWER_TEXT 10
#define CENTERED 2

#ifndef ABS
#define ABS(x) (((x) > 0)? (x) : -(x))
#endif

#define OUTPUT 2
#define MIDDLE 1
#define INPUT 0

static int node2u (int x, int layer);
static int level2v (int level);
static void draw_connection (int m, int level, int nodeu, int nodev,
			     float log_weight);

static int umiddle, voffset;
static int block_radius, lower_text;
static float layer_width[3];
static int layers[3];
static char **labels;
static char **input_labels;
static float struct_height = 1., struct_width = 1.;

int
main(int argc, char **argv)
{
    FILE *param_file;
    int model, norm_model, normalization, n, m, oM, total_m, layer;
    int noutputs, ninputs, nraw, *sv_n, nsvm;
    int kernel_type, multi_type;
    int ubot = 0, utop  = 1020, vbot = 0, vtop = 780;
    float scale;
    int nodeu, nodev, *features;
    int nodea, nodeb, nodeOut;
    int class_a, class_b;
    DEFAULTS_FILE_STRUCT df_args;
    int rv, i;
        
    init_nnsubs(argc, argv, rcsid);
    /********* read command line arguments ************/
    argc = nnargs(argc, argv, struct_plot_argtab, STRUCT_PLOT_TABSIZE);

    /********** open and read parameter file ***********/
    param_file = nn_fopen(struct_param_fname, "r",
			  "for read of svm parameters");
    norm_model = norm_read_paramfile(param_file, &normalization);
    if (norm_model < 0)
      fatal_error("%s: read error %d %s %s", argv[0], norm_model,
		  "on normalization parameter file", struct_param_fname);
    rv = feature_read_paramfile(param_file, &features, &nraw, &ninputs);
    if (rv < 0)
      fatal_error("%s: read error %d %s", argv[0], rv,
		  "on feature selection parameters");
   
    model = svm_read_model_multi(param_file, 0);
    if (model < 0)
      fatal_error("%s: read error %d on model file %s",
		  argv[0], model, struct_param_fname);
    nn_fclose(param_file, struct_param_fname, "for read of svm parameters");

    if(strcmp(struct_defaults_fname, "") != 0){
	read_defaults_file(struct_defaults_fname, &df_args);
	labels = df_args.output_labels;
	input_labels = df_args.input_labels;
    } else {
	input_labels = labels = NULL;
    }

    /************* initialize graphics *************/
    if(no_graphics){
	ginit("no graphics");
    } else {
	ginit(" -geometry 510x450 -title struct_plot");
    }
    ghardcopy(1);
    gwtitle("SVM Structure Plot");
    
    /* get data from model */
    ninputs = *(int *)svm_param_multi(model, "ninputs");
    noutputs = *(int *)svm_param_multi(model, "noutputs");
    nsvm = *(int *)svm_param_multi(model, "n");
    kernel_type = *(int *)svm_param(0, "kernel_type");
    multi_type = *(int *)svm_param_multi(model, "type");

    /* figure out reference positions for actual plotting */
    sv_n = allocate_ivector(nsvm);
    for(i = 0; i < nsvm; i++){
      sv_n[i] = *(int *)svm_param(i, "sv_n");
    }

    umiddle = (utop - ubot)/2;

    /* write out the plot title */
    uvprintf(umiddle, vtop - 20, CENTERED, "%s", struct_param_fname);

    /* write out the covariance types */
    if(!struct_no_text){
      switch(kernel_type){
      case SMO_LINEAR:
      case SMO_LINEAR_USE_SVS:
	uvprintf(umiddle, vbot+20, CENTERED, "Linear Kernels");
	break;
      case SMO_GAUSSIAN:
	uvprintf(umiddle, vbot+20, CENTERED, "Gaussian Kernels");
	break;
      case SMO_POLYNOMIAL:
	uvprintf(umiddle, vbot+20, CENTERED, "Polynomial Kernels");
	break;
      case SMO_INHOMOGENEOUS_POLYNOMIAL:
	uvprintf(umiddle, vbot+20, CENTERED, "Inhomogeneous Polynomial Kernels");
	break;
      }
    }

    vtop -= 20;
    vbot += 40;
    voffset = (vtop - vbot)/2;

    layers[INPUT] = ninputs;
    layers[MIDDLE] = nsvm;
    layers[OUTPUT] = noutputs;
    if(struct_autoscale){
	struct_height = (vtop-vbot-40)/(float)(6*BLOCK_SEPARATION);
	struct_width = (utop-ubot)/(float)((nsvm+2)*BLOCK_SEPARATION);
	
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] =
	      (utop-ubot)/(float)((layers[layer]+2)*BLOCK_SEPARATION);
	}
    } else {
	for(layer = INPUT; layer <= OUTPUT; layer++){
	    layer_width[layer] = struct_width;
	}
    }
    scale = (struct_width < struct_height) ? struct_width : struct_height;
    block_radius = BLOCK_RADIUS*scale;
    lower_text = LOWER_TEXT+block_radius;

    /* write out the class labels */
    nodev = level2v(OUTPUT);
   for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);
	if(!struct_no_text){
	    if(labels != NULL)
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%s",
		       labels[n+1]);
	    else
	      uvprintf(nodeu, nodev+lower_text, CENTERED, "%d", n);
	}
    }

    /* display the output nodes */
    total_m = 0;
    nodev = level2v(OUTPUT);
    for(n = 0; n < noutputs; n++){
	nodeu = node2u(n, OUTPUT);

	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);
    }


    /* draw hidden node circles and connections */
    nodev = level2v(MIDDLE);
    nodeOut = level2v(OUTPUT);
    for(m = 0; m < nsvm; m++){
      /* these things end up in a backwards looking order */
      oM = nsvm - m - 1;
	nodeu = node2u(oM, MIDDLE);

	if(multi_type){
	  class_a = *(int *)svm_param(m, "multi_class_a");
	  nodea = node2u(class_a, OUTPUT);
	  class_b = *(int *)svm_param(m, "multi_class_b");
	  nodeb = node2u(class_b, OUTPUT);
	  draw_connection(oM, MIDDLE, nodea, nodeOut, 1);
	  draw_connection(oM, MIDDLE, nodeb, nodeOut, 1);
	} else {
	  class_a = *(int *)svm_param(m, "multi_class_a");
	  nodea = node2u(class_a, OUTPUT);
	  draw_connection(oM, MIDDLE, nodea, nodeOut, 1);
	}
	/* draw a circle */
	uvcircle(nodeu, nodev, block_radius);
	if(!struct_no_text){
	  uvprintf(nodeu, nodev, CENTERED, "%d", sv_n[m]);
	}
	for(n = 0; n < ninputs; n++){
	    draw_connection(n, INPUT, nodeu, nodev, 0.);
	}
    }
    if(nsvm == 1){
      for(n = 0; n < noutputs; n++){
	draw_connection(0, MIDDLE, node2u(n, OUTPUT), nodeOut, 1);
      }
    }
    
    /* draw input node squares */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	/* draw a filled circle */
	uvdot("foreground", nodeu, nodev, block_radius/2);
    }

    /* write out the input labels */
    nodev = level2v(INPUT);
    for(n = 0; n < ninputs; n++){
	nodeu = node2u(n, INPUT);
	if(!struct_no_text){
	    uvprintf(nodeu, nodev-lower_text, CENTERED,
		     "%s",  get_input_label(n, input_labels, features,
					    normalization));
	}
    }

    gflush();
    gsave(struct_plot_fname, 1);
    gexit(0);/* quit */
    exit_nnsubs();
    exit(0);
}
static int
node2u(int x, int layer)
{
    /* center the node */
    x -= layers[layer]/2;
    /* scale the position for u coordinates*/
    x *= BLOCK_SEPARATION*layer_width[layer];
    /* center the final position on the screen */
    return(x + umiddle);
}

static int
level2v(int level)
{
    return(voffset + (level - 1)*(2*BLOCK_SEPARATION*struct_height));
}

static void
draw_connection(int m, int level, int nodeu, int nodev, float log_weight)
{
    int v1, v2, u1, u2, ulength, vlength, udir, vdir, r;
    int childu, childv;
    childu = node2u(m, level);
    childv = level2v(level);
    
    ulength = ABS(nodeu-childu);
    udir = (nodeu > childu) ? 1 : -1;
    vlength = ABS(nodev-childv);
    vdir = (nodev > childv) ? 1 : -1;
    r = (int)sqrt(((double)ulength*ulength)
		  +(double)vlength*vlength);
    u1 = nodeu-udir*(ulength*block_radius)/r;
    v1 = nodev-vdir*(vlength*block_radius)/r;
    
    if(level == INPUT){
	u2 = childu;
	v2 = childv;
    } else {
	u2 = childu+udir*(ulength*block_radius)/r;
	v2 = childv+vdir*(vlength*block_radius)/r;
    }
    
    
    uvline(u1, v1, u2, v2);
}
