/***--------------------------------------------------------------------------
* NAME:         roc_plot.c
* PURPOSE:      package to plot roc's for a set of data
*
* USAGE:        need to use roc_pack to accumulate data for the model first
* RETURN VALUE:  NONE
*
* EXAMPLE:
*
* GLOBALS:
*
* DESCRIPTION:
*    Functions in this file
*       roc_plot                Plots the roc curve for the last computed roc
*
* SEE ALSO:	roc_pack.c
*
------------------------------------------------------------------------***/
/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */

static char rcsid[] = 
"$Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/roc_plot.c,v 11.2 2001/08/23 20:18:10 craig Exp $";


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <nnlib.h>
#include <get_subs.h>

/**************************************************************************
* NAME: roc_plot()
* DESCRIPTION:
*        Plots the roc plot identified by model_number, class, and reject.
*        xmin,xmax,xstep,ymin,ymax,ystep,line_type determines the plot 
*          characteristics.
*        exp_file_name is the name of the plot file to output
*        plot_flag: 0: generate no window plot and no plot_file
*                   1: generate window plot but no plot_file
*                   2: generate no window plot but plot_file
*                   3: generate window plot and plot_file
*        Outputs roc curve information based on the target class and the 
*          reject rate.  If the currently stored information in the model 
*          is not of the same class and/or reject rate, this call will 
*          trigger a call to roc_batch_compute() to compute the correct 
*          information.
*        The amount of output information is controlled by verbosity flag:
*          0: no output
*          1: output a column of false alarm and a column of detect rate
*          2: output full table
*        Computes roc curve information based on patterns accumulated by
*          roc_batch_train().  The computation is determined by a target
*          class and the fractional rejection rate [0,1].
*
***************************************************************************/

/* declare procedures used */
float roc_get_area();

void roc_plot(model_number, class, reject,
	      xmin, xmax, xstep, ymin, ymax, ystep, 
	      line_type, exp_file_name, plot_flag)
     int model_number;
     int class;
     float reject;
     float xmin, xmax, xstep, ymin, ymax, ystep;
     int line_type;
     char *exp_file_name;
     int plot_flag;
{
  roc *aroc;
  int i;

  roc_param(model_number, &aroc);
  if (aroc==NULL)
    { warning("Invalid model number %d",model_number); return; }

  if (aroc->class!=class || aroc->reject!=reject)
    roc_batch_compute(model_number,class,reject);

  /* plot false_alarm versus detect_rate (length=table_length) */
  {
    int i, j, n;
    int detect_auto_scale_flag = 0;
    char title[128], title_pieces[12][25];
    int symbol_mode = 0, symbol_type = 0; /* no symbols */
    char *symbol_string = NULL;
    int symbol_radius = 0;
    int grid_type = 1; /* full grid */
    int  auto_abscissas = 0; /* x provided by x array */
    float auto_xmin = 0.0, auto_xstep = 1.0; /* ignored */
    float height_fract = 0.9, width_fract = 1.0;
    float xoffset_fract = 0.15, yoffset_fract = 0.1;
    int xlog = 0, ylog = 0; /* linear x and y axes */
    char *xlabel = "%% False Alarm", *ylabel = "%% Detected";
    char *color = "black";
    char *command =
      "-geometry 600x400 -fg black -bg white -title detect_plot ";
    int quit_flag = 0;
    float axis_xmin, axis_xmax, axis_xstep;
    float axis_ymin, axis_ymax, axis_ystep;
    int ubot, vbot, utop, vtop;
    float area;
    float linex[2], liney[2];

    /* plot the curve */
    if (plot_flag==1 || plot_flag==3)
      ginit(command);
    else
      ginit("no graphics");
    if (plot_flag==2 || plot_flag==3)
      ghardcopy(1);
    else
      ghardcopy(0);

    /* get area */
    area= roc_get_area(model_number,class,reject);
    
    sprintf(title_pieces[0], "%s", "target class");
    sprintf(title_pieces[1], "%s", "=");
    sprintf(title_pieces[2], "%d;", class);
    sprintf(title_pieces[3], "%s", "# target class");
    sprintf(title_pieces[4], "%s", "=");
    sprintf(title_pieces[5], "%d;", aroc->ndr[0]);
    sprintf(title_pieces[6], "%s", "#total");
    sprintf(title_pieces[7], "%s", "=");
    sprintf(title_pieces[8], "%d;", aroc->npatterns);
    sprintf(title_pieces[9], "%s", "area");
    sprintf(title_pieces[10],"%s", "=");
    sprintf(title_pieces[11],"%.3f",area);
    
    /* printf("area is %f\n",area);*/
    area = roc_get_percent_fom(model_number, class, reject,
			       0.0,0.4,0.01);
    /* printf("FOM (0 to 40%) is %f\n",area);fflush(stdout);*/

    n = 0;
    for(i = 0; i < 12; i++){
	for(j = 0; j < strlen(title_pieces[i]); j++){
	    title[n++] = title_pieces[i][j];
	}
	title[n++] = ' ';
      }
    title[n] = '\0';

    gwtitle("ROC Plot:");

    /* get size of plot */
    /* no symbols, no lines */
    xygraph_scale(aroc->false_alarm, aroc->detect_rate, aroc->table_length,
		  2, symbol_string, symbol_type, symbol_radius,
		  0, grid_type, "ROC Plot",
		  auto_abscissas, auto_xmin, auto_xstep,
		  height_fract, width_fract, xoffset_fract, yoffset_fract,
		  xlog, detect_auto_scale_flag,
		  xmin, xmax, xstep, "",
		  ylog, detect_auto_scale_flag,
		  ymin, ymax, ystep, "","foreground",
		  &ubot, &vbot, &utop, &vtop,
		  &axis_xmin, &axis_xmax, &axis_xstep,
		  &axis_ymin, &axis_ymax, &axis_ystep);

    /* display label */
    /* uvprintf((utop+ubot)/2, vbot-75, 2, "ROC Plot");*/
    uvprintf((utop+ubot)/2, vbot-100, 2, title);

    /* do real plots using scale gotten from xygraph_scale() */
    xygraph(aroc->false_alarm, aroc->detect_rate, aroc->table_length, 
	    symbol_mode, symbol_string, symbol_type, symbol_radius,
	    line_type, grid_type,exp_file_name,
	    auto_abscissas, auto_xmin, auto_xstep,
	    height_fract, width_fract, xoffset_fract, yoffset_fract,
	    xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
	    ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
	    color);

    /* draw diagonal line */
    linex[0]=1.0; linex[1]=99.0;
    liney[0]=1.0; liney[1]=99.0;
    xygraph(linex, liney, 2,
	    symbol_mode, symbol_string, symbol_type, symbol_radius,
	    line_type, grid_type,exp_file_name,
	    auto_abscissas, auto_xmin, auto_xstep,
	    height_fract, width_fract, xoffset_fract, yoffset_fract,
	    xlog, 0, axis_xmin, axis_xmax, axis_xstep, xlabel,
	    ylog, 0, axis_ymin, axis_ymax, axis_ystep, ylabel,
	    color);

    if (plot_flag==2 || plot_flag==3)
      gsave(exp_file_name,1);
    gexit(quit_flag);

    exit_nnsubs();
    
  }
}  


