/* (c) Copyright 1999 M.I.T. Lincoln Laboratory


Permission is hereby granted, without written agreement or royalty fee,
to use, copy, modify, and distribute this software and its
documentation for any purpose, provided that the above copyright notic
and the following three paragraphs appear in all copies of this software.

IN NO EVENT SHALL M.I.T. LINCOLN LABORATORY BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF M.I.T. LINCOLN
LABORATORY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. M.I.T. LINCOLN
LABORATORY SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
AND NON-INFRINGEMENT.

THE SOFTWARE IS PROVIDED ON AN "AS IS" BASIS AND M.I.T. LINCOLN
LABORATORY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. */
/***---------------------------------------------------------------------------
* NAME:		svm_plot_bound.c
* PURPOSE:	Support vector machine Classifier plot driver (driver) program
* USAGE:
*
* SEE ALSO: 	plot_bound.c svm_pack.c
*
* DESCRIPTION:  Reads in a trained model file and plots the decision
*               region and/or scatter plot in 2 selected dimensions
*
*	read_model	Read model file describing classifier parameters
*	classify	Find class of an input vector using classifier
*	get_outputs	Get address of output vector
*	over_plot	Optional way to add classifier dependent plots
*
* $Header: /data/speech/kukolich/cvsroot/lnknet/src/plot/svm_plot_bound.c,v 11.4 2001/09/11 15:37:59 kukolich Exp $
* $Source: /data/speech/kukolich/cvsroot/lnknet/src/plot/svm_plot_bound.c,v $
* CREATED: 13-Jan-2001 rpl                             MIT LINCOLN LABORATORY
--------------------------------------------------------------------------***/


/* config.h must go first */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <math.h>
#include <nnlib.h>
#include <svm_pack.h>
#include "plot_bound.h"
#include "xlippic.h"
#define SQUARE (0)
#define TRIANGLE (1)
#define PLUS (2)
#define X (3)
#define CIRCLE (7)
#define RADIUS (15)


/* local function */
static void
draw_line (int level, int class_a, int class_b, double slope, double islope,
	   double xintercept, double yintercept, 
	   float bound_xmax, float bound_xmin,
	   float bound_ymax, float bound_ymin);

float *outputs;
int noutputs;
static int sigmoid_fit=0;
static ARG argtab[] =
{
    {"sigmoid_fit", BOOLEAN, (int *)&sigmoid_fit, 
     "use sigmoid to approximate posterior prob for each classifier"},
};
#define TABSIZE (1)

static int multi_model;			/* id of model */

int get_extra_plot_args(int argc, char **argv)
{
    int nargc;
    nargc = nnargs_part(argc, argv, argtab, TABSIZE);
    return nargc;
}

void
read_model (FILE *param_file, int *total_trials, int *ninputs,
	    int *pnoutputs, int **features, int adjust_priors,
	    float *priors_list)
{
 int rv, nraw;
 rv = feature_read_paramfile(param_file, features, &nraw, ninputs);
    if (rv < 0)
      fatal_error("read_model -- read error %d %s", rv,
                  "on feature selection parameters in model file");


    multi_model = svm_read_model_multi(param_file, 1);

    *ninputs = *(int *)svm_param_multi(multi_model,"ninputs");
    *pnoutputs = *(int *)svm_param_multi(multi_model,"nclasses");
    outputs = *(float **)svm_param_multi(multi_model, "outputs");
    noutputs = *pnoutputs;
    if (multi_model < 0) fatal_error ("read_model -- read error %d on model file",
	multi_model);
    svm_test_init_multi(multi_model, adjust_priors, priors_list, sigmoid_fit);
}


int 
classify (float *inputs)
{
    return svm_classify_multi(multi_model,inputs);
}


				/* used only for profile/histogram plots */
float *
get_outputs (void)
{
    float *px;
    px = svm_get_outputs_multi(multi_model);
    return ( px);
}


void
over_plot (int two_d, float internals_scale, int internals_level,
	   char **plot_labels, int x_dimen, int y_dimen,
	   float bound_xmax, float bound_xmin,
	   float bound_ymax, float bound_ymin,
	   int ubot, int utop, int vbot, int vtop)
{
    int nclassifiers, i, j, k, u1, v1;
    int *component_model;
    int kernel_type,sv_n;
    float threshold, cbound;
    float *weight;
    float **support_vectors;
    int *sv_unique;
    float *sv_alpha;
    int a, b;
    double slope, islope;
    double xintercept, yintercept;
    float x_mid;
    
    if (two_d) 
        {
        x_mid = (bound_xmax + bound_xmin)/2.;
	gcolor("foreground");
        /* draw a line for every sub classifier */
        nclassifiers = *(int *)svm_param_multi(multi_model, "n");
        component_model = *(int **)svm_param_multi(multi_model, "component_model");
        for(i=0; i<nclassifiers; i++)
	    {
	    j = component_model[i];
            kernel_type = *(int *)svm_param(j, "kernel_type");
	    a = *(int *)svm_param(j, "multi_class_a");
	    b = *(int *)svm_param(j, "multi_class_b");
            /* only plot separating line if linear */
            if ( kernel_type == SMO_LINEAR)
	       {
	       threshold = *(float *)svm_param(j, "threshold");
	       weight = *(float **)svm_param(j, "weight");
	       if(x_dimen == y_dimen)
                   {
		   slope = 1000;
		   islope = 0;
                   }
	       else 
                   {
		   slope = -weight[x_dimen] / weight[y_dimen];
	 	   if(slope != 0.) 
                            islope = 1.0/slope;
	      	   else 
                            islope = 1000;
		   }
	       xintercept = threshold / weight[x_dimen];
	       yintercept = threshold / weight[y_dimen];
	       draw_line (internals_level, a, b, slope, islope, xintercept, 
			       yintercept, bound_xmax, bound_xmin, bound_ymax,
			       bound_ymin);
		  gflush();
	       }
            /* gauss, polynomial, or linear using support vector classifier */
            else
	        {
		sv_n = *(int *)svm_param(j, "sv_n");
		cbound = *(float *)svm_param(j, "cbound");
		support_vectors = *(float ***)svm_param(j, "support_vectors");
		sv_unique = *(int **)svm_param(j, "sv_unique");
		sv_alpha = *(float **)svm_param(j, "sv_alpha");
		/* plot unique support vectors for this component model 
		---------------------------------------------------------*/
		for (k=0;k<sv_n;k++) 
		    {
		    if ( sv_unique[k] > 0 ) 
			{
			u1 = x2u (support_vectors[k][x_dimen]);
			if(x_dimen == y_dimen) 
			       v1 = y2v(x_mid);
			else 
			    v1 = y2v (support_vectors[k][y_dimen]);
			if (u1 >= ubot && u1 <= utop && v1 >= vbot && v1 <= vtop) 
			    {
			    uvmove(u1, v1);
			    gcolor("foreground");
			    gsymbol (CIRCLE, internals_scale);
			    /* add an x to the circle if unbounded */
			    if (  sv_alpha[k] < cbound )
				{
				gsymbol (X, (int)(RADIUS*internals_scale));
				}
			    gsymbol (CIRCLE, (int)(RADIUS*internals_scale));
			    if (internals_level > 2) 
				{
				uvprintf (u1, v1, 2, "%d", j);
				}
			    }
			}
		    }
		}
	    }
	}
    }


static void
draw_line (int level, int class_a, int class_b, double slope, double islope,
	   double xintercept, double yintercept, 
	   float bound_xmax, float bound_xmin,
	   float bound_ymax, float bound_ymin)
{
    float xdisp[2];			/* values to plot */
    float ydisp[2];
    float x,y;
    int i=0;
    int ix, iy;

    /* determines how line intercepts screen and draws full length
       line across screen */

    if(islope == 0.){
	x = xintercept;
	if (x <= bound_xmax && x >= bound_xmin) {
	    y = bound_ymax;
	    xdisp[i] = x;
	    ydisp[i] = y;
	    i++;

	    y = bound_ymin;
	    xdisp[i] = x;
	    ydisp[i] = y;
	    i++;
	}
    } else if (slope == 0.){
	y = yintercept;
	if ( y <= bound_ymax && y >= bound_ymin) {
	    x = bound_xmin;
	    xdisp[i] = x;
	    ydisp[i] = y;
	    i++;

	    x = bound_xmax;
	    xdisp[i] = x;
	    ydisp[i] = y;
	    i++;
	}
    } else {
	/* check intercept on top */
	y = bound_ymax;
	x = islope*y + xintercept;
	if ( x <= bound_xmax && x >= bound_xmin) {
	    xdisp[i] = x;
	    ydisp[i] = y;
	    i++;
	}

	/* check intercept on bottom */
	y = bound_ymin;
	x = islope*y + xintercept;
	if ( x <= bound_xmax && x >= bound_xmin) {
	    xdisp[i] = x;
	    ydisp[i] = y;
	    i++;
	}

	/* check intercept on left side */
	if ( i < 2 ) {
	    x = bound_xmin;
	    y = slope*x + yintercept;
	    if ( y <= bound_ymax && y >= bound_ymin) {
		xdisp[i] = x;
		ydisp[i] = y;
		i++;
	    }
	}

	/* check intercept on right side */
	if ( i < 2 ) {
	    x = bound_xmax;
	    y = slope*x + yintercept;
	    if ( y <= bound_ymax && y >= bound_ymin) {
		xdisp[i] = x;
		ydisp[i] = y;
		i++;
	    }
	}
    }
    /* plot line across screen */
    if (i >= 2){
	uvline(x2u(xdisp[0]), y2v(ydisp[0]), x2u(xdisp[1]), y2v(ydisp[1]));
	if (level > 2) { 
	    if ( xdisp[0] < xdisp[1]) {
		ix = x2u(xdisp[0]);
		iy = y2v(ydisp[0]);
		if (ydisp[1] > ydisp[0]) iy += 10; else iy -= 30;
	    } else {
		ix = x2u(xdisp[1]);
		iy = y2v(ydisp[1]);
		if (ydisp[0] > ydisp[1]) iy += 10; else iy -= 30;
	    }
	    uvprintf (ix + 20, iy, 1, "%d vs %d", class_a, class_b);
	}
    }
}
