/************************************************************
 *                                                          *
 *  Permission is hereby granted  to  any  individual   or  *
 *  institution   for  use,  copying, or redistribution of  *
 *  the xgobi code and associated documentation,  provided  *
 *  that   such  code  and documentation are not sold  for  *
 *  profit and the  following copyright notice is retained  *
 *  in the code and documentation:                          *
 *     Copyright (c) 1990,1991,1992,1993,1994 Bellcore      *
 *                                                          *
 *  We welcome your questions and comments, and request     *
 *  that you share any modifications with us.               *
 *                                                          *
 *    Deborah F. Swayne            Dianne Cook              *
 *     dfs@bellcore.com         dicook@iastate.edu          *
 *      (201) 829-4263                                      *
 *                                                          *
 ************************************************************/

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <stdio.h>
#include "xincludes.h"
#include "xgobitypes.h"
#include "xgobivars.h"

/* Functions used in this file */
int set_deci();
void set_mono();
void update_lims(), update_sphered(), update_world();
void world_to_plane(), plane_to_screen(), plot_once();
void SetNiceRange(), init_ticks(), update_cprof_plot();
void show_message();
void min_max(), mean_stddev();
void adjust_limits();
void assign_points_to_bins();
/* */

#define NTFORMS 12
char message[512];
#define DOMAIN_ERROR sprintf(message, "Data outside the domain of function.\n")

#define RESTORE 0
#define NEGATIVE 1
#define INVERSE 2
#define LN 3
#define FOURTH_ROOT 4
#define CUBE_ROOT 5
#define SQUARE_ROOT 6
#define SQUARE 7
#define CUBE 8
#define FOURTH_POWER 9
#define SCALE 10
#define STANDARDIZE 11

static char *tform_names[] = {
  "Restore Variable",
  "Negative",
  "Inverse",
  "Natural Log",
  "Fourth Root",
  "Cube Root",
  "Square Root",
  "Square",
  "Cube",
  "Fourth Power",
  "Scale to [0,1]",
  "Standardize"
};
static int *tform_type;
struct {
	Widget cascade;
	Widget btn[NTFORMS];
} tform_menu;

int
transform(xg, cols, ncols, tfnum)
  xgobidata *xg;
  int *cols, ncols;
  int tfnum;
{
  int i, j, n;
  float min, max, diff;
  float mean, stddev;

  switch(tfnum)
  {
    case RESTORE:    /* Restore original values */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
          xg->tform_data[i][j] = xg->raw_data[i][j];

        (void) strcpy(xg->collab_tform[j], xg->collab[j]);
      }
      break;

    case NEGATIVE:    /* Multiply original values by -1 */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
          xg->tform_data[i][j] = (float) -1.0 * xg->raw_data[i][j] ;

        (void)
           sprintf(xg->collab_tform[j], "-%s", xg->collab[j]);
      }
      break;

    case INVERSE:    /* 1/x */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          if (xg->raw_data[i][j] == 0)
          {
            DOMAIN_ERROR;
            show_message(message, xg->shell);
            return(0);
          }
        }
      }

      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          xg->tform_data[i][j] = (float)
            pow((double) xg->raw_data[i][j], (double) (-1.0));
        }

        (void)
           sprintf(xg->collab_tform[j], "1/%s", xg->collab[j]);
      }
      break;

    case LN:    /* Natural log */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          if (xg->raw_data[i][j] <= 0)
          {
            DOMAIN_ERROR;
            show_message(message, xg->shell);
            return(0);
          }
        }
      }
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          xg->tform_data[i][j] = (float)
            log((double) xg->raw_data[i][j]);
        }

        (void)
           sprintf(xg->collab_tform[j], "ln(%s)", xg->collab[j]);
      }
      break;

    case FOURTH_ROOT:    /* Fourth root */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          if (xg->raw_data[i][j] < 0)
          {
            DOMAIN_ERROR;
            show_message(message, xg->shell);
            return(0);
          }
        }
      }

      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          if (xg->raw_data[i][j] < 0)
          {
            xg->tform_data[i][j] = (float)
              - pow( - (double) xg->raw_data[i][j],
                   (double) (1.0/4.0));
          }
          else
          {
            xg->tform_data[i][j] = (float)
              pow((double) xg->raw_data[i][j],
                (double) (1.0/4.0));
          }
        }
        (void)
           sprintf(xg->collab_tform[j], "%s^(1/4)", xg->collab[j]);
      }
      break;

    case CUBE_ROOT:    /* Cube root */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          if (xg->raw_data[i][j] < 0)
          {
            xg->tform_data[i][j] = (float)
              - pow( - (double) xg->raw_data[i][j],
                   (double) (1.0/3.0));
          }
          else
          {
            xg->tform_data[i][j] = (float)
              pow((double) xg->raw_data[i][j],
                (double) (1.0/3.0));
          }
        }
        (void)
           sprintf(xg->collab_tform[j], "%s^(1/3)", xg->collab[j]);
      }
      break;

    case SQUARE_ROOT:    /* Square root */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          if (xg->raw_data[i][j] < 0)
          {
            DOMAIN_ERROR;
            show_message(message, xg->shell);
            return(0);
          }
        }
      }
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
        {
          xg->tform_data[i][j] = (float)
            sqrt((double) xg->raw_data[i][j]);
        }

        (void)
           sprintf(xg->collab_tform[j], "%s^1/2", xg->collab[j]);
      }
      break;

    case SQUARE:    /* Square */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
          xg->tform_data[i][j] = (float)
            pow((double) xg->raw_data[i][j], (double) 2);

        (void)
           sprintf(xg->collab_tform[j], "%s^2", xg->collab[j]);
      }
      break;

    case CUBE:    /* Cube */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
          xg->tform_data[i][j] = (float)
            pow((double) xg->raw_data[i][j], (double) 3);

        (void)
           sprintf(xg->collab_tform[j], "%s^3", xg->collab[j]);
      }
      break;

    case FOURTH_POWER:    /* Fourth Power */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
          xg->tform_data[i][j] = (float)
            pow((double) xg->raw_data[i][j], (double) 4);

        (void)
           sprintf(xg->collab_tform[j], "%s^4", xg->collab[j]);
      }
      break;

    case SCALE:    /* Map onto [0,1] */
      /* First find min and max; they get updated after transformations */
      min_max(xg, xg->raw_data, cols, ncols, &min, &max);
      adjust_limits(&min, &max);
      diff = max - min;

      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
          xg->tform_data[i][j] = (float) (xg->raw_data[i][j] - min)/diff;

        (void)
           sprintf(xg->collab_tform[j], "%s [0,1]", xg->collab[j]);
      }
      break;

    case STANDARDIZE:    /* (x-mean)/sigma */
      /* First find min and max */
      mean_stddev(xg, xg->raw_data, cols, ncols, &min, &max, &mean, &stddev);

      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        for (i=0; i<xg->nrows; i++)
          xg->tform_data[i][j] = (float) (xg->raw_data[i][j] - mean)/stddev;

        (void)
           sprintf(xg->collab_tform[j], "(%s-m)/s", xg->collab[j]);
      }
      break;

  }
  return(1);
}

/* ARGSUSED */
XtCallbackProc
tform_choose_cback(w, xg, cback_data)
  Widget w;
  xgobidata *xg;
  caddr_t cback_data;
{
  int j, n;
  int varno = -1, tfno;
  char *tf_name;
  Widget label = XtParent(XtParent(w));
  int *cols, ncols;
  int groupno;

  if (xg->is_touring && (xg->is_princ_comp || xg->is_pp) )
  {
    sprintf(message,
      "Sorry, it isn't possible to transform variables during\n");
    strcat(message,
      "projection pursuit or when principal components are used.\n");
    show_message(message, xg->shell);
  }
  else
  {
    /*
     * Figure out which variable is to be transformed.
    */
    for (varno=0; varno<xg->ncols_used; varno++)
      if (xg->varlabw[varno] == label)
        break;

    if (varno < 0)
    {
      (void) fprintf(stderr, "error in tform_choose_cback\n");
      return;
    }

    /*
     * Figure out which transformation to do.
    */
    XtVaGetValues(w, XtNlabel, &tf_name, NULL);

    for (tfno=0; tfno<NTFORMS; tfno++)
      if (strcmp(tf_name, tform_names[tfno]) == 0)
        break;

    /*
     * Figure out which columns to transform.
    */
    ncols = 0;
    groupno = xg->vgroup_ids[varno];
    cols = (int *) XtMalloc((Cardinal) xg->ncols_used * sizeof(int));
    for (j=0; j<xg->ncols_used; j++)
    {
      if (xg->vgroup_ids[j] == groupno)
        cols[ncols++] = j;
    }

    /*
     * Call the appropriate transformation program.
    */
    if ( transform(xg, cols, ncols, tfno) )
    {
      update_sphered(xg, cols, ncols);
      update_lims(xg);
      update_world(xg);

      /* Set tform_type[] for transformed columns */
      for (n=0; n<ncols; n++)
        tform_type[cols[n]] = tfno;

      world_to_plane(xg);
      plane_to_screen(xg);
      /*
       * This bit of init_axes() is needed.
      */
      for (n=0; n<ncols; n++)
      {
        j = cols[n];
        xg->nicelim[j].min = xg->lim0[j].min;
        xg->nicelim[j].max = xg->lim0[j].max;
        SetNiceRange(j, xg);
        xg->deci[j] = set_deci(xg->tickdelta[j]);
      }

      if (xg->is_xyplotting)
        init_ticks(&xg->xy_vars, xg);
      else if (xg->is_dotplotting)
        init_ticks(&xg->dotplot_vars, xg);

      if (xg->is_brushing)
      {
        assign_points_to_bins(xg);
        if (xg->brush_mode == transient)
        {
          void reinit_transient_brushing();
          reinit_transient_brushing(xg);
        }
      }

      plot_once(xg);

      if (xg->is_cprof_plotting)
        update_cprof_plot(xg);
    }
    XtFree((XtPointer) cols);
  }
  XtDestroyWidget((Widget) tform_menu.cascade);
}

/* ARGSUSED */
XtEventHandler
add_tform_menu(w, xg, evnt)
/*
 * Add menus for choosing a data transformation.
*/
  Widget w;
  xgobidata *xg;
  XEvent *evnt;
{
  int k, varno;

  for (varno=0; varno<xg->ncols; varno++)
    if (xg->varlabw[varno] == w)
      break;

  tform_menu.cascade = XtVaCreatePopupShell("menu",
    simpleMenuWidgetClass, xg->varlabw[varno],
    NULL);
  if (mono) set_mono(tform_menu.cascade);

  for (k=0; k<NTFORMS; k++)
  {
    tform_menu.btn[k] = XtVaCreateManagedWidget("Command",
      smeBSBObjectClass, tform_menu.cascade,
      XtNlabel, (String) tform_names[k],
      XtNleftMargin, (Dimension) 24,
      NULL);
    if (mono) set_mono(tform_menu.btn[k]);
    XtAddCallback(tform_menu.btn[k], XtNcallback,
      (XtCallbackProc) tform_choose_cback, (XtPointer) xg);
  }

  XtVaSetValues(tform_menu.btn[tform_type[varno]],
    XtNleftBitmap, menu_mark, NULL);
}

void
alloc_transform_types(xg)
  xgobidata *xg;
{
  tform_type = (int *) XtRealloc((char *) tform_type,
    (Cardinal) xg->ncols * sizeof(int));
}

void
init_transform_types(xg)
/*
 * Set each tranformation type to 0;
*/
  xgobidata *xg;
{
  int j;

  for (j=0; j<xg->ncols; j++)
    tform_type[j] = 0;
}
